// -*- c++ -*-
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Iterative Template Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//

#ifndef ITL_GMRES_H
#define ITL_GMRES_H

#include <vector>
#include <algorithm>
#include "itl/itl.h"
#include "itl/givens_rotation.h"
#include "itl/number_traits.h"

namespace itl {

  template <class Vec>
  void ___initialize(Vec& s, typename itl_traits<Vec>::value_type t) {
    std::fill(s.begin(), s.end(), t);
  }

  // Generalized Minimum Residual
  //
  //   This solve the unsymmetric linear system Ax = b using restarted GMRES.
  //
  //   A return value of 0 indicates convergence within the
  //   maximum number of iterations (determined by the iter object).
  //   A return value of 1 indicates a failure to converge.
  //
  //   On instantiating Iteration object outer, the first parameter
  //   Vector w must be the precondtioned one, i.e., solve(M, b, w) where
  //   b is right side of linear system. See test_gmres.cc for example.
  //
  //   See: Y. Saad and M. Schulter. GMRES: A generalized minimum residual
  //   algorithm for solving nonsysmmetric linear systems, SIAM
  //   J. Sci. Statist. Comp.  7(1986), pp, 856-869
  //
  /* required operations: mult,copy,dot_conj,add,scaled,two_norm,tri_solve */
template < class Matrix, class Vector, class VectorB, class Preconditioner, 
	   class Iter, class Basis >
int 
gmres(const Matrix &A, Vector &x, const VectorB &b,
      const Preconditioner &M, int m, Iter& outer, Basis& KS)
{
  typedef typename itl_traits<Vector>::value_type T;

  typedef typename itl_traits<Vector>::gmres_internal_vector TmpVec;
  typedef typename itl_traits<Vector>::internal_vector internal_vector;

  //These must be arithmetically compatible with x
  internal_vector w(size(x)), r(size(x)), u(size(x));

  typedef typename number_traits<T>::magnitude_type Real;
  typedef typename internal_matrix_traits<Real>::Matrix HMat;
  HMat H(m+1, m); //Elements in H must be real type

  TmpVec s(m+1);
  
  std::vector< itl::givens_rotation<T> > rotations(m+1);

  itl::mult(A, itl::scaled(x, -1.0), b, w);

  itl::solve(M, w, r);
  typename Iter::real beta = std::abs(itl::two_norm(r));

  while (! outer.finished(beta)) {

    itl::copy(itl::scaled(r, 1./beta), KS[0]);
    ___initialize(s, 0.0);
    s[0] = beta;

    int i = 0;
    Iter inner(outer.normb(), m, outer.tol(), outer.atol());

    do {
      int k;
      itl::mult(A, KS[i], u);
      itl::solve(M, u, KS[i+1]);

      itl::orthogonalize(KS, H[i], i);

      H(i+1, i) = itl::two_norm(KS[i+1]);
      itl::scale(KS[i+1], 1./H(i+1,i));

      for (k = 0; k < i; k++)
        rotations[k].scalar_apply(H(k,i), H(k+1,i));

      rotations[i] = itl::givens_rotation<T>(H(i,i), H(i+1,i));
      rotations[i].scalar_apply(H(i,i), H(i+1,i));
      rotations[i].scalar_apply(s[i], s[i+1]);
      
      ++inner, ++outer, ++i;

    } while (! inner.finished(std::abs(s[i]))); 

    itl::upper_tri_solve(H, s, i);

    itl::combine(KS, s, x, i);

    itl::mult(A, itl::scaled(x, -1.0), b, w);
    solve(M, w, r);
    beta = std::abs(itl::two_norm(r));
  }
  
  return outer.error_code();
}


}

#endif
