// -*- c++ -*-
//===========================================================================
// DESCRIPTION                                                               
//   This is the file contain the implementations of basic operations.
//   It provides mtl implementations of those operations. Therefore,
//   It requires mtl package. The tested mtl release is 2.1.2
//                                                                           
//---------------------------------------------------------------------------
//                                                                           
// LICENSE AGREEMENT                                                         
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Iterative Template Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//---------------------------------------------------------------------------

#ifndef ITL_MTL_INTERFACE_H
#define ITL_MTL_INTERFACE_H

/*
   
   Other parts in itl dependent upon mtl interface are preconditioners.
   The files are: ilu.h, ilut.h, ssor.h, cholesky.h, preconditioners.h
 */

#include <mtl/dense1D.h>
#include "mtl/matrix.h"
#include "mtl/mtl.h"
#include <complex>
#include <itl/itl_tags.h>
#include <itl/number_traits.h>

namespace itl {
  
  //: The vector type used inside of the ITL routines for work space
  template <class Vec>
  struct itl_traits {
    typedef referencing_object_tag vector_category;
    typedef typename Vec::value_type value_type;
    typedef mtl::dense1D<value_type> internal_vector;
    typedef mtl::dense1D<value_type> gmres_internal_vector;
  };

  template <class MTLMat>
  inline typename MTLMat::size_type nrows(const MTLMat& A)
  { return A.nrows(); }

  template <class Vec>
  inline 
  typename itl::number_traits< typename Vec::value_type >::magnitude_type
  two_norm(const Vec& v) {
    return mtl::two_norm(v);
  }

  //deal with the case when b is a handle, 
  template <class VecA, class VecB>
  inline void copy(const VecA& a, const VecB& b) {
    mtl::copy(a, const_cast<VecB&>(b));
  }
  
  template <class MTLMat, class VecX, class VecY>
  inline void mult(const MTLMat& A, const VecX& x, const VecY& y) {
    mtl::mult(A, x, const_cast<VecY&>(y));
  }

  template <class MTLMat, class VecX, class VecY, class VecZ>
  inline void mult(const MTLMat& A, const VecX& x, const VecY& y, 
		   const VecZ& z) {
    mtl::mult(A, x, y, const_cast<VecZ&>(z));
  }

  template <class VecA, class VecB>
  inline typename VecA::value_type dot(const VecA& a, const VecB& b) {
    return mtl::dot(a, b);
  }
  
  template <class VecA, class VecB>
  inline typename VecA::value_type  dot_conj(const VecA& a, const VecB& b) {
    return mtl::dot_conj(a, b);
  }
  

  template <class VecX, class VecY>
  inline void add(const VecX& x, const VecY& y) {
    mtl::add(x, const_cast<VecY&>(y));
  }

  template <class VecX, class VecY, class VecZ>
  inline void add(const VecX& x, const VecY& y, const VecZ& z) {
    mtl::add(x, y, const_cast<VecZ&>(z));
  }

  template <class VecX, class VecY, class VecZ, class VecR>
  inline void add(const VecX& x, const VecY& y, const VecZ& z, const VecR& r) {
    mtl::add(x, y, z, const_cast<VecR&>(r));
  }

  template <class Scalab, class TTT>
  inline typename Scalab::scaled_type
  scaled(const Scalab& v, TTT t) {
    return mtl::scaled(v, t);
  }

  template <class Vec, class TTT>
    inline void scale(const Vec& v, TTT t) {
    mtl::scale(const_cast<Vec&>(v), t);
  }

  template <class Vector>
  inline typename Vector::size_type
  size(const Vector& x) 
  {
    return x.size();
  }


  template <class Vector, class Size>
  inline void
  resize(Vector& x, const Size& sz) 
  {
    x.resize(sz);
  }


  template <class MTLMat, class VecX, class VecY>
  inline void trans_mult(const MTLMat& A, const VecX& x, const VecY& y) {
    mtl::mult(mtl::trans(A), x, const_cast<VecY&>(y));
  }

  //used inside of GCR and GMRES algorithm

  template <class T>
  struct internal_matrix_traits {
    typedef typename mtl::matrix<T, mtl::rectangle<>,
				 mtl::dense<>,
				 mtl::column_major>::type MTLMat;
  };

  
  template <class Hessenberg, class Vec>
  inline void upper_tri_solve(const Hessenberg& hh, Vec& rs, int i) {
    mtl::tri_solve(mtl::tri_view<mtl::upper>()(hh.sub_matrix(0, i, 0, i)), rs);
  }


}

#endif /*ITL_MTL_INTERFACE_H*/
