/*
* Written by: Thomas Yeo
* 
* Example usage:
* CheckWarpTensorFilter BefTensorImage OutputTensorFile alpha beta gamma
* 
*/

#include "itkImageFileWriter.h"
#include "itkImageFileReader.h"
#include "itkImage.h"
#include "itkImageRegionIterator.h"
#include "itkWarpTensorImageFilter.h"
#include "itkPoint.h"
#include "itkMatrix.h"

#include <getopt.h>
#include <iostream>
#include <string>
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include <math.h>

//#define DEBUG

int main( int argc, char *argv[] )
{   

  if(argc < 6)
    {
    std::cout<<"Usage: CheckWarpTensorFilter BefImage OutputFile alpha beta gamma"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Read in arguments
  const std::string input_file = argv[1];
  const std::string output_file = argv[2];
  float alpha = atof(argv[3]);
  float beta = atof(argv[4]);
  float gamma = atof(argv[5]);

  std::cout<<"input file: "<<input_file<<std::endl;
  std::cout<<"output file: "<<output_file<<std::endl;
  std::cout<<"Rotate by z: "<<gamma<<std::endl;
  std::cout<<"Rotate by y: "<<beta<<std::endl;
  std::cout<<"Rotate by z: "<<alpha<<std::endl;

  //Read in Tensor Image
  typedef itk::Tensor<float, 3> TensorType;
  typedef itk::Image<TensorType, 3> TensorImageType;

  TensorImageType::Pointer tensor_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(tensor_image, input_file);

/*
  const PointType origin = tensor_image->GetOrigin();
  typedef TensorImageType::SpacingType SpacingType;
  SpacingType spacing = tensor_image->GetSpacing();
  typedef TensorImageType::RegionType RegionType;
  RegionType region =  tensor_image->GetLargestPossibleRegion(); 
  typedef TensorImageType::SizeType SizeType;
  SizeType size = region->GetSize();

  float center0 = origin[0] + spacing[0]*size[0]/2.0;
*/




  //Compute Matrix Transformation
  typedef itk::Matrix<float, 3, 3> RotationType;
  RotationType R1;
  R1[0][0] = cos(gamma); R1[0][1] = -sin(gamma); R1[0][2] = 0;
  R1[1][0] = sin(gamma); R1[1][1] = cos(gamma); R1[1][2] = 0;
  R1[2][0] = 0; R1[2][1] = 0; R1[2][2] = 1;
  
#ifdef DEBUG  
  std::cout<<R1(0,0)<<","<< R1(0,1)<< ","<<R1(0,2) <<std::endl;
  std::cout<<R1(1,0)<<","<< R1(1,1)<< ","<<R1(1,2) <<std::endl;
  std::cout<<R1(2,0)<<","<< R1(2,1)<< ","<<R1(2,2) <<std::endl;
#endif
  RotationType R2;
  R2[0][0] = cos(beta);  R2[0][1] = 0; R2[0][2] = sin(beta);
  R2[1][0] = 0;          R2[1][1] = 1; R2[1][2] = 0;
  R2[2][0] = -sin(beta); R2[2][1] = 0; R2[2][2] = cos(beta);

  RotationType R3;
  R3[0][0] = cos(alpha); R3[0][1] = -sin(alpha); R3[0][2] = 0;
  R3[1][0] = sin(alpha); R3[1][1] = cos(alpha);  R3[1][2] = 0;
  R3[2][0] = 0;          R3[2][1] = 0;           R3[2][2] = 1;


  //Create Warp Field
  typedef itk::Vector< float, 3 >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, 3 > DeformationFieldType;
  
  DeformationFieldType::Pointer deformation_field = DeformationFieldType::New();
  deformation_field->SetRegions(tensor_image->GetLargestPossibleRegion());
  deformation_field->SetSpacing(tensor_image->GetSpacing());
  deformation_field->SetOrigin(tensor_image->GetOrigin());
  deformation_field->Allocate();

  //Find Image Center
  typedef itk::Point<float, 3> PointType;
  typedef itk::ImageRegionIterator<TensorImageType> TensorIteratorType;
  
  TensorIteratorType imIt = TensorIteratorType(tensor_image, tensor_image->GetLargestPossibleRegion());
  imIt.GoToEnd(); --imIt;
  PointType EndPt; 
  tensor_image->TransformIndexToPhysicalPoint( imIt.GetIndex(), EndPt );
  
#ifdef DEBUG
  std::cout<<"End Point is: "<<EndPt<<std::endl; 
#endif
  EndPt[0] /= 2;
  EndPt[1] /= 2;
  EndPt[2] /= 2;
#ifdef DEBUG
  std::cout<<"End Point is: "<<EndPt<<std::endl; 
#endif

  //Compute Deformation Field
  typedef itk::ImageRegionIterator<DeformationFieldType> IteratorType;
  IteratorType It = IteratorType(deformation_field, deformation_field->GetLargestPossibleRegion());

  PointType point;
  PointType point2;
  VectorPixelType warp;

  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    // get the point location.
    tensor_image->TransformIndexToPhysicalPoint( It.GetIndex(), point );
    point2[0] = point[0] - EndPt[0];
    point2[1] = point[1] - EndPt[1];
    point2[2] = point[2] - EndPt[2];
    point2 = R3*R2*R1*point2;
    point2[0] = point2[0] + EndPt[0];
    point2[1] = point2[1] + EndPt[1];
    point2[2] = point2[2] + EndPt[2];

    warp[0] = point2[0] - point[0];
    warp[1] = point2[1] - point[1];
    warp[2] = point2[2] - point[2];
    It.Set(warp);
    //transform point to somewhere else
    }

#ifdef DEBUG
  std::cout<<"Point is: "<<point<<std::endl;
  point2 = R1*point;
  std::cout<<"Rotated Point is: "<<point2<<std::endl;
  warp[0] = point2[0] - point[0];
  warp[1] = point2[1] - point[1];
  warp[2] = point2[2] - point[2];
  std::cout<<"Warp Point is: "<<warp<<std::endl;
#endif

  //Performing the warp
  typedef itk::WarpTensorImageFilter<TensorImageType, TensorImageType, DeformationFieldType> WarpTensorImageFilterType;
  WarpTensorImageFilterType::Pointer warper = WarpTensorImageFilterType::New();
  
  warper->SetInput( tensor_image );
  warper->SetOutputSpacing( tensor_image->GetSpacing() );
  warper->SetOutputOrigin( tensor_image->GetOrigin() );
  warper->SetDeformationField(deformation_field);

  try
    {
    warper->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Warper fails." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }
  TensorImageType::Pointer tensor_output = warper->GetOutput();
  
  //Writing the warp
  yeoWriteTensorImage<TensorImageType>(tensor_output, output_file);

  std::cout<<"Warp success!"<<std::endl;
  exit(EXIT_SUCCESS);

}

