
/*
* Written by: Thomas Yeo
* 
* Example usage:
*  CreateRandomWarps input_file output_file mag
* 
*/


#include <getopt.h>
#include <iostream>
#include <string>
#include <time.h>
#include <stdlib.h>
#include <math.h>

#include "GetPot.h"

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"
#include "Debug.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "itkWarpSmoothnessCalculator.h"


/**
   Since ITK 3.8, this filter has been added in global ITK code,
   and it has changed of name,
   please switch the lines if any problems.
*/
// #include "itkWarpJacobianDeterminantFilter.h"
#include "itkDisplacementFieldJacobianDeterminantFilterOld.h"
#include "itkWarpSmoothnessCalculator.h"

#define ImageDimension 3


void PrintHelp (const char* exec)
{
  std::cout << "Usage: " << std::endl;
  std::cout << exec << " <-f fixed image> <-m moving image> <-w warp> <-o output file> <-euc (indicates euclidean interpolation, default log-euc)> <-PPD (if not FSreorientation)>" << std::endl;
  exit (0);
}

int main( int argc, char *argv[] )
{   

  // parse arguments
  GetPot arg (argc, argv);
  if( arg.size() == 1 || arg.search (2,"--help","-h") )
  {
    PrintHelp (arg[0]);    
  }

  const char* fixed_file  = arg.follow("NoFile", "-f");
  const char* moving_file = arg.follow("NoFile", "-m");
  const char* warp_file   = arg.follow("NoFile", "-w");
  const char* output_file = arg.follow("NoFile", "-o");

  const bool PPD = arg.search (1,"-PPD");
  const bool EUC = arg.search (1,"-euc");

  
  std::cout<<"PPD: "<<PPD<<std::endl;
  std::cout<<"EUC: "<<EUC<<std::endl;

  //Some Typedefs
  typedef float PixelType;
  typedef itk::Vector< PixelType, ImageDimension >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, ImageDimension > DeformationFieldType;
  typedef itk::Image<PixelType, ImageDimension> MaskImageType;

  typedef itk::Tensor<float, 3> TensorType;
  typedef itk::Image<TensorType, ImageDimension> TensorImageType;

  //Read in Tensor Images
  std::cout<<"Read Fixed Tensor Image: "<<fixed_file<<std::endl;
  TensorImageType::Pointer fixed_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(fixed_image, fixed_file);
  
  std::cout<<"Read Moving Tensor Image: "<<moving_file<<std::endl;
  TensorImageType::Pointer moving_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(moving_image, moving_file);

  //Read in Warp
  std::cout<<"Read Warps: "<<warp_file<<std::endl;
  DeformationFieldType::Pointer df = DeformationFieldType::New();
  yeoReadImage<DeformationFieldType>(df, warp_file);

  //Log fixed image
  std::cout<<"Log-ing Fixed Image"<<std::endl;
  TensorImageType::Pointer log_fixed_image = TensorImageType::New();
  yeoLogTensorImageNewImage<TensorImageType>(fixed_image, log_fixed_image);

  //Compute foreground mask
  std::cout<<"Compute Foreground mask for fixed image"<<std::endl;
  MaskImageType::Pointer fixed_image_fg = MaskImageType::New();
  yeoGetForegroundMask<TensorImageType, MaskImageType>(fixed_image, fixed_image_fg, itk::NumericTraits<TensorImageType::PixelType>::Zero, 0, 1);
  const double numPixels = yeoSumImage<MaskImageType>(fixed_image_fg);

  //Warp Tensor image
  TensorImageType::Pointer warped_image = TensorImageType::New();
  TensorImageType::Pointer log_warped_image = TensorImageType::New();

  if(!EUC) //need to exponentiate
    {
    std::cout<<"Log-ing moving image"<<std::endl;
    yeoLogTensorImageInPlace<TensorImageType>(moving_image);

    std::cout<<"Warp image"<<std::endl;    
    if(PPD)
      yeoWarpTensorImageNewImagePPD<TensorImageType, DeformationFieldType>(moving_image, df, log_warped_image);
    else
      yeoWarpTensorImageNewImage<TensorImageType, DeformationFieldType>(moving_image, df, log_warped_image);

    std::cout<<"Exp warped image"<<std::endl;
    yeoExpTensorImageNewImage<TensorImageType>(log_warped_image, warped_image);

    yeoExpTensorImageInPlace<TensorImageType>(moving_image);
    }
  else
    {
    std::cout<<"Warp image"<<std::endl;
    if(PPD)
      yeoWarpTensorImageNewImagePPD<TensorImageType, DeformationFieldType>(moving_image, df, warped_image);
    else
      yeoWarpTensorImageNewImage<TensorImageType, DeformationFieldType>(moving_image, df, warped_image);

    std::cout<<"Log-ing warped image"<<std::endl;
    yeoLogTensorImageNewImage<TensorImageType>(warped_image, log_warped_image);
    }

  //std::cout<<"Compute Foreground mask for warped moving image"<<std::endl;
  //MaskImageType::Pointer warped_image_fg = MaskImageType::New();
  //yeoGetForegroundMask<TensorImageType, MaskImageType>(warped_image, warped_image_fg, itk::NumericTraits<TensorImageType::PixelType>::Zero, 0, 1);

  //MaskImageType::Pointer fg_mask = MaskImageType::New();
  //yeoPowerCombine2ImagesNewImage<MaskImageType>(fixed_image_fg, warped_image_fg, fg_mask, 1, 1);
  //const double numPixels = yeoSumImage<MaskImageType>(fg_mask);

  //Compute Statistics
  std::cout<<"Computing statistics and writing to "<<output_file<<" ... "<<std::endl;
  

  //harmonic energy
  std::cout<<"1. Compute harmonic energy: ";
  typedef itk::WarpSmoothnessCalculator<DeformationFieldType> SmoothnessCalculatorType;
  SmoothnessCalculatorType::Pointer m_SmoothnessCalculator = SmoothnessCalculatorType::New();
  m_SmoothnessCalculator->SetImage( df );
  m_SmoothnessCalculator->Compute();
  const double harmonicEnergy = m_SmoothnessCalculator->GetSmoothness();
  
  ofstream myfile;
  myfile.open(output_file, ios::app);
  myfile << harmonicEnergy; myfile << "\n";
  std::cout<<harmonicEnergy<<std::endl;

  //log_mse
  std::cout<<"2. Compute LOG-MSE: ";
  TensorImageType::Pointer temp_tensor_image = TensorImageType::New();
  yeoLinearCombine2ImagesNewImage<TensorImageType>(log_fixed_image, log_warped_image, temp_tensor_image, 1, -1);
  yeoPowerCombine2TensorImagesInPlace<TensorImageType>(temp_tensor_image, temp_tensor_image, 1.0, 1.0);
  yeoMultiplyTensorImageWithImageInPlace<TensorImageType, MaskImageType>(temp_tensor_image, fixed_image_fg);
  const double logMSE = yeoSumTensorImage<TensorImageType>(temp_tensor_image)/numPixels;
  myfile << logMSE; myfile << "\n";
  std::cout<<logMSE<<std::endl;

  //mse
  std::cout<<"3. Compute MSE: ";
  yeoLinearCombine2ImagesNewImage<TensorImageType>(fixed_image, warped_image, temp_tensor_image, 1, -1);
  yeoPowerCombine2TensorImagesInPlace<TensorImageType>(temp_tensor_image, temp_tensor_image, 1.0, 1.0);
  yeoMultiplyTensorImageWithImageInPlace<TensorImageType, MaskImageType>(temp_tensor_image, fixed_image_fg);
  const double MSE = yeoSumTensorImage<TensorImageType>(temp_tensor_image)/numPixels;
  myfile << MSE; myfile << "\n";
  std::cout<<MSE<<std::endl;

  //Compute scalar measures 
  MaskImageType::Pointer fixed_scalar_image  = MaskImageType::New();
  MaskImageType::Pointer warped_scalar_image = MaskImageType::New();
  MaskImageType::Pointer temp_scalar_image   = MaskImageType::New();
  MaskImageType::Pointer moving_scalar_image = MaskImageType::New();

  //FA
  std::cout<<"4. Compute FA: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "fa");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "fa");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double fa = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << fa; myfile << "\n";
  std::cout<<fa<<std::endl;

  std::cout<<"5. Compute wFA: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "fa");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wfa = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wfa; myfile << "\n";
  std::cout<<wfa<<std::endl;

  //LFA
  std::cout<<"6. Compute LFA: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "lfa");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "lfa");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double lfa = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << lfa; myfile << "\n";
  std::cout<<lfa<<std::endl;

  std::cout<<"7. Compute wLFA: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "lfa");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wlfa = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wlfa; myfile << "\n";
  std::cout<<wlfa<<std::endl;
  
  //ADC
  std::cout<<"8. Compute ADC: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "adc");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "adc");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double adc = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << adc; myfile << "\n";
  std::cout<<adc<<std::endl;

  std::cout<<"9. Compute wADC: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "adc");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wadc = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wadc; myfile << "\n";
  std::cout<<wadc<<std::endl;

  //VOL
  std::cout<<"10. Compute VOL: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "vol");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "vol");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double vol = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << vol; myfile << "\n";
  std::cout<<vol<<std::endl;

  std::cout<<"11. Compute wVOL: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "vol");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wvol = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wvol; myfile << "\n";
  std::cout<<wvol<<std::endl;

  //CL
  std::cout<<"12. Compute CL: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "cl");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "cl");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double cl = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << cl; myfile << "\n";
  std::cout<<cl<<std::endl;

  std::cout<<"13. Compute wcl: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "cl");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wcl = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wcl; myfile << "\n";
  std::cout<<wcl<<std::endl;

  //CP
  std::cout<<"14. Compute CP: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "cp");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "cp");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double cp = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << cp; myfile << "\n";
  std::cout<<cp<<std::endl;

  std::cout<<"15. Compute wCP: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "cp");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wcp = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wcp; myfile << "\n";
  std::cout<<wcp<<std::endl;

  //cs
  std::cout<<"16. Compute CS: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "cs");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "cs");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double cs = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << cs; myfile << "\n";
  std::cout<<cs<<std::endl;

  std::cout<<"17. Compute wCS: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "cs");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wcs = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wcs; myfile << "\n";
  std::cout<<wcs<<std::endl;

  //ra
  std::cout<<"18. Compute RA: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "ra");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "ra");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double ra = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << ra; myfile << "\n";
  std::cout<<ra<<std::endl;

  std::cout<<"19. Compute wRA: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "ra");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wra = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wra; myfile << "\n";
  std::cout<<wra<<std::endl;

  //vr
  std::cout<<"20. Compute VR: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "vr");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "vr");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double vr = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << vr; myfile << "\n";
  std::cout<<vr<<std::endl;

  std::cout<<"21. Compute wVR: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "vr");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wvr = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wvr; myfile << "\n";
  std::cout<<wvr<<std::endl;

  //L1
  std::cout<<"22. Compute L1: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "l1");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "l1");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double l1 = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << l1; myfile << "\n";
  std::cout<<l1<<std::endl;

  std::cout<<"23. Compute wL1: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "l1");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wl1 = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wl1; myfile << "\n";
  std::cout<<wl1<<std::endl;

  //L2
  std::cout<<"24. Compute L2: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "l2");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "l2");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double l2 = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << l2; myfile << "\n";
  std::cout<<l2<<std::endl;

  std::cout<<"25. Compute wL2: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "l2");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wl2 = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wl2; myfile << "\n";
  std::cout<<wl2<<std::endl;

  //L3
  std::cout<<"26. Compute L3: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "l3");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "l3");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double l3 = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << l3; myfile << "\n";
  std::cout<<l3<<std::endl;

  std::cout<<"27. Compute wL3: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "l3");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wl3 = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wl3; myfile << "\n";
  std::cout<<wl3<<std::endl;

  //Dispersion
  std::cout<<"28. Compute Disp: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(fixed_image, fixed_scalar_image, "dp");
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(warped_image, warped_scalar_image, "dp");
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double dp = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << dp; myfile << "\n";
  std::cout<<dp<<std::endl;

  std::cout<<"29. Compute wDisp: ";
  yeoTensor2ScalarImage<TensorImageType, MaskImageType>(moving_image, moving_scalar_image, "dp");
  yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(moving_scalar_image, df, warped_scalar_image);
  yeoLinearCombine2ImagesNewImage<MaskImageType>(fixed_scalar_image, warped_scalar_image, temp_scalar_image, 1, -1);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, temp_scalar_image, 1.0, 1.0);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double wdp = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << wdp; myfile << "\n";
  std::cout<<wdp<<std::endl;

  //Overlap
  std::cout<<"30. Compute Overlap: ";
  yeoComputeOverlap2TensorImages<TensorImageType, MaskImageType>(fixed_image, warped_image, temp_scalar_image);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(temp_scalar_image, fixed_image_fg, 1.0, 1.0);
  const double overlap = yeoSumImage<MaskImageType>(temp_scalar_image)/numPixels;
  myfile << overlap; myfile << "\n";
  std::cout<<overlap<<std::endl;


  myfile.close();
  exit(EXIT_SUCCESS);



}

