
/*
* Written by: Thomas Yeo
* 
* Example usage:
*  Compare2WarpFields warp_field1 warp_field2 filename
* 
*/


#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"

#include <getopt.h>
#include <iostream>
#include <fstream>
#include <string>
#include "Debug.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "itkWarpSmoothnessCalculator.h"

#define ImageDimension 3

int main( int argc, char *argv[] )
{   
  
  if(argc < 4)
    {
    std::cout<<"Compare2WarpFields warp_field1 warp_field2 tensor_mask output_filename"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Read in Arguments
  const std::string warp_field1 = argv[1];
  const std::string warp_field2 = argv[2];
  const std::string tensor_filename = argv[3];
  const std::string output_filename = argv[4];
  
  //Some Typedefs
  typedef float PixelType;
  typedef itk::Vector< PixelType, ImageDimension >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, ImageDimension > DeformationFieldType;
  
  std::cout<<"Read in warp field 1: "<<warp_field1<<std::endl;
  DeformationFieldType::Pointer df1 = DeformationFieldType::New();
  yeoReadImage<DeformationFieldType>(df1, warp_field1);

  std::cout<<"Read in warp field 2: "<<warp_field2<<std::endl;
  DeformationFieldType::Pointer df2 = DeformationFieldType::New();
  yeoReadImage<DeformationFieldType>(df2, warp_field2);

  typedef itk::Tensor<float, 3> TensorType;
  typedef itk::Image<TensorType, ImageDimension> TensorImageType;

  //Read in Tensor Image
  std::cout<<"Read Tensor Image: "<<tensor_filename<<std::endl;
  TensorImageType::Pointer tensor_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(tensor_image, tensor_filename);

  //Compute Mask
  typedef itk::Image<PixelType, ImageDimension> MaskImageType;
  MaskImageType::Pointer input_fg = MaskImageType::New();
  yeoGetForegroundMask<TensorImageType, MaskImageType>(tensor_image, input_fg, itk::NumericTraits<TensorImageType::PixelType>::Zero, 0, 1);
  
  //Compute Harmonic Energy
  typedef itk::WarpSmoothnessCalculator<DeformationFieldType> SmoothnessCalculatorType;
  SmoothnessCalculatorType::Pointer SmoothnessCalculator;
  SmoothnessCalculator = SmoothnessCalculatorType::New();
  SmoothnessCalculator->SetImage(df1);
  SmoothnessCalculator->Compute();
  const double harmonicEnergy1 = SmoothnessCalculator->GetSmoothness();
  std::cout<<"Harmonic energy: "<<harmonicEnergy1<<std::endl;

  SmoothnessCalculator->SetImage(df2);
  SmoothnessCalculator->Compute();
  const double harmonicEnergy2 = SmoothnessCalculator->GetSmoothness();
  std::cout<<"Harmonic energy: "<<harmonicEnergy2<<std::endl;


  MaskImageType::Pointer orig_norm_image = MaskImageType::New();
  yeoInnerProductVectorFieldNewImage<DeformationFieldType, MaskImageType>(df1, df1, orig_norm_image);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(orig_norm_image, input_fg, 1.0, 1.0);
  double origSSD = yeoSumImage<MaskImageType>(orig_norm_image);

  yeoPowerCombine2ImagesInPlace<MaskImageType>(orig_norm_image, input_fg, 0.5, 1.0);
  double origDist = yeoSumImage<MaskImageType>(orig_norm_image);

  //Compute Statistics
  yeoLinearCombine2ImagesInPlace<DeformationFieldType>(df1, df2, 1, -1);
  
  MaskImageType::Pointer df_norm_image = MaskImageType::New();
  yeoInnerProductVectorFieldNewImage<DeformationFieldType, MaskImageType>(df1, df1, df_norm_image);

  //Apply Mask to Statistics
  yeoPowerCombine2ImagesInPlace<MaskImageType>(df_norm_image, input_fg, 1.0, 1.0);
  
  double numPixels = yeoSumImage<MaskImageType>(input_fg);
  double totalSSD = yeoSumImage<MaskImageType>(df_norm_image);
  double averageSSD = totalSSD/numPixels;
  double averageOrigSSD = origSSD/numPixels;

  std::cout<<"Average SSD: "<<averageSSD<<", orig SSD: "<<averageOrigSSD<<std::endl;
  
  yeoPowerCombine2ImagesInPlace<MaskImageType>(df_norm_image, input_fg, 0.5, 1.0);
  double averageDist = yeoSumImage<MaskImageType>(df_norm_image)/numPixels;
  double averageOrigDist = origDist/numPixels;
  std::cout<<"Average Dist: "<<averageDist<<", Average Orig Dist: "<<averageOrigDist<<std::endl;

  ofstream myfile;
  myfile.open(output_filename.c_str(), ios::app);
  myfile <<harmonicEnergy1;
  myfile <<"\n";
  myfile <<harmonicEnergy2;
  myfile <<"\n";
  myfile << averageOrigSSD;
  myfile <<"\n";
  myfile << averageSSD;
  myfile << "\n";
  myfile << averageOrigDist;
  myfile << "\n";
  myfile << averageDist;
  myfile << "\n";
  myfile.close();
  exit(EXIT_SUCCESS);
}

