
/*
* Written by: Thomas Yeo
* 
* Example usage:
*      ComputeInverseConsistency warp_field1 warp_field2 output_filename
* 
*/


#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"

#include <getopt.h>
#include <iostream>
#include <fstream>
#include <string>
#include "Debug.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "itkWarpSmoothnessCalculator.h"

#define ImageDimension 3

using namespace std; 

int main( int argc, char *argv[] )
{   
  
  if(argc < 4)
    {
    std::cout<<"ComputeInverseConsistency warp_field1 warp_field2 tensor_image output_filename"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Read in Arguments
  const std::string warp_field1 = argv[1];
  const std::string warp_field2 = argv[2];
  const std::string tensor_filename = argv[3];
  const std::string output_filename = argv[4];
  
  //Some Typedefs
  typedef float PixelType;
  typedef itk::Vector< PixelType, ImageDimension >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, ImageDimension > DeformationFieldType;
  typedef itk::Image< PixelType, ImageDimension> ScalarImageType;
  typedef itk::Tensor<float, 3> TensorType;
  typedef itk::Image<TensorType, ImageDimension> TensorImageType;


  std::cout<<"Read in warp field 1: "<<warp_field1<<std::endl;
  DeformationFieldType::Pointer df1 = DeformationFieldType::New();
  yeoReadImage<DeformationFieldType>(df1, warp_field1);

  std::cout<<"Read in warp field 2: "<<warp_field2<<std::endl;
  DeformationFieldType::Pointer df2 = DeformationFieldType::New();
  yeoReadImage<DeformationFieldType>(df2, warp_field2);
  
  //Read in Tensor Images
  std::cout<<"Read Tensor Image: "<<tensor_filename<<std::endl;
  TensorImageType::Pointer tensor_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(tensor_image, tensor_filename);

  //Compute Harmonic Energy
  typedef itk::WarpSmoothnessCalculator<DeformationFieldType> SmoothnessCalculatorType;
  SmoothnessCalculatorType::Pointer SmoothnessCalculator;
  SmoothnessCalculator = SmoothnessCalculatorType::New();
  SmoothnessCalculator->SetImage(df1);
  SmoothnessCalculator->Compute();
  const double harmonicEnergy1 = SmoothnessCalculator->GetSmoothness();
  std::cout<<"Harmonic energy: "<<harmonicEnergy1<<std::endl;

  SmoothnessCalculator->SetImage(df2);
  SmoothnessCalculator->Compute();
  const double harmonicEnergy2 = SmoothnessCalculator->GetSmoothness();
  std::cout<<"Harmonic energy: "<<harmonicEnergy2<<std::endl;

  std::cout<<"Avg Harmonic energy: "<<(harmonicEnergy1 + harmonicEnergy2)/2<<std::endl;

  //Compute foreground mask
  std::cout<<"Compute Foreground mask for tensor image"<<std::endl;
  ScalarImageType::Pointer mask_image = ScalarImageType::New();
  yeoGetForegroundMask<TensorImageType, ScalarImageType>(tensor_image, mask_image, itk::NumericTraits<TensorImageType::PixelType>::Zero, 0, 1);
  const double numPixels = yeoSumImage<ScalarImageType>(mask_image);
  std::cout<<"Num Pixels: "<<numPixels<<std::endl;


  //Compute Inverse Consistency
  std::cout<<"Compose the warp ..."<<std::endl;
  DeformationFieldType::Pointer out = DeformationFieldType::New();
  yeoWarpImageNewImage<DeformationFieldType, DeformationFieldType>(df2, df1, out);

  std::cout<<"Subtracting original warp ..."<<std::endl;
  yeoLinearCombine2ImagesInPlace<DeformationFieldType>(out, df1, 1.0, 1.0);  

  std::cout<<"Computing Inverse Consistency ..."<<std::endl;
  ScalarImageType::Pointer norm_image = ScalarImageType::New();
  yeoInnerProductVectorFieldNewImage<DeformationFieldType, ScalarImageType>(out, out, norm_image);

  yeoPowerCombine2ImagesInPlace<ScalarImageType>(norm_image, mask_image, 0.5, 1);

  const double maxDev = yeoMaxValue<ScalarImageType>(norm_image);
  std::cout<<"Max Deviation From Identity: "<<maxDev<<std::endl;

  const double meanDev = yeoSumImage<ScalarImageType>(norm_image)/numPixels;
  std::cout<<"Mean Deviation From Identity: "<<meanDev<<std::endl;

  std::ofstream myfile;
  myfile.open(output_filename.c_str(), std::ios::app);
  myfile << (harmonicEnergy1 + harmonicEnergy2)/2;
  myfile <<"\n";
  myfile << maxDev;
  myfile <<"\n";
  myfile << meanDev;
  myfile <<"\n";
  myfile.close();
  exit(EXIT_SUCCESS);
}

