
/*
* Written by: Thomas Yeo
* 
* Example usage:
*  ComputeTensorImagesMSE fixedImage movingImage deformationField MSE_filename LogEuclideanBool
* 
*/


#include "itkImage.h"

#include <getopt.h>
#include <iostream>
#include <string>
#include "Debug.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "itkTensor.h"

#define ImageDimension 3
int main( int argc, char *argv[] )
{   

  if(argc < 6)
    {
    std::cout<<"ComputeTensorImagesMSE fixedImage movingImage deformationField MSE_filename LogEuclideanBool"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Read in Arguments
  const std::string fixedImageFilename = argv[1];
  const std::string movingImageFilename = argv[2];
  const std::string deformationFilename = argv[3];
  const std::string outputFilename = argv[4];
  const int logEuclideanBool = atoi(argv[5]);

  typedef float PixelType;
  typedef itk::Tensor<PixelType, 3> TensorType;
  typedef itk::Image<TensorType, ImageDimension> TensorImageType;
  
  typedef itk::Vector< PixelType, ImageDimension >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, ImageDimension > DeformationFieldType;
  typedef itk::Image<PixelType, ImageDimension> MaskImageType; 

  
  std::cout<<"Reading fixed Image: "<<fixedImageFilename<<std::endl;
  TensorImageType::Pointer fixed_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(fixed_image, fixedImageFilename);

  std::cout<<"Reading moving Image: "<<movingImageFilename<<std::endl;
  TensorImageType::Pointer moving_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(moving_image, movingImageFilename);

  std::cout<<"Reading Deformation Field: "<<deformationFilename<<std::endl;
  DeformationFieldType::Pointer df = DeformationFieldType::New();
  yeoReadImage<DeformationFieldType>(df, deformationFilename);


  if(logEuclideanBool)
    {
    std::cout<<"Log fixed Image "<<std::endl;
    yeoLogTensorImageInPlace<TensorImageType>(fixed_image);
    
    std::cout<<"Log moving Image "<<std::endl;
    yeoLogTensorImageInPlace<TensorImageType>(moving_image);  
    }

  std::cout<<"Warp Tensor Image"<<std::endl;
  TensorImageType::Pointer warped_moving_image = TensorImageType::New();
  yeoWarpTensorImageNewImage<TensorImageType, DeformationFieldType>(moving_image, df, warped_moving_image);
 
  std::cout<<"Computing Fixed Image fg Mask"<<std::endl;
  MaskImageType::Pointer fixedImage_fg = MaskImageType::New();
  yeoGetForegroundMask<TensorImageType, MaskImageType>(fixed_image, fixedImage_fg, itk::NumericTraits<TensorImageType::PixelType>::Zero, 0, 1);
                                            
  std::cout<<"Computing Warped moving image fg Mask"<<std::endl; 
  MaskImageType::Pointer movingImage_fg = MaskImageType::New();
  yeoGetForegroundMask<TensorImageType, MaskImageType>(warped_moving_image, movingImage_fg, itk::NumericTraits<TensorImageType::PixelType>::Zero, 0, 1);
      
  yeoPowerCombine2ImagesInPlace<MaskImageType>(fixedImage_fg, movingImage_fg, 1, 1);

  //compute SSD
  std::cout<<"Computing SSD"<<std::endl;
  yeoLinearCombine2ImagesInPlace<TensorImageType>(warped_moving_image, fixed_image, 1.0, -1.0);
  yeoPowerCombine2TensorImagesInPlace<TensorImageType>(warped_moving_image, warped_moving_image, 1.0, 1.0);

  //apply mask
  yeoMultiplyTensorImageWithImageInPlace<TensorImageType, MaskImageType>(warped_moving_image, fixedImage_fg);
  

  double numPixels = yeoSumImage<MaskImageType>(fixedImage_fg);
  std::cout<<"MSE fixed image vs warped moving image averaged over "<<numPixels<<" pixels: "<<yeoSumTensorImage<TensorImageType>(warped_moving_image)/numPixels<<std::endl;
   
  //write SSD
  MaskImageType::Pointer SSDimage = MaskImageType::New();
  yeoSumTensorImageNewImage<TensorImageType, MaskImageType>(warped_moving_image, SSDimage);
  yeoWriteImage<MaskImageType>(SSDimage, outputFilename);

  std::cout<<"MSE computation success!"<<std::endl;
  exit(EXIT_SUCCESS);
}

