
/*
* Written by: Thomas Yeo, MIT, CSAIL @ Asclepios, INRIA
*
* sent bug reports to ythomas@mit.edu 
* 
* Example usage:
* CreateAtlasFromWarpedImages SUBJECTS_DIR input_filename mask_filename mean_filename warp_filename var_filename ATLAS_SUBJECTS
* 
* CreateAtlasFromWarpedImages
* /auto/cardan/u/cardan/local/tmp-images/yboon/data/ 
* /raw/B0_corrected.nii
* /raw/B0_fg_mask.nii.gz
* /output/warp_field.nhdr
* /auto/cardan/u/cardan/local/tmp-images/yboon/data/atlas/B0_mean0.nhdr
* /auto/cardan/u/cardan/local/tmp-images/yboon/data/atlas/B0_var0.nhdr
* 0
* 5129  5131  5132  5138  5142  5146  5148  5152  5153  5156
*/


#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"
#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkHistogramMatchingImageFilter.h"
#include "UsefulFunctions.h"
#include "itkWarpImageFilter.h"

#include <getopt.h>
#include <iostream>
#include <string>
#include "Debug.h"
//#define DEBUG

#define BASIC_PARMS 7
#define MIN_VAR_PERCENT 1e-6
//#define MIN_VAR 0.01

int main( int argc, char *argv[] )
{   

  if(argc < 7)
    {
    std::cout<<"CreateAtlasFromWarpedImages SUBJECTS_DIR input_filename mask_filename mean_filename warp_filename var_filename ATLAS_SUBJECTS"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  //Read in Arguments
  const std::string SUBJECTS_DIR = argv[1];
  const std::string input_filename = argv[2];
  const std::string mask_filename = argv[3];
  const std::string warp_filename = argv[4];
  const std::string mean_filename = argv[5];
  const std::string var_filename = argv[6];
  int hist_norm = atoi(argv[7]);
  int num_subjects = argc - 1 - BASIC_PARMS;

  std::cout<<"SUBJECTS_DIR: "<<SUBJECTS_DIR<<std::endl;
  std::cout<<"input_filename: "<<input_filename<<std::endl;
  std::cout<<"mask_filename: "<<mask_filename<<std::endl;
  std::cout<<"warp_filename: "<<warp_filename<<std::endl;
  std::cout<<"mean_filename: "<<mean_filename<<std::endl;
  std::cout<<"var_filename: "<<var_filename<<std::endl;
  std::cout<<"Num Subjects: "<<num_subjects<<std::endl;
  std::cout<<"Hist Norm: "<<hist_norm<<std::endl;

  //Check if there's input warp
  int warp_bool = 1;
  std::string warp_comparison;
  warp_comparison = "NULL";
  if(warp_comparison == warp_filename)
    warp_bool = 0;
  std::cout<<"Warp Bool: "<<warp_bool<<std::endl; 

  //Set Image Type for Mean
  typedef float MeanPixelType;
  const unsigned int Dimension = 3;
  typedef itk::Image< MeanPixelType, Dimension >   MeanImageType;
  MeanImageType::Pointer mean_image = MeanImageType::New();

  //Set up mean iterators
  typedef itk::ImageRegionIterator< MeanImageType> MeanIteratorType;
  MeanIteratorType meanIt; //mean image

   //Set Image Type for Var
  typedef MeanPixelType VarPixelType;
  typedef MeanImageType VarImageType;
  VarImageType::Pointer var_image = VarImageType::New();

  //Set up Var iterators
  typedef itk::ImageRegionIterator< VarImageType> VarIteratorType;
  VarIteratorType varIt; //var image

  //Set Image Type for Input
  typedef MeanPixelType InputPixelType;
  typedef MeanImageType InputImageType;

  //Set up Mask 
  InputImageType::Pointer mask_image = InputImageType::New();
  InputImageType::Pointer fg_image = InputImageType::New();

  //Set up Warp
  typedef itk::Vector< float, Dimension >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, Dimension > DeformationFieldType;
  DeformationFieldType::Pointer df = DeformationFieldType::New();


  //Create reference image in case of histogram normalization
  std::string ref_filename;
  ref_filename = SUBJECTS_DIR + argv[BASIC_PARMS+1] + input_filename;
  std::cout<<"Reading Reference: "<<ref_filename<<std::endl;
  
  //Setting up ITK reader for first image
  typedef itk::ImageFileReader< InputImageType > RefImageReaderType;
  RefImageReaderType::Pointer RefImageReader = RefImageReaderType::New();
  RefImageReader->SetFileName(ref_filename);
  
  //Reading in file
  try
    {
    RefImageReader->Update();
    }
  catch(itk::ExceptionObject& err)
    {
    std::cout<<"Image Read Fails"<<std::endl;
    exit(EXIT_FAILURE);
    }
 
  InputImageType::Pointer ref_image =  InputImageType::New();
  //warping if exists
  if(warp_bool)
    {
    InputImageType::Pointer ref_image2 = RefImageReader->GetOutput();
    yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[BASIC_PARMS+1] + warp_filename);
    yeoWarpImageNewImage<InputImageType, DeformationFieldType>(ref_image2, df, ref_image);
    }
  else
    ref_image = RefImageReader->GetOutput();

  //diagnostics
#ifdef DEBUG
  MeanImageType::IndexType pixelIndex;
  pixelIndex[0] = 128;
  pixelIndex[1] = 128;
  pixelIndex[2] = 35;    
  MeanImageType::PixelType pixelValue;

  VarImageType::IndexType varPixelIndex;
  varPixelIndex[0] = 128;
  varPixelIndex[1] = 128;
  varPixelIndex[2] = 35;    
  VarImageType::PixelType varPixelValue;
#endif
  

//First Compute Mean Image
  std::cout<<"==================Computing Mean================="<<std::endl;
  for(int i = BASIC_PARMS+1; i <= BASIC_PARMS + num_subjects; i++) //for each subject
    {
    //create filename
    std::string filename;
    filename = SUBJECTS_DIR + argv[i] + input_filename;
    std::cout<<"Reading: "<<filename<<std::endl;
    
    //Setting up ITK reader
    typedef itk::ImageFileReader< InputImageType > InputImageReaderType;
    InputImageReaderType::Pointer InputImageReader = InputImageReaderType::New();
    InputImageReader->SetFileName( filename);
    
    //Reading in file
    try
      {
      InputImageReader->Update();
      }
    catch(itk::ExceptionObject& err)
      {
      std::cout<<"Image Read Fails"<<std::endl;
      exit(EXIT_FAILURE);
      }

    InputImageType::Pointer input_image = InputImageType::New();
    //warping if exists
    if(warp_bool)
      {
      InputImageType::Pointer input_image2;
      input_image2 = InputImageReader->GetOutput();
      yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[i] + warp_filename);
      yeoWarpImageNewImage<InputImageType, DeformationFieldType>(input_image2, df, input_image);
      }
    else
      input_image = InputImageReader->GetOutput();
    
    //Histogram Normalized!    
    if(hist_norm)
      {
      std::cout<<"Intensity Normalization..."<<std::endl;
      typedef itk::HistogramMatchingImageFilter<InputImageType, InputImageType> MatchingFilterType;
      MatchingFilterType::Pointer matcher = MatchingFilterType::New();
      matcher->SetInput( input_image );
      matcher->SetReferenceImage( ref_image );
      matcher->SetNumberOfHistogramLevels( 1024 );
      matcher->SetNumberOfMatchPoints( 7 );
      matcher->ThresholdAtMeanIntensityOn();
      // Update the matcher
      try
	{
	matcher->Update();
	}
      catch( itk::ExceptionObject& err )
	{
	std::cout << "Could not match the input images." << std::endl;
	std::cout << err << std::endl;
	exit( EXIT_FAILURE );
	}
      input_image = matcher->GetOutput();
      }
    

    //Setting up input Iterator
    typedef itk::ImageRegionConstIterator< InputImageType> InputConstIteratorType;
    InputConstIteratorType inputIt = InputConstIteratorType (input_image, input_image->GetLargestPossibleRegion());

 
    if(i == BASIC_PARMS + 1)
      {
      //set up mean image if it is the first round
      mean_image->SetRegions(input_image->GetLargestPossibleRegion());
      mean_image->SetSpacing(input_image->GetSpacing());
      mean_image->SetOrigin(input_image->GetOrigin());
      mean_image->Allocate();
      
      //Initialize to 0;
      mean_image->FillBuffer(0.0);
      
      //Set up mean_image iterator
      meanIt = MeanIteratorType(mean_image, mean_image->GetLargestPossibleRegion());
      
      //set up var image if it is the first round
      var_image->SetRegions(input_image->GetLargestPossibleRegion());
      var_image->SetSpacing(input_image->GetSpacing());
      var_image->SetOrigin(input_image->GetOrigin());
      var_image->Allocate();
      
      //Initialize to 0;
      var_image->FillBuffer(0.0);

      //Set up var_image iterator
      varIt = VarIteratorType(var_image, var_image->GetLargestPossibleRegion());
      
      //Diagnostics
#ifdef DEBUG
      pixelValue = mean_image->GetPixel(pixelIndex);
      std::cout<<"Mean pixel value: "<<pixelValue<<std::endl;

      varPixelValue = var_image->GetPixel(varPixelIndex);
      std::cout<<"Var pixel value: "<<varPixelValue<<std::endl;
#endif
      }

    //Diagnostics    
#ifdef DEBUG
    InputImageType::IndexType pixelIndex1;
    pixelIndex1[0] = 128;
    pixelIndex1[1] = 128;
    pixelIndex1[2] = 35;
    
    InputImageType::PixelType pixelValue1 = input_image->GetPixel(pixelIndex1);
    std::cout<<"Input pixel value: "<<pixelValue1<<std::endl;
#endif
    
    //Add contribution of each subject to mean
    meanIt.GoToBegin();
    inputIt.GoToBegin();
    while(!meanIt.IsAtEnd())
      {
      if(inputIt.IsAtEnd())
	{
	std::cerr<<"input iterator ends before mean iterator"<<std::endl;
	exit(EXIT_FAILURE);
	}
      
      meanIt.Set( meanIt.Get()  + (MeanPixelType) inputIt.Get() / num_subjects);
      ++meanIt; ++inputIt;
      }
    
    //Diagnostics
#ifdef DEBUG
    pixelValue = mean_image->GetPixel(pixelIndex);
    std::cout<<"Mean pixel value: "<<pixelValue<<std::endl;
#endif

    }
  

//Compute Var Image
  std::cout<<"==================Computing Var================="<<std::endl;
  for(int i = BASIC_PARMS+1; i <= BASIC_PARMS + num_subjects; i++) //for each subject
    {
    //create filename
    std::string filename;
    filename = SUBJECTS_DIR + argv[i] + input_filename;
    std::cout<<"Reading: "<<filename<<std::endl;
    
    //Setting up ITK reader
    typedef itk::ImageFileReader< InputImageType > InputImageReaderType;
    InputImageReaderType::Pointer InputImageReader = InputImageReaderType::New();
    InputImageReader->SetFileName( filename);
    
    //Reading in file
    try
      {
      InputImageReader->Update();
      }
    catch(itk::ExceptionObject& err)
      {
      std::cout<<"Image Read Fails"<<std::endl;
      exit(EXIT_FAILURE);
      }

    InputImageType::Pointer input_image = InputImageType::New();
    //warping if exists
    if(warp_bool)
      {
      InputImageType::Pointer input_image2;
      input_image2 = InputImageReader->GetOutput();
      yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[i] + warp_filename);
      yeoWarpImageNewImage<InputImageType, DeformationFieldType>(input_image2, df, input_image);
      }
    else
      input_image = InputImageReader->GetOutput();
    
    //Histogram Normalized!    
    if(hist_norm)
      {
      std::cout<<"Intensity Normalization..."<<std::endl;
      typedef itk::HistogramMatchingImageFilter<InputImageType, InputImageType> MatchingFilterType;
      MatchingFilterType::Pointer matcher = MatchingFilterType::New();
      matcher->SetInput( input_image );
      matcher->SetReferenceImage( ref_image );
      matcher->SetNumberOfHistogramLevels( 1024 );
      matcher->SetNumberOfMatchPoints( 7 );
      matcher->ThresholdAtMeanIntensityOn();
      // Update the matcher
      try
	{
	matcher->Update();
	}
      catch( itk::ExceptionObject& err )
	{
	std::cout << "Could not match the input images." << std::endl;
	std::cout << err << std::endl;
	exit( EXIT_FAILURE );
	}
      input_image = matcher->GetOutput();
      }

    
    //Setting up input Iterator
    typedef itk::ImageRegionConstIterator< InputImageType> InputConstIteratorType;
    InputConstIteratorType inputIt = InputConstIteratorType (input_image, input_image->GetLargestPossibleRegion());

    

    //Diagnostics    
#ifdef DEBUG
    InputImageType::IndexType pixelIndex1;
    pixelIndex1[0] = 128;
    pixelIndex1[1] = 128;
    pixelIndex1[2] = 35;
    
    InputImageType::PixelType pixelValue1 = input_image->GetPixel(pixelIndex1);
    std::cout<<"Input pixel value: "<<pixelValue1<<std::endl;
#endif
    
    //Add contribution of each subject to variance
    meanIt.GoToBegin();
    varIt.GoToBegin();
    inputIt.GoToBegin();
    while(!varIt.IsAtEnd())
      {
      if(inputIt.IsAtEnd() || meanIt.IsAtEnd())
	{
	std::cerr<<"input or mean iterator ends before var iterator"<<std::endl;
	exit(EXIT_FAILURE);
	}
      
      VarPixelType difference = (VarPixelType) inputIt.Get() - (VarPixelType) meanIt.Get();
      varIt.Set( varIt.Get()  + difference*difference / num_subjects);
      ++meanIt; ++varIt; ++inputIt;
      }
    
    //Diagnostics
#ifdef DEBUG
    varPixelValue = var_image->GetPixel(varPixelIndex);
    std::cout<<"Var pixel value: "<<varPixelValue<<std::endl;
#endif


    }

  //Compute fg image
  std::cout<<"==================Computing foreground image================="<<std::endl;
  yeoCopyImage<InputImageType>(mean_image, fg_image);
  fg_image->FillBuffer(0.0);
  
  for(int i = BASIC_PARMS+1; i <= BASIC_PARMS + num_subjects; i++) //for each subject
    {
    std::string filename;
    filename = SUBJECTS_DIR + argv[i] + mask_filename;
    std::cout<<"Reading: "<<filename<<std::endl;
    
    //Setting up ITK reader
    typedef itk::ImageFileReader< InputImageType > InputImageReaderType;
    InputImageReaderType::Pointer InputImageReader = InputImageReaderType::New();
    InputImageReader->SetFileName( filename);
    
    //Reading in file
    try
      {
      InputImageReader->Update();
      }
    catch(itk::ExceptionObject& err)
      {
      std::cout<<"Image Read Fails"<<std::endl;
      exit(EXIT_FAILURE);
      }
    
    //warping if exists
    if(warp_bool)
      {
      //normalizes mask so that fg is 1.0 and bg is 0.0
      InputImageType::Pointer mask_image2;
      mask_image2 = InputImageReader->GetOutput();
      InputPixelType maxValue = yeoMaxValue<InputImageType>(mask_image2);
      yeoMultiplyImageByConstantInPlace<InputImageType>(mask_image2, 1.0/maxValue);

      //warp image
      yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[i] + warp_filename);
      yeoWarpImageNewImage<InputImageType, DeformationFieldType>(mask_image2, df, mask_image);
      yeoReplaceValueInImageInPlace<InputImageType>(mask_image, itk::NumericTraits<InputImageType::PixelType>::max(), 0, 0); //for parts of mask that are outside, set it to 0.
      }
    else
      {
      //normalizes mask so that fg is 1.0 and bg is 0.0
      mask_image = InputImageReader->GetOutput();
      InputPixelType maxValue = yeoMaxValue<InputImageType>(mask_image);
      yeoMultiplyImageByConstantInPlace<InputImageType>(mask_image, 1.0/maxValue);
      }

    yeoLinearCombine2ImagesInPlace<InputImageType>(fg_image, mask_image, 1.0, 1.0/num_subjects); //note that 1 implies fg...
    }


  //Compute Max variance
  std::cout<<"=========================Computing Max Variance========================="<<std::endl;
  
  //Only use the part of variance image that is completely within brain
  
  //Create Intersection of fg.
  InputImageType::Pointer fg_image2 = InputImageType::New();
  yeoCopyImage<InputImageType>(fg_image, fg_image2);
  yeoReplaceValueInImageInPlace<InputImageType>(fg_image2, 1.0, 0, 1);

  //Mask out variance outside of fg
  VarImageType::Pointer var_image2 = VarImageType::New();
  yeoCopyImage<VarImageType>(var_image, var_image2);
  yeoPowerCombine2ImagesInPlace<VarImageType>(var_image2, fg_image2, 1, 1);
  
  //maximum variance
  float MAX_VAR = 0;
  varIt = VarIteratorType(var_image2, var_image2->GetLargestPossibleRegion());
  for(varIt.GoToBegin(); !varIt.IsAtEnd(); ++varIt)
    {
    if(varIt.Get() > MAX_VAR)
      {
      MAX_VAR = varIt.Get();
      }
    } 
  std::cout<<"Max Variance is: "<<MAX_VAR<<std::endl;
  
  //minimum variance
  float MIN_VAR = MIN_VAR_PERCENT * MAX_VAR;
  std::cout<<"Min Variance is: "<<MIN_VAR<<std::endl;

  //Set minimum variance
  std::cout<<"Regularize variance"<<std::endl;
  varIt = VarIteratorType(var_image, var_image->GetLargestPossibleRegion());

  int k = 0; int j = 0;
  for(varIt.GoToBegin(); !varIt.IsAtEnd(); ++varIt)
    {
    if(varIt.Get() < MIN_VAR)
      {
      varIt.Set(MIN_VAR);
      if(fg_image2->GetPixel(varIt.GetIndex()) == 0)
	k++;
      else
	j++;
      }
    } 
  std::cout<<"Regularize pixels: bg: "<<k<<", fg: "<<j<<std::endl;
  std::cout<<"Out of total fg pixels of: "<<yeoSumImage<InputImageType>(fg_image2)<<std::endl; 


  //Set Background Pixels to high variance
  std::cout<<"Set Background Pixels to high variance"<<std::endl;

  //first compute a background image
  InputImageType::Pointer bg_image = InputImageType::New();
  yeoCopyImage<InputImageType>(mean_image, bg_image);
  bg_image->FillBuffer(1.0);
  yeoLinearCombine2ImagesInPlace<InputImageType>(bg_image, fg_image, 1, -1); 

  //now regularize variance
  yeoPowerCombine2ImagesInPlace<InputImageType>( var_image, fg_image, 1, 1);
  yeoMultiplyImageByConstantInPlace<InputImageType>(bg_image, MAX_VAR);
  yeoLinearCombine2ImagesInPlace<InputImageType>(var_image, bg_image, 1, 1);

 
  //create mean filename
  std::string filename = mean_filename;
  std::cout<<"Writing Mean: "<<filename<<std::endl;

  typedef itk::ImageFileWriter<MeanImageType> WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetInput(mean_image);
  writer->SetFileName(filename);
  
  try
    {
    writer->Update();
    }
  catch(itk::ExceptionObject & err)
    {
    std::cout<<"Mean Write Unsuccessful"<<std::endl;
    return EXIT_FAILURE;
    }
  
  //create var filename
  filename = var_filename;
  std::cout<<"Writing Var: "<<filename<<std::endl;

  typedef itk::ImageFileWriter<VarImageType> VarWriterType;
  VarWriterType::Pointer var_writer = VarWriterType::New();
  var_writer->SetInput(var_image);
  var_writer->SetFileName(filename);
  
  try
    {
    var_writer->Update();
    }
  catch(itk::ExceptionObject & err)
    {
    std::cout<<"Var Write Unsuccessful"<<std::endl;
    return EXIT_FAILURE;
    }
     
   std::cout<<"Atlas Creation successful!"<<std::endl;	
   exit(EXIT_SUCCESS);
}

