
/*
* Written by: Thomas Yeo, MIT, CSAIL @ Asclepios, INRIA
*
* sent bug reports to ythomas@mit.edu
* 
* Example usage:
* CreateAtlasFromWarpedTensorImages SUBJECTS_DIR input_filename mask_filename mean_filename warp_filename var_filename ATLAS_SUBJECTS
* 
* CreateAtlasFromWarpedTensorImages 
* /auto/cardan/u/cardan/local/tmp-images/yboon/data/pierre_dti/  
* /tensors.inr.gz 
* /mask.hdr 
* /output/warp_field.nhdr 
* /auto/cardan/u/cardan/local/tmp-images/yboon/data/tensor_atlas/atlas_mean10.vtk
* /auto/cardan/u/cardan/local/tmp-images/yboon/data/tensor_atlas/atlas_var10.vtk 
* 0 
* 01 02 03 04 05
*
*/



#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"
#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkHistogramMatchingImageFilter.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "itkWarpImageFilter.h"
#include "itkTensor.h"
#include "itkTensorImageIO.h"

#include <getopt.h>
#include <iostream>
#include <string>
#include "Debug.h"
//#define DEBUG

#define BASIC_PARMS 7
#define MIN_VAR_PERCENT 1e-6
//#define MIN_VAR 0.01

int main( int argc, char *argv[] )
{   
  if(argc < 7)
    {
    std::cout<<"CreateAtlasFromWarpedTensorImages SUBJECTS_DIR input_filename mask_filename mean_filename warp_filename var_filename ATLAS_SUBJECTS"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Read in Arguments
  const std::string SUBJECTS_DIR = argv[1];
  const std::string input_filename = argv[2];
  const std::string mask_filename = argv[3];
  const std::string warp_filename = argv[4];
  const std::string mean_filename = argv[5];
  const std::string var_filename = argv[6];
  int hist_norm = atoi(argv[7]);
  int num_subjects = argc - 1 - BASIC_PARMS;

  std::cout<<"SUBJECTS_DIR: "<<SUBJECTS_DIR<<std::endl;
  std::cout<<"input_filename: "<<input_filename<<std::endl;
  std::cout<<"mask_filename: "<<mask_filename<<std::endl;
  std::cout<<"warp_filename: "<<warp_filename<<std::endl;
  std::cout<<"mean_filename: "<<mean_filename<<std::endl;
  std::cout<<"var_filename: "<<var_filename<<std::endl;
  std::cout<<"Num Subjects: "<<num_subjects<<std::endl;
  std::cout<<"Hist Norm: "<<hist_norm<<std::endl;

  //Check if there's input warp
  int warp_bool = 1;
  std::string warp_comparison;
  warp_comparison = "NULL";
  if(warp_comparison == warp_filename)
    warp_bool = 0;
  std::cout<<"Warp Bool: "<<warp_bool<<std::endl; 

  //Set Image Type for Mean
  const unsigned int Dimension = 3;
  typedef itk::Tensor<float, Dimension> TensorType;
  typedef itk::Image<TensorType, 3> ImageType;

  ImageType::Pointer mean_image = ImageType::New();

  //Set Image Pointer for Var
  ImageType::Pointer var_image = ImageType::New();

  //Set up Mask Pointer
  typedef float MaskPixelType;
  typedef itk::Image<MaskPixelType, Dimension> MaskImageType;
  MaskImageType::Pointer mask_image = MaskImageType::New();
  MaskImageType::Pointer fg_image = MaskImageType::New();

  //Set up Warp
  typedef itk::Vector< float, Dimension >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, Dimension > DeformationFieldType;
  DeformationFieldType::Pointer df = DeformationFieldType::New();

  //Create reference image in case of histogram normalization
  std::string ref_filename;
  ref_filename = SUBJECTS_DIR + argv[BASIC_PARMS+1] + input_filename;
  std::cout<<"Reading Reference: "<<ref_filename<<std::endl;

 //diagnostics
#ifdef DEBUG
  ImageType::IndexType pixelIndex;
  pixelIndex[0] = 64;
  pixelIndex[1] = 64;
  pixelIndex[2] = 15;    
  ImageType::PixelType pixelValue;
#endif

  ImageType::Pointer ref_image = ImageType::New();
  if(warp_bool)
    {
    ImageType::Pointer ref_image2 = ImageType::New();
    //read tensor image
    yeoReadTensorImage<ImageType>(ref_image2, ref_filename);
#ifdef DEBUG
    std::cout<<"Before log pixel value: "<<ref_image2->GetPixel(pixelIndex)<<std::endl;
#endif
    yeoLogTensorImageInPlace<ImageType>(ref_image2);
#ifdef DEBUG
    std::cout<<"After log pixel value: "<<ref_image2->GetPixel(pixelIndex)<<std::endl;
#endif
    //read deformation field
    yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[BASIC_PARMS+1] + warp_filename);
    //warp
    yeoWarpTensorImageNewImage<ImageType, DeformationFieldType>(ref_image2, df, ref_image);
    }
  else
    {
    //read tensor image
    yeoReadTensorImage<ImageType>(ref_image, ref_filename);
#ifdef DEBUG
    std::cout<<"Before log pixel value: "<<ref_image->GetPixel(pixelIndex)<<std::endl;
#endif
    yeoLogTensorImageInPlace<ImageType>(ref_image);
#ifdef DEBUG
    std::cout<<"After log pixel value: "<<ref_image->GetPixel(pixelIndex)<<std::endl;
#endif
    }


  //First Compute Mean Image
  std::cout<<"==================Computing Mean================="<<std::endl;
  for(int i = BASIC_PARMS+1; i <= BASIC_PARMS + num_subjects; i++) //for each subject
    {
    //create filename
    std::string filename;
    filename = SUBJECTS_DIR + argv[i] + input_filename;
    std::cout<<"Reading: "<<filename<<std::endl;

    ImageType::Pointer input_image = ImageType::New();
    //warping if exists
    if(warp_bool)
      {
      ImageType::Pointer input_image2;
      yeoReadTensorImage<ImageType>(input_image2, filename);
      yeoLogTensorImageInPlace<ImageType>(input_image2);

      yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[i] + warp_filename);
      yeoWarpTensorImageNewImage<ImageType, DeformationFieldType>(input_image2, df, input_image);
      }
    else
      {
      yeoReadTensorImage<ImageType>(input_image, filename);
#ifdef DEBUG
      std::cout<<"Num of Zero Tensors After Reading: "<<yeoNumZeroTensors<ImageType>(input_image)<<std::endl;
#endif
      yeoLogTensorImageInPlace<ImageType>(input_image);
#ifdef DEBUG
      std::cout<<"Num of Zero Tensors After Log: "<<yeoNumZeroTensors<ImageType>(input_image)<<std::endl;
#endif
      }

    //Histogram Normalized!    
    if(hist_norm)
      {
      std::cerr<<"Normalization not implemented now!"<<std::endl;
      exit(EXIT_FAILURE);
      }
    
    //Compute mean image
    if(i == BASIC_PARMS + 1)
      { 
#ifdef DEBUG
    std::cout<<"Input pixel value: "<<input_image->GetPixel(pixelIndex)<<std::endl;
#endif

      //Initialize mean image to 1/num_subjects of input_image 
      yeoCopyImage<ImageType>(input_image, mean_image);           
      yeoMultiplyImageByConstantInPlace<ImageType>(mean_image, 1.0/num_subjects);

      //Set up var Image
      yeoCreateZeroImage<ImageType, ImageType>(input_image, var_image);
      }
    else
      {
#ifdef DEBUG
    std::cout<<"Input pixel value: "<<input_image->GetPixel(pixelIndex)<<std::endl;
#endif

      yeoLinearCombine2ImagesInPlace<ImageType>(mean_image, input_image, 1, 1.0/num_subjects);
      }  
    
    //Diagnostics
#ifdef DEBUG
    std::cout<<"Mean pixel value: "<<mean_image->GetPixel(pixelIndex)<<std::endl;
    std::cout<<"Num of Zero Tensors in mean_image: "<<yeoNumZeroTensors<ImageType>(mean_image)<<std::endl;
#endif

    }
  
#ifdef DEBUG
    std::cout<<"Original var pixel value: "<<var_image->GetPixel(pixelIndex)<<std::endl;
#endif

  //Compute Var Image
  std::cout<<"==================Computing Var================="<<std::endl;
  for(int i = BASIC_PARMS+1; i <= BASIC_PARMS + num_subjects; i++) //for each subject
    {
    //create filename
    std::string filename;
    filename = SUBJECTS_DIR + argv[i] + input_filename;
    std::cout<<"Reading: "<<filename<<std::endl;
    
    ImageType::Pointer input_image = ImageType::New();
    //warping if exists
    if(warp_bool)
      {
      ImageType::Pointer input_image2;
      yeoReadTensorImage<ImageType>(input_image2, filename);
      yeoLogTensorImageInPlace<ImageType>(input_image2);
      
      yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[i] + warp_filename);
      yeoWarpTensorImageNewImage<ImageType, DeformationFieldType>(input_image2, df, input_image);
      }
    else
      {
      yeoReadTensorImage<ImageType>(input_image, filename);
      yeoLogTensorImageInPlace<ImageType>(input_image);
      }

#ifdef DEBUG
    std::cout<<"Input pixel value: "<<input_image->GetPixel(pixelIndex)<<std::endl;
#endif
    //Histogram Normalized!    
    if(hist_norm)
      {
      std::cerr<<"Normalization not implemented now!"<<std::endl;
      exit(EXIT_FAILURE);
      }
    
    yeoLinearCombine2ImagesInPlace<ImageType>(input_image, mean_image, 1.0, -1.0);
 
#ifdef DEBUG
    std::cout<<"Difference pixel value: "<<input_image->GetPixel(pixelIndex)<<std::endl;
#endif   

    ImageType::Pointer diffSqImage = ImageType::New();
    yeoPowerCombine2TensorImagesNewImage<ImageType>(input_image, input_image, diffSqImage, 1.0, 1.0);
    
#ifdef DEBUG
    std::cout<<"DiffSq pixel value: "<<diffSqImage->GetPixel(pixelIndex)<<std::endl;
#endif

    yeoLinearCombine2ImagesInPlace<ImageType>(var_image, diffSqImage, 1.0, 1.0/num_subjects);

#ifdef DEBUG
    std::cout<<"Var pixel value: "<<var_image->GetPixel(pixelIndex)<<std::endl;
#endif
    }
  
  //Compute fg image
  std::cout<<"==================Computing foreground image================="<<std::endl;
  yeoCreateZeroImage<ImageType, MaskImageType>(mean_image, fg_image);
  
  for(int i = BASIC_PARMS+1; i <= BASIC_PARMS + num_subjects; i++) //for each subject
    {
    std::string filename;
    filename = SUBJECTS_DIR + argv[i] + mask_filename;
    std::cout<<"Reading: "<<filename<<std::endl;
    
    //warping if exists
    if(warp_bool)
      {
      //normalizes mask so that fg is 1.0 and bg is 0.0
      MaskImageType::Pointer mask_image2 = MaskImageType::New();
      yeoReadImage<MaskImageType>(mask_image2, filename);
    
      MaskPixelType maxValue = yeoMaxValue<MaskImageType>(mask_image2);
      yeoMultiplyImageByConstantInPlace<MaskImageType>(mask_image2, 1.0/maxValue);
      
      //warp image
      yeoReadImage<DeformationFieldType>(df, SUBJECTS_DIR + argv[i] + warp_filename);
      yeoWarpImageNewImage<MaskImageType, DeformationFieldType>(mask_image2, df, mask_image);
      yeoReplaceValueInImageInPlace<MaskImageType>(mask_image, itk::NumericTraits<MaskImageType::PixelType>::max(), 0, 0); //for parts of mask that are outside, set it to 0.
      }
    else
      {
      //normalizes mask so that fg is 1.0 and bg is 0.0
      yeoReadImage<MaskImageType>(mask_image, filename);
      MaskPixelType maxValue = yeoMaxValue<MaskImageType>(mask_image);
      yeoMultiplyImageByConstantInPlace<MaskImageType>(mask_image, 1.0/maxValue);
      }
    
    yeoLinearCombine2ImagesInPlace<MaskImageType>(fg_image, mask_image, 1.0, 1.0/num_subjects); //note that 1 implies fg...
    }

  //Compute Max variance
  std::cout<<"=========================Computing Max Variance========================="<<std::endl;

  //Only use the part of variance image that is completely within brain
  
  //Create Intersection of fg.
  MaskImageType::Pointer fg_image2 = MaskImageType::New();
  yeoCopyImage<MaskImageType>(fg_image, fg_image2);
  yeoReplaceValueInImageInPlace<MaskImageType>(fg_image2, 1.0, 0, 1);

  //Mask out variance outside of fg
  ImageType::Pointer var_image2 = ImageType::New();
  yeoCopyImage<ImageType>(var_image, var_image2);
  yeoMultiplyTensorImageWithImageInPlace<ImageType, MaskImageType>(var_image2, fg_image2);
  
  TensorType MAX_VAR = yeoMaxTensorValue<ImageType>(var_image2);
  std::cout<<"Max_var: "<<MAX_VAR<<std::endl;

  //minimum variance
  TensorType MIN_VAR =  float(MIN_VAR_PERCENT) * MAX_VAR;
  std::cout<<"Min Variance is: "<<MIN_VAR<<std::endl;
  
  //Set minimum variance
  std::cout<<"Regularize variance"<<std::endl;
  yeoReplaceValueInTensorImageInPlace<ImageType>(var_image, MIN_VAR, MIN_VAR, 1);

  std::cout<<"Out of total fg pixels of: "<<yeoSumImage<MaskImageType>(fg_image2)<<std::endl; 

#ifdef DEBUG
  std::cout<<"Min Variance is: "<<yeoMinTensorValue<ImageType>(var_image)<<std::endl;
#endif

  //Set Background Pixels to high variance
  std::cout<<"Set Background Pixels to high variance"<<std::endl;

  //first compute a background image
  MaskImageType::Pointer bg_image = MaskImageType::New();
  yeoCreateZeroImage<MaskImageType, MaskImageType>(fg_image, bg_image);
  bg_image->FillBuffer(1.0);
  yeoLinearCombine2ImagesInPlace<MaskImageType>(bg_image, fg_image, 1, -1); 

  //now regularize variance
  yeoMultiplyTensorImageWithImageInPlace<ImageType, MaskImageType>(var_image, fg_image);
  
  ImageType::Pointer bg_variance = ImageType::New();
  yeoCreateZeroImage<ImageType, ImageType>(var_image, bg_variance);
  bg_variance->FillBuffer(MAX_VAR);
  yeoMultiplyTensorImageWithImageInPlace<ImageType, MaskImageType>(bg_variance, bg_image);
  
  yeoLinearCombine2ImagesInPlace<ImageType>(var_image, bg_variance, 1, 1);

  //Exponentiate the mean_image
  
  std::cout<<"Num of Zero Tensors Before Exp Mean: "<<yeoNumZeroTensors<ImageType>(mean_image)<<std::endl;
  yeoExpTensorImageInPlace<ImageType>(mean_image);
  std::cout<<"Num of Zero Tensors After Exp Mean: "<<yeoNumZeroTensors<ImageType>(mean_image)<<std::endl;
  int count = yeoCleanTensorImageInPlace<ImageType>(mean_image);
  std::cout<<"Num of Tensors cleaned from mean_image: "<<count<<std::endl;
  std::cout<<"Num of Zero Tensors in mean_image: "<<yeoNumZeroTensors<ImageType>(mean_image)<<std::endl;


  //Writing Mean Image
  std::string filename = mean_filename;
  std::cout<<"Writing Mean: "<<filename<<std::endl;

  yeoWriteTensorImage<ImageType>(mean_image, filename);

  //Write var image
  filename = var_filename;
  std::cout<<"Writing Var: "<<filename<<std::endl;

  yeoWriteTensorImage<ImageType>(var_image, filename);

  
  std::cout<<"Tensor Atlas Creation successful!"<<std::endl;	
  exit(EXIT_SUCCESS);
}

