
/*
* Written by: Thomas Yeo
* 
* Example usage:
*  CreateRandomWarps input_file output_file mag
* 
*/


#include <getopt.h>
#include <iostream>
#include <string>
#include <time.h>
#include <stdlib.h>
#include <math.h>

#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkImage.h"
#include "Debug.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include <itkRecursiveGaussianImageFilter.h>
#include "itkGaussianDistribution.h"
#include "itkExponentialDeformationFieldImageFilter.h"
/**
   Since ITK 3.8, this filter has been added in global ITK code,
   and it has changed of name,
   please switch the lines if any problems.
*/
// #include "itkWarpJacobianDeterminantFilter.h"
#include "itkDisplacementFieldJacobianDeterminantFilterOld.h"
#include "itkWarpSmoothnessCalculator.h"

#define ImageDimension 3

int main( int argc, char *argv[] )
{   
  
  if(argc < 8)
    {
    std::cout<<"This function reads in a tensor image, creates a random warp and applies the warp to tensor image and add noise"<<std::endl;
    std::cout<<"Usage: CreateRandomWarps input_file output_file output_warp mag noise vec_smooth PPD_bool"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Read in Arguments
  const std::string input_filename = argv[1];
  const std::string noisy_output_filename = argv[2];
  const std::string output_warp = argv[3];

  const float mag = atof(argv[4]);
  const float noise = atof(argv[5]);
  const float vec_smooth = atof(argv[6]);
  const float PPD_bool = atof(argv[7]);

  //Some Typedefs
  typedef float PixelType;
  typedef itk::Vector< PixelType, ImageDimension >    VectorPixelType;
  typedef itk::Image<  VectorPixelType, ImageDimension > DeformationFieldType;
  
  typedef itk::Tensor<float, 3> TensorType;
  typedef itk::Image<TensorType, ImageDimension> TensorImageType;

  //Read in Tensor Image
  std::cout<<"Read Tensor Image: "<<input_filename<<std::endl;
  TensorImageType::Pointer input_image = TensorImageType::New();
  yeoReadTensorImage<TensorImageType>(input_image, input_filename);

  //Create Empty Deformation field
  std::cout<<"Create Empty Deformation Field"<<std::endl;
  DeformationFieldType::Pointer df = DeformationFieldType::New();
  yeoCreateZeroImage<TensorImageType, DeformationFieldType>(input_image, df);
  
  //Initialize Gaussian Distribution Generation
  srand( time(NULL));
  typedef itk::Statistics::GaussianDistribution DistributionType;
  DistributionType::Pointer distributionFunction = DistributionType::New();
  distributionFunction->SetMean( 0.0 );
  distributionFunction->SetVariance( mag );

  //Now Create Random Random Warp Field
  std::cout<<"Create Random Velocity Field"<<std::endl;
  typedef itk::ImageRegionIterator<DeformationFieldType> IteratorType;
  IteratorType It = IteratorType(df, df->GetLargestPossibleRegion());
  
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    VectorPixelType def = It.Get();
    for(unsigned int i = 0; i < VectorPixelType::Dimension; i++)
      {
      def[i] = (float) distributionFunction->EvaluateInverseCDF(rand()/((double)RAND_MAX + 1));
      }
    It.Set(def);
    }   

  //Mask the velocity field, so it's zero in background
  std::cout<<"Masking Random Velocity Field"<<std::endl;
  typedef itk::Image<PixelType, ImageDimension> MaskImageType;
  MaskImageType::Pointer input_fg = MaskImageType::New();
  yeoGetForegroundMask<TensorImageType, MaskImageType>(input_image, input_fg, itk::NumericTraits<TensorImageType::PixelType>::Zero, 0, 1);
  yeoMultiplyTensorImageWithImageInPlace<DeformationFieldType, MaskImageType>(df, input_fg);

  //Smooth the warp field
  typedef itk::RecursiveGaussianImageFilter<DeformationFieldType, DeformationFieldType> SmoothDFType;
  SmoothDFType::Pointer filterx = SmoothDFType::New();
  SmoothDFType::Pointer filtery = SmoothDFType::New();
  SmoothDFType::Pointer filterz = SmoothDFType::New();

  filterx->SetInput( df );
  filterx->SetDirection( 0 ); 
  filterx->SetOrder(SmoothDFType::ZeroOrder );
  filterx->SetSigma(vec_smooth);

  filtery->SetInput( filterx->GetOutput());
  filtery->SetDirection( 1 ); 
  filtery->SetOrder(SmoothDFType::ZeroOrder );
  filtery->SetSigma(vec_smooth);

  filterz->SetInput( filtery->GetOutput());
  filterz->SetDirection( 2 ); 
  filterz->SetOrder(SmoothDFType::ZeroOrder );
  filterz->SetSigma(vec_smooth);

  try
    {
    filterz->Update();
    std::cout<<"Smoothing velocity field"<<std::endl;
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Unexpected error in smoothing velocity field" << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }

  
  //Exponentiate velocity field
  typedef itk::ExponentialDeformationFieldImageFilter< DeformationFieldType, DeformationFieldType >  FieldExponentiatorType;
  FieldExponentiatorType::Pointer Exponentiator = FieldExponentiatorType::New();
  Exponentiator->SetInput( filterz->GetOutput() );
  Exponentiator->AutomaticNumberOfIterationsOn();
  Exponentiator->SetMaximumNumberOfIterations( 2000u );

  try
    {
    Exponentiator->Update();
    std::cout<<"Exponentiating Velocity Field"<<std::endl;
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Unexpected error in exponentiating velocity field" << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }
  
  //Check Jacobian
  std::cout<<"Checking Jacobian"<<std::endl;
//   typedef itk::WarpJacobianDeterminantFilter<DeformationFieldType, MaskImageType> JacobianFilterType;
  typedef itk::DisplacementFieldJacobianDeterminantFilterOld<DeformationFieldType, MaskImageType> JacobianFilterType;
  JacobianFilterType::Pointer JacobianFilter = JacobianFilterType::New();
  JacobianFilter->SetUseImageSpacing( true );

  JacobianFilter->SetInput( Exponentiator->GetOutput());
  JacobianFilter->UpdateLargestPossibleRegion();
  
  const unsigned int numPix = JacobianFilter->GetOutput()->GetLargestPossibleRegion().GetNumberOfPixels();
  
  PixelType* pix_start = JacobianFilter->GetOutput()->GetBufferPointer();
  PixelType* pix_end = pix_start + numPix;
  
  PixelType* jac_ptr;
  
  // Get percentage of det(Jac) below 0
  unsigned int jacBelowZero(0u);
  for (jac_ptr=pix_start; jac_ptr!=pix_end; ++jac_ptr)
    {
    if ( *jac_ptr<=0.0 ) ++jacBelowZero;
    }
  const double jacBelowZeroPrc = static_cast<double>(jacBelowZero) / static_cast<double>(numPix);
 
  //Get min max jacobian
  const double minJac = *(std::min_element (pix_start, pix_end));
  const double maxJac = *(std::max_element (pix_start, pix_end));
  std::cout<<"max|Jac| "<<maxJac<<" - "
	   <<"min|Jac| "<<minJac<<" - "
	   <<"ratio(|Jac|<=0) "<<jacBelowZeroPrc<<std::endl;
  
  //Compute Harmonic Energy
  typedef itk::WarpSmoothnessCalculator<DeformationFieldType> SmoothnessCalculatorType;
  SmoothnessCalculatorType::Pointer SmoothnessCalculator;
  SmoothnessCalculator = SmoothnessCalculatorType::New();
  SmoothnessCalculator->SetImage( Exponentiator->GetOutput());
  SmoothnessCalculator->Compute();
  const double harmonicEnergy = SmoothnessCalculator->GetSmoothness();
  std::cout<<"Harmonic energy: "<<harmonicEnergy<<std::endl;

  //Compute Norm of Displacement field
  MaskImageType::Pointer df_norm_image = MaskImageType::New();
  yeoInnerProductVectorFieldNewImage<DeformationFieldType, MaskImageType>(Exponentiator->GetOutput(), Exponentiator->GetOutput(), df_norm_image);
  yeoPowerCombine2ImagesInPlace<MaskImageType>(df_norm_image, input_fg, 0.5, 1.0);
  
  double numPixels = yeoSumImage<MaskImageType>(input_fg);
  double max_warp = yeoMaxValue<MaskImageType>(df_norm_image);
  double mean_warp = yeoSumImage<MaskImageType>(df_norm_image)/numPixels;

  yeoPowerCombine2ImagesInPlace<MaskImageType>(df_norm_image, df_norm_image, 1.0, 1.0);
  double totalSSD = yeoSumImage<MaskImageType>(df_norm_image);
  double averageSSD = totalSSD/numPixels;

  std::cout<<"Current SSD: "<<averageSSD<<", mean warp: "<<mean_warp<<", max warp: "<<max_warp<<std::endl;

  //Write Out Grid Image
  typedef itk::Image< unsigned char, ImageDimension > GridImageType;
  GridImageType::Pointer gridImage = GridImageType::New();
  gridImage->SetRegions( input_image->GetRequestedRegion() );
  gridImage->SetOrigin( input_image->GetOrigin() );
  gridImage->SetSpacing( input_image->GetSpacing() );
  gridImage->Allocate();
  gridImage->FillBuffer(0);
      
  typedef itk::ImageRegionIteratorWithIndex<GridImageType> GridImageIteratorWithIndex;
  GridImageIteratorWithIndex itergrid = GridImageIteratorWithIndex(gridImage, gridImage->GetRequestedRegion() );

  std::cout<<"Create Grid Image"<<std::endl;
  const int gridspacing(4);
  for (itergrid.GoToBegin(); !itergrid.IsAtEnd(); ++itergrid)
    {
    itk::Index<ImageDimension> index = itergrid.GetIndex();
    
    // produce an xy grid for all z
    if ( (index[0]%gridspacing)==0 ||
	 (index[1]%gridspacing)==0 )
      {
      itergrid.Set( itk::NumericTraits<unsigned char>::max() );
      }
    }

  std::cout<<"Warp Grid Image"<<std::endl;
  typedef itk::WarpImageFilter< GridImageType, GridImageType, DeformationFieldType >  GridWarperType;
  GridWarperType::Pointer gridwarper = GridWarperType::New();
  gridwarper->SetInput( gridImage );
  gridwarper->SetOutputSpacing( input_image->GetSpacing() );
  gridwarper->SetOutputOrigin( input_image->GetOrigin() );
  gridwarper->SetDeformationField( Exponentiator->GetOutput() );
      
  // Write warped grid to file
  typedef itk::ImageFileWriter< GridImageType >  GridWriterType;
  
  GridWriterType::Pointer      gridwriter =  GridWriterType::New();
  gridwriter->SetFileName( "WarpedGridImage.mha" );
  gridwriter->SetInput( gridwarper->GetOutput()   );
  gridwriter->SetUseCompression( true );
  
  try
    {
    gridwriter->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Unexpected error." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }

  //Warp Input Image and Write it out
  std::cout<<"Warping Input Image ... ";
  TensorImageType::Pointer out_image = TensorImageType::New();
  yeoLogTensorImageInPlace<TensorImageType>(input_image);

  if(PPD_bool == 0)
    {
    std::cout<<"using FS"<<std::endl;
    yeoWarpTensorImageNewImage<TensorImageType, DeformationFieldType>(input_image, Exponentiator->GetOutput(), out_image);
    }
  else 
    {
    std::cout<<"using PPD"<<std::endl;
    yeoWarpTensorImageNewImagePPD<TensorImageType, DeformationFieldType>(input_image, Exponentiator->GetOutput(), out_image);
    }

  //Add Noise to output image
  std::cout<<"Adding Noise"<<std::endl;
  typedef itk::ImageRegionIterator< TensorImageType> TensorIteratorType;
  TensorIteratorType It2 = TensorIteratorType(out_image, out_image->GetLargestPossibleRegion());
    
  DistributionType::Pointer noiseFunction = DistributionType::New();
  noiseFunction->SetMean( 0.0 );
  noiseFunction->SetVariance(noise);

  
  for(It2.GoToBegin(); !It2.IsAtEnd(); ++It2)
    {
    
    TensorImageType::PixelType tensor = It2.Get();
    if(tensor != itk::NumericTraits<TensorImageType::PixelType>::Zero) //Not background
      {
      for(int j = 0; j < 6; j++)
	{
	tensor.SetNthComponent(j, tensor.GetNthComponent(j) + (float)noiseFunction->EvaluateInverseCDF(rand()/((double)RAND_MAX + 1)));
	}
      It2.Set(tensor);
      }
    }

  std::cout<<"Writing Output Image: "<<noisy_output_filename<<std::endl;
  yeoExpTensorImageInPlace<TensorImageType>(out_image);
  yeoWriteTensorImage<TensorImageType>(out_image, noisy_output_filename);
  

  //Write Deformation field
  std::cout<<"Writing Output Deformation Field: "<<output_warp<<std::endl;
  yeoWriteImage<DeformationFieldType>(Exponentiator->GetOutput(), output_warp);

  exit(EXIT_SUCCESS);
}

