#ifndef _UsefulFunctions_h_
#define _UsefulFunctions_h_

#include "itkImage.h"
#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkSmoothingRecursiveGaussianImageFilter.h"
#include "itkGradientImageFilter.h"
#include "itkCovariantVector.h"
#include "itkWarpImageFilter.h"
#include "itkTensor.h"
#include "itkNumericTraitsTensorPixel2.h"


#include <getopt.h>
#include <math.h>


/* Written by Thomas Yeo, MIT, CSAIL @ Asclepios, INRIA 
 *
 * Email ythomas@mit.edu for bugs report
 */

/* Summary of functions: 
 
+ yeoCopyImage<ImageType>(fixedImage, copiedImage); -- TensorCertified
+ yeoInnerProductVectorFieldNewImage<DeformationFieldType, NormImageType>(im1, im2, out_image); 
+ yeoCreateZeroImage<ImageType1, ImageType2>(fixedImage, copiedImage); -- Tensor Certified
+ total = yeoSumImage<ImageType>(fixedImage)
+ yeoMultiplyImageByConstantNewImage<ImageType>(in, out, constant); -- Tensor Certified
+ yeoMultiplyImageByConstantInPlace<ImageType>(in, constant); -- Tensor Certified
+ yeoMultiplyImageByConstantInPlace<ImageType>(in, constant); -- Tensor Certified
+ yeoLinearCombine2ImagesNewImage<ImageType>(in1, in2, out, a, b); -- Tensor Certified
+ yeoLinearCombine2ImagesInPlace<ImageType>(in1, in2, a, b); -- Tensor Certified
+ yeoPowerCombine2ImagesNewImage<ImageType>(in1, in2, out, a, b);
+ yeoPowerCombine2ImagesInPlace<ImageType>(in1, in2, a, b);
+ yeoAbsImageNewImage<ImageType>(in, absIm);
+ yeoAbsImageInPlace<ImageType>(im);
+ yeoWriteImage<ImageType>(im, filename);
+ yeoWriteImageCompress<ImageType>(im, filename);
+ yeoReadImage<ImageType>(im, filename);
+ yeoSmoothImageNewImage<ImageType>(im, out, sigma);
+ yeoGradientMagnitudeNewImage<ImageType>(im, out);
+ yeoWarpImageNewImage<ImageType, DeformationFieldType>(im, df, out);
+ yeoReplaceValueInImageInPlace<ImageType>(im, oldValue, newValue, optype);
+ maxValue = yeoMaxValue<ImageType>(im);
*/



/* Usage:  typename ImageType::Pointer copiedImage = ImageType::New();
 *         yeoCopyImage<ImageType>(fixedImage, copiedImage);
*/
template <class TImageType> void yeoCopyImage(const TImageType* im, TImageType* out_image)
{
  if(out_image == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  out_image->CopyInformation(im);
  out_image->SetRegions(im->GetLargestPossibleRegion());
  out_image->SetSpacing(im->GetSpacing());
  out_image->SetOrigin(im->GetOrigin());
  out_image->Allocate();

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(out_image, out_image->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());

  for(outIt.GoToBegin(), It.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
    {
    outIt.Set(It.Get());
    }   
}


/* Usage:
*         yeoInnerProductVectorFieldNewImage<DeformationFieldType, NormImageType>(im1, im2, out_image); 
*/
template <class DeformationFieldType, class NormImageType> void yeoInnerProductVectorFieldNewImage
                (const DeformationFieldType *im1, const DeformationFieldType *im2, NormImageType *out_image)
{
  if(out_image == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im1 == 0)
    {
    std::cout<<"Input Image 1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im2 == 0)
    {
    std::cout<<"Input Image 2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  out_image->SetRegions(im1->GetLargestPossibleRegion());
  out_image->SetSpacing(im1->GetSpacing());
  out_image->SetOrigin(im1->GetOrigin());
  out_image->Allocate();
  
  typedef typename itk::ImageRegionConstIterator<DeformationFieldType> ConstIteratorType;
  ConstIteratorType It1 = ConstIteratorType(im1, im1->GetLargestPossibleRegion());
  ConstIteratorType It2 = ConstIteratorType(im2, im2->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionIterator<NormImageType> IteratorType;
  IteratorType outIt = IteratorType(out_image, out_image->GetLargestPossibleRegion());
  
  for(It1.GoToBegin(), It2.GoToBegin(), outIt.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2, ++outIt)
    {
    typename DeformationFieldType::PixelType p1 = It1.Get();
    typename DeformationFieldType::PixelType p2 = It2.Get();
    
    typename NormImageType::PixelType total = 0;

    for(unsigned int i = 0; i < DeformationFieldType::PixelType::Dimension; i++)
      {
      total += p1[i]*p2[i];
      }
    outIt.Set(total);
    }  
}



/* Usage:  typename ImageType::Pointer copiedImage = ImageType::New();
 *         yeoCreateZeroImage<ImageType1, ImageType2>(fixedImage, copiedImage);
*/
template <class TImageType1, class TImageType2> void yeoCreateZeroImage(const TImageType1* im, TImageType2* out_image)
{
  if(out_image == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  out_image->SetRegions(im->GetLargestPossibleRegion());
  out_image->SetSpacing(im->GetSpacing());
  out_image->SetOrigin(im->GetOrigin());
  out_image->Allocate();
  
  out_image->FillBuffer(itk::NumericTraits<typename TImageType2::PixelType>::Zero);
}



/* Usage: total = yeoSumImage<ImageType>(fixedImage)
*/
template <class TImageType> double yeoSumImage(const TImageType* im)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());

  double total = 0;
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    total += It.Get();
    }
  return total;

}

/* Usage:  typename ImageType::Pointer out = ImageType::New();
 *         yeoMultiplyImageByConstantNewImage<ImageType>(in, out, constant);
*/
template <class TImageType> void yeoMultiplyImageByConstantNewImage(const TImageType* in, TImageType *out, float c)
{
  //First Copy image
  
  if(in == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cout<<"Out Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  yeoCopyImage<TImageType>(in, out);

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(out, out->GetLargestPossibleRegion());

  for(outIt.GoToBegin(); !outIt.IsAtEnd(); ++outIt)
    outIt.Set(outIt.Get()*c);

}

/* Usage:  
 *         yeoMultiplyImageByConstantInPlace<ImageType>(in, constant);
*/
template <class TImageType> void yeoMultiplyImageByConstantInPlace(TImageType* in, float c)
{
  if(in == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(in, in->GetLargestPossibleRegion());

  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    It.Set(c * It.Get());

}


/* Usage:  typename ImageType::Pointer out = ImageType::New();
 *         yeoLinearCombine2ImagesNewImage<ImageType>(in1, in2, out, a, b);
*/
template <class TImageType> void yeoLinearCombine2ImagesNewImage(const TImageType* in1, const TImageType* in2, TImageType *out, float a, float b)
{
  if(in1 == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  if(in2 == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

    if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }


  //First Copy image
  yeoCopyImage<TImageType>(in1, out);

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(out, out->GetLargestPossibleRegion());

  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(in2, in2->GetLargestPossibleRegion());


  for(outIt.GoToBegin(), It2.GoToBegin(); !outIt.IsAtEnd(); ++outIt, ++It2)
    {
    outIt.Set(outIt.Get()*a + It2.Get()*b );
    }

}

/* Usage: 
 *         yeoLinearCombine2ImagesInPlace<ImageType>(in1, in2, a, b);
*/
template <class TImageType> void yeoLinearCombine2ImagesInPlace(TImageType* in1, const TImageType* in2, float a, float b)
{
  if(in1 == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(in2 == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It1 = IteratorType(in1, in1->GetLargestPossibleRegion());
  
  
  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(in2, in2->GetLargestPossibleRegion());
  

  for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
    It1.Set(It1.Get()*a + It2.Get()*b );

}



/* Usage:  typename ImageType::Pointer out = ImageType::New();
 *         yeoPowerCombine2ImagesNewImage<ImageType>(in1, in2, out, a, b);
*/
template <class TImageType> void yeoPowerCombine2ImagesNewImage(const TImageType* in1, const TImageType* in2, TImageType *out, float a, float b)
{

  if(in1 == 0)
    {
    std::cout<<"Input Image 1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(in2 == 0)
    {
    std::cout<<"Input Image 2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cerr<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //First Copy image
  yeoCopyImage(in1, out);

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(out, out->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(in2, in2->GetLargestPossibleRegion());
   
  for(outIt.GoToBegin(), It2.GoToBegin(); !outIt.IsAtEnd(); ++outIt, ++It2)
    outIt.Set(pow(outIt.Get(), a) * pow(It2.Get(), b) );

}

/* Usage: 
 *         yeoPowerCombine2ImagesInPlace<ImageType>(in1, in2, a, b);
 *         
 * InPlace for image 1
*/
template <class TImageType> void yeoPowerCombine2ImagesInPlace(TImageType* in1, const TImageType* in2, float a, float b)
{
  if(in1 == 0)
    {
    std::cout<<"Input Image 1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(in2 == 0)
    {
    std::cout<<"Input Image 2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It1 = IteratorType(in1, in1->GetLargestPossibleRegion());
  
  
  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(in2, in2->GetLargestPossibleRegion());
  

  for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
    It1.Set(pow((float)It1.Get(), a) * pow( (float)It2.Get(), b) );

}



/* Usage:  typename ImageType::Pointer absIm = ImageType::New();
 *         yeoAbsImageNewImage<ImageType>(in, absIm);
*/
template <class TImageType> void yeoAbsImageNewImage(const TImageType* im, TImageType* out_image)
{

 if(im == 0)
   {
   std::cout<<"Input Image 1 is null"<<std::endl;
   exit(EXIT_FAILURE);
   }
 
 if(out_image == 0)
   {
   std::cout<<"Out image is null"<<std::endl;
   exit(EXIT_FAILURE);
   }
 
  out_image->SetRegions(im->GetLargestPossibleRegion());
  out_image->SetSpacing(im->GetSpacing());
  out_image->SetOrigin(im->GetOrigin());
  out_image->Allocate();
  
  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(out_image, out_image->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());

  for(outIt.GoToBegin(), It.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
    {
    outIt.Set(fabs((float) It.Get()));
    }
}

/* Usage:  
 *         yeoAbsImageInPlace<ImageType>(im);
*/
template <class TImageType> void yeoAbsImageInPlace(TImageType* im)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());
  
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    It.Set(fabs((float) It.Get()));
    }
}

/* Usage:
*           yeoWriteImage<ImageType>(im, filename);
*/
template <class TImageType> void yeoWriteImage(const TImageType *im, std::string filename)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef itk::ImageFileWriter< TImageType >  WriterType;
   
  typename WriterType::Pointer      writer =  WriterType::New();
  writer->SetFileName( filename);
  writer->SetInput( im );
  try
    {
    writer->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Unexpected error in writing file." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }
}


/* Usage:
*           yeoWriteImageCompress<ImageType>(im, filename);
*/
template <class TImageType> void yeoWriteImageCompress(const TImageType *im, std::string filename)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef itk::ImageFileWriter< TImageType >  WriterType;
  
  typename WriterType::Pointer      writer =  WriterType::New();
  writer->SetFileName( filename);
  writer->SetInput( im );
  writer->SetUseCompression(1);
  try
    {
    writer->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Unexpected error in writing file." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }
}

/* Usage:   typename ImageType::Pointer im = ImageType::New(); 
*           yeoReadImage<ImageType>(im, filename);
*/
template <class TImageType> void yeoReadImage(TImageType *im, std::string filename)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef itk::ImageFileReader< TImageType >  ReaderType;
  
  typename ReaderType::Pointer      reader =  ReaderType::New();
  reader->SetFileName( filename);
  
  try
    {
    reader->Update();
     }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Unexpected error in reading file." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }
  yeoCopyImage<TImageType>(reader->GetOutput(), im);
}


/* Usage: typename ImageType::Pointer out = ImageType::New();
*         yeoSmoothImageNewImage<ImageType>(im, out, sigma);
*/
template <class TImageType> void yeoSmoothImageNewImage(const TImageType *im, TImageType *out, float sigma)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef itk::SmoothingRecursiveGaussianImageFilter<TImageType, TImageType> SmoothFilterType;
  typename SmoothFilterType::Pointer SmoothFilter = SmoothFilterType::New();

  SmoothFilter->SetInput(im);
  SmoothFilter->SetSigma(sigma);
  
   try
     {
     SmoothFilter->Update();
     }
   catch( itk::ExceptionObject& err )
     {
     std::cout << "Unexpected error in Smoothing Image." << std::endl;
     std::cout << err << std::endl;
     exit( EXIT_FAILURE );
   }
   
   yeoCopyImage<TImageType>(SmoothFilter->GetOutput(), out);   
}

/* Usage: typename ImageType::Pointer out = ImageType::New();
*         yeoGradientMagnitudeNewImage<ImageType>(im, out);
*/
template <class TImageType, unsigned int Dimension> void yeoGradientMagnitudeNewImage(const TImageType *im, TImageType *out)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef itk::GradientImageFilter<TImageType, float, double> GradientFilterType;
  typename GradientFilterType::Pointer GradientFilter = GradientFilterType::New();

  GradientFilter->SetUseImageSpacingOn();
  GradientFilter->SetInput(im);
  
  try
     {
     GradientFilter->Update();
     }
   catch( itk::ExceptionObject& err )
     {
     std::cout << "Unexpected error in Computing Gradient Image." << std::endl;
     std::cout << err << std::endl;
     exit( EXIT_FAILURE );
   }

  //Grab gradImage 
  typedef itk::CovariantVector<double, Dimension> CovariantVectorType;
  typedef itk::Image<CovariantVectorType, Dimension> gradImageType;
  typename gradImageType::Pointer gradImage = GradientFilter->GetOutput();
  
  //Create Output Image
  yeoCopyImage<TImageType>(im, out);

  //Compute OutputImage
  typedef typename itk::ImageRegionConstIterator< gradImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(gradImage, gradImage->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(out, out->GetLargestPossibleRegion());

  for(It.GoToBegin(), outIt.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
     outIt.Set(It.Get().GetNorm());
}


/* Usage:  typename ImageType::Pointer out = ImageType::New();
*         yeoWarpImageNewImage<ImageType, DeformationFieldType>(im, df, out);
*
*  pad edge with zero pixel values
*/
template <class TImageType, class DeformationFieldType> void yeoWarpImageNewImage(const TImageType *im, const DeformationFieldType *df, TImageType *out)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(df == 0)
    {
    std::cout<<"Deformation Field is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef itk::WarpImageFilter < TImageType, TImageType, DeformationFieldType >  WarperType;
  typename WarperType::Pointer warper = WarperType::New();
  
  warper->SetInput(im);
  warper->SetOutputSpacing(im->GetSpacing());
  warper->SetOutputOrigin(im->GetOrigin());
  warper->SetDeformationField(df);
  warper->SetEdgePaddingValue( itk::NumericTraits<typename TImageType::PixelType>::ZeroValue() );
  
  try
    {
    warper->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Warper fails." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }
  yeoCopyImage<TImageType>(warper->GetOutput(), out);   
  
}

/* Usage:  
 *         yeoReplaceValueInImageInPlace<ImageType>(im, oldValue, newValue, optype);
*
* optype: 0 = "=="
*         1 = "<"
*         2 = ">"
*         3 = "<="
*         4 = ">="
*
*/
template <class TImageType> int yeoReplaceValueInImageInPlace(TImageType* im, typename TImageType::PixelType oldValue, typename TImageType::PixelType newValue, int optype)
{
   if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());

  int count = 0;

  switch(optype)
    {
    case 0 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.Get() == oldValue)
	{
	It.Set(newValue);
	count++;
	}
      }
    break;
    }
    case 1 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.Get() < oldValue)
	{
	It.Set(newValue);
	count++;
	}
      }
    break;
    }
    case 2 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.Get() > oldValue)
	{
	It.Set(newValue);
	count++;
	}
      }
    break;
    }
    case 3 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.Get() <= oldValue)
	{
	It.Set(newValue);
	count++;
	}
      }
    break;
    }
    case 4 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.Get() >= oldValue)
	{
	It.Set(newValue);
	count++;
	}
      }
    break;
    }
    default :
    {
    std::cerr<<"Opcode does not exist"<<std::endl;
    exit(EXIT_FAILURE);
    }
    } //end switch
  return count;
}

/* Usage: 
 *         yeoMaxValue<ImageType>(im);
*/
template <class TImageType> typename TImageType::PixelType yeoMaxValue(const TImageType* im)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef typename itk::ImageRegionConstIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());
  
  typedef typename TImageType::PixelType PixelType;
  PixelType maxValue =  itk::NumericTraits<PixelType>::min();

  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.Get() > maxValue)
      maxValue = It.Get();
    }
  return maxValue;
}

 

#endif
