#ifndef _UsefulTensorFunctions_h_
#define _UsefulTensorFunctions_h_

#include "itkTensor.h"
#include "itkTensorImageIO.h"
#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkWarpTensorImageFilter.h"
#include "itkWarpTensorImageWOReorientationFilter.h"
#include "itkNumericTraitsTensorPixel2.h"
#include "UsefulFunctions.h"
#include "itkMatrix.h"


#include "itkTensorToScalarTensorImageFilter.h"
#include "itkTensorToScalarFunction.h"
#include "itkTensorToFAFunction.h"
#include "itkTensorToLogFAFunction.h"
#include "itkTensorToADCFunction.h"
#include "itkTensorToClFunction.h"
#include "itkTensorToCpFunction.h"
#include "itkTensorToCsFunction.h"
#include "itkTensorToRAFunction.h"
#include "itkTensorToVRFunction.h"
#include "itkTensorToVolumeFunction.h"
#include "itkTensorToLambdaFunction.h"
#include "itkTensorToDispersionFunction.h"

#include <getopt.h>
#include <math.h>

/* Written by Thomas Yeo, MIT, CSAIL @ Asclepios, INRIA 
 *
 * Email ythomas@mit.edu for bugs report
 */

/* Summary of functions:
+ count = yeoCleanTensorImageInPlace<ImageType>(im); 
+ count = yeoNumZeroTensors<ImageType>(im);
+ yeoWriteTensorImage<ImageType>(im, filename);
+ yeoReadTensorImage<ImageType>(im, filename);
+ yeoLogTensorImageInPlace<ImageType>(im);
+ yeoLogTensorImageNewImage<ImageType>(im, outIm);
+ yeoExpTensorImageInPlace<ImageType>(im);
+ yeoExpTensorImageNewImage<ImageType>(im, outIm);
+ yeoSqNormTensorImageNewImage<TensorImageType, NormImageType>(im, out_image); 
+ yeoNormTensorImageNewImage<TensorImageType, NormImageType>(im, out_image); --works for vector image as well.
+ yeoMultiplyTensorImageWithImageInPlace<TImageType, ImageType>(im1, im2);
+ yeoMultiplyTensorImageWithImageNewImage<TImageType, ImageType>(im1, im2, outIm);
+ yeoWarpTensorImageNewImage<ImageType, DeformationFieldType>(im, df, out);
+ yeoWarpTensorImageNewImagePPD<ImageType, DeformationFieldType>(im, df, out);
+ yeoWarpTensorImageWOReorientationNewImage<ImageType, DeformationFieldType>(im, df, out);
+ newTensor = yeoPowerCombine2Tensors<TensorType>(tensor1, tensor2, a, b);
+ total = yeoSumTensor<TensorType>(tensor);
+ tensor = yeoAbsTensor<TensorType>(tensor);
+ yeoPowerCombine2TensorImagesNewImage<ImageType>(in1, in2, out, a, b);
+ yeoPowerCombine2TensorImagesInPlace<ImageType>(in1, in2, a, b);
+ yeoReplaceValueInTensorImageInPlace<ImageType>(im, oldValue, newValue, optype);
+ yeoMaxTensorValue<ImageType>(im); 
+ yeoMinTensorValue<ImageType>(im);
+ yeoSumTensorImageNewImage<ImageType, NormImageType>(im, out_image);
+ total = yeoSumTensorImage<ImageType>(im);
+ newMatrix = yeoElemMultiply2Matrices<TMatrix>(TMatrix m1, TMatrix m2);
+ newMatrix = yeoElemDivide2Matrices<TMatrix>(TMatrix m1, TMatrix m2);
+ newMatrix = yeoRepTensor2Matrix<TMatrix, TensorType>(TensorType tensor, TMatrix m);
+ yeoGetForegroundMask<TensorImageType, MaskImageType>(im, mask,delimiter, bg, fg);
+ yeoTensor2ScalarImage<TensorImageType, ScalarImageType>(im, out, scalar_type);
+ yeoComputeOverlap2TensorImages<TensorImageType, ScalarImageType>(im1, im2, out); 
*/


/* Remove non positive symmetric definite matrices and replace them
 * with zero.
 *
 * Usage: 
 *         count = yeoCleanTensorImageInPlace<ImageType>(im);
*/
template <class TImageType> int yeoCleanTensorImageInPlace(TImageType* im)
{
  int count = 0;
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());
 
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.Get()!= itk::NumericTraits<typename TImageType::PixelType>::Zero)
      {
      if(!It.Get().IsPositive() )
	{
	It.Set(itk::NumericTraits<typename TImageType::PixelType>::Zero); 
	count++;
	}
      }
    }
  return count;
}


/*
 * Usage: 
 *         count = yeoNumZeroTensors<ImageType>(im);
*/
template <class TImageType> int yeoNumZeroTensors(const TImageType* im)
{
  int count = 0;
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef typename itk::ImageRegionConstIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());
 
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.Get() == itk::NumericTraits<typename TImageType::PixelType>::Zero)
      {
      count++;
      }
    }
  return count;
}

/* Usage:
*           yeoWriteTensorImage<ImageType>(im, filename);
*/
template <class TImageType> void yeoWriteTensorImage(const TImageType *im, std::string filename)
{

  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename TImageType::PixelType PixelType;
  typedef typename PixelType::ValueType ValueType;

  typedef itk::TensorImageIO<ValueType, TImageType::PixelType::Dimension, 3> TensorIOType; 

  typename TensorIOType::Pointer writer = TensorIOType::New();
  writer->SetInput(im);
  writer->SetFileName(filename);
  try
    {
    writer->Write();
    }
  catch (itk::ExceptionObject &e)
    {
    std::cerr<<"Unexpected error in writing tensor image."<<std::endl;
    std::cerr<< e <<std::endl;
    exit(EXIT_FAILURE);
    } 
   
}

/* Usage:
*           yeoReadTensorImage<ImageType>(im, filename);
*/
template <class TImageType> void yeoReadTensorImage(TImageType *im, std::string filename)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename TImageType::PixelType PixelType;
  typedef typename PixelType::ValueType ValueType;

  typedef itk::TensorImageIO<ValueType, TImageType::PixelType::Dimension, 3> TensorIOType; 

  typename TensorIOType::Pointer reader = TensorIOType::New();
  reader->SetFileName(filename);

  //typename TImageType::Pointer im2;
  //reader->SetInput(im2);

  try
    {
    reader->Read();
    }
  catch (itk::ExceptionObject &e)
    {
    std::cerr<<"Unexpected error in reading tensor image."<<std::endl;
    std::cerr<< e <<std::endl;
    exit(EXIT_FAILURE);
    } 
  yeoCopyImage<TImageType>(reader->GetOutput(), im);
}

/* Usage:
*           yeoLogTensorImageInPlace<ImageType>(im);
*/
template <class TImageType> void yeoLogTensorImageInPlace(TImageType *im)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());

   for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.Get().IsZero())
      It.Set(itk::NumericTraits<typename TImageType::PixelType>::Zero);
    else
      {
      if(!It.Get().IsPositive())
	{
      std::cerr<<"Tensor not positive definite before taking log: "<<It.Get()<<std::endl;
      exit(EXIT_FAILURE);
	}
      else
	{
	It.Set(It.Get().Log());
	}
      }
    }
}

/* Usage:
*           yeoLogTensorImageNewImage<ImageType>(im, outIm);
*/
template <class TImageType> void yeoLogTensorImageNewImage(TImageType *im, TImageType *outIm)
{
  if(outIm == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //Create Output Image
  yeoCopyImage<TImageType>(im, outIm);

  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(outIm, outIm->GetLargestPossibleRegion());
  
  for(It.GoToBegin(), outIt.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
    {
    if(It.Get().IsZero())
      outIt.Set(itk::NumericTraits<typename TImageType::PixelType>::Zero);
    else
      outIt.Set(It.Get().Log());
    }
}


/* Usage:
*           yeoExpTensorImageInPlace<ImageType>(im);
*/
template <class TImageType> void yeoExpTensorImageInPlace(TImageType *im)
{

  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());

   for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.Get().IsZero())
      It.Set(itk::NumericTraits<typename TImageType::PixelType>::Zero);
    else
      It.Set(It.Get().Exp());
    }
}


/* Usage:
*           yeoExpTensorImageNewImage<ImageType>(im, outIm);
*/
template <class TImageType> void yeoExpTensorImageNewImage(const TImageType *im, TImageType *outIm)
{
  if(outIm == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }


  //Create Output Image
  yeoCopyImage<TImageType>(im, outIm);

  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(outIm, outIm->GetLargestPossibleRegion());
  
  for(It.GoToBegin(), outIt.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
    {
    if(It.Get().IsZero())
      outIt.Set(itk::NumericTraits<typename TImageType::PixelType>::Zero);
    else
      outIt.Set(It.Get().Exp());
    }
}

/* Usage:
*         yeoSqNormTensorImageNewImage<TensorImageType, NormImageType>(im, out_image); 
*/
template <class TensorImageType, class NormImageType> void yeoSqNormTensorImageNewImage(const TensorImageType *im, NormImageType *out_image)
{
  if(out_image == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  out_image->SetRegions(im->GetLargestPossibleRegion());
  out_image->SetSpacing(im->GetSpacing());
  out_image->SetOrigin(im->GetOrigin());
  out_image->Allocate();
  
  typedef typename itk::ImageRegionConstIterator<TensorImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionIterator<NormImageType> IteratorType;
  IteratorType outIt = IteratorType(out_image, out_image->GetLargestPossibleRegion());
  
  for(It.GoToBegin(), outIt.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
    {
    outIt.Set(It.Get().GetSquaredNorm());
    }  
}

/* Usage:
*         yeoNormTensorImageNewImage<TensorImageType, NormImageType>(im, out_image); 
*/
template <class TensorImageType, class NormImageType> void yeoNormTensorImageNewImage(const TensorImageType *im, NormImageType *out_image)
{
  if(out_image == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  out_image->SetRegions(im->GetLargestPossibleRegion());
  out_image->SetSpacing(im->GetSpacing());
  out_image->SetOrigin(im->GetOrigin());
  out_image->Allocate();
  
  typedef typename itk::ImageRegionConstIterator<TensorImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionIterator<NormImageType> IteratorType;
  IteratorType outIt = IteratorType(out_image, out_image->GetLargestPossibleRegion());
  
  for(It.GoToBegin(), outIt.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
    {
    outIt.Set(It.Get().GetNorm());
    }  
}



/* Usage:
*  yeoMultiplyTensorImageWithImageInPlace<TImageType, ImageType>(im1, im2);
*/
template <class TensorImageType, class ImageType> void yeoMultiplyTensorImageWithImageInPlace(TensorImageType *im1, const ImageType *im2)
{
   if(im1 == 0)
    {
    std::cout<<"im1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
   if(im2 == 0)
     {
    std::cout<<"im2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

   typedef typename itk::ImageRegionConstIterator<ImageType> ConstIteratorType;
   ConstIteratorType It2 = ConstIteratorType(im2, im2->GetLargestPossibleRegion());
   
   typedef typename itk::ImageRegionIterator<TensorImageType> IteratorType;
   IteratorType It = IteratorType(im1, im1->GetLargestPossibleRegion());
   
   for(It2.GoToBegin(), It.GoToBegin(); !It.IsAtEnd(); ++It, ++It2)
     {
     It.Set(It.Get() * It2.Get());
     }  
}

/* Usage:
* yeoMultiplyTensorImageWithImageNewImage<TImageType, ImageType>(im1, im2, outIm);
*/

template <class TensorImageType, class ImageType> void yeoMultiplyTensorImageWithImageNewImage(TensorImageType *im1, ImageType *im2, TensorImageType *out_image)
{
  if(im1 == 0)
    {
    std::cout<<"im1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im2 == 0)
    {
    std::cout<<"im2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

   if(out_image == 0)
    {
    std::cout<<"outIm is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

   out_image->SetRegions(im1->GetLargestPossibleRegion());
   out_image->SetSpacing(im1->GetSpacing());
   out_image->SetOrigin(im1->GetOrigin());
   out_image->Allocate();
   
   typedef typename itk::ImageRegionConstIterator<ImageType> ConstIteratorType2;
   ConstIteratorType2 It2 = ConstIteratorType2(im2, im2->GetLargestPossibleRegion());
   
   typedef typename itk::ImageRegionConstIterator<TensorImageType> ConstIteratorType1;
   ConstIteratorType1 It1 = ConstIteratorType1(im1, im1->GetLargestPossibleRegion());

   typedef typename itk::ImageRegionIterator<TensorImageType> IteratorType;
   IteratorType It = IteratorType(out_image, out_image->GetLargestPossibleRegion());

   for(It2.GoToBegin(), It1.GoToBegin(), It.GoToBegin(); !It.IsAtEnd(); ++It, ++It2, ++It1)
     {
     It.Set(It2.Get() * It1.Get());
     }  
}



/* Usage:  typename ImageType::Pointer out = ImageType::New();
*         yeoWarpTensorImageNewImage<ImageType, DeformationFieldType>(im, df, out);
*
*  pad edge with zero tensors
*/
template <class TImageType, class DeformationFieldType> void yeoWarpTensorImageNewImage(const TImageType *im, const DeformationFieldType *df, TImageType *out)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(df == 0)
    {
    std::cout<<"Deformation Field is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  

  typedef itk::WarpTensorImageFilter<TImageType, TImageType, DeformationFieldType> WarpTensorImageFilterType;
  typename WarpTensorImageFilterType::Pointer warper = WarpTensorImageFilterType::New();
  
  warper->SetInput( im);
  warper->SetOutputSpacing( df->GetSpacing() );
  warper->SetOutputOrigin( df->GetOrigin() );
  warper->SetDeformationField(df);
  warper->SetEdgePaddingValue( itk::NumericTraits<typename TImageType::PixelType>::Zero );
  warper->SetReorientationStrategy(FS);

  try
    {
    warper->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Warper fails." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }

  yeoCopyImage<TImageType>(warper->GetOutput(), out);   
}

/* Usage:  typename ImageType::Pointer out = ImageType::New();
*         yeoWarpTensorImageNewImagePPD<ImageType, DeformationFieldType>(im, df, out);
*
*  pad edge with zero tensors
*/
template <class TImageType, class DeformationFieldType> void yeoWarpTensorImageNewImagePPD(const TImageType *im, const DeformationFieldType *df, TImageType *out)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(df == 0)
    {
    std::cout<<"Deformation Field is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  

  typedef itk::WarpTensorImageFilter<TImageType, TImageType, DeformationFieldType> WarpTensorImageFilterType;
  typename WarpTensorImageFilterType::Pointer warper = WarpTensorImageFilterType::New();
  
  warper->SetInput( im);
  warper->SetOutputSpacing( df->GetSpacing() );
  warper->SetOutputOrigin( df->GetOrigin() );
  warper->SetDeformationField(df);
  warper->SetEdgePaddingValue( itk::NumericTraits<typename TImageType::PixelType>::Zero );
  warper->SetReorientationStrategy(PPD);

  try
    {
    warper->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Warper fails." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }

  yeoCopyImage<TImageType>(warper->GetOutput(), out);   
}


/* Usage:  typename ImageType::Pointer out = ImageType::New();
*         yeoWarpTensorImageWOReorientationNewImage<ImageType, DeformationFieldType>(im, df, out);
*
*  pad edge with zero tensors
*/
template <class TImageType, class DeformationFieldType> void yeoWarpTensorImageWOReorientationNewImage(const TImageType *im, const DeformationFieldType *df, TImageType *out)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(df == 0)
    {
    std::cout<<"Deformation Field is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  

  typedef itk::WarpTensorImageWOReorientationFilter<TImageType, TImageType, DeformationFieldType> WarpTensorImageFilterType;
  typename WarpTensorImageFilterType::Pointer warper = WarpTensorImageFilterType::New();
  
  warper->SetInput( im);
  warper->SetOutputSpacing( df->GetSpacing() );
  warper->SetOutputOrigin( df->GetOrigin() );
  warper->SetDeformationField(df);
  warper->SetEdgePaddingValue( itk::NumericTraits<typename TImageType::PixelType>::Zero );

  try
    {
    warper->Update();
    }
  catch( itk::ExceptionObject& err )
    {
    std::cout << "Warper fails." << std::endl;
    std::cout << err << std::endl;
    exit( EXIT_FAILURE );
    }

  yeoCopyImage<TImageType>(warper->GetOutput(), out);   
}





/* Usage: 
*          total = yeoPowerCombine2Tensors<TensorType>(tensor1, tensor2, a, b);
*/
template <class TensorType> TensorType yeoPowerCombine2Tensors(TensorType tensor1, TensorType tensor2, float a, float b)
{
  int NDegreesOfFreedom;
  TensorType tensor;

  NDegreesOfFreedom = TensorType::DegreesOfFreedom;
  
  for(int j = 0; j < NDegreesOfFreedom; j++)
    {
      tensor.SetNthComponent(j, static_cast<typename TensorType::RealType>(pow( (double)tensor1.GetNthComponent(j), (double)a)*pow( (double)tensor2.GetNthComponent(j), (double)b)));
    }
  return tensor;
}



/* Usage: 
*          total = yeoSumTensor<TensorType>(tensor);
*/
template <class TensorType> double yeoSumTensor(TensorType tensor)
{
  int NDimension;
  double sum = 0;

  NDimension = TensorType::Dimension;
    
// diagonal elements:
  for( int j=0; j<NDimension; j++ )
    sum += tensor.GetNthComponent (j*(j+1)/2 + j);
          
  // off-diagonal elements
  for( int j=1; j<NDimension; j++ )
    for( int i=0; i<j; i++ )
      sum += 2.0*tensor.GetComponent (i,j);

  return sum;
}

/* Usage: 
*          tensor = yeoAbsTensor<TensorType>(tensor);
*/
template <class TensorType> TensorType yeoAbsTensor(TensorType tensor)
{
  unsigned int DegreesOfFreedom = TensorType::DegreesOfFreedom;
  
  for(unsigned int i = 0; i < DegreesOfFreedom; i++)
    tensor.SetNthComponent(i, fabs(tensor.GetNthComponent(i)));

  return tensor;
}



/* Usage:  typename ImageType::Pointer out = ImageType::New();
 *         yeoPowerCombine2TensorImagesNewImage<ImageType>(in1, in2, out, a, b);
*/
template <class TImageType> void yeoPowerCombine2TensorImagesNewImage(const TImageType* in1, const TImageType* in2, TImageType *out, float a, float b)
{

  if(in1 == 0)
    {
    std::cout<<"Input Image 1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(in2 == 0)
    {
    std::cout<<"Input Image 2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cerr<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  //First Copy image
  yeoCopyImage(in1, out);

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType outIt = IteratorType(out, out->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(in2, in2->GetLargestPossibleRegion());
  
  int NDegreesOfFreedom;
  for(outIt.GoToBegin(), It2.GoToBegin(); !outIt.IsAtEnd(); ++outIt, ++It2)
    {
    if(outIt.IsAtBegin())
      {
      int NDimension = outIt.Get().GetTensorDimension();
      NDegreesOfFreedom = NDimension*(NDimension+1)/2;
      }
    
    //typename TImageType::PixelType tensor;
    typename TImageType::PixelType tensor1 = outIt.Get();
    typename TImageType::PixelType tensor2 = It2.Get();
    
    /*
    for(int j = 0; j < NDegreesOfFreedom; j++)
      {
      tensor.SetNthComponent(j, pow(tensor1.GetNthComponent(j), a)*pow(tensor2.GetNthComponent(j), b));
      }
    */
    outIt.Set(yeoPowerCombine2Tensors<typename TImageType::PixelType>(tensor1, tensor2, a, b));
    } 
}

/* Usage:  typename ImageType::Pointer out = ImageType::New();
 *         yeoPowerCombine2TensorImagesInPlace<ImageType>(in1, in2, a, b);
* In Place for in1
*/
template <class TImageType> void yeoPowerCombine2TensorImagesInPlace(TImageType* in1, const TImageType* in2, float a, float b)
{

  if(in1 == 0)
    {
    std::cout<<"Input Image 1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(in2 == 0)
    {
    std::cout<<"Input Image 2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
 
  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It1 = IteratorType(in1, in1->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< TImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(in2, in2->GetLargestPossibleRegion());
  
  int NDegreesOfFreedom;
  for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
    {
    if(It1.IsAtBegin())
      {
      int NDimension = It1.Get().GetTensorDimension();
      NDegreesOfFreedom = NDimension*(NDimension+1)/2;
      }
    
    typename TImageType::PixelType tensor1 = It1.Get();
    typename TImageType::PixelType tensor2 = It2.Get();
    /*
    for(int j = 0; j < NDegreesOfFreedom; j++)
      {
      tensor1.SetNthComponent(j, pow(tensor1.GetNthComponent(j), a)*pow(tensor2.GetNthComponent(j), b));
      }
    */
    It1.Set(yeoPowerCombine2Tensors<typename TImageType::PixelType>(tensor1, tensor2, a, b));
    }
  
}

/* Usage:  
 *         yeoReplaceValueInTensorImageInPlace<ImageType>(im, oldValue, newValue, optype);
*
* optype: 0 = "=="
*         1 = "<"
*         2 = ">"
*         3 = "<="
*         4 = ">="
*
*/
template <class TImageType> void yeoReplaceValueInTensorImageInPlace(TImageType* im, typename TImageType::PixelType oldValue, typename TImageType::PixelType newValue, int optype)
{
   if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());

  typename TImageType::PixelType tensor;
  typename TImageType::PixelType tensor1;
  int NDegreesOfFreedom;
  switch(optype)
    {
    case 0 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.IsAtBegin())
      {
      int NDimension = It.Get().GetTensorDimension();
      NDegreesOfFreedom = NDimension*(NDimension+1)/2;
      }

      tensor1 = It.Get();      
      for(int j = 0; j < NDegreesOfFreedom; j++)
	{
	if(tensor1.GetNthComponent(j) == oldValue.GetNthComponent(j))
	  tensor.SetNthComponent(j, newValue.GetNthComponent(j));
	else
	  tensor.SetNthComponent(j, tensor1.GetNthComponent(j));
	}
      It.Set(tensor);
      }
    break;
    }
    case 1 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.IsAtBegin())
	{
	int NDimension = It.Get().GetTensorDimension();
	NDegreesOfFreedom = NDimension*(NDimension+1)/2;
	}
      
      tensor1 = It.Get();      
      for(int j = 0; j < NDegreesOfFreedom; j++)
	{
	if(tensor1.GetNthComponent(j) < oldValue.GetNthComponent(j))
	  tensor.SetNthComponent(j, newValue.GetNthComponent(j));
	else
	  tensor.SetNthComponent(j, tensor1.GetNthComponent(j));
	}
      It.Set(tensor);
      }
    break;
    }
    case 2 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.IsAtBegin())
	{
	int NDimension = It.Get().GetTensorDimension();
	NDegreesOfFreedom = NDimension*(NDimension+1)/2;
	}
      
      tensor1 = It.Get();      
      for(int j = 0; j < NDegreesOfFreedom; j++)
	{
	if(tensor1.GetNthComponent(j) > oldValue.GetNthComponent(j))
	  tensor.SetNthComponent(j, newValue.GetNthComponent(j));
	else
	  tensor.SetNthComponent(j, tensor1.GetNthComponent(j));
	}
      It.Set(tensor);
      }
    break;
    }
    case 3 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.IsAtBegin())
	{
	int NDimension = It.Get().GetTensorDimension();
	NDegreesOfFreedom = NDimension*(NDimension+1)/2;
	}
      
      tensor1 = It.Get();      
      for(int j = 0; j < NDegreesOfFreedom; j++)
	{
	if(tensor1.GetNthComponent(j) <= oldValue.GetNthComponent(j))
	  tensor.SetNthComponent(j, newValue.GetNthComponent(j));
	else
	  tensor.SetNthComponent(j, tensor1.GetNthComponent(j));
	}
      It.Set(tensor);
      }
    break;
    }
    case 4 :
    {
    for(It.GoToBegin(); !It.IsAtEnd(); ++It)
      {
      if(It.IsAtBegin())
	{
	int NDimension = It.Get().GetTensorDimension();
	NDegreesOfFreedom = NDimension*(NDimension+1)/2;
	}
      
      tensor1 = It.Get();      
      for(int j = 0; j < NDegreesOfFreedom; j++)
	{
	if(tensor1.GetNthComponent(j) >= oldValue.GetNthComponent(j))
	  tensor.SetNthComponent(j, newValue.GetNthComponent(j));
	else
	  tensor.SetNthComponent(j, tensor1.GetNthComponent(j));
	}
      It.Set(tensor);
      }
    break;
    }    
    default :
    {
    std::cerr<<"Opcode does not exist"<<std::endl;
    exit(EXIT_FAILURE);
    }
    } //end switch 
}

/* Usage: 
 *         yeoMaxTensorValue<ImageType>(im);
*/
template <class TImageType> typename TImageType::PixelType yeoMaxTensorValue(const TImageType* im)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef typename itk::ImageRegionConstIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());
  
  typedef typename TImageType::PixelType TensorType;
  TensorType maxTensor = TensorType( itk::NumericTraits<float>::min() );
  
  int NDegreesOfFreedom;
  TensorType tensor;
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.IsAtBegin())
      {
      int NDimension = It.Get().GetTensorDimension();
      NDegreesOfFreedom = NDimension*(NDimension+1)/2;
      }
    
    tensor = It.Get();      
    for(int j = 0; j < NDegreesOfFreedom; j++)
      {
      if(tensor.GetNthComponent(j) > maxTensor.GetNthComponent(j))
	maxTensor.SetNthComponent(j, tensor.GetNthComponent(j));
      }
    }
  
  return maxTensor;
}

/* Usage: 
 *         yeoMinTensorValue<ImageType>(im);
*/
template <class TImageType> typename TImageType::PixelType yeoMinTensorValue(const TImageType* im)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  typedef typename itk::ImageRegionConstIterator< TImageType> IteratorType;
  IteratorType It = IteratorType(im, im->GetLargestPossibleRegion());
  
  typedef typename TImageType::PixelType TensorType;
  TensorType minTensor = TensorType( itk::NumericTraits<float>::max() );
  
  int NDegreesOfFreedom;
  TensorType tensor;
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.IsAtBegin())
      {
      int NDimension = It.Get().GetTensorDimension();
      NDegreesOfFreedom = NDimension*(NDimension+1)/2;
      }
    
    tensor = It.Get();      
    for(int j = 0; j < NDegreesOfFreedom; j++)
      {
      if(tensor.GetNthComponent(j) < minTensor.GetNthComponent(j))
	minTensor.SetNthComponent(j, tensor.GetNthComponent(j));
      }
    }
  
  return minTensor;
}

/* Usage: 
*        yeoSumTensorImageNewImage<ImageType, NormImageType>(im, out_image);
*/
template <class TImageType, class NormImageType> void yeoSumTensorImageNewImage(const TImageType* im, NormImageType *out_image)
{

  if(out_image == 0)
    {
    std::cout<<"out_image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  out_image->SetRegions(im->GetLargestPossibleRegion());
  out_image->SetSpacing(im->GetSpacing());
  out_image->SetOrigin(im->GetOrigin());
  out_image->Allocate();
  
  typedef typename itk::ImageRegionConstIterator<TImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionIterator<NormImageType> IteratorType;
  IteratorType outIt = IteratorType(out_image, out_image->GetLargestPossibleRegion());
  

  typedef typename TImageType::PixelType TensorType;
  int NDimension; 
  //double sum;
  TensorType tensor;
  
  for(It.GoToBegin(), outIt.GoToBegin(); !It.IsAtEnd(); ++It, ++outIt)
    {
    if(It.IsAtBegin())
      NDimension = It.Get().GetTensorDimension();
    
//    sum = 0;
    tensor = It.Get();
  
/*  
// diagonal elements:
    for( int j=0; j<NDimension; j++ )
      sum += tensor.GetNthComponent (j*(j+1)/2 + j);
          
    // off-diagonal elements
    for( int j=1; j<NDimension; j++ )
      for( int i=0; i<j; i++ )
        sum += 2.0*tensor.GetComponent (i,j);
  */  
    outIt.Set(yeoSumTensor<TensorType>(tensor));
    }    
}

/* Usage: 
*          total = yeoSumTensorImage<ImageType>(im);
*/
template <class TImageType> double yeoSumTensorImage(const TImageType* im)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionConstIterator<TImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());

  typedef typename TImageType::PixelType TensorType;
  int NDimension; double sum;
  TensorType tensor;

  sum = 0;
  for(It.GoToBegin(); !It.IsAtEnd(); ++It)
    {
    if(It.IsAtBegin())
      NDimension = It.Get().GetTensorDimension();
    
    tensor = It.Get();
    sum += yeoSumTensor<TensorType>(tensor);
/*
// diagonal elements:
    for( int j=0; j<NDimension; j++ )
      sum += tensor.GetNthComponent (j*(j+1)/2 + j);
          
    // off-diagonal elements
    for( int j=1; j<NDimension; j++ )
      for( int i=0; i<j; i++ )
	sum += 2.0*tensor.GetComponent (i,j);
  */    
    }
  return(sum);
}

/* Usage: 
*          newMatrix = yeoElemMultiply2Matrices<TMatrix>(TMatrix m1, TMatrix m2);
*/
template <class TMatrix> TMatrix yeoElemMultiply2Matrices(const TMatrix m1, const TMatrix m2)
{
  TMatrix m3;

  unsigned int numRows = TMatrix::RowDimensions;
  unsigned int numCols = TMatrix::ColumnDimensions;

  for(unsigned int i = 0; i < numRows; i++)
    for(unsigned int j = 0; j < numCols; j++)
      m3[i][j] = m1[i][j] * m2[i][j];

  return m3;
}

/* Usage: 
*          newMatrix = yeoElemDivide2Matrices<TMatrix>(TMatrix m1, TMatrix m2);
*/
template <class TMatrix> TMatrix yeoElemDivide2Matrices(const TMatrix m1, const TMatrix m2)
{
  TMatrix m3;

  unsigned int numRows = TMatrix::RowDimensions;
  unsigned int numCols = TMatrix::ColumnDimensions;

  for(unsigned int i = 0; i < numRows; i++)
    for(unsigned int j = 0; j < numCols; j++)
      m3[i][j] = m1[i][j] / m2[i][j];

  return m3;
}

/* Usage: 
*          newMatrix = yeoRepTensor2Matrix<TMatrix, TensorType>(TensorType tensor);
*/
template <class TMatrix, class TensorType> TMatrix yeoRepTensor2Matrix(const TensorType tensor)
{
  TMatrix m2;

  //Assume that tensor is of same dimension with at least the row or
  //columns of m

  unsigned int numRows = TMatrix::RowDimensions;
  unsigned int numCols = TMatrix::ColumnDimensions;
  unsigned int numElems = TensorType::DegreesOfFreedom;

  if(numElems == numRows)
    {
    //repeat tensor numCols times.
    for(unsigned int i = 0; i < numCols; i++)
      for(unsigned int j = 0; j < numRows; j++)
	m2[j][i] = tensor.GetNthComponent(j);
    
    }
  else if(numElems == numCols)
    {
     //repeat tensor numRows times.
    for(unsigned int i = 0; i < numRows; i++)
      for(unsigned int j = 0; j < numCols; j++)
	m2[i][j] = tensor.GetNthComponent(j);
    }
  else
    {
    std::cerr<<"Tensor not of same dimension with either rows or cols of matrix"<<std::endl;
    exit(EXIT_FAILURE);
    }

  return m2;
}


/* Usage: 
*          yeoGetForegroundMask<TensorImageType, MaskImageType>(im, mask, delimiter, bg, fg);
*/
template <class TensorImageType, class MaskImageType> void yeoGetForegroundMask(const TensorImageType *im, MaskImageType *mask, typename TensorImageType::PixelType delimiter, typename MaskImageType::PixelType bg_value, typename MaskImageType::PixelType fg_value )
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  yeoCreateZeroImage<TensorImageType, MaskImageType>(im, mask);

  typedef typename itk::ImageRegionConstIterator< TensorImageType> ConstIteratorType;
  ConstIteratorType It = ConstIteratorType(im, im->GetLargestPossibleRegion());

  typedef typename itk::ImageRegionIterator< MaskImageType> IteratorType;
  IteratorType maskIt = IteratorType(mask, mask->GetLargestPossibleRegion());

  for(It.GoToBegin(); !It.IsAtEnd(); ++It, ++maskIt)
    {
    if(It.Get() == delimiter)
      {
      maskIt.Set(bg_value);
      }
    else
      maskIt.Set(fg_value);
    }

}


/* Usage:
*           yeoTensor2ScalarImage<TensorImageType, ScalarImageType>(im, out, scalar_type);
*/
template <class TensorImageType, class ScalarImageType> void yeoTensor2ScalarImage(const TensorImageType *im, ScalarImageType *out, char *scalar_type)
{
  if(im == 0)
    {
    std::cout<<"Input Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  if(out == 0)
    {
    std::cout<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef itk::TensorToScalarFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> FunctionType;

  typedef itk::TensorToScalarTensorImageFilter<TensorImageType, ScalarImageType> FilterType;
  typename FilterType::Pointer myFilter = FilterType::New();



  bool foundFunction = false;

  if( strcmp (scalar_type, "fa")==0 || strcmp (scalar_type, "FA")==0 )
    {
    typedef itk::TensorToFAFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "lfa")==0 || strcmp (scalar_type, "LFA")==0 )
    {
    typedef itk::TensorToLogFAFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "adc")==0 || strcmp (scalar_type, "ADC")==0 )
    {
    typedef itk::TensorToADCFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "vol")==0 || strcmp (scalar_type, "VOL")==0 )
    {
    typedef itk::TensorToVolumeFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "cl")==0 || strcmp (scalar_type, "CL")==0 )
    {
    typedef itk::TensorToClFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "cp")==0 || strcmp (scalar_type, "CP")==0 )
    {
    typedef itk::TensorToCpFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "cs")==0 || strcmp (scalar_type, "CS")==0 )
    {
    typedef itk::TensorToCsFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "ra")==0 || strcmp (scalar_type, "RA")==0 )
    {
    typedef itk::TensorToRAFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "vr")==0 || strcmp (scalar_type, "VR")==0 )
    {    
    typedef itk::TensorToVRFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }
  
  if( strcmp (scalar_type, "l1")==0 || strcmp (scalar_type, "L1")==0 )
    {
    typedef itk::TensorToLambdaFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFunction->SetLambdaIndex(2);
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "l2")==0 || strcmp (scalar_type, "L2")==0 )
    {
    typedef itk::TensorToLambdaFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFunction->SetLambdaIndex(1);
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }
  
  if( strcmp (scalar_type, "l3")==0 || strcmp (scalar_type, "L3")==0 )
    {
    typedef itk::TensorToLambdaFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFunction->SetLambdaIndex(0);
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if( strcmp (scalar_type, "dp")==0 || strcmp (scalar_type, "DP")==0 )
    {
    typedef itk::TensorToDispersionFunction<typename TensorImageType::PixelType, typename ScalarImageType::PixelType> SpecificFunctionType;
    typename SpecificFunctionType::Pointer myFunction = SpecificFunctionType::New();
    myFilter->SetTensorToScalarFunction (myFunction);
    foundFunction = true;    
    }

  if(foundFunction)
    {
    // compute the function    
    myFilter->SetInput (im);

    try
      {
      myFilter->Update();
      }
    catch (itk::ExceptionObject &e)
      {
      std::cerr<<"Unexpected error in converting tensor image to scalar image."<<std::endl;
      std::cerr << e;
      exit(EXIT_FAILURE);
      }
    yeoCopyImage<ScalarImageType>(myFilter->GetOutput(), out); 

    }
  else
    {
    std::cout<<"Requested Scalar Function Not Found!"<<std::endl;
    exit(EXIT_FAILURE);
    }

}


/* Usage:  typename ScalarImageType::Pointer out = ScalarImageType::New();
 *         yeoComputeOverlap2TensorImages<TensorImageType, ScalarImageType>(im1, im2, out);
*/
template <class TensorImageType, class ScalarImageType> void yeoComputeOverlap2TensorImages(const TensorImageType* im1, const TensorImageType* im2, ScalarImageType *out)
{

  if(im1 == 0)
    {
    std::cout<<"Input Image 1 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(im2 == 0)
    {
    std::cout<<"Input Image 2 is null"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(out == 0)
    {
    std::cerr<<"Output Image is null"<<std::endl;
    exit(EXIT_FAILURE);
    }

  typedef typename itk::ImageRegionIterator< ScalarImageType> IteratorType;
  IteratorType outIt = IteratorType(out, out->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< TensorImageType> ConstIteratorType;
  ConstIteratorType It1 = ConstIteratorType(im1, im1->GetLargestPossibleRegion());
  ConstIteratorType It2 = ConstIteratorType(im2, im2->GetLargestPossibleRegion());


  int NDimension;
  typename ScalarImageType::PixelType overlap;
  typename ScalarImageType::PixelType eigen;
  typename ScalarImageType::PixelType inner_prod;
  typename TensorImageType::PixelType::VectorType vec1;
  typename TensorImageType::PixelType::VectorType vec2;

  for(outIt.GoToBegin(), It1.GoToBegin(), It2.GoToBegin(); !outIt.IsAtEnd(); ++outIt, ++It1, ++It2)
    {
    if(It1.IsAtBegin())
      {
      NDimension = It1.Get().GetTensorDimension();
      }

    typename TensorImageType::PixelType t1 = It1.Get();
    typename TensorImageType::PixelType t2 = It2.Get();
    
    if(!t1.IsZero() && !t2.IsZero())
      {
      overlap = 0;
      eigen = 0;
      
      for (int i = 0; i < NDimension; i++)
	{
	eigen += t1.GetEigenvalue(i) * t2.GetEigenvalue(i);
	
	vec1 = t1.GetEigenvector(i);
	vec2 = t2.GetEigenvector(i);
	inner_prod = 0;
	for (int j = 0; j < NDimension; j++)
	  inner_prod += vec1[j]*vec2[j];
	
	overlap +=  t1.GetEigenvalue(i) * t2.GetEigenvalue(i) * inner_prod * inner_prod;
	}
      
      outIt.Set(overlap/eigen);
      }
    else
      outIt.Set(0.0);

    }
  
 
}


#endif
