#ifndef _itkDiffeomorphicDemonsRegistrationFilterTensor_txx
#define _itkDiffeomorphicDemonsRegistrationFilterTensor_txx
#include "itkDiffeomorphicDemonsRegistrationFilterTensor.h"
#include "itkESMDemonsRegistrationFunctionTensor.h"
#include "itkDivideByConstantImageFilter.h"

#include "itkSmoothingRecursiveGaussianImageFilter.h"
#include "itkImageRegionConstIterator.h"
#include "itkImageRegionIterator.h"
#include "UsefulFunctions.h"
#include "UsefulTensorFunctions.h"
#include "gmm/gmm.h"

//#define DEBUG
#include "Debug.h"
#define STEP_SIZE 1

namespace itk {


/* Usage: 
*          ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(df, filename1, filename2);
*/
template <class DeformationFieldType, class FixedImageType> void ComputeStatisticsDeformationField(const DeformationFieldType *defField, const FixedImageType *fixedImage, std::string filename1, std::string filename2)
{
  typedef typename itk::Image<float, DeformationFieldType::ImageDimension> MaskImageType;

  typename MaskImageType::Pointer DefFieldNormImage = MaskImageType::New();
  yeoNormTensorImageNewImage<DeformationFieldType, MaskImageType>(defField, DefFieldNormImage);
     
  typename MaskImageType::Pointer fixedImage_fg = MaskImageType::New();
  yeoGetForegroundMask<FixedImageType, MaskImageType>(fixedImage, fixedImage_fg, itk::NumericTraits<typename FixedImageType::PixelType>::Zero, 0, 1);
                                                                                     
  double numPixels = yeoSumImage<MaskImageType>(fixedImage_fg);
     
  yeoPowerCombine2ImagesInPlace<MaskImageType>(DefFieldNormImage, fixedImage_fg, 1, 1);
  std::cout<<"Mean Warp: "<<yeoSumImage<MaskImageType>(DefFieldNormImage)/numPixels<<", ";
  std::cout<<"Max Warp: "<<yeoMaxValue<MaskImageType>(DefFieldNormImage)<<", "; 

  if(!filename1.length()==0)
    {
    yeoWriteImage<DeformationFieldType>(defField, filename1);
    }

  if(!filename2.length()==0)
    {
    yeoWriteImage<MaskImageType>(DefFieldNormImage, filename2);
    }
}



/*
 * Default constructor
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::DiffeomorphicDemonsRegistrationFilterTensor()
{
 
  typename DemonsRegistrationFunctionType::Pointer drfp;
  drfp = DemonsRegistrationFunctionType::New();

  this->SetDifferenceFunction( static_cast<FiniteDifferenceFunctionType *>(
                                 drfp.GetPointer() ) );

  m_Multiplier = MultiplyByConstantType::New();
  m_Multiplier->InPlaceOn();

  m_Exponentiator = FieldExponentiatorType::New();
  
  m_Warper = VectorWarperType::New();
  FieldInterpolatorPointer VectorInterpolator =
     FieldInterpolatorType::New();
  m_Warper->SetInterpolator(VectorInterpolator);

  m_Adder = AdderType::New();
  m_Adder->InPlaceOn();

  m_PrevConjField = 0;
  m_OldGradField = 0;
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::SetUseAtlas(bool useAtlas) 
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
    (this->GetDifferenceFunction().GetPointer());
  
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to SymmetricDemonsRegistrationFunction" );
   }
  
  drfp->SetUseAtlas(useAtlas);
}


/*
 * Set the Var image.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::SetVarImage(
  const VarImageType * ptr )
{
  this->ProcessObject::SetNthInput( 3, const_cast< VarImageType * >( ptr ) );
}


/*
 * Get the Var image.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
const typename DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::VarImageType *
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetVarImage() const
{
  return dynamic_cast< const VarImageType * >
    ( this->ProcessObject::GetInput( 3 ) );
}



/*
 * Set the function state values before each iteration
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::InitializeIteration()
{
  // update variables in the equation object
  DemonsRegistrationFunctionType *f = 
    dynamic_cast<DemonsRegistrationFunctionType *>
    (this->GetDifferenceFunction().GetPointer());

  if ( !f )
    {
    itkExceptionMacro(<<"FiniteDifferenceFunction not of type DemonsRegistrationFunctionType");
    }

  f->SetDeformationField( this->GetDeformationField() );



  VarImageConstPointer varPtr = this->GetVarImage();
  f->SetVarImage(varPtr);

  if(!f->GetVarImage() && f->GetUseAtlas())
    {
    std::cerr<<"*************************"<<std::endl;
    std::cerr<<"Diffeomorphic Demons: UseAtlas Boolean is true but VarImage does not exists"<<std::endl;
    std::cerr<<"*************************"<<std::endl;
    exit(EXIT_FAILURE);
    } 

  if(this->GetUseGradientType3() == DemonsRegistrationFunctionType::ConjugateGradient  &&
         (this->m_PrevConjField.IsNull() || 
	  this->GetUpdateBuffer()->GetLargestPossibleRegion().GetSize() != this->m_PrevConjField->GetLargestPossibleRegion().GetSize())) 
    {

    this->m_PrevConjField = DeformationFieldType::New();
    yeoCreateZeroImage<DeformationFieldType, DeformationFieldType>(this->GetUpdateBuffer(), this->m_PrevConjField);

    this->m_OldGradField = DeformationFieldType::New();
    yeoCreateZeroImage<DeformationFieldType, DeformationFieldType>(this->GetUpdateBuffer(), this->m_OldGradField);
   
    }
  
  if(this->GetUseGradientType3() == DemonsRegistrationFunctionType::ConjugateGradient)
     m_CurrentSSD = ComputeSSD(this->GetOutput());


  // call the superclass  implementation ( initializes f )
  Superclass::InitializeIteration();
}


/*
 * Get the metric value from the difference function
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
double
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetMetric() const
{
 
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DiffeomorphicDemonsRegistrationFunction" );
   }
   
  return drfp->GetMetric();
}

/*
 * Get the metric value from the difference function
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
double
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetNormalizedMetric() const
{
 
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DiffeomorphicDemonsRegistrationFunction" );
   }
   
  return drfp->GetNormalizedMetric();
}

/*
 * Get the metric value from the difference function
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
unsigned long
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetNumberOfPixelsProcessed() const
{
 
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DiffeomorphicDemonsRegistrationFunction" );
   }
   
  return drfp->GetNumberOfPixelsProcessed();
}




/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
double
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetIntensityDifferenceThreshold() const
{
 
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DemonsRegistrationFunction" );
   }
   
  return drfp->GetIntensityDifferenceThreshold();
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::SetIntensityDifferenceThreshold(double threshold) 
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to SymmetricDemonsRegistrationFunction" );
   }
   
  drfp->SetIntensityDifferenceThreshold(threshold);
}


/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
double
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetMaximumUpdateStepLength() const
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DemonsRegistrationFunction" );
   }
  
  return drfp->GetMaximumUpdateStepLength();
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::SetMaximumUpdateStepLength(double threshold) 
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to SymmetricDemonsRegistrationFunction" );
   }
  
  drfp->SetMaximumUpdateStepLength(threshold);
}


/*
 * Get the metric value from the difference function
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
const double &
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetRMSChange() const
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DiffeomorphicDemonsRegistrationFunction" );
   }
   
  return drfp->GetRMSChange();
}


/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
typename DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::DeformationFieldPointer
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetUpdateField() const
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DemonsRegistrationFunction" );
   }
  
  return drfp->GetUpdateField();
}


/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
typename DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GradientType
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetUseGradientType() const
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DemonsRegistrationFunction" );
   }
  
  return drfp->GetUseGradientType();
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::SetUseGradientType(GradientType gtype) 
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to SymmetricDemonsRegistrationFunction" );
   }
  
  drfp->SetUseGradientType(gtype);
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
typename DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GradientType2
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetUseGradientType2() const
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DemonsRegistrationFunction" );
   }
  
  return drfp->GetUseGradientType2();
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::SetUseGradientType2(GradientType2 gtype) 
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to SymmetricDemonsRegistrationFunction" );
   }
  
  drfp->SetUseGradientType2(gtype);
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
typename DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GradientType3
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::GetUseGradientType3() const
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to DemonsRegistrationFunction" );
   }
  
  return drfp->GetUseGradientType3();
}

/*
 * 
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::SetUseGradientType3(GradientType3 gtype) 
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
 
  if( !drfp )
   {
   itkExceptionMacro( << 
     "Could not cast difference function to SymmetricDemonsRegistrationFunction" );
   }
  
  drfp->SetUseGradientType3(gtype);
}



template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::AllocateUpdateBuffer()
{
  // The update buffer looks just like the output.
  DeformationFieldPointer output = this->GetOutput();
  DeformationFieldPointer upbuf = this->GetUpdateBuffer();

  upbuf->SetLargestPossibleRegion(output->GetLargestPossibleRegion());
  upbuf->SetRequestedRegion(output->GetRequestedRegion());
  upbuf->SetBufferedRegion(output->GetBufferedRegion());
  upbuf->SetSpacing(output->GetSpacing());
  upbuf->SetOrigin(output->GetOrigin());
  upbuf->Allocate();
}

/*
* At the end of this function. 
* GetUpdateBuffer() now contains the new conj direction. 
* PrevConjField() now contains the new conj field (which will be old next time round)
* m_OldGradField now contains current grad  (which will be old next time round)
* GetUpdateField() remains unchanged.
*/

template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::ComputeNewConjField()
{
  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
    (this->GetDifferenceFunction().GetPointer());
  
  drfp->SetUpdateFieldModified();


//Compute New Conj Field  
  typedef typename itk::Image<double, DeformationFieldType::ImageDimension> MaskImageType;
  typename MaskImageType::Pointer DefFieldNormImage = MaskImageType::New();

  yeoInnerProductVectorFieldNewImage<DeformationFieldType, MaskImageType>(m_OldGradField, m_OldGradField,  DefFieldNormImage);
  double normSqOldGradient = yeoSumImage<MaskImageType>(DefFieldNormImage);

  double factor;
  if(normSqOldGradient != 0)
    {
    yeoInnerProductVectorFieldNewImage<DeformationFieldType, MaskImageType>(this->GetUpdateField(), this->GetUpdateField(), DefFieldNormImage);
    double normSqNewGradient = yeoSumImage<MaskImageType>(DefFieldNormImage);
    
    yeoInnerProductVectorFieldNewImage<DeformationFieldType, MaskImageType>(this->GetUpdateField(), m_OldGradField,  DefFieldNormImage);
    double innerProdOldNewGradient = yeoSumImage<MaskImageType>(DefFieldNormImage);
    
    factor = (normSqNewGradient - innerProdOldNewGradient)/normSqOldGradient;
    }
  else
    factor = 0;

  yeoLinearCombine2ImagesInPlace<DeformationFieldType>(m_OldGradField, m_PrevConjField, 1, factor);
  m_OldGradField->Modified();
  this->GetUpdateBuffer()->Graft(m_OldGradField);
  this->GetUpdateBuffer()->Modified(); //update buffer now contains new conj gradient

  //update old gradient
  yeoCopyImage<DeformationFieldType>(GetUpdateField(), m_OldGradField);

  //update Prev Conj gradient
  yeoCopyImage<DeformationFieldType>(this->GetUpdateBuffer(), m_PrevConjField);

  m_PrevConjField->Modified();
  m_OldGradField->Modified();

  //Now, I want to scale the update to something the same as implied
  //by m_normalizers

  yeoNormTensorImageNewImage<DeformationFieldType, MaskImageType>(this->GetUpdateBuffer(), DefFieldNormImage); 
  double maxValue = yeoMaxValue<MaskImageType>(DefFieldNormImage);
  
  typename DeformationFieldType::Pointer temp_field = DeformationFieldType::New();
  yeoMultiplyImageByConstantNewImage<DeformationFieldType>(this->GetUpdateBuffer(), 
							   temp_field, sqrt(drfp->GetNormalizer())/maxValue * STEP_SIZE);
  this->GetUpdateBuffer()->Graft(temp_field);
  this->GetUpdateBuffer()->Modified();

}

template <class TFixedImage, class TMovingImage, class TDeformationField>
unsigned int
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::ComputeNumberOfSquaringSteps()
{

 double maxnorm2 = 0.0;

 double minpixelspacing = this->GetUpdateBuffer()->GetSpacing()[0];
 for (unsigned int i = 1; i<itkGetStaticConstMacro(ImageDimension); ++i)
   {
   if ( this->GetUpdateBuffer()->GetSpacing()[i] < minpixelspacing )
     minpixelspacing = this->GetUpdateBuffer()->GetSpacing()[i];
   }
      
 typedef ImageRegionConstIterator<DeformationFieldType> InputConstIterator;
 InputConstIterator InputIt = InputConstIterator(this->GetUpdateBuffer(), this->GetUpdateBuffer()->GetLargestPossibleRegion());
 
 for (InputIt.GoToBegin(); !InputIt.IsAtEnd(); ++InputIt)
   {
   double norm2 = InputIt.Get().GetSquaredNorm();
   if (norm2>maxnorm2) maxnorm2=norm2;
   }

 maxnorm2 /= minpixelspacing;

 return(static_cast<unsigned int> (2.0 + 0.5 * vcl_log(maxnorm2)/vnl_math::ln2 ));

}


template <class TFixedImage, class TMovingImage, class TDeformationField>
double
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::ComputeSSD(const DeformationFieldPointer df)
{


  DemonsRegistrationFunctionType *drfp = 
    dynamic_cast<DemonsRegistrationFunctionType *>
    (this->GetDifferenceFunction().GetPointer());
  
  drfp->SetUpdateFieldModified();

  typename FixedImageType::ConstPointer fixedImage = this->GetFixedImage();
  typename MovingImageType::ConstPointer movingImage = this->GetMovingImage();
  
  typename MovingImageType::Pointer warpedMovingImage = MovingImageType::New();
  yeoWarpTensorImageNewImage<MovingImageType, DeformationFieldType>(movingImage, df, warpedMovingImage);

  
  typedef typename itk::Image<double, DeformationFieldType::ImageDimension> MaskImageType;

  //computing mask
  typename MaskImageType::Pointer fixedImage_fg = MaskImageType::New();
  yeoGetForegroundMask<MovingImageType, MaskImageType>(fixedImage, fixedImage_fg, itk::NumericTraits<typename MovingImageType::PixelType>::Zero, 0, 1);
  
  typename MaskImageType::Pointer movingImage_fg = MaskImageType::New();
  yeoGetForegroundMask<MovingImageType, MaskImageType>(warpedMovingImage, movingImage_fg, itk::NumericTraits<typename MovingImageType::PixelType>::Zero, 0, 1);
  
  yeoPowerCombine2ImagesInPlace<MaskImageType>(fixedImage_fg, movingImage_fg, 1, 1);
  
  //compute SSD
  yeoLinearCombine2ImagesInPlace<MovingImageType>(warpedMovingImage, fixedImage, 1.0, -1.0);
  yeoPowerCombine2TensorImagesInPlace<MovingImageType>(warpedMovingImage, warpedMovingImage, 1.0, 1.0);
  
  //apply mask
  yeoMultiplyTensorImageWithImageInPlace<MovingImageType, MaskImageType>(warpedMovingImage, fixedImage_fg);
    
  double numPixels = yeoSumImage<MaskImageType>(fixedImage_fg);
  double SSD;
  
  //normalized SSD
  if(drfp->GetUseAtlas())
    {
    yeoPowerCombine2TensorImagesInPlace<MovingImageType>(warpedMovingImage, drfp->GetVarImage(), 1.0, -1.0);
    }
    
  SSD = yeoSumTensorImage<MovingImageType>(warpedMovingImage)/numPixels;
  std::cout<<"SSD: "<<SSD<<std::endl;
  return SSD;
}

/*
 * Get the metric value from the difference function
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::ApplyUpdate(TimeStepType dt)
{
  time_t start;
  time (&start);


  this->GetUpdateBuffer()->Modified();

  if( this->GetUseGradientType3() == DemonsRegistrationFunctionType::Tom &&  
      this->GetUseGradientType2() != DemonsRegistrationFunctionType::NoRotation ) //if Tom's method (Gauss-Newton/ESM)
    {
  

    DemonsRegistrationFunctionType *drfp = 
      dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
    
    if( !drfp )
      {
      itkExceptionMacro( << 
			 "Could not cast difference function to DemonsRegistrationFunction" );
      }
    
    /*==========================================
* Solving the linear system
========================================*/
    doubleVectorType* m_Residual = drfp->GetResidual();
    SparseWriteMatrixType* m_Grad = drfp->GetGrad();
    
    gmm::iteration iter(0.005);
    iter.set_maxiter(100);
    //iter.set_noisy(1);
    
    long numCols = gmm::mat_ncols(*m_Grad);
      
// ====================================
//initialize X
//=====================================
    doubleVectorType* X = new doubleVectorType(numCols, 0.0);
    typedef typename itk::ImageRegionConstIterator< LongImageType> LongIteratorType;
    LongIteratorType longIt = LongIteratorType(drfp->GetForegroundRef(), drfp->GetForegroundRef()->GetLargestPossibleRegion());
    
    typedef typename itk::ImageRegionIterator< DeformationFieldType> dfIteratorType;
    dfIteratorType dfIt = dfIteratorType(this->GetUpdateBuffer(), this->GetUpdateBuffer()->GetLargestPossibleRegion());
    
    typename DeformationFieldType::PixelType displacement;
    for(longIt.GoToBegin(), dfIt.GoToBegin(); !longIt.IsAtEnd(); ++longIt, ++dfIt)
      {
      if(longIt.Get() >= 0)
	{
	displacement = dfIt.Get();

	(*X)[longIt.Get()*3 + 0] = double(displacement[0]); 
	(*X)[longIt.Get()*3 + 1] = double(displacement[1]);
	(*X)[longIt.Get()*3 + 2] = double(displacement[2]);
	
	dfIt.Set(displacement);
	}
      }

    gmm::clean(*m_Grad, 1E-8);


//Solving Ax = B
/*
    gmm::csc_matrix<double> m_GradRead;
    gmm::copy(*m_Grad, m_GradRead);
    
    time_t start;
    time (&start);
    
    gmm::least_squares_cg(m_GradRead, *X, *m_Residual, iter);   
    
    time_t final;
    final = time(&final);
    double dif;
    dif = difftime(final, start);
    
    delete m_Residual;
    delete m_Grad;
*/


//Solving AtAx = AtB
    
    SparseWriteMatrixType* AtA2 = new SparseWriteMatrixType(numCols, numCols);


    gmm::mult( gmm::transposed(*m_Grad), *m_Grad, *AtA2);

    gmm::csr_matrix<double> AtA;

    gmm::clean(*AtA2, 1E-8);

    gmm::copy(*AtA2, AtA);

    
    doubleVectorType* Atb = new doubleVectorType(numCols, 0.0);

    gmm::mult( gmm::transposed(*m_Grad), *m_Residual, *Atb);


    delete AtA2;


    delete m_Grad;
    delete m_Residual;
 
    gmm::identity_matrix P;

    
    gmm::bicgstab(AtA, *X, *Atb, P, iter);
    //gmm::cg(AtA, *X, *Atb, P, P, iter);
    //gmm::qmr(AtA, *X, *Atb, P, iter);
    //gmm::gmres(AtA, *X, *Atb, P, 50, iter);

    //std::cout<<"Iter: "<<iter.get_iteration()<<std::endl;
    //std::cout<<"Converge: "<<iter.converged()<<std::endl;
   
    delete Atb;
 
//=====================


    std::cout<<"Iter: "<<iter.get_iteration()<<", ";
    std::cout<<"Res: "<<iter.get_res()<<", ";
    std::cout<<"Rhsn: "<<iter.get_rhsnorm()<<", ";
    std::cout<<"Fraction: "<<iter.get_res()/iter.get_rhsnorm()<<std::endl;
 

/*===============================
* Now to get the update
================================*/
    
    typename DeformationFieldType::Pointer defField = DeformationFieldType::New();
    yeoCreateZeroImage<DeformationFieldType, DeformationFieldType>(this->GetUpdateBuffer(), defField);

//    typedef typename itk::ImageRegionConstIterator< LongImageType> LongIteratorType;
//    LongIteratorType longIt = LongIteratorType(drfp->GetForegroundRef(), drfp->GetForegroundRef()->GetLargestPossibleRegion());

//    typedef typename itk::ImageRegionIterator< DeformationFieldType> dfIteratorType;
    dfIt = dfIteratorType(defField, defField->GetLargestPossibleRegion());
    
//    typename DeformationFieldType::PixelType displacement;
    for(longIt.GoToBegin(), dfIt.GoToBegin(); !longIt.IsAtEnd(); ++longIt, ++dfIt)
      {
      if(longIt.Get() >= 0)
	{
	displacement[0] = (*X)[longIt.Get()*3 + 0]; 
	displacement[1] = (*X)[longIt.Get()*3 + 1];
	displacement[2] = (*X)[longIt.Get()*3 + 2];

	dfIt.Set(displacement);
	}
      }

    delete X;

/* ===========================================
*Replace updatebuffer with my buffer!!
========================================*/
    this->GetUpdateBuffer()->Graft(defField); 

    } //end symmetric update
  


  if(this->GetUseGradientType3() == DemonsRegistrationFunctionType::ConjugateGradient)
    {

    DemonsRegistrationFunctionType *drfp = 
      dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
    
    if( !drfp )
      {
      itkExceptionMacro( << 
			 "Could not cast difference function to DemonsRegistrationFunction" );
      }

    /*===========================
* now compute conjugate gradient
==================================*/
    ComputeNewConjField();

    if ( this->GetSmoothUpdateField() )
      {
#ifdef DEBUG
      std::cout<<"Smooth Update Field"<<std::endl;
#endif
      this->SmoothUpdateField();
      }
    
    ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(this->GetUpdateBuffer(), drfp->GetFixedImage(), "", "");
    //std::cout<<std::endl;

    // use time step if necessary
    if ( fabs(dt - 1.0)>1.0e-4 )
      {
      std::cout<<"Using timestep: "<<dt<<std::endl;
      m_Multiplier->SetConstant( dt );
      m_Multiplier->SetInput( this->GetUpdateBuffer() );
      m_Multiplier->GraftOutput( this->GetUpdateBuffer() );
      // in place update
      //m_Multiplier->UpdateLargestPossibleRegion();
      m_Multiplier->Update();
      // graft output back to this->GetUpdateBuffer()
      this->GetUpdateBuffer()->Graft( m_Multiplier->GetOutput() );
      }

    //Compute Number of Times to do exponential
    unsigned int numiter = ComputeNumberOfSquaringSteps();
   

    //Line Search!!
    
    //Set Up exponentiation
    typename DeformationFieldType::Pointer df = DeformationFieldType::New();
    typename DeformationFieldType::Pointer df2 = DeformationFieldType::New();
    typename DeformationFieldType::Pointer bestDF = DeformationFieldType::New();
    yeoMultiplyImageByConstantNewImage<DeformationFieldType>(this->GetUpdateBuffer(), df, 
							     1.0/static_cast<float>(1<<numiter) );


    VectorWarperPointer m_Warper2 = VectorWarperType::New();
    m_Warper2 = VectorWarperType::New();
    FieldInterpolatorPointer VectorInterpolator = FieldInterpolatorType::New();
    m_Warper2->SetInterpolator(VectorInterpolator);
       
    m_Warper2->SetOutputSpacing(df->GetSpacing());
    m_Warper2->SetOutputOrigin(df->GetOrigin());
    
    //Set up Warp composition
    m_Warper->SetOutputSpacing( this->GetUpdateBuffer()->GetSpacing() );
    m_Warper->SetOutputOrigin( this->GetUpdateBuffer()->GetOrigin() );
    m_Warper->SetInterpolator(VectorInterpolator);
  
    //Do the exponentiation
    double SSD = 0;
    double bestSSD = 0;
  
    std::cout<<"Current SSD: "<<m_CurrentSSD<<std::endl;
    for (unsigned int i=0; i<numiter; i++)
      {

      //First Compute 1 exponentiation
      m_Warper2->SetInput(df);
      m_Warper2->SetDeformationField(df);      
      m_Warper2->UpdateLargestPossibleRegion();
     
      yeoLinearCombine2ImagesInPlace<DeformationFieldType>(df, m_Warper2->GetOutput(), 1.0, 1.0);            
      df->Modified();
      
      //Compose With Current Field
      m_Warper->SetInput(this->GetOutput());
      m_Warper->SetDeformationField( df);
      m_Warper->UpdateLargestPossibleRegion();
      
      yeoLinearCombine2ImagesNewImage<DeformationFieldType>(df, m_Warper->GetOutput(), df2, 1.0, 1.0);
      df2->Modified();
       
      if(i >= 1)
	{
	
	SSD = ComputeSSD(df2);	
		
	if(i == 1 || SSD <= bestSSD)
	  {
	  std::cout<<"Improvement"<<std::endl;
	  bestSSD = SSD;
	  yeoCopyImage<DeformationFieldType>(df2, bestDF);
	  }
	}
      }

    std::cout<<"best SSD: "<< bestSSD<<std::endl;
    std::cout<<"SSD: "<< SSD<<std::endl;
  

    //whether improve or not. Will graft...
    //if(bestSSD < m_CurrentSSD)
      this->GraftOutput(bestDF);
       
    this->SetRMSChange( drfp->GetRMSChange() );
           
    m_CurrentSSD = bestSSD;
    
    /*
   * Smooth the deformation field
   */
    if ( this->GetSmoothDeformationField() )
      {
      this->SmoothDeformationField();
      }
    ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(this->GetDeformationField(), drfp->GetFixedImage(), "", "");
    std::cout<<std::endl;

    }//end Conjugate Gradient
  else
    {

    // If we smooth the update buffer before applying it, then the are
    // approximating a viscuous problem as opposed to an elastic problem
    if ( this->GetSmoothUpdateField() )
      {
#ifdef DEBUG
      std::cout<<"Smooth Update Field"<<std::endl;
#endif
      this->SmoothUpdateField();
      }
  
    DemonsRegistrationFunctionType *drfp = 
      dynamic_cast<DemonsRegistrationFunctionType *>
      (this->GetDifferenceFunction().GetPointer());
  
    if( !drfp )
      {
      itkExceptionMacro( << 
			 "Could not cast difference function to DemonsRegistrationFunction" );
      }
  
    std::cout<<"----------------------------"<<std::endl;
    ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(this->GetUpdateBuffer(), drfp->GetFixedImage(), "", "");
  
    // use time step if necessary
    if ( fabs(dt - 1.0)>1.0e-4 )
      {
      std::cout<<"Using timestep: "<<dt<<std::endl;
      m_Multiplier->SetConstant( dt );
      m_Multiplier->SetInput( this->GetUpdateBuffer() );
      m_Multiplier->GraftOutput( this->GetUpdateBuffer() );
      // in place update
      //m_Multiplier->UpdateLargestPossibleRegion();
      m_Multiplier->Update();
      // graft output back to this->GetUpdateBuffer()
      this->GetUpdateBuffer()->Graft( m_Multiplier->GetOutput() );
      }
  
    // compute the exponential
    m_Exponentiator->SetInput( this->GetUpdateBuffer() );
    const double imposedMaxUpStep = this->GetMaximumUpdateStepLength();
    if ( imposedMaxUpStep > 0.0 )
      {
      // max(norm(Phi))/2^N < 0.25*pixelspacing
      const double numiterfloat = 2.0 + vcl_log(imposedMaxUpStep)/vnl_math::ln2;
      unsigned int numiter = 0;
      if ( numiterfloat > 0.0 )
	numiter = static_cast<unsigned int>( 1.0 + numiterfloat );
    
      m_Exponentiator->AutomaticNumberOfIterationsOff();
      m_Exponentiator->SetMaximumNumberOfIterations( numiter );
      }
    else
      {
      m_Exponentiator->AutomaticNumberOfIterationsOn();
      m_Exponentiator->SetMaximumNumberOfIterations( 2000u ); // just a high value
      }
  
    // compose the vector fields
    m_Warper->SetOutputSpacing( this->GetUpdateBuffer()->GetSpacing() );
    m_Warper->SetOutputOrigin( this->GetUpdateBuffer()->GetOrigin() );
  
  
    m_Warper->SetInput( this->GetOutput() );
    m_Warper->SetDeformationField( m_Exponentiator->GetOutput() );
  
    m_Adder->SetInput1( m_Warper->GetOutput() );
    m_Adder->SetInput2( m_Exponentiator->GetOutput() );
  
    //m_Adder->UpdateLargestPossibleRegion();
    m_Adder->GetOutput()->SetRequestedRegion( this->GetOutput()->GetRequestedRegion() );
    m_Adder->Update();
  
    //std::cout<<"out buff spac: "<<this->GetOutput()->GetSpacing()<<std::endl;
    //std::cout<<"up buff spac: "<<this->GetUpdateBuffer()->GetSpacing()<<std::endl;
    //std::cout<<"exp out spac: "<<m_Exponentiator->GetOutput()->GetSpacing()<<std::endl;
    //std::cout<<"warp out spac: "<<m_Warper->GetOutput()->GetSpacing()<<std::endl;
    //std::cout<<"add out spac: "<<m_Adder->GetOutput()->GetSpacing()<<std::endl;
  
    // Region passing stuff
    this->GraftOutput( m_Adder->GetOutput() );
  
  
    this->SetRMSChange( drfp->GetRMSChange() );
  
    /*
   * Smooth the deformation field
   */
    if ( this->GetSmoothDeformationField() )
      {
#ifdef DEBUG
      std::cout<<"Smooth Deformation Field"<<std::endl;
#endif
      this->SmoothDeformationField();
    
      //double var[DeformationFieldType::ImageDimension];
      //for (unsigned int i=0; i<DeformationFieldType::ImageDimension; ++i)
      //   var[i] = vnl_math_sqr( this->GetUpdateFieldStandardDeviations()[i] );
    
      //typedef SmoothingRecursiveGaussianImageFilter<DeformationFieldType,DeformationFieldType> GaussianFilterType;
      //typename GaussianFilterType::Pointer smoother = GaussianFilterType::New();
      //smoother->SetInput( this->GetOutput() );
      //smoother->SetVariance( var );
      //smoother->SetSigma( this->GetUpdateFieldStandardDeviations()[0] );
      //smoother->Update();
    
      //this->GraftOutput( smoother->GetOutput() );
      }
    ComputeStatisticsDeformationField<DeformationFieldType, FixedImageType>(this->GetDeformationField(), drfp->GetFixedImage(), "", "");
    std::cout<<std::endl;
    }


  time_t final;
  final = time(&final);
  double dif;
  dif = difftime(final, start);
  std::cout<<"Time to Apply Update: "<<dif<<" seconds"<<std::endl;
}

template <class TFixedImage, class TMovingImage, class TDeformationField>
void
DiffeomorphicDemonsRegistrationFilterTensor<TFixedImage,TMovingImage,TDeformationField>
::PrintSelf(std::ostream& os, Indent indent) const
{ 
  Superclass::PrintSelf( os, indent );
  os << indent << "Intensity difference threshold: " <<
    this->GetIntensityDifferenceThreshold() << std::endl;
}


} // end namespace itk

#endif
