#ifndef _itkESMDemonsRegistrationFunctionAtlas_h_
#define _itkESMDemonsRegistrationFunctionAtlas_h_

#include "itkESMDemonsRegistrationFunction.h"
#include "itkPoint.h"
#include "itkCovariantVector.h"
#include "itkInterpolateImageFunction.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkCentralDifferenceImageFunction.h"
#include "itkWarpImageFilter.h"

#include "Debug.h"
namespace itk {

/**
 * \class ESMDemonsRegistrationFunctionAtlas
 *
 * This class is a subclass of itkESMDemonsRegistrationFunction.
 * It incorporates the use of an atlas by setting the "noise
 * parameter" to be estimated in the training set. 
 *
 * This class is contributed by Thomas Yeo, CSAIL, MIT
 */

template<class TFixedImage, class TMovingImage, class TDeformationField>
class ITK_EXPORT ESMDemonsRegistrationFunctionAtlas : 
    public ESMDemonsRegistrationFunction< TFixedImage,
                                              TMovingImage, TDeformationField>
{

public:
  /** Standard class typedefs. */
  typedef ESMDemonsRegistrationFunctionAtlas    Self;
  typedef ESMDemonsRegistrationFunction< TFixedImage,
                                         TMovingImage, TDeformationField >    Superclass;
  typedef SmartPointer<Self> Pointer;
  typedef SmartPointer<const Self> ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);
  
  /** Run-time type information (and related methods). */
  itkTypeMacro( ESMDemonsRegistrationFunctionAtlas, 
                ESMDemonsRegistrationFunction );


  // MovingImage image type. 
  typedef typename Superclass::MovingImageType     MovingImageType;
  typedef typename Superclass::MovingImagePointer  MovingImagePointer;

  // FixedImage image type.
  typedef typename Superclass::FixedImageType     FixedImageType;
  typedef typename Superclass::FixedImagePointer  FixedImagePointer;
  typedef typename Superclass::IndexType      IndexType;
  typedef typename Superclass::SizeType       SizeType;
  typedef typename Superclass::SpacingType    SpacingType;
  
  // Deformation field type.
  typedef typename Superclass::DeformationFieldType    DeformationFieldType;
  typedef typename Superclass::DeformationFieldTypePointer   
  DeformationFieldTypePointer;

  // VarImage Type.
  typedef TFixedImage VarImageType;
  typedef typename VarImageType::Pointer VarImagePointer;
  typedef typename VarImageType::ConstPointer VarImageConstPointer;
  // Inherit some enums from the superclass. 
  itkStaticConstMacro(ImageDimension, unsigned int,Superclass::ImageDimension);


  // Inherit some enums from the superclass.
  typedef typename Superclass::PixelType     PixelType;
  typedef typename Superclass::RadiusType    RadiusType;
  typedef typename Superclass::NeighborhoodType    NeighborhoodType;
  typedef typename Superclass::FloatOffsetType  FloatOffsetType;
  typedef typename Superclass::TimeStepType TimeStepType;

  // Interpolator type. 
  typedef typename Superclass::PointType PointType;
  typedef typename Superclass::InterpolatorPointer InterpolatorPointer;

  // Warper type 
  typedef typename Superclass::WarperPointer WarperPointer;


  // Covariant vector type.
  typedef typename Superclass::CovariantVectorType CovariantVectorType;
  
  // Gradient calculator types. 
  typedef typename Superclass::GradientCalculatorType   GradientCalculatorType;
  typedef typename Superclass::GradientCalculatorPointer   GradientCalculatorPointer;



  /** This method is called by a finite difference solver image filter at
   * each pixel that does not lie on a data set boundary */
 virtual PixelType ComputeUpdate(const NeighborhoodType &neighborhood,
                                    void *globalData,
                                   const FloatOffsetType &offset = FloatOffsetType(0.0));
  
/* Whether to use atlas */
  virtual void SetUseAtlas(bool useAtlas){
    m_useAtlas = useAtlas;};
  
  virtual bool GetUseAtlas(void){
    return m_useAtlas;};


/** Set the Var image. */
  virtual void SetVarImage( VarImageConstPointer ptr )
  { m_VarImage = ptr; };
  
  /** Get the Var image. */
  virtual VarImageConstPointer GetVarImage(void) const
  { return m_VarImage; };
  
  virtual void PrintMSE(std::string filename) const;
  virtual void PrintWMSE(std::string filename1, std::string filename2, std::string filename3) const;
  virtual void UpdateVarianceImage(void);
  

protected:
  ESMDemonsRegistrationFunctionAtlas();
  ~ESMDemonsRegistrationFunctionAtlas() {}
  void PrintSelf(std::ostream& os, Indent indent) const;

  /** FixedImage image neighborhood iterator type. */
  typedef ConstNeighborhoodIterator<FixedImageType> FixedImageNeighborhoodIteratorType;

  typedef typename Superclass::GlobalDataStruct GlobalDataStruct;

  bool m_useAtlas;
  /** The Var image. */
  VarImageConstPointer m_VarImage;

  VarImagePointer m_CacheVarianceImage;

  /**
     bad code to be able to link with ITK 3.8 DiffeomorphicDemons updates
  */
  GradientCalculatorPointer GetFixedImageGradientCalculator (void);
  WarperPointer GetMovingImageWarper (void);
  
    


private:
  ESMDemonsRegistrationFunctionAtlas(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented


};


} // end namespace itk




#ifndef ITK_MANUAL_INSTANTIATION
#include "itkESMDemonsRegistrationFunctionAtlas.txx"
#endif

#endif
