#ifndef _itkESMDemonsRegistrationFunctionTensor_h_
#define _itkESMDemonsRegistrationFunctionTensor_h_

#include "itkPDEDeformableRegistrationFunction.h"
#include "itkPoint.h"
#include "itkWarpTensorImageFilter.h"
#include "itkWarpTensorImageWOReorientationFilter.h"
#include "itkMacro.h"
#include "itkMatrix.h"
#include "itkTensorCentralDifferenceImageFunction.h"
#include "itkWarpJacobianFilter.h"
#include "gmm/gmm.h"

#include <vector>

//#define DEBUG
#include "Debug.h"
namespace itk {

/**
 * \class ESMDemonsRegistrationFunction
 *
 * This class provides a substantially faster implementation of the
 * symmetric demons registration force. Speed is improved by keeping
 * a deformed copy of the moving image for gradient evaluation.
 *
 * The moving image should not be saturated. We indeed use
 * NumericTraits<MovingPixelType>::Max() as a special value.
 *
 * This class was contributed by Tom Vercauteren, INRIA
 *
 * \sa SymmetricForcesDemonsRegistrationFunction
 * \sa SymmetricForcesDemonsRegistrationFilter
 * \sa DemonsRegistrationFilter
 * \sa DemonsRegistrationFunction
 * \ingroup FiniteDifferenceFunctions
 *
 */
template<class TFixedImage, class TMovingImage, class TDeformationField>
class ITK_EXPORT ESMDemonsRegistrationFunctionTensor : 
    public PDEDeformableRegistrationFunction< TFixedImage,
                                              TMovingImage, TDeformationField>
{
public:
  /** Standard class typedefs. */
  typedef ESMDemonsRegistrationFunctionTensor    Self;
  typedef PDEDeformableRegistrationFunction< TFixedImage,
                                             TMovingImage, TDeformationField >    Superclass;
  typedef SmartPointer<Self> Pointer;
  typedef SmartPointer<const Self> ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);

  /** Run-time type information (and related methods). */
  itkTypeMacro( ESMDemonsRegistrationFunctionTensor, 
                PDEDeformableRegistrationFunction );

  /** MovingImage image type. */
  typedef typename Superclass::MovingImageType     MovingImageType;
  typedef typename Superclass::MovingImagePointer  MovingImagePointer;

  /** FixedImage image type. */
  typedef typename Superclass::FixedImageType     FixedImageType;
  typedef typename Superclass::FixedImagePointer  FixedImagePointer;
  typedef typename FixedImageType::IndexType      IndexType;
  typedef typename FixedImageType::SizeType       SizeType;
  typedef typename FixedImageType::SpacingType    SpacingType;
  
  /** Deformation field type. */
  typedef typename Superclass::DeformationFieldType    DeformationFieldType;
  typedef typename Superclass::DeformationFieldTypePointer   
  DeformationFieldTypePointer;

  // VarImage Type.
  typedef TFixedImage VarImageType;
  typedef typename VarImageType::Pointer VarImagePointer;
  typedef typename VarImageType::ConstPointer VarImageConstPointer;

  /** Inherit some enums from the superclass. */
  itkStaticConstMacro(ImageDimension, unsigned int,Superclass::ImageDimension);

  /** Inherit some enums from the superclass. */
  //typedef typename Superclass::PixelType     PixelType;
  typedef typename Superclass::PixelType     PixelType;
  typedef typename Superclass::RadiusType    RadiusType;
  typedef typename Superclass::NeighborhoodType    NeighborhoodType;
  typedef typename Superclass::FloatOffsetType  FloatOffsetType;
  typedef typename Superclass::TimeStepType TimeStepType;

  /** Interpolator type. */
/*
  typedef double CoordRepType;
  typedef InterpolateImageFunction<MovingImageType,CoordRepType> InterpolatorType;
  typedef typename InterpolatorType::Pointer         InterpolatorPointer;
  typedef typename InterpolatorType::PointType       PointType;
  typedef LinearInterpolateImageFunction<MovingImageType,CoordRepType>
  DefaultInterpolatorType;
*/
  typedef typename FixedImageType::PointType PointType;

  /** Warper type */
  typedef WarpTensorImageFilter<MovingImageType,MovingImageType,DeformationFieldType> WarperType;
  typedef typename WarperType::Pointer WarperPointer;

  /** Warper type */
  typedef WarpTensorImageWOReorientationFilter<MovingImageType,MovingImageType,DeformationFieldType> WarperWOReorientationType;
  typedef typename WarperWOReorientationType::Pointer WarperWOReorientationPointer;

  /** Covariant vector type. */
  typedef CovariantVector<double,itkGetStaticConstMacro(ImageDimension)> CovariantVectorType;

  /** Gradient calculator types. */
  typedef TensorCentralDifferenceImageFunction<FixedImageType, float> GradientCalculatorType;
  typedef typename GradientCalculatorType::Pointer   GradientCalculatorPointer;


  typedef typename FixedImageType::PixelType TensorPixelType;
  typedef  Matrix<double, TensorPixelType::Dimension*(TensorPixelType::Dimension+1)/2,
		  itkGetStaticConstMacro(ImageDimension)> TensorGradientType;
		   

  //typedef std::vector<TensorGradientType> TensorGradientVectorType;
  //typedef typename itk::Image<TensorGradientVectorType, FixedImageType::ImageDimension> TensorGradientVectorImageType;

  //typedef itk::Tensor<double, TensorPixelType::Dimension> doubleTensor; 
  //typedef std::vector<doubleTensor> DoubleTensorVectorType;
  //typedef typename itk::Image<DoubleTensorVectorType, FixedImageType::ImageDimension> DoubleTensorVectorImageType;


  typedef typename itk::Image<long, FixedImageType::ImageDimension> LongImageType;

  typedef typename gmm::wsvector<double> SparseWriteVectorType;
  typedef typename gmm::row_matrix< SparseWriteVectorType > SparseWriteMatrixType;
  typedef typename std::vector<double> doubleVectorType;


  /** Set the moving image interpolator. */
/*  void SetMovingImageInterpolator( InterpolatorType * ptr )
  { m_MovingImageInterpolator = ptr; m_MovingImageWarper->SetInterpolator( ptr ); }
  */
  

  /** Get the moving image interpolator. */
/*  InterpolatorType * GetMovingImageInterpolator(void)
  { return m_MovingImageInterpolator; }
  */

  /** This class uses a constant timestep of 1. */
  virtual TimeStepType ComputeGlobalTimeStep(void * itkNotUsed(GlobalData)) const
  { return m_TimeStep; }

  /** Return a pointer to a global data structure that is passed to
   * this object from the solver at each calculation.  */
  virtual void *GetGlobalDataPointer() const
  {
    GlobalDataStruct *global = new GlobalDataStruct();
    global->m_SumOfSquaredDifference  = 0.0;
    global->m_NumberOfPixelsProcessed = 0L;
    global->m_SumOfSquaredChange      = 0;
    global->m_NormalizedSumOfSquaredDifference = 0.0;
    return global;
  }

  /** Release memory for global data structure. */
  virtual void ReleaseGlobalDataPointer( void *GlobalData ) const;

  /** Set the object's state before each iteration. */
  virtual void InitializeIteration();

  /** This method is called by a finite difference solver image filter at
   * each pixel that does not lie on a data set boundary */
  virtual PixelType  ComputeUpdate(const NeighborhoodType &neighborhood,
                                   void *globalData,
                                   const FloatOffsetType &offset = FloatOffsetType(0.0));


/* Whether to use atlas */
  virtual void SetUseAtlas(bool useAtlas){
    m_useAtlas = useAtlas;};
  
  virtual bool GetUseAtlas(void){
    return m_useAtlas;};
  
  
/** Set the Var image. */
  virtual void SetVarImage( VarImageConstPointer ptr )
  { m_VarImage = ptr; };

 /** Get the Var image. */
  virtual VarImageConstPointer GetVarImage(void) const
  { return m_VarImage; };

   /** Get the metric value. The metric value is the mean square difference 
   * in intensity between the fixed image and transforming moving image 
   * computed over the the overlapping region between the two images. */
  virtual double GetMetric() const
    { return m_Metric; }

  virtual double GetNormalizedMetric() const
    { return m_NormalizedMetric; }

  virtual unsigned long GetNumberOfPixelsProcessed() const
    { return m_NumberOfPixelsProcessed; }

  /** Get the rms change in deformation field. */
  virtual const double &GetRMSChange() const
    { return m_RMSChange; }

  /** Set/Get the threshold below which the absolute difference of
   * intensity yields a match. When the intensities match between a
   * moving and fixed image pixel, the update vector (for that
   * iteration) will be the zero vector. Default is 0.001. */
  virtual void SetIntensityDifferenceThreshold(double);
  virtual double GetIntensityDifferenceThreshold() const;

  /** Set/Get the maximum update step length. In Thirion this is 0.5.
   *  Setting it to 0 implies no restriction (beware of numerical
   *  instability in this case. */
  virtual void SetMaximumUpdateStepLength(double sm){
     this->m_MaximumUpdateStepLength =sm;};
  virtual double GetMaximumUpdateStepLength() const {
     return this->m_MaximumUpdateStepLength;};

  enum GradientType {
    Symmetric=0,
    Fixed,
    Moving
  };

  enum GradientType2 {
    Rotation=0,
    NoRotation,
    ApproxRotation
  };

 enum GradientType3 {
    Tom=0,
    Xavier,
    ConjugateGradient
  };

  virtual void SetUseGradientType( GradientType gtype )
    { m_UseGradientType = gtype; }
  virtual GradientType GetUseGradientType() const
    { return m_UseGradientType; }

  virtual void SetUseGradientType2( GradientType2 gtype )
    { m_UseGradientType2 = gtype; }
  virtual GradientType2 GetUseGradientType2() const
    { return m_UseGradientType2; }

  virtual void SetUseGradientType3( GradientType3 gtype )
    { m_UseGradientType3 = gtype; }
  virtual GradientType3 GetUseGradientType3() const
    { return m_UseGradientType3; }

  virtual DeformationFieldTypePointer GetUpdateField() const
    { return m_UpdateField; }
  

  virtual SpacingType GetFixedImageSpacing() const {
    return this->m_FixedImageSpacing;};

  virtual PointType GetFixedImageOrigin() const {
    return this->m_FixedImageOrigin;};

  virtual double GetNormalizer() const {
    return this->m_Normalizer;};

  virtual GradientCalculatorPointer GetFixedImageGradientCalculator() const {
    return this->m_FixedImageGradientCalculator;};


  virtual  SparseWriteMatrixType* GetGrad() {
    return m_Grad;};

  virtual  doubleVectorType* GetResidual() {
    return m_Residual;};

  virtual LongImageType* GetForegroundRef() const {
    return m_ForegroundRef;};


  virtual void SetUpdateFieldModified() const{
    m_UpdateField->Modified();
  }



  virtual void PrintMSE(std::string filename) const;
  virtual void PrintWMSE(std::string filename1, std::string filename2, std::string filename3) const;
  virtual void UpdateVarianceImage(void);
  

protected:
  ESMDemonsRegistrationFunctionTensor();
  ~ESMDemonsRegistrationFunctionTensor() {}
  void PrintSelf(std::ostream& os, Indent indent) const;

  /** FixedImage image neighborhood iterator type. */
  typedef ConstNeighborhoodIterator<FixedImageType> FixedImageNeighborhoodIteratorType;

  /** A global data type for this class of equation. Used to store
   * iterators for the fixed image. */
  struct GlobalDataStruct
  {
    double          m_SumOfSquaredDifference;
    unsigned long   m_NumberOfPixelsProcessed;
    double          m_SumOfSquaredChange;
    double m_NormalizedSumOfSquaredDifference;
  };

  bool m_useAtlas;
  /** The Var image. */
  VarImageConstPointer                   m_VarImage;
  
  VarImagePointer m_CacheVarianceImage;


#ifdef DEBUG
  void PrintRadius1FixedImageNeighborhood(IndexType index);
#endif

  /** Cache fixed image information. */
  SpacingType                     m_FixedImageSpacing;
  PointType                       m_FixedImageOrigin;
  double                          m_Normalizer;

  /** Function to compute derivatives of the fixed image. */
  GradientCalculatorPointer       m_FixedImageGradientCalculator;

  GradientType                    m_UseGradientType;
  GradientType2                   m_UseGradientType2;
  GradientType3                   m_UseGradientType3;
  
  /** Function to interpolate the moving image. */
//  InterpolatorPointer             m_MovingImageInterpolator;

  /** Filter to warp moving image for fast gradient computation. */
  WarperPointer                   m_MovingImageWarper;
  WarperWOReorientationPointer    m_MovingImageWOReorientationWarper;

  /*WarpJacobianImage*/
  typedef Matrix<double, Superclass::ImageDimension, Superclass::ImageDimension> JacobianType;
  typedef Image<JacobianType, Superclass::ImageDimension>             JacobianImageType;
  typedef typename JacobianImageType::Pointer             JacobianPointer;
  typedef WarpJacobianFilter<TDeformationField, JacobianImageType>  JacobianFilterType;
  
  typedef itk::Tensor<double, TensorPixelType::Dimension> doubleTensor; 
  typedef itk::Vector<double, PixelType::Dimension>    doubleUpdateType;
  typedef itk::Matrix<double, PixelType::Dimension, PixelType::Dimension> HessianType;

  /** The global timestep. */
  TimeStepType                    m_TimeStep;

  /** Threshold below which the denominator term is considered zero. */
  double                          m_DenominatorThreshold;

  /** Threshold below which two intensity value are assumed to match. */
  double                          m_IntensityDifferenceThreshold;

  /** Maximum update step length in pixels (default is 0.5 as in Thirion). */
  double                          m_MaximumUpdateStepLength;

  /** The metric value is the mean square difference in intensity between
   * the fixed image and transforming moving image computed over the 
   * the overlapping region between the two images. */
  mutable double                  m_Metric;
  mutable double                  m_SumOfSquaredDifference;
  mutable unsigned long           m_NumberOfPixelsProcessed;
  mutable double                  m_RMSChange;
  mutable double                  m_SumOfSquaredChange;
  mutable double                  m_NormalizedSumOfSquaredDifference;
  mutable double                  m_NormalizedMetric;

  /** Mutex lock to protect modification to metric. */
  mutable SimpleFastMutexLock     m_MetricCalculationLock;
  mutable SimpleFastMutexLock     m_GradLock;
  mutable SimpleFastMutexLock     m_ResidualLock;
  mutable SimpleFastMutexLock     m_UpdateFieldLock;

private:
  ESMDemonsRegistrationFunctionTensor(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  //typename TensorGradientVectorImageType::Pointer m_SelfGradientImage;
  //typename DoubleTensorVectorImageType::Pointer m_StdWeightedResidualImage;
  typename LongImageType::Pointer m_ForegroundRef;
  unsigned int m_Iteration;
  //note that in the background pixels, it will be a vector of length
  //0. Other it should be a vector of length 7.

  doubleVectorType *m_Residual; //b
  SparseWriteMatrixType *m_Grad; //A
  typename JacobianImageType::Pointer m_Jacobian;
 
  //create my own update buffer due to dependenceis across neighbors
  typename DeformationFieldType::Pointer m_UpdateField;
  typename FixedImageType::Pointer m_ReorientedFixedImage;

  virtual void ComputeReorientedFixedImage(bool forwardBool);

};


} // end namespace itk

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkESMDemonsRegistrationFunctionTensor.txx"
#endif

#endif
