#ifndef _itkESMDemonsRegistrationFunctionTensor_txx_
#define _itkESMDemonsRegistrationFunctionTensor_txx_

#include "itkESMDemonsRegistrationFunctionTensor.h"
#include "itkExceptionObject.h"
#include "vnl/vnl_math.h"

#include "UsefulTensorFunctions.h"
#include "itkTensor.h"
#include "itkMatrix.h"
#include <math.h>
#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkNumericTraitsTensorPixel2.h"
#include "vnl/vnl_vector.h"
#include "vnl/vnl_matrix.h"
#include "vnl/algo/vnl_cholesky.h" 
#include "itkAffineTensorTransform.h"
#include "gmm/gmm.h"


//#define INI_ON 0 //whether to initialize for m_UseGradientType = Symmetric
#define MIN_VAR 0.001

//#define DEBUG
#include "Debug.h"
namespace itk {

#ifdef DEBUG

template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::PrintRadius1FixedImageNeighborhood(const IndexType index)
{
  //assume dim = 3
  
  std::cout<<"Spacing: "<<this->GetFixedImage()->GetSpacing()<<std::endl;

  IndexType newIndex;

  newIndex = index;
  newIndex[0] -= 1;
  std::cout<<"Pixel Value at "<< -1 << ", " << 0 << ", "<< 0<< ": "<< this->GetFixedImage()->GetPixel(newIndex) <<std::endl;

  newIndex = index;
  newIndex[0] += 1;
  std::cout<<"Pixel Value at "<< 1 << ", " << 0 << ", "<< 0<< ": "<< this->GetFixedImage()->GetPixel(newIndex) <<std::endl;
   
  newIndex = index;
  newIndex[1] -= 1;
  std::cout<<"Pixel Value at "<< 0 << ", " << -1 << ", "<< 0<< ": "<< this->GetFixedImage()->GetPixel(newIndex) <<std::endl;

  newIndex = index;
  newIndex[1] += 1;
  std::cout<<"Pixel Value at "<< 0 << ", " << 1 << ", "<< 0<< ": "<< this->GetFixedImage()->GetPixel(newIndex) <<std::endl;

  newIndex = index;
  newIndex[2] -= 1;
  std::cout<<"Pixel Value at "<< 0 << ", " << 0 << ", "<< -1<< ": "<< this->GetFixedImage()->GetPixel(newIndex) <<std::endl;

  newIndex = index;
  newIndex[2] += 1;
  std::cout<<"Pixel Value at "<< 0 << ", " << 0 << ", "<< 1<< ": "<< this->GetFixedImage()->GetPixel(newIndex) <<std::endl;

  std::cout<<"***********************************"<<std::endl;
}
#endif


/* Usage: 
*          update = yeoSumTensorGradient<TMatrix, TUpdateType>(TMatrix m);
*/
template <class TMatrix, class TUpdateType> TUpdateType yeoSumTensorGradient(const TMatrix m)
{
  //note the weights differently, because off-diagonal elements need
  //to be weighted more!!

  //Also note that we assume tensor dimension is 3 x 3

  TUpdateType update;

  unsigned int dims = TUpdateType::Dimension;
  unsigned int numRows = TMatrix::RowDimensions;
  unsigned int numCols = TMatrix::ColumnDimensions;
  
  if(numRows != 6)
    {
    std::cerr<<"Here we assume Tensor is 3 x 3. You need to change this function!"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  if(numCols!=dims)
    {
    std::cerr<<"numRows of tensor gradient not equal to length of update"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
  //initialize to 0
  update.Fill(0.0);  

  for(unsigned i = 0; i < dims; i++)
    {
    update[i] += m[0][i];
    update[i] += m[2][i];
    update[i] += m[5][i];
    
    update[i] += 2.0*m[1][i];
    update[i] += 2.0*m[3][i];
    update[i] += 2.0*m[4][i];
    }

  return update;
}

/* Usage: 
*          sum = yeoSumUpdate<TUpdateType>(TUpdateType update);
*/
template <class TUpdateType> double yeoSumUpdate(const TUpdateType update)
{
  unsigned int dims = TUpdateType::Dimension;

  double sum = 0;

  for(unsigned int i = 0; i < dims; i++)
    sum += update[i];

  return sum;
}

/* Usage: 
*          tensor = yeoSqrtTensor<TensorType>(t);
*/
template <class TensorType> TensorType yeoSqrtTensor(const TensorType t)
{
  TensorType tensor;

  for(unsigned int i = 0; i < TensorType::NDegreesOfFreedom; i++)
      tensor.SetNthComponent(i, sqrt(t.GetNthComponent(i)));
  
  return tensor;
}

/* Usage: 
*          yeoInsertTensorIntoVector<doubleVectorType, TensorType>(m_Residual, tensor, starting_position);
*/
template <class doubleVectorType, class TensorType> void yeoInsertTensorIntoVector(doubleVectorType* m_Residual, TensorType t, long sp)
{
    
  (*m_Residual)[sp+0] = t.GetNthComponent(0);
  (*m_Residual)[sp+1] = sqrt(2.0)*t.GetNthComponent(1);
  (*m_Residual)[sp+2] = t.GetNthComponent(2);
  (*m_Residual)[sp+3] = sqrt(2.0)*t.GetNthComponent(3);
  (*m_Residual)[sp+4] = sqrt(2.0)*t.GetNthComponent(4);
  (*m_Residual)[sp+5] = t.GetNthComponent(5);
}

/* Usage: 
*          yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>(m_Grad, grad, 
*                                                                                             starting_row, starting col);
*/
template <class SparseWriteMatrixType, class TensorGradientType> void yeoInsertTensorGradientIntoSparseMatrix
                                                                 (SparseWriteMatrixType* m_Grad, TensorGradientType grad, long sr, long sc)
{

  for(int i = 0; i < 3; i++)
    {
    (*m_Grad)(sr+0, sc+i) = grad[0][i];
    (*m_Grad)(sr+1, sc+i) = sqrt(2.0)*grad[1][i];
    (*m_Grad)(sr+2, sc+i) = grad[2][i];
    (*m_Grad)(sr+3, sc+i) = sqrt(2.0)*grad[3][i];
    (*m_Grad)(sr+4, sc+i) = sqrt(2.0)*grad[4][i];
    (*m_Grad)(sr+5, sc+i) = grad[5][i];
    }
}

/* Usage: assume vectortype is actually matrixtype but with only 1 column
*         vector = yeoComputeCrossProductMatrixColumnAndBasisVector<MatrixType, ColumnVectorType>(m, colnum, basis);
*
*/
template <class MatrixType, class ColumnVectorType> ColumnVectorType yeoComputeCrossProductMatrixColumnAndBasisVector(MatrixType m, int colnum, int basis)
{
  ColumnVectorType v;

  if(basis == 0)
    {
    v(0, 0) = 0;
    v(1, 0) = m(2, colnum);
    v(2, 0) = -m(1, colnum);
    }
  else if(basis == 1)
    {
    v(0, 0) = -m(2, colnum);
    v(1, 0) = 0;
    v(2, 0) = m(0, colnum);
    }
  else if(basis == 2)
    {
    v(0, 0) = m(1, colnum);
    v(1, 0) = -m(0, colnum);
    v(2, 0) = 0;
    }
  else
    {
    std::cerr<<"Basis out of range: "<<basis<<std::endl;
    exit(EXIT_FAILURE);
    }

  return v;
}



/* Usage:  assume vectortype is actually matrixtype but with only 1 column
*         matrix = yeoCreateCrossProductMatrix<MatrixType, ColumnVectorType>(v);
*
*/
template<class MatrixType, class ColumnVectorType> MatrixType yeoCreateCrossProductMatrix(ColumnVectorType v) 
{
  MatrixType m;

  m(0, 0) = 0;
  m(0, 1) = -v(2, 0);
  m(0, 2) = v(1, 0);

  m(1, 0) = -m(0, 1);
  m(1, 1) = 0;
  m(1, 2) = -v(0, 0);

  m(2, 0) = -m(0, 2);
  m(2, 1) = -m(1, 2);
  m(2, 2) = 0;

  return m;
}

/* Usage:  
*          movingGradient = yeoReorientateMovingGradient<TensorGradientType, JacobianType>(R, movingGradient);
*/
template<class TensorGradientType, class JacobianType> TensorGradientType yeoReorientateMovingGradient(JacobianType R, TensorGradientType movingGradient)
{
  TensorGradientType output;

  typedef itk::Tensor<double, JacobianType::RowDimensions> doubleTensor;
  
  doubleTensor T;
  
  for(unsigned int col = 0; col < TensorGradientType::ColumnDimensions; col++)
    {
    //first copy onto tensor.
    for(unsigned int row = 0; row < TensorGradientType::RowDimensions; row++)
      T.SetNthComponent(row, movingGradient[row][col]);


    //Then rotate
    T.SetVnlMatrix(T.ApplyMatrix(R.GetTranspose())); //note the RTranspose is because in
                                                                              //ApplyMatrix it sets the tensor to be R*T*R'

    //Then copy to output
     for(unsigned int row = 0; row < TensorGradientType::RowDimensions; row++)
       output[row][col] = T.GetNthComponent(row);
    }
  return output;
}



/* Usage:
*         matrix = yeoComputeOuterProduct<HessianType>(u, v)
*
* u is treated as a column vector
* v is treated as a row vector
*/
template <class HessianType> HessianType yeoComputeOuterProduct(const double *u, const double *v)
{
  HessianType matrix;

  for(unsigned int row = 0; row < HessianType::RowDimensions; row++)
    for(unsigned int col = 0; col < HessianType::ColumnDimensions; col++)
      matrix[row][col] = u[row]*v[col];
  return matrix;
}

/* Usage: 
*          hessian = yeoComputeLeastSquareH<MatrixType, HessianType>(gradMatrix, varMatrix);
*/
template <class MatrixType, class HessianType> HessianType yeoComputeLeastSquareH(const MatrixType matrix, const MatrixType varMatrix)
{
  HessianType hessian; //assume dimensions correct. Check done in main
		       //function. 

  const double *v;

  hessian.Fill(0.0);  
  
  unsigned int row = 0;
  v = matrix[row];
  hessian += yeoComputeOuterProduct<HessianType>(v, v)/varMatrix(row, 0);
  
  row = 1;
  v = matrix[row];
  hessian += yeoComputeOuterProduct<HessianType>(v, v)/varMatrix(row, 0)*2.0;
  
  row = 2;
  v = matrix[row];
  hessian += yeoComputeOuterProduct<HessianType>(v, v)/varMatrix(row, 0);
  
  row = 3;
  v = matrix[row];
  hessian += yeoComputeOuterProduct<HessianType>(v, v)/varMatrix(row, 0)*2.0;
  
  row = 4;
  v = matrix[row];
  hessian += yeoComputeOuterProduct<HessianType>(v, v)/varMatrix(row, 0)*2.0;
  
  row = 5;
  v = matrix[row];
  hessian += yeoComputeOuterProduct<HessianType>(v, v)/varMatrix(row, 0);
  
  return hessian;
}

/* Usage: 
*          matrix = yeoCreateMatrixFromTensor<MatrixType, doubleTensor>(tensor0, tensor1, tensor2);
*/
template <class MatrixType, class doubleTensor> MatrixType yeoCreateMatrixFromTensor(const doubleTensor tensor0, 
										     const doubleTensor tensor1, 
										     const doubleTensor tensor2)
{
  MatrixType matrix;
  
  for(unsigned int i = 0; i < doubleTensor::NDegreesOfFreedom; i++)
    {
    matrix[i][0] = tensor0.GetNthComponent(i);
    }
  for(unsigned int i = 0; i < doubleTensor::NDegreesOfFreedom; i++)
    {
    matrix[i][1] = tensor1.GetNthComponent(i);
    }
 for(unsigned int i = 0; i < doubleTensor::NDegreesOfFreedom; i++)
    {
    matrix[i][2] = tensor2.GetNthComponent(i);
    }

 return matrix;
}


/* Usage: 
*          new_tensor = yeoComputeRotationGrad<TensorType>(tensor, row, col, step)
*/
template <class TensorType> TensorType yeoComputeRotationGrad(const TensorType tensor, int row, int col, double step)
{
  TensorType new_tensor;
  
  typedef vnl_matrix_fixed<double, TensorType::Dimension, TensorType::Dimension>  InternalMatrixType;
  
  InternalMatrixType D = tensor.GetVnlMatrix();

/*  
  InternalMatrixType f;
  f[0][0] = 0; f[0][1] = 0; f[0][2] = 0;
  f[1][0] = 0; f[1][1] = 0; f[1][2] = 0;
  f[2][0] = 0; f[2][1] = 0; f[2][2] = 0;
  
  f[row][col] = step;
  f[col][row] = -step;

  InternalMatrixType temp = f*D;
 new_tensor.SetVnlMatrix( 0.5 *(temp + temp.transpose()));
*/
  
  InternalMatrixType f;
  
  f[0][0] = 0; f[0][1] = 0; f[0][2] = 0;
  f[1][0] = 0; f[1][1] = 0; f[1][2] = 0;
  f[2][0] = 0; f[2][1] = 0; f[2][2] = 0;
  
  for(unsigned int i = 0; i < 3; i++)
    f[row][i] = D[col][i]*step;

  for(unsigned int i = 0; i < 3; i++)
    f[col][i] = D[row][i]*(-step);

  new_tensor.SetVnlMatrix( 0.5 * (f + f.transpose()));

  return new_tensor;
}

/*
 * Default constructor
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::ESMDemonsRegistrationFunctionTensor()
{

  RadiusType r;
  unsigned int j;
  for( j = 0; j < ImageDimension; j++ )
    {
    r[j] = 0;
    }
  this->SetRadius(r);

  m_TimeStep = 1.0;
  m_DenominatorThreshold = 1e-9;
  m_IntensityDifferenceThreshold = 0.001;
  m_MaximumUpdateStepLength = 0.5;
  
  this->SetMovingImage(NULL);
  this->SetFixedImage(NULL);
  m_FixedImageSpacing.Fill( 1.0 );
  m_FixedImageOrigin.Fill( 0.0 );
  m_Normalizer = 0.0;
  m_FixedImageGradientCalculator = GradientCalculatorType::New();

  //default Gradient Type
  this->m_UseGradientType = Symmetric;
  this->m_UseGradientType2 = Rotation;
  this->m_UseGradientType3 = Tom;

/*
  typename DefaultInterpolatorType::Pointer interp =
    DefaultInterpolatorType::New();

  m_MovingImageInterpolator = static_cast<InterpolatorType*>(
    interp.GetPointer() );

  m_MovingImageWarper = WarperType::New();
  m_MovingImageWarper->SetInterpolator( m_MovingImageInterpolator );
  m_MovingImageWarper->SetEdgePaddingValue( NumericTraits<typename MovingImageType::PixelType>::max() );
*/

  m_MovingImageWarper = WarperType::New();
  m_MovingImageWarper->SetEdgePaddingValue(  NumericTraits<typename MovingImageType::PixelType>::Zero );

  m_MovingImageWOReorientationWarper = WarperWOReorientationType::New();
  m_MovingImageWOReorientationWarper->SetEdgePaddingValue(  NumericTraits<typename MovingImageType::PixelType>::Zero );
   

  m_Metric = NumericTraits<double>::max();
  m_NormalizedMetric = NumericTraits<double>::max();
  m_SumOfSquaredDifference = 0.0;
  m_NumberOfPixelsProcessed = 0L;
  m_RMSChange = NumericTraits<double>::max();
  m_SumOfSquaredChange = 0.0;
  m_NormalizedSumOfSquaredDifference = 0.0;

  m_useAtlas = 0;
  m_VarImage = NULL;
  m_CacheVarianceImage = VarImageType::New();
 
  //m_SelfGradientImage = TensorGradientVectorImageType::New();
  //m_StdWeightedResidualImage = DoubleTensorVectorImageType::New();
  m_ForegroundRef = LongImageType::New();
  m_Iteration = 0;

  m_Jacobian = JacobianImageType::New();
  m_UpdateField = DeformationFieldType::New();
  m_ReorientedFixedImage = FixedImageType::New();
}


/*
 * Standard "PrintSelf" method.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::PrintSelf(std::ostream& os, Indent indent) const
{
  Superclass::PrintSelf(os, indent);

  //os << indent << "MovingImageIterpolator: ";
  //os << m_MovingImageInterpolator.GetPointer() << std::endl;
  os << indent << "FixedImageGradientCalculator: ";
  os << m_FixedImageGradientCalculator.GetPointer() << std::endl;
  os << indent << "DenominatorThreshold: ";
  os << m_DenominatorThreshold << std::endl;
  os << indent << "IntensityDifferenceThreshold: ";
  os << m_IntensityDifferenceThreshold << std::endl;

  os << indent << "Metric: ";
  os << m_Metric << std::endl;
  os << indent << "SumOfSquaredDifference: ";
  os << m_SumOfSquaredDifference << std::endl;
  os << indent << "NumberOfPixelsProcessed: ";
  os << m_NumberOfPixelsProcessed << std::endl;
  os << indent << "RMSChange: ";
  os << m_RMSChange << std::endl;
  os << indent << "SumOfSquaredChange: ";
  os << m_SumOfSquaredChange << std::endl;
  os<< indent << "useAtlas: ";
  os<< m_useAtlas << std::endl; 
}

/**
 *
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::SetIntensityDifferenceThreshold(double threshold)
{
  m_IntensityDifferenceThreshold = threshold;
}

/**
 *
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
double
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::GetIntensityDifferenceThreshold() const
{
  return m_IntensityDifferenceThreshold;
}


/**
 *
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::ComputeReorientedFixedImage(bool forwardBool)
{
//if forwardBool = 1, then reorient the "right way"
//else do the opposite. In this case, we are doing the oppposite to
//reorient the fixedImage tensors to be same as moving image.


  yeoCopyImage<FixedImageType>(this->GetFixedImage(), m_ReorientedFixedImage);


  typedef typename itk::ImageRegionIterator< FixedImageType> IteratorType;
  IteratorType It = IteratorType(m_ReorientedFixedImage, m_ReorientedFixedImage->GetLargestPossibleRegion());

  typedef typename itk::ImageRegionConstIterator<JacobianImageType> ConstIteratorType;
  ConstIteratorType constIt = ConstIteratorType(m_Jacobian, m_Jacobian->GetLargestPossibleRegion());


  It.GoToBegin();
  constIt.GoToBegin();
  

  while(!It.IsAtEnd())
    {
    if( It.GetIndex() != constIt.GetIndex())
      {
      std::cout<<"Iterator indices different: "<<It.GetIndex()<<", and "<<constIt.GetIndex()<<std::endl;
      }
    
    typename FixedImageType::PixelType fixedValue = It.Get();
    if(fixedValue != NumericTraits<typename FixedImageType::PixelType>::Zero)
      {

      doubleTensor origVal = static_cast<doubleTensor> (It.Get());
       
      JacobianType J =  constIt.Get();
      JacobianType R;
      JacobianType JJtsqrt;
      
      vnl_matrix_fixed <double, doubleTensor::Dimension, doubleTensor::Dimension >
	JJt = J.GetVnlMatrix() * J.GetTranspose();
      

      doubleTensor T;
      T.SetVnlMatrix(JJt);
      T = T.Sqrt();
      
      for( unsigned int row=0; row<doubleTensor::Dimension; row++)
	{
	for( unsigned int col=0; col<doubleTensor::Dimension; col++)
	  {
	  JJtsqrt(row, col) =  T.GetComponent(row, col);
	  }
	}

      T = T.Inv();
      vnl_matrix_fixed <double, doubleTensor::Dimension, doubleTensor::Dimension >
	Rvnl = T.GetVnlMatrix()*J.GetVnlMatrix();
 
      for( unsigned int row=0; row<doubleTensor::Dimension; row++)
	{
	for( unsigned int col=0; col<doubleTensor::Dimension; col++)
	  {
	  R (row,col) = Rvnl (row,col);
	  }
	}
      
      if(forwardBool)
	origVal.SetVnlMatrix(origVal.ApplyMatrix(R.GetTranspose())); 
      else
	origVal.SetVnlMatrix(origVal.ApplyMatrix(R.GetVnlMatrix()));
      
      It.Set(static_cast<typename FixedImageType::PixelType> (origVal));
      }
    ++It;
    ++constIt;
    }
}


/*
 * Set the function state values before each iteration
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::InitializeIteration()
{ 

#ifdef DEBUG
  std::cout<<"Initializing ESMDemonsRegistrationFunctionTensor"<<std::endl;
#endif
  if( !this->GetMovingImage() || !this->GetFixedImage() )
    {
    itkExceptionMacro( << "MovingImage or FixedImage not set" );
    }

  // cache fixed image information
  m_FixedImageSpacing    = this->GetFixedImage()->GetSpacing();
  m_FixedImageOrigin     = this->GetFixedImage()->GetOrigin();

  // compute the normalizer
  if ( m_MaximumUpdateStepLength > 0.0 )
  {
#ifdef DEBUG
  std::cout<<"Computing m_Normalized"<<std::endl;
#endif
     m_Normalizer = 0.0;
     for( unsigned int k = 0; k < ImageDimension; k++ )
     {
        m_Normalizer += m_FixedImageSpacing[k] * m_FixedImageSpacing[k];
     }
     //note the muliplication by 2.0
     m_Normalizer *= 2.0*2.0 *m_MaximumUpdateStepLength * m_MaximumUpdateStepLength /
        static_cast<double>( ImageDimension );
  }
  else
  {
     // set it to minus one to avoid rounding errors
     m_Normalizer = -1.0;
  }

 

  if(this->m_UseGradientType2 != Rotation)
    {
    m_MovingImageWarper->SetOutputSpacing( this->GetFixedImage()->GetSpacing() );
    m_MovingImageWarper->SetOutputOrigin( this->GetFixedImage()->GetOrigin() );
    m_MovingImageWarper->SetInput( this->GetMovingImage() );
    m_MovingImageWarper->SetDeformationField( this->GetDeformationField() );
    m_MovingImageWarper->GetOutput()->SetRequestedRegion( this->GetDeformationField()->GetRequestedRegion() );
    m_MovingImageWarper->Update();
    }
  else
    { //true rotation gradient requires special warper and Jacobian calculator
    m_MovingImageWOReorientationWarper->SetOutputSpacing( this->GetFixedImage()->GetSpacing() );
    m_MovingImageWOReorientationWarper->SetOutputOrigin( this->GetFixedImage()->GetOrigin() );
    m_MovingImageWOReorientationWarper->SetInput( this->GetMovingImage() );
    m_MovingImageWOReorientationWarper->SetDeformationField( this->GetDeformationField() );
    m_MovingImageWOReorientationWarper->GetOutput()->SetRequestedRegion( this->GetDeformationField()->GetRequestedRegion() );
    m_MovingImageWOReorientationWarper->Update();

    //Run Jacobian Calculator
    typename JacobianFilterType::Pointer jacobianFilter = JacobianFilterType::New();
    jacobianFilter->SetInput( this->GetDeformationField() );
    jacobianFilter->SetUseImageSpacing( true );
    
    try
      {
      jacobianFilter->Update();
      }
    catch(itk::ExceptionObject &e)
      {
      std::cerr << e;
      throw itk::ExceptionObject(__FILE__,__LINE__,"Error in WarpTensorImageFilter::BeforeThreadedGenerateData()");
      }
    
    m_Jacobian = jacobianFilter->GetOutput(); //note that jacobian
					      //filter computes
					      //jacobian + identity

    }


  // setup gradient calculator
  //if(m_UseGradientType == Symmetric && m_UseGradientType2 == Rotation)
  //  {
  //  ComputeReorientedFixedImage(0);
  //  m_FixedImageGradientCalculator->SetInputImage(this->m_ReorientedFixedImage);
  //  }
  //else
  m_FixedImageGradientCalculator->SetInputImage( this->GetFixedImage() );


  // initialize metric computation variables
  m_SumOfSquaredDifference  = 0.0;
  m_NumberOfPixelsProcessed = 0L;
  m_SumOfSquaredChange      = 0.0;
  m_NormalizedSumOfSquaredDifference = 0.0;

  //Reset TensorGradientVector image to all 0 size
  m_Iteration++;
  if(this->m_UseGradientType3 == Tom && this->m_UseGradientType2 != NoRotation) //need to solve linear system.
    {

    if(m_Iteration == 1)
      {

      //allocate new memories
      m_ForegroundRef = LongImageType::New();
      m_ForegroundRef->SetRegions(this->GetFixedImage()->GetLargestPossibleRegion());
      m_ForegroundRef->SetSpacing(this->GetFixedImage()->GetSpacing());
      m_ForegroundRef->SetOrigin(this->GetFixedImage()->GetOrigin());
      m_ForegroundRef->Allocate();
      }
    else if(this->m_ForegroundRef->GetLargestPossibleRegion().GetSize() != this->GetFixedImage()->GetLargestPossibleRegion().GetSize())
      {

      //beginning of new multires level. create new instances
      m_ForegroundRef = LongImageType::New();
      m_ForegroundRef->SetRegions(this->GetFixedImage()->GetLargestPossibleRegion());
      m_ForegroundRef->SetSpacing(this->GetFixedImage()->GetSpacing());
      m_ForegroundRef->SetOrigin(this->GetFixedImage()->GetOrigin());
      m_ForegroundRef->Allocate();

      }
    else
      {
      //nothing
      }
    
    
    long count = 0;

    typedef typename itk::ImageRegionConstIterator< FixedImageType> FixedIteratorType;
    FixedIteratorType fixIt = FixedIteratorType(this->GetFixedImage(), this->GetFixedImage()->GetLargestPossibleRegion());
    
    typedef typename itk::ImageRegionConstIterator< MovingImageType> MovingIteratorType;
    MovingIteratorType movIt;

    if(this->m_UseGradientType2 == Rotation) //checks whether requires reorientation during warp
      movIt = MovingIteratorType(m_MovingImageWOReorientationWarper->GetOutput(), m_MovingImageWOReorientationWarper->GetOutput()->GetLargestPossibleRegion());
    else
      movIt = MovingIteratorType(m_MovingImageWarper->GetOutput(), m_MovingImageWarper->GetOutput()->GetLargestPossibleRegion());


    typedef typename itk::ImageRegionIterator<LongImageType> IteratorType;
    IteratorType It = IteratorType(m_ForegroundRef, m_ForegroundRef->GetLargestPossibleRegion());

    for(It.GoToBegin(), fixIt.GoToBegin(), movIt.GoToBegin(); !It.IsAtEnd(); ++It, ++fixIt, ++movIt)
      {
      if(fixIt.Get() == NumericTraits<typename FixedImageType::PixelType>::Zero &&
	 movIt.Get() == NumericTraits<typename MovingImageType::PixelType>::Zero)
	It.Set(-1);
      else
	{
	It.Set(count);
	count++;
	}
      }

   
    m_Residual = new doubleVectorType((6+3)*count, 0.0);
    m_Grad = new SparseWriteMatrixType((6+3)*count, 3*count);
    }
  
  //clear gradient.
  if(this->m_UseGradientType3 == ConjugateGradient)
    yeoCreateZeroImage<FixedImageType, DeformationFieldType>(this->GetFixedImage(), m_UpdateField);
}


/*
 * Compute update at a non boundary neighbourhood
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
typename ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::PixelType
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::ComputeUpdate(const NeighborhoodType &it, void * gd,
                const FloatOffsetType& itkNotUsed(offset))
{
  if(ImageDimension!=3)
    {
    std::cerr<<"Dimension 2 not handled"<<std::endl;
    exit(EXIT_FAILURE);
    }
  

  bool fizero = 0;
  bool mizero = 0;
  
  GlobalDataStruct *globalData = (GlobalDataStruct *)gd;
  PixelType update;
  IndexType FirstIndex = this->GetFixedImage()->GetLargestPossibleRegion().GetIndex();
  IndexType LastIndex = this->GetFixedImage()->GetLargestPossibleRegion().GetIndex() + 
    this->GetFixedImage()->GetLargestPossibleRegion().GetSize();
  
  const IndexType index = it.GetIndex();
  
  if(!GetVarImage() && m_useAtlas) //Just a check
    {
    std::cerr<<"ESM: UseAtlas Boolean is true but VarImage does not exists"<<std::endl;
    exit(EXIT_FAILURE);
    }
  
/*
  {
  IndexType ReferenceIndex = LastIndex;
  ReferenceIndex[0] = ReferenceIndex[0]/2;
  ReferenceIndex[1] = ReferenceIndex[1]/2;
  ReferenceIndex[2] = ReferenceIndex[2]/2;

  if(FirstIndex == index)
    {
    std::cout<<"position: "<<m_ForegroundRef->GetPixel(ReferenceIndex)<<std::endl;

    ReferenceIndex[0]-= 1;
    std::cout<<"position: "<<m_ForegroundRef->GetPixel(ReferenceIndex)<<std::endl;
    ReferenceIndex[0]+= 2;
    std::cout<<"position: "<<m_ForegroundRef->GetPixel(ReferenceIndex)<<std::endl;
    ReferenceIndex[0]-= 1;

    ReferenceIndex[1]-= 1;
    std::cout<<"position: "<<m_ForegroundRef->GetPixel(ReferenceIndex)<<std::endl;
    ReferenceIndex[1]+= 2;
    std::cout<<"position: "<<m_ForegroundRef->GetPixel(ReferenceIndex)<<std::endl;
    ReferenceIndex[1]-= 1;
 

    ReferenceIndex[2]-= 1;
    std::cout<<"position: "<<m_ForegroundRef->GetPixel(ReferenceIndex)<<std::endl;
    ReferenceIndex[2]+= 2;
    std::cout<<"position: "<<m_ForegroundRef->GetPixel(ReferenceIndex)<<std::endl;

    ReferenceIndex[2]-= 1;
    }
  }
*/

#ifdef DEBUG
  if(GetVarImage())
    {
    typename VarImageType::IndexType referenceIndex2 = GetVarImage()->GetLargestPossibleRegion().GetIndex() 
      +  GetVarImage()->GetLargestPossibleRegion().GetSize();
    
    referenceIndex2[0] = referenceIndex2[0]/2;
    referenceIndex2[1] = referenceIndex2[1]/2;
    referenceIndex2[2] = referenceIndex2[2]/2;
    
    if(index == referenceIndex2)
      {
      std::cout<<"***********"<<std::endl;
      std::cout<<"VarImage["<<referenceIndex2[0]<<","<<referenceIndex2[1]<<","<<referenceIndex2[2]<<"]: "<<GetVarImage()->GetPixel(referenceIndex2)<<std::endl;
      std::cout<<"***********"<<std::endl;
      }
    }
  
  IndexType ReferenceIndex = LastIndex;
  ReferenceIndex[0] = ReferenceIndex[0]/2;
  ReferenceIndex[1] = ReferenceIndex[1]/2;
  ReferenceIndex[2] = ReferenceIndex[2]/2;
  
  //ReferenceIndex = FirstIndex;
#endif

  //typedef itk::Tensor<double, TensorPixelType::Dimension> doubleTensor; 
  //typedef itk::Vector<double, PixelType::Dimension>    doubleUpdateType;
  //typedef itk::Matrix<double, PixelType::Dimension, PixelType::Dimension> HessianType;
  
  // Get fixed image related information
  // Note: no need to check the index is within
  // fixed image buffer. This is done by the external filter.
  typename FixedImageType::PixelType fixedPixValue = this->GetFixedImage()->GetPixel( index );
  if(fixedPixValue == NumericTraits<typename FixedImageType::PixelType>::Zero)
    {
#ifdef DEBUG
    if(index == ReferenceIndex)
      {
      std::cout<<"Fixed Image in bg "<<std::endl;
      }
#endif
    update.Fill( 0.0 );
    fizero = 1;
    }
  const doubleTensor fixedValue = static_cast<doubleTensor> (fixedPixValue);
  TensorGradientType fixedGradient = m_FixedImageGradientCalculator->EvaluateAtIndex( index );
  
  
#ifdef DEBUG 
  if(index == ReferenceIndex)
    {
    std::cout<<"Fixed Gradient"<<fixedGradient<<std::endl;
    std::cout<<"Fixed Value is "<<fixedValue<<std::endl;
    
    if(ReferenceIndex != FirstIndex && (ReferenceIndex[0] != LastIndex[0]-1 || ReferenceIndex[1] != LastIndex[1]-1 || ReferenceIndex[2] != LastIndex[2]-1 ))
      PrintRadius1FixedImageNeighborhood(index);
    }
#endif
  // Get moving image related information
  // check if the point was mapped outside of the moving image using
  // the "special value" NumericTraits<MovingPixelType>::Zero
  typename MovingImageType::PixelType movingPixValue;
  if(this->m_UseGradientType2 == Rotation)
    movingPixValue = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( index );
  else
    movingPixValue = m_MovingImageWarper->GetOutput()->GetPixel( index );

  if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::Zero )
    {
#ifdef DEBUG
    if(index == ReferenceIndex)
      {
      std::cout<<"Out of Bound "<<std::endl;
      }
#endif
    update.Fill( 0.0 );
    mizero = 1;
    if(fizero == 1)
      {
      if(this->m_UseGradientType3 == ConjugateGradient)
	 m_UpdateField->SetPixel(index, update);
      return update;
      }
    // TEST
    // movingPixValue =  NumericTraits<typename MovingImageType::PixelType>::Zero;
    }

  JacobianType J;
  if(this->m_UseGradientType2 == Rotation)
    J = this->m_Jacobian->GetPixel(index);
  JacobianType R;
  JacobianType JJtsqrt;

  if(m_UseGradientType2 == Rotation) //compute associated rotation
    {    
    /*
    typedef itk::AffineTensorTransform< double, ImageDimension >  TransformType;
    typename TransformType::Pointer transform = TransformType::New();
    transform->SetMatrix( J );
    reorientedMovingValue = transform->TransformTensor( movingValue );
*/

    //This should replicate the work above.
    vnl_matrix_fixed <double, doubleTensor::Dimension, doubleTensor::Dimension >
      JJt = J.GetVnlMatrix() * J.GetTranspose();
    
    doubleTensor T;
    T.SetVnlMatrix(JJt);
    T = T.Sqrt();
    
    for( unsigned int row=0; row<doubleTensor::Dimension; row++)
      {
      for( unsigned int col=0; col<doubleTensor::Dimension; col++)
	{
        JJtsqrt(row, col) =  T.GetComponent(row, col);
	}
      }
    
    T = T.Inv();
    vnl_matrix_fixed <double, doubleTensor::Dimension, doubleTensor::Dimension >
      Rvnl = T.GetVnlMatrix()*J.GetVnlMatrix();
    
    for( unsigned int row=0; row<doubleTensor::Dimension; row++)
      {
      for( unsigned int col=0; col<doubleTensor::Dimension; col++)
	{
        R (row,col) = Rvnl (row,col);
	}
      }
    
    }//Compute associated rotation


  //casting the moving tensor
  const doubleTensor movingValue = static_cast<doubleTensor>(movingPixValue);
  
#ifdef DEBUG 
  if(index == ReferenceIndex)
    {
    std::cout<<"First Index: Moving Value is "<<movingValue<<std::endl;
    }
#endif
  /*===================================================
* Extract Moving Image Gradient
======================================================*/
  TensorGradientType usedDoubleGradient;
  
  if ( this->m_UseGradientType != Fixed  )
    {
    
#ifdef DEBUG
    if(index == ReferenceIndex)
      {
      std::cout<<"Computing Moving Gradient"<<std::endl;
      }
#endif
    
    // we don't use a CentralDifferenceImageFunction here to be able to
    // check for NumericTraits<MovingPixelType>::max()
    TensorGradientType movingGradient;
    IndexType tmpIndex = index;
    for( unsigned int dim = 0; dim < ImageDimension; dim++ )
      {
      // bounds checking
      
      if ( FirstIndex[dim]==LastIndex[dim]  || index[dim]<FirstIndex[dim] || index[dim]>=LastIndex[dim] )
	{
#ifdef DEBUG
	std::cout<<"***This should never happen***"<<std::endl;
#endif
	
	for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	  movingGradient[row][dim] = 0.0;
	continue;
	}
      else if ( index[dim] == FirstIndex[dim] )
	{
	// compute derivative
	tmpIndex[dim] += 1;

	if(this->m_UseGradientType2 == Rotation)
	  movingPixValue = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
	else
	  movingPixValue = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );

	if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::Zero )
	  {
	  // weird crunched border case
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    movingGradient[row][dim] = 0.0;
	  //std::cout<<"A weird crunched border case at index: "<<index<<std::endl;
	  }
	else
	  {
	  // forward difference
	  const doubleTensor diff = static_cast<doubleTensor>(movingPixValue) - movingValue;
	  
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    {
	    movingGradient[row][dim] = diff.GetNthComponent(row)/m_FixedImageSpacing[dim];
	    }
	  }
	tmpIndex[dim] -= 1;
	continue;
	}
      else if ( index[dim] == (LastIndex[dim]-1) )
	{
	// compute derivative
	tmpIndex[dim] -= 1;
	if(this->m_UseGradientType2 == Rotation)
	  movingPixValue = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
	else
	  movingPixValue = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );

	if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::Zero )
	  {
	  // weird crunched border case
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    movingGradient[row][dim] = 0.0;
	  //std::cout<<"B weird crunched border case at index: "<<index<<std::endl;
	  }
	else
	  {
	  // backward difference
	  const doubleTensor diff = movingValue - static_cast<doubleTensor>(movingPixValue);
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    {
	    movingGradient[row][dim] = diff.GetNthComponent(row)/m_FixedImageSpacing[dim];
	    }
	  }
	tmpIndex[dim] += 1;
	continue;
	} //End Special Boundary case
      
      
      // compute derivative
      tmpIndex[dim] += 1;
      if(this->m_UseGradientType2 == Rotation)
	movingPixValue = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
      else
	movingPixValue = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
      if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::Zero )
	{
	// backward difference
	
	tmpIndex[dim] -= 2;
	if(this->m_UseGradientType2 == Rotation)
	  movingPixValue = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
	else
	  movingPixValue = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
	if ( movingPixValue == NumericTraits<typename MovingImageType::PixelType>::Zero )
	  {
	  // weird crunched border case
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    movingGradient[row][dim] = 0.0;
	  
	  //bool onborder = false;
	  //for( unsigned int tmpdim = 0; tmpdim < ImageDimension; tmpdim++ )
	  //{
	  //   onborder |= ( index[tmpdim]==FirstIndex[tmpdim] or index[tmpdim]==(LastIndex[tmpdim]-1) );
	  //}
	  //if (onborder)
	  //   std::cout<<"C Weird crunched border case at index: "<<index<<std::endl;
	  //else
	  //   std::cout<<"D Weird crunched border case at index: "<<index<<std::endl;
	  }
	else
	  {
	  // backward difference
	  const doubleTensor diff = movingValue - static_cast<doubleTensor>(movingPixValue);
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    {
	    movingGradient[row][dim] = diff.GetNthComponent(row)/m_FixedImageSpacing[dim];
	    }
#ifdef DEBUG
	  if(index == ReferenceIndex)
	    {
	    std::cout<<"MovingPixelValue2: "<<movingValue<<std::endl;
	    std::cout<<"MovingPixelValue: "<<movingPixValue<<std::endl;
	    std::cout<<"Backward Diff, FixedSpacing: "<<m_FixedImageSpacing[dim]<<std::endl;
	    }
#endif
	  
	  }
	
	}
      else
	{
	//movingGradient[dim] = static_cast<double>( movingPixValue );
	
	tmpIndex[dim] -= 2;
	doubleTensor movingPixValue2;
	if(this->m_UseGradientType2 == Rotation)
	  movingPixValue2 = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
	else
	  movingPixValue2 = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
	
	if ( movingPixValue2 == NumericTraits<typename MovingImageType::PixelType>::Zero )
	  {
	  // forward difference
	  
	  const doubleTensor diff =  static_cast<doubleTensor>(movingPixValue) - movingValue;
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    {
	    movingGradient[row][dim] = diff.GetNthComponent(row)/m_FixedImageSpacing[dim];
	    }
#ifdef DEBUG
	  if(index == ReferenceIndex)
	    {
	    std::cout<<"MovingPixelValue2: "<<movingPixValue<<std::endl;
	    std::cout<<"MovingPixelValue: "<<movingValue<<std::endl;
	    std::cout<<"Forward Diff, FixedSpacing: "<<m_FixedImageSpacing[dim]<<std::endl;
	    }
#endif
	  }
	else
	  {
	  // normal case, central difference
	  const doubleTensor diff =  static_cast<doubleTensor>(movingPixValue) - static_cast<doubleTensor>(movingPixValue2);
	  for(int row = 0; row < MovingImageType::PixelType::DegreesOfFreedom; row++)
	    {
	    movingGradient[row][dim] = 0.5*diff.GetNthComponent(row)/m_FixedImageSpacing[dim];
	    }
#ifdef DEBUG
	  if(index == ReferenceIndex)
	    {
	    std::cout<<"MovingPixelValue2: "<<movingPixValue<<std::endl;
	    std::cout<<"MovingPixelValue: "<<movingPixValue2<<std::endl;
	    std::cout<<"Central Diff, FixedSpacing: "<<m_FixedImageSpacing[dim]<<std::endl;
	    }
#endif
	  }
	}
      tmpIndex[dim] += 1;
      
      }//end of for loop
    
    
/*=============================================
* Reorientate Moving Gradient
============================================*/
    if(m_UseGradientType2 == Rotation) //compute MovingGradient for
      //True Symmetric
      {
      movingGradient = yeoReorientateMovingGradient<TensorGradientType, JacobianType>(R, movingGradient);

      if(m_UseGradientType == Symmetric)
	{ //Average with FixedImage Gradient first
	movingGradient = (movingGradient + fixedGradient)/2.0;
	}
       
      }//reorient tensor
    
    
#ifdef DEBUG 
    if(index == ReferenceIndex)
      {
      std::cout<<"movingGradient: "<<movingGradient<<std:s:endl;
      std::cout<<"********************"<<std::endl;
      }
#endif
    //note, there's no divide by 2, because in the end, we multiply
    //the whole thing by 2. This comes from the gradient having to
    //divide by 2. But the hessian has to be divided by 2 twice, hence
    //we get an overall 4/2 = 2 factor at the end.
    
    //In this current version: we include the divided by 2 because
    //it gives me a headache...
    if (this->m_UseGradientType==Symmetric && this->m_UseGradientType2 != Rotation)
       {
       usedDoubleGradient = (fixedGradient + movingGradient)/2.0;
#ifdef DEBUG
       if(index == ReferenceIndex)
	 {
	 std::cout<<"Symmetric Gradient"<<usedDoubleGradient<<std::endl;
	 }
#endif
       }
    else if (this->m_UseGradientType==Moving || (this->m_UseGradientType == Symmetric && m_UseGradientType2 == Rotation))
      { //for now true symmetric will be using moving gradient...
      usedDoubleGradient = (movingGradient + movingGradient)/2.0;
      }
    else
      {
      itkExceptionMacro(<<"Unknown gradient type");
      }
    }
  else if (this->m_UseGradientType==Fixed)
    {
    usedDoubleGradient = (fixedGradient + fixedGradient)/2.0;
    }
  else
    {
    itkExceptionMacro(<<"Unknown gradient type");
    } //end of big if statement on whether to compute moving gradient type.
  
  // Compute Update.
  // We avoid the mismatch in units between the two terms. 
  // and avoid large step using a normalization term.
  

  doubleTensor reorientedMovingValue = movingValue;
  if(m_UseGradientType2 == Rotation) //reorient movingValue tensor
    {    
    /*
    typedef itk::AffineTensorTransform< double, ImageDimension >  TransformType;
    typename TransformType::Pointer transform = TransformType::New();
    transform->SetMatrix( J );
    reorientedMovingValue = transform->TransformTensor( movingValue );
*/

    //This should replicate the work above.
 
    reorientedMovingValue.SetVnlMatrix(reorientedMovingValue.ApplyMatrix(R.GetTranspose())); //note the RTranspose is because in
                                                                              //ApplyMatrix it sets the tensor to be R*T*R'
#ifdef DEBUG
    IndexType ReferenceIndex = LastIndex;
    ReferenceIndex[0] = ReferenceIndex[0]/2;
    ReferenceIndex[1] = ReferenceIndex[1]/2;
    ReferenceIndex[2] = ReferenceIndex[2]/2;
    if(m_Iteration > 1)
      {
      if(ReferenceIndex == index)
	{
	std::cout<<"Jacobian: "<<this->m_Jacobian->GetPixel(index)<<std::endl;
	std::cout<<"JJt: "<<JJt<<std::endl;
	std::cout<<"JJtsqrt: "<<JJtsqrt<<std::endl;
	std::cout<<"T: "<<T<<std::endl;
	std::cout<<"Rvnl: "<<Rvnl<<std::endl;
	std::cout<<"R: "<<R<<std::endl;
	std::cout<<"movingValue: "<<movingValue<<std::endl;
	std::cout<<"reoriented Moving value: "<<reorientedMovingValue<<std::endl;
	}
      }
#endif
    }//reorient tensor

 
  doubleTensor speedValue;
  if(m_UseGradientType2 == Rotation)
    speedValue = fixedValue - reorientedMovingValue;
  else
    speedValue = fixedValue - movingValue;
  const doubleTensor absSpeedValue =  yeoAbsTensor<doubleTensor>(speedValue);
  const doubleTensor sqSpeedValue = yeoPowerCombine2Tensors<doubleTensor>(speedValue, speedValue, 1, 1);
  
/*===================================================
* Extract Variance
======================================================*/
  doubleTensor varValue; doubleTensor stdValue;
  TensorGradientType varMatrix; TensorGradientType stdMatrix;
  
  if(m_useAtlas)
    {
    varValue =  static_cast<doubleTensor>(GetVarImage()->GetPixel(index));
    }  
  else
    {
    varValue = static_cast<doubleTensor>(sqSpeedValue);
    for(unsigned int row = 0; row < doubleTensor::NDegreesOfFreedom; row++)
      if(varValue.GetNthComponent(row) < MIN_VAR)
	varValue.SetNthComponent(row, MIN_VAR);
    }
  
  if(this->m_UseGradientType3 == Tom && this->m_UseGradientType2 != NoRotation)
    { //need to solve linear system
    stdValue = yeoSqrtTensor<doubleTensor>(varValue);
    stdMatrix = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(stdValue);
    }
  else
    varMatrix = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(varValue);


  double H;
  HessianType hessian;
  doubleUpdateType temp_update;
  if(this->m_UseGradientType2 == NoRotation || this->m_UseGradientType3 == Xavier || this->m_UseGradientType3 == ConjugateGradient)
    {
    /*===================================================
* Compute Update
======================================================*/

//Compute update due to sum of sq difference objective function.
    const TensorGradientType speedMat = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(speedValue);
    TensorGradientType trueGradient = yeoElemMultiply2Matrices<TensorGradientType>(usedDoubleGradient, speedMat);
    trueGradient = yeoElemDivide2Matrices<TensorGradientType>(trueGradient, varMatrix);
    
    temp_update = yeoSumTensorGradient<TensorGradientType, doubleUpdateType>(trueGradient);
    
/*===================================================
* Compute H or Hessian depending on gradient type
======================================================*/
    

    if(this->m_UseGradientType3 == Xavier)
      {

/*===================================================
* Compute H 
======================================================*/
      const TensorGradientType doubleGradientSq = yeoElemMultiply2Matrices<TensorGradientType>(usedDoubleGradient, usedDoubleGradient);
      
      TensorGradientType weightedDoubleGradientSq;
      weightedDoubleGradientSq = yeoElemDivide2Matrices<TensorGradientType>(doubleGradientSq, varMatrix);
    
      const doubleUpdateType H_ndim = yeoSumTensorGradient<TensorGradientType, doubleUpdateType>(weightedDoubleGradientSq);
      H = yeoSumUpdate<doubleUpdateType>(H_ndim);    
      
#ifdef DEBUG 
      if(index == ReferenceIndex)
	{
	std::cout<<"varMatrix: "<<varMatrix<<std::endl;
	std::cout<<"DoubleGradientSq: "<<doubleGradientSq<<std::endl;
	std::cout<<"weightedDoubleGradient: "<< weightedDoubleGradientSq<<std::endl;
	std::cout<<"H_ndim: "<<H_ndim<<std::endl;
	std::cout<<"H: "<<H<<std::endl;
	std::cout<<"temp_update: "<<temp_update<<std::endl;
	std::cout<<"********************"<<std::endl;
	}
#endif
      }
    else if(this->m_UseGradientType3 == Tom) //Tom and no rotation
      {
/*===================================================
* Compute Hessian 
======================================================*/
      
#ifdef DEBUG
      if(index == ReferenceIndex)
	{
	std::cout<<"usedDoubleGradient: "<<usedDoubleGradient<<std::endl;
	std::cout<<"varMatrix: "<<varMatrix<<std::endl;
	}  
#endif
      
      hessian = yeoComputeLeastSquareH<TensorGradientType, HessianType>(usedDoubleGradient, varMatrix);
      
#ifdef DEBUG
      if(index == ReferenceIndex)
	{
	std::cout<<"usedDoubleGradient: "<<usedDoubleGradient<<std::endl;
	std::cout<<"hessian: "<<hessian<<std::endl;
	}  
#endif
      
      //m_Normalizer
      HessianType augmentor;
      augmentor.SetIdentity();
      augmentor = augmentor*1.0/m_Normalizer;
      
      hessian += augmentor;
      
#ifdef DEBUG
      if(index == ReferenceIndex)
	{
	std::cout<<"m_Normalizer: "<<m_Normalizer<<std::endl;
	std::cout<<"augmentor: "<<augmentor<<std::endl;
	std::cout<<"hessian: "<<hessian<<std::endl;
	}  
#endif
      }
    else if(this->m_UseGradientType3 == ConjugateGradient)
      {
      m_UpdateFieldLock.Lock();
      temp_update += m_UpdateField->GetPixel(index);
      m_UpdateField->SetPixel(index, temp_update);
      m_UpdateFieldLock.Unlock();
      }
    else
      {
      std::cerr<<"Should never reach here"<<std::endl;
      exit(EXIT_FAILURE);
      
      }//end of Compute H or hessian
    }//end of compute update
  
  if(this->m_UseGradientType2 != NoRotation && this->m_UseGradientType3 == Tom)//Solve linear system: For Tom and when there's rotation!
    {
    
    long position = m_ForegroundRef->GetPixel(index);
    
    if(position < 0)
      {
      std::cerr<<"Position < 0: "<<position<<std::endl;
      exit(EXIT_FAILURE);
      }
    
#ifdef DEBUG 
    IndexType ReferenceIndex = LastIndex;
    ReferenceIndex[0] = ReferenceIndex[0]/2;
    ReferenceIndex[1] = ReferenceIndex[1]/2;
    ReferenceIndex[2] = ReferenceIndex[2]/2;
    if(ReferenceIndex == index)
      {
      std::cout<<"usedDoubleGradient: "<<usedDoubleGradient<<std::endl;
      std::cout<<"stdMatrix: "<<stdMatrix<<std::endl;
      std::cout<<"position: "<<position<<std::endl;
      std::cout<<"speedValue: "<<speedValue<<std::endl;
      std::cout<<"stdValue: "<<stdValue<<std::endl;
      exit(EXIT_FAILURE);
      }
#endif

    //Fill m_Grad
    m_GradLock.Lock();
    yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
      (m_Grad, 
       yeoElemDivide2Matrices<TensorGradientType>(usedDoubleGradient, stdMatrix), 
       position*6, 
       position*3);
    m_GradLock.Unlock();
    
    //Fill m_Residual
    m_ResidualLock.Lock();
    yeoInsertTensorIntoVector<doubleVectorType, doubleTensor>(m_Residual, 
							      yeoPowerCombine2Tensors<doubleTensor>(speedValue, stdValue, 1, -1), 
							      position*6);       
    
#ifdef DEBUG
    if(index == ReferenceIndex)
      {
      std::cout<<"stdMatrix: "<<stdMatrix<<std::endl;
      exit(EXIT_FAILURE);
      }  
#endif
    
    m_ResidualLock.Unlock();
    }


  

/*===================================================
* Compute Rotation part if necessary
======================================================*/
  if(this->m_UseGradientType2 == ApproxRotation && this->m_UseGradientType != Fixed)
    {
    //Now I am going to compute the part of the update due to rotation
    //you can read: fv_xm as fixed value x direction minus 1
    TensorGradientType rotationGrad;
    typename FixedImageType::PixelType fv_xm;
    typename FixedImageType::PixelType fv_xp;
    typename MovingImageType::PixelType mv_xm;
    typename MovingImageType::PixelType mv_xp;
    doubleTensor var_xm;
    doubleTensor var_xp;
    doubleTensor speedValue_xm;
    doubleTensor speedValue_xp;
    
/*================================
* Compute Tensor Value, Speed Value and var_value in neighborhood
==================================*/
    
    //unfortunately, I have to do it separately for each dimension.
    {
    int dim = 0;
    IndexType tmpIndex = index;
    
    if(index[dim] == FirstIndex[dim])
      {
      fv_xm = NumericTraits<typename FixedImageType::PixelType>::Zero;
      mv_xm = NumericTraits<typename MovingImageType::PixelType>::Zero;
      var_xm = NumericTraits<doubleTensor>::Zero;
      speedValue_xm = static_cast<doubleTensor>(fv_xm) - static_cast<doubleTensor>(mv_xm);
      }
    else
      {
      tmpIndex[dim] -= 1;
      fv_xm = this->GetFixedImage()->GetPixel( tmpIndex );
      
      //if(this->m_UseGradientType == TrueSymmetric)
      //mv_xm = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
      //else
      mv_xm = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );

      speedValue_xm = static_cast<doubleTensor>(fv_xm) - static_cast<doubleTensor>(mv_xm);
      
      if(m_useAtlas)
	var_xm = static_cast<doubleTensor>(GetVarImage()->GetPixel(tmpIndex));
      else
	{
	var_xm = yeoPowerCombine2Tensors<doubleTensor>(speedValue_xm, speedValue_xm, 1, 1);
	for(unsigned int row = 0; row < doubleTensor::NDegreesOfFreedom; row++)
	  if(var_xm.GetNthComponent(row) < MIN_VAR)
	    var_xm.SetNthComponent(row, MIN_VAR);
	}
      
      tmpIndex[dim] += 1;
      }
    
    if(index[dim] == LastIndex[dim]-1)
      {
      fv_xp = NumericTraits<typename FixedImageType::PixelType>::Zero;
      mv_xp = NumericTraits<typename MovingImageType::PixelType>::Zero;
      var_xp = NumericTraits<doubleTensor>::Zero;
      speedValue_xp = static_cast<doubleTensor>(fv_xp) - static_cast<doubleTensor>(mv_xp);
      }
    else
      {
      tmpIndex[dim] += 1;
      fv_xp = this->GetFixedImage()->GetPixel( tmpIndex );

      //if(this->m_UseGradientType == TrueSymmetric)
      //mv_xp = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
      //else
      mv_xp = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
      
      speedValue_xp = static_cast<doubleTensor>(fv_xp) - static_cast<doubleTensor>(mv_xp);
      
      if(m_useAtlas)
	var_xp = static_cast<doubleTensor>(GetVarImage()->GetPixel(tmpIndex));
      else
	{
	var_xp = yeoPowerCombine2Tensors<doubleTensor>(speedValue_xp, speedValue_xp, 1, 1);
	for(unsigned int row = 0; row < doubleTensor::NDegreesOfFreedom; row++)
	  if(var_xp.GetNthComponent(row) < MIN_VAR)
	    var_xp.SetNthComponent(row, MIN_VAR);
	}
      
      tmpIndex[dim] -= 1;
      }
    }//find the tensor values, speed_value and var_value  around the current point in the x-direction
    
    typename FixedImageType::PixelType fv_ym;
    typename FixedImageType::PixelType fv_yp;
    typename MovingImageType::PixelType mv_ym;
    typename MovingImageType::PixelType mv_yp;
    doubleTensor var_ym; 
    doubleTensor var_yp;
    doubleTensor speedValue_ym;
    doubleTensor speedValue_yp;
    
    {
    int dim = 1;
    IndexType tmpIndex = index;
    
    if(index[dim] == FirstIndex[dim])
      {
      fv_ym = NumericTraits<typename FixedImageType::PixelType>::Zero;
      mv_ym = NumericTraits<typename MovingImageType::PixelType>::Zero;
      var_ym = NumericTraits<doubleTensor>::Zero;
      speedValue_ym = static_cast<doubleTensor>(fv_ym) - static_cast<doubleTensor>(mv_ym);
      }
    else
      {
      tmpIndex[dim] -= 1;
      fv_ym = this->GetFixedImage()->GetPixel( tmpIndex );

      //if(this->m_UseGradientType == TrueSymmetric)
      //mv_ym = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
      //else
      mv_ym = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
      
      speedValue_ym = static_cast<doubleTensor>(fv_ym) - static_cast<doubleTensor>(mv_ym);
      
      if(m_useAtlas)
	var_ym = static_cast<doubleTensor>(GetVarImage()->GetPixel(tmpIndex));
      else
	{
	var_ym = yeoPowerCombine2Tensors<doubleTensor>(speedValue_ym, speedValue_ym, 1, 1);
	for(unsigned int row = 0; row < doubleTensor::NDegreesOfFreedom; row++)
	  if(var_ym.GetNthComponent(row) < MIN_VAR)
	    var_ym.SetNthComponent(row, MIN_VAR);
	}
      
      tmpIndex[dim] += 1;
      }
    
    if(index[dim] == LastIndex[dim]-1)
      {
      fv_yp = NumericTraits<typename FixedImageType::PixelType>::Zero;
      mv_yp = NumericTraits<typename MovingImageType::PixelType>::Zero;
      var_yp = NumericTraits<doubleTensor>::Zero;
      speedValue_yp = static_cast<doubleTensor>(fv_yp) - static_cast<doubleTensor>(mv_yp);
      }
    else
      {
      tmpIndex[dim] += 1;
      fv_yp = this->GetFixedImage()->GetPixel( tmpIndex );

      //if(this->m_UseGradientType == TrueSymmetric)
      //mv_yp = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
      //else
      mv_yp = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
      
      speedValue_yp = static_cast<doubleTensor>(fv_yp) - static_cast<doubleTensor>(mv_yp);
      
      if(m_useAtlas)
	var_yp = static_cast<doubleTensor>(GetVarImage()->GetPixel(tmpIndex));
      else
	{
	var_yp = yeoPowerCombine2Tensors<doubleTensor>(speedValue_yp, speedValue_yp, 1, 1);
	for(unsigned int row = 0; row < doubleTensor::NDegreesOfFreedom; row++)
	  if(var_yp.GetNthComponent(row) < MIN_VAR)
	    var_yp.SetNthComponent(row, MIN_VAR);
	}
      
      tmpIndex[dim] -= 1;
      }
    }//find the tensor values, speed_value and var_value  around the current point in the y-direction
    
    typename FixedImageType::PixelType fv_zm;
    typename FixedImageType::PixelType fv_zp;
    typename MovingImageType::PixelType mv_zm;
    typename MovingImageType::PixelType mv_zp;
    doubleTensor var_zm;
    doubleTensor var_zp;
    doubleTensor speedValue_zm;
    doubleTensor speedValue_zp;
    {
    int dim = 2;
    IndexType tmpIndex = index;
    
    if(index[dim] == FirstIndex[dim])
      {
      fv_zm = NumericTraits<typename FixedImageType::PixelType>::Zero;
      mv_zm = NumericTraits<typename MovingImageType::PixelType>::Zero;
      var_zm = NumericTraits<doubleTensor>::Zero;
      speedValue_zm = static_cast<doubleTensor>(fv_zm) - static_cast<doubleTensor>(mv_zm);
      }
    else
      {
      tmpIndex[dim] -= 1;
      fv_zm = this->GetFixedImage()->GetPixel( tmpIndex );

      //if(this->m_UseGradientType == TrueSymmetric)
      //mv_zm = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
      //else
      mv_zm = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
      
      speedValue_zm = static_cast<doubleTensor>(fv_zm) - static_cast<doubleTensor>(mv_zm);
      
      if(m_useAtlas)
	var_zm = static_cast<doubleTensor>(GetVarImage()->GetPixel(tmpIndex));
      else
	{
	var_zm = yeoPowerCombine2Tensors<doubleTensor>(speedValue_zm, speedValue_zm, 1, 1);
	for(unsigned int row = 0; row < doubleTensor::NDegreesOfFreedom; row++)
	  if(var_zm.GetNthComponent(row) < MIN_VAR)
	    var_zm.SetNthComponent(row, MIN_VAR);
	}
      
      tmpIndex[dim] += 1;
      }
    
    if(index[dim] == LastIndex[dim]-1)
      {
      fv_zp = NumericTraits<typename FixedImageType::PixelType>::Zero;
      mv_zp = NumericTraits<typename MovingImageType::PixelType>::Zero;
      var_zp = NumericTraits<doubleTensor>::Zero;
      speedValue_zp = static_cast<doubleTensor>(fv_zp) - static_cast<doubleTensor>(mv_zp);
      }
    else
      {
      tmpIndex[dim] += 1;
      fv_zp = this->GetFixedImage()->GetPixel( tmpIndex );

      //if(this->m_UseGradientType == TrueSymmetric)
      //mv_zp = m_MovingImageWOReorientationWarper->GetOutput()->GetPixel( tmpIndex );
      //else
      mv_zp = m_MovingImageWarper->GetOutput()->GetPixel( tmpIndex );
      
      speedValue_zp = static_cast<doubleTensor>(fv_zp) - static_cast<doubleTensor>(mv_zp);
      
      if(m_useAtlas)
	var_zp = static_cast<doubleTensor>(GetVarImage()->GetPixel(tmpIndex));
      else
	{
	var_zp = yeoPowerCombine2Tensors<doubleTensor>(speedValue_zp, speedValue_zp, 1, 1);
	for(unsigned int row = 0; row < doubleTensor::NDegreesOfFreedom; row++)
	  if(var_zp.GetNthComponent(row) < MIN_VAR)
	    var_zp.SetNthComponent(row, MIN_VAR);
	}
      
      tmpIndex[dim] -= 1;
      }
    }//find the tensor values, speed_value and var_value around the current point in the z-direction
    
#ifdef DEBUG 
    if(index == ReferenceIndex)
      {
      std::cout<<"fv_xm = "<<fv_xm<<std::endl;
      std::cout<<"fv_xp = "<<fv_xp<<std::endl;
      std::cout<<"fv_ym = "<<fv_ym<<std::endl;
      std::cout<<"fv_yp = "<<fv_yp<<std::endl;
      std::cout<<"fv_zm = "<<fv_zm<<std::endl;
      std::cout<<"fv_zp = "<<fv_zp<<std::endl;
      std::cout<<"mv_xm = "<<mv_xm<<std::endl;
      std::cout<<"mv_xp = "<<mv_xp<<std::endl;
      std::cout<<"mv_ym = "<<mv_ym<<std::endl;
      std::cout<<"mv_yp = "<<mv_yp<<std::endl;
      std::cout<<"mv_zm = "<<mv_zm<<std::endl;
      std::cout<<"mv_zp = "<<mv_zp<<std::endl;
      std::cout<<"speedValue_xm = "<<speedValue_xm<<std::endl;
      std::cout<<"speedValue_xp = "<<speedValue_xp<<std::endl;
      std::cout<<"speedValue_ym = "<<speedValue_ym<<std::endl;
      std::cout<<"speedValue_yp = "<<speedValue_yp<<std::endl;
      std::cout<<"speedValue_zm = "<<speedValue_zm<<std::endl;
      std::cout<<"speedValue_zp = "<<speedValue_zp<<std::endl;
      std::cout<<"var_xm = "<<var_xm<<std::endl;
      std::cout<<"var_xp = "<<var_xp<<std::endl;
      std::cout<<"var_ym = "<<var_ym<<std::endl;
      std::cout<<"var_yp = "<<var_yp<<std::endl;
      std::cout<<"var_zm = "<<var_zm<<std::endl;
      std::cout<<"var_zp = "<<var_zp<<std::endl;
      
      std::cout<<"********************"<<std::endl;
      }
#endif
    
/*====================================
* Compute Std for Symmetric
===================================*/
    doubleTensor std_xm;
    doubleTensor std_xp;
    doubleTensor std_ym;
    doubleTensor std_yp;
    doubleTensor std_zm;
    doubleTensor std_zp;
    
    if(m_UseGradientType3 == Tom) //need to solve linear system
      {
      std_xm = yeoSqrtTensor<doubleTensor>(var_xm);
      std_xp = yeoSqrtTensor<doubleTensor>(var_xp);
      std_ym = yeoSqrtTensor<doubleTensor>(var_ym);
      std_yp = yeoSqrtTensor<doubleTensor>(var_yp);
      std_zm = yeoSqrtTensor<doubleTensor>(var_zm);
      std_zp = yeoSqrtTensor<doubleTensor>(var_zp);
      }
      
/*================================
* Compute Rotation Gradient JpT 
==================================*/
      
      //Now I need to compute the 12 terms 
    //you can read: uym as u warp direction, image y direction, minus 1
      
    //first for u
    doubleTensor uym;
    if(fv_ym == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_ym == NumericTraits<typename MovingImageType::PixelType>::Zero)
      uym = NumericTraits<doubleTensor>::Zero;
    else
      uym = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_ym), 1, 0, 0.5/m_FixedImageSpacing[1]);
      
    doubleTensor uyp;
    if(fv_yp == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_yp == NumericTraits<typename MovingImageType::PixelType>::Zero)
      uyp = NumericTraits<doubleTensor>::Zero;
    else
      uyp = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_yp), 1, 0, -0.5/m_FixedImageSpacing[1]);
      
    doubleTensor uzm;
    if(fv_zm == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_zm == NumericTraits<typename MovingImageType::PixelType>::Zero)
      uzm = NumericTraits<doubleTensor>::Zero;
    else
      uzm = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_zm), 2, 0, 0.5/m_FixedImageSpacing[2]);
      
    doubleTensor uzp;
    if(fv_zp == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_zp == NumericTraits<typename MovingImageType::PixelType>::Zero)
      uzp = NumericTraits<doubleTensor>::Zero;
    else
      uzp = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_zp), 2, 0, -0.5/m_FixedImageSpacing[2]);
      
    //then for v
    doubleTensor vxm;
    if(fv_xm == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_xm == NumericTraits<typename MovingImageType::PixelType>::Zero)
      vxm = NumericTraits<doubleTensor>::Zero;
    else
      vxm = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_xm), 1, 0, -0.5/m_FixedImageSpacing[0]);
      
    doubleTensor vxp;
    if(fv_xp == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_xp == NumericTraits<typename MovingImageType::PixelType>::Zero)
      vxp = NumericTraits<doubleTensor>::Zero;
    else
      vxp = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_xp), 1, 0, 0.5/m_FixedImageSpacing[0]);
      
    doubleTensor vzm;
    if(fv_zm == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_zm == NumericTraits<typename MovingImageType::PixelType>::Zero)
      vzm = NumericTraits<doubleTensor>::Zero;
    else
      vzm = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_zm), 2, 1, 0.5/m_FixedImageSpacing[2]);
      
    doubleTensor vzp;
    if(fv_zp == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_zp == NumericTraits<typename MovingImageType::PixelType>::Zero)
      vzp = NumericTraits<doubleTensor>::Zero;
    else
      vzp = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_zp), 2, 1, -0.5/m_FixedImageSpacing[2]);
      
    //then for w
    doubleTensor wxm;
    if(fv_xm == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_xm == NumericTraits<typename MovingImageType::PixelType>::Zero)
      wxm = NumericTraits<doubleTensor>::Zero;
    else
      wxm = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_xm), 2, 0, -0.5/m_FixedImageSpacing[0]);
      
    doubleTensor wxp;
    if(fv_xp == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_xp == NumericTraits<typename MovingImageType::PixelType>::Zero)
      wxp = NumericTraits<doubleTensor>::Zero;
    else
      wxp = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_xp), 2, 0, 0.5/m_FixedImageSpacing[0]);
      
    doubleTensor wym;
    if(fv_ym == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_ym == NumericTraits<typename MovingImageType::PixelType>::Zero)
      wym = NumericTraits<doubleTensor>::Zero;
    else
      wym = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_ym), 2, 1, -0.5/m_FixedImageSpacing[1]);
      
    doubleTensor wyp;
    if(fv_yp == NumericTraits<typename FixedImageType::PixelType>::Zero || mv_yp == NumericTraits<typename MovingImageType::PixelType>::Zero)
      wyp = NumericTraits<doubleTensor>::Zero;
    else
      wyp = yeoComputeRotationGrad<doubleTensor>(static_cast<doubleTensor>(mv_yp), 2, 1, 0.5/m_FixedImageSpacing[1]);
      
      
#ifdef DEBUG 
    if(index == ReferenceIndex)
      {
      std::cout<<"uym = "<<uym<<std::endl;
      std::cout<<"uyp = "<<uyp<<std::endl;
      std::cout<<"uzm = "<<uzm<<std::endl;
      std::cout<<"uzp = "<<uzp<<std::endl;
      std::cout<<"vxm = "<<vxm<<std::endl;
      std::cout<<"vxp = "<<vxp<<std::endl;
      std::cout<<"vzm = "<<vzm<<std::endl;
      std::cout<<"vzp = "<<vzp<<std::endl;
      std::cout<<"wxm = "<<wxm<<std::endl;
      std::cout<<"wxp = "<<wxp<<std::endl;
      std::cout<<"wym = "<<wym<<std::endl;
      std::cout<<"wyp = "<<wyp<<std::endl;
	
      std::cout<<"********************"<<std::endl;
      }
#endif
      
/*================================
* Compute Rotation Gradient JpT normalized by variance
==================================*/
    doubleTensor weightedUym;
    if(var_ym == NumericTraits<doubleTensor>::Zero)
      weightedUym =  NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedUym = yeoPowerCombine2Tensors<doubleTensor>(uym, std_ym, 1, -1);
      else
	weightedUym = yeoPowerCombine2Tensors<doubleTensor>(uym, var_ym, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedUyp;
    if(var_yp == NumericTraits<doubleTensor>::Zero)
      weightedUyp = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedUyp = yeoPowerCombine2Tensors<doubleTensor>(uyp, std_yp, 1, -1);
      else
	weightedUyp = yeoPowerCombine2Tensors<doubleTensor>(uyp, var_yp, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedUzm;
    if(var_zm == NumericTraits<doubleTensor>::Zero)
      weightedUzm = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedUzm = yeoPowerCombine2Tensors<doubleTensor>(uzm, std_zm, 1, -1);
      else
	weightedUzm = yeoPowerCombine2Tensors<doubleTensor>(uzm, var_zm, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedUzp;
    if(var_zp == NumericTraits<doubleTensor>::Zero)
      weightedUzp = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedUzp = yeoPowerCombine2Tensors<doubleTensor>(uzp, std_zp, 1, -1);
      else
	weightedUzp = yeoPowerCombine2Tensors<doubleTensor>(uzp, var_zp, 1, -1); //xavier or conjugate gradient
      }      

    doubleTensor weightedVxm;
    if(var_xm == NumericTraits<doubleTensor>::Zero)
      weightedVxm = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom )
	weightedVxm = yeoPowerCombine2Tensors<doubleTensor>(vxm, std_xm, 1, -1);
      else
	weightedVxm = yeoPowerCombine2Tensors<doubleTensor>(vxm, var_xm, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedVxp;
    if(var_xp == NumericTraits<doubleTensor>::Zero)
      weightedVxp = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedVxp = yeoPowerCombine2Tensors<doubleTensor>(vxp, std_xp, 1, -1);
      else
	weightedVxp = yeoPowerCombine2Tensors<doubleTensor>(vxp, var_xp, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedVzm;
    if(var_zm == NumericTraits<doubleTensor>::Zero)
      weightedVzm = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedVzm = yeoPowerCombine2Tensors<doubleTensor>(vzm, std_zm, 1, -1);
      else
	weightedVzm = yeoPowerCombine2Tensors<doubleTensor>(vzm, var_zm, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedVzp;
    if(var_zp == NumericTraits<doubleTensor>::Zero)
      weightedVzp = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedVzp = yeoPowerCombine2Tensors<doubleTensor>(vzp, std_zp, 1, -1);
      else
	weightedVzp = yeoPowerCombine2Tensors<doubleTensor>(vzp, var_zp, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedWxm;
    if(var_xm == NumericTraits<doubleTensor>::Zero)
      weightedWxm = NumericTraits<doubleTensor>::Zero;
    else	
      {
      if(m_UseGradientType3 == Tom)
	weightedWxm = yeoPowerCombine2Tensors<doubleTensor>(wxm, std_xm, 1, -1);
      else
	weightedWxm = yeoPowerCombine2Tensors<doubleTensor>(wxm, var_xm, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedWxp;
    if(var_xp == NumericTraits<doubleTensor>::Zero)
      weightedWxp = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedWxp = yeoPowerCombine2Tensors<doubleTensor>(wxp, std_xp, 1, -1);
      else
	weightedWxp = yeoPowerCombine2Tensors<doubleTensor>(wxp, var_xp, 1, -1); //xavier or conjugate gradient
      }


    doubleTensor weightedWym;
    if(var_ym == NumericTraits<doubleTensor>::Zero)
      weightedWym = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedWym = yeoPowerCombine2Tensors<doubleTensor>(wym, std_ym, 1, -1);
      else
	weightedWym = yeoPowerCombine2Tensors<doubleTensor>(wym, var_ym, 1, -1); //xavier or conjugate gradient
      }

    doubleTensor weightedWyp;
    if(var_yp == NumericTraits<doubleTensor>::Zero)
      weightedWyp = NumericTraits<doubleTensor>::Zero;
    else
      {
      if(m_UseGradientType3 == Tom)
	weightedWyp = yeoPowerCombine2Tensors<doubleTensor>(wyp, std_yp, 1, -1);
      else
	weightedWyp = yeoPowerCombine2Tensors<doubleTensor>(wyp, var_yp, 1, -1); //xavier or conjugate gradient
      }      

#ifdef DEBUG 
    if(index == ReferenceIndex)
      {
      std::cout<<"weightedUym = "<<weightedUym<<std::endl;
      std::cout<<"weightedUyp = "<<weightedUyp<<std::endl;
      std::cout<<"weightedUzm = "<<weightedUzm<<std::endl;
      std::cout<<"weightedUzp = "<<weightedUzp<<std::endl;
      std::cout<<"weightedVxm = "<<weightedVxm<<std::endl;
      std::cout<<"weightedVxp = "<<weightedVxp<<std::endl;
      std::cout<<"weightedVzm = "<<weightedVzm<<std::endl;
      std::cout<<"weightedVzp = "<<weightedVzp<<std::endl;
      std::cout<<"weightedWxm = "<<weightedWxm<<std::endl;
      std::cout<<"weightedWxp = "<<weightedWxp<<std::endl;
      std::cout<<"weightedWym = "<<weightedWym<<std::endl;
      std::cout<<"weightedWyp = "<<weightedWyp<<std::endl;
      std::cout<<"********************"<<std::endl;
      }
#endif  
      
/*================================
* Compute actual rotation gradient
==================================*/

    if(this->m_UseGradientType3 == Xavier || this->m_UseGradientType3 == ConjugateGradient)
      {
      //compute update due to rotation
      doubleUpdateType temp_update2;
      
      {
      temp_update2[0] = 2.0*yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue_ym,weightedUym,1,1) +
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_yp,weightedUyp,1,1) +
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_zm,weightedUzm,1,1) + 
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_zp,weightedUzp,1,1));
      temp_update2[1] = 2.0*yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue_xm,weightedVxm,1,1)+
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_xp,weightedVxp,1,1)+
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_zm,weightedVzm,1,1)+
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_zp,weightedVzp,1,1));
      temp_update2[2] = 2.0*yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue_xm,weightedWxm,1,1)+
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_xp,weightedWxp,1,1)+
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_ym,weightedWym,1,1)+
						   yeoPowerCombine2Tensors<doubleTensor>(speedValue_yp,weightedWyp,1,1));

      /*================================
* Add actual rotation gradient to original rotation gradient
==================================*/
      if(this->m_UseGradientType3 == Xavier)
	temp_update += temp_update2;
      else
	{
	m_UpdateFieldLock.Lock();
	temp_update2 += this->m_UpdateField->GetPixel(index);
	this->m_UpdateField->SetPixel(index, temp_update2);
	m_UpdateFieldLock.Unlock();
	}

      }//compute update due to rotation
		
#ifdef DEBUG 
      if(index == ReferenceIndex)
	{
	std::cout<<"temp_update2 = "<<temp_update2<<std::endl;
	std::cout<<"********************"<<std::endl;
	}
#endif   
		
#ifdef DEBUG 
      if(index == ReferenceIndex)
	{
	std::cout<<"combined temp_update = "<<temp_update<<std::endl;
	std::cout<<"********************"<<std::endl;
	}
#endif  
	
      if(this->m_UseGradientType3 == Xavier)
	{
/*================================
* Compute Hessian For SymmetricXavier
==================================*/
	TensorGradientType gradxm = yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(NumericTraits<doubleTensor>::Zero, vxm, wxm);
	TensorGradientType gradxp = yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(NumericTraits<doubleTensor>::Zero, vxp, wxp);
	TensorGradientType gradym = yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(uym, NumericTraits<doubleTensor>::Zero, wym);
	TensorGradientType gradyp = yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(uyp, NumericTraits<doubleTensor>::Zero, wyp);
	TensorGradientType gradzm = yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(uzm, vzm, NumericTraits<doubleTensor>::Zero);
	TensorGradientType gradzp = yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(uzp, vzp, NumericTraits<doubleTensor>::Zero);
	
	TensorGradientType varMatrix_xm = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(var_xm);
	TensorGradientType varMatrix_xp = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(var_xp);
	TensorGradientType varMatrix_ym = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(var_ym);
	TensorGradientType varMatrix_yp = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(var_yp);
	TensorGradientType varMatrix_zm = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(var_zm);
	TensorGradientType varMatrix_zp = yeoRepTensor2Matrix<TensorGradientType, doubleTensor>(var_zp);
	
	if(var_xm != NumericTraits<doubleTensor>::Zero)
	  hessian += yeoComputeLeastSquareH<TensorGradientType, HessianType>(gradxm, varMatrix_xm) * 2.0;
	
	if(var_xp != NumericTraits<doubleTensor>::Zero)
	  hessian += yeoComputeLeastSquareH<TensorGradientType, HessianType>(gradxp, varMatrix_xp) * 2.0;
	
	if(var_ym != NumericTraits<doubleTensor>::Zero)
	  hessian += yeoComputeLeastSquareH<TensorGradientType, HessianType>(gradym, varMatrix_ym) * 2.0;
	
	if(var_yp != NumericTraits<doubleTensor>::Zero)
	  hessian += yeoComputeLeastSquareH<TensorGradientType, HessianType>(gradyp, varMatrix_yp) * 2.0;
	
	if(var_zm != NumericTraits<doubleTensor>::Zero)
	  hessian += yeoComputeLeastSquareH<TensorGradientType, HessianType>(gradzm, varMatrix_zm) * 2.0;
	
	if(var_zp != NumericTraits<doubleTensor>::Zero)
	  hessian += yeoComputeLeastSquareH<TensorGradientType, HessianType>(gradzp, varMatrix_zp) * 2.0;
	
#ifdef DEBUG 
	if(index == ReferenceIndex)
	  {
	  std::cout<<"gradxm = "<<gradxm<<std::endl;
	  std::cout<<"gradxp = "<<gradxp<<std::endl;
	  std::cout<<"gradym = "<<gradym<<std::endl;
	  std::cout<<"gradyp = "<<gradyp<<std::endl;
	  std::cout<<"gradzm = "<<gradzm<<std::endl;
	  std::cout<<"gradzp = "<<gradzp<<std::endl;
	  std::cout<<"varMatrix_xm = "<<varMatrix_xm<<std::endl;
	  std::cout<<"varMatrix_xp = "<<varMatrix_xp<<std::endl;
	  std::cout<<"varMatrix_ym = "<<varMatrix_ym<<std::endl;
	  std::cout<<"varMatrix_yp = "<<varMatrix_yp<<std::endl;
	  std::cout<<"varMatrix_zm = "<<varMatrix_zm<<std::endl;
	  std::cout<<"varMatrix_zp = "<<varMatrix_zp<<std::endl;
	  std::cout<<"hessian = "<<hessian<<std::endl;
	  	  
	  std::cout<<"********************"<<std::endl;
	  }
#endif 
	} //end hessian computation for Xavier		
      }//end update 
    else if(m_UseGradientType3 == Tom)
      {

/*
      if(m_UseGradientType == Symmetric)
	{
	weightedUym = weightedUym/2; 
	weightedUyp = weightedUyp/2; 
	weightedUzm = weightedUzm/2;
	weightedUzp = weightedUzp/2;
	weightedVxm = weightedVxm/2;
	weightedVxp = weightedVxp/2;
	weightedVzm = weightedVzm/2;
	weightedVzp = weightedVzp/2;
	weightedWxm = weightedWxm/2;
	weightedWxp = weightedWxp/2;
	weightedWym = weightedWym/2;
	weightedWyp = weightedWyp/2;
	}*/

/*==================================================
*Fill Rotation Component of Sparse Matrix
===================================================*/
      long position = m_ForegroundRef->GetPixel(index);
      m_GradLock.Lock();

      IndexType tmpIndex = index;
	
/*
      IndexType ReferenceIndex = LastIndex;
      ReferenceIndex[0] = 10; //ReferenceIndex[0]/2;
      ReferenceIndex[1] = 12; //ReferenceIndex[1]/2;
      ReferenceIndex[2] = 0; //ReferenceIndex[2]/2;
  */
    
      //xm
      tmpIndex[0] -= 1;

      if(tmpIndex[0] >= FirstIndex[0])
	{
	long pos_xm = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_xm >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(NumericTraits<doubleTensor>::Zero
									 ,weightedVxm, weightedWxm), 
	     6*pos_xm, 
	     3*position);
	  }
	/*
	IndexType ReferenceIndex_xm = ReferenceIndex;
	ReferenceIndex_xm[0] += 1;
	if(ReferenceIndex_xm == index)
	  {
	  std::cout<<"ReferenceIndex_xm: "<<ReferenceIndex_xm<<std::endl;
	  std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(NumericTraits<doubleTensor>::Zero
											      ,weightedVxm, weightedWxm)<<std::endl;
	  std::cout<<"position: "<<position<<std::endl;
	  std::cout<<"6*pos_xm  : "<<6*pos_xm<<std::endl;
	  std::cout<<"3*position: "<<3*position<<std::endl;
	  
	  }*/
	}

    
     

      //xp
      tmpIndex[0] += 2;

      if(tmpIndex[0] < LastIndex[0])
	{
	long pos_xp = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_xp >=0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad,
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(NumericTraits<doubleTensor>::Zero
									 , weightedVxp, weightedWxp),
	     6*pos_xp,
	     3*position);
	  }
/*
	IndexType ReferenceIndex_xp = ReferenceIndex;
	ReferenceIndex_xp[0] -= 1;
	if(ReferenceIndex_xp == index)
	  {
	  std::cout<<"ReferenceIndex_xp: "<<ReferenceIndex_xp<<std::endl;
	  std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(NumericTraits<doubleTensor>::Zero
											      , weightedVxp, weightedWxp) <<std::endl;
	  std::cout<<"position: "<<position<<std::endl;
	  std::cout<<"6*pos_xp  : "<<6*pos_xp<<std::endl;
	  std::cout<<"3*position: "<<3*position<<std::endl;
	  
	  }*/
	}
      tmpIndex[0] -= 1;

     


      //ym
      tmpIndex[1] -= 1;
	
      if(tmpIndex[1] >= FirstIndex[1])
	{
	long pos_ym = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_ym >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUym, 
									 NumericTraits<doubleTensor>::Zero, weightedWym),
	     6*pos_ym, 
	     3*position);
	  }

/*	IndexType ReferenceIndex_ym = ReferenceIndex;
	ReferenceIndex_ym[1] += 1;
	if(ReferenceIndex_ym == index)
	  {
	  std::cout<<"ReferenceIndex_ym: "<<ReferenceIndex_ym<<std::endl;
	  std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUym, 
											      NumericTraits<doubleTensor>::Zero, weightedWym) <<std::endl;
	  std::cout<<"position: "<<position<<std::endl;
	  std::cout<<"6*pos_ym  : "<<6*pos_ym<<std::endl;
	  std::cout<<"3*position: "<<3*position<<std::endl;
	  
	  }*/
	}

     


      //yp
      tmpIndex[1] += 2;
	
      if(tmpIndex[1] < LastIndex[1])
	{
	long pos_yp = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_yp >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUyp, 
									 NumericTraits<doubleTensor>::Zero, weightedWyp),
	     6*pos_yp, 
	     3*position);
	  }

/*	IndexType ReferenceIndex_yp = ReferenceIndex;
	ReferenceIndex_yp[1] -= 1;
	if(ReferenceIndex_yp == index)
	  {
	  std::cout<<"ReferenceIndex_yp: "<<ReferenceIndex_yp<<std::endl;
	  std::cout<<"Insert: "<<  yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUyp, 
											       NumericTraits<doubleTensor>::Zero, weightedWyp) <<std::endl;
	  std::cout<<"position: "<<position<<std::endl;
	  std::cout<<"6*pos_yp  : "<<6*pos_yp<<std::endl;
	  std::cout<<"3*position: "<<3*position<<std::endl;
	  
	  }*/
	}
      tmpIndex[1] -= 1;

   
 
      //zm
      tmpIndex[2] -= 1;

      if(tmpIndex[2] >= FirstIndex[2])
	{
	long pos_zm = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_zm >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUzm, weightedVzm, 
									 NumericTraits<doubleTensor>::Zero),
	     6*pos_zm, 
	     3*position);
	  }

/*	IndexType ReferenceIndex_zm = ReferenceIndex;
	ReferenceIndex_zm[2] += 1;
	if(ReferenceIndex_zm == index)
	  {
	  std::cout<<"ReferenceIndex_zm: "<<ReferenceIndex_zm<<std::endl;
	  std::cout<<"Insert: "<<  yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUzm, weightedVzm, 
											       NumericTraits<doubleTensor>::Zero)<<std::endl;
	  std::cout<<"position: "<<position<<std::endl;
	  std::cout<<"6*pos_zm  : "<<6*pos_zm<<std::endl;
	  std::cout<<"3*position: "<<3*position<<std::endl;	
	  }*/
	}

     

      //zp
      tmpIndex[2] += 2;

      if(tmpIndex[2] < LastIndex[2])
	{
	long pos_zp = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_zp >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUzp, weightedVzp, 
									 NumericTraits<doubleTensor>::Zero),
	     6*pos_zp, 
	     3*position);
	  }	

/*	IndexType ReferenceIndex_zp = ReferenceIndex;
	ReferenceIndex_zp[2] -= 1;
	if(ReferenceIndex_zp == index)
	  {
	  std::cout<<"ReferenceIndex_zp: "<<ReferenceIndex_zp<<std::endl;
	  std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(weightedUzp, weightedVzp, 
											      NumericTraits<doubleTensor>::Zero)<<std::endl; 
	  std::cout<<"position: "<<position<<std::endl;
	  std::cout<<"6*pos_zp  : "<<6*pos_zp<<std::endl;
	  std::cout<<"3*position: "<<3*position<<std::endl;
	  }*/
	}
      tmpIndex[2] -= 1;
      

//Fill the diagonals of m_Grad

      double m_Normalizer_sqrt_inv = 1.0/sqrt(m_Normalizer);
      long numCols = gmm::mat_ncols(*m_Grad);

      (*m_Grad)(2*numCols+3*position+0, 3*position+0) = m_Normalizer_sqrt_inv;
      (*m_Grad)(2*numCols+3*position+1, 3*position+1) = m_Normalizer_sqrt_inv;
      (*m_Grad)(2*numCols+3*position+2, 3*position+2) = m_Normalizer_sqrt_inv;

      m_GradLock.Unlock();
	

      } //Fill in Table for Symmetric Tom

    }//End Symmetric With Rotation Component
  else if(this->m_UseGradientType2 == Rotation  && this->m_UseGradientType != Fixed)
    {

    if(this->m_UseGradientType3 == Xavier)
      {
      std::cout<<"Xavier with Rotation Gradient not implemented!"<<std::endl;
      exit(EXIT_FAILURE);
      }


    /*===========================================
    *Compute TrueSymmetric Rotation Component!!
    ============================================*/
    
    /*====================================================================
    *First Compute beginning part of dR/dJij which is shared by everyone
==========================================================*/
    typedef vnl_matrix_fixed <double, doubleTensor::Dimension, doubleTensor::Dimension > doubleVnlMatrixType;
    typedef vnl_matrix_fixed <double, doubleTensor::Dimension, 1> doubleVnlColVecType;
    
    doubleVnlMatrixType prefactor_vnl;
    double trace = 0;
    for (unsigned int row=0; row<doubleTensor::Dimension; row++)
      trace += JJtsqrt(row, row);

    prefactor_vnl.set_identity();
    prefactor_vnl = prefactor_vnl * trace - JJtsqrt.GetVnlMatrix();
    prefactor_vnl = vnl_matrix_inverse<double>(prefactor_vnl);
    prefactor_vnl = R.GetTranspose()*prefactor_vnl*R.GetVnlMatrix(); //r' (tr(S)I - S)^{-1} r

#ifdef DEBUG
    if(m_Iteration > 1)
      {
       IndexType ReferenceIndex = LastIndex;
       ReferenceIndex[0] = ReferenceIndex[0]/2;
       ReferenceIndex[1] = ReferenceIndex[1]/2;
       ReferenceIndex[2] = ReferenceIndex[2]/2;
       if(ReferenceIndex == index)
	 {
	 std::cout<<"R: "<<R<<std::endl;
	 std::cout<<"JJtsqrt: "<<JJtsqrt<<std::endl;
	 std::cout<<"prefactor: "<<prefactor_vnl<<std::endl;
	 }
      }
#endif
    
    //first column
    //J00 corresponds to Jt00 
    doubleVnlColVecType vJ00 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 0, 0);    
    doubleVnlColVecType postfactor_vnlJ00 = prefactor_vnl * vJ00;
    doubleVnlMatrixType dRdJ00 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ00);

     //J10 corresponds to Jt01 
    doubleVnlColVecType vJ10 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 1, 0);    
    doubleVnlColVecType postfactor_vnlJ10 = prefactor_vnl * vJ10;
    doubleVnlMatrixType dRdJ10 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ10);

    //J20 corresponds to Jt02 
    doubleVnlColVecType vJ20 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 2, 0);    
    doubleVnlColVecType postfactor_vnlJ20 = prefactor_vnl * vJ20;
    doubleVnlMatrixType dRdJ20 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ20);

    //second column
    //J01 corresponds to Jt10 
    doubleVnlColVecType vJ01 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 0, 1);    
    doubleVnlColVecType postfactor_vnlJ01 = prefactor_vnl * vJ01;
    doubleVnlMatrixType dRdJ01 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ01);

    //J11 corresponds to Jt11 
    doubleVnlColVecType vJ11 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 1, 1);    
    doubleVnlColVecType postfactor_vnlJ11 = prefactor_vnl * vJ11;
    doubleVnlMatrixType dRdJ11 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ11);
    
    //J21 corresponds to Jt12 
    doubleVnlColVecType vJ21 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 2, 1);    
    doubleVnlColVecType postfactor_vnlJ21 = prefactor_vnl * vJ21;
    doubleVnlMatrixType dRdJ21 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ21);

    //third column
    //J02 corresponds to Jt20 
    doubleVnlColVecType vJ02 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 0, 2);    
    doubleVnlColVecType postfactor_vnlJ02 = prefactor_vnl * vJ02;
    doubleVnlMatrixType dRdJ02 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ02);

    //J12 corresponds to Jt21 
    doubleVnlColVecType vJ12 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 1, 2);    
    doubleVnlColVecType postfactor_vnlJ12 = prefactor_vnl * vJ12;
    doubleVnlMatrixType dRdJ12 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ12);
    
    //J22 corresponds to Jt22 
    doubleVnlColVecType vJ22 = 
      yeoComputeCrossProductMatrixColumnAndBasisVector<doubleVnlMatrixType, doubleVnlColVecType>(R.GetTranspose(), 2, 2);    
    doubleVnlColVecType postfactor_vnlJ22 = prefactor_vnl * vJ22;
    doubleVnlMatrixType dRdJ22 = -R.GetVnlMatrix() * 
      yeoCreateCrossProductMatrix<doubleVnlMatrixType, doubleVnlColVecType>(postfactor_vnlJ22);

#ifdef DEBUG
    if(m_Iteration > 1)
      {
      IndexType ReferenceIndex = LastIndex;
      ReferenceIndex[0] = ReferenceIndex[0]/2;
      ReferenceIndex[1] = ReferenceIndex[1]/2;
      ReferenceIndex[2] = ReferenceIndex[2]/2;
      if(ReferenceIndex == index)
	{
	std::cout<<"vJ00: "<<vJ00<<std::endl;
	std::cout<<"postfactor_vnlJ00: "<<postfactor_vnlJ00<<std::endl;
	std::cout<<"dRdJ00: "<<dRdJ00<<std::endl;
	std::cout<<"vJ10: "<<vJ10<<std::endl;
	std::cout<<"postfactor_vnlJ10: "<<postfactor_vnlJ10<<std::endl;
	std::cout<<"dRdJ10: "<<dRdJ10<<std::endl;
	std::cout<<"vJ20: "<<vJ20<<std::endl;
	std::cout<<"postfactor_vnlJ20: "<<postfactor_vnlJ20<<std::endl;
	std::cout<<"dRdJ20: "<<dRdJ20<<std::endl;
	std::cout<<"vJ01: "<<vJ01<<std::endl;
	std::cout<<"postfactor_vnlJ01: "<<postfactor_vnlJ01<<std::endl;
	std::cout<<"dRdJ01: "<<dRdJ01<<std::endl;
	std::cout<<"vJ11: "<<vJ11<<std::endl;
	std::cout<<"postfactor_vnlJ11: "<<postfactor_vnlJ11<<std::endl;
	std::cout<<"dRdJ11: "<<dRdJ11<<std::endl;
	std::cout<<"vJ21: "<<vJ21<<std::endl;
	std::cout<<"postfactor_vnlJ21: "<<postfactor_vnlJ21<<std::endl;
	std::cout<<"dRdJ21: "<<dRdJ21<<std::endl;
	std::cout<<"vJ02: "<<vJ02<<std::endl;
	std::cout<<"postfactor_vnlJ02: "<<postfactor_vnlJ02<<std::endl;
	std::cout<<"dRdJ02: "<<dRdJ02<<std::endl;
	std::cout<<"vJ12: "<<vJ12<<std::endl;
	std::cout<<"postfactor_vnlJ12: "<<postfactor_vnlJ12<<std::endl;
	std::cout<<"dRdJ12: "<<dRdJ12<<std::endl;
	std::cout<<"vJ22: "<<vJ22<<std::endl;
	std::cout<<"postfactor_vnlJ22: "<<postfactor_vnlJ22<<std::endl;
	std::cout<<"dRdJ22: "<<dRdJ22<<std::endl;
	}
      }
#endif
/*===================================================
* Retrieve Jacobians of neighbors. jacobian of points outside set to 0.
======================================================*/
    IndexType tmpIndex = index;

    tmpIndex[0] += 1;
    JacobianType Jxp;
    if(tmpIndex[0] < LastIndex[0])
      Jxp = m_Jacobian->GetPixel(tmpIndex);
    else
      Jxp.Fill(0.0);

    tmpIndex[0] -= 2;
    JacobianType Jxm;
    if(tmpIndex[0] >= FirstIndex[0])
      Jxm = m_Jacobian->GetPixel(tmpIndex);
    else
      Jxm.Fill(0.0);
    tmpIndex[0] += 1;
    
    tmpIndex[1] += 1;
    JacobianType Jyp; 
    if(tmpIndex[1] < LastIndex[1])
      Jyp = m_Jacobian->GetPixel(tmpIndex);
    else
      Jyp.Fill(0.0);

    tmpIndex[1] -= 2;
    JacobianType Jym;
    if(tmpIndex[1] >= FirstIndex[1])
      Jym = m_Jacobian->GetPixel(tmpIndex);
    else
      Jym.Fill(0.0);
    tmpIndex[1] += 1;

    tmpIndex[2] += 1;
    JacobianType Jzp;
    if(tmpIndex[2] < LastIndex[2])
      Jzp = m_Jacobian->GetPixel(tmpIndex);
    else
      Jzp.Fill(0.0);

    tmpIndex[2] -= 2;
    JacobianType Jzm;
    if(tmpIndex[2] >= FirstIndex[2])
      Jzm = m_Jacobian->GetPixel(tmpIndex);
    else
      Jzm.Fill(0.0);
    tmpIndex[2] += 1;

/*=================================================================
* Compute Gradient of R: dRdU, dRdV, dRdW
=================================================================*/
  
    //dRdUxp
    doubleVnlMatrixType dRdUxp;
    dRdUxp.fill(0.0);
    dRdUxp += dRdJ00 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(0, 0);
    dRdUxp += dRdJ10 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(1, 0);
    dRdUxp += dRdJ20 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(2, 0);
    
//dRdVxp
    doubleVnlMatrixType dRdVxp;
    dRdVxp.fill(0.0);
    dRdVxp += dRdJ00 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(0, 1);
    dRdVxp += dRdJ10 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(1, 1);
    dRdVxp += dRdJ20 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(2, 1);
    
//dRdWxp
    doubleVnlMatrixType dRdWxp;
    dRdWxp.fill(0.0);
    dRdWxp += dRdJ00 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(0, 2);
    dRdWxp += dRdJ10 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(1, 2);
    dRdWxp += dRdJ20 / (2.0 * this->m_FixedImageSpacing[0]) * Jxp(2, 2);


 //dRdUxm
    doubleVnlMatrixType dRdUxm;
    dRdUxm.fill(0.0);
    dRdUxm += dRdJ00 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(0, 0);
    dRdUxm += dRdJ10 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(1, 0);
    dRdUxm += dRdJ20 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(2, 0);
    
//dRdVxm
    doubleVnlMatrixType dRdVxm;
    dRdVxm.fill(0.0);
    dRdVxm += dRdJ00 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(0, 1);
    dRdVxm += dRdJ10 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(1, 1);
    dRdVxm += dRdJ20 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(2, 1);
    
//dRdWxm
    doubleVnlMatrixType dRdWxm;
    dRdWxm.fill(0.0);
    dRdWxm += dRdJ00 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(0, 2);
    dRdWxm += dRdJ10 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(1, 2);
    dRdWxm += dRdJ20 / (-2.0 * this->m_FixedImageSpacing[0]) * Jxm(2, 2);

    //dRdUyp
    doubleVnlMatrixType dRdUyp;
    dRdUyp.fill(0.0);
    dRdUyp += dRdJ01 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(0, 0);
    dRdUyp += dRdJ11 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(1, 0);
    dRdUyp += dRdJ21 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(2, 0);
    
//dRdVyp
    doubleVnlMatrixType dRdVyp;
    dRdVyp.fill(0.0);
    dRdVyp += dRdJ01 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(0, 1);
    dRdVyp += dRdJ11 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(1, 1);
    dRdVyp += dRdJ21 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(2, 1);
    
//dRdWyp
    doubleVnlMatrixType dRdWyp;
    dRdWyp.fill(0.0);
    dRdWyp += dRdJ01 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(0, 2);
    dRdWyp += dRdJ11 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(1, 2);
    dRdWyp += dRdJ21 / (2.0 * this->m_FixedImageSpacing[1]) * Jyp(2, 2);

 //dRdUym
    doubleVnlMatrixType dRdUym;
    dRdUym.fill(0.0);
    dRdUym += dRdJ01 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(0, 0);
    dRdUym += dRdJ11 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(1, 0);
    dRdUym += dRdJ21 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(2, 0);
    
//dRdVym
    doubleVnlMatrixType dRdVym;
    dRdVym.fill(0.0);
    dRdVym += dRdJ01 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(0, 1);
    dRdVym += dRdJ11 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(1, 1);
    dRdVym += dRdJ21 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(2, 1);
    
//dRdWym
    doubleVnlMatrixType dRdWym;
    dRdWym.fill(0.0);
    dRdWym += dRdJ01 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(0, 2);
    dRdWym += dRdJ11 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(1, 2);
    dRdWym += dRdJ21 / (-2.0 * this->m_FixedImageSpacing[1]) * Jym(2, 2);

  //dRdUzp
    doubleVnlMatrixType dRdUzp;
    dRdUzp.fill(0.0);
    dRdUzp += dRdJ02 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(0, 0);
    dRdUzp += dRdJ12 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(1, 0);
    dRdUzp += dRdJ22 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(2, 0);
    
//dRdVzp
    doubleVnlMatrixType dRdVzp;
    dRdVzp.fill(0.0);
    dRdVzp += dRdJ02 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(0, 1);
    dRdVzp += dRdJ12 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(1, 1);
    dRdVzp += dRdJ22 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(2, 1);
    
//dRdWzp
    doubleVnlMatrixType dRdWzp;
    dRdWzp.fill(0.0);
    dRdWzp += dRdJ02 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(0, 2);
    dRdWzp += dRdJ12 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(1, 2);
    dRdWzp += dRdJ22 / (2.0 * this->m_FixedImageSpacing[2]) * Jzp(2, 2);

 //dRdUzm
    doubleVnlMatrixType dRdUzm;
    dRdUzm.fill(0.0);
    dRdUzm += dRdJ02 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(0, 0);
    dRdUzm += dRdJ12 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(1, 0);
    dRdUzm += dRdJ22 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(2, 0);
    
//dRdVzm
    doubleVnlMatrixType dRdVzm;
    dRdVzm.fill(0.0);
    dRdVzm += dRdJ02 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(0, 1);
    dRdVzm += dRdJ12 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(1, 1);
    dRdVzm += dRdJ22 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(2, 1);
    
//dRdWzm
    doubleVnlMatrixType dRdWzm;
    dRdWzm.fill(0.0);
    dRdWzm += dRdJ02 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(0, 2);
    dRdWzm += dRdJ12 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(1, 2);
    dRdWzm += dRdJ22 / (-2.0 * this->m_FixedImageSpacing[2]) * Jzm(2, 2);

#ifdef DEBUG
    if(m_Iteration > 1)
      {
      IndexType ReferenceIndex = LastIndex;
      ReferenceIndex[0] = ReferenceIndex[0]/2;
      ReferenceIndex[1] = ReferenceIndex[1]/2;
      ReferenceIndex[2] = ReferenceIndex[2]/2;
      if(ReferenceIndex == index)
	{
	std::cout<<"x_spacing: "<<this->m_FixedImageSpacing[0]<<std::endl;
	std::cout<<"y_spacing: "<<this->m_FixedImageSpacing[1]<<std::endl;
	std::cout<<"z_spacing: "<<this->m_FixedImageSpacing[2]<<std::endl;
	std::cout<<"dRdUxp: "<<dRdUxp<<std::endl;
	std::cout<<"dRdVxp: "<<dRdVxp<<std::endl;
	std::cout<<"dRdWxp: "<<dRdWxp<<std::endl;
	std::cout<<"dRdUxm: "<<dRdUxm<<std::endl;
	std::cout<<"dRdVxm: "<<dRdVxm<<std::endl;
	std::cout<<"dRdWxm: "<<dRdWxm<<std::endl;
	std::cout<<"dRdUyp: "<<dRdUyp<<std::endl;
	std::cout<<"dRdVyp: "<<dRdVyp<<std::endl;
	std::cout<<"dRdWyp: "<<dRdWyp<<std::endl;
	std::cout<<"dRdUym: "<<dRdUym<<std::endl;
	std::cout<<"dRdVym: "<<dRdVym<<std::endl;
	std::cout<<"dRdWym: "<<dRdWym<<std::endl;
	std::cout<<"dRdUzp: "<<dRdUzp<<std::endl;
	std::cout<<"dRdVzp: "<<dRdVzp<<std::endl;
	std::cout<<"dRdWzp: "<<dRdWzp<<std::endl;
	std::cout<<"dRdUzm: "<<dRdUzm<<std::endl;
	std::cout<<"dRdVzm: "<<dRdVzm<<std::endl;
	std::cout<<"dRdWzm: "<<dRdWzm<<std::endl;
	}     
      }
#endif

/*============================================
* Compute R' * M * gradR
=============================================*/
    doubleVnlMatrixType tmpMatrix;
    tmpMatrix = R.GetTranspose() * movingValue.GetVnlMatrix();

    doubleVnlMatrixType Rt_M_dRdUxp = tmpMatrix * dRdUxp;
    doubleVnlMatrixType Rt_M_dRdVxp = tmpMatrix * dRdVxp;
    doubleVnlMatrixType Rt_M_dRdWxp = tmpMatrix * dRdWxp;

    doubleVnlMatrixType Rt_M_dRdUxm = tmpMatrix * dRdUxm;
    doubleVnlMatrixType Rt_M_dRdVxm = tmpMatrix * dRdVxm;
    doubleVnlMatrixType Rt_M_dRdWxm = tmpMatrix * dRdWxm;

    doubleVnlMatrixType Rt_M_dRdUyp = tmpMatrix * dRdUyp;
    doubleVnlMatrixType Rt_M_dRdVyp = tmpMatrix * dRdVyp;
    doubleVnlMatrixType Rt_M_dRdWyp = tmpMatrix * dRdWyp;

    doubleVnlMatrixType Rt_M_dRdUym = tmpMatrix * dRdUym;
    doubleVnlMatrixType Rt_M_dRdVym = tmpMatrix * dRdVym;
    doubleVnlMatrixType Rt_M_dRdWym = tmpMatrix * dRdWym;

    doubleVnlMatrixType Rt_M_dRdUzp = tmpMatrix * dRdUzp;
    doubleVnlMatrixType Rt_M_dRdVzp = tmpMatrix * dRdVzp;
    doubleVnlMatrixType Rt_M_dRdWzp = tmpMatrix * dRdWzp;

    doubleVnlMatrixType Rt_M_dRdUzm = tmpMatrix * dRdUzm;
    doubleVnlMatrixType Rt_M_dRdVzm = tmpMatrix * dRdVzm;
    doubleVnlMatrixType Rt_M_dRdWzm = tmpMatrix * dRdWzm;

/*============================================
* Compute entire rotation grad
=============================================*/
   
    doubleTensor gradUxp;
    if(fizero == 1 || mizero == 1)
      gradUxp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradUxp.SetVnlMatrix(Rt_M_dRdUxp.transpose() + Rt_M_dRdUxp);

      if(this->m_UseGradientType3 == Tom) 
	gradUxp = yeoPowerCombine2Tensors<doubleTensor>(gradUxp, stdValue, 1, -1);
      else
	gradUxp = yeoPowerCombine2Tensors<doubleTensor>(gradUxp, varValue, 1, -1);
      }

    doubleTensor gradVxp;
    if(fizero == 1 || mizero == 1)
      gradVxp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradVxp.SetVnlMatrix(Rt_M_dRdVxp.transpose() + Rt_M_dRdVxp);

      if(this->m_UseGradientType3 == Tom) 
	gradVxp = yeoPowerCombine2Tensors<doubleTensor>(gradVxp, stdValue, 1, -1);
      else
	gradVxp = yeoPowerCombine2Tensors<doubleTensor>(gradVxp, varValue, 1, -1);
      }

    doubleTensor gradWxp;
    if(fizero == 1 || mizero == 1)
      gradWxp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradWxp.SetVnlMatrix(Rt_M_dRdWxp.transpose() + Rt_M_dRdWxp);

      if(this->m_UseGradientType3 == Tom) 
	gradWxp = yeoPowerCombine2Tensors<doubleTensor>(gradWxp, stdValue, 1, -1);
      else
	gradWxp = yeoPowerCombine2Tensors<doubleTensor>(gradWxp, varValue, 1, -1);
      }

    doubleTensor gradUxm;
    if(fizero == 1 || mizero == 1)
      gradUxm = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradUxm.SetVnlMatrix(Rt_M_dRdUxm.transpose() + Rt_M_dRdUxm);

      if(this->m_UseGradientType3 == Tom) 
	gradUxm = yeoPowerCombine2Tensors<doubleTensor>(gradUxm, stdValue, 1, -1);
      else
	gradUxm = yeoPowerCombine2Tensors<doubleTensor>(gradUxm, varValue, 1, -1);
      }    

    doubleTensor gradVxm;
    if(fizero == 1 || mizero == 1)
      gradVxm = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradVxm.SetVnlMatrix(Rt_M_dRdVxm.transpose() + Rt_M_dRdVxm);

      if(this->m_UseGradientType3 == Tom) 
	gradVxm = yeoPowerCombine2Tensors<doubleTensor>(gradVxm, stdValue, 1, -1);
      else
	gradVxm = yeoPowerCombine2Tensors<doubleTensor>(gradVxm, varValue, 1, -1);
      }

    doubleTensor gradWxm;
    if(fizero == 1 || mizero == 1)
      gradWxm = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradWxm.SetVnlMatrix(Rt_M_dRdWxm.transpose() + Rt_M_dRdWxm);

      if(this->m_UseGradientType3 == Tom) 
	gradWxm = yeoPowerCombine2Tensors<doubleTensor>(gradWxm, stdValue, 1, -1);
      else
	gradWxm = yeoPowerCombine2Tensors<doubleTensor>(gradWxm, varValue, 1, -1);
      }

    doubleTensor gradUyp;
    if(fizero == 1 || mizero == 1)
      gradUyp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradUyp.SetVnlMatrix(Rt_M_dRdUyp.transpose() + Rt_M_dRdUyp);

      if(this->m_UseGradientType3 == Tom) 
	gradUyp = yeoPowerCombine2Tensors<doubleTensor>(gradUyp, stdValue, 1, -1);
      else
	gradUyp = yeoPowerCombine2Tensors<doubleTensor>(gradUyp, varValue, 1, -1);
      }

    doubleTensor gradVyp;
    if(fizero == 1 || mizero == 1)
      gradVyp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradVyp.SetVnlMatrix(Rt_M_dRdVyp.transpose() + Rt_M_dRdVyp);

      if(this->m_UseGradientType3 == Tom) 
	gradVyp = yeoPowerCombine2Tensors<doubleTensor>(gradVyp, stdValue, 1, -1);
      else
	gradVyp = yeoPowerCombine2Tensors<doubleTensor>(gradVyp, varValue, 1, -1);
      }

    doubleTensor gradWyp;
    if(fizero == 1 || mizero == 1)
      gradWyp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradWyp.SetVnlMatrix(Rt_M_dRdWyp.transpose() + Rt_M_dRdWyp);

      if(this->m_UseGradientType3 == Tom) 
	gradWyp = yeoPowerCombine2Tensors<doubleTensor>(gradWyp, stdValue, 1, -1);
      else
	gradWyp = yeoPowerCombine2Tensors<doubleTensor>(gradWyp, varValue, 1, -1);
      }

    doubleTensor gradUym;
    if(fizero == 1 || mizero == 1)
      gradUym = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradUym.SetVnlMatrix(Rt_M_dRdUym.transpose() + Rt_M_dRdUym);

      if(this->m_UseGradientType3 == Tom) 
	gradUym = yeoPowerCombine2Tensors<doubleTensor>(gradUym, stdValue, 1, -1);
      else
	gradUym = yeoPowerCombine2Tensors<doubleTensor>(gradUym, varValue, 1, -1);
      }

    doubleTensor gradVym;
    if(fizero == 1 || mizero == 1)
      gradVym = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradVym.SetVnlMatrix(Rt_M_dRdVym.transpose() + Rt_M_dRdVym);

      if(this->m_UseGradientType3 == Tom) 
	gradVym = yeoPowerCombine2Tensors<doubleTensor>(gradVym, stdValue, 1, -1);
      else
	gradVym = yeoPowerCombine2Tensors<doubleTensor>(gradVym, varValue, 1, -1);
      }

    doubleTensor gradWym;
    if(fizero == 1 || mizero == 1)
      gradWym = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradWym.SetVnlMatrix(Rt_M_dRdWym.transpose() + Rt_M_dRdWym);

      if(this->m_UseGradientType3 == Tom) 
	gradWym = yeoPowerCombine2Tensors<doubleTensor>(gradWym, stdValue, 1, -1);
      else
	gradWym = yeoPowerCombine2Tensors<doubleTensor>(gradWym, varValue, 1, -1);
      }

    doubleTensor gradUzp;
    if(fizero == 1 || mizero == 1)
      gradUzp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradUzp.SetVnlMatrix(Rt_M_dRdUzp.transpose() + Rt_M_dRdUzp);

      if(this->m_UseGradientType3 == Tom) 
	gradUzp = yeoPowerCombine2Tensors<doubleTensor>(gradUzp, stdValue, 1, -1);
      else
	gradUzp = yeoPowerCombine2Tensors<doubleTensor>(gradUzp, varValue, 1, -1);
      }

    doubleTensor gradVzp;
    if(fizero == 1 || mizero == 1)
      gradVzp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradVzp.SetVnlMatrix(Rt_M_dRdVzp.transpose() + Rt_M_dRdVzp);

      if(this->m_UseGradientType3 == Tom)
	gradVzp = yeoPowerCombine2Tensors<doubleTensor>(gradVzp, stdValue, 1, -1);
      else
	gradVzp = yeoPowerCombine2Tensors<doubleTensor>(gradVzp, varValue, 1, -1);
      }

    doubleTensor gradWzp;
    if(fizero == 1 || mizero == 1)
      gradWzp = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradWzp.SetVnlMatrix(Rt_M_dRdWzp.transpose() + Rt_M_dRdWzp);

      if(this->m_UseGradientType3 == Tom)
	gradWzp = yeoPowerCombine2Tensors<doubleTensor>(gradWzp, stdValue, 1, -1);
      else
	gradWzp = yeoPowerCombine2Tensors<doubleTensor>(gradWzp, varValue, 1, -1);
      }

    doubleTensor gradUzm;
    if(fizero == 1 || mizero == 1)
      gradUzm = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradUzm.SetVnlMatrix(Rt_M_dRdUzm.transpose() + Rt_M_dRdUzm);

      if(this->m_UseGradientType3 == Tom)
	gradUzm = yeoPowerCombine2Tensors<doubleTensor>(gradUzm, stdValue, 1, -1);
      else
	gradUzm = yeoPowerCombine2Tensors<doubleTensor>(gradUzm, varValue, 1, -1);
      }

    doubleTensor gradVzm;
    if(fizero == 1 || mizero == 1)
      gradVzm = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradVzm.SetVnlMatrix(Rt_M_dRdVzm.transpose() + Rt_M_dRdVzm);

      if(this->m_UseGradientType3 == Tom)
	gradVzm = yeoPowerCombine2Tensors<doubleTensor>(gradVzm, stdValue, 1, -1);
      else
	gradVzm = yeoPowerCombine2Tensors<doubleTensor>(gradVzm, varValue, 1, -1);
      }

    doubleTensor gradWzm;
    if(fizero == 1 || mizero == 1)
      gradWzm = NumericTraits<doubleTensor>::Zero;
    else
      {
      gradWzm.SetVnlMatrix(Rt_M_dRdWzm.transpose() + Rt_M_dRdWzm);

      if(this->m_UseGradientType3 == Tom)
	gradWzm = yeoPowerCombine2Tensors<doubleTensor>(gradWzm, stdValue, 1, -1);
      else
      	gradWzm = yeoPowerCombine2Tensors<doubleTensor>(gradWzm, varValue, 1, -1);
      }

/*
    if(this->m_UseGradientType == Symmetric && this->m_UseGradientType3 == Tom)
      { //tom symmetric rotation
      gradUxp = gradUxp/2.0;
      gradVxp = gradVxp/2.0;
      gradWxp = gradWxp/2.0;
      gradUxm = gradUxm/2.0;
      gradVxm = gradVxm/2.0;
      gradWxm = gradWxm/2.0;
      gradUyp = gradUyp/2.0;
      gradVyp = gradVyp/2.0;
      gradWyp = gradWyp/2.0;
      gradUym = gradUym/2.0;
      gradVym = gradVym/2.0;
      gradWym = gradWym/2.0;
      gradUzp = gradUzp/2.0;
      gradVzp = gradVzp/2.0;
      gradWzp = gradWzp/2.0;
      gradUzm = gradUzm/2.0;
      gradVzm = gradVzm/2.0;
      gradWzm = gradWzm/2.0;
      }
*/

/*===============================================
* Insert Rotation Grad into Sparse Matrix
===============================================*/

    if(m_UseGradientType3 == Tom)
      {
      long position = m_ForegroundRef->GetPixel(index);
      m_GradLock.Lock();

      tmpIndex = index;
  
/*
    IndexType ReferenceIndex = LastIndex;
    ReferenceIndex[0] = 10; //ReferenceIndex[0]/2;
    ReferenceIndex[1] = 12; //ReferenceIndex[1]/2;
    ReferenceIndex[2] = 0; //ReferenceIndex[2]/2;
  */
  
      //xm
      tmpIndex[0] -= 1;

      if(tmpIndex[0] >= FirstIndex[0])
	{
	long pos_xm = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_xm >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUxm,gradVxm, gradWxm), 
	     6*position,
	     3*pos_xm);
	  }
 
	/*     if(ReferenceIndex == index)
	{
	std::cout<<"ReferenceIndex: "<<ReferenceIndex<<std::endl;
	std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUxm,gradVxm, gradWxm)<<std::endl;
	std::cout<<"position: "<<position<<std::endl;
	std::cout<<"6*position  : "<<6*position<<std::endl;
	std::cout<<"3*pos_xm: "<<3*pos_xm<<std::endl;
	}*/
	}

      //xp
      tmpIndex[0] += 2;
    
      if(tmpIndex[0] < LastIndex[0])
	{
	long pos_xp = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_xp >=0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad,
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUxp, gradVxp, gradWxp),
	     6*position, 
	     3*pos_xp);
	  }

	/*  if(ReferenceIndex == index)
	{
	std::cout<<"ReferenceIndex: "<<ReferenceIndex<<std::endl;
	std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUxp, gradVxp, gradWxp)<<std::endl;
	std::cout<<"position: "<<position<<std::endl;
	std::cout<<"6*position  : "<<6*position<<std::endl;
	std::cout<<"3*pos_xp: "<<3*pos_xp<<std::endl;
	}*/
	}
      tmpIndex[0] -= 1;


      //ym
      tmpIndex[1] -= 1;

      if(tmpIndex[1] >= FirstIndex[1])
	{
	long pos_ym = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_ym >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUym,gradVym, gradWym), 
	     6*position,
	     3*pos_ym);
	  }
 
	/* if(ReferenceIndex == index)
	{
	std::cout<<"ReferenceIndex: "<<ReferenceIndex<<std::endl;
	std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUym, gradVym, gradWym)<<std::endl;
	std::cout<<"position: "<<position<<std::endl;
	std::cout<<"6*position  : "<<6*position<<std::endl;
	std::cout<<"3*pos_ym: "<<3*pos_ym<<std::endl;
	}*/
	}

  
      //yp
      tmpIndex[1] += 2;
    
      if(tmpIndex[1] < LastIndex[1])
	{
	long pos_yp = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_yp >=0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad,
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUyp, gradVyp, gradWyp),
	     6*position, 
	     3*pos_yp);
	  }

 
	/*  if(ReferenceIndex == index)
	{
	std::cout<<"ReferenceIndex: "<<ReferenceIndex<<std::endl;
	std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUyp, gradVyp, gradWyp)<<std::endl;
	std::cout<<"position: "<<position<<std::endl;
	std::cout<<"6*position  : "<<6*position<<std::endl;
	std::cout<<"3*pos_yp: "<<3*pos_yp<<std::endl;
	}*/
	}
      tmpIndex[1] -= 1;

   


      //zm
      tmpIndex[2] -= 1;

      if(tmpIndex[2] >= FirstIndex[2])
	{
	long pos_zm = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_zm >= 0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad, 
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUzm,gradVzm, gradWzm), 
	     6*position,
	     3*pos_zm);
	  }
  
	/*    if(ReferenceIndex == index)
	{
	std::cout<<"ReferenceIndex: "<<ReferenceIndex<<std::endl;
	std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUzm, gradVzm, gradWzm)<<std::endl;
	std::cout<<"position: "<<position<<std::endl;
	std::cout<<"6*position  : "<<6*position<<std::endl;
	std::cout<<"3*pos_zm: "<<3*pos_zm<<std::endl;
	}*/
	}

   
      //zp
      tmpIndex[2] += 2;
    
      if(tmpIndex[2] < LastIndex[2])
	{
	long pos_zp = m_ForegroundRef->GetPixel(tmpIndex);
	if(pos_zp >=0)
	  {
	  yeoInsertTensorGradientIntoSparseMatrix<SparseWriteMatrixType, TensorGradientType>
	    (m_Grad,
	     yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUzp, gradVzp, gradWzp),
	     6*position, 
	     3*pos_zp);
	  }

   
	/*  if(ReferenceIndex == index)
	{
	std::cout<<"ReferenceIndex: "<<ReferenceIndex<<std::endl;
	std::cout<<"Insert: "<< yeoCreateMatrixFromTensor<TensorGradientType, doubleTensor>(gradUzp, gradVzp, gradWzp)<<std::endl;
	std::cout<<"position: "<<position<<std::endl;
	std::cout<<"6*position  : "<<6*position<<std::endl;
	std::cout<<"3*pos_zp: "<<3*pos_zp<<std::endl;
	}*/
      
	}
      tmpIndex[2] -= 1;
    
  


//Fill the diagonals of m_Grad
      double m_Normalizer_sqrt_inv = 1.0/sqrt(m_Normalizer);
      long numCols = gmm::mat_ncols(*m_Grad);
    
      (*m_Grad)(2*numCols+3*position+0, 3*position+0) = m_Normalizer_sqrt_inv;
      (*m_Grad)(2*numCols+3*position+1, 3*position+1) = m_Normalizer_sqrt_inv;
      (*m_Grad)(2*numCols+3*position+2, 3*position+2) = m_Normalizer_sqrt_inv;
    
      m_GradLock.Unlock();
      } //end computation of rotation part for Symmetric Rotation Tom
    else if(this->m_UseGradientType3 == ConjugateGradient)
      {

      doubleUpdateType update_xm;
      update_xm[0] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradUxm, 1, 1));
      update_xm[1] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradVxm, 1, 1));
      update_xm[2] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradWxm, 1, 1));
					       
      doubleUpdateType update_xp;
      update_xp[0] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradUxp, 1, 1));
      update_xp[1] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradVxp, 1, 1));
      update_xp[2] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradWxp, 1, 1));

      doubleUpdateType update_ym;
      update_ym[0] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradUym, 1, 1));
      update_ym[1] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradVym, 1, 1));
      update_ym[2] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradWym, 1, 1));

      doubleUpdateType update_yp;
      update_yp[0] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradUyp, 1, 1));
      update_yp[1] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradVyp, 1, 1));
      update_yp[2] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradWyp, 1, 1));

      doubleUpdateType update_zm;
      update_zm[0] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradUzm, 1, 1));
      update_zm[1] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradVzm, 1, 1));
      update_zm[2] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradWzm, 1, 1));
					       
      doubleUpdateType update_zp;
      update_zp[0] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradUzp, 1, 1));
      update_zp[1] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradVzp, 1, 1));
      update_zp[2] = 2.0 * yeoSumTensor<doubleTensor>(yeoPowerCombine2Tensors<doubleTensor>(speedValue, gradWzp, 1, 1));

  

      m_UpdateFieldLock.Lock();      
      IndexType tmpIndex = index;
      
//xm
      tmpIndex[0] -= 1;

      if(tmpIndex[0] >= FirstIndex[0])
	{
	update_xm += this->m_UpdateField->GetPixel(tmpIndex);
	this->m_UpdateField->SetPixel(tmpIndex, update_xm);
	}

      //xp
      tmpIndex[0] += 2;
    
      if(tmpIndex[0] < LastIndex[0])
	{
	update_xp += this->m_UpdateField->GetPixel(tmpIndex);
	this->m_UpdateField->SetPixel(tmpIndex, update_xp);
	}
      tmpIndex[0] -= 1;


      //ym
      tmpIndex[1] -= 1;

      if(tmpIndex[1] >= FirstIndex[1])
	{
	update_ym += this->m_UpdateField->GetPixel(tmpIndex);
	this->m_UpdateField->SetPixel(tmpIndex, update_ym);
	}

  
      //yp
      tmpIndex[1] += 2;
    
      if(tmpIndex[1] < LastIndex[1])
	{
	update_yp += this->m_UpdateField->GetPixel(tmpIndex);
	this->m_UpdateField->SetPixel(tmpIndex, update_yp);
	}
      tmpIndex[1] -= 1;   


      //zm
      tmpIndex[2] -= 1;

      if(tmpIndex[2] >= FirstIndex[2])
	{
	update_zm += this->m_UpdateField->GetPixel(tmpIndex);
	this->m_UpdateField->SetPixel(tmpIndex, update_zm);
	}
   
      //zp
      tmpIndex[2] += 2;
    
      if(tmpIndex[2] < LastIndex[2])
	{
	update_zp += this->m_UpdateField->GetPixel(tmpIndex);
	this->m_UpdateField->SetPixel(tmpIndex, update_zp);
	}
      tmpIndex[2] -= 1;

      m_UpdateFieldLock.Unlock();

      }
    else
      {
      std::cout<<"Xavier with Rotation Gradient not implemented!"<<std::endl;
      exit(EXIT_FAILURE);
      }
    } //end true rotation computation when not fixed 
  else
    {
    if(m_UseGradientType != Fixed && m_UseGradientType2 != NoRotation)
      {
      std::cout<<"Should be fixed gradient type or no rotation"<<std::endl;
      exit(EXIT_FAILURE);
      }
    }
    
/*===================================================
* Compute second order update.
======================================================*/
  if ( yeoSumTensor<doubleTensor>(absSpeedValue) <= m_IntensityDifferenceThreshold
       || (this->m_UseGradientType2 != NoRotation && this->m_UseGradientType3 == Tom) || this->m_UseGradientType3 == ConjugateGradient)
    {
    update.Fill( 0.0 );
    }
  else
    {       

    if(m_UseGradientType3 == Xavier) 
      {
      /*===================================
*Compute Xavier's 2nd order Approximation
==========================================*/
      double denom;
      if (  m_Normalizer > 0.0 )//compute normalization
	{
#ifdef DEBUG
	if(index == ReferenceIndex)
	  {
	  std::cout<<"Thirion Normalization"<<usedDoubleGradient<<std::endl;
	  }
#endif    	  
	// "ITK-Thirion" normalization
	denom = H + 1.0/m_Normalizer;
	}
      else
	{
	std::cerr<<"Should always use ITK-Thirion!!"<<std::endl;
	exit(EXIT_FAILURE);
	}
      
      if ( denom < m_DenominatorThreshold ) //apply normalization
	{
	update.Fill( 0.0 );
	}
      else
	{
	
	for( unsigned int j = 0; j < ImageDimension; j++ )
	  {
	  update[j] = temp_update[j] / denom;
	  }
	
#ifdef DEBUG
	if(index == ReferenceIndex)
	  {
	  std::cout<<"m_Normalizer: "<<m_Normalizer<<std::endl;
	  std::cout<<"Denom: "<<denom<<std::endl;
	  std::cout<<"True Gradient: "<<usedDoubleGradient<<std::endl;
	  std::cout<<"temp_update: "<<temp_update<<std::endl;
	  std::cout<<"Final Update: "<<update<<std::endl;
	  std::cout<<"********************"<<std::endl;
	  }
#endif    
	}
      } //end Xavier's computation of update.
    else if(m_UseGradientType3 == Tom) //here should be Tom's no rotation
      {
      if(this->m_UseGradientType2 != NoRotation)
	{
	std::cerr<<"Should not be here when NoRotation"<<std::endl;
	exit(EXIT_FAILURE);
	}
/*===================================================
* Compute Tom's least-square update. 
======================================================*/
      
#ifdef DEBUG
      if(index == ReferenceIndex)
	{
	std::cout<<"**************************"<<std::endl;
	std::cout<<"temp_update: "<<temp_update<<std::endl;
	std::cout<<"hessian: "<<hessian<<std::endl;
	}
#endif
	
      //solving using cholesky
      vnl_cholesky C(hessian.GetVnlMatrix());
      vnl_vector<double> augmented_update;
      C.solve(temp_update.GetVnlVector(), &augmented_update);
	
      for( unsigned int j = 0; j < ImageDimension; j++ )
	{
	update[j] = augmented_update[j];
	}
	
#ifdef DEBUG
      if(index == ReferenceIndex)
	{
	std::cout<<"**************************"<<std::endl;
	std::cout<<"update: "<<update<<std::endl;
	}
#endif

      }
    else
      {
      std::cout<<"**Should not reach here**"<<std::endl;
      exit(EXIT_FAILURE);
      }
    }//end if statement on normalization
    
    
    
    
  // WARNING!! We compute the global data without taking into account the current update step.
  // There are several reasons for that: If an exponential, a smoothing or any other operation
  // is applied on the update field, we cannot compute the newMappedCenterPoint here; and even
  // if we could, this would be an often unnecessary time-consuming
  // task.
  if(fizero == 0 && mizero == 0)
    {
    //only update pixels which are both non background
    if ( globalData )
      {
	
      globalData->m_SumOfSquaredDifference += yeoSumTensor<doubleTensor>(sqSpeedValue);
      globalData->m_NumberOfPixelsProcessed += 1;
      globalData->m_SumOfSquaredChange += update.GetSquaredNorm();
	
	
#ifdef DEBUG 
      if(index == ReferenceIndex)
	{
	std::cout<<"Speed Value:  "<<speedValue<<std::endl;
	std::cout<<"Abs Speed Value:  "<<absSpeedValue<<std::endl;
	std::cout<<"Sq Speed Value:  "<<sqSpeedValue<<std::endl;
	std::cout<<"Var Value:  "<<varValue<<std::endl;
	std::cout<<"Sum Sq Speed Value:  "<<yeoSumTensor<doubleTensor>(sqSpeedValue)<<std::endl;
	std::cout<<"********************************"<<std::endl;
	}
#endif
	
      if(m_useAtlas)
	{
	//const doubleTensor varValue =  static_cast<doubleTensor>(GetVarImage()->GetPixel(index));
	doubleTensor normalizedSqSpeedValue = yeoPowerCombine2Tensors<doubleTensor>(sqSpeedValue, varValue, 1, -1);
	globalData->m_NormalizedSumOfSquaredDifference +=  yeoSumTensor<doubleTensor>(normalizedSqSpeedValue);
	  
#ifdef DEBUG 
	if(index == ReferenceIndex)
	  {
	  std::cout<<"Normalized Sq Speed Value:  "<<normalizedSqSpeedValue<<std::endl;
	  std::cout<<"Sum normalized Sq Speed Value:  "<<yeoSumTensor<doubleTensor>(normalizedSqSpeedValue)<<std::endl;
	  std::cout<<"********************************"<<std::endl;
	  }
#endif
	}
      else
	{
	globalData->m_NormalizedSumOfSquaredDifference +=  yeoSumTensor<doubleTensor>(sqSpeedValue);
	}
      }
    }
    
  return update;
    
}


/*
 * Update the metric and release the per-thread-global data.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::ReleaseGlobalDataPointer( void *gd ) const
{
  GlobalDataStruct * globalData = (GlobalDataStruct *) gd;

  m_MetricCalculationLock.Lock();
  m_SumOfSquaredDifference  += globalData->m_SumOfSquaredDifference;
  m_NumberOfPixelsProcessed += globalData->m_NumberOfPixelsProcessed;
  m_SumOfSquaredChange += globalData->m_SumOfSquaredChange;

  m_NormalizedSumOfSquaredDifference += globalData->m_NormalizedSumOfSquaredDifference;

  //std::cout<<"Number of Pixels:  "<<m_NumberOfPixelsProcessed<<std::endl;
  //std::cout<<"SumOfSquare:  "<<m_SumOfSquaredDifference<<std::endl;
  //std::cout<<"SumOfSquareGlobal: "<<globalData->m_SumOfSquaredDifference<<std::endl;
  if ( m_NumberOfPixelsProcessed )
    {
    m_Metric = m_SumOfSquaredDifference / 
      static_cast<double>( m_NumberOfPixelsProcessed ); 
    m_RMSChange = vcl_sqrt( m_SumOfSquaredChange / 
			    static_cast<double>( m_NumberOfPixelsProcessed ) );
    m_NormalizedMetric = m_NormalizedSumOfSquaredDifference / 
      static_cast<double>( m_NumberOfPixelsProcessed ); 
    //std::cout<<m_Metric<<std::endl;
    }
  m_MetricCalculationLock.Unlock();

  delete globalData;
}



template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::PrintMSE(std::string filename) const
{
  typedef FixedImageType GenericImageType;
  typename GenericImageType::ConstPointer fixedImage  
    =  static_cast<typename GenericImageType::ConstPointer> (this->GetFixedImage());
  this->m_MovingImageWarper->Update();
  typename GenericImageType::ConstPointer movingImage2 
    =  static_cast<typename GenericImageType::ConstPointer> (this->m_MovingImageWarper->GetOutput());

  if(fixedImage->GetLargestPossibleRegion().GetSize() != movingImage2->GetLargestPossibleRegion().GetSize())
    {
    std::cerr<<"Not the same size!!"<<std::endl;
    }
  
  float numPixels = (float) fixedImage->GetRequestedRegion().GetNumberOfPixels();

  typename GenericImageType::Pointer movingImage = GenericImageType::New();
  yeoCopyImage<GenericImageType>(movingImage2, movingImage);

  typedef typename itk::ImageRegionIterator< GenericImageType> IteratorType;
  IteratorType It1 = IteratorType(movingImage, movingImage->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< GenericImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(fixedImage, fixedImage->GetLargestPossibleRegion());
  
  for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
    {
    if(It1.Get() == NumericTraits<typename MovingImageType::PixelType>::Zero)
      {
      It1.Set(It2.Get());
      numPixels--;
      }
    }

/*
  typename GenericImageType::Pointer diffImage = GenericImageType::New();
  //difference
  yeoLinearCombine2ImagesNewImage<GenericImageType>(fixedImage, movingImage, diffImage, 1, -1);
  //square
  yeoPowerCombine2ImagesInPlace<GenericImageType>(diffImage, diffImage, 1, 1);
  //divide by numPixels
  yeoMultiplyImageByConstantInPlace<GenericImageType>(diffImage, 1/numPixels);
  //sum
  std::cout<<"MSE: "<<yeoSumImage<GenericImageType>(diffImage)<<std::endl;

  if(!filename.length()==0)
    {
    yeoWriteImage<GenericImageType>(diffImage, filename);
    }
  
*/
}

template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::PrintWMSE(std::string filename1, std::string filename2, std::string filename3) const
{
  typedef FixedImageType GenericImageType;
  typename GenericImageType::ConstPointer fixedImage  
    =  static_cast<typename GenericImageType::ConstPointer> (this->GetFixedImage());
  this->m_MovingImageWarper->Update();
  typename GenericImageType::ConstPointer movingImage2 
    =  static_cast<typename GenericImageType::ConstPointer> (this->m_MovingImageWarper->GetOutput());
  
  if(fixedImage->GetLargestPossibleRegion().GetSize() != movingImage2->GetLargestPossibleRegion().GetSize())
    {
    std::cerr<<"Not the same size!!"<<std::endl;
    }
  
  float numPixels = (float) fixedImage->GetRequestedRegion().GetNumberOfPixels();
  
  typename GenericImageType::Pointer movingImage = GenericImageType::New();
  yeoCopyImage<GenericImageType>(movingImage2, movingImage);
  
  typedef typename itk::ImageRegionIterator< GenericImageType> IteratorType;
  IteratorType It1 = IteratorType(movingImage, movingImage->GetLargestPossibleRegion());
  
  typedef typename itk::ImageRegionConstIterator< GenericImageType> ConstIteratorType;
  ConstIteratorType It2 = ConstIteratorType(fixedImage, fixedImage->GetLargestPossibleRegion());
  
  for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
    {
    if(It1.Get() == NumericTraits<typename MovingImageType::PixelType>::Zero)
      {
      It1.Set(It2.Get());
      numPixels--;
      }
    }

  /*
  typename GenericImageType::Pointer diffImage = GenericImageType::New();
  //difference
  yeoLinearCombine2ImagesNewImage<GenericImageType>(fixedImage, movingImage, diffImage, 1, -1);
  //square
  yeoPowerCombine2TensorImagesInPlace<GenericImageType>(diffImage, diffImage, 1, 1);
  //divide by var image
  yeoPowerCombine2TensorImagesInPlace<GenericImageType>(diffImage, m_CacheVarianceImage, 1, -1);
 
typedef itk::Image<float, 3> floatImageType;
typename floatImageType::Pointer SqNormImage = floatImageType::New();
yeoSqNormTensorImageNewImage<GenericImageType, floatImageType>(diffImage, SqNormImage); 

  if(!filename1.length()==0)
    {
    yeoWriteTensorImage<GenericImageType>(diffImage, filename1);
    }

  //divide by numPixels
  yeoMultiplyImageByConstantInPlace<GenericImageType>(diffImage, 1/numPixels);
  //sum
  std::cout<<"WMSE: "<<yeoSumImage<GenericImageType>(diffImage)<<std::endl;
  
 
  if(!filename2.length()==0)
    {
    yeoWriteImage<GenericImageType>(m_CacheVarianceImage, filename2);
    }
  if(!filename3.length()==0)
    {
    typename VarImageType::Pointer GradMagImage = VarImageType::New();
    yeoGradientMagnitudeNewImage<VarImageType,itkGetStaticConstMacro(ImageDimension) >(m_CacheVarianceImage, GradMagImage);
    yeoWriteImage<GenericImageType>(GradMagImage, filename3);
    }
*/
}

template <class TFixedImage, class TMovingImage, class TDeformationField>
void
ESMDemonsRegistrationFunctionTensor<TFixedImage,TMovingImage,TDeformationField>
::UpdateVarianceImage(void) 
{
  if(m_useAtlas)
    {
    yeoCopyImage<VarImageType>(m_VarImage, m_CacheVarianceImage);
    }
  else
    {

    std::cerr<<"Trying to set m_CacheVarianceImage when atlas is not used."<<std::endl;
    exit(EXIT_FAILURE);
/*
    //Read fixed image and moving image
    typedef FixedImageType GenericImageType;
    typename GenericImageType::ConstPointer fixedImage  
      =  static_cast<typename GenericImageType::ConstPointer> (this->GetFixedImage());
    
    this->m_MovingImageWarper->Update(); //need to update warper!
    typename GenericImageType::ConstPointer movingImage2 
      =  static_cast<typename GenericImageType::ConstPointer> (this->m_MovingImageWarper->GetOutput());
    
    if(fixedImage->GetLargestPossibleRegion().GetSize() != movingImage2->GetLargestPossibleRegion().GetSize())
      {
      std::cerr<<"Not the same size!!"<<std::endl;
      }

    //Need to reset points of warped moving image which are undefined
    typename GenericImageType::Pointer movingImage = GenericImageType::New();
    yeoCopyImage<GenericImageType>(movingImage2, movingImage);
    
    typedef typename itk::ImageRegionIterator< GenericImageType> IteratorType;
    IteratorType It1 = IteratorType(movingImage, movingImage->GetLargestPossibleRegion());
    
    typedef typename itk::ImageRegionConstIterator< GenericImageType> ConstIteratorType;
    ConstIteratorType It2 = ConstIteratorType(fixedImage, fixedImage->GetLargestPossibleRegion());
    
    for(It1.GoToBegin(), It2.GoToBegin(); !It1.IsAtEnd(); ++It1, ++It2)
      {
      if(It1.Get() == NumericTraits<typename MovingImageType::PixelType>::Zero)
	{
	It1.Set(It2.Get()+1); //0 Variance is a bad idea, since later
			      //on you need to divide by the variance.
	}
      if( fabs(It1.Get() - It2.Get()) <= this->m_IntensityDifferenceThreshold)
	{
	It1.Set(It2.Get()+this->m_IntensityDifferenceThreshold); //Set
								 //limit of variance
	}
      }

    //difference
    yeoLinearCombine2ImagesNewImage<GenericImageType>(fixedImage, movingImage, m_CacheVarianceImage, 1, -1);
    //square
    yeoPowerCombine2TensorImagesInPlace<GenericImageType>(m_CacheVarianceImage, m_CacheVarianceImage, 1, 1);
*/    
    }
}



} // end namespace itk

#endif



