#ifndef _itkMultiResolutionPDEDeformableRegistration2Atlas_txx
#define _itkMultiResolutionPDEDeformableRegistration2Atlas_txx
#include "itkMultiResolutionPDEDeformableRegistration2Atlas.h"

//#include "itkRecursiveMultiResolutionPyramidImageFilter.h"
#include "itkImageRegionIterator.h"
#include "vnl/vnl_math.h"

#include "itkVectorLinearInterpolateNearestNeighborExtrapolateImageFunction.h"


#include "Debug.h"

namespace itk {

/*
 * Default constructor
 */
template <class TFixedImage, class TMovingImage, class TDeformationField, class TRealType>
MultiResolutionPDEDeformableRegistration2Atlas<TFixedImage,TMovingImage,TDeformationField,TRealType>
::MultiResolutionPDEDeformableRegistration2Atlas()
{

  m_VarImagePyramid     = VarImagePyramidType::New();
  m_VarImagePyramid->SetNumberOfLevels(this->GetNumberOfLevels() );

#ifdef DEBUG
  //PrintSelf(std::cout, 3);
  std::cout<<"Calling Constructor of MultiResolutionPDEDeformableRegistration2Atlas"<<std::endl;
#endif
}

/*
 * Standard PrintSelf method.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField, class TRealType>
void
MultiResolutionPDEDeformableRegistration2Atlas<TFixedImage,TMovingImage,TDeformationField,TRealType>
::PrintSelf(std::ostream& os, Indent indent) const
{
  Superclass::PrintSelf(os, indent);

  os << indent << "VarImagePyramid: ";
  os << m_VarImagePyramid.GetPointer() << std::endl;
}


/*
 * Set the fixed image.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField, class TRealType>
void
MultiResolutionPDEDeformableRegistration2Atlas<TFixedImage,TMovingImage,TDeformationField,TRealType >
::SetVarImage(
const VarImageType * ptr )
{
  this->ProcessObject::SetNthInput( 3, const_cast< VarImageType * >( ptr ) );
}


/*
 * Get the fixed image.
 */
template <class TFixedImage, class TMovingImage, class TDeformationField, class TRealType>
const typename MultiResolutionPDEDeformableRegistration2Atlas<TFixedImage,TMovingImage,TDeformationField,TRealType>
::VarImageType *
MultiResolutionPDEDeformableRegistration2Atlas<TFixedImage,TMovingImage,TDeformationField,TRealType>
::GetVarImage(void) const
{
  return dynamic_cast< const VarImageType * >
    ( this->ProcessObject::GetInput( 3 ) );
}

/*
 * Set the number of multi-resolution levels
 */
template <class TFixedImage, class TMovingImage, class TDeformationField, class TRealType>
void
MultiResolutionPDEDeformableRegistration2Atlas<TFixedImage,TMovingImage,TDeformationField,TRealType>
::SetNumberOfLevels(
unsigned int num )
{
  Superclass::SetNumberOfLevels(num);

  if(m_VarImagePyramid && m_VarImagePyramid->GetNumberOfLevels()!=num)
    {
    m_VarImagePyramid->SetNumberOfLevels(this->m_NumberOfLevels);
    }
}


/*
 * Perform a the deformable registration using a multiresolution scheme
 * using an internal mini-pipeline
 *
 *  ref_pyramid ->  registrator  ->  field_expander --|| tempField
 * test_pyramid ->           |                              |
 *                           |                              |
 *                           --------------------------------    
 *
 * A tempField image is used to break the cycle between the
 * registrator and field_expander.
 *
 */                              
template <class TFixedImage, class TMovingImage, class TDeformationField, class TRealType>
void
MultiResolutionPDEDeformableRegistration2Atlas<TFixedImage,TMovingImage,TDeformationField,TRealType>
::GenerateData()
{

  // Check for NULL images and pointers
  MovingImageConstPointer movingImage = this->GetMovingImage();
  FixedImageConstPointer  fixedImage = this->GetFixedImage();
  VarImageConstPointer    varImage = this->GetVarImage();


  if(varImage)
    {
    std::cout<<"Var Image Exists, Will Perform Pyramid"<<std::endl;

#ifdef DEBUG
    typename VarImageType::IndexType referenceIndex = varImage->GetLargestPossibleRegion().GetIndex() +  varImage->GetLargestPossibleRegion().GetSize();

    referenceIndex[0] = referenceIndex[0]/2;
    referenceIndex[1] = referenceIndex[1]/2;
    referenceIndex[2] = referenceIndex[2]/2;
    std::cout<<"***********"<<std::endl;
    std::cout<<"VarImage["<<referenceIndex[0]<<","<<referenceIndex[1]<<","<<referenceIndex[2]<<"]: "<<varImage->GetPixel(referenceIndex)<<std::endl;
    std::cout<<"***********"<<std::endl;
#endif
    }  
  else
    std::cout<<"Var Image Does Not Exist, Will Not Perform Pyramid"<<std::endl;

  if( !movingImage || !fixedImage )
    {
    itkExceptionMacro( << "Fixed and/or moving image not set" );
    }

  if( !this->m_MovingImagePyramid || !this->m_FixedImagePyramid )
    {
    itkExceptionMacro( << "Fixed and/or moving pyramid not set" );
    }

  if( !this->m_RegistrationFilter )
    {
    itkExceptionMacro( << "Registration filter not set" );
    }
  
  if(varImage && !m_VarImagePyramid)
    {
    itkExceptionMacro( << "Var pyramid not set" );
    }


  // Create the image pyramids.
  this->m_MovingImagePyramid->SetInput( movingImage );
  this->m_MovingImagePyramid->UpdateLargestPossibleRegion();

  this->m_FixedImagePyramid->SetInput( fixedImage );
  this->m_FixedImagePyramid->UpdateLargestPossibleRegion();

  if(varImage)
    {
    m_VarImagePyramid->SetInput(varImage);
    m_VarImagePyramid->UpdateLargestPossibleRegion();
    }

  // Initializations
  this->m_CurrentLevel = 0;
  this->m_StopRegistrationFlag = false;

  unsigned int movingLevel = vnl_math_min( (int) this->m_CurrentLevel, 
    (int) this->m_MovingImagePyramid->GetNumberOfLevels() );

  unsigned int fixedLevel = vnl_math_min( (int) this->m_CurrentLevel, 
    (int) this->m_FixedImagePyramid->GetNumberOfLevels() );

  unsigned int varLevel;
  if(varImage)
    {
    varLevel = vnl_math_min( (int) this->m_CurrentLevel, (int) m_VarImagePyramid->GetNumberOfLevels() );
    }


  DeformationFieldPointer tempField = NULL;
  bool lastShrinkFactorsAllOnes = false;

  while ( !this->Halt() )
    {
   
#ifdef DEBUG
    if(tempField.IsNull())
      std::cout<<"temp field is: Null"<<std::endl;
    else
      std::cout<<"temp field is not NULL"<<std::endl;
#endif
    if( this->m_CurrentLevel == 0 )
      {
       // TODO: What to do if there is an input deformation field?
       // Will need a VectorMultiResolutionPyramidImageFilter to downsample it.
      this->m_RegistrationFilter->SetInitialDeformationField( NULL );

      }
    else
      {
      // Resample the field to be the same size as the fixed image
      // at the current level
#ifdef DEBUG
      std::cout<<"Resampling Deformation Field"<<std::endl;
#endif

      this->m_FieldExpander->SetInput( tempField );
      
      typename FloatImageType::Pointer fi = 
        this->m_FixedImagePyramid->GetOutput( fixedLevel );
      this->m_FieldExpander->SetSize( 
        fi->GetLargestPossibleRegion().GetSize() );
      this->m_FieldExpander->SetOutputStartIndex(
        fi->GetLargestPossibleRegion().GetIndex() );
      this->m_FieldExpander->SetOutputOrigin( fi->GetOrigin() );
      this->m_FieldExpander->SetOutputSpacing( fi->GetSpacing());

      this->m_FieldExpander->UpdateLargestPossibleRegion();
      this->m_FieldExpander->SetInput( NULL );
      tempField = this->m_FieldExpander->GetOutput();
      tempField->DisconnectPipeline();

      this->m_RegistrationFilter->SetInitialDeformationField( tempField );

      }
#ifdef DEBUG
    std::cout<<"Resampling Deformation Field End"<<std::endl;
#endif    

    // setup registration filter and pyramids 
    this->m_RegistrationFilter->SetMovingImage( this->m_MovingImagePyramid->GetOutput(movingLevel) );
    this->m_RegistrationFilter->SetFixedImage( this->m_FixedImagePyramid->GetOutput(fixedLevel) );
    
    

    if(varImage)
      {
      DiffeomorphicRegistrationType* temp_registration_filter = static_cast<DiffeomorphicRegistrationType *>(this->m_RegistrationFilter.GetPointer());
      temp_registration_filter->SetVarImage(m_VarImagePyramid->GetOutput(varLevel));
      }
    

    this->m_RegistrationFilter->SetNumberOfIterations(
      this->m_NumberOfIterations[this->m_CurrentLevel] );

    

    // cache shrink factors for computing the next expand factors.
    lastShrinkFactorsAllOnes = true;
    for( unsigned int idim = 0; idim < ImageDimension; idim++ )
      {
      if ( this->m_FixedImagePyramid->GetSchedule()[fixedLevel][idim] > 1 )
        {
        lastShrinkFactorsAllOnes = false;
        break;
        }
      }

    // compute new deformation field
    this->m_RegistrationFilter->UpdateLargestPossibleRegion();
    tempField = this->m_RegistrationFilter->GetOutput();
    tempField->DisconnectPipeline();

    

    // Increment level counter.  
    this->m_CurrentLevel++;
    movingLevel = vnl_math_min( (int) this->m_CurrentLevel, 
      (int) this->m_MovingImagePyramid->GetNumberOfLevels() );
    fixedLevel = vnl_math_min( (int) this->m_CurrentLevel, 
      (int) this->m_FixedImagePyramid->GetNumberOfLevels() );
    
    if(varImage)
      varLevel = vnl_math_min( (int) this->m_CurrentLevel, (int) m_VarImagePyramid->GetNumberOfLevels() );
    

    // Invoke an iteration event.
    this->InvokeEvent( IterationEvent() );

/*
    //After 1st level pyramid
    //Just to check on progress!
    DemonsRegistrationFunctionType *drfp = 
      dynamic_cast<DemonsRegistrationFunctionType *> 
      (m_RegistrationFilter2->GetDifferenceFunction().GetPointer());
    std::cout<<"==========After Update==========="<<std::endl;
    drfp->UpdateVarianceImage();
    drfp->PrintMSE("MSE3.nhdr");
    drfp->PrintWMSE("WMSE3.nhdr", "VarImage3.nhdr", "VarGradMag3.nhdr");
    
    //Compute Diff Image between MSE3-MSE1 
    typename FixedImageType::Pointer MSE1 = FixedImageType::New();
    typename FixedImageType::Pointer MSE3 = FixedImageType::New();
    yeoReadImage<FixedImageType>(MSE1, "MSE1.nhdr");
    yeoReadImage<FixedImageType>(MSE3, "MSE3.nhdr");
    
    yeoLinearCombine2ImagesInPlace<FixedImageType>(MSE3, MSE1, 1, -1);
    yeoWriteImage<FixedImageType>(MSE3, "MSE3mMSE1.nhdr");
    
    //Compute Diff Image between WMSE3-WMSE1
    typename FixedImageType::Pointer WMSE1 = FixedImageType::New();
    typename FixedImageType::Pointer WMSE3 = FixedImageType::New();
    yeoReadImage<FixedImageType>(WMSE1, "WMSE1.nhdr");
    yeoReadImage<FixedImageType>(WMSE3, "WMSE3.nhdr");
    
    yeoLinearCombine2ImagesInPlace<FixedImageType>(WMSE3, WMSE1, 1, -1);
    yeoWriteImage<FixedImageType>(WMSE3, "WMSE3mWMSE1.nhdr");
    
    //Compute diff between VarImage1 and VarImage3 (just to show
    //it's 0.
    typename FixedImageType::Pointer VarImage1 = FixedImageType::New();
    typename FixedImageType::Pointer VarImage3 = FixedImageType::New();
    yeoReadImage<FixedImageType>(VarImage1, "VarImage1.nhdr");
    yeoReadImage<FixedImageType>(VarImage3, "VarImage3.nhdr");
    
    yeoLinearCombine2ImagesInPlace<FixedImageType>(VarImage1, VarImage3, 1, -1);
    yeoAbsImageInPlace<FixedImageType>(VarImage1);
    std::cout<<"Total Difference: "<<yeoSumImage<FixedImageType>(VarImage1)<<std::endl;
    
    //Compute diff between VarImage1 and VarImage3 (just to show
    //it's 0.
    typename FixedImageType::Pointer VarGradMag1 = FixedImageType::New();
    typename FixedImageType::Pointer VarGradMag3 = FixedImageType::New();
    yeoReadImage<FixedImageType>(VarGradMag1, "VarGradMag1.nhdr");
    yeoReadImage<FixedImageType>(VarGradMag3, "VarGradMag3.nhdr");
    
    yeoLinearCombine2ImagesInPlace<FixedImageType>(VarGradMag1, VarGradMag3, 1, -1);
    yeoAbsImageInPlace<FixedImageType>(VarGradMag1);
    std::cout<<"Total Difference: "<<yeoSumImage<FixedImageType>(VarGradMag1)<<std::endl;
    
    exit(EXIT_FAILURE);
*/


    // We can release data from pyramid which are no longer required.
    if ( movingLevel > 0 )
      {
      this->m_MovingImagePyramid->GetOutput( movingLevel - 1 )->ReleaseData();
      }
    if( fixedLevel > 0 )
      {
      this->m_FixedImagePyramid->GetOutput( fixedLevel - 1 )->ReleaseData();
      }
    if(varImage)
      {
      if( varLevel > 0)
	{
	m_VarImagePyramid->GetOutput(varLevel - 1)->ReleaseData();
	}
      }


    } // while not Halt()

    if( !lastShrinkFactorsAllOnes )
      {
      // Some of the last shrink factors are not one
      // graft the output of the expander filter to
      // to output of this filter

      // resample the field to the same size as the fixed image
      this->m_FieldExpander->SetInput( tempField );
      this->m_FieldExpander->SetSize( 
        fixedImage->GetLargestPossibleRegion().GetSize() );
      this->m_FieldExpander->SetOutputStartIndex(
        fixedImage->GetLargestPossibleRegion().GetIndex() );
      this->m_FieldExpander->SetOutputOrigin( fixedImage->GetOrigin() );
      this->m_FieldExpander->SetOutputSpacing( fixedImage->GetSpacing());

      this->m_FieldExpander->UpdateLargestPossibleRegion();
      this->GraftOutput( this->m_FieldExpander->GetOutput() );
      }
    else
      {
      // all the last shrink factors are all ones
      // graft the output of registration filter to
      // to output of this filter
      this->GraftOutput( tempField );
      }

    // Release memory
    this->m_FieldExpander->SetInput( NULL );
    this->m_FieldExpander->GetOutput()->ReleaseData();
    this->m_RegistrationFilter->SetInput( NULL );
    this->m_RegistrationFilter->GetOutput()->ReleaseData();

}



} // end namespace itk

#endif
