#include "itkTensorImageIO.h"

#include "GetPot.h"

#include <itkImage.h>
#include <itkImageFileWriter.h>
#include <itkVector.h>


#include "itkLogTensorImageFilter.h"
#include "itkTensorImageToVectorImage.h"
#include "itkVectorGradientMagnitudeImageFilter.h"

//#include "itkGradientMagnitudeTensorImageFilter.h"

#include <itkDiscreteGaussianImageFilter.h>
#include <itkDerivativeImageFilter.h>
#include <itkImageRegionIteratorWithIndex.h>



void PrintHelp (const char* exec)
{
  std::cout << exec << ": " << std::endl;
  std::cout << "-i [fileIn]" << std::endl;
  std::cout << "-s [sigma]" << std::endl;
  std::cout << "-euc" << std::endl;
  std::cout << "-o [fileOut]" << std::endl;
  exit(0);
}




int main (int narg, char* arg[])
{

  itk::Object::GlobalWarningDisplayOff();
    
  // parse arguments
  GetPot cl (narg, arg);
  if( cl.size() == 1 || cl.search (2,"--help","-h") )
    PrintHelp (cl[0]);

  
  const bool IsInputPresent = cl.search(2,"-i","-I");
  const bool IsOutputPresent = cl.search(2,"-o","-O");
  const bool euclidean = cl.search (1,"-euc");
  const double sigma = cl.follow (1.0, 2, "-s", "-S");
  
  if(!IsInputPresent || !IsOutputPresent)
  {
    std::cerr << "Input file and/or output not set" << std::endl;
    return -1;
  }
  
  
  const char* file_in = cl.follow("NoFile", 2, "-i","-I");
  const char* file_out = cl.follow("NoFile", 2, "-o","-O");
  if(strcmp(file_in,"NoFile")==0 || strcmp(file_out,"NoFile")==0)
  {
    std::cerr << "Input file and/or output not set" << std::endl;
    exit (-1);
  }


  // Read the input:
  typedef double  ScalarType;
  const unsigned int ImageDimension  = 3;
  const unsigned int TensorDimension = 3;
  
  typedef itk::TensorImageIO<ScalarType, TensorDimension, ImageDimension> IOType;
  typedef IOType::TensorImageType                                         TensorImageType;


  TensorImageType::Pointer tensorImage = 0;
  {
    IOType::Pointer myReader = IOType::New();
    myReader->SetFileName (file_in);
    
    std::cout << "Reading: " << file_in;
    std::cout << std::flush;  
    try
    {
      myReader->Read();    
    }
    catch (itk::ExceptionObject &e)
    {
      std::cerr << e;
      exit (-1);    
    }
    std::cout << " Done." << std::endl;

    tensorImage = myReader->GetOutput();
    
  }
  

  /** Specific typedefs */
  typedef itk::LogTensorImageFilter<TensorImageType,TensorImageType>  LogFilterType;
  typedef itk::Image<ScalarType, ImageDimension>                      ScalarImageType;

  /*
  typedef itk::Vector<ScalarType, 6>              VectorType;
  typedef itk::Image<VectorType, ImageDimension>  VectorImageType;
  typedef itk::TensorImageToVectorImage<TensorImageType, VectorImageType>       TensorToVectorFilter;
  typedef itk::VectorGradientMagnitudeImageFilter<VectorImageType, ScalarType>  GradientFilterType;
  */

  typedef itk::DiscreteGaussianImageFilter<TensorImageType, TensorImageType> GaussianFilterType;
  typedef itk::DerivativeImageFilter<TensorImageType, TensorImageType>       DerivativeFilterType;

  /** Set up the pipeline */
  /*
    typedef itk::GradientMagnitudeTensorImageFilter<TensorImageType,ScalarImageType>
    GradientMagnitudeFilterType;
    GradientMagnitudeFilterType::Pointer myGradient = GradientMagnitudeFilterType::New();
  */

  //TensorToVectorFilter::Pointer myConverter = TensorToVectorFilter::New();
  //GradientFilterType::Pointer   myGradient  = GradientFilterType::New();
  //myGradient->SetUsePrincipleComponents (false);

  
  if( !euclidean )
  {
    LogFilterType::Pointer myLog = LogFilterType::New();
    myLog->SetInput ( tensorImage );

    try
    {
      myLog->Update();
    }
    catch (itk::ExceptionObject &e)
    {
      std::cerr << e;
      exit (-1);    
    }

    tensorImage = myLog->GetOutput();
  }



  TensorImageType::Pointer imageX = 0;
  TensorImageType::Pointer imageY = 0;
  TensorImageType::Pointer imageZ = 0;
  

  {
    GaussianFilterType::Pointer   gaussianX   = GaussianFilterType::New();
    DerivativeFilterType::Pointer derivativeX = DerivativeFilterType::New();
    gaussianX->SetVariance (sigma);
    derivativeX->SetDirection (0);

    gaussianX->SetInput (tensorImage);
    derivativeX->SetInput ( gaussianX->GetOutput() );
    
    std::cout << "Gradient X..." << std::flush;
    try
    {
      derivativeX->Update();    
    }
    catch (itk::ExceptionObject &e)
    {
      std::cerr << e;
      exit (-1);    
    }
    std::cout << "Done." << std::endl;

    imageX = derivativeX->GetOutput();
  }


  
  {
    GaussianFilterType::Pointer   gaussianY   = GaussianFilterType::New();
    DerivativeFilterType::Pointer derivativeY = DerivativeFilterType::New();
    gaussianY->SetVariance (sigma);
    derivativeY->SetDirection (1);

    gaussianY->SetInput (tensorImage);
    derivativeY->SetInput ( gaussianY->GetOutput() );
    
    std::cout << "Gradient Y..." << std::flush;
    try
    {
      derivativeY->Update();    
    }
    catch (itk::ExceptionObject &e)
    {
      std::cerr << e;
      exit (-1);    
    }
    std::cout << "Done." << std::endl;

    imageY = derivativeY->GetOutput();
  }

  

  {
    GaussianFilterType::Pointer   gaussianZ   = GaussianFilterType::New();
    DerivativeFilterType::Pointer derivativeZ = DerivativeFilterType::New();
    gaussianZ->SetVariance (sigma);
    derivativeZ->SetDirection (2);

    gaussianZ->SetInput (tensorImage);
    derivativeZ->SetInput ( gaussianZ->GetOutput() );
    
    std::cout << "Gradient Z..." << std::flush;
    try
    {
      derivativeZ->Update();    
    }
    catch (itk::ExceptionObject &e)
    {
      std::cerr << e;
      exit (-1);    
    }
    std::cout << "Done." << std::endl;

    imageZ = derivativeZ->GetOutput();
  }







  // compute the gradient norm
  
  typedef itk::ImageRegionIteratorWithIndex<TensorImageType> TensorImageIteratorType;
  TensorImageIteratorType itX (imageX, imageX->GetLargestPossibleRegion());
  TensorImageIteratorType itY (imageY, imageY->GetLargestPossibleRegion());
  TensorImageIteratorType itZ (imageZ, imageZ->GetLargestPossibleRegion());

  ScalarImageType::Pointer myGradientImage = ScalarImageType::New();
  myGradientImage->SetRegions (imageX->GetLargestPossibleRegion());
  myGradientImage->SetOrigin  (imageX->GetOrigin());
  myGradientImage->SetSpacing (imageX->GetSpacing());
  myGradientImage->Allocate();

  typedef itk::ImageRegionIteratorWithIndex<ScalarImageType> ImageIteratorType;
  ImageIteratorType itGrad(myGradientImage, myGradientImage->GetLargestPossibleRegion());

  while( !itGrad.IsAtEnd() )
  {
    double val = sqrt ( itX.Get().GetSquaredNorm() + itY.Get().GetSquaredNorm() + itZ.Get().GetSquaredNorm() );

    itGrad.Set (val);
    
    ++itGrad;
    ++itX;
    ++itY;
    ++itZ;
  }
  

  
  // save the image:
  typedef itk::ImageFileWriter<ScalarImageType> WriterType;
  WriterType::Pointer myWriter = WriterType::New();
  myWriter->SetFileName(file_out);
  myWriter->SetInput( myGradientImage );
  
  std::cout << "Writing: " << file_out;
  std::cout << std::flush;
  try
  {
    myWriter->Update();
  }
  catch(itk::ExceptionObject &e)
  {
    std::cerr << e;
    exit(-1);
  }
  std::cout << " Done." << std::endl;  
  
  
  return 0;  
}
