#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <basics/Probability.h>

namespace yav {
	GaussianDistribution::GaussianDistribution(const double _mean, const double _sigma, const double _maxSamples) 
{
meanValue=_mean;
standardDeviation=_sigma;
maxHistogramValue=_maxSamples;
minValue=meanValue-3*standardDeviation;
maxValue=meanValue+3*standardDeviation;
numberSamples=120;
}
double GaussianDistribution::getProbability(const double x) const
{
	double v=(x-meanValue)/standardDeviation;
	return exp(-v*v/2.0)/(standardDeviation*sqrt(2.0*M_PI));
}
double GaussianDistribution::getHistogramValue(const double x) const
{
	double v=(x-meanValue)/standardDeviation;
	return (maxHistogramValue*exp(-v*v/2.0)/(standardDeviation*sqrt(2.0*M_PI)));
}
GaussianMixtureDistribution::GaussianMixtureDistribution(
	const unsigned int _numberDistributions,
	const std::vector<double> _meanValue,
	const std::vector<double> _sigma, const std::vector<double> _maxHistogram) 
	: meanValues(_meanValue), standardDeviations(_sigma), maxHistogramValues(_maxHistogram)
{
	numberDistributions=_numberDistributions;
	unsigned int i;
	minValue=meanValues[0]-3*standardDeviations[0];
	maxValue=meanValues[0]+3*standardDeviations[0];
	scaleValue=maxHistogramValues[0];
	for (i=1;i<numberDistributions;++i) 
	{
		if (minValue>(meanValues[i]-3*standardDeviations[i]))
			minValue=meanValues[i]-3*standardDeviations[i];
		if (maxValue<(meanValues[i]+3*standardDeviations[i]))
			maxValue=meanValues[i]+3*standardDeviations[i];
		scaleValue+=maxHistogramValues[i];
	}
	meanValue=0;
	standardDeviation=0;
	for(i=0;i<numberDistributions;++i)
	{
		meanValue+= meanValues[i]*maxHistogramValues[i];
	}
	meanValue/=scaleValue;
	for(i=0;i<numberDistributions;++i)
	{
		standardDeviation+= (standardDeviations[i]*standardDeviations[i]+meanValue*meanValue)*maxHistogramValues[i];
	}
	standardDeviation/=scaleValue;
	standardDeviation-=meanValue*meanValue;
	standardDeviation=sqrt(standardDeviation);
}

double GaussianMixtureDistribution::getProbability(const double x) const
{
	double val,v;
	val=0;
	unsigned int i;
	for (i=0;i<numberDistributions;++i) 
	{
		v=(x-meanValues[i])/standardDeviations[i];
		val+=  maxHistogramValues[i]*exp(-v*v/2.0)/(standardDeviations[i]);
	}
	val/=(double) numberDistributions*scaleValue;
	return (val);
}
double GaussianMixtureDistribution::getHistogramValue(const double x) const
{
	unsigned int i;
	double val,v;
	val=0;
	for (i=0;i<numberDistributions;++i) 
	{
		v=(x-meanValues[i])/standardDeviations[i];
		val+= maxHistogramValues[i]*exp(-v*v/2.0)/(standardDeviations[i]);
	}
	val/=(double) numberDistributions;
	return (val);
}
/// default constructor:
HistogramDistribution::HistogramDistribution(const GenericHistogram *h, 
											 const double parzenWindow,
											 const double quantile)
{
	minValue=h->minValue;
	maxValue=h->maxValue;
	numberSamples=h->nbBuckets;

	std::vector<double> tmpHistogram;
	histogramValue.resize(h->nbBuckets+10);
	tmpHistogram.resize(h->nbBuckets+10);

		unsigned int i,k;
	double threshold,val;
	if (quantile!=0.0) {
		threshold=h->getMaxNbSamples()*quantile;
	}
	else 
		threshold= -1.0;

	int j;
	for (i=0;i<h->nbBuckets;++i) {
		val=(double)h->histogram[i];
		if (val<threshold)
			tmpHistogram[i]=0;
		else
			tmpHistogram[i]=val;
	}
	// now perform Parzen windowing with sigma value


	double totalWeight,newval,weight,x;
	unsigned int delta;
	delta=(unsigned int )((3.0*parzenWindow/h->bucketSize)+.5);
	scaleDistribution=0;
	for(i=0;i<h->nbBuckets;++i)
	{
		newval=tmpHistogram[i];
		totalWeight=1.0;
		for(j=i-1;(j>=0) && ((i-j)<delta);--j)
		{
			x=(i-j)*h->bucketSize;
			weight=exp(-0.5*x*x/(parzenWindow*parzenWindow));
			newval+=weight*tmpHistogram[j];
			totalWeight+=weight;
		}
		for(k=i+1;(k<h->nbBuckets) && ((k-i)<delta);++k)
		{
			x=(k-i)*h->bucketSize;
			weight=exp(-0.5*x*x/(parzenWindow*parzenWindow));
			newval+=weight*tmpHistogram[k];
			totalWeight+=weight;
		}
		newval/=totalWeight;
		histogramValue[i]=  newval;
		scaleDistribution+=newval;
	}
	// normalize histogramValue to sum to 1
	meanValue=0;
	standardDeviation=0;
	double bucketSize=(maxValue-minValue)/(numberSamples);
	for(i=0;i<h->nbBuckets;++i)
	{
		histogramValue[i]/=scaleDistribution;
		meanValue+= (minValue+i*bucketSize)*histogramValue[i];
		standardDeviation += (minValue+i*bucketSize)*(minValue+i*bucketSize)*histogramValue[i];
	}
	standardDeviation-=meanValue*meanValue;
	standardDeviation=sqrt(standardDeviation);
}
double HistogramDistribution::getProbability(const double x) const
{
	if ((x<minValue)|| (x>maxValue))
		return 0.0;
	unsigned int i=(unsigned int)((numberSamples-1)*(x-minValue)/(maxValue-minValue));
	return (histogramValue[i]);
}
double HistogramDistribution::getHistogramValue(const double x) const
{
	if ((x<minValue)|| (x>maxValue))
		return 0.0;
	unsigned int i=(unsigned int)((numberSamples-1)*(x-minValue)/(maxValue-minValue));
	return (histogramValue[i]*scaleDistribution);
}
} // namespace yav

