#ifndef AUTOLINK_H
#define AUTOLINK_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/AutoLink.h"
#include <list>

namespace yav {

  /*! \class AutoLink AutoLink.h basics/AutoLink.h
    \ingroup libBasics
    parent class for every autolinkable object
    any child becomes an double-linked list whose instance automatically
    link and unlink in a static list at creation and destruction time 
  */

template<class T> class AutoLink {

public:
  /*! \class AutoWarn AutoLink.h basics/AutoLink.h
      \ingroup libBasics
      object that want to be warned when AutoLink<T> list gets empty 
  */
  class AutoWarn {
  public:
    /// constructor
    AutoWarn() { AutoLink<T>::warnMe(this); }
    /// destructor
    virtual ~AutoWarn() { AutoLink<T>::doNotWarnMe(this); }

    /// AutoLink<T> gets empty callback function
    virtual void empty(void) = 0;
  };

private:
  /// head of T objects list
  static AutoLink<T> *head;
  /// tail of T objects list
  static AutoLink<T> *tail;
  /// number of T objects
  static unsigned int count;
  /// objects to warn when list get empty
  static std::list<AutoWarn *> warn;

private:
  /// pointer to previous object in list
  AutoLink<T> *prev;
  /// pointer to following object in list
  AutoLink<T> *next;
  /// unique reference for each object
  const unsigned int ref;

public:
  /// returns list head
  static T* getHead(void) { return (T*) head; }
  /// returns list tail
  static T* getTail(void) { return (T*) tail; }
  /// returns number of objects
  static unsigned int getCount(void) { return count; }
  /** returns object whose reference is known, 0 if object does not exist
      @param n object's reference */
  static T* getByRef(unsigned int n);
  /// returns reference that will be given to next object to build
  static unsigned int nextRef(void);

  /** adds 'me' to list of object to warn when list is empty
      @param me object to warn */
  static void warnMe(AutoWarn *me) { warn.push_front(me); }
  /** removes 'me' from list of object to warn when list is empty
      @param me object to warn */
  static void doNotWarnMe(AutoWarn *me) { warn.remove(me); }


public:
  /// returns following object in list (0 if none)
  T* following(void) const { return (T*) next; }
  /// returns previous object in list (0 if none)
  T* previous(void) const { return (T*) prev; }
  /// returns this object id
  unsigned int getRef(void) const { return ref; }

protected:
  /** @name Constructors */
  //@{
  // default constructor: autolink new object at the end of list
  AutoLink();
  /** copy constructor:  autolink new object at the end of list
      @param al copied object (ignored) */
  AutoLink(const AutoLink& al) ;
  // Destructor: unlink object
  //@}

  /** @name Destructor */
  //@{
  /// removes object from list
  ~AutoLink();
  //@}

  /** copy operator
      does  nothing since one instance is discarded and another added
      @param al copied object (ignored) */
  AutoLink& operator=(const AutoLink& al);
};

/// AutoLink<T> class static variables declaration macro
// #if defined(__GNUC__) || defined(WIN32)
#define AUTOLINK_STATIC_VARIABLES(type) \
template < > AutoLink< type > *AutoLink< type >::head = 0;\
template < > AutoLink< type > *AutoLink< type >::tail = 0;\
template < > unsigned int AutoLink< type >::count = 0;\
template < > std::list< AutoLink< type >::AutoWarn * > AutoLink< type >::warn = std::list<  AutoLink< type >::AutoWarn * >();
// #else
// #define AUTOLINK_STATIC_VARIABLES(type) 
// template <typename T> AutoLink<T> *AutoLink<T>::head = 0;
// template <typename T> AutoLink<T> *AutoLink<T>::tail = 0;
// template <typename T> unsigned int AutoLink<T>::count = 0;
// #ifndef WIN32
// template <typename T> std::list<typename AutoLink< T >::AutoWarn *> AutoLink< T >::warn;
// #else
// template <typename T> std::list<typename AutoLink< T >::AutoWarn *> AutoLink< T >::warn;
// #endif
// #endif

#include "AutoLink.C"

} // namespace yav

#endif

// AutoLink< type >* AutoLink< type >::head = 0;
// AutoLink< type >* AutoLink< type >::tail = 0;
// unsigned int AutoLink< type >::count = 0;
// std::list<AutoLink< type >::AutoWarn *> AutoLink< type >::warn
