#ifndef DPLIST_H
#define DPLIST_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/DPList.h"
#include <basics/Exception.h>
#include <list>

namespace yav {

template <class T> class DPListElement;
template <class T> class DPOwner;


/*! \class DPList DPList.h basics/DPList.h
  \ingroup libBasics
  Generic list that warns list owner (DPOwner) when
  elements are added ore removed. A DPList<T> always
  owns pointers on T objects that inherit from DPListElement<T>. 
*/
template <class T> class DPList : public std::list<T *> {
 public:
  /** @name Exception classes */
  //@{
  /*! \class AlreadyInList DPList.h basics/DPList.h
    \ingroup libBasics
    attempt to add an object already in this list
  */
  class AlreadyInList : public Exception {
  public:
    AlreadyInList(const std::string& from): Exception(from,
              "Attempt to add an element that was already available in list") {
}
  };
  /*! \class NullPointer DPList.h basics/DPList.h
    \ingroup libBasics
    attempt to add a null pointer in this list
  */
  class NullPointer : public Exception {
  public:
    NullPointer(const std::string& from): Exception(from,
						    "Attempt to add NULL pointer in list") {
    }
  };
  /*! \class NotFound DPList.h basics/DPList.h
    \ingroup libBasics
    attempt to remove a pointer that is not in this list
  */
  class NotFound : public Exception {
  public:
          NotFound(const std::string& from) :
  Exception(from, "Attempt to remove an element not available in list") {
}
  };
  /*! \class Empty DPList.h basics/DPList.h
    \ingroup libBasics
    attempt to remove a pointer from an empty list
  */
  class Empty : public Exception {
  public:
    Empty(const std::string& from):
  Exception(from, "Attempt to remove an element from an empty list") {
}
  };
  //@}


 private:
  /// the object that owns this DPList
  DPOwner<T> * const owner;


 public:
  /** @name Constructors */
  //@{
  /** build a new DPList in an owner object
      @param _owner the object that owns this DPList */
  DPList(DPOwner<T> *_owner) : owner(_owner) {}

  /** build a new DPList from a list
      @param list list to copy
      @param owner the object that owns thid DPList */
  DPList(const std::list<T *>& list, DPOwner<T> *owner);
  //@}

  /** @name Destructor */
  //@{
  /// destroy each list element (but not pointed objects)
  virtual ~DPList();
  //@}

  /// DPList iterator
  typedef typename std::list<T *>::iterator iterator;
  /// DPList const iterator
  typedef typename std::list<T *>::const_iterator const_iterator;

  /** any list element being deleted call dying
      @param dyingElt element being deleted */
  void dying(T *dyingElt);

  /** appends an element at end of list
      @param t element to append
      @exception AlreadyInList cannot set element twice in same list
      @exception NullPointer   cannot set 0 as list element */
  virtual void append(T *t) throw (AlreadyInList, NullPointer);
  /** appends an element at end of list
      @param t element to append
      @exception AlreadyInList cannot set element twice in same list
      @exception NullPointer   cannot set 0 as list element */
  virtual void operator+=(T *t) throw (AlreadyInList, NullPointer) {
    DPList<T>::append(t);
  }
  /** prepends an element at list beginning
      @param t element to prepend
      @exception AlreadyInList cannot set element twice in same list
      @exception NullPointer   cannot set 0 as list element */
  virtual void prepend(T *t) throw (AlreadyInList, NullPointer);

  /** remove an element from the list
      @param t element to remove
      @exception NotFound element could not be found in list */
  virtual void remove(T *t) throw(NotFound);
  /** removes the first list element equal to t
      @param t element to remove
      @exception NotFound element could not be found in list */
  virtual void operator-=(T *t) throw(NotFound) {
    DPList<T>::remove(t);
  }
  /** removes first list element
      @exception Empty there are no elements in list */
  virtual T *removeFirst(void) throw(Empty);
  /** removes last list element
      @exception Empty there are no elements in list */
  virtual T *removeLast(void) throw(Empty);

  /** returns first list element
      @exception Empty there are no elements in list */
  virtual T *getFirst(void) throw(Empty);
  /** returns last list element
      @exception Empty there are no elements in list */
  virtual T *getLast(void) throw(Empty);

  /// returns this list owner
  const DPOwner<T> *getOwner(void) const;

  /// returns this list owner
  DPOwner<T> *getOwner(void);

};

} // namespace yav

#include <basics/DPOwner.h>
#include <basics/DPListElement.h>

#include "DPList.C"

#endif
