#ifndef DPOWNER_H
#define DPOWNER_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif

#pragma interface "basics/DPOwner.h"
#include <basics/DPList.h>

namespace yav {

template <class T> class DPList;

/*! \class DPOwner DPOwner.h basics/DPOwner.h
  \ingroup libBasics
  an object that owns a DPList (elements member)
*/
template <class T> class DPOwner {
 public:
  /** @name Constructors */
  //@{
  /// default constructor: initializes an empty DPList
  DPOwner() : elements(this) {}
  /** initializes a single element DPList
      @param elt element to put in DPList */
  DPOwner(T *elt) : elements(this) { if (elt) elements += elt; }
  /** initializes from DPList<T>
      @param list DPList<T> to copy */
  DPOwner(const DPList<T>& list) : elements(list) {}
  /** initializes from DList<T *>
      @param list DList<T*> to copy */
  DPOwner(const std::list<T *>& list) : elements(list, this) {}
  /** copy constructor: duplicates DPList
      @param e ojbect to copy */
  DPOwner(const DPOwner<T>& e) : elements(e.elements) { }
  //@}

  /** @name Destructor */
  //@{
  /// destroy DPList: pointed objects are not destroyed
  virtual ~DPOwner();
  //@}

  /** return an iterator on the first element of the DPList 
      @return a iterator on the first element of the DPList */
  virtual typename DPList<T>::const_iterator getFirst(void) const {
    return elements.begin();
  }
  /** return an iterator on the last element of the DPList 
      @return a iterator on the last element of the DPList */
  virtual typename DPList<T>::const_iterator getLast(void) const {
    return elements.end();
  }
      
    /** return an iterator on the first element of the DPList 
      @return a iterator on the first element of the DPList */
  virtual typename DPList<T>::iterator begin(void)  {
    return elements.begin();
  }
  /** return an iterator on the last element of the DPList 
      @return a iterator on the last element of the DPList */
  virtual typename DPList<T>::iterator end(void)  {
    return elements.end();
  }

  /** adds a new element in DPList
      @param elt element to add
      @exception DPList<T>::NullPointer cannot add 0 to DPList
      @exception DPList<T>::AlreadyInList cannot add same element twice in DPList */
  virtual void append(T *elt)
    throw(typename DPList<T>::NullPointer,typename DPList<T>::AlreadyInList) {
    elements.append(elt);
  }
  virtual void operator+=(T *elt)
    throw(typename DPList<T>::NullPointer, typename DPList<T>::AlreadyInList) {
    elements.append(elt);
  }

  /** removes an element from DPList
      @param elt element to remove
      @exception DPList<T>::NotFound element to remove does not exist in DPList */
  virtual void remove(T *elt) throw(typename DPList<T>::NotFound);
  /** removes an element from DPList
      @param elt element to remove
      @exception DPList<T>::NotFound element to remove does not exist in DPList */
  virtual void operator-=(T *elt) throw(typename DPList<T>::NotFound) { remove(elt); }


 private:
  /** default callback function for a DPList element being removed
      (default behaviour delete DPOwner when DPList is empty)
      @param elt element being removed */
  virtual void dying(const T *elt) {
    if(elements.size() == 0) delete this;
  }

 protected:
  /// DPList owned
  DPList<T> elements;

  friend class DPList<T>;
};

} // namespace

#include "DPOwner.C"

#endif
