#ifndef EXCEPTION_H
#define EXCEPTION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4290)
#endif

#ifndef WIN32
//#pragma interface "Exception.h"
#endif
#include <basics/Architecture.h>
#include <stdio.h>
#include <typeinfo>
#include <stdexcept>
#include <stack>
#include <cstdlib>


#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {

/*! \class Exception Exception.h basics/Exception.h
  \ingroup libBasics
  Base exception class
*/
class Exception : public std::exception {

 protected:
  /// error message
  std::string error;
  /// exception thrower stack
  std::stack<std::string> from;


  /// print exception stack on os
  virtual void print(STD_OSTREAM& os) const throw();

  /// exception occured while handling exception: print fatal message
  static void fatal(const char *from) throw();


 public:

  /** @name constructors */
  //@{
  /** new exception constructor
      @param from throwing method */
  Exception(const std::string &from) throw();

  /** new exception constructor
      @param from throwing method
      @param errorMessage exception error message */
  Exception(const std::string &from, const std::string& errorMessage) throw();

  /** copy constructor
      @param org exception to copy */
  Exception(const Exception& org) throw();
  //@}

  /** @name destructor */
  //@{
  /// destructor
  virtual ~Exception() throw();
  //@}

  /// returns error message
  const std::string& getError(void) const throw();

  /// returns exception name
  void getException(std::string&) const throw();

  /// returns stack trace
  const std::stack<std::string>& getStack(void) const throw();

  /// returns exception name string
  operator const char *() const throw();

  /// returns exception name string (overloads std::exception::what)
  virtual const char *what() const throw();

  /** push catcher name 'from' on stack
      @param from method name to push on throwers stack */
  void push(const std::string& from) throw();

  /** prints exception stack on output stream
      @param os output stream
      @param e exception to print
      @returns output stream os */
  friend STD_OSTREAM & operator<< (STD_OSTREAM & os, const Exception &e) throw();
};


// ************************************************************************
// Global Exceptions
//
// Define a bunch of general exceptions not bound to any particular class
// ************************************************************************

/*! \class ErrMemFull Exception.h basics/Exception.h
  \ingroup libBasics
  memory full exception
*/
class ErrMemFull : public Exception {
public:
  /// constructor
  ErrMemFull(const std::string &from) throw() :
    Exception(from, "Memory full: cannot allocate") {};
};

/*! \class ErrDivZero Exception.h basics/Exception.h
  \ingroup libBasics
  zero division exception
*/
class ErrDivZero : public Exception {
public:
  /// constructor
  ErrDivZero(const std::string &from) throw() :
    Exception(from, "Divide by zero exception") {};
  /// copy constructor
  ErrDivZero(const ErrDivZero &edz) throw() :
    Exception(edz) {};
  /// destructor
  virtual ~ErrDivZero() throw() {};
};

/*! \class ErrTypeMismatch Exception.h basics/Exception.h
  \ingroup libBasics
  type mismatch
 */
class ErrTypeMismatch : public Exception {
public:
  /// constructor
  ErrTypeMismatch(const std::string &from) throw() :
    Exception(from, "Type mismatch") {};
};

/*!\class ErrUnknown Exception.h basics/Exception.h
  \ingroup libBasics
  unknown error exception
*/
class ErrUnknown : public Exception {
public:
  /// constructor
  ErrUnknown(const std::string &from) throw() :
    Exception(from, "Unknown error...") {};
};

/*! \class ErrBug Exception.h basics/Exception.h
  \ingroup libBasics
 */
class ErrBug : public Exception {
public:
  // constructor
  ErrBug(const std::string &from) throw() :
    Exception(from, "Bug detected") {};
  virtual ~ErrBug() throw() {};
};

inline void Exception::fatal(const char *from) throw() {
  try {
          CERR << "Fatal: exception received in " << from
         << " while handling exception." << ENDL;
#ifndef __BCPLUSPLUS__
          exit(-1);
#else
          std::exit(-1);
#endif
  }
  catch(...) {
    try {
      CERR << "Fatal: exception received in Exception::fatal while handling exception." << ENDL;
#ifndef __BCPLUSPLUS__
      exit(-1);
#else
      std::exit(-1);
#endif
    }
    catch(...) {
#ifndef __BCPLUSPLUS__
      exit(-1);
#else
      std::exit(-1);
#endif
    }
  }
}

inline Exception::Exception(const std::string& _from, const std::string& msg) throw()
#ifdef __GNUC__
  try
#endif
 : error(msg), from() {
    from.push(_from);
}
#ifdef __GNUC__
catch(...) {
  fatal("Exception::Exception(const std::string&, const std::string&)");
}
#endif

inline Exception::Exception(const std::string& _from) throw()
#ifdef __GNUC__
  try
#endif
 : error(""), from() {
    from.push(_from);
}
#ifdef __GNUC__
catch(...) {
  fatal("Exception::Exception(const std::string&)");
}
#endif


inline Exception::Exception(const Exception& org) throw()
#ifdef __GNUC__
  try
#endif
 : std::exception(org),error(org.error), from(org.from) {
}
#ifdef __GNUC__
catch(...) {
  fatal("Exception::Exception(const Exception&)");
}
#endif


inline Exception::~Exception(void) throw() {
}


inline void Exception::print(STD_OSTREAM& os) const throw() {
  std::string name;
  getException(name);

  try {
    os << "Exception " << name.c_str() << " thrown" << ENDL;
    for(std::stack<std::string> _from = from; !_from.empty(); _from.pop()) {
      os << "  from: " << _from.top().c_str() << ENDL;
    }
    os << error.c_str() << ENDL;
  }
  catch(...) {
    fatal("Exception::print(ostream&)");
  }
}


inline const char *Exception::what(void) const throw() {
  return (const char *) *this;
}


inline void Exception::push(const std::string& org) throw() {
  try {
    from.push(org);
  }
  catch(...) {
    fatal("Exception::addStack(const std::string&)");
  }
}


inline const std::string& Exception::getError(void) const throw() {
  return error;
}

inline void Exception::getException(std::string& name) const throw() {
  CERR << "In Exception" << ENDL;
  try {
    #if (defined __GNUC__) && (__GNUC__ < 3) // GNU C++ compiler class name demangling
    unsigned int nested = 1, i, nb, offset;  // Does not work for gcc-3.2
    std::string one;
    name = "";
    std::string iname = typeid(*this).name();
    if(iname[0] == 'Q') {
      nested = iname[1] - '0';
      iname = std::string(iname, 2, std::string::npos);
    }
    for(i = 0; i < nested; i++) {
      sscanf(iname.c_str(), "%u%n", &nb, &offset);
      iname = std::string(iname, offset, std::string::npos);
      name += std::string(iname, 0, nb);
      if(i + 1 < nested) name += "::";
      iname = std::string(iname, nb, std::string::npos);
    }
#else             // no class name demangling
    name = typeid(*this).name();
#endif
  }
  catch(...) {
    fatal("Exception::getException(std::string&)");
  }
}


inline const std::stack<std::string>& Exception::getStack(void)  const throw() {
  return from;
}


inline Exception::operator const char *() const throw() {
  std::string name;
  getException(name);

  return name.c_str();
}


inline STD_OSTREAM & operator<< ( STD_OSTREAM & os, const Exception &e) throw() {
  try {
          e.print(os);
  }
  catch(...) {
    Exception::fatal("operator<<(ostream&, const Exception&)");
  }
  return os;
}

} // namespace yav

#endif // ifndef Exception.h

