#ifndef GENERALFORMAT_H
#define GENERALFORMAT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/GeneralFormat.h"
#include <basics/Exception.h>
#include <basics/TclString.h>
#include <list>
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Vec2.h>
#include <linearAlgebra/HVec4.h>
#include <linearAlgebra/Rotation2D.h>
#include <linearAlgebra/Rotation3D.h>
#include <basics/ListElement.h>
#include <basics/ObjectsList.h>

namespace yav {

/// module default version
#define DEFAULT_VERSION 1.0

/** The default I/O read mode when no information about the mode is
    given in the header */
#define DEFAULT_GF_MODE GFIO::GF_ASCII

/** Default byte swap information when
    no information is given about the endianness of the binary data */
#define DEFAULT_GF_READ_SWAP false


/// If the compiler does not support stream exceptions ios::failure
#define WITHOUT_STREAM_EXCEPTIONS


class GeneralFormat;

/*! \class GFIO GeneralFormat.h basics/GeneralFormat.h
  \ingroup libBasics
  general format input/output stream
*/
class GFIO {
 public:
  /// defines the way to write or read data
  enum GFMode {
    /// binary file
    GF_BINARY,
    /// ascii file
    GF_ASCII
  };

 protected:
  /// ASCII/BINARY flag
  GFMode mode;

  /// constructor
  GFIO() : mode(GF_ASCII) {}

 public:
         /** get the IO mode : binary or ascii */
         GFMode getMode() const {
                 return mode;
         }

  /** set read flage
      @param mode ASCII/BINARY flag */
  void setMode(GFMode mode);

  /** swap buffer
      @param ptr  buffer adress
      @param size buffer size */
  static void Swap(void *ptr, unsigned int size);
};


/*! \class GFInput GeneralFormat.h basics/GeneralFormat.h
  \ingroup libBasics
  general format input stream
*/
class GFInput : public GFIO, public IFSTREAM {
 protected:
  /// swap bytes flag
  bool swap;
  /// file read
  TclString filename;

 public:
  /// opens an IFSTREAM
  GFInput(const char *filename);

  /// returns read file name
  const TclString& getFileName(void) const { return filename; }

  /** set read flage
      @param mode ASCII/BINARY flag
      @param swap swap bytes flag */
  void setMode(GFMode mode, bool swap);

  bool getSwap() const {
        return swap;
  }

  void readArray(double  *ptr, unsigned int arraySize);
  void readArray(int *ptr, unsigned int arraySize);
  void readArray(short *ptr, unsigned int arraySize);
  void readArray(float *ptr,  unsigned int arraySize);
  void readArray(Vec3<double> *ptr, unsigned int arraySize);
  void readArray(Vec2<double> *ptr, unsigned int arraySize);

};


GFInput& operator>>(GFInput& input, bool& val);
GFInput& operator>>(GFInput& input, char &val);
GFInput& operator>>(GFInput& input, const char *val);
GFInput& operator>>(GFInput& input, TclString& val);
GFInput& operator>>(GFInput& input, int &val);
GFInput& operator>>(GFInput& input, unsigned int &val);
GFInput& operator>>(GFInput& input, short int &val);
GFInput& operator>>(GFInput& input, unsigned short int &val);
GFInput& operator>>(GFInput& input, long int &val);
GFInput& operator>>(GFInput& input, unsigned long int &val);
GFInput& operator>>(GFInput& input, float &val);
GFInput& operator>>(GFInput& input, double &val);
GFInput& operator>>(GFInput& input, Vec3<double> &val);
GFInput& operator>>(GFInput& input, Vec3<unsigned int> &val);
GFInput& operator>>(GFInput& input, Rotation3D &val);
GFInput& operator>>(GFInput& input, Rotation2D &val);
GFInput& operator>>(GFInput& input, Vec2<double> &val);



/*! \class GFOutput GeneralFormat.h basics/GeneralFormat.h 
  \ingroup libBasics
  general format output stream
*/
class GFOutput : public GFIO, public OFSTREAM {
 protected:
  /// file read
  TclString filename;

 public:
  /// opens an OFSTREAM
  GFOutput(const char *filename);

  /// returns written file name
  const TclString& getFileName(void) const { return filename; }

  void writeArray(const int *val, unsigned int arraySize);
  void writeArray(const float *val, unsigned int arraySize);
  void writeArray(const double *val, unsigned int arraySize);
  void writeArray(const Vec2<double> *val, unsigned int arraySize);
  void writeArray(const Vec3<double> *val, unsigned int arraySize);

};

GFOutput& operator<<(GFOutput& output, const bool& val);
GFOutput& operator<<(GFOutput& output, const char &val);
GFOutput& operator<<(GFOutput& output, const char *val);
GFOutput& operator<<(GFOutput& output, const TclString &val);
GFOutput& operator<<(GFOutput& output, const int &val);
GFOutput& operator<<(GFOutput& output, const unsigned int &val);
GFOutput& operator<<(GFOutput& output, const short &val);
GFOutput& operator<<(GFOutput& output, const unsigned short &val);
GFOutput& operator<<(GFOutput& output, const long &val);
GFOutput& operator<<(GFOutput& output, const unsigned long &val);
GFOutput& operator<<(GFOutput& output, const float &val);
GFOutput& operator<<(GFOutput& output, const double &val);
GFOutput& operator<<(GFOutput& output, const Vec3<unsigned int> &val);
GFOutput& operator<<(GFOutput& output, const Vec3<double> &val);
GFOutput& operator<<(GFOutput& output, const Rotation3D &val);
GFOutput& operator<<(GFOutput& output, const Rotation2D &val);
GFOutput& operator<<(GFOutput& output, const Vec2<double> &val);


/*! \class GeneralFormatModuleVersion GeneralFormat.h basics/GeneralFormat.h
  \ingroup libBasics
  an element of GeneralFormatModule version list
*/
class GeneralFormatModuleVersion : public ListElement<GeneralFormatModuleVersion> {
 public:
  /** defines the prototype for a reading function
      @param input The input stream */
  typedef  void (GeneralFormat::*INPUTREADFUNCTION)(GFInput &input);

  /** defines the prototype for a skipping function
      @param input The input stream */
  typedef  void (GeneralFormat::*INPUTSKIPFUNCTION)(GFInput &input);

  /** defines the prototype for a writing function
      @param output The output stream */
  typedef  void (GeneralFormat::*OUTPUTWRITEFUNCTION)(GFOutput &output);

 public:

  /** @name Exception class */
  //@{
  /*! \class EmptyVersion GeneralFormat.h basics/GeneralFormat.h
    \ingroup libBasics
    Exception thrown when read,write,skip routines are not available
   */
  class EmptyVersion : public Exception {
  public:
    EmptyVersion(const std::string& from) :
      Exception(from,
                "Attempt to call a version that has a null reading or writing or skipping functions") {}
  };
  //@}

 private:
  /// the version number
  double version;
  /// the read function
  INPUTREADFUNCTION  read;
  /// the skip function
  INPUTSKIPFUNCTION  skip;
  /// the write function
  OUTPUTWRITEFUNCTION write;

 public:
  /** @name Constructors */
  //@{
  /// default constructor
  GeneralFormatModuleVersion()
    : version(DEFAULT_VERSION), read(0), skip (0), write(0) {}

  /** general constructor
      @param thisversion version number
      @param readFunction  pointer on data read procedure
      @param writeFunction pointer on data write procedure
      @param skipFunction  pointer on data skip procedure */
  GeneralFormatModuleVersion(double thisversion,
                             INPUTREADFUNCTION readFunction,
                             INPUTSKIPFUNCTION skipFunction,
                             OUTPUTWRITEFUNCTION writeFunction)
    : version(thisversion), read(readFunction), skip(skipFunction),
      write(writeFunction) {}
  //@}

  /// returns the version of the module
  double getVersion() const;

  /// sets the read function of the module
  void setReadFunction(const INPUTREADFUNCTION );

  /// sets the skip function of the module
  void setSkipFunction(const INPUTSKIPFUNCTION);

  /// sets the write function of the module
  void setWriteFunction(const OUTPUTWRITEFUNCTION);

  friend class GeneralFormat;
};



/*! \class GeneralFormatModule GeneralFormat.h basics/GeneralFormat.h
  \ingroup libBasics
  an element of GeneralFormat modules list
*/
class GeneralFormatModule : public ListElement<GeneralFormatModule> {
 public:
  /** @name Exception class */
  //@{
  /*!  \class InvalidVersion GeneralFormat.h basics/GeneralFormat.h
    \ingroup libBasics
    Exception thrown when no valid version is available ie when the existing
    versions are all greater than the required version 
  */
  class InvalidVersion : public Exception {
  public:
    InvalidVersion(const std::string& from);
  };
  //@}

 private:
  /// module's name
  TclString moduleName;

  /// different version's list
  ObjectsList<GeneralFormatModuleVersion> versionList;

  /// whether this module should always be saved
  const bool mandatoryWriting;

  /// whether this module should be saved for the current operation
  bool selectedWriting;

  /// whether this module should be  loaded for the current operation
  bool selectedReading;

 public:
  /** @name Constructors */
  //@{
  /// default constructor
  GeneralFormatModule()
    : moduleName("unknown_module"), versionList(),mandatoryWriting(false),
      selectedWriting(false), selectedReading(true)
    {if (mandatoryWriting==true) selectedWriting=true;}

  /** constructor with the name of the module and the mandatory flag
      @param modulename module's name
      @param _mandatoryWriting whether it is mandatory for saving
      @param _selectedWriting whether the module should be written
      @param _selectedReading whether the module should be read */
  GeneralFormatModule(const  char *modulename,
                      const bool _mandatoryWriting=false,
                      const bool _selectedWriting=false,
                      const bool _selectedReading=true)
    : moduleName(modulename),versionList(),mandatoryWriting(_mandatoryWriting),
      selectedWriting(_selectedWriting),selectedReading(_selectedReading)
    { if (mandatoryWriting==true) selectedWriting=true;}
  //@}

  /** This routine returns a pointer of the existing version that best match
      the required version number. If theere is a version with the same  exact
      version number, then it is returned. Otherwise the version having the
      highest version number below the required version number is returned.
      If non is found return 0 and throw an InvalidVersion exception
      @param version  version number that is required
      @exception InvalidVersion version number is not available */
  GeneralFormatModuleVersion *getCorrespondingVersion(const double version)
    throw(InvalidVersion);


  /// Returns the first available version of a module
  GeneralFormatModuleVersion *getFirstVersion();

  /** adds a version of a module in its list
      @param moduleVersion a module version */
  void addModuleVersion(GeneralFormatModuleVersion *moduleVersion);

  /// Returns the name of the module
  TclString getName() const;

  /// enable the saving of the module
  void selectWriting();

  /// disable the saving of the module
  void unselectWriting();

  /// enable the reading of the module
  void selectReading();

  /// disable the reading of the module
  void unselectReading();

  /// Returns true if a module is mandatory for writing
  bool isMandatoryForWriting() const;

  /// returns the writing flag
  bool isSelectedWriting() const;

  /// returns the reading flag
  bool isSelectedReading() const;

  friend class GeneralFormatModuleVersion;
};



/*! \class GeneralFormat GeneralFormat.h basics/GeneralFormat.h
  \ingroup libBasics
  General format reading/writting facility
*/
class GeneralFormat {
 public:
  /** @name Exception class */
  //@{
  /*! \class UnknownModule GeneralFormat.h basics/GeneralFormat.h
    \ingroup libBasics
    no module corresponding to a name
  */
  class UnknownModule : public Exception {
  public:
    UnknownModule(const std::string& from)  :
      Exception(from,
                "Attempt to find a module of unknown name") {}
  };

  /*! \class UnknownFormat GeneralFormat.h basics/GeneralFormat.h
    \ingroup libBasics
    reading a file with unknown format
  */
  class UnknownFormat : public Exception {
    public :
      UnknownFormat(const std::string& from) :
      Exception(from,
                "Attempt to read a file of unknown format (wrong magic number)") {}
  };

  /*! \class FileNotFound GeneralFormat.h basics/GeneralFormat.h
    \ingroup libBasics
    unable to open file
  */
  class FileNotFound : public Exception {
    public :
      FileNotFound(const std::string& from, const std::string& name) :
      Exception(from, "unable to open file " + name) {}
  };
  //@}

 protected:
  /** The name of the object type stored in the header of each file.
      It is used in the magic number in the header file */
  TclString key;
  /** object descriptor header name */
  TclString head;
  /// The highest version number
  double version;
  /// The list of all available module
  ObjectsList<GeneralFormatModule> moduleList;
  /// number of elements in file
  unsigned int numberOfElements;
  /// read file name (or EMPTY_STRING)
  TclString filename;

  /** search for given module by its name
      @param name module name
      @exception UnknownModule module was not found
      @return module reference */
  GeneralFormatModule& moduleByName(const TclString& name)
    throw(UnknownModule);


 public:
  /** give access to the list of modules
      @return an iterator on the list of modules */
         ObjectsList<GeneralFormatModule>::iterator getFirstModule() {
    return moduleList.begin(); }
  /** give access to the last iterator of the  modules list
      @return an iterator on the list of modules */
         ObjectsList<GeneralFormatModule>::iterator getLastModule() {
    return moduleList.end(); }
  /** @name Constructors */
  //@{
  /// default constructor
  GeneralFormat() : key("Unknown Object File"), head("OBJECT"), version (0.0),
    numberOfElements(1), filename("Unnamed") {}
  /** constructor from object name
      @param _key file key string
      @param _head object header in file */
  GeneralFormat(const TclString _key, const TclString _head)
    : key(_key), head(_head), version (0.0), numberOfElements(1),
      filename("Unnamed") {}
  //@}

  virtual ~GeneralFormat() {}
  /** adds a module version in its list. If the module does not
      exists then creates a new module. If the Module version already exists
      then it replaces the 3 I/O functions previously stored
      @param modulename module's name
      @param mandatory whether that module is mandatory for writing
      @param version module version number
      @param readFunction read function
      @param skipFunction skip function
      @param writeFunction write function */
  void addModuleVersion(const TclString modulename,
                        const bool mandatory,
                        const double version,
                        const GeneralFormatModuleVersion::INPUTREADFUNCTION readFunction,
                        const GeneralFormatModuleVersion::INPUTSKIPFUNCTION skipFunction,
                        const GeneralFormatModuleVersion::OUTPUTWRITEFUNCTION writeFunction);

  /** sets the name of the file format stored in the header of each file
      @param key object file key string
      @param head object descriptor head string */
  void setClassName(const char *key, const char *head);
  
  /** returns the first few characters at the beginning of the file */
  TclString getMagic() const;
  /** returns the current filename
      @return the current filename */
  TclString getFilename() const
    { return(filename); }
  /** unselect a given module for reading
      @param moduleName module's name
      @exception UnknownModule unknown module name transmitted */
  void unselectModuleReading(const TclString& moduleName) throw(UnknownModule);
  /** unselect a given module for writing
      @param moduleName module's name
      @exception UnknownModule unknown module name transmitted */
  void unselectModuleWriting(const TclString& moduleName) throw(UnknownModule);
  /** select a given module for reading
      @param moduleName module's name
      @exception UnknownModule unknown module name transmitted */
  void selectModuleReading(const TclString& moduleName) throw(UnknownModule);
  /** select a given module for writing
      @param moduleName module's name
      @exception UnknownModule unknown module name transmitted */
  void selectModuleWriting(const TclString& moduleName) throw(UnknownModule);
  /// select all modules for reading
  void selectAllModulesReading();
  /// select all modules for writing
  void selectAllModulesWriting();
  /// unselect all modules for reading
  void unselectAllReadingModules();
  /// unselect all modules for writing
  void unselectAllWritingModules();
  /** Reads the first line of the file and tests if the type name corresponds
      to the name in GeneralFormat::name. If not throw a UnknownFormat flag.
      Also reads the version number on the file
      @param input input stream
      @param inputVersion version number stored in the header file
      @exception UnknownFormat unknown file format
      @exception ios::failure read failed on input stream */
  void readMagic(STD_ISTREAM &input, double &inputVersion)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (UnknownFormat);
#else
    throw (UnknownFormat, ios::failure);
#endif

    /* Reads the body of a file : first reads the list of modules stored
       in the data. If the module name is unknown throw an UnknownModule
       exception. If the module does not have a read or skip function then
       throws an EmptyVersion exception.  If the file version number is
       too small then throws an  InvalidVersion exception. Seconds it reads
       all modules, skipping the modules that are not selected
       @param currentVersion  input file version
       @param input input stream
       @param mode opened file mode
       @param swap whether swap (big/little endian) is needed
       @exception GeneralFormatModule::InvalidVersion invalid version number
       @exception UnknownModule unknown module
       @exception GeneralFormatModuleVersion::EmptyVersion read version routines is not available
       @exception ios::failure read failed on input stream */
  void readData(const double currentVersion, GFInput &input,
                GFIO::GFMode mode, bool swap)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion, Exception);
#else
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion, ios::failure, Exception);
#endif

    /* Reads a an input file with a variable list of modules names (char *)
       terminated by (char *) 0.
       Call read(GFInput)
       @param filename file to read
       @param ... modules name to read (all if empty)
       @exception GeneralFormatModule::InvalidVersion invalid version number
       @exception UnknownModule unknown module
       @exception GeneralFormatModuleVersion::EmptyVersion read version routines is not available
       @exception ios::failure read failed on input stream
       @exception UnknownFormat unknown format file */
    void  read(char *filename...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion,
             UnknownFormat, FileNotFound, Exception);
#else
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion, ios::failure,
             UnknownFormat, FileNotFound, Exception);
#endif

   /* Reads a a GFInput with a variable list of modules names (char *)
       terminated by (char *) 0. If this list is empty (only 0),
       then all modules in the file will be read. Otherwise, only the
       selected modules will be read, the others will be skipped.
       The a module name is unknown then throws an UnknownModule exception.
       First call readMagic() then
       look for information related to the way the data is stored (binary
       or ASCII) as well as the endianness of the binary data. If none are
       present then used default value. Finally calls readData
       @param filename file to read
       @param ... modules name to read (all if empty)
       @exception GeneralFormatModule::InvalidVersion invalid version number
       @exception UnknownModule unknown module
       @exception GeneralFormatModuleVersion::EmptyVersion read version routines is not available
       @exception ios::failure read failed on input stream
       @exception UnknownFormat unknown format file */
    void  read(GFInput *input...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion,
             UnknownFormat, FileNotFound, Exception);
#else
      throw (GeneralFormatModule::InvalidVersion, UnknownModule,
             GeneralFormatModuleVersion::EmptyVersion, ios::failure,
             UnknownFormat, FileNotFound, Exception);
#endif

    /** Writes the first line of the output file. Store the object type
       and the version number
       @param output output stream
       @param version output file version number
       @exception ios::failure read failed on output stream */
    void writeMagic(GFOutput &output, const double version)
#ifdef WITHOUT_STREAM_EXCEPTIONS
      ;
#else
    throw (ios::failure);
#endif

    /* Writes the body of the file : first writes the names of the modules
       will be saved. Those modules have been either selected previously.
       If the module  does not have a write routine then throws an
       EmptyVersion exception. If the file version number is too small
       then throws an InvalidVersion exception. Finally its successively
       calls the write routine of all selected module.
       @param outputVersion output file version number
       @param output output stream
       @param n number of elements in file
       @exception GeneralFormatModule::InvalidVersion invalid version number
       @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
       @exception ios::failure read failed on output stream */
    void writeData(const double outputVersion , GFOutput &output)
#ifdef WITHOUT_STREAM_EXCEPTIONS
      throw (GeneralFormatModule::InvalidVersion,
             GeneralFormatModuleVersion::EmptyVersion);
#else
    throw (GeneralFormatModule::InvalidVersion,
           GeneralFormatModuleVersion::EmptyVersion, ios::failure);
#endif

    /* Writes a file with a given I/O mode (binary or ASCII).
       This routines only writes mandatory modules with the latest version.
       Calls write(mode,GeneralFormat::version,output(filenemae), (char *)0);
       @param  mode the way to write data : binary or ASCII
       @param outputVersion output file version
       @param filename output file name
       @param n number of elements in file
       @param ... list of module names
       @exception GeneralFormatModule::InvalidVersion invalid version number
       @exception UnknownModule unknown module
       @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
       @exception ios::failure read failed on output stream */
  void write(const GFIO::GFMode mode, const double outputVersion,
             const char *filename, unsigned int n...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion, UnknownModule,
         GeneralFormatModuleVersion::EmptyVersion,ios::failure);
#endif

     /* Writes a file with a given I/O mode (binary or ASCII).
        This routines only writes mandatory modules with the latest version.
        Calls write(mode,GeneralFormat::version,filename, (char *)0);
        @param mode the way to write data : binary or ASCII
        @param filename output file name
        @param n number of elements in file
        @exception GeneralFormatModule::InvalidVersion invalid version number
        @exception UnknownModule unknown module
        @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
        @exception ios::failure read failed on output stream */
  void write(const GFIO::GFMode mode,const char *filename, unsigned int n = 1)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion, UnknownModule,
         GeneralFormatModuleVersion::EmptyVersion, ios::failure);
#endif

  /* Writes a file with a given version number. This routines only writes
     mandatory modules. Calls write(DEFAULT_GF_MODE,outputVersion,
     filename,(char *)0);
     @param  outputVersion output file version
     @param filename output file name
     @param n number of elements in file
     @exception GeneralFormatModule::InvalidVersion invalid version number
     @exception UnknownModule unknown module
     @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
     @exception ios::failure read failed on output stream */
  void write(const double outputVersion,const char  *filename, unsigned int n = 1)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion, UnknownModule,
         GeneralFormatModuleVersion::EmptyVersion, ios::failure);
#endif

  /* Writes a file. This routines only writes
     mandatory modules with the latest version.
     Calls write(DEFAULT_GF_MODE,GeneralFormat::version,filename,
     (char *)0);
     @param filename output file name
     @param n number of elements in file
     @exception GeneralFormatModule::InvalidVersion invalid version number
     @exception UnknownModule unknown module
     @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
     @exception ios::failure read failed on output stream */
  void write(const char  *filename, unsigned int n = 1)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure);
#endif
    /* Writes a output with a given version number and a given I/O mode (binary
       or ASCII) with a variable list of module names (char *) ended with (char *)
       0. If this list is empty (only 0) then only writes mandatory modules.
       Otherwise, writes mandatory modules and the modules whose name appears
       in the variable list. If the name in the list does not correspond to a
       module name, then throws a UnknownModule exception.
       First call writeMagic() then writes some
       information about the I/O mode and the endianness of the computer.
       Finally calls writeData()
       @param  mode the way to write data : binary or ASCII
       @param outputVersion output file version
       @param output
       @param n number of elements in file
       @param ... list of module names
       @exception GeneralFormatModule::InvalidVersion invalid version number
       @exception UnknownModule unknown module
       @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
       @exception ios::failure read failed on output stream */
  void write(const GFIO::GFMode mode, const double outputVersion,
             GFOutput &output, unsigned int n...)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion, UnknownModule,
         GeneralFormatModuleVersion::EmptyVersion,ios::failure);
#endif

     /* Writes a output with a given I/O mode (binary or ASCII).
        This routines only writes mandatory modules with the latest version.
        Calls write(mode,GeneralFormat::version,output, (char *)0);
        @param mode the way to write data : binary or ASCII
        @param output
        @param n number of elements in file
        @exception GeneralFormatModule::InvalidVersion invalid version number
        @exception UnknownModule unknown module
        @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
        @exception ios::failure read failed on output stream */
  void write(const GFIO::GFMode mode, GFOutput &output, unsigned int n = 1)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion, UnknownModule,
         GeneralFormatModuleVersion::EmptyVersion, ios::failure);
#endif

  /* Writes a output with a given version number. This routines only writes
     mandatory modules. Calls write(DEFAULT_GF_MODE,outputVersion,
     output,(char *)0);
     @param  outputVersion output file version
     @param filename output file name
     @param n number of elements in file
     @exception GeneralFormatModule::InvalidVersion invalid version number
     @exception UnknownModule unknown module
     @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
     @exception ios::failure read failed on output stream */
  void write(const double outputVersion, GFOutput &output, unsigned int n = 1)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion, UnknownModule,
           GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion, UnknownModule,
         GeneralFormatModuleVersion::EmptyVersion, ios::failure);
#endif

  /* Writes a output. This routines only writes
     mandatory modules with the latest version.
     Calls write(DEFAULT_GF_MODE,GeneralFormat::version,output,
     (char *)0);
     @param filename output file name
     @param n number of elements in file
     @exception GeneralFormatModule::InvalidVersion invalid version number
     @exception UnknownModule unknown module
     @exception GeneralFormatModuleVersion::EmptyVersion write version routines is not available
     @exception ios::failure read failed on output stream */
  void write(GFOutput &output, unsigned int n = 1)
#ifdef WITHOUT_STREAM_EXCEPTIONS
    throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion);
#else
  throw (GeneralFormatModule::InvalidVersion,UnknownModule,GeneralFormatModuleVersion::EmptyVersion,ios::failure);
#endif

};

#if (defined(_SGI_)||defined(WIN32))
}
yav::GFOutput& endl(yav::GFOutput& output);
#else
yav::GFOutput& endl(yav::GFOutput& output);
}
#endif

#endif
