#ifndef GENERICHISTOGRAM_H
#define GENERICHISTOGRAM_H


#pragma interface "basics/GenericHistogram.h"
#include <basics/Math.h>
#include <basics/Pointed.h>
#include <map>
#include <vector>


namespace yav {

class TclGenericHistogram;


/*! \class GenericHistogram GenericHistogram.h basics/GenericHistogram.h
  \ingroup libBasics
  a class that computes and analyse an histogram from a list of (long
  int=sample value, double = nb of samples or other associated sample value)
*/
class GenericHistogram : public Pointed<GenericHistogram> {
 protected:
  /// the histogram data is stored as a list of  (double=sample value, unsigned long int = nb of samples)  */
 std::map<double,unsigned long int>  data;
 /// the reformated histogram that depends on the bucket size
 unsigned long int *histogram;
 /// the number of buckets
 unsigned int nbBuckets;
 /// the size of each buckets
 double bucketSize;
 /// min data value
 double minValue;
 /// maximum data value
 double maxValue;
 /// mean histogram value
 double meanValue;
 ///  histogram standard deviation value
 double standardDeviation;
 ///  histogram 50 % median value
 double medianValue;
 /// the total number of samples in the original data
 unsigned long int nbSamples;
 public:
 /// the different smoothing filters that can be applied on an histogram
 enum SmoothingFilter {GAUSSIAN_SMOOTHING};
 /** @name constructor */
 //@{
 /** Creates an empty histogram */
 GenericHistogram(){}
 /** Creates an  histogram from a list of (unsigned int, double) pairs
     @param h a list of   (double=sample value, unsigned long int = nb of samples ) describing the histogram*/
 GenericHistogram(const std::map< double, unsigned long int> & h);
  /** Creates an histogram from a list of double values.  When building the
      histogram it counts the number of identical samples
      @param l is a list of   double (=sample value)*/
 GenericHistogram(const std::list< double> & l);
 /** Creates an  histogram from an array of integer and the  of (unsigned int, double) pairs
     @param v an array for each value i between 0 and arraySize-1 it gives the number of samples have the value minVal +i*(maxVal-minVal)/(arraySize-1)
     @param arraySize the size of the array
     @param minVal the value associated with the first entry
     @param maxVal the value associated with the last entry */
 GenericHistogram(const unsigned long int *v,const unsigned int arraySize,
                  double minVal,  double maxVal);
 /** Creates an  histogram as a copy of a given histogram : reallocates all structures
     @param h an histogram */
 GenericHistogram(const GenericHistogram& h);
 /** Creates an  histogram from a file
     @param file name of a file where an histogram is stored */
 GenericHistogram(char * file);
 //@}

 /** @name destructor */
 //@{
 /// empty destructor
 virtual ~GenericHistogram(){}
 //@}
 /// compute the mean, std deviation and median value based on the original data
 void computeStatistics(double min, double max);
 protected:
 /// compute the mean, std deviation and median value based on the original data
 void computeStatistics();
 /// build the histogram from nbBuckets and bucketSize;
 void buildHistogram();
 /** given the data hash table and the min and max data value, it sets
     the value of nbBuckets and bucketSize based on some Heuristics */
 void guessHistogramParameters();
 public:
 /// build the histogram from a given  number of buckets; updates the bucket size
 void buildHistogram( const unsigned int nBuckets) {
   nbBuckets=nBuckets;
   bucketSize=(maxValue-minValue)/nbBuckets;
   buildHistogram();
 }
 /// build the histogram from a given  bucket size; updates the number of buckets
 void buildHistogram(const double size) {
   nbBuckets=(unsigned int)((maxValue-minValue)/size +1);
   bucketSize=size;
   buildHistogram();
 }

 /** returns the number of sample associated with a given entry value
     @param x entry value
     @return the number of samples equal to x (0) if none */
 unsigned long int  getNbSamples (const double  x);
 /** returns the maximum number of sample of the current histogram
     @return the maximum number of samples */
 unsigned long int  getMaxNbSamples () const;

 /** add two histograms
     @param h an histogram that should have nearly the same samples as this
     @return the samples from h are included in the current histogram */
 void  operator += (const GenericHistogram & h) {
   std::map< double, unsigned long int>::iterator it2;
   std::map< double, unsigned long int>::const_iterator it1;
   for(it1=h.data.begin();it1!=h.data.end();it1++)
     {
       it2=data.find(it1->first);
       if (it2!=data.end())
         it2->second+=it1->second;
       else
         {
           data[it1->first]=it1->second;
         }
     }
   if (h.minValue<minValue)
     minValue=h.minValue;
   if (h.maxValue>maxValue)
     maxValue=h.maxValue;
   buildHistogram(bucketSize);
 }

 /** multiply all associated sample value with a given value
    @param lambda a constant multiplied to each associated sample value*/
 void  operator * (const double lambda){
    std::map< double, unsigned long int>::iterator it1;
    for(it1=data.begin();it1!=data.end();it1++)
      it1->second= (unsigned long int) (it1->second*lambda);
    buildHistogram(nbBuckets);
 }
 /** write an histogram in a output stream
    @param out output stream (usually a file)
    @param h an histogram */
 friend STD_OSTREAM & operator<<(STD_OSTREAM& out,const GenericHistogram &h ){
   std::map< double, unsigned long int>::const_iterator it;
   out.precision(3);
   out<<"{ ";
   for(it=h.data.begin();it!=h.data.end();it++)
     out<<"( "<<it->first<<" , "<<it->second<<" ) ";
   out<<"}"<<ENDL;
   return out;
 }
 /// returns the minimum value of the array
 double getMinimumValue() const {
   return minValue;
 }
 /// returns the minimum value of the array
 double getMaximumValue() const {
   return maxValue;
 }
 /// returns the mean  value of the array
 double getMeanValue() const {
   return meanValue;
 }
 /// returns the mean  value of the array
 double getMedianValue() const {
   return medianValue;
 }
 /// returns the standard deviation  value of the array
 double getStandardDeviationValue() const {
   return standardDeviation;
 }
 /// returns a given quantile of the histogram
 double getHistogramQuantile(const double q) const;
 /// returns a given quantile of the histogram
 double getHistogramQuantile(const double q, double min, double max) const;
 /// returns the size of each bucket
 double getBucketSize() const {
   return bucketSize;
 }
 /// returns the size of each bucket
 unsigned int  getNumberBuckets() const {
   return nbBuckets;
 }

 /** write an histogram in an output file
     @param name the filename */
 void write(char* name) const;
  /** smooth an histogram given the smoothing kernel size and the type of smoothing kernel
     @param sigma the half-size of the smoothing kernel
     @param f the type of smoothing filter */
 void smooth(double sigma,SmoothingFilter f=GAUSSIAN_SMOOTHING );

 friend class HistogramDistribution;
 friend class TclGenericHistogram;
};



} // namespace yav

#endif
