#ifndef _IKdTree_
#define _IKdTree_

#include <linearAlgebra/Vec.h>
#include <vector>
#include <algorithm>
#include <functional>

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {


/*! \class IKdTree IKdTree.h basics/IKdTree.h
  \ingroup libBasics
  Template for a kD-Tree of dimension D and type T
  This structure is a binary tree which contains a set of points and can search quickly every points which are included in a given sphere.
  Contrary to KdTree class, the search function used here is non-recursive (thus quicker).
*/
#ifdef __BCPLUSPLUS__
  template<Index Dim, class Real>
#else
  template<Index Dim, class Real=double>
#endif
class IKdTree {
 public:
  /// Construct the kdtree structure for a set of points given as a stl vector of Vect<Dim,Real>
  IKdTree(const std::vector<Vect<Dim,Real> >& pointlist);
  /// Destructor
  ~IKdTree();

  /// search all points in a sphere of center and radius. return number of points found.
  int searchILClosestPoint(const Vect<Dim,Real>& position,
                           const double radius,
                           std::vector<Index>& result);


 private:
  /*! \class IKdTreeNode IKdTree.h basics/IKdTree.h
    A node of the kD-tree
  */
  class IKdTreeNode
  {
  public:
    /// the point
    Vect<Dim,Real> val;
    /// its index in point list
    Index id;
    /// level of the node (profondeur in the tree)
    Index level;
    // left son : index during construction, pointer after (don't remnber why, so don't ask)
    union {int son0i; IKdTreeNode* son0p;};
    // right son : index during construction, pointer after
    union {int son1i; IKdTreeNode* son1p;};

  public:
    /// Constructor : copy point. sub-tree are not initialised
    IKdTreeNode( const Vect<Dim,Real>& v, Index i, Index l=0 )
      : val(v), id(i), level(l)
    {};
  };
  /// the top node of the tree
  IKdTreeNode* topNode;
  /// the vector containing nodes (only used for allocation facility)
  std::vector<IKdTreeNode> nodeList;

  /// current level in the tree used for sorting during construction
  static int level;
  /// Points to point list given to constructor during construction
  static const std::vector<Vect<Dim,Real> >* pointList;


  /*! \class sort_function IKdTree.h basics/IKdTree.h
    \ingroup libBasics
    The sort function used during construction
  */
#ifdef __BCPLUSPLUS__
   class sort_function : public std::binary_function<Index, Index, bool>
   {
   public:bool operator()(const Index& x, const Index& y) const
     {
       const Vect<Dim,Real>& vectX = ((const std::vector<Vect<Dim,Real> >*)pointList)->operator[](x);
       const Vect<Dim,Real>& vectY = ((const std::vector<Vect<Dim,Real> >*)pointList)->operator[](y);

       return (vectX(level) < vectY(level));
     }
   };
   friend class sort_function;
#else
  class sort_function : public std::binary_function<Index, Index, bool> {
  public:bool operator()(const Index& x, const Index& y) const { return ((*pointList)[x])(level) < ((*pointList)[y])(level); }
  };friend class sort_function;
#endif

  /// The recursive function used for constructing the kDTree
  int recursiveCreate(std::vector<Index>::iterator first,
                      std::vector<Index>::iterator last);
  /// The recursive function used for replacing indexes by pointers in nodes just after the construction
  void recursiveReplace(IKdTreeNode* node);

  /// Resizeable Array containing points found in the sphere during search
  IKdTreeNode** _nodesResult;
  /// Allocated size of _nodesResult
  int _nbAllocatedNodesResult;
  /// Resizeable Array containing nodes to check during search
  IKdTreeNode** _nodesToCheck;
  /// Allocated size of _nodesToCheck
  int _nbAllocatedNodesToCheck;
};

template<Index Dim, class Real>
int IKdTree<Dim,Real>::level;

template<Index Dim, class Real>
const std::vector<Vect<Dim,Real> >* IKdTree<Dim,Real>::pointList;


template<Index Dim, class Real>
IKdTree<Dim,Real>::IKdTree(const std::vector<Vect<Dim,Real> >& pointlist)
{
  level=0;
  pointList=&pointlist;

  std::vector<Index> indexList;
  std::vector<Vect<Dim,Real> >::const_iterator it;
  for (it=pointlist.begin(); it!=pointlist.end(); it++)
    indexList.push_back(indexList.size());

  recursiveCreate(indexList.begin(), indexList.end());
  topNode=&(nodeList[0]);
  recursiveReplace(topNode);

  _nbAllocatedNodesResult=10;
  _nodesResult=(IKdTreeNode**)malloc(_nbAllocatedNodesResult*sizeof(IKdTreeNode*));

  _nbAllocatedNodesToCheck=10;
  _nodesToCheck=(IKdTreeNode**)malloc(_nbAllocatedNodesToCheck*sizeof(IKdTreeNode*));

};

template<Index Dim, class Real>
IKdTree<Dim,Real>::~IKdTree()
{
  free(_nodesToCheck);
  free(_nodesResult);
};

/*----------------------------------------*/

template<Index Dim, class Real>
int IKdTree<Dim,Real>::recursiveCreate(std::vector<Index>::iterator first,
                                       std::vector<Index>::iterator last)
{

  int size = last - first;
  std::vector<Index>::iterator first1=first+(size/2);
  std::nth_element(first, first1, last, sort_function());

  int nodeid=nodeList.size();
#ifdef __BCPLUSPLUS__
  const Vect<Dim,Real>& vect =  ((const std::vector<Vect<Dim,Real> >*)pointList)->operator[](*first1);
  nodeList.push_back(IKdTreeNode( vect, *first1, (unsigned int)level ));
#else
  nodeList.push_back(IKdTreeNode( (*pointList)[*first1], *first1, level ));
#endif

  Index oldlevel=level;
  level = (level+1) % Dim;

  if (first1-first > 0) {
    // Attention : ordre important (les pointeurs sur nodeList peuvent etre invalides lors de l'appel recursif)
    Index son=recursiveCreate(first, first1);
    nodeList[nodeid].son0i=son;
  } else {
    nodeList[nodeid].son0i=-1;
  }

  first1++;

  if (last-first1 > 0) {
    Index son=recursiveCreate(first1, last);
    nodeList[nodeid].son1i=son;
  } else {
    nodeList[nodeid].son1i=-1;
  }

  oldlevel=level;

  return nodeid;
};


template<Index Dim, class Real>
void IKdTree<Dim,Real>::recursiveReplace(IKdTreeNode* node)
{
  if (node->son0i!=-1) {
    node->son0p=&(nodeList[node->son0i]);
    recursiveReplace(node->son0p);
  } else {
    node->son0p=0;
  }

  if (node->son1i!=-1) {
    node->son1p=&(nodeList[node->son1i]);
    recursiveReplace(node->son1p);
  } else {
    node->son1p=0;
  }
};

/*-------------------------------------------------------------------*/


template<Index Dim, class Real>
int IKdTree<Dim,Real>::searchILClosestPoint(const Vect<Dim,Real>& center,
                                            const double limit_distance,
                                            std::vector<Index>& result)
{
  IKdTreeNode** nodesResult=_nodesResult;
  int nbAllocatedNodesResult=_nbAllocatedNodesResult;
  IKdTreeNode** nodesToCheck=_nodesToCheck;
  int nbAllocatedNodesToCheck=_nbAllocatedNodesToCheck;

  double limit_distance2=limit_distance*limit_distance;

  int nbNodesResult=0;
  int nbNodesToCheck=1;
  nodesToCheck[0]=topNode;

  int itnodes;
  IKdTreeNode* node;
  Index level;
  double c2;
  double c1;
  double cdist;
  bool b1;
  bool b2;
  IKdTreeNode* nextnode;

  for (itnodes=0;itnodes<nbNodesToCheck;itnodes++) {
    node=nodesToCheck[itnodes];
    while (node) {
      level=node->level;
      c2=node->val(level);
      c1=center(level);
      cdist= ::fabs(c2-c1);

      b1=(c1-limit_distance)<c2;
      b2=(c2-limit_distance)<c1;

      if (b1 && b2 && (Dist2(node->val,center)<=limit_distance2)) {
        if (nbNodesResult==nbAllocatedNodesResult) {
          nbAllocatedNodesResult*=2;
          nodesResult=(IKdTreeNode**)realloc(nodesResult,nbAllocatedNodesResult*sizeof(IKdTreeNode*));
        }
        nodesResult[nbNodesResult]=node;
        nbNodesResult++;
      }

      if (b1) {
        nextnode=node->son0p;
      } else {
        nextnode=0;
      }

      if (b2 && node->son1p) {
        if (nextnode) {
          if (nbNodesToCheck==nbAllocatedNodesToCheck) {
            nbAllocatedNodesToCheck*=2;
            nodesToCheck=(IKdTreeNode**)realloc(nodesToCheck,nbAllocatedNodesToCheck*sizeof(IKdTreeNode*));
          }
          nodesToCheck[nbNodesToCheck]=node->son1p;
          nbNodesToCheck++;
        } else {
          nextnode=node->son1p;
        }
      }
      node=nextnode;
    }
  }

  result.clear();
  for (itnodes=0; itnodes<nbNodesResult; itnodes++)
    result.push_back(nodesResult[itnodes]->id);

  _nodesResult=nodesResult;
  _nbAllocatedNodesResult=nbAllocatedNodesResult;
  _nodesToCheck=nodesToCheck;
  _nbAllocatedNodesToCheck=nbAllocatedNodesToCheck;

  return nbNodesResult;
};

}

#endif
