#ifndef _KdTree_
#define _KdTree_

#include <linearAlgebra/Vec.h>
#include <vector>
#include <algo.h>

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {

/*! \class KdTree KdTree.h basics/KdTree.h
  \ingroup libBasics
  Template for a kD-Tree of dimension D and type T
  This structure is a binary tree which contains a set of points and can search quickly every points which are included in a given sphere and especially the closest point from sphere center.
*/
  template<Index Dim, class Real=double>
    class KdTree {
 public:
  /// Construct the kdtree structure for a set of points given as a stl vector of Vect<Dim,Real>
  KdTree(const std::vector<Vect<Dim,Real> >& point);
  /// Construct the kdtree structure for a set of points given by stl iterators
  template <class InputIterator>
  KdTree(InputIterator first, InputIterator last)
    {std::vector<Vect<Dim,Real> > temp(first,last);
    recursiveCreate(temp, temp.begin(), temp.end(), topNode, 0);}
  /// Copy a KdTree structure
  KdTree(const KdTree<Dim,Real>& kdt);
  /// Destructor
  ~KdTree();

  /// search in a sphere of center and radius given the closest point form sphere center. return false if none has been found.
  bool searchClosestPoint(const Vect<Dim,Real>& center,
                          const double radius,
                          Vect<Dim,Real>& result);
  /// search all points in a sphere of center and radius. return false if none has been found.
  bool searchLClosestPoint(const Vect<Dim,Real>& center,
                          const double radius,
                          std::vector<Vect<Dim,Real> >& result);

 protected:
  /*! \class KdTreeNode KdTree.h basics/KdTree.h
    A node of the kD-tree
  */
  class KdTreeNode
    {
    public:
      /// the point
      Vect<Dim,Real> point;
      /// the top node of left sub-tree
      KdTreeNode *leftSon;
      /// the top node of right sub-tree
      KdTreeNode *rightSon;

    public:
      /// Constructor : copy point and initialise to empty sub-trees (i.e. this node is en end-node)
      KdTreeNode( const Vect<Dim,Real>& v );
      /// Copy Constructor : copy node and sub-trees
      KdTreeNode( const KdTreeNode & node );
      /// Destructor : destruct sub-trees too
      ~KdTreeNode();
    };
  typedef KdTreeNode* KdTreeNodePtr;

 protected:
  /// The top-node of the kDTree
  KdTreeNode * topNode;

 private:
  /// current dimension used for sorting during construction
  static int current_dimension;
  /// The sort function used during construction
  friend struct sort_function : binary_function<Primitive, Primitive, bool> {
    bool operator()(const Primitive& x, const Primitive& y) const { return x(current_dimension) < y(current_dimension); }
  };

 private:
  /// The best node found so far during search
  KdTreeNode * best_node;
  /// The liste of nodes found so far during search
  std::vector<KdTreeNode*> best_nodeL;

  /// The limit distance (sphere radius) used during search
  double limit_distance, limit_distance2;
  /// The sphere center used during search
  Vect<Dim,Real> center;

 private:
  /// The recursive function used for constructing the kDTree
  void recursiveCreate(std::vector<Vect<Dim,Real> >& temp,
                       std::vector<Vect<Dim,Real> >::iterator first,
                       std::vector<Vect<Dim,Real> >::iterator last,
                       KdTreeNodePtr& node,
                       int dimension);
  /// The recursive function used for search of closest point
  void search(KdTreeNode* node, int dimension);
  /// The recursive function used for search of point within a sphere
  void searchL(KdTreeNode* node, int dimension);
};

template<Index Dim, class Real>
KdTree<Dim,Real>::KdTreeNode::KdTreeNode( const Vect<Dim,Real>& v)
  : point(v) {
  leftSon = 0;
  rightSon = 0;
}
/*----------------------------------------*/

template<Index Dim, class Real>
KdTree<Dim,Real>::KdTreeNode::KdTreeNode( const KdTreeNode & node )
  : point(node.point)
{
  if (node.leftSon)
    leftSon = new KdTreeNode(node.leftSon);
  else
    leftSon = 0;

  if (node.rightSon)
    rightSon = new KdTreeNode(node.RightSon);
  else
    rightSon = 0;
};
/*----------------------------------------*/

template<Index Dim, class Real>
KdTree<Dim,Real>::KdTreeNode::~KdTreeNode()
{
  if (rightSon)
    delete rightSon;
  if (leftSon)
    delete leftSon;
};
/*-------------------------------------------------------------------*/

template<Index Dim, class Real>
KdTree<Dim,Real>::KdTree(const std::vector<Vect<Dim,Real> >& point)
{
  std::vector<Vect<Dim,Real> > temp(point);
  recursiveCreate(temp, temp.begin(), temp.end(), topNode, 0);
};
/*----------------------------------------*/

template<Index Dim, class Real>
KdTree<Dim,Real>::KdTree(const KdTree<Dim,Real>& kdt)
{
  topNode=new KdTreeNode(kdt.topNode);
}
/*----------------------------------------*/

template<Index Dim, class Real>
KdTree<Dim,Real>::~KdTree()
{
  delete topNode;
};
/*-------------------------------------------------------------------*/

template<Index Dim, class Real>
void KdTree<Dim,Real>::recursiveCreate(std::vector<Vect<Dim,Real> >& temp,
                                       std::vector<Vect<Dim,Real> >::iterator first,
                                       std::vector<Vect<Dim,Real> >::iterator last,
                                       KdTreeNodePtr& node,
                                       int dimension)
{
  dimension %= Dim;
  current_dimension = dimension;

  int size = last - first;
  std::vector<Vect<Dim,Real> >::iterator middle=first+(size/2);
  nth_element(first, middle, last, sort_function());

  node = new KdTreeNode(*middle);

  if (middle-first > 0)
    recursiveCreate(temp, first, middle, node->leftSon, dimension+1);
  if (last-middle > 1)
    recursiveCreate(temp, middle+1, last, node->rightSon, dimension+1);
};
/*-------------------------------------------------------------------*/

template<Index Dim, class Real>
void KdTree<Dim,Real>::search(KdTreeNode* node, int dimension)
{
  dimension %= Dim;
  int i=dimension;

  double c2=node->point(i);
  double c1=center(i);

  double cdist= ::fabs(c2-c1);

  if (cdist<=limit_distance) {
    double dist2=Dist2(node->point,center);
    if (dist2<=limit_distance2) {
      limit_distance2=dist2;
      limit_distance= ::sqrt(dist2);
      best_node=node;
    }
  }
  if (((c1-limit_distance)<c2) && (node->leftSon))
    search(node->leftSon,dimension+1);
  if (((c2-limit_distance)<c1) && (node->rightSon))
    search(node->rightSon,dimension+1);
};
/*-------------------------------------------------------------------*/

template<Index Dim, class Real>
bool KdTree<Dim,Real>::searchClosestPoint(const Vect<Dim,Real>& position,
                                          const double radius,
                                          Vect<Dim,Real>& result)
{
  center = position;
  limit_distance=radius;
  limit_distance2=radius*radius;
  best_node=0;

  search(topNode,0);
  if (best_node)
    {
      result = best_node->point;
      return(true);
    }
  else
    return(false);
};

/*-------------------------------------------------------------------*/

template<Index Dim, class Real>
void KdTree<Dim,Real>::searchL(KdTreeNode* node, int dimension)
{
  dimension %= Dim;
  int i=dimension;

  double c2=node->point(i);
  double c1=center(i);

  double cdist= ::fabs(c2-c1);

  if ((cdist<=limit_distance) && (Dist2(node->point,center)<=limit_distance2)) {
    best_nodeL.push_back(node);
  }
  if (((c1-limit_distance)<c2) && (node->leftSon))
    searchL(node->leftSon,dimension+1);
  if (((c2-limit_distance)<c1) && (node->rightSon))
    searchL(node->rightSon,dimension+1);
};
/*-------------------------------------------------------------------*/

template<Index Dim, class Real>
bool KdTree<Dim,Real>::searchLClosestPoint(const Vect<Dim,Real>& position,
                                           const double radius,
                                           std::vector<Vect<Dim,Real> >& result)
{
  center = position;
  limit_distance=radius;
  limit_distance2=radius*radius;
  best_nodeL.clear();
  result.clear();

  searchL(topNode,0);

  if (best_nodeL.size())
    {
      std::vector<KdTreeNode*>::iterator it;
      for (it=best_nodeL.begin(); it!=best_nodeL.end(); it++)
        result.push_back((*it)->point);
      return(true);
    }
  else
    return(false);
};

template<Index Dim, class Real>
int KdTree<Dim,Real>::current_dimension;

}

#endif
