#ifndef _NORMAL3D_H_
#define _NORMAL3D_H_


#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "basics/Normal3D.h"
#endif
#include "basics/Constants.h"
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Matrix3x3.h>
#include <linearAlgebra/Matrix4x4.h>
#include <linearAlgebra/Rotation3D.h>

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {

class Normal3D;

double DistAngle(const Normal3D& n1, const Normal3D& n2) throw();
double DistVect(const Normal3D& n1, const Normal3D& n2) throw();
double DistAngleNO(const Normal3D& n1, const Normal3D& n2) throw();
double DistVectNO(const Normal3D& n1, const Normal3D& n2) throw();
double DistTens(const Normal3D& n1, const Normal3D& n2) throw();
//Normal3D operator*(const RotMat3D& m, const Normal3D& v) throw();
// Normal3D operator*(const RigidMat3D& m, const Normal3D& v) throw();
// Normal3D operator*(const RotVec3D& m, const Normal3D& v) throw();
// Normal3D operator*(const RigidVec3D& m, const Normal3D& v) throw();

/*! \class Normal3D Normal3D.h basics/Normal3D.h
  \ingroup libBasics
  \brief 3D rotation Matrix
*/
class Normal3D : protected Vec3<double> {
public:
  /*! \class ErrNonNormal Normal3D.h basics/Normal3D.h
    \ingroup libBasics
    \brief Non rotration matrix excpetion
  */
  class ErrNonNormal : public Exception {
  public:
    // constructor
    ErrNonNormal(const std::string &from) throw() :
      Exception(from, "Non Normal3D exception") {};
  };

  /** @name Constructors */
  //@{
  /// default constructor ; initializes to identity matrix
  Normal3D() throw() : Vec3<double>(VIM_NOINIT)
    {org();}

  /// Copy constructor
  Normal3D(const Normal3D & r) throw() : Vec3<double>(r)
    {};
  /// operator =
  Normal3D& operator=(const Normal3D & r) throw()
    {this->copy((const Vec3<double>&)r);return *this;}

  /// Initialization from a vector
  Normal3D(double v0,double v1,double v2) throw(ErrNonNormal) : Vec3<double>(v0,v1,v2)
    {normalise(*this);}
  /// Initialization from a vector
  Normal3D(const double v[3]) throw(ErrNonNormal) : Vec3<double>(v)
    {normalise(*this);}
  /// Initialization from a vector
  Normal3D(const Vec3<double>& v) throw(ErrNonNormal) : Vec3<double>(v)
    {normalise(*this);}

  /// set rotation matrix from a vector
  Normal3D& operator=(const Vec3<double>& v) throw(ErrNonNormal)
    {normalise(v);return *this;}

  /// Destructor
  ~Normal3D() throw()
    {};
  //@}

  /** @name Data members access */
  //@{
  /// Access to coord i
  double operator()(const int i) const  throw(ErrRange)
    {return Vec3<double>::operator()(i);}

  /// Set to origine (1,0,0)
  void org() throw()
    {val[0]=1;val[1]=0;val[2]=0;}

  /// Set to origine (1,0,0)
  Normal3D& Org() throw()
    {org();return *this;}

  /// Uniform random normal
  void rnd() throw();

  /// Uniform random normal
  Normal3D& Rnd() throw()
    {rnd();return *this;}

  /// Gaussian random normal
  void rndgaussian(const Normal3D& n, double ecarttype) throw();

  /// Gaussian random normal
  Normal3D& RndGaussian(const Normal3D& n, double ecarttype) throw()
    {rndgaussian(n,ecarttype);return *this;}
  //@}

  /** @name Mathematical operations */
  //@{
  /// set normal by normalising a vector
  void normalise(const Vec3<double>& v) throw(ErrNonNormal);
  /// set normal by normalising a vector
  Normal3D& Normalise(const Vec3<double>& v) throw(ErrNonNormal)
    {normalise(v);return *this;}

  /// Intrinsic Distance between normals : angle( n1,n2)
  friend double DistAngle(const Normal3D& n1, const Normal3D& n2) throw();

  /// Euclidian Distance between normals : ||n1-n2||=2.sin(angle(n1,n2)/2)
  friend double DistVect(const Normal3D& n1, const Normal3D& n2) throw();

  /// Intrinsic Distance between non-oriented normals : angle( n1,n2)
  friend double DistAngleNO(const Normal3D& n1, const Normal3D& n2) throw();

  /// Euclidian Distance between non-oriented normals : ||n1-n2||=2.sin(angle(n1,n2)/2)
  friend double DistVectNO(const Normal3D& n1, const Normal3D& n2) throw();

  /// Tensorial Distance between normals : ::sqrt(1-(n1.n2)^2)=sin(angle(n1,n2))
  friend double DistTens(const Normal3D& n1, const Normal3D& n2) throw();

  /// Invert a normal (take the opposite)
  inline void inv(const Normal3D& R) throw()
    {val[0]=-R.val[0];val[1]=-R.val[1];val[2]=-R.val[2];}
  /// Invert a rotation matrix
  inline Normal3D& Inv( const Normal3D& R) throw()
    {this->inv(R); return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  inline Normal3D getInverse(void) const throw()
    {return Normal3D().Inv(*this);}
  /// invert this matrix (this matrix is modified)
  inline void invert(void) throw()
    {this->inv(*this);}

  /// Apply rotation matrix to a normal
  void Apply(const RotMat3D& r, const Normal3D& n) const throw()
    {r.Apply((const Vec3<double>&)n,(Vec3<double>&)(*this));}
  /// Apply rotation matrix to a normal
  void Apply(const RotVec3D& r, const Normal3D& n) const throw()
    {r.Apply((const Vec3<double>&)n,(Vec3<double>&)(*this));}
  //@}

  /// Return canonical rotation R so that R*this=m
  void m2Rnm(const Normal3D& m, RotVec3D& Rnm) const throw();
  /// Return canonical rotation R so that R*m=this
  void m2Rmn(const Normal3D& m, RotVec3D& Rmn) const throw()
    {m.m2Rnm(*this,Rmn);}

  /// Return canonical rotation R so that R*this=m
  void m2Rnm(const Normal3D& m, RotMat3D& Rnm) const throw()
    {RotVec3D tmp;m2Rnm(m,tmp);Rnm=tmp;}
  /// Return canonical rotation R so that R*m=this
  void m2Rmn(const Normal3D& m, RotMat3D& Rmn) const throw()
    {RotVec3D tmp;m2Rmn(m,tmp);Rmn=tmp;}

  void N2mnTens(const Normal3D& m, Vec3<double>& mn) const throw();
  void N2mnVect(const Normal3D& m, Vec3<double>& mn) const throw();
  void N2mnAngl(const Normal3D& m, Vec3<double>& mn) const throw();
  void mnAngl2N(const Normal3D& m, const Vec3<double>& mn) throw();
  void mnTens2N(const Normal3D& m, const Vec3<double>& mn) throw();
  void mnVect2N(const Normal3D& m, const Vec3<double>& mn) throw();

  /// compute approximate mean normal using vectorial mean
  template<class InputIterator>
  void meanVect(InputIterator first, InputIterator last) throw(ErrNonNormal)
  {
    Vec3<double> vectmean;//vectmean.resetToNull();
    for (InputIterator inormal=first; inormal!=last; inormal++)
      vectmean+=*inormal;
    normalise(vectmean);
  }

  /// compute approximate mean normal using tensorial mean
  template<class InputIterator>
  void meanTens(InputIterator first, InputIterator last) throw(ErrNonNormal)
  {
    Sym<3,double> tens,tmp;Matrix<3,double> evect;Vect<3,double> eval;
    for (InputIterator inormal=first; inormal!=last; inormal++)
      tens+=tmp.AutoCorrel(*inormal);
    tens.DiagonalizeSort(evect,eval);
    val[0]=evect(0,0);val[1]=evect(1,0);val[2]=evect(2,0);
  }

  /// compute approximate mean normal using tensorial mean
  template<class InputIterator>
  void meanTensO(InputIterator first, InputIterator last) throw(ErrNonNormal)
  {
    this->meanTens(first,last);
    double s=0;
    for (InputIterator inormal=first; inormal!=last; inormal++)
      s+=DotProd(*this,*inormal);
    if (s<0) {val[0]=-val[0];val[1]=-val[1];val[2]=-val[2];}
  }

#define MAX_NB_ITER_MEAN 20
  /// compute real mean normal using iterative gradient descent
  template<class InputIterator>
  void meanAngl(InputIterator first, InputIterator last) throw(ErrNonNormal)
  {
    //TODO:est-ce vraiment le meilleur point de depart?
    *this=*first;
    for (int iter=0; iter<MAX_NB_ITER_MEAN; iter++) {
      Vec3<double> mmn,tmp;
      int nbnormals=0;
      mmn.resetToNull();
      for (InputIterator inormal=first; inormal!=last; inormal++) {
        inormal->N2mnAngl(*this,tmp);
        mmn+=tmp;
        nbnormals++;
      }
      mmn*=1/(double)nbnormals;
      if (mmn.norm()<0.0000001) break;
      this->mnAngl2N(*this,mmn);
    }
  }

  /// compute variance of a set of normals w.r.t. this normal
  template<class InputIterator>
  double varianceAround(InputIterator first, InputIterator last) throw()
  {
    double var=0,a;int weight=0;
    for (InputIterator inormal=first; inormal!=last; inormal++) {
      a=DistAngle(*this,*inormal);
      var+=a*a;
      weight++;
    }
    return weight ? var/(double)weight : 0;
  }

  /// compute standard-deviation of a set of normals w.r.t. this normal
  template<class InputIterator>
  double sigmaAround(InputIterator first, InputIterator last) throw()
  {
    return ::sqrt(varianceAround(first,last));
  }

  /** @name I/O operations */
  //@{
  /// Print on an output stream
  friend STD_OSTREAM& operator<<(STD_OSTREAM& os, const Normal3D& rot) throw()
    {return os<<(Vec3<double>&) rot;}
  //@}


  static void Test();
};

/// Apply rotation matrix to a vector
//inline Normal3D operator*(const RotMat3D& m, const Normal3D& n) throw()
//  {return ((const Matrix3x3<double>&)m)*((const Vec3<double>&)n);}

} // namespace yav

#endif

