#ifndef OBJECTSLIST_H
#define OBJECTSLIST_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/ObjectsList.h"
#include <basics/Exception.h>
#include <basics/TclString.h>
#include <basics/ListElement.h>
#include <list>

namespace yav {

template <class T> class ObjectsListSTLiterator;
template <class T> class ObjectsListIterator;



/*! \class EmptyObjectsList ObjectsList.h basics/ObjectsList.h
  \ingroup libBasics
  parent class for a class that want to know when an object list becomes empty
*/
template <class T> class EmptyObjectsList {
 protected:
  /// that object kind id
  ObjectsList<T>* observed;

 public:
  /** @name constructors */
  //@{
  /// default constructor: no observed list
  EmptyObjectsList() : observed(0) { }
  /** observe given list
      @param list observed list */
  EmptyObjectsList(ObjectsList<T>& list) : observed(&list) {
    observed->warnMe(this);
  }
  //@}

  /** @name destructors */
  //@{
  virtual ~EmptyObjectsList() {
    if(observed) observed->doNotWarnMe(this);
  }
  //@}

  /// gets empty callback function
  virtual void empty(void) const = 0;
};


/*! \class ObjectsList ObjectsList.h basics/ObjectsList.h
  \ingroup libBasics
  List of object
  Each list element should inherit from ListElement<T> 
*/
 template <class T> class ObjectsList {

 public:
  /** @name Exception classes */
  //@{
   /*! \class NotFound ObjectsList.h basics/ObjectsList.h
    \ingroup libBasics
     object not found in list
   */
  class NotFound : public Exception {
  public:
    NotFound(const std::string& from): Exception(from,
              "Requested element could not be found in list") {
}
  };
  /*! \class AlreadyInList ObjectsList.h basics/ObjectsList.h
    \ingroup libBasics
    attempt to add object twice in same list
  */
  class AlreadyInList : public Exception {
  public:
    AlreadyInList(const std::string& from) : Exception(from,
              "Attempt to add an element that allready belongs to a list") {}
  };
  /*! \class Empty ObjectsList.h basics/ObjectsList.h
    \ingroup libBasics
    attempt to remove an object from an empty list
  */
  class Empty : public Exception {
  public:
    Empty(const std::string& from): Exception(from,
              "Attempt to reach an element in an empty list") {
}

  };
  /*! \class NullPointer ObjectsList.h basics/ObjectsList.h
    \ingroup libBasics
    attempt to add 0 pointer to object list
  */
  class NullPointer: public Exception {
  public:
    NullPointer(const std::string& from): Exception(from,
              "Attempt to add a null pointer element in list") {
}
  };
  //@}

 private:
  /// list first element
  ListElement<T> *head;
  /// list last element
  ListElement<T> *tail;
  /// list number of elements
  unsigned int count;
  /// objects to warn when list get empty
  std::list<EmptyObjectsList<T> *> warn;

 public:
  /** @name Constructors */
  //@{
  /// default constructor: creates an empty list
  ObjectsList() : head(0), tail(0), count(0) {}
  /// copy constructor: copy each list element
  ObjectsList(const ObjectsList<T>& list);
  //@}


  /** @name Destructor */
  //@{
  /// delete each list element
  virtual ~ObjectsList() { 
	 // warn.erase(warn.begin(),warn.end());
	  while(head) delete head; 
  }
  //@}

  /** adds 'me' to list of object to warn when list is empty
      @param me object to warn */
  void warnMe(EmptyObjectsList<T> *me) { warn.push_back(me); }
  /** removes 'me' from list of object to warn when list is empty
      @param me object to warn */
  void doNotWarnMe(EmptyObjectsList<T> *me) { warn.remove(me); }

  /// is list empty ?
  bool isEmpty(void) const { return count ? false : true; }
  /// get list number of elements
  unsigned int getCount(void) const { return count; }
  /// returns first list element
  ListElement<T> *getHead(void) const { return head; }
  /// returns last list element
  ListElement<T> *getTail(void) const { return tail; }

  /** assignement: destruction of this list and copy of new one
      @param  list  list to compare with */
  virtual void operator=(const ObjectsList<T>& list);
  /** list concatenation (this list and given list objects are copied
      and concatenated in a new list)
      @param list list to concatenate*/
  virtual ObjectsList<T> operator+(const ObjectsList<T>& list);
  /** list concatenation (given list objects are copied and concatenated
      to this list)
      @param list list to concatenate*/
  virtual void operator+=(const ObjectsList<T>& list);

  /** appends an element at end of list
      @param obj object to append
      @exception NullPointer obj should not be 0
      @exception AlreadyInList obj is already a member of this list */
  virtual void append(ListElement<T> *obj) throw(NullPointer,AlreadyInList);
  /** appends an element at end of list
      @param obj object to append
      @exception NullPointer obj should not be 0
      @exception AlreadyInList obj is already a member of this list */
  virtual void operator+=(ListElement<T> *obj)
    throw(NullPointer,AlreadyInList) { ObjectsList<T>::append(obj); }
  /** prepends an element at list beginning
      @param obj object to prepend
      @exception NullPointer obj should not be 0
      @exception AlreadyInList obj is already a member of this list */
  virtual void prepend(ListElement<T> *obj) throw(NullPointer,AlreadyInList);

  /** remove an element from the list
      @param obj object to remove
      @exception NotFound obj is not a member of this list */
  virtual void remove(ListElement<T> *obj) throw(NotFound);
  /** remove an element from the list
      @param obj object to remove
      @exception NotFound obj is not a member of this list */
  virtual void operator-=(ListElement<T> *obj) throw(NotFound) { remove(obj); }

  /** removes the first list element
      @exception Empty if there are no elements in list */
  virtual T* removeFirst(void) throw(Empty);
  /** removes last list element
      @exception Empty if there are no elements in list */
  virtual T* removeLast(void) throw(Empty);

  /** returns first list element */
  T* getFirst(void) const;
  /** returns last list element
      @exception Empty if there are no elements in list */
  T* getLast(void) const throw(Empty);
  /** Returns nth list element.
      Warning: this operator is unefficient.
      @param i object rank in list */
  virtual T* operator[](unsigned int i) const
    throw(NotFound /* if n is out of list range */);

  /// list gets empty
  void empty(void);

  public:
  /// returns an iterator on the first element of the list
  ObjectsListSTLiterator<T> begin() const;
  /// returns an iterator on the last element of the list
  ObjectsListSTLiterator<T> end() const;

  /// set ObjectsList<T>::iterator as ObjectsListSTLiterator<T>
  typedef ObjectsListSTLiterator<T> iterator;
 /// set ObjectsList<T>::Iterator as ObjectsListIterator<T>
  typedef ObjectsListIterator<T> Iterator;

  friend class ListElement<T>;
  friend class ObjectsListSTLiterator<T>;
  friend class ObjectsListIterator<T>;
};

 /*! \class ObjectsListIterator ObjectsList.h basics/ObjectsList.h
   \ingroup libBasics
   Iterator on an ObjectsList<T> 
 */
 template <class T> class ObjectsListIterator {
  public:
    /** @name iterator exception class */
    //@{
   /*! \class AccessBeyondList ObjectsList.h basics/ObjectsList.h
     \ingroup libBasics
     accessing an object out of the list
   */
    class AccessBeyondList : public Exception {
    public:
                AccessBeyondList(const std::string& from) :Exception(from,
                        "Attempt to access an element beyond list bounds") {}
                };
    //@}

  private:
    /// list on which to iterate
    const ObjectsList<T> *list;
    /// current list element
    ListElement<T> *current;

  public:
    /** @name Constructors */
    //@{
    /// constructs an iterator for given object list
    ObjectsListIterator(const ObjectsList<T>& l) : list(&l), current(l.head) {}
    /// copy constructor
    ObjectsListIterator(const ObjectsListIterator<T> &t);
    //@}

    /** returns current object then goes to following element
        @exception AccessBeyondList iterator reached list bounds */
    inline T *operator++(int)
      throw(AccessBeyondList);
    /** goes to following element and returns it
        @exception AccessBeyondList iterator reached list bounds */
    inline T *operator++(void)
      throw(AccessBeyondList);
    /** returns current object then goes to previous element
        @exception AccessBeyondList iterator reached list bounds */
    inline T *operator--(int)
      throw(AccessBeyondList);
    /** goes to previous element and returns it
        @exception AccessBeyondList iterator reached list bounds */
    inline T *operator--(void) throw(AccessBeyondList);
    /** returns current element
        @exception AccessBeyondList iterator reached list bounds */
    inline T *operator()(void) const throw(AccessBeyondList);
    /// goes to first list element
    void first(void) { current = list->head; }
    /// goes to last list element
    void last(void) { current = list->tail; }
    /** Goes to nth element and returns it.
        Warning: this operator is unefficient.
        @exception AccessBeyondList iterator reached list bounds */
    inline T *operator[](unsigned int)
      throw(AccessBeyondList);

    /** returns true if list end was reached and there is no
        current element anymore */
    bool isAtEnd(void) const { return current ? false : true; }

    friend class ObjectsList<T>;
    friend class ListElement<T>;
  };

 /*! \class ObjectsListSTLiterator ObjectsList.h basics/ObjectsList.h
   \ingroup libBasics
   STL iterator on an ObjectsList<T> 
 */
 template <class T> class ObjectsListSTLiterator {
  public:
    /** @name iterator exception class */
    //@{
   /*! \class AccessBeyondList ObjectsList.h basics/ObjectsList.h
     \ingroup libBasics
     accessing an object out of the list
   */
    class AccessBeyondList : public Exception {
    public:
      AccessBeyondList(const std::string& from) :Exception(from,
                                                           "Attempt to access an element beyond list bounds") {}
    };
    //@}
    /// iterator category type (required by STL)
    typedef std::bidirectional_iterator_tag iterator_category;
    /// iterator value type (required by STL)
    typedef T* value_type;
    /// iterator difference type (required by STL)
    typedef ptrdiff_t difference_type;
    /// iterator pointer type (required by STL)
    typedef T** pointer;
    /// iterator reference type (required by STL)
    typedef T*  reference;

  protected:
    /// list on which to iterate
    const ObjectsList<T> *list;
    /// current list element
    ListElement<T> *current;

  public:
    /** @name Constructors */
    //@{
    /// constructs an iterator without any information
    ObjectsListSTLiterator() :list(0),current(0) {}
    /// constructs an iterator without any information
    ObjectsListSTLiterator(ListElement<T>  *e);
    /// copy constructor
    ObjectsListSTLiterator(const ObjectsListSTLiterator<T> &t);
    //@}

    /** @name destructor */
    //@{
    virtual ~ObjectsListSTLiterator() {}
    //@}

    /** returns current object then goes to following element */
    inline ObjectsListSTLiterator operator++(int) throw(AccessBeyondList);
    /** goes to following element and returns it */
    inline ObjectsListSTLiterator &operator++(void) throw(AccessBeyondList);
    /** returns current object then goes to previous element */
    inline ObjectsListSTLiterator operator--(int) throw(AccessBeyondList);
    /** goes to previous element and returns it */
    inline ObjectsListSTLiterator &operator--(void) throw(AccessBeyondList);
    /** returns current element */
    inline T *operator*() const;
    /// goes to first list element
    inline void first(void) { current = list->head; }
    /// goes to last list element
    inline void last(void) { current = list->tail; }
    /** returns true if list end was reached and there is no
        current element anymore */
    inline bool isAtEnd(void) const { return current ? false : true; }

    /** returns true if the iterator t points on the same item
        as this iterator  */
    inline bool operator==(const ObjectsListSTLiterator &t)const;
    /** returns true if the iterator t points on a different
        item as this iterator  */
    inline bool operator!=(const ObjectsListSTLiterator &t)const;

    friend class ObjectsList<T>;
    friend class ListElement<T>;
  };

 /*! \class castiterator ObjectsList.h basics/ObjectsList.h
   \ingroup libBasics
   ObjectsList iterator returning V type objects
 */
template <class T,class V> class castiterator : public ObjectsListSTLiterator<T>{
 public:
  /** @name Constructors */
  //@{
  /// constructs an iterator without any information
  castiterator() : ObjectsListSTLiterator<T>(){}
  /// constructs an iterator without any information
  castiterator(ListElement<T>  *e) : ObjectsListSTLiterator<T>(e) {}
  /// copy  from basic constructor
  castiterator(const ObjectsListSTLiterator<T> &t) : ObjectsListSTLiterator<T>(t) {}
  /// copy constructor
  castiterator(const castiterator<T,V> &t) :ObjectsListSTLiterator<T>(t) {}
  //@}
  /** @name destructor */
  //@{
  virtual ~castiterator() {}
  //@}
  /** returns current element */
  virtual  V *operator*() const { return((V *) this->current);}
};

/*! \class CastIterator ObjectsList.h basics/ObjectsList.h
  \ingroup libBasics
  ObjectsList iterator returning V type objects
*/
template <class T,class V> class CastIterator {
 public:
  /*! \class AccessBeyondList ObjectsList.h basics/ObjectsList.h
    \ingroup libBasics
    when accessing an object out of the list
  */
  class AccessBeyondList : public Exception {
  public:
    AccessBeyondList(const std::string& from);
  };
 private:
  const ObjectsList<T> *list;        // list on which to iterate
  ListElement<T> *current;           // current list element
 private:
  ListElement<T> *getNext(ListElement<T> *elt) throw(AccessBeyondList /* if iterator reached list bounds */);
  ListElement<T> *getPrevious(ListElement<T> *elt) throw(AccessBeyondList /* if iterator reached list bounds */);

 public:
  // constructs an iterator for given object list
  CastIterator(const ObjectsList<T> &l);
  CastIterator(const CastIterator &t);

  // returns current object then goes to following element
  V *operator++(int)
    throw(AccessBeyondList /* if iterator reached list bounds */);
  // goes to following element and returns it
  V *operator++(void)
    throw(AccessBeyondList /* if iterator reached list bounds */);
  // returns current object then goes to previous element
  V *operator--(int)
    throw(AccessBeyondList /* if iterator reached list bounds */);
  // goes to previous element and returns it
  V *operator--(void)
    throw(AccessBeyondList /* if iterator reached list bounds */);
  // returns current element
  V *operator()(void) const
    throw(AccessBeyondList /* if iterator reached list bounds */);
  // goes to first list element
  void first(void) { current = list->getHead(); }
  // goes to last list element
  void last(void) { current = list->getTail(); }
  // Goes to nth element and returns it.
  // Warning: this operator is unefficient.
  V *operator[](unsigned int)
    throw(AccessBeyondList /* if iterator reached list bounds */);

  // returns true if list end was reached and there is no
  // current element anymore
  bool isAtEnd(void) const { return current ? false : true; }

  friend class ObjectsList<T>;
};

} // namespace yav

#include "ObjectsList.C"

#endif
