#ifndef POINTER_H
#define POINTER_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "basics/Pointer.h"
#include <basics/Owner.h>
#include <basics/Pointed.h>


namespace yav {


/*! \class ObjectPointer Pointer.h basics/Pointer.h
  \ingroup libBasics
  Smart pointer that resets to 0 when pointed object is deleted.
  A Pointer warns its owner when it resets. 
*/
  template <class T> class ObjectPointer {
 private:
  /// owner object
  Owner<T> *owner;
  /// real pointer
  T *pointer;
  /// indicates if the object T* is being deleted or not
  bool pointerIsDead;
  /// cannot create a pointer on Pointer, use Pointer<T **> instead.
  //  void *operator new(size_t) { return 0; }

  /// cannot copy a Pointer
  ObjectPointer(const ObjectPointer&) : owner(0), pointer(0) {
   pointerIsDead=false;
  }

  /// callback function when the pointed object is being deleted
  void dying(void);

 public:
  /** @name Constructors */
  //@{
  /** constructor from an owner and a pointed object
      @param owner   owner object
      @param pointer pointed object */
  ObjectPointer(Owner<T> *owner, T *pointer);

  /** constructor from an owner (pointer is set to 0)
      @param _owner   owner object */
  ObjectPointer(Owner<T> *_owner) : owner(_owner), pointer(0) {}
  /** constructor from nothing (pointer and owner are set to 0) */
  ObjectPointer() : owner(0), pointer(0) {}

  //@}

  /** @name Destructor */
  //@{
  /// destructor (do not destroy pointed object)
  ~ObjectPointer();
  //@}

  /** assign pointer from a Pointer object
      @param org pointer object */
  ObjectPointer operator=(const ObjectPointer& org) { *this = org.pointer; return *this; }
  /** assign pointer from a T* pointer
      @param t pointer */
  T *operator=(T *t);


  /// returns pointer value
  operator T *() const { return pointer; }
  /// cast pointer to void *
  operator void *() const { return (void *) pointer; }
  /// cast pointer to integer value
  operator size_t () const { return (size_t) pointer; }
  /// dereference pointer
  T& operator*(void) const { return *pointer; }
  /// () operator returns pointer value
  T *operator()(void) const { return pointer; }

  /// cast pointer to boolean
  operator bool () const { return (bool) (pointer!=0); }
  /// returns true if pointer is 0, false otherwise
  bool operator!(void) const { return (bool) !pointer; }

  /// return the owner of the pointer
  Owner<T> *getOwner() const {return owner;}
  /// set the owner
  void setOwner (Owner<T>* _owner){owner = _owner;}
  /// remove the connexion between pointer and pointer
  void removeConnexion() {
	  if (pointer)
	  ((Pointed<T> *) pointer)->pointedAway(this);
  }

  /** adds an integer to pointer (pointer is unchanged)
      @param u integer to add
      @return pointer + u */
  T *operator+(size_t u) { return pointer + u; }
  /** substracts an integer to pointer (pointer is unchanged)
      @param u integer to substract
      @return pointer - u */
  T *operator-(size_t u) { return pointer - u; }
  /** adds an integer to pointer
      @param u integer to add
      @return pointer + u */
  T *operator+=(size_t u) { return pointer += u; }
  /** substracts an integer to pointer
      @param u integer to substract
      @return pointer - u */
  T *operator-=(size_t u) { return pointer -= u; }
  /** pre-increments pointer
      @return pointer + 1 */
  T *operator++(void) { return ++pointer; }
  /** post-increments pointer
      @return pointer */
  T *operator++(int) { return pointer++; }
  /** pre-decrements pointer
      @return pointer - 1 */
  T *operator--(void) { return --pointer; }
  /** post-decrements pointer
      @return pointer */
  T *operator--(int) { return pointer--; }

  /** compares Pointer object to C pointer
      @param org C pointer
      @return true if this < org, false otherwise */
  bool operator<(const T *org) const { return pointer < org; }
  /** compares two pointers objects
      @param org Pointer to comapre with
      @return true if this < org, false otherwise */
  bool operator<(const ObjectPointer& org) const { return pointer < org.pointer; }
  /** compares Pointer object to C pointer
      @param org C pointer
      @return true if this > org, false otherwise */
  bool operator>(const T *org) const { return pointer > org; }
  /** compares two pointers objects
      @param org Pointer to comapre with
      @return true if this > org, false otherwise */
  bool operator>(const ObjectPointer& org) const { return pointer > org.pointer; }
  /** compares Pointer object to C pointer
      @param org C pointer
      @return true if this == org, false otherwise */
  bool operator==(const T *org) const { return pointer == org; }
  /** compares two pointers objects
      @param org Pointer to comapre with
      @return true if this == org, false otherwise */
  bool operator==(const ObjectPointer& org) const { return pointer == org.pointer; }
  /** compares Pointer object to C pointer
      @param org C pointer
      @return true if this != org, false otherwise */
  bool operator!=(const T *org) const { return pointer != org; }
  /** compares two pointers objects
      @param org Pointer to comapre with
      @return true if this != org, false otherwise */
  bool operator!=(const ObjectPointer& org) const { return pointer != org.pointer; }

  /// -> operator returns pointer value
  T *operator->(void) const { return pointer; }

  friend class Pointed<T>;
};

} // namespace yav

#include "Pointer.C"

#endif
