#ifndef PROBABILITY_H
#define PROBABILITY_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <basics/GenericHistogram.h>


namespace yav {
  
  
  /*! \class ProbabilityDensityFunction 
    \ingroup libBasics
   Generic probability density function 
   */
  class ProbabilityDensityFunction 
  {
  protected:
	  /// value below which the pdf is too small to be meaningful 
	  double minValue;
	  /// value above which the pdf is too small to be meaningful 
	  double maxValue;
	  /// indicates the number of samples that is required to scan the range of values
	  unsigned int numberSamples;
	   /// value below which the pdf is too small to be meaningful 
	  double meanValue;
	  /// value below which the pdf is too small to be meaningful 
	  double standardDeviation;
  public:
    
    /// default constructor:
	  ProbabilityDensityFunction(){}
    /// default virtual destructor:
	  virtual ~ProbabilityDensityFunction(){}
	  /// returns the probability of a given sample value
    virtual double getProbability(const double x) const=0;
	  /// returns the histogram value of a given sample value
    virtual double getHistogramValue(const double x) const=0;
	/// returns the value below which the pdf is too small to be meaningful 
	double getMinValue() const {
		return(minValue);
	}
	/// returns value above which the pdf is too small to be meaningful 
	double getMaxValue() const {
		return(maxValue);
	}
	  /// returns the number of samples that is required to scan the range of values
	unsigned int getNumberSamples() const {
		return numberSamples;
	}
	/// returns the mean value of the distribution
	virtual double getMeanValue() const {
		return meanValue;
	}
	/// returns the standard deviation of the distribution
	virtual double getStandardDeviation() const {
		return standardDeviation;
	}
  };

  
  /*! \class GaussianDistribution 
    \ingroup libBasics
   Gaussian Distribution eneric probability density function 
   */
  class GaussianDistribution : public ProbabilityDensityFunction
  {
  protected:
	  /// the maximum value number of samples for the mean value
	  double maxHistogramValue;
  public:
    
    /// default constructor:
	  GaussianDistribution(const double _mean, const double _sigma, const double _maxHistogram);
    /// default virtual destructor:
	  virtual ~GaussianDistribution(){}
	  /// returns the probability of a given sample value
    virtual double getProbability(const double x) const;
	  /// returns the histogram value of a given sample value
    virtual double getHistogramValue(const double x) const;
	
	
	/// returns maximum histogram
	double getMaxHistogram() const {
		return maxHistogramValue;
	}

  };
  
  /*! \class GaussianMixtureDistribution 
    \ingroup libBasics
   Gaussian Mixture Distribution  
   */
  class GaussianMixtureDistribution : public ProbabilityDensityFunction
  {
  protected:
	  /// the number of Gaussian Distributions
	  unsigned int numberDistributions;
	  /// value below which the pdf is too small to be meaningful 
	  std::vector<double> meanValues;
	  /// value below which the pdf is too small to be meaningful 
	  std::vector<double> standardDeviations;
	  /// the maximum histogram value for that Gaussian distribution
	  std::vector<double> maxHistogramValues;
	  /// the scale value to normalize the peaks
	  double scaleValue;
  public:
    
    /// default constructor:
	  GaussianMixtureDistribution(const unsigned int numberGaussians,
		  const std::vector<double> _meanValue,
          const std::vector<double> _sigma, const std::vector<double> _maxHistogram);
    /// default virtual destructor:
	  virtual ~GaussianMixtureDistribution(){}
	  /// returns the probability of a given sample value
    virtual double getProbability(const double x) const;
	/// returns the histogram value of a given sample value
	virtual double getHistogramValue(const double x) const;
	/// returns mean value
	double getMeanValue(const unsigned int i) const {
		return meanValues[i];
	}
	/// returns standard deviation
	double getStandardDeviation(const unsigned int i) const {
		return standardDeviations[i];
	}
	/// returns maximum histogram
	double getMaxHistogram(const unsigned int i) const {
		return maxHistogramValues[i];
	}
	/// returns the number of distributions
	unsigned int getNumberOfDistributions() const {
		return numberDistributions;
	}
  };

  
  /*! \class HistogramDistribution 
    \ingroup libBasics
   Distribution based on an histogram  
   */
  class HistogramDistribution : public ProbabilityDensityFunction
  {
  protected:
	  /// the histogram is stored in an array 
	  std::vector<double> histogramValue;
	  /// the scale value used to insure that the distributions sums to one
	double scaleDistribution;
  public:
    
    /// default constructor:
	  HistogramDistribution(const GenericHistogram *h,const double parzenWindow,
		  const double quantile=0.0); 

    /// default virtual destructor:
	  virtual ~HistogramDistribution(){}
	  /// returns the probability of a given sample value
    virtual double getProbability(const double x) const;
	  /// returns the histogram value of a given sample value
    virtual double getHistogramValue(const double x) const;
	/// the scale value used to insure that the distributions sums to one
	double getScaleDistribution() const {
		return scaleDistribution;
	}
  };

  GaussianMixtureDistribution *fuzzyKMeansWithEMAlgorithm(GenericHistogram *h,
	  std::vector<double> &sampleArray, const unsigned int numberSamples);


} // namespace yav


#endif // ifndef TRANSFORMATION2_H
