
// Properties : inspired from Borland Delphi and C++ Builder
//  designed to hide Get and Set functions for class menber variables
// Exemple :
// class Point {protected: float x;float y;
//              public: void SetX(float); float GetX();
//              public: void SetY(float); float GetY();}
// Point p;
// Here class user must use functions to access x and y : x0=p.GetX() and p.setX(x0)
//
// Now, with properties :
// class Point {protected: float x;float y;
//              public: void SetX(float); float GetX();
//              public: void SetY(float); float GetY();
//              public: PROPERTY_FF(float,X,Point,GetX,SetX);
//              public: PROPERTY_FF(float,Y,Point,GetY,SetY);}
// Point p;
// Now user can directly access x using X property : x0=p.X and p.X=x0
//
// Different property types differs by read and write strategy :
// PROPERTY_V* : use directly value of secondary variable in read access
// PROPERTY_F* : use function in read access
// PROPERTY_N* : don't have read access
// PROPERTY_*V, ... : idem for write access
//
// Then properties can be used to hide functions with no parameters :
// class Point {protected: float x;float y;
//              public:    float GetNorm();
//              public:    PROPERTY_FN(float,Norm,Point,GetNorm);}
// Point p;
// Here norm of point can be accessed using norm=p.Norm;
//  but p.Norm=norm is not possible
//
// Loss of time is about 10% with -O3, 80% with -g

#ifndef _PROPERTY_H
#define _PROPERTY_H

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {


#ifndef __BCPLUSPLUS__

#define PROPERTY_VV(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME) \
class class##PROPNAME { public: \
  inline class##PROPNAME& operator=(const PROPTYPE& v) \
    {((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->VARNAME=v;return *this;}; \
  inline operator PROPTYPE() const \
    {return ((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->VARNAME;}; \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;

#define PROPERTY_VF(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME,SETFUNCTION) \
class class##PROPNAME { public: \
  inline class##PROPNAME& operator=(const PROPTYPE& v) \
    {((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->SETFUNCTION(v);return *this;}; \
  inline operator PROPTYPE() const \
    {return ((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->VARNAME;}; \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;

#define PROPERTY_VN(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME) \
class class##PROPNAME { public: \
  inline operator PROPTYPE() const \
    {return ((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->VARNAME;} \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;

#define PROPERTY_FV(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME,GETFUNCTION) \
class class##PROPNAME { public: \
  inline class##PROPNAME& operator=(const PROPTYPE& v) \
    {((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->VARNAME=v;return *this;} \
  inline operator PROPTYPE() const \
    {return ((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->GETFUNCTION();} \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;

#define PROPERTY_FF(PROPTYPE,PROPNAME,OWNERTYPE,GETFUNCTION,SETFUNCTION) \
class class##PROPNAME { public: \
  inline class##PROPNAME& operator=(const PROPTYPE& v) \
    {((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->SETFUNCTION(v);return *this;} \
  inline operator PROPTYPE() const \
    {return ((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->GETFUNCTION();} \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;

#define PROPERTY_FN(PROPTYPE,PROPNAME,OWNERTYPE,GETFUNCTION) \
class class##PROPNAME { public: \
  inline operator PROPTYPE() const \
    {return ((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->GETFUNCTION();} \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;

#define PROPERTY_NV(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME) \
class class##PROPNAME { public: \
  inline class##PROPNAME& operator=(const PROPTYPE& v) \
    {((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->VARNAME=v;return *this;} \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;

#define PROPERTY_NF(PROPTYPE,PROPNAME,OWNERTYPE,SETFUNCTION) \
class class##PROPNAME { public: \
  inline class##PROPNAME& operator=(const PROPTYPE& v) \
    {((OWNERTYPE*)(this-&(((OWNERTYPE*)0)->PROPNAME)))->SETFUNCTION(v);return *this;} \
}; friend class class##PROPNAME; class##PROPNAME PROPNAME;


#else


#define PROPERTY_VV(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME) \
__property PROPTYPE    PROPNAME = { read=VARNAME, write=VARNAME }

#define PROPERTY_VF(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME,SETFUNCTION) \
__property PROPTYPE    PROPNAME = { read=VARNAME, write=SETFUNCTION }

#define PROPERTY_VN(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME) \
__property PROPTYPE    PROPNAME = { read=VARNAME }

#define PROPERTY_FV(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME,GETFUNCTION) \
__property PROPTYPE    PROPNAME = { read=GETFUNCTION, write=VARNAME }

#define PROPERTY_FF(PROPTYPE,PROPNAME,OWNERTYPE,GETFUNCTION,SETFUNCTION) \
__property PROPTYPE    PROPNAME = { read=GETFUNCTION, write=SETFUNCTION }

#define PROPERTY_FN(PROPTYPE,PROPNAME,OWNERTYPE,GETFUNCTION) \
__property PROPTYPE    PROPNAME = { read=GETFUNCTION }

#define PROPERTY_NV(PROPTYPE,PROPNAME,OWNERTYPE,VARNAME) \
__property PROPTYPE    PROPNAME = { write=VARNAME }

#define PROPERTY_NF(PROPTYPE,PROPNAME,OWNERTYPE,SETFUNCTION) \
__property PROPTYPE    PROPNAME = { write=SETFUNCTION }


#endif //__bcplusplus__


}

#endif

