#ifndef SHAREDMEMORY_H
#define SHAREDMEMORY_H


#pragma interface "basics/SharedMemory.h"

#include <errno.h>
#include <string.h>
#include <basics/Exception.h>

namespace yav {

  /*! \class SharedMemory SharedMemory.h basics/SharedMemory.h
    \ingroup libBasics
  */
class SharedMemory {
 public:
  /** @name Exception classes */
  //@{
  /*! \class CanNotCreateSegment SharedMemory.h basics/SharedMemory.h
    \ingroup libBasics
    Exception when segment couldn't be created */
  class CanNotCreateSegment :  public Exception {
  public:
    CanNotCreateSegment(const std::string& from,const std::string& msg) : Exception
      (from,msg){
    }
  };
  /*! \class CanNotCreateSemaphore SharedMemory.h basics/SharedMemory.h
    \ingroup libBasics
    Exception when segment couldn't be created */
  class CanNotCreateSemaphore :  public Exception {
  public:
    CanNotCreateSemaphore(const std::string& from,const std::string& msg) : Exception
      (from,msg){
    }
  };
  /*! \class IllegalData SharedMemory.h basics/SharedMemory.h 
    \ingroup libBasics
    Exception sent when the data given for segment creation are uncorrect 
  */
  class IllegalData :  public Exception {
  public:
    IllegalData(const std::string& from) : Exception
      (from,"Creating Shared Memory Segment with Null size"){
    }
  };
  //@}

  /** enum type to unambiguous creators */
  enum CreateWith { KEY, SIZE };

 protected:
  bool main; // wether the destruction of the object//
             //cause the destruction of the segment or not
  int shmid; // id of the segment
  int shmsz; // size of the segment
  int shmkey;// key used for both segment and semaphore
  void* shmaddr;// address where is mapped the segment
  int semid; // id of the semaphore

 public:
  /** @name Constructors */
  //@{
  /** constructor creating a new segment with the given key
      and the given size if key or size are 0 an error is generated */
  SharedMemory(int key,int size);
  /** constructor creating a new segment with the given size
      or key according to the value of CreateWith
      @param data the key or size of the segment
      @param mode if KEY then return a already existing segment
      that have been created with this key. If SIZE, create
      segment of that size the key is created automatically
      and can be retrieve with the getKey command*/
  SharedMemory(int data, CreateWith mode);
  //@}

  /** @name destructor */
  //@{
  /// dummy Destructor
  virtual ~SharedMemory();
  //@}

 private:
  /* function that create a segment with the given key and size */
  void createSegment(int key, int size);
  /* function that create a segment with the given key */
  void createSegment(int size);
  /* function that retrieve a segment with the given key */
  void getSegment(int key);


 public:
  /* return the id of the segment */
  inline int getShmid() { return shmid;}
  /* return the id of the semaphore */
  inline int getSemid() { return semid;}
  /* return the key used for both segment and semaphore */
  inline int getKey() { return shmkey;}
  /* return the size of the segment */
  inline int getSize() { return shmsz;}
  /* return the number of attaches */
  short getNbOfAttaches();
 protected:
  /* return the address where the segment is mapped */
  inline void* getAddress() {return shmaddr;}

 public:
  /* functions to read/write the segment with no check of the semaphore */
  inline size_t in(void* buf, int offset, size_t size){
    memcpy(buf,(void*)((size_t)shmaddr+offset),size);
    return size;
  }
  inline size_t ins(char* buf, int offset, size_t size){
    strncpy(buf,(char*)((size_t)shmaddr+offset),size);
    return size;
  }
  inline size_t out(const void* buf, int offset, size_t size){
    memcpy((void*)((size_t)shmaddr+offset),buf,size);
    return size;
  }
  inline size_t outs(const char* buf, int offset, size_t size){
    strncpy((char*)((size_t)shmaddr+offset),buf,size);
    return size;
  }

 public:
  /* functions to read/write the segment checking the semaphore */
 inline size_t read(void* buf, int offset, size_t size){
    getSemaphore();
    memcpy(buf,(void*)((size_t)shmaddr+offset),size);
    releaseSemaphore();
    return size;
  }
 inline size_t reads(char* buf, int offset, size_t size){
    getSemaphore();
    strncpy(buf,(char*)((size_t)shmaddr+offset),size);
    releaseSemaphore();
    return size;
  }
 inline size_t write(const void* buf, int offset, size_t size){
    getSemaphore();
    memcpy((void*)((size_t)shmaddr+offset),buf,size);
    releaseSemaphore();
    return size;
  }
 inline size_t writes(const char* buf, int offset, size_t size){
    getSemaphore();
    strncpy((char*)((size_t)shmaddr+offset),buf,size);
    releaseSemaphore();
    return size;
  }

 public:
 /* function to check if the semaphore is owned by the current process */
 bool checkSemaphore();
 /* function to get the semaphore, process might be queued
    @return 0 on success, -1 otherwise */
 int getSemaphore();
 /* function to get the semaphore, process won't be queued
    if the semaphore is not available then return -1 */
 bool getSemaphoreNoWait();
 /* function to realease the semaphore.
    Warning, be sure you actually own it
    @return 0 on success (always succeed), -1 otherwise */
 int releaseSemaphore();
};



} // namespace yav

#endif /// SHAREDMEMORY_H
