#ifndef STRING_H
#define STRING_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4290)
#endif

#pragma interface "basics/TclString.h"
#include <stdio.h>
#ifndef WIN32
#include <string.h>
#else
#include <string>
#endif
#include <ctype.h>
#include <stdarg.h>
#include <basics/Architecture.h>
#include <basics/Exception.h>
#include <basics/Bool.h>


namespace yav {



/*! \class TclString TclString.h basics/TclString.h
  \ingroup libBasics
  Substitute for C 'char *' object 
*/
class TclString {
public:

  /** @name Exception classes */
  //@{
  /*! \class OutOfString TclString.h basics/TclString.h
    \ingroup libBasics
    index beyond string size */
  class OutOfString : public Exception {
  public:
    /** constructor
        @param from calling procedure */
    OutOfString(const std::string& from) :
      Exception(from,
                "Attempt to access out of string") {}
  };

  /*! \class InvalidIndex TclString.h basics/TclString.h
    \ingroup libBasics
    index value incoherent
  */
  class InvalidIndex : public Exception {
  public:
    /** constructor
        @param from calling procedure */
    InvalidIndex(const std::string& from) :
      Exception(from,
                "Invalid index values in TclString constructor") {}
  };
  //@}

private:
  /// the string itself
  char *str;
  /// string length = strlen(str)+1 (i.e. includes final \0)
  size_t length;

public:
  /** @name Constructors */
  //@{
  /// default constructor
  TclString();
  /** construction from a C string
      @param init initialization string */
  TclString(const char *init);
  /** construction from a single character
      @param c    initialization character */
  TclString(const char c);
  /** construction from a C string, from index start to index stop (included)
      @param init  initialization string
      @param start start index
      @param stop  end index
      @exception InvalidIndex if start > stop
      @exception OutOfString if stop > strlen(init) */
  TclString(const char *init, const unsigned int start, const unsigned int stop)
    throw(InvalidIndex, OutOfString);
  /** construction from a TclString, from index start to index stop (included)
      @param init  initialization string
      @param start start index
      @param stop  end index
      @exception InvalidIndex if start > stop
      @exception OutOfString if stop > strlen(init) */
  TclString(const TclString &init, const unsigned int start, const unsigned int stop)
    throw(InvalidIndex,OutOfString);
  /** construction from a printf-like arguments list
      @param s    printf-like format string
      @param args argument list*/
  TclString(const char *s, va_list args) { isprintf(s, args); }
  /** copy constructor
      @param s    String to copy */
  TclString(const TclString &s);
  //@}

  /** @name Destructor */
  //@{
  /// dealocates string buffer
  ~TclString();
  //@}


private:

  /** allocates a string buffer
      @param size  string buffer size */
  TclString(const size_t size);
  /** allocates a buffer using length field and copy given C string
      @param init initialization string */
  void copy(const char *init);
  /** Realloc string buffer to n bytes.
      If n is less than TclString length, string is truncated.
      If n is 0, buffer is reallocated to exact string length.
      @param n new buffer size*/
  void realloc(size_t n = 0);
  /** Concatenates conc to string, reallocating if needed
      @param conc C string to concatenate */
  void concat(const char *conc);
  /** Concatenates c to string, reallocating if needed
      @param c character to concatenate */
  void concat(const char c);
  /** build string from a printf-like arguments list
      @param s    printf-like format string
      @param args argument list*/
  void isprintf(const char *s, va_list args);


public:
  /** builds string in a printf-like manner
      @param str printf-like format string
      @param ... format arguments */
  void sprintf(const char *str...);

  /// cast to (char *)
  operator char * () const { return str; }

  /** assignement from C string
      @param init initialization string */
  TclString& operator=(const char *init);
  /** assignement from TclString
      @param init initialization string */
  TclString& operator=(const TclString &init);

  /** concatenates a TclString
      @param concat string to append */
  TclString& operator+=(const TclString &concat);
  /** concatenates a C string
      @param concat string to append */
  TclString& operator+=(const char *concat);
  /** concatenates a TclString
      @param init string to append */
  TclString& append(const TclString &init) { return (*this) += init; }
  /** concatenates a C string
      @param init string to append */
  TclString& append(const char *init) { return (*this) += init; }
  /** prepends a TclString
      @param init string to append */
  TclString& prepend(const TclString &init);
  /** prepends a C string
      @param init string to append */
  TclString& prepend(const char *init);

  /** concatenates two TclStrings
      @param s2 string to concatenate to this */
  TclString operator+(const TclString &s2) const;
  /** concatenates a TclString and a C string
      @param s2 string to concatenate to this */
  TclString operator+(const char *s2 ) const;
  /** concatenates a C string and a TclString
      @param s1 C-string
      @param s2 TclString */
  friend TclString operator+(const char *s1, const TclString& s2);

  /** truncates i first characters
      @param i number of characters to remove */
  TclString& truncAtBegining(unsigned int i);
  /** truncates i last characters
      @param i number of characters to remove */
  TclString& truncAtEnd(unsigned int i);

  /// converts TclString to upper case
  void toUpper(void);
  /// converts TclString to lower case
  void toLower(void);
  /// converts TclString to 7 bits ascii characters (& 0x7F)
  void toAscii(void);
  /** converts nth character to upper case
      @param     n           character index
      @exception OutOfString n is out of string range */
  void toUpper(unsigned int n) throw(OutOfString);
  /** converts nth character to lower case
      @param     n           character index
      @exception OutOfString n is out of string range */
  void toLower(unsigned int n) throw(OutOfString);
  /** converts nth character to 7 bit ascii (& 0x7F)
      @param     n           character index
      @exception OutOfString n is out of string range */
  void toAscii(unsigned int n) throw(OutOfString);

#ifndef WIN32
  /** returns nth character reference
      @param     n           character index
      @exception OutOfString n is out of string range */
  char& operator[](int n) const throw(OutOfString);
#endif
  /// returns true if this string object is empty
  bool isEmpty(void) { return length <= 1 ? true : false; }

  /// returns string length
  unsigned int getLength(void) const { return (str ? length - 1 : 0); }


  /** returns true if both s and this have same value
      @param s TclString to compare */
  bool operator==(const TclString &s) const {
    return strcmp(str, s.str) ? false : true;
  }
  /** returns true if both s and this have same value
      @param s C string to compare */
  bool operator==(const char *s) const {
    return strcmp(str, s) ? false : true;
  }
  /** returns false if both s and this have same value
      @param s TclString to compare */
  bool operator!=(const TclString &s) const {
    return strcmp(str, s.str) ? true : false;
  }
  /** returns false if both s and this have same value
      @param s C string to compare */
  bool operator!=(const char *s) const {
    return strcmp(str, s) ? true : false;
  }

  /** returns true if this is less than s in the lexicographic order
      @param s TclString to compare */
  bool operator<(const TclString &s) const {
    return (strcmp(str, s.str)<0);
  }
  /** compares a TclString and a C string
      @param s1 C string
      @param s2 TclString */
  friend bool operator==(const char *s1, const TclString& s2) {
    return strcmp(s1, (char *) s2) ? false : true;
  }
  /** compares a TclString and a C string
      @param s1 C string
      @param s2 TclString */
  friend bool operator!=(const char *s1, const TclString& s2) {
    return strcmp(s1, (char *) s2) ? true : false;
  }

  /** Returns 0 if s and this are equal or a non-zero value otherwise.
      The size of a non-zero result is determined by the sign of the
      difference between the values of the first pair of bytes (both
      interpreted as unsigned char) that differ in the two compared objects
      @param s TclString to compare */
  int compare(const TclString &s) const {
    return strcmp(str, s.str);
  }
  /** Compares at most n first bytes of this and s. Stop if an End Of
      TclString is encountered before n bytes were read.
      Returns 0 if s and this are equal or a non-zero value otherwise.
      The size of a non-zero result is determined by the sign of the
      difference between the values of the first pair of bytes (both
      interpreted as unsigned char) that differ in the two compared objects
      @param s TclString to compare
      @param n maximum number of bytes to compare */
  int compare(const TclString &s, size_t n) const {
    return strncmp(str, s.str, n);
  }
  /** Returns 0 if s and this are equal or a non-zero value otherwise.
      The size of a non-zero result is determined by the sign of the
      difference between the values of the first pair of bytes (both
      interpreted as unsigned char) that differ in the two compared objects
      @param s C string to compare */
  int compare(const char *s) const {
    return strcmp(str, s);
  }
  /** Compares at most n first bytes of this and s. Stop if an End Of
      TclString is encountered before n bytes were read.
      Returns 0 if s and this are equal or a non-zero value otherwise.
      The size of a non-zero result is determined by the sign of the
      difference between the values of the first pair of bytes (both
      interpreted as unsigned char) that differ in the two compared objects
      @param s C string to compare
      @param n maximum number of bytes to compare */
  int compare(const char *s, size_t n) const {
    return strncmp(str, s, n);
  }

  /** converts to an std::string */
  std::string tostring(void) const {
    return std::string(str);
  }

  /** split a string in words splited by given separators
      @param array returned words array
      @param separators string of any word separator to use
      @return number of elements in array */
  int split(TclString**& array, char *separators = " ");

  /** returns the left most part of the string located before the separator character
      @param separator character
      @return the string baseName (EMPTY_STRING is no extension found */
  TclString baseName(char separator = '.') const;

  /** returns the right most part of the string located after the separator character
      @param separator character
      @return the string extension (EMPTY_STRING is no extension found */
  const char *extension(char separator = '.') const;

  /** Parse this string and substitue environnement variables.
      An environnement variable has the name following the '$' sign
      (maybe in between '{' and '}' */
  void expand(void);

public:
  /** returns given C string length
      @param s any C string, even 0 */
  static unsigned int getLength(const char *s) {
    return strlen(s);
  }
  /** returns given TclString length
      @param s any TclString */
  static unsigned int setLength(const TclString& s) {
    return s.getLength();
  }


  /** writes TclString on os and returns os
      @param os stream on which to write
      @param s  TclString to write */
  friend  STD_OSTREAM & operator<<( STD_OSTREAM & os,const TclString& s);
  /** reads TclString from is and returns is
      @param is stream from which to read
      @param s  TclString to read in */
  friend  STD_ISTREAM & operator>>( STD_ISTREAM & is, TclString& s);
};

#ifndef WIN32
/// the "" TclString constant
extern const TclString EMPTY_STRING;

/// the " " TclString constant
extern const TclString SPACE_STRING;

#else
/// the "" TclString constant
static const TclString EMPTY_STRING;

/// the " " TclString constant
static const TclString SPACE_STRING(" ");

#endif
/** writes TclString on os and returns os
    @param os stream on which to write
    @param s  TclString to write */
         STD_OSTREAM & operator<<( STD_OSTREAM & os,const TclString& s);
 /** reads TclString from is and returns is
     @param is stream from which to read
      @param s  TclString to read in */
    STD_ISTREAM & operator>>( STD_ISTREAM & is, TclString& s);

/*! \class CStringCompare TclString.h basics/TclString.h
  \ingroup libBasics
  C string comparison class for stl map 
*/
class CStringCompare {
public:
  /** @name Constructors */
  //@{
  /// default constructor (empty)
  CStringCompare() { }
  /// copy constructor (empty)
  CStringCompare(const CStringCompare&) { }



  /** assignement (empty)
      @return *this */
  CStringCompare& operator=(const CStringCompare&) { return *this; }

  /** strings comparison operator
      @param s1 first string to compare
      @param s2 second string to compare
      @return true if s1 < s2 */
  bool operator()(char *const &s1, char *const &s2) const {
    return strcmp(s1, s2) < 0;
  }
  //@}
};


} // namespace yav


#endif
