#include <classification/ExpectationMaximization.h>
#include <basics/Constants.h>
#include <basics/Math.h>
#include <string>
#include <iostream>
#include <stdio.h>
#include <vector>
#include <algorithm>
#include <limits>

namespace yav{

ExpectationMaximization::ExpectationMaximization(double** data, unsigned int nModes, unsigned int nData, unsigned int nDim, double wCov, unsigned int* dataRange, unsigned int minDim, unsigned int maxIte, double minImpro)
{
	this->data = data;
	this->nModes = nModes;
	this->nData = nData;
	this->nDim = nDim;
	this->wCov = wCov;
	this->dataRange = dataRange;
	this->minDim = minDim;
	this->maxIte = maxIte;
	this->minImpro = minImpro;

	sizeDat = nData*nDim;
	sizeExp = nModes*nData;
	sizeCen = nModes*nDim;
	sizeCov = nDim*nDim;
	isIndData = false;
	isMaxMode = false;
	isMaxData = false;
	isMaxData2= false;

	proba = new double [sizeExp];
	weightMode = new double [nModes];
	centers = new double [sizeCen];
	isReliable = new bool [nData];
	weightRange = new double [nData];
	minRange = new unsigned int [nData];
	maxRange = new unsigned int [nData];
	indData = new unsigned int [nData];
	maxMode = new double [nModes];
	maxData = new double [nData];
	maxData2= new double [nData];
	covMatrices.resize(nModes);

	for (unsigned int i=0; i<nModes; i++)
		covMatrices[i].resize(sizeCov);
}

ExpectationMaximization::~ExpectationMaximization()
{
	covMatrices.clear();
	if (maxData2) delete [] maxData2;
	if (maxData) delete [] maxData;
	if (maxMode) delete [] maxMode;
	if (indData) delete [] indData;
	if (maxRange) delete [] maxRange;
	if (minRange) delete [] minRange;
	if (weightRange) delete [] weightRange;
	if (isReliable) delete [] isReliable;
	if (centers) delete [] centers;
	if (weightMode) delete [] weightMode;
	if (proba) delete [] proba;
}

void ExpectationMaximization::em()
{
	unsigned int i;
	double Lo, Ln;

	invMatrices.resize(nModes);

	for (i=0; i<nModes; i++) 
		invMatrices[i].resize(sizeCov);

	if (dataRange==NULL)
	{
		nDataUsed = nData;

		// By default, all the data are reliable
		for (i=0; i<nData; i++)
		{
			isReliable[i] = true;
			weightRange[i] = 1.0;
			minRange[i] = 0;
			maxRange[i] = nDim-1;
		}
	}
	else
	{
		nDataUsed = 0;

		for (i=0; i<nData; i++)
		{
			if (((dataRange[(2*i)+1]-dataRange[2*i])+1)>=minDim)
			{
				isReliable[i] = true;
				nDataUsed++;
			}
			else
				isReliable[i] = false;

			weightRange[i] = (double)nDim/(double)((dataRange[(2*i)+1]-dataRange[2*i])+1);
			minRange[i] = dataRange[2*i];
			maxRange[i] = dataRange[(2*i)+1];
		}
	}

	// Initialize mode weights, mode centers and covariance matrices
	initEM();

	// EM algorithm
	for (nIte=0; nIte<maxIte; nIte++)
	{
		Estep();
		Mstep();
		Lo = pdf;

		// Check termination condition
		if (((nIte>0) && (fabs(Lo-Ln)<minImpro)) || isnan(Lo))
			break;
		else
			Ln = Lo;
	}

	logLike = Ln;

	invMatrices.clear();
}

void ExpectationMaximization::initEM()
{
	unsigned int i, j;

	FuzzyCMeans* myFCM = new FuzzyCMeans(data,nModes,nData,nDim,dataRange,minDim);

	// Initialize centers by Fuzzy C-Means
	myFCM->fcm();

	for (i=0; i<sizeCen; i++)
		centers[i] = myFCM->getCenters()[i];

	for (i=0; i<nModes; i++)
	{
		// Initialize covariance matrix
		getCovMatrix(myFCM,i);

		// Initialize mode weights
		weightMode[i] = 0.0;

		for (j=0; j<nData; j++)
			if(isReliable[j])
				weightMode[i] += myFCM->getMembership()[(j*nModes)+i];

		weightMode[i] /= nDataUsed;
	}


// CHECK
/*CERR << "INIT Weights:" << ENDL;
for (i=0; i<nModes; i++)
	CERR << weightMode[i] << " ";
CERR << ENDL << ENDL;

CERR << "INIT Centers:" << ENDL;
for (i=0; i<nModes; i++)
{
	for (j=0; j<nDim; j++)
		CERR << centers[(i*nDim)+j] << " ";
	CERR << ENDL;
}
CERR << ENDL;

for (i=0; i<nModes; i++)
{
	CERR << "INIT Covariance matrix #" << i << ENDL;
	for (j=0; j<nDim; j++)
	{
		for (unsigned int k=0; k<nDim; k++)
			CERR << covMatrices[i][(j*nDim)+k] << " ";
		CERR << ENDL;
	}
	CERR << ENDL;
}
CERR << ENDL;*/
// CHECK


	delete myFCM;
}

void ExpectationMaximization::Estep()
{
	unsigned int i, j, k;
	double a, sum, tmp, det;

	double* vector1 = new double [nDim];
	double* vector2 = new double [nDim];
	double* vector3 = new double [nDim];
	double* vector4 = new double [nDim];
	double* coef = new double [nModes];

	a = pow(6.2832,(0.5*(double)nDim));
	pdf = 0.0;

	// Compute inverse matrix and determinant
	for (i=0; i<nModes; i++)
	{
		getInvMatrix(covMatrices,i,det);
		coef[i] = a*sqrt(fabs(det));
	}

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			sum = 0.0;
	
			for (j=0; j<nModes; j++)
			{
				// Retrieve data and center
				for (k=0; k<nDim; k++)
				{
					if ((k>=minRange[i])&&(k<=maxRange[i]))
						vector1[k] = data[i][k];
					else
						vector1[k] = centers[(j*nDim)+k];

					vector2[k] = centers[(j*nDim)+k];
				}
				
				getSubMat(vector1, vector2, vector3);
				getMultMatByMat(invMatrices[j], vector3, vector4, nDim, nDim, i);
				tmp = exp(-0.5*getMultMatByMat(vector3, vector4, i))/coef[j];
				proba[(i*nModes)+j] = weightMode[j]*tmp;
				sum += proba[(i*nModes)+j];
			}

			pdf += log(sum);
	
			for (j=0; j<nModes; j++)
				proba[(i*nModes)+j] /= sum;
		}
		else
			for (j=0; j<nModes; j++)
				proba[(i*nModes)+j] = 0.0;
	}


// CHECK
/*for (i=0; i<nModes; i++)
{
	CERR << "[" << nIte << "] ESTEP Inverse matrix #" << i << ENDL;
	for (j=0; j<nDim; j++)
	{
		for (k=0; k<nDim; k++)
			CERR << invMatrices[i][(j*nDim)+k] << " ";
		CERR << ENDL;
	}
	CERR << ENDL;
}
CERR << ENDL;*/

/*CERR << "[" << nIte << "] ESTEP Proba:" << ENDL;
for (i=0; i<sizeExp; i++)
	CERR << proba[i] << " ";
CERR << ENDL << ENDL;*/
// CHECK


	delete [] coef; delete [] vector4; delete [] vector3; delete [] vector2; delete [] vector1;
}

void ExpectationMaximization::Mstep()
{
	unsigned int i, j, k;

	double* sum = new double[nDim];

	for (i=0; i<nModes; i++)
	{
		// Zeroing stuff
		weightMode[i] = 0.0;

		for (j=0; j<nDim; j++)
		{
			centers[(i*nDim)+j] = 0.0;
			sum[j] = 0.0;
		}
		
		for (j=0; j<sizeCov; j++)
			covMatrices[i][j] = 0.0;

		// Compute weights
		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				weightMode[i] += proba[(j*nModes)+i];
				
				for (k=minRange[j]; k<=maxRange[j]; k++)
				{
					centers[(i*nDim)+k] += (proba[(j*nModes)+i]*data[j][k]);
					sum[k] += proba[(j*nModes)+i];
				}
			}
		}

		for (k=0; k<nDim; k++)
			centers[(i*nDim)+k] /= sum[k];

		// Compute covariance mat
		getCovMatrix(i);

		// Normalize mode weights
		weightMode[i] /= (double)nDataUsed;
	}


// CHECK
/*for (i=0; i<nModes; i++)
{
	CERR << "[" << nIte << "] MSTEP Covariance matrix #" << i << ENDL;
	for (j=0; j<nDim; j++)
	{
		for (k=0; k<nDim; k++)
			CERR << covMatrices[i][(j*nDim)+k] << " ";
		CERR << ENDL;
	}
	CERR << ENDL;
}
CERR << ENDL;*/
// CHECK


}

void ExpectationMaximization::getSubMat(double* mat1, double* mat2, double* result)
{	
	for (unsigned int i=0; i<nDim; i++)
		result[i] = mat1[i]-mat2[i];
}

double ExpectationMaximization::getMultMatByMat(double* mat1, double* mat2, unsigned int &id)
{
	unsigned int i;
	double result;

	result = 0.0;
	
	for (i=0; i<nDim; i++)
		result += weightRange[id]*mat1[i]*mat2[i];

	return result;
}

void ExpectationMaximization::getMultMatByMat(std::vector<double> &mat1, double* mat2, double* result, unsigned int &nRows, unsigned int &n, unsigned int &id)
{
	for (unsigned int i=0; i<nRows; i++)
		result[i] = 0.0;
	
	for (unsigned int i=0; i<nRows; i++)
		for (unsigned int j=0; j<n; j++)
			for (unsigned int k=0; k<1; k++)
				result[i+k] += weightRange[id]*mat1[(i*n)+j]*mat2[j+k];
}

void ExpectationMaximization::getCovMatrix(unsigned int &idMode)
{
	unsigned int i, j, k;
	double dist1, dist2, num, val, trace;

	double* covWeights = new double[sizeCov];

	for (i=0; i<sizeCov; i++)
		covWeights[i] = 0.0;

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			val = proba[(i*nModes)+idMode];

			for (j=minRange[i]; j<=maxRange[i]; j++)
			{
				dist1 = data[i][j]-centers[(idMode*nDim)+j];

				for (k=j; k<=maxRange[i]; k++)
				{
					dist2 = data[i][k]-centers[(idMode*nDim)+k];
					num = val*dist1*dist2;
					covMatrices[idMode][(j*nDim)+k] += num;
					covWeights[(j*nDim)+k] += val;

					if (k>j)
					{
						covMatrices[idMode][(k*nDim)+j] += num;
						covWeights[(k*nDim)+j] += val;
					}
				}
			}
		}
	}

	trace = 0.0;

	for (i=0; i<nDim; i++)
		trace += covMatrices[idMode][(i*nDim)+i]/covWeights[(i*nDim)+i];

	trace *= wCov/nDim;

	for (i=0; i<nDim; i++)
	{
		for (j=0; j<nDim; j++)
		{
			if (i==j)
				covMatrices[idMode][(i*nDim)+j] = (1.0-wCov)*covMatrices[idMode][(i*nDim)+j]/covWeights[(i*nDim)+j]+trace;
			else
				covMatrices[idMode][(i*nDim)+j] *= (1.0-wCov)/covWeights[(i*nDim)+j];
		}
	}

	delete [] covWeights;
}

void ExpectationMaximization::getCovMatrix(FuzzyCMeans* myFCM, unsigned int &idMode)
{
	unsigned int i, j, k;
	double dist1, dist2, num, val, trace;

	double* covWeights = new double[sizeCov];

	for (i=0; i<sizeCov; i++) 
		covWeights[i] = 0.0; 


	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			val = myFCM->getMembership()[(i*nModes)+idMode];

			for (j=minRange[i]; j<=maxRange[i]; j++)
			{
                        dist1 = data[i][j]-centers[(idMode*nDim)+j];

				for (k=j; k<=maxRange[i]; k++)
				{
					dist2 = data[i][k]-centers[(idMode*nDim)+k];
					num = val*dist1*dist2;
					covMatrices[idMode][(j*nDim)+k] += num;
					covWeights[(j*nDim)+k] += val;

					if (k>j)
					{
						covMatrices[idMode][(k*nDim)+j] += num;
						covWeights[(k*nDim)+j] += val;
					}

				}
			}
		}
	}

	trace = 0.0;

	for (i=0; i<nDim; i++)
		trace += covMatrices[idMode][(i*nDim)+i]/covWeights[(i*nDim)+i];

	trace *= wCov/nDim;

	for (i=0; i<nDim; i++)
	{
		for (j=0; j<nDim; j++)
		{
			if (i==j)
				covMatrices[idMode][(i*nDim)+j] =  (1.0-wCov)*covMatrices[idMode][(i*nDim)+j]/covWeights[(i*nDim)+j]+trace;
			else
				covMatrices[idMode][(i*nDim)+j] *= (1.0-wCov)/covWeights[(i*nDim)+j];
		}
	}

	delete [] covWeights; 
}

void ExpectationMaximization::getInvMatrix(std::vector<std::vector<double> > &mat, unsigned int &idMode, double &det)
{
	unsigned int i, j, k, arg;
	double factor, abs;
	bool disp;

	double* invMatrix = new double [sizeCov];
	double* iniMatrix = new double [sizeCov];
	double* matrix1 = new double [nDim];
	double* matrix2 = new double [nDim];
	det = 1.0;
	disp = false;

	// Set inverse mat to identity mat
	for (i=0; i<nDim; i++)
	{
		for (j=0; j<nDim; j++)
		{
			iniMatrix[(i*nDim)+j] = mat[idMode][(i*nDim)+j];

			if(i==j)
				invMatrix[(i*nDim)+j] = 1.0;
			else
				invMatrix[(i*nDim)+j] = 0.0;
		}
	}

	// Compute inverse mat
	for(i=0; i<nDim; i++)
	{
		abs = 0.0;
		arg = 0;

		for(j=i; j<nDim; j++)
		{
			if (fabs(iniMatrix[(j*nDim)+i])>abs)
			{
				abs = fabs(iniMatrix[(j*nDim)+i]);
				arg = j;
			}
		}

		if ((abs==0)&&(!disp))
		{
			CERR << "[" << nIte << "] WARNING: Singular matrix for mode #" << idMode << ENDL;
			disp = true;
		}

		if (i!=arg)
		{
			for(j=0; j<nDim; j++)
			{
				matrix1[j] = iniMatrix[(i*nDim)+j];
				matrix2[j] = invMatrix[(i*nDim)+j];

				iniMatrix[(i*nDim)+j] = iniMatrix[(arg*nDim)+j];
				invMatrix[(i*nDim)+j] = invMatrix[(arg*nDim)+j];
		
				iniMatrix[(arg*nDim)+j] = matrix1[j];
				invMatrix[(arg*nDim)+j] = matrix2[j];
			}
		}

		factor = 1.0/iniMatrix[(i*nDim)+i];
		det *= iniMatrix[(i*nDim)+i];

		for(j=0; j<nDim; j++)
		{
			iniMatrix[(i*nDim)+j] *= factor;
			invMatrix[(i*nDim)+j] *= factor;
		}
	
		for(j=0; j<nDim; j++)
		{
			if (j==i)
				continue;
	
			if (fabs(iniMatrix[(j*nDim)+i])>0)
			{
				factor = iniMatrix[(j*nDim)+i];
	
				for(k=0; k<nDim; k++)
				{
					iniMatrix[(j*nDim)+k] -= (iniMatrix[(i*nDim)+k]*factor);
					invMatrix[(j*nDim)+k] -= (invMatrix[(i*nDim)+k]*factor);
				}
			}
		}
	}

	for (i=0; i<sizeCov; i++)
		invMatrices[idMode][i] = invMatrix[i];

	delete [] matrix1; delete [] matrix2; delete [] iniMatrix; delete [] invMatrix;
}

double ExpectationMaximization::getVaic()
{
	double aic, dof, dim, clu;

	dim = (double)nDim;
	clu = (double)nModes;
	dof = clu-1.0+(clu*dim)+clu*((dim*(dim+1.0))/2.0);
	aic = (2.0*dof)-(2.0*pdf);

	return aic;
}

double ExpectationMaximization::getVaicc()
{
	double aicc, dof, dim, clu, num;

	dim = (double)nDim;
	clu = (double)nModes;
	num = (double)nData;
	dof = clu-1.0+(clu*dim)+clu*((dim*(dim+1.0))/2.0);
	aicc= (2.0*dof*(num/(num-dof-1.0)))-(2.0*pdf);

	return aicc;
}

double ExpectationMaximization::getVbic()
{
	double bic, dof, dim, clu;

	dim = (double)nDim;
	clu = (double)nModes;
	dof = clu-1.0+(clu*dim)+clu*((dim*(dim+1.0))/2.0);
	bic = (dof*log((double)nData))-(2.0*pdf);

	return bic;
}

double ExpectationMaximization::getVo()
{
	unsigned int i;
	double ove, max1, max2;

	// Compute the VOS
	ove = 0.0;

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			max1 = getMaxData()[i];
			max2 = getMaxData2()[i];
			ove += 2.0*(max2/(max1+max2));
		}
	}	

	ove /= (double)nDataUsed;

	return ove;
}

double ExpectationMaximization::getVs()
{
	unsigned int i, j, k, n;
	double up, tmp, sep;

	sep = 0.0;

	for (i=0; i<nModes-1; i++)
	{
		for (j=i+1; j<nModes; j++)
		{
			tmp = 0.0;
			up = 0.0;
			n = 0;

			for (k=0; k<nData; k++)
			{
				if(isReliable[k])
				{
					if(i==getIndData()[k])
					{
						up += proba[(j*nData)+k]/(proba[(i*nData)+k]+proba[(j*nData)+k]);
						n++;
					}
					else if(j==getIndData()[k])
					{
						up += proba[(i*nData)+k]/(proba[(i*nData)+k]+proba[(j*nData)+k]);
						n++;
					}
				}
			}

			tmp = up/(double)n;

			if (tmp>sep)
				sep = tmp;
		}
	}

	return sep;
}

double ExpectationMaximization::getVpi()
{
	unsigned int i, j, k, sum;
	double vpi, tmp1, tmp2, tmp3;

	double* perf = new double [nModes*nData];
	double* average = new double [nDim];

	// Compute the average of all data
	for (i=0; i<nDim; i++)
	{
		tmp1 = 0.0;
		sum = 0;

		for (j=0; j<nData; j++)
		{
			if((isReliable[j])&&(i>=minRange[j])&&(i<=maxRange[j]))
			{
				tmp1 += data[j][i];
				sum++;
			}
		}

		average[i] = tmp1/(double)sum;
	}

	// Get performance index
	for (i=0; i<nModes; i++)
	{
		for (j=0; j<nData; j++)
		{
			perf[(i*nData)+j] = 0.0;

			if(isReliable[j])
			{
				for (k=minRange[j]; k<=maxRange[j]; k++)
				{
					tmp2 = data[j][k]-centers[(i*nDim)+k];
					tmp3 = centers[(i*nDim)+k]-average[k];
					perf[(i*nData)+j] += (weightRange[j]*((tmp2*tmp2)-(tmp3*tmp3)));
				}
			}
		}
	}

	vpi = 0.0;

	for (i=0; i<nModes; i++)
		for (j=0; j<nData; j++)
			if(isReliable[j])
				vpi += proba[(i*nData)+j]*perf[(i*nData)+j];

	delete [] average; delete [] perf;

	return vpi;
}

double ExpectationMaximization::getVfvq()
{
	unsigned int i, j, k;
	double tmp1, tmp2, qe, mind, vfvq;

	// Reconstructed data (x')
	double* data2 = new double [nData*nDim];
	double* sumMF = new double [nDim];

	for (i=0; i<(nData*nDim); i++)
		data2[i] = 0.0;

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			for (j=0; j<nDim; j++)
				sumMF[j] = 0.0;

			for (j=0; j<nModes; j++)
			{
				for (k=minRange[i]; k<=maxRange[i]; k++)
				{
					data2[(i*nDim)+k] += proba[(j*nData)+i]*centers[(j*nDim)+k];
					sumMF[k] += proba[(j*nData)+i];
				}
			}
	
			for (k=0; k<nDim; k++)
				data2[(i*nDim)+k] /= sumMF[k];
		}
	}

	// Quantization error (qe)
	qe = 0.0;

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			tmp2 = 0.0;
	
			for (j=minRange[i]; j<=maxRange[i]; j++)
			{
				tmp1 = data[i][j]-data2[(i*nDim)+j];
				tmp2 += weightRange[i]*tmp1*tmp1; 
			}
			
			qe += tmp2;
		}
	}

	// Minimum distance between modes
	mind = MAX_DOUBLE;

	for (i=0; i<(nModes-1); i++)
	{
		for (j=i+1; j<nModes; j++)
		{
			tmp2 = 0.0;

			for (k=0; k<nDim; k++)
			{
				tmp1 = centers[(i*nDim)+k]-centers[(j*nDim)+k];
				tmp2 += tmp1*tmp1;
			}

			if (sqrt(tmp2)<mind)
				mind = sqrt(tmp2);
		}
	} 
	
	// Compute Vfvq
	vfvq = qe/(nDataUsed*mind);

	delete [] sumMF; delete [] data2;

	return vfvq;
}

double ExpectationMaximization::getVpc()
{
	unsigned int i, j;
	double vpc;

	vpc = 0.0;

	for (i=0; i<nModes; i++)
		for (j=0; j<nData; j++)
			if(isReliable[j])
				vpc += (proba[(i*nData)+j]*proba[(i*nData)+j]);

	vpc = vpc/nDataUsed;

	return vpc;
}

double ExpectationMaximization::getVpe()
{
	unsigned int i, j;
	double vpe;

	vpe = 0.0;

	for (i=0; i<nModes; i++)
		for (j=0; j<nData; j++)
			if(isReliable[j])
				vpe += proba[(i*nData)+j]*log(proba[(i*nData)+j]);

	vpe = -1.0*(vpe/nDataUsed);

	return vpe;
}

unsigned int ExpectationMaximization::getNbrModes()
{
	return nModes;
}

unsigned int ExpectationMaximization::getNbrData()
{
	return nData;
}

unsigned int ExpectationMaximization::getNbrDim()
{
	return nDim;
}

unsigned int ExpectationMaximization::getMaxIte()
{
	return maxIte;
}

unsigned int ExpectationMaximization::getNbrIteUsed()
{
	return nIte;
}

double** ExpectationMaximization::getData()
{
	return data;
}

double* ExpectationMaximization::getProba()
{
	return proba;
}

double* ExpectationMaximization::getCenters()
{
	return centers;
}

double ExpectationMaximization::getLogLike()
{
	return logLike;
}

double ExpectationMaximization::getMinImpro()
{
	return minImpro;
}

double* ExpectationMaximization::getModeWeights()
{
	return weightMode;
}

bool* ExpectationMaximization::getReliable()
{
	return isReliable;
}

std::vector<std::vector<double> > ExpectationMaximization::getCovMatrices()
{
	return covMatrices;
}

double* ExpectationMaximization::getMaxMode()
{
	if(isMaxMode)
		return maxMode;
	else
	{
		unsigned int i, j;
		double temp;
	
		// Compute maximum likelihood by mode
		for (i=0; i<nModes; i++)
		{
			temp = 0.0;
	
			for (j=0; j<nData; j++)
				if ((isReliable[j])&&(proba[(i*nData)+j]>temp))
					temp = proba[(i*nData)+j];
			
			maxMode[i] = temp;
		}

		isMaxMode = true;	
		return maxMode;
	}
}

double* ExpectationMaximization::getMaxData()
{
	if(isMaxData)
		return maxData;
	else
	{
		unsigned int i, j;
		double temp;
	
		// Compute maximum likelihood by data
		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				temp = 0.0;
		
				for (i=0; i<nModes; i++)
					if (proba[(i*nData)+j]>temp)
						temp = proba[(i*nData)+j];
				
				maxData[j] = temp;
			}
			else
				maxData[j] = 0.0;
		}

		isMaxData = true;	
		return maxData;
	}
}

double* ExpectationMaximization::getMaxData2()
{
	if(isMaxData2)
		return maxData2;
	else
	{
		unsigned int i, j;
		double temp;
	
		// Compute second maximum likelihood by data
		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				temp = 0.0;
		
				for (i=0; i<nModes; i++)
					if ((proba[(i*nData)+j]>temp)&&(proba[(i*nData)+j]!=maxData[j]))
						temp = proba[(i*nData)+j];
				
				maxData2[j] = temp;
			}
			else
				maxData2[j] = 0.0;
		}

		isMaxData2 = true;	
		return maxData2;
	}
}

unsigned int* ExpectationMaximization::getIndData()
{
	if(isIndData)
		return indData;
	else
	{
		unsigned int i, j, index;
		double temp;
	
		// Compute maximum likelihood by data
		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				temp = 0.0;
		
				for (i=0; i<nModes; i++)
				{
					if (proba[(j*nModes)+i]>temp)
					{
						temp = proba[(j*nModes)+i];
						index= i;
					}
				}
		
				indData[j] = index;
			}
			else
				indData[j] = 0;
		}

		isIndData = true;	
		return indData;
	}
}
}
