#include <classification/OrientedPoints2FCM.h>

namespace yav{

OrientedPoints2FCM::OrientedPoints2FCM(double** data, unsigned int nClust, unsigned int nData, unsigned int nDim, double wPos, double wOri, double wInt, Metric myMetric, bool once, double expo, unsigned int* dataRange, unsigned int minDim, unsigned int maxIte, double minImpro, double* fuzzyPartition, double* initCluster)
: FuzzyCMeans(data, nClust, nData, nDim, dataRange, minDim, once, expo, maxIte, minImpro, fuzzyPartition, initCluster)
{
	unsigned int i, j, n;
	double delta, stdVarPos, stdVarOri, stdVarInt;

	this->myMetric = myMetric;
	stdVarPos = stdVarOri = stdVarInt = 0.0;
	n = 0;

	mean = new double[nDim];
	var = new double[nDim];
	nRel = new unsigned int[nDim-6];

	for (i=0; i<nDim; i++)
		mean[i] = var[i] = 0.0;

	for (i=0; i<nDim-6; i++)
		nRel[i] = 0;

	// Compute the position, orientation and intensity variance through the different dimensions
	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			n++;

			for (j=0; j<6; j++)
			{
				delta = data[i][j]-mean[j];
				mean[j] = mean[j]+(delta/(double)n);
				var[j] = var[j]+(delta*(data[i][j]-mean[j]));
			}
	
			for (j=6+minRange[i]; j<=6+maxRange[i]; j++)
			{
				nRel[j-6]++;
				delta = data[i][j]-mean[j];
				mean[j] = mean[j]+(delta/(double)nRel[j-6]);
				var[j] = var[j]+(delta*(data[i][j]-mean[j]));
			}
		}
	}

	// Compute the average position and orientation variance
	for (i=0; i<3; i++)
	{
		stdVarPos += var[i];
		stdVarOri += var[i+3];
	}

	for (i=6; i<nDim; i++)
		stdVarInt += var[i]/(double)(nRel[i-6]);

	stdVarPos = stdVarPos/(3.0*((double)n));
	stdVarOri = stdVarOri/(3.0*((double)n));
	stdVarInt = stdVarInt/((double)(nDim-6));
	wPosNorm = wPos/stdVarPos;
	wOriNorm = wOri/stdVarOri;
	wIntNorm = wInt/stdVarInt;
}

OrientedPoints2FCM::~OrientedPoints2FCM()
{
	if (nRel) delete [] nRel;
	if (var) delete [] var;
	if (mean) delete [] mean;
}

void OrientedPoints2FCM::center()
{
	unsigned int i, j, k, offsetData, offsetCoord;
	double norm, totSumMF;
	Vec3<double> barycentre, normOri, rightPart, pointCluster, normalCluster, pointSample, normalSample, matByVect1, centerPos, clusMinSamp, eigenValues, term5, term6;
	Matrix3x3<double> identity, leftPart, crossNormClust, crossNormSamp, invLeftPart, crossPos, oriMatrix, eigenVectors, term3;

	for (i=0; i<(nClust*nDim); i++)
		centers[i] = 0.0;

	// Compute the clusters center
	for (i=0; i<nClust; i++)
	{
		offsetCoord = i*nDim;
		totSumMF = 0.0;

		for (j=0; j<nDim-6; j++)
			sumMF[j] = 0.0;

		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				offsetData = (i*nData)+j;
				mf[offsetData] = pow(memb[offsetData],expo);
	
				switch (myMetric)
				{
				case EUCLIDIAN_POSITION:
	
					for (k=0; k<3; k++)
						centers[offsetCoord+k] += (mf[offsetData]*data[j][k]);
					break;
	
				case EUCLIDIAN_POSITION_ORIENTATION:
	
					for (k=0; k<6; k++)
						centers[offsetCoord+k] += (mf[offsetData]*data[j][k]);
					break;
				default:
					CERR << "WARNING: No metric chosen" << ENDL;
					break;
				}
	
				for (k=6+minRange[j]; k<=6+maxRange[j]; k++)
				{
					centers[offsetCoord+k] += (mf[offsetData]*data[j][k]);
					// In the meantime, compute the sum of mf rows in sumMF
					sumMF[k-6] += mf[offsetData];
				}

				totSumMF += mf[offsetData];
			}
		}

		// Computes orientation vector norm
		normOri[0] = centers[offsetCoord+3];
		normOri[1] = centers[offsetCoord+4];
		normOri[2] = centers[offsetCoord+5];

		norm = normOri.norm();

		for (k=3; k<6; k++)
			centers[offsetCoord+k] /= norm;

		for (k=0; k<6; k++)
			centers[offsetCoord+k] /= totSumMF;

		for (k=6; k<nDim; k++)
			centers[offsetCoord+k] /= sumMF[k-6];
	}
}

void OrientedPoints2FCM::dist()
{
	unsigned int i, j, k, offsetCoord, offsetData;
	double tempValue;
	Vec3<double> pointCluster, normalCluster, pointSample, normalSample, sampMinClu, cluMinSamp;

	// Fill the objArrayput matrix
	// objArray[i][j] is the L2 norm distance between each sample j and each center of cluster i
	for (i=0; i<nClust; i++)
	{
		offsetCoord = i*nDim;

		for (j=0; j<nData; j++)
		{
			offsetData = (i*nData)+j;
			objArray[offsetData] = 0.0;

			if(isReliable[j])
			{
				switch (myMetric)
				{
				case EUCLIDIAN_POSITION:
	
					for (k=0; k<3; k++)
					{
						tempValue = data[j][k]-centers[offsetCoord+k];
						objArray[offsetData] += (wPosNorm*tempValue*tempValue);
					}
					break;
	
				case EUCLIDIAN_POSITION_ORIENTATION:
	
					for (k=0; k<3; k++)
					{
						tempValue = data[j][k]-centers[offsetCoord+k];
						objArray[offsetData] += (wPosNorm*tempValue*tempValue);
					}
					for (k=3; k<6; k++)
					{
						tempValue = data[j][k]-centers[offsetCoord+k];
						objArray[offsetData] += (wOriNorm*tempValue*tempValue);
					}
					break;
	
				default:

					CERR << "WARNING: No metric chosen" << ENDL;
					break;
				}
		
				for (k=6+minRange[j]; k<=6+maxRange[j]; k++)
				{
					tempValue = data[j][k]-centers[offsetCoord+k];
					objArray[offsetData] += (weightRange[j]*wIntNorm*tempValue*tempValue);
				}
			}
		}
	}
}
}
