#ifndef probaMAXIMIZATION_H
#define probaMAXIMIZATION_H

#include <vector>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <classification/FuzzyCMeans.h>

namespace yav{

class ExpectationMaximization
{
 protected:
	/// Number of iterations used
	unsigned int nIte;
	/// Maximum number of iterations allowed
	unsigned int maxIte;
	/// Number of modes
	unsigned int nModes;
	/// Number of data
	unsigned int nData;
	/// Data dimension
	unsigned int nDim;
	/// Data minimum dimension (to be defined as reliable)
	unsigned int minDim;
	/// If data range specified, number of data defined as reliable
	unsigned int nDataUsed;
	/// Size of data in memory
	unsigned int sizeDat;
	/// Size of probability in memory
	unsigned int sizeExp;
	/// Size of centers in memory
	unsigned int sizeCen;
	/// Size of covariance in memory
	unsigned int sizeCov;
	/// Log likelihood of estimated modes
	double logLike;
	/// Minimum of improvement between 2 iterations
	double minImpro;
	/// Probability density function (PDF)
	double pdf;
	/// Weights to regularize the covariance matrix
	double wCov;
	/// Specify whether indData has been assigned
	bool isIndData;
	/// Specify whether maxMode has been assigned
	bool isMaxMode;
	/// Specify whether maxData has been assigned
	bool isMaxData;
	/// Specify whether maxData2 has been assigned
	bool isMaxData2;
	/// Estimated covariance matrices
	std::vector<std::vector<double> > covMatrices;
	/// Inverse matrices
	std::vector<std::vector<double> > invMatrices;

	/// For each data, range of data to be taken into account [min1 max1 min2 max2...]
	unsigned int* dataRange;
	/// For each data, index of the mode with the highest probability
	unsigned int* indData;
	/// For each data, index of the first dimension to be taken into account
	unsigned int* minRange;
	/// For each data, index of last dimension to be taken into account
	unsigned int* maxRange;
	/// For each data, probability to belong to modes
	double* proba;
	/// For each mode, center coordinates
	double* centers;
	/// Estimated mode weights
	double* weightMode;
	/// For each data, weighting factor for the center computation
	double* weightRange;
	/// For each mode, data maximum probability
	double* maxMode;
	/// For each data, maximum probability
	double* maxData;
	/// For each data, second maximum probability
	double* maxData2;
	/// For each data, boolean specifying whether the data is reliable
	bool* isReliable;

	/// Data
	double** data;

	/** Initializes the EM algorithm.*/
	void initEM();

	/** Computes the Expectation step.*/
	void Estep();

	/** Computes the Maximization step.*/
	void Mstep();

	/** Computes a matrix substraction.*/
	void getSubMat(double* mat1, double* mat2, double* result);

	/** Computes the multiplication of a matrix MxN by a matrix NxP.*/
	void getMultMatByMat(std::vector<double> &mat1, double* mat2, double* result, unsigned int &nRows, unsigned int &n, unsigned int &id);

	/** Computes the multiplication of a matrix 1xN by a matrix Nx1.*/
	double getMultMatByMat(double* mat1, double* mat2, unsigned int &id);

	/** Computes the covariance matrix.*/
	void getCovMatrix(unsigned int &idMode);
	void getCovMatrix(FuzzyCMeans* myFCM, unsigned int &idMode);

	/** Computes the inverse matrix and the determinant.*/
	void getInvMatrix(std::vector<std::vector<double> > &mat, unsigned int &idMode, double &determinant);
 public:
	/** @name Constructors */
	/** Initializes the Expectation-Maximization algorithm.
      @param data data (data[number_of_data][number_of_coordinates])
      @param nModes number of modes
      @param nData number of data
      @param nDim data dimension
	@param wCov weight to regularize the covariance matrix
	@param dataRange for each data, range to be taken into account [min1 max1 min2 max2...]
      @param minDim data minimum dimension (to be defined as reliable, default=1)
      @param maxIte maximum number of iterations (default=1000)
      @param minImpro minimum improvement between 2 iterations (default=0.00001)*/
	ExpectationMaximization(double** data, unsigned int nModes, unsigned int nData, unsigned int nDim, double wCov=0.0, unsigned int* dataRange=NULL, unsigned int minDim=1, unsigned int maxIte=1000, double minImpro=0.00001);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	virtual ~ExpectationMaximization();

	/** @name Other functions */
	/** Launches the algorithm. The process stops when the maximum number of iterations maxIte is reached, or when the minimum amount of improvement between two consecutive iterations is less than MIN_IMPRO.*/
	void em();

	/** Returns the Akaike's information criterion (AIC).*/
	double getVaic();

	/** Returns the second-order Akaike's information criterion (AICc).*/
	double getVaicc();

	/** Returns the Bayesian information criterion (BIC).*/
	double getVbic();

	/** Returns the overlap.*/
	double getVo();

	/** Returns the separation.*/
	double getVs();

	/** Returns the performance index (Vpi).*/
	double getVpi();

	/** Returns the fuzzy vector quantization (Vfvq).*/
	double getVfvq();

	/** Returns the partition coefficient (Vpc).*/
	double getVpc();

	/** Returns the partition entropy (Vpe).*/	
	double getVpe();

	/** Returns the number of modes.*/
	unsigned int getNbrModes();

	/** Returns the number of data.*/
	unsigned int getNbrData();

	/** Returns the data dimension.*/
	unsigned int getNbrDim();

	/** Returns the maximum number of iterations.*/
	unsigned int getMaxIte();

	/** Returns the number of iterations used.*/
	unsigned int getNbrIteUsed();

	/** Returns the data.*/
	double** getData();

	/** Returns the probability for each data to belong to a mode. The values 0 and 1 indicate no probability and full probability respectively. Values between 0 and 1 indicate that the data has partial probability in a mode.*/
	double* getProba();

	/** Returns the mode centers.*/
	double* getCenters();

	/** Returns the log likelihood of estimated modes.*/
	double getLogLike();

	/** Returns the minimum amount of improvement between 2 iterations.*/
	double getMinImpro();

	/** Returns the estimated mode weights.*/
	double* getModeWeights();

	/** Returns the reliable data.*/
	bool* getReliable();

	/** Returns the estimated covariance matrices.*/
	std::vector<std::vector<double> > getCovMatrices();

	/** Returns the data maximum probability for each mode.*/
	double* getMaxMode();

	/** Returns the data maximum probability for each data.*/
	double* getMaxData();

	/** Returns the data second maximum probability for each data.*/
	double* getMaxData2();

	/** Returns the index of the mode with the highest likelihood for each data.*/
	unsigned int* getIndData();
};

}

#endif
