#ifndef ORIENTEDPOINTS2FCM_H
#define ORIENTEDPOINTS2FCM_H

#include <classification/FuzzyCMeans.h>
#include <basics/TclString.h>
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Matrix3x3.h>

namespace yav{

/// Metrics used with intensities
enum Metric
{
	/// Only position
	EUCLIDIAN_POSITION,
	/// Both position and orientation
	EUCLIDIAN_POSITION_ORIENTATION,
};

/** ORIENTEDPOINTS2FCM class inherits from FUZZYCMEANS and is used to take the position and orientation of points into account. Position should take the 3 first dimensions of DATA ([0 1 2]) and orientation the 3 following dimensions ([3 4 5]). The next dimensions are voxel intensities */
class OrientedPoints2FCM : public FuzzyCMeans
{
 protected:
	/// Normalized position weight
	double wPosNorm;
	/// Normalized orientation weight
	double wOriNorm;
	/// Normalized intensity weight
	double wIntNorm;
	/// Metric used with intensities
	Metric myMetric;

	/// For each dimension, number of reliable data
	unsigned int* nRel;
	/// Mean of the data
	double* mean;
	/// Variance of the data
	double* var;
 public:
	/** @name Constructor */
	/** Initializes the Fuzzy C-Means algorithm.
      @param data data (data[number_of_data][number_of_coordinates])
      @param nClust number of clusters
      @param nData number of data
      @param nDim data dimension
      @param wPos position weight [0 1] used as a trade-off between position, orientation and intensity
	@param wOri orientation weight [0 1] used as a trade-off between position, orientation and intensity
	@param wInt intensity weight [0 1] used as a trade-off between position, orientation and intensity
	@param myMetric metric used with intensities
      @param once boolean taken into account for the pseudo-random numbers generation used by U. Should be always true (default=true). This boolean is supposed to be false only when used by static FINDBESTNUMBERCLUSTERS function.
      @param expo fuzziness exponent (default=2.0)
	@param dataRange for each data, range to be taken into account [min1 max1 min2 max2...]
      @param minDim data minimum dimension to be considered as reliable (default=1)
      @param maxIte maximum number of iteration allowed (default=100)
      @param minImpro minimum amount of improvement (default=0.00001)
      @param fuzzyPartition initial membership (default=NULL)
      @param initCluster initial cluster centers (default=NULL) */
	OrientedPoints2FCM(double** data, unsigned int nClust, unsigned int nData, unsigned int nDim, double wPos=1.0, double wOri=1.0, double wInt=1.0, Metric myMetric=EUCLIDIAN_POSITION, bool once=true, double expo=2.0, unsigned int* dataRange=NULL, unsigned int minDim=1, unsigned int maxIte=100, double minImpro=0.00001, double* fuzzyPartition=NULL, double* initCluster=NULL);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	~OrientedPoints2FCM();

	/** @name Other functions */
	/** Overwritten function CENTERFCM from the FUZZYCMEANS class */
	void center();

	/** Overwritten function DISTFCM from the FUZZYCMEANS class */
	void dist();
};

}
#endif
