#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4290)
#endif

#pragma implementation "framework/Transformation.h"
#include <framework/Transformation.h>
#include <framework/PCATransformation.h>


namespace yav {


  // default constructor:
  Transformation::Transformation() {}

  // default destructor:
  Transformation::~Transformation() {}

  // returns the type of the transformation:
  TclString Transformation::getType() const {
    return _type;
  }

  // -------------------------------------------------------------------------


  // default constructor:
  MatrixTransformation::MatrixTransformation() 
    : _current(NULL), _deformed(NULL) {
    inputType = NONE;
  }

  // constructor from pairs of points:
  MatrixTransformation::
  MatrixTransformation(PointSetIterator* current,
			PointSetIterator* deformed) 
    : _current(current), _deformed(deformed) {

    inputType = POINTPAIRSET;
  }

  // constructor from given 4x4 matrix
  MatrixTransformation::MatrixTransformation(const Matrix4x4<double>& mat)
    : _current(NULL),
      _deformed(NULL),
      _mat(mat) {

    inputType = MATRIX;
  }


  // default destructor:
  MatrixTransformation::~MatrixTransformation() {}


  // returns the current point set iterator:
  PointSetIterator* MatrixTransformation::getCurrentShape() const {
    return _current;
  }

  // returns the deformed point set iterator:
  PointSetIterator* MatrixTransformation::getDeformedShape() const {
    return _deformed;
  }

  // sets the current point set iterator:
  void MatrixTransformation::setCurrentShape(PointSetIterator* current) {
    _current = current;
    inputType = POINTPAIRSET;
  }

  // sets the deformed point set iterator:
  void MatrixTransformation::setDeformedShape(PointSetIterator* deformed) {
    _deformed = deformed;
    inputType = POINTPAIRSET;
  }


  // returns the transformation matrix:
  Matrix4x4<double> MatrixTransformation::getMatrix() const {
    return _mat;
  }

  // set the transformation matrix:
  void MatrixTransformation::setMatrix(const Matrix4x4<double>& mat) {
    _mat = mat;
    inputType = MATRIX;
  }

  // apply the transformation:
  void MatrixTransformation::apply(PointSetIterator* set) {

    if (inputType != NONE) {

      for(set->first(); !set->isAtEnd(); (*set)++)
	set->set(_mat * set->get());

    } else {
      // **************** ERROR
    }
  }


  // -------------------------------------------------------------------------


  // default constructor:
  RigidTransformation::RigidTransformation() { _type = "rigid"; }
  RigidTransformation::RigidTransformation(PointSetIterator* current,
					     PointSetIterator* deformed)
    : MatrixTransformation(current, deformed) { _type = "rigid"; }
  // default destructor:
  RigidTransformation::~RigidTransformation() {}

  // computes the rigid transformation:
  void RigidTransformation::compute() {

    unsigned int N;
    int i, imin;
    double dmin;
    Vect<4,double> d;
    Vec3<double> Gm, Gs;
    Quaternion x, y;
    
    /* compute barycenters Gm and Gs */
    N = _current->barycenter(Gm);
    _deformed->barycenter(Gs);
    
    if(N > 5) {
      /* Compute the rotation quaternion from model to scene
       * in barycentric coordinates (xi for model and yi for scene) */
      Matrix4x4<double> A, y_tilde, x_anti, xy;
      x[0] = y[0] = 0.0;
      for(_current->first(), _deformed->first();
	  !_current->isAtEnd();
	  (*_current)++, (*_deformed)++) {

	if (_current->getWeight()>0.0) {
	  x[1] = _current->getX() - Gm.getX();
	  x[2] = _current->getY() - Gm.getY();
	  x[3] = _current->getZ() - Gm.getZ();
	  y[1] = _deformed->getX() - Gs.getX();
	  y[2] = _deformed->getY() - Gs.getY();
	  y[3] = _deformed->getZ() - Gs.getZ();
	  
	  /* xy = tilde4(y) + anti4(x) */
	  y.toTilde(y_tilde);
	  x.toAnti(x_anti);
	  xy = x_anti;
	  xy += y_tilde;
	  
	  /* add (xy)^t .(xy) to A */
	  x_anti = xy;
	  x_anti.transpose();
	  y_tilde = x_anti * xy;
	  A += y_tilde;
	}
      }
      
      /* Compute the eigenvector associated to the minimal eigenvalue */
      Matrix4x4<double> V;
      A.computeEigenV(V, d);
      for(i=1, imin=0, dmin=d[0]; i<4; i++)
	if( d[i] < dmin) {
	  dmin = d[i];
	  imin = i;
	}
      for(i=0; i<4; i++) x[i] = V.getValue(i, imin);
      
      dmin = x.norm();
      if (dmin != 1.0)
	for(i=0; i<4; i++) x[i] /= dmin;
      x.toTilde(y_tilde);
      x.toAnti(x_anti);
      A = y_tilde;
      A *= x_anti;
      
      Matrix3x3<double> R(A.getValue(1, 1), A.getValue(1, 2), A.getValue(1, 3),
			  A.getValue(2, 1), A.getValue(2, 2), A.getValue(2, 3),
			  A.getValue(3, 1), A.getValue(3, 2), A.getValue(3, 3));
      
      /* Translation : T = Gs - R * Gm */
      Vec3<double> T(Gs - (R * Gm));
      
      /* create homogeneous matrix */
      //from transformation.c:
      //  trsf[0][3] = T.x; trsf[1][3] = T.y; trsf[2][3] = T.z;
      //  trsf[3][0] = trsf[3][1] = trsf[3][2] = 0.0; trsf[3][3] = 1.0;
      _mat = R;
      _mat = T;
      _mat.setValue(3, 0, 0.0);
      _mat.setValue(3, 1, 0.0);
      _mat.setValue(3, 2, 0.0);
      _mat.setValue(3, 3, 1.0);
    }
    else {
      _mat.resetToIdentity();
    }

  }


  // -------------------------------------------------------------------------


  // default constructor:
  SimilarityTransformation::SimilarityTransformation() {
    _type = "similarity";
  }
  SimilarityTransformation::
  SimilarityTransformation(PointSetIterator* current, 
			    PointSetIterator* deformed)
    : MatrixTransformation(current, deformed) {
    _type = "similarity";
  }
  // default destructor:
  SimilarityTransformation::~SimilarityTransformation() {}

  // computes the similarity transformation:
  void SimilarityTransformation::compute() {
    unsigned int N;
    int i, imin;
    double dmin, s1, s2;
    Vect<4,double> d;
    Vec3<double> Gm, Gs;
    Quaternion x, y;

    /* compute barycenters Gm and Gs */
    N = _current->barycenter(Gm);
    _deformed->barycenter(Gs);

    if(N > 6) {
      /* Compute the rotation quaternion from model to scene
       * in barycentric coordinates (xi for model and yi for scene) */
      Matrix4x4<double> A, y_tilde, x_anti, xy;
      x[0] = y[0] = 0.0;
      s1 = s2 = 0.0;
      for(_current->first(), _deformed->first();
	  !_current->isAtEnd();
	  (*_current)++, (*_deformed)++) {

	if (_current->getWeight()>0.0) {
	  x[1] = _current->getX() - Gm.getX();
	  x[2] = _current->getY() - Gm.getY();
	  x[3] = _current->getZ() - Gm.getZ();
	  y[1] = _deformed->getX() - Gs.getX();
	  y[2] = _deformed->getY() - Gs.getY();
	  y[3] = _deformed->getZ() - Gs.getZ();
	  
	  s1 += x[1]*x[1] + x[2]*x[2] + x[3]*x[3];
	  s2 += y[1]*y[1] + y[2]*y[2] + y[3]*y[3];
	  
	  /* xy = tilde4(y) + anti4(x) */
	  y.toTilde(y_tilde);
	  x.toAnti(x_anti);
	  xy = x_anti;
	  xy += y_tilde;
	  
	  /* add (xy)^t .(xy) to A */
	  x_anti = xy;
	  x_anti.transpose();
	  y_tilde = x_anti * xy;
	  A += y_tilde;
	}
      }

      /* Compute the eigenvector associated to the minimal eigenvalue */
      Matrix4x4<double> V;
      A.computeEigenV(V, d);
      for(i=1, imin=0, dmin=d[0]; i<4; i++)
	if( d[i] < dmin) {
	  dmin = d[i];
	  imin = i;
	}
      for(i=0; i<4; i++) x[i] = V.getValue(i, imin);
      
      dmin = x.norm();
      if (dmin != 1.0)
	for(i=0; i<4; i++) x[i] /= dmin;
      x.toTilde(y_tilde);
      x.toAnti(x_anti);
      A = y_tilde;
      A *= x_anti;
      
      Matrix3x3<double> R(A.getValue(1, 1), A.getValue(1, 2), A.getValue(1, 3),
			  A.getValue(2, 1), A.getValue(2, 2), A.getValue(2, 3),
			  A.getValue(3, 1), A.getValue(3, 2), A.getValue(3, 3));
      
      /* compute scale parameter */
      double tmp = (s1 - s2) / (s1 + s2 - d[imin]);
      dmin = ::sqrt(1.0 + tmp*tmp) - tmp;
      
      Matrix3x3<double> S;
      for(i = 0; i < 3; i++) S.setValue(i, i, dmin);
      S *= R;
      
      /* Translation : T = Gs - RS * Gm */
      Vec3<double> T(Gs - (S * Gm));
      
      /* create homogeneous matrix */
      _mat = S;
      _mat = T;
      _mat.setValue(3, 0, 0.0);
      _mat.setValue(3, 1, 0.0);
      _mat.setValue(3, 2, 0.0);
      _mat.setValue(3, 3, 1.0);
    } else

      _mat.resetToIdentity();
    
  }


  // -------------------------------------------------------------------------


  // default constructor:
  AffineTransformation::AffineTransformation() { _type = "affine"; }
  AffineTransformation::AffineTransformation(PointSetIterator* current,
					       PointSetIterator* deformed) 
    : MatrixTransformation(current, deformed) {
    _type = "affine";
  }
  // default destructor:
  AffineTransformation::~AffineTransformation() {}

  // computes the affine transformation:
  void AffineTransformation::compute() {

    unsigned int i = 0, j, k, N;
    Vec3<double>Gx, Gy;
    Vect<6,double> z, Gz, EVal; // z = [x,y]
    
    /* compute barycenters Gm and Gs */
    N = _current->barycenter(Gx);
    _deformed->barycenter(Gy);

    for(k=0;k<3;++k){
      Gz(k) = Gx(k);
      Gz(k+3) = Gy(k);
    }
    
    if(N > 6) {
      Matrix<6,double> SigmaMat, EVect;
      SigmaMat.Null();
      
      for(_current->first(), _deformed->first();
	  !_current->isAtEnd();
	  (*_current)++, (*_deformed)++) {

	if (_current->getWeight()>0) {
	  ++i;
	  z[0] = _current->getX();
	  z[1] = _current->getY();
	  z[2] = _current->getZ();
	  z[3] = _deformed->getX();
	  z[4] = _deformed->getY();
	  z[5] = _deformed->getZ();
	  /* Barycentric coord */
	  z -= Gz;
	  
	  for(j=0;j<6;++j)
	    for(k=0;k<6;++k)
	      SigmaMat(j,k) += _current->getWeight() * z(j) * z(k);
	}
      }
      
      Sym<6,double> Sigma( SigmaMat);
      // possible exception
      Sigma.DiagonalizeSort(EVect, EVal);
      
      Matrix3x3<double> X1, X2, A;
      
      for(j=0;j<3;++j)
	for(k=0;k<3;++k){
	  X1(j,k) = EVect(j, k+3);
	  X2(j,k) = EVect(j+3, k+3);
	}
      X1.transpose();
      X2.transpose();
      
      // possible exception
      X2.invert();
      
      A.Mul(-X2, X1);
      
      // translation
      Vec3<double> t = Gy - A * Gx;
      
      // Ok compute final mat4
      for(j=0;j<3;j++)
	for(k=0;k<3;k++)
	  _mat(j, k) = A(j,k); 
      for(j=0;j<3;j++){
	_mat(j,3) = t(j);
	_mat(3,j) = 0.0;
      }
      _mat(3,3) = 1.0;
    }
    
    else {
      _mat.resetToIdentity();
    } 
  }


} // namespace yav
