#ifndef CONSTRAINT_H
#define CONSTRAINT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "framework/Constraint.h"
#include <basics/Exception.h>
#include <basics/TclString.h>
#include <basics/ListElement.h>
#include <basics/GeneralFormat.h>
#include <map>

namespace yav {
/** \class Constraint Constraint.h basics/Constraint.h
    \ingroup libBasics
    \brief defines a constraint that is actived depending on its counter value */
class Constraint : public ListElement<Constraint> {
 private:
  /// the counter is used to control the activation of the constraint 
  unsigned int counter;
  /** the frequency value. The constraint is activated when 
      counter%frequency==0 */
  unsigned int frequency;
  /// the name of the constraint 
  TclString name;
 public:

  /** @name destructor */
  //@{
  /// empty
  virtual ~Constraint() {}
  //@}

  /** @name constructor */
  //@{
  /**  Builds a Constraint given its frequency */
  Constraint(const unsigned int freq,const char *n): frequency(freq), name(n) {
  counter=0;}
  //@}
  /** defines the prototype of function that creates a RangeInformation object from a file pointer
      @param input The input stream */
  typedef  Constraint * (*CONSTRAINTREADFUNCTION)(GFInput &input); 
  /** a hash table that associates a read function given a keyword */
  static std::map<char *,CONSTRAINTREADFUNCTION,CStringCompare> readFunctionMap; 
 /** return a Constraint given a keyword and a pointer on an input file
     @param key a keyword that indicates the type of Constraint
     @param input a pointer on an input file
     @return a range information object corresponding to the input file */
  static  Constraint * createFromFile(const char *key,GFInput &input) {
    std::map<char *,CONSTRAINTREADFUNCTION,CStringCompare>::iterator it;
    it=readFunctionMap.find((char *)key);
    if (it==readFunctionMap.end())
      return 0;
    else
      {
	return (*((*it).second))(input);
      }
  } 
  /// resets the counter to zero
  void resetCounter() {
  counter=0;}

  /// sets the frequency of a constraint
  void setFrequency(const unsigned int freq) {
  frequency=freq;}

  /// gets the frequency of a constraint
  unsigned int  getFrequency() const {
    return(frequency);}

  /// gets the frequency of a constraint
  unsigned int  getCounter() const {
    return(counter);}

  /// returns the name of the constraint
  char *getName() const {
    return((char *)name);}
  /** returns true if  counter%frequency==0  and false otherwise
      In all cases, increment the counter */
  bool isActive() {
    if (counter%frequency==0)
      {
	counter=1;
	return(true);
      }
    else
      {
	counter++;
	return(false);
      }
  } 
  /** write in a file the data stored in a constraint object
      @param output a pointer on a file */
  virtual void writeInFile(GFOutput &output){}
  /// the apply() function that actually applies the constraint
  virtual void apply(){}
  /// for force constraint compute the product of the stiffness matrix of a given displacement
  virtual void computeDfDx() {}
  /// for force constraint add the stiffness matrix of the constraint into the matrix
  virtual void addStiffnessMatrix() {}

};

/** an object (usually a deformable model) where force and position
    constraints can be applied */
class ConstraintContainer {
 public:
   /** @name constructor */
  //@{
  /** builds a new ConstraintContainer */
  ConstraintContainer(){}
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~ConstraintContainer(){}
  //@}
 protected:
 /** a list of force constraints that modifies the external force field of
   a deformable model*/
  ObjectsList<Constraint> forceConstraintList;
  /** a list of position constraints that modifies the geometry or
   the topology of a deformable model*/
  ObjectsList<Constraint> positionConstraintList;
public:
  /** scan all existing force constraint in forceConstraintList
      and call the apply function */
  virtual void applyForceConstraint();
   /** scan all existing force constraint in forceConstraintList
      and call the addStiffnessMatrix function */
  virtual void applyAddStiffnessMatrix();
  /** scan all existing position constraint in forceConstraintList
      and call the apply function */
  virtual void applyPositionConstraint();
  /** add a constraint to the list of force constraints */
  void addForceConstraint(Constraint *c) {
    forceConstraintList.append(c); }
  /** add a constraint to the list of POSITION constraints */
  void addPositionConstraint(Constraint *c) {
    positionConstraintList.append(c); }
  /// returns an iterator on the first force constraint
  ObjectsList<Constraint>::iterator firstForceConstraint() const {
    return(forceConstraintList.begin());}
  /// returns an iterator on the last force constraint
  ObjectsList<Constraint>::iterator lastForceConstraint() const {
    return(forceConstraintList.end());}
  /// returns an iterator on the first position constraint
  ObjectsList<Constraint>::iterator firstPositionConstraint() const {
    return(positionConstraintList.begin());}
  /// returns an iterator on the last position constraint
  ObjectsList<Constraint>::iterator lastPositionConstraint() const {
    return(positionConstraintList.end());}
  // get the positionConstraint given its name
  Constraint * getPositionConstraint(TclString name);
  // get the forceConstraint given its name
  Constraint * getForceConstraint(TclString name);

};

}


#endif
