#ifndef ITERATEDACTIVEOBJECT_H
#define ITERATEDACTIVEOBJECT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 )
#endif

#include <basics/ListElement.h>

namespace yav {

class IteratedActiveObjectManager;
//! Prototype for an active object synchronized by a manager. 
/*! \class ActiveGroupManager ActiveGroupManager.h basics/ActiveGroupManager.h
\ingroup libBasics
Each IteratedActiveObject has a callback function (the iterate function)
which is iteratively called by its manager. The object evolves in a synchronous
manner which the other IteratedActiveObjects stored in the manager. 
*/
class IteratedActiveObject : public ListElement<IteratedActiveObject> {
public : 
//! The constructor
/*!
Create a new IteratedActiveObject and store it in a group Manager
\sa
*/
  IteratedActiveObject(IteratedActiveObjectManager *agm=0);
//! The destructor
/*!
Removes itself from the list of ActiveObject stored in the group Manager
\sa
*/
  virtual ~IteratedActiveObject(){}
  
protected:
/** indicates if the IteratedActiveObject is actually active or not */
  bool active;
/// the total number of iterations
  unsigned int iterationNumber;
/// the total time (machine time) spent for activating the object
  unsigned LONGINT computationTime;
/// the time on which the equations were integrated (~nbIterations*timeStep)
  double integratedTime;
/// the  manager (that calls its callback function iterate()) of this object 
  IteratedActiveObjectManager *manager;
public:
/** routine that performs some  iterations on an ActiveObject
  @param n the number of iterations to performed. Can be set to zero */
  virtual void iterate(const unsigned int n)=0;
/// return the value of the active boolean value
  bool isActive() const {
    return(active);}
/// return the value of the active boolean value
  bool getActivation() const {
    return(active);}
/// set the value of the active boolean value
  void setActivation(const bool activation)  {
    active=activation;}
/// returns the time spent at deforming the deformable model
  unsigned LONGINT getComputationTime() const {
    return(computationTime);}
/// set to 0 the time spent at deforming the deformable model
  void resetComputationTime()  {
    computationTime=0;}
/// get the total number of iterations 
  unsigned int  getNbIterations() const  {
    return(iterationNumber);}
/// reset the total number of iterations 
  void resetNbIterations() {
    iterationNumber=0;}
/// get the integrated time
  double getIntegratedTime(void) const {
    return(integratedTime);}
/// set the integrated time
  void setIntegratedTime(const double t) {
    integratedTime = t;}
/// returns the manager of this object (may be null)
  IteratedActiveObjectManager *getManager() const {
    return( manager);
  }
/// returns the rank of the object in the list of ActiveObjects store in the manager
  unsigned int getRankInList() const {
    return ListElement<IteratedActiveObject>::getRank();
  }
/** change the manager of the object
  @param scene new object scene */
  void changeManager(IteratedActiveObjectManager *agm);
  
  friend class IteratedActiveObjectManager;
  
};

} // namespace yav


#endif
