#ifndef PCATRANSFORMATION_H
#define PCATRANSFORMATION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "framework/PCATransformation.h"


#include <linearAlgebra/Vec4.h>
#include <linearAlgebra/Matrix4x4.h>
#include <framework/PointSet.h>
#include <framework/Transformation.h>
#include <basics/GeneralFormat.h>


namespace yav {


  class PCATransformation;

  
  /*! \class PCAData PCATransformation.h basics/PCATransformation.h
    \ingroup libBasics
    PCAData class:
  */
  class PCAData {

    friend class PCATransformation;

  public:

    /// default constructor:
    PCAData();
    /// read PCA data from a file:
    PCAData(PointSetIterator* reference,
	     TclString filename,
	     double* limit = NULL);
    /// read PCA data from arrays:
    PCAData(PointSetIterator* reference,
	     unsigned int nbModes, unsigned int nbPoints,
	     double* vectors, double* values,
	     double* limit = NULL);
    /// default destructor:
    ~PCAData();


    /// read PCA data from a C++ file:
    void read_from_file(PointSetIterator* reference, TclString filename);
    /// read PCA data from a yav++ GFInput:
    void read_from_GFInput(GFInput& input);
    /// write PCA data to a yav++ GFoutput:
    void write_to_GFOutput(GFOutput& output);

    /// read PCA data from arrays:
    void read_from_arrays(PointSetIterator* reference,
			 unsigned int nbModes,unsigned int nbPoints,
			 double* vectors, double* values);

    /// returns the reference point set iterator:
    PointSetIterator* getReferenceShape() const;

    /// sets the reference point set iterator:
    void setReferenceShape(PointSetIterator* reference);

    /// returns the number of modes:
    unsigned int getNbModes() const;
    
    /// returns the number of points:
    unsigned int getNbPoints() const;

	///returns the array that indicates the limit of each mode
	const double *getModeLimit() const {
		return _limit;
	}
	/// sets the limit of each mode
	void setModeLimit( const double *limit);

	///return the array of eigenValues
	const double *getEigenValues() const {
		return _eigenValues;
	}
	///return the array of eigenVectors
	const double *getEigenVectors() const {
		return _eigenVectors;
	}

  protected:

    // reference shape:
    PointSetIterator* _reference;

    // number of modes:
    unsigned int _nbModes;
    // number of points:
    unsigned int _nbPoints;

    // eigen vectors and eigen values:
    double* _eigenVectors;
    double* _eigenValues;

    // limit on the projections:
    double* _limit;


  }; // PCAData


  /*! \class PCATransformation PCATransformation.h basics/PCATransformation.h
    \ingroup libBasics
    PCA Transformation class
  */
  class PCATransformation : public MatrixTransformation {
  public:

    // default constructor:
    PCATransformation();
    PCATransformation(PointSetIterator* current,
		       PointSetIterator* deformed,
		       PCAData* pcaData);
    // default destructor:
    virtual ~PCATransformation();


    // returns the associated PCAData:
    PCAData* getPCAData();
    // sets the current point set iterator:
    void setPCAData(PCAData* pcaData);

    /// return the number of useful Modes
	unsigned int getNbUsefulModes() {
		return nbUsefulModes;
	}
    /// sets the number of useful Modes
	void setNbUsefulModes(unsigned int _nUModes) {
		nbUsefulModes=_nUModes;
	}

    // computes the transformation:
    virtual void compute();

    // apply the transformation:
    virtual void apply(PointSetIterator* set);


  protected:

    // PCA data:
    PCAData* _pcaData;
	/// array that store the amplitude of each mode
	double *db;
	/// indicates how many modes should be considered to compute the deformation
	unsigned int nbUsefulModes;
    
    

  }; // PCATransformation



  
} // namespace yav


#endif // ifndef TRANSFORMATION_H
