#ifndef TRANSFORMATION_H
#define TRANSFORMATION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "framework/Transformation.h"

#include <linearAlgebra/Vec4.h>
#include <linearAlgebra/Matrix4x4.h>
#include <framework/PointSet.h>
#include <basics/TclString.h>


namespace yav {
  
  
  /*! \class Transformation Transformation.h basics/Transformation.h
    \ingroup libBasics
    base class Transformation:   
   */
  class Transformation {
  public:
    
    /// default constructor:
    Transformation();
    /// default virtual destructor:
    virtual ~Transformation();
    
    /// computes the transformation:
    virtual void compute() = 0;
    
    /// applies the transformation:
    virtual void apply(PointSetIterator* set) = 0;

    /// returns the type of the transformation:
    virtual TclString getType() const;

  protected:
    /// type of the transformation
    TclString _type;  
  };


  /*!  \class MatrixTransformation Transformation.h basics/Transformation.h
    \ingroup libBasics
    Matrix Transformation class:
    apply to the points in the input PointSetIteratora transformation matrix
    either given as an input argument or computed from a PointPairSetIterator.
  */
  class MatrixTransformation : public Transformation {
  public:

    /// default constructor
    MatrixTransformation();
    /// constructor from pairs of points
    MatrixTransformation(PointSetIterator* current,
			  PointSetIterator* deformed);
    /// constructor from given 4x4 matrix
    MatrixTransformation(const Matrix4x4<double>& mat);
    /// default destructor:
    virtual ~MatrixTransformation();

    /// returns the current point set iterator:
    PointSetIterator* getCurrentShape() const;
    /// returns the deformed point set iterator:
    PointSetIterator* getDeformedShape() const;
    /// sets the current point set iterator:
    void setCurrentShape(PointSetIterator* current);
    /// sets the deformed point set iterator:
    void setDeformedShape(PointSetIterator* deformed);

    /// returns the transformation matrix:
    virtual Matrix4x4<double> getMatrix() const;
    /// sets the transformation matrix:
    virtual void setMatrix(const Matrix4x4<double>& mat);

    /// applies the transformation:
    virtual void apply(PointSetIterator* set);
    

  protected:

    /// point set iterators:
    PointSetIterator* _current;
    PointSetIterator* _deformed;

    /// homogeneous transformation (rigid, similarity, affine)
    Matrix4x4<double> _mat;
    
    /// flag indicating which input data is to be used:
    enum {NONE = 0, MATRIX = 1, POINTPAIRSET = 2};
    int inputType;
  };


  /*! \class RigidTransformation Transformation.h basics/Transformation.h
    \ingroup libBasics
    Rigid Transformation class (LSE)
  */
  class RigidTransformation : public MatrixTransformation {
  public:

    /// default constructor:
    RigidTransformation();
    RigidTransformation(PointSetIterator* current,
			 PointSetIterator* deformed);
    /// default destructor:
    virtual ~RigidTransformation();

    /// computes the transformation:
    virtual void compute();

  }; // RigidTransformation


  /*! \class SimilarityTransformation Transformation.h basics/Transformation.h
    \ingroup libBasics
    Similarity Transformation class
   */
  class SimilarityTransformation : public MatrixTransformation {
  public:

    /// default constructor:
    SimilarityTransformation();
    SimilarityTransformation(PointSetIterator* current,
			      PointSetIterator* deformed);
    /// default destructor:
    virtual ~SimilarityTransformation();

    /// computes the transformation:
    virtual void compute();

  }; // SimilarityTransformation


  /*! \class AffineTransformation Transformation.h basics/Transformation.h
    \ingroup libBasics
    Affine Transformation class
  */
  class AffineTransformation : public MatrixTransformation {
  public:

    /// default constructor:
    AffineTransformation();
    AffineTransformation(PointSetIterator* current,
			  PointSetIterator* deformed);
    /// default destructor:
    virtual ~AffineTransformation();

    /// computes the transformation:
    virtual void compute();

  }; // AffineTransformation  

  
} // namespace yav


#endif // ifndef TRANSFORMATION2_H
