//**********************************************************************
//
//  TestReg  ------ Test of Rigid LSQ Registration -- 
//
//  (c) Xavier PENNEC -- INRIA -- April 1996 / December 1998
//  
//**********************************************************************

#include <linearAlgebra/Rotation3D.h>
#include <linearAlgebra/Matrix4x4.h>
#include "BasicsTest.h"

using namespace yav;


const Index ITER=100000;


#include <list>

// Rigid to mat4
// ----------------------------------------------------------------------
void RigidToMat4( const RotMat3D& Rot, const Vec3<double>& Trans, Matrix4x4<double>& M)
{
  M.Null();
  for(Index i=0;i<3;i++)
    for(Index j=0;j<3;j++)
      M(i,j) = Rot(i,j);
  for(Index k=0;k<3;k++)
    M(k,3) = Trans(k);
  M(3,3) =1.;
}


// Rigid transfo from L2 to L1
// ----------------------------------------------------------------------
double LSQRegistration( std::list<Vec3<double> >& L1, 
			std::list<Vec3<double> >& L2, 
			RotMat3D& Rot, Vec3<double>& Trans)
{
  std::list<Vec3<double> >::iterator it1, it2;

  // Barycenter computation 
  Vec3<double> G1,G2;
  G1.Null(); G2.Null();
  int n=0;
  for(it1 = L1.begin(), it2 = L2.begin(); 
      (it1 != L1.end()) && (it2 != L2.end()); it1++, it2++){ 
    G1 += (*it1);
    G2 += (*it2);
    n++; };
  G1 *= 1./n;
  G2 *= 1./n;
  
  // Rotation computation
  double s1=0.0, s2=0.0;
  Matrix3x3<double> K; K.Null();

  for(it1 = L1.begin(), it2 = L2.begin(); 
      (it1 != L1.end()) && (it2 != L2.end()); it1++, it2++){
    Vec3<double> tmp1, tmp2;
    Mat<3,3,double> Ki;

    // Baricentric coordinates
    tmp1.Sub( *it1 , G1);
    tmp2.Sub( *it2 , G2);

    // scales
    s1 += DotProd( tmp1, tmp1);
    s2 += DotProd( tmp2, tmp2);
    
    // Correlation matrix
    Ki.Mul( tmp1.ColMat(), tmp2.RowMat());
    K += Ki;
  }
  
  // maximize tr(R.Ktrp) with SVD
  try{
    double residue = Rot.LSQ(K);
    
    // Trans = G1 - R.G2
    Trans.Mul( (const Matrix3x3<double>) Rot, -G2);
    Trans += G1;
    
    return (s1+s2 -2.0*residue)/((double) n);
  }
  catch(ErrConvergence& exc){
    exc.push("LSQRegistration( std::list<A> L1, std::list<A> L2, RotMat3D Rot, Vec3<double> Trans)");
    throw exc;
  }
  catch(RotMat3D::ErrNonRotMat& exc){
    exc.push("LSQRegistration( std::list<A> L1, std::list<A> L2, RotMat3D Rot, Vec3<double> Trans)");
    throw exc;
  }
  catch(NonUniqueOptimum& exc){
    exc.push("LSQRegistration( std::list<A> L1, std::list<A> L2, RotMat3D Rot, Vec3<double> Trans)");
    throw exc;
  }
}

// Rigid transfo from L2 to L1
// ----------------------------------------------------------------------
double LSQRegistration( std::list<Vec3<double> >& L1, std::list<Vec3<double> >& L2, Matrix4x4<double>& M)
{
  RotMat3D R;
  Vec3<double> T;
  double tmp = LSQRegistration(L1,L2,R,T);
  RigidToMat4( R,T,M);
  return tmp;
}


//----------------------------------------------------------------------
// Test the LSQ Registration
//----------------------------------------------------------------------

int TestReg(void){
  
  const Index NumPts = 15;
  const double sig = 5.0;

  // Test of LSQRegistration
  double maxdiff_r = 0, maxdiff_t=0;
  for(Index i=0; i<ITER;i++){
    
    // Random transfo
    RotMat3D R,Restim;
    R.Rnd();
    
    Vec3<double> T, Testim;
    T.Rnd( -128., 128.);
    
    // Compute two sets of matched points
    std::list<Vec3<double> > L1, L2;
    for(Index k=0; k<NumPts; k++){
      // Random point in im2
      Vec3<double> pt2; // = new( Vec3<double> );
      pt2.Rnd( -256., 256.);
      
      // transform point into im1
      Vec3<double> pt1; // = new( Vec3<double> );
      pt1.Mul( (const Matrix3x3<double>) R, pt2);
      pt1 += T;
      
      // Add a noise
      Vec3<double> noise;
      pt1 += noise.Rnd(-sig,sig);
      pt2 += noise.Rnd(-sig,sig);
      
      // Save in lists: pt1 and pt2 are copied 
      L2.push_back( pt2 );
      L1.push_back( pt1 );

    }
    
    LSQRegistration(L1, L2, Restim, Testim); 

    // Lists are automatically freed with its elements (as they are not pointers)
    

    
    double dr = Dist(R, Restim);
    double dt = Dist(T, Testim);

    maxdiff_r = max( maxdiff_r,  dr);
    maxdiff_t = max( maxdiff_t,  dt);
	
    if ( ( maxdiff_t > 3.0 * sig  / sqrt(NumPts/10.) ) 
	 || ( maxdiff_r > 0.02 * sig / sqrt(NumPts/10.) )) {
      std::cerr << "Original rotation : "<< std::endl << R;
      std::cerr << "Estimated matrix : "<< std::endl << Restim << std::endl;
      std::cerr << "Original translation : "<< std::endl << T;
      std::cerr << "Estimated translation : "<< std::endl << Testim << std::endl;
      
      throw ErrBug("LSQRot( Mat<m,n>&, Mat<n,n>&)");
    }
  
  }
    
      
  std::cout << std::endl;
  std::cout << "Function LSQRegistration( list, list, R, t) appears OK" << std::endl;
  std::cout << NumPts << " random points in an image of 512x512x512 with a noise of sigma = " << sig << std::endl;
  std::cout << "  maxdiff R  is " << maxdiff_r << " rad ( " << ITER << " trials)" << std::endl;
  std::cout << "  maxdiff t  is " << maxdiff_t << " vox ( " << ITER << " trials)" << std::endl;

  return 0;
}


int BasicsRegTest(int argc, char *argv[]){
  try{
    TestReg();
  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return(1);
  }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }
  return 0;
}
