//**********************************************************************
//
//  TestSVD.cc  ------ Test of svd operations -- 
//
//  (c) Xavier PENNEC -- INRIA -- April 1996 / December 1998
//  
//**********************************************************************

#include <linearAlgebra/Rotation3D.h>
#include "BasicsTest.h"

using namespace yav;

#include <list>

//----------------------------------------------------------------------
//  R_LSQ_SVD : input matrix K, output the rotation matrix R maximizing 
//              C = trace( R. K\trp ) (K and R are (n,n) matrices).     
//              residue is the value at the maximum                       
//----------------------------------------------------------------------

template<Index n,class Real>
Real LSQRot( const Matrix<n,Real>& K, Matrix<n,Real>& R)
{
  Matrix<n,Real> U,V,W;
  Vect<n,Real> w;
  
  try{ K.SVD(U, w, V); }
  catch(ErrConvergence& exc){
    exc.push("LSQRot(RealMat<m,n,Real>& K, RealMatrix<n,Real> R");
    throw exc;
  }

  V.Trp(V);
  R.Mul( U, V);  // R = U.Vtrp

  Real s = R.Det();
  // If s = det(U.Vtrp) id not +1 or -1 : error 
  if ( fabs( fabs(s) -1.) >  EPS(s)*100. ){
    std::cerr << "det = " << s << "      |det|-1 = " << fabs(s) -1. << std::endl;
    // throw RotMat3D::ErrNonRotMat("LSQRot : non unit determinant");
    throw ErrDivZero("LSQRot : non unit determinant");
  }
  // If s = -1, we should multiply by s = -1 both the smallest 
  // singular value of diag and the corresponding ligne in R 
  if (s < 0.0){
    // find smallest singular value
    s = w(0); 
    Index i_dmax = 0;
    for(Index i=1;i<n;i++)
      if( w(i) < s ) {
	s = w(i);
	i_dmax = i;
      }
    // Multiply the sing. val and the corresponding row in V by -1 
    w(i_dmax) *= -1.0;
    for(Index i=0;i<n;i++)
      V(i_dmax,i) *= -1.0;
    // and recompute the rotation
    R.Mul( U, V);  // R = U.Vtrp
  }

  
  // Grep for the number of singular (0) values
  {
    // Find the max 
    Real smax = 0;
    for(Index i=0;i<n;i++)
      smax = max( smax, w(i));
    
    // counting small singular values 
    int num = 0;
    if ( smax < Xlinalg_Small_Val<Real>() )
      num = n;
    else
      for(Index i=0;i<n;i++)
	//if ( ( (Real) w(i)/100. + smax) == ((Real) smax)) num++;
	if( w(i)/100. < smax * Xlinalg_Small_Val<Real>() ) num++;
    
    if( num > 1){
      std::cerr << "*** WARNING: Matrix has %d null singular values" << num << std::endl;
      throw NonUniqueOptimum("LSQRot");
    }
  }
  

  // compute the residue 
  {
    Real residue =0.0;
    for(Index i=0;i<n;i++) 
      residue += w(i);
    return residue;
  }
}



//----------------------------------------------------------------------
// Test the SVD
//----------------------------------------------------------------------
const Index ITER=100000;

template <class Real>
void TestSVD(void){
  
  { // Test the SVD for m = n
    Real maxdiff = 0, maxdiff_V =0, maxdiff_U=0;
    for(Index i=0; i<ITER;i++){
      Matrix<5,Real> A,U,V,W,Vtrp,Utrp,Id;
      Vect<5,Real> w;
      
      A.Rnd();
      try{
	A.SVD(U,w,V);
      }
      catch( ErrConvergence& e){
	// If the algorithm does not converge, print this curious matrix
	std::cerr << "SVD failed for the matrix:\n" << A << std::endl;
	throw e;
      }

      // Reconstruction of the matrix
      Vtrp.Trp(V);
      W.Diag(w);
      
      W.Mul(W,Vtrp);
      W.Mul(U,W);
      
      Utrp.Trp(U);
      Id.Id();
      Real du = max( DistLinf( U * Utrp, Id), DistLinf(Id, Utrp *U));
      maxdiff_U = max( maxdiff_U,  du);
      Real dv = max( DistLinf( V * Vtrp, Id), DistLinf(Id, Vtrp *V));
      maxdiff_V = max( maxdiff_V,  dv);
      Real d = DistLinf( A,W);
      maxdiff = max( maxdiff,d);
      
      if ( maxdiff > EPS(maxdiff) *100.0 ){
	std::cout << "Original matrix : A"<< std::endl << A;
	std::cout << "Reconstructed matrix : "<< std::endl << W;
	std::cout << "U"<< std::endl << U << "w : "<< w;
	std::cout << "V" << std::endl << V << std::endl;

	std::cout << "  maxdiff      A / U.Diag(w).Vtrp is " 
	     << maxdiff << " ( iter:" << i << " )" << std::endl;
	std::cout << "  maxdiff    V.Vtrp / Vtrp.V / Id is " 
	     << maxdiff_V << " ( iter:" << i << ")" << std::endl;
	std::cout << "  maxdiff    U.Utrp / Utrp.U / Id is " 
	     << maxdiff_V << " ( iter:" << i << " )" << std::endl;
	  throw ErrBug("Matrix<n>::SVD( Matrix<n>&, Vect<n>&, Matrix<n>&)");
      }
    }
    
    std::cout << std::endl;
    std::cout << "Function Matrix<n>::SVD( Matrix<n>&, Vect<n>&, Matrix<n>&) appears OK" << std::endl;
    std::cout << "  maxdiff      A / U.Diag(w).Vtrp is " 
	 << maxdiff << " ( " << ITER << " trials)" << std::endl;
    std::cout << "  maxdiff    V.Vtrp / Vtrp.V / Id is " 
	 << maxdiff_V << " ( " << ITER << " trials)" << std::endl;
    std::cout << "  maxdiff    U.Utrp / Utrp.U / Id is " 
	 << maxdiff_V << " ( " << ITER << " trials)" << std::endl;
  }

 { // Test the SVD for m > n
    Real maxdiff = 0, maxdiff_V =0, maxdiff_U=0;
    for(Index i=0; i<ITER;i++){
      Mat<5,3,Real> A,U, Arec;
      Matrix<3,Real> V,W,Vtrp,Id;
      Mat<3,5,Real> Utrp;
      Vect<3,Real> w;
      
      A.Rnd();
      try{
	A.SVD(U,w,V);
      }
      catch( ErrConvergence& e){
	// If the algorithm does not converge, print this curious matrix
	std::cerr << "SVD failed for the matrix:\n" << A << std::endl;
	throw e;
      }

      // Reconstruction of the matrix
      Vtrp.Trp(V);
      W.Diag(w);
      
      W.Mul(W,Vtrp);
      Arec.Mul(U,W);
      
      Utrp.Trp(U);
      Id.Id();
      Real du = DistLinf(Id, Utrp *U);
      maxdiff_U = max( maxdiff_U,  du);
      Real dv = max( DistLinf( V * Vtrp, Id), DistLinf(Id, Vtrp *V));
      maxdiff_V = max( maxdiff_V,  dv);
      Real d = DistLinf( A,Arec);
      maxdiff = max( maxdiff,d);
      
      if ( maxdiff > EPS( maxdiff) *100.0 ){
	std::cout << "  maxdiff      A / U.Diag(w).Vtrp is " 
	     << maxdiff << " ( iter:" << i << " )" << std::endl;
	std::cout << "  maxdiff    V.Vtrp / Vtrp.V / Id is " 
	     << maxdiff_V << " ( iter:" << i << " )" << std::endl;
	std::cout << "  maxdiff             Utrp.U / Id is " 
	     << maxdiff_V << " ( iter:" << i << " )" << std::endl;

	throw ErrBug("Mat<m,n>::SVD( Mat<m,n>&, Vect<n>&, Matrix<n>&) (m>n)");
      }
    }
    
    std::cout << std::endl;
    std::cout << "Function Mat<m,n>::SVD( Mat<m,n>&, Vect<n>&, Matrix<n>&) appears OK (m>n)" << std::endl;
    std::cout << "  maxdiff      A / U.Diag(w).Vtrp is " 
	 << maxdiff << " ( " << ITER << " trials)" << std::endl;
    std::cout << "  maxdiff    V.Vtrp / Vtrp.V / Id is " 
	 << maxdiff_V << " ( " << ITER << " trials)" << std::endl;
    std::cout << "  maxdiff             Utrp.U / Id is " 
	 << maxdiff_V << " ( " << ITER << " trials)" << std::endl;
  }

    { // Test the SVD for m < n
    Real maxdiff = 0, maxdiff_V =0, maxdiff_U=0;
    for(Index i=0; i<ITER;i++){
      Mat<3,5,Real> A,U, Arec;
      Matrix<5,Real> V,W,Vtrp,Id;
      Mat<5,3,Real> Utrp;
      Vect<5,Real> w;
      
      A.Rnd();
      try{
	A.SVD(U,w,V);
      }
      catch( ErrConvergence& e){
	// If the algorithm does not converge, print this curious matrix
	std::cerr << "SVD failed for the matrix:\n" << A << std::endl;
	throw e;
      }

      // Reconstruction of the matrix
      Vtrp.Trp(V);
      W.Diag(w);
      
      W.Mul(W,Vtrp);
      Arec.Mul(U,W);
      
      Utrp.Trp(U);
      Id.Id();
      Real du = DistLinf(Id, Utrp *U);
      maxdiff_U = max( maxdiff_U,  du);
      Real dv = max( DistLinf( V * Vtrp, Id), DistLinf(Id, Vtrp *V));
      maxdiff_V = max( maxdiff_V,  dv);
      Real d = DistLinf( A,Arec);
      maxdiff = max( maxdiff,d);
      
      if ( maxdiff > EPS( maxdiff) *100.0 ){
	std::cout << "Original matrix : A"<< std::endl << A;
	std::cout << "Reconstructed matrix : "<< std::endl << Arec;
	std::cout << "U"<< std::endl << U << "w : "<< w;
	std::cout << "V" << std::endl << V;

	std::cout << "d = " << d << "      dv = " << dv << "      du = " << du 
	     << std::endl << std::endl << std::endl;

	throw ErrBug("Mat<m,n>::SVD( Mat<m,n>&, Vect<n>&, Matrix<n>&) (m<n)");
      }
    }
  
    std::cout << std::endl;
    std::cout << "Function Mat<m,n>::SVD( Mat<m,n>&, Vect<n>&, Matrix<n>&) appears OK (m<n)" << std::endl;
    std::cout << "  maxdiff      A / U.Diag(w).Vtrp is " 
	 << maxdiff << " ( " << ITER << " trials)" << std::endl;
    std::cout << "  maxdiff    V.Vtrp / Vtrp.V / Id is " 
	 << maxdiff_V << " ( " << ITER << " trials)" << std::endl;
    std::cout << "  maxdiff             Utrp.U / Id is " 
	 << maxdiff_V << " ( " << ITER << " trials)" << std::endl;
    

  }


  { // Test of LSQRot
    Real maxdiff = 0;
    for(Index i=0; i<ITER;i++){
      RotMat3D R,Restim;
      R.Rnd();R.Rnd();R.Rnd();R.Rnd();R.Rnd();R.Rnd();

      Matrix3x3<Real> K(R);

      Restim.LSQ(K);
      Matrix3x3<Real> Kestim (Restim);

      Real d = DistLinf(Kestim, K);
      maxdiff = max( maxdiff,  d);

      if ( maxdiff > EPS( maxdiff) *100.0 ){
	std::cout << "Original rotation : K"<< std::endl << K;
	std::cout << "Estimated matrix : "<< std::endl << Restim << std::endl;

	throw ErrBug("LSQRot( Mat<m,n>&, Matrix<n>&)");
      }
    }
  
    std::cout << std::endl;
    std::cout << "Function LSQRot( Mat<m,n,Real>&, Matrix<n,Real>&) appears OK" << std::endl;
    std::cout << "  maxdiff R / LSQRot(R) is " << maxdiff << " ( " << ITER << " trials)" << std::endl;

  }
    
  
}

int BasicsSVDTest(int argc, char *argv[]){
  try{
    
    clock_t begin=0;
    clock_t end=0;
    
    begin=clock();
    TestSVD<float>();
    end=clock();
    std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC)  << std::endl;
    
    begin=clock();
    TestSVD<double>();
    end=clock();
    std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC) << std::endl;
  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return(1);
  }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return(1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in main:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return(1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in main:end, exiting\n";
    return(1);
  }
  return 0;
}
