// ************************************************************************
//
//   TestVect.cc  ------  Test of static vector class  (Vect.hh) 
//
//   C++ translation of a part of K_linalg (c) X. Pennec INRIA 1994 / 1995
//
//	(c) Xavier Pennec -- INRIA -- Ferbuary 1997
//
// ************************************************************************
//
//  Vect<dim>::Vect( Index )
//  Vect<dim>::Rnd
//  Vect<dim>::Null
//  Add( Vect<dim>, Vect<dim>, Vect<dim> )
//  Sub( Vect<dim>, Vect<dim>, Vect<dim> )
//  Mul( Real, Vect<dim>, Vect<dim> )
//  DotProd( Vect<dim>, Vect<dim>, Real )
//  Real DotProd( Vect<dim>, Vect<dim> )
//  Real Vect<dim>::Norm()
//  Real Dist( Vect<dim>, Vect<dim> ) 
//  Real Vect<dim>::NormL1()
//  Real DistL1( Vect<dim>, Vect<dim> ) 
//  Real Vect<dim>::NormLinf()
//  Real DistLinf( Vect<dim>, Vect<dim> ) 
//  Real& operator()( Index i ); 
//
// ************************************************************************


#include <linearAlgebra/Vec.h>
#include "BasicsTest.h"

using namespace yav;

const Index Ntries = 100000;


// Add two vectors (procedure)
// ----------------------------------------------------------------------
template<Index dim, class Real>
inline void Add( const Vect<dim, Real>& u, 
		 const Vect<dim, Real>& v, 
		 Vect<dim, Real>& res) throw()
{
  for( Index i=0; i < dim; i++)
    res(i) = u(i) + v(i);
};

// Substract two vectors
// ----------------------------------------------------------------------
template<Index dim, class Real>
inline void Sub( const Vect<dim, Real>& u, 
	  const Vect<dim, Real>& v, 
	  Vect<dim, Real>& res) throw()
{
  for( Index i=0; i < dim; i++)
    res(i) = u(i) - v(i);
};


// Inner product (procedure)
// ----------------------------------------------------------------------
template<Index dim, class Real>
inline void DotProd( const Vect<dim,Real>& u, 
		     const Vect<dim,Real>& v, 
		     Real& res) throw()
{
  register Real tmpres = 0;
  for( Index i=0; i < dim; i++)  tmpres += u(i) * v(i);
  res = tmpres;
};



// Test functions (template of type Real)
// ----------------------------------------------------------------------
template<class Real=double>
class rigid{
public:
  Vect<6,Real> f;
  Vect<3,Real>& r;
  Vect<3,Real>& t;

  rigid() : r( (Vect<3,Real>&) f), t( (Vect<3,Real>&) f(3) ) {};
};

template <class Real>
int TestVect(void){
  try{
    rigid<Real> f;

    std::cout << std::endl << "******************************************" <<std::endl; 
    std::cout << "Verifying TEMPLATE CLASS Vect<dim,"<<classname<Real>()<<"> " << std::endl ;
    std::cout << "******************************************" << std::endl << std::endl;

    
    std::cout << "Verify initialization to basis vector\n";
    {
      Vect<3> e1(0), e2(1), e3(2);
      std::cout << "e1 : " << e1 ;
      std::cout << "e2 : " << e2 ;
      std::cout << "e3 : " << e3 ;

#undef RANGE_CHECK
#ifdef RANGE_CHECK
      try {Vect<3,Real> e4(3);
      throw ErrBug("Range checking is bugged in Real& Vect<dim,Real>::Vect( Index )");
      }
      catch( ErrRange& exc){
	std::cout << "Range checking OK for Real& Vect<dim,"<<classname<Real>()<<
	  ">::Vect( Index )\n"; 
      }
#endif
      std::cout << std::endl;
    }
    
    f.f.Rnd(-10,10);
    std::cout << "f : 6D random vector(-10,10):\t" << f.f ;
    std::cout << "r : ref on the 3 first coords:\t" << f.r;
    std::cout << "t : ref on the 3 last coords:\t\t\t\t    " << f.t << std::endl;
    
    // This should be errors at compilation
    // Add( f.r, f.t, f.f);
    // Add( f.f, f.t, f.f);
    
    f.r.Null();
    std::cout << "r set to null:\t\t\t" << f.r ;
    std::cout << "verify effect on f:\t\t" << f.f << std::endl;
    
    f.t.Rnd();
    std::cout << "t set to Rnd(-1,1):\t\t\t " << f.t;
    std::cout << "verify effect on f:\t\t" << f.f << std::endl;
    
    f.f.Rnd(-10,10);
    Vect<3,Real> sum;
    std::cout << "f : 6D random vector(-10,10):\t" << f.f ;
    Add( f.r, f.t, sum);
    std::cout << "sum = r + t :\t\t\t" << sum;
    Sub( f.r, f.t, sum);
    std::cout << "diff = r - t :\t\t\t" << sum;
    f.t *= (Real) 10.0;
    std::cout << "t *= 10 :\t\t\t" << f.t;
    
    // Test of DotProd, Norms and distances
    // -----------------------------------------------------
    {
      Real dot;
      DotProd( f.r, f.t, dot);
      std::cout << "< r | t > =:\t\t\t" << dot << std::endl;
      if ( fabs( (dot - DotProd( f.r, f.t))/dot ) > 1e-6 ){
	std::cerr.precision(30);
	std::cerr << "Real DotProd( f.r, f.t) = " << DotProd( f.r, f.t) << std::endl;
	std::cerr << "DotProd( f.r, f.t, dot)   = " << dot << std::endl;
        throw ErrBug("DotProd function does not give the same result as procedure");
      }
      std::cout << "\nDotProd function gives the same result as the procedure\n";
    }
    
    // Test of element access in write and read mode
    // -----------------------------------------------------
    {
      Vect<3,Real> u;
      std::cout << std::endl;
#ifdef RANGE_CHECK
      // test negative index
      int i= -1;
      try{ 
	u(i) = 0.0;
	throw ErrBug("Range checking is bugged in Real& Vect<dim,Real>::operator()");
      }
      catch( ErrRange& e){};

      // Test out of range
      try{
	u(3) = 0;
	throw ErrBug("Range checking is bugged in Real& Vect<dim,Real>::operator()");
      }
      catch( ErrRange& e){};
      std::cout << "Range checking appears OK for "<<classname<Real>()<<
	"& Vect<dim,"<<classname<Real>()<<">::operator()\n"; 
#endif
      // Test if set and read properly
      u(0) = 1; u(1) = -2; u(2) = -3;
      if ( ( u(0) != 1.) || ( u(1) != -2.) || (u(2) != -3))
	throw ErrBug("Member  Real& Vect<dim,Real>::operator()  is bugged");
      std::cout << std::endl;
    }
    
    // Test of Euclidean Norms and distances
    // -----------------------------------------------------
    {
      RndInit();
      Real n,d;
      Vect<3> zero;  zero.Null();
      Vect<3> u; u(0) = 1; u(1) = -2; u(2) = -3;
      
      std::cout << "Vector u: \t" << u;
      std::cout << std::endl ;
      std::cout << "Euclidean norm and distance:";
      n = u.Norm();
      d = Dist( u, zero );
      std::cout << " \tNorm(u) = " << n
	   << " \tDist(u,0) = " << d 
	   << " \tdiff = " << d - n << std::endl;
      if ( fabs( n - sqrt(14.))/n > EPS( n) ){
	std::cerr.precision(30);
	std::cerr << "u.Norm() = " << n << "      different from" <<std::endl;
	std::cerr << "sqrt(14) = " << sqrt(14.)  << std::endl;
	throw ErrBug("Function Vect<dim,Real>::Norm() is bugged");
      }
      if ( fabs(d - sqrt(14.)) >  EPS(d)  ){
	std::cerr.precision(30);
	std::cerr << "Dist( u, zero ) = " << d << "different from 14.0" <<std::endl;
	std::cerr << "Dist( u, zero ) = " <<  Dist( u, zero ) << std::endl;
	throw ErrBug("Function Dist( Vect<dim,Real>, Vect<dim> ) is bugged");      
      }
      std::cout << "Functions Vect<dim,"<<classname<Real>()<<
	">::Norm() and Dist( Vect<dim,"<<classname<Real>()<<
	">, Vect<dim,"<<classname<Real>()<<"> ) appear OK" << std::endl;
      
      // Test of L1 Norms and distances
      // -----------------------------------------------------
      std::cout << std::endl ;
      std::cout << "L1 norm and distance:\t";
      n = u.NormL1();
      d = DistL1( u, zero );
      std::cout << " \tNormL1(u) = " << n
	   << " \t\tDistL1(u,0) = " << d 
	   << " \tdiff = " << d - n << std::endl;
      if ( fabs(n - 6.) >  EPS(n)  )
	throw ErrBug("Function Vect<dim,Real>::NormL1() is bugged");
      if ( fabs(d - 6.) >  EPS(d)  )
	throw ErrBug("Function DistL1( Vect<dim,Real>, Vect<dim,Real> ) is bugged");
      std::cout << "Functions Vect<dim,"<<classname<Real>()<<
	">::NormL1() and DistL1( Vect<dim,"<<classname<Real>()<<
	">, Vect<dim,"<<classname<Real>()<<"> ) appear OK" << std::endl;
      
      // Test of Linf Norms and distances
      // -----------------------------------------------------
      std::cout << std::endl ;
      std::cout << "Linfinite norm and distance:";
      n = u.NormLinf();
      d = DistLinf( u, zero );
      std::cout << " \tNormLinf(u) = " << n
	   << " \tDistLinf(u,0) = " << d 
	   << " \tdiff = " << d - n << std::endl;
      if ( fabs( n - 3.) >  EPS(n)  )
	throw ErrBug("Function Vect<dim,Real>::NormLinf() is bugged");
      if ( fabs( d - 3.) > EPS(d)  )
	throw ErrBug("Function DistLinf( Vect<dim,Real>, Vect<dim,Real> ) is bugged");
      std::cout << "Functions Vect<dim,"<<classname<Real>()<<
	">::NormLinf() and DistLinf( Vect<dim,"<<classname<Real>()<<
	">, Vect<dim,"<<classname<Real>()<<"> ) appear OK" << std::endl;
    }
    
    // Test of Add, operator +, operator +=
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      double maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<6, Real> u, v, upv, upv_func, upv_op;
	double dfunc, dop, dope;
	u.Rnd(); v.Rnd();
	Add(u, v, upv);
	upv_func.Add(u,v);
	upv_op = u+v;
	u += v;
	
	dop = Dist( upv, upv_op);
	dfunc = Dist( upv, upv_func);
	dope = Dist( upv, u);
	maxdiff = max( maxdiff, max( dop, max(dfunc, dope)));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Function Vect::Add(Vect&, Vect& ) or Vect::operator + ( Vect&, Vect&) or Vect::operator += ( Vect& )  is bugged");
      }
      std::cout <<  "Addition:    maxdiff = " << maxdiff << std::endl;
      std::cout << "Function Add, operator + and operator += appears oK" << std::endl;
    }

    // Test of Add, operator +, operator += (Vect<4>)
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      double maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<4, Real> u, v, upv, upv_func, upv_op;
	double dfunc, dop, dope;
	u.Rnd(); v.Rnd();
	Add(u, v, upv);
	upv_func.Add(u,v);
	upv_op = u+v;
	u += v;
	
	dop = Dist( upv, upv_op);
	dfunc = Dist( upv, upv_func);
	dope = Dist( upv, u);
	maxdiff = max( maxdiff, max( dop, max(dfunc, dope)));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Function Vect<4>::Add(Vect<4>&, Vect<4>& ) or Vect<4>::operator + ( Vect<4>&, Vect<4>&) or Vect<4>::operator += ( Vect<4>& )  is bugged");
      }
      std::cout <<  "Addition (Vect<4>):    maxdiff = " << maxdiff << std::endl;
      std::cout << "Function Add, operator + and operator += (Vect<4>) appears oK" << std::endl;
    }

    // Test of Add, operator +, operator += (Vect<3>)
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      double maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<3, Real> u, v, upv, upv_func, upv_op;
	double dfunc, dop, dope;
	u.Rnd(); v.Rnd();
	Add(u, v, upv);
	upv_func.Add(u,v);
	upv_op = u+v;
	u += v;
	
	dop = Dist( upv, upv_op);
	dfunc = Dist( upv, upv_func);
	dope = Dist( upv, u);
	maxdiff = max( maxdiff, max( dop, max(dfunc, dope)));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Function Vect<3>::Add(Vect<3>&, Vect<3>& ) or Vect<3>::operator + ( Vect<3>&, Vect<3>&) or Vect<3>::operator += ( Vect<3>& )  is bugged");
      }
      std::cout <<  "Addition (Vect<3>):    maxdiff = " << maxdiff << std::endl;
      std::cout << "Function Add, operator + and operator += (Vect<3>) appears oK" << std::endl;
    }


    // Test of Sub, operator -, operator -=, unary operator -
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      double maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<3, Real> u, v, umv, umv_func, umv_op, upmv;
	double dfunc, dop, dope, dpmv;
	u.Rnd(); v.Rnd();
	Sub(u, v, umv);
	umv_func.Sub(u,v);
	umv_op = u-v;
	upmv.Add( u, -v);
	u -= v;
	
	dop = Dist( umv, umv_op);
	dfunc = Dist( umv, umv_func);
	dope = Dist( umv, u);
	dpmv = Dist(umv, upmv);

	maxdiff = max( maxdiff, max( dop, max(dfunc, max( dope, dpmv) )));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Function Vect::Sub(Vect&, Vect& ) or Vect::operator - ( Vect&, Vect&) or Vect::operator -= ( Vect& )  is bugged");
      }
      std::cout <<  "Subtraction:    maxdiff = " << maxdiff << std::endl;
      std::cout << "Function Sub, operator - and operator -= appears oK" << std::endl;
    }

    // Test of Sub, operator -, operator -=, unary operator - (Vect<4>)
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      double maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<4, Real> u, v, umv, umv_func, umv_op, upmv;
	double dfunc, dop, dope, dpmv;
	u.Rnd(); v.Rnd();
	Sub(u, v, umv);
	umv_func.Sub(u,v);
	umv_op = u-v;upmv.Add( u, -v);
	u -= v;
	
	dop = Dist( umv, umv_op);
	dfunc = Dist( umv, umv_func);
	dope = Dist( umv, u);
	dpmv = Dist(umv, upmv);

	maxdiff = max( maxdiff, max( dop, max(dfunc, max( dope, dpmv) )));
	
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Function Vect::Sub(Vect<4>&, Vect<4>& ) or Vect<4>::operator - ( Vect<4>&, Vect<4>&) or Vect<4>::operator -= ( Vect<4>& )  is bugged");
      }
      std::cout <<  "Subtraction (Vect<4>):    maxdiff = " << maxdiff << std::endl;
      std::cout << "Function Sub, operator - and operator -= (Vect<4>) appears oK" << std::endl;
    }
    
    // Test of Sub, operator -, operator -=, unary operator - (Vect<3>)
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      double maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<3, Real> u, v, umv, umv_func, umv_op, upmv;
	double dfunc, dop, dope, dpmv;
	u.Rnd(); v.Rnd();
	Sub(u, v, umv);
	umv_func.Sub(u,v);
	umv_op = u-v;
	upmv.Add( u, -v);
	u -= v;
	
	dop = Dist( umv, umv_op);
	dfunc = Dist( umv, umv_func);
	dope = Dist( umv, u);
	dpmv = Dist(umv, upmv);

	maxdiff = max( maxdiff, max( dop, max(dfunc, max( dope, dpmv) )));
		
	if ( maxdiff > EPS( maxdiff)  )
	  throw ErrBug("Function Vect<3>::Sub(Vect<3>&, Vect<3>& ) or Vect<3>::operator - ( Vect<3>&, Vect<3>&) or Vect<3>::operator -= ( Vect<3>& )  is bugged");
      }
      std::cout <<  "Subtraction (Vect<3>):    maxdiff = " << maxdiff << std::endl;
      std::cout << "Function Sub, operator - and operator -= (Vect<3>) appears oK" << std::endl;
    }


    // Test of Scalar Mul and operator *, *=
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      Real maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<5,Real> u, ru, ur, ufr, rfu;
	Real r = 20.0 * ( rand() / (RAND_MAX+1.0) );
	u.Rnd();
	ru.Mul(r, u);
	ur.Mul(u,r);
	rfu = r * u; 
	ufr = rfu; //ufr = u * r;
	u *= r;
	maxdiff = max( maxdiff, max( DistLinf( u, ru), DistLinf( u, ur)));
	maxdiff = max( maxdiff, max( DistLinf( u, rfu), DistLinf( u, ufr)));

	if ( maxdiff >  EPS( maxdiff) )
	  throw ErrBug("Function Vect& Mul( Real, Vect&) or Real& Mul( Vect&, Real) or  Vect::operator *= ( Real ) or Vect operator *(Vect&, Vect&)  is bugged");
      }
      std::cout << "Scalar multiplication :  maxdiff : " << maxdiff << std::endl;
      std::cout << "Functions (scalar) Mul and operators * and *= appears OK"
	   << std::endl;
    }

    // Test of Scalar Mul and operator *, *=  (Vect<4>)
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      Real maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<4,Real> u, ru, ur, ufr, rfu;
	Real r = 20.0 * ( rand() / (RAND_MAX+1.0) );
	u.Rnd();;
	ru.Mul(r, u);
	ur.Mul(u,r);
	rfu = r * u;
	ufr = rfu;
	u *= r;
	maxdiff = max( maxdiff, max( DistLinf( u, ru), DistLinf( u, ur)));
	maxdiff = max( maxdiff, max( DistLinf( u, rfu), DistLinf( u, ufr)));

	if ( maxdiff >  EPS( maxdiff) )
	  throw ErrBug("Function Vect<4>& Mul( Real, Vect<4>&) or Real& Mul( Vect<4>&, Real) or  Vect<4>::operator *= ( Real ) or Vect<4> operator *(Vect<4>&, Vect<4>&)  is bugged");
      }
      std::cout << "Scalar multiplication (Vect<4>):  maxdiff : " << maxdiff << std::endl;
      std::cout << "Functions (scalar) Mul and operators * and *= (Vect<4>)appears OK"
	   << std::endl;
    }

    // Test of Scalar Mul and operator *, *=  (Vect<3>)
    // -----------------------------------------------------
    {
      RndInit();
      std::cout << std::endl;
      
      Real maxdiff =0;
      for(Index i=0; i< Ntries; i++){
	Vect<3,Real> u, ru, ur, ufr, rfu;
	Real r = 20.0 * ( rand() / (RAND_MAX+1.0) );
	u.Rnd();;
	ru.Mul(r, u);
	ur.Mul(u,r);
	rfu = r * u;
	ufr = rfu;
	u *= r;
	maxdiff = max( maxdiff, max( DistLinf( u, ru), DistLinf( u, ur)));
	maxdiff = max( maxdiff, max( DistLinf( u, rfu), DistLinf( u, ufr)));

	if ( maxdiff >  EPS( maxdiff) )
	  throw ErrBug("Function Vect<3>& Mul( Real, Vect<3>&) or Real& Mul( Vect<3>&, Real) or  Vect<3>::operator *= ( Real ) or Vect<3> operator *(Vect<3>&, Vect<3>&)  is bugged");
      }
      std::cout << "Scalar multiplication (Vect<3>):  maxdiff : " << maxdiff << std::endl;
      std::cout << "Functions (scalar) Mul and operators * and *= (Vect<3>)appears OK"
	   << std::endl;
    }

    
  }
  catch( Exception& e){
    e.push( "main:end");
    std::cerr << e << std::endl;
    return( 1);
    }
  catch( std::exception& e){
    std::cerr << "Trapped an unexpected stdlib exception in TestVect:end, exiting\n";
    std::cerr << e.what() << std::endl;
    return( 1);
  }
  catch( std::exception *e){
    std::cerr << "Trapped an unexpected stdlib exception in TestVect:end, exiting\n";
    std::cerr << e->what() << std::endl;
    return( 1);
  }
  catch(...){
    std::cerr << "Trapped an unexpected exception in TestVect:end, exiting\n";
    return ( 1);
  }

  std::cout << std::endl << std::endl << "No major bug detected\n";
  
  std::cout << std::endl << "******************************************" <<std::endl; 
  std::cout << "TEMPLATE CLASS Vect<dim,"<<classname<Real>()<<"> APPEARS TO BE OK" << std::endl ;
  std::cout << "******************************************" << std::endl;
  
  return 0;
}



#include <time.h>

int BasicsVectTest(int argc, char *argv[]){
  clock_t begin=0;
  clock_t end=0;
  
  begin=clock();
  int test_double = TestVect<double>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC)  << std::endl;
  
  begin=clock();
  int test_float= TestVect<float>();
  end=clock();
  std::cout << "Clock time: " << (end-begin)/ ( (double) CLOCKS_PER_SEC) << std::endl;
  
  return( test_double && test_float );

}
