#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifdef USE_PETSC
#include "petsc.h"
#endif

#pragma implementation "linearAlgebra/GenericSparseMatrix.h"

#include <linearAlgebra/GenericSparseMatrix.h>
#include <linearAlgebra/MTLSparseMatrix.h>
#include <linearAlgebra/PETSCSparseMatrix.h>


// Select linear algebra library according to compiler flags
#define LINALG_LIB "MTL" // We have to see what is the default behavior
#ifdef USE_MTL
#undef  LINALG_LIB
#define LINALG_LIB "MTL"
#endif
#ifdef USE_PETSC
#undef  LINALG_LIB
#define LINALG_LIB "PETSC"
#endif


namespace yav
{

  GenericSparseMatrix *createSparseMatrix(const GenericSparseMatrix::SparseMatrixType smt,
					  const unsigned int sizeRows, 
					  const unsigned int sizeColumns,
					  const unsigned int nbNonZeroDiagonal,
					  const unsigned int nbNonZeroOffDiagonal)
  {
    GenericSparseMatrix* matrix = 0;
	
    if ( !strcmp( LINALG_LIB, "MTL" ) ) 
    {
#ifdef USE_MTL
      if (smt==GenericSparseMatrix::SPARSE)
	matrix=new MTLSparseMatrix(sizeRows,sizeColumns);
      else if (smt==GenericSparseMatrix::SPARSE_SYMMETRIC)
	matrix=new MTLSymmetricSparseMatrix(sizeRows,sizeColumns);
#endif
    }

    if ( !strcmp( LINALG_LIB, "PETSC" ) )
    {
#ifdef USE_PETSC
      // TODO: PETSC_SPARSE_SYMMETRIC
      if ( smt == GenericSparseMatrix::SPARSE ||
	   smt == GenericSparseMatrix::SPARSE_SYMMETRIC )
	matrix = new PETSCSparseMatrix( sizeRows, sizeColumns, nbNonZeroDiagonal, nbNonZeroOffDiagonal );
#endif
    }

    return matrix;
  }


  GenericSparseVector *createSparseVector(const unsigned int sizeRows)
  {
    GenericSparseVector* vector = 0;

    if ( !strcmp( LINALG_LIB, "MTL") )
    {
#ifdef USE_MTL
      vector=new MTLSparseVector( sizeRows );
#endif
    }

    if ( !strcmp( LINALG_LIB, "PETSC") )
    {
#ifdef USE_PETSC
      vector=new PETSCSparseVector( sizeRows );
#endif
    }

    return vector;
  }



  void initializeLinearAlgebraLibrary(  int* argc, char*** args )
  {
#ifdef USE_PETSC

    int npro, myid;
    int myargc = *argc;
    char** myargv;

    PetscInitialize( argc, args, (char*) 0, "Parallel YAV++ package\n\n" );

    MPI_Comm_size(MPI_COMM_WORLD, &npro);
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);

    if ( npro != 1 )
    { 
      myargc = 2;
      myargv = (char**) malloc( sizeof(char*) );
      myargv[0] = (char*) malloc( (strlen((*args)[0])+1) * sizeof(char) );
      strcpy( myargv[0], (*args)[0] );
      if( myid == 0 )
      {
	myargv[1] = (char*) malloc( (strlen((*args)[1])+1) * sizeof(char) );
	strcpy( myargv[1], (*args)[1] );
      }
      else 
      {
	myargv[1] = (char*) malloc( (strlen((*args)[2])+1) * sizeof(char) );
	strcpy( myargv[1], (*args)[2] );
      }

      (*argc) = myargc;
      args = &myargv;
    }

#endif
  }



  void finalizeLinearAlgebraLibrary( void )
  {
#ifdef USE_PETSC
    PetscFinalize();
#endif    
  }

} // namespace yav

