#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifdef USE_PETSC

#pragma implementation "linearAlgebra/PETSCSparseMatrix.h"

#include <linearAlgebra/PETSCSparseMatrix.h>

#include "petscvec.h"
#include "petscksp.h"
#include "petscao.h"


using namespace std;

namespace yav 
{


  PETSCSparseVector::PETSCSparseVector( const unsigned int nbRows )
  {
    VecCreateMPI( PETSC_COMM_WORLD, PETSC_DECIDE, nbRows, &(this->vector) );
    VecZeroEntries( this->vector );
  }


  PETSCSparseVector::~PETSCSparseVector()
  {
    VecDestroy( vector );
  }


  /// add a value to a vector
  void PETSCSparseVector::addVectorElement( const unsigned int row,
					    const double value )
  {
    VecSetValue( this->vector, row, (PetscScalar) value, ADD_VALUES );

    // Assembly: should be done at the very end I think
//     VecAssemblyBegin( this->vector );
//     VecAssemblyEnd( this->vector );
  }


  /// set a value to a vector
  void PETSCSparseVector::setVectorElement( const unsigned int row,
					    const double value )
  {
    VecSetValue( this->vector, row, (PetscScalar) value, INSERT_VALUES);

    // Assembly: should be done at the very end I think
//     VecAssemblyBegin( this->vector );
//     VecAssemblyEnd( this->vector );
  }


  /// get vector value
  double PETSCSparseVector::getVectorElement( const unsigned int row ) const
  {
    PetscScalar v[1]; // Read values
    PetscInt ni[1];   // Indices where to read the values
    ni[0] = row;

    VecGetValues( this->vector, (PetscInt) 1, ni, v );

    return v[0];
  }


  /// set vector to zero
  void PETSCSparseVector::resetVector()
  {
    VecZeroEntries( this->vector );
  }


  /// Assemble matrix : the matrix can no longer be modified
  void PETSCSparseVector::assembleVector()
  {
    VecAssemblyBegin( this->vector );
    VecAssemblyEnd( this->vector );
  }


  /// Get vector as array of double
  const double* PETSCSparseVector::getVectorAsArray()
  {
    double* a;
    VecGetArray( this->vector, (PetscScalar**) &a );
    return a;
  }

  // restore the array of double (free the memory). MUST be done after
  // getVectorAsArray
  void PETSCSparseVector::restoreVectorArray( double* array )
  {
    VecRestoreArray( this->vector, &array );
  }


  /// get vector size
  unsigned int PETSCSparseVector::getSize() const
  {
    PetscInt s;
    VecGetSize( this->vector, &s );
    return s;
  }


  // Save vector in file
  void PETSCSparseVector::saveInFile( const std::string filename ) const
  {
    PetscViewer viewer;
    PetscViewerASCIIOpen( PETSC_COMM_WORLD, filename.c_str(), &viewer);
    VecView( this->vector, viewer );
    PetscViewerFlush( viewer );
    PetscViewerDestroy( viewer );
  }


// -----------------------------------------------------------------------------


  /// Create a matrix of a given number of rows and columns
  PETSCSparseMatrix::PETSCSparseMatrix( const unsigned int sizeRows, 
					const unsigned int sizeColumns,
					const unsigned int nbNonZeroDiagonal,
					const unsigned int nbNonZeroOffDiagonal )
  {
    // DETECT NB PROC

    MatCreateMPIAIJ( PETSC_COMM_WORLD, 
		     PETSC_DECIDE, PETSC_DECIDE,
		     (PetscInt) sizeRows, (PetscInt) sizeColumns,
		     nbNonZeroDiagonal, PETSC_NULL,
		     nbNonZeroOffDiagonal, PETSC_NULL,
		     &(this->matrix) );

    MatZeroEntries( this->matrix );
    MatAssemblyBegin( this->matrix, MAT_FLUSH_ASSEMBLY );
    MatAssemblyEnd( this->matrix, MAT_FLUSH_ASSEMBLY );

    this->solver = NULL;
    this->hasMatrixChanged = true;
  }


  /// destructor
  PETSCSparseMatrix::~PETSCSparseMatrix()
  {
    if ( this->solver )
      KSPDestroy( this->solver );
    MatDestroy( this->matrix );
  }


  /// set all matrix elements to zero
  void PETSCSparseMatrix::resetMatrix()
  {
    MatZeroEntries( this->matrix );
    this->hasMatrixChanged = true;
  }


  /// add a given value to an entry in the matrix
  void PETSCSparseMatrix::addMatrixElement( const unsigned int row,
					    const unsigned int column,
					    const double val )
  {
    MatSetValue( this->matrix, (PetscInt) row, (PetscInt) column, (PetscScalar) val, ADD_VALUES );

    MatAssemblyBegin( this->matrix, MAT_FLUSH_ASSEMBLY );
    MatAssemblyEnd( this->matrix, MAT_FLUSH_ASSEMBLY );
    this->hasMatrixChanged = true;
  }


  /// add a given value to an entry in the matrix and its transpose entry
  void PETSCSparseMatrix::addSymmetricMatrixElement( const unsigned int row,
						     const unsigned int column,
						     const double val )
  {
    
    MatSetValue( this->matrix, (PetscInt) row, (PetscInt) column, (PetscScalar) val, ADD_VALUES );
    if ( row != column )
      MatSetValue( this->matrix, (PetscInt) column, (PetscInt) row, (PetscScalar) val, ADD_VALUES );

    MatAssemblyBegin( this->matrix, MAT_FLUSH_ASSEMBLY );
    MatAssemblyEnd( this->matrix, MAT_FLUSH_ASSEMBLY );
    this->hasMatrixChanged = true;
  }


  // Save matrix in file
  void PETSCSparseMatrix::saveInFile( const std::string filename ) const
  {
    PetscViewer viewer;
    PetscViewerASCIIOpen( PETSC_COMM_WORLD, filename.c_str(), &viewer);
    MatView( this->matrix, viewer );
    PetscViewerFlush( viewer );
    PetscViewerDestroy( viewer );
  }


  /// set a given value to an entry in the matrix
  void PETSCSparseMatrix::setMatrixElement( const unsigned int row,
					    const unsigned int column,
					    const double val )
  {
    MatSetValue( this->matrix, (PetscInt) row, (PetscInt) column, (PetscScalar) val, INSERT_VALUES );

    MatAssemblyBegin( this->matrix, MAT_FLUSH_ASSEMBLY );
    MatAssemblyEnd( this->matrix, MAT_FLUSH_ASSEMBLY );
    this->hasMatrixChanged = true;
  }


  /// set a given value to an entry in the matrix
  void PETSCSparseMatrix::setSymmetricMatrixElement( const unsigned int row,
						     const unsigned int column,
						     const double val )
  {
    MatSetValue( this->matrix, (PetscInt) row, (PetscInt) column, (PetscScalar) val, INSERT_VALUES );

    if ( row != column )
      MatSetValue( this->matrix, (PetscInt) column, (PetscInt) row, (PetscScalar) val, INSERT_VALUES );

    MatAssemblyBegin( this->matrix, MAT_FLUSH_ASSEMBLY );
    MatAssemblyEnd( this->matrix, MAT_FLUSH_ASSEMBLY );
    this->hasMatrixChanged = true;
  }


  /// get a given value to an entry in the matrix
  double PETSCSparseMatrix::getMatrixElement( const unsigned int row,
					      const unsigned int column ) const
  {
    PetscScalar v[1];

    PetscInt r[1], c[1];
    r[0] = row;
    c[0] = column;

    MatGetValues( this->matrix, (PetscInt) 1, r, (PetscInt) 1, c, v);

    return v[0];
  }


  /// add matrices 
  void PETSCSparseMatrix::matrixAddMatrix( const GenericSparseMatrix *m )
  {
    // Will only be efficient if one has the SAME_NONZERO_PATTERN or SUBSET_NONZERO_PATTERN
    MatAXPY( this->matrix, (PetscScalar) 1.0, ((PETSCSparseMatrix*)(m))->matrix, DIFFERENT_NONZERO_PATTERN );
    this->hasMatrixChanged = true;
  }


  /// multiply all matrix elements with a scalar  
  void PETSCSparseMatrix::scalarMultiplyMatrix(const double s){
    MatScale( this->matrix, (PetscScalar) s );
    this->hasMatrixChanged = true;
  }

	
  /// Assemble matrix : the matrix can no longer be modified
  void PETSCSparseMatrix::assembleMatrix()
  {
    MatAssemblyBegin( this->matrix, MAT_FINAL_ASSEMBLY );
    MatAssemblyEnd( this->matrix, MAT_FINAL_ASSEMBLY );
  }

  /// Solve a linear system of equations given the sovler type
  void PETSCSparseMatrix::solveIterativeMethod( GenericSparseVector* rightHandSide,
						GenericSparseVector* result,
						const unsigned int maximumNumberIterations,
						const double tolerance,
						double &residual )
  {

    // Who really assembles the matrix, the user or the solver?
    PetscTruth assembled;
    MatAssembled( this->matrix, &assembled );
    if ( assembled == PETSC_FALSE )
      this->assembleMatrix();

    if ( this->hasMatrixChanged )
    {
      if ( this->solver )
	KSPDestroy( this->solver );

      // Create KSP linear solver
      KSPCreate( PETSC_COMM_WORLD, &this->solver );
      KSPSetOperators( this->solver, this->matrix, this->matrix, DIFFERENT_NONZERO_PATTERN );

      // Set preconditioner type
      KSPGetPC( this->solver, &this->preconditioner );
      switch ( this->pt )
      {
      case NO_PRECONDITIONER:
	PCSetType( this->preconditioner, PCNONE );
	break;
      case CHOLESKY:
	PCSetType( this->preconditioner, PCCHOLESKY );
	break;
      case SSOR:
	PCSetType( this->preconditioner, PCSOR );
	break;
      case ILU:
	PCSetType( this->preconditioner, PCILU );
	PCFactorSetLevels( this->preconditioner, (PetscInt) 0 );
	break;
      case ILUT:
	PCSetType( this->preconditioner, PCILU );
	PCFactorSetUseDropTolerance( this->preconditioner, 0.001, 0.1, 40 );
	break;
      default:
	std::cerr << "[WRNING] No preconditioner defined\n";
	PCSetType( this->preconditioner, PCNONE );
	break;
      }

      // Set solver type
      switch ( this->st )
      {
      case GenericSparseMatrix::CONJUGATED_GRADIENT:
	KSPSetType( this->solver, KSPCG );
	break;
      case GenericSparseMatrix::BICONJUGATED_GRADIENT:
	KSPSetType( this->solver, KSPBICG );
	break;
      case GenericSparseMatrix::BICONJUGATED_GRADIENT_STABILIZED:
	KSPSetType( this->solver, KSPBCGS );
	break;
      case GenericSparseMatrix::GMRES:
	KSPSetType( this->solver, KSPGMRES );
	break;
      default:
	std::cerr << "[WARNING] Unknown solver type. Conjugated gradient selected\n";
	KSPSetType( solver, KSPCG );
	break;
      }

      KSPSetTolerances( this->solver, tolerance, 1.0e-50, 1.0e5, maximumNumberIterations );
      KSPSetFromOptions( this->solver );
      KSPSetUp( this->solver );

      this->hasMatrixChanged = false;
    }

    // Solve the sysem
    KSPSolve( this->solver, 
	      ((PETSCSparseVector*) rightHandSide)->vector,
	      ((PETSCSparseVector*) result)->vector );

    KSPGetResidualNorm( this->solver, (PetscReal*) &residual );
  }


} // namespace yav

#endif
