#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4243 )
#endif

#ifndef WIN32
#pragma implementation "linearAlgebra/Rotation3D.h"
#endif
#include "linearAlgebra/Rotation3D.h"

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {


#define TEST_CONSTRAINTS


// value of PI up to the machine precision.
// ----------------------------------------------------------------------
#include <math.h>
#include <basics/Architecture.h>

// const double M_PI = 4.*atan(1.0);
const double TH_TINY = 1e-5;
const double PI_TH_TINY = M_PI - TH_TINY;
const double TH_SMALL = 1e-1;







// Rotation Matrix <- Rotation Vector
// ----------------------------------------------------------------------
void RotMat3DToRotVec3D( const RotMat3D& rot, RotVec3D& r )
{
#ifdef TEST_CONSTRAINTS
  // Verify that it is a rotation matrix
  {
    float small_val;Xlinalg_Small_Val<float>(small_val);
    Matrix3x3<double> rtt, tmp, Id;
    rtt.Trp( rot);
    tmp.Mul( rot, rtt);
    Id.Id();
    float f=(float)DistLinf( Id, tmp);
    if( f >  small_val ) {
      CERR << "RotMat3DToRotVec3D : R.Rt <> Id : " << f << ">" << small_val << ENDL;CERR.flush();
      throw RotMat3D::ErrNonRotMat("RotMat3DToRotVec3D : R.Rt <> Id");
    }
  }
#endif

  // Compute the rotation vector
  Vec3<double> sin_r;
  double d, sq, fact;
  double c = 0.5 * (rot(0,0) + rot(1,1) + rot(2,2) - 1.0); // cos(theta)
  if ((d = ::fabs(c)) > 1.0)
    c /= d;
  double theta = acos(c);

  if ( ::fabs(theta) <= TH_TINY){
    // R close to identity : DL around theta = 0
    fact = 0.5 + theta*theta/12.0;
    r.setValue( ( rot(2,1) - rot(1,2) ) * fact ,
                ( rot(0,2) - rot(2,0) ) * fact ,
                ( rot(1,0) - rot(0,1) ) * fact );
    return;
  }
  else
    if ( ::fabs(theta - M_PI) <= TH_SMALL) {
      // R is close to a symetry : DL around theta = M_PI
      // set values with diag terms of n.nt
      for (int i = 0; i < 3; i++) {
        // approx order 0
        // c = (rot[i][i] - 1.0)*0.5;
        sq = 1.0+ (rot(i,i) - 1.0)/(1-c);
        // 0 <=  [ sq = (n[i])^2 ] <= 1
        if ( sq < 0.0) sq = 0.;
        if (sq > 1.) sq = 1.;
        r.setVal(i, ::sqrt(sq));
      }
      ((Vec3<double>&)r) *= theta/ r.Norm();
      // set signs with off-diag terms of n.nt
      if ( (rot(0,1) + rot(1,0) ) < 0.0)
        r.setVal(1, -r(1) );
      if ( (rot(0,2) + rot(2,0) ) < 0.0)
        r.setVal(2, -r(2) );

      // determine wether r = +/- theta*n
      sin_r(0) = ( rot(2,1) - rot(1,2) );
      sin_r(1) = ( rot(0,2) - rot(2,0) );
      sin_r(2) = ( rot(1,0) - rot(0,1) );
      // determine the most significant term
      Index k = 0;
      if ( ::fabs(sin_r(1)) > ::fabs(sin_r(k))) k = 1;
      if ( ::fabs(sin_r(2)) > ::fabs(sin_r(k))) k = 2;
      // choose the sign
      if ( sin_r(k) * r(k) < 0.0 )
        // They do not have the same signs : r = -v
        ((Vec3<double>&)r) *= -1;
      return;
    }
    else { // At last no numerical problem
      fact = 0.5 * theta / sin(theta);
      r.setVal(0, ( rot(2,1) - rot(1,2) ) * fact );
      r.setVal(1, ( rot(0,2) - rot(2,0) ) * fact );
      r.setVal(2, ( rot(1,0) - rot(0,1) ) * fact );
      return;
    }
}


// Rotation Vector -> Rotation Matrix
// ----------------------------------------------------------------------
void RotVec3DToRotMat3D( const RotVec3D& r, RotMat3D& rot) 
{
  double c, s, k1, k2;

  double theta2 = r.Norm2();
  double theta = ::sqrt( theta2 );
  if (theta > TH_TINY){
    c = cos(theta);
    s = sin(theta);
    k1 = s / theta;
    k2 = (1 - c) / theta2;
  }
  else {  // Taylor expension around theta = 0
    k2 = 1.0/2.0 - theta2/24.0;
    c = 1.0 - theta2*k2;
    k1 = 1.0 - theta2/6;
  }

  /* I + M*Mt */
  for (Index i = 0; i < 3; i++)
    for (Index j = 0; j <= i; j++){
      rot.setValue(i,j, k2 * r(i) * r(j) );
      if (i != j)
        rot.setValue(j,i, rot(i,j));
      else
        rot.setValue(i,i, rot(i,i) + c );
    }
  double aux = k1 * r(2);
  rot.setValue(0,1, rot(0,1) - aux );
  rot.setValue(1,0, rot(1,0) + aux );
  aux = k1 * r(1);
  rot.setValue(0,2, rot(0,2) + aux );
  rot.setValue(2,0, rot(2,0) - aux );
  aux = k1 * r(0);
  rot.setValue(1,2, rot(1,2) - aux );
  rot.setValue(2,1, rot(2,1) + aux );
}


// Rotation Vector -> Rotation Quaternion
// ----------------------------------------------------------------------
void RotVec3DToUnitQuat( const RotVec3D& r, UnitQuat& q) 
{
  double theta = r.Norm();

  if (theta > TH_TINY){
    q.RealPart() = cos(theta/2.0);
    double s = sin(theta/2.0)/theta;
    q.VectPart().Mul( s, r);
  }
  else { // Taylor expansion around theta = 0
    double a = 1.0-theta*theta/8.0;
    q.RealPart() = a;
    double s = 1.0/3.0 + a/6.0;
    q.VectPart().Mul( s, r);
  }

#ifdef TEST_CONSTRAINTS
  {
    double small_val;Xlinalg_Small_Val<double>(small_val);
    double theta = q.Norm();
    if ( ::fabs( theta - 1.0) > small_val )
      COUT << "Warning : non unit quaternion in RotVec3DToUnitQuat : ||q|| -1 = "
           << theta -1.0 << ENDL;
    // Ensure normalisation
    q *= 1.0/q.Norm();
  }
#endif

}


// Rotation Vector <- Rotation Quaternion
// ----------------------------------------------------------------------
void UnitQuatToRotVec3D( const UnitQuat& q, RotVec3D& r) 
{

#ifdef TEST_CONSTRAINTS
  {
    double small_val;Xlinalg_Small_Val<double>(small_val);
    double theta = q.Norm();
    if ( ::fabs( theta - 1.0) > small_val )
      COUT << "Warning : non unit quaternion in UnitQuatToRotVec3D ||q|| -1 = "
           << theta -1.0 << ENDL;
  }
#endif

  double f;

  if( fabs( q.RealPart() ) < TH_TINY) {
    // we are around theta = Pi
    // theta = 2 arccos(a) = pi - a(2+a^2/3 + 3a^4/20) + O(a^6)
    double a= q.RealPart(), a2 = a*a, a4 = a2*a2;
    double theta = a*(2.0 + a2/3.0+3/20*a4);
    if ( a>=0 )
      f = (M_PI - theta) / q.VectPart().Norm();
    else
      f =  -(M_PI + theta) / q.VectPart().Norm();
  }
  else {
    // compute  2*sgn(alpha)
    double eps;
    if ( q.RealPart() < 0.0)
      eps = -2.0;
    else
      eps = 2.0;
    
    double b = q.VectPart().Norm();
    if ( b> 1.0) b=1.;
    if ( b > TH_TINY)
      f = eps*asin(b)/b;
    else // DL of f= th/sin(theta/2) for theta = 0 at order 4
      f = eps*(1.0+b*b/6+3*b*b*b*b/40);
  }


  r.Mul( f, q.VectPart());
}

//  Rotation Quaternion -> Rotation Matrix
// ----------------------------------------------------------------------
void UnitQuatToRotMat3D( const UnitQuat& q, RotMat3D& rot) 
{
#ifdef TEST_CONSTRAINTS
    {
      double small_val;Xlinalg_Small_Val<double>(small_val);
      double theta = q.Norm();
      if ( ::fabs( theta - 1.0) > small_val )
        COUT << "Warning : non unit quaternion in  UnitQuatToRotMat3D||q|| -1 = "
             << theta -1.0 << ENDL;
    }
#endif

    //double nq = q.Norm();
    //if ( ::fabs(nq-1.) > TH_TINY)     // was (nq < TH_TINY) || (nq > 10.0)
    //  throw UnitQuat::ErrNonUnitQuat(" UnitQuatToRotMat3D");

    // We could correct but this is not const
    // if ( nq != 1.0) q *= 1.0/nq;

    Matrix<4,double> qt, qa, qq;
    TildeLeft( q, qt);
    Anti( q, qa);
    qq.Mul( qt,qa);
    for (Index i=0;i<3;i++)
      for(Index j=0;j<3;j++)
        rot.setValue(i,j, qq(i+1,j+1));
}







// **********************************************************************
// Operations specific to 3D rotations
// **********************************************************************


//  LSQRot : input matrix K, output the rotation matrix R maximizing
//              C = trace( R. K\trp ) (K and R are (n,n) matrices).
//           residue is the value at the maximum
//----------------------------------------------------------------------
double RotMat3D::LSQ(const Matrix3x3<double>& K) throw(NonUniqueOptimum,RotMat3D::ErrNonRotMat,ErrConvergence)
{
  Index i;
  double small_val;Xlinalg_Small_Val<double>(small_val);
  const Index dim=3;

  Vec3<double> w;
  Matrix3x3<double> U,V;

  try{ K.SVD(U, w, V); }
  catch(ErrConvergence& exc){
    exc.push("double RotMat3D::LSQ( Mat<3>& K )");
    throw exc;
  }

  V.transpose();

  this->Mul(U,V);  // R = U.Vtrp

  double s = this->Det();
  // If s = det(U.Vtrp) id not +1 or -1 : error
  if ( ::fabs( ::fabs(s) -1.) >  small_val*100. ){
    COUT << "det = " << s << "      |det|-1 = " << ::fabs(s) -1. << ENDL;
    throw RotMat3D::ErrNonRotMat("LSQRot : non unit determinant");
  }

  // If s = -1, we should multiply by s = -1 both the smallest
  // singular value of diag and the corresponding ligne in R
  if (s < 0.0){
    // find smallest singular value
    s = w(0);
    Index i_dmax = 0;
    for(i=1;i<dim;i++)
      if( w(i) < s ) {
        s = w(i);
        i_dmax = i;
      }
    // Multiply the sing. val and the corresponding row in V by -1
    w(i_dmax) *= -1.0;
    for(i=0;i<dim;i++) V(i_dmax,i) *= -1.0;
    // and recompute the rotation
    this->Mul( U, V);  // R = U.Vtrp
  }

  // Grep for the number of singular (0) values
  {
    // Find the max
    double smax = 0;
    for(i=0;i<dim;i++)
      if (w(i)>smax) smax = w(i);

    // counting small singular values
    int num = 0;
    if ( smax < small_val )
      num = dim;
    else
      for(i=0;i<dim;i++)
        if ( ( (double) w(i)/100. + smax) == ((double) smax)) num++;

    if( num > 1){
      CERR << "*** WARNING: Matrix has " << num << " null singular values" << ENDL;
      throw NonUniqueOptimum("LSQRot");
    }
  }

  // compute the residue
  {
    double residue =0.0;
    for(i=0;i<dim;i++)
      residue += w(i);
    return residue;
  }
}

// Intrinsic Distance between rotation matrices : angle( R1trp.R2)
// ----------------------------------------------------------------------
double RotMat3D::Norm() const throw()
{
  double tmp = (Trace() -1.0)*0.5;
  if (tmp > 1.) tmp = 1.;
  if (tmp < -1.) tmp = -1.;
  return acos( tmp);
}

double Dist( const RotMat3D& R1, const RotMat3D& R2) throw()
{
  RotMat3D R;
  R.Trp(R1);
  R.Mul( R, R2);
  return R.Norm();
}


void RotMat3D::rnd(void) throw()
{
  UnitQuat q;
  *this=q.Rnd();
}


// Intrinsic Distance between rotation vectors : angle( r1\inv o r2)
// ----------------------------------------------------------------------
double Dist( const RotVec3D& r1, const RotVec3D& r2) throw()
{
  UnitQuat q1(r1), q2(r2);
  QuatInv(q1, q1);
  QuatProd(q1,q2,q1);
  RotVec3D r(q1);
  return( r.Norm() );
}

void RotVec3D::rnd(void) throw()
{
  UnitQuat q;
  *this=q.Rnd();
}

void RotVec3D::comp( const RotVec3D& r1, const RotVec3D& r2)  throw()
{
  UnitQuat q1(r1), q2(r2), q;
  *this=q.Comp(q1,q2);
}


/// Replace vector in definition domain (where norm(v)<Pi)
// ----------------------------------------------------------------------
void RotVec3D::Domain() throw()
{
  // Previous setting was to keep the axis, but set the angle to Pi
  // if( Norm() > M_PI ) ((Vec3<double>&)*this) *= M_PI/Norm();

  double theta =  Vec3<double>::Norm();
  int k = - (int) floor( 0.5 + theta/(2.0*M_PI));
  if (k != 0) // Out of the domain : theta >= Pi
    Vec3<double>(*this) *=(1.0 + 2.0 * M_PI *((double) k) / theta);
}


// **********************************************************************
void RotVec3D::J_Rt_r(const Vec3<double>& x, Matrix3x3<double>& dRx_dr) const throw()
{
    register double theta, theta2, costheta, sintheta,f, g, df,dg;
    //    Vec3<double> rt;
    Matrix3x3<double> xtild,rtild,rtxt,tmp1,tmp2;
    int i,j;

    theta2 = Norm2();
    theta = ::sqrt(theta2);
    sintheta = sin(theta);
    costheta = cos(theta);
    if (theta > TH_TINY) {
        /* No numerical problem */
        f = sintheta / theta;
        g = (1 - costheta) / theta2;
        df = (costheta - f) / theta2;
        dg = (f - 2 * g) / theta2;
    } else { /* Close to Id : DL around theta = 0 */
        f = 1 - theta2/6.0;
        g = 0.5 - theta2/24.0;
        df = -1.0 / 3.0 + theta2/30.0;
        dg = (theta2/15.0 - 1.0 )/ 12.0;
    }

    Tilde(x,xtild);
    Tilde((const Vec3<double>&)*this,rtild);

    ((Matrix<3,double>&)rtxt).Mul(x.ColMat(),this->RowMat());
    theta=rtxt.Trace();
    for(i=0;i<3;i++)
        rtxt(i,i) -= theta;

    for(i=0;i<3;i++)
        for(j=0;j<3;j++) {
            theta = val[i]*val[j];
            tmp1(i,j) = -df*theta +g*rtild(i,j);
            tmp2(i,j) = -dg*theta;
        }
    for(i=0;i<3;i++){
        tmp1(i,i) -= f;
        tmp2(i,i) -= 2.0*g;
    }

    dRx_dr.Mul(xtild,tmp1);
    tmp1.Mul(rtxt,tmp2);
    dRx_dr.Add(dRx_dr,tmp1);
}






// Init from a Quat (Normalize it)
// ----------------------------------------------------------------------
UnitQuat::UnitQuat( const Vect<4> q) throw(ErrDivZero)
{
  double nn = q.Norm();
  double small_val;Xlinalg_Small_Val<double>(small_val);
  if( nn <  small_val )
    throw ErrDivZero("UnitQuat::UnitQuat( const Vect<4> )");
  this->Mul( 1./nn, q);
}

//  Uniform Random UnitQuat (i.e. 3D Rotation)
//----------------------------------------------------------------------
void UnitQuat::rnd(void) throw()
{
  double nn;
  do {
    nn =0.0;
    for(Index i=0; i<4;i++) {
      double tmp = val[i] = 2.0*( rand() / (RAND_MAX+1.0) )-1.0;
      nn += tmp*tmp;
    }
  }
  while  ((nn > 1.0) || (nn < TH_TINY));
  nn = ::sqrt(nn);
  *this *= 1./nn;
}

// Intrinsic Distance between UnitQuat : angle( q1, q2)
// ----------------------------------------------------------------------
double Dist( const UnitQuat& q1, const UnitQuat& q2) throw()
{
  double tmp = DotProd((Quaternion&) q1,(Quaternion&) q2)/q1.Norm()/q2.Norm();
  if (tmp > 1.) tmp = 1.;
  if (tmp < -1.) tmp = -1.;
  return acos( tmp );
};

// Intrinsic Distance between Rotation Quaternions: angle( q1, q2)*2 mod M_PI
// ----------------------------------------------------------------------
double DistRot( const UnitQuat& q1, const UnitQuat& q2) throw()
{
  double tmp = ::fabs(DotProd((Quaternion&) q1,(Quaternion&) q2)/q1.Norm()/q2.Norm());
  if (tmp > 1.) tmp = 1.;
  return 2.0*acos( tmp );
};


void QuatProd(const UnitQuat& q1, const UnitQuat& q2, UnitQuat& q) throw()
  {QuatProd( (Quaternion&) q1, (Quaternion&) q2,(Quaternion&) q);}

void QuatConj(const UnitQuat& q, UnitQuat& barq) throw()
  {QuatConj( (Quaternion&) q, (Quaternion&) barq);}

void QuatInv(const UnitQuat& q, UnitQuat& invq) throw()
  {QuatConj((Quaternion&) q, (Quaternion&) invq);}

void TildeLeft(const UnitQuat& q, Matrix<4,double>& Q) throw()
  {TildeLeft((Quaternion&) q,Q);}

void Anti(const UnitQuat& q, Matrix<4,double>& Q) throw()
  {Anti((Quaternion&) q,Q);}

void TildeRight(const UnitQuat& q, Matrix<4,double>& Q) throw()
  {TildeRight((Quaternion&) q, Q);}

} // namespace yav
