#ifndef GENERICSPARSEMATRIX_H
#define GENERICSPARSEMATRIX_H

#pragma interface "linearAlgebra/GenericSparseMatrix.h"
#include <basics/Math.h>
#include <basics/TclString.h>

namespace yav {

  class GenericSparseVector;

  class GenericSparseMatrix {
  public:
    /// the type of sparse matrix
    enum SparseMatrixType {
      SPARSE,
      SPARSE_SYMMETRIC
    };
    /// the nature of the preconditionner
    enum PreconditionerType {
      NO_PRECONDITIONER,
      CHOLESKY,
      SSOR,
      ILU,
      ILUT
    };
    /// the selected solver method
    enum SolverType {
      CONJUGATED_GRADIENT,
      BICONJUGATED_GRADIENT,
      BICONJUGATED_GRADIENT_STABILIZED,
      GMRES
    };
    /// void constructor 
    GenericSparseMatrix() {
    }
    /// void destructor
    virtual ~GenericSparseMatrix() {
    }
    /// initializes some static variables associated with the linear algebra library
    virtual void initializeLibrary()=0;

    /// set all matrix elements to zero
    virtual void resetMatrix()=0;

    /// save matrix in file
    virtual void saveInFile(const std::string filename) const=0;

    /// add a given value to an entry in the matrix
    virtual void addMatrixElement(const unsigned int row,
				  const  unsigned int column,const double val)=0;

    /// add a given value to an entry in the matrix and its transpose entry 
    virtual void addSymmetricMatrixElement(const unsigned int row,
					   const  unsigned int column,const double val)=0;

    /// set a given value to an entry in the matrix
    virtual void setMatrixElement(const unsigned int row,
				  const  unsigned int column,const double val)=0;

    /// set a given value to an entry in the matrix and its transpose entry 
    virtual void setSymmetricMatrixElement(const unsigned int row,
					   const  unsigned int column,const double val)=0;

    /// get a given value to an entry in the matrix
    virtual double getMatrixElement(const unsigned int row,
				    const  unsigned int column) const=0;

    /// multiply all matrix elements with a scalar  
    virtual void matrixAddMatrix(const GenericSparseMatrix *m)=0;

    /// add multiply all matrix elements with a scalar  
    virtual void scalarMultiplyMatrix(const double s)=0;

    /// assemble matrix : the matrix can no longer be modified 
    virtual void assembleMatrix()=0;

    /// build a preconditioner with the associated matrix
    virtual void setPreconditioner(const PreconditionerType pt)=0;

    /// set the type of the solver
    virtual void setSolverType( const SolverType st)=0;

    /** solve a linear system of equations given the solver type (iterative methods),
	the right hand term and the maximum number of iterations and a
	tolerance number to terminate the algorithm 
	@param residual the residual to the iterative method*/
    virtual void solveIterativeMethod(GenericSparseVector *rightHandSide,
				      GenericSparseVector *result,
				      const unsigned int maximumNumberIterations,
				      const double tolerance, double &residual)=0;

  };



  class GenericSparseVector {
  public:
    /// void constructor 
    GenericSparseVector() {
    }
    /// void destructor
    virtual ~GenericSparseVector() {
    }
    /// add a value to a vector
    virtual void addVectorElement(const unsigned int row,const  double value)=0; 

    /// set a value to a vector
    virtual void setVectorElement(const unsigned int row,const  double value)=0; 

    /// get vector value
    virtual double getVectorElement(const unsigned int row) const=0; 

    // get vector as an array of double
    virtual const double *getVectorAsArray()=0;

    // restore the array of double (free the memory).
    virtual void restoreVectorArray(double* array)=0;

    /// get vector size
    virtual unsigned int getSize() const=0; 

    /// set vector to zero
    virtual void resetVector()=0; 

    /// assemble vector: to be done after all calls to set/addVectorElement
    virtual void assembleVector()=0;

    /// save the vector in file
    virtual void saveInFile(const std::string filename) const=0;
  };
	

  /** Factory object that creates a sparse matrix of a given type
      @param smt the type of matrix
      @param sizeRows the number of rows
      @param sizeColumns the number of columns
      @return a generic sparse matrix of a given type. Return 0 if key do not match any known value */
  GenericSparseMatrix *createSparseMatrix( const GenericSparseMatrix::SparseMatrixType smt,
					   const unsigned int sizeRows, 
					   const unsigned int sizeColumns,
					   const unsigned int nbNonZeroDiagonal=0,
					   const unsigned int nbNonZeroOffDiagonal=0 );

  /** Factory object that creates a sparse vector of a given type
      @param sizeRow the number of rows
      @return a generic sparse vector of a given type. Return 0 if key do not match any known value */
  GenericSparseVector *createSparseVector( const unsigned int sizeRows );

  /** Initialize the linear algebra library 
   * @param argc pointer towards the number of arguments that have been passed
   *             through the command line
   * @param args arguments passed through the command line */
  void initializeLinearAlgebraLibrary( int* argc, char*** args );


  /** Finilize the linear algebra libray */
  void finalizeLinearAlgebraLibrary();

}
#endif
