// General comments : see Vec.h

#ifndef _HVEC4_H
#define _HVEC4_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Hvec4.h"
#endif
#include "linearAlgebra/Vec.h"
#include "linearAlgebra/Mat.h"
#include "linearAlgebra/Vec3.h"
#include "linearAlgebra/Matrix3x3.h"

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {

template <class T> class HVec4;
template <class T> class Matrix4x4;

/****************************************************************************/
/* HVec4 Specialisation                                                      */
/****************************************************************************/
template <class T> HVec4<T> operator+(const HVec4<T> &p, const HVec4<T> &q) throw();
template <class T> HVec4<T> operator-(const HVec4<T> &p, const HVec4<T> &q) throw();
template <class T> STD_OSTREAM& operator<<(STD_OSTREAM &s, const HVec4<T> &m) throw();
template <class T> HVec4<T> operator*(const _Mat<4,4,T> &m, const HVec4<T> &v) throw();
template <class T> HVec4<T> operator*(const HVec4<T> &v, const _Mat<4,4,T> &m) throw();
template <class T> inline HVec4<T> operator*(const Matrix<4,T> &m, const HVec4<T> &v) throw();
template <class T> inline HVec4<T> operator*(const HVec4<T> &v, const Matrix<4,T> &m) throw();

/*! \class HVec4 HVec4.h linearAlgebra/HVec4
  \ingroup liblinearAlgebra
  template for homogeneous vector of dimension 4 and type T
*/
template <class T> class HVec4 : protected Vect<4,T> {
public:
  /** @name Constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  inline HVec4() throw()
    : Vect<4,T>(VIM_NOINIT)
    {
#ifndef UNINIT_ZERO
    this->val[0] = this->val[1] = this->val[2] = (T) 0;
    this->val[3] = (T) 1;
#endif
    }
  /** constructor - initializes to 0 if b==VIM_INIT */
  inline HVec4(VectInitMode b) throw()
    : Vect<4,T>(VIM_NOINIT)
  {
    if (b==VIM_INIT) {
      this->val[0] = this->val[1] = this->val[2] = (T) 0;
      this->val[3] = (T) 1;
    }
  }

  /** constructor: initialize to vector (x,y)
      @param x first vector coordinate
      @param y second vector coordinate
      @param z third vector coordinate
      @param t
  */
  HVec4(T x, T y, T z, T t) throw();
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  HVec4(const T coord[3]) throw();

  /** constructor from 3D vector initializes to (x, y, z, 1)
      @param v 3D vector (x, y, z) */
  HVec4(const Vec3<T> &v) throw();

#ifndef __GNUC__
#ifndef WIN32
  HVec4(const HVec4<T> &v) throw()
    : Vect<4,T>(VIM_NOINIT)
    {copy(v);}
  HVec4<T>& operator=(const HVec4<T>& v)
    {copy(v);return *this;}
#endif
#endif

   /** copy constructor with cast
       @param v vector to copy */
  template<class T2> HVec4<T>(const HVec4<T2>& v) throw()
    : Vect<4,T>(VIM_NOINIT)
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> HVec4<T>& operator=(const HVec4<T2>& v) throw()
    {copy(v);return *this;}
  //@}


  /** @name Data members access */
  //@{
  /** copy with cast as a member function
      @param p vector to copy */
  template<class T2> void copy(const HVec4<T2>& p) throw()
    {this->val[0] = (T) p(0); this->val[1] = (T) p(1); this->val[2] = (T) p(2); this->val[3] = (T) p(3);}

  /** set this vector to vector (x,y,z)
      @param x first vector coordinate
      @param y second vector coordinate
      @param z third vector coordinate
      @param t
  */
  void setValue(T x, T y, T z, T t) throw();

  /// returns X coordinate
  T getX(void) const throw();

  /// returns Y coordinate
  T getY(void) const throw();

  /// returns Z coordinate
  T getZ(void) const throw();

  /// returns T coordinate
  T getT(void) const throw();

  // functions inherited from vector that must not be used
  /** returns ith component of this matrix considered as a vector
      @param i component number (0<= i < D) */
//   T& operator()(int i) throw(ErrRange)
//     {return Vect<4,T>::operator()(i);}

  /** returns ith component of this considered as a vector (const version)
      @param i component number (0<= i < D) */
  T operator()(int i) const throw(ErrRange)
    {return Vect<4,T>::operator()(i);}

  /** acces to element i (non normalized)
      usefull for matrix4x4 element acces via operator [][]
      @param i element number */
  T& operator[](int i) throw(ErrRange)
    {return Vect<4,T>::operator[](i);}

  /// cast an homogenous vector to a 3D vector
  operator Vec3<T>() const throw();
  //@}

  /** @name mathematical operators */
  //@{
  /** vector sum (this is not modified)
      @param p vector to add
      @param q
      @return this + v */
  friend HVec4<T> operator+ FRIEND_TEMPLATE_ARG(const HVec4<T> &p, const HVec4<T> &q) throw();

  /** vector sum (this is modified)
      @param v vector to add */
  HVec4<T>& operator+=(const HVec4<T>& v) throw();

  /** vector substraction (this is not modified)
      @param p vector to add
      @param q vector to add
      @return this - v */
  friend HVec4<T> operator- FRIEND_TEMPLATE_ARG(const HVec4<T> &p, const HVec4<T> &q) throw();

  /** vector substraction (this is modified)
      @param v vector to substract */
  HVec4<T>& operator-=(const HVec4<T>& v) throw();

  /// normalize homogeneous coordinates so that v[3] == 1.0
  void normalize(void) throw();
  //@}

  /** @name Matrix operations */
  //@{
  HVec4<T>& Mul(const Matrix<4,T> &m, const HVec4<T> &v) throw();
  HVec4<T>& Mul(const HVec4<T> &v, const Matrix<4,T> &m) throw();
  friend HVec4<T> operator* FRIEND_TEMPLATE_ARG(const Matrix<4,T> &m, const HVec4<T> &v) throw();
  friend HVec4<T> operator* FRIEND_TEMPLATE_ARG(const HVec4<T> &v, const Matrix<4,T> &m) throw();
  //@}

  /** @name I/O operations */
  //@{
//   /** writes vector on outstream as "(x, y, z, t)"
//       param s outstream
//       param c vector
//       return s */
//   // friend STD_OSTREAM& operator<<  FRIEND_TEMPLATE_ARG(STD_OSTREAM &s, const HVec4<T> &m) throw();
  //@}

  static void Test() throw();
};

} // namespace yav


/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

namespace yav {

/****************************************************************************/
/* HVec4 Specialisation                                                      */
/****************************************************************************/
// constructors
template <class T> inline HVec4<T>::HVec4(T x, T y, T z, T t) throw()
  : Vect<4,T>(VIM_NOINIT)
{
  this->val[0]=x; this->val[1]=y; this->val[2]=z;
  this->val[3]=t;
}
template <class T> inline HVec4<T>::HVec4(const T coord[3]) throw()
  : Vect<4,T>(VIM_NOINIT)
{
  this->val[0]=coord[0]; this->val[1]=coord[1]; this->val[2]=coord[2]; this->val[3]=(T) 1;
}

template <class T> inline HVec4<T>::HVec4(const Vec3<T> &p) throw()
  : Vect<4,T>(VIM_NOINIT)
{
    this->val[0] = p(0); this->val[1] = p(1); this->val[2] = p(2);
    this->val[3] = (T) 1;
}

template <class T> inline void HVec4<T>::setValue(T x, T y, T z, T t) throw()
{
    this->val[0]=x; this->val[1]=y; this->val[2]=z;
    this->val[3]=t;
}
template <class T> inline T HVec4<T>::getX(void) const throw() {return this->val[0];}
template <class T> inline T HVec4<T>::getY(void) const throw() {return this->val[1];}
template <class T> inline T HVec4<T>::getZ(void) const throw() {return this->val[2];}
template <class T> inline T HVec4<T>::getT(void) const throw() {return this->val[3];}

template <class T> inline HVec4<T>::operator Vec3<T>() const throw()
{
  return Vec3<T>(this->val[0]/this->val[3], this->val[1]/this->val[3], this->val[2]/this->val[3]);
}

template <class T> inline Vec3<T>::operator HVec4<T>() throw()
{
  return HVec4<T>(this->val[0], this->val[1], this->val[2], 1);
}


// normalize homogeneous coordinates so that v[3] == 1.0
template <class T> inline void HVec4<T>::normalize(void) throw() {
  if(this->val[3] != (T) 1) {
    this->val[0] /= this->val[3];
    this->val[1] /= this->val[3];
    this->val[2] /= this->val[3];
    this->val[3] = (T) 1;
  }
}


// vectorial + operator
template <class T> inline HVec4<T> operator+(const HVec4<T> &p, const HVec4<T> &q) throw()
{
  return HVec4<T>(p.val[0] / p.val[3] + q.val[0] / q.val[3],
                  p.val[1] / p.val[3] + q.val[1] / q.val[3],
                  p.val[2] / p.val[3] + q.val[2] / q.val[3],
                  (T) 1);
}

// vectorial += operator
template <class T> inline HVec4<T>& HVec4<T>::operator+=(const HVec4<T> &v) throw()
{
  normalize();
  this->val[0] += v.val[0] / v.val[3];
  this->val[1] += v.val[1] / v.val[3];
  this->val[2] += v.val[2] / v.val[3];
  return *this;
}

// vectorial - operator
template <class T> inline HVec4<T> operator-(const HVec4<T> &p, const HVec4<T> &q) throw()
{
  return HVec4<T>(p.val[0] / p.val[3] - q.val[0] / q.val[3],
                  p.val[1] / p.val[3] - q.val[1] / q.val[3],
                  p.val[2] / p.val[3] - q.val[2] / q.val[3],
                  (T) 1);
}

// vectorial -= operator
template <class T> inline HVec4<T>& HVec4<T>::operator-=(const HVec4<T> &v) throw()
{
  normalize();
  this->val[0] -= v.val[0] / v.val[3];
  this->val[1] -= v.val[1] / v.val[3];
  this->val[2] -= v.val[2] / v.val[3];
  return *this;
}


template <class T>  inline HVec4<T>& HVec4<T>::Mul(const Matrix<4,T> &m, const HVec4<T> &v) throw()
{((Mat<4,1,T>*)this)->Mul(m,(const Mat<4,1,T>&)v);return *this;}
template <class T>  inline HVec4<T>& HVec4<T>::Mul(const HVec4<T> &v, const Matrix<4,T> &m) throw()
{((Mat<1,4,T>*)this)->Mul((const Mat<1,4,T>&)v,m);return *this;}
template <class T> inline HVec4<T> operator*(const Matrix<4,T> &m, const HVec4<T> &v) throw()
{return HVec4<T>(VIM_NOINIT).Mul(m,v);}
template <class T> inline HVec4<T> operator*(const HVec4<T> &v, const Matrix<4,T> &m) throw()
{return HVec4<T>(VIM_NOINIT).Mul(v,m);}


template <class T> STD_OSTREAM& operator<<(STD_OSTREAM &s, const HVec4<T> &m) throw()
{
  s<<"(";
  for (Index i=0; i<4; i++) {
    s<<m(i)<<" ";
  }
  s<<")";
  return s;
}


template <class T> void HVec4<T>::Test() throw()
{
  T A[3]={1,2,3};
  Vec3<T>aa(2,3,4);
  HVec4<float> df;HVec4<double> dd;HVec4<int> di;

  HVec4<T> a(A),b(aa),c(VIM_INIT),d(a),e(5,6,7,8),ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "HVec4<" << TName << ">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1, 2, 3, 1 : " << a << ENDL;
  COUT << "b must be 2, 3, 4, 1 : " << b << ENDL;
  COUT << "c must be 0, 0, 0, 1 : " << c << ENDL;
  COUT << "d must be as a    : " << d << ENDL;
  COUT << "e must be 5, 6, 7, 8 : " << e << ENDL;
  COUT << "ef must be as 0 : " << ef << ENDL;
  COUT << "ed must be as 0 : " << ed << ENDL;
  COUT << "ei must be as 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << df << ENDL;
  d=dd;COUT << "d=dd : " << dd << ENDL;
  d=di;COUT << "d=di : " << di << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  COUT << "d[1] : " << d[1] << ENDL;
  COUT << "d(1) : " << d(1) << ENDL;
  d.setValue(2,3,4,5);COUT << "d.setValue(2,3,4,5); : " << d << ENDL;
  COUT << "d.get.() : " << d.getX() << " " << d.getY() << " " << d.getZ() << " " << d.getT() << ENDL;
  // d.Null();COUT << "d.Null() : " << d << ENDL;
//   d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
//   T*D=d.getValues();
//   COUT << "getValues() : " << D[0] << " " << D[1] << " " << D[2] << ENDL;
//   const T*E=d;
//   COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
//   COUT << "a.ColMat : " << a.ColMat() << ENDL;
//   COUT << "a.RowMat : " << a.RowMat() << ENDL;
  //  COUT << "operator Vec3<T> d : " << (Vec3<T>) d << ENDL;

  COUT << "Math" << ENDL;
  // COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
//   COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
//   COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
//   COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
//   COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
//   COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
//   COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

//   COUT << "+a : " << +a << ENDL;
//   COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  // COUT << "a*2 : " << a*((T)2) << ENDL;
//   COUT << "2*a : " << ((T)2)*a << ENDL;
//   COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  e.normalize();COUT << "e.normalize() : " << e << ENDL;
//   d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
//   d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

//   d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
//   d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

//   COUT << "dot(a,b) : " << dot(a,b) << ENDL;
//   COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
//   COUT << "norm2(a) : " << a.norm2() << ENDL;
//   COUT << "norm(a) : " << a.norm() << ENDL;
//   COUT << "Norm2(a) : " << a.Norm2() << ENDL;
//   COUT << "Norm(a) : " << a.Norm() << ENDL;
//   COUT << "NormL1(a) : " << a.NormL1() << ENDL;
//   COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
//   COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
//   COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
//   COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
//   COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

//   d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
//   d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

//   d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;

  COUT << ENDL;
}


} // namespace yav


#endif
