#ifndef MTLSPARSEMATRIX_H
#define MTLSPARSEMATRIX_H

#ifdef USE_MTL

#pragma interface "linearAlgebra/MTLSparseMatrix.h"
#include <linearAlgebra/GenericSparseMatrix.h>
#include "mtl/mtl.h"
#include <mtl/matrix.h>
#include <mtl/utils.h>



namespace yav {

  typedef mtl::matrix<double, mtl::rectangle<>, mtl::compressed<>, mtl::row_major>::type MTLMatrix;
  typedef mtl::matrix<double, mtl::symmetric<mtl::upper>, mtl::array< mtl::compressed<int,mtl::internal,mtl::index_from_zero> >, mtl::row_major>::type MTLSymmetricMatrix;


  class MTLSparseVector : public GenericSparseVector{
  protected:
    mtl::dense1D<double> *vector;
  public:
    /// void constructor 
    MTLSparseVector(const unsigned int nbRows); 

    /// void destructor
    virtual ~MTLSparseVector();

    /// add a value to a vector
    virtual void addVectorElement(const unsigned int row,const  double value); 

    /// set a value to a vector
    virtual void setVectorElement(const unsigned int row,const  double value); 

    /// set vector to zero
    virtual void resetVector(); 
 
    /// assemble vector: to be done after all calls to set/addVectorElement
    virtual void assembleVector() {};

    /// get vector value
    virtual double getVectorElement(const unsigned int row) const; 

    // get vector as an array of double
    virtual const double *getVectorAsArray();

    // restore the array of double (free the memory).
    virtual void restoreVectorArray( double* array ) {};


    /// get vector size
    virtual unsigned int getSize() const; 

    /// save vector in file
    virtual void saveInFile(const std::string filename) const;

    friend class MTLSparseMatrix;
    friend class MTLSymmetricSparseMatrix;

  };

  class MTLSparseMatrix : public GenericSparseMatrix{
  protected:
    MTLMatrix  *matrix;
    PreconditionerType pt;
    SolverType st;
  public:

    /** Create a matrix of a given number of rows and columns */ 
    MTLSparseMatrix(const unsigned int sizeRows, 
		    const unsigned int sizeColumns);

    /// void destructor
    virtual ~MTLSparseMatrix();
    /// initializes some static variables associated with the linear algebra library
    virtual void initializeLibrary(){}

    /// set all matrix elements to zero
    virtual void resetMatrix();

    /// save matrix in file
    virtual void saveInFile(const std::string filename) const;

    /// add a given value to an entry in the matrix
    virtual void addMatrixElement(const unsigned int row,
				  const  unsigned int column,const double val);

    /// add a given value to an entry in the matrix and its transpose entry 
    virtual void addSymmetricMatrixElement(const unsigned int row,
					   const  unsigned int column,const double val);

    /// set a given value to an entry in the matrix
    virtual void setMatrixElement(const unsigned int row,
				  const  unsigned int column,const double val);

    /// set a given value to an entry in the matrix and its transpose entry 
    virtual void setSymmetricMatrixElement(const unsigned int row,
					   const  unsigned int column,const double val);

    /// get a given value to an entry in the matrix
    virtual double getMatrixElement(const unsigned int row,
				    const  unsigned int column) const;

    /// multiply all matrix elements with a scalar  
    virtual void matrixAddMatrix(const GenericSparseMatrix *m);

    /// add multiply all matrix elements with a scalar  
    virtual void scalarMultiplyMatrix(const double s);

    /// Assemble matrix : the matrix can no longer be modified 
    virtual void assembleMatrix(){}

    /// build a preconditioner with the associated matrix
    virtual void setPreconditioner(const PreconditionerType _pt){
      pt=_pt;
    }

    /// set solver type
    virtual void setSolverType( const SolverType _st )
      {
	st=_st;
      };

    /** solve a linear system of equations given the solver type (iterative methods),
	the right hand term and the maximum number of iterations and a
	tolerance number to terminate the algorithm */
    virtual void solveIterativeMethod(GenericSparseVector *rightHandSide,
				      GenericSparseVector *result,
				      const unsigned int maximumNumberIterations,
				      const double tolerance, double &residual);
  };

  
  class MTLSymmetricSparseMatrix : public GenericSparseMatrix{

  protected:
    PreconditionerType pt;
    MTLSymmetricMatrix  *matrix;
    SolverType st;
  public:

    /** Create a matrix of a given number of rows and columns */ 
    MTLSymmetricSparseMatrix(const unsigned int sizeRows, 
			     const unsigned int sizeColumns);

    /// void destructor
    virtual ~MTLSymmetricSparseMatrix();
    /// initializes some static variables associated with the linear algebra library
    virtual void initializeLibrary(){}

    /// set all matrix elements to zero
    virtual void resetMatrix();
		
    /// save matrix in file
    virtual void saveInFile(const std::string filename) const;

    /// add a given value to an entry in the matrix
    virtual void addMatrixElement(const unsigned int row,
				  const  unsigned int column,const double val);

    /// add a given value to an entry in the matrix and its transpose entry 
    virtual void addSymmetricMatrixElement(const unsigned int row,
					   const  unsigned int column,const double val) {
      addMatrixElement(row,column,val);
    }

    /// set a given value to an entry in the matrix
    virtual void setMatrixElement(const unsigned int row,
				  const  unsigned int column,const double val);
		
    /// set a given value to an entry in the matrix and its transpose entry 
    virtual void setSymmetricMatrixElement(const unsigned int row,
					   const  unsigned int column,const double val) {
      setMatrixElement(row,column,val);
    }

    /// get a given value to an entry in the matrix
    virtual double getMatrixElement(const unsigned int row,
				    const  unsigned int column) const;

    /// multiply all matrix elements with a scalar  
    virtual void matrixAddMatrix(const GenericSparseMatrix *m);

    /// add multiply all matrix elements with a scalar  
    virtual void scalarMultiplyMatrix(const double s);

    /// Assemble matrix : the matrix can no longer be modified 
    virtual void assembleMatrix(){}

    /// build a preconditioner with the associated matrix
    virtual void setPreconditioner(const PreconditionerType _pt){
      pt=_pt;
    }

    /// set solver type
    virtual void setSolverType( const SolverType _st )
      {
	st=_st;
      };

    /** solve a linear system of equations given the solver type (iterative methods),
	the right hand term and the maximum number of iterations and a
	tolerance number to terminate the algorithm */
    virtual void solveIterativeMethod(GenericSparseVector *rightHandSide,
				      GenericSparseVector *result,
				      const unsigned int maximumNumberIterations,
				      const double tolerance, double &residual);
  };
}

#endif

#endif

