
// General comments : see Vec.h

#ifndef _MAT_H
#define _MAT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4290)
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Mat.h"
#endif
#include "linearAlgebra/Vec.h"



namespace yav {


template <Index L, Index C, class T> class _Mat;

template <Index L, Index C, class T> STD_OSTREAM& operator<<(STD_OSTREAM &s, const _Mat<L,C,T> &c) throw();
template <Index L, Index C, class T> STD_ISTREAM& operator>>(STD_ISTREAM &s, _Mat<L,C,T> &c) throw();

/*! \class _Mat Mat.h linearAlgebra/Mat.h
  \ingroup liblinearAlgebra
  base class for template for a matrix of dimension LxC and type T
*/
template <Index L, Index C, class T> class _Mat
{
protected:
  /** @name Constructors */
  //@{
  //  /** @name Vectorial redefined constructors */
  //  //@{
  //  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  //  vec_Inline _Mat() throw() : _Vect<L*C,T>()
  //    {}
  //  /** constructor - initializes to 0 if b==VIM_INIT */
  //  explicit vec_Inline _Mat(VectInitMode b) throw() : _Vect<L*C,T>(b)
  //    {}
  //  //@}
  /// Default constructor : initializes to 0 if UNINIT_ZERO not defined
  vec_Inline _Mat() throw()
    {
#ifndef UNINIT_ZERO
      resetTo(0);
#endif
    }
  /// Constructor : initializes to 0 if b==VIM_INIT
  vec_Inline _Mat(VectInitMode b) throw()
    {if (b==VIM_INIT) resetTo(0);}

#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline _Mat(const _Mat<L,C, T>& v) throw()
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline _Mat<L,C, T>& operator=(const _Mat<L,C, T>& v) throw()
    {copy(v);return *this;}
#endif

  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline _Mat(const _Mat<L,C, T2>& v) throw()
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline _Mat<L,C, T>& operator=(const _Mat<L,C, T2>& v) throw()
    {copy(v);return *this;}
  //@}

public:
  vec_Inline T getVal(int i) const
    {return val[i];}
  vec_Inline void setVal(int i, T v)
    {val[i]=v;}

  /** @name Data members access */
  //@{
  //   /** copy as a member function
  //       @param v vector to copy */
  //  void copy(const _Mat<L,C,T>& v) throw()
  //     {for (Index i=0; i<L*C; i++) val[i] = v.getVal(i);}
  /** copy with cast as a member function
      @param v vector to copy */
  template<class T2> vec_Inline void copy(const _Mat<L,C,T2>& v) throw()
    {for (Index i=0; i<L*C; i++) val[i] = (T) v.getVal(i);}

  /** set each element of this vector to value
    @param value elements new value */
  void resetTo(T value) throw()
    {for (Index i=0; i<L*C; i++) val[i]=value;}
  /// set this vector to the null vector
  vec_Inline void  resetToNull(void) throw()
    {resetTo(0);}

  //   /** setValues with a T[L*C] array */
  //   vec_Inline void setValues(const T v[L*C]) throw()
  //     { for (Index i=0; i<L*C; i++) val[i]=v[i]; }
  /** setValues with a T2[D] array */
  //   template<class T2> vec_Inline void setValues(const T2 v[L*C]) throw()
  //     {for (Index i=0; i<L*C; i++) val[i]=(T) v[i];}

  /** returns a T* (usefull for some extern functions, like OGL ones) */
  vec_Inline T* getValues(void) throw()
    {return val;}
  /** returns a const T* (usefull for some extern functions, like OGL ones) */
  vec_Inline const T* getValues(void) const throw()
    {return val;}

  /** cast vector in a T*
   */
  vec_Inline operator const T*() const throw()
    {return (const T*) val;}

// #ifndef WIN32
  /** setValues with a T[L][C] array */
  vec_Inline void setValues(const T v[L][C]) throw()
    {for (Index i=0; i<L; i++) for (Index j=0; j<C; j++) val[i*C+j]=v[i][j];}
// #endif
//   /** setValues with a T2[D] array */
//   template<class T2> vec_Inline void setValues(const T2 v[L][C]) throw()
//     {for (Index i=0; i<L; i++) for (Index j=0; j<C; j++) val[i*C+j]=(T) v[i][j];}

  /** set one element value
      @param i element line
      @param j element column
      @param value element new value */
  void setValue(int i, int j, T value) throw(ErrRange);

  /** access a matrix element
      @param i element line
      @param j element column
      @return this[i][j] */
  T getValue(int i, int j) const throw(ErrRange);

  /** access ithline vector
      @param i line number
      @return line vector */
  Vect<C,T> operator[](unsigned int i) const throw(ErrRange);

  /** access a matrix element
      @param i element line
      @param j element column
      @return this[i][j] */
  T& operator()(int i, int j) throw(ErrRange);
  /** access a matrix element (const version)
      @param i element line
      @param j element column
      @return this[i][j] */
  T operator()(int i, int j) const throw(ErrRange);
  //@}

  /** @name mathematical operators */
  //@{
  /** squared norm
      @return ||this||^2 */
  vec_Inline T norm2(void) const throw()
    {return vec_norm2(val,L*C);}
  /** norm
    @return ||this|| */
  vec_Inline T norm(void) const throw()
    {return vec_norm(val,L*C);}
  /** L1 norm
      @return sum abs(v_i) */
  vec_Inline T normL1(void) const throw()
    {return vec_normL1(val,L*C);}
  /** Linf norm
    @return max abs(v_i) */
  vec_Inline T normLinf(void) const throw()
    {return vec_normLinf(val,L*C);}
  /** squared norm
      @return ||this||^2 */
  vec_Inline T Norm2(void) const throw()
    {return vec_norm2(val,L*C);}
  /** norm
    @return ||this|| */
  vec_Inline T Norm(void) const throw()
    {return vec_norm(val,L*C);}
  /** L1 norm
      @return sum abs(v_i) */
  vec_Inline T NormL1(void) const throw()
    {return vec_normL1(val,L*C);}
  /** Linf norm
    @return max abs(v_i) */
  vec_Inline T NormLinf(void) const throw()
    {return vec_normLinf(val,L*C);}
  
  /** add q to this vector */
  vec_Inline void add(const _Mat<L,C, T> &q) throw()
    {vec_add(val,L*C,q.val);}
  /** substract q to this vector */
  vec_Inline void sub(const _Mat<L,C, T> &q) throw()
    {vec_sub(val,L*C,q.val);}
  /** set each coordinate of this vector to product of this and q same coordinate */
  vec_Inline void cmul(const _Mat<L,C, T> &q) throw()
    {vec_cmul(val,L*C,q.val);}
  /** multiply this vector with a */
  vec_Inline void mul(const T a) throw()
    {vec_mul(val,L*C,a);}
  /** divide this vector with a */
  vec_Inline void div(const T a) throw(ErrDivZero)
    {vec_div(val,L*C,a);}
  
  /** set this vector to -p */
  vec_Inline void opposite(const _Mat<L,C, T> &p) throw()
    {vec_opposite(val,L*C,p.val);}
  /** set this vector to p+q */
  vec_Inline void add(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {vec_add(val,L*C,p.val,q.val);}
  /** set this vector to p-q */
  vec_Inline void sub(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {vec_sub(val,L*C,p.val,q.val);}
  /** set each coordinate of this vector to product of p and q same coordinate */
  vec_Inline void cmul(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {vec_cmul(val,L*C,p.val,q.val);}
  /** set this vector to p*a */
  vec_Inline void mul(const _Mat<L,C, T> &p, const T a) throw()
    {vec_mul(val,L*C,p.val,a);}
  /** set this vector to p/a */
  void div(const _Mat<L,C, T> &p, const T a) throw(ErrDivZero)
    {vec_div(val,L*C,p.val,a);}
  
  /** randomize vector */
  vec_Inline void rnd(T min = (T)-1., T max = (T)1.) throw()
    {vec_rnd(val,L*C,min,max);}
  /** randomize vector : isotropic gaussian centred noise */
  vec_Inline void rndgaussian(T standard_deviation = 1.) throw()
    {vec_rndgaussian(val,L*C,standard_deviation);}
  
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates */
  vec_Inline void min(const _Mat<L,C, T> &p) throw()
    {vec_min(val,L*C,p.val);}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates */
  vec_Inline void max(const _Mat<L,C, T> &p) throw()
    {vec_max(val,L*C,p.val);}
  
  /** dot product
      @param p first vector
      @param q second vector
      @return p.q */
  vec_Inline static T dot(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {return vec_dot(L*C,p.val,q.val);}
  /// Square Euclidian Distance
  vec_Inline static T dist2( const _Mat<L,C, T>& p, const _Mat<L,C, T>& q) throw()
    {return vec_dist2(L*C,p.val,q.val);}
  /// Euclidian Distance
  vec_Inline static T dist( const _Mat<L,C, T>& p, const _Mat<L,C, T>& q) throw()
    {return vec_dist(L*C,p.val,q.val);}
  /// L1 Distance
  vec_Inline static T distL1( const _Mat<L,C, T>& p, const _Mat<L,C, T>& q) throw()
    {return vec_distL1(L*C,p.val,q.val);}
  /// Linf Distance
  vec_Inline static T distLinf( const _Mat<L,C, T>& p, const _Mat<L,C, T>& q) throw()
    {return vec_distLinf(L*C,p.val,q.val);}
  
  /// operator ==
  vec_Inline static bool isEqual(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {return vec_isEqual(L*C,p.val,q.val);}
  /// operator !=
  vec_Inline static bool isDifferent(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {return vec_isDifferent(L*C,p.val,q.val);}
  /// operator <
  vec_Inline static bool isSmaller(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {return vec_isSmaller(L*C,p.val,q.val);}
  /// operator >
  vec_Inline static bool isGreater(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {return vec_isSmaller(L*C,q.val,p.val);}
  /// operator <=
  vec_Inline static bool isSmallerOrEqual(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {return !vec_isSmaller(L*C,q.val,p.val);}
  /// operator >=
  vec_Inline static bool isGreaterOrEqual(const _Mat<L,C, T> &p, const _Mat<L,C, T> &q) throw()
    {return !vec_isSmaller(L*C,p.val,q.val);}


  // TODO : vec_Inlines pour les fonctions qui suivent ?????????????
  /** matrix product (check if src and dest doesn't match before multiplication)
      @param m1 first matrix
      @param m2 second matrix
      @return m1m2 */
  template <Index K> void mulmat(const _Mat<L,K,T> &m1, const _Mat<K,C,T> &m2) throw()
  {
    if ( ((_Mat<L,C,T>*) (&m1) == this) || ((const _Mat<L,C,T>*) (&m2) == this) ) {
      _Mat<L,C,T> tmp(VIM_NOINIT);
      tmp.private_mulmat(m1,m2);
      this->copy(tmp);
    } else {
      this->private_mulmat(m1,m2);
    }
  }

  /** matrix/sym product
      @param m1 symetric matrix
      @param m2 second matrix
      @return m1m2 */
  void mulmatsym(const Sym<L,T> &m1, const _Mat<L,C,T> &m2) throw()
  {
    if ( ((_Mat<L,C,T>*) (&m2) == this) ) {
      _Mat<L,C,T> tmp(VIM_NOINIT);
      tmp.private_mulmatsym(m1,m2);
      this->copy(tmp);
    } else {
      this->private_mulmatsym(m1,m2);
    }
  }
  /** matrix/sym product
      @param m1 symetric matrix
      @param m2 second matrix
      @return m1m2 */
  void mulmatsym(const _Mat<L,C,T> &m1, const Sym<C,T> &m2) throw()
  {
    if ( ((_Mat<L,C,T>*) (&m1) == this) ) {
      _Mat<L,C,T> tmp(VIM_NOINIT);
      tmp.private_mulmatsym(m1,m2);
      this->copy(tmp);
    } else {
      this->private_mulmatsym(m1,m2);
    }
  }

  /** matrix transpose
      @param m matrix
      @return this = m^t */
  void trp(const _Mat<C,L,T> &m) throw();

  /** Action of a matrix on a symmetric matrix: Wout = this.Win.Trp(this)
      @param Win in matrix
      @param Wout out matrix */
  void symaction(const Sym<C,T>& Win, Sym<L, T>& Wout) const throw();

  /** Singular Value Decomposition : this = U . Diag(W) . Vtrp
      @param U LxC matrix
      @param W singular values vector
      @param V CxC matrix */
  void svd(Mat<L,C,T>& U, Vect<C,T>& W, Mat<C,C,T>& V) const throw(ErrConvergence,ErrUnknown);
  void svd2(Mat<L,C,T>& U, Vect<C,T>& W, Mat<C,C,T>& V) const throw(ErrConvergence,ErrUnknown);

  /** Action of a matrix on a symmetric matrix: Wout = this.Win.Trp(this)
      @param Win in matrix
      @param Wout out matrix */
  void SymAction(const Sym<C,T>& Win, Sym<L, T>& Wout) const throw();

  /** Singular Value Decomposition : this = U . Diag(W) . Vtrp
      @param U LxC matrix
      @param W singular values vector
      @param V CxC matrix */
  void SVD(Mat<L,C,T>& U, Vect<C,T>& W, Mat<C,C,T>& V) const throw(ErrConvergence,ErrUnknown);

  /** Set Matrix to Pseudo Inverse of matrix A using SVD
      @param A mtraix to psseudo-invert
      @return number of singular values */
  int PseudoInv( const Mat<C,L,T>& A) throw(ErrConvergence);
  //@}

  /** @name I/O operations */
  //@{
  /** writes matrix on outstream as "((x00 x01 x02) (x10 x11 x12) (x20 x21 x22))"
      @param s outstream
      @param m matrix
      @return s */
  //friend STD_OSTREAM& operator<< FRIEND_TEMPLATE_ARG(STD_OSTREAM &s, const _Mat<L,C,T> &m) throw();
  /** reads matrix from instream as "x00 x01 x02 x10 x11 x12 x20 x21 x22"
      @param s instream
      @param m matrix
      @return s */
  //friend STD_ISTREAM& operator>> FRIEND_TEMPLATE_ARG(STD_ISTREAM &s, _Mat<L,C,T> &m) throw();
  //@}

private:
  // TODO : vec_Inlines pour les fonctions qui suivent ?????????????
  /** matrix internal product (doesn't check if src and dest match)
      @param a first matrix
      @param b second matrix
      @return ab */
  template <Index K> void private_mulmat(const _Mat<L,K,T> &m1, const _Mat<K,C,T> &m2) throw()
  {
    T *res = val;
    for(register const T *m1i = m1.getValues(); m1i < m1.getValues() + L*K; m1i += K)
      for (register const T *m2j = m2.getValues(); m2j < m2.getValues() + C; m2j++) {
        register T cur_val = 0. ;
        for(register const T *m2kj = m2j,*m1ik = m1i; m1ik < m1i+K ; m2kj += C, m1ik++)
          cur_val += (*m1ik) * (*m2kj);
        *res++ = cur_val;
      }
  }


  /** matrix/sym internal product (doesn't check if src and dest match)
      @param a symetric matrix
      @param b second matrix
      @return ab */
  void private_mulmatsym(const Sym<L,T> &m1, const _Mat<L,C,T> &m2) throw()
  {
    T *res = val;
    const T *m1i; Index i;
    for(m1i = m1.getValues(), i=0; i < L; i++, m1i+=i) {
      const T *m2j; Index j;
      for (m2j = m2.getValues(), j=0; j < C; m2j++, j++) {
        T cur_val = 0. ;
        const T *m2kj,*m1ik; Index k;
        for(m2kj = m2j,m1ik = m1i, k=0; k<i ; m2kj += C, m1ik++, k++)
          cur_val += (*m1ik) * (*m2kj);
        for(; k<L ; m2kj += C, k++, m1ik+=k)
          cur_val += (*m1ik) * (*m2kj);
        *res = cur_val;
        res++;
      }
    }
  }
  /** matrix/sym internal product (doesn't check if src and dest match)
      @param a symetric matrix
      @param b second matrix
      @return ab */
  void private_mulmatsym(const _Mat<L,C,T> &m1, const Sym<C,T> &m2) throw()
  {
    T *res = val;
    const T *m1i; Index i;
    for(m1i = m1.getValues(), i=0; i < L; m1i+=C, i++) {
      const T *m2j; Index j;
      for (m2j = m2.getValues(), j=0; j < C; j++, m2j+=j) {
        T cur_val = 0.;
        const T *m2kj,*m1ik; Index k;
        for(m2kj = m2j, m1ik = m1i, k=0; k<j ; m2kj++, m1ik++, k++)
          cur_val += (*m1ik) * (*m2kj);
        for(; k<C ; m1ik+=1, k++, m2kj+=k)
          cur_val += (*m1ik) * (*m2kj);
        *res++ = cur_val;
      }
    }
  }

  /** matrix internal transpose (check if src and dest doesn't match)
      @param m matrix
      @return this = m^t */
  void private_trp(const _Mat<C,L,T> &m) throw();

  /** Internal action of a matrix on a symmetric matrix: Wout = this.Win.Trp(this)
      @param Win in matrix
      @param Wout out matrix */
  void private_symaction(const Sym<C,T>& Win, Sym<L, T>& Wout) const throw();

protected:
  /// vector data
  T val[L*C];
};






/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template <Index L, Index C, class T> vec_Inline Mat<L,C, T> operator+(const Mat<L,C, T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template <Index L, Index C, class T> vec_Inline Mat<L,C, T> operator-(const Mat<L,C, T> &p) throw()
  {return Mat<L,C, T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template <Index L, Index C, class T> vec_Inline Mat<L,C, T> operator+(const Mat<L,C, T> &p, const Mat <L,C, T> &q) throw()
  {return Mat<L,C, T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template <Index L, Index C, class T> vec_Inline Mat<L,C, T> operator-(const Mat<L,C, T> &p, const Mat <L,C, T> &q) throw()
  {return Mat<L,C, T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template <Index L, Index C, class T> vec_Inline Mat<L,C, T> operator*(const Mat<L,C, T> &p, const T a) throw()
  {return Mat<L,C, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template <Index L, Index C, class T> vec_Inline Mat<L,C, T> operator*(const T a, const Mat<L,C, T> &p) throw()
  {return Mat<L,C, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template <Index L, Index C, class T> vec_Inline Mat<L,C, T> operator/(const Mat<L,C, T> &p, const T a) throw(ErrDivZero)
  {return Mat<L,C, T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template <Index L, Index C, class T> vec_Inline T dot(const Mat<L,C, T>& u, const Mat<L,C, T> &v) throw()
  {return Mat<L,C, T>::dot(u,v);}
/// Square Euclidian Distance
template <Index L, Index C, class T> vec_Inline T dist2( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::dist2(u,v);}
/// Euclidian Distance
template <Index L, Index C, class T> vec_Inline T dist( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::dist(u,v);}
/// L1 Distance
template <Index L, Index C, class T> vec_Inline T distL1( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::distL1(u,v);}
/// Linf Distance
template <Index L, Index C, class T> vec_Inline T distLinf( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template <Index L, Index C, class T> vec_Inline T DotProd(const Mat<L,C, T> &p, const Mat<L,C, T> &q) throw()
  {return Mat<L,C, T>::dot(p,q);}
/// Square Euclidian Distance
template <Index L, Index C, class T> vec_Inline T Dist2( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::dist2(u,v);}
/// Euclidian Distance
template <Index L, Index C, class T> vec_Inline T Dist( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::dist(u,v);}
/// L1 Distance
template <Index L, Index C, class T> vec_Inline T DistL1( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::distL1(u,v);}
/// Linf Distance
template <Index L, Index C, class T> vec_Inline T DistLinf( const Mat<L,C, T>& u, const Mat<L,C, T>& v) throw()
  {return Mat<L,C, T>::distLinf(u,v);}
/** operator == */
template <Index L, Index C, class T> vec_Inline bool operator == (const Mat<L,C, T>& u, const Mat<L,C, T> &v) throw()
  {return Mat<L,C, T>::isEqual(u,v);}
/** operator != */
template <Index L, Index C, class T> vec_Inline bool operator != (const Mat<L,C, T>& u, const Mat<L,C, T> &v) throw()
  {return Mat<L,C, T>::isDifferent(u,v);}
/** operator < */
template <Index L, Index C, class T> vec_Inline bool operator < (const Mat<L,C, T>& u, const Mat<L,C, T> &v) throw()
  {return Mat<L,C, T>::isSmaller(u,v);}
/** operator > */
template <Index L, Index C, class T> vec_Inline bool operator > (const Mat<L,C, T>& u, const Mat<L,C, T> &v) throw()
  {return Mat<L,C, T>::isGreater(u,v);}
/** operator <= */
template <Index L, Index C, class T> vec_Inline bool operator <= (const Mat<L,C, T>& u, const Mat<L,C, T> &v) throw()
  {return Mat<L,C, T>::isSmallerOrEqual(u,v);}
/** operator >= */
template <Index L, Index C, class T> vec_Inline bool operator >= (const Mat<L,C, T>& u, const Mat<L,C, T> &v) throw()
  {return Mat<L,C, T>::isGreaterOrEqual(u,v);}
//@}

#ifndef WIN32
/** matrix product (check if src and dest doesn't match before multiplication)
    @param m1 first matrix
    @param m2 second matrix
    @return m1m2 */
template <Index L, Index K, Index C, class T> Mat<L,C,T> operator*(const Mat<L,K,T>& m1, const Mat<K,C,T>& m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
#else
/** matrix product (check if src and dest doesn't match before multiplication)
    @param a first matrix
    @param b second matrix
    @return ab */
template <Index L, Index C, class T> Mat<L,C,T> operator*(const Mat<L,1,T>& m1, const Mat<1,C,T>& m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
/** matrix product (check if src and dest doesn't match before multiplication)
    @param a first matrix
    @param b second matrix
    @return ab */
template <Index L, Index C, class T> Mat<L,C,T> operator*(const Mat<L,2,T>& m1, const Mat<2,C,T>& m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
/** matrix product (check if src and dest doesn't match before multiplication)
    @param a first matrix
    @param b second matrix
    @return ab */
template <Index L, Index C, class T> Mat<L,C,T> operator*(const Mat<L,3,T>& m1, const Mat<3,C,T>& m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
/** matrix product (check if src and dest doesn't match before multiplication)
    @param a first matrix
    @param b second matrix
    @return ab */
template <Index L, Index C, class T> Mat<L,C,T> operator*(const Mat<L,4,T>& m1, const Mat<4,C,T>& m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
/** matrix product (check if src and dest doesn't match before multiplication)
    @param a first matrix
    @param b second matrix
    @return ab */
template <Index L, Index C, class T> Mat<L,C,T> operator*(const Mat<L,5,T>& m1, const Mat<5,C,T>& m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
/** matrix product (check if src and dest doesn't match before multiplication)
    @param a first matrix
    @param b second matrix
    @return ab */
template <Index L, Index C, class T> Mat<L,C,T> operator*(const Mat<L,6,T>& m1, const Mat<6,C,T>& m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
#endif

/** matrix/sym product (check if src and dest doesn't match before multiplication)
    @param m1 first matrix
    @param m2 second matrix
    @return m1m2 */
template <Index L, Index C, class T>
Mat<L,C,T>& operator*(const Sym<L,T> &m1, const _Mat<L,C,T> &m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}
/** matrix/sym product (check if src and dest doesn't match before multiplication)
    @param m1 first matrix
    @param m2 second matrix
    @return m1m2 */
template <Index L, Index C, class T>
Mat<L,C,T>& operator*(const _Mat<L,C,T> &m1, const Sym<C,T> &m2) throw()
  {return Mat<L,C,T>(VIM_NOINIT).Mul(m1,m2);}

/*! \class Mat Mat.h linearAlgebra/Mat.h
  \ingroup liblinearAlgebra
  template class for a matrix of dimension LxC and type T
*/
template <Index L, Index C, class T> class Mat : public _Mat<L,C,T>
{
public:
  /** @name Constructors */
  //@{
#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline Mat(const Mat<L,C, T>& v) throw() : _Mat<L,C,T>(VIM_NOINIT)
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Mat<L,C, T>& operator=(const Mat<L,C, T>& v) throw()
    {copy(v);return *this;}
#endif

  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Mat(const Mat<L,C, T2>& v) throw() : _Mat<L,C,T>(VIM_NOINIT)
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Mat<L,C, T>& operator=(const Mat<L,C, T2>& v) throw()
    {copy(v);return *this;}

  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Mat() throw() : _Mat<L,C,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Mat(VectInitMode b) throw() : _Mat<L,C,T>(b)
    {}
  //@}

// #ifndef WIN32
  /** constructor: initialize to matrix
      @param matrix matrix coordinates array */
  Mat(const T matrix[L][C]) throw() : _Mat<L,C,T>(VIM_NOINIT)
    {setValues(matrix);}
// #endif

//   /** constructor: initialize to matrix
//        @param coord matrix coordinates array */
//    template<class T2> Mat(const T2 matrix[L][C]) throw() : _Mat<L,C,T>(VIM_NOINIT)
//     {setValues(matrix);}

  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Mat<L,C, T>& Minus(const Mat<L,C, T> &p) throw()
    {vec_opposite(this->val,L*C,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Mat<L,C, T>& Add(const Mat<L,C, T> &p, const Mat<L,C, T> &q) throw()
    {vec_add(this->val,L*C,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Mat<L,C, T>& Sub(const Mat<L,C, T> &p, const Mat<L,C, T> &q) throw()
    {vec_sub(this->val,L*C,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Mat<L,C, T>& CMul(const Mat<L,C, T> &q) throw()
    {vec_cmul(this->val,L*C,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Mat<L,C, T>& CMul(const Mat<L,C, T> &p, const Mat<L,C, T> &q) throw()
    {vec_cmul(this->val,L*C,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Mat<L,C, T>& Mul(const Mat<L,C, T> &p, const T a) throw()
    {vec_mul(this->val,L*C,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Mat<L,C, T>& Mul(const T a, const Mat<L,C, T> &p) throw()
    {vec_mul(this->val,L*C,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Mat<L,C, T>& Div(const Mat<L,C, T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,L*C,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Mat<L,C, T>& operator+=(const Mat<L,C, T> &q) throw()
    {vec_add(this->val,L*C,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Mat<L,C, T>& operator-=(const Mat<L,C, T> &q) throw()
    {vec_sub(this->val,L*C,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Mat<L,C, T>& operator*=(const T a) throw()
    {vec_mul(this->val,L*C,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Mat<L,C, T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,L*C,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Mat<L,C, T>& Rnd(T min = (T)-1., T max = (T)1.) throw()
    {vec_rnd(this->val,L*C,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Mat<L,C, T>& RndGaussian(T standard_deviation = (T)1.) throw()
    {vec_rndgaussian(this->val,L*C,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<L*C,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Mat<L,C, T>& Min(const Mat<L,C, T>& p) throw()
    {vec_min(this->val,L*C,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<L*C,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Mat<L,C, T>& Max(const Mat<L,C, T>& p) throw()
    {vec_max(this->val,L*C,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Mat<L,C, T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Mat<L,C, T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Mat<L,C, T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** @name matrix operations */
  //@{
  // TODO : vec_Inlines pour les fonctions qui suivent ?????????????
  /** matrix product (check if src and dest doesn't match before multiplication)
      @param m1 first matrix
      @param m2 second matrix
      @return m1m2 */
  template <Index K> Mat<L,C,T>& Mul(const _Mat<L,K,T> &m1, const _Mat<K,C,T> &m2) throw()
    {mulmat(m1,m2);return *this;}

  /** matrix/sym product (check if src and dest doesn't match before multiplication)
      @param m1 first matrix
      @param m2 second matrix
      @return m1m2 */
  Mat<L,C,T>& Mul(const Sym<L,T> &m1, const _Mat<L,C,T> &m2) throw()
    {mulmatsym(m1,m2);return *this;}
  /** matrix/sym product (check if src and dest doesn't match before multiplication)
      @param m1 first matrix
      @param m2 second matrix
      @return m1m2 */
  Mat<L,C,T>& Mul(const _Mat<L,C,T> &m1, const Sym<C,T> &m2) throw()
    {mulmatsym(m1,m2);return *this;}

  /** TODO */
  Mat<L,C,T>& operator*=(const _Mat<C,C,T>& m) throw()
    {mulmat(*this,m);return *this;}
  
  /** TODO */
  Mat<L,C,T>& operator*=(const Sym<C,T>& m) throw()
    {mulmatsym(*this,m);return *this;}

  /** TODO */
  Mat<L,C,T>& Trp(const _Mat<C,L,T>& m) throw()
    {trp(m);return *this;}

  /** TODO */
  Mat<C,L,T> getTransposed(void) const throw()
    {return Mat<C,L,T>(VIM_NOINIT).Trp(*this);}
  //@}

  static void Test() throw();
};













/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template <Index D, class T> vec_Inline Matrix<D, T> operator+(const Matrix<D, T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template <Index D, class T> vec_Inline Matrix<D, T> operator-(const Matrix<D, T> &p) throw()
  {return Matrix<D, T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template <Index D, class T> vec_Inline Matrix<D, T> operator+(const Matrix<D, T> &p, const Matrix <D, T> &q) throw()
  {return Matrix<D, T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template <Index D, class T> vec_Inline Matrix<D, T> operator-(const Matrix<D, T> &p, const Matrix <D, T> &q) throw()
  {return Matrix<D, T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template <Index D, class T> vec_Inline Matrix<D, T> operator*(const Matrix<D, T> &p, const T a) throw()
  {return Matrix<D, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template <Index D, class T> vec_Inline Matrix<D, T> operator*(const T a, const Matrix<D, T> &p) throw()
  {return Matrix<D, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template <Index D, class T> vec_Inline Matrix<D, T> operator/(const Matrix<D, T> &p, const T a) throw(ErrDivZero)
  {return Matrix<D, T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template <Index D, class T> vec_Inline T dot(const Matrix<D, T>& u, const Matrix<D, T> &v) throw()
  {return Matrix<D, T>::dot(u,v);}
/// Square Euclidian Distance
template <Index D, class T> vec_Inline T dist2( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::dist2(u,v);}
/// Euclidian Distance
template <Index D, class T> vec_Inline T dist( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::dist(u,v);}
/// L1 Distance
template <Index D, class T> vec_Inline T distL1( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::distL1(u,v);}
/// Linf Distance
template <Index D, class T> vec_Inline T distLinf( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template <Index D, class T> vec_Inline T DotProd(const Matrix<D, T> &p, const Matrix<D, T> &q) throw()
  {return Matrix<D, T>::dot(p,q);}
/// Square Euclidian Distance
template <Index D, class T> vec_Inline T Dist2( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::dist2(u,v);}
/// Euclidian Distance
template <Index D, class T> vec_Inline T Dist( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::dist(u,v);}
/// L1 Distance
template <Index D, class T> vec_Inline T DistL1( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::distL1(u,v);}
/// Linf Distance
template <Index D, class T> vec_Inline T DistLinf( const Matrix<D, T>& u, const Matrix<D, T>& v) throw()
  {return Matrix<D, T>::distLinf(u,v);}
/** operator == */
template <Index D, class T> vec_Inline bool operator == (const Matrix<D, T>& u, const Matrix<D, T> &v) throw()
  {return Matrix<D, T>::isEqual(u,v);}
/** operator != */
template <Index D, class T> vec_Inline bool operator != (const Matrix<D, T>& u, const Matrix<D, T> &v) throw()
  {return Matrix<D, T>::isDifferent(u,v);}
/** operator < */
template <Index D, class T> vec_Inline bool operator < (const Matrix<D, T>& u, const Matrix<D, T> &v) throw()
  {return Matrix<D, T>::isSmaller(u,v);}
/** operator > */
template <Index D, class T> vec_Inline bool operator > (const Matrix<D, T>& u, const Matrix<D, T> &v) throw()
  {return Matrix<D, T>::isGreater(u,v);}
/** operator <= */
template <Index D, class T> vec_Inline bool operator <= (const Matrix<D, T>& u, const Matrix<D, T> &v) throw()
  {return Matrix<D, T>::isSmallerOrEqual(u,v);}
/** operator >= */
template <Index D, class T> vec_Inline bool operator >= (const Matrix<D, T>& u, const Matrix<D, T> &v) throw()
  {return Matrix<D, T>::isGreaterOrEqual(u,v);}
//@}
/** @name Square matrix redefined friend functions */
//@{
template <Index D, class T> vec_Inline Matrix<D, T> operator*(const Matrix<D, T> &m1, const Matrix<D, T> &m2) throw()
  {return Matrix<D, T>(VIM_NOINIT).Mul(m1,m2);}
//@}

template <Index D, class T> vec_Inline Matrix<D, T> operator*(const Sym<D, T> &m1, const Matrix<D, T> &m2) throw()
  {return Matrix<D, T>(VIM_NOINIT).Mul(m1,m2);}
template <Index D, class T> vec_Inline Matrix<D, T> operator*(const Matrix<D, T> &m1, const Sym<D, T> &m2) throw()
  {return Matrix<D, T>(VIM_NOINIT).Mul(m1,m2);}

template<Index D, class T> void Sym2Mat( const Sym<D,T>& W, Mat<D,D,T>& M) throw();
template<Index D, class T> void Mat2Sym( const Mat<D,D,T>& M, Sym<D,T>& W) throw();

/*! \class Matrix Mat.h linearAlgebra/Mat.h
  \ingroup liblinearAlgebra
  template for a square matrix of dimension D and type T
*/
template <Index D, class T> class Matrix : public Mat<D,D,T> {
public:
  /** @name Constructors */
  //@{
  /** @name Vectorial redefined constructors */
  //@{
  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  vec_Inline Matrix() throw() : Mat<D,D,T>()
    {}
  /** constructor - initializes to 0 if b==VIM_INIT */
  explicit vec_Inline Matrix(VectInitMode b) throw() : Mat<D,D,T>(b)
    {}
  //@}
  /** @name Square matrix redefined constructors */
  //@{
  /// Set to diag( v )
  explicit vec_Inline Matrix(const Vect<D,T>& v) throw() : Mat<D,D,T>(VIM_NOINIT)
    {diag(v);}
  
#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline Matrix(const Mat<D,D,T>& v) throw() : Mat<D,D,T>(VIM_NOINIT)
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Matrix<D, T>& operator=(const Mat<D,D,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Matrix(const Mat<D,D,T2>& v) throw() : Mat<D,D,T>(VIM_NOINIT)
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Matrix<D, T>& operator=(const Mat<D,D,T2>& v) throw()
    {copy(v);return *this;}
  
#ifndef WIN32
  /// Constructor : set to matrix coord
  vec_Inline Matrix(const T coord[D][D]) throw() : Mat<D,D,T>(VIM_NOINIT)
    {setValues(coord);}
#endif
  /// Constructor : set to matrix coord
  template <class T2> vec_Inline Matrix(const T2 coord[D][D]) throw() : Mat<D,D,T>(VIM_NOINIT)
    {setValues(coord);}
  
  /** Creates a square matrix from a symetric matrix
      @param W symetric matrix */
  explicit vec_Inline Matrix(const Sym<D,T>& W) throw() : Mat<D,D,T>(VIM_NOINIT)
    {Sym2Mat<D,T>(W,*this);}
  /** Set from a symetric matrix
      @param W symetric matrix */
  vec_Inline Matrix<D, T>& operator=(const Sym<D,T>& W) throw()
    {Sym2Mat<D,T>(W,*this);return *this;}
  //@}

  /** @name Data members access */
  //@{
  /// set this matrix to the identity matrix
  void resetToIdentity(const T v=1.0) throw();

  /// creates a diagonal matrix using a vector for diagonal terms
  void diag(const Vect<D,T>& v) throw();

  /// set this matrix to the tensor product of v1 and v2
  void tensorProduct(const Vect<D,T>& v1, const Vect<D,T>& v2) throw();
  //@}


  /** @name mathematical operators */
  //@{
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix<D, T>& Minus(const Matrix<D, T> &p) throw()
    {vec_opposite(this->val,D*D,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Matrix<D, T>& Add(const Matrix<D, T> &p, const Matrix<D, T> &q) throw()
    {vec_add(this->val,D*D,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Matrix<D, T>& Sub(const Matrix<D, T> &p, const Matrix<D, T> &q) throw()
    {vec_sub(this->val,D*D,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Matrix<D, T>& CMul(const Matrix<D, T> &q) throw()
    {vec_cmul(this->val,D*D,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Matrix<D, T>& CMul(const Matrix<D, T> &p, const Matrix<D, T> &q) throw()
    {vec_cmul(this->val,D*D,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix<D, T>& Mul(const Matrix<D, T> &p, const T a) throw()
    {vec_mul(this->val,D*D,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Matrix<D, T>& Mul(const T a, const Matrix<D, T> &p) throw()
    {vec_mul(this->val,D*D,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Matrix<D, T>& Div(const Matrix<D, T> &p, const T a) throw(ErrDivZero)
    {vec_div(this->val,D*D,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Matrix<D, T>& operator+=(const Matrix<D, T> &q) throw()
    {vec_add(this->val,D*D,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Matrix<D, T>& operator-=(const Matrix<D, T> &q) throw()
    {vec_sub(this->val,D*D,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Matrix<D, T>& operator*=(const T a) throw()
    {vec_mul(this->val,D*D,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Matrix<D, T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(this->val,D*D,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Matrix<D, T>& Rnd(T min = (T)-1., T max = (T)1.) throw()
    {vec_rnd(this->val,D*D,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Matrix<D, T>& RndGaussian(T standard_deviation = (T)1.) throw()
    {vec_rndgaussian(this->val,D*D,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<D*D,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Matrix<D, T>& Min(const Matrix<D, T>& p) throw()
    {vec_min(this->val,D*D,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<D*D,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Matrix<D, T>& Max(const Matrix<D, T>& p) throw()
    {vec_max(this->val,D*D,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Matrix<D, T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Matrix<D, T>& Null() throw()
  {resetTo((T)0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Matrix<D, T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** set this matrix to inverse of a given matrix
    @return this */
  void inv(const Matrix<D,T> &m) throw(ErrMatSingular);
#ifdef TODO
invert rend matrice nulle en cas de singularite
#endif

  /** computes the determninant
      @return det matrix determinant */
  T det(void) const throw();

  /** computes the trace
      @return trace matrix trace */
  T trace(void) const throw();

  /** @name Square matrix redefined menber functions */
  //@{
  /** set this matrix to inverse of a given matrix
      @return this */
  vec_Inline Matrix<D, T>& Inv(const Matrix<D, T>& m) throw(ErrMatSingular)
    {this->inv(m);return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  vec_Inline Matrix<D, T> getInverse(void) const throw(ErrMatSingular)
    {return Matrix<D, T>(VIM_NOINIT).Inv(*this);}
  /// invert this matrix (this matrix is modified)
  vec_Inline void invert(void) throw(ErrMatSingular)
    {this->Inv(*this);}
  
  /** set this matrix to transposed of a given matrix
      @return this */
  vec_Inline Matrix<D, T>& Trp(const Matrix<D, T>& m) throw()
    {this->trp(m);return *this;}
  /** get this matrix transposed matrix (this matrix is NOT modifyed)
      @return transpose matrix */
  vec_Inline Matrix<D, T> getTransposed(void) const throw()
    {return Matrix<D, T>(VIM_NOINIT).Trp(*this);}
  /// transpose this matrix (this matrix is modified)
  vec_Inline void transpose(void) throw()
    {this->Trp(*this);}
  
  /// Matrix multiplication
  vec_Inline Matrix<D, T>& operator*=(const Matrix<D, T>& m) throw()
    {mulmat(*this,m);return *this;}
  /// Matrix multiplication
  Matrix<D, T>& Mul(const Matrix<D, T> &m1, const Matrix<D, T> &m2) throw()
    {mulmat(m1,m2);return *this;}
  
  /** computes the determninant
      @return det matrix determinant */
  vec_Inline T Det(void) const throw()
    {return this->det();}
  /** computes the trace
      @return trace matrix trace */
  vec_Inline T Trace(void) const throw()
    {return this->trace();}
  
  /// set this matrix to the identity matrix
  vec_Inline Matrix<D, T>& Id(const T v=1.0) throw()
    {this->resetToIdentity(v);return *this;}
  /// Set to diag( v )
  vec_Inline Matrix<D, T>& Diag(const Vect<D,T>& v) throw()
    {this->diag(v);return *this;}
  /// Set to tensor product v1*v2
  vec_Inline Matrix<D, T>& TensorProduct(const Vect<D,T>& v1,const Vect<D,T>& v2) throw()
    {this->tensorProduct(v1,v2);return *this;}
  //@}

//#ifndef WIN32
  // TODO : vec_Inlines pour les fonctions qui suivent ?????????????
  /** matrix product (check if src and dest doesn't match before multiplication)
      @param m1 first matrix
      @param m2 second matrix
      @return m1m2 */
  template <Index K> Matrix<D,T>& Mul(const _Mat<D,K,T> &m1, const _Mat<K,D,T> &m2) throw()
    {mulmat(m1,m2);return *this;}
//#else
/*   /\** matrix product (check if src and dest doesn't match before multiplication) */
/*       @param a first matrix */
/*       @param b second matrix */
/*       @return ab *\/ */
/*    Matrix<D,T>& Mul(const _Mat<D,1,T> &m1, const _Mat<1,D,T> &m2) throw() */
/*     {mulmat(m1,m2);return *this;} */
/*   /\** matrix product (check if src and dest doesn't match before multiplication) */
/*       @param a first matrix */
/*       @param b second matrix */
/*       @return ab *\/ */
/*    Matrix<D,T>& Mul(const _Mat<D,2,T> &m1, const _Mat<2,D,T> &m2) throw() */
/*     {mulmat(m1,m2);return *this;} */
/*   /\** matrix product (check if src and dest doesn't match before multiplication) */
/*       @param a first matrix */
/*       @param b second matrix */
/*       @return ab *\/ */
/*    Matrix<D,T>& Mul(const _Mat<D,3,T> &m1, const _Mat<3,D,T> &m2) throw() */
/*     {mulmat(m1,m2);return *this;} */
/*   /\** matrix product (check if src and dest doesn't match before multiplication) */
/*       @param a first matrix */
/*       @param b second matrix */
/*       @return ab *\/ */
/*    Matrix<D,T>& Mul(const _Mat<D,4,T> &m1, const _Mat<4,D,T> &m2) throw() */
/*     {mulmat(m1,m2);return *this;} */
/*   /\** matrix product (check if src and dest doesn't match before multiplication) */
/*       @param a first matrix */
/*       @param b second matrix */
/*       @return ab *\/ */
/*    Matrix<D,T>& Mul(const _Mat<D,5,T> &m1, const _Mat<5,D,T> &m2) throw() */
/*     {mulmat(m1,m2);return *this;} */
/*   /\** matrix product (check if src and dest doesn't match before multiplication) */
/*       @param a first matrix */
/*       @param b second matrix */
/*       @return ab *\/ */
/*    Matrix<D,T>& Mul(const _Mat<D,6,T> &m1, const _Mat<6,D,T> &m2) throw() */
/*     {mulmat(m1,m2);return *this;} */
/* #endif */
  /** matrix/sym product (check if src and dest doesn't match before multiplication)
      @param m1 first matrix
      @param m2 second matrix
      @return m1m2 */
  Matrix<D,T>& Mul(const Sym<D,T> &m1, const _Mat<D,D,T> &m2) throw()
    {mulmatsym(m1,m2);return *this;}
  /** matrix/sym product (check if src and dest doesn't match before multiplication)
      @param m1 first matrix
      @param m2 second matrix
      @return m1m2 */
  Matrix<D,T>& Mul(const _Mat<D,D,T> &m1, const Sym<D,T> &m2) throw()
    {mulmatsym(m1,m2);return *this;}


  /** eigenvalues extraction (only for SYMETRIC Matrices... see Sym::Diagonalise... implemented for compatibility with old Matrixnxn classes)
      @param eigenVecttors eigen vectors returned
      @param eigenValues eigen values returned
      @return true if matrix could be diagonalized, false otherwise */
  bool computeEigenV(Matrix<D,T> &eigenVecttors, Vect<D,T> &eigenValues) throw();

  friend class Sym<D,T>;

  static void Test() throw();
private:
  /// used by Jacobi method (diagonalisation of symetric matrices)
  void ROTATE(Index i, Index j, Index k, Index l, T s, T tau) throw();
};


} // namespace yav


/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

#include "linearAlgebra/Sym.h"

namespace yav {

// access to data
template <Index L, Index C, class T> vec_Inline void _Mat<L,C,T>::setValue(int i, int j, T value) throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (j<0) || (i >= (int)L) || (j >= (int)C)  )
    throw  ErrRange("Mat::setValue(Index,Index)");
#endif
  val[i*C+j]=value;
}

template <Index L, Index C, class T> vec_Inline T _Mat<L,C,T>::getValue(int i, int j) const throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (j<0) || (i >= (int)L) || (j >= (int)C) )
    throw  ErrRange("Mat::getValue(Index,Index)");
#endif
  return val[i*C+j];
}

template <Index L, Index C, class T> vec_Inline Vect<C,T> _Mat<L,C,T>::operator[](unsigned int i) const throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (i >= (int)L)  )
    throw  ErrRange("Mat::operator[](Index)");
#endif
  return Vect<C,T>(val+i*C);
}

template <Index L, Index C, class T> vec_Inline T& _Mat<L,C,T>::operator()(int i, int j) throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (j<0) || (i >= (int)L) || (j >= (int)C)  )
    throw  ErrRange("Mat::operator()(Index,Index)");
#endif
  return val[i*C+j];
}

template <Index L, Index C, class T> vec_Inline T _Mat<L,C,T>::operator()(int i, int j) const throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (j<0) || (i >= (int)L) || (j >= (int)C)  )
    throw  ErrRange("Mat::operator()(Index,Index) const");
#endif
  return val[i*C+j];
}

// mathematical operations
// TODO : vec_Inlines pour les fonctions qui suivent ?????????????
template <Index L, Index C, class T> void _Mat<L,C,T>::private_trp(const _Mat<C,L,T> &A) throw()
{
  const T* M1 = A.getValues();
  T* M2 = val;
  for(register const T *m1i = M1; m1i < M1+L; m1i++)
    for(register const T *m1ji = m1i; m1ji < M1+C*L; m1ji += L)
      *M2++ = *m1ji;
}

template <Index L, Index C, class T> void _Mat<L,C,T>::trp(const _Mat<C,L,T> &A) throw()
{
  if ( ((_Mat<L,C,T>*) (&A) == this) ) {
    _Mat<L,C,T> tmp(VIM_NOINIT);
    tmp.private_trp(A);
    this->copy(tmp);
  } else {
    this->private_trp(A);
  }
}

template <Index L, Index C, class T> void _Mat<L,C,T>::private_symaction(const Sym<C,T>& Win, Sym<L, T>& Wout) const throw()
{
  for (Index i=0; i<L; i++)
    for (Index j=0; j<=i; j++) {
      T res=0;
      for (Index k=0; k<C; k++) {
        T Aik=val[i*C+k];
        T Ajk=val[j*C+k];
        for (Index l=0; l<k; l++)
          res+=Win(k,l)*(Aik*val[j*C+l]+Ajk*val[i*C+l]);
        res+=Win(k,k)*(Aik*Ajk);
      }
      Wout(i,j)=res;
    }
}

 template <Index L, Index C, class T> vec_Inline  void _Mat<L,C,T>::symaction(const Sym<C,T>& Win, Sym<L, T>& Wout) const throw()
{
  if ( &Win == (const Sym<C,T> *) &Wout  ) {
    Sym<L,T> tmp(VIM_NOINIT);
    private_symaction(Win,tmp);
    Wout=tmp;
  } else {
    private_symaction(Win,Wout);
  }
}

 template <Index L, Index C, class T> vec_Inline void  _Mat<L,C,T>::SymAction(const Sym<C,T>& Win, Sym<L, T>& Wout) const throw()
{symaction(Win,Wout);}

// I/O operations
// TODO : vec_Inline pour ces fonctions ????????
template <Index L, Index C, class T> STD_OSTREAM& operator<<(STD_OSTREAM &s, const _Mat<L,C,T> &m) throw()
{
  s<<"(";
  for (Index i=0; i<L; i++) {
    s<<"(";
    for (Index j=0; j<C; j++)
      s<<m(i,j)<<" ";
    s<<")\n";
  }
  s<<")";
  return s;
}

template<class T> const T sign( const T& a){return (( a < 0 ) ? -1 : 1);}
template<class T> const T Abs( const T& a){return (( a < 0 ) ? -a : a);}

template<class Real >
vec_Inline void Pythag(const Real f, const Real h, Real& c, Real& s, Real &z)
{
  Real nf=Abs(f),nh=Abs(h);
  if (nf > nh){
    Real tmp =nh/nf;
    Real sq = ::sqrt(1.0+ tmp*tmp);
    c = sign(f) / sq;
    s = sign(h) * tmp / sq;
    z = nf*sq;
  }
  else{
    if (nh > nf){
      Real tmp=nf/nh;
      Real sq = ::sqrt(1.0+ tmp*tmp);
      c = sign(f) * tmp / sq;
      s = sign(h) / sq;
      z = nh*sq;
    }
    else { // nh == nf
      const Real sq2 = ::sqrt(2.0);
      z = sq2 * nh;
      c = sign(f) / sq2;
      s = sign(h) / sq2;
    }
  }
}

template <class T> void private_svd(/*Mat<L,C,T>&*/ T*U, /*Vect<C,T>&*/ T*w, /*Mat<C,C,T>&*/ T*V, /*Vect<C,T>*/ T*Buffer, int m/*L*/, int n/*C*/) throw(ErrConvergence,ErrUnknown)
{
  T small_val;Xlinalg_Small_Val<T>(small_val);

  int i,j,k,l=0,jj;
  T c,f,h,s,x,y,z;
  T anorm=0.0,g=0.0,scale=0.0;

  const int MAXIT=130;

  for (i=0;i<n;i++) {
    l=i+1;
    Buffer[i]=scale*g;
    g=s=scale=0.0;
    if (i < m) {
      for(k=i;k<m;k++) scale += Abs(U[k*n+i]);
      if (scale > small_val ) {  // Addon: was if (scale)
        for(k=i;k<m;k++) {
          T tmp = ( U[k*n+i] /= scale );
          s += tmp*tmp;
        }
        f=U[i*n+i];
        g = -sign(f) * ::sqrt(s);
        h=f*g-s;
        U[i*n+i]=f-g;
        for(j=l;j<n;j++) {
          s=0.0;
          for(k=i;k<m;k++) s += U[k*n+i]*U[k*n+j];
          f= s/h;
          for(k=i;k<m;k++) U[k*n+j] += f*U[k*n+i];
        }
        for(k=i;k<m;k++) U[k*n+i] *= scale;
      }
    }
    w[i]=scale*g;
    g=s=scale=0.0;
    if ( (i < m) && (i != n-1)) {
      for(k=l;k<n;k++) scale += Abs(U[i*n+k]);
      if (scale > small_val ) { // Addon: was if (scale)
        for(k=l;k<n;k++) {
          T tmp = (U[i*n+k] /= scale);
          s += tmp*tmp;
        }
        f=U[i*n+l];
        g = - sign(f) * ::sqrt(s);
        h=f*g-s;
        U[i*n+l]=f-g;
        for(k=l;k<n;k++) Buffer[k]=U[i*n+k]/h;
        for(j=l;j<m;j++) {
          s = 0.0;
          for(k=l;k<n;k++) s += U[j*n+k]*U[i*n+k];
          for(k=l;k<n;k++) U[j*n+k] += s*Buffer[k];
        }
        for(k=l;k<n;k++) U[i*n+k] *= scale;
      }
    }
    T temp=(Abs(w[i])+Abs(Buffer[i]));
    if (temp>anorm) anorm=temp;
  }
  // TODO : l=0;
  for(i=n-1;i>=0;i--){
    if (i < n-1) {
      if ( Abs(g) > small_val ) { // Addon: was if (g)
        for(j=l;j<n;j++)
          V[j*n+i]=(U[i*n+j]/U[i*n+l])/g;
        for(j=l;j<n;j++) {
          s = 0.0;
          for(k=l;k<n;k++) s += U[i*n+k]*V[k*n+j];
          for(k=l;k<n;k++) V[k*n+j] += s*V[k*n+i];
        }
      }
      for(j=l;j<n;j++) V[i*n+j]=V[j*n+i]=0.0;
    }
    V[i*n+i]=1.0;
    g=Buffer[i];
    l=i;
  }
  for(i=(m<n?m:n)-1;i>=0;i--) {
    l=i+1;
    g=w[i];
    for(j=l;j<n;j++) U[i*n+j]=0.0;
    if ( Abs(g) > small_val ) { // Addon: was if (scale)
      g=1.0/g;
      for(j=l;j<n;j++) {
        s = 0.0;
        for(k=l;k<m;k++) s += U[k*n+i]*U[k*n+j];
        f=(s/U[i*n+i])*g;
        for(k=i;k<m;k++) U[k*n+j] += f*U[k*n+i];
      }
      for(j=i;j<m;j++) U[j*n+i] *= g;
    } else
      for(j=i;j<m;j++) U[j*n+i]=0.0;
    ++U[i*n+i];
  }
  for(k=n-1;k>=0;k--){
    for(int its=0;its<=MAXIT;its++) {
      int flag=1, nm=0;
      for(l=k;l>=0;l--) {
        nm=l-1; // Buffer[0] should always be 0;
        if ( ( ((T) Abs(Buffer[l]))+ anorm) == ((T) anorm)) {
          flag=0;
          break;
        }
        if ( (((T) Abs(w[nm]))+ anorm) == ((T) anorm)) break;
      }
      if (flag) {
        c=0.0;
        s=1.0;
        for(i=l;i<=k;i++) {
          f=s*Buffer[i];
          Buffer[i]=c*Buffer[i];
          if ( (((T) Abs(f)) + anorm) == ((T) anorm)) break;
          g=w[i];

          Pythag(g,f,c,s,h);
          w[i]=h;
          s = -s;

          for(j=0;j<m;j++) {
            y=U[j*n+nm];
            z=U[j*n+i];
            U[j*n+nm]=y*c+z*s;
            U[j*n+i]=z*c-y*s;
          }
        }
      }
      z=w[k];
      if (l == k) {
        if (z < 0.0) {
          w[k] = -z;
          for(j=0;j<n;j++) V[j*n+k]=(-V[j*n+k]);
        }
        break;
      }

      if( its >= MAXIT ){
        CERR << "SVD did not converge in " << its << " iterations"<<ENDL;
        throw ErrConvergence("svd");
      }

      x=w[l];
      nm=k-1;
      y=w[nm];
      g=Buffer[nm];
      h=Buffer[k];
      f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);

      // g= ::sqrt(f*f+1.0);
      if ( Abs(f) > 1.0 )
        g = Abs(f) * ::sqrt( 1.0 + 1.0/(f*f));
      else
        g = ::sqrt( 1.0 + f*f);

      f=((x-z)*(x+z)+h*((y/(f+ sign(f) * Abs(g)))-h))/x;
      c=s=1.0;
      for(j=l;j<=nm;j++) {
        i=j+1;
        g=Buffer[i];
        y=w[i];
        h=s*g;
        g=c*g;

        Pythag(f,h,c,s,z);
        Buffer[j]=z;

        f=x*c+g*s;
        g=g*c-x*s;
        h=y*s;
        y=y*c;
        for(jj=0;jj<n;jj++) {
          x=V[jj*n+j];
          z=V[jj*n+i];
          V[jj*n+j]=x*c+z*s;
          V[jj*n+i]=z*c-x*s;
        }

        Pythag( f,h,c,s,z);
        w[j]=z;

        f=(c*g)+(s*y);
        x=(c*y)-(s*g);
        for(jj=0;jj<m;jj++) {
          y=U[jj*n+j];
          z=U[jj*n+i];
          U[jj*n+j]=y*c+z*s;
          U[jj*n+i]=z*c-y*s;
        }
      }
      Buffer[l]=0.0;
      Buffer[k]=f;
      w[k]=x;
    }
  }
}

template <Index L, Index C, class T> void _Mat<L,C,T>::svd(Mat<L,C,T>& U, Vect<C,T>& W, Mat<C,C,T>& V) const throw(ErrConvergence,ErrUnknown)
{
  U.copy((const Mat<L,C,T>&)*this);
  Vect<C,T> tmp;
  private_svd(U.getValues(),W.getValues(),V.getValues(),tmp.getValues(),L,C);
}

#ifdef TODO
template <Index L, Index C, class T> void _Mat<L,C,T>::svd2(Mat<L,C,T>& U, Vect<C,T>& W, Mat<C,C,T>& VVV) const throw(ErrConvergence,ErrUnknown)
{
  Index i,j;
  const Mat<L,C,T>& A=(const Mat<L,C,T>&)*this;
  Matrix<C,T>& V=(Matrix<C,T>&)VVV;

  Mat<C,L,T> At(VIM_NOINIT);
  At.Trp(A);

  Sym<L,T> AAt(A*At);
  Matrix<L,T> U1(VIM_NOINIT);
  Vect<L,T> W1(VIM_NOINIT);
  AAt.DiagonalizeSort(U1,W1);

  Sym<C,T> AtA(At*A);
  AtA.DiagonalizeSort(V,W);

  for (j=0; j<C; j++) {
    if (W[j]<0) W[j]=0;
    W[j]= ::sqrt(W[j]);
  }

  for (i=0; i<L; i++) {
    for (j=0; j<C; j++) {
      U[i*n+j]=(j<L) ? U1(i,j) : 0;
    }
  }

  int possibiliteOK;
  T distance;

  for (int poss=0; poss<(1 << C); poss++){
    Vect<C,T> newW(W);

    for (i=0; i<C; i++)
      if (poss & (1 << i))
        newW[i]=-W[i];

    Mat<L,C,T> newA(U*Matrix<C,T>(newW)*V.getTransposed());

    T dist = Dist(A,newA);
    if ( (!poss) || (dist < distance) ){
      distance = dist;
      possibiliteOK = poss;
    }
  }

  for (j=0; j<C; j++)
    if (possibiliteOK & (1 << j))
      for (i=0; i<L; i++)
        U[i*n+j]=-U[i*n+j];
}
#endif

template <Index L, Index C, class T> vec_Inline void _Mat<L,C,T>::SVD(Mat<L,C,T>& U, Vect<C,T>& W, Mat<C,C,T>& V) const throw(ErrConvergence,ErrUnknown)
{svd(U,W,V);}

template<Index L, Index C, class T> vec_Inline int _Mat<L,C,T>::PseudoInv( const Mat<C,L,T>& A) throw(ErrConvergence)
{
  try{
    Mat<C,L,T> U(VIM_NOINIT);
    Vect<L,T> w(VIM_NOINIT);
    Mat<L,L,T> V(VIM_NOINIT);

    A.SVD(U,w,V);

    /* compute max singular value */
    T max = ::fabs( w(0) );
    for(Index i=1; i< L; i++){
      T tmp;
      if ( (tmp= ::fabs( w(i) ) ) > max) max = tmp;
    }

    /* Exeption for the null matrix */
    if ( ((T) max) == ((T) 0.0)) {
      this->resetTo(0);
      return L;
    }

    T Small_Val;Xlinalg_Small_Val<T>(Small_Val);
    const T Precision_Factor = 1e3;

    /* Ok, now we can divide by max */
    int normal_val = 0; /* count of non-null singular values */
    {
      Mat<L,C,T> Utrp(VIM_NOINIT);
      Utrp.Trp(U);

      for(Index i=0; i<L; i++){
        T& wi = w(i);
        T g = ::fabs( wi ) / Precision_Factor;
        //if ( ((T) (g + max))  == ((T) max) )
        if ( g <  max * Small_Val )
          wi =0.0;
        else {
          wi = 1.0/wi;
          normal_val++;
        }

        // inv = V.Winv.Utrp
        for(Index j=0; j<C; j++)
          Utrp(i,j) *= wi;
      }

      // inv = V. (Winv.Utrp)
      this->mulmat(V, Utrp);


      return ( (C<L ? C : L) - normal_val);
    }
  }
  catch( ErrConvergence& exc ){
    exc.push("Mat<m,n,T>::PseudoInv( const Mat<n,m,T>& A) ");
    throw exc;
  }
};


template <Index L, Index C, class T> STD_ISTREAM& operator>>(STD_ISTREAM &s,_Mat<L,C,T> &m) throw()
{
  T tmp;
  for (Index i=0; i<L; i++) {
    for (Index j=0; j<C; j++) {
      s >> tmp;
      m(i,j)=tmp;
    }
  }
  return s;
}


template <Index L, Index C, class T> void Mat<L,C,T>::Test() throw()
{
  float Af[L][C];double Ad[L][C];int Ai[L][C];
  T A[L][C];for (Index k=0; k<L; k++) for (Index l=0; l<C; l++) {A[k][l]=k*C+l+1;Af[k][l]=k*C+l+1;Ad[k][l]=k*C+l+1;Ai[k][l]=k*C+l+1;}
  Mat<L,C,float> df;Mat<L,C,double> dd;Mat<L,C,int> di;

  Mat<L,C,T> a(A),b,c(VIM_INIT),d(a),/*af(Af),ad(Ad),ai(Ai),*/ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Mat<"<<L<<","<<C<<","<<TName<<">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
//   COUT << "af must be 1,2,... : " << af << ENDL;
//   COUT << "ad must be 1,2,... : " << ad << ENDL;
//   COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be 0,0,... : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  COUT << "d[0] : " << d[0] << ENDL;
  d[0][0]=25;COUT << "d[0][0]=25;  d[0][0] : " << d[0][0] << ENDL;
  d(0,0)=26;COUT << "d(0,0)=26;  d(0,0) : " << d(0,0) << ENDL;
  d.setValue(0,0,11);COUT << "d.setValue(0,0,11);  d.getValue(0,0) : " << d.getValue(0,0) << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*Dd=d.getValues();
  COUT << "getValues() : " << Dd[0] << " " << Dd[1] << " " << Dd[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  d.setValues(A);COUT << "d.setValues(T[1,2,3...]) : " << d << ENDL;
//   d.setValues(Af);COUT << "d.setValues(float[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ad);COUT << "d.setValues(double[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ai);COUT << "d.setValues(int[1,2,3...]) : " << d << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;

  Mat<C,L,T> m;COUT << "m.Trp(d) : " << m.Trp(d) << ENDL;
  COUT << "d.getTransposed() : " << d.getTransposed() << ENDL;

  {Mat<L,1,T> m;Mat<1,C,T> n;COUT << "Mat<L,1,T>*Mat<1,C,T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}
  {Mat<L,2,T> m;Mat<2,C,T> n;COUT << "Mat<L,2,T>*Mat<2,C,T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}
  {Mat<L,3,T> m;Mat<3,C,T> n;COUT << "Mat<L,3,T>*Mat<3,C,T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}
  {Mat<L,4,T> m;Mat<4,C,T> n;COUT << "Mat<L,4,T>*Mat<4,C,T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}
  {Mat<L,5,T> m;Mat<5,C,T> n;COUT << "Mat<L,5,T>*Mat<5,C,T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}
  {Mat<L,6,T> m;Mat<6,C,T> n;COUT << "Mat<L,6,T>*Mat<6,C,T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}

  {d.Rnd();Sym<C,T> S1;Sym<L,T> S2;d.SymAction(S1,S2);COUT << "SymAction" << ENDL;}
  {d.Rnd();Mat<L,C,T> S1;Vect<C,T>W;Mat<C,C,T> S2;d.SVD(S1,W,S2);COUT << "SVD" << ENDL;}
  {d.Rnd();Mat<C,L,T> S;d.PseudoInv(S);COUT << "PseudoInv" << ENDL;}

  COUT << ENDL;
}








// *********************************************************************
// Lower/Upper matrix class for system solving and matrix inversion
// Rewritten from NRC
// *********************************************************************

/*! \class LUMat Mat.h linearAlgebra/Mat.h
  \ingroup liblinearAlgebra
  Lower/Upper matrix class for system solving and matrix inversion
  Rewritten from NRC
*/
template<Index D, class T>
class LUMat{
  Matrix<D,T> mat;      // Upper (+diag) is U and lower part is L
  Index rowPerm[D];             // index of row permutations
  int permSign;                 // signature of permutation // was d

public:
  // Construct the LU decomposition of a square matrix A
  // ----------------------------------------------------------------------
  LUMat( const Matrix<D,T>& A) throw(ErrMatSingular);

  // Destructor
  // ----------------------------------------------------------------------
  ~LUMat() throw() {};

  // LU backsubstitution for system A*x = b - adapted from NRC
  //    Matrix A is in LU form -- Result x replaces the vector b
  // ----------------------------------------------------------------------
  void BackSub( Vect<D,T>& xb) throw();

  // Determinant of a matrix in LU form
  // ----------------------------------------------------------------------
  T Det() const throw();

};

// *********************************************************************
// Implementation of Lower/Upper matrix class operations
// *********************************************************************

// Construct the LU decomposition of a square matrix A
// Adapted, rewritten and debugged from NRC
// ----------------------------------------------------------------------
template<class T>
void private_LUMat(T*mat, Index D, Index* rowPerm, int& permSign, T* Buffer) throw(ErrMatSingular)
{
  Index i,j,k;
  permSign = 1;         // No row interchanged yet

  const T Precision_Factor = 1e3;
  T Small_Val;Xlinalg_Small_Val<T>(Small_Val);

  // Getting scale information
  for(i=0; i<D; i++) {
    T scale=0.0;
    T tmp;
    for (j=0; j<D; j++)
      if ((tmp=(T) ::fabs( mat[i*D+j] )) > scale) scale = tmp;

    if ( scale/ Precision_Factor < Small_Val ){  // Addon: was "if(scale)"
      // Singular matrix : clean-up and throw exception
      for(j=0; j<D*D; j++) mat[j]=0.0;
      for(j=0; j<D; j++) rowPerm[j] = 0;
      throw ErrMatSingular("LUMat<D,T>( Matrix<D,T>& )");
    }
    Buffer[i] = 1.0/scale;
  }

  for (j=0; j<D; j++) {
    for (i=0; i<j; i++) {
      T sum = mat[i*D+j];
      for (k=0; k<i; k++)
        sum -= mat[i*D+k] * mat[k*D+j];
      mat[i*D+j] = sum;
    }
    T scale=0.0;
    Index imax=0;
    for (Index ii=j; ii<D; ii++) {
      T sum = mat[ii*D+j];
      for (Index k=0; k<j; k++)
        sum -= mat[ii*D+k] * mat[k*D+j];
      mat[ii*D+j] = sum;
      register T tmp = Buffer[ii] * ::fabs(sum);
      if ( tmp > scale ) {
        scale = tmp;
        imax = ii;
      }
    }
    if (j != imax) {
      for (Index k=0; k < D; k++) {
        register T tmp = mat[imax*D+k];
        mat[imax*D+k] = mat[j*D+k];
        mat[j*D+k] = tmp;
      }
      permSign *= -1;
      Buffer[imax] = Buffer[j];
    }
    rowPerm[j] = imax;
    if ( ::fabs(mat[j*D+j]) <  Small_Val ) // Addon
      mat[j*D+j] = 2.*Small_Val;
    // WAS: if ( ((T) mat[j*D+j]) == ((T) 0.0) )
    //      mat[j*D+j] = TINY= Small_Val * 1e-10;
    if (j != D) {
      register T tmp = 1.0/( mat[j*D+j] );
      for (Index i=j+1; i<D; i++)
        mat[i*D+j] *= tmp;
    }
  }
}

template<Index D, class T>
LUMat<D,T>::LUMat( const Matrix<D,T>& A) throw(ErrMatSingular)
  : mat(A)
{
  Vect<D,T> Buffer;
  private_LUMat(mat.getValues(), D, rowPerm, permSign, Buffer.getValues());
};


// LU backsubstitution for system A*x = b - adapted from NRC
//    Matrix A is in LU form -- Result x replaces the vector b
// ----------------------------------------------------------------------
template<class T>
void private_BackSub(T*mat, Index D, Index* rowPerm, T* xb)
{
  int ii= -1;

  {
    for (Index i=0; i<D; i++) {
      Index ip = rowPerm[i];
      register T sum = xb[ip];
      xb[ip] = xb[i];
      if (ii != -1)
        for (Index j=ii; j<i; j++)
          sum -= mat[i*D+j] * xb[j];
      else
        if (sum != 0)
          ii=i;
      xb[i] = sum;
    }
  }
  for(int i = D-1; i>=0; i--){ //for(Index i = D-1; i<D; i--)
    register T sum = xb[i];
    for (Index j=i+1; j<D; j++)
      sum -= mat[i*D+j] * xb[j];
    xb[i] = sum / mat[i*D+i];
  }
}

template<Index D, class T>
void LUMat<D,T>::BackSub( Vect<D,T>& xb) throw() {
  private_BackSub(mat.getValues(), D, rowPerm, xb.getValues());
};

// Determinant
// ----------------------------------------------------------------------
template<Index D, class T>
vec_Inline T LUMat<D,T>::Det() const throw()
{
  T d=permSign;
  for(Index j=0; j<D; j++)
    d *= mat(j,j);
  return d;
}










// mathematical operations
template <Index D, class T> vec_Inline void Matrix<D,T>::resetToIdentity(const T v) throw()
{
  for (Index i=0; i<D; i++) for (Index j=0; j<D; j++) if (i==j) this->val[i*D+j]=v; else this->val[i*D+j]=0;
}

template <Index D, class T> vec_Inline void Matrix<D,T>::diag(const Vect<D,T>& v) throw()
{
  Null();for (Index i=0; i<D; i++) this->val[i*(D+1)]=v[i];
}

template <Index D, class T> vec_Inline void Matrix<D,T>::tensorProduct(const Vect<D,T>& v1, const Vect<D,T>& v2) throw()
{
  for (Index i=0; i<D; i++) for (Index j=0; j<D; j++) this->val[i*D+j]=v1[i]*v2[j];
}


// mathematical operations
template <Index D, class T> vec_Inline void Matrix<D,T>::inv(const Matrix<D,T> &m) throw(ErrMatSingular)
{
  // Decompose A in LU form.
  try{
    LUMat<D,T> LU( m );
    for(Index j=0; j<D; j++) {
      // Init current column to ej
      Vect<D,T> current_column(j);
      // solve A.ej' = ej
      LU.BackSub( current_column );
      // copy result into Ainv.
      // This is safe since we don't use the matrix A anymore.
      for(Index i=0; i<D; i++) this->val[i*D+j] = current_column(i);
    }
  }
  catch( ErrMatSingular& exc ){
    exc.push("Inv( Mat, Mat )");
    throw exc;
  }
}


template <Index D, class T> vec_Inline T Matrix<D,T>::det() const throw()
{
  // Decompose A in LU form.
  try{
    LUMat<D,T> LU( *this );
    return LU.Det();
  }
  catch( ErrMatSingular& ){
    // Singularity in LU decomposition means that a column is null
    return 0.0;
  }
}

template <Index D, class T> vec_Inline T Matrix<D,T>::trace() const throw()
{
  T t=0;
  for (Index i=0; i<D; i++) t+=this->val[i*(D+1)];
  return t;
}

template <Index D, class T> vec_Inline bool Matrix<D,T>::computeEigenV(Matrix<D,T> &eigenVecttors, Vect<D,T> &eigenValues) throw()
{
  try {
    Sym<D,T>(*this).DiagonalizeSort(eigenVecttors,eigenValues);
  }
  catch(ErrConvergence& ){
    return false;
  }
  return true;
}

template <Index D, class T> void Matrix<D,T>::Test() throw()
{
  float Af[D][D];double Ad[D][D];int Ai[D][D];
  T A[D][D];for (Index k=0; k<D; k++) for (Index l=0; l<D; l++) {A[k][l]=k*D+l+1;Af[k][l]=k*D+l+1;Ad[k][l]=k*D+l+1;Ai[k][l]=k*D+l+1;}
  Matrix<D,float> df;Matrix<D,double> dd;Matrix<D,int> di;
  Vect<D,T> bb(0);

  Matrix<D,T> a(A),b(bb),c(VIM_INIT),d(a),/*af(Af),ad(Ad),ai(Ai),*/ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Matrix<"<<D<<","<<TName<<">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
//   COUT << "af must be 1,2,... : " << af << ENDL;
//   COUT << "ad must be 1,2,... : " << ad << ENDL;
//   COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be diag(1,0,...) : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  COUT << "d[0] : " << d[0] << ENDL;
  d[0][0]=25;COUT << "d[0][0]=25;  d[0][0] : " << d[0][0] << ENDL;
  d(0,0)=26;COUT << "d(0,0)=26;  d(0,0) : " << d(0,0) << ENDL;
  d.setValue(0,0,11);COUT << "d.setValue(0,0,11);  d.getValue(0,0) : " << d.getValue(0,0) << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*Dd=d.getValues();
  COUT << "getValues() : " << Dd[0] << " " << Dd[1] << " " << Dd[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  d.setValues(A);COUT << "d.setValues(T[1,2,3...]) : " << d << ENDL;
//   d.setValues(Af);COUT << "d.setValues(float[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ad);COUT << "d.setValues(double[1,2,3...]) : " << d << ENDL;
//   d.setValues(Ai);COUT << "d.setValues(int[1,2,3...]) : " << d << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;

  Matrix<D,T> m;COUT << "m.Trp(d) : " << m.Trp(d) << ENDL;
  d.transpose();COUT << "d.transpose() : " << d << ENDL;
  COUT << "d.getTransposed() : " << d.getTransposed() << ENDL;

  {Matrix<D,T> m;Matrix<D,T> n;COUT << "Matrix<D,T>*Matrix<D,T> : " << m*n << "  Mul : " << d.Mul(m,n) << ENDL;}

  {d.Rnd();Sym<D,T> S1;Sym<D,T> S2;d.SymAction(S1,S2);COUT << "SymAction" << ENDL;}
  {d.Rnd();Matrix<D,T> S1;Vect<D,T>W;Matrix<D,T> S2;d.SVD(S1,W,S2);COUT << "SVD" << ENDL;}
  {d.Rnd();Matrix<D,T> S;d.PseudoInv(S);COUT << "PseudoInv" << ENDL;}

  {d.Rnd();d.Inv(d);COUT << "Inv" << ENDL;}
  {d.Rnd();COUT << "getInverse" << d.getInverse() << ENDL;}
  {d.Rnd();d.invert();COUT << "invert" << ENDL;}

  {d.Rnd();COUT << "Det" << d.Det() << ENDL;}
  {d.Rnd();COUT << "Trace" << d.Trace() << ENDL;}

  {d.Rnd();COUT << "Id" << d.Id() << ENDL;}
  {d.Rnd();Vect<D,T>v;COUT << "Diag" << d.Diag(v) << ENDL;}
  {d.Rnd();Vect<D,T>v;COUT << "TensorProduct" << d.TensorProduct(v,v) << ENDL;}

  {Sym<D,T> m;m.Rnd();d=m;COUT << "operator=Sym<D,T>" << d << "=" << m << ENDL;}
  {Sym<D,T> m;m.Rnd();Matrix<D,T> d(m);COUT << "Constructeur(Sym<D,T>)" << d << "=" << m << ENDL;}

  COUT << ENDL;
}


} // namespace yav

#endif
