#ifndef PETSCSPARSEMATRIX_H
#define PETSCSPARSEMATRIX_H

#ifdef USE_PETSC

#pragma interface "linearAlgebra/PETSCSparseMatrix.h"
#include <linearAlgebra/GenericSparseMatrix.h>

#include "petscvec.h"
#include "petscksp.h"
#include "petscao.h"

namespace yav
{

  class PETSCSparseVector : public GenericSparseVector
  {
		
  protected:
    Vec vector;

  public:
    /// void constructor 
    PETSCSparseVector( const unsigned int nbRows );

    /// void destructor
    virtual ~PETSCSparseVector();

    /// add a value to a vector
    virtual void addVectorElement( const unsigned int row, const double value ); 

    /// set a value to a vector
    virtual void setVectorElement( const unsigned int row, const double value ); 

    /// get vector value
    virtual double getVectorElement( const unsigned int row ) const; 

    /// set vector to zero
    virtual void resetVector(); 

    /// assemble vector: to be done after all calls to set/addVectorElement
    virtual void assembleVector();

    // get vector as an array of double
    virtual const double* getVectorAsArray();

    // restore the array of double (free the memory). MUST be done after
    // getVectorAsArray
    virtual void restoreVectorArray( double* array );

    /// get vector size
    virtual unsigned int getSize() const; 

    /// save vector in file
    virtual void saveInFile(const std::string filename) const;
 		
    friend class PETSCSparseMatrix;
  };



// -----------------------------------------------------------------------------



  class PETSCSparseMatrix : public GenericSparseMatrix
  {
		
  protected:
    Mat matrix;
    PreconditionerType pt;
    SolverType st;

    // Variables specific to PETSC
    KSP solver;
    PC preconditioner;
    bool hasMatrixChanged;
    
  public:
    /// Create a matrix of a given number of rows and columns
    PETSCSparseMatrix( const unsigned int sizeRows, 
		       const unsigned int sizeColumns,
		       const unsigned int nbNonZeroDiagonal = 0,
		       const unsigned int nbNonZeroOffDiagonal = 0 );

    /// void destructor
    virtual ~PETSCSparseMatrix();

    /// initializes some static variables associated with the linear algebra library
    virtual void initializeLibrary(){}

    /// set all matrix elements to zero
    virtual void resetMatrix();

    /// add a given value to an entry in the matrix
    virtual void addMatrixElement( const unsigned int row,
				   const unsigned int column,
				   const double val );

    /// add a given value to an entry in the matrix and its transpose entry
    virtual void addSymmetricMatrixElement( const unsigned int row,
					    const unsigned int column,
					    const double val );

    /// set a given value to an entry in the matrix
    virtual void setMatrixElement( const unsigned int row,
				   const unsigned int column,
				   const double val );

    /// set a given value to an entry in the matrix and its transpose entry 
    virtual void setSymmetricMatrixElement( const unsigned int row,
					    const unsigned int column,
					    const double val );

    /// get a given value to an entry in the matrix
    virtual double getMatrixElement( const unsigned int row,
				     const unsigned int column ) const;

    /// add matrices  
    virtual void matrixAddMatrix( const GenericSparseMatrix *m );

    /// add multiply all matrix elements with a scalar  
    virtual void scalarMultiplyMatrix( const double s );

    /// assemble matrix : the matrix can no longer be modified 
    virtual void assembleMatrix();

    /// build a preconditioner with the associated matrix
    virtual void setPreconditioner( const PreconditionerType _pt )
      {
	pt=_pt;
      };

    /// set solver type
    virtual void setSolverType( const SolverType _st )
      {
	st=_st;
      };

    /// save matrix in file
    virtual void saveInFile(const std::string filename) const;

    /** 
     * Solve a linear system of equations given the solver type
     * (iterative methods), the right hand term and the maximum number
     * of iterations and a tolerance number to terminate the algorithm 
     */
    virtual void solveIterativeMethod( GenericSparseVector* rightHandSide,
				       GenericSparseVector* result,
				       const unsigned int maximumNumberIterations,
				       const double tolerance,
				       double &residual );

  };


} // End of namespace yav

#endif

#endif

