#ifndef _RIGID3D_H_
#define _RIGID3D_H_

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Rigid3D.h"
#endif
#include "linearAlgebra/Rotation3D.h"
#include "linearAlgebra/Matrix4x4.h"
#include "linearAlgebra/Matrix3x3.h"

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {

// forward declarations
class RigidMat3D;
class RigidVec3D;

typedef RigidMat3D Rigid3D;
typedef RigidVec3D RigidVect;

/// Conversion from a rigid matrix to a rigid vector
void RigidMat3DToRigidVec3D( const RigidMat3D& rig, RigidVec3D& r) throw();
/// Conversion from a rigid vector to a rigid matrix
void RigidVec3DToRigidMat3D( const RigidVec3D& r, RigidMat3D& rig) throw();


/*! \class RigidMat3D Rigid3D.h linearAlgebra/Rigid3D.h
  \ingroup liblinearAlgebra
  3D rigid transformation Matrix
*/
 class RigidMat3D : protected Matrix4x4<double> {
 public:
   /*! \class ErrNonRigidMat Rigid3D.h linearAlgebra/Rigid3D.h
    \ingroup liblinearAlgebra
     Non rigid matrix excpetion
   */
   class ErrNonRigidMat : public Exception {
  public:
    // constructor
    ErrNonRigidMat(const std::string &from) throw() :
      Exception(from, "Non rigid matrix exception") {};
  };

  /** @name Constructors */
  //@{
  /// default constructor ; initializes to identity matrix
  RigidMat3D() throw() : Matrix4x4<double>(VIM_NOINIT)
    {this->Id();}

  /// Copy constructor
  RigidMat3D(const RigidMat3D & r) throw() : Matrix4x4<double>(r)
    {};
  /// operator =
  RigidMat3D& operator=(const RigidMat3D & r) throw()
    {this->copy((const Matrix4x4<double>&)r);return *this;}

  /// set from rotation and translation part
  RigidMat3D(const RotMat3D& rot, const Vec3<double>& trans) throw() : Matrix4x4<double>(rot,trans)
    {}

  friend void RigidVec3DToRigidMat3D( const RigidVec3D& r, RigidMat3D& rig) throw();
  friend void RigidMat3DToRigidVec3D( const RigidMat3D& rig, RigidVec3D& r) throw();

  /// construct rigid matrix from a rigid vector
  RigidMat3D(const RigidVec3D& r) throw() : Matrix4x4<double>(VIM_NOINIT)
    {RigidVec3DToRigidMat3D(r,*this);}
  /// set rigid matrix from a rigid vector
  RigidMat3D& operator=(const RigidVec3D& r) throw()
    {RigidVec3DToRigidMat3D(r,*this);return *this;}

  /// construct rigid matrix from an affine matrix (LSQ Vectorial  distance - extrinsic)
  template <class T2>
  RigidMat3D(T2 matrix[4][4]) throw(NonUniqueOptimum,ErrNonRigidMat,ErrConvergence) : Matrix4x4<double>(VIM_NOINIT)
    {val[12]=val[13]=val[14]=0;val[15]=1;
    try {LSQ(Matrix4x4<double>(matrix));} catch(ErrConvergence& exc){exc.push("RigidMat3D::RigidMat3D( Mat<3>& K )");throw exc;}}
  /// construct rigid matrix from an affine matrix (LSQ Vectorial  distance - extrinsic)
  template <class T2>
  RigidMat3D(const Matrix4x4<T2>& K) throw(NonUniqueOptimum,ErrNonRigidMat,ErrConvergence) : Matrix4x4<double>(VIM_NOINIT)
    {val[12]=val[13]=val[14]=0;val[15]=1;
    try {LSQ(Matrix4x4<double>(K));} catch(ErrConvergence& exc){exc.push("RigidMat3D::RigidMat3D( Mat<3>& K )");throw exc;}}
  /// set rigid matrix from an affine matrix (LSQ Vectorial  distance - extrinsic)
  template <class T2>
  RigidMat3D& operator=(const Matrix4x4<T2>& K) throw(NonUniqueOptimum,ErrNonRigidMat,ErrConvergence)
    {try {LSQ(Matrix4x4<double>(K));} catch(ErrConvergence& exc){exc.push("RigidMat3D::operator=( Mat<3>& K )");throw exc;} return *this;}

  /// Constructor : load transfo from a file in O8 format
  RigidMat3D(const char* FileName) throw(ErrUnknown)
    : Matrix4x4<double>(VIM_NOINIT)
     {val[12]=val[13]=val[14]=0;val[15]=1;readO8(FileName);}
  /// Constructor : load transfo from a file in O8 format
  RigidMat3D(IFSTREAM& fin) throw(ErrUnknown)
    : Matrix4x4<double>(VIM_NOINIT)
     {val[12]=val[13]=val[14]=0;val[15]=1;readO8(fin);}

  /// Destructor
  ~RigidMat3D() throw()
    {};
  //@}

  /** @name Data members access */
  //@{
  /// Access to matrix element i,j
  double operator()(const int i, const int j) const  throw(ErrRange)
    {return Matrix4x4<double>::operator()(i,j);}
  /// Access to matrix element i,j
  Vect<3,double> operator[](const int i) const  throw(ErrRange)
    {return Matrix4x4<double>::operator[](i);}
  /// Access to matrix element i,j
  double getValue(const int i, const int j) const  throw(ErrRange)
    {return Matrix4x4<double>::getValue(i,j);}

  ///
  Matrix4x4<double>::resetToIdentity;
  ///
  RigidMat3D& Id() throw()
    {resetToIdentity();return *this;}

  // Conversion to matrix
  // ----------------------------------------------------------------------
  //operator const Matrix4x4<double>&() const throw()
  //  { return (const Matrix4x4<double>&) *this;}
  // Conversion to a const Matrix4x4<double> does not work
  // -> Use a member function
  const Matrix4x4<double>& mat() const throw()
   { return (const Matrix4x4<double>&) *this;}

  /// set rotation part
  RigidMat3D& operator= (const RotMat3D& rot) throw();
  /// set translation part
  RigidMat3D& operator= (const Vec3<double>& trans) throw();

  /// get COPY of rotation part
  RotMat3D getRotPartCopy() const throw()
   {return RotMat3D(Matrix3x3<double>(val[0],val[1],val[2],val[4],val[5],val[6],val[8],val[9],val[10]));}
  /// get COPY of rotation part
  Vec3<double> getTransPartCopy() const throw()
   {return Vec3<double>(val[3],val[7],val[11]);}
  //@}

  /** @name Mathematical operations */
  //@{
  /** Closest Rigidation to an affine matrix (Vectorial  distance - extrinsic)
      @param K affine matrix
      @return residue
   */
  double LSQ(const Matrix4x4<double>& K) throw(NonUniqueOptimum,ErrNonRigidMat,ErrConvergence);

   /** @param K affine matrix
       @param ModelMean
       @param SceneMean */
  RigidMat3D& LSQ(const Matrix3x3<double>& K,
                  const Vec3<double>& ModelMean,
                  const Vec3<double>& SceneMean)
    throw(NonUniqueOptimum,ErrNonRigidMat,ErrConvergence);

#ifdef TODO
  /// Intrinsic Distance between rigid matrices : angle( R1trp.R2)
  friend double Dist(const RigidMat3D& R1, const RigidMat3D& R2) throw();
#endif

  /** Extrinsinc Euclidean Distance between rigid matrices
      as in the Matrix4x4<double> vector space. For rigids :
      dist(R1,R2)^2 = 6 - 2 tr(R1trp.R2) */
  friend double DistEuc( const RigidMat3D& R1, const RigidMat3D& R2) throw()
    {return Dist( (Matrix4x4<double>&) R1, (Matrix4x4<double>&) R2);}

#ifdef TODO
  ///  Uniform Random 3D Rigid matrix
  RigidMat3D& Rnd(void) throw();
#endif

  /// Invert a rigid matrix
  void inv( const RigidMat3D& R) throw();
  /// Invert a rigid matrix
  inline RigidMat3D& Inv( const RigidMat3D& R) throw()
    {this->inv(R); return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  inline RigidMat3D getInverse(void) const throw()
    {return RigidMat3D().Inv(*this);}
  /// invert this matrix (this matrix is modified)
  inline void invert(void) throw()
    {this->inv(*this);}

  /// Apply rigid matrix to a vect
  void Apply(const Vec3<double>& In, Vec3<double>& Out) const throw()
    {Out=((const Matrix4x4<double>&)*this) * In;}
  /// Apply rigid matrix to a vector
  friend inline Vec3<double> operator*(const RigidMat3D& m, const Vec3<double>& v)
    {return ((const Matrix4x4<double>&)m)*v;}

  /// Compose rigid matrices
  inline void comp(const RigidMat3D& R1, const RigidMat3D& R2)  throw()
    {this->Mul(R1,R2);}
  /// Compose rigid matrices
  inline RigidMat3D& Comp(const RigidMat3D& R1, const RigidMat3D& R2)  throw()
    {this->comp(R1,R2); return *this;}
  /// Matrix multiplication
  inline RigidMat3D& operator*=(const RigidMat3D& m) throw()
    {comp(*this,m);return *this;}
  /// Matrix multiplication
  friend inline RigidMat3D operator*(const RigidMat3D& m, const RigidMat3D& v)
    {return RigidMat3D().Comp(m,v);}

#define MAX_NB_ITER_MEAN 20
  // compute mean transfo of given list in this
   /// @param first
   /// @param last
  template<class InputIterator>
  void mean(InputIterator first, InputIterator last) throw()
    {
      *this=*first;
      for (int iter=0; iter<MAX_NB_ITER_MEAN; iter++) {
        RigidMat3D minv;minv.inv(*this);
        Vect<6,double> dm;
        int n=0;
        InputIterator irigidmat;
        for (irigidmat=first;irigidmat!=last;irigidmat++){
          dm+=(const Vect<6,double>&) RigidVec3D(minv*(*irigidmat));
          n+=1;
        }
        dm*=1/(double)n;
        if (dm.norm2()<1e-12) break;
        *this*=RigidMat3D(dm);
      }
    }
   /// compute covariance of given list of transfo around this
   /// @param first
   /// @param last
   /// @param covar
  template<class InputIterator>
  void covararound(InputIterator first, InputIterator last, Sym<6,double>& covar) const throw()
    {
      Vect<6,double> vtmp;
      Sym<6,double> tmp;
      RigidMat3D minv;minv.inv(*this);
      covar.Null();
      int n=0;
      InputIterator irigidmat;
      for (irigidmat=first;irigidmat!=last;irigidmat++){
        vtmp=(const Vect<6,double>&) RigidVec3D(minv*(*irigidmat));
        tmp.AutoCorrel(vtmp);
        covar+=tmp;
        n+=1;
      }
      covar*=1/(double)n;
    }
  //@}

  /** @name I/O operations */
  //@{
  /// Print on an output stream
  friend STD_OSTREAM& operator<<(STD_OSTREAM& os, const RigidMat3D& rig) throw()
    {return os<<(Matrix4x4<double>&) rig;}

  /// read a Matrix4x4 from a file using old O8 format
  void readO8(IFSTREAM& fin) throw(ErrUnknown);
  /// read a Matrix4x4 from a file using old O8 format
  void readO8(const char* FileName) throw(ErrUnknown);
  /// write a Matrix4x4 in a file using old O8 format
  void writeO8(OFSTREAM& fout) const throw(ErrUnknown);
  /// write a Matrix4x4 in a file using old O8 format
  void writeO8(const char* FileName) const throw(ErrUnknown);

  /// read a Matrix4x4 from a file using old O8 format
  void ReadO8(IFSTREAM& fin) throw(ErrUnknown)
     {readO8(fin);}
  /// read a Matrix4x4 from a file using old O8 format
  void ReadO8(const char* FileName) throw(ErrUnknown)
     {readO8(FileName);}
  /// write a Matrix4x4 in a file using old O8 format
  void WriteO8(OFSTREAM& fout) const throw(ErrUnknown)
     {writeO8(fout);}
  /// write a Matrix4x4 in a file using old O8 format
  void WriteO8(const char* FileName) const throw(ErrUnknown)
     {writeO8(FileName);}

  /// read a Matrix4x4 from a file using old O8 format
  template <class List>
  static void readO8List(IFSTREAM& fin, List& list)
    {
      while (fin && (!fin.eof())) {
        try {
          list.push_back(RigidMat3D(fin));
        } catch(...) {
          break;
        }
      }
    }
  /// read a Matrix4x4 from a file using old O8 format
  template <class List>
  static void readO8List(const char* FileName, List& list)
     {
       IFSTREAM fin(FileName);
       RigidMat3D::readO8List(fin,list);
     }

  /// write a Matrix4x4 in a file using old O8 format
  template <class Iterator>
  static void writeO8List(OFSTREAM& fout, Iterator first, Iterator last) throw(ErrUnknown)
    {
      for (Iterator it=first; it!=last; it++) {
        it->writeO8(fout);
      }
    }
  /// write a Matrix4x4 in a file using old O8 format
  template <class Iterator>
  void writeO8List(const char* FileName, Iterator first, Iterator last) throw(ErrUnknown)
     {
       OFSTREAM fout(FileName);
       RigidMat3D::writeO8List(fout,first,last);
     }
   //@}
};





// **********************************************************************
// 3D RigidMat vector
// **********************************************************************
/*! \class RigidVec3D Rigid3D.h linearAlgebra/Rigid3D.h
  \ingroup liblinearAlgebra
  3D RigidMat vector
*/

class RigidVec3D : protected Vect<6,double> {
public:
  /** @name Constructors */
  //@{
  /// Default constructor
  RigidVec3D() throw() : Vect<6,double>()
    {};

  /// Copy Constructor
  RigidVec3D(const RigidVec3D & rv) throw() : Vect<6,double>(rv)
    {};
  /// Operator =
  RigidVec3D& operator=(const RigidVec3D & rv) throw()
    {(Vect<6,double>&)(*this)=rv;return *this;}

  friend void RigidVec3DToRigidMat3D( const RigidVec3D& r, RigidMat3D& rig) throw();
  friend void RigidMat3DToRigidVec3D( const RigidMat3D& rig, RigidVec3D& r ) throw();

  /// constructor from a rigid matrix
  RigidVec3D(const RigidMat3D& rig) throw() : Vect<6,double>(VIM_NOINIT)
    {RigidMat3DToRigidVec3D(rig,*this);}
  /// conversion from a rigid matrix
  RigidVec3D& operator= (const RigidMat3D& rig) throw()
    {RigidMat3DToRigidVec3D(rig,*this);return *this;}

  /// Constructor from values
  template <class T2>
  RigidVec3D(T2 rx, T2 ry, T2 rz, T2 tx, T2 ty, T2 tz) throw()
    : Vect<6,double>(VIM_NOINIT)
    {val[0]=rx;val[1]=ry;val[2]=rz;
    val[3]=tx;val[4]=ty;val[5]=tz;
    Domain();}
  /// Constructor from values
  template <class T2>
  RigidVec3D(const T2 v[6]) throw() : Vect<6,double>(v)
    {Domain();}
  /// Constructor from Vec3<double>
  template <class T2>
  RigidVec3D(const Vect<6,T2>& v) throw() : Vect<6,double>(v)
    {Domain();}
  /// Constructor from Vec3<double>
  template <class T2>
  RigidVec3D& operator=(const Vect<6,T2>& v) throw()
    {(Vect<6,double>&)*this=v;Domain();return *this;}

  /// Constructor : load transfo vector from a file
  RigidVec3D(const char* FileName) throw(ErrUnknown)
    : Vect<6,double>(VIM_NOINIT)
     {readV6(FileName);}
  /// Constructor : load transfo vector from a file
  RigidVec3D(IFSTREAM& fin) throw(ErrUnknown)
    : Vect<6,double>(VIM_NOINIT)
     {readV6(fin);}

  /// Destructor
  ~RigidVec3D() throw()
    {};
  //@}

  /** @name Access to data */
  //@{
  /// Access to an element
  double operator()(const int i) const throw(ErrRange) {
    return Vect<6,double>::operator()(i);}

  Vect<6,double>::resetToNull;

  vec_Inline void resetToIdentity()throw()
    {resetToNull();}
  vec_Inline RigidVec3D& Id() throw()
    {resetToNull();return *this;}

  vec_Inline const RotVec3D& RotPart() const
    {return ((const RotVec3D&)val[0]);}
  vec_Inline const Vec3<double>& TransPart() const
    {return ((const Vec3<double>&)val[3]);}

  vec_Inline RotVec3D& RotPart()
    {return ((RotVec3D&)val[0]);}
  vec_Inline Vec3<double>& TransPart()
    {return ((Vec3<double>&)val[3]);}
  //@}

  /** @name Mathematical operations */
  //@{
  /// Replace vector in definition domain (where norm(v)<Pi)
  void Domain() throw()
    {}

  /// Intrinsic Norm (distance to Id) for rigid : angle( r ) = || r ||
  double Norm( void) const throw()
    {return Vect<6,double>::Norm();}

#ifdef TODO
  /// Intrinsic Distance between rigid vectors : angle( r1\inv o r2)
  friend double Dist( const RigidVec3D& r1, const RigidVec3D& r2) throw();

  ///  Uniform Random 3D Rigidation
  RigidVec3D& Rnd(void) throw();
#endif

  /// Invert a rigid vector
  void inv( const RigidVec3D& r) throw();
  /// Invert a rigid vector
  inline RigidVec3D& Inv( const RigidVec3D& R) throw()
    {this->inv(R); return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  inline RigidVec3D getInverse(void) const throw()
    {return RigidVec3D().Inv(*this);}
  /// invert this matrix (this matrix is modified)
  inline void invert(void) throw()
    {this->inv(*this);}

  /// Apply rigid matrix to a vect
  void Apply(const Vec3<double>& In, Vec3<double>& Out) const throw()
    { RigidMat3D(*this).Apply(In, Out); }
  /// Apply rigid matrix to a vector
  friend inline Vec3<double> operator*(const RigidVec3D& m, const Vec3<double>& v)
    {return RigidMat3D(m)*v;}

  /// Compose rigid vectors
  inline void comp( const RigidVec3D& r1, const RigidVec3D& r2)  throw();
  /// Compose rigid vectors
  inline RigidVec3D& Comp( const RigidVec3D& r1, const RigidVec3D& r2)  throw()
    {comp(r1,r2);return *this;}
  /// Matrix multiplication
  inline RigidVec3D& operator*=(const RigidVec3D& m) throw()
    {comp(*this,m);return *this;}
  /// Matrix multiplication
  friend inline RigidMat3D operator*(const RigidVec3D& m, const RigidVec3D& v)
    {return RigidVec3D().Comp(m,v);}

  /// Jacobien
  void J_Rt_r(const Vec3<double>& x, Mat<3,6,double>& dRx_dr) const throw();
  //@}

  /** @name I/O operations */
  //@{
  /// Print on an output stream
  friend STD_OSTREAM& operator<<( STD_OSTREAM& os, const RigidVec3D& rig) throw()
    {return os<<(Vect<6,double>&) rig;}
  //@}


  /// read a Rigid Vector from a file
  void readV6(IFSTREAM& fin) throw(ErrUnknown);
  /// read a Rigid Vector from a file
  void readV6(const char* FileName) throw(ErrUnknown);
  /// write a Rigid Vector in a file
  void writeV6(OFSTREAM& fout) const throw(ErrUnknown);
  /// write a Rigid Vector in a file
  void writeV6(const char* FileName) const throw(ErrUnknown);

  /// read a Rigid Vector from a file
  void ReadV6(IFSTREAM& fin) throw(ErrUnknown)
     {readV6(fin);}
  /// read a Rigid Vector from a file
  void ReadV6(const char* FileName) throw(ErrUnknown)
     {readV6(FileName);}
  /// write a Rigid Vector in a file
  void WriteV6(OFSTREAM& fout) const throw(ErrUnknown)
     {writeV6(fout);}
  /// write a Rigid Vector in a file
  void WriteV6(const char* FileName) const throw(ErrUnknown)
     {writeV6(FileName);}

  /// read a Rigid Vector from a file
  template <class List>
  static void readV6List(IFSTREAM& fin, List& list)
    {
      while (fin && (!fin.eof())) {
        try {
          list.push_back(RigidVec3D(fin));
        } catch(...) {
          break;
        }
      }
    }
  /// read a Rigid Vector from a file
  template <class List>
  static void readV6List(const char* FileName, List& list)
     {
       IFSTREAM fin(FileName);
       RigidVec3D::readV6List(fin,list);
     }

  /// write a Rigid Vector in a file
  template <class Iterator>
  static void writeV6List(OFSTREAM& fout, Iterator first, Iterator last) throw(ErrUnknown)
    {
      for (Iterator it=first; it!=last; it++) {
        it->writeV6(fout);
      }
    }
  /// write a Rigid Vector in a file
  template <class Iterator>
  void writeV6List(const char* FileName, Iterator first, Iterator last) throw(ErrUnknown)
     {
       OFSTREAM fout(FileName);
       RigidVec3D::writeV6List(fout,first,last);
     }
};

} // namespace

#endif
