#ifndef _ROTATION3D_H_
#define _ROTATION3D_H_


#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Rotation3D.h"
#endif
#include "basics/Constants.h"
#include "linearAlgebra/Vec3.h"
#include "linearAlgebra/Vec4.h"
#include "linearAlgebra/Matrix3x3.h"

#ifdef WIN32
#include "basics/export.h"
#endif


namespace yav {


// forward declarations
class RotVec3D;
class RotMat3D;
class UnitQuat;

typedef RotVec3D RotVect;


/// Conversion from a rotation matrix to a rotation vector
void RotMat3DToRotVec3D( const RotMat3D& rot, RotVec3D& r);
  // no RotMat3DToUnitQuat function
  //
/// Conversion from a rotation vector to a rotation matrix
void RotVec3DToRotMat3D( const RotVec3D& r, RotMat3D& rot);
/// Conversion from a rotation vector to a unit quaternion
void RotVec3DToUnitQuat( const RotVec3D& r, UnitQuat& q);
/// Conversion from a unit quaternion to a rotation vector
void UnitQuatToRotVec3D( const UnitQuat& q, RotVec3D& r);
/// Conversion from a unit quaternion to a rotation matrix
void UnitQuatToRotMat3D( const UnitQuat& q, RotMat3D& rot);


Vec3<double> operator*(const RotMat3D& m, const Vec3<double>& v);
RotMat3D operator*(const RotMat3D& m, const RotMat3D& v);

/*! \class RotMat3D Rotation3D.h linearAlgebra/Rotation3D.h
  \ingroup liblinearAlgebra
  3D rotation Matrix
*/
class RotMat3D : protected Matrix3x3<double> {
public:
  /*! \class ErrNonRotMat Rotation3D.h linearAlgebra/Rotation3D.h
    \ingroup liblinearAlgebra
    Non rotration matrix excpetion
  */
  class ErrNonRotMat : public Exception {
  public:
    // constructor
    ErrNonRotMat(const std::string &from) throw() :
      Exception(from, "Non rotation matrix exception") {};
  };

  /** @name Constructors */
  //@{
  /// default constructor ; initializes to identity matrix
  RotMat3D() throw() : Matrix3x3<double>(VIM_NOINIT)
    {this->Id();}

  /// Copy constructor
  RotMat3D(const RotMat3D & r) throw() : Matrix3x3<double>(r)
    {};
  /// operator =
  RotMat3D& operator=(const RotMat3D & r) throw()
    {this->copy((const Matrix3x3<double>&)r);return *this;}

  friend void RotVec3DToRotMat3D( const RotVec3D& r, RotMat3D& rot); 
  friend void RotMat3DToRotVec3D( const RotMat3D& rot, RotVec3D& r);
  friend void UnitQuatToRotMat3D( const UnitQuat& q, RotMat3D& rot);

  /// construct rotation matrix from a rotation vector
  RotMat3D(const RotVec3D& r) throw() : Matrix3x3<double>(VIM_NOINIT)
    {RotVec3DToRotMat3D(r,*this);}
  /// set rotation matrix from a rotation vector
  RotMat3D& operator=(const RotVec3D& r) throw()
    {RotVec3DToRotMat3D(r,*this);return *this;}

  /// construct rotation matrix from a unit quaternion
  RotMat3D(const UnitQuat& q) throw() : Matrix3x3<double>(VIM_NOINIT)
    {UnitQuatToRotMat3D(q,*this);}
  /// set rotation matrix from a unit quaternion
  RotMat3D& operator=(const UnitQuat& q) throw()
    {UnitQuatToRotMat3D(q,*this);return *this;}

  /** Initialization from 3 column vectors
      @param v1 first column vector
      @param v2 second column vector
      @param v3 third column vector
      @exception NonRotationMatrix the determinant is not close to 1 enough */
  RotMat3D(const Vec3<double>& v1,
           const Vec3<double>& v2,
           const Vec3<double>& v3 ) throw(NonUniqueOptimum,ErrNonRotMat,ErrConvergence) : Matrix3x3<double>(VIM_NOINIT)
    {try {LSQ(Matrix3x3<double>(v1,v2,v3));} catch(ErrConvergence& exc){exc.push("RotMat3D::RotMat3D( Mat<3>& K )");throw exc;}}

  /// construct rotation matrix from an affine matrix (LSQ Vectorial  distance - extrinsic)
  template <class T2>
  explicit RotMat3D(T2 matrix[3][3]) throw(NonUniqueOptimum,ErrNonRotMat,ErrConvergence) : Matrix3x3<double>(VIM_NOINIT)
    {try {LSQ(Matrix3x3<double>(matrix));} catch(ErrConvergence& exc){exc.push("RotMat3D::RotMat3D( Mat<3>& K )");throw exc;}}
  /// construct rotation matrix from an affine matrix (LSQ Vectorial  distance - extrinsic)
  template <class T2>
  explicit RotMat3D(const Matrix3x3<T2>& K) throw(NonUniqueOptimum,ErrNonRotMat,ErrConvergence) : Matrix3x3<double>(VIM_NOINIT)
    {try {LSQ(Matrix3x3<double>(K));} catch(ErrConvergence& exc){exc.push("RotMat3D::RotMat3D( Mat<3>& K )");throw exc;}}
  /// set rotation matrix from an affine matrix (LSQ Vectorial  distance - extrinsic)
  template <class T2>
  RotMat3D& operator=(const Matrix3x3<T2>& K) throw(NonUniqueOptimum,ErrNonRotMat,ErrConvergence)
    {try {LSQ(Matrix3x3<double>(K));} catch(ErrConvergence& exc){exc.push("RotMat3D::operator=( Mat<3>& K )");throw exc;}return *this;}

  /// Destructor
  ~RotMat3D() throw()
    {};
  //@}

  /** @name Data members access */
  //@{
  /// Access to matrix element i,j
  double operator()(const int i, const int j) const  throw(ErrRange)
    {return Matrix3x3<double>::operator()(i,j);}
  /// Access to matrix element i,j
  Vect<3,double> operator[](const int i) const  throw(ErrRange)
    {return Matrix3x3<double>::operator[](i);}
  /// Access to matrix element i,j
  double getValue(const int i, const int j) const  throw(ErrRange)
    {return Matrix3x3<double>::getValue(i,j);}

  ///
  Matrix3x3<double>::resetToIdentity;
  ///
  RotMat3D& Id() throw()
    {resetToIdentity();return *this;}

  // Conversion to matrix (because of protected inheritance, but g++ don't want to use them)
  // ----------------------------------------------------------------------
  //operator const Matrix3x3<double>&() const throw()
  //  { return (const Matrix3x3<double>&) *this;}
  // Conversion to a const Matrix3x3<double> does not work
  // -> Use a member function
  // const Matrix3x3<double>& mat() const throw()
  //   { return (const Matrix3x3<double>&) *this;}
  //@}

  /** @name Mathematical operations */
  //@{
  /** Closest Rotation to an affine matrix (Vectorial  distance - extrinsic)
      @param K affine matrix
      @return residue
   */
  double LSQ(const Matrix3x3<double>& K) throw(NonUniqueOptimum,ErrNonRotMat,ErrConvergence);

  /// Intrinsic Distance between rotation matrices : angle( R1trp.R2)
  friend double Dist(const RotMat3D& R1, const RotMat3D& R2) throw();

  /// Intrinsic norm of rotation matrix : angle( R)
  double Norm() const throw();

  /** Extrinsinc Euclidean Distance between rotation matrices
      as in the Matrix3x3<double> vector space. For rotations :
      dist(R1,R2)^2 = 6 - 2 tr(R1trp.R2) */
  friend double DistEuc( const RotMat3D& R1, const RotMat3D& R2) throw()
    {return Dist( (Matrix3x3<double>&) R1, (Matrix3x3<double>&) R2);}

  ///  Uniform Random 3D Rotation
  void rnd() throw();

  ///  Uniform Random 3D Rotation
  RotMat3D& Rnd() throw()
    {rnd();return *this;}

  /** set this matrix to transposed of a given matrix
      @return this */
  void trp(const RotMat3D& m) throw()
    {((Matrix3x3<double>&)*this).trp((const Matrix3x3<double>&)m);}
  /** set this matrix to transposed of a given matrix
      @return this */
  inline RotMat3D& Trp(const RotMat3D& m) throw()
    {this->trp(m);return *this;}
  /// transpose this matrix (this matrix is modified)
  inline void transpose() throw()
    {this->Trp(*this);}
  /** get this matrix transposed matrix (this matrix is NOT modified)
      @return transposed matrix */
  inline RotMat3D getTransposed() const throw()
    {return RotMat3D().Trp(*this);}

  /// Invert a rotation matrix
  inline void inv(const RotMat3D& R) throw()
    {this->trp(R);}
  /// Invert a rotation matrix
  inline RotMat3D& Inv( const RotMat3D& R) throw()
    {this->inv(R); return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  inline RotMat3D getInverse() const throw()
    {return RotMat3D().Inv(*this);}
  /// invert this matrix (this matrix is modified)
  inline void invert() throw()
    {this->inv(*this);}

  /// Apply rotation matrix to a vect
  void Apply(const Vec3<double>& In, Vec3<double>& Out) const throw()
    {Out.Mul((const Matrix3x3<double>&)*this,In);}
  /// Apply rotation matrix to a vector
  friend inline Vec3<double> operator*(const RotMat3D& m, const Vec3<double>& v)
    {return ((const Matrix3x3<double>&)m)*v;}

  /// Compose rotation matrices
  inline void comp(const RotMat3D& R1, const RotMat3D& R2)  throw()
    {this->Mul(R1,R2);}
  /// Compose rotation matrices
  inline RotMat3D& Comp(const RotMat3D& R1, const RotMat3D& R2)  throw()
    {this->comp(R1,R2); return *this;}
  /// Compose rotation matrices
  inline RotMat3D& operator*=(const RotMat3D& m) throw()
    {comp(*this,m);return *this;}
  /// Compose rotation matrices
  friend inline RotMat3D operator*(const RotMat3D& m, const RotMat3D& v)
    {return RotMat3D().Comp(m,v);}

#define MAX_NB_ITER_MEAN 20
  /// set this as the mean of a list of rotations using iterative computation
  template<class InputIterator>
  void mean(InputIterator first, InputIterator last) throw()
    {
      *this=*first;
      for (int iter=0; iter<MAX_NB_ITER_MEAN; iter++) {
        RotMat3D minv;minv.inv(*this);
        Vec3<double> dm;
        int n=0;
        InputIterator irotmat;
        for (irotmat=first;irotmat!=last;irotmat++){
          dm+=(const Vec3<double>&) RotVec3D(minv*(*irotmat));
          n+=1;
        }
        dm*=1/n;
        if (dm.norm2()<1e-6) break;
        *this*=RotMat3D(dm);
      }
    }

  /// set this as the approximated mean of a list of rotations using direct computation (SVD method)
  template<class InputIterator>
  void meanSVD(InputIterator first, InputIterator last) throw()
    {
      Matrix3x3<double> rm;
      InputIterator irotmat;
      for (irotmat=first;irotmat!=last;irotmat++){
        rm+=(const Matrix3x3<double>&) (*irotmat);
      }
      LSQ(rm);
    }
  // compute covariance of given list of rotations around this
  /// @param first
  /// @param last
  /// @param covar
  template<class InputIterator>
  void covararound(InputIterator first, InputIterator last, Sym<3,double>& covar) const throw()
    {
      Vect<3,double> vtmp;
      Sym<3,double> tmp;
      RotMat3D minv;minv.inv(*this);
      covar.Null();
      int n=0;
      InputIterator irotmat;
      for (irotmat=first;irotmat!=last;irotmat++){
        vtmp=(const Vect<3,double>&) RotVec3D(minv*(*irotmat));
        //cout << vtmp << endl;
        tmp.AutoCorrel(vtmp);
        covar+=tmp;
        n+=1;
      }
      covar*=1/(double)n;
    }
  // compute variance of given list of rotations around this
  /// @param first
  /// @param last
  template<class InputIterator>
  double vararound(InputIterator first, InputIterator last) const throw()
    {
      Vect<3,double> vtmp;
      RotMat3D minv;minv.inv(*this);
      double var=0;
      int n=0;
      InputIterator irotmat;
      for (irotmat=first;irotmat!=last;irotmat++){
        vtmp=(const Vect<3,double>&) RotVec3D(minv*(*irotmat));
        var+=vtmp.Norm2();
        n+=1;
      }
      return var/n;
    }
  // compute standard deviation of given list of rotations around this
  /// @param first
  /// @param last
  template<class InputIterator>
  double sigmaaround(InputIterator first, InputIterator last) const throw()
    {return ::sqrt(vararound(first,last));}
  //@}

  /** @name I/O operations */
  //@{
  /// Print on an output stream
  friend STD_OSTREAM& operator<<(STD_OSTREAM& os, const RotMat3D& rot) throw()
    {return os<<(Matrix3x3<double>&) rot;}
  //@}
};





// **********************************************************************
// 3D Rotation vector
// **********************************************************************
/*! \class RotVec3D Rotation3D.h linearAlgebra/Rotation3D.h
  \ingroup liblinearAlgebra
 */
class RotVec3D : protected Vec3<double> {
public:
  /** @name Constructors */
  //@{
  /// Default constructor
  RotVec3D() throw() : Vec3<double>()
    {};

  /// Copy Constructor
  RotVec3D(const RotVec3D & rv) throw() : Vec3<double>(rv)
    {};
  /// Operator =
  RotVec3D& operator=(const RotVec3D & rv) throw()
    {(Vec3<double>&)(*this)=rv;return *this;}

  friend void RotVec3DToRotMat3D( const RotVec3D& r, RotMat3D& rot);
  friend void RotMat3DToRotVec3D( const RotMat3D& rot, RotVec3D& r );
  friend void RotVec3DToUnitQuat( const RotVec3D& r, UnitQuat& q);
  friend void UnitQuatToRotVec3D( const UnitQuat& q, RotVec3D& r);

  /// constructor from a rotation matrix
  RotVec3D(const RotMat3D& rot) throw() : Vec3<double>(VIM_NOINIT)
    {RotMat3DToRotVec3D(rot,*this);}
  /// conversion from a rotation matrix
  RotVec3D& operator= (const RotMat3D& rot) throw()
    {RotMat3DToRotVec3D(rot,*this);return *this;}

  /// constructor from a unit quaternion
  RotVec3D(const UnitQuat& rot) throw() : Vec3<double>(VIM_NOINIT)
    {UnitQuatToRotVec3D(rot,*this);}
  /// conversion from a unit quaternion
  RotVec3D& operator= (const UnitQuat& rot) throw()
    {UnitQuatToRotVec3D(rot,*this);return *this;}

  /// Constructor from values
  template <class T2>
  RotVec3D(T2 rx, T2 ry, T2 rz) throw() : Vec3<double>(rx, ry, rz)
    {Domain();}
  /// Constructor from Vec3<double>
  template <class T2>
  RotVec3D(const Vec3<T2>& v) throw() : Vec3<double>(v)
    {Domain();}
  /// Constructor from values
  template <class T2>
  RotVec3D(const T2 v[3]) throw() : Vec3<double>(v)
    {Domain();}

  /// Destructor
  ~RotVec3D() throw()
    {};
  //@}

  /** @name Access to data */
  //@{
  /// Access to an element
  double operator()(const int i) const throw(ErrRange) {
    return Vec3<double>::operator()(i);}
  //@}

  /** @name Mathematical operations */
  //@{
  /// Replace vector in definition domain (where norm(v)<Pi)
  void Domain() throw();

  /// Intrinsic Distance between rotation vectors : angle( r1^-1 o r2)
  friend double Dist( const RotVec3D& r1, const RotVec3D& r2) throw();

  /// Intrinsic Norm (distance to Id) for rotation : angle( r ) = || r ||
  double Norm( void) const throw()
    {return Vec3<double>::Norm();}

  ///  Uniform Random 3D Rotation
  void rnd() throw();

  ///  Uniform Random 3D Rotation
  RotVec3D& Rnd() throw()
    {rnd();return *this;}

  /// Invert a rotation matrix
  inline void inv(const RotVec3D& R) throw()
    {this->Mul(-1.,R);}
  /// Invert a rotation matrix
  inline RotVec3D& Inv( const RotVec3D& R) throw()
    {this->inv(R); return *this;}
  /** get this matrix inverse matrix (this matrix is NOT modifyed)
      @return inverse matrix */
  inline RotVec3D getInverse() const throw()
    {return RotVec3D().Inv(*this);}
  /// invert this matrix (this matrix is modified)
  inline void invert() throw()
    {this->inv(*this);}

  /// Compose rotation vectors
  void comp( const RotVec3D& r1, const RotVec3D& r2)  throw();
  /// Compose rotation vectors
  inline RotVec3D& Comp( const RotVec3D& r1, const RotVec3D& r2)  throw()
    {this->comp(r1,r2);return *this;}
  /// Compose rotation matrices
  inline RotVec3D& operator*=(const RotVec3D& m) throw()
    {comp(*this,m);return *this;}
  /// Compose rotation matrices
  friend inline RotVec3D operator*(const RotVec3D& m, const RotVec3D& v)
    {return RotVec3D().Comp(m,v);}

  /// Apply rotation matrix to a vect
  void Apply(const Vec3<double>& In, Vec3<double>& Out) const throw()
    { RotMat3D(*this).Apply(In, Out); }
  /// Apply rotation matrix to a vector
  friend inline Vec3<double> operator*(const RotVec3D& m, const Vec3<double>& v)
    {return RotMat3D(m)*v;}

  /// Jacobien
  void J_Rt_r(const Vec3<double>& x, Matrix3x3<double>& dRx_dr) const throw();
  //@}

  /** @name I/O operations */
  //@{
  /// Print on an output stream
  friend STD_OSTREAM& operator<<( STD_OSTREAM& os, const RotVec3D& rot) throw()
    {return os<<(Vec3<double>&) rot;}
  //@}
};








// **********************************************************************
// Unit Quaternions
// **********************************************************************

void QuatProd(const UnitQuat& q1, const UnitQuat& q2, UnitQuat& q) throw();
void QuatConj(const UnitQuat& q, UnitQuat& barq) throw();
void QuatInv(const UnitQuat& q, UnitQuat& invq) throw();
void TildeLeft(const UnitQuat& q, Matrix<4,double>& Q) throw();
void Anti(const UnitQuat& q, Matrix<4,double>& Q) throw();
void TildeRight(const UnitQuat& q, Matrix<4,double>& Q) throw();

/*! \class UnitQuat Rotation3D.h linearAlgebra/Rotation3D.h
  \ingroup liblinearAlgebra
  Unit Quaternions
*/

class UnitQuat : protected Quaternion {
public:
  /*! \class ErrNonUnitQuat Rotation3D.h linearAlgebra/Rotation3D.h
    \ingroup liblinearAlgebra
   */
  class ErrNonUnitQuat : public Exception {
  public:
    // constructor
    ErrNonUnitQuat(const std::string &from) throw() :
      Exception(from, "Non unit quaternion exception") {};
  };


  /** @name Constructors */
  //@{
  /// Default constructor: set to (1, 0,0,0)
  UnitQuat() throw() : Quaternion(1,0,0,0)
    {};

  /// Init from a Quat (Normalize it)
  UnitQuat(const Vect<4> q) throw(ErrDivZero);

  /// Copy constructor
  UnitQuat(const UnitQuat& q) throw() : Quaternion(q)
    {};
  /// operator=
  UnitQuat& operator=( const UnitQuat& q)
    {*this=(const Quaternion&)q;return *this;}

  friend void RotVec3DToUnitQuat( const RotVec3D& r, UnitQuat& q);
  friend void UnitQuatToRotVec3D( const UnitQuat& q, RotVec3D& r);
  friend void UnitQuatToRotMat3D( const UnitQuat& q, RotMat3D& rot);

  /// constructor from a rotation matrix
  UnitQuat(const RotMat3D& rot) throw() : Quaternion(VIM_NOINIT)
    {RotVec3D r(rot); RotVec3DToUnitQuat( r, *this);}
  /// conversion from a rotation matrix
  UnitQuat& operator= (const RotMat3D& rot) throw()
    {RotVec3D r(rot); RotVec3DToUnitQuat( r, *this); return *this;}

  /// constructor from a rotation matrix
  UnitQuat(const RotVec3D& rot) throw() : Quaternion(VIM_NOINIT)
    {RotVec3DToUnitQuat(rot,*this);}
  /// conversion from a rotation matrix
  UnitQuat& operator= (const RotVec3D& rot) throw()
    {RotVec3DToUnitQuat(rot,*this);return *this;}

  /// Destructor
  ~UnitQuat() throw()
    {};
  //@}

  /** @name Access to data */
  //@{
protected:
  /// Access to the unit quat as a (real, Vector) pair.
  double& RealPart() throw()
    {return val[0];}
  /// Access to the unit quat as a (real, Vector) pair.
  Vec3<double>& VectPart() throw()
    {return (Vec3<double>&) val[1]; }
public:
  /// Access to the unit quat as a (real, Vector) pair.
  double RealPart() const throw()
    {return val[0];}
  /// Access to the unit quat as a (real, Vector) pair.
  const Vec3<double>& VectPart() const throw()
    {return (const Vec3<double>&) val[1]; }
  //@}

  /** @name Mathematical operations */
  //@{
  /// Quaternion product: q = q1 * q2
  //friend void QuatProd(const UnitQuat& q1, const UnitQuat& q2, UnitQuat& q) throw();
  /// Conjugate quaternion: barq = (alpha, -beta)
  //friend void QuatConj(const UnitQuat& q, UnitQuat& barq) throw();
  /// Inverse quaternion: invq = barq for unit quaternions
  //friend void QuatInv(const UnitQuat& q, UnitQuat& invq) throw();
  /// Antisymmetric matrices associated with the Quaternion product:
  //friend void TildeLeft(const UnitQuat& q, Matrix<4,double>& Q) throw();
  /// Antisymmetric matrices associated with the Quaternion product:
  //friend void Anti(const UnitQuat& q, Matrix<4,double>& Q) throw();
  /// Antisymmetric matrices associated with the Quaternion product:
  //friend void TildeRight(const UnitQuat& q, Matrix<4,double>& Q) throw();

  ///  Uniform Random UnitQuat (i.e. 3D Rotation)
  void rnd() throw();
  ///  Uniform Random UnitQuat (i.e. 3D Rotation)
  UnitQuat& Rnd() throw()
    {rnd();return *this;}

  /// Intrinsic Distance between UnitQuat (not rotations !!): angle( q1, q2)
  friend double Dist( const UnitQuat& q1, const UnitQuat& q2) throw();

  /// Intrinsic Distance between Rotation Quaternions: angle( q1, q2)*2 mod Pi
  friend double DistRot( const UnitQuat& q1, const UnitQuat& q2) throw();


  /// Extrinsinc Euclidean Distance between UnitQuat (distance in the Vect<4> vector space).
  friend double DistEuc( const UnitQuat& q1, const UnitQuat& q2) throw()
    { return Dist((Quaternion&) q1,(Quaternion&) q2); }


  /// Compose rotation quaternions
  UnitQuat& Comp( const UnitQuat& q1, const UnitQuat& q2)  throw()
    {QuatProd( q1,q2,*this); return *this;}

  /// Invert a rotation matrices
  UnitQuat& Inv( const UnitQuat& q)  throw()
    {QuatConj(q, *this); return *this;}

  /// approximated mean rotation using direct computation (quaternions method)
  template<class InputIterator>
  void meanQuat(InputIterator first, InputIterator last) throw()
    {
      Vec4<double> qm;
      InputIterator irotquat;
      for (irotquat=first;irotquat!=last;irotquat++){
        qm+=(const Vec4<double>&) (*irotquat);
      }
      //TODO:verifier divisions par zero
      ((Vec4<double>&)*this).Mul(qm,1/qm.norm());
    }
  //@}

  /** @name I/O operations */
  //@{
  /// Print on an output stream
  friend STD_OSTREAM& operator<<( STD_OSTREAM& os, const UnitQuat& q) throw()
    {return os<<(Quaternion&) q;}
  //@}
};


} // namespace yav

#endif
