// General comments :
// Adapted from Vec*.h and Matrix*.h in liblinearAlgebra and from libXLinalg
//  in general, methods from liblinearAlgebra start with a lower-case letter and
//  returns void, and methods from libXLinalg start with a upper-case letter
//  and returns reference to the object. Most functions are defined in these
//  2 different ways.
// Some functions must be redefined for each child class (constructors,
//  functions which return type is something like Vec<D,t>) => I use a
//  perl-script to do this easily. Please ask me
//  when making any change in Vec*.[hC] or Mat*.[hC] files.
// Designed to compile with gcc and MS-VisualC++6.0. Consequences :
//  MS-VC++ doesn't accept template class specialisation
//   => we use derivation
//   exemple : we can't specialise Vect<3,T> in order to declare some
//    3d-specific functions => we derive a class Vec3<T>
//  MS-VC++ does not resolve expressions such as vec[2] when both
//   operator[] and operator (const T*) are defined => suppressed (const T*)
//   use getValue() instead
//                                                        00/10 sgranger
//
//
// MAY HAVE A STRANGE BEHAVIOR ON INTEGERS
//                                                        98/06 jcl
//
// Compiling with -DRANGE_CHECK will provide range checking on arrays access.
//  This is designed to be used in a verification step only since this adds
//  an extra computation time.
// Compiling with -DCHECK enables all checkings (defaults)
//                -DNO_CHECK disables all checkings
// Compiling with -DUNINIT_ZERO will NOT initialize memory to zero
//                                                        98/08 xpennec
//

#ifndef _VEC_H
#define _VEC_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/Vec.h"
#endif
#include <linearAlgebra/cVec.h>

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {


#ifndef __GNUC__
#undef min
#undef max
template <class T> T min(T x, T y){if ( x < y) return x; else return y;}
template <class T> T max(T x, T y){if ( x > y) return x; else return y;}
#endif

#ifndef __GNUC__
#define FRIEND_TEMPLATE_ARG
#else
#define FRIEND_TEMPLATE_ARG <>
#endif

#define CHECK
#ifdef NO_CHECK
#undef CHECK
#endif

#ifdef CHECK
#ifndef RANGE_CHECK
#define RANGE_CHECK
#endif
#endif

///
enum VectInitMode {VIM_NOINIT, VIM_INIT};


/*! \class ErrConvergence Vec.h linearAlgebra/Vec.h
  \ingroup liblinearAlgebra
 */
class ErrConvergence : public Exception {
public:
  // constructor
  ErrConvergence(const std::string &from) throw() :
    Exception(from, "Algorithm did not converge") {};
  virtual ~ErrConvergence() throw() {};
};

/*! \class NonUniqueOptimum Vec.h linearAlgebra/Vec.h
  \ingroup liblinearAlgebra
 */
class NonUniqueOptimum : public Exception {
public:
  // constructor
  NonUniqueOptimum(const std::string &from) throw() :
    Exception(from, "Solution is not unique") {};
  virtual ~NonUniqueOptimum() throw() {};
};

/*! \class ErrMatSingular Vec.h linearAlgebra/Vec.h
  \ingroup liblinearAlgebra
 */
class ErrMatSingular : public Exception {
 public:
  // constructor
  ErrMatSingular(const std::string &from) throw() :
    Exception(from, "Singular matrix exception") {};
  virtual ~ErrMatSingular() throw() {};
};


template <Index D, class T> class Vect;
template <Index L, Index C, class T> class Mat;
template <Index D, class T> class Matrix;
template <Index D, class T> class Sym;


/** @name Vectorial redefined friend functions */
//@{
/** unary +
    @param p vector
    @return +p */
template <Index D, class T> vec_Inline Vect<D, T> operator+(const Vect<D, T> &p) throw()
  {return p;}
/** unary -
    @param p vector
    @return -p */
template <Index D, class T> vec_Inline Vect<D, T> operator-(const Vect<D, T> &p) throw()
  {return Vect<D, T>(VIM_NOINIT).Minus(p);}
/** vector sum
    @param p first vector
    @param q second vector
    @return p + q */
template <Index D, class T> vec_Inline Vect<D, T> operator+(const Vect<D, T> &p, const Vect <D, T> &q) throw()
  {return Vect<D, T>(VIM_NOINIT).Add(p,q);}
/** vector substraction
    @param p first vector
    @param q second vector
    @return p - q */
template <Index D, class T> vec_Inline Vect<D, T> operator-(const Vect<D, T> &p, const Vect <D, T> &q) throw()
  {return Vect<D, T>(VIM_NOINIT).Sub(p,q);}
/** vector multiplication by scalar
    @param p vector
    @param a scalar
    @return ap */
template <Index D, class T> vec_Inline Vect<D, T> operator*(const Vect<D, T> &p, const T a) throw()
  {return Vect<D, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar multiplication
    @param a scalar
    @param p vector
    @return ap */
template <Index D, class T> vec_Inline Vect<D, T> operator*(const T a, const Vect<D, T> &p) throw()
  {return Vect<D, T>(VIM_NOINIT).Mul(p,a);}
/** vector scalar division
    @param p vector
    @param a scalar
    @return p / a */
template <Index D, class T> vec_Inline Vect<D, T> operator/(const Vect<D, T> &p, const T a) throw(ErrDivZero)
  {return Vect<D, T>(VIM_NOINIT).Div(p,a);}
/** dot product
    @param u first vector
    @param v second vector
    @return u.v */
template <Index D, class T> vec_Inline T dot(const Vect<D, T>& u, const Vect<D, T> &v) throw()
  {return Vect<D, T>::dot(u,v);}
/// Square Euclidian Distance
template <Index D, class T> vec_Inline T dist2( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::dist2(u,v);}
/// Euclidian Distance
template <Index D, class T> vec_Inline T dist( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::dist(u,v);}
/// L1 Distance
template <Index D, class T> vec_Inline T distL1( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::distL1(u,v);}
/// Linf Distance
template <Index D, class T> vec_Inline T distLinf( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::distLinf(u,v);}
/** dot product
    @param p first vector
    @param q second vector
    @return p.q */
template <Index D, class T> vec_Inline T DotProd(const Vect<D, T> &p, const Vect<D, T> &q) throw()
  {return Vect<D, T>::dot(p,q);}
/// Square Euclidian Distance
template <Index D, class T> vec_Inline T Dist2( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::dist2(u,v);}
/// Euclidian Distance
template <Index D, class T> vec_Inline T Dist( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::dist(u,v);}
/// L1 Distance
template <Index D, class T> vec_Inline T DistL1( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::distL1(u,v);}
/// Linf Distance
template <Index D, class T> vec_Inline T DistLinf( const Vect<D, T>& u, const Vect<D, T>& v) throw()
  {return Vect<D, T>::distLinf(u,v);}
/** operator == */
template <Index D, class T> vec_Inline bool operator == (const Vect<D, T>& u, const Vect<D, T> &v) throw()
  {return Vect<D, T>::isEqual(u,v);}
/** operator != */
template <Index D, class T> vec_Inline bool operator != (const Vect<D, T>& u, const Vect<D, T> &v) throw()
  {return Vect<D, T>::isDifferent(u,v);}
/** operator < */
template <Index D, class T> vec_Inline bool operator < (const Vect<D, T>& u, const Vect<D, T> &v) throw()
  {return Vect<D, T>::isSmaller(u,v);}
/** operator > */
template <Index D, class T> vec_Inline bool operator > (const Vect<D, T>& u, const Vect<D, T> &v) throw()
  {return Vect<D, T>::isGreater(u,v);}
/** operator <= */
template <Index D, class T> vec_Inline bool operator <= (const Vect<D, T>& u, const Vect<D, T> &v) throw()
  {return Vect<D, T>::isSmallerOrEqual(u,v);}
/** operator >= */
template <Index D, class T> vec_Inline bool operator >= (const Vect<D, T>& u, const Vect<D, T> &v) throw()
  {return Vect<D, T>::isGreaterOrEqual(u,v);}
//@}

#ifdef __GNUC__
template <Index D, Index K, class T> vec_Inline Vect<D,T> operator*(const Mat<D,K,T>& m, const Vect<K,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, Index K, class T> vec_Inline Vect<D,T> operator*(const Vect<K,T>& v, const Mat<K,D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}
#else
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Mat<D,1,T>& m, const Vect<1,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Vect<1,T>& v, const Mat<1,D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Mat<D,2,T>& m, const Vect<2,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Vect<2,T>& v, const Mat<2,D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Mat<D,3,T>& m, const Vect<3,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Vect<3,T>& v, const Mat<3,D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Mat<D,4,T>& m, const Vect<4,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Vect<4,T>& v, const Mat<4,D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Mat<D,5,T>& m, const Vect<5,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Vect<5,T>& v, const Mat<5,D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Mat<D,6,T>& m, const Vect<6,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Vect<6,T>& v, const Mat<6,D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}
#endif

template <Index D, class T> vec_Inline Vect<D,T> operator*(const Sym<D,T>& m, const Vect<D,T>& v) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(m,v);}
template <Index D, class T> vec_Inline Vect<D,T> operator*(const Vect<D,T>& v, const Sym<D,T>& m) throw()
  {return Vect<D,T>(VIM_NOINIT).Mul(v,m);}

template <Index D, class T> STD_OSTREAM& operator<<(STD_OSTREAM &s, const Vect<D,T> &v) throw();
template <Index D, class T> STD_ISTREAM& operator>>(STD_ISTREAM &s, Vect<D,T> &v) throw();


/*! \class Vect Vec.h linearAlgebra/Vec.h
  \ingroup liblinearAlgebra
  template for a vector of dimension D and type T */
template <Index D, class T=double> class Vect
{
public:
  /** @name Constructors */
  //@{
  //  /** @name Vectorial redefined constructors */
  //  //@{
  //  /** default constructor - initializes to 0 if UNINIT_ZERO not defined */
  //  vec_Inline Vect() throw() : _Vect<D,T>()
  //    {}
  //  /** constructor - initializes to 0 if b==VIM_INIT */
  //  explicit vec_Inline Vect(VectInitMode b) throw() : _Vect<D,T>(b)
  //    {}
  //  //@}
  vec_Inline Vect() throw()
    {
#ifndef UNINIT_ZERO
      resetTo(0);
#endif
    }
  /// Constructor : initializes to 0 if b==VIM_INIT
  vec_Inline Vect(VectInitMode b) throw()
    {if (b==VIM_INIT) resetTo(0);}

  //  /** @name Vector redefined constructors */
  //  //@{
#ifdef  __GNUC__
  //  /** copy constructor
  //      @param v vector to copy */
  //  vec_Inline Vect(const Vect<D,T>& v) throw() : _Vect<D,T>(v)
  //    {}
  //  /** operator =
  //      @param v vector to copy
  //      @return this */
  //  vec_Inline Vect<D, T>& operator=(const Vect<D,T>& v) throw()
  //    {copy(v);return *this;}
#endif
  //  /** copy constructor with cast
  //      @param v vector to copy */
  //  template<class T2> vec_Inline Vect(const Vect<D,T2>& v) throw() : _Vect<D,T>(v)
  //    {}
  //  /** operator = with cast
  //      @param v vector to copy
  //      @return this */
  //  template<class T2> vec_Inline Vect<D, T>& operator=(const Vect<D,T2>& v) throw()
  //    {copy(v);return *this;}
  //  

  //  /** constructor: initialize to vector (coord[0], coord[1], ...)
  //      @param coord vector coordinates array */
  //  template<class T2> vec_Inline Vect(const T2 coord[D]) throw() : _Vect<D,T>(coord)
  //    {}
  //  
  //  /** constructor: initializes element i with 1, other with 0 (canonical basis vector i) */
  //  explicit vec_Inline Vect(int i) throw(ErrRange) : _Vect<D,T>(i)
  //    {}
  //  //@}
  /** @name Vector redefined constructors */
  //@{
#ifndef WIN32
  /** copy constructor
      @param v vector to copy */
  vec_Inline Vect(const Vect<D,T>& v) throw()
    {copy(v);}
  /** operator =
      @param v vector to copy
      @return this */
  vec_Inline Vect<D,T>& operator=(const Vect<D,T>& v) throw()
    {copy(v);return *this;}
#endif
  /** copy constructor with cast
      @param v vector to copy */
  template<class T2> vec_Inline Vect(const Vect<D,T2>& v) throw()
    {copy(v);}
  /** operator = with cast
      @param v vector to copy
      @return this */
  template<class T2> vec_Inline Vect<D, T>& operator=(const Vect<D,T2>& v) throw()
    {copy(v);return *this;}

#ifndef WIN32
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  vec_Inline Vect(const T coord[D]) throw()
    {setValues(coord);}
#endif
  /** constructor: initialize to vector (coord[0], coord[1], ...)
      @param coord vector coordinates array */
  template<class T2> vec_Inline Vect(const T2 coord[D]) throw()
    {setValues(coord);}

  /** constructor: initializes element i with 1, other with 0 (canonical basis vector i) */
  explicit vec_Inline Vect(int i) throw(ErrRange)
    {basisVector(i);}
  //@}

  vec_Inline T getVal(int i) const
    {return val[i];}
  vec_Inline void setVal(int i, T v)
    {val[i]=v;}

  /** @name Data members access */
  //@{
  //   /** copy as a member function
  //       @param v vector to copy */
  //  void copy(const Vect<D,T>& v) throw()
  //     {for (Index i=0; i<D; i++) val[i] = v.getVal(i);}
  /** copy with cast as a member function
      @param v vector to copy */
  template<class T2> vec_Inline void copy(const Vect<D,T2>& v) throw()
    {for (Index i=0; i<D; i++) val[i] = (T) v.getVal(i);}

  /** set each element of this vector to value
    @param value elements new value */
  void resetTo(T value) throw()
    {for (Index i=0; i<D; i++) val[i]=value;}
  /// set this vector to the null vector
  vec_Inline void  resetToNull(void) throw()
    {resetTo(0);}

  //   /** setValues with a T[D] array */
  //   vec_Inline void setValues(const T v[D]) throw()
  //     { for (Index i=0; i<D; i++) val[i]=v[i]; }
  /** setValues with a T2[D] array */
  template<class T2> vec_Inline void setValues(const T2 v[D]) throw()
    {for (Index i=0; i<D; i++) val[i]=(T) v[i];}

  /** returns a T* (usefull for some extern functions, like OGL ones) */
  vec_Inline T* getValues(void) throw()
    {return val;}
  /** returns a const T* (usefull for some extern functions, like OGL ones) */
  vec_Inline const T* getValues(void) const throw()
    {return val;}

  /** cast vector in a T*
   */
  vec_Inline operator const T*() const throw()
    {return (const T*) val;}

  /** initializes element i with 1, other with 0 (canonical basis vector i)
      @param i component number (0<= i < D) */
  void basisVector(int i) throw(ErrRange)
    {vec_basisVector(val,D,i);}

//   /** returns ith component of this vector
//       @param i component number (0<= i < D) */
//   T getValue(int i) const throw(ErrRange);
//   /** set ith component
//       @param i component number (0<= i < D) */
//   void setValue(int i, T v) throw(ErrRange);

  /** returns ith component of this vector
      @param i component number (0<= i < D) */
  T& operator[](int i) throw(ErrRange)
     {return vec_getRefVal(val,D,i);}
  /** returns ith component of this vector
      @param i component number (0<= i < D) */
  //  T operator[](int i) const throw(ErrRange);
  /** returns ith component of this vector
      @param i component number (0<= i < D) */
  T& operator()(int i) throw(ErrRange)
     {return vec_getRefVal(val,D,i);}
  /** returns ith component of this vector (const version)
      @param i component number (0<= i < D) */
  T operator()(int i) const throw(ErrRange)
     {return vec_getRefVal(val,D,i);}

  /// consider this vector as a col matrice
  vec_Inline const Mat<D,1,T>& ColMat() const throw()
    { return (Mat<D,1,T>&) *this;}
  /// consider this vector as a row matrice
  vec_Inline const Mat<1,D,T>& RowMat() const throw()
    { return (Mat<1,D,T>&) *this;}
  //@}

  /** @name mathematical operators */
  //@{
  /** squared norm
      @return ||this||^2 */
  vec_Inline T norm2(void) const throw()
    {return vec_norm2(val,D);}
  /** norm
    @return ||this|| */
  vec_Inline T norm(void) const throw()
    {return vec_norm(val,D);}
  /** L1 norm
      @return sum abs(v_i) */
  vec_Inline T normL1(void) const throw()
    {return vec_normL1(val,D);}
  /** Linf norm
    @return max abs(v_i) */
  vec_Inline T normLinf(void) const throw()
    {return vec_normLinf(val,D);}
  /** squared norm
      @return ||this||^2 */
  vec_Inline T Norm2(void) const throw()
    {return vec_norm2(val,D);}
  /** norm
    @return ||this|| */
  vec_Inline T Norm(void) const throw()
    {return vec_norm(val,D);}
  /** L1 norm
      @return sum abs(v_i) */
  vec_Inline T NormL1(void) const throw()
    {return vec_normL1(val,D);}
  /** Linf norm
    @return max abs(v_i) */
  vec_Inline T NormLinf(void) const throw()
    {return vec_normLinf(val,D);}
  
  /** add q to this vector */
  vec_Inline void add(const Vect<D, T> &q) throw()
    {vec_add(val,D,q.val);}
  /** substract q to this vector */
  vec_Inline void sub(const Vect<D, T> &q) throw()
    {vec_sub(val,D,q.val);}
  /** set each coordinate of this vector to product of this and q same coordinate */
  vec_Inline void cmul(const Vect<D, T> &q) throw()
    {vec_cmul(val,D,q.val);}
  /** multiply this vector with a */
  vec_Inline void mul(const T a) throw()
    {vec_mul(val,D,a);}
  /** divide this vector with a */
  vec_Inline void div(const T a) throw(ErrDivZero)
    {vec_div(val,D,a);}
  
  /** set this vector to -p */
  vec_Inline void opposite(const Vect<D, T> &p) throw()
    {vec_opposite(val,D,p.val);}
  /** set this vector to p+q */
  vec_Inline void add(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {vec_add(val,D,p.val,q.val);}
  /** set this vector to p-q */
  vec_Inline void sub(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {vec_sub(val,D,p.val,q.val);}
  /** set each coordinate of this vector to product of p and q same coordinate */
  vec_Inline void cmul(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {vec_cmul(val,D,p.val,q.val);}
  /** set this vector to p*a */
  vec_Inline void mul(const Vect<D, T> &p, const T a) throw()
    {vec_mul(val,D,p.val,a);}
  /** set this vector to p/a */
  void div(const Vect<D, T> &p, const T a) throw(ErrDivZero)
    {vec_div(val,D,p.val,a);}
  
  /** randomize vector */
  vec_Inline void rnd(T min = (T)-1., T max = (T)1.) throw()
    {vec_rnd(val,D,min,max);}
  /** randomize vector : isotropic gaussian centred noise */
  vec_Inline void rndgaussian(T standard_deviation = (T)1.) throw()
    {vec_rndgaussian(val,D,standard_deviation);}
  
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates */
  vec_Inline void min(const Vect<D, T> &p) throw()
    {vec_min(val,D,p.val);}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates */
  vec_Inline void max(const Vect<D, T> &p) throw()
    {vec_max(val,D,p.val);}
  
  /** dot product
      @param p first vector
      @param q second vector
      @return p.q */
  vec_Inline static T dot(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {return vec_dot(D,p.val,q.val);}
  /// Square Euclidian Distance
  vec_Inline static T dist2( const Vect<D, T>& p, const Vect<D, T>& q) throw()
    {return vec_dist2(D,p.val,q.val);}
  /// Euclidian Distance
  vec_Inline static T dist( const Vect<D, T>& p, const Vect<D, T>& q) throw()
    {return vec_dist(D,p.val,q.val);}
  /// L1 Distance
  vec_Inline static T distL1( const Vect<D, T>& p, const Vect<D, T>& q) throw()
    {return vec_distL1(D,p.val,q.val);}
  /// Linf Distance
  vec_Inline static T distLinf( const Vect<D, T>& p, const Vect<D, T>& q) throw()
    {return vec_distLinf(D,p.val,q.val);}
  
  /// operator ==
  vec_Inline static bool isEqual(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {return vec_isEqual(D,p.val,q.val);}
  /// operator !=
  vec_Inline static bool isDifferent(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {return vec_isDifferent(D,p.val,q.val);}
  /// operator <
  vec_Inline static bool isSmaller(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {return vec_isSmaller(D,p.val,q.val);}
  /// operator >
  vec_Inline static bool isGreater(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {return vec_isSmaller(D,q.val,p.val);}
  /// operator <=
  vec_Inline static bool isSmallerOrEqual(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {return !vec_isSmaller(D,q.val,p.val);}
  /// operator >=
  vec_Inline static bool isGreaterOrEqual(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {return !vec_isSmaller(D,p.val,q.val);}
  /** @name Vectorial redefined menber functions */
  //@{
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vect<D, T>& Minus(const Vect<D, T> &p) throw()
    {vec_opposite(val,D,p.val);return *this;}
  /** set this vector to p+q
      @return this = p + q */
  vec_Inline Vect<D, T>& Add(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {vec_add(val,D,p.val,q.val);return *this;}
  /** set this vector to p-q
      @return this = p - q */
  vec_Inline Vect<D, T>& Sub(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {vec_sub(val,D,p.val,q.val);return *this;}
  /** set each coordinate of this vector to product of this and p same coordinate
      @return this = this.*q */
  vec_Inline Vect<D, T>& CMul(const Vect<D, T> &q) throw()
    {vec_cmul(val,D,q.val);return *this;}
  /** set each coordinate of this vector to product of q and p same coordinate
      @return this = p.*q */
  vec_Inline Vect<D, T>& CMul(const Vect<D, T> &p, const Vect<D, T> &q) throw()
    {vec_cmul(val,D,p.val,q.val);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vect<D, T>& Mul(const Vect<D, T> &p, const T a) throw()
    {vec_mul(val,D,p.val,a);return *this;}
  /** set this vector to p*a
      @return this = p*a */
  vec_Inline Vect<D, T>& Mul(const T a, const Vect<D, T> &p) throw()
    {vec_mul(val,D,p.val,a);return *this;}
  /** set this vector to p/a
      @return this = p/a */
  vec_Inline Vect<D, T>& Div(const Vect<D, T> &p, const T a) throw(ErrDivZero)
    {vec_div(val,D,p.val,a);return *this;}
  /** adds vector q to vector p
      @param q vector to add
      @return p + q */
  vec_Inline Vect<D, T>& operator+=(const Vect<D, T> &q) throw()
    {vec_add(val,D,q.val);return *this;}
  /** substracts vector q from vector p
      @param q vector to substract
      @return p - q */
  vec_Inline Vect<D, T>& operator-=(const Vect<D, T> &q) throw()
    {vec_sub(val,D,q.val);return *this;}
  /** multiplies vector p by scalar a
      @param a scalar
      @return ap */
  vec_Inline Vect<D, T>& operator*=(const T a) throw()
    {vec_mul(val,D,a);return *this;}
  /** divides vector p by scalar a
      @param a scalar
      @return p / a */
  vec_Inline Vect<D, T>& operator/=(const T a) throw(ErrDivZero)
    {vec_div(val,D,a);return *this;}
  /** randomize vector
      @return this */
  vec_Inline Vect<D, T>& Rnd(T min = (T)-1., T max = (T)1.) throw()
    {vec_rnd(val,D,min,max);return *this;}
  /** randomize vector : isotropic gaussian centred noise
      @return this */
  vec_Inline Vect<D, T>& RndGaussian(T standard_deviation = (T)1.) throw()
    {vec_rndgaussian(val,D,standard_deviation);return *this;}
  /** takes the X,Y or Z value of p if they are less than its own X,Y,Z coordinates
      @return _Vect<D,T>(min(x,p.x),min(y,p.y),...) */
  vec_Inline Vect<D, T>& Min(const Vect<D, T>& p) throw()
    {vec_min(val,D,p.val);return *this;}
  /** takes the X,Y or Z value of p if they are greater than its own X,Y,Z coordinates
      @return _Vect<D,T>(max(x,p.x),max(y,p.y),...) */
  vec_Inline Vect<D, T>& Max(const Vect<D, T>& p) throw()
    {vec_max(val,D,p.val);return *this;}
  /** set each element of this vector to value
      @param v value
      @return this */
  vec_Inline Vect<D, T>& Reset(T v) throw()
    {resetTo(v);return *this;}
  /** set this vector to null vector
      @return this */
  vec_Inline Vect<D, T>& Null() throw()
    {resetTo(0);return *this;}
  /** set this vector to mean of a given list of vectors
      @return this */
  template<class InputIterator>
  vec_Inline Vect<D, T>& Mean(InputIterator first, InputIterator last) throw()
    {mean(first,last);return *this;}
  //@}

  /** @name I/O operations */
  //@{
  /** writes vector on outstream as "(x, y, z)"
      @param s outstream
      @param c vector
      @return s */
  //friend STD_OSTREAM& operator<<  FRIEND_TEMPLATE_ARG(STD_OSTREAM &s, const Vect<D,T> &c) throw();
  /** reads vector from instream as "x y z"
      @param s instream
      @param c vector
      @return s */
  //friend STD_ISTREAM& operator>>  FRIEND_TEMPLATE_ARG(STD_ISTREAM &s, Vect<D,T> &c) throw();
  //@}

  /** @name Matrix operations */
  //@{
#ifndef WIN32
  template <Index K> vec_Inline Vect<D,T>& Mul(const Mat<D,K,T> &m, const Vect<K,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<K,1,T>&)v);return *this;}
  template <Index K> vec_Inline Vect<D,T>& Mul(const Vect<K,T> &v, const Mat<K,D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,K,T>&)v,m);return *this;}
#else
  vec_Inline Vect<D,T>& Mul(const Mat<D,1,T> &m, const Vect<1,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<1,1,T>&)v);return *this;}
  vec_Inline Vect<D,T>& Mul(const Vect<1,T> &v, const Mat<1,D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,1,T>&)v,m);return *this;}
  vec_Inline Vect<D,T>& Mul(const Mat<D,2,T> &m, const Vect<2,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<2,1,T>&)v);return *this;}
  vec_Inline Vect<D,T>& Mul(const Vect<2,T> &v, const Mat<2,D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,2,T>&)v,m);return *this;}
  vec_Inline Vect<D,T>& Mul(const Mat<D,3,T> &m, const Vect<3,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<3,1,T>&)v);return *this;}
  vec_Inline Vect<D,T>& Mul(const Vect<3,T> &v, const Mat<3,D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,3,T>&)v,m);return *this;}
  vec_Inline Vect<D,T>& Mul(const Mat<D,4,T> &m, const Vect<4,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<4,1,T>&)v);return *this;}
  vec_Inline Vect<D,T>& Mul(const Vect<4,T> &v, const Mat<4,D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,4,T>&)v,m);return *this;}
  vec_Inline Vect<D,T>& Mul(const Mat<D,5,T> &m, const Vect<5,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<5,1,T>&)v);return *this;}
  vec_Inline Vect<D,T>& Mul(const Vect<5,T> &v, const Mat<5,D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,5,T>&)v,m);return *this;}
  vec_Inline Vect<D,T>& Mul(const Mat<D,6,T> &m, const Vect<6,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<6,1,T>&)v);return *this;}
  vec_Inline Vect<D,T>& Mul(const Vect<6,T> &v, const Mat<6,D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,6,T>&)v,m);return *this;}
#endif

  vec_Inline Vect<D,T>& Mul(const Sym<D,T> &m, const Vect<D,T> &v) throw()
    {((Mat<D,1,T>*)this)->Mul(m,(const Mat<D,1,T>&)v);return *this;}
  vec_Inline Vect<D,T>& Mul(const Vect<D,T> &v, const Sym<D,T> &m) throw()
    {((Mat<1,D,T>*)this)->Mul((const Mat<1,D,T>&)v,m);return *this;}
  //@}

  /** @name Statistics */
  //@{
  /// set this as the mean of a list of vectors
  template<class InputIterator>
  void mean(InputIterator first, InputIterator last) throw()
    {
      this->resetToNull();
      int n=0;
      InputIterator ivec;
      for (ivec=first;ivec!=last;ivec++){
        *this+=*ivec;
        n+=1;
      }
      *this*=1/n;
    }
  // compute covariance of given list of vectors around this
  template<class InputIterator>
  void covarAround(InputIterator first, InputIterator last, Sym<D,T>& covar) const throw()
    {
      Vect<D,T> vtmp;
      Sym<D,T> tmp;
      int n=0;
      InputIterator ivec;
      for (ivec=first;ivec!=last;ivec++){
        vtmp.Sub(*ivec,*this);
        tmp.AutoCorrel(vtmp);
        covar+=tmp;
        n+=1;
      }
      covar*=1/(double)n;
    }
  // compute covariance of given list of vectors around this
  template<class InputIterator>
  Sym<D,T> covarAround(InputIterator first, InputIterator last) const throw()
    {
      Sym<D,T> tmp;covarAround(first, last, tmp);
      return tmp;
    }
  // compute variance of given list of vectors around this
  template<class InputIterator>
  double varAround(InputIterator first, InputIterator last) const throw()
    {
      double var=0;
      int n=0;
      InputIterator ivec;
      for (ivec=first;ivec!=last;ivec++){
        var+=Dist2(*ivec,*this);
        n+=1;
      }
      return var/n;
    }
  // compute standard deviation of given list of rotations around this
  template<class InputIterator>
  double sigmaAround(InputIterator first, InputIterator last) const throw()
    {return ::sqrt(varAround(first,last));}
  //@}

  static void Test() throw();

protected:
  /// vector data
  T val[D];
};

} // namespace yav

/*][** NODOC ****************************************************************/
/****************************************************************************/

/*]]**************************************************************************/

namespace yav {

template <Index D, class T> vec_Inline STD_OSTREAM& operator<<(STD_OSTREAM &s, const Vect<D,T> &c) throw()
{
  s<<"("<<c(0);
  for (Index i=1; i<D; i++) s<<", "<< c(i);
  return s<< ")";
}

template <Index D, class T> vec_Inline STD_ISTREAM& operator>>(STD_ISTREAM &s,Vect<D,T> &c) throw()
{
  for (Index i=0; i<D; i++) s>> c(i);
  return s;
}

template <Index D, class T> void Vect<D,T>::Test() throw()
{
  float Af[D];double Ad[D];int Ai[D];
  T A[D];for (Index k=0; k<D; k++) {A[k]=k+1;Af[k]=k+1;Ad[k]=k+1;Ai[k]=k+1;}
  Vect<D,float> df;Vect<D,double> dd;Vect<D,int> di;

  Vect<D,T> a(A),b(0),c(VIM_INIT),d(a),af(Af),ad(Ad),ai(Ai),ef(df),ed(dd),ei(di);
  T v;char TName[100];Xlinalg_Class_Name(v,TName);
  COUT << "Vec<" <<D << "," << TName << ">" << ENDL;
  COUT << "Constructors" << ENDL;
  COUT << "a must be 1,2,... : " << a << ENDL;
  COUT << "af must be 1,2,... : " << af << ENDL;
  COUT << "ad must be 1,2,... : " << ad << ENDL;
  COUT << "ai must be 1,2,... : " << ai << ENDL;
  COUT << "b must be 1,0,... : " << b << ENDL;
  COUT << "c must be 0,0,... : " << c << ENDL;
  COUT << "d must be as a (copy) : " << d << ENDL;
  COUT << "ef must be 0 : " << ef << ENDL;
  COUT << "ed must be 0 : " << ed << ENDL;
  COUT << "ei must be 0 : " << ei << ENDL;

  COUT << "operator=" << ENDL;
  d=c;COUT << "d=c : " << d << ENDL;
  d=df;COUT << "d=df : " << d << ENDL;
  d=dd;COUT << "d=dd : " << d << ENDL;
  d=di;COUT << "d=di : " << d << ENDL;

  COUT << "Access to Data" << ENDL;
  d.copy(b);COUT << "d.copy(b)  : " << d << ENDL;
  d[0]=1;COUT << "d[0]=1 : d[0] " << d[0] << ENDL;
  d(0)=1;COUT << "d(0)=1 : d(0) " << d(0) << ENDL;
  d.Null();COUT << "d.Null() : " << d << ENDL;
  d.resetTo(2);COUT << "d.resetTo(2) : " << d << ENDL;
  T*Dd=d.getValues();
  COUT << "getValues() : " << Dd[0] << " " << Dd[1] << " " << Dd[2] << ENDL;
  const T*E=d;
  COUT << "(const T*) d : " << E[0] << " " << E[1] << " " << E[2] << ENDL;
  COUT << "a.ColMat : " << a.ColMat() << ENDL;
  COUT << "a.RowMat : " << a.RowMat() << ENDL;

  COUT << "Math" << ENDL;
  COUT << "d.Minus(a) : " << d.Minus(a) << ENDL;
  COUT << "d.Add(a,b) : " << d.Add(a,b) << ENDL;
  COUT << "d.Sub(a,b) : " << d.Sub(a,b) << ENDL;
  COUT << "d.Mul(a,2) : " << d.Mul(a,(T)2) << ENDL;
  COUT << "d.Mul(2,a) : " << d.Mul((T)2,a) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;
  COUT << "d.Div(a,2) : " << d.Div(a,(T)2) << ENDL;

  COUT << "+a : " << +a << ENDL;
  COUT << "-a : " << -a << ENDL;
  COUT << "a+b : " << a+b << ENDL;
  COUT << "a-b : " << a-b << ENDL;
  COUT << "a*2 : " << a*((T)2) << ENDL;
  COUT << "2*a : " << ((T)2)*a << ENDL;
  COUT << "a/2 : " << a/((T)2) << ENDL;

  d.copy(b);d+=a;COUT << "d=b; d+=a : " << d << ENDL;
  d.copy(b);d-=a;COUT << "d=b; d-=a : " << d << ENDL;
  d.copy(b);d*=(T)2;COUT << "d=b; d*=2 : " << d << ENDL;
  d.copy(b);d/=(T)2;COUT << "d=b; d/=2 : " << d << ENDL;

  d.copy(b);COUT << "d=b; d==b : " << (d==b) << ENDL;
  d.copy(b);COUT << "d=b; d!=b : " << (d!=b) << ENDL;

  COUT << "dot(a,b) : " << dot(a,b) << ENDL;
  COUT << "DotProd(a,b) : " << DotProd(a,b) << ENDL;
  COUT << "norm2(a) : " << a.norm2() << ENDL;
  COUT << "norm(a) : " << a.norm() << ENDL;
  COUT << "Norm2(a) : " << a.Norm2() << ENDL;
  COUT << "Norm(a) : " << a.Norm() << ENDL;
  COUT << "NormL1(a) : " << a.NormL1() << ENDL;
  COUT << "NormLinf(a) : " << a.NormLinf() << ENDL;
  COUT << "Dist2(a,b) : " << Dist2(a,b) << ENDL;
  COUT << "Dist(a,b) : " << Dist(a,b) << ENDL;
  COUT << "DistL1(a,b) : " << DistL1(a,b) << ENDL;
  COUT << "DistLinf(a,b) : " << DistLinf(a,b) << ENDL;

  d.resetTo(2); d.min(a); COUT << "d.resetTo(2); d.min(a) : " << d << ENDL;
  d.resetTo(2); d.max(a); COUT << "d.resetTo(2); d.max(a) : " << d << ENDL;

  d.Rnd(2); COUT << "d.Rnd(2) : " << d << ENDL;
  d.RndGaussian(2); COUT << "d.RndGaussian(2) : " << d << ENDL;

  {Mat<D,1,T> m;Vect<1,T> g; COUT << "Mat<D,1,T> m*Vect<1,T> g : " << m*g << "  Mul : " << c.Mul(m,g) << ENDL;}
  {Mat<1,D,T> m;Vect<1,T> g; COUT << "Vect<1,T> g*Mat<D,1,T> m : " << g*m << "  Mul : " << c.Mul(g,m) << ENDL;}
  {Mat<D,2,T> m;Vect<2,T> g; COUT << "Mat<D,2,T> m*Vect<2,T> g : " << m*g << "  Mul : " << c.Mul(m,g) << ENDL;}
  {Mat<2,D,T> m;Vect<2,T> g; COUT << "Vect<2,T> g*Mat<D,2,T> m : " << g*m << "  Mul : " << c.Mul(g,m) << ENDL;}
  {Mat<D,3,T> m;Vect<3,T> g; COUT << "Mat<D,3,T> m*Vect<3,T> g : " << m*g << "  Mul : " << c.Mul(m,g) << ENDL;}
  {Mat<3,D,T> m;Vect<3,T> g; COUT << "Vect<3,T> g*Mat<D,3,T> m : " << g*m << "  Mul : " << c.Mul(g,m) << ENDL;}
  {Mat<D,4,T> m;Vect<4,T> g; COUT << "Mat<D,4,T> m*Vect<4,T> g : " << m*g << "  Mul : " << c.Mul(m,g) << ENDL;}
  {Mat<4,D,T> m;Vect<4,T> g; COUT << "Vect<4,T> g*Mat<D,4,T> m : " << g*m << "  Mul : " << c.Mul(g,m) << ENDL;}
  {Mat<D,5,T> m;Vect<5,T> g; COUT << "Mat<D,5,T> m*Vect<5,T> g : " << m*g << "  Mul : " << c.Mul(m,g) << ENDL;}
  {Mat<5,D,T> m;Vect<5,T> g; COUT << "Vect<5,T> g*Mat<D,5,T> m : " << g*m << "  Mul : " << c.Mul(g,m) << ENDL;}
  {Mat<D,6,T> m;Vect<6,T> g; COUT << "Mat<D,6,T> m*Vect<6,T> g : " << m*g << "  Mul : " << c.Mul(m,g) << ENDL;}
  {Mat<6,D,T> m;Vect<6,T> g; COUT << "Vect<6,T> g*Mat<D,6,T> m : " << g*m << "  Mul : " << c.Mul(g,m) << ENDL;}

  COUT << ENDL;
}


} // namespace yav


#endif
