#ifndef _CVEC_H
#define _CVEC_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4290)
#endif

#ifndef WIN32
#pragma interface "linearAlgebra/cVec.h"
#endif
#include <math.h>
#include <stdlib.h>     // for rand() and srand() declaration
#include <time.h>       // for time() declaration
#include <basics/Exception.h>
#include <basics/Bool.h>
#include <basics/Constants.h>
#include <basics/Architecture.h>

#ifdef WIN32
#include "basics/export.h"
#endif

namespace yav {

#define vec_Inline inline

  /*! \class ErrRange cVec.h linearAlgebra/cVec.h
    \ingroup liblinearAlgebra
   */
class ErrRange: public Exception {
 public:
  // constructor
  ErrRange(const std::string &from) throw() :
    Exception(from, "Range error (access to an out-of-bound element)") {};
  virtual ~ErrRange() throw() {};
};

vec_Inline void RndInit(){srand((long int) time( NULL) );}
vec_Inline double UniformRnd() {return rand()/(RAND_MAX+1.0);}
vec_Inline double GaussianRnd() {return ( ::sqrt(-2.0*log(UniformRnd()))*cos(2*M_PI*UniformRnd()));}

template<class T> void Xlinalg_Small_Val(T&);
template<class T> void Xlinalg_Class_Name(T&,char*);

///
typedef unsigned int Index;



template <class T> vec_Inline void vec_add(T*val, Index D, const T*q) throw()
{
  for (Index i=0; i<D; i++) val[i]+=q[i];
}
template <class T> vec_Inline void vec_sub(T*val, Index D, const T*q) throw()
{
  for (Index i=0; i<D; i++) val[i]-=q[i];
}
template <class T> vec_Inline void vec_cmul(T*val, Index D, const T*q) throw()
{
  for (Index i=0; i<D; i++) val[i]*=q[i];
}
template <class T> vec_Inline void vec_mul(T*val, Index D, const T a) throw()
{
  for (Index i=0; i<D; i++) val[i]*=a;
}
template <class T> vec_Inline void vec_div(T*val, Index D, const T a) throw(ErrDivZero)
{
  T smallval; Xlinalg_Small_Val(smallval);
  if ( ::fabs(a) < smallval * vec_norm(val,D) )
    throw ErrDivZero("vec_div(T*val, Index D, const T a)");
  T b=1/a;
  for (Index i=0; i<D; i++) val[i]*=b;
}

template <class T> vec_Inline void vec_opposite(T*val, Index D, const T*p) throw()
{
  for (Index i=0; i<D; i++) val[i]=-p[i];
}
template <class T> vec_Inline void vec_add(T*val, Index D, const T*p, const T*q) throw()
{
  for (Index i=0; i<D; i++) val[i]=p[i]+q[i];
}
template <class T> vec_Inline void vec_add(T*val, Index D, const T*p, const T q) throw()
{
  for (Index i=0; i<D; i++) val[i]=p[i]+q;
}

template <class T> vec_Inline void vec_sub(T*val, Index D, const T*p, const T*q) throw()
{
  for (Index i=0; i<D; i++) val[i]=p[i]-q[i];
}
template <class T> vec_Inline void vec_cmul(T*val, Index D, const T*p, const T*q) throw()
{
  for (Index i=0; i<D; i++) val[i]=p[i]*q[i];
}
template <class T> vec_Inline void vec_mul(T*val, Index D, const T*p, const T a) throw()
{
  for (Index i=0; i<D; i++) val[i]=p[i]*a;
}
template <class T> vec_Inline void vec_div(T*val, Index D, const T*p, const T a) throw(ErrDivZero)
{
  T smallval; Xlinalg_Small_Val(smallval);
  if ( ::fabs(a) < smallval * vec_norm(p,D) )
    throw ErrDivZero("vec_div(T*val, Index D, const T*p, const T a)");

  for (Index i=0; i<D; i++) val[i]=p[i]/a;
}

template <class T> vec_Inline void vec_rnd(T*val, Index D, T min, T max) throw()
{
  T range = max - min;
  for( Index i=0; i<D; i++) val[i] = (T)(min + UniformRnd()*range);
}

template <class T> vec_Inline void vec_rndgaussian(T*val, Index D, T standard_deviation) throw()
{
  for( Index i=0; i<D; i++) val[i] = (T)(GaussianRnd()*standard_deviation);
}

// min and max
template <class T> vec_Inline void vec_min(T*val, Index D, const T*p) throw()
{
  for (Index i=0; i<D; i++) if (p[i]<val[i]) val[i]=p[i];
}
template <class T> vec_Inline void vec_max(T*val, Index D, const T*p) throw()
{
  for (Index i=0; i<D; i++) if (p[i]>val[i]) val[i]=p[i];
}

template <class T> vec_Inline T vec_norm2(const T*val, Index D) throw()
{
  T sum = 0, tmp;
  for( Index i=0; i < D; i++){
    tmp = val[i];
    sum += tmp * tmp;
  }
  return sum;
}
template <class T> vec_Inline T vec_norm(const T*val, Index D) throw()
{
        return ::sqrt(vec_norm2(val,D));
}
template <class T> vec_Inline T vec_normL1(const T*val, Index D) throw()
{
  T sum = 0;
  for( Index i=0; i < D; i++){
    sum += ::fabs(val[i]);
  }
  return sum;
}
template <class T> vec_Inline T vec_normLinf(const T*val, Index D) throw()
{
  T max = 0, tmp;
  for( Index i=0; i < D; i++){
    tmp = ::fabs(val[i]);
    if (tmp>max) max=tmp;
  }
  return max;
}

template <class T> vec_Inline T vec_dot(Index D, const T*u, const T*v) throw()
{
  T sum = 0;
  for( Index i=0; i < D; i++){
    sum += u[i]*v[i];
  }
  return sum;
}

template <class T> vec_Inline T vec_dist2(Index D, const T*u, const T*v) throw()
{
  T sum = 0, tmp;
  for( Index i=0; i < D; i++){
    tmp = u[i]-v[i];
    sum += tmp * tmp;
  }
  return sum;
}
template <class T> vec_Inline T vec_dist(Index D, const T*u, const T*v) throw()
{
        return ::sqrt(vec_dist2(D,u,v));
}
template <class T> vec_Inline T vec_distL1(Index D, const T*u, const T*v) throw()
{
  T sum = 0;
  for( Index i=0; i < D; i++){
    sum += ::fabs(u[i]-v[i]);
  }
  return sum;
}
template <class T> vec_Inline T vec_distLinf(Index D, const T*u, const T*v) throw()
{
  T max = 0, tmp;
  for( Index i=0; i < D; i++){
    tmp = ::fabs(u[i]-v[i]);
    if (tmp>max) max=tmp;
  }
  return max;
}
// boolean operators
template <class T> vec_Inline bool vec_isEqual(Index D, const T*p, const T*q) throw()
{
  for (Index i=0; i<D; i++) if (p[i]!=q[i]) return false;
  return true;
}
template <class T> vec_Inline bool vec_isDifferent(Index D, const T*p, const T*q) throw()
{
  for (Index i=0; i<D; i++) if (p[i]!=q[i]) return true;
  return false;
}
template <class T> vec_Inline bool vec_isSmaller(Index D, const T*p, const T*q) throw()
{
  for (Index i=0; i<D-1; i++){
    if (p[i]>q[i]) return false;
    if (p[i]<q[i]) return true;
  }
  if (p[D-1]<q[D-1]) return true;
  return false;
}










template <class T> vec_Inline void vec_basisVector(T*val, Index D, int i) throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (i>=(int)D) )
    throw  ErrRange("Vect::Vect(Index)");
#endif
  for (int j=0; j<(int)D; j++) val[j]=(T) (i==j);
}

// template <class T> vec_Inline T vec_getConstVal(T*val, Index D, int i) const throw(ErrRange)
// {
// #ifdef RANGE_CHECK
//   if ( (i<0) || (i>=(int)D) )
//     throw  ErrRange("Vect::getValue(Index)");
// #endif
//   return val[i];
// }

// template <class T> vec_Inline void vec_setValue(T*val, Index D, int i, T v) throw(ErrRange)
// {
// #ifdef RANGE_CHECK
//   if ( (i<0) || (i>=(int)D) )
//     throw  ErrRange("Vect::setValue(Index)");
// #endif
//   val[i]=v;
// }

template <class T> vec_Inline T& vec_getRefVal(T*val, Index D, int i) throw(ErrRange)
{
#ifdef RANGE_CHECK
  if ( (i<0) || (i>=(int)D) )
    throw  ErrRange("Vect::operator[](Index)");
#endif
  return val[i];
}

/****************************************************************************/
/* I/O operations                                                           */
/****************************************************************************/
template <class T> vec_Inline STD_OSTREAM& vec_out(STD_OSTREAM &s, const T*c, Index D) throw()
{
  s<<"("<<c[0];
  for (Index i=1; i<D; i++) s<<", "<< c[i];
  return s<< ")";
}

template <class T> vec_Inline STD_ISTREAM& vec_in(STD_ISTREAM &s, T*c, Index D) throw()
{
  for (Index i=0; i<D; i++) s>> c[i];
  return s;
}

}

#endif

