#ifndef ACTIVEVERTEX_H
#define ACTIVEVERTEX_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif

#pragma interface "graphics/ActiveVertex.h"
#include <math.h>
#include <basics/Bool.h>
#include <basics/AutoLink.h>
#include <linearAlgebra/Vec2.h>
#include <linearAlgebra/Vec3.h>
#include <graphics/RangeData.h>
#include <vector>

namespace yav {

/** \class ActiveVertex  
    \brief deformable model vertex dynamic behaviour
    \ingroup Graphics */
template <class T> class ActiveVertex {
 public:
  /// the total size of the force array
#define nbForceType 6

  /// the types of forces
  enum ForceType {
    INTERNAL=0,
    EXTERNAL = 1,
    AVERAGED_EXTERNAL = 2,
    GLOBAL = 3,
    GLOBALLY_CONSTRAINED = 4,
    TOTAL = 5
  };

 protected:
  /// the vertex position at the previous time step
  T previousPosition;
  /// the reference shape position for shape control
  T shapePosition;
  /// the force vertex array
  T force[nbForceType];
  /// weight of internal forces
  double alpha;
  /// default internal force weight
  static const double defaultAlpha;
  /// weight of external forces
  double beta;
  /// default external force weight
  static const double defaultBeta;
  /// vertex weight in global force computation
  double weight;
  /// confidence in data closest point
  double dataConfidence;
  /// the damping factor
  double damping;
  /// the default damping factor
  static const double defaultDamping;
  /// the damping factor computed from the internal force
  double elasticDamping;
  /// the default elastic damping factor
  static const double defaultElasticDamping;
  /// the maximum distance at which a point can be attracted
  double forceRange;
  /// If a vertex is moving or not
  bool fixed;
  /// scale for computing the internal and external force
  unsigned int rigidity;
  /// default rigidity value
  static const unsigned int defaultRigidity;
  /// The maximum rigidity value for a vertex
  unsigned int maximumRigidity;
  /// The locality parameter
  double locality;
  /// The default locality parameter
  static const double defaultLocality;
  /// The information stored for computing the closest point
  RangeInformation *rangeInformation;
 public:
  /** storing additional scalar in vertex : should only be used for 
      internal force  computation */
  std::vector<double> scalarArray;
  /** storing additional vectors in vertex : should only be used for 
      internal force  computation */
  std::vector<T > vectorArray;
  /// set the external force range
  void setForceRange(const double theRange)
    { forceRange=theRange;}
  /// set the internal force weight
  void setAlpha(const double theAlpha)
    // { alpha = theAlpha < 0.0 ? 0.0 : (theAlpha > 1.0 ? 1.0 : theAlpha);}
  { alpha = theAlpha;}
  /// set the external force weight
  void setBeta(const double theBeta)
    // { beta = theBeta < 0.0 ? 0.0 : (theBeta > 1.0 ? 1.0 : theBeta);}
  { beta = theBeta;}
  /// set the damping value 
  void setDamping(const double theDamping)
    { damping=theDamping;}
  /// set the elastic damping value 
  void setElasticDamping(const double theDamping)
    { elasticDamping=theDamping;}
  /// returns the weight used to compute the global transformation
  inline double getWeight(void) const { return weight;}
  /// sets the weight used to compute the global transformation
  inline void setWeight(const double w) { 
    weight = w < 0.0 ? 0.0 : (w > 1.0 ? 1.0 : w);}
    /// returns the weight used to compute the global transformation
  inline double getDataConfidence(void) const { return dataConfidence;}
  /// sets the weight used to compute the global transformation
  inline void setDataConfidence(const double c) { 
    dataConfidence = c;}
  /// set the rigidity value
  void setRigidity(const unsigned int theRigidity) {
    rigidity = theRigidity > maximumRigidity ? maximumRigidity : theRigidity;
  }
  /// set the maximum rigidity value
  void setMaximumRigidity(const unsigned int theRigidity)
    { maximumRigidity = theRigidity;}
  /// set the locality value
  virtual void setLocality(const double l)
    { locality = l < 0.0 ? 0.0 : (l > 1.0 ? 1.0 : l); }
  /// fix a vertex ie prevent it from moving
  virtual void fixVertex()
    { fixed=true;}
  /// free a vertex ie allow it to move
  virtual void unfixVertex()
    { fixed=false;}
  /// set the mobility of a vertex
  virtual void setMobility(const bool m) {
    fixed=m;}
  /// add a force of type ft
  inline void addForce(const T& f, ForceType ft) { force[ft]+= f; }
   /// multiply a force of type ft with a scalar
  inline void mulForce(const double sc, ForceType ft) { force[ft]*= sc; }
  /// set force of type ft
  inline void setForce(const T& f, ForceType ft) { force[ft] = f; }

  /// returns the external force range
  inline double getForceRange() const
    { return(forceRange);}
  /// returns the internal force weight
  inline double getAlpha() const
    { return(alpha);}
  /// returns the external force weight
  inline double getBeta() const
    { return(beta);}
  /// returns the damping factor
  inline double getDamping() const
    { return(damping);}
  /// returns the elastic damping factor
  inline double getElasticDamping() const
    { return(elasticDamping);}
  /// returns the rigidity value
  inline unsigned int getRigidity() const
    { return(rigidity);}
  /// returns the maximum allowed rigidity value
  inline unsigned int getMaximumRigidity() const
    { return(maximumRigidity);}
  /// returns the locality of a vertex
  inline double getLocality() const 
    { return(locality);}  
  /// returns the mobility boolean
  inline bool isFixed() const
    {return(fixed); }
  inline bool getMobility() const
    {return(fixed); }

  /// returns the previous position of a vertex
  inline const T& getPreviousPosition() const {
    return(previousPosition);  }
  /// returns the shape position of a vertex
  inline const T& getShapePosition() const {
    return(shapePosition);  }
  /// returns the force applied on a vertex
  inline const T& getForceValue(const ForceType ft) const {
    return(force[(unsigned int) ft]);  }
  /// set new shape position
  inline void setShapePosition(const T &p) {
    shapePosition = p;  }
  /// set new previous position
  inline void setPreviousPosition(const T &p) {
    previousPosition = p;  }
  /// returns a pointer on a range information stored at that vertex
  inline RangeInformation *getRangeInformation() const {
    return(rangeInformation);  }
  /// replace the current information with a given RangeInformation *
  inline void setRangeInformation( RangeInformation *information)  {
    if (rangeInformation)
      delete rangeInformation;
    rangeInformation=information;}
  /// Constructor that initializes each vertex value
  ActiveVertex(void);
  /// Destructor : free the rangeInformation pointer if necessary
  virtual ~ActiveVertex();
};


typedef ActiveVertex < Vec2<double> > ActiveVertex2D ;
typedef ActiveVertex < Vec3<double> > ActiveVertex3D ;





/** \class InternalForceType
    \brief internal force computation descriptor
    \ingroup Graphics */
template <class MT, class MV> class InternalForceType : public 
AutoLink <InternalForceType<MT, MV> > {
 public: 
	 typedef void (MT::*MESHFUNCTION) (void);
	 typedef void (MV::*VERTEXFUNCTION) (void);
	 typedef void (MV::*VERTEXIMPLICITFUNCTION) (unsigned int);

 private:
  /** the function to be called when this type of
      internal force is assigned to all vertices */
  MESHFUNCTION setGlobalFunction;
  /** the function to be called when this type of
      internal force is assigned to a vertex */
  VERTEXFUNCTION setLocalFunction;
  /** the function to be called just before computing 
      the internal force on a single vertex. Usually
      computes geometric entities such a normal vector or curvature */
  MESHFUNCTION precomputeGlobalFunction;
  /** the function to be called just before computing 
      the internal force on all vertices. Usually
      computes geometric entities such a normal vector or curvature */
  VERTEXFUNCTION precomputeLocalFunction;
  /** the function to be called for computing 
      the internal force on all vertices */
  MESHFUNCTION computeGlobalFunction;
  /** the function to be called for computing 
      the internal force on a vertex */
  VERTEXFUNCTION computeFunction;
  /** the function to be called for updating the rigidity matrix 
      corresponding to an internal force on all vertices */
  MESHFUNCTION computeGlobalImplicitFunction;
  /** the function to be called for updating the rigidity matrix 
      corresponding to an internal force on a vertex. The index of the
      vertex in the rigidity matrix as given as parameter*/
  VERTEXIMPLICITFUNCTION computeImplicitFunction;
  /// the name of the internal force 
  TclString name;

 public:
  /// returns the setGlobal function
  MESHFUNCTION getSetGlobalFunction(void) const {
    return setGlobalFunction;
  }
  /// returns the compute function
  VERTEXFUNCTION getComputeFunction(void) const {
    return computeFunction;
  }
  /// returns the precomputeGlobal function
  MESHFUNCTION getPrecomputeGlobalFunction(void) const {
    return precomputeGlobalFunction;
  }
  /// returns the precomputeLocal function
  VERTEXFUNCTION getPrecomputeLocalFunction(void) const {
    return precomputeLocalFunction;
  }
  /// returns the name of internal force
  char *getForceName() const {
    return((char *) name);  }
  /// constructor that assign all functions
  InternalForceType( void (MT::*setG) (void),void (MV::*setL) (void),
		     void (MT::*pg) (void),
		     void (MV::*pl) (void),void (MT::*cgf) (void),
		     void (MV::*cf) (void),void (MT::*cgif) (void),
		     void (MV::*cif) (unsigned int),
		     TclString n): setGlobalFunction(setG),
  setLocalFunction(setL),precomputeGlobalFunction (pg), 
  precomputeLocalFunction(pl) ,computeGlobalFunction(cgf),computeFunction(cf),
  computeGlobalImplicitFunction(cgif), computeImplicitFunction(cif),name(n){ }

  /// destructor
  virtual ~InternalForceType() {}

  /** returns force type that matches given name
      @param name researched force name
      @return 0 if force type does no exist, pointer on class otherwise */
  static InternalForceType *getByName(const char *name);

};

/** \class InternalForceItem
    \ingroup Graphics
*/
template <class T, class V> class InternalForceItem :
  public ListElement <InternalForceItem<T,V> > {

  /// an internal force can be applied on all vertices or on a vertex
  enum FunctionScope {GLOBAL,LOCAL};
  /// the scope of that force
  FunctionScope functionScope;
  // the type of internal force
  InternalForceType<T,V> *internalForceType;
  // the vertex where localInternalForceFunction applies
  V *vertex;

 public:
  // constructor for internal force with a global scope
  InternalForceItem(InternalForceType<T,V> *);
  // constructor for internal force with a local scope
  InternalForceItem(InternalForceType<T,V> *, V *);
  /// returns the scope of the internal force 
  FunctionScope getForceScope() const {
    return(functionScope);}
  /// returns the type of internal force 
  InternalForceType<T,V> *getForceType() const {
    return(internalForceType);}
  
};

} // namespace yav


#include <graphics/ActiveVertex.txx>

#endif
