#ifndef CAMERA_H
#define CAMERA_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifdef GRAPHICS

#pragma interface "graphics/Camera.h"
#ifndef WIN32
#include <sys/time.h>
#endif
#include <basics/Bool.h>
#include <linearAlgebra/Vec2.h>
#include <basics/TclString.h>
#include <framework/Activable.h>
#include <graphics/CameraTypes.h>

namespace yav {

/// Default scene margin
#define DEFAULT_MARGIN 0
/// default camera radius
#define DEFAULT_CAMERA_RADIUS 1.0
/// default camera scale
#define DEFAULT_CAMERA_SCALE 1.0
/// default camera width
#define DEFAULT_CAMERA_WIDTH 400
/// default camera height
#define DEFAULT_CAMERA_HEIGHT 400

#define AUTO_TIMING 200000
#define TRACKBALL 0.8
#define SQR_TRACKBALL (TRACKBALL*TRACKBALL)

/** \class Camera
    \brief cameras parent class
    \ingroup Graphics
*/
template <class VEC, class ROT, class SCENE, class DATA, class CAMERA>
class Camera : public CameraTypes {
 public:
  /// the type of function used to translate items of a Data
  typedef void (DATA::*TranslationSubObjectFunction)(const VEC& );
  /// the type of function used to rotate items of a Data
  typedef void (DATA::*RotationSubObjectFunction)(const ROT & );
  /// the type of function used to scale items of a Data3D
  typedef void (DATA::*ScaleSubObjectFunction)(double  );

 protected:
  /// the scene this camera looks at
  SCENE *scene;
  /// field of view center
  VEC center;  
  /// radius of a sphere of center "center" that encloses the scene
  double radius;
  /// scale factor
  double scale;
  /// height in pixel of the image plane
  unsigned int width;
  /// width in pixel of the image plane
  unsigned int height;
  /// margin to define the viewport
  unsigned int margin;
  /// background color
  float backgroundColor[3];
  /// what to do when  the left button is pushed
  SELECTION_MODE selectionMode;
  /// whether the camera should decide when to focus on scene
  bool autoFocus;
  /// was the camera focus done at least once ?
  bool focusedYet;
  /// wether to display selected object bounding box
  bool showBoundingBox;


  /// function called for translation when motionMode==MOVE_CAMERA_OBJECT 
  TranslationSubObjectFunction translationFunction;
  /// function called for translation when motionMode==MOVE_CAMERA_OBJECT 
  RotationSubObjectFunction rotationFunction;
  /// function called for scale when motionMode==MOVE_CAMERA_OBJECT 
  ScaleSubObjectFunction scaleFunction;
  

  /// flag for displaying a boundingbox
  CameraTypes::RedrawOrigin bboxRedrawFlag;
  /// should we move the object or the scene ?
  MOTION_MODE motionMode;

  /// each mouse button state
  static BUTTON_STATE B[3];
  
  /// bounding box color
  float boundingBoxColor[3];
  /// bounding box line width
  float boundingBoxWidth;
  /// selected object
  DATA *selectedObject;

  ///viewing mode used (rotation or navigation)
  MOVE_MODE moveMode;

  



  /// last mouse click instant 
  static LONGINT clickTime;
  /// camera last render duration
  static LONGINT renderDuration;
  /// camera last click duration
  static LONGINT clickDuration;

  /// trackball click point when a button was pressed
  static VEC pressPoint;
  /// last trackball position
  static VEC lastPoint;
 protected:
  /** get trackball location function of mouse pointer location
      @param x pixel X coordinate 
      @param y pixel Y coordinate
      @return trackball vector */
  virtual VEC getTrackballVector(int x, int y) const = 0;
  /** get trackball rotation from mouse pointer move and store
      results in autoRotation matrix
      @param a mouse click point
      @param b mouse release point */
  virtual ROT getTrackballRotation(const VEC& a, const VEC& b) = 0;
  /** get trackball scale from mouse pointer move and store
      results in autoRotation matrix
      @param a mouse click point
      @param b mouse release point */
  virtual double getTrackballScale(const VEC& a, const VEC& b) = 0;
  /** get trackball translation from mouse pointer move
      @param a mouse click point
      @param b mouse release point
      @return translation vector */
  virtual VEC getTrackballTranslation(const VEC& a, const VEC& b) = 0;
   /** get trackball translation from mouse pointer move in the Z direction
      @param a mouse click point
      @return translation vector */
  virtual VEC getTrackballZTranslation(const double a) = 0;


 public:
  /** @name Constructor */
  //@{
  /** creates a new camera
      @param scn
      @param h height in pixel size
      @param w width in pixel size
      @param scale default scale value */
  Camera(SCENE *scn = 0,
	 unsigned int h = DEFAULT_CAMERA_HEIGHT,
	 unsigned int w = DEFAULT_CAMERA_WIDTH,
	 double scale = DEFAULT_CAMERA_SCALE);
  //@}

  /** @name Destructor */
  //@{
  virtual ~Camera();
  //@}

  /// returns dynamic camera type
  virtual CAMERA *getType(void) { return dynamic_cast<CAMERA *>(this); }

  /// returns camera's scene
  virtual SCENE *getScene(void) const { return scene; }
  /// returns scene center
  VEC getCenter(void) const { return center; }
  /// returns scene radius
  double getRadius(void) const { return radius; }
  /// returns camera scale
  double getScale(void) const { return scale; }
  /// set camera scale
  void setScale(const double s)  { scale=s; }
  /// returns camera width (in pixels)
  unsigned int getWidth(void) const { return width; }
  /// returns camera height (in pixels)
  unsigned int getHeight(void) const { return height; }
  /// returns margin
  unsigned int getMargin(void) const { return margin; }
  /// returns camera autofocus flag
  bool getAutoFocus(void) const { return autoFocus; }
  /** returns camera background color
      @param color color index to return
      @return color value (between 0.0 and 1.0) */
  float getBackgroundColor(BASE_COLOR color) { return backgroundColor[color]; }
  /** returns camera background color (each component value is between 0.0 and 1.0)
      @param r returned red value
      @param g returned green value
      @param b returned blue value */
  void getBackgroundColor(float &r, float &g, float &b) {
    r = backgroundColor[BASE_RED]; 
    g = backgroundColor[BASE_GREEN]; 
    b = backgroundColor[BASE_BLUE]; 
  }
  /** set background color (each component is between 0 and 1)
      @param r red component
      @param g green component
      @param b blue component */
  virtual void setBackgroundColor(float r, float g, float b);
  /// set camera move mode
  void setMoveMode(MOVE_MODE mm) { moveMode = mm; }
  /** set camera center
      @param c new center */
  void setCenter(VEC c) { center = c; }  
  /** set the camera radius */
  void setRadius(double r) { radius = r;}
  /** set the camera width */
  void setWidth(unsigned int w) { width = w;}
  /** set the camera height */
  void setHeight(unsigned int h) { height = h;}
  /// returns camera motion mode
  MOTION_MODE getMotionMode(void) const { return motionMode; }
  /// returns selection mode
  SELECTION_MODE getSelectionMode() const { return(selectionMode); }
 /** set selection mode
      @param mm Camera selection mode */
  void setSelectionMode(SELECTION_MODE mm) { selectionMode = mm; }
  /// set motion mode to move camera
  void setMotionMode(MOTION_MODE mm) { motionMode = mm; }

  /// display selected object bounding box
  virtual void displayBoundingBox();
  // hides all bounding boxes
  virtual void hideBoundingBox();
  /// returns show bounding box flag
  bool getShowBoundingBox() const { return(showBoundingBox);}
  /** set bounding box color (each component is between 0 and 1)
      @param r red component
      @param g green component
      @param b blue component */
  void setBoundingBoxColor(float r, float g, float b) {
    boundingBoxColor[BASE_RED] = (r < 0.0f ? 0.0f : (r > 1.0f ? 1.0f : r));
    boundingBoxColor[BASE_GREEN] = (g < 0.0f ? 0.0f : (g > 1.0f ? 1.0f : g));
    boundingBoxColor[BASE_BLUE] = (b < 0.0f ? 0.0f : (b > 1.0f ? 1.0f : b));
    redraw();
  }
  /** retrieve bounding box color (each component is between 0 and 1)
      @param r red component returned
      @param g green component returned
      @param b blue component  returned */
  void getBoundingBoxColor(float &r, float &g,float &b) const {
    r = boundingBoxColor[BASE_RED];
    g = boundingBoxColor[BASE_GREEN];
    b = boundingBoxColor[BASE_BLUE];
  }
  /** set bounding box line width (should be > 0.0)
      @param w line width */
  void setBoundingBoxWidth(float w) {
    if(w > 0.0) {
      boundingBoxWidth = w;
      redraw();
    }
  }
  /** returns bounding box line width
      @param w line width returned */
  void getBoundingBoxWidth(float &w) const { w = boundingBoxWidth;}
  /// returns next bounding box redraw kind
  CameraTypes::RedrawOrigin getBboxRedrawFlag() const { return(bboxRedrawFlag); }
  /// set next bounding box redraw kind
  void setBboxRedrawFlag(const CameraTypes::RedrawOrigin flag) {
    bboxRedrawFlag = flag;
  }
  /** empty virtual function invoked to draw a boundingBox
      @param m object whose bounding box should be drawn */
  virtual void drawBoundingBox(DATA *m) const {}

   /** retrieve selected object
      @return selected object (can be 0) */
  virtual DATA *getSelectedObject(void) const {
    return selectedObject;
  }
  /** select a new object
      @param obj object to select */
  virtual void setSelectedObject(DATA *obj);
  /** unselect an object
      @param obj object to unselect */
  virtual void unsetSelectedObject (DATA *obj);
  /** set translation function on subobject
      @param sf new translation function */  
  void setTranslationFunction(TranslationSubObjectFunction sf) {
    translationFunction = sf;
  }
  /** set rotation function on subobject
      @param sf new rotation function */  
  void setRotationFunction(RotationSubObjectFunction sf) {
    rotationFunction = sf;
  }
  /** set scale function on subobject
      @param sf new scale function */  
  void setScaleFunction(ScaleSubObjectFunction sf) {
    scaleFunction = sf;
  }
  
	  
  /** subobject selection method
      @param vec mouse click point
      @param cam clicked camera */
  void select(const Vec2<double> vec, CAMERA *cam){}
 
 
  /** translates selected object
      @param t translation vector */
  virtual void translateObject(VEC &t);
  /** scales selected object
      @param s scale factor */
  virtual void scaleObject(double s);
  /** rotates selected object
      @param r rotation matrix */
  virtual void rotateObject(const ROT &r);

  /// focus if needed
  virtual void focus(void); 
  /// focus whatever autofocus says
  virtual void doFocus(void) = 0;
  /** activates auto focus (if on == true) or desactivates autofocus
      (on == false) */
  void setAutoFocus(bool on) { autoFocus = on; }

  /// camera render callback
  virtual void redraw(void);
  
  /// resize camera to current height and width
  virtual void resize(void);
  /** resize camera
      @param w new camera width
      @param h new camera height */
  virtual void resize(unsigned int w, unsigned int h);

  /** rotates camera
      @param rot rotation matrix */
  virtual void rotateCamera(ROT &rot) = 0;
  /** translates camera center
      @param t translation vector */
  virtual void translateCamera(VEC &t) = 0;
  /** scales camera
      @param s scale */
  virtual void scaleCamera(double s) = 0;

  /** camera callback to catch trackball rotation
      @param rot rotation matrix */
  virtual void rotateCameraObject(ROT &rot) {
    if ((selectedObject) && (rotationFunction))
      (selectedObject->getData()->*rotationFunction)(rot);
  }
  /** camera callback to catch trackball translation
      @param t translation vector */
  virtual void translateCameraObject(VEC &t) {
    if ((selectedObject) && (translationFunction))
      (selectedObject->getData()->*translationFunction)(t);
  }
  /** camera callback to catch trackball scale
      @param s scale */
  virtual void scaleCameraObject(double s) {
    if ((selectedObject) && (scaleFunction))
      (selectedObject->getData()->*scaleFunction)(s);
  }


  /** mouse button was pressed
      @param w which button
      @param k which key modifier
      @param x mouse x coordinate
      @param y mouse y coordinate */
  virtual void buttonPressed(CameraTypes::WHICH_BUTTON w, CameraTypes::KEY_MODIFIER k,
		     int x, int y);
  /** mouse button was released
      @param w which button
      @param x x mouse coordinate
      @param y mouse y coordinate */
  virtual void buttonReleased(CameraTypes::WHICH_BUTTON w, int x, int y);
  /** mouse pointer moved
      @param x X mouse location
      @param y Y mouse location */
  virtual void mouseMoved(int x, int y);

  /** returns the state (pressed or released) of a mouse button 
  @param w the mouse button */
  BUTTON_STATE getButtonState(CameraTypes::WHICH_BUTTON w) const {
    return (B[w]);
  }
  
};

} // namespace yav

#include <graphics/Scene.h>
#include <graphics/Data.h>

#include "Camera.txx"

#endif  // ifdef GRAPHICS

#endif  // ifdef CAMERA_H
