#ifndef DATA_H
#define DATA_H

#pragma interface "graphics/Data.h"

#include <basics/Exception.h>
#include <basics/TclString.h>
#include <graphics/CameraTypes.h>
#include <linearAlgebra/Vec2.h>
#include <basics/ListElement.h>
#ifdef TCL_INTERFACE
#include <framework/Interface.h>
#endif

namespace yav {


/** \class Data
    \brief data parent class
    \ingroup Graphics
*/
template <class SCENE, class DATA>
class Data : public ListElement<DATA> {
 protected:
  /// the scene where  the object is
  SCENE *scene;
  /// only one data of that kind is allowed in a scene if exclusive is true
  bool exclusive;
 public:
  /** @name constructor */
  //@{
  /** builds a new Data
      @param scene scene to create data in
      @param exclusive */
  Data(SCENE *scene, bool exclusive = false);
  //@}
  
  /** @name destructor */
  //@{
  /// delete interface (if any) and this
  virtual ~Data(){}
  //@}
  /// returns dynamic data type
  DATA *getData(void) { return dynamic_cast<DATA *>(this); }
  /// returns scene this object is in
  SCENE *getScene() const { return scene; }
  /// move object in new scene
  virtual void setScene(SCENE *scene);
  /// returns exclusive flag
  bool getExclusivity(void) const { return exclusive; }
};

#if (defined GRAPHICS || defined TCL_INTERFACE)
/** \class DataTypes
    \brief data base types
    \ingroup Graphics
*/
class DataTypes : public Client {
 public:
   /// base class for mouse click information
  struct MouseClick {};
};

/** \class GraphicsData
    \brief data parent class
    \ingroup Graphics
*/
template <class VEC, class ROT, class SCENE, class CAMERA, class DATA>
class GraphicsData : public Data<SCENE,DATA>,public DataTypes {
 public:
  /// a type of function that is called when the scene is redrawn
  typedef void (*DATA_REDRAW_FUNCTION)(DATA *,CAMERA *,void *);
  /// a structure 
  struct RedrawCallback {
    DATA_REDRAW_FUNCTION redrawFunction;
    /// store a pointer to be sent by the redrawFunction
    void *clientData;
  };
 protected:
  /// The centroid of the object
  VEC center;
  /// object orientation
  ROT orientation;
  /// The radius of the circumscribing circle
  double radius;
   /// object's boundingbox
  VEC boundingBox[2];
  /// a list of callback functions that are called during the redraw of a scene
  std::list<RedrawCallback *> callbackList;
  
 public:
  /** @name constructor */
  //@{
  /** builds a new Data
      @param scene scene to create data in
      @param exclusive
  */
  GraphicsData(SCENE *scene, bool exclusive = false);
  //@}
  
  /** @name destructor */
  //@{
  /// delete interface (if any) and this
  virtual ~GraphicsData();
  //@}
  
  /// returns object's center 
  double getRadius(void) const { return(radius); }
  /// returns object's center 
  VEC getCenter(void) const { return(center); }
  /// returns object's orientation
  const ROT& getOrientation(void) const { return orientation; }
  /// add a redraw callback function in the list
  void addRedrawCallback(DATA_REDRAW_FUNCTION f,void *c);
  /// removes a redraw callback function in the list
  void removeRedrawCallback(DATA_REDRAW_FUNCTION f);

  /** translates the object
      @param tr translation vector */
  virtual void translate(const VEC& tr) {}
  /** rotates the object
      @param rotation rotation matrix */
  virtual void rotate(const ROT& rotation) {}
  /** scales the object
      @param scale scale facto */
  virtual void scale(const double scale) {}

 public:
 
 /// draw on all cameras the display of that object including the redraw callbacks
  void redrawAll(CAMERA *);
  /// draw on all cameras the display of that object
  virtual void redraw(CAMERA *) = 0;
  /** Returns the graphic object center and weight.
      The weight is a measure of the number of primitives to allow
      barycentric computation of the whole scene center.
  */
  virtual void updateCenter(VEC& center, unsigned int& n) = 0;
  /** returns the graphic object radius given its center
      @return object radius */
  virtual double updateRadius(const VEC& center) = 0;
 
  /** translates the bounding box
      @param tr translation vector */
  virtual void translateBoundingBox(const VEC& tr);
  /** rotates the bounding box
      @param r rotation matrix */
  virtual void rotateBoundingBox(const ROT& r);
  /** scales the bounding box
      @param scale scale factor */
  virtual void scaleBoundingBox(double scale);
  /** returns the bounding box */
  const VEC *getBoundingBox() const { return(boundingBox); }
  /// returns the RedrawFlag


  /** routine called when a mouse button is pressed and that it is not handle
      for scaling, selecting, rotating or translating the camera or an object */
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  CAMERA *cam,int x,int y) {}
  /** routine called when a mouse button is released and that it is not handle
      for scaling, selecting, rotating or translating the camera or an object */
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				   CameraTypes::KEY_MODIFIER modifier,
				   CAMERA *cam,int x,int y) {}
  /** routine called when a mouse button is moved and that it is not handle
      for scaling, selecting, rotating or translating the camera or an object */
  virtual void mouseButtonMoved(CAMERA *cam,int x,int y	) {}
  /// function to be called when the object appearance is modified
  void redrawScene();
  /// function to be called when the object appearance is modified
  void focusScene();
  
};
#endif  //(defined GRAPHICS || defined OPENGL)




} // namespace yav

#include "Data.txx"

#endif // ifndef DATA_H



