#ifndef INTERNALFORCE_H
#define INTERNALFORCE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics/InternalForce.h"
#include <basics/Exception.h>

namespace yav {

/** \class InternalForce
    \ingroup Graphics
    internal force computation descriptor : Create an autolink object
    with two functions : one for computing internal forces on the mesh
    and one for computing the internal forces on each vertex */
template <class MT, class MC, class MV> class InternalForce : public 
AutoLink <InternalForce<MT, MC, MV> > {
 public: 
  /** a function in the global scope with a mesh as an argument */
  typedef void (*MESHFUNCTION) (MT *);
  /** a function in the global scope with a cell as an argument */
  typedef void (*CELLFUNCTION) (MC *);
  /** a function in the global scope with a vertex as an argument */
  typedef void (*VERTEXFUNCTION) (MV *);
  /** a function in the global scope with a mesh as an argument that returns a floating value : the internal energy */
  typedef double (*MESHENERGYFUNCTION) (MT *);
 protected:
  /** the function to be called when this type of
      internal force is assigned to all vertices */
  MESHFUNCTION setInternalForce;
  /** the function to be called for computing the internal force 
      on all vertices */
  MESHFUNCTION computeInternalForce;
  /** the function to be called when this type of
      internal force is not applied anymore */
  MESHFUNCTION unsetInternalForce;
  /** the function to be called when an implicit scheme is used  */
  MESHFUNCTION addStiffnessMatrix;
    /** the function to be called when an implicit scheme is used  */
  MESHFUNCTION computeImplicit;
  /** the function to be called when an implicit scheme is used and the internal conjugated gradient is used */
  MESHFUNCTION computeDfDx;
  /** the function to be called when this type of
      internal force is assigned to a newly created cell */
  CELLFUNCTION setCellInternalForce;
  /** the function to be called when a cell using this type of
      internal force is removed from the mesh */
  CELLFUNCTION unsetCellInternalForce;
  /** the function to be called when this type of
      internal force is computed on a vertex */
  VERTEXFUNCTION computeLocalInternalForce;
  /** the function to be called to get the whole internal energy
      associated with this internal force */
  MESHENERGYFUNCTION computeEnergy;
  /// the name of the internal force 
  TclString name;
 public:
  /// returns the name of internal force
  char *getForceName() const {
    return((char *) name);  }
  /** returns the function to be called when this type of
      internal force is assigned to all vertices */ 
  MESHFUNCTION getSetInternalForce() const {
    return setInternalForce;
  }
  /** returns the function to be called when this type of
      internal force is not applied anymore */
  MESHFUNCTION getUnsetInternalForce() const {
    return unsetInternalForce;
  }
  /** returns the function to be called when this type of
      internal force is assigned to a newly created cell */
  CELLFUNCTION getSetCellInternalForce() const {
    return setCellInternalForce;
  }
  /** returns the function to be called when a cell using
      this type of internal force is removed from the mesh */
  CELLFUNCTION getUnsetCellInternalForce() const {
    return unsetCellInternalForce;
  }
  /** returns the function to be called for computing the internal force 
      on all vertices */
  MESHFUNCTION getComputeInternalForce() const {
    return computeInternalForce;
  }
    /** returns the function to be called for computing the internal force 
      on all vertices */
  MESHFUNCTION getComputeImplicitInternalForce() const {
    return computeImplicit;
  }
  /** returns the function to be called when this type of
      internal force is computed on a vertex */
  VERTEXFUNCTION getComputeLocalInternalForce() const {
    return computeLocalInternalForce;
  }
  /** returns the function to be called to get the whole internal energy
      associated with this internal force */
  MESHENERGYFUNCTION getComputeEnergy() const {
    return computeEnergy;
  }
  /** returns the function to be called when an implicit scheme is used  */
  MESHFUNCTION getAddStiffnessMatrixInternalForce() const {
    return addStiffnessMatrix;
  }
    /** returns the function to be called when an implicit scheme is used  */
  MESHFUNCTION getComputeDfDxInternalForce() const {
    return computeDfDx;
  }
  /// constructor that assign all functions
  InternalForce( void (*setG) (MT *),void (*computeG) (MT *),
		 void (*unsetG) (MT *),void (*computeL) (MV *),
		 TclString n,void (*addStiff)(MT*)=0,void (*dfdx)(MT*)=0,double (*computeE) (MT *)=0,void (*compImp)(MT*)=0,void (*setC)(MC*)=0, void (*unsetC)(MC*)=0) :
    setInternalForce(setG), computeInternalForce(computeG), unsetInternalForce(unsetG),addStiffnessMatrix(addStiff),computeImplicit(compImp),
 computeDfDx(dfdx), 
    setCellInternalForce(setC),unsetCellInternalForce(unsetC),computeLocalInternalForce(computeL),computeEnergy(computeE),
    name(n) {
  }

  /// destructor
  virtual ~InternalForce() {}

  /** returns force type that matches given name
      @param name researched force name
      @return 0 if force type does no exist, pointer on class otherwise */
  static InternalForce *getByName(const char *name) {
    AutoLink <InternalForce<MT, MC, MV> > *el;
    for(el = AutoLink<InternalForce<MT, MC, MV> >::getHead(); el;
	el = el->following()) {
      if(!strcmp(name, ((InternalForce<MT, MC, MV> *) el)->name))
	return (InternalForce<MT, MC, MV> *) el;
    }
    return 0;
  }

};

} // namespace yav

#endif
