#ifndef SCENE_H
#define SCENE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics/Scene.h"
#include <graphics/Camera.h>
#include <graphics/Data.h>
#include <framework/Client.h>
#include <basics/Timer.h>

namespace yav {

#define SAFETY_MARGIN_TO_GET_PROPER_FRAME_RATE 0.8

/** \class Scene
    \ingroup Graphics
    \brief world description
*/
 template <class DATA,  class USERINTERFACE, class SCENE>
class Scene :  public Client {
  protected:
  /// list of data in scene
  ObjectsList< DATA > dataList;
  /// list of userinterfaces in scene
  ObjectsList< USERINTERFACE > userInterfaceList;
 public:
  /** @name constructor */
  //@{
  /// initializes scene with nothing inside
  Scene() : Client(){
  }
  //@}

  /** @name destructor */
  //@{
  /// destroy each object in list
  virtual ~Scene();
  //@}

  /// returns dynamic scene type
  virtual SCENE *getType(void) { return dynamic_cast<SCENE *>(this); }


  /** returns true if an object requiring exclusivity exists in scene,
      false otherwise */
  bool getExclusivity(void) const;
  

  /** adds a new data to this scene
      @param d data to add
      @exception ObjectsList<DATA>::AlreadyInList cannot add the same object twice
      @exception ObjectsList<DATA>::NullPointer cannot add a 0 pointer */
  virtual void addData(DATA *d) /*throw (typename ObjectsList< DATA >::AlreadyInList, 
			  typename ObjectsList< DATA >::NullPointer)*/;
  /** adds a new data to this scene
      @param d data to add
      @exception ObjectsList<DATA>::AlreadyInList cannot add the same object twice
      @exception ObjectsList<DATA>::NullPointer cannot add a 0 pointer */
  virtual void operator+=(DATA *d) throw (typename ObjectsList< DATA >::AlreadyInList, 
					  typename ObjectsList< DATA >::NullPointer)
    { addData(d); }
 
  /** removes a data from this scene
      @param d data to remove
      @exception ObjectsList<DATA>::NotFound d does not exist in scene */
  virtual void removeData(DATA *d) /*throw (typename ObjectsList< DATA >::NotFound)*/;
  /** removes a data from this scene
      @param d data to remove
      @exception ObjectsList<DATA>::NotFound d does not exist in scene */
  virtual void operator-=(DATA *d) throw (typename ObjectsList< DATA >::NotFound)
    {removeData(d);}
    
 
  /** adds a new user interface to this scene
      @param d user interface to add
      @exception ObjectsList<DEFORMABLE>::AlreadyInList cannot add the same object twice
      @exception ObjectsList<DEFORMABLE>::NullPointer cannot add a 0 pointer */
  virtual void addUserInterface(USERINTERFACE *d) 
    /*    throw (typename ObjectsList< USERINTERFACE >::AlreadyInList,
	  typename ObjectsList< USERINTERFACE >::NullPointer)*/;

  /** removes a user Interface from this scene
      @param ui user interface to remove
      @exception ObjectsList<USERINTERFACE>:NotFound ui does not exist in scenee */
  virtual void removeUserInterface(USERINTERFACE *ui)
    /*throw (typename ObjectsList< USERINTERFACE >::NotFound)*/;
  /** removes a user interface from this scene
      @param ui user interface data to remove
      @exception ObjectsList<USERINTERFACE>:NotFound ui does not exist in scene*/
  virtual void operator-=(USERINTERFACE *ui)
    throw (typename ObjectsList< USERINTERFACE >::NotFound)
    { removeUserInterface(ui); }


  /// returns an iterator initialized on first user interface
  typename ObjectsList< USERINTERFACE >::iterator firstUserInterface(void) const {
    return userInterfaceList.begin();
  }
  /// returns an iterator initialized on last user interface
  typename ObjectsList< USERINTERFACE >::iterator lastUserInterface(void) const {
    return userInterfaceList.end();
  }

  /// returns first user interface
  USERINTERFACE *getFirstUserInterface(void) const {
    return userInterfaceList.getFirst();
  }
  /// returns an iterator initialized on first data
  typename ObjectsList< DATA >::iterator firstData(void) const {
    return dataList.begin();
  }
  /// returns an iterator initialized on last data
  typename ObjectsList< DATA >::iterator lastData(void) const {
    return dataList.end();
  }
  
 
  /// returns first data
  DATA *getFirstData(void) const { return dataList.getFirst(); }
  
 


  /** reads a file into a scene
      @param filename file name
      @return true if file was succesfully read, false otherwise */
  virtual bool read(const TclString &filename) = 0;
};

#if (defined GRAPHICS)
/** \class GraphicsScene
    \ingroup Graphics
    \brief world description
*/
  template <class VEC,class DATA,  class USERINTERFACE, class CAMERA, class SCENE>
    class GraphicsScene : public Scene<DATA,USERINTERFACE,SCENE> {
 protected:
  /// list of camera on this scene
  ObjectsList< CAMERA > cameraList;

 public:
  /** @name constructor */
  //@{
  /// initializes scene with nothing inside
  GraphicsScene() : Scene<DATA,USERINTERFACE,SCENE>()
     {
     }
  //@}

  /** @name destructor */
  //@{
  /// destroy each object in list
  virtual ~GraphicsScene(){}
  //@}
 
  /// focus all scene cameras
  void focus(void);
  /// redraw all scene cameras
  virtual void redraw(void);
 /** computes whole scene mean center and radius
      @param center returned scene center
      @param radius returned scene radius */
  virtual void getCenterRadius(VEC& center, double& radius);

  /** @name Exception */
  //@{
  /** \class EmptyGraphicsObject
      \ingroup Graphics
      \brief using a 0 pointer on a Data object
  */
  class EmptyGraphicsObject : public Exception {
  public:
    EmptyGraphicsObject(const std::string& from) : 
      Exception(from,
		"Using a null pointer on a Data object") {}
  };
  //@}

  /** add a new camera to visualize this scene
      @param cam new camera*/
  virtual void addCamera(CAMERA *cam) { cameraList.append(cam); }
  /** removes a camera from this scene
      @param cam camera to remove
      @exception ObjectsList<CAMERA>::NotFound cam does not exist in this scene */
  virtual void removeCamera(CAMERA *cam) throw (typename ObjectsList< CAMERA >::NotFound)
    { cameraList.remove(cam); }
  

  /// returns an iterator initialized on first scene camera
  typename ObjectsList< CAMERA >::iterator firstCamera(void) const {
    return cameraList.begin();
  }
  /// returns an iterator initialized on the last scene camera
  typename ObjectsList< CAMERA >::iterator lastCamera(void) const {
    return cameraList.end();
  }

};
#endif  //(defined GRAPHICS || defined OPENGL)

} // namespace yav

#include "Scene.txx"

#endif // ifndef SCENE_H


