#ifndef SCENEITEM_H
#define SCENEITEM_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics/SceneItem.h"
#include <list>
#include <basics/AutoLink.h>
#include <modules/TclObjectKind.h>
#include <graphics/FileScene.h>


namespace yav {

/** \class SceneItem
    \ingroup Graphics
    \brief An item that can be created and added into a scene */
template <class S> class SceneItem : public TclObjectKind,
                                     public AutoLink< SceneItem<S> > 
  {
  protected:
    /// file extensions list
    std::list<const char *> extensions;
    /// object construction mode
    int mode;
    /// Tcl scripts for specific creation
    TclString creationScript;
    
  public:
    /** @name constructors */
    //@{
    /** creates a new object kind
	@param m module creating this ObjectKind
	@param type objects type
	@param name kind name
	@param realName kind real name ie that can be understood by the user
	@param createFromArgs pointer on an object creation function
	@param objectsList pointer on a function returning a list of this kind objects
	@param _mode describes if this Object can be started directly from scratch
	@param synopsis describes the parameter to provide at the new command in order to create an object 
	@exception InvalidParameter no argument may be 0 */
    SceneItem(TclModule *m,
	      const std::type_info& type,
	      const char *name,
	      const char *realName,
	      int (*createFromArgs)(TclModule *, int, char **),
	      int (*objectsList)(TclModule *, int, char **),
	      int _mode = 1,
	      const char *synopsis = 0)
      throw(TclObjectKind::InvalidParameters)
      : TclObjectKind(m, type, name, realName, createFromArgs, objectsList, synopsis),
      mode(_mode),creationScript(EMPTY_STRING) {
    }
 
 /** creates a new object kind and associate a read function
      @param m
      @param type objects type
      @param name kind name
      @param realName kind real name ie that can be understood by the user
      @param createFromArgs pointer on an object creation function
      @param objectsList pointer on a function returning a list of this kind objects
      @param _mode describes if this Object can be started directly from scratch
      @param key file key
      @param fext file extension. If several file extensions may be used, then each extension should be separed by a comma like in ".trian,.tr3D"
      @param readf file read function
      @param magicFunction magic number test function
      @param synopsis describes the parameter to provide at the new command in order to create an object 
      @exception InvalidParameter no argument may be 0 */
  SceneItem(TclModule *m,
	    const std::type_info& type,
	    const char *name,
	    const char *realName,
	    int (*createFromArgs)(TclModule *, int, char **),
	    int (*objectsList)(TclModule *, int, char **),
	    int _mode,
	    const char *key,
	    const char *fext,
	    typename FileScene< S >::READFILE readf,
	    typename FileScene< S >::TESTMAGICNUMBER magicFunction,
	    const char *synopsis)
    throw(TclObjectKind::InvalidParameters)
    : TclObjectKind(m, type, name, realName,createFromArgs, objectsList, synopsis),
    mode(_mode) ,creationScript(EMPTY_STRING)
    {
      char fileExt[200];
      strcpy(fileExt,fext);
      char *p,*pstart;
      p=pstart=fileExt;
      do {
	for (;(*p!='\0')&& (*p!=',');p++);
	if (*p==',')
	  {
	    *p='\0';
	    char *ext= new char[200];
	    strcpy(ext,pstart);
	    extensions.push_back(ext);
	    (void) new FileScene< S >(key, ext, readf, magicFunction);
	    ++p;
	    pstart=p;
	  }
	else
	  {
	    char *ext= new char[200];
	    strcpy(ext,pstart);
	    extensions.push_back(ext);
	    (void) new FileScene< S >(key, ext, readf, magicFunction);
	  }
      } while (*p!='\0');
    }
    //@}

  /** @name destructor */
  //@{
  /// destroy each object in list
  virtual ~SceneItem() {
    std::list<const char *>::iterator it=extensions.begin();
    for(;it!=extensions.end();it++)
      delete [] (char *) (*it);
  }
  //@}

  /// returns file extensions list for that kind of object
  const std::list<const char *>& getExtensions(void) const {
    return extensions;
  }
  /// returns object creation mode
  int getMode(void) const { return mode; }
  /// set an additional way of creating an object
  void addMode(int newMode) const { mode= mode | newMode; }
  /// returns the object creation script
  TclString getCreationScript(void) const { return creationScript; }
  /// sets the object creation script
  void setCreationScript(const TclString s)  {creationScript=s; }
};


} // namespace yav

#endif
