#ifndef BASICCAMERABACKGROUND_H
#define BASICCAMERABACKGROUND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#ifdef WIN32
#include <windows.h>
#endif
#include <graphics2D/Camera2D.h>


namespace yav {

#define SLICE_NAME	"2D Image"

/** \class BasicCameraBackground
    \ingroup Graphics2D
    \brief A single image Background on a 2D camera */
class BasicCameraBackground : public Camera2DBackground {
public:
  /// the two different ways of coding each pixel of a texture
  enum Format {TEXTURE_RGBA, TEXTURE_LUMINANCE};

 protected:
  /// the name of the background type
  TclString sliceName;
   /// if the texture has been binded
  bool binded;
  /// the texture ID
  GLuint id;
  /// texture buffer 
  void *buffer;
  /// number of rows: must be a power of two
  unsigned int nrows;
  /// number of columns: must be a power of two
  unsigned int ncols;
  /// the number of rows that are really meaningful
  unsigned int nbRealRows;
  /// the number of columns that are really meaningful
  unsigned int nbRealColumns;
   /// pixel format
  Format format;
  /// if mipmapping should be used
  bool useMipmapping;
  /// the number of levels of the mipmapping
  unsigned int mipmapLevel;
  /// if the display list should be update
  bool updateDisplayList;
 public:
  /** @name constructor */
  //@{
	 /** constructor based an image filename
	 @param cam the camera associated with this background
	  @param filename the name of an SGI format image */ 
  BasicCameraBackground(Camera2D *cam,const char *filename);
/** constructor based an image buffer
	 @param cam the camera associated with this background
	  @param b a buffer of integer 
	  @param nr the number of rows
	  @param nc the number of columns
	  @param f the format of the input image */ 
  BasicCameraBackground(Camera2D *cam,unsigned int *b, const unsigned int nr, const unsigned int nc,
	    Format f=TEXTURE_RGBA);
  //@}

  /** @name destructor */
  //@{
  virtual ~BasicCameraBackground(){
	  if (buffer)
	    delete []  ((unsigned int *)buffer);
  }
  //@}
   protected:
  /** enlarge a given texture such that its size is a power of 2 and store
      it in the texture buffer */
  void enlargeTexture(const void *b, const unsigned int nr, 
		      const unsigned int nc);
  /// create the display list associated with the texture
  void createTexture();
  /// initialize variables
  void init();
public:
	/// update the OpenGL texture based on the new texture buffer
  void updateTexture() const;
  /// return the nb of texture rows 
  unsigned int getNbRows() const { return nrows;}
  /// return the nb of texture columns
  unsigned int getNbColumns() const { return ncols;}
  /// return the nb of meaningful texture rows 
  unsigned int getNbRealRows() const { return nbRealRows;}
  /// return the nb of meaningful texture columns
  unsigned int getNbRealColumns() const { return nbRealColumns;}
   /// use this texture
  void enableTexture() {
    if (binded==false)
      createTexture();
    glBindTexture(GL_TEXTURE_2D, id);
    glEnable(GL_TEXTURE_2D);
  }
  /// stop using this texture
  void disableTexture() const {
    glDisable(GL_TEXTURE_2D);
  }
  /// draws the image in the camera
  virtual void redraw();
  /** defines the two corners of the camera viewport 
  (corresponding to the image size) */
  virtual void getViewportSize(Vec2<double> &vmin,Vec2<double> &vmax);
  /** returns the normal vector and a point lying on the image plane
  @param org the plane origin (output)
  @param normal the plane normal (output)*/
  virtual void getImagePlane(Vec3<double> &org, Vec3<double> &normal);
  /*** transforms 3D coordinates into plane coordinates 
  @param p an input 3D point lying in the image plane
  @param pt output the associated 2D coordinates in the image reference frame 
  @return returns true if the 3D lies effectively on the image plane */
  virtual bool map3DWorldTo2DWorldCoordinates(const Vec3<double> &p, Vec2<double> &pt);
  /*** transforms 2D plane coordinates into image coordinates. The image coordinates
  may have 4 components (x,y,z and time) for 4D images
  @param pt an input 2D point in plane coordinates 
  @param pixel output the pixel coordinates taking into account pixel size. The pixel
  coordinates may have negative coordinates */
  virtual void map2DWorldToImageCoordinates(const Vec2<double> &pt, Vec4<double> &pixel);
  /*** transforms 2D plane coordinates into 3D world coordinates. 
  @param pt an input 2D point in plane coordinates 
  @param p output the 3D world coordinates taking into account the plane orientation and pixel size */
  virtual void map2DWorldTo3DWorldCoordinates(const Vec2<double> &pt, Vec3<double> &p);
  /*** provide the slice name to characterize the slice type */
  virtual const char *getSliceName() const;
  /*** returns the slice number if applicable */
  virtual int getSliceNumber() const;
  /*** returns  a label associated with current mouse position */
  virtual const char *getLabel(const Vec2<double> &pt);

  /** routine called when a mouse button is pressed and that it is not handle
  for scaling, selecting, rotating or translating the camera or an object */
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
	  CameraTypes::KEY_MODIFIER modifier,
	  Camera2D *cam,int x,int y){}
	  /** routine called when a mouse button is released and that it is not handle
	  for scaling, selecting, rotating or translating the camera or an object */
	  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
		  CameraTypes::KEY_MODIFIER modifier,
		  Camera2D *cam,int x,int y){}
		  /** routine called when a mouse button is moved and that it is not handle
		  for scaling, selecting, rotating or translating the camera or an object */
		  virtual void mouseButtonMoved(Camera2D *cam,int x,int y	){}
};

}

#endif
