#ifndef CAMERA2D_H
#define CAMERA2D_H

#pragma interface "graphics2D/Camera2D.h"
#ifndef WIN32
#include <sys/time.h>
#endif
#include <basics/Bool.h>
#include <basics/ListElement.h>
#include <linearAlgebra/Vec2.h>
#include <linearAlgebra/Rotation2D.h>
#include <linearAlgebra/Matrix4x4.h>
#include <graphics/Camera.h>
#include <graphics/CameraTypes.h>

#ifdef _MSC_VER
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>

namespace yav {

/// default camera center
#define DEFAULT_CAMERA2D_CENTER Vec2<double>(0.0, 0.0);

#define CAMERA2D_INITIAL_ASPECT_RATIO 0.7

class Scene2D;
class Camera2D;
class Data2D;
class Camera2DActivation;

/** \class SliceCameraBackground
    \ingroup Graphics2D
    \brief Background image in a slice camera */
class Camera2DBackground {
 protected:
  /// the slice camera attached to the background
  Camera2D *camera;
 public:
  /** @name constructor */
  //@{
  Camera2DBackground(Camera2D *cam);
  //@}

  /** @name destructor */
  //@{
  virtual ~Camera2DBackground();
  //@}
  /// draws the image in the camera
  virtual void redraw()=0;
  /** defines the two corners of the camera viewport 
  (corresponding to the image size) */
  virtual void getViewportSize(Vec2<double> &vmin,Vec2<double> &vmax)=0;
  /** returns the normal vector and a point lying on the image plane
  @param org the plane origin (output)
  @param normal the plane normal (output)*/
  virtual void getImagePlane(Vec3<double> &org, Vec3<double> &normal)=0;
  /*** transforms 3D coordinates into 2D plane coordinates 
  @param p an input 3D point lying in the image plane
  @param pt output the associated 2D coordinates in the image reference frame 
  @return returns true if the 3D lies effectively on the image plane */
  virtual bool map3DWorldTo2DWorldCoordinates(const Vec3<double> &p, Vec2<double> &pt)=0;
  /*** transforms 2D plane coordinates into 3D world coordinates. 
  @param pt an input 2D point in plane (2D world) coordinates 
  @param p output the 3D world coordinates taking into account the plane orientation and pixel size */
  virtual void map2DWorldTo3DWorldCoordinates(const Vec2<double> &pt, Vec3<double> &p)=0;
  /*** provide the slice name to characterize the slice type */
  virtual const char *getSliceName() const=0;
  /*** returns the slice number if applicable */
  virtual int getSliceNumber() const=0;
  /*** returns the 2D camera associated with the background*/
  virtual Camera2D *getCamera() const {
	  return camera;
  }
  
  /** routine called when a mouse button is pressed and that it is not handle
      for scaling, selecting, rotating or translating the camera or an object */
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int x,int y)=0;
  /** routine called when a mouse button is released and that it is not handle
      for scaling, selecting, rotating or translating the camera or an object */
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				   CameraTypes::KEY_MODIFIER modifier,
				   Camera2D *cam,int x,int y)=0;
  /** routine called when a mouse button is moved and that it is not handle
      for scaling, selecting, rotating or translating the camera or an object */
  virtual void mouseButtonMoved(Camera2D *cam,int x,int y	)=0;
};

/** \class Camera2D
    \ingroup Graphics2D
    \brief General 2D camera. It gives no practival implementation */
class Camera2D :  public ListElement<Camera2D>, 
                  public Camera<Vec2<double>, Rotation2D, Scene2D,Data2D, Camera2D> {
public:
  /// the type of function used to select sub-items of a Data
  typedef void (*SelectionFunctionType)(Data2D *,const Vec2<double> pixel,
				    Camera2D *cam);

 private:
  /// common constructors code
  void initCamera(void);
 protected:
 /** called before the destruction of a camera : removes stencil display list */
  void destroyCamera();

 protected:
  /// the camera background
  Camera2DBackground *background;
  /// a callback function for user defined selection
  SelectionFunctionType selectionFunction;
  /// is the camera being deleted ?
  bool beingDestroyed;
   /// an object that handles 
  Camera2DActivation *cameraActivation;
  /// whether to use stencil Buffer for display
  bool stencilBuffer;
  /// the display list associated with a stencil
  GLuint stencilBufferDisplayList;

  // display list of the default font
  GLuint defaultFont;
  // display list of the large font
  GLuint largeFont;
  // display list of the small font
  GLuint smallFont;
   /// rotation to apply while auto-moving
  Rotation2D autoRotation;
  /// translation to apply while auto-moving
  Vec2<double>  autoTranslation;
  ///  Frame Ratestring
  char stringFrameRate[100];
  /// copyright string
  char stringCopyright[100];
  ///  if copyright should be displayed
  bool showCopyright;
  /// Counting number of redraws 
  unsigned int count;
  ///  Measuring the time between two redraws action
  Timer globalTimer;
  ///  if frame rate should be displayed
  bool showFrameRate;
  /// the field depth ratio multiplied by the radius
  double scale;
  /** 3D mouse position  corresponding to the 3D point defining the 3D line defined when a mouse button is clicked, the other extremity of the line being the center of the camera*/
  Vec2<double> mouse;
   /* Projection matrix */
  Matrix4x4<double> matTransform;
  /// perspective center
  Vec2<double> refCenter;
  /** defines the coordinates of the two corners of camera (parameters used 
  in glOrtho2D )*/
  Vec2<double> lowerCorner, upperCorner;

#ifdef __GNUC__
  /// The initial aspect ratio when displaying 2D graphical objects
  static const  double initialAspectRatio;
#endif
 
  /** @name protected constructor */
  //@{
  /// constructor that does not create any scene
  Camera2D(Camera2D *);
  //@}

 public:
  /** @name Constructor */
  //@{
  /** Creates a new camera. The camera visualize given scene. If scn is 0
      the camera visualize the first scene in global scene list (if any)
      or a new scene is created.
      @param scn the graphics scene
      @param h height in pixel size
      @param w width in pixel size
      @param scale default scale value */
  Camera2D(Scene2D *scn = 0, unsigned int h = 0, unsigned int w = 0,
	   double scale = DEFAULT_CAMERA_SCALE);
  //@}

  /** @name Destructor */
  //@{
  virtual ~Camera2D();
  //@}
 
  /** set scale function on subobject
      @param sf new scale function */  
  void setSelectionFunction(SelectionFunctionType sf){
	  selectionFunction=sf;
  }
  
  /** subobject selection method
      @param vec mouse click point
      @param cam clicked camera */
  virtual void select(const Vec2<double> vec, Camera2D *cam){
  if (selectionFunction)
      (*selectionFunction) (selectedObject,vec,cam);
  }
  	
  /** set default font */
  void setDefaultFont(GLuint f) {
	  defaultFont=f;
  }
   /** set small font */
  void setSmallFont(GLuint f) {
	  smallFont=f;
  } /** set large font */
  void setLargeFont(GLuint f) {
	  largeFont=f;
  }
/** camera callback to catch trackball rotation
      @param rot rotation matrix */
  virtual void rotateCameraObject(Rotation2D &rot);
  /** camera callback to catch trackball translation
      @param t translation vector */
  virtual void translateCameraObject(Vec2<double> &t);
  /// returns camera center
  Vec2<double> getRefCenter(void) const { return refCenter; }
/// return the object dealing with the camera activation
  Camera2DActivation *getCameraActivation() const;
  /// set the object dealing with the camera activation
  void setCameraActivation(Camera2DActivation *activation); 
     /** updates the internal camera parameters after the camera size has changed
      @param reset wether tp init the transformation matrix to identity */
  virtual void setViewingParameters(bool reset=true);
  /** set the OGL projection matrix 
      @param reset wether tp init the transformation matrix to identity */
  virtual void setProjectionParameters(bool reset=true);
  /// returns the camera background
  Camera2DBackground *getBackground() const;
  /// sets the camera background
  void setBackground(const Camera2DBackground *b);
  /** mouse button was pressed
      @param w which button
      @param k which key modifier
      @param x mouse x coordinate
      @param y mouse y coordinate */
  virtual void buttonPressed(CameraTypes::WHICH_BUTTON w, CameraTypes::KEY_MODIFIER k,
		     int x, int y);
  /** mouse button was released
      @param w which button
      @param x x mouse coordinate
      @param y mouse y coordinate */
  virtual void buttonReleased(CameraTypes::WHICH_BUTTON w, int x, int y);
  /** mouse pointer moved
      @param x X mouse location
      @param y Y mouse location */
  virtual void mouseMoved(int x, int y);
 /** Redraw all existing 3D objects in the 3D scene  and eventually 
      objects belonging to the camera (clipping planes, spot loghts,..) */
  virtual void redraw(void);
  /// camera creation callback
  virtual void init(void);
  /// resize camera to current height and width
  virtual void resize(void);

  /** resize camera
      @param w new camera width
      @param h new camera height */
  virtual void resize(unsigned int w, unsigned int h) {
    Camera<Vec2<double>, Rotation2D, Scene2D, Data2D, Camera2D>::resize(w, h);
  }
 /** returns the 4x4 matrix in homogeneous coordinate corresponding to 
      the actual rotation, translation and scale performed. It corresponds
      to the Viewing matrix in OpenGL */
  Matrix4x4<double> getMatTransform(void) const { return matTransform; }
  /** set the 4x4 Viewing matrix in homogeneous coordinate */
  void setMatTransform(Matrix4x4<double> mat){ matTransform = mat; }
    /** updates the Viewing matrix based on refcenter, center and the 
      matTransform matrix */
  virtual void applyTransformation();
 /** rotates camera
     @param rot rotation matrix (unsused) */
  virtual void rotateCamera(Rotation2D& rot);
  /** translates camera center
      @param t translation vector */
  virtual void translateCamera(Vec2<double> &t);
  /** scales camera
      @param s scale */
  virtual void scaleCamera(double s);
 
  /** initialize  the viewport and basic OpenGL parameters */
  virtual void initViewport(void);

  /// focus whatever autofocus says
  virtual void doFocus(void);

  /** converts screen coordinate to cartesian coordinate
      @param pixel pixel value
      @param position returned cartesian coordinate */
  void mapScreenTo2DWorldCoordinates(const Vec2<double> &pixel, Vec2<double> &position);

  /** converts cartesian coordinate to screen coordinate
      @param position position in cartesian coordinate
      @param pixel returned screen coordinate */
  void map2DWorldToScreenCoordinates(const Vec2<double> &position, Vec2<double> &pixel);
  /*** sets the background color of the camera
       @param r the red component
       @param g the green component
       @param b the blue component */
  virtual void setBackgroundColor(float r, float g, float b);
  /** print a string on the screen 
      @param posx the position in the window (column index) 
      @param posy the position in the window (row index) 
      @param str the null terminated string
      @param font must be either defaultFont, smallFont or largeFont */
  void printString(unsigned int posx,unsigned int posy, char *str, GLuint font=0);
  /// change the display of the copyright
  void toggleCopyright();
  /// change the display of the frame rate
  void toggleFrameRate();
  ///is the camera being destroyed?
  bool isCameraBeingDestroyed() const;





 protected:
	 
  /** set stencilBuffer flag value
      @param nstb new stencilBuffer value */
  void setStencilBuffer(bool nstb);

  /** build stencilBufferDisplayList
    initialize GL_COLOR_WRITEMASK and GL_DEPHT_WRITEMASK to 0x1
     */
  void buildStencilBufferDisplayList();
  /** get trackball location function of mouse pointer location
      @param x pixel X coordinate 
      @param y pixel Y coordinate
      @return trackball vector */
  virtual Vec2<double> getTrackballVector(int x, int y) const {
    return Vec2<double>(x, y);
  }
  /** get trackball rotation from mouse pointer move and store
      results in autoRotation matrix
      @param a mouse click point
      @param b mouse release point */
  virtual Rotation2D getTrackballRotation(const Vec2<double>&a,
					  const Vec2<double>&b);
  /** get trackball scale from mouse pointer move and store
      results in autoRotation matrix
      @param a mouse click point
      @param b mouse release point */
  virtual double getTrackballScale(const Vec2<double>&a, const Vec2<double>&b);
  /** get trackball translation from mouse pointer move
      @param a mouse click point
      @param b mouse release point
      @return translation vector */
  virtual Vec2<double> getTrackballTranslation(const Vec2<double>& a,
				       const Vec2<double>& b);
 /** get trackball translation from mouse pointer movein the Z direction
      @param a mouse displacement in the Z direction orthogonal to image plane
      @return translation vector */
  virtual Vec2<double> getTrackballZTranslation(const double a) {
    return (Vec2<double>(0.0,0.0));
 }
  public:
  friend class TclCamera2D;
  friend class Camera2DBackground;
};


} // namespace yav

#include <graphics2D/Scene2D.h> 

#endif  // ifndef CAMERA2D_H
