#ifndef SCALECAMERA_H
#define SCALECAMERA_H

#pragma interface "graphics2D/ScaleCamera.h"
#include <basics/TclString.h>
#include <basics/Pointed.h>
#include <framework/Activable.h>
#include <basics/DPOwner.h>
#include <graphics2D/Scalable.h>


namespace yav {


/** \class ScaleCamera
    \ingroup Graphics2D
    an interface class defining generic commands for a scale 
    connected to a set of scalables */
class ScaleCamera : public Activable,  
		    public DPOwner<Scalable>,
		    public Pointed<ScaleCamera>
{
 public:
 /** @name constructor */
  //@{
  /** builds a scale object */
  ScaleCamera() :  DPOwner<Scalable>() {}
 /** builds a scale object that manages given list of scalables
      @param list object this scale commands */
  ScaleCamera(std::list<Scalable *>& list) : DPOwner<Scalable>(list){}    
 /** builds a scale object that manages a given scalable
     @param elt object this scale commands */
  ScaleCamera(Scalable *elt) : DPOwner<Scalable>(elt){}    
  //@}

  /** @name destructor */
  //@{
  /// destroys associated TCL scale
  virtual ~ScaleCamera(){}
  //@}
  
 public:
  /** set scale label
      @param lab new scale label */
  virtual void setLabel(const TclString& lab)=0;
  /** set scale bounds
      @param min scale min value
      @param max scale min value */
  virtual void setBounds(double min, double max)=0;
  /** set a new value to scale
      @param val new scale value */
  virtual void set(double val)=0;
  /** set a new value to scale but do not call scale command
      @param val new scale value */
  virtual void to(double val)=0;
  /// set bounds of the scale to min/max of all managed scalables
  virtual void configureToMax(void)=0;
  /// returns scale value
  virtual double get(void) const=0;
  /** checks if the current value is between the min and max value;
      if not then set the value to (min+max)/2 */
  virtual void checkValue(void)=0;
  
  /// activates scale, overloads Activate::on
  virtual void on(void)=0;

  /** callback function for a DPList element being removed
      @param elt element being removed */
  virtual void dying(const Scalable *elt)=0;
  /** adds a new scalable to scale object
      @param sclb scalable to command */
  virtual void operator+=(Scalable *sclb)
    throw(DPList<Scalable>::NullPointer, DPList<Scalable>::AlreadyInList)  {
    DPOwner<Scalable>::operator+=(sclb);}

  /** removes a scalable from a scale object
      @param sclb scalable to release */
  virtual void operator-=(Scalable *sclb) throw(DPList<Scalable>::NotFound) {
    DPOwner<Scalable>::operator-=(sclb);}

 /** reset scale bounds and label
     @param cam calling viewer
     @param lab new scale label (if not EMPTY_STRING) */
  virtual void reset(Scalable *cam,
		     const TclString& lab)= 0;

  /// returns iterator on handled scalable list
  const DPList<Scalable>& getObjects(void) const {
    return elements;
  }
  /// returns scale type
  virtual const std::type_info& getType(void) const =0;

  /// Activation callback: activates scale up and down
  virtual void activation(void){}
};


}


#endif
