#ifndef SLICECAMERA_H
#define SLICECAMERA_H


#pragma interface "graphics2D/SliceCamera.h"
#ifdef GRAPHICS
#include <graphics2D/Camera2D.h>
#include <graphics2D/Sliceable.h>
#include <graphics2D/ScaleCamera.h>
#include <basics/Owner.h>
#include <basics/Rotation3D.h>

#endif

namespace yav {


#ifdef TCL_INTERFACE
class TclSliceCamera;
#endif

#ifdef GRAPHICS
class SliceCamera;

/** \class SliceCameraBackground
    \ingroup Graphics2D
    \brief Background image in a slice camera */
class SliceCameraBackground {
 protected:
  /// the slice camera attached to the background
  SliceCamera *slicecam;
 public:
  /** @name constructor */
  //@{
  SliceCameraBackground(SliceCamera *cam){slicecam=cam;}
  //@}

  /** @name destructor */
  //@{
  virtual ~SliceCameraBackground(){}
  //@}

  /** Given a point in image (voxel) coordinates returns the position in 
      screen coordinates */
  virtual void getPixelPositionFromImageCoordinates(const unsigned int vx,
						    const unsigned int vy,
						    const unsigned int vz,
						    int &x, int&y) const=0;
  /** Given a point in world coordinates returns the position in 
      screen coordinates */
  virtual void getPixelPositionFromWorldCoordinates(const double px,
						    const double py,
						    const double pz,
						    int &x, int&y) const=0;

  /// return the position in image coordinates of a pixel located at (x,y)
  virtual void getImageCoordinates(const int x,const int y, 
				   unsigned int &vx, unsigned int &vy, 
				   unsigned int &vz ) const =0;  
 /// return the position in world coordinates of a pixel located at (x,y)
  virtual void getWorldCoordinates(const int x,const int y, 
				   double &px, double &py, 
				   double &pz ) const=0;  
  /// returns  the X, Y and Z axis in the world coordinates (from image coordinates) 
  virtual Rotation3D getWorldFrame() const=0;
  /// returns the offset position of the lower corner of the image 
  virtual Vec3<double> getImageOffset() const=0;
  /// return the label coresponding to pixel (x,y)
  virtual TclString getLabel(int &x,int &y)=0;
  /// return the viewport of the background
  virtual const unsigned int *getVisibleArea() const=0;
  /** return the 3 scales in the X, Y,Z direction that is used between
      image and world coordinates */
  virtual const double *getScaleVector() =0;
  /** return the 3 voxel sizes in the X, Y,Z direction */
  virtual const double *getVoxelSize() =0;
  /// set visible area to whole image
  virtual void setVisibleArea(void)=0;
  /// get the minimum intensity value in the background
  virtual double getMinIntensity() const=0;
  /// get the maximum intensity value in the background
  virtual double getMaxIntensity() const=0;
  /// get the array of 255 char corresponding to the red colormap
  virtual  const unsigned char *getRedColorMap() const=0;
  /// get the array of 255 char corresponding to the green colormap
  virtual  const unsigned char *getGreenColorMap() const=0;
  /// get the array of 255 char corresponding to the blue colormap
  virtual  const unsigned char *getBlueColorMap() const=0;
  /** set visible area to given bounds (truncate to image size if needed)
      @param x1 visible area X first coordinate
      @param y1 visible area Y first coordinate
      @param z1 visible area Z first coordinate
      @param x2 visible area X second coordinate
      @param y2 visible area Y second coordinate
      @param z2 visible area Z second coordinate */
  virtual void setVisibleArea(unsigned int x1, unsigned int y1, unsigned int z1,
			      unsigned int x2, unsigned int y2, unsigned int z2)=0;
  /// return the SliceCamera attached to the background
  SliceCamera *getCamera() const {return(slicecam);}
};
#endif

/** \class SliceCamera
    \ingroup Graphics2D
    \brief 2D camera that can change displayed scene slice */
class SliceCamera    
#ifdef GRAPHICS
: public Camera2D,public Owner<ScaleCamera>,
  public Scalable
#endif
{
 public:
  /** the different geometry of a slice of a volumetric image */
  enum SLICE_ORIENTATION { 
    /// slice with X (column) = constant
    SLICE_X = 0,
    /// slice with Y (row) = constant
    SLICE_Y = 1,
    /// slice with Z (plane) = constant
    SLICE_Z = 2, 
    /// slice with alpha (angle in XY) = constant
    SLICE_ALPHA = 3
  };

  /// the different image kinds
  enum GEOMETRY {
    /// standard 3D image
    CARTESIAN,
    /// cylindrical geometry (Z = plane angle, Y = elevation, Z = in plane radius)
    CYLINDRICAL,
    /// spherical geometry (Z = plane angle, Y = in plane angle, X = radius)
    SPHERICAL
  };
#ifdef GRAPHICS
 protected:
  /// visible slice
  double slice;
  /// slice thickness
  double thickness;
  /// visible slice geometry
  SLICE_ORIENTATION geometry;
  /// flag indicating whether to die on inrimage deletion
  bool dieOnBackground;
  /// the camera background (if any) or 0
  SliceCameraBackground *background;
  /// offset when drawing with a visible area
  Vec2<double> offset;
 public:
  /** @name constructor */
  //@{
  /** build new camera
      @param scene viewn scene
      @param slice slice to display
      @param geom slice geometry
      @param zoom scale factor
      @param sc a slider controlling the slice value */
  SliceCamera(Scene2D *scene, double slice = 0.0,
	      SLICE_ORIENTATION geom = SLICE_Z,
	      double zoom = 1.0, ScaleCamera *sc=0);
  //@}

  /** @name destructor */
  //@{
  virtual ~SliceCamera();
  //@}

  /// dying inrimage callback
  void backgroundDying(void);
  /// set the background of the slice camera
  void setBackground(SliceCameraBackground *_background);
  /// recompute the display list of all sliceable in the scene
  void updateSliceables();
  /// recompute the display list of all sliceable in the scene
  virtual void update();
  
  /// returns the (slice) scale attached to the camera
  ScaleCamera *getScaleCamera() const;
  /// sets the (slice) scale attached to the camera
  void setScaleCamera(ScaleCamera *sc);
  /// resets the (slice) scale attached to the camera
  void resetScaleCamera();

  /// slicescale owner callback
  virtual void dying(const ObjectPointer<ScaleCamera>& obj) {}
  /** Scalable<SliceScale> objectName method callback
      @return object name */
  virtual TclString objectName(void) const;
  /** slice scale value changed
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleValue(const TclScale *scale, double val);
  /** returns slice name string by geometry
      @param geom slice geometry
      @return slice name */
  virtual const char *sliceName(SliceCamera::SLICE_ORIENTATION geom) const;
  /** returns geometry by slice name string
      @param name slice name
      @return slice geometry */
  virtual SliceCamera::SLICE_ORIENTATION geomByName(const char *name) const;
 
  /// returns slice number
  double getSlice(void) const { return slice; }
  /// returns slice thickness
  double getThickness(void) const { return thickness; }
  /// returns slice scene geometry
  SLICE_ORIENTATION getGeometry(void) const { return geometry; }

  /** set new slice thickness */
  void setThickness(double d) { thickness = d; }
  /** Change slice geometry. Cause recompute method to be called.
      Set slice number to 0.0 */
  void setGeometry(SLICE_ORIENTATION geom);
  /// shows new slice
  void setSlice(double val);

  /** translates camera center and call sliceable objects recompute method
      @param t translation vector */
  virtual void translateCamera(Vec2<double> &t);
  /** scales camera and call sliceable objects recompute method
      @param s scale */
  virtual void scaleCamera(double s);

  /// returns slice camera background
  SliceCameraBackground *getBackground(void) const    {
    return background;
  }
  /// set a new inrimage in scene
  void changeBackground(SliceCameraBackground *b);
  /** get min possible scale value for this scalable
      @param scale calling scale object
      @return min possible value */
  virtual double getMin(const TclScale *scale) const;
  
  /** get max possible scale value for this scalable
      @param scale calling scale object
      @return max possible value */
  virtual double getMax(const TclScale *scale) const;
  /** get the offset for drawing background */
  Vec2<double> getOffset() const {
   return offset;
  }


  /// overloads the computation of the center and scale of the camera
  void doFocus(void);
  /// reset camera size to background size and focus camera on background center
  virtual void focusOnBackground(void);

  
#ifdef TCL_INTERFACE
  /// returns camera interface
  TclSliceCamera *getInterface(void) const;

  friend class TclSliceCamera;
#endif
#endif
};




}



#endif // ifndef SLICECAMERA_H
