#ifndef SURFACEVIEW_H
#define SURFACEVIEW_H

#pragma interface "graphics2D/SurfaceView.h"

#include <graphics2D/SurfaceSliceable.h>
#include <graphics2D/Camera2D.h>

namespace yav {


/** \class SurfaceView
    \ingroup Graphics2D
    \brief a class allowing to display a surface in a slice camera */
class SurfaceView : public Data2D, public Owner<SurfaceSliceable>
{
 public:
  /** the different type of interaction when the left mouse button is pressed,
      moved and released */
  enum INTERACTION_MODE { GRAB_MODE, 
	 POSITION_CONSTRAINT_MODE, 
	 MOVE_POSITION_CONSTRAINT_MODE, 
	 REMOVE_POSITION_CONSTRAINT_MODE};

  /** a structure where are stored points that should be interpolated by the surface */
  struct PointPosition {
    /// point position
    Vec3<double> position;
    /// the identifier of the associated constraint
    unsigned int ref;
  };

 protected:
  /** the maximum square distance (in world coordinares) between a mouse
      click position and a Position constraint location */
  static double selectThreshold;
  /** whether surface trace should be displayed using zone material colors
      or default trace color */
  bool useZonesColor;
  /// default trace width
  float traceWidth;
  /// default trace color
  float  traceColor[3];
  /// the way to interact with a surface model
  INTERACTION_MODE interactionMode;
  /// the list of position constraint associated with this model
  std::list<PointPosition *> constraintList;
  /*** the selected constraint used when moving position constraint mode is active */
  PointPosition *selectedConstraint;
  /// position of the 2 points of a segment when it is required to draw a segment
  Vec2<double> segmentPoint[2];
  /// whether it is necessary to draw a segment
  bool drawSegment;
 public:
  /** @name constructor */
  //@{
  /** build a SurfaceView from a surface sliceable and a 2D scene 
      @param _scene a 2D scene 
      @param _mesh a surface mesh */
  SurfaceView(Scene2D *_scene,SurfaceSliceable *_mesh);
  //@}

  /** @name destructor */
  //@{
  /// destroys a SM2 View
  virtual ~SurfaceView();
  //@}

  /// get the current action when the left mouse button is activated
  INTERACTION_MODE getInteractionMode() const {
    return interactionMode;
  }
  /// set the current action when the left mouse button is activated
  void setInteractionMode(const INTERACTION_MODE mode) {
    interactionMode=mode;
  }
  
  /// get the width of the trace
  float getTraceWidth() const {
    return traceWidth;
  }
  /// set the width of the trace
  void setTraceWidth(const float w) {
   traceWidth=w;
  }
    /// set the color of the trace
  float getTraceColor(unsigned int i) const {
    return traceColor[i%3];
  }
  void setTraceColor(const float col[3]) {
   traceColor[0]=col[0];
   traceColor[1]=col[1];
   traceColor[2]=col[2];
  }
  /// indicates if zones are displayed with the same color or not
  bool getUseZonesColor() const{
    return useZonesColor;
  }

  
  /// set the boolean  controlling the zones color
  void setUseZonesColor(const bool b) {
    useZonesColor=b;
  }
 
 /** callback called when a button is pressed in a camera
     @param button indicates which button is pressed
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int mx,int my);
 /** callback called when a button is released in a camera
     @param button indicates which button is released
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				   CameraTypes::KEY_MODIFIER modifier,
				     Camera2D *cam,int mx,int my){}
 /** callback called when a button is moved after having been pressed
     @param cam the 2D camera where occured the click button
     @param x first coordinate of the pixel clicked 
     @param y second coordinate of the pixel clicked */ 
  virtual void mouseButtonMoved(Camera2D *cam,int x,int y);

  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec2<double>& center, unsigned int& n)  {
    center = Vec2<double>(0.0, 0.0);
    n = 0;
  }
  /// ask the surface to update the display of its slice
  void updateSliceDisplay() {
    getPointer()->updateSliceDisplay();
  }
  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec2<double>& center)  { return 0.0; }
  /** Return MAXDOUBLE, an image is considered as background
      and may not be selected */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const {
    return -1.0;
  }


  /** default redraw function for sliceable objects
      @param cam camera to redraw */
  virtual void redraw(Camera2D *cam);
  /// returns the mesh referenced by this object
  SurfaceSliceable *getSurface() const {
    return Owner<SurfaceSliceable>::getPointer();
  }
 protected:
  /** a mesh owned by this object is being destroyed
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<SurfaceSliceable>& obj);
  
};


} // namespace yav


#endif
