#ifndef TCLCAMERA2D_H
#define TCLCAMERA2D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics2D/TclCamera2D.h"


#include <tcl.h>
#include <tk.h>
#include <modules/togl.h>
#include <modules/TclObject.h>
#include <modules/TclObjectKind.h>
#include <graphics/TclSceneObject.h>
#include <graphics2D/Camera2D.h>

#define DEFAULT_WIDTH		400
#define DEFAULT_HEIGHT		400

namespace yav {

class TclScene2D;
class ClippingPlane;
class Light;

 int createCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);
 int displayCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);
 int resizeCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);
 int destroyCamera2D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);



/** \class TclCamera2D
    \ingroup Graphics2D
    \brief 2D Camera interface using a TCL canvas widget */
class TclCamera2D : public TclSceneObject {
protected:
	/// parent window name
	TclString parentFrame;
	/// OpenGL buffer window name
	TclString cameraFrame;
	/// menu window name
	TclString menuFrame;
	/// submenu window name
	TclString submenuFrame;
	/// buttons window name
	TclString buttonFrame;
  	
    /// whether each new frame should be dumped to a file to create a movie
    bool record;
    /// movie frame number
	unsigned int frameNumber;
	/// movie base file name
	TclString movieName;
	/** in general the mouse motion is only tracked when a button is pressed. This 
	flag allows to track permanently the mouse motion (when background image is present) */
	bool alwaysTrackMouseMotion;

	/// the Togl widget associated with this object
	struct Togl *widget;
  
  // constructors
public:
	/** @name Constructor */
	//@{
	/** builds and display a new camera
	@param m module to build TCL object in
	@param parent the widget name where the camera is to be inserted
	@param scn scene the camera visualizes
	@param type dynamic object type
	@param cam client (if already created) */
	TclCamera2D(TclModule* m, const TclString& parent,TclScene2D *scn,
	const unsigned int w=DEFAULT_WIDTH, const unsigned int h=DEFAULT_HEIGHT,
	const std::type_info& type = typeid(TclCamera2D),
		Camera2D *cam = 0 );
	//@}
	
	/** @name Desstructor */
	//@{
	/// Destroy TclCamera2D object and associated TCL canvas
	virtual ~TclCamera2D();
	//@}
	
	/** returns camera2D client */
	Camera2D *getClient(void) const {
		return (Camera2D *) client;
	}
		/** sets the togl widget associated with the TclCamera3D */ 
	void setWidget( Togl *w) {
		widget=w;
	}

	/// parent window name
	const char *getParent() const {
		return (const char *) parentFrame;
	}
	/// OpenGL buffer window name
	const char *getCameraFrame() const {
		return (const char *) cameraFrame;
	}
	/// menu window name
	const char *getMenuFrame() const {
		return (const char *) menuFrame;
	}
	/// submenu window name
	const char *getSubMenuFrame() const {
		return (const char *) submenuFrame;
	}
	/// buttons window name
	const char *getButtonFrame() const {
		return (const char *) buttonFrame;
	}
	/// movie file name
	const TclString &getMovieName() const {
		return movieName;
	}
	/// return the current frame number and increment it
	unsigned int getFrameNumberAndIncrement() {
		return (frameNumber++);
	}
	/// Bind the OpenGL rendering context to the specified Togl widget
	virtual void makeCurrent(void) const;
	/// Call the glXSwapBuffers() to swap the buffer when using double buffering
	void swapBuffers(void) const;
	/** sets the viewport to (w,h) and updates camera parameters */
	void resize(unsigned int w, unsigned int h);
	/** dump camera window content in an xwd file
      @param filename dump file name */
  virtual void dumpToFile(const TclString& filename);
  /** dump camera window content in an PPM file
      @param filename dump file name */
  virtual void dumpToPPM(const TclString& filename);
/** set the action to perform each time the mouse is moved
      @param on if true then a specific action is performed each time the mouse is moved. If false then unbind the action */
  virtual void bindMouseMotion(bool on);

	/// whether the camera is being destroyed to avoid multiple destruction
	virtual bool isCameraBeingDestroyed() const{
		return getClient()->isCameraBeingDestroyed();
	}
		/// returns the current width
	virtual unsigned int getWidth() const {
		return getClient()->getWidth();
	}
	/// return the current widget height
	virtual unsigned int getHeight() const {
		return getClient()->getHeight();
	}

	/** implies that the mouse motion event should be  tracked whenever it enters the widget */
	void alwaysTrackMouseMotionOn(); 

	/** implies that the mouse motion event should be  tracked only when a mouse button is pressed (default) */
	void alwaysTrackMouseMotionOff(); 

	/** indicates is the mouse motion event is always tracked when it enters the widget */
	bool getAlwaysTrackMouseMotion() const; 

	/** selects a scene object
	@param data object to select */
	void setSelectedObject(Data2D *data);
	
	// install /uninstall the menu of a selected object
	/** Eventually install the Tcl menus associated with a 2D object
	@param obj the selected object */
	virtual void applyObjectSelection (Data2D *obj);
	/** Eventually Uninstall the Tcl menus associated with a 2D object
	@param obj the  object to be unselected */
	virtual void applyObjectUnselection (Data2D *obj);
	
	/** start camera recording
	@param name base file name */
	void startMovie(const char *name);
	/** stop camera recording */
	void stopMovie(void);
	/// is the camera recording  each frame ?
	bool isRecording() const {
		return record;
	} 
	/** pause camera recording
    @param state pause if state = true, record if state = false */
	void pauseMovie(bool state);

	void setMotionMode(Camera2D::MOTION_MODE mm) { 
		getClient()->motionMode = mm; }
	
	
	/// returns camera scene
	TclScene2D *getScene(void) const;
	
	/// -destroy TCL option
	virtual int destroyCmd(int argc, char **argv);
	int viewPointCmd(int argc, char **argv);

	/// -dumpToPPM TCL option
	int dumpToPPMCmd(int argc, char **argv);

	/// -bgColor TCL option
	int bgColorCmd(int argc, char **argv);
	/// -scene TCL option
	int sceneCmd(int argc, char **argv);
	/// -width TCL option
	int widthCmd(int argc, char **argv);
	/// -height TCL option
	int heightCmd(int argc, char **argv);	
	/// -alwaysTrackMouseMotion TCL option
	int alwaysTrackMouseMotionCmd(int argc, char **argv);
	/// -motionMode TCL option
	int motionModeCmd(int argc, char **argv);
	/// -mouseMotion TCL option
	int mouseMotionCmd(int argc, char **argv);
	/// -transformationMatrix TCL option
	int transformationMatrixCmd(int argc, char **argv);
	/// -center TCL option
	int centerCmd(int argc, char **argv);
	/// -refCenter TCL option
	int refCenterCmd(int argc, char **argv);
	/// -scaleFactor TCL option
	int scaleFactorCmd(int argc, char **argv);
	/// -focus TCL option
	int focusCmd(int argc, char **argv);
	

	/// -widget TCL option
	int widgetCmd(int argc, char **argv);
	/// -menu TCL option
	int menuCmd(int argc, char **argv);
	/// -objects TCL option
	int objectsCmd(int argc, char **argv);

	/// -stencilBuffer TCL option
	int stencilBufferCmd(int argc, char **argv);
	
  /// -resize TCL option
  int resizeCmd(int argc, char **argv);
	 /// -redraw TCL option
  int redrawCmd(int argc, char **argv);
  /// -select TCL option
  int selectCmd(int argc, char **argv);

  /// -bindMotion TCL option
  int bindMotionCmd(int argc, char **argv);

  /// -keyPlus TCL option
  int keyPlusCmd(int argc, char **argv);
  /// -keyMinus TCL option
  int keyMinusCmd(int argc, char **argv);
  /// -keyDown TCL option
  int keyDownCmd(int argc, char **argv);
  /// -keyUp TCL option
  int keyUpCmd(int argc, char **argv);
  /// -keyRight TCL option
  int keyRightCmd(int argc, char **argv);
  /// -keyLeft TCL option
  int keyLeftCmd(int argc, char **argv);
 /// -key TCL option
  int keyCmd(int argc, char **argv);

  /// -button1Released TCL option
  int button1ReleasedCmd(int argc, char **argv);
  /// -button2Released TCL option
  int button2ReleasedCmd(int argc, char **argv);
  /// -button3Released TCL option
  int button3ReleasedCmd(int argc, char **argv);
  /// -button1Pressed TCL option
  int button1PressedCmd(int argc, char **argv);
  /// -button2Pressed TCL option
  int button2PressedCmd(int argc, char **argv);
  /// -button3Pressed TCL option
  int button3PressedCmd(int argc, char **argv);
  /// -shiftButton1Pressed TCL option
  int shiftButton1PressedCmd(int argc, char **argv);
  /// -shiftButton2Pressed TCL option
  int shiftButton2PressedCmd(int argc, char **argv);
  /// -shiftButton3Pressed TCL option
  int shiftButton3PressedCmd(int argc, char **argv);
  /// -ctrlButton1Pressed TCL option
  int ctrlButton1PressedCmd(int argc, char **argv);
  /// -ctrlButton2Pressed TCL option
  int ctrlButton2PressedCmd(int argc, char **argv);
  /// -ctrlButton3Pressed TCL option
  int ctrlButton3PressedCmd(int argc, char **argv);
  /// -record TCL command option
  int recordCmd(int argc, char **argv);

 
/** creates a new 3D camera
      @param m module to create camera in
      @param argc command line number of arguments 
      @param argv command line arguments */
  static int create(TclModule *m, int argc, char **argv);
  /** build a list of all available cameras */
  static int list(TclModule *m, int argc, char **argv);
  /// creates the Tcl commands when the module is loaded
  static void createCommands(TclModule *m);

public:
  /// redraw camera only
  virtual void redraw(void);
  /** sets the viewport to (width,height) and updates camera parameters */
  void resize(void);
  /// The action performed when the deformation loop of the scene is activated
  virtual void startComputation();
  /// The action performed when the deformation loop of the scene is de-activated
  virtual void stopComputation();
	
};


}


#endif // TCLCAMERA2D_H
