#ifndef TCLSCALE_H
#define TCLSCALE_H

#pragma interface "graphics2D/TclScale.h"
#include <graphics2D/ScaleCamera.h>
#include <modules/TclSon.h>
#include <framework/Activable.h>


namespace yav {


/** \class TclScale
    \ingroup Graphics2D
    \brief C++ representation of a TCL scale */
class TclScale : public TclSon, public TclObject,
                 public ScaleCamera {
 protected:
  /// max scale value
  double max;
  /// min scale value
  double min;
  /// scale value
  double value;
  /// maximum number of steps when activated
  unsigned int maxSteps;
 public:
  /** @name constructor */
  //@{
  /** builds a scale object from an existing TCL scale
      @param type dynamic type
      @param scaleName TCL scale name (TCL scale must exist)
      @param module module this Scale is created in 
      @param menu the pop-up menu associated with the scale*/
  TclScale(const std::type_info& type, const TclString& scaleName,
	   TclModule *module,TclString menu="graphics2D::scalePopupMenu");
 /** builds a scale object from an existing TCL scale
      manage given list of scalables 
      @param type dynamic object type
      @param module module this SliceScale is created in
      @param scaleName TCL scale name
      @param list scenes this scale commands 
      @param menu the pop-up menu associated with the scale*/
  TclScale(const std::type_info &type, const TclString& scaleName,
		TclModule *module, std::list<Scalable *>& list,TclString menu="graphics2D::scalePopupMenu");
 /** builds a scale object from an existing TCL scale
      manage a given scalable
      @param type dynamic object type
      @param module module this SliceScale is created in
      @param scaleName TCL scale name
      @param elt scalable this scale commands 
      @param menu the pop-up menu associated with the scale*/
  TclScale(const std::type_info &type, const TclString& scaleName,
		TclModule *module, Scalable *elt,TclString menu="graphics2D::scalePopupMenu");
  //@}

  /** @name destructor */
  //@{
  /// destroys associated TCL scale
  virtual ~TclScale();
  //@}
  
  /** set scale label
      @param lab new scale label */
  void setLabel(const TclString& lab);

  /// set bounds of the scale to min/max of all managed scalables
  void configureToMax(void);

  /** set scale bounds
      @param min scale min value
      @param max scale min value */
  void setBounds(double min, double max);

  /** TCL slider moved: call update command if needed
      @param val TCL slider value */
  virtual void command(double val);

  /** set a new value to scale
      @param val new scale value */
  virtual void set(double val);
  
  /** set a new value to scale but do not call scale command
      @param val new scale value */
  virtual void to(double val);
  
  /// returns scale value
  virtual double get(void) const;

  /** checks if the current value is between the min and max value;
      if not then set the value to (min+max)/2 */
  virtual void checkValue(void);
  
  /// activates scale, overloads Activate::on
  virtual void on(void);

  /** callback function for a DPList element being removed
      @param elt element being removed */
  virtual void dying(const Scalable *elt);
  /** adds a new viewer to scale object
      @param inrv viewer to command */
  virtual void operator+=(Scalable *inrv)
    throw(DPList<Scalable>::NullPointer, DPList<Scalable>::AlreadyInList) ;

  /** removes a viewer from a scale object
      @param inrv viewer to release */
  virtual void operator-=(Scalable *inrv) throw(DPList<Scalable>::NotFound);

  /// fix ambigous AutoLink::getRef method call
  virtual unsigned int getRef(void) const {
    return TclObject::getRef();
  }

  /// fix ambigous AutoLink::getHead method call
  static TclScale *getHead(void) {
    return (TclScale *) TclObject::getHead(typeid(TclScale));
  }

  /// fix ambigous AutoLink::following method call
  virtual TclScale *following(void) const {
    return (TclScale *) TclObject::following();
  }
  
  /// returns scale type
  virtual const std::type_info& getType(void) const { return typeid(TclScale); }

  /// -destroy TCL option
  int destroyCmd(int argc, char **argv);
  /// - TCL option
  int widgetCmd(int argc, char **argv);
  /// -name TCL option
  int getCmd(int argc, char **argv);
  /// -get TCL option
  int setCmd(int argc, char **argv);
  /// -set TCL option
  int commandCmd(int argc, char **argv);
  /// -activate TCL option
  int activateCmd(int argc, char **argv);
  /// -from TCL option
  int fromCmd(int argc, char **argv);
  /// -to TCL option
  int toCmd(int argc, char **argv);
  /// -maxSteps TCL option
  int maxStepsCmd(int argc, char **argv);
  /// -resolution TCL option
  int resolutionCmd(int argc, char **argv);
  /// TCL scalables command
  static int scalablesCmd(TclModule *m, int argc, char **argv);
  /// TCL -objects option
  int objectsCmd(int argc, char **argv);
  /// TCL -add option
  int addCmd(int argc, char **argv);
  /// TCL -remove option
  int removeCmd(int argc, char **argv);
  /// TCL -die option
  int dieCmd(int argc, char **argv);
 private:
  /// indicates whether this scale should kill itself when its DPList becomes empty
  bool dieWhenEmpty;
  /// init tcl commands
  void init(const std::type_info& type,TclString menu);
 protected:
  /// has the scale already been updated ?
  bool firstupdate;

  /// Activation callback: activates scale up and down
  virtual void activation(void);

  /// callback function for scales when slider moved
  virtual void update(double val) const {}

 private:
  /// activation direction 
  short int activationDir;
};


} // namespace yav


#endif
