#ifndef CAMERA3D_H
#define CAMERA3D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#pragma interface "graphics3D/Camera3D.h"
#include <basics/Bool.h>
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Vec2.h>
#include <basics/ListElement.h>
#include <linearAlgebra/Matrix3x3.h>
#include <linearAlgebra/Rotation3D.h>
#include <linearAlgebra/Matrix4x4.h>
#include <graphics/Camera.h>
#include <graphics3D/Light.h>
#ifdef _MSC_VER
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>




namespace yav {
class Light;
class ClippingPlane;
/************************************************************************/
/************************************************************************/
/**                           CLASS Camera                             **/
/************************************************************************/
/************************************************************************/

/// the default fog range parameter
#define DEFAULT_FOG_RANGE 1.0

#define DEFAULT_CAMERA3D_CENTER Vec3<double>(4.0, 0.0, 1.0);
/// Default scene field of view
#define DEFAULT_FIELD_OF_VIEW 30.0
// represent the ratio of the radius of a graphical object
// used for picking an object
#define RATIO_OBJECT_SELECTION 0.1
/// Default scene depth ratio
#define DEFAULT_FIELD_DEPTH_RATIO 5.0

class Scene3D;
class Data3D;
class Camera3DActivation;

/// the offset to draw lines on top of rendered objects

#ifdef _LINUX_
#define lineDisplacementOffset 0.05
#else
#define lineDisplacementOffset 0.01
#endif 

#define fontType TclObj *

/** \class Camera3D
    \ingroup Graphics3D
    \brief Generic camera object for 3D Objects */
 class Camera3D : public ListElement<Camera3D>,
	 public Camera<Vec3<double>, Rotation3D, Scene3D,Data3D, Camera3D>
 {
 public:
	 /// the type of function used to select sub-items of a Data
	 typedef void (*Selection3DFunctionType)(Data3D *,const Vec2<double> pixel,
		 Camera3D *cam);


  protected:
 /** called before the destruction of a camera : removes clipping planes,
 lights and stencil display list */
  void destroyCamera();
  /// is the camera being deleted ?
  bool beingDestroyed;
  /// a callback function for user defined selection
  Selection3DFunctionType selectionFunction;
  /// should we compute the shading on both sides?
  bool useTwoSides;
  /// should we compute the true specularity component in the shading computation
  bool useTrueSpecularity;
  /// whether to use fog for inrimage display
  bool fog;
  /// whether to use stencil Buffer for display
  bool stencilBuffer;
  /// the display list associated with a stencil
  GLuint stencilBufferDisplayList;

  /// fog range
  double fogRange;
  /// moved from Togl to implement openGl stereo on capable hardware
  int StereoFlag;
 
  /// vector defining the displacement between the stereo cameras
  Vec3<double> eyeOffset;

  // display list of the default font
  GLuint defaultFont;
  // display list of the large font
  GLuint largeFont;
  // display list of the small font
  GLuint smallFont;
  
  /// Counting number of redraws 
  unsigned int count;
  ///  Frame Ratestring
  char stringFrameRate[100];
  /// copyright string
  char stringCopyright[100];
  ///  if copyright should be displayed
  bool showCopyright;
  ///  if frame rate should be displayed
  bool showFrameRate;
  /// auto-move mode used (none, rotation, translation or scale)
  AUTO_MOVE_MODE autoMoveMode;
  /// scale factor to apply while auto-moving
  double autoScale;
  /* The angle from which is seen the scene*/
  double fieldOfView;
  /* The ratio defining the field depth as a function of the radius */
  double fieldDepthRatio;
  /// the field depth ratio multiplied by the radius
  double dl;
  /** the expansion in the X direction computed as dl*tan(fieldOfView)/(width/2) */
  double expansionX;
  /** the expansion in the X direction computed as dl*tan(fieldOfView)/(height/2) */
  double expansionY;
 
  /** 3D mouse position  corresponding to the 3D point defining the 3D line defined when a mouse button is clicked, the other extremity of the line being the center of the camera*/
  Vec3<double> mouse;
    /// allocated clipping planes
  std::list<ClippingPlane *> clipPlanes;
  /// allocated light
  std::list<Light *> lights;
   /// rotation to apply while auto-moving
  Rotation3D autoRotation;
  /// translation to apply while auto-moving
  Vec3<double>  autoTranslation;
   /* Projection matrix */
  Matrix4x4<double> matTransform;
  /// camera position
  Vec3<double> eye;
  /// perspective center
  Vec3<double> refCenter;
  ///  Measuring the time between two redraws action
  Timer globalTimer;
 /// an object that handles 
  Camera3DActivation *cameraActivation;
 
 protected:
  /** @name constructor */
  //@{
  /** builds a camera3D by calling the Camera() constructor 
      @param c unused parameter */
  Camera3D(Camera3D *c);

 public:
  /** Creates a new camera. The camera visualize given scene. If scn is 0
      @param scn the scene where the camera should look at. If scn is 0 the camera visualize the first scene in global scene list (if any) or a new scene is created */
  Camera3D(Scene3D *scn = 0);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~Camera3D();
  //@}

 /** set scale function on subobject
      @param sf new scale function */  
  void setSelectionFunction(Selection3DFunctionType sf);
	
  /** set default font */
  void setDefaultFont(GLuint f) {
	  defaultFont=f;
  }
   /** set small font */
  void setSmallFont(GLuint f) {
	  smallFont=f;
  } /** set large font */
  void setLargeFont(GLuint f) {
	  largeFont=f;
  }
  
  /** subobject selection method
      @param vec mouse click point
      @param cam clicked camera */
  virtual void select(const Vec2<double> vec, Camera3D *cam);
 
  /** mouse button was pressed
      @param w which button
      @param k which key modifier
      @param x mouse x coordinate
      @param y mouse y coordinate */
  virtual void buttonPressed(CameraTypes::WHICH_BUTTON w, CameraTypes::KEY_MODIFIER k,
		     int x, int y);
  /** mouse button was released
      @param w which button
      @param x x mouse coordinate
      @param y mouse y coordinate */
  virtual void buttonReleased(CameraTypes::WHICH_BUTTON w, int x, int y);
  /** mouse pointer moved
      @param x X mouse location
      @param y Y mouse location */
  virtual void mouseMoved(int x, int y);
  /** activates a camera for auto-move
      @param amm auto-move mode */
  virtual void startAutoMove(AUTO_MOVE_MODE amm);
  /// stops camera auto-move
  virtual void stopAutoMove(void);
  /// move the camera or object according to the current auto move mode
  virtual void applyAutoMove();
  /// return the object dealing with the camera activation
  Camera3DActivation *getCameraActivation() const;
  /// set the object dealing with the camera activation
  void setCameraActivation(Camera3DActivation *activation); 
  /** updates the Viewing matrix based on refcenter, center and the 
      matTransform matrix */
  virtual void applyTransformation();
  /// The routines that computes the 3D lines associated with a given pixel
  Vec3<double> computeEye(void);
  /// returns the reference center of the camera
  Vec3<double> getRefCenter(void) const { return refCenter; }
  /// set the reference center of the camera
  void setRefCenter(Vec3<double> rc) { refCenter = rc; }
  /** returns the 4x4 matrix in homogeneous coordinate corresponding to 
      the actual rotation, translation and scale performed. It corresponds
      to the Viewing matrix in OpenGL */
  Matrix4x4<double> getMatTransform(void) const { return matTransform; }
  /** set the 4x4 Viewing matrix in homogeneous coordinate */
  void setMatTransform(Matrix4x4<double> mat){ matTransform = mat; }
  /// computes expansionX, expansionY and dl variables given the new height and width of the camera
  void setParameters(void);
  /** returns the mouse position vector given the  a pixel position.
      The mouse position is a 3D point defining the 3D line defined by a camera pixel, the other extremity of the line being the center of the camera
      @param pixel the pixel position
      @return the mouse position in world coordinates */
  Vec3<double> getMousePosition(const Vec2<double> &pixel);
  /** returns the 3D line associated with a pixel of the camera
      @param pixel the pixel position
      @param eye the coordinate of the camera center
      @param dir the unit vector designed the direction of the 3D line pointing from the camera center */
  void getEyeDirectionVector(const Vec2<double> &pixel, 
			     Vec3<double>&eye, Vec3<double>&dir);
  /** returns the field of view of the camera */
  double getFieldOfView() const {
    return(fieldOfView); }
  /** set the field of view of the camera */
  void setFieldOfView(const double f) {
   fieldOfView=f; }
  /** returns the field of view of the camera */
  double getFieldDepthRatio() const {
    return(fieldDepthRatio); }
  /** set the field of view of the camera */
  void setFieldDepthRatio(const double f) {
    fieldDepthRatio=f; }
  /** compose rotation matrix of camera with the rotation defined by 
      (axis, angle)
      @param axis the unit vector describing the rotation axis
      @param angle the angle of rotation in radian */
  void rotate(const Vec3<double>& axis,double angle);
 
  /** rotates selected object
      @param r rotation matrix */
  virtual void rotateObject(const Rotation3D &r);

  /** rotates camera
     @param rot rotation matrix (unsused) */
  virtual void rotateCamera(Rotation3D& rot);
  /** translates camera center
      @param t translation vector */
  virtual void translateCamera(Vec3<double> &t);
  /** scales camera
      @param s scale */
  virtual void scaleCamera(double s);

  /// focus whatever autofocus says
  void doFocus(void);
  /** focus the camera on a given region defined as a sphere
      @param c the region center
      @param _radius the region radius */
  void focusOnRegion(Vec3<double> &c,double _radius);

  /// function called at the creation of the Camera3D : initialized as empty
  virtual void init(void);
  /// function that defines the lights applied with the camera : initialized as empty
  virtual void defineLights(void);
  /// function that use given lights : initialized as empty
  virtual void useLights(void) const;
 

  /** transforms a translation vector from screen coordinate to World coordinates
      @param t the translation vector in screen coordinates */
  Vec3<double> mapToScreen(const Vec3<double>& t);
  /** Draw a bounding box around a 3D object
      @param obj the object around which the bounding box is drawn */
  virtual void drawBoundingBox(Data3D *obj) const;

  /** Redraw all existing 3D objects in the 3D scene  and eventually 
      objects belonging to the camera (clipping planes, spot loghts,..) */
  virtual void redraw(void);
  
  /** initialize  the viewport and basic OpenGL parameters */
  virtual void initViewport(void);
  /** sets the viewport to (width,height) and updates camera parameters */
  virtual void resize(void);
  /** sets the viewport to (width,height) and updates camera parameters */
  virtual void resize(unsigned int w, unsigned int h);
  /*** sets the background color of the camera
       @param r the red component
       @param g the green component
       @param b the blue component */
  virtual void setBackgroundColor(float r, float g, float b);
  /** updates the internal camera parameters after the camera size has changed
      @param reset wether tp init the transformation matrix to identity */
  virtual void setViewingParameters(bool reset=true);
  /** set the OGL projection matrix 
      @param reset wether tp init the transformation matrix to identity */
  virtual void setProjectionParameters(bool reset=true);

  /** set new fog flag value
      @param nfog new fog value */
  void setFog(bool nfog);
  /** print a string on the screen 
      @param posx the position in the window (column index) 
      @param posy the position in the window (row index) 
      @param str the null terminated string
      @param font must be either defaultFont, smallFont or largeFont */
  void printString(double posx,double posy, char *str, GLuint font=0);


  /** set stencilBuffer flag value
      @param nstb new stencilBuffer value */
  void setStencilBuffer(bool nstb);

  /** build stencilBufferDisplayList
    initialize GL_COLOR_WRITEMASK and GL_DEPHT_WRITEMASK to 0x1
     */
  void buildStencilBufferDisplayList();


  /// returns fog flag
  bool getFog(void) const;

  /** set new specularity flag value
      @param specularity new specularity value */
  void setSpecularity(bool specularity);

  /// returns true specularity flag
  bool getSpecularity() const;

  /** set front or front and back rendering
      @param flag front (if false) or front and back (if true) flag */
  void setRenderFrontAndBack(bool flag);

  /// returns front and back rendering flag
 bool getUseTwoSides() const;

  // set/get the offset between eyes for stereo mode.
  // As mean human have the two eyes horizontally separated by 7cm,
  // the default value is (0.035, 0, 0).
  void setEyeOffset(const Vec3<double> &v);
  Vec3<double> getEyeOffset(void);

  /** camera callback to catch trackball rotation
      @param rot rotation matrix */
  virtual void rotateCameraObject(Rotation3D &rot);
  /** camera callback to catch trackball translation
      @param t translation vector */
  virtual void translateCameraObject(Vec3<double> &t);

  /// change the display of the copyright
  void toggleCopyright();
  /// change the display of the frame rate
  void toggleFrameRate();
 /** remove a light in the list of lights */
  void removeLight(Light *l); 
  /** add a light in the list of lights */
  void addLight(Light *l); 
  /** creates a new light and add it to the list */
  Light *newLight(); 
  /** get a light from index */
  Light *getLight(unsigned int n);
  /** returns the maximum number of lights */
  unsigned int getNumberOfLights(){
	  return lights.size();
  }
  /// returns iterator on first clipping plane
  std::list<ClippingPlane *>::iterator getFirstClippingPlane();
  /// returns iterator on last clipping plane
  std::list<ClippingPlane *>::iterator getLastClippingPlane();
  /// returns iterator on first clipping plane
  std::list<Light *>::iterator getFirstLight();
  /// returns iterator on last clipping plane
  std::list<Light *>::iterator getLastLight();

  /** creates a new clipping plane and add it to the list */
  ClippingPlane *newClippingPlane(); 
  /** get a clipping plane from index */
  ClippingPlane *getClippingPlane(unsigned int n);
  /** returns the maximum number of clipping planes */
  unsigned int getNumberOfClippingPlanes(){
	  return clipPlanes.size();
  }
  /** remove a light in the list of clipping planes */
  void removeClippingPlane(ClippingPlane *l); 
  ///is the camera being destroyed?
  bool isCameraBeingDestroyed() const;

  /// returns an unused OpenGL clipping plane id
  GLenum getFreeClippingPlane(void) const;
  /// returns an unused OpenGL light id
  GLenum getFreeLight(void) const;

protected:
  /** get trackball location function of mouse pointer location
      @param x pixel X coordinate 
      @param y pixel Y coordinate
      @return trackball vector */
  virtual Vec3<double> getTrackballVector(int x, int y) const;
  /** get trackball rotation from mouse pointer move and store
      results in autoRotation matrix
      @param a mouse click point
      @param b mouse release point */
  virtual Rotation3D getTrackballRotation(const Vec3<double>&a,
					  const Vec3<double>&b);
  /** get trackball scale from mouse pointer move and store
      results in autoRotation matrix
      @param a mouse click point
      @param b mouse release point */
  virtual double getTrackballScale(const Vec3<double>&a, const Vec3<double>&b);
  /** get trackball translation from mouse pointer move
      @param a mouse click point
      @param b mouse release point
      @return translation vector */
  virtual Vec3<double> getTrackballTranslation(const Vec3<double>& a,
					       const Vec3<double>& b);
  /** get trackball translation from mouse pointer movein the Z direction
      @param a mouse displacement in the Z direction orthogonal to image plane
      @return translation vector */
  virtual Vec3<double> getTrackballZTranslation(const double a);
  /** transforms a pixel position in standard camera coordinates 
      @param pixel the pixel position in the camera frame
      @param wx the X mouse position in standard camera coodinates
      @param wy the Y mouse position in standard camera coodinates */
  void mapworld2(const Vec2<double> &pixel, double *wx, double *wy);

 private:
  /** adds the camera in the corresponding scene and sets all protected 
      parameters to their initial values */
  void initCamera(void);
public:
  /** get state of mouse buttons */
  BUTTON_STATE getButtonState(WHICH_BUTTON w)
    {return B[w];}
 public:
  friend class ClippingPlane;
  friend class TclCamera3D;
  
};

} // namespace yav

#include <graphics3D/Scene3D.h> 



#endif
