#ifndef CLIPPINGPLANE_H
#define CLIPPINGPLANE_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/ClippingPlane.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <graphics3D/Plane.h>


namespace yav {

/** \class ClippingPlane
    \ingroup Graphics3D
    \brief an OpenGL clipping plane */
class ClippingPlane : public Plane {
 protected:
  /// OpenGL clipping plane id
  GLenum index;
  /// is plane active
  bool active;
  /// whether the clipping plane is being displaced
  bool moving;
 public:
  /** @name destructor */
  //@{
  /** construct a new clipping plane passing through scene center and
      orthogonal to (0, 0, 1)
      @param _index the OpenGL index of the clipping plane
	  @param z the z value of the plane (initialized from a cemera as refCenter.getZ()) */
  ClippingPlane(unsigned int _index, double z);
  /** construct a new clipping plane with plane equation
      equation[0] x + equation[1] y + equation[2] z + equation[3] = 0
      @param _index the OpenGL index of the clipping plane
      @param equation four plane equation coefficients */
  ClippingPlane(unsigned int _index, const GLdouble *equation);
  /** construct a new clipping plane with passing through P and
      orthogonal to n
      @param _index the OpenGL index of the clipping plane
      @param P one clipping plane point
      @param n clipping plane normal */
  ClippingPlane(unsigned int _index, const Vec3<double>& P, const Vec3<double>& n);
  //@}
  
  /** @name constructors */
  //@{
  ~ClippingPlane();
  //@}

  /// enable (use) clipping plane
  void enable(void);
  /// disable clipping plane
  void disable(void);

  /// returns OpenGL clipping plane index
  GLenum getIndex(void) const { return index; }

  /// returns clipping plane active flag
  bool isActive(void) const { return active; }

  /// whether the clipping plane is being displaced
  bool isMoving(void) const { return moving; }
  /// indicates that the clipping plane is moving
  void movePlane() { moving=true; }
  /// indicates that the clipping plane is not moving
  void fixPlane() { moving=false; }
  /** set new plane equation: eqn[0] x + eqn[1] y + eqn[2] z = 0
      @param eqn plane equation */
  virtual void setEquation(const double *eqn);

  /** set new plane equation for plane passing through P with
      normal n
      @param P one plane point
      @param n plane normal */
  virtual void setEquation(const Vec3<double>& P, const Vec3<double>& n);

  /** translates plane in space
      @param t translation vector */
  virtual void translate(const Vec3<double>& t);

  /** rotates plane in space
      @param r rotation matrix */
  virtual void rotate(const Rotation3D& r);

  /** rotates plane in space around one point
      @param r rotation matrix
      @param P rotation center */
  virtual void rotate(const Rotation3D& r, const Vec3<double>& P);

  /// enables clipping plane when camera request it
  void putInCamera(void) const;

};

} // namespace yav



#endif // ifdef CLIPPINGPLANE_H
