#ifndef COLOR_H
#define COLOR_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Color.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <basics/Exception.h>
#include <modules/TclModule.h>
#include <graphics3D/Data3D.h>
#include <graphics3D/Material.h>

namespace yav {


/** \class Color
    \ingroup Graphics3D
    \brief Class for Managing color strings and names */
class Color {
 public:
  /// string representation mode
  enum StringMode {
    /// uninitialized
    NONE,
    /// #RGB mode
    THREE_DIGITS,
    /// #RRGGBB mode
    SIX_DIGITS,
    /// #RRRGGGBBB mode
    NINE_DIGITS,
    /// #RRRRGGGGBBBB mode
    TWELVE_DIGITS
  };

  /** @name exception */
  //@{
/** \class InvalidString
    \ingroup Graphics3D
    \brief invalid color string transmitted */
  class InvalidString : public Exception {
  public:
    InvalidString(const std::string& from);
  };
  //@}

 protected:
  /// RGBA color components
  GLfloat color[4];
  /// color string
  char *str;
  /// str mode
  StringMode mode;

 public:
  /** @name constructors */
  //@{
  /// default (white) color
  Color() : str(0), mode(NONE) {
    color[0] = color[1] = color[2] = color[3] = 1.0f;
  }
  /** creates a new color, each component should be between 0 and 1
      @param r red color component
      @param g green color component
      @param b blue color component
      @param a alpha component */
  Color(GLfloat r, GLfloat g, GLfloat b, GLfloat a = 1.0f)
    : str(0), mode(NONE) {
    setColor(r, g, b, a);
  }
  /** creates a color from a string
      @param src color string, should be #RGB, #RRGGBB, #RRRGGGBBB or #RRRRGGGGBBBB
      @exception InvalidString src does not represent a color */
  Color(const char *src) throw(InvalidString);
  /** creates a new color from material color
      @param mat reference material */
  Color(const Material& mat);
  /// copy constructor
  Color(const Color& c);
  //@}

  /** @name destructor */
  //@{
  ~Color() { if(str) delete[] str; }
  //@}

  /** set color, each component should be between 0 and 1
      @param r red color component
      @param g green color component
      @param b blue color component
      @param a alpha component */
  void setColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a = 1.0f) {
    color[0] = r < 0.0f ? 0.0f : (r > 1.0f ? 1.0f : r);
    color[1] = g < 0.0f ? 0.0f : (g > 1.0f ? 1.0f : g);
    color[2] = b < 0.0f ? 0.0f : (b > 1.0f ? 1.0f : b);
    color[3] = a < 0.0f ? 0.0f : (a > 1.0f ? 1.0f : a);
    mode = NONE;
  }

  /// assignement operator
  Color& operator=(const Color& c);

  /// returns color values as a 4 components array
  const GLfloat *getColor(void) const { return color; }
  /// returns color red component
  GLfloat getRed(void) const { return color[0]; }
  /// returns color green component
  GLfloat getGreen(void) const { return color[1]; }
  /// returns color blue component
  GLfloat getBlue(void) const { return color[2]; }
  /// returns color alpha component
  GLfloat getAlpha(void) const { return color[3]; }

  /// set color in GL window
  void glSet(void) const { glColor4fv(color); }
  // convert RGB color to HSV color
  void RGBtoHSV(Color & col);
  // convert HSV color to RGB color
  void HSVtoRGB(Color & col);

  /** set color value from a #RGB, #RRGGBB, #RRRGGGBBB or #RRRRGGGGBBBB
      string
      @param src string to parse
      @return true if src is valid, false otherwise */
  bool scan(const char *src);

  /* print color value as a #RGB, #RRGGBB, #RRRGGGBBB or #RRRRGGGGBBBB
      string */
  const char *print(StringMode mode = SIX_DIGITS);

  /// -color TCL command
  int colorCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);

  /// -color option synopsis
  static const char *synopsis;
};


} // namespace yav


#endif // ifndef COLOR_H
