#ifndef FORCEFEEDBACKINTERFACE3D_H
#define FORCEFEEDBACKINTERFACE3D_H

#pragma interface "graphics3D/ForceFeedbackInterface3D.h"

#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Matrix3x3.h>
#include <basics/Pointed.h>
#include <linearAlgebra/Rotation3D.h>
#include <graphics3D/UserInterface3D.h>

namespace yav {



/** \class LaparoscopicInterface3D
    \ingroup Graphics3D
    3D laparoscopic Interface that correspond to a specific user interface.
    It provides a member fonction to get the fixed point of the device
    
    have an "getPosition" and "setForce" routines that are  suitable 
    for interaction with a Scene2D */
class LaparoscopicInterface3D : public UserInterface3D, public Pointed <LaparoscopicInterface3D> {
 public:
  /// the different types of surgical tools that can be simulated with the LIE
  enum SurgicalToolType {PLIERS, CAUTERY, CAVITRON, CLAMP, CAMERA} ;
 protected:
  /// the type of surgical instrument simulated by the LIE
  SurgicalToolType toolType;
  /// the rotation to apply on all coordinates
  Rotation3D rotation;
  /// the translation to apply on all coordinates
  Vec3<double> translation;
  ///  scale factor to apply on all coordinates
  double coordScale;
  ///  scale factor to apply on all force vectors
  double forceScale;
  /// length of the cylinder defining the LIE
  double length;
  /// radius of the cylinder defining the LIE
  double LIEradius;
  /// the gripper length
  double gripperLength;
  /// the gripper width
  double gripperWidth;
 public:
  /** @name constructor */
  //@{
  /** builds a new Laparoscopic Interface
      @param scene  scene where the laparoscopic  interface object should be added (if scene is 0, the first existing scene is used or a scene is created== 0)
      @param l the tool length
      @param r the tool LIEradius
      @param gl the gripper length
      @param gw the gripper width
 */
  LaparoscopicInterface3D(Scene3D *scene,double l=1.0, double r=-1,double gl=-1,double gw=-1);
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~LaparoscopicInterface3D(){}
  //@}

  /// returns the translation applied on all coordinates
  virtual Vec3<double> getTranslation() const{
    return(translation);}
  /// sets the translation applied on all coordinates
  virtual void setTranslation(const Vec3<double> &t)  {
    translation=t; }
  /// returns the translation applied on all coordinates
  virtual Rotation3D getRotation() const{
    return(rotation);}
  /// sets the rotation applied on all coordinates
  virtual void setRotation(const Rotation3D  &r)  {
    rotation=r; }
  /// returns the type of surgical tool simulated by the LIE
  SurgicalToolType getSurgicalToolType () const {
    return(toolType); }
  /// returns the scale factor applied on all coordinates
  virtual double getLength() const{
    return(length);}
  /// returns the gripper length
  virtual double getGripperLength() const{
    return(gripperLength);}
  /// returns the gripper width
  virtual double getGripperWidth() const{
    return(gripperWidth);}
  /// returns the radius of the cylinder defining the LIE main body
  virtual double getLIEradius() const{
    return(LIEradius);}
  /// returns the scale factor applied on all force vectors
  virtual double getForceScale() const{
    return(forceScale);}
  /// returns the scale factor applied on all coordinates
  virtual double getCoordScale() const{
    return(coordScale);}
  /// sets the LIE tool length
  virtual void setLength(const double &s)  {
    length=s; }
  /// sets the LIE tool LIEradius
  virtual void setLIEradius(const double &s)  {
    LIEradius=s; }
  /// sets the gripper length
  virtual void setGripperLength(const double &s)  {
    gripperLength=s; }
  /// sets the gripper width
  virtual void setGripperWidth(const double &s)  {
    gripperWidth=s; }
  /// sets the scale applied on all force vectors
  virtual void setForceScale(const double &s)  {
    forceScale=s; }
  /// sets the scale applied on all coordinates
  virtual void setCoordScale(const double &s)  {
    double sf = s/coordScale;
    length*=sf;
    LIEradius*=sf;
    gripperLength*=sf;
    gripperWidth*=sf;
    coordScale=s;
  }
  /// returns the type of surgical tool simulated by the LIE
  virtual void  setSurgicalToolType (const SurgicalToolType st)  {
    toolType=st; }
  /// send a force to the LIE device
  virtual void setForce(const Vec3<double> &f) = 0;
  /// returns the fixed point of the laparoscopic device
  virtual Vec3<double> getFixedPoint() =0;
  /// returns the angle of rotation of the gripper along the tool axis
  virtual double getAxisAngle() =0;
  /// returns the end point of the laparoscopic device
  virtual Vec3<double> getEndPoint()=0;
  /// get the gripper axis (Y Axis)
  virtual Vec3<double> getGripperAxis()=0;
  inline Vec3<double> getYAxis(){return getGripperAxis();}
  /// get the Z Axis
  virtual Vec3<double> getZAxis(){
    Vec3<double> Z = getEndPoint()-getFixedPoint();
    double l = Z.norm();
    if(l!=0.0) return Z/l;
    else return Vec3<double>();
  }
  // get the X Axis
  virtual Vec3<double> getXAxis(){
    return cross(getYAxis(),getZAxis());
  }
  /// get the gripper end point
  virtual Vec3<double> getGripperEndPoint(){return getEndPoint()+getZAxis()*getGripperLength()*cos(getGripperAngle());}
  /// returns the angle of the gripper in radian unit (0 = closed)
  virtual double getGripperAngle()=0;
  /// returns the previous angle of rotation of the gripper along the tool axis
  virtual double getPreviousAxisAngle(){return getAxisAngle();}
  /// returns the previous end point of the laparoscopic device
  virtual Vec3<double> getPreviousEndPoint(){return getEndPoint();}
  /// returns the previous angle of the gripper in radian unit (0 = closed)
  virtual double getPreviousGripperAngle(){return getGripperAngle();};
  /// get the previous gripper axis (Y Axis)
  virtual Vec3<double> getPreviousGripperAxis(){return getGripperAxis();}
  inline Vec3<double> getPreviousYAxis(){return getPreviousGripperAxis();}
  /// get the old Z Axis
  virtual Vec3<double> getPreviousZAxis(){return getZAxis();}
  // get the old Y Axis
  virtual Vec3<double> getPreviousXAxis(){return getXAxis();}
  /// get the old gripper end point
  virtual Vec3<double> getPreviousGripperEndPoint(){return getPreviousEndPoint()+getPreviousZAxis()*getGripperLength()*cos(getPreviousGripperAngle());}
  /// returns the number ID of the laparoscopic device
  virtual int getDeviceNumber()const=0;
  /** test is the gripper at the tip of the laparoscopic
      device is opened or closed */
  virtual bool isGripperOpened()=0;
  /** add a new constraint to send to iLIE
      @param n the normal of the plane
      @param d the distance to 0
   */
  virtual void addConstraint(Vec3<double> n, double d){};

  friend class Scene3D;
};



/** \class PhantomInterface3D
    \ingroup Graphics3D
    \brief 3D PHANToM Interface **/
class PhantomInterface3D : public UserInterface3D {
 public:
  /// the different types of surgical tools that can be simulated with the PHANToM
  enum SurgicalToolType { POINTER } ;
 protected:
  /// the type of surgical instrument simulated by the PHANToM
  SurgicalToolType toolType;
  /// the rotation to apply on all coordinates
  Rotation3D rotation;
  /// the translation to apply on all coordinates
  Vec3<double> translation;
  ///  scale factor to apply on all coordinates
  double coordScale;
  ///  scale factor to apply on all force vectors
  double forceScale;

  /// length of the phantom stylus
  double stylusLength;
  /// radius of the cylinder defining the stylus (actually stylusRadius == tipRadius)
  double stylusRadius;
  /// length of the phantom tip
  double tipLength;
  /// radius of the cone defining the tip (actually stylusRadius == tipRadius)
  double tipRadius;
  /// length of the phantom stylus+tip. This is a dependent parameter. Do not set directly.
  double phantomLength;
  /// the radius of the phantom, i.e. max(r_stlyus, r_tip). This is a dependent parameter. Do not set directly. (actually stylusRadius == tipRadius)
  double phantomRadius;

 public:
  /** @name constructor */
  //@{
  /** builds a new PHANToM Interface
      @param scene scene to which the PHANToM interface object will be added (if scene is 0, the first existing scene is used or a scene is created== 0)
      @param sl length of the phantom stylus
      @param sr radius of the cylinder defining the stylus
      @param tl length of the phantom tip
      @param tr radius of the cone defining the tip
 */
  PhantomInterface3D(Scene3D *scene, double sl, double sr, double tl, double tr); 
  //@}
  
  /** @name destructor */
  //@{
  /// empty
  virtual ~PhantomInterface3D(){}
  //@}

  /// returns the translation applied on all coordinates
  virtual Vec3<double> getTranslation() const {
    return(translation);
  }
  /// sets the translation applied on all coordinates
  virtual void setTranslation(const Vec3<double> &t) {
    translation=t;
  }
  /// returns the translation applied on all coordinates
  virtual Rotation3D getRotation() const {
    return(rotation);}
  /// sets the rotation applied on all coordinates
  virtual void setRotation(const Rotation3D  &r) {
    rotation=r;
  }
  /// returns the type of surgical tool simulated by the PHANToM
  SurgicalToolType getSurgicalToolType () const {
    return(toolType);
  }
  /// returns the scale factor applied on all coordinates
  virtual double getScale() const {
    return(coordScale);
  }
  /// returns the scale factor applied on all force vectors
  virtual double getForceScale() const {
    return(forceScale);
  }
  /// sets the scale factor applied on all coordinates
  virtual void setScale(const double &s) {
    coordScale = s;
  }
  /// sets the scale applied on all force vectors
  virtual void setForceScale(const double &s) {
    forceScale = s;
  }
  /// sets the type of surgical tool simulated by the PHANToM
  virtual void  setSurgicalToolType (const SurgicalToolType st) {
    toolType=st;
  }

  /// get the tip position at a given simulation step (0=current, -1=previous, ...)
  virtual Vec3<double> getTipPosition(int n=0) = 0;
  /// get the bottom position at a given simulation step (0=current, -1=previous, ...)
  virtual Vec3<double> getBottomPosition(int n=0) = 0;
  /// get the rotation matrix at a given simulation step (0=current, -1=previous, ...)
  virtual Rotation3D getRotationMatrix(int n=0) = 0;
  /// get the stylus orientation at a given simulation step (0=current, -1=previous, ...)
  virtual Vec3<double> getStylusOrientation(int n=0) = 0;
  /// get the switch orientation at a given simulation step (0=current, -1=previous, ...)
  virtual Vec3<double> getSwitchOrientation(int n=0) = 0;
  /// get the PHANToM button status at a given simulation step (0=current, -1=previous, ...)
  virtual bool isStylusButtonPressed(int n=0) = 0;
  /// get the time stamp of a given simulation step (0=current, -1=previous, ...)
  virtual double getTimeStamp(int n=0) = 0;

  /// send a force to the PHANToM device
  virtual void setForce(const Vec3<double> &f)=0;


  /// returns the phantom length
  virtual double getPhantomLength() const{
    return(phantomLength);}
  /// returns the phantom radius
  virtual double getPhantomRadius() const{
    return(phantomRadius);}

  /// returns the stylus length
  virtual double getStylusLength() const{
    return(stylusLength);}
  /// returns the stylus radius
  virtual double getStylusRadius() const{
    return(stylusRadius);}
  /// sets the stylus length
  virtual void setStylusLength(const double &l)  {
    stylusLength=l; 
    phantomLength = tipLength + stylusLength;}
  /// sets the stylus radius
  virtual void setStylusRadius(const double &r)  {
    stylusRadius=r;
    phantomRadius = (tipRadius > stylusRadius)? tipRadius : stylusRadius;}

  /// returns the tip length
  virtual double getTipLength() const{
    return(tipLength);}
  /// returns the tip radius
  virtual double getTipRadius() const{
    return(tipRadius);}
  /// sets the tip length
  virtual void setTipLength(const double &l)  {
    tipLength=l; 
    phantomLength = tipLength + stylusLength;}
  /// sets the tip radius
  virtual void setTipRadius(const double &r)  {
    tipRadius=r; 
    phantomRadius = (tipRadius > stylusRadius)? tipRadius : stylusRadius;}

  // TMP BLOCK: auxiliary variables, used for displaying things
  Vec3<double> proxyPos;  
  Vec3<double> prevTipPos;
  // TMP BLOCK: display aux points
  Vec3<double> auxPoint1; 
  Vec3<double> auxPoint2; 
  // TMP BLOCK: display line
  Vec3<double> lineP1; 
  Vec3<double> lineP2; 
  // TMP BLOCK: display triangle
  Vec3<double> trianV1; 
  Vec3<double> trianV2; 
  Vec3<double> trianV3; 
  // TMP BLOCK: display triangles
  int trNum; // TMP: num of triangles in trV?[]
  Vec3<double> trV1[100]; 
  Vec3<double> trV2[100]; 
  Vec3<double> trV3[100]; 

  friend class Scene3D;
};


typedef BasicUserInterface3DIterator<LaparoscopicInterface3D> LaparoscopicInterface3DIterator;
typedef BasicUserInterface3DIterator<PhantomInterface3D> PhantomInterface3DIterator;


} // namespace yav


#endif // FORCEFEEDBACKINTERFACE3D_H
