#ifndef GEOMETRY_H
#define GEOMETRY_H

#pragma interface "graphics3D/Geometry.h"
#include <linearAlgebra/Vec3.h>

namespace yav {

/** \class VectorialBase
    \ingroup Graphics3D
    \brief 3D orthogonal vectorial base */
class VectorialBase {
 protected:
  /// first base vector
  Vec3<double> vx;
  /// second base vector
  Vec3<double> vy;
  /// third base vector
  Vec3<double> vz;
  /// vector component less than epsilon are considered equal to 0
  static double epsilon;

 public:
  /** @name constructors */
  //@{
  /** builds default base */
  VectorialBase();
  /** Builds base from on vector. Resulting base is normalized if v
      is normalized
      @param v first base vector */
  VectorialBase(const Vec3<double>& v);
  //@}
  /// normalize each base vector
  void normalize(void);

  /// returns first base vector
  const Vec3<double>& getVX() { return vx; }
  /// returns second base vector
  const Vec3<double>& getVY() { return vy; }
  /// returns third base vector
  const Vec3<double>& getVZ() { return vz; }
};


/** \class Distance
    \ingroup Graphics3D
    \brief distance between several geometric primitives  */
class Distance {
 public:
  /** the confidence interval that is considered for detecting
      intersections between primitives */
  static double epsilon;
  /** returns the distance between a 3D point and a 3D Line
      @param org a point on the 3D line
      @param dir the direction vector of the 3D line (may not be unit vector )
      @param point the 3D point 
      @return the distance between the 3D point and the 3D line*/
  static double pointToLine(const Vec3<double>& org,
			    const Vec3<double>& dir,
			    const Vec3<double>& point);

  /** returns the distance between a 3D line and a 3D segment 
      @param org a point on the 3D line
      @param dir the direction vector of the 3D line (may not be unit vector )
      @param segA the first segment extremity 
      @param segB the second  segment extremity 
      @return the distance between the 3D line and the 3D segment */
  static double segmentToLine(const Vec3<double>& org,
			      const Vec3<double>& dir,
			      const Vec3<double>& segA,
			      const Vec3<double>& segB);

  /** returns the distance between a 3D line and a triangle 
      @param org a point on the 3D line
      @param dir the direction vector of the 3D line (may not be unit vector )
      @param trA the first triangle vertex
      @param trB the second triangle vertex
      @param trC the third triangle vertex
      @return the distance between the 3D line and the triangle */
  static double triangleToLine(const Vec3<double>& org,
			       const Vec3<double>& dir,
			       const Vec3<double>& trA,
			       const Vec3<double>& trB,
			       const Vec3<double>& trC);
  
};



}

#endif // ifndef GEOMETRY_H
