#ifndef LIGHT_H
#define LIGHT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Light.h"
#ifdef _MSC_VER
#include <basics/Architecture.h>
#include <windows.h>
#endif
#include <GL/gl.h>
#include <linearAlgebra/Rotation3D.h>
#include <modules/TclModule.h>

namespace yav {

class Camera3D;
class LightFantom;

/** \class Light
    \ingroup Graphics3D
    \brief OpenGL light */
class Light {
 protected:
  /// light graphic primitive
  LightFantom *fantom;
  /// OpenGL light id
  GLenum index;
  /// is light active
  bool active;

  /// light ambient coefficients
  GLfloat ambient[4];
  /// light diffuse coefficients
  GLfloat diffuse[4];
  /// light specular coefficients
  GLfloat specular[4];

  /// light source position
  GLfloat position[4];

  /// spot direction
  GLfloat direction[3];
  /// spotlight exponent
  GLfloat exponent;
  /// spotlight cutoff (in degrees)
  GLfloat cutoff;
  /// spotlight constant, linear and quardatic attenuation factors
  GLfloat attenuation[3];

 public:
  /** @name constructors */
  //@{
  /** default (white) light
	 @param ind GL index of light */
  Light(GLenum ind );
  /** creates a new light
      @param ind GL index of light 
	  @param cam camera owning this light
      @param ambient ambient light 4 coefficients
      @param diffuse diffuse light 4 coefficients
      @param specular specular light 4 coefficients
      @param position light position 4 coefficients
      @param direction spotlight direction 3 coefficients
      @param exponent spotlight exponent
      @param cutoff spotlight cutoff (in degrees)
      @param attenuation spotlight 3 attenuation factors
   */
  Light(GLenum ind, const GLfloat *ambient, const GLfloat *diffuse,
	const GLfloat *specular, const GLfloat *position,
	const GLfloat *direction = 0,
	GLfloat exponent = 0.0, GLfloat cutoff = 180.0,
	const GLfloat *attenuation = 0);  
  //@}
  
  /** @name destructor */
  //@{
  /// destroys display list
  virtual ~Light();
  //@}


  /** set new light attributes
      @param ambient ambient light 4 coefficients
      @param diffuse diffuse light 4 coefficients
      @param specular specular light 4 coefficients
      @param position light position 4 coefficients
      @param direction spotlight direction 3 coefficients
      @param exponent spotlight exponent
      @param cutoff spotlight cutoff (in degrees)
      @param attenuation spotlight 3 attenuation factors
   */
  void setLight(const GLfloat *ambient, const GLfloat *diffuse,
		const GLfloat *specular, const GLfloat *position,
		const GLfloat *direction = 0,
		GLfloat exponent = 0.0, GLfloat cutoff = 180.0,
		const GLfloat *attenuation = 0);

  /// returns OpenGL light index
  GLenum getIndex(void) const { return index; }
  /// returns light active flag
  bool isActive(void) const { return active; }
  /// returns light ambient coefficients
  const GLfloat *getAmbient(void) const;
  /// returns light diffuse coefficients
  const GLfloat *getDiffuse(void) const;
  /// returns light specular coefficients
  const GLfloat *getSpecular(void) const;
  /// returns light position
  const GLfloat *getPosition(void) const;
  /// returns spotlight direction
  const GLfloat *getDirection(void) const;
  /// returns spotlight expoenent
  GLfloat getExponent(void) const;
  /// returns spotlight cutoff (in degrees)
  GLfloat getCutoff(void) const;
  /// returns light attenuation factors (constant, linear and quadratic)
  const GLfloat *getAttenuation(void) const;

  /** set light ambient coefficients
      @param ambient 4 coefficients array */
  void setAmbient(const GLfloat *ambient);

  /** set light diffuse coefficients
      @param diffuse 4 coefficients array */
  void setDiffuse(const GLfloat *diffuse);

  /** set light specular coefficients
      @param specular 4 coefficients array */
  void setSpecular(const GLfloat *specular);

  /** set light position coefficients
      @param position 4 coefficients array (light is at infinity if position[3] == 0.0) */
  void setPosition(const GLfloat *position);

  /** set spotlight direction coefficients
      @param direction 3 coefficients array */
  void setDirection(const GLfloat *direction);

  /** set spotlight exponent
      @param exponent exponent value */
  void setExponent(const GLfloat exponent);

  /** set spotlight cutoff
      @param cutoff cutoff value (in degrees) */
  void setCutoff(const GLfloat cutoff);

  /** set light attenuation coefficients
      @param attenuation 3 coefficients (attenuation[0] = constant, attenuation[1] = linear and attenuation[2] = quadratic) */
  void setAttenuation(const GLfloat *attenuation);

  /// recomputes if needed and executes material display list
  void glSet(void);

  /// enable (use) light
  void enable(void);
  /// disable light plane
  void disable(void);

  /** choose whether to display a graphical representation fo
      this spot light
      @param disp display fantom if disp == true */
  void displayFantom(bool disp);
	
  /** return the associated fantom that displays the position and orientation
  of the light */
  LightFantom *getFantom() const {
	return fantom;
  }
/** set the fantom associated with the light */
 void  setFantom(LightFantom *l);


  /// set spot light in scene
  void putInCamera(void);
 
  friend class LightFantom;

};

} // namespace yav


#endif  // ifndef LIGHT_H

