#ifndef MATERIAL_H
#define MATERIAL_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Material.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <modules/TclModule.h>
#include <graphics3D/Data3D.h>
#include <graphics3D/Sphere.h>

namespace yav {

class Color;


/** \class Material
    \ingroup Graphics3D
    \brief OpenGL Material */
class Material {
 protected:
  /// material ambient light coefficients
  GLfloat ambient[4];
  /// material diffuse light coefficients
  GLfloat diffuse[4];
  /// material specular light coefficients
  GLfloat specular[4];
  /// material shininess
  GLfloat shininess;
  /// material light emission coefficients
  GLfloat emission[4];

  /// material ambient light coefficients for back side
  GLfloat ambientBack[4];
  /// material diffuse light coefficients for back side
  GLfloat diffuseBack[4];
  /// material specular light coefficients for back side
  GLfloat specularBack[4];
  /// material shininess for back side
  GLfloat shininessBack;
  /// material light emission coefficients for back side
  GLfloat emissionBack[4];

  /// display list index
  GLuint list;

  /// emissive light flag
  bool lighted;
  /// whether the material is transparent
  bool transparent;
 public:
  /** @name constructors */
  //@{
  /// default material
  Material();
  /** creates a new material
      @param ambient material ambient light reflexion 4 coefficients
      @param diffuse material diffuse light reflexion 4 coefficients
      @param specular material specular light reflexion 4 coefficients
      @param shininess material shininess */
  Material(const GLfloat *ambient, const GLfloat *diffuse,
	   const GLfloat *specular, GLfloat shininess);
  /** set a material from a new color
      @param color material color */
  Material(const Color& color);
  //@}

  /** @name destructor */
  //@{
  /// destroys display list
  ~Material();
  //@}
  /** indicates is the material has an alpha component (opacity) less than 1.0 */
  bool isTransparent() const {
	  return transparent;
  }
  /** set a material
      @param ambient material ambient light reflexion 4 coefficients
      @param diffuse material diffuse light reflexion 4 coefficients
      @param specular material specular light reflexion 4 coefficients
      @param shininess material shininess */
  void setMaterial(const GLfloat *ambient, const GLfloat *diffuse,
		   const GLfloat *specular, GLfloat shininess);

  /** set  material for back side
      @param ambient material ambient light reflexion 4 coefficients
      @param diffuse material diffuse light reflexion 4 coefficients
      @param specular material specular light reflexion 4 coefficients
      @param shininess material shininess */
  void setBackMaterial(const GLfloat *ambient, const GLfloat *diffuse,
		       const GLfloat *specular, GLfloat shininess);

  /** set a material from a new color
      @param color material color */
  void setMaterial(const Color& color);

  /** set ambient material coefficients
      @param ambient 4 coefficients array */
  void setAmbient(const GLfloat *ambient);

  /** set diffuse material coefficients
      @param diffuse 4 coefficients array */
  void setDiffuse(const GLfloat *diffuse);

  /** set specular material coefficients
      @param specular 4 coefficients array */
  void setSpecular(const GLfloat *specular);

  /** set material shininess
      @param shininess material shininess */
  void setShininess(const GLfloat shininess);

  /** set emission material coefficients
      @param specular 4 coefficients array */
  void setEmission(const GLfloat *specular);

 /** set a material from a new color for back side 
      @param color material color */
  void setBackMaterial(const Color& color);

  /** set ambient material coefficients for back side 
      @param ambient 4 coefficients array */
  void setBackAmbient(const GLfloat *ambient);

  /** set diffuse material coefficients for back side 
      @param diffuse 4 coefficients array */
  void setBackDiffuse(const GLfloat *diffuse);

  /** set specular material coefficients for back side 
      @param specular 4 coefficients array */
  void setBackSpecular(const GLfloat *specular);

  /** set material shininess for back side 
      @param shininess material shininess */
  void setBackShininess(const GLfloat shininess);

  /** set emission material coefficients for back side 
      @param specular 4 coefficients array */
  void setBackEmission(const GLfloat *specular);
  /** returns lighted flag
      @param l new light state */
  void setLight(bool l);

  /** set material transparency
      @param alpha transparency factor (0 is opaque and 1 is invisible) */
  void setAlpha(double alpha);
    /** set material transparency for back side 
      @param alpha transparency factor (0 is opaque and 1 is invisible) */
  void setBackAlpha(double alpha);


  /// recomputes if needed executes material display list
  void glSet(void);

  /// returns material ambient light coefficients
  const GLfloat *getAmbient(void) const { return ambient; }
  /// returns material diffuse light coefficients
  const GLfloat *getDiffuse(void) const { return diffuse; }
  /// returns material specular light coefficients
  const GLfloat *getSpecular(void) const { return specular; }
  /// returns material shininess
  GLfloat getShininess(void) const { return shininess; }
  /// returns material specular light coefficients
  const GLfloat *getEmission(void) const { return emission; }
  /// returns lighted flag
  bool getLight(void) const { return lighted; }
  /// returns material transparency coefficient
  double getAlpha(void) const { return diffuse[3]; }

 /// returns material ambient light coefficients
  const GLfloat *getBackAmbient(void) const { return ambient; }
  /// returns material diffuse light coefficients
  const GLfloat *getBackDiffuse(void) const { return diffuse; }
  /// returns material specular light coefficients
  const GLfloat *getBackSpecular(void) const { return specular; }
  /// returns material shininess
  GLfloat getBackShininess(void) const { return shininess; }
  /// returns material specular light coefficients
  const GLfloat *getBackEmission(void) const { return emission; }
  /// returns material transparency coefficient
  double getBackAlpha(void) const { return diffuse[3]; }

  /// -material TCL command
  int materialCmd(TclModule *m, Data3D *data, int argc, char **argv,
		  int i = 0);

  /// -material TCL command
  int materialBackCmd(TclModule *m, Data3D *data, int argc, char **argv,
		  int i = 0);

  /// -material option synopsis
  static const char *synopsis;
};


} // namespace yav


#include <graphics3D/Color.h>


#endif // ifndef MATERIAL_H
