#ifndef OGLBOXCOLLISION_H
#define OGLBOXCOLLISION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif
#pragma interface "graphics3D/OGLBoxCollision.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <list>
#include <GL/gl.h>
#include <GL/glu.h>
#include <linearAlgebra/Vec3.h>
#include <basics/Constants.h>
#ifdef TCL_INTERFACE
#include <modules/TclModule.h>
#endif
#include <graphics3D/BoxCollision.h>

namespace yav {

/** \class OGLBoxCollision
    \ingroup Graphics3D
    a class that is used to detect collision between a box and an
    geometrical object described by an OpenGL display List. This display list
    must include tags through pushName()/popName() commands */
template <class OBJ, class DATA> class OGLBoxCollision : public BoxCollision<DATA> {
  public :
  /** \class CollisionData
      \ingroup Graphics3D
  */
  class CollisionData {
    public :
    GLuint zMin,zMax;
    GLuint *names;
    GLuint namesDepth;
    CollisionData(GLuint nD){
      namesDepth = nD;
    names = new GLuint[nD];
    };
    ~CollisionData(){
      delete [] names;
    }
  };
 protected:
  /** the display list that is called to detect collision */
  GLuint collisionDisplayList;
  // if we want to use OGL's display lists or not
  bool useCollisionDisplayList;


#ifndef WIN32
  /// the maximum buffer size
  static const unsigned int bufferSize=600;
  /// the selection buffer
  GLuint selectionBuffer[bufferSize];
#else
GLuint selectionBuffer[600];
#endif
 public: 
  /// the list of tags corresponding to each collision
  std::list<CollisionData*> collisionDataList;
 public:
  /** @name destructor */
  //@{
  virtual ~OGLBoxCollision();
  //@}
  /** detect collision with a 3D box : reset and update collisionList by
      storing the tags corresponding to the intersection OpenGL primitives 
      @param p1 the first point of box's axis
      @param p2 the second point of box's axis
      @param r1 the square width 
      @param p3 the up direction (optional = 1,0,0)
      @param r2 the square lenght in p3 direction (optional = r1)
      @return a boolean that is true if there were a collision */
  virtual bool detectCollisionWithBox(Vec3<double> p1,Vec3<double> p2, double r1, Vec3<double> p3 = Vec3<double>(1,0,0),double r2=-1);
  /** detect collision with a 3D triangle : reset and update collisionList by
      storing the tags corresponding to the intersection OpenGL primitives 
      @param org the top of triangle
      @param dest the middle of triangle's base
      @param up the width direction
      @param base the triangle's base length
      @param width the triangle's width
      @return a boolean that is true if there were a collision */
  virtual bool detectCollisionWithTriangle(Vec3<double> org,Vec3<double> dest, Vec3<double> up, double base,double width);

  virtual void geometryHasChanged();
  /// call (and build if needed) object's collision display list
  void callCollisionDisplayList(void);
  /// delete collision display list
  void destroyCollisionDisplayList();
  /// get the use of display list for collision detection
  bool getUseCollisionDisplayList(){return useCollisionDisplayList;}
  /// set useCollisionDisplayList's value
  void setUseCollisionDisplayList(bool ucdl){
    useCollisionDisplayList = ucdl;
  }
  #ifdef TCL_INTERFACE
    /// -useCollisionDisplayList TCL command
  int useCollisionDisplayListCmd(TclModule *m, int argc, char **argv);
  #endif

 private:
  void buildCollisionDataList(int hits);

 protected:
  OBJ* objectCollided;
 public:

  typedef void (*UPDATEGEOMETRYFUNCTION)(OBJ* obj);
  typedef void (*BUILDCOLLISIONLISTFUNCTION)(OGLBoxCollision<OBJ, DATA>* box, OBJ* obj);

  /** @name constructor */
  //@{
  /** Build a new OGLBoxCollision object
      @param object the object defining the geometry
      @param updateG the function updating the geometry
      @param buildCL the function filling the collisionList according
      to the data present in collisionDataList
  */
  OGLBoxCollision(OBJ* object,void (*updateG)(OBJ*),void (*buildCL)(OGLBoxCollision<OBJ, DATA>*,OBJ*));
  //@}

  UPDATEGEOMETRYFUNCTION updateGeometry;
  BUILDCOLLISIONLISTFUNCTION buildCollisionList;
};

}

#include "OGLBoxCollision.txx"

#endif // OGLBOXCOLLISION_H
