#ifndef PLANE_H
#define PLANE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Plane.h"
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Rotation3D.h>

namespace yav {

/** \class Plane
    \ingroup Graphics3D
    \brief a geometric plane */
class Plane {
 protected:
  /** plane equation is
      equation[0] x + equation[1] y + equation[2] z + equation[3] = 0 */
  double equation[4];

 public:
  /** @name destructor */
  //@{
  /** construct a new plane passing through (0, 0, 0) and
      orthogonal to (0, 0, 1) */
  Plane();
  /** construct a new plane with equation
      equation[0] x + equation[1] y + equation[2] z + equation[3] = 0
      @param equation four plane equation coefficients */
  Plane(const double *equation);
  /** construct a new plane with passing through P and
      orthogonal to n
      @param P one plane point
      @param n plane normal */
  Plane(const Vec3<double>& P, const Vec3<double>& n);
  //@}
  
  /** @name constructors */
  //@{
  virtual ~Plane() {}
  //@}

  /// returns plane equation
  const double *getEquation(void) const { return equation; }

  /// returns a point belonging to the plane
  Vec3<double> getPoint(void) const;

  /** returns plane point closest to P (orthogonal projection of P on the plane)
      @param P space point */
  Vec3<double> getPoint(const Vec3<double> &P) const;

  /// returns plane normal
  Vec3<double> getNormal(void) const;  

  /** set new plane equation to: eqn[0] x + eqn[1] y + eqn[2] z + eqn[3] = 0
      @param eqn plane equation */
  virtual void setEquation(const double *eqn);

  /** set new plane equation for plane passing through P with
      normal n
      @param P one plane point
      @param n plane normal */
  virtual void setEquation(const Vec3<double>& P, const Vec3<double>& n);

  /** translates plane in space
      @param t translation vector */
  virtual void translate(const Vec3<double>& t);

  /** rotates plane in space
      @param r rotation matrix */
  virtual void rotate(const Rotation3D& r);

  /** rotates plane in space around one point
      @param r rotation matrix
      @param P rotation center */
  virtual void rotate(const Rotation3D& r, const Vec3<double>& P);
};


} // namespace yav


#endif // ifdef PLANE_H
