#ifndef RECTANGLE3D_H
#define RECTANGLE3D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Rectangle3D.h"
#include <graphics3D/Data3D.h>
#include <graphics3D/Texture2D.h>
#include <graphics3D/Surface.h>

namespace yav {

/** \class Rectangle3D
    \ingroup Graphics3D
    \brief class of a 3D rectangle that can be displayed with different rendering methods */
class Rectangle3D : public Data3D, public Surface {
 private:
  /// the scale factor for the s and t texture coordinates
  double sscale,tscale;
  /// the four vertices of  the rectangle
  Vec3<double> point[4];
  /// if the display should be updated
  bool updateDisplayList;
  /// if the normal should be flipped for display
  bool flipNormal;
 public:
  /** @name constructors */
  //@{
  /** create a rectangle two diagonal points and its normal vector
      @param scene the 3D scene ehere the rectangle is stored
      @param p1 first diagonal point of the rectangle
      @param p2 first diagonal point of the rectangle
      @param v1 the direction of the first axis of the rectangle
      @param v2 the direction of the second axis of the rectangle */
  Rectangle3D(Scene3D *scene=0, 
	      const Vec3<double> &p1=Vec3<double>(0,0,0), 
	      const Vec3<double> &p2=Vec3<double>(1,1,0), 
	      const Vec3<double> &v1=Vec3<double>(1,0,0),
	      const Vec3<double> &v2=Vec3<double>(0,1,0));
  //@}

  /** @name destructor */
  //@{
  /// destroys display list
  ~Rectangle3D();
  //@}
 private:
  // initialize variables
  void init();
  void surfaceHasChanged();
 public:
  inline bool normalFlipped(void) const {
    return(flipNormal);}
  /// redraw with texture mapping
  void textureRedraw();
  /// create display list of flat display
  void buildFlat();
  /// create display list of wireframe display
  void buildWireframe();
  /// create display list of solid display
  void buildSolid();
  //// get the normal vector of the rectangle
  Vec3<double> getNormal() const {
    Vec3<double> normal;
    normal=cross<double>(point[1]-point[0],point[3]-point[0]);
    return(normal/normal.norm());
  }
  /// get the centroid of the rectangle
  Vec3<double> getCenter() const {
    return((point[0]+point[1]+point[2]+point[3])/4.0);
  }
  /// get the two length sizes
  void getSize(double &length, double &width) const;
  /// get vertex nb i (i<4)
  Vec3<double> getVertex(const unsigned int i) const {
    if (i<4)
      return point[i];
    else
      return Vec3<double>();
  }
  /// the texture scale coordinates
  void getTextureScale (double &s, double &t) const {
    s=sscale; t=tscale;
  }
  /// the texture scale coordinates
  void setTextureScale (const double s, const double t)  {
    sscale=s; tscale=t;
  }

  /** translates the object
      @param tr translation vector */
  virtual void translate(const Vec3<double>& tr) {
    for(unsigned int i=0;i<4;++i)
      point[i]+=tr;
    surfaceHasChanged();
  }
  /** rotates the object
      @param rotation rotation matrix */
  virtual void rotate(const Rotation3D& rotation) {
    Vec3<double> center=(point[0]+point[1]+point[2]+point[3])/4.0;
    for(unsigned int i=0;i<4;++i)
      point[i]=center+ rotation*(point[i]-center);
    surfaceHasChanged();
  }
  /** scales the object
      @param scale scale facto */
  virtual void scale(const double scale){
    Vec3<double> center=(point[0]+point[1]+point[2]+point[3])/4.0;
    for(unsigned int i=0;i<4;++i)
      point[i]=center+ scale*(point[i]-center);
    surfaceHasChanged();
  }

  /// all graphics objects can be drawn
  virtual void redraw(Camera3D *);
  /** Returns the graphic object center and weight.
      The weight is a measure of the number of primitives to allow
      barycentric computation of the whole scene center.
      @param center returned center 
      @param n returned barycentric weight */
  virtual void updateCenter(Vec3<double>& center, unsigned int& n){
    n=4;
    center=(point[0]+point[1]+point[2]+point[3])/4.0;
  }
  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec3<double>& center){
    Vec3<double> dp;
    double radius=0,val;
    for(unsigned int i=0;i<4;++i)
      {
	dp=point[i]-center;
	val=dp.norm2();
	if (val>radius)
	  radius=val;
      }
    return(sqrt(radius));
  }
  /// change the normal direction for display
  void flipNormals() {
    if (flipNormal==true)
      flipNormal=false;
    else
      flipNormal=true;
  }
  /** The routine that allows the camera to select an object
      by computing the distance of an object from its radial line
      @param mc 3D line descriptor
      @return distance to line */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const;
  /** set object redraw order for transparency sorting
      @param order redraw order (should be 1 or 2) */
  virtual void setRedrawOrder(unsigned int order) {
    redrawOrder=order;}
};

}

#endif
