#ifndef SURFACE_H
#define SURFACE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Surface.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <modules/TclModule.h>
#include <graphics3D/Data3D.h>
#include <graphics3D/Color.h>
#include <graphics3D/Material.h>
#include <graphics3D/Texture2D.h>
#include <graphics3D/Texture1D.h>
#include <basics/Owner.h>


namespace yav {


class Camera3D;

/** \class Surface
    \ingroup Graphics3D
    \brief any surface object */
class Surface : public Owner<Texture2D>,  public Owner<Texture1D>  {
 public:

  /** defines the type of member function that can be called
      when drawing a rendering mode */
  typedef void (*RedrawRenderingMode)(Camera3D *c,Surface *s, void *d);
  
  /** a data structure describing a given rendering mode. It includes
      all necessary information for the Tcl interface */
  struct Rendering {
    /// name of the rendering
    TclString renderingName;
    /// name of the option associated with the Tcl command -render
    TclString optionName;
    /// the function called for redrawing the surface
    RedrawRenderingMode redraw;
    /// additional information sent to the redraw function : could be 0
    void *clientData;
  };
  
  static const TclString NONE, WIREFRAME, FLAT, SOLID, GOURAUD, WIREFRAME_SOLID,
    HIDDEN_LINES;
  /// display modes
  enum RenderMode {
    /// do not display
    RM_NONE,
    /// display wireframe
    RM_WIREFRAME,
    /// display solid flat surface
    RM_FLAT,
    /// display solid smoothed surface
    RM_SOLID,
    /// display solid surface plus wireframe
    RM_WIREFRAME_SOLID,
    /// display wireframe with hidden lines
    RM_HIDDEN_LINES,
  };

 protected:

  /// wether we really need a specific flat display List
  bool useFlat;
  /// surface material
  Material mat;
  /// wireframe color
  Color color;
  /// wireframe line width
  GLfloat width;

  /// wireframe display list
  GLuint wireframeList;
  /// solid display list
  GLuint solidList;
  /// flat display list
  GLuint flatList;

  /// a texture applied on the surface
  ObjectPointer<Texture1D> texture1D;
  /// a texture applied on the surface
  ObjectPointer<Texture2D> texture2D;
  /// a texture applied on the surface corresponding to the specular component
  ObjectPointer<Texture2D> highlightTexture;
  /// if we should use texture mapping
  bool useTexture1D;
 /// if we should use texture mapping
  bool useTexture2D;
  /// if we should use texture highlighting instead of Gouraud specularity
  bool useTextureHighlight;
  /// if the texture map should replace lighting or being modulated with it
  bool textureReplaceLighting;
  // if we want to use OGL's display lists or not
  bool useDisplayLists;
  
  /// describe how to interpolate texture
  Texture2D::TextureInterpolation texture2DInterpolation;
  /// describe how to extrapolate texture
  Texture2D::TextureExtrapolation texture2DExtrapolation;
  /// describe how to apply the texture
  Texture2D::TextureMode texture2DMode;

 /// describe how to interpolate texture
  Texture1D::TextureInterpolation texture1DInterpolation;
  /// describe how to extrapolate texture
  Texture1D::TextureExtrapolation texture1DExtrapolation;
  /// describe how to apply the texture
  Texture1D::TextureMode texture1DMode;

  /// s texture parameter vector when using automatic texture coordinate generate with a plane
  double sTextureParameters[4];
  /// t texture parameter vector when using automatic texture coordinate generate with a plane
  double tTextureParameters[4];

  /// builds object's wireframe display list
  void buildWireframeDisplayList(void);
  /// builds object's flat display list
  void buildFlatDisplayList(void);
  /// builds object's solid list
  void buildSolidDisplayList(void);

  /// call (and build if needed) object's wireframe display list
  void callWireframeDisplayList(void);
  /// call (and build if needed) object's flat display list
  void callFlatDisplayList(void);
  /// call (and build if needed) object's solid display list
  void callSolidDisplayList(void);
  
  /// destroy if needed object's wireframe display list
  void destroyWireframeDisplayList(void);
  /// destroy if needed object's flat display list
  void destroyFlatDisplayList(void);
  /// destroy if needed object's solid list
  void destroySolidDisplayList(void);
  
  /// the list of possible rendering for that surface */
  std::list<Rendering *> renderingList;
  /// the rendering mode that is currently used
  Rendering *selectedRendering;
  
private:
  /// initialize the list of rendering and select one rendering
  void  initRenderingList(void);
  /// destroy the list of rendering 
  void  destroyRenderingList(void);
  
  

  /// redraw function for the "wireframe" rendering
  void wireframeLinesRedraw(Camera3D *cam);
  /// redraw function for the "hidden lines" rendering
  void wireframeHiddenRedraw(Camera3D *cam);

  /// redraw function for the "flat" rendering
  void solidFlatRedraw(Camera3D *cam);
  /// redraw function for the "solid" rendering
  void solidGouraudRedraw(Camera3D *cam);
  
 public:
  /** @name constructors */
  //@{
  /// default (white) surface
  Surface();
  /** creates a new surface with given material
      @param ambient surface ambient light reflexion 4 coefficients
      @param diffuse surface diffuse light reflexion 4 coefficients
      @param specular surface specular light reflexion 4 coefficients
      @param shininess surface shininess */  
  Surface(const GLfloat *ambient, const GLfloat *diffuse,
	  const GLfloat *specular, GLfloat shininess);
  /// copy constructor
  Surface(const Surface& org);
  //@}

  /** @name destructor */
  //@{
  virtual ~Surface();
  //@}

  /** add a rendering mode to the current list
      @param renderingName defines the name of the rendering mode
      @param optionName defines the name used in the Tcl interface to designate the mode
      @param redrawFn the callback function called for the redraw
      @param ClientData additional pointer used for the redraw function 
  */
  void addRenderingMode(TclString renderingName,TclString optionName,
			RedrawRenderingMode redrawFn, void *ClientData);

  /// remove a Rendering mode from the List
  void removeRenderingMode(Rendering *r);

  /** returns a pointer to the rendering structure having Tcl option name n
      if no rendering has name n then returns 0 */
  Rendering *getRenderingFromOptionName(TclString n);
  /** returns a pointer to the rendering structure having name n
      if no rendering has name n then returns 0 */
  Rendering *getRenderingFromName(TclString n);
  /// returns current render mode
  const Rendering *getRenderMode(void) const { return selectedRendering; }
  /// set a  rendering mode gven its name
  virtual void setRendering(const TclString& name);
  /// set a  rendering mode 
  virtual void setRendering(Rendering *r);
  /** call surface display list for all pre-defined rendering modes
      @param cam camera being redrawn
      @param rm
  */
  void redrawBasic(Camera3D *cam,RenderMode rm);
  /** draw a colorbar associated with 1d texture in the opengl window
	@param windowHeight the height of the camera
	@param windowWidth the width of the camera
	@param colorbarHeight the height of the colorbar
	@param colorbarWidth the width of the colorbar
	@param posx the position of the colorbar in the x direction
	@param posy the position of the colorbar in the y direction
	@param minLabel the label corresponding to the min value
	@param maxLabel the label corresponding to the max value */
  void drawColorbar(Camera3D *cam,
	  double colorbarHeight,double colorbarWidth,
	  double posx,double posy,char *minLabel,char *maxLabel);
  
  /// returns wireframe color
  Color& getColor(void)  { return color; }
  /// returns surface material
  Material& getMaterial(void)   { return mat; }
  /// returns wireframe line width
  GLfloat getLineWidth(void) const { return width; }
  /** set wireframe line width
      @param width set new line width */
  void setLineWidth(GLfloat width);
  /// get textureMode of the 2D texture
  Texture2D::TextureMode getTexture2DMode(){return texture2DMode;}
  /// get textureMode of the 1D texture
  Texture1D::TextureMode getTexture1DMode(){return texture1DMode;}
  /// set textureMode of the 2D texture
  void setTexture2DMode(Texture2D::TextureMode tm);
  /// set textureMode of the 1D texture
  void setTexture1DMode(Texture1D::TextureMode tm);
  /// set 2D texture
  void setTexture2D(Texture2D *t) {
    texture2D=t;
    if (useTexture2D)
      surfaceChanged();
  }
  /// set 1D texture
  void setTexture1D(Texture1D *t) {
    texture1D=t;
    if (useTexture1D)
      surfaceChanged();
  }
  /// set highlight texture
  void setHighlightTexture(Texture2D *t) {
    highlightTexture=t;
    if (useTextureHighlight)
      surfaceChanged();
  }
  /// get 2D texture
  Texture2D *getTexture2D() const {
    return texture2D();
  }
  /// get 2D texture
  Texture1D *getTexture1D() const {
    return texture1D();
  }
  /// get highlight texture
  Texture2D *getHighlightTexture() const {
    return highlightTexture();
  }
  /// if 2D texture should be used
  bool usingTexture2D() const {
    return useTexture2D;
  }
  /// if 2D texture should be used
  bool usingTexture1D() const {
    return useTexture1D;
  }
  /// if texture highlight should be used
  bool usingTextureHighlight() const {
    return useTextureHighlight;
  }
  /// set the use of 2D texture mapping
  void setTexture2DUsage(const bool b) {
    if (b!=useTexture2D)
      {
	useTexture2D=b;
	surfaceChanged();
      }
  }
  /// set the use of 2D texture mapping
  void setTexture1DUsage(const bool b) {
    if (b!=useTexture1D)
      {
	useTexture1D=b;
	surfaceChanged();
      }
  }
  /// set the use of texture highlight
  void setTextureHighlightUsage(const bool b) {
    if (b!=useTextureHighlight)
      {
	useTextureHighlight=b;
	surfaceChanged();
      }
  }
  /// if lighting should replace lighting or should be modulated with lighting
  bool replaceLighting() const {
    return textureReplaceLighting;
  }
  /// set the use of lighting vs texture mapping
  void setLightingTextureUsage(const bool b) {
    if (b!=textureReplaceLighting)
      {
	textureReplaceLighting=b;
	surfaceChanged();
      }
  }
  /** an 2D Texture is being destroyed
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<Texture2D>& obj);
 /** a 1D Texture is being destroyed
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<Texture1D>& obj);
  /// delete display lists when surface changes
  void surfaceChanged(void);
  /// redraw the surface
  void redraw(Camera3D *cam);

  /// wireframe display list creation callback
  virtual void buildWireframe(void) = 0;

  /// solid display list creation callback
  virtual void buildSolid(void) = 0;

  /// flat display list creation callback
  virtual void buildFlat(void) {
    buildSolid(); }

  /** set object redraw order for transparency sorting
      @param order redraw order (should be 1 or 2) */
  virtual void setRedrawOrder(unsigned int order) = 0;

  /// -render TCL command
  int renderCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);

  /// -renderOption TCL command
  int renderOptionCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);

  /// -texture2D TCL command
  int texture2DCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);

  /// -texture1D TCL command
  int texture1DCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);
  /// -texture TCL command
  int highlightTextureCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);
  /// -useDisplayList TCL command
  int useDisplayListsCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);

  /// -texture2DOption TCL command
  int texture2DOptionCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);
  /// -texture1DOption TCL command
  int texture1DOptionCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0);

  /// -color TCL command
  int colorCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0) {
    return color.colorCmd(m,data,argc,argv,i);
  }
  /// -material TCL command
  int materialCmd(TclModule *m, Data3D *data, int argc, char **argv, int i = 0) {
    return mat.materialCmd(m,data,argc,argv,i);
  }

  /// -render option synopsis
  static const char *synopsis;

  /// -texture2DOption option synopsis
  static const char *texture2DSynopsis;

  /// -texture1DOption option synopsis
  static const char *texture1DSynopsis;
};


  /// redraw function for the "hide" rendering
  void dummySurfaceRedraw(Camera3D *cam,Surface *s,void *);
  /// redraw function for the "wireframe" rendering
  void wireframeSurfaceRedraw(Camera3D *cam,Surface *s,void *);
  /// redraw function for the "flat" rendering
  void flatSurfaceRedraw(Camera3D *cam,Surface *s,void *);
  /// redraw function for the "solid" rendering
  void gouraudSurfaceRedraw(Camera3D *cam,Surface *s,void *);
  /// redraw function for the "wireframesolid" rendering
  void wireGouraudSurfaceRedraw(Camera3D *cam,Surface *s,void *);
  /// redraw function for the "hidden" rendering
  void hiddenSurfaceRedraw(Camera3D *cam,Surface *s,void *);
 



} // namespace yav

#endif // ifndef SURFACE_H
