#ifndef TclCamera3D_H
#define TclCamera3D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif



#pragma interface "graphics3D/TclCamera3D.h"


#include <tcl.h>
#include <tk.h>
#include <modules/togl.h>
#include <graphics/TclSceneObject.h>
#include <graphics3D/Camera3D.h>


namespace yav {

class TclScene3D;
class ClippingPlane;
class Light;


 int createCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);
 int displayCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);
 int resizeCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);
 int destroyCamera3D(ClientData clientData, Tcl_Interp *interp, int objc,
        Tcl_Obj *const *objv);


/** \class TclCamera3D
\ingroup Graphics3D
\brief 3D Tcl Camera interface using a Togl widget */
class TclCamera3D :  public TclSceneObject 
{	

protected:
	/// parent window name
	TclString parentFrame;
	/// OpenGL buffer window name
	TclString cameraFrame;
	/// menu window name
	TclString menuFrame;
	/// submenu window name
	TclString submenuFrame;
	/// buttons window name
	TclString buttonFrame;

	/// whether each new frame should be dumped to a file to create a movie
	bool record;
	/// movie frame number
	unsigned int frameNumber;
	/// movie base file name
	TclString movieName;

	/// the Togl widget associated with this object
	struct Togl *widget;

public:
	/** @name Constructor */
	//@{
	/** builds a new 3D OpenGL/TCL camera
      @param m module to build camera in
      @param parent parent window to create camera in (if EMPTY_STRING, camera is created in a new toplevel window)
      @param scene visualized scene interface
      @param stereo if true, allocates a new stereo buffer
      @param type dynamic object type
      @param cam client (if already created)
      @exception InvalidCommandLine invalid command line argument transmitted to Togl command
	  @exception CannotMakeWindowExist unable to create TCL window */
	TclCamera3D(TclModule *m, const TclString& parent, TclScene3D *scene,
		bool stereo = false,
		const std::type_info& type = typeid(TclCamera3D), Camera3D *cam = 0);
	//@}
	

	/** @name Desstructor */
	//@{
	/// Destroy TclCam2D object and associated TCL canvas
	virtual ~TclCamera3D();
	//@}

	/** returns camera2D client */
	Camera3D *getClient(void) const {
		return (Camera3D *) client;
	}
	/** sets the togl widget associated with the TclCamera3D */ 
	void setWidget( Togl *w) {
		widget=w;
	}
	
	/// parent window name
	const char *getParentFrame() const {
		return (const char *) parentFrame;
	}
	/// OpenGL buffer window name
	const char *getCameraFrame() const {
		return (const char *) cameraFrame;
	}
	/// menu window name
	const char *getMenuFrame() const {
		return (const char *) menuFrame;
	}
	/// submenu window name
	const char *getSubMenuFrame() const {
		return (const char *) submenuFrame;
	}
	/// buttons window name
	const char *getButtonFrame() const {
		return (const char *) buttonFrame;
	}
	/** dump camera window content in an xwd file
	@param filename dump file name */
	virtual void dumpToFile(const TclString& filename);
    /** dump camera window content in an PPM file
     @param filename dump file name */
    virtual void dumpToPPM(const TclString& filename);

	/// movie file name
	const TclString &getMovieName() const {
		return movieName;
	}
	/// return the current frame number and increment it
	unsigned int getFrameNumberAndIncrement() {
		return (frameNumber++);
	}
	/// Bind the OpenGL rendering context to the specified Togl widget
	virtual void makeCurrent(void) const;
	/// Call the glXSwapBuffers() to swap the buffer when using double buffering
	void swapBuffers(void) const;
	/** sets the viewport to (w,h) and updates camera parameters */
	void resize(unsigned int w, unsigned int h);

	/** selects a scene object
	@param data object to select */
	void setSelectedObject(Data3D *data);

	// install /uninstall the menu of a selected object
	/** Eventually install the Tcl menus associated with a 3D object
	@param obj the selected object */
	virtual void applyObjectSelection (Data3D *obj);
	/** Eventually Uninstall the Tcl menus associated with a 3D object
	@param obj the  object to be unselected */
	virtual void applyObjectUnselection (Data3D *obj);
	/** start camera recording
	@param name base file name */
	void startMovie(const char *name);
	/** stop camera recording */
	void stopMovie(void);
	/// is the camera recording  each frame ?
	bool isRecording() const {
		return record;
	} 
	/** pause camera recording
	@param state pause if state = true, record if state = false */
	void pauseMovie(bool state);
	/// redraw camera and eventually record the display in an image to create a movie
	virtual void redraw(void);

	/// returns camera scene
	TclScene3D *getScene(void) const;
	// Callback setup functions
	/// redraw scene when idle
	void redrawCamera(void);
	/** sets the viewport to (width,height) and updates camera parameters */
	void resize(void);
	/// The action performed when the deformation loop of the scene is activated
	virtual void startComputation();
	/// The action performed when the deformation loop of the scene is de-activated
	virtual void stopComputation();

	// Miscellaneous
	/// dummy routine that is called by Tk_CreateTimerHandler
	virtual void timer_routine(void) {}
	/// returns the current width
	virtual unsigned int getWidth() const {
		return getClient()->getWidth();
	}
	/// return the current widget height
	virtual unsigned int getHeight() const {
		return getClient()->getHeight();
	}
	/// whether the camera is being destroyed to avoid multiple destruction
	virtual bool isCameraBeingDestroyed() const{
		return getClient()->isCameraBeingDestroyed();
	}

	// Query functions

	/// returns the name of the parent widget
	const TclString& getParent(void) const { return parentFrame; }


	/** set the action to perform each time the mouse is moved
	@param on if true then a specific action is performed each time the mouse is moved. If false then unbind the action */
	virtual void bindMouseMotion(bool on);


	/** creates a new 3D camera
	@param m module to create camera in
	@param argc command line number of arguments 
	@param argv command line arguments */
	static int create(TclModule *m, int argc, char **argv);
	/** build a list of all available cameras */
	static int list(TclModule *m, int argc, char **argv);
	/// creates the Tcl commands when the module is loaded
	static void createCommands(TclModule *m);

	/// -destroy TCL option
	virtual int destroyCmd(int argc, char **argv);
	int viewPointCmd(int argc, char **argv);
	
        /// -dumpToPPM TCL option
        int dumpToPPMCmd(int argc, char **argv);

	/// -clippingPlanes TCL option
	int clippingPlanesCmd(int argc, char **argv);
	/// -clippingPlaneCreate TCL option
	int clippingPlaneCreateCmd(int argc, char **argv);
	/// -clippingPlaneRemove TCL option
	int clippingPlaneRemoveCmd(int argc, char **argv);
	/// -clippingPlaneEdit TCL option
	int clippingPlaneEditCmd(int argc, char **argv);
	/// -useTwoSides TCL option
	int useTwoSidesCmd(int argc, char **argv);
	/// -lightCreate TCL option
	int lightCreateCmd(int argc, char **argv);
	/// -lightRemove TCL option
	int lightRemoveCmd(int argc, char **argv);
	/// -lightEdit TCL option
	int lightEditCmd(int argc, char **argv);
	/// -lights TCL option
	int lightsCmd(int argc, char **argv);
	/// -boundingBoxColor TCL option
	int boundingBoxColorCmd(int argc, char **argv);
	/// -boundingBoxWidth TCL option
	int boundingBoxWidthCmd(int argc, char **argv);
	/// -boundingBoxDisplay TCL option
	int boundingBoxDisplayCmd(int argc, char **argv);
	/// -bgColor TCL option
	int bgColorCmd(int argc, char **argv);
	/// -scene TCL option
	int sceneCmd(int argc, char **argv);
	/// -width TCL option
	int widthCmd(int argc, char **argv);
	/// -height TCL option
	int heightCmd(int argc, char **argv);
	/// -eyeOffset TCL option
	int eyeOffsetCmd(int argc, char **argv);
	/// -fieldOfView TCL option
	int fieldOfViewCmd(int argc, char **argv);
	/// -motionMode TCL option
	int motionModeCmd(int argc, char **argv);
	/// -mouseMotion TCL option
	int mouseMotionCmd(int argc, char **argv);
	/// -transformationMatrix TCL option
	int transformationMatrixCmd(int argc, char **argv);
	/// -center TCL option
	int centerCmd(int argc, char **argv);
	/// -refCenter TCL option
	int refCenterCmd(int argc, char **argv);
	/// -scaleFactor TCL option
	int scaleFactorCmd(int argc, char **argv);
	/// -focus TCL option
	int focusCmd(int argc, char **argv);
	/// -focusOnRegion TCL option
	int focusOnRegionCmd(int argc, char **argv);
	/// -autofocus TCL option
	int autofocusCmd(int argc, char **argv);
	/// -widget TCL option
	int widgetCmd(int argc, char **argv);
	/// -menu TCL option
	int menuCmd(int argc, char **argv);
	/// -objects TCL option
	int objectsCmd(int argc, char **argv);
	/// -fog TCL option
	int fogCmd(int argc, char **argv);
	/// -fogRange TCL option
	int fogRangeCmd(int argc, char **argv);
	/// -stencilBuffer TCL option
	int stencilBufferCmd(int argc, char **argv);
	/// -specularity TCL option
	int specularityCmd(int argc, char **argv);
	/// -frontAndBack TCL option
	int renderFrontAndBackCmd(int argc, char **argv);

	/// -redraw TCL option
	int redrawCmd(int argc, char **argv);
	/// -select TCL option
	int selectCmd(int argc, char **argv);

	/// -bindMotion TCL option
	int bindMotionCmd(int argc, char **argv);

	/// -keyPlus TCL option
	int keyPlusCmd(int argc, char **argv);
	/// -keyMinus TCL option
	int keyMinusCmd(int argc, char **argv);
	/// -keyDown TCL option
	int keyDownCmd(int argc, char **argv);
	/// -keyUp TCL option
	int keyUpCmd(int argc, char **argv);
	/// -keyRight TCL option
	int keyRightCmd(int argc, char **argv);
	/// -keyLeft TCL option
	int keyLeftCmd(int argc, char **argv);
	/// -key TCL option
	int keyCmd(int argc, char **argv);

	/// -button1Released TCL option
	int button1ReleasedCmd(int argc, char **argv);
	/// -button2Released TCL option
	int button2ReleasedCmd(int argc, char **argv);
	/// -button3Released TCL option
	int button3ReleasedCmd(int argc, char **argv);
	/// -button1Pressed TCL option
	int button1PressedCmd(int argc, char **argv);
	/// -button2Pressed TCL option
	int button2PressedCmd(int argc, char **argv);
	/// -button3Pressed TCL option
	int button3PressedCmd(int argc, char **argv);
	/// -shiftButton1Pressed TCL option
	int shiftButton1PressedCmd(int argc, char **argv);
	/// -shiftButton2Pressed TCL option
	int shiftButton2PressedCmd(int argc, char **argv);
	/// -shiftButton3Pressed TCL option
	int shiftButton3PressedCmd(int argc, char **argv);
	/// -ctrlButton1Pressed TCL option
	int ctrlButton1PressedCmd(int argc, char **argv);
	/// -ctrlButton2Pressed TCL option
	int ctrlButton2PressedCmd(int argc, char **argv);
	/// -ctrlButton3Pressed TCL option
	int ctrlButton3PressedCmd(int argc, char **argv);
	/// -record TCL command option
	int recordCmd(int argc, char **argv);

	/// -mouseline TCL command option
	int mouselineCmd(int argc, char **argv);


private:

	// friends
	/** destroy the widget and the rendering context when possible 
	@param clientData pointer to a Togl casted as ClientData */
	static void destroyContext(ClientData clientData);
	/** render the scene by calling Camera3D::redraw() 
	@param clientData pointer to a Togl casted as ClientData */
	static void redrawWhenIdle(ClientData clientData);
	/** render in the overlay planes by calling redraw_overlay()
	@param clientData pointer to a Togl casted as ClientData */  
	static void renderOverlayWhenIdle(ClientData clientData);
	/** resize camera
	@param clientData pointer to a Togl casted as ClientData */
	static void resizeWhenIdle(ClientData clientData);
	/** destroy the widget 
	@param clientData pointer to a Togl casted as ClientData */    
	static void Destroy(char *clientData);

	/**  Gets called from Tk_CreateTimerHandler.
	@param clientData pointer to a Togl casted as ClientData   */
	static void Timer(ClientData);


};


}


#endif // TclCamera3D_H
