#ifndef TEXTURE1D_H
#define TEXTURE1D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Texture1D.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <basics/Pointed.h>

namespace yav {


/** \class Texture1D
    \ingroup Graphics3D
    \brief a class defining a 2D texture according to OpenGL */
class Texture1D : public Pointed<Texture1D> {
 public:
  /// the generic texture that can be created directly
  enum GenericTexture { TEMPERATURE, TEMPERATURE_REVERSE, BLUE2YELLOW, JET, JET_REVERSE, ACTIVATION, HEART, HEART_REVERSE, RANDOM, BRB, OTHER};
  /// the two different ways of coding each pixel of a texture
  enum Format {TEXTURE_RGBA, TEXTURE_LUMINANCE};
  /** the two different types of texture interpolation */
  enum TextureInterpolation { LINEAR, NEAREST};
  /** the two different types of texture interpolation */
  enum TextureExtrapolation { CLAMP, REPEAT};
  /** the four different types of texture mapping */
  enum TextureMode { COORDINATES, SPHERICAL_MAPPING,PLANAR_MAPPING, CAMERA_MAPPING};
 protected:
  /// if the texture has been binded
  bool binded;
  /// the texture ID
  GLuint id;
  /// texture buffer 
  void *buffer;
  /// number of columns: must be a power of two
  unsigned int ncols;
  /// the number of columns that are really meaningful
  unsigned int nbRealColumns;
  /// pixel format
  Format format;
  /// the type of texture 
   GenericTexture textureType;
  /// if the display list should be update
  bool updateDisplayList;
 public:
  /** @name constructors */
  //@{
  /** build a texture from a buffer 
   @param b a buffer where each pixel is each coded with RGBA or ABGR
   @param nc the number of columns (may not be a power of 2) 
   @param f the format of each pixel */ 
  Texture1D(unsigned int *b,  const unsigned int nc,
	    Format f=TEXTURE_RGBA);
  /** build a texture of a given type
      @param type the type of generic texture
      @param size the number of rows and columns of the texture 
  */
  Texture1D(GenericTexture type,const unsigned int size);
  /** build a texture from a GIF image file
      @param filename file name of a GIF file */
  Texture1D(char *filename);
  //@}

  /** @name destructor */
  //@{
  /// delete a 2D Texture 
  virtual ~Texture1D();
  //@}
 protected:
  /** enlarge a given texture such that its size is a power of 2 and store
      it in the texture buffer */
  void enlargeTexture(const void *b, 
		      const unsigned int nc);
  /// create the display list associated with the texture
  void createTexture();
  /// initialize variables
  void init();
 public:
  /** draw a texture of a given type
      @param type the type of generic texture
  */
  void drawGenericTexture(GenericTexture type);
  /// get the texture type
  GenericTexture getTextureType() const {
   return textureType;
 }
  /// texture buffer 
  void *getTextureBuffer() {
	return buffer;
   }
  /// update the OpenGL texture based on the new texture buffer
  void updateTexture() const;
  /// return the nb of texture columns
  unsigned int getNbColumns() const { return ncols;}
  /// return the nb of meaningful texture columns
  unsigned int getNbRealColumns() const { return nbRealColumns;}
  /** get the RGBA color of a texture pixel 
      @param col the column index of a pixel
      @param color the array of Red Green Blue and Alpha value*/
  void getColor(const unsigned int col,
		unsigned char *color) const;
  /// return the texture ID
  GLuint getId() const {
    return id; }
  /// enable texture coordinates generation
  void enableTextureCoordinatesGeneration() const {
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
  }
  /// disable texture coordinates generation
  void disableTextureCoordinatesGeneration() const {
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T); 
  }
  /// use this texture
  void enableTexture(TextureMode m) {
    if (binded==false)
      createTexture();
    glBindTexture(GL_TEXTURE_1D, id);
    
    glEnable(GL_TEXTURE_1D);
    if (m!=COORDINATES)
      {
	glEnable(GL_TEXTURE_GEN_S);
      }
  }
  /// stop using this texture
  void disableTexture(TextureMode m) const {
      
    glDisable(GL_TEXTURE_1D);

    if (m!=COORDINATES)
      {
	glDisable(GL_TEXTURE_GEN_S);
      }
  }
  /// set the type of texture interpolation
  void interpolateTexture(TextureInterpolation i) const {
    if (i==LINEAR)
      {
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      }
    else if (i==NEAREST)
      {
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      }
  }
  /// set the type of texture extrapolation
  void extrapolateTexture(TextureExtrapolation i) const {
    if (i==CLAMP)
      {
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      }
    else 
      {
	glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_REPEAT);
      }
  }
  /// set the type of texture interpolation
  void applyTextureMode(TextureMode i, const double s[4], const double t[4]) const {
    if (i==SPHERICAL_MAPPING)
      {
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
      }	
    else if (i==PLANAR_MAPPING)
      {
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
       	glTexGendv(GL_S, GL_OBJECT_PLANE, s);
      }
    else if (i==CAMERA_MAPPING)
      {
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGendv(GL_S, GL_EYE_PLANE, s);
      }
  }

};


}


#endif
