#ifndef TEXTURE2D_H
#define TEXTURE2D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "graphics3D/Texture2D.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <basics/Pointed.h>

namespace yav {


/** \class Texture2D
    \ingroup Graphics3D
    \brief a class defining a 2D texture according to OpenGL */
class Texture2D : public Pointed<Texture2D> {
 public:
  /// the generic texture that can be created directly
  enum GenericTexture { CHECKER, HALO, GAUSSIAN};
  /** the two different types of texture interpolation */  /// the two different ways of coding each pixel of a texture
  enum Format {TEXTURE_RGBA, TEXTURE_LUMINANCE};

  enum TextureInterpolation { LINEAR, NEAREST};
  /** the two different types of texture interpolation */
  enum TextureExtrapolation { CLAMP, REPEAT};
  /** the four different types of texture mapping */
  enum TextureMode { COORDINATES, SPHERICAL_MAPPING,PLANAR_MAPPING, CAMERA_MAPPING};
 protected:
  /// if the texture has been binded
  bool binded;
  /// the texture ID
  GLuint id;
  /// texture buffer 
  void *buffer;
  /// number of rows: must be a power of two
  unsigned int nrows;
  /// number of columns: must be a power of two
  unsigned int ncols;
  /// the number of rows that are really meaningful
  unsigned int nbRealRows;
  /// the number of columns that are really meaningful
  unsigned int nbRealColumns;
  /// pixel format
  Format format;
  /// if mipmapping should be used
  bool useMipmapping;
  /// the number of levels of the mipmapping
  unsigned int mipmapLevel;
  /// if the display list should be update
  bool updateDisplayList;
 public:
  /** @name constructors */
  //@{
  /** build a texture from a buffer 
   @param b a buffer where each pixel is each coded with RGBA or ABGR
   @param nr the number of rows (may not be a power of 2) 
   @param nc the number of columns (may not be a power of 2) 
   @param f the format of each pixel */ 
  Texture2D(unsigned int *b, const unsigned int nr, const unsigned int nc,
	    Format f=TEXTURE_RGBA);
  /** build a texture of a given type
      @param type the type of generic texture
      @param size the number of rows and columns of the texture 
      @param exponent for the gaussian and halo texture controls the spread of the gaussian
      @param radius for the halo texture controls the radius of the halo (as default it is a quarter of the size )
  */
  Texture2D(GenericTexture type,const unsigned int size, const double exponent=1.0, const double radius=1.0);
  /** build a texture from a GIF image file
      @param filename file name of a GIF file */
  Texture2D(char *filename);
  //@}

  /** @name destructor */
  //@{
  /// delete a 2D Texture 
  virtual ~Texture2D();
  //@}
 protected:
  /** enlarge a given texture such that its size is a power of 2 and store
      it in the texture buffer */
  void enlargeTexture(const void *b, const unsigned int nr, 
		      const unsigned int nc);
  /// create the display list associated with the texture
  void createTexture();
  /// initialize variables
  void init();
 public:
  /** reads an RGB color image of format SGI Iris from its name and
      returns a pointer on the allocated buffer
      @param name image file name
      @param width the X size of the image
      @param height the Y size of the image
      @param components the number of components (1 for bw, 3 for RGB and 4 for RGBA) */
  static unsigned int *read_texture(char *name, unsigned int *width, unsigned int *height, unsigned int *components);
  /** draw a texture of a given type
      @param type the type of generic texture
      @param exponent for the gaussian and halo texture controls the spread of the gaussian
      @param radius for the halo texture controls the radius of the halo (as default it is a quarter of the size )
  */
  void drawGenericTexture(GenericTexture type,const double exponent=1.0, const double radius=1.0);
  /// update the OpenGL texture based on the new texture buffer
  void updateTexture() const;
  /// return the nb of texture rows 
  unsigned int getNbRows() const { return nrows;}
  /// return the nb of texture columns
  unsigned int getNbColumns() const { return ncols;}
  /// return the nb of meaningful texture rows 
  unsigned int getNbRealRows() const { return nbRealRows;}
  /// return the nb of meaningful texture columns
  unsigned int getNbRealColumns() const { return nbRealColumns;}
  /** get the RGBA color of a texture pixel 
      @param row the row index of a pixel
      @param col the column index of a pixel
      @param color the array of Red Green Blue and Alpha value*/
  void getColor(const unsigned int row, 
		const unsigned int col,
		unsigned char *color) const;
  /// return the texture ID
  GLuint getId() const {
    return id; }
  /// enable texture coordinates generation
  void enableTextureCoordinatesGeneration() const {
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
  }
  /// disable texture coordinates generation
  void disableTextureCoordinatesGeneration() const {
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T); 
  }
  /// use this texture
  void enableTexture(TextureMode m) {
    if (binded==false)
      createTexture();
    glBindTexture(GL_TEXTURE_2D, id);
    
    glEnable(GL_TEXTURE_2D);
    if (m!=COORDINATES)
      {
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
      }
  }
  /// stop using this texture
  void disableTexture(TextureMode m) const {
      
    glDisable(GL_TEXTURE_2D);

    if (m!=COORDINATES)
      {
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
      }
  }
  /// set the type of texture interpolation
  void interpolateTexture(TextureInterpolation i) const {
    if (i==LINEAR)
      {
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      }
    else if (i==NEAREST)
      {
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      }
  }
  /// set the type of texture extrapolation
  void extrapolateTexture(TextureExtrapolation i) const {
    if (i==CLAMP)
      {
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      }
    else 
      {
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
      }
  }
  /// set the type of texture interpolation
  void applyTextureMode(TextureMode i, const double s[4], const double t[4]) const {
    if (i==SPHERICAL_MAPPING)
      {
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
      }	
    else if (i==PLANAR_MAPPING)
      {
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
       	glTexGendv(GL_S, GL_OBJECT_PLANE, s);
	glTexGendv(GL_T, GL_OBJECT_PLANE, t);  
      }
    else if (i==CAMERA_MAPPING)
      {
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGendv(GL_S, GL_EYE_PLANE, s);
	glTexGendv(GL_T, GL_EYE_PLANE, t);
      }
  }

};


}


#endif
