
#include "vtkActor.h"
#include "vtkConeSource.h"
#include "vtkGlyph3D.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper.h"
#include "vtkRenderWindow.h"
#include "vtkRenderer.h"
#include "vtkSphereSource.h"
#include "vtkObjectFactory.h"

#include <gtk/gtk.h>
#include <assert.h>


#include "GtkRenderWindowInteractor.h"
#include "GtkRenderWindow.h"

static void
init_vtk_stuff (void)
{
  // Required for Gtk/Vtk interface
  vtkObjectFactory::RegisterFactory (GtkInteractorFactory::New ());
  vtkObjectFactory::RegisterFactory (GtkRenderWindowFactory::New ());
}

static vtkRenderer *
build_vtk_ui (void)
{
  // vtkRenderer provides an abstract specification for renderers. Rendering is
  // the process of converting geometry, a specification for lights, and a
  // camera view into an image.
  vtkRenderer *ren;
  // vtkActor is used to represent an entity in a rendering scene.
  vtkActor *sphereActor1, *spikeActor1;
  // vtkSphereSource creates a sphere (represented by polygons) of specified
  // radius centered at the origin.
  vtkSphereSource *sphere;
  // vtkConeSource creates a cone centered at a specified point and pointing in
  // a specified direction.
  vtkConeSource *cone;
  // vtkGlyph3D is a filter that copies a geometric representation (called a
  // glyph) to every point in the input dataset. The glyph is defined with
  // polygonal data from a source filter input.
  vtkGlyph3D *glyph;
  // vtkPolyDataMapper is a class that maps polygonal data (i.e., vtkPolyData)
  // to graphics primitives. vtkPolyDataMapper serves as a superclass for
  // device-specific poly data mappers, that actually do the mapping to the
  // rendering/graphics hardware/software.
  vtkPolyDataMapper *sphereMapper, *spikeMapper;
 

  // Create a renderer window 
  ren = vtkRenderer::New();

  // Create a sphere
  sphere = vtkSphereSource::New();
  sphere->SetThetaResolution(10);
  sphere->SetPhiResolution(10);
  
  // Create the corresponding mapper and actor
  sphereMapper = vtkPolyDataMapper::New();
  sphereMapper->SetInput(sphere->GetOutput());
  sphereActor1 = vtkActor::New();
  sphereActor1->SetMapper(sphereMapper);

  // Create a cone
  cone = vtkConeSource::New();

  // Create the glyph. The glyph will copy the cone at each vertex of the
  // sphere
  glyph = vtkGlyph3D::New();
  glyph->SetInput(sphere->GetOutput());
  glyph->SetSource(cone->GetOutput());
  // Specify whether to use vector or normal to perform vector operations.  We
  // want to use normal vectors so that cones point towards outside the sphere
  glyph->SetVectorModeToUseNormal();
  //glyph->SetVectorModeToUseVector();

  // Either scale by scalar or by vector/normal magnitude.
  glyph->SetScaleModeToScaleByVector();
  // glyph->SetScaleModeToScaleByScalar();

  // Specify scale factor to scale object by.
  glyph->SetScaleFactor(0.25);

  // Create mapper and actor for the spike
  spikeMapper = vtkPolyDataMapper::New();
  spikeMapper->SetInput(glyph->GetOutput());
  spikeActor1 = vtkActor::New();
  spikeActor1->SetMapper(spikeMapper);

  // Add the sphere and spike within the scene
  ren->AddActor(sphereActor1);
  ren->AddActor(spikeActor1);
  ren->SetBackground(0.4,0.1,0.2);

  return ren;
}

static void
init_gtk_stuff (int *argc, char **argv[])
{
  gtk_init (argc, argv);
}

static void
clicked_cb (GtkButton *btn, gpointer data)
{
  assert (data == 0);
  /* trigger end of main loop */
  gtk_main_quit ();
}

static gboolean
delete_cb (GtkWidget *widget, GdkEvent *event, gpointer data)
{
  assert (data == 0);
  /* trigger end of main loop */
  gtk_main_quit ();
  return TRUE;
}

static GtkWidget *
build_gtk_ui (GtkWidget *area, vtkRenderWindow *renWin)
{
  GtkWidget *win, *button, *vbox;
  win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  // This corresponds to the top right window cross click: This click will
  // delete the window and send delete-event. We have to end the program
  // afterwards.
  g_signal_connect (G_OBJECT (win), "delete-event",
 	 	    G_CALLBACK (delete_cb), 0);
  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER (win), vbox);

  // Add the input vtk window (which is a GTK_WIDGET), request minimum size
  gtk_widget_set_size_request (GTK_WIDGET (area), 300, 300);
  gtk_box_pack_start (GTK_BOX (vbox), area, TRUE, TRUE, 0);

  button = gtk_button_new_with_label (" exit ");
  gtk_box_pack_end (GTK_BOX (vbox), button, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (button), "clicked", 
		    G_CALLBACK (clicked_cb), 0);

  gtk_widget_show_all (GTK_WIDGET (win));
  return area;
}

int main (int argc, char *argv[])
{
  vtkRenderWindow *renWin;
  vtkRenderer *ren;
  vtkRenderWindowInteractor *iren;

  // Mostly gtk_init
  init_vtk_stuff ();

  init_gtk_stuff (&argc, &argv);

  renWin = vtkRenderWindow::New();

  ren = build_vtk_ui ();

  renWin->AddRenderer(ren);

  iren = vtkRenderWindowInteractor::New();
  iren->SetRenderWindow(renWin);
  iren->Initialize ();

  // Builds the Gtk ui. GTK_WIDGET (renWin->GetGenericWindowId ()) is the vtk
  // window
  build_gtk_ui (GTK_WIDGET (renWin->GetGenericWindowId ()), renWin);

  // Launches the ui.
  gtk_main ();

  return 0;
}
