#ifndef MIPS_VIEW_H
#define MIPS_VIEW_H

#include <vector>
#include <string>
#include <utility>
#include <gtk/gtk.h>

namespace mips {

class ViewSpace;

/** \class View
    \ingroup MipsGUI
*/
class View 
{
  // PUBLIC TYPES
  //
 public:
  /** View interaction style enumeration. */
  enum InteractionStyleIds {
    NO_INTERACTION,
    SELECT_INTERACTION,
    WINDOW_LEVEL_INTERACTION,
    FULL_PAGE_INTERACTION,
    MEASURE_INTERACTION,
    ZOOM_INTERACTION
  };

 public:
  /** Create a View.  */
  View();

  /** Destroy the View and unreference the GtkWidget.  */
  virtual ~View();

  /** Send a render message to the associated rendering mechanism. */
  virtual void Render()=0;

  /** Update the the display parameters based on the data information. */
  virtual void Update()=0;

  /** Return the Gtk widget in which the View is displayed*/
  virtual GtkWidget * GetGtkWidget(); 

  /** Pack this widget into the view. */
  void AddGtkWidget (GtkWidget *widget);

  /** Render OffScreen and return a GdkPixbuf containing the result. */
  virtual GdkPixbuf*  CreatePixbuf (unsigned int p_width, unsigned int p_height); 

  /** Return true if the view contains the given widget in it's widget tree. */
  virtual bool IsContaining(GtkWidget* p_widget);

  /** Set the name of the view.*/
  void SetName(const char* p_name);
  /** Get the name of the view.*/
  const char * GetName();  
  
  /** Toggle view to a selected graphic mode */
  virtual void SetSelected(); 

  /** Toggle view to a selected graphic mode */
  virtual void SetUnselected();

  /** Return true if the view is in the selected mode */
  virtual bool IsSelected();

  /** Set/Get annotations methods*/
  void         SetUpLeftAnnotation (const char *p_annotation);  
  const char * GetUpLeftAnnotation ();  
  void         SetUpRightAnnotation (const char *p_annotation);  
  const char * GetUpRightAnnotation ();  
  void         SetNorthAnnotation (const char *p_annotation);  
  const char * GetNorthAnnotation ();  
  void         SetSouthAnnotation (const char *p_annotation);  
  const char * GetSouthAnnotation ();  
  void         SetEastAnnotation  (const char *p_annotation);
  const char * GetEastAnnotation  ();  
  void         SetOuestAnnotation (const char *p_annotation);  
  const char * GetOuestAnnotation ();
  
  /** Request the widget to have a minimum size */
  virtual void SetSize(unsigned int p_width, unsigned int p_height);

  /** Set the user interaction type..*/
  void         SetInteractionStyle(unsigned int p_style);
  /** Get the user interaction type..*/
  unsigned int GetInteractionStyle() const;

  /** Copy the given point to the current point.*/
  void          SetCurrentPoint(const double p_point[3]);
  /** Return the current point.*/
  const double* GetCurrentPoint() const;
  /** Return the current point.*/
  void          GetCurrentPoint(double p_point[3]) const;
  /** Return the current point.*/
  const char*   GetCurrentPointStringValue() ;
  
  /** Set (copy) the windows level from given view window level.*/
  virtual void SetWindowLevelFrom(View* p_view){};
  /** Set (copy) the current from given view current point.*/
  void SetCurrentPointFrom(const View* p_view);

    
  void         AddMouseButtonPressObserver(GCallback,  void *p_observer);  
  void         AddFullPageObserver        (GCallback,  void *p_observer);
  void         AddCurrentPointObserver    (GCallback,  void *p_observer);
  void         AddWindowLevelObserver     (GCallback,  void *p_observer);

 private:
  void OnMouseButtonPress(GtkWidget *p_viewWidget,
                          GdkEventButton *p_event);
  static gboolean OnMouseButtonPressCbk (GtkWidget *p_viewWidget,
                                         GdkEventButton *p_event,
                                         gpointer p_view);
  
 protected:
  std::vector< std::pair<GCallback, void *> >    m_mouseButtonPressObservers;  
  std::vector< std::pair<GCallback, void *> >    m_fullPageObservers;  
  std::vector< std::pair<GCallback, void *> >    m_currentPointObservers;  
  std::vector< std::pair<GCallback, void *> >    m_windowLevelObservers;  

 private: 
  std::string           m_name;
  GtkWidget            *m_topWidget;
  GtkWidget            *m_alignment;
  GdkColor             *m_unselectedColor;
  GdkColor             *m_selectedColor;
  bool                  m_selected;
  unsigned int          m_interactionStyle;
  double                m_currentPoint[3];
  std::string           m_upLeftAnnotation;
  std::string           m_upRightAnnotation;
  std::string           m_northAnnotation;
  std::string           m_southAnnotation;
  std::string           m_eastAnnotation;
  std::string           m_ouestAnnotation;

  /** Class initialisation. If renderer is given it is used, otherwise it is 
      created. */
  virtual void Initialize();

  /** Get a pixbuf from the widget. Assert the widget is realized. */
  virtual GdkPixbuf* GetPixbufFromWidget (GtkWidget* p_widget);
  static void viewport_unrealize_cb (GtkWidget *widget,
				     gpointer user_data);

  /** When the current point is modified, that method is the place for child
      classes to do all their updates. */
  virtual void UpdatePosition (){};

  /** Return the value of the current point in double. Child classes have to
      override that method to get the correct value. */
  virtual double GetCurrentPointDoubleValue () {return 0;};

  /** Refresh annotations on the view.*/
  virtual void UpdateAnnotations(){};

};

} //namespace mips

#endif /* MIPS_VIEW_H */
