#ifndef MIPS_VIEWSPACE_H
#define MIPS_VIEWSPACE_H

#include <gtk/gtk.h>
#include <vector>
#include <map>
#include "mipsViewSpaceLayout.h"

namespace mips {

class GladeGui;
class View;
class ViewSpaceLayout;
class DataSet;

/** \class ViewSpace
    \ingroup MipsGUI
*/
class ViewSpace 
{
  // PUBLIC TYPES
  //
 public:
  /** Array containing an ordered list of views */
  typedef std::vector<View*> ViewArrayType;

  /** A ViewId is the index number of view in a ViewArrayType. The first View 
      in an ArrayView have an Id equal to 0. If Id is equal to -1 the Id is 
      considered as a NULL pointer. */
  typedef int ViewId;

  /** Array containing an ordered list of ViewId */
  typedef std::vector<ViewId> ViewIdArrayType;

  /** Describe the arrangement of some specific views in a tab. In fact, puts
      views in correspondance with the cells of a layout in witch they have
      to be arranged.*/
  struct Page {
	ViewIdArrayType viewIds;
	ViewSpaceLayout layout;
  };

// PUBLIC METHODS
  
 public:
  /** Create a ViewSpace.  */
  ViewSpace();

  /** Destroy the ViewSpace.  */
  ~ViewSpace();

  /** Return the main widget.*/
  GtkWidget * GetTopWidget();

  /** Display the given view in a one view layout. */
  void DisplayOneView (View* p_view);
  
  /** Display all views in the view space. */
  void DisplayAllViews ();

  /** Create a new page in the view space. */
  void AddPage (const Page* p_page);
  
  /** Display a page of the viewspace. */
  void DisplayPage (unsigned int p_pageNumber);

  /** Create a new page in the view space. */
  void AddDataSetPage (DataSet* p_dataSet);
  
  /** Update and display a page of the viewspace. */
  void DisplayDataSetPage (DataSet* p_dataSet);

  /** Add a new view in the ViewSpace, in the specified DataSet page.
      @param p_dataSet : The dataset for witch a DataSetPage have already been
      created.
      @param p_view : The view to add in the page.
      @param p_position : The position of the view in the layout. If -1 the
      view will be added at the end of the list.
  */
  void AddViewInDataSetPage(DataSet* p_dataSet, View * p_view, int p_position = -1);

  /** Set the layout of the specified DataSet page.*/
  void SetDataSetPageLayout(DataSet* p_dataSet, ViewSpaceLayout p_layout);

  /** Add a new view in the viewspace. This method do not attach the given view
      to a page.
      @param p_view : The view to add in the viewspace.
      @return : The view id or -1 if the view cannot be added.
   */
  int AddView(View * p_view);

  /** Define an new selection of one view */
  void SelectOneView(View* p_view);

  /** Set the tool used for interactions on views.
   @param p_tool : One of the View::InteractionsStyleIds (SELECT_INTERACTION,
   WINDOW_LEVEL_INTERACTION ...) */  
  void         SetInteractionTool(unsigned int p_tool);
  unsigned int GetInteractionTool();

  /** Add the view in the given group. The view will be removed from the group
      it belonged to before.
      @param p_groupId : the group where the view have to be added. If -1 the
      view will not belong to a group anymore. If the group id don't correspond
      to an already created group, do nothing.*/
  void         SetViewGroup (View* p_view, int p_groupId);
  /** Return the group the view belongs to.*/
  int          GetViewGroup (View* p_view);
  /** Remove a View from the group it belongs*/
  void         RemoveViewFromGroup(View* p_view);  
  /** Create a new empty group of views. Return the group id.*/
  unsigned int CreateViewGroup ();  

  // ATTRUBUTES
 private:
  /** All the views managed by the viewspace */
  ViewArrayType m_views;

  /** The list of the pages that are *not* DataSet pages and *not* the 
      specific tabs (one_view page, all_views page). 
      m_tabsViewArrangements[0] is the tab labeled "1" for the user and the 
      tab number 2 (the third one). */
  std::vector<Page> m_pages;

  /** The list of the DataSet pages ids that associate the DataSet to the
      corresponding pages */
  std::map<DataSet*, unsigned int> m_dataSetPageIds;

  /** The array of groups of views. */
  std::vector< ViewArrayType > m_groups;  

  /** The tab that was the current tab before the actual current tab. 
      @return : the number of the previous page (starting from 0) */
  int     m_previousPage;
  
  /** True if the viewspace is currently displaying a view in full page.*/
  bool m_isFullPageOn;
  
  /** True if under the call of OnWindowLevelChange to block infinite loops.*/
  bool m_isOnWindowLevelChange;

  /** Point to the last selected view in the currently selected views.*/
  View* m_lastSelectedView;

  /** The manager in charge of the viewspace glade project. */
  GladeGui*   m_gladeGui;
  
  /** The main widget of the ViewSpace.*/
  GtkWidget* m_topWidget;

  /** The GtkTableWidget used to display pages */
  GtkWidget* m_table;

   /** Toolbar buttons*/
  GtkToolItem *m_selectorButton;
  GtkToolItem *m_windowLevelButton;  
  GtkToolItem *m_fullPageButton;  
  GtkToolItem *m_measureButton;  
  GtkToolItem *m_zoomOutButton;  
  
  // PRIVATE METHODS
 private:
  // Remove all views from the table.
  void DetachAllViews();  

  /// Display the given views under the given layout in the current Table
  void DisplayTable (ViewIdArrayType& p_viewIdArray, ViewSpaceLayout& p_layout);

  /// Display the given widgets under the given layout in the given table
  void DisplayWidgetsOnTable (GtkTable* p_table, std::vector<GtkWidget*> p_widgets,
                              ViewSpaceLayout& p_layout);

  /// Forward the selected Interaction Tool to all the views of the ViewSpace
  void SetViewsInteractionStyle(unsigned int p_tool);  

  /// Refresh the labels that display the position of the current point
  /// of the last selected view
  void RefreshPositionDisplay();
  
 //
  // CALLBACKS
  //
  void        OnToolbarButtonToggled   (GtkToggleToolButton *p_toggleToolButton);
  static void OnToolbarButtonToggledCbk(GtkToggleToolButton *p_toggleToolButton,
                                        gpointer p_viewSpace);

  // CALLBACKS FROM THE VIEWS
  
  /** Used by Views to forward mouse events. When a View is clicked the viewspace
   have to manage the selection state of the other views*/
  void         OnViewMouseButtonPress    (View* p_view,
                                          GdkEventButton *p_event);
  static void  OnViewMouseButtonPressCbk (View* p_view,
                                          GdkEventButton *p_event,
                                          void * p_viewSpace);

  // Calls when a view have to be displayed on FullPage
  void         OnFullPageEvent    (View* p_view);
  static void  OnFullPageEventCbk (View* p_view, void * p_viewSpace);

  // Calls when a view have change its current point
  void         OnCurrentPointChange    (View* p_view);
  static void  OnCurrentPointChangeCbk (View* p_view, void * p_viewSpace);
  
  // Calls when a view have change its window level
  void         OnWindowLevelChange    (View* p_view);
  static void  OnWindowLevelChangeCbk (View* p_view, void * p_viewSpace);
  
};

} //namespace mips

#endif /* MIPS_VIEWSPACE_H */
