#include "mipsVtkViewImage2D.h"
#include "mipsVtkViewImage2DCommand.h"
#include "mipsVtkViewImage2DFullCommand.h"
#include "mipsInteractorStyleImage2D.h"
#include "mipsVtkView.h"
#include "vtkCamera.h"
#include "vtkProperty2D.h"
#include "vtkRendererCollection.h"
#include "vtkImageReslice.h"
#include "vtkTransform.h"
#include "vtkImageMapToWindowLevelColors.h"
#include "vtkImageActor.h"
#include "vtkImageData.h"
#include "vtkTextActor.h"
#include "vtkActor.h"
#include "vtkPolyDataMapper.h"
#include "vtkLineSource.h"
#include "vtkTextProperty.h"
#include <sstream>
#include <assert.h>
#include <math.h>

namespace mips
{

/** The inner pipeline is :
    vtkImageData > ( vtkImageReslice >
    vtkImageMapToWindowLevelColors) > vtkImageActor
*/
VtkViewImage2D::VtkViewImage2D()
  :VtkView(),m_FirstRender(1), m_showCurrentPoint(true), m_showDirections(true),
   m_showSliceNumber(true)
{
  m_currentImage = NULL;
      
  m_ImageReslicer = vtkImageReslice::New();
  m_ImageActor = vtkImageActor::New();
  m_WindowLevel = vtkImageMapToWindowLevelColors::New();     

  // set up the vtk pipeline  
  m_ImageReslicer->SetOutputDimensionality(2);
  m_ImageReslicer->InterpolateOff();
  m_WindowLevel->SetInput(m_ImageReslicer->GetOutput());

  this->AddActor(m_ImageActor);

  // Interactor Style
  this->InitInteractorStyle(View::SELECT_INTERACTION);
  
  // Setting interactor Style
  
  vtkRendererCollection* collec = this->GetVtkRenderWindow()->GetRenderers();
  collec->InitTraversal();
  m_Renderer = collec->GetNextItem();
  m_Renderer->SetBackground(0.0,0.0,0.0);
  
  m_orientation = AXIAL_ID;
  m_MaxDim = -1;
  double pos[3] = {0,0,0};  
  this->SetCurrentPoint(pos);  
  m_2DPosition[0] = 0;
  m_2DPosition[1] = 0;

  // Initialize cursor lines
  m_horizontalLineSource =  vtkLineSource::New();
  vtkPolyDataMapper* mapper =  vtkPolyDataMapper::New();
  m_horizontalLineActor = vtkActor::New();
  m_horizontalLineActor->SetMapper(mapper);
  this->AddActor(m_horizontalLineActor);  
  mapper->SetInput(m_horizontalLineSource->GetOutput());

  m_verticalLineSource =  vtkLineSource::New();
  mapper      =  vtkPolyDataMapper::New();
  m_verticalLineActor = vtkActor::New();
  m_verticalLineActor->SetMapper(mapper);
  this->AddActor(m_verticalLineActor);  
  mapper->SetInput(m_verticalLineSource->GetOutput());

}

  
VtkViewImage2D::~VtkViewImage2D()
{
  if (m_currentImage != NULL)
    m_currentImage->Delete();
  m_InteractorStyleImage->Delete();
  m_ImageActor->Delete();
  m_WindowLevel->Delete();
  m_ImageReslicer->Delete();
}


void
VtkViewImage2D::Update()
{
  this->UpdateImageActor();
  this->ResetCurrentPoint();  
  this->UpdatePosition();
  //this->PrintSelf(std::cout, vtkIndent(0));      
}


void
VtkViewImage2D::UpdateImageActor()
{
  m_ImageReslicer->GetOutput()->UpdateInformation();  
  m_ImageActor->GetInput()->SetUpdateExtent(
    m_ImageReslicer->GetOutput()->GetWholeExtent());
  m_ImageActor->SetDisplayExtent(m_ImageReslicer->GetOutput()->GetWholeExtent());
  m_FirstRender=true;
  this->InitializeImagePositionAndSize();
}

  
int VtkViewImage2D::SetListOfFiles(ListImageType listOfFiles)
{
  if(listOfFiles.size() == 0)
  {
    std::cerr << "No Image in the list." << std::endl;
    return -1;
  }
  else
  {
    m_ListOfImages = listOfFiles;	
    m_FirstRender = 1;
    return 0;
	
  }
}
  
  
void VtkViewImage2D::InitializeImagePositionAndSize()
{
  if (m_FirstRender)
  {
	
    m_ImageActor->GetInput()->Update();    

    double Pos[3];
    // Get the current position of the camera:
    m_Renderer->GetActiveCamera()->GetPosition(Pos);	
	
    double bnds[6];
    // Get the bounds of the image: coodinates in the real world
    m_ImageActor->GetBounds(bnds);
	
    // extension of the image:
    double xs = (bnds[1] - bnds[0]);
    double ys = (bnds[3] - bnds[2]);
	
    double bnds_x = xs/2.0;
    double bnds_y = ys/2.0;
	
    m_Renderer->GetActiveCamera()->SetParallelScale(xs < ys ? bnds_y:bnds_x); 
	
    m_Renderer->GetActiveCamera()->SetPosition(bnds_x+bnds[0],bnds_y+bnds[2],Pos[2]);
    m_Renderer->GetActiveCamera()->SetFocalPoint(bnds_x+bnds[0],bnds_y+bnds[2],0.0);
	
    m_Renderer->GetActiveCamera()->OrthogonalizeViewUp();
    m_Renderer->ResetCameraClippingRange();
	
    m_Renderer->GetActiveCamera()->ParallelProjectionOn();

    m_FirstRender = 0;  
  }
}

  
void VtkViewImage2D::SetFullInteraction()
{
  vtkInteractorStyleImage *interactor = vtkInteractorStyleImage::New();
  /// Attach the vtkCommand : basically, attach an
  /// observer for basic events (move, click, key-pressed etc)
  VtkViewImage2DFullCommand *cbk = VtkViewImage2DFullCommand::New();  
  cbk->IV = this;  
  interactor->AddObserver(vtkCommand::WindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::StartWindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::ResetWindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::EndWindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::KeyPressEvent, cbk);
  interactor->AddObserver(vtkCommand::EnterEvent, cbk);
  interactor->AddObserver(vtkCommand::LeaveEvent, cbk);
  cbk->Delete();  
  this->SetInteractorStyle(interactor);
}


void
VtkViewImage2D::InitInteractorStyle(InteractionStyleIds p_style)
{
  InteractorStyleImage2D *interactor = InteractorStyleImage2D::New();
  interactor->SetView(this);  
  this->SetInteractionStyle(p_style);
  /// Attach the vtkCommand : basically, attach an
  /// observer for basic events (move, click, key-pressed etc)
  VtkViewImage2DCommand *cbk = VtkViewImage2DCommand::New();  
  cbk->SetView(this);
  interactor->AddObserver(vtkCommand::KeyPressEvent, cbk);
  interactor->AddObserver(vtkCommand::WindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::StartWindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::ResetWindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::EndWindowLevelEvent, cbk);
  interactor->AddObserver(vtkCommand::PickEvent, cbk);
  interactor->AddObserver(vtkCommand::StartPickEvent, cbk);
  interactor->AddObserver(vtkCommand::EndPickEvent, cbk);
  interactor->AddObserver(VtkViewImage2DCommand::StartZSliceMoveEvent, cbk);
  interactor->AddObserver(VtkViewImage2DCommand::ZSliceMoveEvent, cbk);
  interactor->AddObserver(VtkViewImage2DCommand::EndZSliceMoveEvent, cbk);
  interactor->AddObserver(VtkViewImage2DCommand::StartMeasureEvent, cbk);
  interactor->AddObserver(VtkViewImage2DCommand::MeasureEvent, cbk);
  interactor->AddObserver(VtkViewImage2DCommand::EndMeasureEvent, cbk);
  interactor->AddObserver(VtkViewImage2DCommand::FullPageEvent, cbk);
  cbk->Delete();  
  this->SetInteractorStyle(interactor);
}


void
VtkViewImage2D::SetShowDirections(bool p_showDirections)
{
  m_showDirections = p_showDirections;
  if(!m_showDirections) {
    this->SetNorthAnnotation("");    
    this->SetSouthAnnotation("");    
    this->SetEastAnnotation ("");    
    this->SetOuestAnnotation("");
  }  
}
bool
VtkViewImage2D::GetShowDirections()
{
  return m_showDirections;  
}
void
VtkViewImage2D::SetShowCurrentPoint(bool p_showCurrentPoint)
{
  m_showCurrentPoint = p_showCurrentPoint;
  if (!m_showCurrentPoint) {
    m_horizontalLineSource->SetPoint1(0, 0, 0);
    m_horizontalLineSource->SetPoint2(0, 0, 0);   
    m_verticalLineSource->SetPoint1(0, 0, 0);   
    m_verticalLineSource->SetPoint2(0, 0, 0);
  }  
}
bool
VtkViewImage2D::GetShowCurrentPoint()
{
  return m_showCurrentPoint;  
}
void
VtkViewImage2D::SetShowSliceNumber(bool p_showSliceNumber)
{
  m_showSliceNumber = p_showSliceNumber;
  if (!m_showSliceNumber) {
    this->SetUpRightAnnotation("");
  }  
}
bool
VtkViewImage2D::GetShowSliceNumber()
{
  return m_showSliceNumber;  
}

void
VtkViewImage2D::DisplayInFullPage()
{
  void (*callback) (View*, void*);
  // Forward the signal to the observers 
  for (unsigned int i = 0; i < m_fullPageObservers.size() ; i++ ) {
    callback = (void (*) (View*, void*))
      (m_fullPageObservers[i].first);
   (*callback) (this, m_fullPageObservers[i].second);    
  }
}

int
VtkViewImage2D::GetWholeZMin()
{
  return 0;  
}

int
VtkViewImage2D::GetWholeZMax()
{
  if(!m_currentImage)
    return 0;  
  int* dims = m_currentImage->GetDimensions();
  assert(m_orientation<NB_DIRECTION_IDS);  
  int axis = this->GetOrthogonalAxis(m_orientation);
  return dims[axis]-1;
}

float
VtkViewImage2D::GetWholeMinPosition(unsigned int p_axis)
{
  if(!m_currentImage)
    return -VTK_LARGE_FLOAT;
  
  return m_currentImage->GetBounds()[p_axis * 2] -
    m_currentImage->GetSpacing()[p_axis] / 2;
}
float
VtkViewImage2D::GetWholeMaxPosition(unsigned int p_axis)
{
  if(!m_currentImage)
    return VTK_LARGE_FLOAT;  

  return m_currentImage->GetBounds()[p_axis * 2 + 1] +
    m_currentImage->GetSpacing()[p_axis] / 2;
}

int
VtkViewImage2D::GetZSlice()
{
  return this->GetSlice(m_orientation);  
}

void
VtkViewImage2D::SetZSlice(int p_zslice)
{
  SetSlice(m_orientation,p_zslice);  
}

void
VtkViewImage2D::GetCurrentVoxelCoordinates(int p_coordinates[3])
{
  p_coordinates[this->GetOrthogonalAxis(SAGITTAL_ID)] = this->GetSlice(SAGITTAL_ID);  
  p_coordinates[this->GetOrthogonalAxis(CORONAL_ID)]  = this->GetSlice(CORONAL_ID);  
  p_coordinates[this->GetOrthogonalAxis(AXIAL_ID)]    = this->GetSlice(AXIAL_ID);  
}

int
VtkViewImage2D::GetSlice(unsigned int p_plan)
{
  assert(p_plan < NB_DIRECTION_IDS);  
  if(!m_currentImage)
    return 0;

  const double* pos     = this->GetCurrentPoint();
  double*       spacing = m_currentImage->GetSpacing();
  double*       origin  = m_currentImage->GetOrigin();

  int   axis     = this->GetOrthogonalAxis(p_plan);
  float soft_pos = pos[axis];
  float pos_max  = this->GetWholeMaxPosition(axis);
  float pos_min  = this->GetWholeMinPosition(axis);

  // Treat extreme position at the end of the last pixel
  if ((soft_pos > pos_max-0.005) && (soft_pos < pos_max+0.005)){
    soft_pos = pos_max-0.005;
  }  
  if ((soft_pos > pos_min-0.005) && (soft_pos < pos_min+0.005)){
    soft_pos = pos_min+0.005;
  }
  return (int)rint((soft_pos-origin[axis])/spacing[axis]);    
}

unsigned int
VtkViewImage2D::GetOrthogonalAxis(unsigned int p_plan)
{
  assert(p_plan<NB_DIRECTION_IDS);  
    
  switch(p_plan)
  {
  case SAGITTAL_ID: 
    return X_ID;    
  case CORONAL_ID: 
    return Y_ID;    
  case AXIAL_ID:   
    return Z_ID;    
  }
  return 0;    
}


void
VtkViewImage2D::SetSlice(unsigned int p_plan, int p_zslice)
{
  if(!m_currentImage)
    return;
  assert(p_plan<NB_DIRECTION_IDS);  
  unsigned int axis = GetOrthogonalAxis(p_plan);
  
  double* spacing = m_currentImage->GetSpacing();
  int*    dims = m_currentImage->GetDimensions();
  double* origin = m_currentImage->GetOrigin();
  int     slice = p_zslice;
  
  if(slice >= dims[axis])  slice = dims[axis]-1;      
  if(slice <  0)           slice = 0;

  double pos[3];
  this->GetCurrentPoint(pos);
  pos[axis] = origin[axis] + slice * spacing[axis];
  SetCurrentPoint(pos);  
}

void
VtkViewImage2D::UpdatePosition ()
{
  double  x,y,max_x,max_y, min_x, min_y;
  double pos[3];
  this->GetCurrentPoint(pos);

  switch (m_orientation) {
  case SAGITTAL_ID:
   m_ImageReslicer->SetResliceAxesOrigin(pos[0],0,0);
   x = (float)pos[1]*-1;
   y = (float)pos[2];
   max_x = this->GetWholeMaxPosition(1) * -1;
   max_y = this->GetWholeMaxPosition(2);
   min_x = this->GetWholeMinPosition(1) * -1;
   min_y = this->GetWholeMinPosition(2);
   break;
  case CORONAL_ID:
   m_ImageReslicer->SetResliceAxesOrigin(0,pos[1],0);
   x = (float)pos[0];
   y = (float)pos[2];
   max_x = this->GetWholeMaxPosition(0);
   max_y = this->GetWholeMaxPosition(2);
   min_x = this->GetWholeMinPosition(0);
   min_y = this->GetWholeMinPosition(2);
   break;
  case AXIAL_ID:
   m_ImageReslicer->SetResliceAxesOrigin(0,0,pos[2]);
   x = (float)pos[0];
   y = (float)pos[1]*-1;
   max_x = this->GetWholeMaxPosition(0);
   max_y = this->GetWholeMaxPosition(1) * -1;
   min_x = this->GetWholeMinPosition(0);
   min_y = this->GetWholeMinPosition(1) * -1;
   break;   
  }
  if(m_showCurrentPoint) {    
    m_horizontalLineSource->SetPoint1(min_x, y, 0);   
    m_horizontalLineSource->SetPoint2(max_x, y, 0);   
    m_verticalLineSource->SetPoint1(x, min_y, 0);   
    m_verticalLineSource->SetPoint2(x, max_y, 0);
  }
  
    
  // Update annotations
  if(m_showSliceNumber) {    
    std::ostringstream os;
    os << this->GetZSlice() << "/" << this->GetWholeZMax();  
    this->SetUpRightAnnotation(os.str().c_str());
  }
}


double
VtkViewImage2D::GetCurrentPointDoubleValue () 
{
  if (!m_currentImage) return -1;
  
  int coordinates[3];
  this->GetCurrentVoxelCoordinates(coordinates);  
  void *scalar = m_currentImage->GetScalarPointer(coordinates);
  
  switch (m_currentImage->GetScalarType()) {
  case VTK_DOUBLE:
    return (*(double*)scalar);
  case VTK_FLOAT:
    return (double) (*(float*)scalar);
  case VTK_UNSIGNED_LONG:
    return (double) (*(unsigned long*)scalar);
  case VTK_LONG:
    return (double) (*(long*)scalar);
  case VTK_UNSIGNED_INT:
    return (double) (*(unsigned int*)scalar);
  case VTK_INT:
    return (double) (*(int*)scalar);
  case VTK_UNSIGNED_SHORT:
    return (double) (*(unsigned short*)scalar);
  case VTK_SHORT:
    return (double) (*(short*)scalar);
  case VTK_UNSIGNED_CHAR:
    return (double) (*(unsigned char*)scalar);
  case VTK_CHAR:
    return (double) (*(char*)scalar);
  }    
  return -1;        
}

float
VtkViewImage2D::GetColorWindow()
{
  return m_WindowLevel->GetWindow();
}
float 
VtkViewImage2D::GetColorLevel()
{
  return m_WindowLevel->GetLevel();
}
void  
VtkViewImage2D::SetColorWindow(float s)
{
  m_WindowLevel->SetWindow(s);
  // Forward the signal to the observers 
  void (*callback) (View*, void*);
  for (unsigned int i = 0; i < m_windowLevelObservers.size() ; i++ ) {
    callback = (void (*) (View*, void*))
      (m_windowLevelObservers[i].first);
   (*callback) (this, m_windowLevelObservers[i].second);    
  }
}
void   
VtkViewImage2D::SetColorLevel(float s)
{
  m_WindowLevel->SetLevel(s);
  // Forward the signal to the observers 
  void (*callback) (View*, void*);
  for (unsigned int i = 0; i < m_windowLevelObservers.size() ; i++ ) {
    callback = (void (*) (View*, void*))
      (m_windowLevelObservers[i].first);
   (*callback) (this, m_windowLevelObservers[i].second);    
  }
}

void
VtkViewImage2D::SetWindowLevelFrom(View* p_view)
{
  VtkViewImage2D* image_view = dynamic_cast<VtkViewImage2D*>(p_view);
  
  if (image_view)  {    
    this->SetColorWindow(image_view->GetColorWindow());    
    this->SetColorLevel(image_view->GetColorLevel());
  }  
}

void VtkViewImage2D::SetTransform(vtkTransform* p_transform)
{
  m_ImageReslicer->SetResliceTransform(p_transform);
}


void VtkViewImage2D::SetImage(vtkImageData* image)
{
  if(!image)  return;

  m_currentImage = image;
  m_ImageReslicer->SetInput(m_currentImage);
  m_ImageActor->SetInput(m_WindowLevel->GetOutput());

  this->SetOrientation(m_orientation);  
  m_FirstRender = true;

  this->InitializeImagePositionAndSize();
}


void VtkViewImage2D::SetOrientation(unsigned int p_orientation)
{
  if(p_orientation > NB_PLAN_IDS - 1)
    return;
  
  m_orientation = p_orientation;
  
  if(!m_currentImage)
    return;

  switch(m_orientation)
  {
  case AXIAL_ID:    // Z fixed
    m_ImageReslicer->SetResliceAxesDirectionCosines(1, 0, 0,
                                                    0,-1, 0,
                                                    0, 0,-1);
    if (m_showDirections) {      
      this->SetNorthAnnotation("A");    
      this->SetSouthAnnotation("P");    
      this->SetEastAnnotation ("R");    
      this->SetOuestAnnotation("L");
    }    
    break;    
  case SAGITTAL_ID: // X fixed	
    m_ImageReslicer->SetResliceAxesDirectionCosines( 0,-1,0,
                                                     0, 0,1,
                                                    -1, 0,0);
    if (m_showDirections) {      
      this->SetNorthAnnotation("S");    
      this->SetSouthAnnotation("I");    
      this->SetEastAnnotation ("P");    
      this->SetOuestAnnotation("A");
    }    
    break;
  case CORONAL_ID:  // Y fixed
    m_ImageReslicer->SetResliceAxesDirectionCosines(1, 0,0,
                                                    0, 0,1,
                                                    0,-1,0);
    if (m_showDirections) {      
      this->SetNorthAnnotation("S");    
      this->SetSouthAnnotation("I");    
      this->SetEastAnnotation ("R");    
      this->SetOuestAnnotation("L");
    }    
    break;	
  }
  
  m_FirstRender = 1;
  this->UpdateImageActor();
  this->UpdatePosition();
}

void
VtkViewImage2D::SetInterpolationMode(int i)
{
  m_ImageActor->SetInterpolate(i);
}

int
VtkViewImage2D::GetInterpolationMode(void)
{
  return m_ImageActor->GetInterpolate();
}

void
VtkViewImage2D::ChangeVolume(int vol)
{	
  if ( vol > (int)(m_ListOfImages.size()) - 1 )
    return;
  
  vtkImageData* im = m_ListOfImages[vol];
  
  if(im)
    this->SetImage(im);	
}

void
VtkViewImage2D::ResetWindowLevel()
{
  if (!m_currentImage) return;
  
  m_currentImage->UpdateInformation();
  m_currentImage->SetUpdateExtent(m_currentImage->GetWholeExtent());
  m_currentImage->Update();
  double *range = m_currentImage->GetScalarRange();
  this->SetColorWindow(range[1] - range[0]);
  this->SetColorLevel(0.5 * (range[1] + range[0]));
  this->Render();
}

void
VtkViewImage2D::ResetCurrentPoint()
{
  if (!m_currentImage) return;

  double* spacing = m_currentImage->GetSpacing();
  int*    dim     = m_currentImage->GetDimensions();
  double* origin  = m_currentImage->GetOrigin();

  double  pos[3] = 
    {
      origin[0] + (dim[0]-1) / 2.0 * spacing[0],
      origin[1] + (dim[1]-1) / 2.0 * spacing[1],
      origin[2] + (dim[2]-1) / 2.0 * spacing[2]
    };

  this->SetCurrentPoint(pos);  
}

vtkImageActor*
VtkViewImage2D::GetImageActor()
{
  return m_ImageActor;  
}

void
VtkViewImage2D::PrintSelf(ostream& os, vtkIndent indent)
{
  os << indent << "Orientation = " << m_orientation << "\n";  
  double* bounds   = 0;
  double* spacing  = 0;
  double* origin   = 0;
  int*    extent   = 0;
  int*    wextent  = 0;
  int*    uextent  = 0;
  m_currentImage->Update();  
  m_ImageReslicer->GetOutput()->Update();  
  m_WindowLevel->GetOutput()->Update();
  
  m_currentImage->GetBounds();
  m_ImageReslicer->GetOutput()->GetBounds();
  m_WindowLevel->GetOutput()->GetBounds();
  m_ImageActor->GetBounds();
    
  extent  = m_currentImage->GetExtent();  
  uextent = m_currentImage->GetUpdateExtent();  
  wextent = m_currentImage->GetWholeExtent();  
  origin  = m_currentImage->GetOrigin();
  spacing = m_currentImage->GetSpacing();
  os << indent << "InputImage   - "<<m_currentImage->GetNumberOfPoints()<<" points.\n"
     << "  extent : ("
     << extent[0] << "/" << extent[1] << " ; "
     << extent[2] << "/" << extent[3] << " ; "
     << extent[4] << "/" << extent[5] << ") \n"  
     << " uextent : ("
     << uextent[0] << "/" << uextent[1] << " ; "
     << uextent[2] << "/" << uextent[3] << " ; "
     << uextent[4] << "/" << uextent[5] << ") \n"  
     << " wextent : ("
     << wextent[0] << "/" << wextent[1] << " ; "
     << wextent[2] << "/" << wextent[3] << " ; "
     << wextent[4] << "/" << wextent[5] << ") \n"  
     << "  origin : ("
     << origin[0] << " , " << origin[1] << " , " << origin[2] << ")\n"
     << "  spacing : ("
     << spacing[0] << " , " << spacing[1] << " , " << spacing[2] << ")\n";
  
  extent  = m_ImageReslicer->GetOutput()->GetExtent();  
  uextent = m_ImageReslicer->GetOutput()->GetUpdateExtent();  
  wextent = m_ImageReslicer->GetOutput()->GetWholeExtent();  
  origin  = m_ImageReslicer->GetOutput()->GetOrigin();
  spacing = m_ImageReslicer->GetOutput()->GetSpacing();
  os << indent << "Resliced Image   - "<<m_ImageReslicer->GetOutput()->GetNumberOfPoints()<<" points.\n"
     << "  extent : ("
     << extent[0] << "/" << extent[1] << " ; "
     << extent[2] << "/" << extent[3] << " ; "
     << extent[4] << "/" << extent[5] << ") \n" 
     << " uextent : ("
     << uextent[0] << "/" << uextent[1] << " ; "
     << uextent[2] << "/" << uextent[3] << " ; "
     << uextent[4] << "/" << uextent[5] << ") \n"  
     << " wextent : ("
     << wextent[0] << "/" << wextent[1] << " ; "
     << wextent[2] << "/" << wextent[3] << " ; "
     << wextent[4] << "/" << wextent[5] << ") \n"  
     << "  origin : ("
     << origin[0] << " , " << origin[1] << " , " << origin[2] << ")\n"
     << "  spacing : ("
     << spacing[0] << " , " << spacing[1] << " , " << spacing[2] << ")\n";
  
  extent  = m_WindowLevel->GetOutput()->GetExtent();  
  uextent = m_WindowLevel->GetOutput()->GetUpdateExtent();  
  wextent = m_WindowLevel->GetOutput()->GetWholeExtent();  
  origin  = m_WindowLevel->GetOutput()->GetOrigin();
  spacing = m_WindowLevel->GetOutput()->GetSpacing();
  os << indent << "WindowLevel Image   - "<<m_WindowLevel->GetOutput()->GetNumberOfPoints()<<" points.\n"
     << "  extent : ("
     << extent[0] << "/" << extent[1] << " ; "
     << extent[2] << "/" << extent[3] << " ; "
     << extent[4] << "/" << extent[5] << ") \n"
     << " uextent : ("
     << uextent[0] << "/" << uextent[1] << " ; "
     << uextent[2] << "/" << uextent[3] << " ; "
     << uextent[4] << "/" << uextent[5] << ") \n"  
     << " wextent : ("
     << wextent[0] << "/" << wextent[1] << " ; "
     << wextent[2] << "/" << wextent[3] << " ; "
     << wextent[4] << "/" << wextent[5] << ") \n"  
     << "  origin : ("
     << origin[0] << " , " << origin[1] << " , " << origin[2] << ")\n"
     << "  spacing : ("
     << spacing[0] << " , " << spacing[1] << " , " << spacing[2] << ")\n";
  
  bounds = m_currentImage->GetBounds();  
  os << indent << "InputImage \n"
     << "           bounds : ("
     << bounds[0] << "/" << bounds[1] << " ; "
     << bounds[2] << "/" << bounds[3] << " ; "
     << bounds[4] << "/" << bounds[5] << ") \n";
  bounds = m_ImageReslicer->GetOutput()->GetBounds();  
  os << indent << "Resliced Image \n"
     << "           bounds : ("
     << bounds[0] << "/" << bounds[1] << " ; "
     << bounds[2] << "/" << bounds[3] << " ; "
     << bounds[4] << "/" << bounds[5] << ") \n";
  bounds = m_WindowLevel->GetOutput()->GetBounds();  
  os << indent << "WindowLevel Image \n"
     << "           bounds : ("
     << bounds[0] << "/" << bounds[1] << " ; "
     << bounds[2] << "/" << bounds[3] << " ; "
     << bounds[4] << "/" << bounds[5] << ") \n";
  bounds = m_ImageActor->GetBounds();  
  os << indent << "Image Actor  bounds : ("
     << bounds[0] << "/" << bounds[1] << " ; "
     << bounds[2] << "/" << bounds[3] << " ; "
     << bounds[4] << "/" << bounds[5] << ") \n";

}

} // end of namespace
