#ifndef _MIPS_VTKVIEWIMAGE2D_H_
#define _MIPS_VTKVIEWIMAGE2D_H_

#include<vector>

// mips includes
#include<mipsVtkView.h>

class vtkImageActor;
class vtkTextActor;
class vtkActor;
class vtkLineSource;
class vtkTextMapper;
class vtkImageReslice;
class vtkImageMapToWindowLevelColors;
class vtkImageData;
class vtkTransform;

namespace mips
{

class InteractorStyleImage2D;
class InrimageDemoGui;
class InrimageDemoGui;

/** \class VtkViewImage2D mipsVtkViewImage2D.h "mipsVtkViewImage2D.h"
    \brief 3D image 2D slice viewer
    \author: Pierre Fillard & Marc Traina
    
    This class allows to view 2D slices of 3D images. Images have to be
    vtkImageData */

class VtkViewImage2D: public VtkView
{
 public:
  /// Indices use in Vtk referentials corresponding to the standard x, y and z unitary vectors
  enum DirectionIds {
    X_ID,          
    Y_ID,       
    Z_ID,
    NB_DIRECTION_IDS
  };
  /// Ids of the 2D plan displayed in the view. Each 2D plan is defined with regard to one of
  /// the 3 directions (x,y,z).
  enum PlanIds {
    SAGITTAL_ID,        
    CORONAL_ID,         
    AXIAL_ID,           
    NB_PLAN_IDS  // The number of PlanIds
  };
  
 public:
  
  // typedef:
  typedef std::vector<vtkImageData*>          ListImageType; 
  typedef ListImageType::iterator             ListImageIteratorType;
    
  
  /// default constructor
  VtkViewImage2D();    
  /// default destructor
  ~VtkViewImage2D();

  /** Update the the display parameters based on the data information. */
  virtual void Update();

  void PrintSelf(ostream& os, vtkIndent indent);  

  void SetShowDirections(bool p_showDirections);
  bool GetShowDirections();
  void SetShowCurrentPoint(bool p_showCurrentPoint);
  bool GetShowCurrentPoint();  
  void SetShowSliceNumber(bool p_showSliceNumber);
  bool GetShowSliceNumber();  

  /// Set a list of image filenames to be loaded and rendered
  /// @param listOfFiles a list of VTK files
  int SetListOfFiles(ListImageType listOfFiles);
  /// Get the current image
  vtkImageData *GetInput() { return m_currentImage; }
  /// Set the current image
  void SetImage(vtkImageData* image);
  /// Return the vtkImageActor of the view
  vtkImageActor* GetImageActor();
  /// Set a transform to apply to visualize the image.
  void SetTransform(vtkTransform* p_transform);  
  /// Set the orientation of the image
  /// @param orientation 0 is XY, 1 is ZY, 2 is XZ
  void SetOrientation(unsigned int orientation);
  /// Get the orientation of the current image
  unsigned int GetOrientation(void) {return m_orientation;};
  /// Add a file
  int AddImage(vtkImageData* image);
    
  /// Get the id of the first slice available relative to the current orientation.
  int GetWholeZMin();
  /// Get the id of the last slice available relative to the current orientation.
  int GetWholeZMax();
  /// Get the begining position of the first pixel in the given axis. Note : the
  /// position (0,0,0) is the center of the first pixel (0,0,0) so the return value can be
  /// negative, depending of the origin of the image.
  float GetWholeMinPosition(unsigned int p_axis);
  /// Get the end position of the last pixel in the given axis.
  float GetWholeMaxPosition(unsigned int p_axis);
  

  
  /// Get the current slice. Z is relative to the displayed plan
  int GetZSlice();  
  /// Set the Slice to display. Z is relative to the displayed plan
  void SetZSlice(int p_slice);
  /// Get the slice number within the current point is located.
  /// @param p_orientation : the desired slice type AXIAL, SAGITTAL, CORONAL
  int  GetSlice(unsigned int p_orientation);  
  /// Set the slice number within the current point is located.
  /// @param p_slice : the desired slice number.
  /// @param p_orientation : the desired slice type AXIAL, SAGITTAL, CORONAL.
  void SetSlice(unsigned int p_orientation, int p_slice);  
  /// Reset the current point to the center of the image in the 3 axes
  void ResetCurrentPoint();  
  
  /// Set/Get window/level for mapping pixels to colors.
  float GetColorWindow();
  /// Set/Get window/level for mapping pixels to colors.
  float GetColorLevel();
  /// Set/Get window/level for mapping pixels to colors.
  void SetColorWindow(float s);
  /// Set/Get window/level for mapping pixels to colors.
  void SetColorLevel(float s);
  /// Reset window level to calculated default value
  void ResetWindowLevel();
  /** Set (copy) the windows level from given view window level.*/
  virtual void SetWindowLevelFrom(View* p_view);
  
  // Set/Get the position in world coordinates of the rendering window.
  //int *GetPosition() {return this->GetPosition();};
  //void SetPosition(int a,int b) {this->SetPosition(a,b);};
  //virtual void SetPosition(int a[2]);
    
  // Set/Get the size of the window in screen coordinates.
  //void SetSize(int a,int b) {this->SetSize(a,b);};
  //virtual void SetSize(int a[2]);
    
  /// Set/Get the interpolation method (1 or 0)
  void SetInterpolationMode(int i);
  /// Set/Get the interpolation method (1 or 0)
  int GetInterpolationMode(void);
    
  /// Update volume among vtkImageData list
  void ChangeVolume(int vol);

/*   void Set2DPosition(float x, float y) */
/*     {m_2DPosition[0] = x; m_2DPosition[1] = y;} */
/*   float *Get2DPosition() {return m_2DPosition;} */

  /// Setup full interaction on the view (Windowing, zoom, reset ...)
  void SetFullInteraction(void);

  /// Send to registered FullPageObservers the signal that the view have to
  /// be displayed in Full page mode.
  void DisplayInFullPage();  
  
 protected:
  /// Initialize rendering parameters only once. m_FirstRender controls if the
  /// initialization has been done or not
  void InitializeImagePositionAndSize();
/*   /// Clamp current position, so that it fits in the image dimensions */
/*   void ClampPosition(); */
  
 private:
  void           InitInteractorStyle(InteractionStyleIds p_style);
  // Update data information and set the update and display extents.
  // Reset the camera position.
  void           UpdateImageActor();

  /// Get the id of the axis orthogonal to the given plan
  /// (x for sagittal, y for coronal and z for axial)
  unsigned int   GetOrthogonalAxis(unsigned int p_plan);
  /// Fill the p_coodirdinate parameter with the coordinates of the voxel
  /// corresponding to the current point
  void           GetCurrentVoxelCoordinates(int p_coordinates[3]);

  /* Overridden methods. */
  virtual void   UpdatePosition ();
  virtual double GetCurrentPointDoubleValue ();

 private:

  /// VTK current image (the one being shown)
  vtkImageData*                   m_currentImage;
  
  /// VTK Renderer:
  vtkRenderer*                    m_Renderer;
  /// Interactor and interactor style:
  InteractorStyleImage2D*         m_InteractorStyleImage;
    
  /// to map the image to the current window level
  vtkImageMapToWindowLevelColors* m_WindowLevel;
    
  /// vtkImageReslice is the swiss-army-knife of image geometry filters: It can
  /// permute, rotate, flip, scale, resample, deform, and pad image data in any
  /// combination with reasonably high efficiency.
  vtkImageReslice*                m_ImageReslicer;
    
  /// actors and mappers:
  vtkImageActor                  *m_ImageActor; 
  vtkActor                       *m_horizontalLineActor;
  vtkLineSource                  *m_horizontalLineSource;  
  vtkActor                       *m_verticalLineActor;
  vtkLineSource                  *m_verticalLineSource;  
  
    
  int                             m_FirstRender;
  bool                            m_showCurrentPoint;
  bool                            m_showDirections;  
  bool                            m_showSliceNumber;  

  /// List of inputed images:
  ListImageType                   m_ListOfImages; 
  
  /// Orientation: 0 is XY, 1 is ZY, 2 is XZ
  unsigned int                    m_orientation;
    
  /// max dim:
  int                             m_MaxDim;

  /// Current 2D position of the mouse within the view
  float                           m_2DPosition[2];
};
  
  
} // end of namespace

#endif
