#include "mipsIterationUnit.h"
#include <assert.h>
#include <iostream>
#ifdef WIN32
#  include <algorithm>
#endif

#include "mipsIterationScheduler.h"

using namespace std;
namespace mips {

IterationScheduler::IterationScheduler() {
  m_isRunning=false;
  m_stopRunningAsked=false;

  m_iterationsDone=0;
  m_collectionPeriod=1;

  m_runningTimer=new Timer();
  m_intermediateTimer=new Timer();

  this->ResetCollectionInformations();

  m_printIterationRateAsked=false;
  m_iterationRatePrintOutputStream=&std::cout;
}

IterationScheduler::~IterationScheduler() {
  delete m_runningTimer;
  delete m_intermediateTimer;
}
  
void IterationScheduler::ResetCollectionInformations() {
  m_intermediateIterations=0;
  m_intermediateTimer->Reset();
  m_collectedTime=0;
  m_collectedIterations=0;
  IterationScheduler::IterationUnitListType::iterator it;
  for (it=m_unitList.begin() ; it!=m_unitList.end() ; it++) {
	(*it)->ResetIterationRateData();
  }
}


void IterationScheduler::Iterate (unsigned long nb_iter)
{
  if (m_isRunning) return;
  this->StartTheRun();

  while ((m_iterationsDone<nb_iter) && (!m_stopRunningAsked)) {
	this->Step();
  }

  this->StopTheRun();
}


void IterationScheduler::Run ()
{
  if (m_isRunning) return;
  this->StartTheRun();

  do {
	this->Step();
  } while (!m_stopRunningAsked);

  this->StopTheRun();
}

void IterationScheduler::StartTheRun ()
{
  // Reset variables
  m_isRunning=true;
  m_stopRunningAsked=false;
  this->ResetCollectionInformations();
  m_runningTimer->Reset();
  m_iterationsDone=0;
  m_currentUnitIterator=m_unitList.end();
}

void IterationScheduler::Step ()
{
  assert(m_isRunning);
  if (m_stopRunningAsked) return;
  
  // Test if we are in the case of the END of an iteration
  if (m_currentUnitIterator==m_unitList.end()) {		 
	m_currentUnitIterator=m_unitList.begin();
  }
  
  // Test if we are in the case of the BODY of an iteration
  //  if(m_currentUnitIterator!=m_unitList.end()) {
	m_runningTimer->Start();
	m_intermediateTimer->Start();
	(*m_currentUnitIterator)->Run();
	m_runningTimer->Stop();
	m_intermediateTimer->Stop();
	m_currentUnitIterator++;
	//  }
  
  // Test if we are in the case of the END of an iteration
  if (m_currentUnitIterator==m_unitList.end()) {		 
	m_iterationsDone++; 
	m_intermediateIterations++;
	this->CollectIterationRateInformations();
  }
}

void IterationScheduler::StopTheRun ()
{
  m_isRunning=false;
  m_stopRunningAsked=false;
}

void IterationScheduler::AddUnit (IterationUnit * unit)
{
  m_unitList.push_back(unit);
}

void IterationScheduler::Stop ()
{
   m_stopRunningAsked=true;
}

bool IterationScheduler::IsStopAsked ()
{
   return m_stopRunningAsked;
}

bool IterationScheduler::IsRunning ()
{
   return m_isRunning;
}
void IterationScheduler::CollectIterationRateInformations() {
  if (m_collectionPeriod-m_intermediateTimer->GetTime() > 1e-4) return;
  
  m_collectedIterations=m_intermediateIterations;
  m_intermediateIterations=0;
  m_collectedTime=m_intermediateTimer->GetTime();
  m_intermediateTimer->Reset();
  
  IterationUnitListType::iterator it;
  for (it=m_unitList.begin() ; it!=m_unitList.end() ; it++) {
	(*it)->CollectIterationRateData();
  }

  if (m_printIterationRateAsked) {
	  this->PrintIterationRateInformations(*m_iterationRatePrintOutputStream);
  }
}

double IterationScheduler::GetRunningTime() {
  return m_runningTimer->GetTime();
}

unsigned long IterationScheduler::GetIterationsDone() {
  return m_iterationsDone;
}

void IterationScheduler::SetCollectPeriod(double period) {
  m_collectionPeriod=period;
}

double IterationScheduler::GetLastCollectPeriodTime() {
  return m_collectedTime;
}

unsigned long IterationScheduler::GetLastCollectPeriodIterations() {
  return m_collectedIterations;
}

double IterationScheduler::GetGlobalTimePerIteration() {
  return m_runningTimer->GetTime()/m_iterationsDone;
}

double IterationScheduler::GetGlobalIterationFrequency() {
  return m_iterationsDone/m_runningTimer->GetTime();
}

double IterationScheduler::GetInstantTimePerIteration() {
  return m_collectedTime/m_collectedIterations;
}

double IterationScheduler::GetInstantIterationFrequency() {
  return m_collectedIterations/m_collectedTime;
}

double IterationScheduler::GetInstantUnitLoadRatio(IterationUnit* unit_ptr) {
  IterationUnitListType::iterator unit_it;
  unit_it = find(m_unitList.begin(),m_unitList.end(),unit_ptr);
  double mean_unit_running_time=(*unit_it)->GetRunningTime()/m_collectedIterations;
  double mean_collected_time=m_collectedTime/m_collectedIterations;
  return mean_unit_running_time/mean_collected_time;
}

double IterationScheduler::GetInstantUnitActivationLoadRatio(IterationUnit* unit_ptr) {
  IterationUnitListType::iterator unit_it;
  unit_it = find(m_unitList.begin(),m_unitList.end(),unit_ptr);
  unsigned long nb_activations=(*unit_it)->GetNumberOfActivations();
  double load_ratio=double(nb_activations)/m_collectedIterations;
  return load_ratio;
}


void IterationScheduler::ActivatePrintIterationRate(std::ostream& dest_stream) {
  m_printIterationRateAsked=true;
  m_iterationRatePrintOutputStream=&dest_stream;
}

void IterationScheduler::StopPrintIterationRate(){
  m_printIterationRateAsked=false;
}

void IterationScheduler::PrintIterationRateInformations(std::ostream& os) {
  os << " --> Iteration rate on " << m_collectedIterations << " iterations : "
	 <<this->GetInstantIterationFrequency() << " Hz ("
	 << GetInstantTimePerIteration()*1000 <<" ms/iter.)"<<std::endl;
  os << "       unit iteration rate ratio :" << std::endl;
  IterationUnitListType::iterator unit_it;
  for (unit_it=m_unitList.begin() ; unit_it!=m_unitList.end() ; unit_it++) {
	os << "         " <<(*unit_it)->GetLabel()<<" -> "
	   <<this->GetInstantUnitLoadRatio(*unit_it)*100<<" % of the elapsed time"
	   << " (" <<  (*unit_it)->GetNumberOfActivations() << " activations - "
	   << this->GetInstantUnitActivationLoadRatio(*unit_it)*100 << "% of iterations)"
	   << std::endl;
  }  
  os << std::endl<< std::endl;
}

} //namespace mips
