#ifndef MIPS_ITERATION_SCHEDULER_H
#define MIPS_ITERATION_SCHEDULER_H

#include <list>
#include <iostream>

namespace mips {

class IterationUnit;
class Timer;

/** \class IterationScheduler
    \ingroup MipsViewer
*/
class IterationScheduler {
 public:
  /** Construct an IterationScheduler. */
  IterationScheduler();
  /** Detroy an IterationScheduler. */
  virtual ~IterationScheduler();
  /** Iterate nb_iter over the added units. Iteration can be stopped by calling 
	  the Stop method. */
  virtual void Iterate(unsigned long nb_iter=1);
  /** Iterate over the added units until the Stop method is called. */
  virtual void Run();
  /** Ask to stop the current iteration loop. This method only register the 
	  request and returns without waiting the effective end of the loop. The
	  loop will stop effectively when the current running unit will 
	  terminate. */
  virtual void Stop();
  /** Inform if a stop request is registered and the current running unit is
	  not yet terminated. */
  virtual bool IsStopAsked();
  /** Inform on the running state of the scheduler. If a stop request have been
	  done, but not yet satisfied, the scheduler is considered as still running. */
  virtual bool IsRunning();
  /** Add an IterationUnit at the end of the loop. */
  virtual void AddUnit(IterationUnit * unit);


  // Global Iteration rate informations methods 
  //
  /** Return the running time corresponding to the last call of an Iterate
	  or Run method. If the IterationScheduler is running, the returned time 
	  is the elapsed time since the call of Iterate or Run method. Otherwise, it is
	  the time where the Scheduler was considered as running after the last 
	  call to an Iterate or Run method. 
	  @see IsRunning */
  virtual double GetRunningTime();
  /** Return the number of iterations that the IterationsScheduler have done
	  durring the time it was considered as running.
	  @see : IsRunning method.
      @see : GetRunningTime */
  virtual unsigned long GetIterationsDone();

  // Collected Iteration rate informations methods 
  //
  /** Define the minimum time between two iteration rate informations 
	  collection. The collection is done at the end of the iteration loop.
	  So, the time between two collections is depending of the time
	  taken by each iteration. In the other hand, the number of iterations
	  done between two collection is an integer. The default CollectPeriod
	  value is 1 second.*/
  virtual void SetCollectPeriod(double);
  /** Return the time of the last collect period.*/
  virtual double GetLastCollectPeriodTime();
  /** Return the number of iterations done in the last collect period .*/
  virtual unsigned long GetLastCollectPeriodIterations();

  // Iteration rate informations access methods
  //
  /** Return the mean time spend per iteration calculated over all the 
	  iterations that the IterationsScheduler have done durring the time 
	  it was considered as running. */
  virtual double GetGlobalTimePerIteration();
  /** Return the mean iteration rate or frequency calculated over all the 
	  iterations that the IterationsScheduler have done durring the time 
	  it was considered as running. */
  virtual double GetGlobalIterationFrequency();
  /** Return the time spend per iteration calculated over the last collected
	  period. */
  virtual double GetInstantTimePerIteration();
  /** Return the iteration rate or frequency calculated over the last collected
	  period. */
  virtual double GetInstantIterationFrequency();
  /** Return for a specific IterationUnit, the load impact ratio on an 
	  iteration of the last collected period. */
  virtual double GetInstantUnitLoadRatio(IterationUnit* unit_ptr);
   /** Return for a specific IterationUnit, the effective load impact ratio
	   on an iteration of the last collected period, calculated over the 
	   iterations where the IterationUnit have been activated. */
  virtual double GetInstantUnitActivationLoadRatio(IterationUnit* unit_ptr);

  // Iteration rate printing methods
  //
  /** Activate a formated print of the Iteration rate informations at 
	  each periodic collection. */
  virtual void ActivatePrintIterationRate(std::ostream& dest_stream = std::cout);
  /** Disactivate the print iteration rate informations. */
  virtual void StopPrintIterationRate();
  /** Formated print of the iteration rate informations on the given stream. */
  void PrintIterationRateInformations(std::ostream& dest_stream = std::cout);

 protected:
  /** Used by Iterate and Run methods to initialize the run environnement.*/
  void StartTheRun ();

  /** Run one Unit of an iteration. Run and Iteration methods are only loops 
	  over the Step method.If m_stopRunningAsked, do nothing. Assert that 
	  m_isRunning is true.*/
  void Step();

  /** Used by Iterate and Run methods to finalize the run environnement.*/
  void StopTheRun ();


 private:
  typedef std::list<IterationUnit*> IterationUnitListType;
  
  /** Save and reset temporary values of the iteration rate informations at the 
	  maximum rate corresponding of the defined collect period. */
  void CollectIterationRateInformations ();
  /** Reset the attributes needed for iteration rate informations collection */
  void ResetCollectionInformations() ;




  IterationUnitListType m_unitList;

  bool m_isRunning;
  bool m_stopRunningAsked;

  // Global Iteration rate informations attributs 
  //
  /** Timer to manage time the scheduler iterates */
  Timer* m_runningTimer;
  /** Number of iterations done. */
  unsigned int m_iterationsDone;

  /** The next unit to run or currently running */
  IterationUnitListType::iterator m_currentUnitIterator;


  // Collected Iteration rate informations attributs
  //
  /** Asked collecting period in seconds */
  double m_collectionPeriod; 

  /** Timer to manage intermediate time between collections */
  Timer* m_intermediateTimer;
  /** Intermediate number of iterations done between collections. */
  unsigned int m_intermediateIterations;

  /** Last collected time*/
  double m_collectedTime;
  /** Last collected iterations. */
  unsigned int m_collectedIterations;

  // Iteration rate printing attributes
  //
  /** Register if printing of Iteration rate informations is required */
  bool m_printIterationRateAsked;
  /** Output stream where the automatic printing have to be done */
  std::ostream* m_iterationRatePrintOutputStream;
};

 
} // namespace mips
#endif /* MIPS_ITERATION_SCHEDULER_H */
