#ifndef MIPS_ITERATION_UNIT_H
#define MIPS_ITERATION_UNIT_H

#include <string>
#include <iostream>

#include "mipsIterationScheduler.h"
#include "mipsTimer.h"

namespace mips {

/** \class IterationUnit
    \ingroup MipsViewer
*/
class IterationUnit 
{
 public:
  /** Construct an IterationUnit */
  IterationUnit();

  /** Destroy an IterationUnit */
  virtual ~IterationUnit();

  /** Run the given function if the activation 
	  constraints are realised. The activation constraint depend on the 
	  activation period and the number of non activated runs set by the 
	  corresponding methods.Do nothing otherwise */
  void Run();

  /** Set the period the IterationUnit have to be activated. */
  void SetActivationPeriod(unsigned int);
  /** Set the number of runs that will not be effectively done */
  void SetNumberOfNonActivatedRuns(unsigned int);

  /** Set the Unit label to define it clearly to the end user.*/
  virtual void SetLabel(const std::string& label);
  /** Get the Unit label.*/
  virtual const std::string& GetLabel();
  /** Standard print the object atributes values to the spcified stream.*/
  virtual void PrintSelf(std::ostream& os=std::cout);
  
  /** Get the running time previously collected. */
  double GetRunningTime();
  /** Get how many times the IterationUnit have been effectively activated */
  unsigned long GetNumberOfActivations(); 
  
  

 private:
  

  /** Call the C function. This method is called by IterationUnit::Run method  */
  virtual void CallFunction()=0;
  
  std::string m_label;
  unsigned int m_activationPeriod;
  unsigned int m_numberOfNonActivatedRuns;

  /** Running time since last collection. */
  mips::Timer m_runningTimer;
  /** number of activations since last collection. */
  unsigned long m_currentNumberOfActivations;

  /** collected running time */
  double m_runningTime;
  /** collected number of activations */
  unsigned long m_numberOfActivations;

  //---------------- Friend Zone ------------------

  /* Only the  IterationScheduler::CollectIterationRateInformations() and 
	 IterationScheduler::ResetCollectionInformations() methods are
	 is autorized to acces private members. In fact : respectively 
	 CollectIterationRateData and ResetIterationRateData member functions */
  friend void IterationScheduler::CollectIterationRateInformations();
  friend void IterationScheduler::ResetCollectionInformations();
  
  /** Save and reset running time and number of runs since the last
	  refresh or the construction of the object. */
  void CollectIterationRateData();

  /** Reset temporary and collected running times and number of activations .
   */
  void ResetIterationRateData();
};

} //namespace mips

#endif /* MIPS_ITERATION_UNIT_H */
