#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "modules/TclObject.h"
#include <modules/TclObject.h>
#ifdef __GNUC__
#if ( __GNUC__ < 3 )
#include <strstream>
#else
#include <sstream>
#endif // __GNUC__ < 3
#endif

namespace yav {


TclObject::TclObject(const std::type_info& info, TclModule *m) {
  objectKind = TclObjectKind::findByKind(info);
  init();
}


TclObject::TclObject(const TclObject& object) {
  objectKind = object.objectKind;
  init();
}


void TclObject::init(void) {
  *objectKind += this;
  if(TclModule::verbose) COUT << "  creating object " << name() << ENDL;
}


TclObject::~TclObject() {
  if(TclModule::verbose) COUT << "  deleting object " << name() << ENDL;
  if(objectKind->module) deleteTCLCommand();
}


void TclObject::buildTCLCommand(void) const {
  TclString cmdname;
  cmdname.sprintf("%s%u", objectKind->getName(), getRef());
  Tcl_CreateCommand(objectKind->module->getInterp(), (char *) cmdname,
		    (Tcl_CmdProc *) TclObject::genericCmd, (ClientData) this, 0);
}


void TclObject::deleteTCLCommand(void) const {
  TclString cmdname;
  cmdname.sprintf("%s%u", objectKind->getName(), getRef());
  Tcl_DeleteCommand(objectKind->module->getInterp(), (char *) cmdname);
}


int TclObject::list(TclModule *m, int argc, char **argv) {
  TclObjectKind *kind;
  TclString list, one;

  kind = TclObjectKind::findByName(argv[0]);
  if(kind) {
    ObjectsList<TclObject>::iterator it = kind->objects.begin();
    for(; it!=kind->objects.end(); it++) {
      one.sprintf("%s%u ", kind->getName(), (*it)->getRef());
      list += one;
    }
    m->setResult(list);
  }
  else {
    m->setResult("no such object %s", argv[0]);
  }

  return TCL_OK;
}


int TclObject::create(TclModule *m, int argc, char **argv) {
  TclString res;
  res.sprintf("no creation function provided for %s", argv[0]);
  m->setResult(res);
  return TCL_ERROR;
}


int TclObject::error( int argc, char **argv) {
  TclString res=objectKind->getOptionSynopsis();
  res+= " while executing ";
  int i;
  for (i=0;i<argc;++i)
   res += SPACE_STRING + argv[i];
  CERR << res << ENDL;
  objectKind->module->setResult(res);
  return TCL_ERROR;
}
int TclObject::creationSynopsisCmd(TclModule *m, int argc, char **argv) {
  setResult("%s", getObjectKind()->getCreationSynopsis());
  return TCL_OK;
}

int TclObject::kindCmd(int argc, char **argv) {
  objectKind->module->setResult(objectKind->getName());
  return TCL_OK; 
}
int TclObject::realNameCmd(int argc, char **argv) {
  objectKind->module->setResult(objectKind->getRealName());
  return TCL_OK; 
}

int TclObject::exec(Tcl_Interp *interp, int argc, char **argv) {
   int res;
   
   try{
      if(!objectKind->options.empty()) {
	 if(argc < 2) return error(argc,argv);
	 
	 std::map< char *, CommandOption *, CStringCompare >::iterator it =
	    objectKind->options.find(argv[1]);
	 if(it != objectKind->options.end()) {
	    int (TclObject::*cmd)(int, char **) = it->second->getOptionName();
	    if (cmd)
	      res = (this->*cmd)(argc - 2, argv + 2);
	    else {
	      int (*cmd)(TclObject*,int, char **) = it->second->getGlobalOptionName();
	      res = (*cmd)(this,argc - 2, argv + 2);
	    }
	    if(res == COMMAND_SYNOPSIS) {
	       TclString error;
	       error.sprintf("usage: %s%u %s", objectKind->getName(), getRef(), argv[1]);
	       if(it->second->getOptionSynopsis())
		    error += SPACE_STRING + it->second->getOptionSynopsis() + " while executing ";
		   int i;
		   for (i=0;i<argc;++i)
			   error += SPACE_STRING + argv[i];
		    objectKind->module->setResult(error);
	    }
	    return res;
	 }
	 if(!objectKind->globalCommand) return error(argc,argv);
      }
      
      if(objectKind->globalCommand) {
	 res = (*objectKind->globalCommand)(this, interp, argc, argv);
	 if(res == COMMAND_SYNOPSIS) return error(argc,argv);
	 else return res;
      }
      
      return TCL_OK;
   } catch(Exception& e){
#ifdef __GNUC__
#if  ( __GNUC__ <3 )
     std::ostrstream os;
     os << e << ENDS;
     setResult("%s",os.str());
#else
     std::ostringstream os;
     os << e << ENDS;
     setResult("%s",os.str().c_str());
#endif // __GNUC__ < 3
#else
     setResult("Exception %s thrown from %s\n%s", e.what(), (const char *)e.getStack().top().c_str(),
	      (const char *) e.getError().c_str()); 
#endif
     return TCL_ERROR;
   } catch(std::exception& e){
      setResult("%s",e.what());
      return TCL_ERROR;
   } catch(...){
      setResult("Unknown error occured.");
      return TCL_ERROR;
   }
}
   

int TclObject::genericCmd(ClientData cd, Tcl_Interp *interp, int argc,
			  char **argv) {
  TclObject *cmd = (TclObject *) cd;
  return cmd->exec(interp, argc, argv);
}


void TclObject::setResult(char *s...) const {
  va_list args;

  // format string
  va_start(args, s);
  TclString str(s, args);
  va_end(args);
  
  // eval command
  objectKind->module->setResult(str);
}



int TclObject::tickle(char *s...) const {
  va_list args;

  // format string
  va_start(args, s);
  TclString str(s, args);
  va_end(args);
  
  // eval command
  return objectKind->module->tickle((const TclString) str);
}


void TclObject::name(TclString& objid) const {
  objid.sprintf("%s%u", objectKind->getName(), getRef());
}


TclString TclObject::name(void) const {
  TclString id;
  id.sprintf("%s%u", objectKind->getName(), getRef());
  return id;
}


void TclObject::setObjectKind(const std::type_info& info, TclModule *m) {
  TclObjectKind *kind;
  kind = TclObjectKind::findByKind(info);

  if(kind) {
    if(!kind->module) {
      if(!m) kind->module = objectKind->module;
      else kind->module = m;
    }
    //if(objectKind->module) deleteTCLCommand();
    *objectKind -= this;
    *kind += this;
  }
}


TclObject *TclObject::getHead(const std::type_info& type) {
  TclObjectKind *kind = TclObjectKind::findByKind(type);
  if(kind) return kind->objects.getFirst();
  return 0;
}


TclObject *TclObject::getTail(const std::type_info& type) {
  TclObjectKind *kind = TclObjectKind::findByKind(type);
  if(kind) return kind->objects.getLast();
  return 0;
}


void TclObject::done(const std::type_info& type) {
  TclObjectKind *kind = TclObjectKind::findByKind(type);
  kind->created(this);
}


void TclObject::deleted(const std::type_info& type) {
  TclObjectKind *kind = TclObjectKind::findByKind(type);
  kind->deleted(this);
}


} // namespace yav

