#ifndef TCLMODULE_H
#define TCLMODULE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#endif

#pragma interface "modules/TclModule.h"
#include <basics/Architecture.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <basics/Exception.h>
#include <list>
#include <map>
#include <basics/TclString.h>
#include <basics/AutoLink.h>
#include <tcl.h>
#include <tk.h>

#ifndef WIN32
#include <dlfcn.h>
#include <sys/utsname.h>
#define HANDLE_TYPE void *

namespace yav {

#else
#include <wtypes.h>
#define dlopen(A,B) LoadLibrary(A)
#define dlclose(A) FreeLibrary(A)
#define dlsym(A,B) GetProcAddress(A,B)
#define HANDLE_TYPE HINSTANCE

namespace yav {

#endif

#define COMMAND_SYNOPSIS -1

class TclModule;
class TclObject;
class TclObjectKind;


/// pointer on a TCL command
typedef int (*COMMAND) (ClientData, Tcl_Interp *, int, char **);
/// pointer on a display function
typedef void (*DISPLAY_FUNCTION) (const TclModule&, const char *str);
/// pointer on a TCL commands evaluation function
typedef int (*EVAL_FUNCTION) (const TclModule&, const char *str);
/// pointer on a main loop function
typedef void (*LOOP_FUNCTION) (unsigned int);
/// pointer on a function that initialises a module
typedef void (*INIT_MODULE_FUNCTION) (TclModule *);


/** \class TclStringCompare
    \ingroup Modules
    \brief Tclstring comparison class for stl map */
class TclStringCompare {
public:
  /** @name Constructors */
  //@{
  /// default constructor (empty)
  TclStringCompare() { }
  /// copy constructor (empty)
  TclStringCompare(const TclStringCompare&) { }


  
  /** assignement (empty)
      @return *this */
  TclStringCompare& operator=(const TclStringCompare&) { return *this; }

  /** strings comparison operator
      @name s1 first string to compare
      @name s2 second string to compare
      @return true if s1 < s2 */
  bool operator()(const TclString &s1, const TclString &s2) const {
    return (s1<s2);
  }
  //@}
};


/** \class CommandOption
    \ingroup Modules
    \brief Class for TCL command option description */
class CommandOption {
 private:
  /// command option synopsis
  const char *synopsis;
  /// command option effect description
  const char *description;
  /// object method associated to command
  int (TclObject::*fct)(int, char **);
  /// object method associated to command
  int (*fctGlobal)(TclObject *,int, char **);

 public:
  /** @name Constructor */
  //@{  
  /** buids a command option descriptor from its synopsis, description strings and object callback method
      @name cmd object callback method
      @name syn command option synopsis
      @name desc command option description */  
  CommandOption(int (TclObject::*cmd)(int, char **), const char *syn = 0,
		const char *desc = 0)
    : synopsis(syn), description(desc) , fct(cmd) {
    fctGlobal=0;
  }
  //@}
  /** buids a command option descriptor from its synopsis, description strings and object callback method
      @name cmd global object callback method
      @name syn command option synopsis
      @name desc command option description */  
  CommandOption(int (*cmd)(TclObject *,int, char **), const char *syn = 0,
		const char *desc = 0)
    : synopsis(syn), description(desc) , fctGlobal(cmd) {
    fct=0;
  }
  //@}

  /** @name Destructor */
  //@{
  /// empty
  ~CommandOption()  {
    if(synopsis) delete [] ((char *)synopsis);
    if(description) delete [] ((char *)description);
  }
  //@}

  /// return object callback method
  int (TclObject::*getOptionName(void))(int, char **) { return fct; }
  /// return object callback method
  int (*getGlobalOptionName(void))(TclObject *,int, char **) { return fctGlobal; }
  /// return command option synopsis
  const char *getOptionSynopsis(void) { return synopsis; }
  /// return command option description
  const char *getOptionDescription(void) { return description; }
};



/** \class CommandInformation
    \ingroup Modules
    \brief Class for TCL command description */
class CommandInformation {
 protected:
  /// command name
  const char *name;
  /// command synopsis
  const char *synopsis;
  /// commandeffect description
  const char *description;
  /// command options hashtable (can be empty if globalCommand != 0)
  std::map < char *, CommandOption *, CStringCompare > *options;
  /// command callback function (can be 0 if options is not empty)
  int(*globalCommand)(TclModule *m, int argc, char **argv);
  /// command module
  TclModule *module;

 public:
  /** @name Constructors */
  //@{
  /** builds a new command calling back one global function
      @name m command's module
      @name name command's name
      @name fct command callback function
      @name _synopsis command synopsis
      @name _description command description string */
  CommandInformation(TclModule *m, const char *_name,
	  int(*fct)(TclModule *m, int argc, char **argv),
	  const char *_synopsis = 0, const char *_description = 0);

  /** builds a new command calling back one object method per option
      @name _name command's name
      @name _options command options hashtable
      @name _description command description string */
  CommandInformation(const char *_name,
	  std::map < char *, CommandOption *, CStringCompare > *_options,
	  const char *_description = 0);
  //@}

  /** @name Destructor */
  //@{
  /// empty
  ~CommandInformation() {}
  //@}

  /// returns command callback function (0 if none was set)
  int (*getCommand(void) const)(TclModule *m, int argc, char **argv) {
    return globalCommand;
  }
  /// returns command synopsis
  TclString *getSynopsis(void) const;
  const char *getDescription(void) const {
    return description;
  }
  /// return command options hashtable
  std::map < char *, CommandOption *, CStringCompare > *
    getOptions(void) const { return options; }
  
  
  /** execute command with given arguments
      @name argc arguments number
      @name argv arguments array */
  int exec(int argc, char **argv) const;

  /// TCL callback command
  static int genericCmd(ClientData cd, Tcl_Interp *interp,
			int argc, char **argv);
};


/** \class TclModule
    \ingroup Modules
    \brief Module description class */
class TclModule : public AutoLink<TclModule> {
public:
  /** @name Exception classes */
  //@{
/** \class DlopenFailed
    \ingroup Modules
    \brief dynamic library loading failed */
  class DlopenFailed : public Exception {
  public:
    DlopenFailed(const std::string &mname, const std::string &from);
  };

/** \class ModuleNotFound
    \ingroup Modules
    \brief named module could not be found */
  class ModuleNotFound : public Exception {
  public:
    ModuleNotFound(const std::string &mname, const std::string &from);
  };

/** \class ModuleAlreadyExists
    \ingroup Modules
    \brief attempt to load a module already loaded */
  class ModuleAlreadyExists : public Exception {
  public:
    ModuleAlreadyExists(const std::string &mname, const std::string &from);
  };

/** \class NoInitFunction
    \ingroup Modules
    \brief no initialization routine found for module */
  class NoInitFunction : public Exception {
  public:
    NoInitFunction(const std::string &mname, const std::string &fname,
		   const std::string &from);
  };

/** \class ShellNotFound
    \ingroup Modules
    \brief no TCL shell file found for module */
  class ShellNotFound : public Exception {
  public:
    ShellNotFound(const std::string& sname, const std::string& from);
  };
  //@}
public:
	/*
	/// a map that associate with a module name a static function that creates ObjectKinds
	static std::map< char *, INIT_MODULE_FUNCTION,CStringCompare> initMap;
	*/
	static std::map< char *, INIT_MODULE_FUNCTION,CStringCompare> &getModuleNameMap();
	/// name of TclModule
	static const char *tclModuleName;
protected:
  /// module TCL interpretor
  Tcl_Interp *interp;
  /// module name
  const TclString name;
  /// module dynamic library handle
  HANDLE_TYPE handle;
  /// module TCL script was loaded in the module init function
  bool embededTclScript;

  /// module TCL commands hashtable
  std::map< char *, CommandInformation *, CStringCompare > commands;
  
  /// object descriptors constructed by this module
  std::list<TclObjectKind *> descriptor;

  /// display function callback to display messages
  static DISPLAY_FUNCTION displayFct;
  /// TCL display function callback
  static TclString TCLdisplayCommandName;
  /// evaluation function callback to evaluate TCL commands
  static EVAL_FUNCTION evalFct;
  /// TCL evaluation function callback
  static TclString TCLevalCommandName;
  /// main application loop function
  static LOOP_FUNCTION loopFct;

public:
  /** @name Constructor */
  //@{
  /** builds a new module
      @name interp TCL interpretor to evaluate this module commands
      @name name module name
      @name handle dynamic library handle (null if module was not dynamicaly loaded) */
  TclModule(Tcl_Interp *interp, const TclString& name, HANDLE_TYPE handle = 0)
    throw(NoInitFunction);
  //@}

  /** @name Destructor */
  //@{
  /// deletes module and any attached TCL command
  ~TclModule();
  //@}

  /// returns module name
  const TclString& getName(void) const { return name; }
  /// returns module commands interpretor
  Tcl_Interp *getInterp(void) const { return interp; }
  /** returns module dynamic library handle (0 if module was statically
      linked) */
  HANDLE_TYPE getHandle(void) const { return handle; }
  /// States that the tcl script corresponding to the module (if any) has been loaded
  void tclScriptLoaded();
  /// display all module TCL commands name
  void displayFunctionList(void);
  /// returns module commands hashtable
  std::map< char *, CommandInformation *, CStringCompare >& getCommands(void) {
    return commands;
  }

  /// returns main application module
  TclModule *getMainModule(void) const { return coremodule; }

  /** builds a new Tcl command with a single callback function
      @param name command name
      @param fct command callback function
      @param synopsis command synopsis
      @param description command effect description */
  void newCommand(const char *name, int (*fct)(TclModule *, int, char **),
		  const char *synopsis = 0, const char *description = 0);

  /** registers a new Tcl command (not created here) in commands list
      @param name command name
      @param options command options hashtable
      @param description command effect description */
  void newCommand(const char *name,
		  std::map < char *, CommandOption *, CStringCompare > *options,
		  const char *description = 0);

  /** display a message calling display function
      @param str message to display */
  void display(const TclString& str) const {
    display((char *) str);
  }
  /** display a message calling display function
      @param str message to display */
  void display(const char *str) const;
  /** display a TCL command calling eval function
      @param str TCL command to evaluate */
  int eval(const TclString& str) const {
    return eval((char*) str);
  }
  /** display a TCL command calling eval function
      @param str TCL command to evaluate */
  int eval(const char *str) const;
  /** set a new display function
      @param fct display function pointer */
  static void setDisplayFunction(DISPLAY_FUNCTION fct) {
    displayFct = fct;
    TCLdisplayCommandName = EMPTY_STRING;
  }
  /** set a new TCL commands evaluation function
      @param fct evaluation function pointer */
  static void setEvalFunction(EVAL_FUNCTION fct) {
    evalFct = fct;
    TCLevalCommandName = EMPTY_STRING;
  }
  /** set a new display function
      @param name TCL display function name */
  static void setTCLDisplayFunction(const TclString& name) {
    displayFct = 0;
    TCLdisplayCommandName = name;
  }
  /** set a new TCL commands evaluation function
      @param name TCL evaluation function name */
  static void setTCLEvalFunction(const TclString& name) {
    evalFct = 0;
    TCLevalCommandName = name;
  }
  /// returns current display function
  static DISPLAY_FUNCTION getDisplayFunction(void) {
    return displayFct;
  }
  /// load the tcl shell associated with the module
  void loadShellScript();
  /// returns current evaluation function
  static EVAL_FUNCTION getEvalFunction(void) {
    return evalFct;
  }
  /// returns current TCL display function name
  static const TclString& getTCLDisplayFunction(void) {
    return TCLdisplayCommandName;
  }
  /// returns current TCL evaluation function name
  static const TclString& getTCLEvalFunction(void) {
    return TCLevalCommandName;
  }
  /** set main loop to handle TK events
      @param fct main loop function pointer (cannot be 0) */
  static void setMainLoop(LOOP_FUNCTION fct) {
    if(fct) loopFct = fct;
  }
  /// returns main loop function
  static LOOP_FUNCTION getMainLoop(void) {
    return loopFct;
  }

  /// reset result in module interpretor
  void resetResult(void) { Tcl_ResetResult(interp); }
 
  /** set command result in module interpretor
      @param str result */
  void setResult(const TclString& str) const {
    Tcl_SetResult(interp, (char *) str, TCL_VOLATILE);
  }

  /** set command result in module interpretor
      @param str result */
  void setResult(const char *str) const {
    Tcl_SetResult(interp, (char *) str, TCL_VOLATILE);
  }

  /** Format given string in a printf-like manner and set
      result in interp
      @name s printf-like format string
      @name ... arguments */
  void setResult(char *s...) const;

  /** Format given string in a printf-like manner and evaluates
      command in module interpretor.
      @name s printf-like format string
      @name ... arguments
      @return TCL_OK if the command successfully evaluated, TCL_ERROR otherwise */
  int tickle(const char *s... /* printf-like format string */) const;

  /** Evaluates given string in module interpretor.
      @param s string to evaluate
      @return TCL_OK if the command successfully evaluated, TCL_ERROR otherwise */
  int tickle(const TclString &s) const;

  /** Loads script file and evaluates it in module interpretor.
      File is looked for in YAV_SHELLS environnement variable,
      DefaultShellPath directory, and current directory.
      @param sname script file name
      @exception ShellNotFound script file was not found*/
  void loadSource(const TclString& sname) throw(ShellNotFound);

  /** test a TCL window existence
      @param wname TCL window name
      @return true if wname exists for this module, false otherwise */
  bool windowExists(char *wname) const {
    return ((wname != 0) && Tk_NameToWindow(interp, wname,
					       Tk_MainWindow(interp)))
      ? true : false;
  }

  /** test a TCL window existence
      @param wname TCL window name
      @return true if wname exists for this module, false otherwise */
  bool windowExists(const TclString& wname) const {
    return ((wname != EMPTY_STRING) && Tk_NameToWindow(interp, (char *) wname,
						       Tk_MainWindow(interp)))
      ? true : false;
  }

  /// adds a new object descriptor created for this module
  void newDescriptor(TclObjectKind *ok) { descriptor.push_back(ok); }

private:
  /** Tries to perform symbolic resolution given a function name.
      Attempt several C++ compilers mangling resolution for a function whose
      name is known and taking an only  TclModule * argument.
      @param  initFct function name
      @return pointer to the function body if resolution succedded, or 0 otherwise */
  void (*resolveSymbol(const TclString& initFct))(TclModule*);


public:

	typedef int (*MODULEMAINPROC)(TclModule *, int, char **);
  /** creates a new module 'name' by dynamically loading file libname.so
      from YAV_SHELLS, TclDefaultShellsPath or current directory
      @param interp module TCL interpretor
      @param name module name
      @exception DlopenFailed dlopen call failed to open dynamic library
      @exception ModuleNotFound libname.so dynamic library could not be found
      @exception ModuleAlreadyExists module was already loaded */
  static TclModule *loadNewModule(Tcl_Interp *interp, const TclString& name)
    throw(DlopenFailed, ModuleNotFound, ModuleAlreadyExists, NoInitFunction);

  /** Returns TCL shells search paths. This is the concatenation of
      .yavrc shells fields, YAV_SHELLS environment variable,
      and default shells paths.
      @param path returned paths separated by ':' */
  static void shellSearchPath(TclString& path);
  /** Returns modules search paths. This is the concatenation of
      .yavrc libraries fields, TCL_MODULES environment variable,
      LD_LIBRARY_PATH environment variable and default modules paths.
      @param path returned paths separated by ':' */
  static void moduleSearchPath(TclString& path);
  /** Looks for a dynamic library for module given by name.
      Library name is lib{MODULE_NAME}.so. Search path is
      TCL_MODULES environnment variable, LD_LIBRARY_PATH
      environnement variable and DefaultModulesPath
      @param name module name
      @return a pointer on a TclString containing the full library
      path name or 0 if the library could not be found.*/
  static TclString *findLibrary(const TclString& name);
  /** returns TclModule pointer on module given by its name or 0 if
      module could not be found
      @param name module name */
  static TclModule *findModuleByName(const TclString& name);
  /** initializes TCL, TK, core module, user modules and call application
      main procedure
      @param mainProc application main procedure
      @param verbose whether to print lots of informations...
      @param argc number of command line arguments
      @param argv command line arguments list: null terminated list of strings indicating module names
  */
  static int init(MODULEMAINPROC mainProc,
		  bool verbose, int argc, char **argv...);
  /** initializes core module for other modules management
      @param interp core module TCL interpretor
      @param name module name
      @return core module pointer */
  static TclModule *init(Tcl_Interp *interp, const TclString& name) {
    return new TclModule(interp, name);
  }
  /// infinite loop on TK events
  static void run(void) { while(true) (*loopFct)(1); }
  /** n-loop on TK events
      @param n number of iteration */
  static void run(unsigned int n) {
    (*loopFct)(n);
  }
  /** call any module cleanup procedure and exit
      @param ret shell return value */
  static void exit(int ret = 0);
  /** print a prompt in text interface */
  static void Prompt(Tcl_Interp *interp, int partial); 

  /** @name boolean management */
  //@{
  /// returns "true" if b is true and "false" otherwise
  static const char *boolString(bool b);

  /** set value to true and return true if str is "true",
      set value to false and return true if str is "false",
      return false otherwise
      @param str string
      @param value str boolean value (if str is "true" or "false")
      @return true if str is "true" or "false", false otherwise */
  static bool boolValue(char *str, bool& value);
  //@}

  /** @name TCL static commands */
  //@{
  /** default display function: sends message on cerr
      @param m calling module
      @param str message to display */
  static void DefaultDisplay(const TclModule &m, const char *str);
  /** default TCL commands evaluation function: calls Tcl_RecordAndEval
      @param m calling module
      @param str command to evaluate
      @return TCL_OK if command succesfully evaluated, TCL_ERROR otherwise */
  static int DefaultEval(const TclModule &m, const char *str);
  /** default main loop: handles TK events then activates active modules
      @param n number of loops */
  static void DefaultMainLoop(unsigned int n);

  /// TCL evaluate command
  static int Evaluate(TclModule *m, int argc, char **argv);
  /// TCL display command
  static int Display(TclModule *m, int argc, char **argv);
  /// TCL exit command
  static int Exit(TclModule *m, int argc, char **argv);
  /// TCL module command
  static int Module(TclModule *m, int argc, char **argv);
  /// TCL functions command
  static int Functions(TclModule *m, int argc, char **argv);
  /// TCL options command
  static int Options(TclModule *m, int argc, char **argv);
  /// TCL isModuelAvailable command
  static int IsModuleAvailable(TclModule *m, int argc, char **argv);
  /// TCL isModuleLoaded command
  static int IsModuleLoaded(TclModule *m, int argc, char **argv);
  /// TCL help command
  static int Help(TclModule *m, int argc, char **argv);
  /// TCL loadSource command
  static int LoadSource(TclModule *m, int argc, char **argv);
  /// TCL setDisplayFunction command
  static int SetDisplayMessageFunction(TclModule *m, int argc,
					   char **argv);
  /// TCL limited version command
  static int LimitedVersion(TclModule *m, int argc,
					   char **argv);
  /// TCL setEvalFunction command
  static int SetEvalFunction(TclModule *m, int argc, char **argv);

  /// allow to activate activable object inside a script
  static void activateWhenIdle(ClientData);
  /// TCL script command
  static int StartScript(TclModule *m, int argc, char **argv);
  /// TCL end command
  static int EndScript(TclModule *m, int argc, char **argv);
  /// TCL flushEvents command
  static int FlushEvents(TclModule *m, int argc, char **argv);
  /// TCL activateActivables command
  static int ActivateActivables(TclModule *m, int argc, char **argv);
  /// TCL mainLoop command
  static int MainLoop(TclModule *m, int argc, char **argv);
  /// TCL shellsPath command
  static int Shell(TclModule *m, int argc, char **argv);
  /// TCL xgrab command
  static int xgrabCmd(TclModule *m, int argc, char **argv);
  //@}

 private:
  /// the "true" and "false" string constants
  static const char *booleans[2];
  /// main application number of command line arguments
  static int initargc;
  /// main application command line arguments
  static char **initargv;
  /// number of modules
  static unsigned int nmodules;
  /// module names
  static char **modulenames;
  /// application core module
  static TclModule *coremodule;
  /// application main procedure
  static int (*mainProc)(TclModule *, int, char **);
  /// Tk_Main hook procedure
  static int main(Tcl_Interp *);

 public:
  /// default path for libraries
  static char *DefaultModulesPath;
  /// default path for TCL shells
  static char *DefaultShellsPath;
  /// default ressource file directory
  static char *DefaultRessourceFileLocation;
  /// prioritary path for libraries
  static char *RessourceModulesPath;
  /// prioritary path for TCL shells
  static char *RessourceShellsPath;
  /// verbose mode
  static bool verbose;
 /// if the modules should be loaded dynamically (dlopen) or statically
  static bool loadDynamically;
  /** Read ressource file.
      Ressource file is searched in:
        - if rfile == 0: ${HOME}/.yavrc
                         DEFAULT_RESSOURCE_FILE_LOCATION/.yavrc
	- if rfile != 0: rfile
                         ${HOME}/rfile
			 DEFAULT_RESSOURCE_FILE_LOCATION/rfile
      @param rfile ressource file name
      @param init returned initialization String
      @param nbModules returned number of modules
      @param modules returned module names array
      @param verbose
      @return true if a file was found and successfully loaded,
      false otherwise */
  static bool readrc(char *rfile, TclString& init,
					 unsigned int& nbModules, 
					 TclString *& modules, 
					 bool verbose = false);
};

/** the routine that initialize the tclModule TclModule */
void Tclmodule_Init(TclModule *m);

/** \class TclModuleInit
    \ingroup Modules
    \brief dummy class whose constructor updates the initMap hash table by inserting Tclmodule_Init with the tclModule keyword */
class TclModuleInit {
public:
	/** actually register Tclmodule_Init in the hash table */
	TclModuleInit() {
		TclModule::getModuleNameMap()[(char *)TclModule::tclModuleName]=Tclmodule_Init;
	}
};


} // namespace yav

#endif

