#ifndef TCLOBJECT_H
#define TCLOBJECT_H

#pragma interface "modules/TclObject.h"
#include <basics/ListElement.h>
#include <modules/TclModule.h>
#include <modules/TclObjectKind.h>
#include <tcl.h>

namespace yav {


/** \class TclObject
    \ingroup Modules
    \brief base class for any object that is linked with a TCL command */
class TclObject : public ListElement<TclObject> {
 protected:
  /// TCL object class descriptor
  TclObjectKind *objectKind;
  /// objects unique reference for that kind
  unsigned int ref;
  /// Nickname of the object
  //TclString nickname;

  /// builds a TCL command named 'objectName%u'
  void buildTCLCommand(void) const;

  /// deletes TCL command 'objectName%u'
  void deleteTCLCommand(void) const;

  /** method that should be called at the end of any
      TclObject constructor to call all WarnedObjects callback
      @param id built object type */
  void done(const std::type_info& id);

  /** method that should be called at the end of any
      TclObject destructor to call all WarnedObjects callback
      @param id built object type */
  void deleted(const std::type_info& id);

 public:
  /// command option pointer
  typedef int (TclObject::*OPTION)(int, char **);

  /// command option pointer
  typedef int (*GLOBAL_OPTION)(TclObject *,int, char **);

  /** @name Constructors */
  //@{
  /** Build a new TCL object. A TCL command name
      'objectName%u' where '%u' is object reference (available through
      getRef()) is created
      @param info object kind
      @param m module this object is created in */
  TclObject(const std::type_info& info, TclModule *m = 0);
  
  /** copy construction
      @param object object to copy */
  TclObject(const TclObject& object);
  //@}

  /** @name Destructor */
  //@{
  // destroy associated TCL command
  virtual ~TclObject();
  //@}
  /// -kind command option valid for all TclObject
  int kindCmd(int argc, char **argv);

  /// -realname command option valid for all TclObject
  int realNameCmd(int argc, char **argv);

  /** returns objects unique id in objects kind list */
  unsigned int getRef(void) const { return ref; }

  /** returns unique object identifier string
      @param objid returned name */
  virtual void name(TclString& objid) const;

  /** returns unique object identifier string
      @return object name */
  virtual TclString name(void) const;

  /// get command function associated to this object
  int (*getCommand(void))(TclObject *, Tcl_Interp*, int, char **) {
    return objectKind->getCommand();
  }

  /// set a new command function for this object
  void setCommand(int (*cmd)(TclObject *, Tcl_Interp*, int, char **)) {
    objectKind->setCommand(cmd);
  }

  /// get options map
  std::map < char *, CommandOption *, CStringCompare > *getOptions(void) {
    return &objectKind->options;
  }

  /**  returns an error message with all command options and input command
     @param argc command options number
      @param argv command options  */
  int error(int argc, char **argv);

  /** executes command
      @param interp TCL interpretor to execute command
      @param argc command options number
      @param argv command options
      @return TCL_OK if command succesfully evaluated, TCL_ERROR otherwise */
  int exec(Tcl_Interp *interp, int argc, char **argv);

  /// reset object module's result
  void resetResult(void) { objectKind->module->resetResult(); }

  /** set object module's result
      @param str result string */
  void setResult(const TclString& str) const {
    objectKind->module->setResult(str);
  }

  /** set object module's result
      @param str result string */
  void setResult(const char *str) const {
    objectKind->module->setResult(str);
  }

  /** set object module's result
      @param s printf-like format string */
  void setResult(char *s...) const;

  /** evaluates string in object module's TCL interpretor
      @param str string to evaluate
      @return command evaluation TCL return code */
  int tickle(const TclString& str) const {
    return objectKind->module->tickle(str);
  }

  /** formata string and evaluates it in object module's TCL interpretor
      @param s printf-like format string
  */
  int tickle(char *s...) const;

  /** returns object base name
      Tcl command name is 'getObjectName()getRef()' */
  const char *getObjectName(void) const { return objectKind->getName(); }

  /// returns class module
  TclModule *getModule(void) const { return objectKind->module; }

  /** builds a TclObject<T> objects list in module interpretor
      @param m module whose interpretor result string will be assingned
      @param argc command line number of arguments
      @param argv command line arguments
      @return TCL_OK */
  static int list(TclModule *m, int argc, char **argv);

  /** Default creation function for building a new T object from a TCL
      command line. It should be overloaded by T::create. It returns
      an error message in m indicating that it cannot create T.
      @param m module whose interpretor result string will be assingned
      @param argc number of command line arguments
      @param argv command line arguments
      @return TCL_OK */
  static int create(TclModule *m, int argc, char **argv);

  /// TCL creationsynopsis command
  int creationSynopsisCmd(TclModule *m, int argc, char **argv);
  /// TCL callback command
  static int genericCmd(ClientData cd, Tcl_Interp *interp,
			int argc, char **argv);

  /** returns first object of that kind */
  TclObject *getHead(void) const {
    return objectKind->objects.getFirst();
  }

  /** returns last object of that kind */
  TclObject *getTail(void) const {
    return objectKind->objects.getLast();
  }

  /** returns object of given kind
      @param type object type to retrieve */
  static TclObject *getHead(const std::type_info& type);

  /** returns last object of given kind
      @param type object type to retrieve */
  static TclObject *getTail(const std::type_info& type);

  /** returns number of object of that kind */
  unsigned int getCount(void) const {
    return objectKind->objects.getCount();
  }

  /// get this object class descriptor
  TclObjectKind *getObjectKind(void) const { return objectKind; }

 protected:
  /// constructors common code
  void init(void);

  /** set a new base object type for derivating objects
      @param info new object type
      @param m module for new object kind */
  void setObjectKind(const std::type_info& info, TclModule *m = 0);

  friend class TclObjectKind;
};


} // namespace yav

#endif      // ifndef TCLOBJECT_H
