#ifndef TCLOBJECTKIND_H
#define TCLOBJECTKIND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "modules/TclObjectKind.h"
#include <basics/Exception.h>
#include <typeinfo>
#include <list>
#include <basics/AutoLink.h>
#include <basics/ListElement.h>
#include <modules/TclModule.h>


namespace yav {


  /** \class WarnedObject
      \ingroup Modules
      \brief an object that want to be called back if a given id object is created or deleted */
  class WarnedObject {
  protected:
    /// that object kind id
    std::list<const std::type_info *> type;
    
  public:
    /** @name constructors */
    //@{
    /** warn this object when an object of typeid t1 is created or deleted
	@param t1 typeid of objects to watch */
    WarnedObject(const std::type_info& t1);
    /** warn this object when an object of typeid t1 or t2 is
	created or deleted
	@param t1 typeid of objects to watch
	@param t2 typeid of objects to watch */
    WarnedObject(const std::type_info& t1, const std::type_info& t2);
    /** warn this object when an object of typeid t1, t2 or t3 is
	created or deleted
	@param t1 typeid of objects to watch
	@param t2 typeid of objects to watch
	@param t3 typeid of objects to watch */
    WarnedObject(const std::type_info& t1, const std::type_info& t2,
		 const std::type_info& t3);
    //@}
    /// destructor
    virtual ~WarnedObject();

    /** adds an object type to watch
	@param t type to watch */
    void warnMe(const std::type_info& t);
    /** removes an object type to watch
	@param t type to remove */
    void doNotWarnMe(const std::type_info& t);

    /// gets empty callback function
    virtual void emptyObjectsList(void) = 0;
    /** new object callback function
	@param obj created object
	@param id created object real type */
    virtual void createdObject(TclObject *obj, const std::type_info& id) = 0;
    /** deleted object callback function
	@param obj object being deleted
	@param id created object real type */
    virtual void deletedObject(TclObject *obj, const std::type_info& id) = 0;
  };
  
  class TclObject;
  
  
/** \class TclObjectKind
    \ingroup Modules
    \brief A TCL object descriptor */
  class TclObjectKind : public AutoLink< TclObjectKind >,
    public EmptyObjectsList<TclObject> 
    {
      
    protected:
      /// object name
      TclString name;
      /// object real name i.e. name that can be unsderstood by the user
      TclString realName;
      /// pointer on a creation function
      int (*createFromArgs)(TclModule *, int, char **);
      /// pointer on a function returning a list of this kind objects
      int (*objectsList)(TclModule *, int, char **);
      /// TCL global command (if any)
      int (*globalCommand)(TclObject *, Tcl_Interp*, int, char **);
      /// hashtable with (char *) keys and (int (T::*)(int, char **)) values
      std::map < char *, CommandOption *, CStringCompare > options;
      /// class module
      TclModule *module;
      /// objects of that kind list
      ObjectsList<TclObject> objects;
      /// that object kind id
      const std::type_info& objectsType;
      /// objects to warn when list get empty
      std::list<WarnedObject *> warn;
      /// new operator command synopsis
      TclString newSynopsis;
      /// the list of object kinds
      std::list<const char *> childList;
      
      /** add a TclObject to objects list
	  @param obj object whose type should be typeid */
      void operator+=(TclObject *obj);
      
      /** removes a TclObject from objects list
	  @param obj object whose type should be typeid */
      void operator-=(TclObject *obj) throw(ObjectsList<TclObject>::NotFound);

      /** an object of that kind is being created
	  @param obj object being created */
      void created(TclObject *obj);
      /** an object of that kind is being deleted
	  @param obj object being deleted */
      void deleted(TclObject *obj);
      /** add the name of a child object kind in the list of childs */
      void addChild(const char *childName) {
	childList.push_back(childName);
      }
      /** retrieve a TclObject given its name . First it tests if this name corresponds to one of its TclObject.
	  If not, it recursively call the same function on its child object kind.
	  This is a recursive search procedure which is depth first */
      TclObject *getChildObject(char *objectName); 
    public:
      /** @name exception class */
      //@{
      /** \class InvalidParameters
          \ingroup Modules
      */
      class InvalidParameters : public Exception {
      public:
	InvalidParameters(const std::string& from)
	  : Exception(from,
		      "no argument may be 0") {}
      };
      //@}
      
      /** @name constructor */
      //@{
      /** creates a new object kind
	  @param module module creating this ObjectKind
	  @param type typeid describing objects of that kind dynamic type
	  @param name kind name
	  @param realName kind real name ie that can be understood by the user
	  @param synopsis new command synopsis
	  @param createFromArgs pointer on an object creation function
	  @param objectsList pointer on a function returning a list of this kind objects
	  @exception InvalidParameter no argument may be 0 */
      TclObjectKind(TclModule *module,
		    const std::type_info& type,
		    const char *name,
		    const char *realName,
		    const char *synopsis,
		    int (*createFromArgs)(TclModule *, int, char **),
		    int (*objectsList)(TclModule *, int, char **))
	throw(InvalidParameters);

      /** creates a new object kind
	  @param module module creating this ObjectKind
	  @param type typeid describing objects of that kind dynamic type
	  @param motherType typeid of a mother class whose option commands will be added 
	  @param name kind name
	  @param realName kind real name ie that can be understood by the user
	  @param synopsis new command synopsis
	  @param createFromArgs pointer on an object creation function
	  @param objectsList pointer on a function returning a list of this kind objects
	  @exception InvalidParameter no argument may be 0 */
      TclObjectKind(TclModule *module,
		    const std::type_info& type,
		    const std::type_info& motherType,
		    const char *name,
		    const char *realName,
		    const char *synopsis,
		    int (*createFromArgs)(TclModule *, int, char **),
		    int (*objectsList)(TclModule *, int, char **))
	throw(InvalidParameters);


      /** creates a new object kind
	  @param module module creating this ObjectKind
	  @param type typeid describing objects of that kind dynamic type
	  @param name kind name
	  @param realName kind real name ie that can be understood by the user
	  @param createFromArgs pointer on an object creation function
	  @param objectsList pointer on a function returning a list of this kind objects
	  @param synopsis new command synopsis
	  @exception InvalidParameter no argument may be 0 */
      TclObjectKind(TclModule *module,
		    const std::type_info& type,
		    const char *name,
		    const char *realName,
		    int (*createFromArgs)(TclModule *, int, char **),
		    int (*objectsList)(TclModule *, int, char **),
		    const char *synopsis = 0)
	throw(InvalidParameters);
      //@}
      
      /** @name destructor */
      //@{
      /// removes any warn and objects of that kind
      ~TclObjectKind();
      //@}
      
      /// delete any object of that kind
      void deleteAllObjects(void);

      /// returns item name
      const char *getName(void) const { return name; }
      /// returns item real name
      const char *getRealName(void) const { return realName; }
      
      /** adds a new option to this command
	  @param name option name
	  @param cmd command option callback method defined as a member fonction 
	  @param synopsis option synopsis
	  @param description option description */
      void newOption(char *name, int (TclObject::*cmd)(int argc, char **argv),
		     const char *synopsis = 0, const char *description = 0);

      /** adds a new option to this command
	  @param name option name
	  @param opt CommandOption pointer filled with the right values */
      void newOption(char *name, CommandOption *opt);

      /** adds a new option to this command
	  @param name option name
	  @param cmd command option callback method defined as a static method
	  @param synopsis option synopsis
	  @param description option description */
      void newOption(char *name, int (*cmd)(TclObject *,int argc, char **argv),
		     const char *synopsis = 0, const char *description = 0);
      
      /// returns module
      TclModule *getModule(void) const { return module; }
      /// set new module
      void setModule(TclModule *m) { module = m; }
      
      /// get options map
      std::map < char *, CommandOption *, CStringCompare > *getOptions(void) {
	return &options;
      }
      
      /// returns ref of next object to be created
      unsigned int getNextRef(void) const;

      /// get command function associated to this object
      int (*getCommand(void))(TclObject *, Tcl_Interp*, int, char **) {
	return globalCommand;
      }
      
      /// set a new command function for this object
      void setCommand(int (*cmd)(TclObject *, Tcl_Interp*, int, char **)) {
	globalCommand = cmd;
      }
      
      /// set new command synopsis
      void setNewSynopsis(const char *synopsis) { newSynopsis = synopsis; }
      
      /// get synopsis
      const char * getCreationSynopsis(void) const { return newSynopsis; }
      /// returns a string that details the possible options
	  TclString getOptionSynopsis(void) const;
      /// returns iterator on first object
      ObjectsList<TclObject>::iterator firstObject(void) const {
	return (objects.begin());
      }
      /// returns iterator on last object
      ObjectsList<TclObject>::iterator lastObject(void) const {
	return (objects.end());
      }
      /// returns objects list first element
      TclObject *getFirstObject(void) const { return objects.getFirst(); }
      /// returns objects list last element
      TclObject *getLastObject(void) const { return objects.getLast(); }
      
      /** adds 'me' to list of object to warn when list is empty
	  @param me object to warn */
      void warnMe(WarnedObject *me) { warn.push_back(me); }
      /** removes 'me' from list of object to warn when list is empty
	  @param me object to warn */
      void doNotWarnMe(WarnedObject *me) { warn.remove(me); }
      
      /// objects list gets empty callback function
      virtual void empty(void) const;
      
      /** returns object with reference n (or 0 if it does not exist)
	  @param type object type
	  @param n object reference */
      static TclObject *getObject(const std::type_info &type, unsigned int n);
      
      /** returns object (or 0 if it does not exist)
	  @param name object name */
      static TclObject *getObject(TclString name);
      
      /** returns object whose name is in argv[i] (or 0 if it does no
	  exist). If module is not 0, a message is displayed in TCL
	  interpretor
	  @param type object type
	  @param argc number of arguments
      @param argv aruments
      @param i argument index supposed to name a TCL object */
      static TclObject *getObject(const std::type_info &type,
				  int argc, char **argv, int i = 1);
      
      /** returns given type object name
	  @param type wanted object type */
      static const char *getObjectName(const std::type_info &type);
      
      /** retrieve TclObjectKind object for objects with given type */
      static TclObjectKind *findByKind(const std::type_info& type);
      /** retrieve TclObjectKind object for objects with given name */
      static TclObjectKind *findByName(const char *name);
      
      /// TCL new command
      static int newCmd(TclModule *m, int argc, char **argv);
      /// TCL which command
      static int whichCmd(TclModule *m, int argc, char **argv);
       
      /** adds options from TclObject class whose type is given to this class
	  @param src source class */
      void addOptions(const std::type_info& src);
      
      friend class TclObject;
    };
  
  

} // namespace yav

#include <modules/TclObject.h>

#endif     // ifndef TCLOBJECTKIND_H
