#ifndef TCLSUBOBJECT_H
#define TCLSUBOBJECT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 4355)
#endif
#pragma interface "modules/TclSubObject.h"
#include <map>
#include <modules/TclModule.h>
#include <modules/TclObjectKind.h>


namespace yav {


template <class CLIENT> class TclSubObject;

/** \class CommandSubOption
    \ingroup Modules
    \brief sub object command option structure */
template <class CLIENT>
class CommandSubOption {
 private:
  /// command option synopsis
  const char *synopsis;
  /// command option effect description
  const char *description;
  /// object method associated to command
  int (TclSubObject<CLIENT>::*fct)(CLIENT *, int, char **);

 public:
  /** @name Constructor */
  //@{  
  /** buids a command option descriptor from its synopsis, description strings and object callback method
      @name cmd object callback method
      @name syn command option synopsis
      @name desc command option description */
  CommandSubOption(int (TclSubObject<CLIENT>::*cmd)(CLIENT *, int, char **),
		   const char *syn = 0, const char *desc = 0)
    : synopsis(syn), description(desc) , fct(cmd) {}
  //@}

  /** @name Destructor */
  //@{
  /// empty
  ~CommandSubOption()  {
    if(synopsis) delete[] synopsis;
    if(description) delete[] description;
  }
  //@}

  /// return object callback method
  int (TclSubObject<CLIENT>::*getOption(void))(CLIENT *, int, char **) {
    return fct;
  }
  /// return command option synopsis
  const char *getOptionSynopsis(void) { return synopsis; }
  /// return command option description
  const char *getOptionDescription(void) { return description; }
};

/** \class TclSubObject
    \ingroup Modules
    \brief sub object interface parent class */
template <class CLIENT>
class TclSubObject {
 protected:
  /// sub object commands list
  static std::map<char *, CommandSubOption<CLIENT> *, CStringCompare> suboptions;

 public:
  /// empty constructor
  TclSubObject() {
  }
  /// empty destructor
  virtual ~TclSubObject() {}

  /// sub object commands prototype
  typedef int (TclSubObject<CLIENT>::*SUBOPTION)(CLIENT *,
						 int argc, char **argv);

  /** adds a new option to this command
      @param name option name
      @param cmd command option callback method
      @param synopsis option synopsis
      @param description option description */
  static void newSubOption(char *name,
			   int (TclSubObject<CLIENT>::*cmd)(CLIENT *,
							    int argc,
							    char **argv),
			   const char *synopsis = 0,
			   const char *description = 0) {
    suboptions[name] = new CommandSubOption<CLIENT> (cmd, synopsis,
						     description);
  }


  /** executes command
      @param ptr subobject to apply function on
      @param argc command options number
      @param argv command options
      @return TCL_OK if command succesfully evaluated, TCL_ERROR otherwise */
  int exec(CLIENT *ptr, int argc, char **argv) {
    typename std::map< char *, CommandSubOption<CLIENT> *, CStringCompare >::iterator
      it = suboptions.find(argv[0]);

    if(it != suboptions.end()) {
      int (TclSubObject<CLIENT>::*cmd)(CLIENT *, int, char **) =
	it->second->getOption();
      int res = (this->*cmd)(ptr, argc - 1, argv + 1);
      if(res == COMMAND_SYNOPSIS) {
	TclString error;
	error.sprintf("usage: %s%u %s %s",
		      getObjectKind()->getName(), getRef(),
		      argv[-1], argv[0]);
	if(it->second->getOptionSynopsis())
	  error += SPACE_STRING + it->second->getOptionSynopsis();
	getObjectKind()->getModule()->setResult(error);
	return TCL_ERROR;
      }
      return res;
    }
    else {
      TclString res;
      typename std::map< char *, CommandSubOption<CLIENT> *, CStringCompare >::iterator
	it;

      res.sprintf("usage: %s%u %s < ",
		  getObjectKind()->getName(), getRef(), argv[-1]);
      for(it = suboptions.begin(); it != suboptions.end(); it++) {
	res += it->first;
	res += " | ";
      }
      res = TclString(res, 0, res.getLength() - 3);
      res += ">";
      
      getObjectKind()->getModule()->setResult(res);
      return TCL_ERROR;
    }
  }

  /// returns main object object kind
  virtual TclObjectKind *getObjectKind(void) const = 0;
  /// returns main object reference
  virtual unsigned int getRef(void) const = 0;
};

  template < class CLIENT > std::map<char *, CommandSubOption<CLIENT> *, CStringCompare>  TclSubObject<CLIENT>::suboptions =  std::map<char *, CommandSubOption<CLIENT> *, CStringCompare>();

}


#endif  // ifndef TCLSUBOBJECT_H

