#ifdef USE_PETSC
#include "petsc.h"
#endif
#include <stdlib.h>
#include <string.h>
#if ( __GNUC__ < 3 ) &&  (!defined(WIN32))
#include <iostream.h>
#include <fstream.h>
#else
#include <iostream>
#include <fstream>
#endif
#include <tcl.h>
#include <tk.h>
#include <modules/TclModule.h>

#include <linearAlgebra/GenericSparseMatrix.h>

#ifdef EMBEDED_TCL_SCRIPTS
#include "yav++.tcl.h"
#endif

using namespace yav;

#ifndef DefaultRessourcesFile
#define DefaultRessourcesFile "/user/epidaure/.yavrc"
#endif


static void readStdin(ClientData, int);
static void Prompt(Tcl_Interp *, int);
static TclString command, initrc = EMPTY_STRING;
Tcl_Interp *globinterp;
static TclString *modules = 0;
static unsigned int nbModules = 0;


static int letsgo(TclModule *core, int argc, char **argv);
static void usage(char *name);

static char releaseInfo[] = "\
\n\
                        Yet Another Viewer\n\
\n\
A script interpretor for volume images processing and deformable models.\n\
         (in fact an extensible C++ front-end for TCL/TK)\n\
\n\
                        Version 1.4 (beta) \n\
\n\
authors:\n\
  Johan Montagnat (Johan.Montagnat@sophia.inria.fr)\n\
  Herv�Delingette (Herve.Delingette@sophia.inria.fr)\n\
contributors:\n\
  Clement Forest\n\
  Celine Fouard\n\
  Sebastien Granger\n\
  Alexandre Guimond\n\
  Jean-Didier Lemarechal\n\
  Jean-Christophe Lombardo\n\
  Gregoire Malandain\n\
  Guillaume Picinbono\n\
  Alain Pitiot\n\
  Nicolas Scapel\n\
  Maxime Sermesant\n\
  Jonathan Stoeckel\n\
\n\
  INRIA Sophia-Antipolis, Epidaure project\n\
  2004 route des lucioles, BP93, 06902 Sophia-Antipolis, France\n\
  http://www.inria.fr/epidaure/Epidaure-eng.html\n\
\n\
Using TCL/TK 8.1\n\
  http://www.scriptics.com/software/8.1.html\n\
\n\
OpenGL TK widget inspired from Togl v1.4, a Tk OpenGL widget\n\
  Brian Paul and Ben Bederson\n\
  http://www.cs.umd.edu/~bederson/Togl.html\n\
\n";



/* Main: let's go */
int main(int argc, char **argv) {
  /* initialize interface */
  initializeLinearAlgebraLibrary( &argc, &argv );
  try {
    char *rfile = ".yavrc";
    bool rcopt = false, verbose = false;

    if(argc > 1 && !strcmp(argv[1], "-help")) usage(argv[0]);
    for(int i = 1; i < argc; i++) {
      if(!strcmp(argv[i], "-rc")) {
	rcopt = true;
	if(++i >= argc) rfile = 0;
	else {
	  if(argv[i][0] == '-') rfile = 0;
	  else rfile = argv[i];
	}
      }
      else if(!strcmp(argv[i], "-verbose")) {
	verbose = true;
      }
      else if(!strcmp(argv[i], "-version")) {
	COUT << releaseInfo << ENDL;
      }
    }

    if(!rcopt || rfile) {
	  TclModule::readrc(rfile, initrc, nbModules, modules, verbose);
	}

    TclModule::init(letsgo, verbose, argc, argv, 0);

  }
  catch(...) {
    CERR << "yav: fatal: Unexpected exception caught" << ENDL;
    TclModule::exit(-1);
  }
  finalizeLinearAlgebraLibrary();
  return 0;
}


void usage(char *name) {
  fprintf(stderr, "usage: %s [script_name|[-text|-graphic]] [-init %%s] [-help]\n\
         Launch yav++ (Yet Another Viewer) that interprets tcl commands\n\
         -text|-graphic : select either text or graphic (tk) interface\n\
         script_name    : evaluate given tcl script\n\
         -init %%s       : evaluate given string before anything else\n\
         -quiet         : do not print out any message (exept critical)\n\
         -static        : load all modules statically \n\
         -verbose       : print out lot of message (libraries search paths...)\n\
         -load %%s...    : load given modules\n\
         -rc [%%s]       : load given ressource file (none if no ressource file\n\
                          name is given). By default, $HOME/.yavrc file is read.\n\
         -version       : print out release informations\n",
	  name);
  exit(0);
}




/* Starting point called by TK */
int letsgo(TclModule *core, int argc, char **argv) {
  Tcl_Channel inChannel;
  int i, nscript=0;
  int ninit = 0;
  enum {
    NONE,
    TEXT,
    GRAPHIC,
    SCRIPT
  } interactive = NONE;
  bool quiet = false;
  bool verbose = false;
  bool dynamic = true;


  
#if TK_MAJOR_VERSION>=8 && TK_MINOR_VERSION>=5
  core->eval ("::ttk::setTheme clam");
#endif //TK_MAJOR_VERSION>=8 && TK_MINOR_VERSION>=5
  
  
  for(i = 1; i < argc; i++) {
    if(!strcmp(argv[i], "-quiet")) quiet = true;
    else if(!strcmp(argv[i], "-verbose")) verbose = true;
    else if(!strcmp(argv[i], "-static")) dynamic = false;
  }

  if(nbModules > 0 && !quiet && !verbose) {
    COUT << argv[0] << ": loading: ";
    COUT.flush();
  }
  for(i = 0; i < (int) nbModules; i++) {
    try {
      if(verbose)
	COUT << "-------------------------------------------------------" << ENDL
	     << "Adding module " << modules[i] << "..." << ENDL;
      if (dynamic)
	(void) core->loadNewModule(core->getInterp(), modules[i]);
      else
	(void) new TclModule(core->getInterp(), modules[i]);
      if(!quiet && !verbose) {
	COUT << modules[i] << " ";
	COUT.flush();
      }
      if(verbose) COUT << "... module " << modules[i]
		       << " successfully added." << ENDL;
    }

    catch(TclModule::ModuleAlreadyExists&) {
      if(!quiet) {
	COUT << ENDL << "yav++: warning: attempt to load module "
	     << modules[i] << " twice" << ENDL;
      }
    }
    catch(Exception& e) {
      CERR << ENDL << e << ENDL;
      exit(-1);
    }
  }
  if(nbModules > 0 && !quiet && !verbose) COUT << "done" << ENDL;
  if(verbose)
    COUT << "modules loaded" << ENDL
	 << "-------------------------------------------------------" << ENDL;
  if(modules) {
    delete[] modules;
    nbModules = 0;
  }

  for(i = 1; i < argc; i++) {
    if(strlen(argv[i]) > 0) {
      if(argv[i][0] == '-') {
	if(!strcmp(argv[i], "-text")) {
	  // no buffer on stdout, to allow interactive pipe (with Matlab)
	  setvbuf(stdout,0,_IONBF,0);
	  if(interactive == NONE) interactive = TEXT;
	  else usage(argv[0]);
	}
	else if(!strcmp(argv[i], "-graphic")) {
	  if(interactive == NONE) interactive = GRAPHIC;
	  else usage(argv[0]);
	}
	else if(!strcmp(argv[i], "-init")) {
	  if(++i < argc && argv[i][0] != '-')
	    ninit = i;
	}
	else if(!strcmp(argv[i], "-load")) {
	  if(++i < argc) {
	    int j;
	    for(j = i; j < argc && argv[j][0] != '-'; j++) {
	      try {
		(void) core->loadNewModule(core->getInterp(), argv[j]);
	      }

	catch(TclModule::ModuleAlreadyExists&) {
		if(!quiet) {
		  COUT << ENDL << "yav++: warning: attempt to load module "
		       << argv[j] << " twice" << ENDL;
		}
	      }
	      catch(Exception& e) {
		CERR << ENDL << e << ENDL;
		exit(-1);
	      }
	    }
	    i = j - 1;
	  }
	}
	else if(!strcmp(argv[i], "-quiet")) ;
	else if(!strcmp(argv[i], "-static")) ;
	else if(!strcmp(argv[i], "-verbose")) ;
	else if(!strcmp(argv[i], "-version")) ;
	else if(!strcmp(argv[i], "-rc")) {
	  if(i + 1 < argc && argv[i+1][0] != '-') i++;
	}
	else {
	  fprintf(stderr, "yav: error: invalid option \'%s\'\n", argv[i]);
	  TclModule::exit(-1);
	}
      }
      else {
	if(interactive == NONE) {
	  nscript = i;
	  interactive = SCRIPT;
	}
	else usage(argv[0]);
      }
    }
  }

  if(interactive == NONE) interactive = GRAPHIC;
  globinterp = core->getInterp();


#ifdef EMBEDED_TCL_SCRIPTS
  /* do not load tcl scripts dynamically anymore */
  Tcl_Eval(core->getInterp(), yav___tcl);
#else
  try {
    core->loadSource("yav++.tcl");
  }
  catch(TclModule::ShellNotFound&) {
    CERR << "Could not find yav++.tcl or tclModules.tcl" << ENDL
	 << "Set YAV_SHELLS variable to proper value." << ENDL;
    TclModule::exit(-1);
  }
#endif

  /* launch graphic interface */
  if(interactive == GRAPHIC) core->tickle("yav::interface");

  if(initrc != EMPTY_STRING) {
    if(core->tickle((char *) initrc) == TCL_ERROR) {
      CERR << argv[0] << " warning: error while executing .yavrc init command."
	   << ENDL;
    }
  }

  /* handle initialization command */
  if(ninit > 0) {
    TclString str;
    str.sprintf("modules::evaluate \"%s\"",argv[ninit]);
    core->tickle(str);
  }
  core->tickle("wm withdraw .");

  /* start yav */
  switch(interactive) {

  case NONE:
    break;

  case GRAPHIC:
    TclModule::run();
    break;

  case TEXT:
    Tcl_SetVar(globinterp, "tcl_interactive", "1", TCL_GLOBAL_ONLY);
    inChannel = Tcl_GetStdChannel(TCL_STDIN);
    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, readStdin,
			     (ClientData) inChannel);
    Prompt(globinterp, 0);
    Tcl_ResetResult(globinterp);

    while(1) Tcl_DoOneEvent(0);

  case SCRIPT:
    core->tickle("script %s", argv[nscript]);
    TclModule::exit(0);
    break;
  }

  return TCL_OK;
}





/* read stdin in text interface */
static void readStdin(ClientData clientData, int mask) {
  static int partial = 0;
  Tcl_Channel outChannel, inChannel = (Tcl_Channel) clientData;
  Tcl_DString line;

  Tcl_DStringInit(&line);

  if(Tcl_Gets(inChannel, &line) < 0) {
    outChannel = Tcl_GetChannel(globinterp, "stdout", NULL);
    Tcl_Write(outChannel, "\n", 1);
    Tcl_Exit(0);
    return;
  }

  
  command += Tcl_DStringValue(&line);
  TclString cmd = command + "\n";
  Tcl_DStringFree(&line);
    
  if(!Tcl_CommandComplete((char *) cmd)) partial = 1;
  else {
    partial = 0;
    Tcl_CreateChannelHandler(inChannel, 0, readStdin, (ClientData) inChannel);
    Tcl_DStringInit(&line);
    switch(Tcl_RecordAndEval(globinterp, (char *) cmd, TCL_EVAL_GLOBAL)) {
    case TCL_OK:
      break;
    case TCL_ERROR:
      Tcl_DStringAppend(&line, "*Error* ", -1);
      break;
    case TCL_RETURN:
      Tcl_DStringAppend(&line, "*Return* ", -1);
      break;
    case TCL_BREAK:
      Tcl_DStringAppend(&line, "*Break* ", -1);
      break;
    case TCL_CONTINUE:
      Tcl_DStringAppend(&line, "*Continue* ", -1);
      break;
    }
    Tcl_DStringAppend(&line, globinterp->result, -1);
    if(*Tcl_DStringValue(&line)) puts(Tcl_DStringValue(&line));
    Tcl_DStringFree(&line);
    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, readStdin,
			     (ClientData) inChannel);
    command = "";
  }

  Prompt(globinterp, partial);
  Tcl_ResetResult(globinterp);
}



/* print a prompt in text interface */
static void Prompt(Tcl_Interp *interp, int partial) {
  Tcl_Channel outChannel = Tcl_GetChannel(interp, "stdout", NULL);

  if(outChannel) {
    if(partial) Tcl_Write(outChannel, "...> ", 5);
    else Tcl_Write(outChannel, "> ", 2);
    Tcl_Flush(outChannel);
  }
}
