/* epi_morpho.c 2.1 5/14/98 */

#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>

#ifndef NO_PROTO
static void E_LogicDilCond( char *tab1, char *tab2, char *tab3, int taille );
static void E_LogicEroCond( char *tab1, char *tab3, int taille );

static int E_DilationSingle( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par );
static int E_ErosionSingle( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par );
static int ErosionBin( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par );
static int ErosionNdg( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par );
static int DilationBin( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par );
static int DilationNdg( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par );
#else
static void E_LogicDilCond();
static void E_LogicEroCond();
static int E_DilationSingle();
static int E_ErosionSingle();
static int ErosionBin();
static int ErosionNdg();
static int DilationBin();
static int DilationNdg();
#endif

/* Morphologie : valeurs par defaut pour la structure de parametres.

   L'operation a realiser (champ operation de la structure) est
   une dilatation (DILATION).
   Le nombre d'iterations a faire (champ order de la structure)
   est mis a 1.
   L'element structurant (champ connexite de la structure) est
   le 26-voisinage (CONNEXE_26).
   La dimension du traitement (champ mode de la structure) est
   3D (MODE_3D).
*/

#if defined(_ANSI_)
void E_InitMorphology( e_ParamMorphoPtr par /* parameters */ )
#else
void E_InitMorphology( par )
e_ParamMorphoPtr par; /* parameters */
#endif
{
	/*------------------------------------------------------------*/
	/*   set the default parameters for morphology operations     */
	/*------------------------------------------------------------*/
	par->operation = DILATION;
	par->order = 1;
	par->connexite = CONNEXE_26;
	par->mode = MODE_3D;
	par->binaire = FALSE;
	par->scnd_op = UNKNOWN;
}

/* Dilatation.

   Dans le cas des images binaires, la dilatation se definit
   comme le lieu des points ou l'element structurant heurte 
   l'objet. 
   Dans le cas des images en niveaux de gris, la valeur d'un
   point apres dilatation est egale a la valeur maximale 
   rencontree dans l'element structurant.

   Les elements structurants implementes sont des voisinages
   elementaires, specifies par le champ connexite de la structure
   de parametre :
   - CONNEXE_26 : 26-voisinage, c'est-a-dire les points partageant
     un sommet, une arete ou une face avec le point central.
   - CONNEXE_6 : 6-voisinage, c'est-a-dire les points partageant
     une face avec le point central.
   - CONNEXE_10 : 10-voisinage, c'est-a-dire les points partageant
     une arete ou une face avec le point central dans le plan, et
     les points partageant une face avec le point central dans les
     plans superieurs et inferieurs.

   L'operation peut etre limitee au plan grace au champ mode de
   la structure de parametre :
   - MODE_2D : on ne considere que la restriction au plan de 
     l'element strcturant
   - MODE_3D : on considere l'element structurant 3D

   Le nombre d'iterations a realiser avec un element structurant
   est donne par le champ order de la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_Dilation( t_ImagePtr theIm    /* input image */, 
	       t_ImagePtr resIm     /* result image */, 
	       e_ParamMorphoPtr par /* parameters */ )
#else
int E_Dilation( theIm, resIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	t_Image outIm;
	int i;
	char *buffer;
	
        E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_Dilation", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_Dilation", "output image not allocated");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_Dilation", 
               "input and output have different dimensions");
	E_ERR( (par->order < 1), E_ERRFATAL, FALSE, "E_Dilation", 
	     "error in parameters");
	E_ERR( ((par->connexite!=CONNEXE_26) && (par->connexite!=CONNEXE_10) && (par->connexite!=CONNEXE_6)), E_ERRFATAL,
	     FALSE, "E_Dilation", "error in parameters");


	/* first case */
	/**************/
	if ( (par->order == 1) && (theIm->buf != resIm->buf) ) {
		E_DilationSingle( theIm, resIm, par );
		return ( TRUE );
	}
	/* second case */
	/***************/
	if ( (par->order == 1) && (theIm->buf == resIm->buf) ) {
		T_CopyHeaderIm( resIm, &outIm );
		T_AllocImage( &outIm );
		E_ISERR(FALSE,"E_Dilation",(char*)NULL);
		E_DilationSingle( theIm, &outIm, par );
		buffer = theIm->buf;
		theIm->buf = resIm->buf = outIm.buf;
		outIm.buf = buffer;
		T_FreeImage( &outIm );
		return( TRUE );
	}
	/* third and fourth cases */
	/* par->order > 1         */
	/**************************/
	T_CopyHeaderIm( resIm, &outIm );
	T_AllocImage( &outIm );
	E_ISERR(FALSE,"E_Dilation",(char*)NULL);

	E_DilationSingle( theIm, &outIm, par );
	for (i=0; i<par->order-2; i++) {
		E_DilationSingle( &outIm, resIm, par );
		buffer = outIm.buf;
		outIm.buf = resIm->buf;
		resIm->buf = buffer;
	}
	E_DilationSingle( &outIm, resIm, par );
	T_FreeImage( &outIm );
	return( TRUE );
}

/* Erosion.

   Dans le cas des images binaires, l'erosion se definit
   comme le lieu des points ou l'element structurant est 
   inclus dans l'objet. 
   Dans le cas des images en niveaux de gris, la valeur d'un
   point apres erosion est egale a la valeur minimale rencontree
   dans l'element structurant.

   Les elements structurants implementes sont des voisinages
   elementaires, specifies par le champ connexite de la structure
   de parametre :
   - CONNEXE_26 : 26-voisinage, c'est-a-dire les points partageant
     un sommet, une arete ou une face avec le point central.
   - CONNEXE_6 : 6-voisinage, c'est-a-dire les points partageant
     une face avec le point central.
   - CONNEXE_10 : 10-voisinage, c'est-a-dire les points partageant
     une arete ou une face avec le point central dans le plan, et
     les points partageant une face avec le point central dans les
     plans superieurs et inferieurs.

   L'operation peut etre limitee au plan grace au champ mode de
   la structure de parametre :
   - MODE_2D : on ne considere que la restriction au plan de 
     l'element strcturant
   - MODE_3D : on considere l'element structurant 3D

   Le nombre d'iterations a realiser avec un element structurant
   est donne par le champ order de la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_Erosion( t_ImagePtr theIm    /* input image */, 
	       t_ImagePtr resIm     /* result image */, 
	       e_ParamMorphoPtr par /* parameters */ )
#else
int E_Erosion(theIm, resIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	t_Image outIm;
	int i;
	char *buffer;
	
        E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_Erosion", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_Erosion", "output image not allocated");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_Erosion", 
               "input and output have different dimensions");
	E_ERR( (par->order < 1), E_ERRFATAL, FALSE, "E_Erosion", 
	     "error in parameters");
	E_ERR( ((par->connexite!=CONNEXE_26) && (par->connexite!=CONNEXE_10) && (par->connexite!=CONNEXE_6)), E_ERRFATAL,
	     FALSE, "E_Erosion", "error in parameters");


	/* first case */
	/**************/
	if ( (par->order == 1) && (theIm->buf != resIm->buf) ) {
		E_ErosionSingle( theIm, resIm, par );
		return ( TRUE );
	}
	/* second case */
	/***************/
	if ( (par->order == 1) && (theIm->buf == resIm->buf) ) {
		T_CopyHeaderIm( resIm, &outIm );
		T_AllocImage( &outIm );
		E_ISERR(FALSE,"E_Erosion",(char*)NULL);
		E_ErosionSingle( theIm, &outIm, par );
		buffer = theIm->buf;
		theIm->buf = resIm->buf = outIm.buf;
		outIm.buf = buffer;
		T_FreeImage( &outIm );
		return( TRUE );
	}
	/* third and fourth cases */
	/* par->order > 1         */
	/**************************/
	T_CopyHeaderIm( resIm, &outIm );
	T_AllocImage( &outIm );
	E_ISERR(FALSE,"E_Erosion",(char*)NULL);

	E_ErosionSingle( theIm, &outIm, par );
	for (i=0; i<par->order-2; i++) {
		E_ErosionSingle( &outIm, resIm, par );
		buffer = outIm.buf;
		outIm.buf = resIm->buf;
		resIm->buf = buffer;
	}
	E_ErosionSingle( &outIm, resIm, par );
	T_FreeImage( &outIm );
	return( TRUE );
}

/* Fermeture morphologique.

   La fermeture est definie abusivement comme etant une dilatation
   suivie d'une erosion (cette definition n'est vraie que
   dans le cas des elements structurants symetriques, ce qui
   est verifie ici).

   La fermeture supprime les parties sombres plus petites que 
   l'element structurant dans l'image traitee.

   Les operations morphologiques sont realisees avec les valeurs
   donnees par la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_Closing( t_ImagePtr theIm    /* input image */, 
	       t_ImagePtr resIm     /* result image */, 
	       e_ParamMorphoPtr par /* parameters */ )
#else
int E_Closing(theIm, resIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_Closing", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_Closing", "output image not allocated");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_Closing", 
               "input and output have different dimensions");
	
	E_Dilation( theIm, resIm, par );
	E_ISERR(FALSE,"E_Closing",(char*)NULL);
	E_Erosion( resIm, resIm, par );
	E_ISERR(FALSE,"E_Closing",(char*)NULL);
	return( TRUE );
}

/* Ouverture morphologique.

   L'ouverture est definie abusivement comme etant une erosion
   suivie d'une dilatation (cette definition n'est vraie que
   dans le cas des elements structurants symetriques, ce qui
   est verifie ici).

   L'ouverture supprime les parties claires plus petites que 
   l'element structurant dans l'image traitee.

   Les operations morphologiques sont realisees avec les valeurs
   donnees par la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_Opening( t_ImagePtr theIm    /* input image */, 
	       t_ImagePtr resIm     /* result image */, 
	       e_ParamMorphoPtr par /* parameters */ )
#else
int E_Opening(theIm, resIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_Opening", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_Opening", "output image not allocated");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_Opening", 
               "input and output have different dimensions");
	
	E_Erosion( theIm, resIm, par );
	E_ISERR(FALSE,"E_Opening",(char*)NULL);
	E_Dilation( resIm, resIm, par );
	E_ISERR(FALSE,"E_Opening",(char*)NULL);
	return( TRUE );
}

/* Calcul d'une dilatation conditionnelle.

   La dilatation conditionnelle de theIm par rapport a supIm
   force la dilatation a ne se produire qu'a la frontiere
   entre theIm et supIm.

   Les operations conditionnelles sont realisees par des operations
   logiques sur les buffers, il vaut mieux les utiliser sur des images
   binaires

   Les operations morphologiques sont realisees avec les valeurs
   donnees par la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_CDilation( t_ImagePtr theIm    /* input image */, 
		t_ImagePtr resIm     /* result image */, 
		t_ImagePtr supIm     /* auxiliary image */,
		e_ParamMorphoPtr par /* parameters */ )
#else
int E_CDilation(theIm, resIm, supIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
t_ImagePtr supIm;     /* auxiliary image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	t_Image outIm;
	int i;
	char *buffer;

	E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_CDilation", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_CDilation", "output image not allocated");
        E_ERR( (supIm==NULL) || (supIm->buf==NULL), E_ERRNULL, FALSE,
             "E_CDilation", "output image not allocated");
	E_ERR( (supIm->buf == theIm->buf) || (supIm->buf == resIm->buf), E_ERRNULL, FALSE,
	      "E_CDilation", "support image equal to input or output image");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_CDilation", 
               "input and output have different dimensions");
        E_ERR( (theIm->tx!=supIm->tx) || (theIm->ty!=supIm->ty) || (theIm->tz!=supIm->tz),
               E_ERRFATAL, FALSE, "E_CDilation", 
               "input and support have different dimensions");
	
	/* first case */
	/**************/
	if ( (par->order == 1) && (theIm->buf != resIm->buf) ) {
		E_DilationSingle( theIm, resIm, par );
		E_ISERR(FALSE,"E_CDilation",(char*)NULL);
		E_LogicDilCond(theIm->buf, supIm->buf, resIm->buf, T_ImageSize(theIm));
		return ( TRUE );
	}
	/* second case */
	/***************/
	if ( (par->order == 1) && (theIm->buf == resIm->buf) ) {
		T_CopyHeaderIm( resIm, &outIm );
		T_AllocImage( &outIm );
		E_ISERR(FALSE,"E_CDilation",(char*)NULL);
		E_DilationSingle( theIm, &outIm, par );
		E_ISERR(FALSE,"E_CDilation",(char*)NULL);
		E_LogicDilCond(theIm->buf, supIm->buf, outIm.buf, T_ImageSize(theIm));
		buffer = theIm->buf;
		theIm->buf = resIm->buf = outIm.buf;
		outIm.buf = buffer;
		T_FreeImage( &outIm );
		return( TRUE );
	}
	/* third and fourth cases */
	/* par->order > 1         */
	/**************************/
	T_CopyHeaderIm( resIm, &outIm );
	T_AllocImage( &outIm );
	E_ISERR(FALSE,"E_Dilation",(char*)NULL);

	E_DilationSingle( theIm, &outIm, par );
	E_ISERR(FALSE,"E_CDilation",(char*)NULL);
	E_LogicDilCond(theIm->buf, supIm->buf, outIm.buf, T_ImageSize(theIm));
	for (i=0; i<par->order-2; i++) {
		E_DilationSingle( &outIm, resIm, par );
		E_ISERR(FALSE,"E_CDilation",(char*)NULL);
		E_LogicDilCond(outIm.buf, supIm->buf, resIm->buf, T_ImageSize(theIm));
		buffer = outIm.buf;
		outIm.buf = resIm->buf;
		resIm->buf = buffer;
	}
	E_DilationSingle( &outIm, resIm, par );
	E_ISERR(FALSE,"E_CDilation",(char*)NULL);
	E_LogicDilCond(outIm.buf, supIm->buf, resIm->buf, T_ImageSize(theIm));
	T_FreeImage( &outIm );
	return( TRUE );
}

/* Calcul d'une erosion conditionnelle.

   L'erosion conditionnelle de theIm par rapport a supIm
   force l'erosion a ne se produire qu'a la frontiere
   entre theIm et supIm.

   Les operations conditionnelles sont realisees par des operations
   logiques sur les buffers, il vaut mieux les utiliser sur des images
   binaires

   Les operations morphologiques sont realisees avec les valeurs
   donnees par la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_CErosion( t_ImagePtr theIm     /* input image */, 
		t_ImagePtr resIm     /* result image */, 
		t_ImagePtr supIm     /* auxiliary image */,
		e_ParamMorphoPtr par /* parameters */ )
#else
int E_CErosion( theIm, resIm, supIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
t_ImagePtr supIm;     /* auxiliary image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_CErosion", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_CErosion", "output image not allocated");
        E_ERR( (supIm==NULL) || (supIm->buf==NULL), E_ERRNULL, FALSE,
             "E_CErosion", "output image not allocated");
	E_ERR( (supIm->buf == theIm->buf) || (supIm->buf == resIm->buf), E_ERRNULL, FALSE,
	      "E_CErosion", "support image equal to input or output image");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_CErosion", 
               "input and output have different dimensions");
        E_ERR( (theIm->tx!=supIm->tx) || (theIm->ty!=supIm->ty) || (theIm->tz!=supIm->tz),
               E_ERRFATAL, FALSE, "E_CErosion", 
               "input and support have different dimensions");
	

	E_CDilation( supIm, resIm, theIm, par );
	E_ISERR(FALSE,"E_CErosion",(char*)NULL);
	E_LogicEroCond(theIm->buf, resIm->buf, T_ImageSize(theIm));
 	return( TRUE );
}


#if defined(_ANSI_)
static void E_LogicDilCond(char *tab1, char *tab2, char *tab3, int taille)
#else
static void E_LogicDilCond(tab1, tab2, tab3, taille)
char *tab1;
char *tab2;
char *tab3;
int taille;
#endif
{
  register int i;
  /*   tab3 : tableau contenant le dilate de X   */
  /*   tab2 : tableau contenant le support Y     */
  /*   tab1 : tableau contenant l'objet X        */
  /***********************************************/
  for (i=0;i<taille;i++){
    *tab3 = *tab1++ | ( *tab3 & *tab2++ );
    tab3++;
  }
}

#if defined(_ANSI_)
static void E_LogicEroCond(char *tab1, char *tab3, int taille)
#else
static void E_LogicEroCond(tab1, tab3, taille)
char *tab1;
char *tab3;
int taille;
#endif
{
  register int i;
  /*   tab3 : tableau contenant le dilate de Y   */
  /*   tab1 : tableau contenant l'objet X        */
  /***********************************************/
  for (i=0;i<taille;i++) {
    *tab3 = *tab1++ & ( *tab3  ^ 0xff );
    tab3++;
  }
}

/* Calcul du gradient morphologique.

   Le laplacien morphologique est defini comme etant :
              dilatee - erodee
   Noter le lien avec la definition de la derivee premiere au sens
   des differences finies.

   Les operations morphologiques sont realisees avec les valeurs
   donnees par la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_MGradient( t_ImagePtr theIm    /* input image */, 
		t_ImagePtr resIm     /* result image */, 
		e_ParamMorphoPtr par /* parameters */ )
#else
int E_MGradient(theIm, resIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	t_Image outIm;
	register int i, j, k;
	reeltype val;
	
        E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_MGradient", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_MGradient", "output image not allocated");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_MGradient", 
               "input and output have different dimensions");
	E_ERR( ((par->connexite!=CONNEXE_26) && (par->connexite!=CONNEXE_10) && (par->connexite!=CONNEXE_6)), E_ERRFATAL,
	     FALSE, "E_MGradient", "error in parameters");


	T_CopyHeaderIm( resIm, &outIm );
	T_AllocImage( &outIm );
	E_ISERR(FALSE,"E_MGradient",(char*)NULL);
	E_Erosion(theIm, &outIm, par);
	E_ISERR(FALSE,"E_MGradient",(char*)NULL);
	E_Dilation(theIm, resIm, par);
	E_ISERR(FALSE,"E_MGradient",(char*)NULL);

	switch ( resIm->type ) {
	case TYPEIM_256 :
	  {
	    unsigned char *br, *bo;
	    k = theIm->tz * theIm->ty * theIm->tx;
	    br = (unsigned char *)(resIm->buf);
	    bo = (unsigned char *)(outIm.buf);
	    for (i=0; i<k; i++)
	      br[i] = (unsigned char)( br[i] - *(bo++) );
	  }
	  break;
	default :
	  for (k=0; k<theIm->tz; k++)
	    for (j=0; j<theIm->ty; j++)
	      for (i=0; i<theIm->tx; i++) {
		/* val = (T_Get3Point(i, j, k, resIm) - T_Get3Point(i, j, k, &outIm))/2.0; */
		val = (T_Get3Point(i, j, k, resIm) - T_Get3Point(i, j, k, &outIm));
		T_Set3Point(i, j, k, resIm, (reeltype)val, (char)MODE_SET );
	      }
	}
	T_FreeImage( &outIm );
	return( TRUE );
}

/* Filtre (morphologique) rehausseur de contraste.

   On calcule l'erodee et la dilatee de l'image d'entree.
   Un point de l'image resultat aura pour valeur, la valeur 
   "la plus proche" entre les valeurs correspondantes dans
   les images dilatee et erodee, c'est-a-dire :
   - dilatee   si (dilatee - image) < (image - erodee)
   - erodee    si (dilatee - image) > (image - erodee)
   - inchangee sinon (la valeur initiale est entre les 2)

   Les operations morphologiques sont realisees avec les valeurs
   donnees par la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_MContrast( t_ImagePtr theIm    /* input image */, 
	       t_ImagePtr resIm     /* result image */, 
	       e_ParamMorphoPtr par /* parameters */ )
#else
int E_MContrast(theIm, resIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
	t_Image outIm, auxIm;
	register int i, j, k;
	reeltype valOri, valEro, valDil, dif1, dif2;
	
        E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	     "E_MContrast", "input image not allocated");
        E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
             "E_MContrast", "output image not allocated");
        E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
               E_ERRFATAL, FALSE, "E_MContrast", 
               "input and output have different dimensions");
	E_ERR( (par->order < 1), E_ERRFATAL, FALSE, "E_MContrast", 
	     "error in parameters");
	E_ERR( ((par->connexite!=CONNEXE_26) && (par->connexite!=CONNEXE_10) && (par->connexite!=CONNEXE_6)), E_ERRFATAL,
	     FALSE, "E_MContrast", "error in parameters");


	T_CopyHeaderIm( theIm, &outIm );
	T_AllocImage( &outIm );
	E_ISERR(FALSE,"E_MContrast",(char*)NULL);
	T_CopyHeaderIm( theIm, &auxIm );
	if (theIm->buf == resIm->buf) 
		T_AllocImage( &auxIm );
	else
		auxIm.buf = resIm->buf;
	E_ISERR(FALSE,"E_MContrast",(char*)NULL);

	E_Erosion(theIm, &outIm, par);
	E_ISERR(FALSE,"E_MContrast",(char*)NULL);
	E_Dilation(theIm, &auxIm, par);
	E_ISERR(FALSE,"E_MContrast",(char*)NULL);

	switch ( theIm->type ) {
	case TYPEIM_256 :
	  if ( resIm->type == TYPEIM_256 ) {
	    unsigned char *bi, *bo, *br, *ba;
	    register unsigned char vi, vo, va, d1, d2;
	    bi = (unsigned char*)(theIm->buf);
	    bo = (unsigned char*)(outIm.buf);
	    ba = (unsigned char*)(auxIm.buf);
	    br = (unsigned char*)(resIm->buf);
	    k = theIm->tz * theIm->ty * theIm->tx;
	    for (i=0; i<k; i++) {
	      vi = *(bi++);   vo = *(bo++);   va = *(ba++);
	      d1 = vi - vo;   d2 = va - vi;
	      if (d1 < d2) {
		*(br++) = vo;
	      }
	      else if (d1 > d2) {
		*(br++) = va;
	      }
	      else
		*(br++) = vi;
	    }
	  }
	  else {
	    unsigned char *bi, *bo, *ba;
	    register unsigned char vi, vo, va, d1, d2;
	    bi = (unsigned char*)(theIm->buf);
	    bo = (unsigned char*)(outIm.buf);
	    ba = (unsigned char*)(auxIm.buf);
	    for (k=0; k<theIm->tz; k++)
	      for (j=0; j<theIm->ty; j++)
		for (i=0; i<theIm->tx; i++) {
		  vi = *(bi++);   vo = *(bo++);   va = *(ba++);
		  d1 = vi - vo;   d2 = va - vi;
		  if (d1 < d2) {
		    T_Set3Point(i, j, k, resIm, (reeltype)vo, (char)MODE_SET );
		  }
		  else if (d1 > d2) {
		    T_Set3Point(i, j, k, resIm, (reeltype)va, (char)MODE_SET );
		  }
		  else
		    T_Set3Point(i, j, k, resIm, (reeltype)vi, (char)MODE_SET );
		}
	  }
	  break;
	default :
	  for (k=0; k<theIm->tz; k++)
	    for (j=0; j<theIm->ty; j++)
	      for (i=0; i<theIm->tx; i++) {
		valOri = T_Get3Point(i, j, k, theIm);
		valEro = T_Get3Point(i, j, k, &outIm);
		valDil = T_Get3Point(i, j, k, &auxIm);
		dif1 = valOri - valEro;
		dif2 = valDil - valOri;
		if (dif1 < dif2) {
		  T_Set3Point(i, j, k, resIm, (reeltype)valEro, (char)MODE_SET );
		}
		else if (dif1 > dif2) {
		  T_Set3Point(i, j, k, resIm, (reeltype)valDil, (char)MODE_SET );
		}
		else
		  T_Set3Point(i, j, k, resIm, (reeltype)valOri, (char)MODE_SET );
	      }
	}
	if (theIm->buf == resIm->buf) 
	  T_FreeImage( &auxIm );
	T_FreeImage( &outIm );
	return( TRUE );
}

/* Calcul du laplacien morphologique, suivant d'une operation eventuelle.

   Le laplacien morphologique est defini comme etant :
              dilatee - 2 * image + erodee
   Noter le lien avec la definition de la derivee seconde au sens
   des differences finies.

   Si le champ scnd_op de la structure est different de ZEROCROSS
   ou GRADIENT, l'image resultat contient le laplacien.
   Si ce champ est egal a ZEROCROSS, l'image resultat contient les
   passages par zero du laplacien. Si il est egal a GRADIENT,
   elle contient egalement les passages par zero, mais avec la valeur
   du gradient (morphologique : dilatee - erodee) associe.

   Les operations morphologiques sont realisees avec les valeurs
   donnees par la structure de parametre.

RETURN
   FALSE en cas d'erreur.
*/

#if defined(_ANSI_)
int E_MLaplacian( t_ImagePtr theIm    /* input image */, 
	       t_ImagePtr resIm     /* result image */, 
	       e_ParamMorphoPtr par /* parameters */ )
#else
int E_MLaplacian(theIm, resIm, par )
t_ImagePtr theIm;     /* input image */
t_ImagePtr resIm;     /* result image */
e_ParamMorphoPtr par; /* parameters */
#endif
{
  t_Image eroIm, dilIm, lapIm;
  register int i, j, k;
  float lap;
  int dim;
  
  E_ERR( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL, FALSE,
	 "E_MLaplacian", "input image not allocated");
  E_ERR( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL, FALSE,
	 "E_MLaplacian", "output image not allocated");
  E_ERR( (theIm->tx!=resIm->tx) || (theIm->ty!=resIm->ty) || (theIm->tz!=resIm->tz),
	 E_ERRFATAL, FALSE, "E_MLaplacian", 
	 "input and output have different dimensions");
  E_ERR( (par->order < 1), E_ERRFATAL, FALSE, "E_MLaplacian", 
	 "error in parameters");
  E_ERR( ((par->connexite!=CONNEXE_26) && (par->connexite!=CONNEXE_10) && (par->connexite!=CONNEXE_6)), E_ERRFATAL,
	 FALSE, "E_MLaplacian", "error in parameters");
  
  dim = theIm->tz * theIm->ty * theIm->tx;
  
  eroIm.buf = dilIm.buf = lapIm.buf = (char*)NULL;
  
  if ( (par->scnd_op == GRADIENT) || (par->scnd_op == ZEROCROSS) ) {
    T_CopyHeaderIm( theIm, &lapIm );
    lapIm.type = TYPEIM_FLOAT;
    T_AllocImage( &lapIm );
    E_REPERR( goto END_E_MLaplacian );
  } else {
    T_CopyHeaderIm( resIm, &lapIm );
    lapIm.buf = resIm->buf;
  }
  
  T_CopyHeaderIm( theIm, &eroIm );
  T_AllocImage( &eroIm );
  E_REPERR( goto END_E_MLaplacian );
  T_CopyHeaderIm( theIm, &dilIm );
  T_AllocImage( &dilIm );
  E_REPERR( goto END_E_MLaplacian );
  
  E_Erosion(theIm, &eroIm, par);
  E_REPERR( goto END_E_MLaplacian );
  E_Dilation(theIm, &dilIm, par);
  E_REPERR( goto END_E_MLaplacian );
  
  switch ( theIm->type ) {
  case TYPEIM_256 :
    {
      unsigned char *bi, *be, *bd;
      register int vi, vo, va, ilap;
      bi = (unsigned char*)(theIm->buf);
      be = (unsigned char*)(eroIm.buf);
      bd = (unsigned char*)(dilIm.buf);
      switch ( lapIm.type ) {
      case TYPEIM_S16B :
	{
	  short int *tmp;
	  tmp = (short int*)(lapIm.buf);
	  for (i=0; i<dim; i++) {
	    vi = *(bi++);   vo = *(be++);   va = *(bd++);
	    ilap  = vo + va - vi - vi;
	    *tmp++ = (short int)ilap;
	  }}
	break;
      case TYPEIM_FLOAT :
	{
	  float *tmp;
	  tmp = (float*)(lapIm.buf);
	  for (i=0; i<dim; i++) {
	    vi = *(bi++);   vo = *(be++);   va = *(bd++);
	    ilap  = vo + va - vi - vi;
	    *tmp++ = (float)(ilap/255.0);
	  }}
	break;
      default :
	for (k=0; k<theIm->tz; k++)
	  for (j=0; j<theIm->ty; j++)
	    for (i=0; i<theIm->tx; i++) {
	      vi = *(bi++);   vo = *(be++);   va = *(bd++);
	      lap = (float) (vo - 2 *vi + va);
	      T_Set3Point(i, j, k, &lapIm, (reeltype)(lap/255.0), (char)MODE_SET );
	    }
      }}
    break;
  case TYPEIM_16B :
    {
      unsigned short int *bi, *be, *bd;
      register int vi, vo, va, ilap;
      bi = (unsigned short int*)(theIm->buf);
      be = (unsigned short int*)(eroIm.buf);
      bd = (unsigned short int*)(dilIm.buf);
      switch ( lapIm.type ) {
      case TYPEIM_S16B :
	{
	  short int *tmp;
	  tmp = (short int*)(lapIm.buf);
	  for (i=0; i<dim; i++) {
	    vi = *(bi++);   vo = *(be++);   va = *(bd++);
	    ilap  = vo + va - vi - vi;
	    *tmp++ = (short int)ilap;
	  }}
	break;
      case TYPEIM_FLOAT :
	{
	  float *tmp;
	  tmp = (float*)(lapIm.buf);
	  for (i=0; i<dim; i++) {
	    vi = *(bi++);   vo = *(be++);   va = *(bd++);
	    ilap  = vo + va - vi - vi;
	    *tmp++ = (float)(ilap/255.0);
	  }}
	break;
      default :
	for (k=0; k<theIm->tz; k++)
	  for (j=0; j<theIm->ty; j++)
	    for (i=0; i<theIm->tx; i++) {
	      vi = *(bi++);   vo = *(be++);   va = *(bd++);
	      lap = (float) (vo - 2*vi + va);
	      T_Set3Point(i, j, k, &lapIm, (reeltype)(lap/255.0), (char)MODE_SET );
	    }
      }}
    break;
  case TYPEIM_S16B :
    {
      short int *bi, *be, *bd;
      register int vi, vo, va, ilap;
      bi = (short int*)(theIm->buf);
      be = (short int*)(eroIm.buf);
      bd = (short int*)(dilIm.buf);
      switch ( lapIm.type ) {
      case TYPEIM_S16B :
	{
	  short int *tmp;
	  tmp = (short int*)(lapIm.buf);
	  for (i=0; i<dim; i++) {
	    vi = *(bi++);   vo = *(be++);   va = *(bd++);
	    ilap  = vo + va - vi - vi;
	    *tmp++ = (short int)ilap;
	  }}
	break;
      case TYPEIM_FLOAT :
	{
	  float *tmp;
	  tmp = (float*)(lapIm.buf);
	  for (i=0; i<dim; i++) {
	    vi = *(bi++);   vo = *(be++);   va = *(bd++);
	    ilap  = vo + va - vi - vi;
	    *tmp++ = (float)(ilap/255.0);
	  }}
	break;
      default :
	for (k=0; k<theIm->tz; k++)
	  for (j=0; j<theIm->ty; j++)
	    for (i=0; i<theIm->tx; i++) {
	      vi = *(bi++);   vo = *(be++);   va = *(bd++);
	      lap = (float) (vo - 2.0*((float)vi) + (float)va);
	      T_Set3Point(i, j, k, &lapIm, (reeltype)(lap/255.0), (char)MODE_SET );
	    }
      }}
    break;
  default :	
    {
      reeltype valOri, valEro, valDil;
      for (k=0; k<theIm->tz; k++)
	for (j=0; j<theIm->ty; j++)
	  for (i=0; i<theIm->tx; i++) {
	    valOri = T_Get3Point(i, j, k, theIm);
	    valEro = T_Get3Point(i, j, k, &eroIm);
	    valDil = T_Get3Point(i, j, k, &dilIm);
	    lap = valDil - 2 * valOri + valEro;
	    T_Set3Point(i, j, k, &lapIm, (reeltype)lap, (char)MODE_SET );
	  }}
  }
  
  if ( (par->scnd_op != GRADIENT) && (par->scnd_op != ZEROCROSS) ) {
    T_FreeImage( &dilIm );
    T_FreeImage( &eroIm );
    return( TRUE );
  }
  
  E_ZeroCrossings( &lapIm, resIm, par->mode );
  T_FreeImage( &lapIm ); lapIm.buf = (char*)NULL;
  if ( par->scnd_op == ZEROCROSS ) {
    T_FreeImage( &dilIm );
    T_FreeImage( &eroIm );
    return( TRUE );
  }
  
  /*--- reste le cas ou par->scnd_op == GRADIENT ---*/
  
  switch ( theIm->type ) {
  case TYPEIM_256 :
    {
      unsigned char *be, *bd;
      register int ve, vd, vz;
      be = (unsigned char*)(eroIm.buf);
      bd = (unsigned char*)(dilIm.buf);
      switch ( resIm->type ) {
      case TYPEIM_256 :
	{
	  unsigned char *tmp;
	  tmp = (unsigned char*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (unsigned char)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_16B :
	{
	  unsigned short int *tmp;
	  tmp = (unsigned short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (unsigned short int)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_S16B :
	{
	  short int *tmp;
	  tmp = (short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (short int)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_FLOAT :
	{
	  float *tmp;
	  tmp = (float*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    /* vz = *tmp; */
	    if (*tmp > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (float)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      default :
	for (k=0; k<theIm->tz; k++)
	  for (j=0; j<theIm->ty; j++)
	    for (i=0; i<theIm->tx; i++) {
	      ve = *(be++);   vd = *(bd++);
	      if ( T_Get3Point(i, j, k, resIm) > 0.0 )
		T_Set3Point(i, j, k, resIm, (reeltype)((vd-ve)/255.0), (char)MODE_SET );
	    }
      }}
    break;
  case TYPEIM_16B :
    {
      unsigned short int *be, *bd;
      register int ve, vd, vz;
      be = (unsigned short int*)(eroIm.buf);
      bd = (unsigned short int*)(dilIm.buf);
      switch ( resIm->type ) {
      case TYPEIM_256 :
	{
	  unsigned char *tmp;
	  tmp = (unsigned char*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (unsigned char)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_16B :
	{
	  unsigned short int *tmp;
	  tmp = (unsigned short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (unsigned short int)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_S16B :
	{
	  short int *tmp;
	  tmp = (short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (short int)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_FLOAT :
	{
	  float *tmp;
	  tmp = (float*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    /* vz = *tmp; */
	    if (*tmp > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (float)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      default :
	for (k=0; k<theIm->tz; k++)
	  for (j=0; j<theIm->ty; j++)
	    for (i=0; i<theIm->tx; i++) {
	      ve = *(be++);   vd = *(bd++);
	      if ( T_Get3Point(i, j, k, resIm) > 0.0 )
		T_Set3Point(i, j, k, resIm, (reeltype)((vd-ve)/255.0), (char)MODE_SET );
	    }
      }}
    break;
  case TYPEIM_S16B :
    {
      short int *be, *bd;
      register int ve, vd, vz;
      be = (short int *)(eroIm.buf);
      bd = (short int *)(dilIm.buf);
      switch ( resIm->type ) {
      case TYPEIM_256 :
	{
	  unsigned char *tmp;
	  tmp = (unsigned char*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (unsigned char)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_16B :
	{
	  unsigned short int *tmp;
	  tmp = (unsigned short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (unsigned short int)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_S16B :
	{
	  short int *tmp;
	  tmp = (short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    vz = *tmp;
	    if (vz > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (short int)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_FLOAT :
	{
	  float *tmp;
	  tmp = (float*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    /* vz = *tmp; */
	    if (*tmp > 0) {
	      ve = *(be++);   vd = *(bd++);
	      *tmp++ = (float)(vd-ve);
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      default :
	for (k=0; k<theIm->tz; k++)
	  for (j=0; j<theIm->ty; j++)
	    for (i=0; i<theIm->tx; i++) {
	      ve = *(be++);   vd = *(bd++);
	      if ( T_Get3Point(i, j, k, resIm) > 0.0 )
		T_Set3Point(i, j, k, resIm, (reeltype)((vd-ve)/255.0), (char)MODE_SET );
	    }
      }}
    break;
  case TYPEIM_FLOAT :
    {
      float *be, *bd;
      /* register int ve, vd, vz; */
      be = (float*)(eroIm.buf);
      bd = (float*)(dilIm.buf);
      switch ( resIm->type ) {
      case TYPEIM_256 :
	{
	  unsigned char *tmp;
	  tmp = (unsigned char*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    /* vz = *tmp; */
	    if (*tmp > 0) {
	      /* ve = *(be++);   vd = *(bd++); */
	      *tmp++ = (unsigned char)(*(bd++) - *(be++)); /* (vd-ve); */
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_16B :
	{
	  unsigned short int *tmp;
	  tmp = (unsigned short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    /* vz = *tmp; */
	    if (*tmp > 0) {
	      /* ve = *(be++);   vd = *(bd++); */
	      *tmp++ = (unsigned short int)(*(bd++) - *(be++)); /* (vd-ve); */
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_S16B :
	{
	  short int *tmp;
	  tmp = (short int*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    /* vz = *tmp; */
	    if (*tmp > 0) {
	      /* ve = *(be++);   vd = *(bd++); */
	      *tmp++ = (short int)(*(bd++) - *(be++)); /* (vd-ve); */
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      case TYPEIM_FLOAT :
	{
	  float *tmp;
	  tmp = (float*)(resIm->buf);
	  for (i=0; i<dim; i++) {
	    /* vz = *tmp; */
	    if (*tmp > 0) {
	      /* ve = *(be++);   vd = *(bd++); */
	      *tmp++ = (float)(*(bd++) - *(be++)); /* (vd-ve); */
	    } else {
	      tmp++; be++; bd++;
	    }
	  }}
	break;
      default :
	for (k=0; k<theIm->tz; k++)
	  for (j=0; j<theIm->ty; j++)
	    for (i=0; i<theIm->tx; i++) {
	      /* ve = *(be++);   vd = *(bd++); */
	      if ( T_Get3Point(i, j, k, resIm) > 0.0 )
		T_Set3Point(i, j, k, resIm, (reeltype)((*(bd++) - *(be++))/255.0), (char)MODE_SET );
	    }
      }}
    break;
  default :
    {
      reeltype tmp;	
      for (k=0; k<resIm->tz; k++)
	for (j=0; j<resIm->ty; j++)
	  for (i=0; i<resIm->tx; i++) {
	    if ( T_Get3Point(i, j, k, resIm) > 0.0 ) {
	      tmp = T_Get3Point(i, j, k, &dilIm) - T_Get3Point(i, j, k, &eroIm);
	      T_Set3Point(i, j, k, resIm, (reeltype)tmp, (char)MODE_SET );
	    }
	  }}
  }
  T_FreeImage( &dilIm );
  T_FreeImage( &eroIm );
  return( TRUE );
  
  END_E_MLaplacian :
    T_FreeImage( &dilIm );
  T_FreeImage( &eroIm );
  if ( (par->scnd_op == GRADIENT) || (par->scnd_op == ZEROCROSS) ) 
    T_FreeImage( &lapIm );
  E_ISERR(FALSE,"E_MLaplacian",(char*)NULL);
  return( TRUE );
}


#ifdef COMPILE_UNUSED

#if defined(_ANSI_)
static int _E_DilationSingle( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int _E_DilationSingle(theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamMorphoPtr par ;
#endif
{
	/*------------------------------------------------------------*/
	/*                                                            */
	/*   morphology : a single dilation                           */
	/*                                                            */
	/*------------------------------------------------------------*/
	t_Image supIm;
	reeltype max;
	register int x,y,z;

	switch (par->connexite) {
	case CONNEXE_26 :
	case CONNEXE_10 :
		T_CopyHeaderIm( resIm, &supIm );
		supIm.buf = resIm->buf;
		break;
	case CONNEXE_06 :
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
		break;
	}

	/*   maxima computation along X   */
	/**********************************/
	if ( theIm->tx > 1 ) {
		for ( z = 0; z < theIm->tz; z++ ) 
			for ( y = 0; y < theIm->ty; y++ ) {
				for ( x = 0; x < theIm->tx-1; x++ ) {
					max = EPI_MAX( T_Get3Point( x, y, z, theIm ), T_Get3Point( x+1, y, z, theIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
				T_Set3Point( theIm->tx-1, y, z, resIm, T_Get3Point( theIm->tx-1, y, z, theIm ), MODE_SET );
 				for ( x = theIm->tx-1; x > 0; x-- ) {
					max = EPI_MAX( T_Get3Point( x-1, y, z, theIm ), T_Get3Point( x, y, z, resIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
			}
	}
	/*   maxima computation along Y   */
	/**********************************/
	if ( theIm->ty > 1 ) {
		for ( z = 0; z < theIm->tz; z++ ) 
			for ( x = 0; x < theIm->tx; x++ ) {
				for ( y = 0; y < theIm->ty-1; y++ ) {
					max = EPI_MAX( T_Get3Point( x, y, z, resIm ), T_Get3Point( x, y+1, z, &supIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
 				for ( y = theIm->ty-1; y > 0; y-- ) {
					max = EPI_MAX( T_Get3Point( x, y-1, z, &supIm ), T_Get3Point( x, y, z, resIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
			}
	}
	
	if (par->mode == MODE_2D) 
                return( TRUE );

	/*   maxima computation along Z   */
	/**********************************/
	if ( par->connexite == CONNEXE_10 ) {
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
	}
	if ( theIm->tz > 1 ) {
		for ( y = 0; y < theIm->ty; y++ ) 
			for ( x = 0; x < theIm->tx; x++ ) {
				for ( z = 0; z < theIm->tz-1; z++ ) {
					max = EPI_MAX( T_Get3Point( x, y, z, resIm ), T_Get3Point( x, y, z+1, &supIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
 				for ( z = theIm->tz-1; z > 0; z-- ) {
					max = EPI_MAX( T_Get3Point( x, y, z-1, &supIm ), T_Get3Point( x, y, z, resIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
			}
	}

	return(TRUE);
}

#if defined(_ANSI_)
static int _E_ErosionSingle( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int _E_ErosionSingle(theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamMorphoPtr par ;
#endif
{
	/*------------------------------------------------------------*/
	/*                                                            */
	/*   morphology : a single erosion                            */
	/*                                                            */
	/*------------------------------------------------------------*/
	t_Image supIm;
	reeltype max;
	register int x,y,z;

	switch (par->connexite) {
	case CONNEXE_26 :
	case CONNEXE_10 :
		T_CopyHeaderIm( resIm, &supIm );
		supIm.buf = resIm->buf;
		break;
	case CONNEXE_06 :
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
		break;
	}

	/*   maxima computation along X   */
	/**********************************/
	if ( theIm->tx > 1 ) {
		for ( z = 0; z < theIm->tz; z++ ) 
			for ( y = 0; y < theIm->ty; y++ ) {
				for ( x = 0; x < theIm->tx-1; x++ ) {
					max = EPI_MIN( T_Get3Point( x, y, z, theIm ), T_Get3Point( x+1, y, z, theIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
				T_Set3Point( theIm->tx-1, y, z, resIm, T_Get3Point( theIm->tx-1, y, z, theIm ), MODE_SET );
 				for ( x = theIm->tx-1; x > 0; x-- ) {
					max = EPI_MIN( T_Get3Point( x-1, y, z, theIm ), T_Get3Point( x, y, z, resIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
			}
	}
	/*   maxima computation along Y   */
	/**********************************/
	if ( theIm->ty > 1 ) {
		for ( z = 0; z < theIm->tz; z++ ) 
			for ( x = 0; x < theIm->tx; x++ ) {
				for ( y = 0; y < theIm->ty-1; y++ ) {
					max = EPI_MIN( T_Get3Point( x, y, z, resIm ), T_Get3Point( x, y+1, z, &supIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
 				for ( y = theIm->ty-1; y > 0; y-- ) {
					max = EPI_MIN( T_Get3Point( x, y-1, z, &supIm ), T_Get3Point( x, y, z, resIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
			}
	}

	if (par->mode == MODE_2D) 
                return( TRUE );

	/*   maxima computation along Z   */
	/**********************************/
	if ( par->connexite == CONNEXE_10 ) {
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
	}
	if ( theIm->tz > 1 ) {
		for ( y = 0; y < theIm->ty; y++ ) 
			for ( x = 0; x < theIm->tx; x++ ) {
				for ( z = 0; z < theIm->tz-1; z++ ) {
					max = EPI_MIN( T_Get3Point( x, y, z, resIm ), T_Get3Point( x, y, z+1, &supIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
 				for ( z = theIm->tz-1; z > 0; z-- ) {
					max = EPI_MIN( T_Get3Point( x, y, z-1, &supIm ), T_Get3Point( x, y, z, resIm ) );
					T_Set3Point( x, y, z, resIm, (reeltype)max, (char)MODE_SET );
				}
			}
	}

	return(TRUE);
}
#endif /* COMPILE_UNUSED */



#ifdef _ANSI_
static int E_DilationSingle( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int E_DilationSingle( theIm, resIm, par )
t_ImagePtr theIm, resIm;
e_ParamMorphoPtr par;
#endif
{
  /*---------------------------------------------------------------*
   *           Single Dilation and Erosion                         *
   *             Version Amelioree                                 *
   *                YH 14/10/93                                    *
   *---------------------------------------------------------------*/
    
  t_Image supIm;
  reeltype max;
  register int x,y,z;
  EPI_Imaccess ima1_access, ima2_access;
  reeltype val1, val2;
  int retour;
  
  if (theIm->type == resIm->type) {
	  switch ( theIm->type ) {
	  case TYPEIM_256 :
		  if ( par->binaire == TRUE )
			  retour = DilationBin( theIm, resIm, par );
		  else
			  retour = DilationNdg( theIm, resIm, par );
		  return( retour );
	  case TYPEIM_16B :
		  retour = DilationNdg( theIm, resIm, par );
		  return( retour );
	  case TYPEIM_S16B :
		  retour = DilationNdg( theIm, resIm, par );
		  return( retour );
	  }
  }

  switch (par->connexite) {
  case CONNEXE_26 :
  case CONNEXE_10 :
    T_CopyHeaderIm( resIm, &supIm );
    supIm.buf = resIm->buf;
    break;
  case CONNEXE_06 :
    T_CopyHeaderIm( theIm, &supIm );
    supIm.buf = theIm->buf;
    break;
  }
  
  /*   maxima computation along X   */
  /**********************************/
  
  if ( theIm->tx > 1 ) {
    /*--------
      initialize the access to the input images
      -------*/
    EPI_init_access_point(&ima1_access, theIm);
    EPI_init_access_point(&ima2_access, resIm);
    
    for ( z = 0; z < theIm->tz; z++ ) {
      for ( y = 0; y < theIm->ty; y++ ) {
	/*------- 
	  premier parcours
	  -----*/	
	for ( x = 0; x < theIm->tx-1; x++ ) {
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_go_to_next_x_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val2);
	  max = EPI_MAX( val1, val2);
	  EPI_set_current_point(&ima2_access, max);
	  EPI_go_to_next_x_point(&ima2_access);
	}
	/*----- 
	  x = theIm->tx -1 val2 = theIm(x, y, z)
	  ------*/
	EPI_set_current_point(&ima2_access, val2);
	
	/*--------
	  deuxieme parcours
	  --------*/
	for ( x = theIm->tx-1; x > 0; x-- ) {
	  EPI_go_to_prev_x_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  max = EPI_MAX(val1, val2);
	  EPI_set_current_point(&ima2_access, max);
	  EPI_go_to_prev_x_point(&ima2_access);
	}
	
	/*------
	  sortie des deux parcours x = 0.
	  se positionner au debut de la ligne suivante
	  ------*/
	EPI_go_to_next_y_point(&ima1_access);
	EPI_go_to_next_y_point(&ima2_access);	
      }
    }
  }
  
  /*   maxima computation along Y   */
  /**********************************/
  if ( theIm->ty > 1 ) {
    
    EPI_init_access_point(&ima1_access, &supIm);
    EPI_go_to_point(&ima2_access, 0, 0, 0);
    
    for ( z = 0; z < theIm->tz; z++ ) {
      for ( x = 0; x < theIm->tx; x++ ) {
	for ( y = 0; y < theIm->ty-1; y++ ) {
	  EPI_go_to_next_y_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  max = EPI_MAX(val1, val2);
	  EPI_set_current_point(&ima2_access, max);
	  EPI_go_to_next_y_point(&ima2_access);
	}
	for ( y = theIm->ty-1; y > 0; y-- ) {
	  EPI_go_to_prev_y_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  max = EPI_MAX(val1, val2);
	  EPI_set_current_point(&ima2_access, max);
	  EPI_go_to_prev_y_point(&ima2_access);
	}
	EPI_go_to_next_x_point(&ima1_access);
	EPI_go_to_next_x_point(&ima2_access);	
      }
      /*-----
	le point atteint est (0, 1, z), il faut se positionner
	au debut du plan suivant i.e (0, 0, z + 1)
	-----*/
      EPI_go_to_prev_y_point(&ima1_access);
      EPI_go_to_next_z_point(&ima1_access);
      EPI_go_to_prev_y_point(&ima2_access);
      EPI_go_to_next_z_point(&ima2_access);
    }
  }
  
  if (par->mode == MODE_2D) 
	  return( TRUE );

  /*   maxima computation along Z   */
  /**********************************/
  if ( par->connexite == CONNEXE_10 ) {
	  T_CopyHeaderIm( theIm, &supIm );
	  supIm.buf = theIm->buf;
	  EPI_init_access_point(&ima1_access, &supIm);
  }
  if ( theIm->tz > 1 ) {
    EPI_go_to_point(&ima1_access, 0, 0, 0);
    EPI_go_to_point(&ima2_access, 0, 0, 0);
    
    for ( y = 0; y < theIm->ty; y++ ) {
      for ( x = 0; x < theIm->tx; x++ ) {
	
	for ( z = 0; z < theIm->tz-1; z++ ) {
	  EPI_go_to_next_z_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  max = EPI_MAX(val1, val2);
	  EPI_set_current_point(&ima2_access, max);
	  EPI_go_to_next_z_point(&ima2_access);
	}
	for ( z = theIm->tz-1; z > 0; z-- ) {
	  EPI_go_to_prev_z_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  max = EPI_MAX(val1, val2);
	  EPI_set_current_point(&ima2_access, max);
	  EPI_go_to_prev_z_point(&ima2_access);
	}
	EPI_go_to_next_x_point(&ima1_access);
	EPI_go_to_next_x_point(&ima2_access);
      }
    }
  }
  return(TRUE);  
}

#ifdef _ANSI_
static int E_ErosionSingle( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int E_ErosionSingle( theIm, resIm, par )
t_ImagePtr theIm, resIm;
e_ParamMorphoPtr par;
#endif
{
  t_Image supIm;
  reeltype min;
  register int x,y,z;
  EPI_Imaccess ima1_access, ima2_access;
  reeltype val1, val2;
  int retour;

  if (theIm->type == resIm->type) {
	  switch ( theIm->type ) {
	  case TYPEIM_256 :
		  if ( par->binaire == TRUE )
			  retour = ErosionBin( theIm, resIm, par );
		  else
			  retour = ErosionNdg( theIm, resIm, par );
		  return( retour );
	  case TYPEIM_16B :
		  retour = ErosionNdg( theIm, resIm, par );
		  return( retour );
	  case TYPEIM_S16B :
		  retour = ErosionNdg( theIm, resIm, par );
		  return( retour );
	  }
  }

  switch (par->connexite) {
  case CONNEXE_26 :
  case CONNEXE_10 :
    T_CopyHeaderIm( resIm, &supIm );
    supIm.buf = resIm->buf;
    break;
  case CONNEXE_06 :
    T_CopyHeaderIm( theIm, &supIm );
    supIm.buf = theIm->buf;
    break;
  }
  
  /*   minima computation along X   */
  /**********************************/
  
  if ( theIm->tx > 1 ) {
    /*--------
      initialize the access to the input images
      -------*/
    EPI_init_access_point(&ima1_access, theIm);
    EPI_init_access_point(&ima2_access, resIm);
    
    for ( z = 0; z < theIm->tz; z++ ) {
      for ( y = 0; y < theIm->ty; y++ ) {
	/*------- 
	  premier parcours
	  -----*/	
	for ( x = 0; x < theIm->tx-1; x++ ) {
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_go_to_next_x_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val2);
	  min = EPI_MIN( val1, val2);
	  EPI_set_current_point(&ima2_access, min);
	  EPI_go_to_next_x_point(&ima2_access);
	}
	/*----- 
	  x = theIm->tx -1 val2 = theIm(x, y, z)
	  ------*/
	EPI_set_current_point(&ima2_access, val2);
	
	/*--------
	  deuxieme parcours
	  --------*/
	for ( x = theIm->tx-1; x > 0; x-- ) {
	  EPI_go_to_prev_x_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  min = EPI_MIN(val1, val2);
	  EPI_set_current_point(&ima2_access, min);
	  EPI_go_to_prev_x_point(&ima2_access);
	}
	
	/*------
	  sortie des deux parcours x = 0.
	  se positionner au debut de la ligne suivante
	  ------*/
	EPI_go_to_next_y_point(&ima1_access);
	EPI_go_to_next_y_point(&ima2_access);	
      }
    }
  }
  
  /*   minima computation along Y   */
  /**********************************/
  if ( theIm->ty > 1 ) {
    
    EPI_init_access_point(&ima1_access, &supIm);
    EPI_go_to_point(&ima2_access, 0, 0, 0);
    
    for ( z = 0; z < theIm->tz; z++ ) {
      for ( x = 0; x < theIm->tx; x++ ) {
	for ( y = 0; y < theIm->ty-1; y++ ) {
	  EPI_go_to_next_y_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  min = EPI_MIN(val1, val2);
	  EPI_set_current_point(&ima2_access, min);
	  EPI_go_to_next_y_point(&ima2_access);
	}
	for ( y = theIm->ty-1; y > 0; y-- ) {
	  EPI_go_to_prev_y_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  min = EPI_MIN(val1, val2);
	  EPI_set_current_point(&ima2_access, min);
	  EPI_go_to_prev_y_point(&ima2_access);
	}
	EPI_go_to_next_x_point(&ima1_access);
	EPI_go_to_next_x_point(&ima2_access);	
      }
      /*-----
	le point atteint est (0, 1, z), il faut se positionner
	au debut du plan suivant i.e (0, 0, z + 1)
	-----*/
      EPI_go_to_prev_y_point(&ima1_access);
      EPI_go_to_next_z_point(&ima1_access);
      EPI_go_to_prev_y_point(&ima2_access);
      EPI_go_to_next_z_point(&ima2_access);
    }
  }
  
  if (par->mode == MODE_2D) 
	  return( TRUE );
	
  /*   minima computation along Z   */
  /**********************************/
  if ( par->connexite == CONNEXE_10 ) {
	  T_CopyHeaderIm( theIm, &supIm );
	  supIm.buf = theIm->buf;
	  EPI_init_access_point(&ima1_access, &supIm);
  }
  if ( theIm->tz > 1 ) {
    EPI_go_to_point(&ima1_access, 0, 0, 0);
    EPI_go_to_point(&ima2_access, 0, 0, 0);
    
    for ( y = 0; y < theIm->ty; y++ ) {
      for ( x = 0; x < theIm->tx; x++ ) {
	
	for ( z = 0; z < theIm->tz-1; z++ ) {
	  EPI_go_to_next_z_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  min = EPI_MIN(val1, val2);
	  EPI_set_current_point(&ima2_access, min);
	  EPI_go_to_next_z_point(&ima2_access);
	}
	for ( z = theIm->tz-1; z > 0; z-- ) {
	  EPI_go_to_prev_z_point(&ima1_access);
	  EPI_inq_current_point(&ima1_access, &val1);
	  EPI_inq_current_point(&ima2_access, &val2);
	  min = EPI_MIN(val1, val2);
	  EPI_set_current_point(&ima2_access, min);
	  EPI_go_to_prev_z_point(&ima2_access);
	}
	EPI_go_to_next_x_point(&ima1_access);
	EPI_go_to_next_x_point(&ima2_access);
      }
    }
  }
  return( TRUE );
}

#if defined(_ANSI_)
static int ErosionBin( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int ErosionBin( theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamMorphoPtr par ;
#endif
{
	t_Image supIm;
	register int x, y, z, i, j;
	int dx, dy, dz, dx1, dy1, dz1, dxy;
  
	dx = theIm->tx;   dy = theIm->ty;   dz = theIm->tz;   dxy = dx*dy;
	dx1 = dx - 1;   dy1 = dy - 1;   dz1 = dz - 1;
	switch (par->connexite) {
	case CONNEXE_26 :
	case CONNEXE_10 :
		T_CopyHeaderIm( resIm, &supIm );
		supIm.buf = resIm->buf;
		break;
	case CONNEXE_06 :
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
		break;
	}

	if ( theIm->type != resIm->type) return( FALSE );
	switch( theIm->type ) {
	case TYPEIM_256 :
		{
                unsigned char *bi, *bo, *ba;
		bi = (unsigned char*)(theIm->buf);
		bo = (unsigned char*)(resIm->buf);
		ba = (unsigned char*)(supIm.buf);
		/*--- minima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ )
					bo[i] = (bi[j] < bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- ) {
					if (bo[i] == 0) continue;
					if (bi[j] > 0 ) continue;
					bo[i] = 0;
				}
			}
		}
		/*--- minima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) {
					if (bo[i] == 0) continue;
					if (ba[j] > 0 ) continue;
					bo[i] = 0;
				}
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) {
					if (bo[i] == 0) continue;
					if (ba[j] > 0 ) continue;
					bo[i] = 0;
				}
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- minima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) {
					if (bo[i] == 0) continue;
					if (ba[j] > 0 ) continue;
					bo[i] = 0;
				}
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) {
					if (bo[i] == 0) continue;
					if (ba[j] > 0 ) continue;
					bo[i] = 0;
				}
			}
		}}
		break;
	default :
		return( FALSE );
	}
	return( TRUE );
}

#if defined(_ANSI_)
static int ErosionNdg( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int ErosionNdg( theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamMorphoPtr par ;
#endif
{
	t_Image supIm;
	register int x, y, z, i, j;
	int dx, dy, dz, dx1, dy1, dz1, dxy;
  
	dx = theIm->tx;   dy = theIm->ty;   dz = theIm->tz;   dxy = dx*dy;
	dx1 = dx - 1;   dy1 = dy - 1;   dz1 = dz - 1;
	switch (par->connexite) {
	case CONNEXE_26 :
	case CONNEXE_10 :
		T_CopyHeaderIm( resIm, &supIm );
		supIm.buf = resIm->buf;
		break;
	case CONNEXE_06 :
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
		break;
	}

	if ( theIm->type != resIm->type) return( FALSE );
	switch( theIm->type ) {
	case TYPEIM_256 :
		{
                unsigned char *bi, *bo, *ba;
		bi = (unsigned char*)(theIm->buf);
		bo = (unsigned char*)(resIm->buf);
		ba = (unsigned char*)(supIm.buf);
		/*--- minima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ ) 
					bo[i] = (bi[j] < bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- )
					if (bi[j] < bo[i]) bo[i] = bi[j];
			}
		}
		/*--- minima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- minima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
			}
		}}
		break;
	case TYPEIM_16B :
		{
                unsigned short int *bi, *bo, *ba;
		bi = (unsigned short int*)(theIm->buf);
		bo = (unsigned short int*)(resIm->buf);
		ba = (unsigned short int*)(supIm.buf);
		/*--- minima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ ) 
					bo[i] = (bi[j] < bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- )
					if (bi[j] < bo[i]) bo[i] = bi[j];
			}
		}
		/*--- minima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- minima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
			}
		}}
		break;
	case TYPEIM_S16B :
		{
                short int *bi, *bo, *ba;
		bi = (short int*)(theIm->buf);
		bo = (short int*)(resIm->buf);
		ba = (short int*)(supIm.buf);
		/*--- minima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ ) 
					bo[i] = (bi[j] < bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- )
					if (bi[j] < bo[i]) bo[i] = bi[j];
			}
		}
		/*--- minima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- minima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) 
					if (ba[j] < bo[i]) bo[i] = ba[j];
			}
		}}
		break;
	default :
		return( FALSE );
	}
	return( TRUE );
}

#if defined(_ANSI_)
static int DilationBin( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int DilationBin( theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamMorphoPtr par ;
#endif
{
	t_Image supIm;
	register int x, y, z, i, j;
	int dx, dy, dz, dx1, dy1, dz1, dxy;
  
	dx = theIm->tx;   dy = theIm->ty;   dz = theIm->tz;   dxy = dx*dy;
	dx1 = dx - 1;   dy1 = dy - 1;   dz1 = dz - 1;
	switch (par->connexite) {
	case CONNEXE_26 :
	case CONNEXE_10 :
		T_CopyHeaderIm( resIm, &supIm );
		supIm.buf = resIm->buf;
		break;
	case CONNEXE_06 :
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
		break;
	}

	if ( theIm->type != resIm->type) return( FALSE );
	switch( theIm->type ) {
	case TYPEIM_256 :
		{
                unsigned char *bi, *bo, *ba;
		bi = (unsigned char*)(theIm->buf);
		bo = (unsigned char*)(resIm->buf);
		ba = (unsigned char*)(supIm.buf);
		/*--- maxima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ )
					bo[i] = (bi[j] > bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- ) {
					if (bo[i] > 0) continue;
					if (bi[j] == 0 ) continue;
					bo[i] = 255;
				}
			}
		}
		/*--- maxima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) {
					if (bo[i] > 0) continue;
					if (ba[j] == 0 ) continue;
					bo[i] = 255;
				}
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) {
					if (bo[i] > 0) continue;
					if (ba[j] == 0 ) continue;
					bo[i] = 255;
				}
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- maxima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) {
					if (bo[i] > 0) continue;
					if (ba[j] == 0 ) continue;
					bo[i] = 255;
				}
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) {
					if (bo[i] > 0) continue;
					if (ba[j] == 0 ) continue;
					bo[i] = 255;
				}
			}
		}}
		break;
	default :
		return( FALSE );
	}
	return( TRUE );
}

#if defined(_ANSI_)
static int DilationNdg( t_ImagePtr theIm, t_ImagePtr resIm, e_ParamMorphoPtr par )
#else
static int DilationNdg( theIm, resIm, par )
t_ImagePtr theIm;
t_ImagePtr resIm;
e_ParamMorphoPtr par ;
#endif
{
	t_Image supIm;
	register int x, y, z, i, j;
	int dx, dy, dz, dx1, dy1, dz1, dxy;
  
	dx = theIm->tx;   dy = theIm->ty;   dz = theIm->tz;   dxy = dx*dy;
	dx1 = dx - 1;   dy1 = dy - 1;   dz1 = dz - 1;
	switch (par->connexite) {
	case CONNEXE_26 :
	case CONNEXE_10 :
		T_CopyHeaderIm( resIm, &supIm );
		supIm.buf = resIm->buf;
		break;
	case CONNEXE_06 :
		T_CopyHeaderIm( theIm, &supIm );
		supIm.buf = theIm->buf;
		break;
	}

	if ( theIm->type != resIm->type) return( FALSE );
	switch( theIm->type ) {
	case TYPEIM_256 :
		{
                unsigned char *bi, *bo, *ba;
		bi = (unsigned char*)(theIm->buf);
		bo = (unsigned char*)(resIm->buf);
		ba = (unsigned char*)(supIm.buf);
		/*--- maxima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ ) 
					bo[i] = (bi[j] > bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- )
					if (bi[j] > bo[i]) bo[i] = bi[j];
			}
		}
		/*--- maxima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- maxima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
			}
		}}
		break;
	case TYPEIM_16B :
		{
                unsigned short int *bi, *bo, *ba;
		bi = (unsigned short int*)(theIm->buf);
		bo = (unsigned short int*)(resIm->buf);
		ba = (unsigned short int*)(supIm.buf);
		/*--- maxima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ ) 
					bo[i] = (bi[j] > bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- )
					if (bi[j] > bo[i]) bo[i] = bi[j];
			}
		}
		/*--- maxima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- maxima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
			}
		}}
		break;
	case TYPEIM_S16B :
		{
                short int *bi, *bo, *ba;
		bi = (short int*)(theIm->buf);
		bo = (short int*)(resIm->buf);
		ba = (short int*)(supIm.buf);
		/*--- maxima computation along X ---*/
		if ( dx > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( y = 0; y < dy; y++ ) {
				i = z*dxy + y*dx;   j = i+1;
				for ( x = 0; x < dx1; x++, i++, j++ ) 
					bo[i] = (bi[j] > bi[i]) ? bi[j] : bi[i];
				bo[i] = bi[i];  /* i = z*dxy + y*dx + dx-1 */
				j = i-1;
				for ( x = dx1; x > 0; x--, i--, j-- )
					if (bi[j] > bo[i]) bo[i] = bi[j];
			}
		}
		/*--- maxima computation along Y ---*/
		if ( dy > 1 ) {
			for ( z = 0; z < dz; z++ ) 
			for ( x = 0; x < dx; x++ ) {
				i = z*dxy + x;   j = i+dx;
				for ( y = 0; y < dy1; y++, i+=dx, j+=dx ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
				j = i-dx;
				for ( y = dy1; y > 0; y--, i-=dx, j-=dx ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
			}
		}
		if (par->mode == MODE_2D) 
			break;
		/*--- maxima computation along Z ---*/
		if ( par->connexite == CONNEXE_10 ) {
			T_CopyHeaderIm( theIm, &supIm );
			supIm.buf = theIm->buf;
		}
		if ( dz > 1 ) {
			for ( y = 0; y < dy; y++ )
			for ( x = 0; x < dx; x++ ) {
				i = y*dx + x;   j = i+dxy;
				for ( z = 0; z < dz1; z++, i+=dxy, j+=dxy ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
				j = i-dxy;
				for ( z = dz1; z > 0; z--, i-=dxy, j-=dxy ) 
					if (ba[j] > bo[i]) bo[i] = ba[j];
			}
		}}
		break;
	default :
		return( FALSE );
	}
	return( TRUE );
}
