/* epi_reech.c 2.1 5/14/98 */
/*==========================================================================
 *
 * File : $EPI/src/epi_reech.c
 * Module : ?
 * $ Id : ? 
 * Test file : $EPI/test/test.im $EPI/test/test.tr
 * Author : J-P Thirion & Andre' Gueziec
 * Creation: March 9 1992
 *===========================================================================
 *   (c) Copyright 1992, Inria, all right reserved
 *===========================================================================
 *
 *------------------------------Description----------------------------------
 *
 * Fonctions de reechantillonage d'images
 *
 *-----------------------------Modification----------------------------------
 *
 * juillet 1993 : G. Malandain
 *                introduction du type short int : TYPEIM_S16B
 *                fonction modifiee :
 *                - _T_ReechPlanFunc
 */

/*==========================================================================*/
/*
	epi_reech.c

  Resample 3D images

  Resample a 3D image into another 3D image with a 3D geometric transform,
  using spline interpolation

  (c) Jean-Philippe THIRION & Andre' Gueziec -- INRIA -- 9/3/92
*/
/*==========================================================================*/

#if defined(INC_MIN) /* inclusions minimales */
 /* rajouter -DINC_MIN dans les options de compilation */
 /* GM 12/09/94                                        */
#include <epi_kernel.ei>
#include <epi_reech.ei>
#include <epi_var.ee>
#include <epi_spline.ee>
#else /* inclusions maximales */
#include <epidaure.h>
#include <epidaure.ee>
#include <epidaure.ei>
#endif

/*=======================================================================
 * int E_Reech3D(	t_ImagePtr theIm,
 *					t_Mat4Ptr trans,
 *					int type,
 *					t_ImagePtr resIm )
 *
 * Action :
 * Reechantillonage d'une image tri-dimensionnelle.
 *
 * Inclure dans l'ordre :
 * <epidaure.h>
 *
 * Parametres de sortie :
 * resIm : buffer d'image re'sultat
 *
 * Parametres d'entree :
 * theIm : Image a reechantillonner
 * trans : matrice de transformation en coordonnees homogenes
 * type : type de reechantillonnage (E_LINEAR ou E_SPLINE)
 *
 * Parametres d'entree/sortie :
 * aucun.
 *
 * Retour :
 * 0 si une erreur c'est produite, 1 si tout c'est bien passe.
 *
 * Erreurs :
 * pas assez de memoire pour l'image resultat
 * type de reechantillonnage inconnu
 *
 * Observations :
 * mettre la variable globale "verbose" a FALSE (0) pour ne pas avoir
 * le numero de plan traite' affiche' au cour du traitement. Pour le
 * mode E_SPLINE, l'image en entree est l'image des coefficients spline.
 *
 * Exemple :
 * t_Image theIm;
 * t_Mat4Ptr trans= NULL;
 *
 * T_FReadMat4( NULL, "transfo.file", &trans );
 * T_Init3Image( &theIm, 0,0,0,TYPEIM_UNKNOWN, "image.file" );
 * T_ReadInrim( &theIm );
 * E_Reech3D( &theIm, trans, E_LINEAR, &theIm );
 * T_WriteInrim( &theIm );
 * T_FreeImage( &theIm );
 * T_FreeMat4( &trans );
 *
 *========================================================================
 */

/*
  Simplified procedure for 3D image resampling
  */


#if defined(_ANSI_)
int E_Reech3D( t_ImagePtr theIm, t_Mat4Ptr trans, int type, t_ImagePtr resIm )
#else
int E_Reech3D(theIm, trans, type, resIm )
t_ImagePtr theIm;
t_Mat4Ptr trans;
int type;
t_ImagePtr resIm ;

#endif

{
  e_ReechPar rPar;
  
  E_ERR( theIm->buf == NULL, E_ERRNULL, FALSE, "E_Reech3D", 
	"input image not allocated" );
  
  E_DefaultReech( &rPar );
  rPar.type= type;
  rPar.trsf= *trans;
  rPar.verbose= TRUE;
  
  E_ReechAllPar( theIm, &rPar, resIm );
  E_ISERR( FALSE, "E_Reech3D",NULL);
  return( TRUE );
}

/*=======================================================================
 * int E_ReechAllPar(	t_ImagePtr theIm,
 *						e_ReechParPtr rPar,
 *						t_ImagePtr resIm )
 *
 * Action :
 * Reechantillonage d'une image tri-dimensionnelle, avec l'ensemble des
 * parametres de disponibles.
 *
 * Inclure dans l'ordre :
 * <epidaure.h>
 *
 * Parametres de sortie :
 * resIm : buffer d'image re'sultat
 *
 * Parametres d'entree :
 * theIm : Image a reechantillonner
 * rPar : parametres de reechantillonnages
 *
 * Parametres d'entree/sortie :
 * aucun.
 *
 * Retour :
 * 0 si une erreur c'est produite, 1 si tout c'est bien passe.
 *
 * Erreurs :
 * pas assez de memoire pour l'image resultat
 * type de reechantillonnage inconnu
 *
 * Observations :
 * initialiser les valeurs par defaut de rPar avec E_DefaultReech()
 *
 * Exemple :
 * t_Image theIm;
 * e_ReechPar rPar;
 * t_Mat4 trans;
 *
 * T_Init3Image( &theIm, 0,0,0,TYPEIM_UNKNOWN, "image.file" );
 * T_ReadInrim( &theIm );
 * ...
 * E_DefaultReech( &rPar );
 * rPar.trsf= trans;
 * E_ReechAllPar( &theIm, &rPar, &theIm );
 * T_WriteInrim( &theIm );
 * T_FreeImage( &theIm );
 *
 *========================================================================
 */

/* 
   High level function for 3D image resampling
   */

#if defined(_ANSI_)
int E_ReechAllPar( t_ImagePtr theIm, e_ReechParPtr rPar, t_ImagePtr resIm )
#else
int E_ReechAllPar(theIm, rPar, resIm )
t_ImagePtr theIm;
e_ReechParPtr rPar;
t_ImagePtr resIm ;

#endif

{
  ReechParPtr_func_type func= NULL;
  t_ReechPar par;
  
  E_ERR( theIm->buf == NULL, E_ERRNULL, FALSE, "E_Reech3D", 
	"input image not allocated" );
  
  switch( rPar->type ){
    
  case E_LINEAR: func= (ReechParPtr_func_type) T_Interp3D; break;
  case E_SPLINE: func= (ReechParPtr_func_type) T_InterSP3; break;
  default: 
    E_XERR(E_ERRUNKNOWN,FALSE,"E_Reech3D","wrong type of resampling");
    break;
  }
  
  par.verbose= rPar->verbose;
  _T_ReechFuncBegin( theIm, func, &par );
  _T_Reech3DFunc( theIm, &rPar->trsf, resIm, &par );
  _T_ReechFuncEnd( &par );
  
  E_ISERR( FALSE, "E_Reech3D",NULL);
  return( TRUE );
}

/*=======================================================================
 * int E_DefaultReech( e_ReechParPtr rPar )
 *
 * Action :
 * initialise avec les parametres par de'faut pour le reechantillonnage
 *
 * Inclure dans l'ordre :
 * <epidaure.h>
 *
 * Parametres de sortie : (aucun)
 * Parametres d'entree : (aucun)
 * Parametres d'entree/sortie : rPar
 *
 * Retour : (aucun)
 * Erreurs : (aucune)
 *
 *========================================================================
 */

/*
  Initialize e_ReechPar parameters with default values
  */

#if defined(_ANSI_)
void E_DefaultReech( e_ReechParPtr rPar )
#else
void E_DefaultReech(rPar )
e_ReechParPtr rPar ;

#endif

{
  T_4MIdentity( &rPar->trsf );
  rPar->type= E_LINEAR;
  rPar->verbose= FALSE;
}

/*=======================================================================
 * int E_ReechPlan(	t_ImagePtr theIm,
 *					t_Mat4Ptr trans,
 *					int type,
 *					reeltype offset,
 *					t_ImagePtr resIm )
 *
 * Action :
 * Reechantillonage d'une image plane a partir d'une image tri-dimensionnelle.
 *
 * Inclure dans l'ordre :
 * <epidaure.h>
 *
 * Parametres de sortie :
 * resIm : Image 2D re'sultat
 *
 * Parametres d'entree :
 * theIm : Image a reechantillonner
 * trans : matrice de transformation en coordonnees homogenes
 * type : type de reechantillonnage (E_LINEAR ou E_SPLINE)
 * offset : offset du plan par rapport a l'origine, dans la direction de
 * l'axe z.
 *
 * Parametres d'entree/sortie :
 * aucun.
 *
 * Retour :
 * 0 si une erreur c'est produite, 1 si tout c'est bien passe.
 *
 * Erreurs :
 * pas assez de memoire pour allouer l'image resultat
 * type de reechantillonnage inconnu
 *
 * Observations :
 * Si l'image resultat est une image 3D, alors c'est le plan d'image
 * correspondant a la partie entiere de l'offset qui est charge. Pour
 * le mode E_SPLINE, l'image en entree est une image des coefficients
 * splines
 *
 * Exemple :
 * t_Image theIm;
 * t_Image plan;
 * t_Mat4Ptr trans= NULL;
 *
 * T_FReadMat4( NULL, "transfo.file", &trans );
 * T_Init3Image( &theIm, 0,0,0,TYPEIM_UNKNOWN, "image.file" );
 * T_ReadInrim( &theIm );
 * T_InitImage( &plan, theIm.tx, theIm.ty, theIm.type, "plan.file" );
 * E_ReechPlan( &theIm, trans, E_LINEAR, theIm.tz/2.0,&plan );
 * T_WriteInrim( &plan );
 * T_FreeImage( &theIm );
 * T_FreeImage( &plan );
 * T_FreeMat4( &trans );
 *
 *========================================================================
 */

/*
  Resample a single slice from a 3D image
  */

#if defined(_ANSI_)
int E_ReechPlan( t_ImagePtr theIm, t_Mat4Ptr trans, int type, reeltype offset, t_ImagePtr resIm )
#else
int E_ReechPlan(theIm, trans, type, offset, resIm )
t_ImagePtr theIm;
t_Mat4Ptr trans;
int type;
reeltype offset;
t_ImagePtr resIm ;

#endif

{
  t_Image auxIm;
  t_ImagePtr anIm;
  t_ReechPar par;
  ReechParPtr_func_type func= NULL;
  
  E_ERR( theIm->buf == NULL, E_ERRNULL, FALSE, "E_ReechPlan", 
	"input image not allocated" );
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
    E_ISERR( FALSE, "E_ReechPlan",NULL);
    anIm= &auxIm;
  }
  else anIm= resIm;
  
  switch( type ){
    
  case E_LINEAR: func= (ReechParPtr_func_type) T_Interp3D; break;
  case E_SPLINE: func= (ReechParPtr_func_type) T_InterSP3; break;
  default: 
    E_XERR(E_ERRUNKNOWN,FALSE,"E_ReechPlan","wrong type of resampling");
    break;
  }
  
  _T_ReechFuncBegin( theIm, func, &par );
  _T_ReechPlanFunc(theIm,trans,offset,(int) offset,anIm,&par);
  _T_ReechFuncEnd( &par );
  
  E_ISERR( FALSE, "E_ReechPlan",NULL);
  
  if ( anIm != resIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  
  return( TRUE );
}

/*----------------------------------------------------------------------*/
/*
  Resample an image using an interpolation function
*/
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void _T_Reech3DFunc( t_ImagePtr theIm, t_Mat4Ptr tr, t_ImagePtr resIm, t_ReechParPtr par )
#else
void _T_Reech3DFunc(theIm, tr, resIm, par )
t_ImagePtr theIm;
t_Mat4Ptr tr;
t_ImagePtr resIm;
t_ReechParPtr par ;

#endif

{
  int k;
  t_Image auxIm;
  t_ImagePtr anIm;
  
  if ( (resIm->buf == NULL) || (resIm->buf == theIm->buf) ){
    T_CopyHeaderIm( resIm, &auxIm );
    T_AllocImage( &auxIm );
    E_ISERRN( "_T_Reech3DFunc", NULL );
    anIm= &auxIm;
  }
  else anIm= resIm;
  
  
  for( k=0; k< anIm->tz; k++ ){
    
    if ( par->verbose ){ 
      sprintf( tmpStr1, " plane number %d\r", k );
      T_Mess( tmpStr1, NULL );
    }
    
    _T_ReechPlanFunc( theIm, tr, (reeltype) k, k, anIm, par );
    E_REPERR( T_FreeImage(anIm) );
    E_ISERRN( "_T_Reech3DFunc", NULL );
  }
  
  if ( anIm != resIm ){
    T_FreeImage( resIm );
    resIm->buf= anIm->buf;
  }
  
}


/*---------------------------------------------------------------------------*/
/*
  Resample a 2D image plane from a 3D image.

  It uses a 3D transform, and the offset along to the Z direction.
  The plane numbered "plane" of the
  result image is loaded with the result. For a 2D image, plane should
  be set to zero. For the computation of a 3D image plane by plane, 
  "plane" must be equal to offset. The result image must be 
  different from the input image.
  */
/*---------------------------------------------------------------------------*/

  
#if defined(_ANSI_)
  void _T_ReechPlanFunc(
t_ImagePtr theIm,
t_Mat4Ptr tr,
reeltype offset, /* offset along to the z axis */
int plane,  /* number of the plane in the result image */
t_ImagePtr resIm,
t_ReechParPtr par
)
#else
  void _T_ReechPlanFunc(theIm, tr, offset, plane, resIm, par )
t_ImagePtr theIm;
t_Mat4Ptr tr;
reeltype offset;
int plane;
t_ImagePtr resIm;
t_ReechParPtr par ;

#endif

{
  int i,j;
  t_3Point pt;
  reeltype value;
  int pos;
  
  unsigned char *ubyte_pt= NULL;
  unsigned short *ushort_pt= NULL;
  short int *sshort_pt= NULL;
  reeltype *float_pt= NULL;
  
  E_ERRN( (theIm==NULL) || (theIm->buf==NULL), E_ERRNULL,
	 "_T_ReechPlanFunc", "input image not allocated" );
  
  E_ERRN( (resIm==NULL) || (resIm->buf==NULL), E_ERRNULL,
	 "_T_ReechPlanFunc", "output image not allocated" );
  
  E_ERRN( (theIm->buf==resIm->buf), E_ERRFATAL,
	 "_T_ReechPlanFunc", 
	 "output image must be different from input image" );
  /*
    for( j=0; j< resIm->ty; j++ )
    for( i=0; i< resIm->tx; i++ ){
    
    pt.x= (reeltype) i;
    pt.y= (reeltype) j;
    pt.z= offset;
    
    T_4MApply( &pt, &pt, tr );
    
    if (    (pt.x >= 0) && (pt.x <= theIm->tx-1) &&
    (pt.y >= 0) && (pt.y <= theIm->ty-1) &&
    (pt.z >= 0) && (pt.z <= theIm->tz-1) ){
    
    value= par->func( pt.x, pt.y, pt.z, theIm, par );
    
    if ( resIm->type != TYPEIM_FLOAT ){
    if ( value < 0. ) value= 0.;
    else if ( value > 1.0 ) value= 1.0;
    }
    
    T_Set3Point( i,j,plane, resIm, value, MODE_SET );
    }
    }
    */ 
  plane = plane % resIm->tz;
  pos= plane * resIm->tx * resIm->ty;
  
  switch(resIm -> type) {
  case TYPEIM_256: ubyte_pt=((unsigned char *)(resIm->buf))+pos; break;
  case TYPEIM_16B: ushort_pt=((unsigned short *)(resIm->buf))+pos;break;
  case TYPEIM_S16B: sshort_pt=((short int*)(resIm->buf))+pos;break;
  case TYPEIM_FLOAT: float_pt=((reeltype *) (resIm->buf))+pos; break;
  default: break;
  }
  
  for( j=0; j< resIm->ty; j++) 
    for( i=0; i< resIm->tx; i++ ){
      
      pt.x= (reeltype) i;
      pt.y= (reeltype) j;
      pt.z= offset;
      
      T_4MApply( &pt, &pt, tr );
      
      if ((pt.x >= 0.0) && (pt.x <= theIm->tx-1) &&
	  (pt.y >= 0.0) && (pt.y <= theIm->ty-1) &&
	  (pt.z >= 0.0) && (pt.z <= theIm->tz-1)) {
	
	value= par->func( pt.x, pt.y, pt.z, theIm, par );
	
	if ( resIm->type != TYPEIM_FLOAT ){
	  if ( value < 0. ) value= 0.;
	  else if ( value > 1.0 ) value= 1.0;
	}
	
	switch(resIm->type) {
	  
	case TYPEIM_256: *ubyte_pt= (unsigned char) (value*255.0); break;
	case TYPEIM_16B: *ushort_pt= (unsigned short) (value*65535.0); break;
	case TYPEIM_S16B: *sshort_pt= (short int) (value*32767.0); break;
	case TYPEIM_FLOAT: *float_pt= value; break;
	  
	default:
	  T_Set3Point( i,j,plane, resIm, value, MODE_SET ); break;
	}
      }
      
      switch( resIm->type ){
      case TYPEIM_256: ubyte_pt += 1; break;
      case TYPEIM_16B: ushort_pt += 1; break;
      case TYPEIM_S16B: sshort_pt += 1; break;
      case TYPEIM_FLOAT: float_pt += 1; break;
      }
    }
}


/*---------------------------------------------------------------------------*/
/*
  Set the globals and perform precomputations
  */ 
/*---------------------------------------------------------------------------*/
  
#if defined(_ANSI_)
void _T_ReechFuncBegin( 
t_ImagePtr theIm, /* image, or spline control point */
ReechParPtr_func_type func, 
t_ReechParPtr par 
)
#else
  void _T_ReechFuncBegin(theIm, func, par )
t_ImagePtr theIm;
ReechParPtr_func_type func;
t_ReechParPtr par ;

#endif

{
  if ( func == (ReechParPtr_func_type) T_InterSP3 ){
    par->SPline.De= 4;
    ES_InitSpline(	&par->SPline, 
		  theIm->tx, theIm->ty, theIm->tz,
		  theIm->tx, theIm->ty, theIm->tz );
  }
  par->func= func;
}

/*----------------------------------------------------------------------*/
/*
  Free structures used for re-sampling
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void _T_ReechFuncEnd( t_ReechParPtr par )
#else
void _T_ReechFuncEnd(par )
t_ReechParPtr par ;

#endif

{
  if ( par->func == (ReechParPtr_func_type) T_InterSP3 ) ES_FreeSpline( &par->SPline );
  par->func= NULL;
}

/*----------------------------------------------------------------------*/
/*
  Perform the spline resampling
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
reeltype T_InterSP3( reeltype x, reeltype y, reeltype z, t_ImagePtr theIm, t_ReechParPtr par )
#else
reeltype T_InterSP3(x, y, z, theIm, par )
reeltype x;
reeltype y;
reeltype z;
t_ImagePtr theIm;
t_ReechParPtr par ;

#endif

{
  reeltype value;
  
  ES_ContinuousF( &par->SPline, theIm, x, y, z, &value );
  
  /*
    t_3Derivates theDer;
    ES_computeDerivates( &par->SPline, &theDer, (float *) theIm->buf,
    xx, yy, zz, theIm->tx, theIm->ty, theIm->tz );
    ES_ContinuousFDerivates( &par->SPline, &theDer, (float *) theIm->buf,
    x, y, z, theIm->tx, theIm->ty, theIm->tz );
    value= theDer.f;
    */
  
  return( value );
}

/*----------------------------------------------------------------------*/
/*
  Test program for the 3D image resampling
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_TestReech3D( void )
#else
void T_TestReech3D()

#endif

{
  t_Image theIm;
  t_Mat4Ptr trans= NULL;
  t_ListPtr matList= NULL;
  e_ReechPar rPar;
  
  T_FReadList( (FILE *) NULL, "test.tr", &matList );
  T_ListItemErr( matList, (t_ObjHand) &trans, OBJ_MAT4 );
  
  E_ISERRN( "T_TestReech3D", NULL );
  
  T_Init3Image( &theIm, 0,0,0,TYPEIM_UNKNOWN, "test.im" );
  T_ReadInrim( &theIm );
  E_ISERRN( "T_TestReech3D", NULL );
  
  T_SetImName( &theIm, "tmp.im");			/* apply direct transform	*/
  E_DefaultReech( &rPar );
  rPar.type= E_LINEAR;
  rPar.verbose= TRUE;
  rPar.trsf= *trans;
  E_ReechAllPar( &theIm, &rPar, &theIm );
  T_WriteInrim( &theIm );
  
  T_SetImName( &theIm, "tmp2.im" );		/* invert the transform		*/
  T_4MInvert( &rPar.trsf, &rPar.trsf );
  E_ReechAllPar( &theIm, &rPar, &theIm );
  T_WriteInrim( &theIm );
  
  T_FreeImage( &theIm );
  T_FreeList( &matList );
  E_ISERRN( "T_TestReech3D", NULL );
}

/*----------------------------------------------------------------------*/
/*
  Test program for 3D image resampling with splines
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_TestReech3DSpline( void )
#else
void T_TestReech3DSpline()

#endif

{
  t_Image theIm;
  t_Image theCtrl;
  t_Mat4Ptr trans= NULL;
  t_ListPtr matList = NULL;
  
  T_Init3Image( &theIm, 0,0,0,TYPEIM_UNKNOWN, "test.im" );
  T_Init3Image( &theCtrl, 0,0,0,TYPEIM_UNKNOWN, "tmp.ctrl" );
  
  T_FReadList( (FILE *) NULL, "test.tr", &matList );
  T_ListItemErr( matList, (t_ObjHand) &trans, OBJ_MAT4 );
  E_REPERR( goto END_TestReech3DSpline );
  
  T_ReadInrim( &theIm );
  E_REPERR( goto END_TestReech3DSpline );
  
  T_Init3Image( &theCtrl, theIm.tx, theIm.ty, theIm.tz,
	       TYPEIM_FLOAT, "tmp.ctrl" );
  T_AllocImage( &theCtrl );
  E_REPERR( goto END_TestReech3DSpline );
  
  /* compute control points	*/
  
  T_Mess( "Compute spline control points\n", NULL );
  E_CoefSpline( &theIm, &theCtrl, (float) 0.01, (float) 0.01, (float) 0.01 );
  T_FreeImage( &theIm );
  E_REPERR( goto END_TestReech3DSpline );
  
  /* apply direct transform	*/
  
  T_Mess( "Apply direct transform\n", NULL );
  T_SetImName( &theIm, "tmp.im");
  E_Reech3D( &theCtrl, trans, E_SPLINE, &theIm );
  E_REPERR( goto END_TestReech3DSpline );
  T_WriteInrim( &theIm );
  
  /* compute control points	*/
  
  T_Mess( "Compute spline control points\n", NULL );
  E_CoefSpline( &theIm, &theCtrl, (float) 0.01, (float) 0.01, (float) 0.01 );
  T_FreeImage( &theIm );
  E_REPERR( goto END_TestReech3DSpline );
  
  /* apply invert transform	*/
  
  T_Mess( "Apply inverse transform\n", NULL );
  T_SetImName( &theIm, "tmp2.im" );
  T_4MInvert( trans, trans );
  E_Reech3D( &theCtrl, trans, E_SPLINE, &theIm );
  E_REPERR( goto END_TestReech3DSpline );
  T_WriteInrim( &theIm );
  
  T_Mess( "End of the computation\n", NULL );
  
 END_TestReech3DSpline:
  
  T_FreeImage( &theIm );
  T_FreeImage( &theCtrl );
  T_FreeMat4( &trans );
  
  E_ISERRN( "T_TestReech3DSpline", NULL );
}

/*----------------------------------------------------------------------*/
/*
  Test program for the plane image resampling
  */
/*----------------------------------------------------------------------*/

#if defined(_ANSI_)
void T_TestReechPlan( void )
#else
void T_TestReechPlan()

#endif

{
  t_Image theIm;
  t_Image plan;
  t_Mat4Ptr trans= NULL;
  
  T_FReadMat4( (FILE *) NULL, "test.tr", &trans );
  E_ISERRN( "T_TestReechPlan", NULL );
  
  T_Init3Image( &theIm, 0,0,0,TYPEIM_UNKNOWN, "test.im" );
  T_ReadInrim( &theIm );
  
  T_InitImage( &plan, theIm.tx, theIm.ty, theIm.type, "tmp.im" );
  E_ISERRN( "T_TestReechPlan", NULL );
  
  E_ReechPlan( &theIm, trans, E_LINEAR, (reeltype) (theIm.tz/2.0),&plan );
  
  T_WriteInrim( &plan );
  T_FreeImage( &theIm );
  T_FreeImage( &plan );
  T_FreeMat4( &trans );
}

/*----------------------------------------------------------------------*/

