#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "imagedata/IntensityProfile.h"
#ifdef WIN32
#include <windows.h>
#include <float.h>
#endif
#ifdef GRAPHICS
#include <GL/gl.h>
#endif
#include <inrimage/ImageGeometry.h>
#include <imagedata/IntensityProfile.h>

#include <algorithm>

namespace yav {

ImageVoxel::ImageVoxel(int *_pos, const Vec3<double>& subpos, double _value, double _distance, unsigned int _index)
{
	vpos[0] = _pos[0];
	vpos[1] = _pos[1];
	vpos[2] = _pos[2];
	spos = subpos;
	value = _value;
	distance = _distance;
	index = _index;
}

ImageVoxel::~ImageVoxel()
{}

IntensityProfile::IntensityProfile(double max, bool (IntensityProfile::*_stopMethod)(ImageVoxel *, Voxel&))
  : inr(0), maxDistance(max), startvoxel(0), stopMethod(_stopMethod)
{}


IntensityProfile::~IntensityProfile()
{
	if(startvoxel)
		clear();
}

ImageVoxel * IntensityProfile::getVoxel(int i) const
{
	if(i >= 0)
		return positive[i];
	else
		return negative[-i-1];
}

ImageVoxel * IntensityProfile::operator[](int i) const
{
	return(getVoxel(i));
}

void IntensityProfile::clear(void)
{
	delete startvoxel;
	startvoxel = 0;
	std::vector<ImageVoxel *>::iterator it;
	for(it = positive.begin(); it != positive.end(); it++) delete *it;
	positive.erase(positive.begin(), positive.end());
	for(it = negative.begin(); it != negative.end(); it++) delete *it;
	negative.erase(negative.begin(), negative.end());
}

double IntensityProfile::getIntensity(double d)
{
	std::vector<ImageVoxel *>::iterator it;
	double l, val, previousDistance, previousVal, currentDistance;
	bool found;

	if (!startvoxel)
		return 0.0;

	if ((d<getMinVoxelDistance()) || (d>getMaxVoxelDistance()))
		return 0.0;

	else if (d>=0.0)
	{
		found = false;
		previousDistance = startvoxel->getDistance();
		previousVal=startvoxel->getValue();

		if (d==0.0)
			return previousVal;

		for(it = positive.begin(); (found==false) && (it != positive.end()); it++) 
		{
			currentDistance = (*it)->getDistance();

			if (d==currentDistance)
			{
				found = true;
				val = (*it)->getValue();
			}
			else if (d<currentDistance)
			{
				l = (d-previousDistance)/(currentDistance-previousDistance);
				val = previousVal+l*((*it)->getValue()-previousVal);
				found = true;
			}
	
			previousDistance = currentDistance;
			previousVal = (*it)->getValue();
		}

		if (found==false)
			return 0.0;
	}
	else
	{
		d = -d;
		found = false;
		previousDistance = startvoxel->getDistance();
		previousVal=startvoxel->getValue();

		for(it = negative.begin(); (found==false) && (it != negative.end()); it++) 
		{
			currentDistance = (*it)->getDistance();

			if (d==currentDistance)
			{
				found = true;
				val = (*it)->getValue();
			}
			else if (d<currentDistance)
			{
				l = (d-previousDistance)/(currentDistance-previousDistance);
				val = previousVal+l*((*it)->getValue()-previousVal);
				found = true;
			}

			previousDistance = currentDistance;
			previousVal = (*it)->getValue();
	  	}

		if (found==false)
			return 0.0;
	}

	return val;
}

bool IntensityProfile::stopFromDistance(ImageVoxel *voxel, Voxel& iv)
{
	if(voxel->getDistance()>maxDistance)
		return true;
	else
		return false;
}
double IntensityProfile::getMinVoxelDistance() const{
	if (negative.size()>0)
		return -negative[negative.size()-1]->getDistance();
	else if (positive.size()>0)
		return positive[0]->getDistance();
	else 
		return 0.0;
}
double IntensityProfile::getMaxVoxelDistance() const{
	if (positive.size()>0)
		return positive[positive.size()-1]->getDistance();
	else if (negative.size()>0)
		return -negative[0]->getDistance();
	else 
		return 0.0;
}
double IntensityProfile::getGradient(int i)
{
	if (i>=0)
		return positiveGradient[i];
	else 
		return negativeGradient[-i-1];
}

double IntensityProfile::getGradient(double d)
{
	unsigned int minSize, maxSize, i;
	double previousDistance, previousVal, currentDistance, l;

	minSize= negative.size();
	maxSize= positive.size();
	
	if ((d<getMinVoxelDistance()) || (d>getMaxVoxelDistance()))
		return 0.0;
	else if (d>=0.0)
	{
		i = 1;
		previousDistance = 0.0;
		previousVal = positiveGradient[0];

		do
		{
			currentDistance = positive[i]->getDistance();

			if (d<=currentDistance)
			{
				l = (d-previousDistance)/(currentDistance-previousDistance);
				return previousVal+l*(positiveGradient[i]-previousVal);
			}
	
			previousVal = positiveGradient[i];
			previousDistance = currentDistance;
			i++;
		}
		while ((i<maxSize));
	}
	else
	{
		i = 0;
		previousDistance = 0.0;
		previousVal = positiveGradient[0];
	
		do
		{
			currentDistance = -negative[i]->getDistance();

			if (d>=currentDistance)
			{
				l = (d-previousDistance)/(currentDistance-previousDistance);
				return previousVal+l*(negativeGradient[i]-previousVal);
			}

			previousVal=negativeGradient[i];
			previousDistance=currentDistance;
			i++;
		}
		while ((i<minSize));
	}
	// we should never access this line, this is only to suppress a compilation warning.
	return 0.0;
}
double IntensityProfile::computeGradientAtIndex(const int index,const double sigma) {
	int j, n;
	double val, referenceDistance, x, y, sumxy, sumx, sumy, sumxx;
	int minIndex = -(int)(negative.size()+1);
	int	maxIndex = (int) positive.size();

	referenceDistance = getDistance(index);
	n = 1;
	sumy = getVoxel(index)->getValue();
	sumx = sumxx = sumxy = 0.0;
	j = index-1;

	while (j>minIndex)
	{
		x = getDistance(j);

		if (fabs(x-referenceDistance)<sigma)
		{
			sumx += x;
			y = getVoxel(j)->getValue();
			sumy += y;
			sumxy += x*y;
			sumxx += x*x;
			n++;
			j--;
		}
		else
			j=minIndex;
	}

	j=index+1;

	while (j<maxIndex)
	{
		x = getDistance(j);

		if (fabs(x-referenceDistance)<sigma)
		{
			sumx += x;
			y = getVoxel(j)->getValue();
			sumy += y;
			sumxy += x*y;
			sumxx += x*x;
			n++;
			j++;
		}
		else
			j=maxIndex;
	}

	val = (n*sumxy-sumx*sumy)/(n*sumxx-sumx*sumx);
	return val;
}


double IntensityProfile::computeGradientAtDistance(const double distance,const double sigma) {
std::vector<ImageVoxel *>::iterator it;
	double l, val, previousDistance, previousVal,currentDistance,d;
	bool found;
	int index;
	d=distance;

	if (!startvoxel)
	{
		std::cout << "WARNING: Intensity profile does not have starting voxel (set to 0)" << std::endl;
		return 0.0;
	}

	if (fabs(d)>maxDistance)
		return 0.0;

	else if (d>=0.0)
	{
		found = false;
		previousDistance = 0.0;

		if (d==0.0)
			return computeGradientAtIndex(0,sigma);

		for(index=0,it = positive.begin(); (found==false) && (it != positive.end()); it++,index++) 
		{
			currentDistance = (*it)->getDistance();

			if (d==currentDistance)
			{
				found = true;
				val =computeGradientAtIndex(index,sigma);
			}
			else if (d<currentDistance)
			{
				l = (d-previousDistance)/(currentDistance-previousDistance);
				previousVal=computeGradientAtIndex(index-1,sigma);
				val = previousVal+l*(computeGradientAtIndex(index,sigma)-previousVal);
				found = true;
			}
			previousDistance = currentDistance;
		}

		if (found==false)
			return 0.0;
	}
	else
	{
		d = -d;
		found = false;
		previousDistance = 0.0;

		for(index= -1,it = negative.begin(); (found==false) && (it != negative.end()); it++,index--) 
		{
			currentDistance = (*it)->getDistance();

			if (d==currentDistance)
			{
				found = true;
				val = (*it)->getValue();
			}
			else if (d<currentDistance)
			{
				l = (d-previousDistance)/(currentDistance-previousDistance);
				previousVal=computeGradientAtIndex(index+1,sigma);
				val = previousVal+l*(computeGradientAtIndex(index,sigma)-previousVal);
				found = true;
			}

			previousDistance = currentDistance;
	  	}

		if (found==false)
			return 0.0;
	}

	return val;
}

void IntensityProfile::computeGradient(double sigma) 
{
	int minIndex, maxIndex;
	int i, j, n;
	double val, referenceDistance, x, y, sumxy, sumx, sumy, sumxx;

	minIndex = -(int)(negative.size()+1);
	maxIndex = positive.size();
	negativeGradient.resize(negative.size());
	positiveGradient.resize(positive.size());

	for(i = minIndex+1; i<maxIndex; ++i)
	{
		
		referenceDistance = getDistance(i);
		n = 1;
		sumy = getVoxel(i)->getValue();
		sumx = sumxx = sumxy = 0.0;
		j = i-1;

		while (j>minIndex)
		{
			x = getDistance(j);

			if (fabs(x-referenceDistance)<sigma)
			{
				sumx += x;
				y = getVoxel(j)->getValue();
				sumy += y;
				sumxy += x*y;
				sumxx += x*x;
				n++;
				j--;
			}
			else
				j=minIndex;
		}

		j=i+1;

		while (j<maxIndex)
		{
			x = getDistance(j);

			if (fabs(x-referenceDistance)<sigma)
			{
				sumx += x;
				y = getVoxel(j)->getValue();
				sumy += y;
				sumxy += x*y;
				sumxx += x*x;
				n++;
				j++;
			}
			else
				j=maxIndex;
		}

		val = (n*sumxy-sumx*sumy)/(n*sumxx-sumx*sumx);
		
//		val=computeGradientAtIndex(i,sigma);

		if (i<0)
			negativeGradient[-i-1] = val;
		else 
			positiveGradient[i] = val;
	}
#ifdef old
	int minIndex, maxIndex, i, j;
	double variance, value, weight, distance, referenceDistance;

	variance = 2.0*sigma*sigma;
	minIndex = 0-(negative.size());
	maxIndex = positive.size();
	negativeGradient.resize(negative.size());
	positiveGradient.resize(positive.size());

	for(i = minIndex+1; i<maxIndex; ++i)
	{
		j = i-1;
		value = 0.0;
		referenceDistance = getDistance(i);
		distance = 0.0;

		while (j>minIndex)
		{
			distance = getDistance(j)-referenceDistance;
			weight = -distance*exp(-(distance*distance)/variance);
			value += weight*distance;
			j--;
		}

		while (distance > -2.0*sigma)
		{
			distance -= sigma/3.0;
			weight = -distance*exp(-(distance*distance)/variance);
			value += weight*distance;
		}

		distance = 0.0;
		j = i+1;

		while (j<maxIndex)
		{
			distance = getDistance(j)-referenceDistance;
			weight = -distance*exp(-(distance*distance)/variance);
			value += weight*distance;
			j++;
		} 

		while (distance < 2.0*sigma)
		{
			distance += sigma/3.0;
			weight = -distance*exp(-(distance*distance)/variance);
			value += weight*distance;
		}

		if (i<0)
			negativeGradient[-i] = 2.0*value/variance;
		else 
			positiveGradient[i] = 2.0*value/variance;
	}
#endif
}

double IntensityProfile::getDistance(int i)
{
	if (i>=0)
		return positive[i]->getDistance();
	else
		return -negative[-i-1]->getDistance();
}

void IntensityProfile::compute(double _maxDistance, const Vec3<double>& org, const Vec3<double>& dir, SIDE side)
{
	double max, dist, a,dBorder;
	ImageVoxel *current;
	bool inImage, stop;
	int vpos[3], i;
	
	if(startvoxel)
		clear();

	if(_maxDistance>0.0)
		max = maxDistance = _maxDistance;
	else
		max = maxDistance;

	startvoxel = 0;
	Vec3<double> pos(org);
	origin = pos;
	direction = dir;
	ImageGeometry *converter = inr->newConverter();
	converter->set(origin.getX(), origin.getY(), origin.getZ());
	dist = 0.0;
	converter->toImage(vpos);
	Vec3<double> dp(direction);

	if(dp.getX() == 0.0)
		dp[0] = ImageGeometry::epsilon;

	if(dp.getY() == 0.0)
		dp[1] = ImageGeometry::epsilon;

	if(dp.getZ() == 0.0)
		dp[2] = ImageGeometry::epsilon;

	if(converter->in())
	{
		startvoxel = new ImageVoxel(vpos, pos, converter->interpolate(), 0.0, 0);
		start(converter->getVoxel());
		current = new ImageVoxel(vpos, pos, converter->interpolate(), 0.0, 0);
		positive.push_back(current);
		stop = (this->*stopMethod)(current, converter->getVoxel());
	}
	else {
		std::cout << "WARNING: Intensity profile coordinates out of image bounds (x=" << origin.getX() << ", y=" << origin.getY() << ", z=" << origin.getZ() << ")" << std::endl;
		dBorder=converter->nextBorderVoxel(dp);

	}


	// scan normal side
	if(side==NORMAL || side==BOTH)
	{
		i = 0;
		stop = false;
		dist = 0.0;
		if (startvoxel==0) {
			// the forward intensity profile intersect the image
			if (dBorder>0) {
				pos += dBorder * dp;
				dist += dBorder;
				if ((dist>max) || (!converter->in()))
					stop=true;
				else {
					converter->toImage(vpos);
					startvoxel = new ImageVoxel(vpos, pos, converter->interpolate(), dBorder, 0);
					start(converter->getVoxel());
					current = new ImageVoxel(vpos, pos, converter->interpolate(), dBorder, 0);
					positive.push_back(current);
					stop = (this->*stopMethod)(current, converter->getVoxel());
				}

			} else
				stop=true;
		}
	
		while(!stop)
		{
			a = converter->nextVoxel(dp);
			pos += a * dp;
			dist += a;
			inImage=converter->in();
			stop = (dist>max) || (startvoxel && !inImage);

			if ((!stop) && (inImage))
			{
				converter->toImage(vpos);
				current = new ImageVoxel(vpos, pos, converter->interpolate(), dist, ++i);
				positive.push_back(current);
				stop = (this->*stopMethod)(current, converter->getVoxel());
			}
		}
	}
	// scan inverse normal side
	if(side==INVERSE || side==BOTH)
	{
		pos = origin;
		dp *= -1.0;
		i = 0;
		stop = false;
		dist = 0.0;
		if (startvoxel==0) {
			// the backward intensity profile intersect the image
			if (dBorder<0) {
				pos -= dBorder * dp;
				dist -= dBorder;
				if ((dist>max) || (!converter->in()))
					stop=true;
				else {
					converter->toImage(vpos);
					startvoxel = new ImageVoxel(vpos, pos, converter->interpolate(), -dBorder, -1);
					start(converter->getVoxel());
					current = new ImageVoxel(vpos, pos, converter->interpolate(), -dBorder, -1);
					negative.push_back(current);
					stop = (this->*stopMethod)(current, converter->getVoxel());
					i= -1;
				}

			} else
				stop=true;
		} else
			converter->set(origin.getX(), origin.getY(), origin.getZ());

	
		while(!stop)
		{
			a = converter->nextVoxel(dp);
			pos += a * dp;
			dist += a;	
			inImage = converter->in();
			stop = (dist>max) || (startvoxel && !inImage);

			if ((!stop) && (inImage))
			{
				converter->toImage(vpos);
				current = new ImageVoxel(vpos, pos, converter->interpolate(), dist, --i);
				negative.push_back(current);
				stop = (this->*stopMethod)(current, converter->getVoxel());
			}
		}
	}

	delete converter;
}
#ifdef GRAPHICS
void IntensityProfile::glSet(void) const
{
	float x0, y0, z0, x1, y1, z1, x2, x3, y2, y3, X0;
	double r0, r1, theta0, theta1;
	int i;
	ImageVoxel * im;
	
	glDisable(GL_LIGHTING);
	glLineWidth(1.0);
	glColor3f(1.0, 1.0, 1.0);

	for(i = -((int)negativeArraySize()-1); i<(int)positiveArraySize(); i++)
	{
		im = operator[](i);

		switch(inr->getGeometry())
		{
			case Inrimage::CARTESIAN:
				x0 = ((float)im->getX())*inr->getVX();
				x1 = ((float)(im->getX()+1.0))*inr->getVX();
				y0 = ((float)im->getY())*inr->getVY();
				y1 = ((float)(im->getY()+1.0))*inr->getVY();
				z0 = ((float)im->getZ())*inr->getVZ();
				z1 = ((float)(im->getZ()+1.0))*inr->getVZ();
				glBegin(GL_LINE_LOOP);
				glVertex3f(x0, y0, z0);
				glVertex3f(x1, y0, z0);
				glVertex3f(x1, y1, z0);
				glVertex3f(x0, y1, z0);
				glEnd();
				glBegin(GL_LINE_LOOP);
				glVertex3f(x0, y0, z1);
				glVertex3f(x1, y0, z1);
				glVertex3f(x1, y1, z1);
				glVertex3f(x0, y1, z1);
				glEnd();
				glBegin(GL_LINES);
				glVertex3f(x0, y0, z0);
				glVertex3f(x0, y0, z1);
				glVertex3f(x1, y0, z0);
				glVertex3f(x1, y0, z1);
				glVertex3f(x1, y1, z0);
				glVertex3f(x1, y1, z1);
				glVertex3f(x0, y1, z1);
				glVertex3f(x0, y1, z0);
				glEnd();
				break;
			case Inrimage::CYLINDRICAL:
				X0 = (float)(inr->getX()*inr->getVX()/2.0);
				theta0 = ((double)im->getZ()-0.5)*M_PI/(double)inr->getZ();
				theta1 = theta0+(M_PI/(double) inr->getZ());

				if(im->getX()>inr->getX()/2-1)
					r0 = im->getX() * inr->getVX() - X0;
				else
				{
					theta0 += (float)M_PI;
					theta1 += (float)M_PI;
					r0 = X0-(im->getX()+1)*inr->getVX();
				}

				r1 = r0+inr->getVX();
				x0 = (float)(X0-r0*sin(theta0));
				y0 = (float)(X0+r0*cos(theta0));
				x1 = (float)(X0-r1*sin(theta0));
				y1 = (float)(X0+r1*cos(theta0));
				x2 = (float)(X0-r1*sin(theta1));
				y2 = (float)(X0+r1*cos(theta1));
				x3 = (float)(X0-r0*sin(theta1));
				y3 = (float)(X0+r0*cos(theta1));
				z0 = ((float)(im->getY())*inr->getVY());
				z1 = ((float)((im->getY()+1.0))*inr->getVY());
				glBegin(GL_LINE_LOOP);
				glVertex3f(x0, y0, z0);
				glVertex3f(x1, y1, z0);
				glVertex3f(x2, y2, z0);
				glVertex3f(x3, y3, z0);
				glEnd();
				glBegin(GL_LINE_LOOP);
				glVertex3f(x0, y0, z1);
				glVertex3f(x1, y1, z1);
				glVertex3f(x2, y2, z1);
				glVertex3f(x3, y3, z1);
				glEnd();
				glBegin(GL_LINES);
				glVertex3f(x0, y0, z0);
				glVertex3f(x0, y0, z1);
				glVertex3f(x1, y1, z0);
				glVertex3f(x1, y1, z1);
				glVertex3f(x2, y2, z0);
				glVertex3f(x2, y2, z1);
				glVertex3f(x3, y3, z1);
				glVertex3f(x3, y3, z0);
				glEnd();
				break;
			case Inrimage::SPHERICAL:
				break;
		}
	}

	glEnable(GL_LIGHTING);
}
#endif
void VoxelMax::start(Voxel&)
{
	maxVoxel = startvoxel;
}

bool VoxelMax::stopFromDistanceAndStore(ImageVoxel * voxel, Voxel&)
{
	if(voxel->getDistance()>maxDistance)
		return true;
	else
	{
		if(voxel->getValue()>maxVoxel->getValue()) 
		maxVoxel = voxel;
		return false;
	}
}

IntensityProfileSet::IntensityProfileSet(Inrimage *inr, const std::vector< Vec3<double> > &pos, const std::vector< Vec3<double> > &dir, const double dist, const double step) : step(step)
{
	std::vector<Vec3 <double> >::const_iterator itp, itd;
	IntensityProfile * ils;
	unsigned int i;

	for(i=0,itp = pos.begin(),itd=dir.begin(); (itp!=pos.end())&&(itd!=dir.end()); itp++,itd++,i++)
	{
		ils = new IntensityProfile();
		ils->setIntensity(inr);
		ils->compute(dist,(*itp),(*itd));
		il.push_back(ils);
	}
	// Store the length of the intensity profiles. Will bw used by "-draw" command
	//realSizeIntensityProfile = ((unsigned int)dist)-2;
	realSizeIntensityProfile = (unsigned int)dist;
}

IntensityProfileSet::~IntensityProfileSet()
{
	std::vector<IntensityProfile *>::iterator it;

	for(it = il.begin(); it!=il.end(); it++)
		delete (*it);

	il.erase(il.begin(), il.end());
}

Inrimage * IntensityProfileSet::createIntensityProfileImage(const IMAGETYPE imageType) const
{
	int h, hh;
	unsigned i, j, sizeIP;
	Inrimage *tinr;
	double k, bound, *** buffer;
	std::vector<IntensityProfile *>::const_iterator it;

	// Intensity profiles size taking the step into account
	sizeIP = (unsigned int) floor((double)realSizeIntensityProfile/step);
	tinr = new Inrimage(getNumberIntensityProfiles(),sizeIP,1,Inrimage::WT_DOUBLE);
	buffer = (double ***)tinr->getArray();

	switch (imageType)
	{
		case POSITIVE_PROFILES : 
			for(i=0,it=il.begin(); it!=il.end(); it++,i++)
				for(j=0,k=0.0; j<sizeIP; ++j,k+=step)
					buffer[0][j][i] = (*it)->getIntensity(k);
			break;
		case NEGATIVE_PROFILES :
			bound = (-1.0*(double)sizeIP*step)+step;

			for(i=0,it=il.begin(); it!=il.end(); it++,i++)
				for(j=0,k=bound; j<sizeIP; ++j,k+=step)
					buffer[0][j][i] = (*it)->getIntensity(k);
			break;
		case ALL_PROFILES :
			h = realSizeIntensityProfile;
			h += 1-(h%2);
			hh = h/2;
			bound = -1.0*(double)hh;

			for(i=0,it=il.begin(); it!=il.end(); it++,i++)
				for(j=0,k=bound; j<sizeIP; ++j,k+=step)
					buffer[0][j][i] = (*it)->getIntensity(k);
			break;
	}

	return tinr;
}

FuzzyCMeans * IntensityProfileSet::classifyByFuzzyCMeans(unsigned int nClust, IMAGETYPE imageType)
{
	// Declarations
	unsigned int i, j, nIP, sizeIP, temp;
	int h, hh, temp2;
	double k, bound, minRange, maxRange, ** initialArrayIP;
	std::vector<IntensityProfile *>::const_iterator it;
	FuzzyCMeans * outputFCM;
	unsigned int * validityRange;

	// Put the intensity profiles in a double array in order to call the Fuzzy C-Means clustering algorithm
	sizeIP = (unsigned int) floor((double)realSizeIntensityProfile/step);
	nIP = getNumberIntensityProfiles();
	initialArrayIP = new double * [nIP];
	validityRange = new unsigned int [2*nIP];

	for(i=0; i<nIP; i++)
		initialArrayIP[i] = new double[sizeIP];

	switch (imageType)
	{
		case POSITIVE_PROFILES : 
			for(i=0, it=il.begin(); it!=il.end(); it++, i++)
			{
				minRange = (*it)->getMinVoxelDistance()/step;
				maxRange = (*it)->getMaxVoxelDistance()/step;

				if (minRange<0.0)
					validityRange[2*i] = 0;
				else
				{
					temp = (unsigned int)(minRange/step);

					if (temp>(sizeIP-1))
						validityRange[2*i] = sizeIP-1;
					else
						validityRange[2*i] = temp;
				}

				if (maxRange<0.0)
					validityRange[(2*i)+1] = 0;
				else
				{
					temp = (unsigned int)(maxRange/step);

					if (temp>(sizeIP-1))
						validityRange[(2*i)+1] = sizeIP-1;
					else
						validityRange[(2*i)+1] = temp;
				}

				for(j=0,k=0.0; j<sizeIP; ++j,k+=step)
					initialArrayIP[i][j] = (*it)->getIntensity(k);
			}
			break;
		case NEGATIVE_PROFILES :
			bound = (-1.0*(double)sizeIP*step)+step;

			for(i=0, it=il.begin(); it!=il.end(); it++, i++)
			{
				minRange = (*it)->getMinVoxelDistance()/step;
				maxRange = (*it)->getMaxVoxelDistance()/step;

				if (minRange>=0.0)
					validityRange[2*i] = sizeIP-1;
				else
				{
					temp = (unsigned int)(-minRange/step);

					if (temp>(sizeIP-1))
						validityRange[2*i] = 0;
					else
						validityRange[2*i] = sizeIP-1-temp;
				}

				if (maxRange>=0.0)
					validityRange[(2*i)+1] = sizeIP-1;
				else
				{
					temp = (unsigned int)(-maxRange/step);

					if (temp>(sizeIP-1))
						validityRange[(2*i)+1] = 0;
					else
						validityRange[(2*i)+1] = sizeIP-1-temp;
				}

				for(j=0,k=bound; j<sizeIP; j++,k+=step)
					initialArrayIP[i][j] = (*it)->getIntensity(k);
			}
			break;
		case ALL_PROFILES :
			h = realSizeIntensityProfile;
			h += 1-(h%2);
			hh = h/2;
			bound = -1.0*(double)hh;

			for(i=0, it=il.begin(); it!=il.end(); it++, i++)
			{
				minRange = (*it)->getMinVoxelDistance()/step;
				maxRange = (*it)->getMaxVoxelDistance()/step;

				if (minRange<=bound)
					validityRange[2*i] = 0;
				else
				{
					temp2 = ((sizeIP-1)/2)+(int)(minRange/step);

					if (temp2>(int)(sizeIP-1))
						validityRange[2*i] = sizeIP-1;
					else
						validityRange[2*i] = temp2;
				}

				if (maxRange>=-bound)
					validityRange[(2*i)+1] = sizeIP-1;
				else
				{
					temp2 = ((sizeIP-1)/2)+(int)(maxRange/step);

					if (temp2>(int)(sizeIP-1))
						validityRange[(2*i)+1] = sizeIP-1;
					else
						validityRange[(2*i)+1] = temp2;
				}

				for(j=0,k=bound; j<sizeIP; j++,k+=step)
					initialArrayIP[i][j] = (*it)->getIntensity(k);
			}
			break;
	}

	// Call the fuzzy function
	outputFCM = new FuzzyCMeans(initialArrayIP, nClust, nIP, sizeIP, validityRange);
	outputFCM->fcm();

	return outputFCM;
}

Inrimage* IntensityProfileSet::sortFCMClassification(FuzzyCMeans* outputFCM)
{
	int i, j, k, idClust, nIP, dimIP, nClust, count;
	double tmp;
	std::vector<IntensityProfile *>::const_iterator it;

	nIP = outputFCM->getNbrData();
	dimIP = outputFCM->getNbrDim();
	nClust = outputFCM->getNbrClust();

	// Attribuate data to the highest probability cluster
	int* posMax = new int [nIP];

	for (j=0; j<nIP; j++)
	{
		tmp = 0.0;

		for (i=0; i<nClust; i++)
		{
			if (outputFCM->getMembership()[(i*nIP)+j]>tmp)
			{
				tmp = outputFCM->getMembership()[(i*nIP)+j];
				idClust = i;
			}

		}

		posMax[j] = idClust;
	}

	// Calculate the mean of each cluster
	double* mean1 = new double [nClust];
	double* mean2 = new double [nClust];
	double* table = new double [nClust];

	for (i=0; i<nClust; i++)
	{
		tmp = 0.0;

		for (j=0; j<dimIP; j++)
			tmp += outputFCM->getCenters()[(i*dimIP)+j];

		mean1[i] = mean2[i] = tmp/dimIP;
	}

	// Sort the clusters by ascending intensity value
	std::sort(mean2, mean2+nClust);

	// Create the look-up table with increasing cluster intensities
	for (i=0; i<nClust; i++)
		for (j=0; j<nClust; j++)
			if (mean2[i]==mean1[j])
				table[i] = j;

	// Gather intensity profiles by cluster
	count = 0;
	double** data = new double * [dimIP];

	for (i=0; i<dimIP; i++)
		data[i] = new double[nIP];

	for (i=nClust-1; i>=0; i--)
	{
		for (j=0; j<nIP; j++)
		{
			if (posMax[j]==table[i])
			{
				for (k=0; k<dimIP; k++)
					data[k][count] = outputFCM->getData()[j][k];

				count++;
			}
		}
	}

	// Create the clusters image
	Inrimage* image = new Inrimage(nIP, dimIP, 1, Inrimage::WT_DOUBLE);
	double*** array = (double***)image->getArray();	

	for (i=0; i<dimIP; i++)
		for (j=0; j<nIP; j++)
			array[0][i][j] = data[i][j];

	return image;
}

Inrimage* IntensityProfileSet::sortEMClassification(ExpectationMaximization* outputEM)
{
	int i, j, k, idClust, nIP, dimIP, nClust, count;
	double tmp;
	std::vector<IntensityProfile *>::const_iterator it;

	nIP = outputEM->getNbrData();
	dimIP = outputEM->getNbrDim();
	nClust = outputEM->getNbrModes();

	// Attribuate data only to the highest probability mode
	int* posMax = new int [nIP];

	for (j=0; j<nIP; j++)
	{
		tmp = 0.0;

		for (i=0; i<nClust; i++)
		{
			if (outputEM->getProba()[(i*nIP)+j]>tmp)
			{
				tmp = outputEM->getProba()[(i*nIP)+j];
				idClust = i;
			}

		}

		posMax[j] = idClust;
	}

	// Calculate the mean of each mode
	double* mean1 = new double [nClust];
	double* mean2 = new double [nClust];
	double* table = new double [nClust];

	for (i=0; i<nClust; i++)
	{
		tmp = 0.0;

		for (j=0; j<dimIP; j++)
			tmp += outputEM->getCenters()[(i*dimIP)+j];

		mean1[i] = mean2[i] = tmp/dimIP;
	}

	// Sort the modes by ascending intensity value	
	std::sort(mean2, mean2+nClust);

	// Create the look-up table with increasing mode intensities
	for (i=0; i<nClust; i++)
	{
		for (j=0; j<nClust; j++)
		{
			if (mean2[i]==mean1[j])
				table[i] = j;
		}
	}

	// Gather intensity profiles by mode
	count = 0;
	double** data = new double * [dimIP];

	for (i=0; i<dimIP; i++)
		data[i] = new double[nIP];

	for (i=nClust-1; i>=0; i--)
	{
		for (j=0; j<nIP; j++)
		{
			if (posMax[j]==table[i])
			{
				for (k=0; k<dimIP; k++)
					data[k][count] = outputEM->getData()[j][k];

				count++;
			}
		}
	}

	// Create the clusters image
	Inrimage* image = new Inrimage(nIP, dimIP, 1, Inrimage::WT_DOUBLE);
	double*** array = (double***)image->getArray();	

	for (i=0; i<dimIP; i++)
		for (j=0; j<nIP; j++)
			array[0][i][j] = data[i][j];

	return image;
}
} // namespace yav
