#ifndef IMAGE3DVIEWER_H
#define IMAGE3DVIEWER_H

#include <inrimage/Image3DBackgroundManager.h>
#include <graphics3D/Data3D.h>

namespace yav {

/** \class ImageData
    \ingroup ImageData
    \brief 3D image data */
class Image3DViewer : public Image3DViewerInterface, public Data3D
{
public:
  /// grid display mode
  enum GRID_DISPLAY {
    /// only slice frame
    HIDE,
    /// only slice frame
    FRAME,
    /// pixels grid
    GRID,
    /// voxels
    VOXELS
  };
 protected:
  /// inrimage bounding box color
  float boxColor[3];
  /// inrimage line width
  double boxWidth;
  /// slices grid color
  float gridColor[3];
  /// slices grid line width
  double gridWidth;
  /// if we should show or hide black pixels
  bool showBlackPixels;
 /// grid display mode
  GRID_DISPLAY gridDisplay;
  /// constructors common code
  void init(void);

 public:
  /** @name constructors */
  //@{
  /** Builds a new image data from given inrimage.
      @param inr image to display
      @param scene scene to create image in
      @param allCameras whether to track all cameras or only ones that display image inrimage */
  Image3DViewer(Scene3D *scene,Image3DBackgroundManager *manager);
  //@}

  /** @name destructor */
  //@{
  /// destroys image data (but not concerned inrimages)
  virtual ~Image3DViewer(){}
  //@}

  /** get the image being displayed */
  Inrimage *getImage() const {
    return manager->getInrimage();
  }
 /// ask the Image3DViewer to be redrawn
  virtual void redrawViewer(){
	  redrawScene();
  }
  /// all graphics objects can be drawn
  virtual void redraw(Camera3D *);

  /** Returns the graphic object center and weight.
      The weight is a measure of the number of primitives to allow
      barycentric computation of the whole scene center.
      @param center returned center 
      @param n returned barycentric weight */
  virtual void updateCenter(Vec3<double>& center, unsigned int& n);

  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec3<double>& center);
  
  /** The routine that allows the camera to select an object
      by computing the distance of an object from its radial line
      @param mc 3D line descriptor
      @return distance to line */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const;

  /** returns R, G and B bounding box color (each component is between 0 and 1)
      @param r returned red component 
      @param g returned green component
      @param b returned blue component */
  void getBoxColor(float &r, float &g, float &b) const;

  /// returns bounding box line width
  double getBoxWidth(void) const { return boxWidth; }

  /** returns R, G and B slices grid color (each component is between 0 and 1)
      @param r returned red component 
      @param g returned green component
      @param b returned blue component */
  void getGridColor(float &r, float &g, float &b) const;

  /// returns slices grid line width
  double getGridWidth(void) const { return gridWidth; }

  /** Set R, G and B bounding box (each component should be between 0 and 1)
      @param r red component 
      @param g green component
      @param b blue component */
  void setBoxColor(float r, float g, float b);

  /** set bounding box line width
      @param width line width (should be > 0.0) */
  void setBoxWidth(double width);
 
  /** Set grid R, G and B components (each component should be between 0 and 1)
      @param r red component 
      @param g green component
      @param b blue component */
  void setGridColor(float r, float g, float b);

  /** set slices grid line width
      @param width line width (should be > 0.0) */
  void setGridWidth(double width);
  /** indicates if the black pixels should be shown or not */
  bool AreBlackPixelsShown() const {
    return showBlackPixels;}
  /** set the boolean indicating if the black pixels should be shown or not */
  void setShowBlackPixels(const bool b); 
  

  /// get current grid display mode
  GRID_DISPLAY getGridDisplay(void) const {
	  return gridDisplay; }
  /// set new grid display mode
  void setGridDisplay(GRID_DISPLAY gd);
 public:
/** translate a 3D contour
      @param t the 3D translation vector */
  void translate(const Vec3<double>& t);
  /** scale a 3D contour
      @param scale the scale factor (should be greater than 0) */
  void scale(const double scale);
   /** rotate a 3D contour around its center 
      @param rotation the 3D rotation */ 
  void rotate(const Rotation3D &rotation);
  
};

} // namespace yav




#endif  // ifndef IMAGE3DVIEWER_H


