#ifndef IMAGEDATA_H
#define IMAGEDATA_H

#pragma interface "imagedata/ImageData.h"
#include <map>
#ifdef OPENGL
#include <GL/gl.h>
#endif
#include <basics/Exception.h>
#include <basics/FileInfo.h>
#include <graphics3D/Data3D.h>
#include <graphics2D/SliceCamera.h>
#include <inrimage/Inrimage.h>
#include <inrimage/InrimageSliceInterface.h>
#include <graphics2D/Sliceable.h>
#include <math.h>

namespace yav {


#ifdef TCL_INTERFACE
class TclImageData;
#endif


/** \class ImageData
    \ingroup ImageData
    \brief 3D image data */
class ImageData : public Data3D {
 protected:
  /// storing the minimum intensity value of the image
  Voxel *glmin;
  /// storing the maximum intensity value of the image
  Voxel *glmax;
  /// intensity image
  Inrimage *image;

#ifdef OPENGL  
  /// inrimage bounding box color
  float boxColor[3];
  /// inrimage line width
  double boxWidth;
  /// slices grid color
  float gridColor[3];
  /// slices grid line width
  double gridWidth;
  /// if we should show or hide black pixels
  bool showBlackPixels;
#endif

  /// whether to track all cameras or only ones that display image inrimage
  bool allCameras;
  
  /// constructors common code
  void init(void);

 public:
  /** @name constructors */
  //@{
  /** Builds a new image data from given inrimage.
      @param inr image to display
      @param scene scene to create image in
      @param allCameras whether to track all cameras or only ones that display image inrimage */
  ImageData(Inrimage *inr, Scene3D *scene = 0, bool allCameras = false);

  /** Builds a new image data from file name.
      @param fileName inrimage file name
      @param scene scene to create image in
      @exception Inrimage::ImageNotFound file not found 
      @exception Inrimage::UnknownImageType unknown image file format
      @exception Inrimage::InvalidHeader invalid inrimage header
      @exception Inrimage::InvalidData invalid inrimage data
      @exception Inrimage::UnknownType unhandled inrimage type
      @param allCameras whether to track all cameras or only ones that display image inrimage */
  ImageData(const TclString& fileName, Scene3D *scene = 0,
	    bool allCameras = false)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);
  //@}

  /** @name destructor */
  //@{
  /// destroys image data (but not concerned inrimages)
  virtual ~ImageData();
  //@}

  /// returns X dimension
  unsigned int getX(void) const { return image->getX(); }
  /// returns X dimension
  unsigned int ncols(void) const { return image->getX(); }
  /// returns Y dimension
  unsigned int getY(void) const { return image->getY(); } 
  /// returns Y dimension
  unsigned int nrows(void) const { return image->getY(); }
  /// returns Z dimension
  unsigned int getZ(void) const { return image->getZ(); }
  /// returns Z dimension
  unsigned int nplanes(void) const { return image->getZ(); }

  /// returns X voxel size
  double getVX(void) const { return image->getVX(); }
  /// returns X voxel size
  double getVC(void) const { return image->getVX(); }
  /// returns Y voxel size
  double getVY(void) const { return image->getVY(); }
  /// returns Y voxel size
  double getVR(void) const { return image->getVY(); }
  /// returns Z voxel size
  double getVZ(void) const { return image->getVZ(); }
  /// returns Z voxel size
  double getVP(void) const { return image->getVZ(); }
  /** get the image being displayed */
  Inrimage *getImage() const {
    return image;
  }
  /** modifies min grey value for volume rendering */
  void setMinGreyValue(const double val);
  /** modifies min grey value for volume rendering */
  void setMaxGreyValue(const double val);

  /// all graphics objects can be drawn
  virtual void redraw(Camera3D *);

  /** Returns the graphic object center and weight.
      The weight is a measure of the number of primitives to allow
      barycentric computation of the whole scene center.
      @param center returned center 
      @param n returned barycentric weight */
  virtual void updateCenter(Vec3<double>& center, unsigned int& n);

  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec3<double>& center);
  
  /** The routine that allows the camera to select an object
      by computing the distance of an object from its radial line
      @param mc 3D line descriptor
      @return distance to line */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const;

  /** returns R, G and B bounding box color (each component is between 0 and 1)
      @param r returned red component 
      @param g returned green component
      @param b returned blue component */
  void getBoxColor(float &r, float &g, float &b) const;

  /// returns bounding box line width
  double getBoxWidth(void) const { return boxWidth; }

  /** returns R, G and B slices grid color (each component is between 0 and 1)
      @param r returned red component 
      @param g returned green component
      @param b returned blue component */
  void getGridColor(float &r, float &g, float &b) const;

  /// returns slices grid line width
  double getGridWidth(void) const { return gridWidth; }

  /** Set R, G and B bounding box (each component should be between 0 and 1)
      @param r red component 
      @param g green component
      @param b blue component */
  void setBoxColor(float r, float g, float b);

  /** set bounding box line width
      @param width line width (should be > 0.0) */
  void setBoxWidth(double width);
 
  /** Set grid R, G and B components (each component should be between 0 and 1)
      @param r red component 
      @param g green component
      @param b blue component */
  void setGridColor(float r, float g, float b);

  /** set slices grid line width
      @param width line width (should be > 0.0) */
  void setGridWidth(double width);
  /** indicates if the black pixels should be shown or not */
  bool AreBlackPixelsShown() const {
    return showBlackPixels;}
  /** set the boolean indicating if the black pixels should be shown or not */
  void setShowBlackPixels(const bool b); 
  /** set which inrimage should be displayed
      @param showImage*/
  virtual bool display(bool showImage);

 

  /// Volume rendering texturing modes
  enum VOLUME_RENDERING_TYPE {
    /// Use 2-D textures. Fast on all platforms, but not that convincing
   VR_2D,
    /// Use 3-D textures. Extremely fast, but only with hardware support for 3D textures
   VR_3D,
  };
  
  
  /// returns display resolution
  unsigned int getResolution(void) const { return resolution; }
  /** set display resolution
      @param resolution display resolution (should be > 0) */
  void setResolution(unsigned int resolution);

  /// set slices blending behavior
  void setBlending(bool blending, double alpha = 0.3333333333333);
  /** return alpha blending factor */
  double getAlpha() const {
    return alpha;
  }
  /// indicates if the volume rendering is to be performed in 2D or 3D */
  void setVolumeRenderingType(const VOLUME_RENDERING_TYPE volType);

  /// clear all display lists
  void clearLists(void);
  /// clear all display lists for volume rendering
  void clearVolumeLists(void);


  
  /// volume rendering blending modes
  enum BLENDING {
    /// flow of light through a transparent material
    OVER,
    /// attenuate light shining through the material along the view direction (X-ray)
    ATTENUATE,
    /// Maximum Intensity Projection
    MIP,
    /// OVER operator from back to front
    UNDER,
  };

  /// returns volume rendering number of slices
  unsigned int getNSlices(void) const { return nslices; }

  /// returns volume rendering blend mode
  BLENDING getVolumeBlending(void) const { return volumeBlending; }

  /** set number of slices to perform volume rendering
      @param nslices number of slice (truncated between 1 and 4 * max(xdim, ydim, zdim) */
  void setNSlices(unsigned int nslices);

  /** set volume blending mode
      @param blend new volume blending mode */
  void setVolumeBlending(ImageData::BLENDING blend);

  /** set voxels transparency from voxel intensity for volume rendering
      @param trs 256 elements array with alpha value for each intensity value
      @param ot overall blending factor */
  void setVolumeTransparency(double *trs, float ot);


  /** function called when the inrimage is modified
      @param i an inrimage */
   virtual void imageUpdate(const Inrimage *i);
  /** function called when the inrimage is deleted
      @param i an inrimage */
   virtual void imageDestroyed(const Inrimage *i);
 protected:
  /** set new intensity image and resize data image
      @param inr intensity image */
  void setImage(Inrimage *inr);
  /** removed all display list stored for each slice
      @param inr intensity image */
  void resetImage(Inrimage *inr);

#ifdef OPENGL
 protected:


  /// which image to display
  bool showImage;

  /** computes volume rendering display list */
  GLuint volumeDisplayList(); 
  /** computes one slice display list
      @param b slice camera background*/
  GLuint sliceDisplayList(const SliceCameraBackground *b);
  /** computes one slice grid display list
      @param slice the slice number
      @param geom slice geometry */
  GLuint gridDisplayList(unsigned int slice,
			 SliceCamera::SLICE_ORIENTATION geom) const;

  /// delete all ImageData::Slice structures
  virtual void clearSlices(void);

  /// grid display mode
  GRID_DISPLAY gridDisplay;
  /// slices display mode
  SLICES_DISPLAY slicesDisplay;
  /// display resolution
  unsigned int resolution;

  /// use blending ?
  bool blending;
  /// blending factor
  double alpha;
  /// the current orientation used for 2D volume rendering
  SliceCamera::SLICE_ORIENTATION sliceOrientationTexture2D;
  /// OpenGL display list for volume rendering mode
  GLuint volumeList; 
  ///  OpenGL display lists created to be used in volume rendering using 2D textures
  GLuint *textures;
  /// volume rendering blending mode
  BLENDING volumeBlending;
  /// number of slices for volume rendering
  unsigned int nslices;
  /// clip plane equations
  GLdouble clipPlaneEqn[6][4];
  /// volume rendering lookup table
  float *lookupTable;
  /** either 2D or 3D volume rendering type */
  VOLUME_RENDERING_TYPE volumeRenderingType;
  /// texture bg color
  float textbg[4];
  /// texture map mode
  GLfloat textmap;
 public:
/** translate a 3D contour
      @param t the 3D translation vector */
  void translate(const Vec3<double>& t);
  /** scale a 3D contour
      @param scale the scale factor (should be greater than 0) */
  void scale(const double scale);
   /** rotate a 3D contour around its center 
      @param rotation the 3D rotation */ 
  void rotate(const Rotation3D &rotation);
  /** return the texture background 4 components */
  const float *getTextureBackground() const {
    return textbg;
  }
  /** return texture environement mode (modulate, decal,...) */
  GLfloat getTextMap(){
     return textmap;
   }
  /** return color table of size 256 */
   const float *getLookupTable() {
     return lookupTable;
   }
   /** save the lookup table in a file */
   void writeLookupTable(const TclString lookupFileName);
   /** read the lookup table from a file */
   void readLookupTable(const TclString lookupFileName);

 private:
   /** create 2D volume rendering textures and map them on slices */
   void createVRTexture2D(Vec3<double> &n);
   /** display 2D volume rendering slices using X direction */
   void displaySliceX(unsigned int slice);
   /** display 2D volume rendering slices using Y direction */
   void displaySliceY(unsigned int slice);
   /** display 2D volume rendering slices using Z direction */
   void displaySliceZ(unsigned int slice);
   /** create 2D colored quads  from RGB volumetric image */
   void buildRGBQuads(Inrimage *inr, unsigned int slice,
		      SliceCamera::SLICE_ORIENTATION geom) const;
   /** create 2D slice textures from RGB volumetric image */   
   void buildRGBTexture(Inrimage *inr, unsigned int slice,
			SliceCamera::SLICE_ORIENTATION geom) const;
   /** create 3D texture from RGB volumetric image */   
   void build3DRGBTexture(Inrimage *inr);
   /** create a direct trihedron from a single vector */
   static void computeOrthonormalBase(const Vec3<double>& n,
				      Vec3<double>& i, Vec3<double>& j);
#endif
  

   friend class TclImageData;
};

} // namespace yav




#endif  // ifndef IMAGEDATA_H


