#ifndef IMAGELINE_H
#define IMAGELINE_H
#pragma interface "imagedata/IntensityProfile.h"

#include <vector>
#include <limits>
#include <basics/Exception.h>
#include <linearAlgebra/Vec3.h>
#include <basics/Math.h>
#include <classification/FuzzyCMeans.h>
#include <classification/ExpectationMaximization.h>
#include <inrimage/Inrimage.h>

namespace yav {

class IntensityProfile;

/** \class ImageVoxel
    \ingroup ImageData
    \brief one voxel descriptor */
class ImageVoxel
{
 protected:
	/// Voxel coordinates
	unsigned int vpos[3];
	/// Subvoxel coordinates (in cartesian space)
	Vec3<double> spos;
	/// Voxel value converted to a double
	double value;
	/// Distance from line origin
	double distance;
	/// Index
	unsigned int index;
 public:
	/** @name constructor */
	/** Builds this voxel
	@param pos 3 dimensional array giving voxel coordinates (each coordinate should be >= 0)
	@param subpos cartesian coordinate of sub-voxel point
	@param value voxel value (converted to a double)
	@param distance voxel distance from line origin 
	@param index the index of the voxel in the image line  */
	ImageVoxel(int *pos, const Vec3<double>& subpos, double value, double distance, unsigned int index);
	
	/** @name destructor */
	~ImageVoxel();
	
	/// returns voxel X coordinate (voxel column)
	unsigned int getX(void) const { return vpos[0]; }

	/// returns voxel Y coordinate (voxel row)
	unsigned int getY(void) const { return vpos[1]; }

	/// returns voxel Z coordinate (voxel plane)
	unsigned int getZ(void) const { return vpos[2]; }

	/// returns voxel distance to origin
	double getDistance(void) const { return distance; }

	/// returns voxel value
	double getValue(void) const { return value; }

	/// returns voxel position
	Vec3<unsigned int>   getVoxelPosition(void) const { return Vec3<unsigned int>(vpos[0],vpos[1],vpos[2]); }

	/// returns sub-voxel position
	const Vec3<double>& getPosition(void) const { return spos; }

	/// set the value of the voxel
	void setValue(const double val) { value = val; }
};

/** \class IntensityProfile
    \ingroup ImageData
    \brief voxels line descriptor */
class IntensityProfile
{ 
 protected:
	/// image from which this line is extracted
	Inrimage *inr;
	/// line starting point in euclidian coordinates
	Vec3<double> origin;
	/// line direction vector
	Vec3<double> direction;
	/// maximal line segment length
	double maxDistance;
	/// line starting voxel
	ImageVoxel *startvoxel;
	/// line voxels in direction
	std::vector<ImageVoxel *> positive;
	/// line voxels in -direction
	std::vector<ImageVoxel *> negative;
	/// gradient in direction
	std::vector<double> positiveGradient;
	/// gradient in -direction
	std::vector<double> negativeGradient;
	/// pointer to stop condition
	bool (IntensityProfile::*stopMethod)(ImageVoxel *voxel, Voxel&);
 public:
	/** @name constructor */
	/** initialize a new line with stopFromDistance default stop method
	@param max maximal line segment length
	@param stop method used */
	IntensityProfile(double max = 0.0, bool (IntensityProfile::*stop)(ImageVoxel *, Voxel&) = &IntensityProfile::stopFromDistance);

	/** @name destructor */
	/// destroys any line voxel
	virtual ~IntensityProfile();

	/// segment or half segment direction
	enum SIDE
	{
		/// half segment in direction
		NORMAL,
		/// half segment in -direction
		INVERSE,
		/// complete segment
		BOTH
	};

	/** get line voxel operator
	@param i voxel index (0 for line starting point, > 0 for voxels in direction and < 0 for voxels in -direction
	@return desired voxel or 0 if voxel index exceed segment length */
	ImageVoxel * operator[](int i) const;
	
	/** return a  voxel given its index
	@param i voxel index (0 for line starting point, > 0 for voxels in direction and < 0 for voxels in -direction
	@return desired voxel or 0 if voxel index exceed segment length */
	ImageVoxel * getVoxel(int i) const;


	/** returns the intensity corresponding to a given distance value performs a linear interpolation between image voxels. returns 0 it fabs(dist)>maxDistance */
	double getIntensity(double dist);
	
	/** returns the intensity gradient corresponding to a given distance value performs a linear interpolation between image gradient returns 0 if fabs(dist)>maxDistance */
	double getGradient(double dist);
	
	/** returns the intensity gradient corresponding to a given distance value performs a linear interpolation between image gradient returns 0 if fabs(dist)>maxDistance */
	double getGradient(int i);

	/** return the minimum distance at which a voxel is intersected (may be a positive value) */
	double getMinVoxelDistance() const;
	/** return the maximum distance at which a voxel is intersected (may be a negative value) */
	double getMaxVoxelDistance() const;

	/** returns the distance from the origin voxel */
	double getDistance(int i);
	/** compute new intensity profile composed of voxels with given origin and direction
	@param distance the maximum distance along the line to search for the optimal voxel
	@param org line origin in euclidian coordinates
	@param dir line direction vector
	@param side full or half segment computation flag */
	void compute(double distance, const Vec3<double>& org, const Vec3<double>& dir, SIDE side = BOTH);
	
	/** compute the gradient of image intensity profile 
	@param sigma the standard deviation for the derivative of Gaussian*/
	void computeGradient(double sigma);

	/** compute the gradient at one given index of the profile  
	@param index the (signed) index at which the gradient is computed
	@param sigma the standard deviation for the derivative of Gaussian
	@return the gradient value at that index*/
	double computeGradientAtIndex(const int index,const double sigma);

	/** compute the gradient at one given distance from the origin  
	@param distance the (signed) distance at which the gradient is computed
	@param sigma the standard deviation for the derivative of Gaussian
	@return the gradient value at that distance using a linear interpolation between gradients computed at 2 profile indices*/
	double computeGradientAtDistance(const double distance,const double sigma);

	/** default stop computation method: stops line segment computation when current voxel distance from origin exceeds maxDistance
	@param voxel last computed voxel index
	@param iv an image type voxel for any use in stop */
	bool stopFromDistance(ImageVoxel *voxel, Voxel& iv);
	
	/** callback procedure called when first line voxel is set
	@param voxel an image type voxel for any use in start */
	virtual void start(Voxel& voxel) {}
	
	/// sets the new image intensity
	void setIntensity(Inrimage *_inr) { inr = _inr; }

	/// destroys all line voxels
	void clear(void);
	
	/// returns line origin
	const Vec3<double>& getOrigin(void) const { return origin; }

	/// returns line direction
	const Vec3<double>& getDirection(void) const { return direction; }

	/// returns max line length
	double getMaxDistance(void) const { return maxDistance; }
	
	/** set max line length
	@param max new line lenght */
	void setMaxDistance(double max) { maxDistance = max; }

	/** -----
	@param org */
	void setOrigin(Vec3<double> &org) { origin = org; }

	/** -----
	@param d*/
	void setDirection(Vec3<double> &d) { direction = d; }
	
	///returns the number of elements for the positive array
	unsigned int positiveArraySize() const { return (startvoxel ? (positive.size()) : 0); }

	/// returns the number of elements for the negative array
	unsigned int negativeArraySize() const { return (startvoxel ? (negative.size()+1) : 0); }
#ifdef GRAPHICS
	/// display discrete line in an OpenGL window
	void glSet(void) const;
#endif
};

/** \class VoxelMax
    \ingroup ImageData
    \brief max voxel value in line */
class VoxelMax : public IntensityProfile
{
 protected:
	/// max line intensity point
	ImageVoxel *maxVoxel;
 public:
	/** @name constructor */
	/** initialize a new line with stopFromDistance default stop method
	@param max maximal line segment length
	@param stop method used */
	VoxelMax(double max = 0.0, bool (VoxelMax::*stop)(ImageVoxel *, Voxel&) = &VoxelMax::stopFromDistanceAndStore) : IntensityProfile(max, (bool (IntensityProfile::*)(ImageVoxel *, Voxel&)) stop) {}
	
	/** @name destructor */
	/// destroys any line voxel
	virtual ~VoxelMax() {}
	
	/** stop computation method: stops line segment computation when current voxel distance from origin exceeds maxDistance and stores max encountered voxel value
	@param voxel last computed voxel index
	@param iv an image type voxel for any use in stop */
	bool stopFromDistanceAndStore(ImageVoxel *voxel, Voxel& iv);
	
	/** callback procedure called when first line voxel is set
	@param voxel an image type voxel for any use in start */
	virtual void start(Voxel &voxel);
	
	/// returns max intensity voxel
	ImageVoxel *getMaxVoxel(void) const { return maxVoxel; }
};


/** \class IntensityProfileSet
    \ingroup ImageData
    \brief a set of IntensityProfile */
class IntensityProfileSet
{ 
 protected:
	/// the array of intensity profiles
	std::vector<IntensityProfile *> il;
	/// the length of the intensity profiles;
	unsigned int realSizeIntensityProfile;
	/// step along the intensity profile;
	const double step;
public:
	/// segment or half segment direction
	enum IMAGETYPE
	{
		/// half segment in direction
		POSITIVE_PROFILES,
		/// half segment in -direction
		NEGATIVE_PROFILES,
		/// complete segment
		ALL_PROFILES
	};

	/** @name constructor */
	/** Create the object from a vector of IntensityProfiles (copy the array into the new one) 
	@param ils array of IntensityProfile */
	IntensityProfileSet(std::vector<IntensityProfile *> &ils) : il(ils), step(1.0) {}

	/** Copy constructor (copy the array into the new one) 
	@param ils IntensityProfileSet */
	IntensityProfileSet(IntensityProfileSet &ils) : il(ils.il), step(1.0) {}

	/** creates an array of intensity profiles composed of voxels given an image, a list of position and  normal directions and the half-distance 
	@param inr a volumetric image 
	@param pos a list of position vector corresponding to the profile origin
	@param dir a list of direction vectors for each profile 
	@param dist the half distance of all intensity profiles
	@param step step along the intensity profile (default=1.0) */
	IntensityProfileSet(Inrimage *inr, const std::vector< Vec3<double> > &pos, const std::vector< Vec3<double> > &dir, const double dist, const double step = 1.0);

	/** get the length of the intensity profiles */
	unsigned int getSizeIntensityProfiles() { return realSizeIntensityProfile; }

	/** empty constructor  */
	IntensityProfileSet() : step(1.0) {}
	
	/** @name destructor */
	/// destroys the array of IntensityProfile after destroying each object
	virtual ~IntensityProfileSet();

	/// returns the first iterator  of image line pointed by the interface
	std::vector<IntensityProfile *>::const_iterator getFirstIntensityProfile() const { return il.begin(); }

	/// returns the last iterator of image line pointed by the interface
	std::vector<IntensityProfile *>::const_iterator getLastIntensityProfile() const { return il.end(); }

	/// adds an image line in the set 
	void addIntensityProfile(IntensityProfile *i){ il.push_back(i); }

	/// returns the number of image lines in the list
	unsigned int getNumberIntensityProfiles() const { return il.size(); }

	/// returns the array of intensity profiles
	std::vector<IntensityProfile *> getIntensityProfileSet() { return il; }
	
	/** creates a 2D image by putting side by side all intensity profiles
	@param imageType specifies if only the positive side or negative side or both should be considered when generating the image
	@return a 2D image of size width (= number of profile) x height. The intensity are sampled regularly from the images pointed by each intensity profile */  
	Inrimage * createIntensityProfileImage(const IMAGETYPE imageType) const; 
	
	/** classify the intensity profiles using the Fuzzy C-Means clustering algorithm
	@param nbrClusters number of clusters used for the classification
	@param imageType specifies if only the positive side or negative side or both should be considered when generating the image
	@return pointer on a FuzzyCMeans class object */
	FuzzyCMeans * classifyByFuzzyCMeans(unsigned int nbrClusters, IMAGETYPE imageType);
	
	/** sort the Fuzzy C-Means classification
	@param outputFCM pointer on FuzzyCMeans
	@return pointer on Inrimage composed of classified intensity profiles */
	Inrimage * sortFCMClassification(FuzzyCMeans * outputFCM);

	/** sort the Expectation-Maximization classification
	@param outputEM pointer on ExpectationMaximization
	@return pointer on Inrimage composed of classified intensity profiles */
	Inrimage * sortEMClassification(ExpectationMaximization * outputEM);
};
}
#endif
