#ifndef TCLIMAGELINE_H
#define TCLIMAGELINE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma interface "imagedata/TclIntensityProfile.h"

#include <linearAlgebra/Vec3.h>
#include <inrimage/Inrimage.h>
#include <modules/TclObject.h>
#include <imagedata/RangeInrimage.h>
#include <imagedata/IntensityProfile.h>
#include <list>

namespace yav {

/** \class TclIntensityProfile
    \ingroup ImageData
    \brief Tcl interface of an Intensity Profile  */
class TclIntensityProfile :  public TclObject
{
 protected:
	/// the intensity profile
	IntensityProfile il;
	/// the index used for the display of image line
	unsigned int displayIndex;
 public:
	/** @name constructor */
	/** creates a Tcl Interface of an IntensityProfile by copying a previously existing IntensityProfile object
	@param m module to create TCL object
	@param il2 another image line object */
	TclIntensityProfile(TclModule *m, IntensityProfile &il2) : TclObject(typeid(TclIntensityProfile),m),il(il2)
	{ displayIndex = 0; done(typeid(TclIntensityProfile)); }

	/** creates a Tcl Interface of an IntensityProfile object given an image, 
	an origin position and a direction vector and the half-distance 
	@param m module to create TCL object
	@param inr a volumetric image 
	@param pos the origin position of the intensity profile 
	@param dir the direction of the intensity profile in the image
	@param dist the half distance of the intensity profile */
	TclIntensityProfile( TclModule *m,Inrimage *inr, Vec3<double> pos,Vec3<double> dir,double dist) : TclObject(typeid(TclIntensityProfile),m),il()
	{
		displayIndex = 0;
		il.setIntensity(inr);
		il.compute(dist,pos,dir);
		done(typeid(TclIntensityProfile));
	}

	/** @name destructor */
	/// destroys this interface
	virtual ~TclIntensityProfile() {}
	
	/// returns the reference of image line pointed by the interface
	const IntensityProfile *getIntensityProfile() const { return &il; }
	
	/// new TCL command callback
	static int list(TclModule *m, int argc, char **argv) { return TclObject::list(m, argc, argv); }
	
	/// new TCL command callback
	static int create(TclModule *m, int argc, char **argv);
	
	/// -destroy option
	int destroyCmd(int argc, char **argv);
	
	/// -draw option
	int drawCmd(int argc, char **argv);
	
	/// -dump option
	int dumpCmd(int argc, char **argv);
	
	/// -dumpGradient option
	int dumpGradientCmd(int argc, char **argv);
};

/** \class TclIntensityProfileSet
    \ingroup ImageData
    \brief Tcl interface of a set of Intensity Profiles  */
class TclIntensityProfileSet :  public TclObject
{
 protected:
	/// the list of intensity profiles
	IntensityProfileSet ils;
 public :
	/** @name constructor */
	/** creates a Tcl Interface of an IntensityProfile by copying a previously existing set of IntensityProfile objects
	@param m module to create TCL object
	@param il2 an array of image line objects */
	TclIntensityProfileSet(TclModule *m, std::vector<IntensityProfile *> &il2) : TclObject(typeid(TclIntensityProfileSet),m),ils(il2) { done(typeid(TclIntensityProfileSet)); }

	/** creates a Tcl Interface of an IntensityProfile by copying a previously existing IntensityProfileSet object
	@param m module to create TCL object
	@param il2 another image line set object */
	TclIntensityProfileSet(TclModule *m, IntensityProfileSet &il2) : TclObject(typeid(TclIntensityProfileSet),m),ils(il2) { done(typeid(TclIntensityProfileSet)); }

	/** creates a Tcl Interface of a set of intensity profiles using voxels given an image, 
	a list of osition and  direction vectors and the half-distance 
	@param m module to create TCL object
	@param inr a volumetric image 
	@param pos a list of position vector corresponding to the profile origin
	@param dir a list of direction vectors for each profile 
	@param dist the half distance of all intensity profiles
	@param step step along the intensity profile (default=1.0) */
	TclIntensityProfileSet( TclModule *m,Inrimage *inr, const std::vector< Vec3<double> > &pos, const std::vector< Vec3<double> > &dir, const double dist, const double step = 1.0);

	/** creates a Tcl Interface of an empty set of IntensityProfiles 
	@param m module to create TCL object */
	TclIntensityProfileSet( TclModule *m); 
	
	/** @name destructor */
	/// destroys this interface
	virtual ~TclIntensityProfileSet();
	
	/// returns the first iterator  of image line pointed by the interface
	std::vector<IntensityProfile *>::const_iterator getFirstIntensityProfile() const
	{ return ils.getFirstIntensityProfile(); }

	/// returns the last iterator of image line pointed by the interface
	std::vector<IntensityProfile *>::const_iterator  getLastIntensityProfile() const
	{ return ils.getLastIntensityProfile(); }

	/// adds an image line in the set 
	void addIntensityProfile(IntensityProfile *i) { ils.addIntensityProfile(i); }

	/// returns the number of image lines in the list
	unsigned int getNumberIntensityProfiles() const { return ils.getNumberIntensityProfiles(); }
	
	/// new TCL command callback
	static int list(TclModule *m, int argc, char **argv) { return TclObject::list(m, argc, argv); }

	/// new TCL command callback
	static int create(TclModule *m, int argc, char **argv);

	/// -addIntensityProfile option
	int addIntensityProfileCmd(int argc, char **argv);

	/// -destroy option
	int destroyCmd(int argc, char **argv);

	/// -draw option
	int drawCmd(int argc, char **argv);

	/// -fuzzyCMeans option
	int fuzzyCMeansCmd(int argc, char **argv);

	/// returns the list of intensity profiles
	IntensityProfileSet* getIntensityProfileSet() { return &ils; }
};
}
#endif
