#include "mipsVtkViewImage2D.h"
#include "mipsViewSpace.h"
#include "mipsViewSpaceLayout.h"
#include "vtkTransform.h"
#include "vtkImageReslice.h"
#include "vtkMatrix4x4.h"
#include "vtkImageData.h"
#include "vtkIndent.h"
#include "assert.h"

#include "mipsImageDataSet.h"

namespace mips {

ImageDataSet::ImageDataSet() :
  DataSet(), m_vtkImageData(0)
{
  m_reslicer  = vtkImageReslice::New();
  m_reslicer->InterpolateOff();
  m_reslicer->AutoCropOutputOn();
  m_transform = vtkTransform::New();  
}
  
ImageDataSet::~ImageDataSet()
{
  m_reslicer->Delete();
  m_transform->Delete();  
}

vtkMatrix4x4*  
ImageDataSet::GetOriginalTransform()
{
  return m_transform->GetMatrix();
}

void
ImageDataSet::SetOriginalTransform(vtkMatrix4x4* p_matrix)
{
  m_transform->Concatenate(p_matrix);   
  m_reslicer->SetResliceAxes(m_transform->GetMatrix());
  m_reslicer->GetOutput()->UpdateInformation();
  m_reslicer->GetOutput()->SetUpdateExtent(m_reslicer->GetOutput()->GetWholeExtent());
  m_reslicer->GetOutput()->Update();
}

vtkImageReslice*
ImageDataSet::GetReslicer()
{
  if (!m_vtkImageData) return 0;
  
  return m_reslicer;  
}


vtkImageData*
ImageDataSet::GetVtkImageData()
{
  if (!m_vtkImageData)      return 0;

  return   this->GetReslicer()->GetOutput();  
}

void
ImageDataSet::SetVtkImageData(vtkImageData* p_image)
{
  if (m_vtkImageData) m_vtkImageData->Delete();
  
  m_vtkImageData = p_image;
  m_vtkImageData->UpdateInformation();
  m_reslicer->SetInput(m_vtkImageData);  
  m_reslicer->GetOutput()->UpdateInformation();
  m_reslicer->GetOutput()->SetUpdateExtent(m_reslicer->GetOutput()->GetWholeExtent());
  m_reslicer->GetOutput()->Update();
}

void
ImageDataSet::CreatePageInViewSpace(ViewSpace* p_viewSpace)
{
  VtkViewImage2D* view = 0;
  int group_id = p_viewSpace->CreateViewGroup();

  p_viewSpace->AddDataSetPage(this);
  p_viewSpace->SetDataSetPageLayout(this, ViewSpaceLayout::GetGridLayout(2,2));  

  view = new VtkViewImage2D();
  p_viewSpace->AddViewInDataSetPage(this,view);
  p_viewSpace->SetViewGroup(view, group_id);  
  view->SetImage(this->GetVtkImageData());
  view->SetName(this->GetFileBaseName());  
  view->SetOrientation(VtkViewImage2D::AXIAL_ID);
  view->ResetCurrentPoint();
  view->ResetWindowLevel();

  view = new VtkViewImage2D();
  p_viewSpace->AddViewInDataSetPage(this,view,2);
  p_viewSpace->SetViewGroup(view, group_id);  
  view->SetImage(this->GetVtkImageData());
  view->SetName(this->GetFileBaseName());  
  view->SetOrientation(VtkViewImage2D::CORONAL_ID);
  view->ResetCurrentPoint();
  view->ResetWindowLevel();  

  view = new VtkViewImage2D();
  p_viewSpace->AddViewInDataSetPage(this,view,3);
  p_viewSpace->SetViewGroup(view, group_id);  
  view->SetImage(this->GetVtkImageData());
  view->SetName(this->GetFileBaseName());  
  view->SetOrientation(VtkViewImage2D::SAGITTAL_ID);
  view->ResetCurrentPoint();
  view->ResetWindowLevel();  

//  this->PrintSelf(std::cout, vtkIndent(0));
}



void
ImageDataSet::PrintSelf(ostream& os, vtkIndent indent)
{
  double* bounds   = 0;
  double* spacing  = 0;
  double* origin   = 0;
  int*    extent   = 0;
  int*    wextent  = 0;
  int*    uextent  = 0;
  
  m_vtkImageData->Update();  
  m_reslicer->GetOutput()->Update();  
  
  m_vtkImageData->GetBounds();  
  m_reslicer->GetOutput()->GetBounds();
    
  extent  = m_vtkImageData->GetExtent();  
  uextent = m_vtkImageData->GetUpdateExtent();  
  wextent = m_vtkImageData->GetWholeExtent();  
  origin  = m_vtkImageData->GetOrigin();
  spacing = m_vtkImageData->GetSpacing();
  os << indent << "ImageDataSet :\n";
  indent=indent.GetNextIndent();  
  vtkIndent indent2 =indent2.GetNextIndent();
  os << indent << "InputImage   - "<<m_vtkImageData->GetNumberOfPoints()<<" points.\n"
     << indent2 << " extent : ("
     << extent[0] << "/" << extent[1] << " ; "
     << extent[2] << "/" << extent[3] << " ; "
     << extent[4] << "/" << extent[5] << ") \n"  
     << indent2 << "uextent : ("
     << uextent[0] << "/" << uextent[1] << " ; "
     << uextent[2] << "/" << uextent[3] << " ; "
     << uextent[4] << "/" << uextent[5] << ") \n"  
     << indent2 << "wextent : ("
     << wextent[0] << "/" << wextent[1] << " ; "
     << wextent[2] << "/" << wextent[3] << " ; "
     << wextent[4] << "/" << wextent[5] << ") \n"  
     << indent2 << " origin : ("
     << origin[0] << " , " << origin[1] << " , " << origin[2] << ")\n"
     << indent2 << " spacing : ("
     << spacing[0] << " , " << spacing[1] << " , " << spacing[2] << ")\n";
  
  extent  = m_reslicer->GetOutput()->GetExtent();  
  uextent = m_reslicer->GetOutput()->GetUpdateExtent();  
  wextent = m_reslicer->GetOutput()->GetWholeExtent();  
  origin  = m_reslicer->GetOutput()->GetOrigin();
  spacing = m_reslicer->GetOutput()->GetSpacing();
  os << indent << "Output Image   - "<<m_reslicer->GetOutput()->GetNumberOfPoints()<<" points.\n"
     << indent2 << " extent : ("
     << extent[0] << "/" << extent[1] << " ; "
     << extent[2] << "/" << extent[3] << " ; "
     << extent[4] << "/" << extent[5] << ") \n" 
     << indent2 << "uextent : ("
     << uextent[0] << "/" << uextent[1] << " ; "
     << uextent[2] << "/" << uextent[3] << " ; "
     << uextent[4] << "/" << uextent[5] << ") \n"  
     << indent2 << "wextent : ("
     << wextent[0] << "/" << wextent[1] << " ; "
     << wextent[2] << "/" << wextent[3] << " ; "
     << wextent[4] << "/" << wextent[5] << ") \n"  
     << indent2 << " origin : ("
     << origin[0] << " , " << origin[1] << " , " << origin[2] << ")\n"
     << indent2 << " spacing : ("
     << spacing[0] << " , " << spacing[1] << " , " << spacing[2] << ")\n";
  
  bounds = m_vtkImageData->GetBounds();  
  os << indent << "InputImage \n"
     << indent2 << "bounds : ("
     << bounds[0] << "/" << bounds[1] << " ; "
     << bounds[2] << "/" << bounds[3] << " ; "
     << bounds[4] << "/" << bounds[5] << ") \n";
  bounds = m_reslicer->GetOutput()->GetBounds();  
  os << indent << "Output Image \n"
     << indent2 << "bounds : ("
     << bounds[0] << "/" << bounds[1] << " ; "
     << bounds[2] << "/" << bounds[3] << " ; "
     << bounds[4] << "/" << bounds[5] << ") \n";

  os << indent << "Reslice Axes : \n";
  if (m_reslicer->GetResliceAxes()) {    
    m_reslicer->GetResliceAxes()->PrintSelf(os, indent2);
  }
  else {
    os << indent2 << "No reslice axes matrix. \n";    
  }
    
}


}
