#include <gtk/gtk.h>
#include "mipsGladeGui.h"
#include "mipsViewSpace.h"
#include "mipsDataSetManager.h"
#include "mipsImageDataSet.h"
#include "mipsOrientationDialog.h"
#include "vtkImageData.h"
#include <iostream>
#include <assert.h>

#include "mipsImageViewer.h"

// LOCAL CALLBACK FUNCTION DECLARATIONS
static gboolean delete_event  (GtkWidget *widget,
                               GdkEvent  *event,
                               gpointer   data );
static void     destroy       (GtkWidget *widget,
                               gpointer   data );

// CLASS DEFINITION
namespace mips {

enum {
  DATASET_LIST_DATASET_POINTER_COLUMN,  // DataSet*
  DATASET_LIST_TYPE_PIXMAP_COLUMN,      // Icon associated to the type of the file (2D image, 3DImage, Mesh ...)
  DATASET_LIST_FILE_NAME_COLUMN,        // File name of the dataset
  DATASET_LIST_NB_COLUMNS               // Number of columns in the model.
};


ImageViewer::ImageViewer()
  : m_fileListTreeView(0)
{
  m_gladeGui = new mips::GladeGui("mipsImageViewer.glade");
  m_viewSpace = new  mips::ViewSpace;
  
  /* Set the default window size and callbacks*/
  m_mainWindow = m_gladeGui->GetWidget("main_window");
  gtk_window_set_title(GTK_WINDOW(m_mainWindow),"Mips Image Viewer");
  gtk_window_resize(GTK_WINDOW(m_mainWindow), 700,500);
  /* When the window is given the "delete_event" signal (this is given
   * by the window manager, usually by the "close" option, or on the
   * titlebar). */
  g_signal_connect (G_OBJECT (m_mainWindow), "delete_event",
                    G_CALLBACK (delete_event), NULL);
    
  /* Here we connect the "destroy" event to a signal handler.  
   * This event occurs when we call gtk_widget_destroy() on the window,
   * or if we return FALSE in the "delete_event" callback. */
  g_signal_connect (G_OBJECT (m_mainWindow), "destroy",
                    G_CALLBACK (destroy), NULL);
  
  /* Attach the viewspace to the viewer */
  GtkWidget *mipsviewer_hpaned;
  mipsviewer_hpaned = m_gladeGui->GetWidget("mipsviewer_hpaned");
  gtk_paned_add2(GTK_PANED(mipsviewer_hpaned), m_viewSpace->GetTopWidget());

  /* Initialise the main GtkTreeView file list */
  this->CreateFileList();

  // Initialize menu items
  GtkWidget* menuItem = m_gladeGui->GetWidget("save_as_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("orientation_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("properties_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("cut_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("copy_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("paste_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("modify_windowing_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("modify_windowing_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("create_preset_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("create_preset_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  menuItem = m_gladeGui->GetWidget("delete_preset_menu_item");  
  gtk_widget_set_sensitive (menuItem,FALSE);
  
  /* Initialise callbacks */
  m_gladeGui->SignalConnect("open_menu_item", "activate",
                            G_CALLBACK (ImageViewer::OnOpenMenuItemActivateCbk),
                            this);
  m_gladeGui->SignalConnect("orientation_menu_item", "activate",
                            G_CALLBACK (ImageViewer::OnOrientationMenuItemActivateCbk),
                            this);
}

void
ImageViewer::CreateFileList()
{
  m_fileListTreeView =
    GTK_TREE_VIEW(m_gladeGui->GetWidget("main_file_list_treeview"));

  GtkTreeSelection* tree_selection = gtk_tree_view_get_selection(m_fileListTreeView);  
  gtk_tree_selection_set_mode(tree_selection, GTK_SELECTION_BROWSE);
  g_signal_connect(tree_selection, "changed", (GCallback) ImageViewer::OnFileListSelectionChangedCbk, this);
  
  //
  // Finalize the TreeView column widgets creation.
  //
  GtkTreeViewColumn *column=0;
  GtkCellRenderer   *renderer=0;
  //       FILE_NAME column
  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new_with_attributes ("File Name",
                                                     renderer,
                                                     "text",
                                                     DATASET_LIST_FILE_NAME_COLUMN,
                                                     NULL);
  gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);    
  gtk_tree_view_insert_column (m_fileListTreeView, column,
                               DATASET_LIST_FILE_NAME_COLUMN);
  //       FILE_TYPE_PIXMAP column
  renderer = gtk_cell_renderer_pixbuf_new ();
  column = gtk_tree_view_column_new_with_attributes ("Type",
                                                     renderer,
                                                     "pixbuf",
                                                     DATASET_LIST_TYPE_PIXMAP_COLUMN,
                                                     NULL);
  gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);    
  gtk_tree_view_insert_column (m_fileListTreeView, column,
                               DATASET_LIST_TYPE_PIXMAP_COLUMN);

  
}

void
ImageViewer::RefreshFileList (DataSet* p_selectedDataSet)
{
  // Re create a new model with the list of DataSets stored in the DataSets Manager
  GtkTreeModel* fileListTreeModel =
    GTK_TREE_MODEL(gtk_list_store_new (DATASET_LIST_NB_COLUMNS,
                                       G_TYPE_POINTER,
                                       GDK_TYPE_PIXBUF, 
                                       G_TYPE_STRING));
  GtkTreeIter iter;
  GtkTreePath * path =0;
  unsigned int i =0;
  const std::vector<DataSet*>& datasets = mips::DataSetManager::GetDataSets();
  for (i=0; i<datasets.size() ; i++) {
    // Add a line in model
    gtk_list_store_append(GTK_LIST_STORE(fileListTreeModel), &iter);
    gtk_list_store_set (GTK_LIST_STORE(fileListTreeModel), &iter,
                        DATASET_LIST_DATASET_POINTER_COLUMN, datasets[i],
                        DATASET_LIST_TYPE_PIXMAP_COLUMN, 0,//datasets[i]->GetPixmap(),
                        DATASET_LIST_FILE_NAME_COLUMN, datasets[i]->GetFileBaseName() ,
                        -1);
    // store the path of the dataset to select if corresponding
    if (datasets[i] == p_selectedDataSet) {
      path = gtk_tree_model_get_path(fileListTreeModel,  &iter);
    }    
  }
  gtk_tree_view_set_model(m_fileListTreeView, fileListTreeModel);

  // Select the dataset in the TreeView => activate the OnFileListSelectionChangedCbk  
  gtk_tree_selection_select_path(gtk_tree_view_get_selection(m_fileListTreeView),
                                 path);
  
  gtk_tree_path_free(path);  
}

// Hack to follow the memory release after on path_list in the
// OnFileListSelectionChanged following method.
// See gtk doc :
// http://developer.gnome.org/doc/API/2.0/gtk/GtkTreeSelection.html#gtk-tree-selection-get-selected-rows
void  _gtk_tree_path_free (gpointer data,gpointer user_data)
{
  gtk_tree_path_free((GtkTreePath*)data);  
}
DataSet*
ImageViewer::GetSelectedDataSet()
{
  GtkTreeSelection* selection = gtk_tree_view_get_selection(m_fileListTreeView);

  GtkTreeIter   iter;
  GtkTreePath*  path = 0;
  GtkTreeModel* model = 0;
  DataSet*      ds = 0;  
  // Getting an iterator on the model row corresponding to the first selected
  // row in the TreeView
  GList* path_list = gtk_tree_selection_get_selected_rows(selection,
                                                          &model);
  if (path_list == NULL) return 0;
  
  path = (GtkTreePath*)g_list_nth_data(path_list,0);
  gtk_tree_model_get_iter(model, &iter, path);
  g_list_foreach (path_list, _gtk_tree_path_free, NULL);
  g_list_free (path_list);

  // Getting the DataSet
  gtk_tree_model_get(model, &iter,
                     DATASET_LIST_DATASET_POINTER_COLUMN, &ds,
                     -1);
  return ds;  
}

vtkImageData*
ImageViewer::GetImageDataFromDataSet(DataSet* p_dataSet)
{
  ImageDataSet* imageDS= dynamic_cast<ImageDataSet*>(p_dataSet);
  if (imageDS)
    return imageDS->GetVtkImageData();  
  return 0;  
}

void
ImageViewer::OnOpenMenuItemActivate ()
{
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new ("Open File",
                                        GTK_WINDOW(m_mainWindow),
                                        GTK_FILE_CHOOSER_ACTION_OPEN,
                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                        NULL);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
  {
    char *filename;    
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    DataSet* ds = DataSetManager::OpenDataSet(filename);    
    if(! ds) {     
      GtkWidget* err_dialog =
        gtk_message_dialog_new (GTK_WINDOW(m_mainWindow),
                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                GTK_MESSAGE_ERROR,
                                GTK_BUTTONS_CLOSE,
                                "Error loading file '%s'",
                                filename);
      gtk_dialog_run (GTK_DIALOG (err_dialog));
      gtk_widget_destroy (err_dialog);
    }
    else {
      this->RefreshFileList(ds);
      ds->CreatePageInViewSpace(m_viewSpace);
      m_viewSpace->DisplayDataSetPage(ds);      
      GtkWidget* menuItem = m_gladeGui->GetWidget("orientation_menu_item");  
      gtk_widget_set_sensitive (menuItem,TRUE);
    }
    g_free (filename);
  }
  gtk_widget_destroy (dialog);
}

void
ImageViewer::OnOpenMenuItemActivateCbk (GtkMenuItem *p_menuitem, gpointer p_imageViewer)
{
  static_cast<ImageViewer*>(p_imageViewer)->OnOpenMenuItemActivate();
}


void
ImageViewer::OnOrientationMenuItemActivate ()
{
  OrientationDialog dialog(m_mainWindow);
  ImageDataSet* image_ds = static_cast<ImageDataSet*>(this->GetSelectedDataSet());
  if (!image_ds) return;
  if (dialog.Run(image_ds) == GTK_RESPONSE_OK)
  {
    // Refresh the page
    m_viewSpace->DisplayDataSetPage(image_ds);    
  }  
}
void
ImageViewer::OnOrientationMenuItemActivateCbk (GtkMenuItem *p_menuitem, gpointer p_imageViewer)
{
  static_cast<ImageViewer*>(p_imageViewer)->OnOrientationMenuItemActivate();
}


void
ImageViewer::OnFileListSelectionChanged(GtkTreeSelection* p_treeSelection)
{
  DataSet*      ds = this->GetSelectedDataSet();  
  m_viewSpace->DisplayDataSetPage(ds);  

  // Change the title of the main_window to the name of the selected DataSet
  if (ds) {    
    gtk_window_set_title(GTK_WINDOW(m_mainWindow),
                         ds->GetFileBaseName());
  }
  else {
    gtk_window_set_title(GTK_WINDOW(m_mainWindow),"MipsImageViewer");    
  }  
}

void
ImageViewer::OnFileListSelectionChangedCbk(GtkTreeSelection *p_treeSelection,
                                           gpointer p_imageViewer)
{
  static_cast<ImageViewer*>(p_imageViewer)->OnFileListSelectionChanged(p_treeSelection);
}

} //namespace mips


// LOCAL CALLBACK FUNCTION DEFINITIONS
static gboolean delete_event  (GtkWidget *widget,
                               GdkEvent  *event,
                               gpointer   data )
{
  /* If you return FALSE in the "delete_event" signal handler,
   * GTK will emit the "destroy" signal. Returning TRUE means
   * you don't want the window to be destroyed.
   * This is useful for popping up 'are you sure you want to quit?'
   * type dialogs. */

  // g_print ("delete event occurred\n");
  gtk_main_quit ();

  /* Change TRUE to FALSE and the main window will be destroyed with
   * a "delete_event". */
  return TRUE;
}

static void destroy( GtkWidget *widget,
                     gpointer   data )
{
  // g_print ("destroy event occurred\n");
  gtk_main_quit ();
}

