#include "mipsGladeGui.h"
#include "mipsImageDataSet.h"
#include "mipsVtkViewImage2D.h"
#include "vtkImageData.h"
#include "vtkTransform.h"

#include "mipsOrientationDialog.h"

namespace mips {

enum DirectionIds {
  _UNDEFINED_,          
  _ANTERIOR_,       
  _POSTERIOR_,
  _INFERIOR_,
  _SUPERIOR_,
  _LEFT_,
  _RIGHT_
};

  
OrientationDialog::OrientationDialog (GtkWidget* p_parentTopWidget)
  :m_parentTopWidget(p_parentTopWidget), m_transform(0)
{  
  // Load the main viewspace interface
  m_gladeGui = new mips::GladeGui("mipsOrientationDialog.glade");

  // Get and initialize the orientationDialog
  m_orientationDialogWidget = m_gladeGui->GetWidget("orientation_dialog");
  gtk_widget_hide_all(m_orientationDialogWidget);

  // Add views
  m_xyAxialView = new VtkViewImage2D();
  m_xyAxialView->SetOrientation(VtkViewImage2D::AXIAL_ID);
  gtk_container_add (GTK_CONTAINER(m_gladeGui->GetWidget("xy_axial_viewport")),
                     m_xyAxialView->GetGtkWidget());
  m_yzSagittalView = new VtkViewImage2D();
  m_yzSagittalView->SetOrientation(VtkViewImage2D::SAGITTAL_ID);
  gtk_container_add (GTK_CONTAINER(m_gladeGui->GetWidget("yz_sagittal_viewport")),
                     m_yzSagittalView->GetGtkWidget());  
}

OrientationDialog::~OrientationDialog ()
{
  delete m_gladeGui;  
}

gint
OrientationDialog::Run(ImageDataSet* p_image)
{
  // Set image in the views
  m_xyAxialView->SetImage(p_image->GetVtkImageData());
  m_xyAxialView->ResetWindowLevel();
  m_xyAxialView->ResetCurrentPoint();
  m_xyAxialView->SetShowDirections(false);
  m_xyAxialView->SetShowCurrentPoint(false);  
  m_xyAxialView->SetShowSliceNumber(false);  
  m_yzSagittalView->SetImage(p_image->GetVtkImageData());
  m_yzSagittalView->ResetWindowLevel();
  m_yzSagittalView->ResetCurrentPoint();
  m_yzSagittalView->SetShowDirections(false);
  m_yzSagittalView->SetShowCurrentPoint(false);
  m_yzSagittalView->SetShowSliceNumber(false);

  // Set the combo boxes
  m_xStartComboBox = GTK_COMBO_BOX(m_gladeGui->GetWidget("x_start_combobox"));
  m_yStartComboBox = GTK_COMBO_BOX(m_gladeGui->GetWidget("y_start_combobox"));
  m_zEndComboBox   = GTK_COMBO_BOX(m_gladeGui->GetWidget("z_end_combobox"));
  gtk_combo_box_set_active(m_xStartComboBox, 0);
  gtk_combo_box_set_active(m_yStartComboBox, 0);
  gtk_combo_box_set_active(m_zEndComboBox,   0);

  // Set transient now to take care of reparent operations on the ViewSpace
  gtk_window_set_transient_for(GTK_WINDOW(m_orientationDialogWidget),
                               GTK_WINDOW(gtk_widget_get_toplevel(m_parentTopWidget)));
  // Reset transformation
  if(m_transform)   m_transform->Delete();
  m_transform = vtkTransform::New();

  // Show dialog
  gtk_widget_show_all(m_orientationDialogWidget);
  gint response = gtk_dialog_run(GTK_DIALOG(m_orientationDialogWidget));  
  if (response == GTK_RESPONSE_OK){    
    this->ComputeTransform(gtk_combo_box_get_active(m_xStartComboBox),
                           gtk_combo_box_get_active(m_yStartComboBox),
                           gtk_combo_box_get_active(m_zEndComboBox));    
    p_image->SetOriginalTransform(m_transform->GetMatrix());
  }

  // Hide the dialog again
  gtk_widget_hide(m_orientationDialogWidget);

  return response;  
}

int
_inverseOrientation(int p_orientation)
{
  switch (p_orientation) { 
  case _RIGHT_:
    return _LEFT_;
  case _LEFT_:
    return _RIGHT_;
  case _ANTERIOR_:
    return _POSTERIOR_;
  case _POSTERIOR_:
    return _ANTERIOR_;
  case _SUPERIOR_:
    return _INFERIOR_;
  case _INFERIOR_:
    return _SUPERIOR_;
  case _UNDEFINED_:
    return _UNDEFINED_;
  default:
    std::cout<<"ERROR in _inverseOrientation function in mipsOrientationDialog.cxx :"
             << p_orientation << "\n";
  }
  exit(1);
}

void
OrientationDialog::ComputeTransform(int p_x_start,
                                    int p_y_start,
                                    int p_z_end)
{
  unsigned int nb_modifications = 0;
  bool modified;
  bool fliped = false;      
  do {
    modified = false;    
    // Rotate around axis Z (Axial)
    // We want p_y_start = _ANTERIOR_
    switch (p_y_start) { 
    case _RIGHT_:
      m_transform->RotateZ(90);
      p_y_start = p_x_start;
      p_x_start = _RIGHT_;
      nb_modifications++;
      modified = true;
      break;    
    case _POSTERIOR_:
      m_transform->RotateZ(90);    
      m_transform->RotateZ(90);
      p_y_start = _ANTERIOR_;
      p_x_start = _inverseOrientation(p_x_start);
      nb_modifications++;
      modified = true;
      break;    
    case _LEFT_:
      m_transform->RotateZ(-90);
      p_y_start =  _inverseOrientation(p_x_start);
      p_x_start = _RIGHT_;
      nb_modifications++;
      modified = true;
      break;    
    }

    // Rotate around axis Y (Coronal)
    // We want p_x_start = _RIGHT_
    switch (p_x_start) {
    case _INFERIOR_:
      m_transform->RotateY(90);    
      p_x_start = p_z_end;
      p_z_end   = _SUPERIOR_;
      nb_modifications++;
      modified = true;
      break;    
    case _LEFT_:
      m_transform->RotateY(90);    
      m_transform->RotateY(90);
      p_x_start = _RIGHT_;
      p_z_end   = _inverseOrientation(p_z_end);
      nb_modifications++;
      modified = true;
      break;    
    case _SUPERIOR_:
      m_transform->RotateY(-90);
      p_x_start =  _inverseOrientation(p_z_end);
      p_z_end   = _SUPERIOR_;
      nb_modifications++;
      modified = true;
      break;    
    }
  
    switch (p_z_end) {
    case _POSTERIOR_:
      m_transform->RotateX(90);    
      p_z_end   = p_y_start;
      p_y_start = _ANTERIOR_;
      nb_modifications++;
      modified = true;
      break;    
    case _INFERIOR_:
      m_transform->RotateX(90);    
      m_transform->RotateX(90);
      p_z_end   = _SUPERIOR_;    
      p_y_start = _inverseOrientation(p_y_start);
      nb_modifications++;
      modified = true;
      break;    
    case _ANTERIOR_:
      m_transform->RotateX(-90);
      p_z_end   = _inverseOrientation(p_y_start);
      p_y_start = _ANTERIOR_;    
      nb_modifications++;
      modified = true;
      break;    
    }

    if (!fliped && (nb_modifications > 3)) {
      this->FlipY();
      p_y_start = _inverseOrientation(p_y_start);
      modified = true;
      nb_modifications = 1; // We reset the number of modifications
      fliped = true;      
    }    
  } while (modified && (nb_modifications <=3 ));

  // If more than 3 modifications have been done.
  // the algorithm have a problem.
  if (nb_modifications > 3) {    
    std::cout << "Warning : Orientation algorithm failed !\n";
  }
}

void
OrientationDialog::FlipY()
{
  // Play with transforms
  vtkMatrix4x4* mat = vtkMatrix4x4::New();
  mat->SetElement (0,0,  1); // x
  mat->SetElement (1,0,  0);
  mat->SetElement (2,0,  0);
  mat->SetElement (3,0,  0);
  mat->SetElement (0,1,  0); // y
  mat->SetElement (1,1,  -1);
  mat->SetElement (2,1,  0);
  mat->SetElement (3,1,  0);  
  mat->SetElement (0,2,  0); // z
  mat->SetElement (1,2,  0);
  mat->SetElement (2,2,  1);
  mat->SetElement (3,2,  0);
  // Translation
  mat->SetElement (0,3,  0);
  mat->SetElement (1,3,  0);
  mat->SetElement (2,3,  0);
  mat->SetElement (3,3,  1);
  m_transform->Concatenate(mat);
  mat->Delete();  
}

void
OrientationDialog::RotateX90(vtkImageData* p_image)
{

  double* spacing = p_image->GetSpacing();
  int*    dim     = p_image->GetDimensions();
  double* origin  = p_image->GetOrigin();

  // Play with transforms
  vtkMatrix4x4* mat = vtkMatrix4x4::New();
  mat->SetElement (0,0,  1); // x
  mat->SetElement (1,0,  0);
  mat->SetElement (2,0,  0);
  mat->SetElement (3,0,  0);
  mat->SetElement (0,1,  0); // y
  mat->SetElement (1,1,  0);
  mat->SetElement (2,1, -1);
  mat->SetElement (3,1,  0);  
  mat->SetElement (0,2,  0); // z
  mat->SetElement (1,2,  1);
  mat->SetElement (2,2,  0);
  mat->SetElement (3,2,  0);
  // Translation
  mat->SetElement (0,3,  0);
  mat->SetElement (1,3,  origin[1] + origin[2]);
  mat->SetElement (2,3,  origin[2] + origin[1]+ dim[1]*spacing[1] );
  mat->SetElement (3,3,  1);
  mat->Invert();  
  m_transform->Concatenate(mat);
  mat->Delete();  
}

void
OrientationDialog::RotateY90(vtkImageData* p_image)
{

  double* spacing = p_image->GetSpacing();
  int*    dim     = p_image->GetDimensions();
  double* origin  = p_image->GetOrigin();

  // Play with transforms
  vtkMatrix4x4* mat = vtkMatrix4x4::New();
  mat->SetElement (0,0,  0); // x
  mat->SetElement (1,0,  0);
  mat->SetElement (2,0,  1);
  mat->SetElement (3,0,  0);
  mat->SetElement (0,1,  0); // y
  mat->SetElement (1,1,  1);
  mat->SetElement (2,1,  0);
  mat->SetElement (3,1,  0);  
  mat->SetElement (0,2, -1); // z
  mat->SetElement (1,2,  0);
  mat->SetElement (2,2,  0);
  mat->SetElement (3,2,  0);
  // Translation
  mat->SetElement (0,3,  origin[0] + origin[2]+ dim[2]*spacing[2] );
  mat->SetElement (1,3,  0);
  mat->SetElement (2,3,  origin[2] + origin[0]);
  mat->SetElement (3,3,  1);
  m_transform->Concatenate(mat);
  mat->Delete();  
}

void
OrientationDialog::RotateZ90(vtkImageData* p_image)
{

  double* spacing = p_image->GetSpacing();
  int*    dim     = p_image->GetDimensions();
  double* origin  = p_image->GetOrigin();

  // Play with transforms
  vtkMatrix4x4* mat = vtkMatrix4x4::New();
  mat->SetElement (0,0,  0); // x
  mat->SetElement (1,0, -1);
  mat->SetElement (2,0,  0);
  mat->SetElement (3,0,  0);
  mat->SetElement (0,1,  1); // y
  mat->SetElement (1,1,  0);
  mat->SetElement (2,1,  0);
  mat->SetElement (3,1,  0);  
  mat->SetElement (0,2,  0); // z
  mat->SetElement (1,2,  0);
  mat->SetElement (2,2,  1);
  mat->SetElement (3,2,  0);
  // Translation
  mat->SetElement (0,3,  origin[0] + origin[1]+ dim[1]*spacing[1] );
  mat->SetElement (1,3,  origin[1] - origin[0]);
  mat->SetElement (2,3,  0);
  mat->SetElement (3,3,  1);
  mat->Invert();
  m_transform->Concatenate(mat);
  mat->Delete();  
}

} //namespace mips
