/******************************************************************************/
/* ChamferMap.h                                                               */
/* Template class managing the chamfer map calculation on a buffer            */
/******************************************************************************/

#ifndef CHAMFERMAP_H
#define CHAMFERMAP_H



#ifdef _YAV_ // namespace for INIRA
#include "ChamferMask.h"
namespace yav {
#define DISTANCEMAP_API 
#else
#include "distancemap/ChamferMask.h"
#include "distancemap/distancemapAPI.h"
#endif

  enum { // Initialization flags :  
    // used for chamfer map initialization on Large Data (LDA in Amira)
    _DistanceMap_NI,    // normal initialization
    
    // keep the old values of the image's border during initialization, for :
    _DistanceMap_FS,   // the first Slices
    _DistanceMap_FL,   // the first Lines
    _DistanceMap_FC,   // first Columns
    _DistanceMap_FSL,  // both the first Slices and fisrt Lines
    _DistanceMap_FSC,  // both the first Slices and first Columns
    _DistanceMap_FLC,  // both the first Lines  and first Columns
    _DistanceMap_FSLC, // the first Slices and the first Lines and the first Columns

    _DistanceMap_LS,   // the last Slices
    _DistanceMap_LL,   // the last Lines
    _DistanceMap_LC,   // the last Columns
    _DistanceMap_LSL,  // both the last Slices and last Lines
    _DistanceMap_LSC,  // both the last Slices and last Columns
    _DistanceMap_LLC,  // both the last Lines  and last Columns
    _DistanceMap_LSLC, // the last Slices and the last Lines and the last Columns

  };

/* class ChamferMap : template class to compute distance/norme map on a buffer */

template <class DATATYPE>
class DISTANCEMAP_API ChamferMap
{
public :
  /* Constructors */
  //
  ChamferMap(DATATYPE infinity, int maskSize = 3, int maskDim = 3, float voxelSize = 1.0);

  //
  ChamferMap(DATATYPE infinity, float dy, float dz, float voxelSize = 1.0);

  /* Destructor */
  //
  ~ChamferMap();

  /* Accessors */
  int getIBegin();
  int getJBegin();
  int getKBegin();


  /* Public methods */
  //
  void computeChamferMap(DATATYPE * buffer, const int * dims);
  //
  void computeChamferMapOut(DATATYPE * buffer, const int * dims);
  //
  void computeChamferMapInOut(DATATYPE * buffer, const int * dims);
  //
  void initMasksOffsets(const int * dims);
  //
  float getExactValue(DATATYPE * buffer, int pos);
  //
  float * getExactMap(DATATYPE * buffer, const int * dims);
  //
  float getMaxRelativeError();

  /* Methods for the inside of the object */
  //
  void initBoundaries(DATATYPE * buffer, const int * dims, int flag = _DistanceMap_NI);
  //
  void forwardRun(DATATYPE * buffer, const int * dims);
  //

  void backwardRun(DATATYPE * buffer, const int * dims);
  /* Methods for the outside of the object */
  //
  void initBoundariesOut(DATATYPE * buffer, const int * dims, int flag = _DistanceMap_NI);
  //
  void forwardRunOut(DATATYPE * buffer, const int * dims);
  //
  void backwardRunOut(DATATYPE * buffer, const int * dims);
  /* Methods for both the inside and the outside of the object */
  //
  void initBoundariesInOut(DATATYPE * buffer, const int * dims, int flag = _DistanceMap_NI);
  //
  void forwardRunInOut(DATATYPE * buffer, const int * dims);
  //
  void backwardRunInOut(DATATYPE * buffer, const int * dims);


protected :
  /* Protected methods */

  /* Methods for the inside of the object */
  //
  void _initFirstSlices (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastSlices  (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initFirstLines  (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastLines   (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initFirstColumns(DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastColumns (DATATYPE ** buffer, const int * dims, int flag);

  //
  void _forwardRunFirstSlices (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunFirstLines  (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunLastLines   (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunFirstColumns(DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunLastColumns (DATATYPE ** buffer, const int * dims);

  //
  void _backwardRunLastSlices  (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunLastLines   (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunFirstLines  (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunLastColumns (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunFirstColumns(DATATYPE ** buffer, const int * dims);

  /* Methods for the outside of the object */
  //
  void _initFirstSlicesOut (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastSlicesOut  (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initFirstLinesOut  (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastLinesOut   (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initFirstColumnsOut(DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastColumnsOut (DATATYPE ** buffer, const int * dims, int flag);

  //
  void _forwardRunFirstSlicesOut (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunFirstLinesOut  (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunLastLinesOut   (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunFirstColumnsOut(DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunLastColumnsOut (DATATYPE ** buffer, const int * dims);

  //
  void _backwardRunLastSlicesOut  (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunLastLinesOut   (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunFirstLinesOut  (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunLastColumnsOut (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunFirstColumnsOut(DATATYPE ** buffer, const int * dims);


  /* Methods for both the inside and the outside of the object */
  //
  void _initFirstSlicesInOut (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastSlicesInOut  (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initFirstLinesInOut  (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastLinesInOut   (DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initFirstColumnsInOut(DATATYPE ** buffer, const int * dims, int flag);
  //
  void _initLastColumnsInOut  (DATATYPE ** buffer, const int * dims, int flag);

  //
  void _forwardRunFirstSlicesInOut (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunFirstLinesInOut  (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunLastLinesInOut   (DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunFirstColumnsInOut(DATATYPE ** buffer, const int * dims);
  //
  void _forwardRunLastColumnsInOut (DATATYPE ** buffer, const int * dims);

  //
  void _backwardRunLastSlicesInOut  (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunLastLinesInOut   (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunFirstLinesInOut  (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunLastColumnsInOut (DATATYPE ** buffer, const int * dims);
  //
  void _backwardRunFirstColumnsInOut(DATATYPE ** buffer, const int * dims);


  /* Fields */
  //
  ChamferMask * mask;
  //
  ChamferMask * maskInit;
  //
  DATATYPE _infinity;
  //
  float voxelSize;

#ifdef _YAV_
//#include "ChamferMapOpt.h"

// Originally, the header was separated so that the code within ChamferMapOpt.h
// and ChamferMapOpt.cpp was not provided to Amira. Now, so that doxygen
// doesn't screw up, ChamferMapOpt.h has been merged within this file.

public :
  /** @name Constructors */
  //@{
  ///
  ChamferMap(DATATYPE infinity, int flag = POSITION_PROPAGATION, 
		int maskSize = 3, int maskDim = 3, float voxelSize = 1.0);
  ///
  ChamferMap(DATATYPE infinity, int flag,
		float dy, float dz, float voxelSize = 1.0);
  //@}

  /** @name Public methods */
  //@{

  ///
  void initOutput(DATATYPE * distMap, DATATYPE * buffer, const int * dims);

  ///
  void computeChamferMap(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  ///
  void computeChamferMapOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  ///
  void computeChamferMapInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims);

  /** @name Methods for the inside of the object */
  //@{
  ///
  void initBoundaries(DATATYPE * distMap, DATATYPE * buffer, const int * dims, int flag = _DistanceMap_NI);
  ///
  void forwardRun(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  ///
  void backwardRun(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  //@}
  /** @name Methods for the outside of the object */
  //@{
  ///
  void initBoundariesOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims, int flag = _DistanceMap_NI);
  ///
  void forwardRunOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  ///
  void backwardRunOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  //@}
  /** @name Methods for both the inside and the outside of the object */
  //@{
  ///
  void initBoundariesInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims, int flag = _DistanceMap_NI);
  ///
  void forwardRunInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  ///
  void backwardRunInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims);
  //@}

protected :
  /** @name Protected methods */
  //@{

  /** @name Methods for the inside of the object */
  //@{
  ///
  void _initFirstSlices (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastSlices  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initFirstLines  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastLines   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initFirstColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastColumns (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);

  ///
  void _forwardRunFirstSlices (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunFirstLines  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunLastLines   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunFirstColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunLastColumns (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);

  ///
  void _backwardRunLastSlices  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunLastLines   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunFirstLines  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunLastColumns (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunFirstColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  //@}

  /** @name Methods for the outside of the object */
  //@{
  ///
  void _initFirstSlicesOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastSlicesOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initFirstLinesOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastLinesOut   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initFirstColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastColumnsOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);

  ///
  void _forwardRunFirstSlicesOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunFirstLinesOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunLastLinesOut   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunFirstColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunLastColumnsOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);

  ///
  void _backwardRunLastSlicesOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunLastLinesOut   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunFirstLinesOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunLastColumnsOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunFirstColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);

  //@}

  /** @name Methods for both the inside and the outside of the object */
  //@{
  ///
  void _initFirstSlicesInOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastSlicesInOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initFirstLinesInOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastLinesInOut   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initFirstColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);
  ///
  void _initLastColumnsInOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag);

  ///
  void _forwardRunFirstSlicesInOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunFirstLinesInOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunLastLinesInOut   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunFirstColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _forwardRunLastColumnsInOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);

  ///
  void _backwardRunLastSlicesInOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunLastLinesInOut   (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunFirstLinesInOut  (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunLastColumnsInOut (DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  ///
  void _backwardRunFirstColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims);
  //@}

  /** @name Fields */
  //@{
  ///
  int optFlag;
///
int currentPos;
  //@}
#endif

};

#ifdef _YAV_
} // namespace yav
#endif

#include "ChamferMap.cpp"


#ifdef _YAV_
#include "ChamferMapOpt.cpp"
#endif

#endif 


