/****************************************************************************/
/* ChamferMapOpt.cpp                                                           */
/****************************************************************************/

#ifdef _YAV_ // namespace for INRIA
namespace yav {
#endif

/****************************************************************************/
/*                                                                          */
/*                               CONSTRUCTORS                               */
/*                                                                          */
/****************************************************************************/
template <class DATATYPE>
ChamferMap<DATATYPE>::ChamferMap(DATATYPE infinity, int flag,
				       int maskSize, int maskDim, float voxelSize)
  //  : ChamferMap<DATATYPE>(infinity, maskSize, maskDim, voxelSize)
{
  this->_infinity = infinity;
  this->voxelSize = voxelSize;
  maskInit = new ChamferMask(maskSize, maskDim, 1);
  mask = new ChamferMask(maskSize, maskDim, 0);
  optFlag = flag;
}

template <class DATATYPE>
ChamferMap<DATATYPE>::ChamferMap(DATATYPE infinity, int flag,
				       float dy, float dz, float voxelSize)
  //  : ChamferMap<DATATYPE>(infinity, dy, dz, voxelSize)
{
  MaskCalculation * myTool;
  this->_infinity = infinity;
  this->voxelSize = voxelSize;

  if ((dy - 1.0) < 0.000001)
    myTool = (MaskCalcAnisotropZ *)  new MaskCalcAnisotropZ(dz);
  else
    myTool = (MaskCalcAnisotropYZ *) new MaskCalcAnisotropYZ(dy, dz);
  
  myTool->wheightsCalc("maskResult/aMaskCalc", 20, 1);
  unsigned char *** maskCoefs = myTool->getFinalMask();

  maskInit = new ChamferMask(3, maskCoefs, myTool->getEpsilon());
 
  for (int k = 0; k < 2; k++)
    for (int j = 0; j < 2; j++)
      for (int i = 0; i < 2; i++)
	maskCoefs[i][j][k] = 2*maskCoefs[i][j][k];
  
  mask = new ChamferMask(3, maskCoefs, 2*myTool->getEpsilon());
  optFlag = flag;
}

/****************************************************************************/
/*                                                                          */
/*                             DESTRUCTORS                                  */
/*                                                                          */
/****************************************************************************/
// template <class DATATYPE>
// ChamferMap<DATATYPE>::~ChamferMap()
// {
// }


/****************************************************************************/
/*                                                                          */
/*                             PUBLIC METHODS                               */
/*                                                                          */
/****************************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::initOutput(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  int i,j;
  int dimT = dims[0]*dims[1]*dims[2];
  switch (optFlag)
    {
    case POSITION_PROPAGATION :
      for (i = 0; i < dimT; i++) 
	{
	  if (distMap[i] > 0)
	    buffer[i] = 1;
	  else
	    buffer[i] = -1;
	}
      break;
    case COLOR_PROPAGATION :
      for (i = 0; i < dimT; i++)
	buffer[i] = distMap[i];
      for ( j = 0; j < dimT; j++)
	{
	  if (buffer[j] > (DATATYPE) 0)
	    distMap[j] = 0;
	  else
	    distMap[j] = 1;
	}
      break;
    default :
      break;
    } // end switch(optFlag)
}
/****************************************************************************/
/*                             General Methods                              */
/****************************************************************************/
/*Inside ********************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::computeChamferMap(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  initMasksOffsets(dims);
  initOutput(distMap, buffer, dims);
  initBoundaries(distMap, buffer, dims);
  forwardRun(distMap, buffer, dims);
  backwardRun(distMap, buffer, dims);
}

/*Outside *******************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::computeChamferMapOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  initMasksOffsets(dims);
  initOutput(distMap, buffer, dims);
  initBoundariesOut(distMap, buffer, dims);
  forwardRunOut(distMap, buffer, dims);
  backwardRunOut(distMap, buffer, dims);

}
/*Both inside and outside ***************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::computeChamferMapInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  initMasksOffsets(dims);
  initOutput(distMap, buffer, dims);
  initBoundariesInOut(distMap, buffer, dims);
  forwardRunInOut(distMap, buffer, dims);
  backwardRunInOut(distMap, buffer, dims);
}

/****************************************************************************/
/*                             Initialization                               */
/****************************************************************************/
/*Inside ********************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::initBoundaries(DATATYPE * distMap, DATATYPE * buffer, const int * dims, int flag)
{
  int i, j, k, m;
  
  DATATYPE * pixel = distMap;
  DATATYPE * neighbor;
  DATATYPE * colorPixel = buffer;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  int maskSize                 = maskInit->getMaskSize();
  int * neighborhood           = maskInit->getNeighborhood();
  unsigned char * chamferMask  = maskInit->getChamferMask();


  currentPos = 0;
  // for (k = 0; k < kBegin; k++)
  _initFirstSlices (&pixel, &colorPixel, dims, flag);
  for (k = kBegin; k < dims[2] - kBegin; k++)
    {
      // for (j = 0; j < jBegin; j++)
      _initFirstLines (&pixel, &colorPixel, dims, flag);
      for (j = jBegin; j < dims[1] - jBegin; j++)
	{
	  // for (i = 0; i < iBegin; i++)
	  _initFirstColumns (&pixel, &colorPixel, dims, flag);
	  for (i = iBegin; i < dims[0] - iBegin; i++)
	    {
	      // we consider the object pixels (the others remain to zero)
	      if (*pixel > 0 )
		{
		  // if the pixel is not on the border, it gets the infinity value
		  *pixel = _infinity;
		  
		  for (m = 0; m < maskSize; m++) {
		    // take the closest pixel to the background 
		    // in the forward neighborhood
		    neighbor = pixel + neighborhood[m];
		    if ((*neighbor <= 0) && ((DATATYPE) chamferMask[m] < *pixel)) {
		      *pixel = (DATATYPE) chamferMask[m];
		      colorNeighbor = colorPixel + neighborhood[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  *colorPixel = (DATATYPE) (currentPos + neighborhood[m]);
			  break;
			default :
			  break;
			}
		    }
		    
		    // and in the backward one
		    neighbor = pixel - neighborhood[m];
		    if ((*neighbor <= 0) && ((DATATYPE) chamferMask[m] < *pixel)) {
		      *pixel = (DATATYPE) chamferMask[m];
		      colorNeighbor = colorPixel - neighborhood[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  *colorPixel = (DATATYPE) (currentPos - neighborhood[m]);
			  break;
			default :
			  break;
			} 
		    } 
		  }
		}
	      else 
		*pixel = 0;
	      
	      pixel++;
	      colorPixel++;
	      currentPos++;
	    }
	  // for (i = dims[0] - iBegin; i < dims[0]; i++)
	  _initLastColumns (&pixel, &colorPixel, dims, flag);
	}
      // for (j = dims[1] - jBegin; j < dims[1]; j++)
      _initLastLines (&pixel, &colorPixel, dims, flag);
    }
  // for (k = dims[2] - kBegin; k < dims[2]; k++)
  _initLastSlices (&pixel, &colorPixel, dims, flag);
  

}


/*Outside *******************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::initBoundariesOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims, int flag)
{
  int i, j, k, m, p;

  DATATYPE * pixel = distMap;
  DATATYPE * neighbor;
  DATATYPE * colorPixel = buffer;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  int maskSize                 = maskInit->getMaskSize();
  int * neighborhood           = maskInit->getNeighborhood();
  unsigned char * chamferMask  = maskInit->getChamferMask();

  // for (k = 0; k < kBegin; k++)
  _initFirstSlicesOut (&pixel, &colorPixel, dims, flag);
  for (k = kBegin; k < dims[2] - kBegin; k++)
    {
      // for (j = 0; j < jBegin; j++)
      _initFirstLinesOut (&pixel, &colorPixel, dims, flag);
      for (j = jBegin; j < dims[1] - jBegin; j++)
	{
	  // for (i = 0; i < iBegin; i++)
	  _initFirstColumnsOut (&pixel, &colorPixel, dims, flag);
	  for (i = iBegin; i < dims[0] - iBegin; i++)
	    {
	      // we consider the background pixels (the others remain unchanged)
	      if (*pixel <= 0 )
		{
		  // if the pixel is not on the border, it gets the infinity value
		  *pixel = - _infinity;
		  
		  for (m = 0; m < maskSize; m++) {
		    // take the closest pixel to the foreground 
		    // in the forward neighborhood
		    neighbor = pixel + neighborhood[m];
		    if ((*neighbor > 0) && ( -((DATATYPE) chamferMask[m]) > *pixel)) {
		      *pixel = - ( (DATATYPE) chamferMask[m] );
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  colorNeighbor = colorPixel + neighborhood[m];
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  p = i + j*dims[0] + k*dims[0]*dims[1];
			  *colorPixel = (DATATYPE) - (p + neighborhood[m]);
			  break;
			default :
			  break;
			}
		    }
		    // and in the backward one
		    neighbor = pixel - neighborhood[m];
		    if ((*neighbor > 0) && ( -((DATATYPE) chamferMask[m]) > *pixel)) {
		      *pixel = - ( (DATATYPE) chamferMask[m] );
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  colorNeighbor = colorPixel - neighborhood[m];
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  p = i + j*dims[0] + k*dims[0]*dims[1];
			  *colorPixel = (DATATYPE) - (p - neighborhood[m]);
			  break;
			default :
			  break;
			}
		    }
		  }
		}
	      pixel++;
	      colorPixel++;
	    }
	  // for (i = dims[0] - iBegin; i < dims[0]; i++)
	  _initLastColumnsOut (&pixel, &colorPixel, dims, flag);
	}
      // for (j = dims[1] - jBegin; j < dims[1]; j++)
      _initLastLinesOut (&pixel, &colorPixel, dims, flag);
    }
  // for (k = dims[2] - kBegin; k < dims[2]; k++)
  _initLastSlicesOut (&pixel, &colorPixel, dims, flag);
  
}

/*Both inside and outside ***************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::initBoundariesInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims, int flag)
{
  int i, j, k, m, p;

  DATATYPE * pixel = distMap;
  DATATYPE * neighbor;
  DATATYPE * colorPixel = buffer;
  DATATYPE * colorNeighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  int maskSize                 = maskInit->getMaskSize();
  int * neighborhood           = maskInit->getNeighborhood();
  unsigned char * chamferMask  = maskInit->getChamferMask();

  // for (k = 0; k < kBegin; k++)
  _initFirstSlicesInOut (&pixel, &colorPixel, dims, flag);
  for (k = kBegin; k < dims[2] - kBegin; k++)
    {
      // for (j = 0; j < jBegin; j++)
      _initFirstLinesInOut (&pixel, &colorPixel, dims, flag);
      for (j = jBegin; j < dims[1] - jBegin; j++)
	{
	  // for (i = 0; i < iBegin; i++)
	  _initFirstColumnsInOut (&pixel, &colorPixel, dims, flag);
	  for (i = iBegin; i < dims[0] - iBegin; i++)
	    {
	      p = i + j*dims[0] + k*dims[0]*dims[1];
	      // foreground
	      if (*pixel > 0 )
		{
		  // if the pixel is not on the border, it gets the infinity value
		  *pixel = _infinity;
		  
		  for (m = 0; m < maskSize; m++) {
		    // take the closest pixel to the background 
		    // in the forward neighborhood
		    neighbor = pixel + neighborhood[m];
		    if ((*neighbor <= 0) && ((DATATYPE) chamferMask[m] < *pixel)) {
		      *pixel = (DATATYPE) chamferMask[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  colorNeighbor = colorPixel + neighborhood[m];
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  *colorPixel = (DATATYPE) (p + neighborhood[m]);
			  break;
			default :
			  break;
			}
		    }		    
		    // and in the backward one
		    neighbor = pixel - neighborhood[m];
		    if ((*neighbor <= 0) && ((DATATYPE) chamferMask[m] < *pixel)) {
		      *pixel = (DATATYPE) chamferMask[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  colorNeighbor = colorPixel - neighborhood[m];
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  *colorPixel = (DATATYPE) (p - neighborhood[m]);
			}
		    }
		  }
		}
	      else // background
		{
		  // if the pixel is not on the border, it gets the infinity value
		  *pixel = - _infinity;
		  
		  for (m = 0; m < maskSize; m++) {
		    // take the closest pixel to the foreground 
		    // in the forward neighborhood
		    neighbor = pixel + neighborhood[m];
		    if ((*neighbor > 0) && ( -((DATATYPE) chamferMask[m]) > *pixel)) {
		      *pixel = - ( (DATATYPE) chamferMask[m] );
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  colorNeighbor = colorPixel + neighborhood[m];
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  *colorPixel = (DATATYPE) - (p + neighborhood[m]);
			  break;
			default :
			  break;
			}
		    }
		    // and in the backward one
		    neighbor = pixel - neighborhood[m];
		    if ((*neighbor > 0) && ( -((DATATYPE) chamferMask[m]) > *pixel)) {
		      *pixel = - ( (DATATYPE) chamferMask[m] );
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  colorNeighbor = colorPixel - neighborhood[m];
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  *colorPixel = (DATATYPE) - (p - neighborhood[m]);
			  break;
			default :
			  break;
			}
		    }
		  }
		}
	      pixel++;
	      colorPixel++;
	    }
	  // for (i = dims[0] - iBegin; i < dims[0]; i++)
	  _initLastColumnsInOut (&pixel, &colorPixel, dims, flag);
	}
      // for (j = dims[1] - jBegin; j < dims[1]; j++)
      _initLastLinesInOut (&pixel, &colorPixel, dims, flag);
    }
  // for (k = dims[2] - kBegin; k < dims[2]; k++)
  _initLastSlicesInOut (&pixel, &colorPixel, dims, flag);
  
}

/****************************************************************************/
/*                             Forward Running                              */
/****************************************************************************/
/*Inside ********************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::forwardRun(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  int i, j, k, m;
  DATATYPE newValue;
  DATATYPE * neighbor;
  DATATYPE * pixel = distMap;
  DATATYPE * colorPixel = buffer;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  
  int maskSize                 = mask->getMaskSize();
  int * neighborhood           = mask->getNeighborhood();
  unsigned char * chamferMask  = mask->getChamferMask();

  currentPos = 0;
  // for (k = 0; k < kBegin; k++)
  _forwardRunFirstSlices (&pixel, &colorPixel, dims);
  for (k = kBegin; k < dims[2]; k++) 
    {
      // for (j = 0; j < jBegin; j++)
      _forwardRunFirstLines (&pixel, &colorPixel, dims);
      for (j = jBegin; j < dims[1] - jBegin; j++)
	{
	  // for (i = 0; i < iBegin; i++)
	  _forwardRunFirstColumns (&pixel, &colorPixel, dims);
	  for (i = iBegin; i < dims[0] - iBegin; i++)
	    {
	      // we consider the object pixels (the others remain to zero)
	      if (*pixel > 0)
		{
		  // take the closest pixel to the background in the forward neighborhood
		  for (m = 0; m < maskSize; m++) {
		    neighbor = pixel - neighborhood[m];
		    if ((DATATYPE) chamferMask[m] < (_infinity - *neighbor))
		      newValue = *neighbor + (DATATYPE) chamferMask[m];
		    else
		      newValue = _infinity;
		    if (newValue < *pixel) {
		      *pixel = newValue;
		      colorNeighbor = colorPixel - neighborhood[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor <= 0)
			    *colorPixel = (DATATYPE)(currentPos - neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }
		  } 
		}
	      pixel++;
	      colorPixel++;
	      currentPos++;
	    }
	  // for (i = dims[0] - iBegin; i < dims[0]; i++)
	  _forwardRunLastColumns (&pixel, &colorPixel, dims);
	}
      // for (j = dims[1] - jBegin; j < dims[1]; j++)
      _forwardRunLastLines (&pixel, &colorPixel, dims);
    }
}

/*Outside *******************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::forwardRunOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  int i, j, k, m;
  DATATYPE newValue;
  DATATYPE * neighbor;
  DATATYPE * pixel = distMap;
  DATATYPE * colorPixel = buffer;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  
  int maskSize                 = mask->getMaskSize();
  int * neighborhood           = mask->getNeighborhood();
  unsigned char * chamferMask  = mask->getChamferMask();


  // for (k = 0; k < kBegin; k++)
  _forwardRunFirstSlicesOut (&pixel, &colorPixel, dims);
  for (k = kBegin; k < dims[2]; k++) 
    {
      // for (j = 0; j < jBegin; j++)
      _forwardRunFirstLinesOut (&pixel, &colorPixel, dims);
      for (j = jBegin; j < dims[1] - jBegin; j++)
	{
	  // for (i = 0; i < iBegin; i++)
	  _forwardRunFirstColumnsOut (&pixel, &colorPixel, dims);
	  for (i = iBegin; i < dims[0] - iBegin; i++)
	    {
	      // we consider the object pixels (the others remain to zero)
	      if (*pixel <= 0)
		{
		  // take the closest pixel to the background in the forward neighborhood
		  for (m = 0; m < maskSize; m++) {
		    neighbor = pixel - neighborhood[m];
		    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) chamferMask[m]) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) chamferMask[m];
		    else
		      newValue = - _infinity;
		    if (newValue > *pixel) {
		      *pixel = newValue;
		      colorNeighbor = colorPixel - neighborhood[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor > 0)
			    *colorPixel = (DATATYPE) - (i+j*dims[0]+k*dims[0]*dims[1] - neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }
		  } 
		}
	      pixel++;
	      colorPixel++;
	    }
	  // for (i = dims[0] - iBegin; i < dims[0]; i++)
	  _forwardRunLastColumnsOut (&pixel, &colorPixel, dims);
	}
      // for (j = dims[1] - jBegin; j < dims[1]; j++)
      _forwardRunLastLinesOut (&pixel, &colorPixel, dims);
    }
}

/*Both inside and outside ***************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::forwardRunInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  int i, j, k, m, p;
  DATATYPE newValue;
  DATATYPE * neighbor;
  DATATYPE * pixel = distMap;
  DATATYPE * colorPixel = buffer;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  
  int maskSize                 = mask->getMaskSize();
  int * neighborhood           = mask->getNeighborhood();
  unsigned char * chamferMask  = mask->getChamferMask();

  // for (k = 0; k < kBegin; k++)
  _forwardRunFirstSlicesInOut (&pixel, &colorPixel, dims);
  for (k = kBegin; k < dims[2]; k++) 
    {
      // for (j = 0; j < jBegin; j++)
      _forwardRunFirstLinesInOut (&pixel, &colorPixel, dims);
      for (j = jBegin; j < dims[1] - jBegin; j++)
	{
	  // for (i = 0; i < iBegin; i++)
	  _forwardRunFirstColumnsInOut (&pixel, &colorPixel, dims);
	  for (i = iBegin; i < dims[0] - iBegin; i++)
	    {
	      p = i + j*dims[0] + k*dims[0]*dims[1];
	      for (m = 0; m < maskSize; m++) {
		neighbor = pixel - neighborhood[m];
		colorNeighbor = colorPixel - neighborhood[m];
		if (*pixel > 0) // object pixels 
		  {
		    if ((DATATYPE) chamferMask[m] < (_infinity - max(*neighbor, (DATATYPE) 0)))
		      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) chamferMask[m];
		    else
		      newValue = _infinity;
		    if (newValue < *pixel) {
		      *pixel = newValue;
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor <= 0)
			    *colorPixel = (DATATYPE) (p - neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }
		  }
		else            // background pixels
		  {
		    if (min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) chamferMask[m]) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) chamferMask[m];
		    else
		      newValue = - _infinity;
		    if (newValue > *pixel) {
		      *pixel = newValue;
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor > 0)
			    *colorPixel = (DATATYPE) - (p - neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }
		  }
	      }
	      pixel++;
	      colorPixel++;
	    }
	  // for (i = dims[0] - iBegin; i < dims[0]; i++)
	  _forwardRunLastColumnsInOut (&pixel, &colorPixel, dims);
	}
      // for (j = dims[1] - jBegin; j < dims[1]; j++)
      _forwardRunLastLinesInOut (&pixel, &colorPixel, dims);
    }
}

/****************************************************************************/
/*                             Backward Running                             */
/****************************************************************************/
/*Inside ********************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::backwardRun(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  int i, j, k, m;
  DATATYPE newValue;
  DATATYPE *  neighbor;
  currentPos = dims[0] * dims[1] * dims[2] - 1;
  DATATYPE * pixel = distMap + currentPos;
  DATATYPE * colorPixel = buffer + currentPos;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  
  int maskSize                 = mask->getMaskSize();
  int * neighborhood           = mask->getNeighborhood();
  unsigned char * chamferMask  = mask->getChamferMask();

  // for (k = dims[2]; k > dims[2] - kBegin; k--)
  _backwardRunLastSlices (&pixel, &colorPixel, dims);
  for (k = dims[2] - kBegin; k > 0; k--) 
    {
      // for (j = dims[1]; j > dims[1] - jBegin; j--)
      _backwardRunLastLines (&pixel, &colorPixel, dims);
      for (j = dims[1] - jBegin; j > jBegin; j--)
	{
	  // for (i = dims[0]; i > dims[0] - iBegin; i--)
	  _backwardRunLastColumns (&pixel, &colorPixel, dims);
	  for (i = dims[0] - iBegin; i > iBegin; i--)
	    {
	      // we consider the object pixels (the others remain to zero)
	      if (*pixel > 0)
		{
		  // take the closest pixel to the background in the backward neighborhood
		  for (m = 0; m < maskSize; m++) {
		    neighbor = pixel + neighborhood[m];
		    if ((DATATYPE) chamferMask[m] < (_infinity - *neighbor))
		      newValue = *neighbor + (DATATYPE) chamferMask[m];
		    else
		      newValue = _infinity;
		    
		    if (newValue < *pixel) {
		      *pixel = newValue;
		      colorNeighbor = colorPixel + neighborhood[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor == 0)
			    *colorPixel = (DATATYPE) (currentPos + neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }
		  } 
		}
	      pixel--;
	      colorPixel--;
	      currentPos--;
	    }
	  // for (i = iBegin ; i > 0; i--)
	  _backwardRunFirstColumns (&pixel, &colorPixel, dims);
	}
      // for (j = jBegin; j > 0; j--)
      _backwardRunFirstLines (&pixel, &colorPixel, dims);
    }
}

/*Outside *******************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::backwardRunOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  int i, j, k, m;
  DATATYPE newValue;
  DATATYPE *  neighbor;
  DATATYPE * pixel = distMap + dims[0] * dims[1] * dims[2] - 1;
  DATATYPE * colorPixel = buffer + dims[0] * dims[1] * dims[2] - 1;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  
  int maskSize                 = mask->getMaskSize();
  int * neighborhood           = mask->getNeighborhood();
  unsigned char * chamferMask  = mask->getChamferMask();

  // for (k = dims[2]; k > dims[2] - kBegin; k--)
  _backwardRunLastSlicesOut (&pixel, &colorPixel, dims);
  for (k = dims[2] - kBegin; k > 0; k--) 
    {
      // for (j = dims[1]; j > dims[1] - jBegin; j--)
      _backwardRunLastLinesOut (&pixel, &colorPixel, dims);
      for (j = dims[1] - jBegin; j > jBegin; j--)
	{
	  // for (i = dims[0]; i > dims[0] - iBegin; i--)
	  _backwardRunLastColumnsOut (&pixel, &colorPixel, dims);
	  for (i = dims[0] - iBegin; i > iBegin; i--)
	    {
	      // we consider the object pixels (the others remain to zero)
	      if (*pixel <= 0)
		{
		  // take the closest pixel to the foreground in the backward neighborhood
		  for (m = 0; m < maskSize; m++) {
		    neighbor = pixel + neighborhood[m];
		    
		    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) chamferMask[m]) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) chamferMask[m];
		    else
		      newValue = - _infinity;
		    
		    if (newValue > *pixel) {
		      *pixel = newValue;
		      colorNeighbor = colorPixel + neighborhood[m];
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor > 0)
			    *colorPixel = (DATATYPE) - (i+j*dims[0]+k*dims[1] + neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }  
		  }
		}
	      pixel--;
	      colorPixel --;
	    }
	  // for (i = iBegin ; i > 0; i--)
	  _backwardRunFirstColumnsOut (&pixel, &colorPixel, dims);
	}
      // for (j = jBegin; j > 0; j--)
      _backwardRunFirstLinesOut (&pixel, &colorPixel, dims);
    }
}

/*Both inside and outside ***************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::backwardRunInOut(DATATYPE * distMap, DATATYPE * buffer, const int * dims)
{
  int i, j, k, m, p;
  DATATYPE newValue;
  DATATYPE *  neighbor;
  DATATYPE * pixel = distMap + dims[0] * dims[1] * dims[2] - 1;
  DATATYPE * colorPixel = buffer + dims[0] * dims[1] * dims[2] - 1;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  
  int maskSize                 = mask->getMaskSize();
  int * neighborhood           = mask->getNeighborhood();
  unsigned char * chamferMask  = mask->getChamferMask();

  // for (k = dims[2]; k > dims[2] - kBegin; k--)
  _backwardRunLastSlicesInOut (&pixel, &colorPixel, dims);
  for (k = dims[2] - kBegin; k > 0; k--) 
    {
      // for (j = dims[1]; j > dims[1] - jBegin; j--)
      _backwardRunLastLinesInOut (&pixel, &colorPixel, dims);
      for (j = dims[1] - jBegin; j > jBegin; j--)
	{
	  // for (i = dims[0]; i > dims[0] - iBegin; i--)
	  _backwardRunLastColumnsInOut (&pixel, &colorPixel, dims);
	  for (i = dims[0] - iBegin; i > iBegin; i--)
	    {
	      p = i + j*dims[0] + k*dims[0]*dims[1];
	      // take the closest pixel to the background in the backward neighborhood
	      for (m = 0; m < maskSize; m++) {
		neighbor = pixel + neighborhood[m];
		colorNeighbor = colorPixel + neighborhood[m];
		if (*pixel > 0)  // foreground
		  {
		    if ((DATATYPE) chamferMask[m] < (_infinity - max(*neighbor, (DATATYPE) 0)))
		      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) chamferMask[m];
		    else
		      newValue = _infinity;
		    
		    if (newValue < *pixel) {
		      *pixel = newValue;
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor <= 0)
			    *colorPixel = (DATATYPE) (p + neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }
		  }
		else              //background
		  {
		    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) chamferMask[m]) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) chamferMask[m];
		    else
		      newValue = - _infinity;
		    
		    if (newValue > *pixel) {
		      *pixel = newValue;
		      switch(optFlag)
			{
			case COLOR_PROPAGATION :
			  *colorPixel = *colorNeighbor;
			  break;
			case POSITION_PROPAGATION :
			  if (*colorNeighbor > 0)
			    *colorPixel = (DATATYPE) - (p + neighborhood[m]);
			  else
			    *colorPixel = *colorNeighbor;
			  break;
			default :
			  break;
			}
		    }
		  }
	      }
	      pixel--;
	      colorPixel--;
	    }
	  // for (i = iBegin ; i > 0; i--)
	  _backwardRunFirstColumnsInOut (&pixel, &colorPixel, dims);
	}
      // for (j = jBegin; j > 0; j--)
      _backwardRunFirstLinesInOut (&pixel, &colorPixel, dims);
    }
}


/****************************************************************************/
/*                                                                          */
/*                             PROTECTED METHODS                            */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/*                             Initialization                               */
/****************************************************************************/
/*Inside ********************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstSlices(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;
 
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();


  // if keepOldVal == true, keep the old value of the pixel (do not set it to infinity)
  bool keepOldVal = (flag == _DistanceMap_FS)  || (flag == _DistanceMap_FSL) || 
                    (flag == _DistanceMap_FSC) || (flag == _DistanceMap_FSLC);
    

  for (k = 0; k < kBegin; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
	      
	      pixel = *distMap;
	      colorPixel = *buffer;
	      if (*pixel > 0)
		{
		  if (! keepOldVal) 
		    *pixel = _infinity;
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if (maskElmt.exist) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
				  *pixel = (DATATYPE) maskElmt.value;
				  colorNeighbor = colorPixel + maskElmt.offset;
				  switch(optFlag)
				    {
				    case COLOR_PROPAGATION :
				      *colorPixel = *colorNeighbor;
				      break;
				    case POSITION_PROPAGATION :
				      *colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
				      break;
				    default :
				      break;
				    }
				}

			      }
			    }
			}
		    }
		  // forward mask
		  for (z = 0; z <= k; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
				  *pixel = (DATATYPE) maskElmt.value;
				  colorNeighbor = colorPixel - maskElmt.offset;
				  switch(optFlag)
				    {
				    case COLOR_PROPAGATION :
				      *colorPixel = *colorNeighbor;
				      break;
				    case POSITION_PROPAGATION :
				      *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
				      break;
				    default :
				      break;
				    }
				}
			      }
			    }
			}
		    }
		}
	      else
		*pixel = 0;
	      
	      (*distMap)++;
	      (*buffer)++;
	      currentPos++;
	    }
	}
    }
 
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastSlices(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt; 
 
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LS)  || (flag == _DistanceMap_LSL) || 
                    (flag == _DistanceMap_LSC) || (flag == _DistanceMap_LSLC);
  
  for (k = dims[2] - kBegin; k < dims[2]; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
		  
	      pixel = *distMap;
	      colorPixel = *buffer;
	      if (*pixel > 0)
		{
		  if (!keepOldVal)
		    *pixel = _infinity;
		  // backward mask
		  for (z = 0; z <= (dims[2] - 1) - k; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
				  colorNeighbor = colorPixel + maskElmt.offset;
				  *pixel = (DATATYPE) maskElmt.value;
				  switch(optFlag)
				    {
				    case COLOR_PROPAGATION :
				      *colorPixel = *colorNeighbor;
				      break;
				    case POSITION_PROPAGATION :
				      *colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
				      break;
				    default :
				      break;
				    }
				}
			      }
			    }
			}
		    }
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
				  colorNeighbor = colorPixel - maskElmt.offset;
				  *pixel = (DATATYPE) maskElmt.value;
				  switch(optFlag)
				    {
				    case COLOR_PROPAGATION :
				      *colorPixel = *colorNeighbor;
				      break;
				    case POSITION_PROPAGATION :
				      *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
				      break;
				    default :
				      break;
				    }
				}
			      }
			    }
			}
		    }
		}
	      else
		*pixel = 0;
	      
	      (*distMap)++;
	      (*buffer)++;
	      currentPos++;
	    }
	}
    }
}


template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstLines(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

 
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FL)  || (flag == _DistanceMap_FSL) ||
                    (flag == _DistanceMap_FLC) || (flag == _DistanceMap_FSLC);

  for (j = 0; j < jBegin; j++)
    {
      yMin = j;
      yMax = jBegin;
      
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }

	  pixel = *distMap;
	  colorPixel = *buffer;
	  if (*pixel > 0)
	    {
	      if (!keepOldVal)
		*pixel = _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
			      *pixel = (DATATYPE) maskElmt.value;
			      colorNeighbor = colorPixel + maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  *colorPixel = (DATATYPE) (DATATYPE) (currentPos + maskElmt.offset);
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if (maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor <= 0) && ( (DATATYPE) maskElmt.value < *pixel) ) {
			      *pixel = (DATATYPE) maskElmt.value;
			      colorNeighbor = colorPixel - maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		}
	    }
	  
	  else
	    *pixel = 0;
	  
	  (*distMap) ++;
	  (*buffer) ++;
	  currentPos++;
	}
    }
}


template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastLines(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;
 
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LL)  || (flag == _DistanceMap_LSL) ||
                    (flag == _DistanceMap_LLC) || (flag == _DistanceMap_LSLC);

  for (j = dims[1] - jBegin; j < dims[1]; j++)
    {
      yMin = jBegin;
      yMax = (dims[1] - 1) - j;
      
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }

	  pixel = *distMap;
	  colorPixel = *buffer;

	  if (*pixel > 0)
	    {
	      if (!keepOldVal)
		*pixel = _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
			      *pixel = (DATATYPE) maskElmt.value;
			      colorNeighbor = colorPixel + maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  *colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
			      *pixel = (DATATYPE) maskElmt.value;
			      colorNeighbor = colorPixel - maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		}
	    }
	  else
	    *pixel = 0;
 
	  (*distMap) ++;
	  (*buffer) ++;
	  currentPos++;
	}
    }
}


template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;
  
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FC)  || (flag == _DistanceMap_FSC) ||
                    (flag == _DistanceMap_FLC) || (flag == _DistanceMap_FSLC);


  for (i = 0; i < iBegin; i++)
    {
      pixel = *distMap;
      colorPixel = *buffer;

      if (*pixel > 0)
	{
	  if (!keepOldVal)
	    *pixel = _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // backward mask
		  for (x = - i; x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
			  *pixel = (DATATYPE) maskElmt.value;
			  colorNeighbor = colorPixel + maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION :
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION : 
			      *colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
			      break;
			    default :
			      break;
			    }
			}
		      }
		    }
		  // forward mask
		  for (x = -iBegin; x <= i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
			  *pixel = (DATATYPE) maskElmt.value;
			  colorNeighbor = colorPixel - maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION : 
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION :
			      *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
			    default :
			      break;
			    }
			}
		      }
		    }
		}
	    }
	}
      else
	*pixel = 0;
      
      (*distMap) ++;
      (*buffer) ++;
      currentPos++;
    }
}


template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LC)  || (flag == _DistanceMap_LSC) ||
                    (flag == _DistanceMap_LLC) || (flag == _DistanceMap_LSLC);


  for (i = dims[0] - iBegin; i < dims[0]; i++)
    {
      pixel = *distMap;
      colorPixel = *buffer;

      if (*pixel > 0)
	{
	  if (!keepOldVal)
	    *pixel = _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // bacward mask
		  for (x = - iBegin; x <= (dims[0] - 1) - i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
			  *pixel = (DATATYPE) maskElmt.value;
			  colorNeighbor = colorPixel + maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION :
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION :
			      *colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
			      break;
			    }
			}
		      }
		    }
		  // forward mask
		  for (x = -((dims[0] - 1) - i); x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) ) {
			  *pixel = (DATATYPE) maskElmt.value;
			  colorNeighbor = colorPixel - maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION :
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION :
			      *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
			      break;
			    default :
			      break;
			    }
			}
		      }
		    }
		}
	    }
	}
      else
	*pixel = 0;
      
      (*distMap) ++;
      (*buffer)++;
      currentPos++;
    }
}

/*Outside *******************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstSlicesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;
 
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FS)  || (flag == _DistanceMap_FSL) || 
                    (flag == _DistanceMap_FSC) || (flag == _DistanceMap_FSLC);


  for (k = 0; k < kBegin; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
	      
	      pixel = *distMap;
	      if (*pixel <= 0)
		{
		  if (!keepOldVal)
		    *pixel = - _infinity;
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if (maskElmt.exist) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor > 0) && 
				     ( - ((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		  // forward mask
		  for (z = 0; z <= k; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor > 0) && 
				     (- ((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		}
	      (*distMap)++;
	      (*buffer) ++;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastSlicesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt; 
 
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();
  
  bool keepOldVal = (flag == _DistanceMap_LS)  || (flag == _DistanceMap_LSL) || 
                    (flag == _DistanceMap_LSC) || (flag == _DistanceMap_LSLC);

  for (k = dims[2] - kBegin; k < dims[2]; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
		  
	      pixel = *distMap;
	      if (*pixel <= 0)
		{
		  if (!keepOldVal)
		    *pixel = - _infinity;
		  // forward mask
		  for (z = 0; z <= (dims[2] - 1) - k; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor > 0) && 
				     (-((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor > 0) && 
				     (- ((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		}
	      (*distMap)++;
	      (*buffer) ++;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstLinesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

 
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FL)  || (flag == _DistanceMap_FSL) ||
                    (flag == _DistanceMap_FLC) || (flag == _DistanceMap_FSLC);

  for (j = 0; j < jBegin; j++)
    {
      yMin = j;
      yMax = jBegin;
      
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }

	  pixel = *distMap;
	  if (*pixel <= 0)
	    {
	      if (!keepOldVal)
		*pixel = - _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = - ((DATATYPE) maskElmt.value);
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if (maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = - ((DATATYPE) maskElmt.value);
			  }
			}
		    }
		}
	    } 
	  (*distMap) ++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastLinesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;
 
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LL)  || (flag == _DistanceMap_LSL) ||
                    (flag == _DistanceMap_LLC) || (flag == _DistanceMap_LSLC);

  for (j = dims[1] - jBegin; j < dims[1]; j++)
    {
      yMin = jBegin;
      yMax = (dims[1] - 1) - j;
      
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }

	  pixel = *distMap;
	  if (*pixel <= 0)
	    {
	      if (!keepOldVal)
		*pixel = - _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = - ((DATATYPE) maskElmt.value);
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = -((DATATYPE) maskElmt.value);
			  }
			}
		    }
		}
	    } 
	  (*distMap) ++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;
  
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FC)  || (flag == _DistanceMap_FSC) ||
                    (flag == _DistanceMap_FLC) || (flag == _DistanceMap_FSLC);

  for (i = 0; i < iBegin; i++)
    {
      pixel = *distMap;
      if (*pixel <= 0)
	{
	  if (!keepOldVal)
	    *pixel = - _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // backward mask
		  for (x = - i; x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		  // forward mask
		  for (x = -iBegin; x <= i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		}
	    }
	}      
      (*distMap) ++;
      (*buffer) ++;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LC)  || (flag == _DistanceMap_LSC) ||
                    (flag == _DistanceMap_LLC) || (flag == _DistanceMap_LSLC);

  for (i = dims[0] - iBegin; i < dims[0]; i++)
    {
      pixel = *distMap;
      if (*pixel <= 0)
	{
	  if (!keepOldVal)
	    *pixel = - _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // bacward mask
		  for (x = - iBegin; x <= (dims[0] - 1) - i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		  // forward mask
		  for (x = -((dims[0] - 1) - i); x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		}
	    }
	}      
      (*distMap) ++;
      (*buffer) ++;
    }
}


/*Both inside and outside ***************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstSlicesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;
 
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FS)  || (flag == _DistanceMap_FSL) || 
                    (flag == _DistanceMap_FSC) || (flag == _DistanceMap_FSLC);

  for (k = 0; k < kBegin; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
	      
	      pixel = *distMap;
	      if (*pixel > 0) // foreground
		{
		  if (!keepOldVal)
		    *pixel = _infinity;
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if (maskElmt.exist) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
				  *pixel = (DATATYPE) maskElmt.value;
			      }
			    }
			}
		    }
		  // forward mask
		  for (z = 0; z <= k; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
				  *pixel = (DATATYPE) maskElmt.value;
			      }
			    }
			}
		    }
		}
	      else // background
		{
		  if (!keepOldVal)
		    *pixel = - _infinity;
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if (maskElmt.exist) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor > 0) && 
				     ( - ((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		  // forward mask
		  for (z = 0; z <= k; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor > 0) && 
				     (- ((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		}


	      (*distMap)++;
	      (*buffer) ++;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastSlicesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt; 
 
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LS)  || (flag == _DistanceMap_LSL) || 
                    (flag == _DistanceMap_LSC) || (flag == _DistanceMap_LSLC);
    
  for (k = dims[2] - kBegin; k < dims[2]; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
		  
	      pixel = *distMap;
	      if (*pixel > 0) // foreground
		{
		  if (!keepOldVal)
		    *pixel = _infinity;
		  // forward mask
		  for (z = 0; z <= (dims[2] - 1) - k; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
				  *pixel = (DATATYPE) maskElmt.value;
			      }
			    }
			}
		    }
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
				  *pixel = (DATATYPE) maskElmt.value;
			      }
			    }
			}
		    }
		}
	      else      // background
		{
		  if (!keepOldVal)
		    *pixel = - _infinity;
		  // forward mask
		  for (z = 0; z <= (dims[2] - 1) - k; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel + maskElmt.offset;
				if ( (*neighbor > 0) && 
				     (-((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		  // backward mask
		  for (z = 0; z <= kBegin; z++)
		    {
		      for (y = - yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x++)
			    {
			      maskInit->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( (*neighbor > 0) && 
				     (- ((DATATYPE) maskElmt.value) > *pixel) )
				  *pixel = - ((DATATYPE) maskElmt.value);
			      }
			    }
			}
		    }
		}
	      (*distMap)++;
	      (*buffer) ++;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstLinesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

 
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FL)  || (flag == _DistanceMap_FSL) ||
                    (flag == _DistanceMap_FLC) || (flag == _DistanceMap_FSLC);

  for (j = 0; j < jBegin; j++)
    {
      yMin = j;
      yMax = jBegin;
      
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }

	  pixel = *distMap;
	  if (*pixel > 0) // foreground
	    {
	      if (!keepOldVal)
		*pixel = _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
			      *pixel = (DATATYPE) maskElmt.value;
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if (maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor <= 0) && ( (DATATYPE) maskElmt.value < *pixel) )
			      *pixel = (DATATYPE) maskElmt.value;
			  }
			}
		    }
		}
	    }
	  else // background
	    {
	      if (!keepOldVal)
		*pixel = - _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = - ((DATATYPE) maskElmt.value);
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if (maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = - ((DATATYPE) maskElmt.value);
			  }
			}
		    }
		}
	    } 	    
	  (*distMap) ++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastLinesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;
 
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LL)  || (flag == _DistanceMap_LSL) ||
                    (flag == _DistanceMap_LLC) || (flag == _DistanceMap_LSLC);

  for (j = dims[1] - jBegin; j < dims[1]; j++)
    {
      yMin = jBegin;
      yMax = (dims[1] - 1) - j;
      
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }

	  pixel = *distMap;
	  if (*pixel > 0)    // foreground
	    {
	      if (!keepOldVal)
		*pixel = _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
			      *pixel = (DATATYPE) maskElmt.value;
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
			      *pixel = (DATATYPE) maskElmt.value;
			  }
			}
		    }
		}
	    }
	  else // background
	    {
	      if (!keepOldVal)
		*pixel = - _infinity;
	      for (z = 0; z <= kBegin; z++)
		{
		  // backward mask
		  for (y = - yMin; y <= yMax; y ++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = - ((DATATYPE) maskElmt.value);
			  }
			}
		    }
		  // forward mask
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  maskInit->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( (*neighbor > 0) && 
				 (- ((DATATYPE) maskElmt.value) > *pixel) )
			      *pixel = -((DATATYPE) maskElmt.value);
			  }
			}
		    }
		}
	    } 
	  (*distMap) ++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initFirstColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;
  
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_FC)  || (flag == _DistanceMap_FSC) ||
                    (flag == _DistanceMap_FLC) || (flag == _DistanceMap_FSLC);

  for (i = 0; i < iBegin; i++)
    {
      pixel = *distMap;
      if (*pixel > 0)  // foreground
	{
	  if (!keepOldVal)
	    *pixel = _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // backward mask
		  for (x = - i; x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
			  *pixel = (DATATYPE) maskElmt.value;
		      }
		    }
		  // forward mask
		  for (x = -iBegin; x <= i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
			  *pixel = (DATATYPE) maskElmt.value;
		      }
		    }
		}
	    }
	}
      else // background
	{
	  if (!keepOldVal)
	    *pixel = - _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // backward mask
		  for (x = - i; x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		  // forward mask
		  for (x = -iBegin; x <= i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		}
	    }
	}      
      (*distMap) ++;
      (*buffer) ++;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_initLastColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims, int flag)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = maskInit->getIBegin();
  int jBegin = maskInit->getJBegin();
  int kBegin = maskInit->getKBegin();

  bool keepOldVal = (flag == _DistanceMap_LC)  || (flag == _DistanceMap_LSC) ||
                    (flag == _DistanceMap_LLC) || (flag == _DistanceMap_LSLC);

  for (i = dims[0] - iBegin; i < dims[0]; i++)
    {
      pixel = *distMap;
      if (*pixel > 0) // foreground
	{
	  if (!keepOldVal)
	    *pixel = _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // bacward mask
		  for (x = - iBegin; x <= (dims[0] - 1) - i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
			  *pixel = (DATATYPE) maskElmt.value;
		      }
		    }
		  // forward mask
		  for (x = -((dims[0] - 1) - i); x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor <= 0) && ((DATATYPE) maskElmt.value < *pixel) )
			  *pixel = (DATATYPE) maskElmt.value;
		      }
		    }
		}
	    }
	}
      else   // background
	{
	  if (!keepOldVal)
	    *pixel = - _infinity;
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y ++)
		{
		  // bacward mask
		  for (x = - iBegin; x <= (dims[0] - 1) - i; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel + maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		  // forward mask
		  for (x = -((dims[0] - 1) - i); x <= iBegin; x++)
		    {
		      maskInit->getPrivatMask(x, y, z, &maskElmt);
		      if ( maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( (*neighbor > 0) && 
			     (- ((DATATYPE) maskElmt.value) > *pixel) )
			  *pixel = - ((DATATYPE) maskElmt.value);
		      }
		    }
		}
	    }
	}      
      (*distMap) ++;
      (*buffer) ++;
    }
}

/****************************************************************************/
/*                             Forward Running                              */
/****************************************************************************/
/*Inside ********************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstSlices(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (k = 0; k < kBegin; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
	      
	      pixel = *distMap;
	      colorPixel = *buffer;

	      if (*pixel > 0)
		{
		  for (z = 0; z <= k; z++)
		    {
		      for (y = -yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x ++)
			    {
			      mask->getPrivatMask(x, y, z, &maskElmt);
			      if (maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
				  newValue = *neighbor + (DATATYPE) maskElmt.value;
				else 
				  newValue = _infinity;
				if (newValue < *pixel) {
				  *pixel = newValue;
				  colorNeighbor = colorPixel - maskElmt.offset;
				  switch(optFlag)
				    {
				    case COLOR_PROPAGATION :
				      *colorPixel = *colorNeighbor;
				      break;
				    case POSITION_PROPAGATION :
				      if (*colorNeighbor <= 0)
					*colorPixel = (DATATYPE)(currentPos - maskElmt.offset);
				      else
					*colorPixel = *colorNeighbor;
				      break;
				    default :
				      break;
				    }
				}
			      }
			    }
			}
		    }
		}
	      (*distMap)++;
	      (*buffer) ++;
	      currentPos++;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstLines(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (j = 0; j < jBegin; j++)
    {
      yMin = j;
      yMax = jBegin;

      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }
	  
	  pixel = *distMap;
	  colorPixel = *buffer;

	  if (*pixel > 0)
	    {
	      for (z = 0; z <= kBegin; z++)
		{
		  for (y = -yMax; y <= yMin; y++)
		    {
		      for (x = -xMax; x <= xMin; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if (maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			      newValue = *neighbor + (DATATYPE) maskElmt.value;
			    else
			      newValue = _infinity;
			    if (newValue < *pixel) {
			      *pixel = newValue;
			      colorNeighbor = colorPixel - maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  if (*colorNeighbor <= 0)
				    *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
				  else
					*colorPixel = *colorNeighbor;
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		}
	    }
	  (*distMap)++;
	  (*buffer) ++;
	  currentPos++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunLastLines(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;  
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  

  for (j = dims[1] - jBegin; j < dims[1]; j++)
    {
      yMin = jBegin;
      yMax = (dims[1] - 1) - j;
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }
	  
	  pixel = *distMap;
	  colorPixel = *buffer;

	  if (*pixel > 0)
	    {
	      for (z = 0; z <= kBegin; z++)
		{
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {		      
			    neighbor = pixel - maskElmt.offset;
			    if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			      newValue = *neighbor + (DATATYPE) maskElmt.value;
			    else
			      newValue = _infinity;
			    if (newValue < *pixel) {
			      *pixel = newValue;
			      colorNeighbor = colorPixel - maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  if (*colorNeighbor <= 0)
				    *colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
				  else
				    *colorPixel = *colorNeighbor;
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		}
	    }
	  (*distMap)++;
	  (*buffer) ++;
	  currentPos++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (i = 0; i < iBegin; i++)
    {
      pixel = *distMap;
      colorPixel = *buffer;
      if (*pixel > 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = -jBegin; y <= jBegin; y++)
		{
		  for (x = -iBegin; x <= i; x ++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			  newValue = *neighbor + (DATATYPE) maskElmt.value;
			else
			  newValue =_infinity;
			if (newValue < *pixel) {
			  *pixel = newValue;
			  colorNeighbor = colorPixel - maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION :
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION :
			      if (*colorNeighbor <= 0)
				*colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
			      else
				*colorPixel = *colorNeighbor;
			      break;
			    default :
			      break;
			    }
			}
		      }
		    }
		}
	    }
	}
      (*distMap)++;
      (*buffer) ++;
      currentPos++;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunLastColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;
  
  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (i = dims[0] - iBegin; i < dims[0]; i++)
    {
      pixel = *distMap;
      colorPixel = *buffer;

      if (*pixel > 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y++)
		{
		  for (x = - ((dims[0] - 1) - i); x <= iBegin; x++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			  newValue = *neighbor + (DATATYPE) maskElmt.value;
			else
			  newValue = _infinity;
			if (newValue < *pixel) {
			  *pixel = newValue;
			  colorNeighbor = colorPixel - maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION :
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION :
			      if (*colorNeighbor <= 0)
				*colorPixel = (DATATYPE) (currentPos - maskElmt.offset);
			      else
				*colorPixel = *colorNeighbor;
			      break;
			    default :
			      break;
			    }
			}
		      }
		    }
		}
	    }
	}
      (*distMap)++;
      (*buffer) ++;
      currentPos++;
    }
}

/*Outside *******************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstSlicesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (k = 0; k < kBegin; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
	      
	      pixel = *distMap;
	      if (*pixel <= 0)
		{
		  for (z = 0; z <= k; z++)
		    {
		      for (y = -yMax; y <= yMin; y++)
			{
			  for (x = - xMax; x <= xMin; x ++)
			    {
			      mask->getPrivatMask(x, y, z, &maskElmt);
			      if (maskElmt.exist ) {
				neighbor = pixel - maskElmt.offset;
				if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
				  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
				else 
				  newValue = - _infinity;
				if (newValue > *pixel)
				  *pixel = newValue;
			      }
			    }
			}
		    }
		}
	      (*distMap)++;
	      (*buffer) ++;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstLinesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (j = 0; j < jBegin; j++)
    {
      yMin = j;
      yMax = jBegin;

      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }
	  
	  pixel = *distMap;
	  if (*pixel <= 0)
	    {
	      for (z = 0; z <= kBegin; z++)
		{
		  for (y = -yMax; y <= yMin; y++)
		    {
		      for (x = -xMax; x <= xMin; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if (maskElmt.exist ) {
			    neighbor = pixel - maskElmt.offset;
			    if ( *neighbor > (- _infinity + (DATATYPE) maskElmt.value) )
			      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			    else
			      newValue = - _infinity;
			    if (newValue > *pixel)
			      *pixel = newValue;
			  }
			}
		    }
		}
	    }
	  (*distMap)++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunLastLinesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;  
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  

  for (j = dims[1] - jBegin; j < dims[1]; j++)
    {
      yMin = jBegin;
      yMax = (dims[1] - 1) - j;
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }
	  
	  pixel = *distMap;
	  if (*pixel <= 0)
	    {
	      for (z = 0; z <= kBegin; z++)
		{
		  for (y = - yMax; y <= yMin; y++)
		    {
		      for (x = - xMax; x <= xMin; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {		      
			    neighbor = pixel - maskElmt.offset;
			    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			    else
			      newValue = - _infinity;
			    if (newValue > *pixel)
			      *pixel = newValue;
			  }
			}
		    }
		}
	    }
	  (*distMap)++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (i = 0; i < iBegin; i++)
    {
      pixel = *distMap;
      if (*pixel <= 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = -jBegin; y <= jBegin; y++)
		{
		  for (x = -iBegin; x <= i; x ++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = - _infinity;
			if (newValue > *pixel)
			  *pixel = newValue;
		      }
		    }
		}
	    }
	}
      (*distMap)++;
      (*buffer)++;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunLastColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  
  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (i = dims[0] - iBegin; i < dims[0]; i++)
    {
      pixel = *distMap;
      if (*pixel <= 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = - jBegin; y <= jBegin; y++)
		{
		  for (x = - ((dims[0] - 1) - i); x <= iBegin; x++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist ) {
			neighbor = pixel - maskElmt.offset;
			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = - _infinity;
			if (newValue > *pixel)
			  *pixel = newValue;
		      }
		    }
		}
	    }
	}
      (*distMap)++;
      (*buffer) ++;
    }
}


/*Both inside and outside ***************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstSlicesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (k = 0; k < kBegin; k++)
    {
      for (j = 0; j < dims[1]; j++)
	{
	  if (j < jBegin) {                // first lines
	    yMin = j;
	    yMax = jBegin;
	  }
	  else if (j < dims[1] - jBegin) {    // other lines
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else  {                          // last lines
	    yMin = jBegin;
	    yMax = (dims[1] - 1) - j;
	  }
	  
	  for (i = 0; i < dims[0]; i++)
	    {
	      if (i < iBegin) {            // first columns
		xMin = i;
		xMax = iBegin;
	      }
	      else if (i < dims[0] - iBegin){ //rest of the line
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {                       // last columns
		xMin = iBegin;
		xMax = (dims[0] - 1) - i;
	      }
	      
	      pixel = *distMap;

	      for (z = 0; z <= k; z++) {
		for (y = -yMax; y <= yMin; y++) {
		  for (x = - xMax; x <= xMin; x ++) {
		    mask->getPrivatMask(x, y, z, &maskElmt);
		    if (maskElmt.exist ) 
		      {				  
			neighbor = pixel - maskElmt.offset;
			if (*pixel > 0) // foreground
			  {
			    if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
			      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
			    else 
			      newValue = _infinity;
			    if (newValue < *pixel)
			      *pixel = newValue;
			  }
			else // background
			  {
			    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			    else 
			      newValue = - _infinity;
			    if (newValue > *pixel)
			      *pixel = newValue;
			  }
		      }
		  }
		}
	      }
	      
	      (*distMap)++;
	      (*buffer) ++;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstLinesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (j = 0; j < jBegin; j++)
    {
      yMin = j;
      yMax = jBegin;

      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }
	  
	  pixel = *distMap;
	  
	  for (z = 0; z <= kBegin; z++) {
	    for (y = -yMax; y <= yMin; y++) {
	      for (x = -xMax; x <= xMin; x++) {
		mask->getPrivatMask(x, y, z, &maskElmt);
		if (maskElmt.exist ) 
		  {
		    neighbor = pixel - maskElmt.offset;
		    if (*pixel > 0) // foreground
		      {
			if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
			  newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
			else
			  newValue = _infinity;
			if (newValue < *pixel)
			  *pixel = newValue;
		      }
		    else            // background
		      {
			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = - _infinity;
			if (newValue > *pixel)
			  *pixel = newValue;
		      }
		  }
	      }
	    }
	  }
	  (*distMap)++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunLastLinesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;  
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();
  

  for (j = dims[1] - jBegin; j < dims[1]; j++)
    {
      yMin = jBegin;
      yMax = (dims[1] - 1) - j;
      for (i = 0; i < dims[0]; i++)
	{
	  if (i < iBegin) {            // first columns
	    xMin = i;
	    xMax = iBegin;
	  }
	  else if (i < dims[0] - iBegin){ //rest of the line
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {                       // last columns
	    xMin = iBegin;
	    xMax = (dims[0] - 1) - i;
	  }
	  
	  pixel = *distMap;
	
	  for (z = 0; z <= kBegin; z++) {
	    for (y = - yMax; y <= yMin; y++) {
	      for (x = - xMax; x <= xMin; x++) {
		mask->getPrivatMask(x, y, z, &maskElmt);
		if ( maskElmt.exist ) 
		  {		      
		    neighbor = pixel - maskElmt.offset;

		    if (*pixel > 0) // foreground
		      {
			if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
			  newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
			else
			  newValue = _infinity;
			if (newValue < *pixel)
			  *pixel = newValue;
		      }
		    else            // background
		      {
			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = - _infinity;
			if (newValue > *pixel)
			  *pixel = newValue;
		      }
		  }
	      }
	    }
	  }
	  (*distMap)++;
	  (*buffer) ++;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunFirstColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (i = 0; i < iBegin; i++)
    {
      pixel = *distMap;

      for (z = 0; z <= kBegin; z++) {
	for (y = -jBegin; y <= jBegin; y++) {
	  for (x = -iBegin; x <= i; x ++) {
	    mask->getPrivatMask(x, y, z, &maskElmt);
	    if (maskElmt.exist ) 
	      {
		neighbor = pixel - maskElmt.offset;

		if (*pixel >0) // foreground
		  {
		    if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
		      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
		    else
		      newValue =_infinity;
		    if (newValue < *pixel)
		      *pixel = newValue;
		  }
		else           // background
		  {
		    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
		    else
		      newValue = - _infinity;
		    if (newValue > *pixel)
		      *pixel = newValue;
		  }
	      }
	  }
	}
      }
      (*distMap)++;
      (*buffer) ++;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_forwardRunLastColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  
  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (i = dims[0] - iBegin; i < dims[0]; i++)
    {
      pixel = *distMap;
      for (z = 0; z <= kBegin; z++) {
	for (y = - jBegin; y <= jBegin; y++) {
	  for (x = - ((dims[0] - 1) - i); x <= iBegin; x++) {
	    mask->getPrivatMask(x, y, z, &maskElmt);
	    if (maskElmt.exist ) 
	      {
		neighbor = pixel - maskElmt.offset;
		
		if (*pixel > 0) // foreground
		  {
		    if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
		      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
		    else
		      newValue = _infinity;
		    
		    if (newValue < *pixel)
		      *pixel = newValue;
		  }
		else            // background
		  {
		    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
		    else
		      newValue = - _infinity;
		    if (newValue > *pixel)
		      *pixel = newValue;
		  }
	      }
	  }
	}
      }
      (*distMap)++;
      (*buffer) ++;
    }
}


/****************************************************************************/
/*                             Backward Running                             */
/****************************************************************************/
/*Inside ********************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastSlices(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (k = dims[2]; k > dims[2] - kBegin; k--)
    {
      for (j = dims[1]; j > 0; j --)
	{
	  if (j > dims[1] - jBegin) {
	    yMin = jBegin;
	    yMax = dims[1] - j;
	  }
	  else if (j > jBegin) {
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else {
	    yMin = j - 1;
	    yMax = jBegin;
	  }
	  
	  for (i = dims[0]; i > 0; i--)
	    {
	      if (i > dims[0] - iBegin) {
		xMin = iBegin;
		xMax = dims[0] - i;
	      }
	      else if (i > iBegin) {
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {
		xMin = i - 1;
		xMax = iBegin;
	      }
	      
	      pixel = *distMap;
	      colorPixel = *buffer;

	      if (*pixel > 0)
		{
		  for (z  = 0; z <= dims[2] - k; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      mask->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel + maskElmt.offset;
				if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
				  newValue = *neighbor + (DATATYPE) maskElmt.value;
				else
				  newValue =_infinity;
				
				if (newValue < *pixel) {
				  *pixel = newValue;
				  colorNeighbor = colorPixel + maskElmt.offset;
				  switch(optFlag)
				    {
				    case COLOR_PROPAGATION :
				      *colorPixel = *colorNeighbor;
				      break;
				    case POSITION_PROPAGATION :
				      if (*colorNeighbor == 0)
					*colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
				      else
					*colorPixel = *colorNeighbor;
				      break;
				    default :
				      break;
				    }
				}
			      }
			    }
			}
		    }
		}
	      (*distMap)--;
	      (*buffer) --;
	      currentPos--;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastLines(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (j = dims[1]; j > dims[1] - jBegin; j--)
    {
      yMin = jBegin;
      yMax = dims[1] - j;
      for (i = dims[0]; i > 0; i--)
	{
	  if (i > dims[0] - iBegin) {
	    xMin = iBegin;
	    xMax = dims[0] - i;
	  }
	  else if (i > iBegin) {
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {
	    xMin = i - 1;
	    xMax = iBegin;
	  }
	  
	  pixel = *distMap;
	  colorPixel = *buffer;

	  if (*pixel > 0)
	    {
	      for (z  = 0; z <= kBegin; z++)
		{
		  for (y = -yMin; y <= yMax; y++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			      newValue = *neighbor + (DATATYPE) maskElmt.value;
			    else
			      newValue = _infinity;
			    
			    if (newValue < *pixel) {
			      *pixel = newValue;
			      colorNeighbor = colorPixel + maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  if (*colorNeighbor == 0)
				    *colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
				  else
				    *colorPixel = *colorNeighbor;
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		}
	    }
	  (*distMap)--;
	  (*buffer)--;
	  currentPos--;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunFirstLines(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (j = jBegin; j > 0; j--)
    {
      yMin = j - 1;
      yMax = jBegin;

      for (i = dims[0]; i > 0; i--)
	{
	  if (i > dims[0] - iBegin) {
	    xMin = iBegin;
	    xMax = dims[0] - i;
	  }
	  else if (i > iBegin) {
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {
	    xMin = i - 1;
	    xMax = iBegin;
	  }
	  
	  pixel = *distMap;
	  colorPixel = *buffer;

	  if (*pixel > 0)
	    {
	      for (z = 0; z <= kBegin; z++)
		{
		  for (y = - yMin; y <= yMax; y++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			      newValue = *neighbor + (DATATYPE) maskElmt.value;
			    else
			      newValue =_infinity;
			    
			    if (newValue < *pixel) {
			      *pixel = newValue;
			      colorNeighbor = colorPixel + maskElmt.offset;
			      switch(optFlag)
				{
				case COLOR_PROPAGATION :
				  *colorPixel = *colorNeighbor;
				  break;
				case POSITION_PROPAGATION :
				  if (*colorNeighbor == 0)
				    *colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
				  else
				    *colorPixel = *colorNeighbor;
				  break;
				default :
				  break;
				}
			    }
			  }
			}
		    }
		}
	    } 
	  (*distMap)--;
	  (*buffer) --;
	  currentPos--;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

   /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

 for (i =  dims[0]; i > dims[0] - iBegin; i--)
    {
      pixel = *distMap;
      colorPixel = *buffer;

      if (*pixel > 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = -jBegin; y <= jBegin; y++)
		{
		  for (x = -iBegin; x <= dims[0] - i; x++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist) {
			neighbor = pixel + maskElmt.offset;
			if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			  newValue = *neighbor + (DATATYPE) maskElmt.value;
			else
			  newValue = _infinity;
			if (newValue < *pixel) {
			  *pixel = newValue;
			  colorNeighbor = colorPixel + maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION :
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION :
			      if (*colorNeighbor == 0)
				*colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
			      else
				*colorPixel = *colorNeighbor;
			      break;
			    default :
			      break;
			    }
			}
		      }
		    }
		}
	    }
	}
      (*distMap)--;
      (*buffer)--;
      currentPos--;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunFirstColumns(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;
  DATATYPE * colorPixel;
  DATATYPE * colorNeighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (i = iBegin; i > 0; i--)
    {
      pixel = *distMap;
      colorPixel = *buffer;

      if (*pixel > 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = -jBegin; y <= jBegin; y++)
		{
		  for (x = -(i-1); x <= iBegin; x++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist) {
			neighbor = pixel + maskElmt.offset;

			if ((DATATYPE) maskElmt.value < (_infinity - *neighbor))
			  newValue = *neighbor + (DATATYPE) maskElmt.value;
			else
			  newValue = _infinity;

			if (newValue < *pixel) {
			  *pixel = newValue;
			  colorNeighbor = colorPixel + maskElmt.offset;
			  switch(optFlag)
			    {
			    case COLOR_PROPAGATION :
			      *colorPixel = *colorNeighbor;
			      break;
			    case POSITION_PROPAGATION :
			      if (*colorNeighbor == 0)
				*colorPixel = (DATATYPE) (currentPos + maskElmt.offset);
			      else
				*colorPixel = *colorNeighbor;
			      break;
			    default :
			      break;
			    }
			}
		      }
		    }
		}
	    }
	}
      (*distMap)--;
      (*buffer)--;
      currentPos--;
    }
}

/*Outside *******************************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastSlicesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (k = dims[2]; k > dims[2] - kBegin; k--)
    {
      for (j = dims[1]; j > 0; j --)
	{
	  if (j > dims[1] - jBegin) {
	    yMin = jBegin;
	    yMax = dims[1] - j;
	  }
	  else if (j > jBegin) {
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else {
	    yMin = j - 1;
	    yMax = jBegin;
	  }
	  
	  for (i = dims[0]; i > 0; i--)
	    {
	      if (i > dims[0] - iBegin) {
		xMin = iBegin;
		xMax = dims[0] - i;
	      }
	      else if (i > iBegin) {
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {
		xMin = i - 1;
		xMax = iBegin;
	      }
	      
	      pixel = *distMap;
	      if (*pixel <= 0)
		{
		  for (z  = 0; z <= dims[2] - k; z++)
		    {
		      for (y = -yMin; y <= yMax; y++)
			{
			  for (x = - xMin; x <= xMax; x++)
			    {
			      mask->getPrivatMask(x, y, z, &maskElmt);
			      if ( maskElmt.exist ) {
				neighbor = pixel + maskElmt.offset;
				if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
				  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
				else
				  newValue = - _infinity;
				
				if (newValue > *pixel)
				  *pixel = newValue;
			      }
			    }
			}
		    }
		}
	      (*distMap)--;
	      (*buffer) --;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastLinesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (j = dims[1]; j > dims[1] - jBegin; j--)
    {
      yMin = jBegin;
      yMax = dims[1] - j;
      for (i = dims[0]; i > 0; i--)
	{
	  if (i > dims[0] - iBegin) {
	    xMin = iBegin;
	    xMax = dims[0] - i;
	  }
	  else if (i > iBegin) {
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {
	    xMin = i - 1;
	    xMax = iBegin;
	  }
	  
	  pixel = *distMap;
	  if (*pixel <= 0)
	    {
	      for (z  = 0; z <= kBegin; z++)
		{
		  for (y = -yMin; y <= yMax; y++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			    else
			      newValue = - _infinity;
			    
			    if (newValue > *pixel)
			      *pixel = newValue;
			  }
			}
		    }
		}
	    }
	  (*distMap)--;
	  (*buffer) --;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunFirstLinesOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (j = jBegin; j > 0; j--)
    {
      yMin = j - 1;
      yMax = jBegin;

      for (i = dims[0]; i > 0; i--)
	{
	  if (i > dims[0] - iBegin) {
	    xMin = iBegin;
	    xMax = dims[0] - i;
	  }
	  else if (i > iBegin) {
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {
	    xMin = i - 1;
	    xMax = iBegin;
	  }
	  
	  pixel = *distMap;
	  if (*pixel <= 0)
	    {
	      for (z = 0; z <= kBegin; z++)
		{
		  for (y = - yMin; y <= yMax; y++)
		    {
		      for (x = - xMin; x <= xMax; x++)
			{
			  mask->getPrivatMask(x, y, z, &maskElmt);
			  if ( maskElmt.exist ) {
			    neighbor = pixel + maskElmt.offset;
			    
			    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			    else
			      newValue = -_infinity;
			    
			    if (newValue > *pixel)
			      *pixel = newValue;    
			  }
			}
		    }
		}
	    } 
	  (*distMap)--;
	  (*buffer) --;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

   /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

 for (i =  dims[0]; i > dims[0] - iBegin; i--)
    {
      pixel = *distMap;
      if (*pixel <= 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = -jBegin; y <= jBegin; y++)
		{
		  for (x = -iBegin; x <= dims[0] - i; x++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist) {
			neighbor = pixel + maskElmt.offset;
			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = - _infinity;
			if (newValue > *pixel)
			  *pixel = newValue;
		      }
		    }
		}
	    }
	}
      (*distMap)--;
      (*buffer) --;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunFirstColumnsOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (i = iBegin; i > 0; i--)
    {
      pixel = *distMap;
      if (*pixel <= 0)
	{
	  for (z = 0; z <= kBegin; z++)
	    {
	      for (y = -jBegin; y <= jBegin; y++)
		{
		  for (x = -(i-1); x <= iBegin; x++)
		    {
		      mask->getPrivatMask(x, y, z, &maskElmt);
		      if (maskElmt.exist) {
			neighbor = pixel + maskElmt.offset;

			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = - _infinity;

			if (newValue > *pixel)
			  *pixel = newValue;
		      }
		    }
		}
	    }
	}
      (*distMap)--;
      (*buffer) --;
    }
}


/*Both inside and outside ***************************************************/
template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastSlicesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j, k;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (k = dims[2]; k > dims[2] - kBegin; k--)
    {
      for (j = dims[1]; j > 0; j --)
	{
	  if (j > dims[1] - jBegin) {
	    yMin = jBegin;
	    yMax = dims[1] - j;
	  }
	  else if (j > jBegin) {
	    yMin = jBegin;
	    yMax = jBegin;
	  }
	  else {
	    yMin = j - 1;
	    yMax = jBegin;
	  }
	  
	  for (i = dims[0]; i > 0; i--)
	    {
	      if (i > dims[0] - iBegin) {
		xMin = iBegin;
		xMax = dims[0] - i;
	      }
	      else if (i > iBegin) {
		xMin = iBegin;
		xMax = iBegin;
	      }
	      else {
		xMin = i - 1;
		xMax = iBegin;
	      }
	      
	      pixel = *distMap;
	      
	      for (z  = 0; z <= dims[2] - k; z++) {
		for (y = -yMin; y <= yMax; y++) {
		  for (x = - xMin; x <= xMax; x++) {
		    mask->getPrivatMask(x, y, z, &maskElmt);
		    if ( maskElmt.exist ) 
		      {
			neighbor = pixel + maskElmt.offset;
			
			if (*pixel > 0) // foreground 
			  {
			    if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
			      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
			    else
			      newValue =_infinity;
			    
			    if (newValue < *pixel)
			      *pixel = newValue;
			  }
			else            // background
			  {
			    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			    else
			      newValue = - _infinity;
			    
			    if (newValue > *pixel)
			      *pixel = newValue;
			  }
		      }
		  }
		}
	      }
	      (*distMap)--;
	      (*buffer) --;
	    }
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastLinesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (j = dims[1]; j > dims[1] - jBegin; j--)
    {
      yMin = jBegin;
      yMax = dims[1] - j;
      for (i = dims[0]; i > 0; i--)
	{
	  if (i > dims[0] - iBegin) {
	    xMin = iBegin;
	    xMax = dims[0] - i;
	  }
	  else if (i > iBegin) {
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {
	    xMin = i - 1;
	    xMax = iBegin;
	  }
	  
	  pixel = *distMap;
	  for (z  = 0; z <= kBegin; z++) {
	    for (y = -yMin; y <= yMax; y++) {
	      for (x = - xMin; x <= xMax; x++) {
		mask->getPrivatMask(x, y, z, &maskElmt);
		if ( maskElmt.exist ) 
		  {
		    neighbor = pixel + maskElmt.offset;
		    if (*pixel > 0) // foreground
		      {
			if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
			  newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
			else
			  newValue = _infinity;
			
			if (newValue < *pixel)
			  *pixel = newValue;
		      }
		    else
		      {
			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = - _infinity;
			
			if (newValue > *pixel)
			  *pixel = newValue;
		      }
		  }
	      }
	    }
	  }
	  (*distMap)--;
	  (*buffer) --;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunFirstLinesInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i, j;
  int x, y, z;
  int xMin, xMax, yMin, yMax;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();


  for (j = jBegin; j > 0; j--)
    {
      yMin = j - 1;
      yMax = jBegin;

      for (i = dims[0]; i > 0; i--)
	{
	  if (i > dims[0] - iBegin) {
	    xMin = iBegin;
	    xMax = dims[0] - i;
	  }
	  else if (i > iBegin) {
	    xMin = iBegin;
	    xMax = iBegin;
	  }
	  else {
	    xMin = i - 1;
	    xMax = iBegin;
	  }
	  
	  pixel = *distMap;
	  
	  for (z = 0; z <= kBegin; z++) {
	    for (y = - yMin; y <= yMax; y++) {
	      for (x = - xMin; x <= xMax; x++) {
		mask->getPrivatMask(x, y, z, &maskElmt);
		if ( maskElmt.exist ) 
		  {
		    neighbor = pixel + maskElmt.offset;
		    if (*pixel > 0)  // foreground
		      {
			if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
			  newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
			else
			  newValue =_infinity;
			
			if (newValue < *pixel)
			  *pixel = newValue;    
		      }
		    else
		      {
			if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
			  newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
			else
			  newValue = -_infinity;
			
			if (newValue > *pixel)
			  *pixel = newValue;    
		      }
		  }
	      }
	    }
	  } 
	  (*distMap)--;
	  (*buffer)--;
	}
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunLastColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;

   /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

 for (i =  dims[0]; i > dims[0] - iBegin; i--)
    {
      pixel = *distMap;

      for (z = 0; z <= kBegin; z++) {
	for (y = -jBegin; y <= jBegin; y++) {
	  for (x = -iBegin; x <= dims[0] - i; x++) {
	    mask->getPrivatMask(x, y, z, &maskElmt);
	    if (maskElmt.exist) 
	      {
		neighbor = pixel + maskElmt.offset;
		if (*pixel > 0)  // foreground
		  {
		    if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
		      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
		    else
		      newValue = _infinity;
		    if (newValue < *pixel)
		      *pixel = newValue;
		  }
		else              // background
		  {
		    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
		    else
		      newValue = - _infinity;
		    if (newValue > *pixel)
		      *pixel = newValue;
		  }
	      }
	  }
	}
      }
      (*distMap)--;
      (*buffer) --;
    }
}

template <class DATATYPE>
void
ChamferMap<DATATYPE>::_backwardRunFirstColumnsInOut(DATATYPE ** distMap, DATATYPE ** buffer, const int * dims)
{
  int i;
  int x, y, z;
  chamferElmt maskElmt;

  DATATYPE newValue;
  DATATYPE * pixel;
  DATATYPE * neighbor;


  /* mask fields */
  int iBegin = mask->getIBegin();
  int jBegin = mask->getJBegin();
  int kBegin = mask->getKBegin();

  for (i = iBegin; i > 0; i--)
    {
      pixel = *distMap;
	  
      for (z = 0; z <= kBegin; z++) {
	for (y = -jBegin; y <= jBegin; y++) {
	  for (x = -(i-1); x <= iBegin; x++) {
	    mask->getPrivatMask(x, y, z, &maskElmt);
	    if (maskElmt.exist) 
	      {
		neighbor = pixel + maskElmt.offset;
		
		if (*pixel > 0)  // foreground
 		  {
		    if ((DATATYPE) maskElmt.value < (_infinity - max(*neighbor, (DATATYPE) 0)))
		      newValue = max(*neighbor, (DATATYPE) 0) + (DATATYPE) maskElmt.value;
		    else
		      newValue = _infinity;
		    
		    if (newValue < *pixel)
		      *pixel = newValue;
		  }
		else             // background
		  {
		    if ( min(*neighbor, (DATATYPE) 0) > (- _infinity + (DATATYPE) maskElmt.value) )
		      newValue = min(*neighbor, (DATATYPE) 0) - (DATATYPE) maskElmt.value;
		    else
		      newValue = - _infinity;
		    
		    if (newValue > *pixel)
		      *pixel = newValue;
		  }
	      }
	  }
	}
      }
      (*distMap)--;
      (*buffer)--;
    }
}



#ifdef _YAV_
} // namespace yav
#endif
