#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma implementation "inrimage/Inrimage.h"
#include <inrimage/Inrimage.h>
#include <inrimage/Interpolator.h>
#include <inrimage/ImageGeometry.h>
#include <stdlib.h>

#ifndef WIN32
#include <sys/time.h>
#else
#include <time.h>
#endif

namespace yav {

Inrimage::UnknownType::UnknownType(const std::string& from)
  : Exception(from, "Unhandled inrimage type") {
}

Inrimage::InvalidArguments::InvalidArguments(const std::string& from,
					     const std::string& error)
  : Exception(from, error) {
}

Inrimage::InvalidCompressType::InvalidCompressType(const std::string& from)
  : Exception(from,"attempt to write image in an invalid compress type") {
}
   
Inrimage::MissingCompressInfo::MissingCompressInfo(const std::string& from)
  : Exception(from,"missing compression information in header") {
}
  
  /** set GL buffer bounds (for non-floating scalar inrimages)
      @param min min GL value
      @param max max GL value
      @param whole image is concerned if z < 0, only indicated plane otherwise */
  template <class B, class V>
    void setBufferBounds(B ,Inrimage &inr, V *min, V *max, int z){
    unsigned int i, start, s;
  double a;
  V *vmin, *vmax;
  B *ptr, minb, maxb;

  if(z < 0) {
    start = 0;
    s = inr.getX() * inr.getY() * inr.getZ();
  }
  else {
    start = inr.getX() * inr.getY() * z;
    s = inr.getX() * inr.getY();
  }

  vmin = new V(minTypeValue<B>(0));
  vmax = new V(maxTypeValue<B>(0));
  minb = min->getValue();
  maxb = max->getValue();

  a = (vmax->getValueAsDouble() - vmin->getValueAsDouble()) /
    (double) (maxb - minb);

  for(i = 0, ptr = ((B *) inr.getData()) + start; i < s; ++i, ++ptr) {
    if(*ptr < minb) *ptr = vmin->getValue();
    else if(*ptr > maxb) *ptr = vmax->getValue();
    else *ptr = (B) ((*ptr - minb) * a + vmin->getValue() + .5);
  }

  delete vmin;
  delete vmax;
}
  /** build GL buffer bounds (for non-floating scalar inrimages)
	@param b dummy parameter 
	@param buffer where to store the histogram
	@param bSize the buffer size 
	@param minVal  the minimum  value of the type B
	@param maxVal  the maximum  value of the type B */
    template <class B>
      void buildHistogram(B ,const Inrimage *inr,unsigned long int  * &buffer,unsigned int &bSize, double &minVal,
			  double &maxVal) 
      {
	
	B minV=minTypeValue<B>(0);
	minVal= (double) minV;
	B maxV=maxTypeValue<B>(0);
	maxVal= (double) maxV;
	bSize=maxV-minV+1;
	buffer= new unsigned long int[bSize];
	memset(buffer, 0, bSize * sizeof(unsigned long int)); 
	
	B *buf = (B *) inr->getData();
	B *end = buf + inr->getX() * inr->getY()* inr->getZ() * inr->getVdim();
	for(; buf < end;) {
	  buffer[*buf - minV]++;
	  buf++;
	}
      }



  /** set GL buffer bounds (for floating scalar inrimages)
      @param min min GL value
      @param max max GL value
      @param whole image is concerned if z < 0, only indicated plane otherwise */
 template <class B, class V>
    void setFloatBufferBounds(B a,Inrimage &inr, V *min, V *max, int z) {
  unsigned int i, start, s;
  B *ptr, minb, maxb;

  if(z < 0) {
    start = 0;
    s = inr.getX() * inr.getY() * inr.getZ();
  }
  else {
    start = inr.getX() * inr.getY()  * z;
    s = inr.getX() * inr.getY() ;
  }

  minb = min->getValue();
  maxb = max->getValue();

  for(i = 0, ptr = ((B *) inr.getData()) + start; i < s; ++i, ++ptr) {
    if(*ptr < minb) *ptr = minb;
    else if(*ptr > maxb) *ptr = maxb;
  }
}

  /** slit one vectorial buffer in different scalar buffers
      @param inrs returned scalar inrimages */
  template <class T>
    void splitVectorsBuffer(T a,const Inrimage &image,Inrimage **inrs) {
  unsigned int i;
  unsigned long size;

  size = image.getX() * image.getY() * image.getZ();
  if(image.getVectorialMode() == VM_INTERLACED) {
    T *from, *end;
    T **tos;
    tos = new T*[image.getVdim()];
    for(i = 0; i < image.getVdim(); i++) tos[i] = (T *) inrs[i]->getData();

    from = (T *) image.getData();
    end = from + image.getVdim() * size;
    for(; from < end;)
      for(i = 0; i < image.getVdim(); i++) *tos[i]++ = *from++;

    delete[] tos;
  }

  else {
    for(i = 0; i < image.getVdim(); i++) {
      memcpy((char *) inrs[i]->getData(), (const char *) image.getData() + i * image.getWdim() * size,
	     size * image.getWdim());
    }
  }
}

  /** set one vectorial dimension from one image
      @param inr scalar inrimage
      @param z vectorial dimension to assign from inr */

  template <class FROM, class TO>
    void setComponentFrom(FROM , TO ,Inrimage &image, const Inrimage *inr, unsigned int z){
	  unsigned long size;
  FROM *from, *end;
  TO *to;

  size =  image.getX() * image.getY() * image.getZ();
  from = (FROM *) inr->getData();
  end = from + size;
  if(image.getVectorialMode() == VM_INTERLACED) {
    for(to = (TO *) image.getData() + z; from < end; from++, to += image.getVdim())
      *to = (TO) *from;
  }
  else {
    for(to = (TO *) image.getData() + z * size; from < end; from++, to++)
      *to = (TO) *from;
  }
}

  /** computes GL buffer bounds for vectorial inrimages (GL value is vector norm)
      @param min returned min GL value
      @param max returned max GL value */
  template <class B, class V>
  void computeVectorialBufferBounds(B ,const Inrimage &inr, V *vmin, V *vmax) {
    unsigned int i, size;
    
    size = inr.getX() * inr.getY() * inr.getZ();
    
    B *buf, *end;
    double f;
    buf = (B *) inr.getData();
    end = buf + size * inr.getVdim();
    B mn = maxTypeValue<B>(0);
    B mx = minTypeValue<B>(0);
    for(; buf < end;) {
      f = 0.0;
      for(i = 0; i < inr.getVdim(); i++, buf++)
	f += *buf * *buf;
      f = sqrt(f);
      
      if(f < mn) mn = (B) f;
      if(f > mx) mx = (B) f;
    }
    vmin->setValue(mn);
    vmax->setValue(mx);
  }


  /** computes GL buffer bounds for scalar inrimages
      @param min returned min GL value
      @param max returned max GL value */
  
  template <class B, class V>
  void computeBufferBounds(B,const Inrimage &inr,V *min, V *max) {
    unsigned int i, size;
    
    size = inr.getX() * inr.getY() * inr.getZ() * inr.getVdim();
    
    B *buf = (B *) inr.getData();
    B mn = maxTypeValue<B>(0);
    B mx = minTypeValue<B>(0);
    for(i = 0; i < size; i++, buf++) {
      if(*buf < mn) mn = *buf;
      if(*buf > mx) mx = *buf;
    }
    min->setValue(mn);
    max->setValue(mx);
  }
  

bool Inrimage::inside(const int x, const int y, const int z) {
  if(x >= 0 && (unsigned int) x < xdim &&
     y >= 0 && (unsigned int) y < ydim &&
     z >= 0 && (unsigned int) z < zdim)
    return true;
  else
    return false;}


#define RESCALEBUFFER(TYPE1,TYPE2,A,B,C,D,E,F) rescaleBuffer((TYPE1)0,(TYPE2)0,A,B,C,D,E,F)

Inrimage::Inrimage(const TclString& _name, bool b,
		   VECTORIAL_MODE vm, VECTORIAL_MODE fvm)
  throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	Inrimage::InvalidHeader, Inrimage::InvalidData,
	Inrimage::UnknownType, InvalidVoxelSize)
    : BaseInrimage(_name == EMPTY_STRING ? 0 : (char *) _name, b, vm, fvm), Pointed<Inrimage>() , 
      name(_name),type(WT_UNSIGNED_CHAR),  geom(CARTESIAN), interpolator(0),
      extrapolator(new ToZeroExtrapolator()) {
    char constructor1[] = "Inrimage::Inrimage(const TclString&, bool, VECTORIAL_MODE)";
  array.threeD = 0;
  for(int i=0;i<6;++i) region[i]=-1;
  
  char tmp[100];
  if(userField("RESTORE_TYPE",tmp,100)){
     removeUserField("RESTORE_TYPE");
     double min=0;
     double max=0;
     BaseInrimage img(*this);
     int wd=getWdim();
     setType(WORD_TYPE(atoi(tmp)),getVdim(),img.getVectorialMode());
     if(!userField("RESTORE_MIN",tmp,100)) throw MissingCompressInfo(constructor1);
     removeUserField("RESTORE_MIN");
     min=atof(tmp);
     if(!userField("RESTORE_MAX",tmp,100)) throw MissingCompressInfo(constructor1);
     removeUserField("RESTORE_MAX");
     max=atof(tmp);

     switch(wd){
	  case 1:
	   switch(type){
	      case WT_SIGNED_SHORT:
		 img.RESCALEBUFFER(unsigned char,short,0,255,min,max,0,*this);
		 break;
	      case WT_UNSIGNED_SHORT:
		 img.RESCALEBUFFER(unsigned char,unsigned short,0,255,min,max,0,*this);
		 break;
	      case WT_SIGNED_INT:
		 img.RESCALEBUFFER(unsigned char,int,0,255,min,max,0,*this);
		 break;
	      case WT_UNSIGNED_INT:
		 img.RESCALEBUFFER(unsigned char,unsigned int,0,255,min,max,0,*this);
		 break;
	      case WT_SIGNED_LONG:
		 img.RESCALEBUFFER(unsigned char,LONGINT,0,255,min,max,0,*this);
		 break;
	      case WT_UNSIGNED_LONG:
		 img.RESCALEBUFFER(unsigned char,unsigned LONGINT,0,255,min,max,0,*this);
		 break;
	      case WT_FLOAT:
	      case WT_FLOAT_VECTOR:
		 img.RESCALEBUFFER(unsigned char,float,0,255,min,max,0,*this);
		 break;
	      case WT_DOUBLE:
	      case WT_DOUBLE_VECTOR:
		 img.RESCALEBUFFER(unsigned char,double,0,255,min,max,0,*this);
		 break;
	      default:
		 break;
	   }
	   break;
	case 2:
	   switch(type){
	      case WT_SIGNED_SHORT:
		 img.RESCALEBUFFER(unsigned short,short,0,65535,min,max,0,*this);
		 break;
	      case WT_UNSIGNED_SHORT:
		 img.RESCALEBUFFER(unsigned short,unsigned short,0,65535,min,max,0,*this);
		 break;
	      case WT_SIGNED_INT:
		 img.RESCALEBUFFER(unsigned short,int,0,65535,min,max,0,*this);
		 break;
	      case WT_UNSIGNED_INT:
		 img.RESCALEBUFFER(unsigned short,unsigned int,0,65535,min,max,0,*this);
		 break;
	      case WT_SIGNED_LONG:
		 img.RESCALEBUFFER(unsigned short,LONGINT,0,65535,min,max,0,*this);
		 break;
	      case WT_UNSIGNED_LONG:
		 img.RESCALEBUFFER(unsigned short,unsigned LONGINT,0,65535,min,max,0,*this);
		 break;
	      case WT_FLOAT:
	      case WT_FLOAT_VECTOR:
		 img.RESCALEBUFFER(unsigned short,float,0,65535,min,max,0,*this);
		 break;
	      case WT_DOUBLE:
	      case WT_DOUBLE_VECTOR:
		 img.RESCALEBUFFER(unsigned short,double,0,65535,min,max,0,*this);
		 break;
	      default:
		 break;
	   }
	   break;
	default:
	   throw InvalidCompressType(constructor1);
	   break;
     }
  }
  
  try {
    setType();
    queryGeometry();
    setInterpolator(Interpolator::trilinearInterpolatorByType(geom, type,
							      vdim));
  }  
  catch(UnknownType& e) {
    e.push(constructor1);
    throw e;
  }

  try {
    buildIndex();
  }
  catch(NotEnoughMemory& e) {
    e.push(constructor1);
    throw e;
  }
}



Inrimage::Inrimage(unsigned int xdim, unsigned int ydim, unsigned int zdim,
		   unsigned int vdim, unsigned int wdim, WORD_KIND wordKind,
		   SIGN sign, float vx, float vy, float vz)
  throw(InvalidSize, UnknownType, NotEnoughMemory, InvalidVoxelSize)
    : BaseInrimage(xdim, ydim, zdim, vdim, wdim, wordKind, sign, vx, vy, vz), Pointed<Inrimage>(), name("noname.inr.gz"),type(WT_UNSIGNED_CHAR), geom(CARTESIAN),
      interpolator(0), extrapolator(new ToZeroExtrapolator()) {
    char constructor2[] = "Inrimage::Inrimage(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, WORD_KIND, SIGN)";

  array.threeD = 0;
  for(int i=0;i<6;++i) region[i]=-1;
  
  try {
    setType();
    queryGeometry();
    setInterpolator(Interpolator::trilinearInterpolatorByType(geom, type,
							      vdim));
  }
  catch(UnknownType& e) {
    e.push(constructor2);
    throw e;
  }

  try {
    buildIndex();
  }
  catch(NotEnoughMemory& e) {
    e.push(constructor2);
    throw e;
  }
}



Inrimage::Inrimage(unsigned int xdim, unsigned int ydim, unsigned int zdim,
		   void *data, unsigned int vdim, unsigned int wdim,
		   WORD_KIND wordKind, SIGN sign, float vx, float vy, float vz)
  throw(Inrimage::InvalidSize, Inrimage::UnknownType, InvalidVoxelSize) 
    : BaseInrimage(xdim, ydim, zdim, data, vdim, wdim, wordKind, sign, vx, vy, vz), Pointed<Inrimage>(), name("noname.inr.gz"),type(WT_UNSIGNED_CHAR), geom(CARTESIAN),
      interpolator(0), extrapolator(new ToZeroExtrapolator()) {

  char constructor3[] = "Inrimage::Inrimage(unsigned int, unsigned int, unsigned int, void *, unsigned int, unsigned int, WORD_KIND, SIGN)";

  array.threeD = 0; 
  for(int i=0;i<6;++i) region[i]=-1;
  
  try {
    setType();
    queryGeometry();
    setInterpolator(Interpolator::trilinearInterpolatorByType(geom, type,
							      vdim));
  }
  catch(UnknownType& e) {
    e.push(constructor3);
    throw e;
  }

  if(data) {
    try {
      buildIndex();
    }
    catch(NotEnoughMemory& e) {
      e.push(constructor3);
      throw e;
    }
  }
}



Inrimage::Inrimage(unsigned int xdim, unsigned int ydim, unsigned int zdim,
		   WORD_TYPE wt, unsigned int _vdim, VECTORIAL_MODE vm,
		   float vx, float vy, float vz)
  throw(InvalidSize, UnknownType, NotEnoughMemory, InvalidVoxelSize) 
    : BaseInrimage(xdim, ydim, zdim, (void *) 0, 1, 1, WK_FIXED, SGN_SIGNED, vx, vy, vz), Pointed<Inrimage>(),
	  name("noname.inr.gz"),type(wt), geom(CARTESIAN), interpolator(0),
      extrapolator(new ToZeroExtrapolator()) {

  char constructor4[] = "Inrimage::Inrimage(unsigned int, unsigned int, unsigned int, WORD_TYPE, unsigned int)";

  array.threeD = 0; 
  for(int i=0;i<6;++i) region[i]=-1;
  
  if(_vdim == 1) {
    if(type == WT_FLOAT_VECTOR) type = WT_FLOAT;
    if(type == WT_DOUBLE_VECTOR) type = WT_DOUBLE;
  }

  setType(type, _vdim, vm);
  setInterpolator(Interpolator::trilinearInterpolatorByType(geom, type, vdim));
  
  if(type ==  WT_FLOAT_VECTOR || type == WT_DOUBLE_VECTOR) {
    if(vm != VM_SCALAR) vectMode = vm;
    else throw Inrimage::UnknownType(constructor4);
  }

  if(vdim == 0) throw Inrimage::UnknownType(constructor4);
  
  try {
    allocate();
    buildIndex();
  }
  catch(NotEnoughMemory& e) {
    e.push(constructor4);
    throw e;
  }  
}


// Creates a vectorial inrimage from scalar inrimages
Inrimage::Inrimage(unsigned int n, Inrimage **inrs, WORD_TYPE wt,
		   VECTORIAL_MODE vm, float vx, float vy, float vz)
  throw(InvalidSize, InvalidArguments, NotEnoughMemory, InvalidVoxelSize)
    : BaseInrimage(1, 1, 1, (void *) 0, 1, 1, WK_FLOAT, SGN_SIGNED, vx, vy, vz), Pointed<Inrimage>(), name("noname.inr.gz"), geom(CARTESIAN),interpolator(0),
      extrapolator(new ToZeroExtrapolator()) {

  char constructor5[] = "Inrimage::Inrimage(unsigned int, Inrimage **, WORD_TYPE, VECTORIAL_MODE)";

  array.threeD = 0;
  for(int i=0;i<6;++i) region[i]=-1;
  
  if(n == 0) throw InvalidArguments(constructor5);
  if(wt == WT_FLOAT_VECTOR) {
    type = WT_FLOAT_VECTOR;
    wdim = 4;
  }
  else if(wt == WT_DOUBLE_VECTOR) {
    type = WT_DOUBLE_VECTOR;
    wdim = 8;
  }
  else throw InvalidArguments(constructor5);

  vdim = n;
  if(vm == VM_SCALAR) 
    throw Inrimage::UnknownType(constructor5);
  else
    vectMode=vm;

  xdim = inrs[0]->xdim;
  ydim = inrs[0]->ydim;
  zdim = inrs[0]->zdim;


  try {
    allocate();
  }
  catch(NotEnoughMemory& e) {
    e.push(constructor5);
    throw e;
  }

  try {
    setComponentsFrom(inrs);
  }
  catch(Exception& e) {
    e.push(constructor5);
    throw e;
  }

}


/// Creates an replica of img.
Inrimage::Inrimage(const Inrimage& img)
   throw(InvalidSize, InvalidVectorialMode, InvalidVoxelSize, NotEnoughMemory)
    : BaseInrimage(img), Pointed<Inrimage>(),
   name(img.name.baseName()+"-copy.inr.gz"), interpolator(img.interpolator->clone()),
      extrapolator(img.extrapolator->clone()) {

  char constructor6[] = "Inrimage::Inrimage(const Inrimage&)";

  array.threeD = 0; 
  for(int i=0;i<6;++i) region[i]=-1;
  
  geom = img.getGeometry();
  setType(img.getType(), img.getVdim(), img.getVectorialMode());
  try {
    buildIndex() ;
  }
  catch(NotEnoughMemory& e) {
    e.push(constructor6);
    throw e;
  }
}


Inrimage::~Inrimage() {
  deleteArray();
  delete interpolator;
  delete extrapolator;
}


void Inrimage::setRegion(int x1,int y1,int z1,int x2,int y2,int z2){
   region[0]=x1*getVX();
   region[1]=y1*getVY();
   region[2]=z1*getVZ();
   region[3]=x2*getVX();
   region[4]=y2*getVY();
   region[5]=z2*getVZ();
}
   
void Inrimage::getRegion(int& x1,int& y1,int& z1,int& x2,int& y2,int& z2){
   x1=region[0]<0?-1:int(region[0]/getVX()+0.5);
   y1=region[1]<0?-1:int(region[1]/getVY()+0.5);
   z1=region[2]<0?-1:int(region[2]/getVZ()+0.5);
   x2=region[3]<0?-1:int(region[3]/getVX()+0.5);
   y2=region[4]<0?-1:int(region[4]/getVY()+0.5);
   z2=region[5]<0?-1:int(region[5]/getVZ()+0.5);
}
   
void Inrimage::copyRegion(Inrimage const& img){
   region[0]=img.region[0];
   region[1]=img.region[1];
   region[2]=img.region[2];
   region[3]=img.region[3];
   region[4]=img.region[4];
   region[5]=img.region[5];
}
   
void Inrimage::getWorkRegion(unsigned int& x1,unsigned int& y1,unsigned int& z1,unsigned int& x2,unsigned int& y2,unsigned int& z2) const{
   x1=region[0]<0?0:(unsigned int)(region[0]/getVX()+0.5);
   y1=region[1]<0?0:(unsigned int)(region[1]/getVY()+0.5);
   z1=region[2]<0?0:(unsigned int)(region[2]/getVZ()+0.5);
   x2=region[3]<0||region[3]>getX()*getVX()?getX():(unsigned int)(region[3]/getVX()+0.5);
   y2=region[4]<0||region[4]>getY()*getVY()?getY():(unsigned int)(region[4]/getVY()+0.5);
   z2=region[5]<0||region[5]>getZ()*getVZ()?getZ():(unsigned int)(region[5]/getVZ()+0.5);
}

bool Inrimage::regionIsWholeImage() const{
   return region[0]<=0 && region[1]<=0 && region[2]<=0 &&
      (region[3]<0 || region[3]>getX()*getVX()) &&
      (region[4]<0 || region[4]>getY()*getVY()) &&
      (region[5]<0 || region[5]>getZ()*getVZ());
}
   
/// Make *this an replica of img.
Inrimage& Inrimage::operator=(const Inrimage& img)
   throw(InvalidSize, InvalidVectorialMode, InvalidVoxelSize,
	 NotEnoughMemory) {

  char from[] = "Inrimage::operator=(const Inrimage&)";
  for(int i=0;i<6;++i) region[i]=img.region[i];
  
  if( &img != this ) {
    try {
      BaseInrimage::operator=(img) ;
    }
    catch(BaseInrimage::InvalidSize &e) {
      e.push(from);
      throw e;
    }
    catch(BaseInrimage::InvalidVectorialMode &e) {
      e.push(from);
      throw e;
    }
    catch(BaseInrimage::InvalidVoxelSize &e) {
      e.push(from);
      throw e;
    }
    catch(BaseInrimage::NotEnoughMemory &e) {
      e.push(from);
      throw e;
    }
    
    setType(img.getType(), img.getVdim(), img.getVectorialMode());
    setInterpolator(Interpolator::trilinearInterpolatorByType(geom,
							      type,
							      vdim));
    geom = img.getGeometry();
    try {
      buildIndex() ;
    }
    catch(NotEnoughMemory &e) {
      e.push(from);
      throw e;
    }
  }

  return *this ;
}   


void Inrimage::write(const char *_name)
  throw(Inrimage::UnableToCreateFile,
	Inrimage::UnableToWriteHeader, Inrimage::UnableToWriteData) {
  char from[] = "Inrimage::write(const char *&)";

  try {
    if(!userField("GEOMETRY", 0, 0)) {
      switch(geom) {
      case CYLINDRICAL:
	addUserField("GEOMETRY", "CYLINDRICAL");
	break;
      case SPHERICAL:
	addUserField("GEOMETRY", "SPHERICAL");
	break;
      default:
	addUserField("GEOMETRY", "CARTESIAN");
	break;
      }
    }
    name=_name;
    BaseInrimage::write((char *) _name);
  }
  catch(BaseInrimage::UnableToCreateFile &e) {
    e.push(from);
    throw e;
  }
  catch(BaseInrimage::UnableToWriteHeader &e) {
    e.push(from);
    throw e;
  }
  catch(BaseInrimage::UnableToWriteData &e) {
    e.push(from);
    throw e;
  }
}


void Inrimage::write(const char *_name, CompressType ct)
   throw(BaseInrimage::UnableToCreateFile,
	 BaseInrimage::UnableToWriteHeader, BaseInrimage::UnableToWriteData,
	 InvalidCompressType) {
  char from[] = "void Inrimage::write(const char *name,CompressType ct)";

  try {
     BaseInrimage img(1,1,1);
     Voxel *imin, *imax;
     TclString tmp;

     img.setVoxelSize(getVX(),getVY(),getVZ());
     imin = Voxel::voxelByType(type);
     imax = Voxel::voxelByType(type);
     getBounds(imin, imax, false);
     
     if(!img.userField("GEOMETRY", 0, 0)) {
       switch(geom) {
       case CYLINDRICAL:
	 img.addUserField("GEOMETRY", "CYLINDRICAL");
	 break;
       case SPHERICAL:
	 img.addUserField("GEOMETRY", "SPHERICAL");
	 break;
       default:
	 img.addUserField("GEOMETRY", "CARTESIAN");
	 break;
       }
     }
     imin->print(tmp);
     img.addUserField("RESTORE_MIN",(char*)tmp);
     imax->print(tmp);
     img.addUserField("RESTORE_MAX",(char*)tmp);
     tmp.sprintf("%d",type);
     img.addUserField("RESTORE_TYPE",tmp);
     switch(ct){
	case CT_8BITS:
	   img.resize(getX(),getY(),getZ(),getVdim(),1,WK_FIXED,SGN_UNSIGNED);
	   img.setVectorialMode(getVectorialMode());
	   switch(type) {
	      case WT_SIGNED_SHORT:
		 RESCALEBUFFER(short,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      case WT_UNSIGNED_SHORT:
		 RESCALEBUFFER(unsigned short,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      case WT_SIGNED_INT:
		 RESCALEBUFFER(int,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      case WT_UNSIGNED_INT:
		 RESCALEBUFFER(unsigned int,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      case WT_SIGNED_LONG:
		 RESCALEBUFFER(LONGINT,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      case WT_UNSIGNED_LONG:
		 RESCALEBUFFER(unsigned LONGINT,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      case WT_FLOAT:
	      case WT_FLOAT_VECTOR:
		 RESCALEBUFFER(float,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      case WT_DOUBLE:
	      case WT_DOUBLE_VECTOR:
		 RESCALEBUFFER(double,unsigned char,imin->getValueAsDouble(),imax->getValueAsDouble(),0,255,0.5,img);
		 break;
	      default:
		 break;
	   }
	   break;
	case CT_16BITS:
	   img.resize(getX(),getY(),getZ(),getVdim(),2,WK_FIXED,SGN_UNSIGNED);
	   img.setVectorialMode(getVectorialMode());
	   switch(type) {
	      case WT_SIGNED_INT:
		 RESCALEBUFFER(int,unsigned short,imin->getValueAsDouble(),imax->getValueAsDouble(),0,65535,0.5,img);
		 break;
	      case WT_UNSIGNED_INT:
		 RESCALEBUFFER(unsigned int,unsigned short,imin->getValueAsDouble(),imax->getValueAsDouble(),0,65535,0.5,img);
		 break;
	      case WT_SIGNED_LONG:
		 RESCALEBUFFER(LONGINT,unsigned short,imin->getValueAsDouble(),imax->getValueAsDouble(),0,65535,0.5,img);
		 break;
	      case WT_UNSIGNED_LONG:
		 RESCALEBUFFER(unsigned LONGINT,unsigned short,imin->getValueAsDouble(),imax->getValueAsDouble(),0,65535,0.5,img);
		 break;
	      case WT_FLOAT:
	      case WT_FLOAT_VECTOR:
		 RESCALEBUFFER(float,unsigned short,imin->getValueAsDouble(),imax->getValueAsDouble(),0,65535,0.5,img);
		 break;
	      case WT_DOUBLE:
	      case WT_DOUBLE_VECTOR:
		 RESCALEBUFFER(double,unsigned short,imin->getValueAsDouble(),imax->getValueAsDouble(),0,65535,0.5,img);
		 break;
	      default:
		 break;
	   }
	   break;
	default:
	   throw InvalidCompressType("void BaseInrimage::write(const char *name,CompressType ct) const");
	   break;
     }
     img.write(_name);
     delete imin;
     delete imax;
  }
  catch(BaseInrimage::UnableToCreateFile &e) {
     e.push(from);
     throw e;
  }
  catch(BaseInrimage::UnableToWriteHeader &e) {
     e.push(from);
     throw e;
  }
  catch(BaseInrimage::UnableToWriteData &e) {
     e.push(from);
     throw e;
  }
}


void Inrimage::resize(unsigned int _xdim,
		      unsigned int _ydim,
		      unsigned int _zdim,
		      unsigned int _vdim,
		      unsigned int _wdim,
		      WORD_KIND _wordKind,
		      SIGN _sign)
  throw(Inrimage::InvalidSize, Inrimage::NotEnoughMemory) {

  char from[] = "Inrimage::resize";

  try {
    BaseInrimage::resize(_xdim, _ydim, _zdim, _vdim, _wdim, _wordKind, _sign);
  }
  catch(BaseInrimage::InvalidSize &e) {
    e.push(from);
    throw e;
  }
  catch(BaseInrimage::NotEnoughMemory &e) {
    e.push(from);
    throw e;
  }
  try {
    setType();
  }
  catch(UnknownType e) {
    throw InvalidSize("Inrimage::resize");
  }
}
  


void Inrimage::setWords(WORD_KIND _wordKind, SIGN _sign)
  throw(Inrimage::NotEnoughMemory) {
   
  char from[] = "Inrimage::setWords";

  try {
    BaseInrimage::setWords(_wordKind, _sign);
  }
  catch(BaseInrimage::NotEnoughMemory &e) {
    e.push(from);
    throw e;
  }
  setType();
}
  


// Changes image vectorial mode. A vectorial image buffer can be
// stored either interlaced (default) or non interlaced.
void Inrimage::setVectorialMode(VECTORIAL_MODE vm)
  throw(Inrimage::InvalidVectorialMode)
{
  char from[] = "Inrimage::setVectorialMode";

  if(vm!=vectMode) {
    deleteArray();
    try {
      BaseInrimage::setVectorialMode(vm);
    }
    catch(BaseInrimage::InvalidVectorialMode &e) {
      e.push(from);
      throw e;
    }
      if(vectMode == VM_NON_INTERLACED) index4D();
      else index3D();
  }

}



void Inrimage::setType(void) throw(Inrimage::UnknownType) {
  char from[] = "setType(void)";

  /* vectorial type */
  if(vdim > 1) {
    if(wdim == 1) {
      if ( sign == SGN_UNSIGNED && vdim == 3 ) type = WT_RGB;
      else if ( sign == SGN_UNSIGNED && vdim == 4 ) type = WT_RGBA;
      else throw Inrimage::UnknownType(from);
    }
    else if(wordKind == WK_FLOAT) {
      if(wdim == 4) type = WT_FLOAT_VECTOR;
      else if(wdim == 8) type = WT_DOUBLE_VECTOR;
      else throw Inrimage::UnknownType(from);
    }
    else throw Inrimage::UnknownType(from);
  }

  /* scalar type */
  else {
    if(wordKind == WK_FIXED) {
      if(wdim == 1)  {
	if(sign == SGN_SIGNED) type = WT_SIGNED_CHAR;
	else if(sign == SGN_UNSIGNED) type = WT_UNSIGNED_CHAR;
	else throw Inrimage::UnknownType(from);
      }
      else if(wdim == 2) {
	if(sign == SGN_SIGNED) type = WT_SIGNED_SHORT;
	else if(sign == SGN_UNSIGNED) type = WT_UNSIGNED_SHORT;
	else throw Inrimage::UnknownType(from);
      }
      else if(wdim == 4) {
	if(sign == SGN_SIGNED) type = WT_SIGNED_INT;
	else if(sign == SGN_UNSIGNED) type = WT_UNSIGNED_INT;
	else throw Inrimage::UnknownType(from);
      }
      else if(wdim == 8) {
	if(sign == SGN_SIGNED) type = WT_SIGNED_LONG;
	else if(sign == SGN_UNSIGNED) type = WT_UNSIGNED_LONG;
	else throw Inrimage::UnknownType(from);
      }
      else throw Inrimage::UnknownType(from);
    }
    else if(wordKind == WK_FLOAT) {
      if(wdim == 4) type = WT_FLOAT;
      else if(wdim == 8) type = WT_DOUBLE;
      else throw Inrimage::UnknownType(from);
    }
    else throw Inrimage::UnknownType(from);
  }

}



void Inrimage::setType(WORD_TYPE t, unsigned int v,
		       VECTORIAL_MODE vm, bool rsz) {
  type = t;
  size_t vd = 0;
  size_t wd = 0;
  WORD_KIND wk=WK_FIXED;
  SIGN sg=SGN_UNSIGNED;
  
  switch(type) {
  case WT_SIGNED_CHAR:
    vd = 1;
    wd = 1;
    wk = WK_FIXED;
    sg = SGN_SIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_UNSIGNED_CHAR:
    vd = 1;
    wd = 1;
    wk = WK_FIXED;
    sg = SGN_UNSIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_UNSIGNED_SHORT:
    vd = 1;
    wd = 2;
    wk = WK_FIXED;
    sg = SGN_UNSIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_SIGNED_SHORT:
    vd = 1;
    wd = 2;
    wk = WK_FIXED;
    sg = SGN_SIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_UNSIGNED_INT:
    vd = 1;
    wd = 4;
    wk = WK_FIXED;
    sg = SGN_UNSIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_SIGNED_INT:
    vd = 1;
    wd = 4;
    wk = WK_FIXED;
    sg = SGN_SIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_UNSIGNED_LONG:
    vd = 1;
    wd = 8;
    wk = WK_FIXED;
    vectMode = VM_SCALAR;
    break;
  case WT_SIGNED_LONG:
    vd = 1;
    wd = 8;
    wk = WK_FIXED;
    sg = SGN_UNSIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_FLOAT:
    vd = 1;
    wd = 4;
    wk = WK_FLOAT;
    sg = SGN_SIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_DOUBLE:
    vd = 1;
    wd = 8;
    wk = WK_FLOAT;
    sg = SGN_SIGNED;
    vectMode = VM_SCALAR;
    break;
  case WT_RGB:
    vd = 3;
    wd = 1;
    wk = WK_FIXED;
    sg = SGN_UNSIGNED;
    vectMode = VM_INTERLACED;
    break;
  case WT_RGBA:
    vd = 4;
    wd = 1;
    wk = WK_FIXED;
    sg = SGN_UNSIGNED;
    vectMode = VM_INTERLACED;
    break;
  case WT_FLOAT_VECTOR:
    vd = v;
    wd = 4;
    wk = WK_FLOAT;
    sg = SGN_SIGNED;
    vectMode = vm;
    break;
  case WT_DOUBLE_VECTOR:
    vd = v;
    wd = 8;
    wk = WK_FLOAT;
    sg = SGN_SIGNED;
    vectMode = vm;
    break;
  }
  
  if(rsz) resize(xdim, ydim, zdim, vd, wd, wk, sg);
  else {
    vdim = vd;
    wordKind = wk;
    sign = sg;
    wdim = wd;
  }

  
  if(dynamic_cast<TrilinInterpolator const*>(interpolator) ||
     dynamic_cast<TrilinVectInterpolator const*>(interpolator) ||
     dynamic_cast<CylTrilinInterpolator const*>(interpolator) ||
     dynamic_cast<SphTrilinInterpolator const*>(interpolator))
    setInterpolator(Interpolator::trilinearInterpolatorByType(geom,
							      type,
							      vdim));
}   

unsigned int Inrimage::sizeofWords(void) const {
  switch(type) {
  case WT_SIGNED_CHAR:
  case WT_UNSIGNED_CHAR:
  case WT_RGB:
  case WT_RGBA:
    return sizeof(unsigned char);    
  case WT_UNSIGNED_SHORT:
    return sizeof(unsigned short int);
  case WT_SIGNED_SHORT:
    return sizeof(short int);
  case WT_UNSIGNED_INT:
    return sizeof(unsigned int);
  case WT_SIGNED_INT:
    return sizeof(int);
  case WT_UNSIGNED_LONG:
    return sizeof(unsigned LONGINT);
  case WT_SIGNED_LONG:
    return sizeof(LONGINT);
  case WT_FLOAT:
  case WT_FLOAT_VECTOR:
    return sizeof(float);
  case WT_DOUBLE:
  case WT_DOUBLE_VECTOR:
    return sizeof(double);
  default:
    return 0;
  }
}


void Inrimage::index3D(void) throw(NotEnoughMemory) {
  unsigned char **t1, *t2;
  unsigned int i, j, stype;
  stype = sizeofVoxels();

#ifdef __GNUC__
  try{
#endif
     array.threeD = (void ***) new unsigned char[zdim * sizeof(void **) +
						ydim * zdim * sizeof(void *)];
#ifdef __GNUC__
  } catch(std::bad_alloc&){
    throw NotEnoughMemory("Inrimage::index3D");
  }
#endif

  for(i = 0, t1 = (unsigned char **) (array.threeD) + zdim; i < zdim;
      i++, t1 += ydim)
    ((unsigned char ***) array.threeD)[i] = t1;

  for(i = 0, t2 = (unsigned char *) data; i < zdim; i++)
    for(j = 0; j < ydim; j++, t2 += xdim * stype)
      array.threeD[i][j] = t2;
}



void Inrimage::index4D(void) throw(NotEnoughMemory) {
  unsigned char ***t1, **t2, *t3;
  unsigned int i, j, k, stype;
  stype = sizeofWords();

#ifdef __GNUC__
  try{
#endif
     array.fourD = (void ****) new char[vdim * sizeof(void ***) +
				       vdim * zdim * sizeof(void **) +
				       vdim * zdim * ydim * sizeof(void *)];
#ifdef __GNUC__
  }catch(std::bad_alloc&){
     throw NotEnoughMemory("Inrimage::index4D");
  }
#endif

  for(i = 0, t1 = (unsigned char ***) (array.fourD) + vdim; i < vdim;
      i++, t1 += zdim)
    ((unsigned char ****) array.fourD)[i] = t1;

  for(i = 0, t2 = (unsigned char **) array.fourD + vdim * (1 + zdim);
      i < vdim; i++) {
    for(j = 0; j < zdim; j++, t2 += ydim) {
      ((unsigned char ****) array.fourD)[i][j] = t2;
    }
  }

  for(i = 0, t3 = (unsigned char *) data; i < vdim; i++) {
    for(j = 0; j < zdim; j++) {
      for(k = 0; k < ydim; k++, t3 += xdim * stype) {
	((unsigned char ****) array.fourD)[i][j][k] = t3;
      }
    }
  }
}



void Inrimage::buildIndex()  throw(NotEnoughMemory) {
  deleteArray();
  try {
    if(vectMode == VM_NON_INTERLACED) index4D();
    else index3D();
  }
  catch(NotEnoughMemory& e) {
    e.push("Inrimage::buildIndex");
    throw e;
  }
}

void Inrimage::copyHeader(const Inrimage *inr){
  translation=inr->getTranslationFromConstImage();
  rotation=inr->getRotationMatrixFromConstImage();
}

void Inrimage::allocate(void) throw(NotEnoughMemory) {
  try {
    BaseInrimage::allocate();
  }
  catch(BaseInrimage::NotEnoughMemory& e) {
    e.push("Inrimage::allocate");
    throw e;
  }
  try {
    buildIndex();
  }
  catch(Inrimage::NotEnoughMemory& e) {
    e.push("Inrimage::allocate");
    throw e;
  }
}


void Inrimage::free(void) {
  BaseInrimage::free();
  deleteArray();
}


void *Inrimage::setData(void *_data)
  throw(NotEnoughMemory) {

  deleteArray();
  void *old = BaseInrimage::setData(_data);
  try{
    buildIndex();
  }
  catch(NotEnoughMemory& e) {
    e.push("Inrimage::setData");
    throw e;
  }

  return old;
}


void Inrimage::deleteArray(void) {
  if(vectMode == VM_NON_INTERLACED) {
    if(array.fourD) delete[] array.fourD;
    array.fourD = 0;
  }
  else {
    if(array.threeD) delete[] array.threeD;
    array.threeD = 0;
  }
}


void Inrimage::queryGeometry(void) {
  char __name[100];
  if(userField("GEOMETRY", __name, 100)) {
    if(!strcmp(__name, "CARTESIAN")) geom = CARTESIAN;
    if(!strcmp(__name, "CYLINDRICAL")) geom = CYLINDRICAL;
    if(!strcmp(__name, "SPHERICAL")) geom = SPHERICAL;
  }
  else {
    // for compatibility transition only
    if(userField("ALPHA", 0, 0)) geom = CYLINDRICAL;
    else geom = CARTESIAN;
  }
}

void Inrimage::setInterpolator(Interpolator* i){
  Interpolator* o=const_cast<Interpolator*>(interpolator);
  if (o)
    delete o;
  interpolator=i;
}

void Inrimage::setExtrapolator(Extrapolator* e){
  Extrapolator* o=const_cast<Extrapolator*>(extrapolator);
  if (o)
    delete o;
  extrapolator=e;
}

Voxel* Inrimage::getRealVoxel(double x, double y, double z) const {
  return &getRealVoxel(x, y, z, *Voxel::voxelByType(type));
}

Voxel* Inrimage::getRealVoxel(Vec3<double> pos) const {
  return &getRealVoxel(pos.getX(),pos.getY(),pos.getZ(), *Voxel::voxelByType(type));
}


bool Inrimage::in(const Vec3<double>& pos) const {
  if (useTransformation) {
    Vec3<double> p=pos;
    /// transform from world coordinates into image coordinates
    p -= translation;
    p = rotation.getTransposed()*p;
    return((p[0] >= (-0.5*vx)) && (p[0] <= ((((double)xdim)-0.5)*vx)) &&
	   (p[1] >= (-0.5*vy)) && (p[1] <= ((((double)ydim)-0.5)*vy)) &&
	   (p[2] >= (-0.5*vz)) && (p[2] <= ((((double)zdim)-0.5)*vz)));  
  }
  else
    return((pos[0] >= (-0.5*vx)) && (pos[0] <= ((((double)xdim)-0.5)*vx)) &&
	   (pos[1] >= (-0.5*vy)) && (pos[1] <= ((((double)ydim)-0.5)*vy)) &&
	   (pos[2] >= (-0.5*vz)) && (pos[2] <= ((((double)zdim)-0.5)*vz)));  
}

  
Voxel& Inrimage::getRealVoxel(double x, double y, double z, Voxel& v) const {
  if (useTransformation) {
    /// transform from world coordinates into image coordinates
    Vec3<double> p=Vec3<double>(x,y,z)-translation;
    p = rotation.getTransposed()*p;
    x=p.getX(); y=p.getY(); z=p.getZ();
  }
  if((x >= (-0.5*vx)) && (x <= ((((double)xdim)-0.5)*vx)) &&
     (y >= (-0.5*vy)) && (y <= ((((double)ydim)-0.5)*vy)) &&
     (z >= (-0.5*vz)) && (z <= ((((double)zdim)-0.5)*vz)))
    return (*interpolator)(*this, x, y, z, v);
  else
    return (*extrapolator)(*this, x, y, z, v);
}

bool Inrimage::map3DWorldToVoxelCoordinates(const Vec3<double> pos,Vec3<double> &res) const{
Vec3<double> p;
	if (useTransformation) {
    /// transform from world coordinates into image coordinates
    p=pos-translation;
    p = rotation.getTransposed()*p;
  } else 
	  p=pos;
res=Vec3<double>(p.getX()/vx,p.getY()/vy,p.getZ()/vz);
 return((p[0] >= (-0.5*vx)) && (p[0] <= ((((double)xdim)-0.5)*vx)) &&
	   (p[1] >= (-0.5*vy)) && (p[1] <= ((((double)ydim)-0.5)*vy)) &&
	   (p[2] >= (-0.5*vz)) && (p[2] <= ((((double)zdim)-0.5)*vz)));  
}
void Inrimage::map3DVoxelToWorldCoordinates(const Vec3<double> pos,Vec3<double> &res) const
{
	res=Vec3<double>(pos.getX()*vx,pos.getY()*vy,pos.getZ()*vz);
	if (useTransformation) {
		res=rotation*res;
		res+=translation;
	} 
}

Voxel& Inrimage::getRealVoxelImageSpace(double x, double y, double z, Voxel& v) const {
  if((x >= (-0.5*vx)) && (x <= ((((double)xdim)-0.5)*vx)) &&
     (y >= (-0.5*vy)) && (y <= ((((double)ydim)-0.5)*vy)) &&
     (z >= (-0.5*vz)) && (z <= ((((double)zdim)-0.5)*vz)))
    return (*interpolator)(*this, x, y, z, v);
  else
    return (*extrapolator)(*this, x, y, z, v);
}
 
Voxel *Inrimage::getVoxel(unsigned int x, unsigned int y, unsigned int z)
  const {

  switch(type) {
  case WT_SIGNED_CHAR:
    return new Voxel8s(((char ***) array.threeD)[z][y][x]);
  case WT_UNSIGNED_CHAR:
    return new Voxel8u(((unsigned char ***) array.threeD)[z][y][x]);
  case WT_UNSIGNED_SHORT:
    return new Voxel16u(((unsigned short int ***) array.threeD)[z][y][x]);
  case WT_SIGNED_SHORT:
    return new Voxel16s(((short int ***) array.threeD)[z][y][x]);
  case WT_UNSIGNED_INT:
    return new Voxel32u(((unsigned int ***) array.threeD)[z][y][x]);
  case WT_SIGNED_INT:
    return new Voxel32s(((int ***) array.threeD)[z][y][x]);
  case WT_UNSIGNED_LONG:
    return new Voxel64u(((unsigned LONGINT ***) array.threeD)[z][y][x]);
  case WT_SIGNED_LONG:
    return new Voxel64s(((LONGINT ***) array.threeD)[z][y][x]);
  case WT_FLOAT:
    return new VoxelFloat(((float ***) array.threeD)[z][y][x]);
  case WT_DOUBLE:
    return new VoxelDouble(((double ***) array.threeD)[z][y][x]);
  case WT_RGB:
    return new VoxelRGB(((unsigned char ***) array.threeD)[z][y][3*x],
			((unsigned char ***) array.threeD)[z][y][3*x+1],
			((unsigned char ***) array.threeD)[z][y][3*x+2]);
  case WT_RGBA:
#if (defined (_ALPHA_) || defined (_LINUX_))
    return new VoxelRGBA(((unsigned char ***) array.threeD)[z][y][4*x+3],
			 ((unsigned char ***) array.threeD)[z][y][4*x+2],
			 ((unsigned char ***) array.threeD)[z][y][4*x+1],
			 ((unsigned char ***) array.threeD)[z][y][4*x]);
#else
    return new VoxelRGBA(((unsigned char ***) array.threeD)[z][y][4*x],
			 ((unsigned char ***) array.threeD)[z][y][4*x+1],
			 ((unsigned char ***) array.threeD)[z][y][4*x+2],
			 ((unsigned char ***) array.threeD)[z][y][4*x+3]);
#endif
  case WT_FLOAT_VECTOR:
    if(vectMode == VM_INTERLACED) {
      return new VoxelFV(((float ***) array.threeD)[z][y] + vdim*x, vdim);
    }
    else {
      float *val = new float[vdim];
      for(unsigned int i = 0; i < vdim; i++)
	val[i] = ((float ****) array.fourD)[i][z][y][x];
      Voxel *v = new VoxelFV(val, vdim);
      delete val;
      return v;
    }
    break;
  case WT_DOUBLE_VECTOR:
     if(vectMode == VM_INTERLACED) {
      return new VoxelDV(((double ***) array.threeD)[z][y] + vdim*x, vdim);
    }
    else {
      double *val = new double[vdim];
      for(unsigned int i = 0; i < vdim; i++)
	val[i] = ((double ****) array.fourD)[i][z][y][x];
      Voxel *v = new VoxelDV(val, vdim);
      delete val;
      return v;
    }
   break;
  }

  return 0;
}


Voxel& Inrimage::getVoxel(unsigned int x, unsigned int y, unsigned int z,
			Voxel &v) const {

  switch(type) {
  case WT_SIGNED_CHAR:
    (static_cast<Voxel8s *>(&v))->setValue(((char ***) array.threeD)[z][y][x]);
    break;
  case WT_UNSIGNED_CHAR:
    (static_cast<Voxel8u *>(&v))->setValue(((unsigned char ***) array.threeD)[z][y][x]);
    break;
  case WT_UNSIGNED_SHORT:
    (static_cast<Voxel16u *>(&v))->setValue(((unsigned short int ***) array.threeD)[z][y][x]);
    break;
  case WT_SIGNED_SHORT:
    (static_cast<Voxel16s *>(&v))->setValue(((short int ***) array.threeD)[z][y][x]);
    break;
  case WT_UNSIGNED_INT:
    (static_cast<Voxel32u *>(&v))->setValue(((unsigned int ***) array.threeD)[z][y][x]);
    break;
  case WT_SIGNED_INT:
    (static_cast<Voxel32s *>(&v))->setValue(((int ***) array.threeD)[z][y][x]);
    break;
  case WT_UNSIGNED_LONG:
    (static_cast<Voxel64u *>(&v))->setValue(((unsigned LONGINT ***) array.threeD)[z][y][x]);
    break;
  case WT_SIGNED_LONG:
    (static_cast<Voxel64s *>(&v))->setValue(((LONGINT ***) array.threeD)[z][y][x]);
    break;
  case WT_FLOAT:
    (static_cast<VoxelFloat *>(&v))->setValue(((float ***) array.threeD)[z][y][x]);
    break;
  case WT_DOUBLE:
    (static_cast<VoxelDouble *>(&v))->setValue(((double ***) array.threeD)[z][y][x]);
    break;
  case WT_RGB:
    (static_cast<VoxelRGB *>(&v))->setValue(
                       ((unsigned char ***) array.threeD)[z][y][3 * x],
		       ((unsigned char ***) array.threeD)[z][y][3 * x + 1],
		       ((unsigned char ***) array.threeD)[z][y][3 * x + 2]);
    break;
  case WT_RGBA:
#if (defined (_ALPHA_) || defined (_LINUX_))
    (static_cast<VoxelRGBA *>(&v))->setValue(
                       ((unsigned char ***) array.threeD)[z][y][3 * x + 3],
		       ((unsigned char ***) array.threeD)[z][y][3 * x + 2],
		       ((unsigned char ***) array.threeD)[z][y][3 * x + 1],
		       ((unsigned char ***) array.threeD)[z][y][3 * x]);
    break;
#else
    (static_cast<VoxelRGBA *>(&v))->setValue(
                       ((unsigned char ***) array.threeD)[z][y][3 * x],
		       ((unsigned char ***) array.threeD)[z][y][3 * x + 1],
		       ((unsigned char ***) array.threeD)[z][y][3 * x + 2],
		       ((unsigned char ***) array.threeD)[z][y][3 * x + 3]);
    break;
#endif
  case WT_FLOAT_VECTOR:
    if(vectMode == VM_INTERLACED) {
      (static_cast<VoxelFV *>(&v))->setValue(((float ***) array.threeD)[z][y] + x*vdim, vdim);
    }
    else {
      VoxelFV *fv = static_cast<VoxelFV *>(&v);
      float *f = ((float ****) array.threeD)[0][z][y] + x;
      unsigned long size = zdim * ydim * xdim;
      for(unsigned int i = 0; i < vdim; i++, f += size) fv->operator[](i) = *f;
    }
    break;
  case WT_DOUBLE_VECTOR:
    if(vectMode == VM_INTERLACED) {
      (static_cast<VoxelDV *>(&v))->setValue(((double ***) array.threeD)[z][y] + x*vdim, vdim);
    }
    else {
      VoxelDV *fv = static_cast<VoxelDV *>(&v);
      double *f = ((double ****) array.threeD)[0][z][y] + x;
      unsigned long size = zdim * ydim * xdim;
      for(unsigned int i = 0; i < vdim; i++, f += size) fv->operator[](i) = *f;
    }
    break;
  }
  return v;
}


void Inrimage::setVoxel(unsigned int x, unsigned int y, unsigned int z,
			const Voxel& v) {
  unsigned int i;
  const VoxelRGB *vrgb;
  const VoxelRGBA *vrgba;
  const VoxelFV *vfv;
  const VoxelDV *vdv;

  switch(type) {
  case WT_SIGNED_CHAR:
    ((char ***) array.threeD)[z][y][x] =
      static_cast<const Voxel8s *>(&v)->getValue();
    break;
  case WT_UNSIGNED_CHAR:
    ((unsigned char ***) array.threeD)[z][y][x] =
      static_cast<const Voxel8u *>(&v)->getValue();
    break;
  case WT_UNSIGNED_SHORT:
    ((unsigned short int ***) array.threeD)[z][y][x] =
      static_cast<const Voxel16u *>(&v)->getValue();
    break;
  case WT_SIGNED_SHORT:
    ((short int ***) array.threeD)[z][y][x] =
      static_cast<const Voxel16s *>(&v)->getValue();
    break;
  case WT_UNSIGNED_INT:
    ((unsigned int ***) array.threeD)[z][y][x] =
      static_cast<const Voxel32u *>(&v)->getValue();
    break;
  case WT_SIGNED_INT:
    ((int ***) array.threeD)[z][y][x] =
      static_cast<const Voxel32s *>(&v)->getValue();
    break;
  case WT_UNSIGNED_LONG:
    ((unsigned LONGINT ***) array.threeD)[z][y][x] =
      static_cast<const Voxel64u *>(&v)->getValue();
    break;
  case WT_SIGNED_LONG:
    ((LONGINT ***) array.threeD)[z][y][x] =
      static_cast<const Voxel64s *>(&v)->getValue();
    break;
  case WT_FLOAT:
    ((float ***) array.threeD)[z][y][x] =
      static_cast<const VoxelFloat *>(&v)->getValue();
    break;
  case WT_DOUBLE:
    ((double ***) array.threeD)[z][y][x] =
      static_cast<const VoxelDouble *>(&v)->getValue();
    break;
  case WT_RGB:
    vrgb = static_cast<const VoxelRGB *>(&v);
    ((unsigned char ***) array.threeD)[z][y][3 * x] = vrgb->getValue()[0];
    ((unsigned char ***) array.threeD)[z][y][3 * x + 1] = vrgb->getValue()[1];
    ((unsigned char ***) array.threeD)[z][y][3 * x + 2] = vrgb->getValue()[2];
    break;
  case WT_RGBA:
    vrgba = static_cast<const VoxelRGBA *>(&v);
#if (defined (_ALPHA_) || defined (_LINUX_))
    ((unsigned char ***) array.threeD)[z][y][3 * x + 3] = vrgba->getValue()[0];
    ((unsigned char ***) array.threeD)[z][y][3 * x + 2] = vrgba->getValue()[1];
    ((unsigned char ***) array.threeD)[z][y][3 * x + 1] = vrgba->getValue()[2];
    ((unsigned char ***) array.threeD)[z][y][3 * x] = vrgba->getValue()[3];
#else
    ((unsigned char ***) array.threeD)[z][y][3 * x] = vrgba->getValue()[0];
    ((unsigned char ***) array.threeD)[z][y][3 * x + 1] = vrgba->getValue()[1];
    ((unsigned char ***) array.threeD)[z][y][3 * x + 2] = vrgba->getValue()[2];
    ((unsigned char ***) array.threeD)[z][y][3 * x + 3] = vrgba->getValue()[3];
#endif
    break;
  case WT_FLOAT_VECTOR:
    vfv = static_cast<const VoxelFV *>(&v);
    if(vectMode == VM_INTERLACED) {
      for(i = 0; i < vdim; i++)
	((float ***) array.threeD)[z][y][vdim * x + i] = vfv->getValue()[i];
    }
    else {
      for(i = 0; i < vdim; i++)
	((float ****) array.fourD)[i][z][y][x] = vfv->getValue()[i];
    }
    break;
  case WT_DOUBLE_VECTOR:
    vdv = static_cast<const VoxelDV *>(&v);
    if(vectMode == VM_INTERLACED) {
      for(i = 0; i < vdim; i++)
	((double ***) array.threeD)[z][y][vdim * x + i] = vdv->getValue()[i];
    }
    else {
      for(i = 0; i < vdim; i++)
	((double ****) array.fourD)[i][z][y][x] = vdv->getValue()[i];
    }
    break;
  default:
    break;
  }
}


Voxel *Inrimage::checkVoxel(unsigned int x,
			    unsigned int y,
			    unsigned int z) const {
  unsigned int rx, ry, rz;

  rx = xdim - 1;
  if(x < rx) rx = x;
  ry = ydim - 1;
  if(y < ry) ry = y;
  rz = zdim - 1;
  if(z < rz) rz = z;

  return getVoxel(rx, ry, rz);
}

void Inrimage::isolateRegion(unsigned int x1, unsigned int y1, unsigned int z1,
			     unsigned int x2, unsigned int y2, unsigned int z2)
  throw(Inrimage::InvalidArguments) {

  unsigned int j, i,k,regionSize,stype;
  unsigned char *c;
  int nx, ny, nz;
  char from[] = "Inrimage::isolateRegion";

  if(x2 >= xdim || y2 >= ydim || z2 >= zdim) throw InvalidArguments(from);
  nx = x2 - x1 + 1;
  ny = y2 - y1 + 1;
  nz = z2 - z1 + 1;
  if(nx <= 0 || ny <= 0 || nz <= 0) throw InvalidArguments(from);

  stype = sizeofVoxels();
  /// set to zero planes bewteen 0 and z1
  regionSize=z1*xdim*ydim*stype;
  memset ((void *) data,0,regionSize);
  /// set to zero planes bewteen z2+1 and getZ()-1
  regionSize=(zdim-z2-1)*xdim*ydim*stype;
  c=(unsigned char *)data+z2*xdim*ydim*stype;
  memset ((void *) c,0,regionSize);

  for(c=(unsigned char *)data+z1*xdim*ydim*stype,k=z1;k<=z2;k++) {
    for(j = 0; j<ydim; j++) {
      for(i = 0; i<xdim; i++,c+=stype) {
	if ((i<x1) || (i>x2) || (j<y1) || (j>y2))
	  {
	    memset ((void *) c,0,stype);
	  }
      }
    }
  }
}

void Inrimage::zeroRegion(unsigned int x1, unsigned int y1, unsigned int z1,
			  unsigned int x2, unsigned int y2, unsigned int z2)
  throw(Inrimage::InvalidArguments) {

  unsigned int j,k,stype, plane, plane1, line, line1;
  unsigned char *sbuf;
  int nx, ny, nz;
  char from[] = "Inrimage::zeroRegion";

  if(x2 >= xdim || y2 >= ydim || z2 >= zdim) throw InvalidArguments(from);
  nx = x2 - x1 + 1;
  ny = y2 - y1 + 1;
  nz = z2 - z1 + 1;
  if(nx <= 0 || ny <= 0 || nz <= 0) throw InvalidArguments(from);

  stype = sizeofVoxels();
  plane = xdim * ydim * stype;
  plane1 = nx * ny * stype;
  line = xdim * stype;
  line1 = nx * stype;
  
  for(k = z1; k <= z2; k++) {
    sbuf = (unsigned char *) data + k * plane + y1 * line + x1 * stype;
    for(j = y1; j <= y2; j++, sbuf += line)
      memset((void *) sbuf, 0 , line1);
  }

}

void Inrimage::crop(unsigned int x1, unsigned int y1, unsigned int z1,
		    unsigned int x2, unsigned int y2, unsigned int z2)
  throw(Inrimage::InvalidArguments) {

  unsigned int j, k, stype, plane, plane1, line, line1;
  unsigned char *sbuf, *dbuf, *nbuf;
  int nx, ny, nz;
  char from[] = "Inrimage::crop";

  if(x2 >= xdim || y2 >= ydim || z2 >= zdim) throw InvalidArguments(from);
  nx = x2 - x1 + 1;
  ny = y2 - y1 + 1;
  nz = z2 - z1 + 1;
  if(nx <= 0 || ny <= 0 || nz <= 0) throw InvalidArguments(from);

  stype = sizeofVoxels();
  nbuf = new unsigned char[nx * ny * nz * stype];

  plane = xdim * ydim * stype;
  plane1 = nx * ny * stype;
  line = xdim * stype;
  line1 = nx * stype;

  for(k = z1; k <= z2; k++) {
    sbuf = (unsigned char *) data + k * plane + y1 * line + x1 * stype;
    dbuf = (unsigned char *) nbuf + (k-z1) * plane1;
    for(j = y1; j <= y2; j++, sbuf += line, dbuf += line1)
      memcpy((void *) dbuf, (const void *) sbuf, line1);
  }

  xdim = nx;
  ydim = ny;
  zdim = nz;
  delete[] (unsigned char *) setData((void *) nbuf );
  /// add offset
  translation+=Vec3<double>(x1*vx,y1*vy,z1*vz);

}


#define COMPUTEVECTORIALBUFFERBOUND(TYPE1,TYPE2,INR,A,B) computeVectorialBufferBounds((TYPE1)0,INR,A,B)


#define COMPUTEBUFFERBOUND(TYPE1,TYPE2,INR,A,B) computeBufferBounds((TYPE1)0,INR,A,B)


void Inrimage::getBounds(Voxel *min, Voxel *max, bool vectorNorm) const {
  switch(type) {
  case WT_UNSIGNED_CHAR:
    COMPUTEBUFFERBOUND(unsigned char, Voxel8u, *this, dynamic_cast<Voxel8u *>(min),
					     dynamic_cast<Voxel8u *>(max));
    break;
  case WT_SIGNED_CHAR:
    COMPUTEBUFFERBOUND(char, Voxel8s, *this, dynamic_cast<Voxel8s *>(min),
					     dynamic_cast<Voxel8s *>(max));
    break;
  case WT_SIGNED_SHORT:
    COMPUTEBUFFERBOUND(short int, Voxel16s, *this, dynamic_cast<Voxel16s *>(min),
					     dynamic_cast<Voxel16s *>(max));
    break;
  case WT_UNSIGNED_SHORT:
    COMPUTEBUFFERBOUND(unsigned short int, Voxel16u, *this,dynamic_cast<Voxel16u *>(min),
						      dynamic_cast<Voxel16u *>(max));
    break;
  case WT_SIGNED_INT:
    COMPUTEBUFFERBOUND(int, Voxel32s, *this,dynamic_cast<Voxel32s *>(min),
				       dynamic_cast<Voxel32s *>(max));
    break;
  case WT_UNSIGNED_INT:
    COMPUTEBUFFERBOUND(unsigned int, Voxel32u, *this,dynamic_cast<Voxel32u *>(min),
						dynamic_cast<Voxel32u *>(max));
    break;
  case WT_SIGNED_LONG:
    COMPUTEBUFFERBOUND(LONGINT, Voxel64s, *this,dynamic_cast<Voxel64s *>(min),
					dynamic_cast<Voxel64s *>(max));
    break;
  case WT_UNSIGNED_LONG:
    COMPUTEBUFFERBOUND(unsigned LONGINT, Voxel64u, *this,dynamic_cast<Voxel64u *>(min),
						 dynamic_cast<Voxel64u *>(max));
    break;
  case WT_FLOAT:
    COMPUTEBUFFERBOUND(float, VoxelFloat, *this,dynamic_cast<VoxelFloat *>(min),
					   dynamic_cast<VoxelFloat *>(max));
    break;
  case WT_DOUBLE:
    COMPUTEBUFFERBOUND(double, VoxelDouble, *this,dynamic_cast<VoxelDouble *>(min),
					     dynamic_cast<VoxelDouble *>(max));
    break;
  case WT_FLOAT_VECTOR:
    if(vectorNorm) 
      COMPUTEVECTORIALBUFFERBOUND(float, VoxelFloat, *this,dynamic_cast<VoxelFloat *>(min),
						      dynamic_cast<VoxelFloat *>(max));
    else
      COMPUTEBUFFERBOUND(float, VoxelFloat, *this,dynamic_cast<VoxelFloat *>(min),
					     dynamic_cast<VoxelFloat *>(max));
    break;
  case WT_DOUBLE_VECTOR:
    if(vectorNorm) 
      COMPUTEVECTORIALBUFFERBOUND(double, VoxelDouble, *this,dynamic_cast<VoxelDouble *>(min),
							dynamic_cast<VoxelDouble *>(max));
    else
      COMPUTEBUFFERBOUND(double, VoxelDouble, *this,dynamic_cast<VoxelDouble *>(min),
					       dynamic_cast<VoxelDouble *>(max));      
    break;
  case WT_RGB:
    dynamic_cast<VoxelRGB *>(min)->setMin();
    dynamic_cast<VoxelRGB *>(max)->setMax();
    break;
  case WT_RGBA:
    dynamic_cast<VoxelRGBA *>(min)->setMin();
    dynamic_cast<VoxelRGBA *>(max)->setMax();
    break;

  default:
    break;
  }
}



#define SETBUFFERBOUNDS(TYPE1,TYPE2,INR,A,B,C) setBufferBounds((TYPE1)0,INR,A,B,C)


#define SETFLOATBUFFERBOUNDS(TYPE1,TYPE2,INR,A,B,C) setFloatBufferBounds((TYPE1)0,INR,A,B,C)





void Inrimage::setBounds(Voxel *min, Voxel *max, int z) {
  switch(type) {
  case WT_SIGNED_CHAR:
    SETBUFFERBOUNDS(char, Voxel8s, *this, dynamic_cast<Voxel8s *>(min),
					    dynamic_cast<Voxel8s *>(max), z);
    break;
  case WT_UNSIGNED_CHAR:
    SETBUFFERBOUNDS(unsigned char, Voxel8u, *this, dynamic_cast<Voxel8u *>(min),
					    dynamic_cast<Voxel8u *>(max), z);
    break;
  case WT_SIGNED_SHORT:
    SETBUFFERBOUNDS(short int, Voxel16s,*this, dynamic_cast<Voxel16s *>(min),
					 dynamic_cast<Voxel16s *>(max), z);
    break;
  case WT_UNSIGNED_SHORT:
    SETBUFFERBOUNDS(unsigned short int, Voxel16u,*this, dynamic_cast<Voxel16u *>(min),
						  dynamic_cast<Voxel16u *>(max), z);
    break;
  case WT_SIGNED_INT:
    SETBUFFERBOUNDS(int, Voxel32s,*this, dynamic_cast<Voxel32s *>(min),
				   dynamic_cast<Voxel32s *>(max), z);
    break;
  case WT_UNSIGNED_INT:
    SETBUFFERBOUNDS(unsigned int, Voxel32u,*this, dynamic_cast<Voxel32u *>(min),
					    dynamic_cast<Voxel32u *>(max), z);
    break;
  case WT_SIGNED_LONG:
    SETBUFFERBOUNDS(LONGINT, Voxel64s,*this, dynamic_cast<Voxel64s *>(min),
				    dynamic_cast<Voxel64s *>(max), z);
    break;
  case WT_UNSIGNED_LONG:
    SETBUFFERBOUNDS(unsigned LONGINT, Voxel64u,*this, dynamic_cast<Voxel64u *>(min),
					     dynamic_cast<Voxel64u *>(max), z);
    break;
  case WT_FLOAT:
    SETFLOATBUFFERBOUNDS(float, VoxelFloat,*this, dynamic_cast<VoxelFloat *>(min),
					    dynamic_cast<VoxelFloat *>(max), z);
    break;
  case WT_DOUBLE:
    SETFLOATBUFFERBOUNDS(double, VoxelDouble,*this, dynamic_cast<VoxelDouble *>(min),
					      dynamic_cast<VoxelDouble *>(max),
					      z);
    break;

  case WT_FLOAT_VECTOR:
  case WT_DOUBLE_VECTOR:
  case WT_RGB:
  case WT_RGBA:
    break;

  default:
    break;
  }  
}



void Inrimage::flip(SLICE_ORIENTATION dir) {
  unsigned int k, y, z, plane, line, stype;
  char tmp;
  unsigned char *dbuf, *sbuf, *end;

  stype = sizeofVoxels();
  plane = xdim * ydim * stype;
  line = xdim * stype;

  switch(dir) {
  case SLICE_X:
    for(z = 0; z < zdim; z++) {
      for(y = 0; y < ydim; y++) {
	sbuf = (unsigned char *) data + z * plane + y * line;
	dbuf = sbuf + line - stype;
	for(; sbuf < dbuf; dbuf -= 2*stype) {
	  for(k = 0; k < stype; k++) {
	    tmp = *sbuf;
	    *sbuf++ = *dbuf;
	    *dbuf++ = tmp;
	  }
	}
      }
    }
    break;
	
  case SLICE_Y:
    for(z = 0; z < zdim; z++) {
      for(y = 0; y < ydim / 2; y++) {
	sbuf = (unsigned char *) data + z * plane + y * line;
	end = sbuf + line;
	dbuf = (unsigned char *) data + z * plane + (ydim - 1 - y) * line;
	for(; sbuf < end;) {
	  tmp = *sbuf;
	  *sbuf++ = *dbuf;
	  *dbuf++ = tmp;
	}
      }
    }
    break;

  case SLICE_Z:
    for(z = 0; z < zdim / 2; z++) {
      sbuf = (unsigned char *) data + z * plane;
      end = sbuf + plane;
      dbuf = (unsigned char *) data + (zdim - 1 - z) * plane;
      for(; sbuf < end;) {
	for(k = 0; k < stype; k++) {
	  tmp = *sbuf;
	  *sbuf++ = *dbuf;
	  *dbuf++ = tmp;
	}
      }
    }
    break;

  default:
    break;
  }

}


void Inrimage::switchGeometry(SLICE_ORIENTATION geom1,
			      SLICE_ORIENTATION geom2) {
  if(geom1 == geom2) return;
  if(geom1 > geom2) {
    SLICE_ORIENTATION g = geom1;
    geom1 = geom2;
    geom2 = g;
  }

  Inrimage *inr;
  Voxel *v = Voxel::voxelByType(getType());
  if(geom1 == SLICE_X) {
    // geom1 == SLICE_X, geom2 == SLICE_Y
    if(geom2 == SLICE_Y) {
      inr = new Inrimage(getY(), getX(), getZ(), getType());
      for(unsigned int k = 0; k < getZ(); k++) {
	for(unsigned int j = 0; j < getY(); j++) {
	  for(unsigned int i = 0; i < getX(); i++) {
	    getVoxel(i, j, k, *v);
	    inr->setVoxel(j, i, k, *v);
	  }
	}
      }
      float t = vx;
      vx = vy;
      vy = t;
    }
    // geom1 == SLICE_X, geom2 == SLICE_Z
    else {
      inr = new Inrimage(getZ(), getY(), getX(), getType());
      for(unsigned int k = 0; k < getZ(); k++) {
	for(unsigned int j = 0; j < getY(); j++) {
	  for(unsigned int i = 0; i < getX(); i++) {
	    getVoxel(i, j, k, *v);
	    inr->setVoxel(k, j, i, *v);
	  }
	}
      }
      float t = vx;
      vx = vz;
      vz = t;
    }
  }
  // geom1 == SLICE_Y, geom2 == SLICE_Z
  else {
    inr = new Inrimage(getX(), getZ(), getY(), getType());
    for(unsigned int k = 0; k < getZ(); k++) {
      for(unsigned int j = 0; j < getY(); j++) {
	for(unsigned int i = 0; i < getX(); i++) {
	  getVoxel(i, j, k, *v);
	  inr->setVoxel(i, k, j, *v);
	}
      }
      float t = vz;
      vz = vy;
      vy = t;
    }
  }
  delete v;

  xdim = inr->getX();
  ydim = inr->getY();
  zdim = inr->getZ();
  delete[] (unsigned char *) setData(inr->getData());
  inr->setData(0);
  delete inr;
}


void Inrimage::mel(const Inrimage *inr,
		   int xdest, int ydest, int zdest,
		   int xorg, int yorg, int zorg,
		   unsigned int dx, unsigned int dy, unsigned int dz)
  throw(Inrimage::InvalidArguments) {

  unsigned char *orgptr, *destptr;
  char fromstr[] = "Inrimage::mel", invdest[] = "invalid destination box";
  unsigned int from[3], to[3], j, k, lineSize, orgPlane, orgLine,
    destPlane, destLine;

  if(inr->type != type) throw InvalidArguments(fromstr, "another image type was expected in this method");
  if(xorg > (int) inr->xdim || yorg > (int) inr->ydim ||
     zorg > (int) inr->zdim)
    throw InvalidArguments(fromstr, "invalid source box");
  if(xdest > (int)xdim || ydest > (int)ydim || zdest > (int)zdim)
    throw InvalidArguments(fromstr, invdest);

  if(xorg < 0) {
    from[0] = 0;
    dx += xorg;
  }
  else from[0] = xorg;
  if(yorg < 0) {
    from[1] = 0;
    dy += yorg;
  }
  else from[1] = yorg;
  if(zorg < 0) {
    from[2] = 0;
    dz += zorg;
  }
  else from[2] = zorg;

  if(from[0] + dx > inr->xdim) dx = inr->xdim - from[0];
  if(from[1] + dy > inr->ydim) dy = inr->ydim - from[1];
  if(from[2] + dz > inr->zdim) dz = inr->zdim - from[2];

  if(xdest < 0) {
    if((int)(xdest + dx) < 0) throw InvalidArguments(fromstr, invdest);
    dx += xdest;
    from[0] -= xdest;
    to[0] = 0;
  }
  else {
    to[0] = xdest;
    if(xdest + dx > xdim) dx = xdim - xdest;
  }
  if(ydest < 0) {
    if((int)(ydest + dy) < 0) throw InvalidArguments(fromstr, invdest);
    dy += ydest;
    from[1] -= ydest;
    to[1] = 0;
  }
  else {
    to[1] = ydest;
    if(ydest + dy > ydim) dy = ydim - ydest;
  }
  if(zdest < 0) {
    if((int)(zdest + dz) < 0) throw InvalidArguments(fromstr, invdest);
    dz += zdest;
    from[2] -= zdest;
    to[2] = 0;
  }
  else {
    to[2] = zdest;
    if(zdest + dz > zdim) dz = zdim - zdest;
  }

  lineSize = dx * sizeofVoxels();
  orgLine = inr->xdim * sizeofVoxels();
  orgPlane = inr->ydim * orgLine;
  destLine = xdim * sizeofVoxels();
  destPlane = ydim * destLine;

  for(k = 0; k < dz; k++) {
    orgptr = (unsigned char *) inr->data + (from[2]+k) * orgPlane +
      from[1] * orgLine + from[0] * sizeofVoxels();
    destptr = (unsigned char *) data + (to[2]+k) * destPlane +
      to[1] * destLine + to[0] * sizeofVoxels();
    for(j = 0; j < dy; j++, orgptr += orgLine, destptr += destLine)
      memcpy((void *) destptr, (const void *) orgptr, lineSize);
  }
}

// create a noisy cube inside this inrimage
void Inrimage::noise(unsigned int x1, unsigned int y1, unsigned int z1,
		     unsigned int x2, unsigned int y2, unsigned int z2,
		     Noise kind, double mean, double dev)
  throw(Inrimage::InvalidArguments) {

#define MIN_VAL 1e-5

#ifndef WIN32
  struct timeval tv;
#endif
  unsigned int x, y, z;
  double a, b, c, m;
  char from[] = "Inrimage::noise";
  Voxel *v;

  if(x1 > x2 || y1 > y2 || z1 > z2)
    throw InvalidArguments(from, "invalid source box");
  if(x2 >= xdim || y2 >= ydim || z2 >= zdim)
    throw InvalidArguments(from, "invalid destination box");

  v = Voxel::voxelByType(type);
#ifndef WIN32
  gettimeofday(&tv, 0);
  srand(tv.tv_usec);
#else
  srand((unsigned int)time(0));
#endif
  for(z = z1; z <= z2; z++) {
    for(y = y1; y <= y2; y++) {
      for(x = x1; x <= x2; x++) {
	b = (double) rand() / (double) RAND_MAX;
	if(b < MIN_VAL) b = MIN_VAL;
	c = (double) rand() / (double) RAND_MAX;
	a = -2.0 * log(b);
	if(a < 0.0) a = 0.0;
	else a *= cos(2.0 * M_PI * c);

	switch(kind) {
	case WHITE_NOISE:
	  a = a * dev + mean;
	  break;
	case GAUSSIAN_NOISE:
	  getVoxel(x, y, z, *v);
	  m = v->getValueAsDouble();
	  a = a * dev + m;
	  break;
	case US_NOISE:
	  getVoxel(x, y, z, *v);
	  m = v->getValueAsDouble();
	  a = m + sqrt(m) * a * dev;
	  break;
	}
	if(a < v->minValue()) a = v->minValue();
	else if(a > v->maxValue()) a = v->maxValue();
	v->setValueFromDouble(a);
	setVoxel(x, y, z, *v);
      }
    }
  }
  
  delete v;

#undef MIN_VAL
}



bool Inrimage::vectorial(WORD_TYPE type) {
  return type == WT_RGB || type == WT_RGBA ||
    type == WT_FLOAT_VECTOR || type == WT_DOUBLE_VECTOR;
}


void Inrimage::eval(char *func,int nmbr,Var_list *op)
throw(Inrimage::InvalidArguments) {  /*****************************************/
  /* parse image                           */
  /*****************************************/
  char from[] = "Inrimage::eval";
  FunctionParser Func;
  int res,i,j,k;
  char errorpos[255];
  Var_list vars1,vars2,vars3,*w_op;
  int x,y,z;
  int sx,sy,sz;
  double* values;
  char * func_extended;
  char *func_work;
  char *func_op;
  char *func_op_w;
  int par_num;
  char Ops_s[] = "+-*/^=<>|&";
  char Ops[] = "^*+";
  char Ops_[] = " /-"; 
  int func_len;
  Inrimage **im_ar=0;
  Voxel **vox=0;
  Voxel *res_voxel = Voxel::voxelByType(getType());
  if(strlen(func)==0)
    throw InvalidArguments(from, "Empty function string");
  
  
  /********  pre processor ********/
  /* Adds parenthesis to supply   */
  /* opertor priority:            */
  /* ^,/or*,-or+                  */
  /********************************/
  /* allocate strings long enough to hold extra parenthesis */
  func_len=(int)(strlen(func)*5);
  func_extended=new(char[func_len]);
  func_work=new(char[func_len]);
  func_op=new(char[func_len]);
  func_op_w=new(char[func_len]);
  strcpy(func_extended,func);
  strcpy(func_work,func_extended);
  strcpy(func_op,func_extended);
  
  /* Have func_op contain O if position contains ANY operator */
  func_op[0]=' ';
  // operator is never on first positon
  /* loop over string */
  for(j=1;j<(int)strlen(func_work);j++){ 
    func_op[j]=' ';
    /* loop over operator characters */
    for(i=0;i<(int)strlen(Ops_s);i++){
      if(func_work[j]==Ops_s[i]){ 
	func_op[j]='O'; 
	//prevent negation to be treated as operator
	if((func_op[j-1]=='O')&&(func_work[j]=='-'))
	  func_op[j]=' ';
      }
    }
  }  
  strcpy(func_op_w,func_op);
  COUT<<func_op<<ENDL;   
  for(i=0;i<(int)strlen(Ops);i++){
    for(j=1;j<(int)strlen(func_work)-1;j++){
      if(((func_work[j]==Ops[i])||(func_work[j]==Ops_[i]))
	 &&(func_op_w[j]=='O')){
	k=j-1;
	par_num=0;
	while((k>0)&&
	      ((func_op_w[k]!='O')||(par_num<0))&&
	      !((par_num==0)&&(func_work[k]=='('))){
	  if(func_work[k]=='(') par_num++;
	  if(func_work[k]==')') par_num--;
	  k--;
	}
	if(k>0) k++;
	strcpy(func_op,func_op_w);
	strcpy(func_extended,func_work);
	strcpy(func_op_w+k+1,func_op+k);
	strcpy(func_work+k+1,func_extended+k);
	func_work[k]='(';
	func_op_w[k]=' ';
	k=j+2;
	if(k>(int)strlen(func_work))
	  k--;
	par_num=0;
	while((k<(int)strlen(func_work))&&
	      ((func_op_w[k]!='O')||(par_num>0))&&
	      !((par_num==0)&&(func_work[k]==')'))){
	  if(func_work[k]=='(') par_num++;
	  if(func_work[k]==')') par_num--; 
	  k++;
	}      
	if(k>=(int)strlen(func_work)){
	  func_work[k]=')';
	  func_op_w[k]=' ';
	  func_work[k+1]='\0';
	  func_op_w[k+1]='\0';
	}
	else{
	  strcpy(func_op,func_op_w);
	  strcpy(func_extended,func_work);
	  strcpy(func_op_w+k+1,func_op+k);
	  strcpy(func_work+k+1,func_extended+k);
	  func_work[k]=')';
	  func_op_w[k]=' ';
	}
	j++;
      }  
    }
  }
  strcpy(func_extended,func_work);
  /* pre-processor ends here */

  /* fill arrays of Voxel objects and Inrimage object pointers */
  if(nmbr>0){ 
    vox=new(Voxel*[nmbr]);
    im_ar=new(Inrimage *[nmbr]);
    w_op=op;
    for(i=0;i<nmbr;im_ar[i]=(Inrimage*)w_op->im,
	  vox[i] = Voxel::voxelByType(((Inrimage *)(w_op->im))->getType()),
	  i++,w_op=w_op->next);
  }

  /* prepare standar variables x,y,z */
  vars1.next=&vars2;
  strcpy(vars1.name,"x");
  vars2.next=&vars3;
  strcpy(vars2.name,"y");
  vars3.next=op;
  strcpy(vars3.name,"z");
  /* get image size */
  sx=(int)getX();
  sy=(int)getY();
  sz=(int)getZ();

  /* parse(pre-compile) function, evaluates only global functions (like mean) */
  res = Func.Parse(func_extended, &vars1,sx,sy,sz);
  if(Func.ErrorMsg()!=NULL){
    COUT<<Func.ErrorMsg()<<ENDL;
    for(i=0;i<res;errorpos[i]=' ',i++);
    errorpos[res]='^';
    errorpos[res+1]='\0';
    COUT<<func_extended<<ENDL<<errorpos<<ENDL;
    throw InvalidArguments(from, "Error in function string");
    return ;
  }

  /*****************************************/
  /* process result image                  */
  /*****************************************/
  
  values=new double[nmbr+3];
  for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
    for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
      for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
	for(i=0;i<nmbr;i++){ im_ar[i]->getVoxel(x, y, z, *vox[i]);
	  values[i+3] = vox[i]->getValueAsDouble();
        }
        res_voxel->setValueFromDouble(Func.Eval(values));
	setVoxel(x, y, z, *res_voxel);
      }
    }
  }
  
  /* delete all allocated stuff */
  for(i=0;i<nmbr;i++) delete(vox[i]);
  if(nmbr>0){
    delete(vox);
    delete(im_ar);
  }
  delete(res_voxel);
  delete(func_extended);
  delete(func_work);
  delete(func_op);
  delete(func_op_w);
  delete(values);
  return;
}

bool Inrimage::floating(WORD_TYPE type) {
  return type == WT_FLOAT || type == WT_DOUBLE ||
    type == WT_FLOAT_VECTOR || type == WT_DOUBLE_VECTOR;
}



#define SETCOMPONENTFROM(TYPE1,TYPE2,INR,A,B) setComponentFrom((TYPE1)0,(TYPE2)0,INR,A,B)



// Set vectorial components from an array of scalar inrimages
void Inrimage::setComponentsFrom(Inrimage **inrs)
  throw(InvalidSize, InvalidArguments) {

  char from[] = "Inrimage::setComponentsFrom";
  unsigned int i;

  if(type == WT_FLOAT_VECTOR || type == WT_DOUBLE_VECTOR) {

    for(i = 0; i < vdim; i++) {
      if(inrs[i]->xdim != xdim || inrs[i]->ydim != ydim ||
	 inrs[i]->zdim != zdim) throw InvalidSize(from);
      if(vectorial(inrs[i]->type)) throw InvalidArguments(from);
    }
		
    if(type == WT_FLOAT_VECTOR) {
      for(i = 0; i < vdim; i++) {
	switch(inrs[i]->type) {
	case WT_SIGNED_CHAR:
	  SETCOMPONENTFROM(char, float,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_CHAR:
	  SETCOMPONENTFROM(unsigned char, float,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_SHORT:
	  SETCOMPONENTFROM(unsigned short int,float,*this, inrs[i], i);
	  break;
	case WT_SIGNED_SHORT:
	  SETCOMPONENTFROM(short int, float,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_INT:
	  SETCOMPONENTFROM(unsigned int, float,*this,inrs[i], i);
	  break;
	case WT_SIGNED_INT:
	  SETCOMPONENTFROM(int, float,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_LONG:
	  SETCOMPONENTFROM(unsigned LONGINT, float,*this,inrs[i], i);
	  break;
	case WT_SIGNED_LONG:
	  SETCOMPONENTFROM(LONGINT, float,*this,inrs[i], i);
	  break;
	case WT_FLOAT:
	  SETCOMPONENTFROM(float, float,*this,inrs[i], i);
	  break;
	case WT_DOUBLE:
	  SETCOMPONENTFROM(double, float,*this,inrs[i], i);
	  break;
	default:
	  break;
	}
      }
    }
    else {
      for(i = 0; i < vdim; i++) {
	switch(inrs[i]->type) {
	case WT_SIGNED_CHAR:
	  SETCOMPONENTFROM(char, double,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_CHAR:
	  SETCOMPONENTFROM(unsigned char, double,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_SHORT:
	  SETCOMPONENTFROM(unsigned short int, double,*this,inrs[i], i);
	  break;
	case WT_SIGNED_SHORT:
	  SETCOMPONENTFROM(short int, double,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_INT:
	  SETCOMPONENTFROM(unsigned int, double,*this,inrs[i], i);
	  break;
	case WT_SIGNED_INT:
	  SETCOMPONENTFROM(int, double,*this,inrs[i], i);
	  break;
	case WT_UNSIGNED_LONG:
	  SETCOMPONENTFROM(unsigned LONGINT, double,*this,inrs[i], i);
	  break;
	case WT_SIGNED_LONG:
	  SETCOMPONENTFROM(LONGINT, double,*this,inrs[i], i);
	  break;
	case WT_FLOAT:
	  SETCOMPONENTFROM(float, double,*this,inrs[i], i);
	  break;
	case WT_DOUBLE:
	  SETCOMPONENTFROM(double, double,*this,inrs[i], i);
	  break;
	default:
	  break;
	}
      }
    }
  }
}




#define SPLITVECTORS(TYPE,INR,A) splitVectorsBuffer((TYPE)0,INR,A)


Inrimage **Inrimage::splitVectors(void) const {
  Inrimage **inrs;
  unsigned int i;

  inrs = new Inrimage*[vdim];

  switch(type) {

  case WT_RGB:
  case WT_RGBA:
    for(i = 0; i < vdim; i++)
      inrs[i] = new Inrimage(xdim, ydim, zdim, WT_UNSIGNED_CHAR);
    SPLITVECTORS(unsigned char,*this,inrs);
    break;

  case WT_FLOAT_VECTOR:
    for(i = 0; i < vdim; i++)
      inrs[i] = new Inrimage(xdim, ydim, zdim, WT_FLOAT);
    SPLITVECTORS(float,*this,inrs);
    break;

  case WT_DOUBLE_VECTOR:
    for(i = 0; i < vdim; i++)
      inrs[i] = new Inrimage(xdim, ydim, zdim, WT_DOUBLE);
    SPLITVECTORS(double,*this,inrs);
    break;

  default:
    inrs[0] = new Inrimage(*this);
    break;
  }

  for(i = 0; i < vdim; i++) inrs[i]->setVoxelSize(vx,vy,vz);
  return inrs;
}


void Inrimage::line3D(Vec3<double> p1, Vec3<double> p2,
		    const Voxel& v,bool forceProjection) 
{
	if (forceProjection) {
		if (p1[0]<0) {
			p1[0]=0;
		} 
		if (p1[0]>=getX()) {
			p1[0]=getX()-1;
		}
		if (p1[1]<0) {
			p1[1]=0;
		}
		if (p1[1]>=getY()) {
			p1[1]=getY()-1;
		}
		if (p1[2]<0) {
			p1[2]=0;
		} 
		if (p1[2]>=getZ()) {
			p1[2]=getZ()-1;
		}
		if (p2[0]<0) {
			p2[0]=0;
		} 
		if (p2[0]>=getX()) {
			p2[0]=getX()-1;
		}
		if (p2[1]<0) {
			p2[1]=0;
		}
		if (p2[1]>=getY()) {
			p2[1]=getY()-1;
		}
		if (p2[2]<0) {
			p2[2]=0;
		} 
		if (p2[2]>=getZ()) {
			p2[2]=getZ()-1;
		}
	}

	Vec3<double> pos(p1);
	Vec3<double> direction = p2-p1;
	double maxDist=direction.norm();
	Vec3<double> dp=direction/maxDist;

	ImageGeometry *converter = newConverter();
	converter->set(pos.getX(), pos.getY(), pos.getZ());


	if(dp.getX() == 0.0) dp[0] = ImageGeometry::epsilon;
	if(dp.getY() == 0.0) dp[1] = ImageGeometry::epsilon;
	if(dp.getZ() == 0.0) dp[2] = ImageGeometry::epsilon;

	double dist = 0.0;
	int vpos[3];
	double a;
	converter->toImage(vpos);
	if (converter->in()==true) {
		setVoxel(vpos[0],vpos[1],vpos[2],v);
	}

	while(dist<maxDist && (converter->in()==true)) {
		a = converter->nextVoxel(dp);
		pos += a * dp;
		dist += a;
		converter->toImage(vpos);
		if (converter->in()) {
			setVoxel(vpos[0],vpos[1],vpos[2],v);
		}
	}

	delete converter;

}

void Inrimage::line(int x1, int y1, int x2, int y2, unsigned int z,
		    const Voxel& v) {
  int i, j, k, ok, my, My, mx, Mx, mk, Mk;
  double r;
    
  if(x1 == x2) {
    if(x2 >= 0 && x2 < (int) xdim) {
      if(y1 >= y2) {
	my = y2;
	My = y1;
      }
      else {
	my = y1;
	My = y2;
      }
      if(my < 0) my = 0;
      if(My >= (int) ydim) My = (int) ydim - 1;
      for(i = my; i <= My; i++) setVoxel(x2, i, z, v);
    }
  }

  else {
    if(x2 < x1) {
      r = (double) (y1 - y2) / (double) (x1 - x2);
      ok = y2;
      if(x2 + 1 < 0) mx = 0;
      else mx = x2 + 1;
      if(x1 >= (int) xdim) Mx = (int) xdim-1;
      else Mx = x1;
      for(i = mx; i <= Mx; i++) {
	k = (int) ((double) (i - x2) * r + (double) y2 + 0.5);
	if(ok >= k) {
	  if(k < 0) mk = 0;
	  else mk = k;
	  if(ok >= (int) ydim) Mk = (int) ydim - 1;
	  else Mk = ok;
	  for(j = mk; j <= Mk; j++) setVoxel(i, j, z, v);
	}
	else {
	  if(ok < 0) mk = 0;
	  else mk = ok;
	  if(k >= (int) ydim) Mk = (int) ydim - 1;
	  else Mk = k;
	  for(j = mk; j <= Mk; j++) setVoxel(i, j, z, v);
	}
	ok = k;
      }
    }
    else {
      r = (double) (y2 - y1) / (double) (x2 - x1);
      ok = y1;
      if(x1 + 1 < 0) mx = 0;
      else mx = x1 + 1;
      if(x2 >= (int) xdim) Mx = (int) xdim - 1;
      else Mx = x2;
      for(i = mx; i <= Mx; i++) {
	k = (int) ((double) (i - x1) * r + (double) y1 + 0.5);
	if(ok >= k) {
	  if(k < 0) mk = 0;
	  else mk = k;
	  if(ok >= (int) ydim) Mk = (int) ydim - 1;
	  else Mk = ok;
	  for(j = mk; j <= Mk; j++) setVoxel(i, j, z, v);
	}
	else {
	  if(ok < 0) mk = 0;
	  else mk = ok;
	  if(k >= (int) ydim) Mk = (int) ydim - 1;
	  else Mk = k;
	  for(j = mk; j <= Mk; j++) setVoxel(i, j, z, v);
	}
	ok = k;
      }
    }
  }
  
}


ImageGeometry *Inrimage::newConverter(void) const {
  switch(geom) {
  case CARTESIAN:
    return new CartesianGeometry(*this);
  case CYLINDRICAL:
    return new CylindricalGeometry(*this);
  case SPHERICAL:
    return new SphericalGeometry(*this);
  }
  return 0;
}

GenericHistogram *Inrimage::createHistogram() const {
  unsigned long int *b;
  double minVal,maxVal;
  unsigned int bsize;
  GenericHistogram *h;
  switch(type) {
  case WT_UNSIGNED_CHAR:
    buildHistogram((unsigned char)0,this,b,bsize,minVal,maxVal); 
    h= new GenericHistogram(b,bsize,minVal,maxVal);
    delete [] b;
    break;
  case WT_SIGNED_SHORT:
    buildHistogram((short int)0,this,b,bsize,minVal,maxVal); 
    h= new GenericHistogram(b,bsize,minVal,maxVal);
    delete [] b;
    break;
  case WT_UNSIGNED_SHORT:
    buildHistogram((unsigned short int)0,this,b,bsize,minVal,maxVal); 
    h= new GenericHistogram(b,bsize,minVal,maxVal);
    delete [] b;
    break;
  default:
    {
      Voxel *v = Voxel::voxelByType(getType());
      double key;
      std::map< double, unsigned long int> histo;
      std::map< double, unsigned long int>::iterator it;

      for(unsigned int k = 0; k < getZ(); k++) {
	for(unsigned int j = 0; j < getY(); j++) {
	  for(unsigned int i = 0; i < getX(); i++) {
	    getVoxel(i, j, k, *v);
	    key=v->getValueAsDouble();
	    it=histo.find(key);
	    if (it==histo.end())
	      {
		histo[key]=1;
	      }
	    else
	      {	
		histo[key]=it->second+1;
	      }
	  }
	}
      }
      h= new GenericHistogram(histo);
      histo.erase(histo.begin(),histo.end());
    }
  }
  return h;
  
}

GenericHistogram *Inrimage::createMaskedHistogram( const Inrimage* mask ) const {
 // unsigned long int *b;
  //double minVal,maxVal;
  //unsigned int bsize;
  GenericHistogram *h;
  Voxel *v = Voxel::voxelByType(getType());
  Voxel *m = Voxel::voxelByType( mask->getType() );
  long   roi;
  double key;
  std::map< double, unsigned long int> histo;
  std::map< double, unsigned long int>::iterator it;

  for(unsigned int k = 0; k < getZ(); k++) {
    for(unsigned int j = 0; j < getY(); j++) {
      for(unsigned int i = 0; i < getX(); i++) {
	getVoxel(i, j, k, *v);
	key=v->getValueAsDouble();
	mask->getVoxel(i, j, k, *m);
	roi=m->getValueAsLong();
	if ( roi != 0 )
	{
	  it=histo.find(key);
	  if (it==histo.end())
	  {
	    histo[key]=1;
	  }
	  else
	  {	
	    histo[key]=it->second+1;
	  }
	}
      }
    }
  }
  h= new GenericHistogram(histo);
  histo.erase(histo.begin(),histo.end());

  return h;  
}


VoxelIterator Inrimage::getVoxelIterator() {
  VoxelIteratorObject *vi;
  switch(type){
  case WT_UNSIGNED_CHAR:
    vi = (VoxelIteratorObject *) new VoxelIteratorScalar<unsigned char>(this);
    break;
  case WT_SIGNED_CHAR:
    vi = new VoxelIteratorScalar<char>(this);
    break;
  case WT_SIGNED_SHORT:
    vi = new VoxelIteratorScalar<short int>(this);
    break;
  case WT_UNSIGNED_SHORT:
    vi = new VoxelIteratorScalar<unsigned short int>(this);
    break;
  case WT_SIGNED_INT:
    vi = new VoxelIteratorScalar<int>(this);
    break;
  case WT_UNSIGNED_INT:
    vi = new VoxelIteratorScalar<int>(this);
    break;
  case WT_SIGNED_LONG:
    vi = new VoxelIteratorScalar<long int>(this);
    break;
  case WT_UNSIGNED_LONG:
    vi = new VoxelIteratorScalar<unsigned long int>(this);
    break;
  case WT_FLOAT:
  case WT_FLOAT_VECTOR:
    vi = new VoxelIteratorScalar<float>(this);
    break;
  case WT_DOUBLE:
  case WT_DOUBLE_VECTOR:
    vi = new VoxelIteratorScalar<double>(this);
    break;
  default:
    vi = new VoxelIteratorScalar<unsigned char>(this);
    break;
  }
  return (*vi);
}
}
