
/*
 * Threholding procedures with one or two thresholds
   one threshold : background is defined by I(x) <  min
                   foreground is defined by I(x) >= min
   two threshold : background is defined by I(x) <  min || I(x) >  max
                   foreground is defined by I(x) >= min && I(x) <= max
 
 * if the max voxel is NULL, perform a simple thresholding
 */

template <class TYPE, class V> void thresholdBuffer( TYPE *theBuf, 
						     const unsigned int size,
						     V *vmin, V *vmax, 
						     V *vbg, V *vfg )
{
  unsigned int i;
  TYPE *data = theBuf;
  TYPE min = vmin->getValue();
  TYPE max;
  TYPE bg = vbg->getValue();
  TYPE fg = vfg->getValue();

  if ( vmax != NULL ) {
     max = vmax->getValue();
     for ( i=0; i<size; i++, data++ ) 
       	*data = ( *data < min || *data > max ) ? bg : fg;
  }
  else {
      for ( i=0; i<size; i++, data++ ) 
       	*data = ( *data < min ) ? bg : fg;
  }
}





template <class TYPE1, class TYPE2, class V1> void thresholdBufferToBuffer( TYPE1 *theBuf, 
						     TYPE2 *resBuf, const unsigned int size,
						     V1 *vmin, V1 *vmax, 
						     const TYPE2 bg, const TYPE2 fg )
{
  unsigned int i;
  TYPE1 *data = theBuf;
  TYPE2 *res = resBuf;
  TYPE1 min = vmin->getValue();
  TYPE1 max;

  if ( vmax != NULL ) {
     max = vmax->getValue();
     for ( i=0; i<size; i++, data++, res++ ) 
       	*res = ( *data < min || *data > max ) ? bg : fg;
  }
  else {
      for ( i=0; i<size; i++, data++, res++ ) 
       	*res = ( *data < min ) ? bg : fg;
  }
}







template <class TYPE2, class V2> void thresholdImageToBuffer( TYPE2 *T,
							     Inrimage *inr,
							     Inrimage::WORD_TYPE ntype,
							     Voxel *min,
							     Voxel *max,
							     V2 *vbg2,
							     V2 *vfg2 )
  throw( ErrTypeMismatch )
{
  char *from = "thresholdImageToBuffer( TYPE2*, Inrimage*, Inrimage::WORD_TYPE, Voxel*, Voxel*, V2*, V2* )";
  TYPE2 bg2 = vbg2->getValue();
  TYPE2 fg2 = vfg2->getValue();
  unsigned int size = inr->getX() * inr->getY() * inr->getZ() * inr->getVdim();
  TYPE2 *resBuf;
  
  resBuf = new TYPE2 [size];

  switch( inr->getType() ) {
  default :
    delete[] resBuf;
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    thresholdBufferToBuffer( (unsigned char *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel8u *>(min), dynamic_cast<Voxel8u *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    thresholdBufferToBuffer( (char *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel8s *>(min), dynamic_cast<Voxel8s *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    thresholdBufferToBuffer( (unsigned short int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel16u *>(min), dynamic_cast<Voxel16u *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    thresholdBufferToBuffer( (short int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel16s *>(min), dynamic_cast<Voxel16s *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    thresholdBufferToBuffer( (unsigned int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel32u *>(min), dynamic_cast<Voxel32u *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_INT:
    thresholdBufferToBuffer( (int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel32s *>(min), dynamic_cast<Voxel32s *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    thresholdBufferToBuffer( (unsigned LONGINT *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel64u *>(min), dynamic_cast<Voxel64u *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_LONG:
    thresholdBufferToBuffer( (LONGINT *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel64s *>(min), dynamic_cast<Voxel64s *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_FLOAT:
     thresholdBufferToBuffer( (float *)inr->getData(), resBuf, size, 
			    dynamic_cast<VoxelFloat *>(min), dynamic_cast<VoxelFloat *>(max), bg2, fg2 );
    break;
  case Inrimage::WT_DOUBLE:
    thresholdBufferToBuffer( (double *)inr->getData(), resBuf, size, 
			    dynamic_cast<VoxelDouble *>(min), dynamic_cast<VoxelDouble *>(max), bg2, fg2 );
    break;
  }

  inr->setType( ntype, inr->getVdim(), inr->getVectorialMode(), false );
  delete[] (unsigned char *) inr->setData( resBuf );
}













/*
 * "Binarize" procedures
   the background remains the background
   while the other values are changed into foreground
 *
 */

/* one buffer is binarized 
 */
template <class TYPE, class V> void binarizeBuffer( TYPE *theBuf,
						    const unsigned int size,
						     V *vbg1, V *vbg2, V *vfg2 )
{
  unsigned int i;
  TYPE *data = theBuf;
  TYPE bg1 = vbg1->getValue();
  TYPE bg2 = vbg2->getValue();
  TYPE fg  = vfg2->getValue();

  if ( bg1 == bg2 ) {
    for ( i=0; i<size; i++, data++ ) {
      if ( *data != bg1 ) *data = fg;
    }
  }
  else {
    for ( i=0; i<size; i++, data++ ) {
      *data = ( *data == bg1 ) ? bg2 : fg;
    }
  }
}




/* one buffer is binarized into an other one 
 */

template <class TYPE1, class TYPE2, class V1> void binarizeBufferToBuffer( TYPE1 *theBuf,
									   TYPE2 *resBuf,
									   const unsigned int size,
									   V1 *vbg1,
									   const TYPE2 bg2,
									   const TYPE2 fg2 )
{
  unsigned int i;
  TYPE1 *data = theBuf;
  TYPE2 *res  = resBuf;
  TYPE1 bg1 = vbg1->getValue();
  for ( i=0; i<size; i++, data++, res++ ) {
    *res = ( *data == bg1 ) ? bg2 : fg2;
  }
}




/* one image is binarized into a buffer
 */

template <class TYPE2, class V2> void binarizeImageToBuffer( TYPE2 *T,
							     Inrimage *inr,
							     Inrimage::WORD_TYPE ntype,
							     Voxel *vbg1,
							     V2 *vbg2,
							     V2 *vfg2 )
  throw( ErrTypeMismatch )
{
  char *from = "binarizeImageToBuffer( TYPE2*, Inrimage*, Inrimage::WORD_TYPE, Voxel*, V2*, V2* )";
  TYPE2 bg2 = vbg2->getValue();
  TYPE2 fg2 = vfg2->getValue();
  unsigned int size = inr->getX() * inr->getY() * inr->getZ() * inr->getVdim();
  TYPE2 *resBuf;
  
  resBuf = new TYPE2 [size];

  switch( inr->getType() ) {
  default :
    delete[] resBuf;
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    binarizeBufferToBuffer( (unsigned char *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel8u *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    binarizeBufferToBuffer( (char *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel8s *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    binarizeBufferToBuffer( (unsigned short int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel16u *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    binarizeBufferToBuffer( (short int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel16s *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    binarizeBufferToBuffer( (unsigned int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel32u *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_INT:
    binarizeBufferToBuffer( (int *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel32s *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    binarizeBufferToBuffer( (unsigned LONGINT *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel64u *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_SIGNED_LONG:
    binarizeBufferToBuffer( (LONGINT *)inr->getData(), resBuf, size, 
			    dynamic_cast<Voxel64s *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_FLOAT:
     binarizeBufferToBuffer( (float *)inr->getData(), resBuf, size, 
			    dynamic_cast<VoxelFloat *>(vbg1), bg2, fg2 );
    break;
  case Inrimage::WT_DOUBLE:
    binarizeBufferToBuffer( (double *)inr->getData(), resBuf, size, 
			    dynamic_cast<VoxelDouble *>(vbg1), bg2, fg2 );
    break;
  }

  inr->setType( ntype, inr->getVdim(), inr->getVectorialMode(), false );
  delete[] (unsigned char *) inr->setData( resBuf );
}










/*
 * "Ternarize" procedures
   It is nothing but a double thresholding
   points with value strictly lower than the low threshold are set to value_1
   points with value larger or equal than the low threshold but 
               strictly lower than the high threshold are set to value_2
   points with value larger or equal than th3 high threshold are set to value_3
 *
 */

/* one buffer is ternarized
 */
template <class TYPE, class V> void ternarizeBuffer( TYPE *theBuf,
						     const unsigned int size,
						     V *vlow, V *vhigh,
						     V *v1, V *v2, V *v3 )
{
  unsigned int i;
  TYPE *data = theBuf;
  TYPE low  = vlow->getValue();
  TYPE high = vhigh->getValue();
  TYPE val1 = v1->getValue();
  TYPE val2 = v2->getValue();
  TYPE val3 = v3->getValue();

  for ( i=0; i<size; i++, data++ ) {
    if ( *data >= high ) {
      *data = val3;
    }
    else if ( *data >= low ) {
      *data = val2;
    }
    else {
      *data = val1;
    }
  }
}




/* one buffer is ternarized into an other one 
 */

template <class TYPE1, class TYPE2, class V1> void ternarizeBufferToBuffer( TYPE1 *theBuf,
								      TYPE2 *resBuf,
								      const unsigned int size,
								      V1 *vlow,
								      V1 *vhigh,
								      const TYPE2 val1,
								      const TYPE2 val2,
								      const TYPE2 val3 )
{
  unsigned int i;
  TYPE1 *data = theBuf;
  TYPE2 *res  = resBuf;
  TYPE1 low  = vlow->getValue();
  TYPE1 high = vhigh->getValue();
  for ( i=0; i<size; i++, data++, res++ ) {
    if ( *data >= high ) {
      *res = val3;
    }
    else if ( *data >= low ) {
      *res = val2;
    }
    else {
      *res = val1;
    }
  }
}




/* one image is ternarized into a buffer
 */

template <class TYPE2, class V2> void ternarizeImageToBuffer( TYPE2 *T,
							      Inrimage *inr,
							      Inrimage::WORD_TYPE ntype,
							      Voxel *vlow,
							      Voxel *vhigh,
							      V2 *v1, V2 *v2, V2 *v3 )
  throw( ErrTypeMismatch )
{
  char *from = "ternarizeImageToBuffer( TYPE2*, Inrimage*, Inrimage::WORD_TYPE, Voxel*, Voxel*, V2*, V2*, V2* )";
  TYPE2 val1 = v1->getValue();
  TYPE2 val2 = v2->getValue();
  TYPE2 val3 = v3->getValue();
  unsigned int size = inr->getX() * inr->getY() * inr->getZ() * inr->getVdim();
  TYPE2 *resBuf;
  
  resBuf = new TYPE2 [size];

  switch( inr->getType() ) {
  default :
    delete[] resBuf;
    throw ErrTypeMismatch( from );
    return;
  case Inrimage::WT_UNSIGNED_CHAR:
    ternarizeBufferToBuffer( (unsigned char *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel8u *>(vlow), dynamic_cast<Voxel8u *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_SIGNED_CHAR:
    ternarizeBufferToBuffer( (char *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel8s *>(vlow), dynamic_cast<Voxel8s *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_UNSIGNED_SHORT:
    ternarizeBufferToBuffer( (unsigned short int *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel16u *>(vlow), dynamic_cast<Voxel16u *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_SIGNED_SHORT:
    ternarizeBufferToBuffer( (short int *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel16s *>(vlow), dynamic_cast<Voxel16s *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_UNSIGNED_INT:
    ternarizeBufferToBuffer( (unsigned int *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel32u *>(vlow), dynamic_cast<Voxel32u *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_SIGNED_INT:
    ternarizeBufferToBuffer( (int *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel32s *>(vlow), dynamic_cast<Voxel32s *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_UNSIGNED_LONG:
    ternarizeBufferToBuffer( (unsigned LONGINT *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel64u *>(vlow), dynamic_cast<Voxel64u *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_SIGNED_LONG:
    ternarizeBufferToBuffer( (LONGINT *)inr->getData(), resBuf, size, 
			     dynamic_cast<Voxel64s *>(vlow), dynamic_cast<Voxel64s *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_FLOAT:
    ternarizeBufferToBuffer( (float *)inr->getData(), resBuf, size, 
			     dynamic_cast<VoxelFloat *>(vlow), 
			     dynamic_cast<VoxelFloat *>(vhigh), 
			     val1, val2, val3 );
    break;
  case Inrimage::WT_DOUBLE:
    ternarizeBufferToBuffer( (double *)inr->getData(), resBuf, size, 
			     dynamic_cast<VoxelDouble *>(vlow), 
			     dynamic_cast<VoxelDouble *>(vhigh), 
			     val1, val2, val3 );
    break;
  }

  inr->setType( ntype, inr->getVdim(), inr->getVectorialMode(), false );
  delete[] (unsigned char *) inr->setData( resBuf );
}
