#include <inrimage/anisotropicDiffusion.h>
#include <inrimage/recursiveFilters.h>

namespace yav {
#ifdef USE_MTL
void anisotropicDiffusion(const Inrimage4D *inr, Inrimage4D *& output, const float & dt, const float & tmax, const float & q, const bool smooth, 
			  const float & sigma, 
			  const DIFFUSION_FUNCTION & p, 
			  const bool disk) {
   
  unsigned int nbiter = (unsigned int) floor(tmax/dt+0.5), w, x, y, z, t;
  Inrimage * extrait = inr->getFirstInrimage();
  unsigned int xdim=extrait->getX(), ydim=extrait->getY(), zdim=extrait->getZ(), wdim=inr->getTimeDimension();
  double vx = 1.0/extrait->getX(), vy = 1.0/extrait->getVY(), vz = 1.0/extrait->getVZ();
  
  Inrimage::WORD_TYPE wt = inr->getType();
  if(wt != Inrimage::WT_FLOAT) {
    for(w=0;w<wdim;w++) {
      output->getInrimage(w)->convertCast(Inrimage::WT_FLOAT);
    }
  }
  
  float **** buf = (float ****) output->getArray();
  unsigned int ms = xdim * ydim * zdim * wdim;
  float k;
  float ts = sigma*sigma/2;
  MTLVector u1(ms), du(ms);
  MTLVecIt i;

  if(!disk) {
    MTLVector u2(ms), u3(ms), u4(ms);
    
    for(t=0;t<nbiter;t++) {
      CERR << " " << (unsigned int) ((float) t / (tmax/dt) * 100.0) << " %" << "\r";
      
      i=u1.begin();
      for (w=0; w<wdim; w++) {
	for (z=0; z<zdim; z++) {
	  for (y=0; y<ydim; y++) {
	    for (x=0; x<xdim; x++) {
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  }
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u1, ts, 4.0);
	derive(du,us,vx);
      }
      else derive(du,u1,vx); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u1, du, dt, 4.0, xdim);
      
      i=u2.begin();    
      for (w=0; w<wdim; w++) {
	for (z=0; z<zdim; z++) {
	  for (x=0; x<xdim; x++) {
	    for (y=0; y<ydim; y++) {
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  }
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u2, ts, 4.0);
	derive(du,us,vy);
      }
      else derive(du,u2,vy); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u2, du, dt, 4.0, ydim);
      
      i=u3.begin();
      for (w=0; w<wdim; w++) {
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {
	    for (z=0; z<zdim; z++) {	
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  } 
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u3, ts, 4.0);
	derive(du,us,vz);
      }
      else derive(du,u3,vz); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u3, du, dt, 4.0, zdim);
      
      i=u4.begin();
      for (z=0; z<zdim; z++) {
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {	  
	    for (w=0; w<wdim; w++) {	
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  } 
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u4, ts, 4.0);
	derive(du,us);
      }
      else derive(du,u4); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u4, du, dt, 4.0, wdim);
      
      i=u1.begin();    
      for (w=0; w<wdim; w++) {
	for (z=0; z<zdim; z++) {
	  for (y=0; y<ydim; y++) {
	    for (x=0; x<xdim; x++) {
	      buf[w][z][y][x] = *i / 4.0;
	      ++i;
	    }    
	  }
	}
      }
      
      i=u2.begin();
      for (w=0; w<wdim; w++) { 
	for (z=0; z<zdim; z++) {
	  for (x=0; x<xdim; x++) {
	    for (y=0; y<ydim; y++) {  
	      buf[w][z][y][x] += *i / 4.0;
	      ++i;
	    }
	  }
	}
      }

      i=u3.begin();
      for (w=0; w<wdim; w++) { 
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {
	    for (z=0; z<zdim; z++) {	
	      buf[w][z][y][x] += *i / 4.0;
	      ++i;
	    }
	  } 
	}
      }
      
      i=u4.begin();
      for (z=0; z<zdim; z++) {
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {
	    for (w=0; w<wdim; w++) { 	
	      buf[w][z][y][x] += *i / 4.0;
	      ++i;
	    }
	  } 
	}
      }
    }
  }
  
  else {
    for(t=0;t<nbiter;t++) {
      CERR << " " << (unsigned int) ((float) t / (tmax/dt) * 100.0) << " %" << "\r";
      
      i=u1.begin();
      for (w=0; w<wdim; w++) {
	for (z=0; z<zdim; z++) {
	  for (y=0; y<ydim; y++) {
	    for (x=0; x<xdim; x++) {
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  }
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u1, ts, 4.0);
	derive(du,us,vx);
      }
      else derive(du,u1,vx); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u1, du, dt, 4.0, xdim);
      writeMTLVector("anisodiff.tmp1~",u1);
      
      i=u1.begin();    
      for (w=0; w<wdim; w++) {
	for (z=0; z<zdim; z++) {
	  for (x=0; x<xdim; x++) {
	    for (y=0; y<ydim; y++) {
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  }
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u1, ts, 4.0);
	derive(du,us,vy);
      }
      else derive(du,u1,vy); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u1, du, dt, 4.0, ydim);
      writeMTLVector("anisodiff.tmp2~",u1);
            
      i=u1.begin();
      for (w=0; w<wdim; w++) {
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {
	    for (z=0; z<zdim; z++) {	
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  } 
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u1, ts, 4.0);
	derive(du,us,vz);
      }
      else derive(du,u1,vz); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u1, du, dt, 4.0, zdim);
      writeMTLVector("anisodiff.tmp3~",u1);
      
      i=u1.begin();
      for (z=0; z<zdim; z++) {
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {	  
	    for (w=0; w<wdim; w++) {	
	      *i = buf[w][z][y][x];
	      ++i;
	    }
	  } 
	}
      }
      if(smooth) {
	MTLVector us(ms);
	AOSGauss(us, u1, ts, 4.0);
	derive(du,us);
      }
      else derive(du,u1); 
      k = quantile(du, q);
      diffusivity(du, k, p);
      AOS(u1, du, dt, 4.0, wdim);
      
      i=u1.begin();
      for (z=0; z<zdim; z++) {
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {
	    for (w=0; w<wdim; w++) { 	
	      buf[w][z][y][x] = *i / 4.0;
	      ++i;
	    }
	  } 
	}
      }
      
      readMTLVector("anisodiff.tmp1~",u1);
      remove("anisodiff.tmp1~");
      i=u1.begin();    
      for (w=0; w<wdim; w++) {
	for (z=0; z<zdim; z++) {
	  for (y=0; y<ydim; y++) {
	    for (x=0; x<xdim; x++) {
	      buf[w][z][y][x] += *i / 4.0;
	      ++i;
	    }    
	  }
	}
      }

      readMTLVector("anisodiff.tmp2~",u1);
      remove("anisodiff.tmp2~");
      i=u1.begin();
      for (w=0; w<wdim; w++) { 
	for (z=0; z<zdim; z++) {
	  for (x=0; x<xdim; x++) {
	    for (y=0; y<ydim; y++) {  
	      buf[w][z][y][x] += *i / 4.0;
	      ++i;
	    }
	  }
	}
      }

      readMTLVector("anisodiff.tmp3~",u1);
      remove("anisodiff.tmp3~");
      i=u1.begin();
      for (w=0; w<wdim; w++) { 
	for (x=0; x<xdim; x++) {
	  for (y=0; y<ydim; y++) {
	    for (z=0; z<zdim; z++) {	
	      buf[w][z][y][x] += *i / 4.0;
	      ++i;
	    }
	  } 
	}
      }
    }
  }
  
  if (wt != Inrimage::WT_FLOAT) {
    for (w=0; w<wdim; w++) {
      output->getInrimage(w)->convertCast(wt);
    }
  }
  CERR << "Fini." << ENDL;    
}
#endif
} // namespace yav
