#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#include <string.h>

#include "fld.h"




/** fld - AVS field  file format */
/** Magic header for fld */
#define FLD_MAGIC "# AV"




int testFldHeader(char *magic,const char *name) {

  if( !strncmp(magic, FLD_MAGIC, strlen(FLD_MAGIC))) 
    return 0;
  else 
    return -1;
}

PTRIMAGE_FORMAT createFldFormat() {
  PTRIMAGE_FORMAT f=(PTRIMAGE_FORMAT) ImageIO_alloc(sizeof(IMAGE_FORMAT));

  f->testImageFormat=&testFldHeader;
  f->readImageHeader=&readFldImage;
  f->writeImage=&writeFldImage;
  strcpy(f->fileExtension,".fld");
  strcpy(f->realName,"Fld");
  return f;
}


/* get a string from a file and discard the ending newline character
   if any */
static char *fgetns(char *str, int n, _image *im ) 
{
  char *ret;
  int l;

  ret = ImageIO_gets( im, str, n );
  
  if(!ret) return NULL;
  l = strlen(str);
  if(l > 0 && str[l-1] == '\n') str[l-1] = '\0';
  return ret;
}


/*
       The  AVS field  format is a simple file format with an ASCII header
       color image file format.

       See http://astronomy.swin.edu.au/~pbourke/geomformats/field/ for a more complete description

       The definition is as follows:

       - A  "magic  number" for identifying the file type is "# AVS" 

       - Description of the image dimension with the ndim keyword

       - Description of the image size with the dim1,dim2 and dim3 keywords

       - Description of the voxel size with the x scale,y scale and z scale keywords

       - Description of the embedding space dimension (must be 3)

       - Description of the type of word with the keyword data (for instance byte)

       - Description of the number of vector compoenents with the keyword veclen (1 for scalar, 3 for 3D vector )

       - Description of the image geometry with the keyword field  (uniform for cartesian geometry)

       - comments are written after the # character

       The end of the header is indicated with the sequence of 3 bytes 0x0a 0x0c 0x0c */
int readFldImage(const char *name,_image *im)
{
  char string[256];
  char end=0;
  char buffer[4];
  
  fgetns( string, 255, im );
  if ( strncmp(string, FLD_MAGIC, strlen(FLD_MAGIC) ) ) {
    fprintf( stderr, "readFldImage: bad magic string in \'%s\'\n", name );
    return( -1 );
  }
  
  do {
    fgetns( string, 255, im );
    
    if ((string[0]==12) && (string[1]==12))
      end=1;
    else {
      if (!strncmp(string,"ndim",4)) {
	unsigned int ndim;
	sscanf(string,"ndim=%ud",&ndim);
	if (ndim!=3) {
	  fprintf( stderr, "readFldImage: wrong dimension  \'%ud\'\n", ndim );
	  return( -1 );
	}
      }
    else if  (!strncmp(string,"dim1",4)) {
      sscanf(string,"dim1=%ud",  &im->xdim );
    }
    else if  (!strncmp(string,"dim2",4)) {
      sscanf(string,"dim2=%ud",  &im->ydim );
    }
    else if  (!strncmp(string,"dim3",4)) {
      sscanf(string,"dim3=%ud",  &im->zdim );
    }
    else if  (!strncmp(string,"x scale",7)) {
      sscanf(string,"x scale = %lf",  &im->vx );
    }
    else if  (!strncmp(string,"y scale",7)) {
      sscanf(string,"y scale = %lf",  &im->vy );
    }
    else if  (!strncmp(string,"z scale",7)) {
      sscanf(string,"z scale = %lf",  &im->vz );
    }
    else if (!strncmp(string,"nspace",6)) {
      unsigned int nspace;
      sscanf(string,"nspace=%u",&nspace);
      if (nspace!=3) {
	fprintf( stderr, "readFldImage: wrong space   \'%u\'\n", nspace );
	return( -1 );
      }
    }
    else if  (!strncmp(string,"veclen",6)) {
      sscanf(string,"veclen=%u",  &im->vdim );
    }
    else if  (!strncmp(string,"data",4)) {
      char data[255];
      sscanf(string,"data=%s", data);
      if (!strcmp(data,"byte")) {
	im->wordKind = WK_FIXED;
	im->sign = SGN_UNSIGNED;
	im->wdim = 1;
      }
      else  {
	fprintf( stderr, "readFldImage: data type  \'%s\'\n", data );
	return( -1 );
      }

    }
    else if  (!strncmp(string,"field",5)) {
      char field[255];
      sscanf(string,"field=%s", field);
      /** decide to stop reading the header with the fgetn function */
      end=1;
      if (strcmp(field,"uniform")) {
	fprintf( stderr, "readFldImage: field type  \'%s\'\n", field );
	return( -1 );
      }

    }
    else  if ( string[0] != '#' ) {
      fprintf( stderr, "readFldImage: unknown keyword%s\'\n", string );
      return( -1 );
    }
    }
    
    


  } while ( end==0 );


  /* find the end of the buffer corresponding to the 0x0f 0x0f bytes */  
  ImageIO_read( im, buffer, 2);
  while ( (buffer[0]!=12) || (buffer[1]!=12) ) {
    buffer[0]=buffer[1];
    ImageIO_read( im, buffer+1, 1);
  }

  im->data = ImageIO_alloc( im->xdim*im->ydim*im->zdim );		     
  
  ImageIO_read( im, im->data, im->xdim*im->ydim*im->zdim  );
  
  return 1;
}


int writeFldImage( char *name,_image *im )
{
  char string[256];

  if ( im->xdim <= 0 || im->ydim <= 0 || im->zdim != 3 ) {
    fprintf( stderr, "writeFldImage: bad dimensions, unable to write '%s'\n", name );
    return -1;
  }
  if ( im->wordKind != WK_FIXED || im->sign != SGN_UNSIGNED 
       || ( im->wdim != 1  ) ) {
    fprintf( stderr, "writeFldImage: bad type, unable to write '%s'\n", name );
    return -1;  
  }
  
  _openWriteImage( im, name );
  
  if(!im->fd) {
    fprintf(stderr, "writeFldImage: error: unable to open file \'%s\'\n", name );
    return ImageIO_OPENING;
  }
  
  sprintf( string, "# AVS field file\n");
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "ndim=3         # number of dimension is the field\n" );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "dim1=%d         # dimension of axis 1\n",im->xdim );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "dim2=%d         # dimension of axis 2\n",im->ydim );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "dim3=%d         # dimension of axis 3\n",im->zdim );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "x scale = %f mm/voxel\n",im->vx );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "y scale = %f mm/voxel\n",im->vy );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "z scale = %f mm/voxel\n",im->vz );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "nspace=3                # number of physical coordinates per point\n" );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "veclen=%d                # number of components at each point\n",im->vdim );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "data=byte               # native format of sgi-irix\n" );
  ImageIO_write( im, string, strlen( string ) );
  sprintf( string, "field=uniform           # field type (uniform, rectilinear, irregular)" );
  ImageIO_write( im, string, strlen( string ) );
  string[0]=10;
  string[1]=12;
  string[2]=12;
  ImageIO_write( im, string, 3 );

  ImageIO_write( im, im->data, im->xdim*im->ydim*3*im->wdim );
  ImageIO_close( im );
  im->openMode = OM_CLOSE;

  return 1;
}

