#include <math.h>
#include <inrimage/fparser.h>
#include <inrimage/Inrimage.h>


namespace yav {

#define DEBUG


static const char* const Funcs[]=
{
    "abs","exp","ceil","floor","log","sqrt",
    "sin","cos","tan","asin","acos","atan",
    "sinh","cosh","tanh","asinh","acosh","atanh",
    0
};

static const char* const Globs[]=
{
  "mean", "min", "max", "sum", "med", "std", "var", 0
};

static const char* const Ops[]=
{
      "|<", "|>","<=", ">=","==", "<", ">", "&&","||", "+", "-", "*", "/", "^"
};

static const unsigned char
    cImmed= 1,
    cNeg  = 2,
    cFirstOp=3,
    cMin  = 3,
    cMax  = 4,
    cSmalT= 5,
    cLargT= 6,
    cLogic= 7,
    cSmal = 8,
    cLarg = 9,  
    cAnd  = 10,
    cOr   =11,
    cAdd  = 12,
    cSub  = 13,
    cMul  = 14,
    cDiv  = 15,
    cPow  = 16,
    cOps = 14,
 
    cAbs  = 17,
    cExp  = 18,
    cCeil = 19,
    cFloor= 20,
    cLog  = 21,
    cSqrt = 22,
    cSin  = 23,
    cCos  = 24,
    cTan  = 25,
    cAsin = 26,
    cAcos = 27,
    cAtan = 28,
    cSinh = 29,
    cCosh = 30,
    cTanh = 31,
    cAsinh= 32,
    cAcosh= 33,
    cAtanh= 34,
    VarBegin=35,
    gMean = 0,
    gMin  = 1,
    gMax  = 2,
    gSum  = 3,
    gMed  = 4,
    gStd  = 5,
    gVar  = 6,
    gNone =10;


int  VT_Compare(const void* e1, const void* e2 ) 
{ 

  if ( *((double*) e1 ) >  *((double*) e2 ) ) return 1; 
  if ( *((double*) e1 ) <  *((double*) e2 ) ) return -1; 
  return 0; 

}


// check if s1 is equal to s2 up to strlen(s1)
static inline bool IsEqual(const char* s1,const char* s2)
{ 
  int end;
  end=strcspn(s2,"+-*/^=()<>|&");
  if(end<1) return false; 
  for(int i=0;i<end;i++) if(s1[i]!=s2[i]) return false;
	return true;
}

// check if s1 is equal to s2 up to strlen(s1)
static inline bool IsEqualOp(const char* s1,const char* s2)
{ 
  int end;
  int i;
  end=strspn(s2,"+-*/^=<>|&");
  /* special case for negation operator */
  if(end>1)
    {
      if(s2[end-1]=='-')
	end=end-1;
    }
  if(end<1) return false;
  for(i=0;i<end;i++) if(s1[i]!=s2[i]) return false;
	if(i==(int)strlen(s1)) return true; else return false;
}


// Return operator number
static inline unsigned char OperatorNumber(char* f)
{
    for(int i=0;i<cOps;i++)
        if(IsEqualOp(Ops[i],f))
            return i+cFirstOp;
    return 0;
}



// Return math function number
static inline unsigned char MathFunctionNumber(char* f)
{
    for(int i=0;Funcs[i];i++)
        if(IsEqual(Funcs[i],f))
            return i+cAbs;
    return 0;
}



// Return math function number
static inline unsigned char MathGlobalNumber(char* f)
{
    for(int i=0;Globs[i];i++)
        if(IsEqual(Globs[i],f))
            return i;
    return gNone;
}


//Return variable number
static inline unsigned char VariableNumber(char* f,Var_list *v)
{	    
  Var_list *c_var;
  int i =0;
  c_var=v;
  while(c_var!=NULL){
    if(IsEqual(c_var->name,f)){
      return i+VarBegin;      
    }           
    c_var=c_var->next;
    i++;
  }	  
  return 0;	      
}

//---------------------------------------------------------------------------
// Constructors and destructors
//---------------------------------------------------------------------------
//===========================================================================
FunctionParser::FunctionParser():
    VarAmnt(0),ParseErrorType(-1),EvalErrorType(0),sx(0), sy(0), sz(0),Variable(0) 
{
}

FunctionParser::~FunctionParser()
{
}

FunctionParser::Comp::Comp():
    ByteCode(0),ByteCodeSize(0),Immed(0),ImmedSize(0),Stack(0),StackSize(0) { }

FunctionParser::Comp::~Comp()
{
    if(ByteCode) { delete[] ByteCode; ByteCode=0; }
    if(Immed) { delete[] Immed; Immed=0; }
    if(Stack) { delete[] Stack; Stack=0; }
}

//---------------------------------------------------------------------------
// Function parsing
//---------------------------------------------------------------------------
//===========================================================================
static const char* ParseErrorMessage[]=
{
    "Syntax error",                        // 0
    "Mismatched parenthesis",              // 1
    "Missing ')'",                         // 2
    "Empty parentheses",                   // 3
    "Syntax error. Operator expected",     // 4
    "Not enough memory"                    // 5
};

static void StrCpyRemovingSpaces(char* Dest,const char* Source);
static int RealIndexPos(const char* s,int i);

int FunctionParser::Parse(const char* Function,Var_list *Vars,int mx,int my, int mz)
{
    Var_list *c_var;
    c_var=Variable=Vars;
    VarAmnt=0;
    while(c_var!=NULL){
      VarAmnt++;
      c_var=c_var->next;
    }
    sx=mx;sy=my;sz=mz;

    char* Func=new char[strlen(Function)+1];
    StrCpyRemovingSpaces(Func,Function);

    int Result=CheckSyntax(Func);
    if(Result>=0) return RealIndexPos(Function,Result);

    if(!Compile(Func)) return strlen(Function)+1;
    delete(Func);

    ParseErrorType=-1;
    return -1;
}

// Syntax checking aux functions
// -----------------------------
// Copy string but without spaces
static void StrCpyRemovingSpaces(char* Dest,const char* Source)
{
    int i,j;
    for(i=j=0;Source[i];i++)
        if(!isspace(Source[i])) Dest[j++]=Source[i];
    Dest[j]=0;
}

// Return index to original string when the index is to the stripped version
static int RealIndexPos(const char* s,int Pos)
{
    int i,Ind=0;
    for(i=0;i<Pos;i++,Ind++)
        while(s[Ind]==' ') Ind++;
    while(s[Ind]==' ') Ind++;
    return Ind;
}

// Is given char an operator?
inline static bool IsOperator(char *f)
{
    for(int i=0;i<cOps;i++)
      if(IsEqualOp(Ops[i],f)){
	return true;
      }
    return false;

}

// Is given char a valid variable?
inline bool FunctionParser::IsVariable(int c)
{
  return 0;
}

//---------------------------------------------------------------------------
// Check function string syntax
// ----------------------------
int FunctionParser::CheckSyntax(char* Function)
{
    int Ind=0,ParenthCnt=0,c,i;
    char* Ptr;
    Var_list *c_var;

    while(true)
    {
        c=Function[Ind];

// Check for valid operand (must appear)

        // Check for leading -
        if(c=='-') c=Function[++Ind];
        if(c==0) { ParseErrorType=0; return Ind; }
        // Check for math function
        for(i=0;Funcs[i];i++)
        {
            if(IsEqual(Funcs[i],&Function[Ind]))
            {
                Ind+=strlen(Funcs[i]);
                c=Function[Ind];
                if(c!='(') { ParseErrorType=0; return Ind; }
                break;
            }
        }
        // Check for global function
        for(i=0;Globs[i];i++)
        {
            if(IsEqual(Globs[i],&Function[Ind]))
            {
                Ind+=strlen(Globs[i]);
                c=Function[Ind];
                if(c!='(') { 
		  ParseErrorType=0;
		  return Ind; }
                break;
            }
        }


        // Check for opening parenthesis
        if(c=='(') { ParenthCnt++; Ind++; continue; }
        // Check for number
        if(isdigit(c) || (c=='.' && isdigit(Function[Ind+1])))
        {
            strtod(&Function[Ind],&Ptr);
            Ind+=int(Ptr-&Function[Ind]);
            c=Function[Ind];
        }
        else
	  // Check for variable    
	  {
	    c_var=Variable;
	    while(c_var!=NULL){
	      if(IsEqual(c_var->name,&Function[Ind])){
                Ind+=strlen(c_var->name);
                c=Function[Ind];
		break;
	      }
	      c_var=c_var->next;
	    }
	    if(c_var==NULL){ParseErrorType=0; return Ind;}
	  }
        // Check for closing parenthesis
        while(c==')')
        {
            ParenthCnt--;
            if(ParenthCnt<0) { ParseErrorType=1; return Ind; }
            if(Function[Ind-1]=='(') { ParseErrorType=3; return Ind; }
            c=Function[++Ind];
        }

// If we get here, we have a legal operand and now a legal operator or
// end of string must follow

        // Check for EOS
        if(c==0) break; // The only way to end the checking loop without error
        // Check for operator
        for(i=0;i<cOps;i++)
        {
            if(IsEqualOp(Ops[i],&Function[Ind]))
            {
                Ind+=strlen(Ops[i]);
                c=Function[Ind];
                break;
            }
        }
	if(i==cOps) { ParseErrorType=4; return Ind; }
	

// If we get here, we have an operand and an operator; the next loop will
// check for another operand (must appear)
    } // while

    // Check that all opened parentheses are also closed
    if(ParenthCnt>0) { ParseErrorType=2; return Ind; }

// The string is ok
    ParseErrorType=-1;
    return -1;
}

// Compile function string to bytecode
// -----------------------------------
bool FunctionParser::Compile(char* Function)
{
    unsigned FLen=strlen(Function)-1;
 
    if(Comp.ByteCode) { delete[] Comp.ByteCode; Comp.ByteCode=0; }
    if(Comp.Immed) { delete[] Comp.Immed; Comp.Immed=0; }
    if(Comp.Stack) { delete[] Comp.Stack; Comp.Stack=0; }

// Compile to nowhere to get the size of the bytecode
    Comp.ByteCodeSize=Comp.ImmedSize=Comp.StackSize=Comp.StackPtr=0;
    CompileSubstr(Function,0,FLen);

    if(Comp.ByteCodeSize) Comp.ByteCode=new unsigned char[Comp.ByteCodeSize];
    if(Comp.ImmedSize) Comp.Immed=new double[Comp.ImmedSize];
    if(Comp.StackSize) Comp.Stack=new double[Comp.StackSize];
    if((Comp.ByteCodeSize && !Comp.ByteCode) ||
       (Comp.ImmedSize && !Comp.Immed) || 
       (Comp.StackSize && !Comp.Stack))
    {
        ParseErrorType=5; return false;
    }

// Compile
    Comp.ByteCodeSize=Comp.ImmedSize=Comp.StackSize=Comp.StackPtr=0;
    CompileSubstr(Function,0,FLen);

    return true;
}

void FunctionParser::AddCompiledByte(unsigned char c)
{
    if(Comp.ByteCode)
        Comp.ByteCode[Comp.ByteCodeSize] = c;
    Comp.ByteCodeSize++;
}

unsigned char FunctionParser::MathItemNumber(char* F)
{
    if(isdigit(F[0]) || F[0]=='.') // Number
    {
        if(Comp.Immed)
            Comp.Immed[Comp.ImmedSize]=atof(F);
        Comp.ImmedSize++;
        return cImmed;
    }
    return VariableNumber(F,Variable);
}

static bool CompletelyEnclosed(char* F,int b,int e)
{
    int i, cnt;

    if(F[b]=='(' && F[e]==')')
    {
        cnt=1;
        for(i=b+1; i<e; i++)
        {
            if(F[i]=='(') cnt++;
            else if(F[i]==')') cnt--;
            if(cnt==0) break;
        }
        if(i==e) return true;
    }
    return false;
}

void FunctionParser::CompileSubstr(char* F,int b,int e)
{
  if(CompletelyEnclosed(F, b, e))
  {	
	CompileSubstr(F, b+1, e-1);
	return;
  }
  
  if(F[b]=='-' && CompletelyEnclosed(F, b+1, e)){
    CompileSubstr(F, b+2, e-1);
    AddCompiledByte(cNeg);
    return;
  }

  if(isalpha(F[b])){
    unsigned char n=MathFunctionNumber(&F[b]);
    if(n>0){
      int b2=b;
      while(F[b2]!='(') b2++;
      if(CompletelyEnclosed(F, b2, e)){
	CompileSubstr(F, b2+1, e-1);
	AddCompiledByte(n);
	return;
      }
    }
    n=MathGlobalNumber(&F[b]);
    if(n!=gNone){
      int b2=b;
      double sub_res=1.0,tmp_res;
      FunctionParser Func;
      char func_str[512];
      double *values;
      int x,y,z,i;
      Var_list *w_op;
      double *list_tmp,*w_list_tmp;            
      Inrimage **im_ar=0;
      VoxelIteratorObject **vox=0;
	  
      while(F[b2]!='(') b2++;
      if(CompletelyEnclosed(F, b2, e)){
	if(Comp.Immed){
	  strcpy(func_str,&(F[b2+1]));
	  func_str[e-1-b2]='\0';
	  Func.Parse(func_str,Variable,sx,sy,sz);
	  values=new double[VarAmnt];
	  if(VarAmnt>3){ 
	    vox=new(VoxelIteratorObject*[VarAmnt-3]);
	    im_ar=new(Inrimage *[VarAmnt-3]);
	    w_op=Variable;
	    w_op=w_op->next; w_op=w_op->next; w_op=w_op->next;
	    for(i=0;i<VarAmnt-3;im_ar[i]=(Inrimage*)w_op->im,
		  vox[i] = &(((Inrimage *)w_op->im)->getVoxelIterator()),
		  i++,w_op=w_op->next);
	  }
	  if(n==gSum){
	    tmp_res=0.0;
	    for(i=0;i<VarAmnt-3;i++){ 
	      vox[i]->first();
	    }
	    for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
	      for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
		for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
		  for(i=0;i<VarAmnt-3;i++){ 
		    values[i+3] = vox[i]->operator*();
		    vox[i]->operator++();
		  }
		  tmp_res+=Func.Eval(values);
		}
	      }
	    }
	    sub_res=tmp_res;
	  }
	  if(n==gMean){
	    tmp_res=0.0; 
	    for(i=0;i<VarAmnt-3;i++){ 
	      vox[i]->first();
	    }
	    for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
	      for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
		for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
		  for(i=0;i<VarAmnt-3;i++){ 
		   values[i+3] = vox[i]->operator*();
		    vox[i]->operator++();
		  }
		  tmp_res+=Func.Eval(values);
		}
	      }
	    }
	    sub_res=tmp_res/(sx*sy*sz);
	  }
	  if(n==gMax){
	    sub_res=minTypeValue((double)0);
	    for(i=0;i<VarAmnt-3;i++){ 
	      vox[i]->first();
	    }
	    for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
	      for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
		for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
		  for(i=0;i<VarAmnt-3;i++){ 
		   values[i+3] = vox[i]->operator*();
		    vox[i]->operator++();
		  }
		  tmp_res=Func.Eval(values);
		  if(tmp_res>sub_res) sub_res=tmp_res;   
		}
	      }
	    }
	  }
	  if(n==gMin){
	    sub_res=maxTypeValue((double)0);;
	    for(i=0;i<VarAmnt-3;i++){ 
	      vox[i]->first();
	    }
	    for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
	      for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
		for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
		  for(i=0;i<VarAmnt-3;i++){ 
		   values[i+3] = vox[i]->operator*();
		    vox[i]->operator++();
		  }
		  tmp_res=Func.Eval(values);
		  if(tmp_res<sub_res) sub_res=tmp_res; 
		}
	      }
	    }
	  }
	  
	  if(n==gMed){
	    sub_res=maxTypeValue((double)0);;
	    list_tmp= new double[sx*sy*sz];
	    w_list_tmp=list_tmp;
	    for(i=0;i<VarAmnt-3;i++){ 
	      vox[i]->first();
	    }
	    for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
	      for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
		for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
		  for(i=0;i<VarAmnt-3;i++){ 
		    values[i+3] = vox[i]->operator*();
		    vox[i]->operator++();
		  }
		  *(w_list_tmp++)=Func.Eval(values);   
		}
	      }
	    }
	    qsort(list_tmp,long(sx)*sy*sz, sizeof(double), VT_Compare); 
	    sub_res=list_tmp[(int)((double(sx)*sy*sz) / 2.0 + 0.5)];
	    delete list_tmp;
	  }
	  if(n==gStd){
	    double x_=0.0,xx_=0.0;
	    double nr=double(sx)*double(sy)*double(sz);
	    for(i=0;i<VarAmnt-3;i++){ 
	      vox[i]->first();
	    }
	    for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
	      for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
		for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
		  for(i=0;i<VarAmnt-3;i++){ 
		    values[i+3] = vox[i]->operator*();
		    vox[i]->operator++();
		  }
		  tmp_res=Func.Eval(values);
		  x_+=tmp_res;
		  xx_+=(tmp_res*tmp_res);
		}
	      }
	    }
	    sub_res=sqrt((1.0/(nr-1.0))*(xx_-(x_*x_)/nr));
	  }
	  if(n==gVar){
	    double xs=0.0,xxs=0.0;
	    double nr=double(sx)*double(sy)*double(sz);
	    for(i=0;i<VarAmnt-3;i++){ 
	      vox[i]->first();
	    }
	    for(x=0,values[0]=(double)x;x<sx;x++,values[0]=(double)x){
	      for(y=0,values[1]=(double)y;y<sy;y++,values[1]=(double)y){
		for(z=0,values[2]=(double)z;z<sz;z++,values[2]=(double)z){ 
		  for(i=0;i<VarAmnt-3;i++){ 
		    values[i+3] = vox[i]->operator*();
		    vox[i]->operator++();
		  }
		  tmp_res=Func.Eval(values);
		  xs+=tmp_res;
		  xxs+=(tmp_res*tmp_res);
		}
	      }
	    }
	    sub_res=(1.0/(nr-1.0))*(xxs-(xs*xs)/nr);
	  }
	  for(i=0;i<VarAmnt-3;i++) vox[i]->free();
	  if(VarAmnt-3>0){
	    delete(vox);
	    delete(im_ar);
	  }
	  delete(values);
	}
	if(Comp.Immed)
	  Comp.Immed[Comp.ImmedSize]=sub_res;
	Comp.ImmedSize++;
	AddCompiledByte(cImmed);
	Comp.StackPtr++; if(Comp.StackPtr>Comp.StackSize) Comp.StackSize++;
	return;
      }
    }   
  }



  int op, i, cnt, nr;
  for(op=0; op<cOps; op++){
    cnt=0;
    for(i=e; i>b; i--){
      if(F[i]==')') cnt++;
      else if(F[i]=='(') cnt--;
      if(cnt==0 && IsOperator(&(F[i]))&& !(IsOperator(&(F[i-1]))) &&
	 !(F[i]=='-' && (IsOperator(&(F[i-1])) || F[i-1]=='('))){
	CompileSubstr(F, b, i-1);
	nr=OperatorNumber(&(F[i]));
	i=i+strlen(Ops[nr-cFirstOp])-1;
	CompileSubstr(F, i+1, e);
	AddCompiledByte(nr);
	Comp.StackPtr--;
	return;
      }
    }
  }

  int b2=b;
  if(F[b]=='-') b2++;
  AddCompiledByte(MathItemNumber(&F[b2]));
  Comp.StackPtr++; if(Comp.StackPtr>Comp.StackSize) Comp.StackSize++;
  if(b2>b) AddCompiledByte(cNeg);
}

// Return parse error message
// --------------------------
const char* FunctionParser::ErrorMsg(void)
{
    if(ParseErrorType>=0) return ParseErrorMessage[ParseErrorType];
    return 0;
}

//---------------------------------------------------------------------------
// Function evaluation
//---------------------------------------------------------------------------
//===========================================================================
double FunctionParser::Eval(double x,...)
{
    double* Vars=0;
    if(VarAmnt>0)
    {
        Vars=new double[VarAmnt];
        Vars[0]=x;
    }

    va_list ap;
    va_start(ap,x);
    for(int i=1;i<VarAmnt;i++) Vars[i]=va_arg(ap,double);
    va_end(ap);

    double Res=Eval(Vars);
    if(Vars) delete[] Vars;
    return Res;
}

double FunctionParser::Eval(double* Vars)
{
    unsigned IP, DP=0;
    int SP=-1;

    for(IP=0; IP<Comp.ByteCodeSize; IP++)
    {
        switch(Comp.ByteCode[IP])
        {
          case cImmed: Comp.Stack[++SP]=Comp.Immed[DP++]; break;
          case   cNeg: Comp.Stack[SP]=-Comp.Stack[SP]; break;
          case   cAdd: Comp.Stack[SP-1]+=Comp.Stack[SP]; SP--; break;
          case   cSub: Comp.Stack[SP-1]-=Comp.Stack[SP]; SP--; break;
          case   cMul: Comp.Stack[SP-1]*=Comp.Stack[SP]; SP--; break;
          case   cDiv: if(Comp.Stack[SP]==0) { EvalErrorType=1; return 0; }
                       Comp.Stack[SP-1]/=Comp.Stack[SP]; SP--; break;
          case   cPow: Comp.Stack[SP-1]=pow(Comp.Stack[SP-1],Comp.Stack[SP]);
                       SP--; break;
	case  cSmal:  Comp.Stack[SP-1]=(Comp.Stack[SP-1]<Comp.Stack[SP]); SP--; break;
          case  cLarg:  Comp.Stack[SP-1]=(Comp.Stack[SP-1]>Comp.Stack[SP]); SP--; break;  
	case  cAnd:  Comp.Stack[SP-1]=(Comp.Stack[SP-1]&&Comp.Stack[SP]); SP--; break;
          case  cOr:  Comp.Stack[SP-1]=(Comp.Stack[SP-1]||Comp.Stack[SP]); SP--; break;  
	case  cLogic:  Comp.Stack[SP-1]=(Comp.Stack[SP-1]==Comp.Stack[SP]); SP--; break;
          case  cSmalT:   Comp.Stack[SP-1]=(Comp.Stack[SP-1]<=Comp.Stack[SP]); SP--; break;
	  case  cLargT:  Comp.Stack[SP-1]=(Comp.Stack[SP-1]>=Comp.Stack[SP]); SP--; break;  
	case   cMin :  Comp.Stack[SP-1]=(Comp.Stack[SP-1]<Comp.Stack[SP])?
	                         Comp.Stack[SP-1]:Comp.Stack[SP];SP-- ;break;
	   case cMax   :Comp.Stack[SP-1]=(Comp.Stack[SP-1]>Comp.Stack[SP])?
			 Comp.Stack[SP-1]:Comp.Stack[SP]; SP--;break;
          case   cAbs: Comp.Stack[SP]=fabs(Comp.Stack[SP]); break;
          case   cExp: Comp.Stack[SP]=exp(Comp.Stack[SP]); break;
          case  cCeil: Comp.Stack[SP]=ceil(Comp.Stack[SP]); break;
          case cFloor: Comp.Stack[SP]=floor(Comp.Stack[SP]); break;
          case   cLog: if(Comp.Stack[SP]<=0) { EvalErrorType=3; return 0; }
                       Comp.Stack[SP]=log(Comp.Stack[SP]); break;
          case  cSqrt: if(Comp.Stack[SP]<0) { EvalErrorType=2; return 0; }
                       Comp.Stack[SP]=sqrt(Comp.Stack[SP]); break;
          case   cSin: Comp.Stack[SP]=sin(Comp.Stack[SP]); break;
          case   cCos: Comp.Stack[SP]=cos(Comp.Stack[SP]); break;
          case   cTan: Comp.Stack[SP]=tan(Comp.Stack[SP]); break;
          case  cAsin: if(Comp.Stack[SP]<-1 || Comp.Stack[SP]>1)
                       { EvalErrorType=4; return 0; }
                       Comp.Stack[SP]=asin(Comp.Stack[SP]); break;
          case  cAcos: if(Comp.Stack[SP]<-1 || Comp.Stack[SP]>1)
                       { EvalErrorType=4; return 0; }
                       Comp.Stack[SP]=acos(Comp.Stack[SP]); break;
          case  cAtan: Comp.Stack[SP]=atan(Comp.Stack[SP]); break;
          case  cSinh: Comp.Stack[SP]=sinh(Comp.Stack[SP]); break;
          case  cCosh: Comp.Stack[SP]=cosh(Comp.Stack[SP]); break;
          case  cTanh: Comp.Stack[SP]=tanh(Comp.Stack[SP]); break;
	    /*     does not compile under linux */ 
/*        case cAsinh: Comp.Stack[SP]=asinh(Comp.Stack[SP]); break;
          case cAcosh: Comp.Stack[SP]=acosh(Comp.Stack[SP]); break;
          case cAtanh: Comp.Stack[SP]=atanh(Comp.Stack[SP]); break;*/
          default:
              Comp.Stack[++SP]=Vars[Comp.ByteCode[IP]-VarBegin];
        }
    }

    EvalErrorType=0;
    return Comp.Stack[0];
}

} // namespace yav
