
#ifndef BASEINRIMAGE_H
#define BASEINRIMAGE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma interface "inrimage/BaseInrimage.h"
#include <inrimage/ImageIO.h>
#include <basics/Constants.h>
#include <linearAlgebra/Vec3.h>
#include <linearAlgebra/Rotation3D.h>
#include <linearAlgebra/Matrix4x4.h>

#ifdef LIBINRIMAGE
#include <stdexcept>
#else
#include <basics/TclString.h>
#include <basics/Exception.h>
#endif


namespace yav {



/** \class ImageInformation
    \ingroup Inrimage
    \brief Base class for additional information that may be stored in an image */
class ImageInformation {
 public:
  virtual ~ImageInformation() {}
};


/** \class BaseInrimage
    \ingroup Inrimage
    \brief Base class for inrimages */
class BaseInrimage {
 public:
  /** @name Exception classes */
  //@{
#ifdef LIBINRIMAGE
  /** \class ImageException
      \ingroup Inrimage
  */
  class ImageException : public std::exception {
  public:
    ImageException() {}
    virtual ~ImageException() {}
  };
  /** \class ImageNotFound
      \ingroup Inrimage */
  class ImageNotFound : public ImageException {
  public:
    ImageNotFound() : ImageException() {}
    virtual ~ImageNotFound() {}
  };
  /** \class UnknownImageType
      \ingroup Inrimage */
  class UnknownImageType : public ImageException {
  public:
    UnknownImageType() : ImageException() {}
  };
  /** \class InvalidHeader
      \ingroup Inrimage */
  class InvalidHeader : public ImageException {
  public:
    InvalidHeader() : ImageException() {}
  };
/** \class InvalidData
    \ingroup Inrimage */
  class InvalidData : public ImageException {
  public:
    InvalidData() : ImageException() {}
  };
/** \class UnableToCreateFile
    \ingroup Inrimage */
  class UnableToCreateFile : public ImageException {
  public:
    UnableToCreateFile() : ImageException() {}
  };
/** \class UnableToWriteHeader
    \ingroup Inrimage */
  class UnableToWriteHeader : public ImageException {
  public:
    UnableToWriteHeader() : ImageException() {}
  };
/** \class UnableToWriteData
    \ingroup Inrimage */
  class UnableToWriteData : public ImageException {
  public:
    UnableToWriteData() : ImageException() {}
  };
/** \class NotEnoughMemory
    \ingroup Inrimage */
  class NotEnoughMemory : public ImageException {
  public:
    NotEnoughMemory() : ImageException() {}
  };
/** \class InvalidSize
    \ingroup Inrimage */
  class InvalidSize : public ImageException {
  public:
    InvalidSize() : ImageException() {}
  };
/** \class InvalidVoxelSize
    \ingroup Inrimage */
  class InvalidVoxelSize : public ImageException {
  public:
    InvalidVoxelSize() : ImageException() {}
  };
/** \class InvalidVectorialMode
    \ingroup Inrimage */
  class InvalidVectorialMode : public ImageException {
  public:
    InvalidVectorialMode() : ImageException() {}
  };
#else
/** \class ImageNotFound
    \ingroup Inrimage */
  class ImageNotFound : public Exception {
  public:
    ImageNotFound(const std::string& from);
  };

/** \class UnknownImageType
    \ingroup Inrimage
    Unknown image file format */
  class UnknownImageType : public Exception {
  public:
    UnknownImageType(const std::string& from);
  };

/** \class InvalidHeader
    \ingroup Inrimage */
  class InvalidHeader : public Exception {
  public:
    InvalidHeader(const std::string& from);
  };
  
/** \class InvalidData
    \ingroup Inrimage */
  class InvalidData : public Exception {
  public:
    InvalidData(const std::string& from);
  };

/** \class UnableToCreateFile
    \ingroup Inrimage
    unable to create file for writting */
  class UnableToCreateFile : public Exception {
  public:
    UnableToCreateFile(const std::string& from);
  };

/** \class UnableToWriteHeader
    \ingroup Inrimage */
  class UnableToWriteHeader : public Exception {
  public:
    UnableToWriteHeader(const std::string& from);
  };

/** \class UnableToWriteData
    \ingroup Inrimage */
  class UnableToWriteData : public Exception {
  public:
    UnableToWriteData(const std::string& from);
  };

/** \class NotEnoughMemory
    \ingroup Inrimage */
  class NotEnoughMemory : public Exception {
  public:
    NotEnoughMemory(const std::string& from);
  };

/** \class InvalidSize
    \ingroup Inrimage
    \brief image size cannot be 0 or negative */
  class InvalidSize : public Exception {
  public:
    InvalidSize(const std::string& from);
  };

/** \class InvalidVoxelSize
    \ingroup Inrimage
    \brief voxel size cannot be 0 or negative */
  class InvalidVoxelSize : public Exception {
  public:
    InvalidVoxelSize(const std::string& from);
  };

/** \class InvalidVectorialMode
    \ingroup Inrimage
    \brief attempt to set vectorial mode to a scalar image or opposite */
  class InvalidVectorialMode : public Exception {
  public:
    InvalidVectorialMode(const std::string& from);
  };
#endif
  //@}

 protected:
  /** a 4x4 matrix used to transform from voxel coordinates to world 
      coordinates. It includse the voxel size, translation and rotation */
  Matrix4x4<double> transformationMatrix;
  /// rotation matrix between image and  world coordinates
  Rotation3D rotation;
  /// translation of the origin  between image and world coordinates
  Vec3<double> translation;
  /** to optimize computation, this boolean indicates if the rotation 
       (resp. translation) is different from the identity matrix 
       (resp. null vector) */
  bool useTransformation;

  /// Image x dimension (number of columns)
  unsigned int xdim;
  /// Image y dimension (number of rows)
  unsigned int ydim;
  /// Image z dimension (number of planes)
  unsigned int zdim;
  /// Image vectorial dimension
  unsigned int vdim;

  /// Image voxel size in x dimension
  float vx; 
  /// Image voxel size in y dimension
  float vy;
  /// Image voxel size in z dimension
  float vz; 


  /// Image data buffer
  void *data;

  /// Image word size (in bytes)
  unsigned int wdim;
  /// Image word kind
  WORD_KIND wordKind;
  /// Image word sign
  SIGN sign; 

  /// Data buffer vectors are interlaced or non interlaced
  VECTORIAL_MODE vectMode;

  /** User defined strings array.
      The user can use any internal purpose string.
      Each string is written at then end of header
      after a '#' character. */
  char **user;
  /// Number of user strings
  unsigned int nuser;

  /// format of image source file
  PTRIMAGE_FORMAT imageFormat;

  /// Optional additional information
  ImageInformation *info;

 private:
  /** update the 4x4 transformation matrix once the translation
      or rotation have been modified */
  void updateTransformationMatrix();

 public:
  /// get image x dimension
  unsigned int getX() const { return xdim; }
  /// get image y dimension
  unsigned int getY() const { return ydim; }
  /// get image z dimension
  unsigned int getZ() const { return zdim; }
  /// set image x dimension
  void setX(const unsigned int x) { xdim = x; }
  /// set image y dimension
  void setY(const unsigned int y) { ydim = y; }
  /// set image z dimension
  void setZ(const unsigned int z) { zdim = z; }
  
  /// set image vectorial dimension
  void setVdim(const unsigned int v) { vdim = v; }
  /// get image vectorial dimension
  unsigned int getVdim(void) const { return vdim; }
  /// get image words size (in bytes)
  unsigned int getWdim(void) const { return wdim; }

  /// get image voxel size in x dimension
  float getVX() const { return vx; }
  /// get image voxel size in y dimension
  float getVY() const { return vy; }
  /// get image voxel size in z dimension
  float getVZ() const { return vz; }

  /// get image translation (offset) in the x dimension
  float getTX() const { return vx; }
  /// get image translation (offset) in the y dimension
  float getTY() const { return vy; }
  /// get image translation (offset) in the z dimension
  float getTZ() const { return vz; }

  /// get image data buffer
  void *getData() const { return data; }


  /// returns efficiently the 4x4 transformation matrix
  Matrix4x4<double> &getTransformationMatrix() {
    return transformationMatrix;
  }
  /// set the transformation matrix
  inline void setTransformationMatrix(const Matrix4x4<double> &mat) {
    transformationMatrix=mat;
  }
  /// returns the rotation matrix
  Rotation3D &getRotationMatrix()  {
    return rotation;
  }
  /// returns the translation (offset) vector
  Vec3<double> &getTranslation() {
    return translation;
  }

  /// returns the 4x4 transformation matrix
  Matrix4x4<double> getTransformationMatrixFromConstImage()  const {
    return transformationMatrix;
  }
  /// returns the rotation matrix
  Rotation3D getRotationMatrixFromConstImage()  const {
    return rotation;
  }
  /// returns the translation (offset) vector
  Vec3<double> getTranslationFromConstImage()  const {
    return translation;
  }

  /** compose the current rotation with a new rotation matrix */
  void composeRotation(const Rotation3D& rot) {
    rotation *= rot;
    updateTransformationMatrix();
  }
  /** set the rotation matrix */
  void setRotation(const Rotation3D& rot) {
    rotation = rot;
    updateTransformationMatrix();
  }
  /** add to the current translation (offset) vector a given vector */
  void addTranslation(const Vec3<double>& off) {
    translation +=off;
    updateTransformationMatrix();
  }
  /** set the current offset vector */
  void setTranslation(const Vec3<double>& off) {
    translation =off;
    updateTransformationMatrix();
  }
  /// whether the translation and rotation are non trivial
  inline bool useTransformationMatrix() const {
    return useTransformation;
  }
  /// Turn on the use of the transformation matrix
  void useTransformationMatrixOn() {
    useTransformation = true;
  }
  void useTransformationMatrixOff() {
    useTransformation = false;
  }



  /// get image word size
  unsigned int getWordSize() const { return wdim; }
  /// get word kind
  WORD_KIND getWordKind() const { return wordKind; }
  /// get word sign
  SIGN getWordSign() const { return sign; }
  /// get VECTORIAL_MODE
  VECTORIAL_MODE getVectorialMode() const { return vectMode; }

  /// get image file format
  PTRIMAGE_FORMAT getFileFormat(void) const { return imageFormat; }
  /** set a new image file format for saving
      @param f new file format */
  void setFileFormat(PTRIMAGE_FORMAT f) { imageFormat = f; }

  /** @name Constructors */
  //@{
  /** Reads inrimage from file whose path is given as argument.
      File can be an INRIMAGE(4) gzipped or not, an analyse, a GIF or
      an RGB irix file. If image has a vectorial type, it can be read
      interlaced (default) or non interlaced by giving
      WM_NON_INTERLACED as third argument. Ih headerOnly parameter
      is true, only image header is read.
      @param name inrimage file name
      @param headerOnly whether to read header or full image
      @param vm interlaced or non-interlaced mode for vectorial image
      @param filevm if VM_NON_INTERLACED, data is considered as non-interlaced
      @exception ImageNotFound image file was not found
      @exception UnknownImageType unknown image file format
      @exception InvalidData invalid inriamge header read
      @exception InvalidData invalid inriamge data read */
  BaseInrimage(const char *name, bool headerOnly = false,
	       VECTORIAL_MODE vm = VM_INTERLACED,
	       VECTORIAL_MODE filevm = VM_INTERLACED)
    throw(ImageNotFound, UnknownImageType, InvalidHeader, InvalidData);

  /** Creates an inrimage with given size and word type.
      The inrimage data buffer is allocated.
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize image dimension should be > 0
      @exception NotEnoughMemory unable to allocate memory to store image
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  BaseInrimage(unsigned int xdim, unsigned int ydim,  unsigned int zdim,
	       unsigned int vdim = 1, unsigned int wdim = 1, 
	       WORD_KIND wordKind = WK_FIXED, SIGN sign = SGN_SIGNED,
	       float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(InvalidSize, NotEnoughMemory, InvalidVoxelSize);

  /** Creates an inrimage with given size and word type.
      If data is 0, the image is not allocated.
      If data is non 0, image data points to it.
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param data image data buffer
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize image dimension should be > 0
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  BaseInrimage(unsigned int xdim, unsigned int ydim, unsigned int zdim,
	       void *data,  unsigned int vdim = 1, unsigned int wdim = 1,
	       WORD_KIND wordKind = WK_FIXED, SIGN sign = SGN_SIGNED,
	       float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(InvalidSize, InvalidVoxelSize);

  /** creates an replica of img
      @param img image to copy
      @exception InvalidSize image dimension should be > 0
      @exception InvalidVectorialMode attempt to set a vector mode to a scalar image or opposite
      @exception InvalidVoxelSize voxels size should be > 0.0
      @exception NotEnoughMemory unable to allocate memory to store image */
  explicit
  BaseInrimage(const BaseInrimage& img)
     throw(InvalidSize, InvalidVectorialMode, InvalidVoxelSize,
	   NotEnoughMemory);
  //@}

  /** @name Destructors */
  //@{
  /// frees image buffer
  virtual ~BaseInrimage();
  //@}

  /** Writes inrimage to file whose path is given as argument.
      If file name ends with '.gz' suffix, file is gzipped.
      @param name inrimage file name
      @exception UnableToCreateFile file creation failed for writting
      @exception UnableToWriteHeader write error encountered while writting header
      @exception UnableToWriteData write error encountered while writting data */
  void write(const char *name) const
    throw(UnableToCreateFile, UnableToWriteHeader,
	  UnableToWriteData);

  /** Makes *this an replica of img.
      @param img image to copy 
      @exception InvalidSize image dimension should be > 0
      @exception InvalidVectorialMode attempt to set a vector mode to a scalar image or opposite
      @exception InvalidVoxelSize voxels size should be > 0.0
      @exception NotEnoughMemory unable to allocate memory to store image */
  BaseInrimage& operator=(const BaseInrimage& img)
     throw(InvalidSize, InvalidVectorialMode, InvalidVoxelSize,
	   NotEnoughMemory);
      
  /** Allocates and fills with zero image data buffer. If data
      is not null, buffer is first fried.
      @exception NotEnoughMemory unable to allocate memory to store image */
  virtual void allocate(void) throw(NotEnoughMemory);

  /// frees image data buffer
  virtual void free(void);

  /** set a new image voxel size
      @param vx X voxel size
      @param vy Y voxel size
      @param vz Z voxel size
      @exception InvalidVoxelSize invalid voxel size */
  void setVoxelSize(float vx, float vy, float vz)
    throw(InvalidVoxelSize);
// returns the image voxel size in a Vec3<float>(vx,vy,vz)
  Vec3<float> getVoxelSize(void){
    return (Vec3<float>(vx,vy,vz));
  }
// returns the image dimensions in a Vec3<unsigned int>(xdim,ydim,zdim)
  Vec3<unsigned int> getImageDimension(){
    return (Vec3<unsigned int>(xdim,ydim,zdim));
  }

  /** set a new data buffer. Returns old image data buffer.
      Use setVectorialMode(VECTORIAL_MODE vm) if vectMode has changed 
      @param data new data buffer */
  virtual void *setData(void *data);

  /** Changes image vectorial mode. A vectorial image buffer can be
      stored either interlaced (default) or non interlaced.
      @param vm new image vectorial mode
      @exception InvalidVectorialMode attempt to set a vector mode to a scalar image or opposite */
  virtual void setVectorialMode(VECTORIAL_MODE vm)
    throw(InvalidVectorialMode);
 
  /** toggles buffer to produce an interlaced vector image (for
      vectorial images only) */
  virtual void setInterlacedData(void);

  /** toggles buffer to produce an interlaced vector image (for
      vectorial images only) */
  virtual void setNonInterlacedData(void);

  /** set user defined parameters strings to write into image file
      header.
      @param n number of strings
      @param params strings array */
  void setUserParams(unsigned int n, char **params);

  /** Changes image size. If image data buffer existed it is fried
      and reallocated.
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @exception InvalidSize image dimension should be > 0
      @exception NotEnoughMemory unable to allocate memory to store image */
  virtual void resize(unsigned int xdim, unsigned int ydim,
		      unsigned int zdim, unsigned int vdim = 1,
		      unsigned int wdim = 1,
		      WORD_KIND wordKind = WK_FIXED,
		      SIGN sign = SGN_SIGNED)
    throw(InvalidSize, NotEnoughMemory);

  /** Changes image word type. If image data buffer existed it is fried
      and reallocated.
      @param wordKind image word kind
      @param sign image word type
      @exception NotEnoughMemory unable to allocate memory to store image */
  virtual void setWords(WORD_KIND wordKind = WK_FIXED,
			SIGN sign = SGN_SIGNED) throw(NotEnoughMemory);


  /** Converts an _image structure to a BaseInrimage
      @param img _image structure
      @exception InvalidSize image dimension should be > 0
      @exception InvalidVoxelSize voxels size should be > 0.0
      @exception InvalidVectorialMode attempt to set a vector mode to a scalar image or opposite
      @return BaseInrimage object with same data buffer than img */
  static BaseInrimage *_image2BaseInrimage(const _image& img)
    throw(BaseInrimage::InvalidSize, BaseInrimage::InvalidVoxelSize,
	  BaseInrimage::InvalidVectorialMode);

  /** Converts a BaseImage structure to an _image
      @param bimg BaseInrimage object
      @return _image structure with same data buffer than bimg */
  static _image *BaseInrimage2_image(const BaseInrimage& bimg);

  /** Check whether given field exists in user defined strings.
      Looks for a field with following syntax: name [= val]
      If name is found and val buffer is not 0, field value
      is returned in val.
      @param name researched field name
      @param val returned field value (buffer must have been previously allocated)
      @param n value buffer length
      @return true if field exists and false otherwise */
  bool userField(const char *name, char *val,  unsigned int n) const;

  /** Check whether given field exists in user defined strings.
      Return true if field exists and false otherwise.
      Looks for a field with following syntax: name [= val]
      If name is found and val buffer is not NULL, field value
      is set to val.
      @param name researched field name
      @param val value to give to field
      @return true if field exists and false otherwise */
  bool updateUserField(const char *name, const char *val) const ;

  /** Add a given field to the user defined strings.
      Looks for a field with following syntax: name [= val]
      If name is found and val buffer is not NULL, field value
      is modified. Else, the field is added.
      @param name researched field name
      @param val value to give to field */
  void addUserField(const char *name, const char *val);

  /** Remove a given field to the user defined strings.
      Looks for a field with following syntax: name [= val]
      If name is found it is removed for the user field list.
      @param name researched field name
      @return true if field was found and false otherwise */
  bool removeUserField(const char *name);

  /** Rescale data values into another image.
      <b>img</b> needs to be already initialized to proper format.
      <b>fmin</b> and <b>fmax</b> are min and max values of this image's buffer.
      <b>tmin</b> and <b>tmax</b> are min and max values of resulting image's buffer.
      @param fmin minimum value of image
      @param fmax maximum value of image
      @param tmin minimum value of image
      @param tmax maximum value of image
      @param offset offset to add to every value.
      @param img image to put result in */
  template <class FROM, class TO>
    void rescaleBuffer(FROM , TO , double fmin,double fmax,double tmin,double tmax,double offset,BaseInrimage& img){
    
    TO *to=(TO *) img.getData();
    FROM *from=(FROM *) data;
    unsigned LONGINT size = img.getX() * img.getY() * img.getZ() * img.getVdim();
    
    if(fmax==fmin) fmax++;
    if(tmax==tmin) tmax++;
    double const frange=fmax-fmin;
    double const trange=tmax-tmin;
    double const rr=trange/frange;
    tmin+=offset;
    
    for(unsigned LONGINT i=0;i<size;i++,from++,to++) *to=(TO)((*from-fmin)*rr+tmin);
  }
  
  
  /** compute grey level bounds of image buffer (valid for image scalar
      types only)
      @param min returned min value
      @param max returned max value
  */
  void getBounds(double &min, double &max) const;
  
 private:
  /** open inrimage file and read header
      @param name inrimage file name
      @return inrimage header, data is 0
      @exception ImageNotFound image file was not found
      @exception InvalidHeader invalid inrimage header encountered
      @exception UnknownImageType unknown file format */
  _image *readHeader(const char *name)
    throw(ImageNotFound, InvalidHeader, UnknownImageType);
  
  template <class T>
    void internalGetBounds(T , double &min, double &max) const {
    
    T *buf = (T *) data;
    T *end = buf + xdim * ydim * zdim;
    
    
    min = MAX_DOUBLE;
    max = MIN_DOUBLE;
    while(buf < end) {
      if((double) *buf > max) max = (double) *buf;
      if((double) *buf < min) min = (double) *buf;
      buf++;
    }
  }
};
 
 
#ifdef MINC_FILES
/** \class MincInformation
    \ingroup Inrimage
    \brief Additional information contained in minc files */
class MincInformation : public ImageInformation {
 protected:
  /// original minc file name
  TclString sourceName;
  /// start vector
  double start[3];
  /// step vector
  double step[3];
  /// X cosine directions
  double Xcosine[3];
  /// Y cosine directions
  double Ycosine[3];
  /// Z cosine directions
  double Zcosine[3];
  
 public:
  /** builds a minc information object
      @param sourceName source minc file name
      @param startx origin vector X coordinate
      @param starty origin vector Y coordinate
      @param startz origin vector Z coordinate
      @param stepx step vector X coordinate
      @param stepy step vector Y coordinate
      @param stepz step vector Z coordinate
      @param Xcosine 3 dimensional array containing X axis cosine directions
      @param Ycosine 3 dimensional array containing Y axis cosine directions
      @param Zcosine 3 dimensional array containing Z axis cosine directions
  */
  MincInformation(const char *sourceName,
		  double startx, double starty, double startz,
		  double stepx, double stepy, double stepz,
		  double *Xcosine, double *Ycosine, double *Zcosine);
  /// destroys minc information object
  virtual ~MincInformation() {}

  /// returns origin vector X coordinate
  double getStartX(void) const { return start[0]; }
  /// returns origin vector Y coordinate
  double getStartY(void) const { return start[1]; }
  /// returns origin vector Z coordinate
  double getStartZ(void) const { return start[2]; }
  
  /// returns step vector X coordinate
  double getStepX(void) const { return step[0]; }
  /// returns step vector Y coordinate
  double getStepY(void) const { return step[1]; }
  /// returns step vector Z coordinate
  double getStepZ(void) const { return step[2]; }
  
  /// returns source minc file name
  const TclString& getSourceFilename(void) const { return sourceName; }

  /// returns a 3 dimension array containing X cosine directions
  const double *getXcosines(void) const { return Xcosine; }
  /// returns a 3 dimension array containing Y cosine directions
  const double *getYcosines(void) const { return Ycosine; }
  /// returns a 3 dimension array containing Z cosine directions
  const double *getZcosines(void) const { return Zcosine; }
};
#endif

/*
template <class FROM, class TO>
void BaseInrimage::rescaleBuffer(double fmin,double fmax,double tmin,double tmax,double offset,BaseInrimage& img) 
*/


} // namespace yav
  
#endif


