#ifndef CONNECTINGELEMENT_H
#define CONNECTINGELEMENT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma interface "inrimage/ConnectingElement.h"

#include <inrimage/Inrimage.h>

#include <inrimage/connectivity.h>


namespace yav {

typedef int typeSeedPoint[3];


/** @name Connected components extraction based functions
 */

//@{



  /// The type of output that is desired when computing the connected component
  enum COMPONENT_OUTPUT {
    /// Only the binary image is outputed
    COMPONENT_BINARY,
    /** An unsigned short image is outputed corresponding
	to the different labels of the connected components */
    COMPONENT_LABEL
  };
 


  /// The parameter of the connected component computation
  /// not to be used (will disappear soon)
  struct CONNEXE_PARAMETER {
    /// minimum size of connected components
    unsigned int min_size;
    /// maximum number of connected components
    int max_nbcc;
    /// output: binary image or label image
    COMPONENT_OUTPUT output;
  };

CONNEXE_PARAMETER *connectedParameterAlloc();










/** \class ConnectingElement 
    \ingroup Inrimage
    \brief Class for connecting elements 

    This class is used for connected components based procedure. It is named
    ConnectingElement after StructuringElement.
*/
class ConnectingElement {

  /** @name Exception classes 
   */
  //@{
 public:
#ifdef LIBINRIMAGE
/** \class ElementException
    \ingroup Inrimage
*/
  class ElementException : public std::exception {
  public:
    ElementException() {}
    virtual ~ElementException() {}
  };
#endif // #ifdef LIBINRIMAGE
  //@}

 protected:
  /** @name Parameters
   */
  //@{  
  /// connectivity
  CONNECTIVITY_TYPE connectivity;
  /// the minimal size of selected connected components to be valid
  int minimal_size;
  /// the minimal number of points upper the high threshold
  /// that a connected component must contain to be valid
  int minimal_size_high_threshold;
  /// maximal number of connected components to be
  /// retained. If smaller than the number of valid connected components,
  /// then only the larger components are selected. 
  /// If this number is greater than 1 but smaller than the largest encoding
  /// label (eg 65535 with work_type= Inrimage::WT_UNSIGNED_SHORT),
  /// the connected components are sorted by size, even if there are less
  /// connected components thna this number. Thus the largest will be indexed
  /// by 1, etc.
  int maximal_number_of_components;
  /// type of output: if COMPONENT_LABEL, components are keeped indexed,
  /// if COMPONENT_BINARY, the ouput image is binarized.
  COMPONENT_OUTPUT output;
  /// type of the work image (intermediary labels)
  Inrimage::WORD_TYPE work_type;
  /// number of classes in the equivalence array
  unsigned int cc_array_size;

  /// number of seed points (if any)
  unsigned int number_seed_points;
  /// seed points
  /// only the connected components containing seed points are kept
  typeSeedPoint *seeds;
  
  //@}
  
  private:
  /// initialisation des champs
  void init( void );

 public:

  /** @name Public methods
   */
  //@{  
  CONNECTIVITY_TYPE getConnectivity( void ) { return connectivity; }
  int getMinimalSize( void ) { return minimal_size; }
  int getMinimalSizeHighThreshold( void ) { return minimal_size_high_threshold; }
  int getMaximalNumberOfComponents( void ) { return maximal_number_of_components; }
  COMPONENT_OUTPUT getOutputType( void ) { return output; }
  Inrimage::WORD_TYPE getWorkType( void ) { return work_type; }
  unsigned int getCCArraySize( void ) { return cc_array_size; }

  unsigned int getNumberSeedPoints( void ) { return number_seed_points; }
  typeSeedPoint *getSeeds( void ) { return seeds; }

  void setConnectivity( CONNECTIVITY_TYPE C ) { connectivity = C; }
  void setConnectivity( int c );
  void setMinimalSize( int s ) { minimal_size = s; }
  void setMinimalSizeHighThreshold( int s ) { minimal_size_high_threshold = s; }
  void setMaximalNumberOfComponents( int n ) { maximal_number_of_components = n; }
  void setOutputType( COMPONENT_OUTPUT O ) { output = O; }
  void setWorkType( Inrimage::WORD_TYPE T );
  void getCCArraySize( unsigned int s ) { cc_array_size = s; }

  void addSeedPoint( int x, int y, int z = -1 );
  void markSeedPoints( Inrimage *inr, Voxel *v )
    throw( ErrTypeMismatch );

  //@}
  
 public:
  /** @name Constructors */
  //@{
  ConnectingElement( const CONNECTIVITY_TYPE C=TWENTY_SIX_CONNECTIVITY,
		     const unsigned int min_size_low_threshold = 1,
		     const unsigned int min_size_high_threshold = 1,
		     const int max_number_of_components = -1,
		     const COMPONENT_OUTPUT type_output = COMPONENT_BINARY,
		     const Inrimage::WORD_TYPE label_type = Inrimage::WT_UNSIGNED_SHORT );
  ConnectingElement( const int c,
		     const unsigned int min_size_low_threshold = 1,
		     const unsigned int min_size_high_threshold = 1,
		     const int max_number_of_components = -1,
		     const COMPONENT_OUTPUT type_output = COMPONENT_BINARY,
		     const Inrimage::WORD_TYPE label_type = Inrimage::WT_UNSIGNED_SHORT );
  //@}
  
  /** @name Destructor */
  //@{
  /// deallocates image buffer
  virtual ~ConnectingElement();
  //@}



}; // end of class ConnectingElement
















/** Applies a hysteresis thresholding and cast the result image.
    Such a thresholding keeps the connected components,
    made of points whose value is larger or equal to 'low threshold',
    and which contains at least one (or some, see parameters)
    point whose value is larger or equal to 'high threshold'.
    @param inr
    @param low_threshold  self explanatory
    @param high_threshold self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void hysteresis( Inrimage *inr, 
		 double low_threshold,  double high_threshold,
		 ConnectingElement *CE=NULL,
		 Voxel *bg = NULL, Voxel *fg=NULL )
  throw( Exception );

/** Applies a hysteresis thresholding and cast the result image.
    Such a thresholding keeps the connected components,
    made of points whose value is larger or equal to 'low threshold',
    and which contains at least one (or some, see parameters)
    point whose value is larger or equal to 'high threshold'.
    @param inr
    @param newtype wanted type of result image
    @param low_threshold  self explanatory
    @param high_threshold self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void hysteresis( Inrimage *inr, 
		 Inrimage::WORD_TYPE newtype,
		 double low_threshold,  double high_threshold,
		 ConnectingElement *CE=NULL,
		 Voxel *bg = NULL, Voxel *fg=NULL )
  throw( Exception );
 
/** Applies a hysteresis thresholding and cast the result image.
    Such a thresholding keeps the connected components,
    made of points whose value is larger or equal to 'low threshold',
    and which contains at least one (or some, see parameters)
    point whose value is larger or equal to 'high threshold'.
    @param inr
    @param low_threshold  self explanatory
    @param high_threshold self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void hysteresis( Inrimage *inr, 
		 Voxel *low_threshold,  Voxel *high_threshold,
		 ConnectingElement *CE=NULL,
		 Voxel *bg = NULL, Voxel *fg=NULL )
  throw( Exception );
 
/** Applies a hysteresis thresholding and cast the result image.
    Such a thresholding keeps the connected components,
    made of points whose value is larger or equal to 'low threshold',
    and which contains at least one (or some, see parameters)
    point whose value is larger or equal to 'high threshold'.
    @param inr
    @param newtype wanted type of result image
    @param low_threshold  self explanatory
    @param high_threshold self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void hysteresis( Inrimage *inr, 
		 Inrimage::WORD_TYPE newtype,
		 Voxel *low_threshold,  Voxel *high_threshold,
		 ConnectingElement *CE=NULL,
		 Voxel *bg=NULL, Voxel *fg=NULL )
  throw( Exception );
 





/** Extract the connected components.
    First the input image is thresholded, 
    if the voxels values are denoted by v
    Background is defined by: v <  min
    Foreground is defined by: v >= min
    Then the connected components of the foreground are computed
    and selected with respect to extraction parameters CE.
    @param inr
    @param min self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void connectedComponents( Inrimage *inr, 
			  double min,
			  ConnectingElement *CE=NULL,
			  Voxel *bg=NULL, Voxel *fg=NULL )
  throw( Exception );
 
/** Extract the connected components and cast the result image.
    First the input image is thresholded, 
    if the voxels values are denoted by v
    Background is defined by: v <  min
    Foreground is defined by: v >= min
    Then the connected components of the foreground are computed
    and selected with respect to extraction parameters CE.
    @param inr
    @param newtype wanted type of result image
    @param min self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void connectedComponents( Inrimage *inr, 
			  Inrimage::WORD_TYPE newtype,
			  double min,
			  ConnectingElement *CE=NULL,
			  Voxel *bg=NULL, Voxel *fg=NULL )
   throw( Exception );
 
/** Extract the connected components.
    First the input image is thresholded, 
    if the voxels values are denoted by v
    Background is defined by: v <  min ( || v >  max  if max is provided)
    Foreground is defined by: v >= min ( && v <= max  if max is provided)
    Then the connected components of the foreground are computed
    and selected with respect to extraction parameters CE.
    @param inr
    @param min self explanatory
    @param max self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void connectedComponents( Inrimage *inr, 
			  double min,  double max,
			  ConnectingElement *CE=NULL,
			  Voxel *bg=NULL, Voxel *fg=NULL )
  throw( Exception );
 
/** Extract the connected components and cast the result image.
    First the input image is thresholded, 
    if the voxels values are denoted by v
    Background is defined by: v <  min ( || v >  max  if max is provided)
    Foreground is defined by: v >= min ( && v <= max  if max is provided)
    Then the connected components of the foreground are computed
    and selected with respect to extraction parameters CE.
    @param inr
    @param newtype wanted type of result image
    @param min self explanatory
    @param max self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void connectedComponents( Inrimage *inr, 
			  Inrimage::WORD_TYPE newtype,
			  double min,  double max,
			  ConnectingElement *CE=NULL,
			  Voxel *bg=NULL, Voxel *fg=NULL )
  throw( Exception );

/** Extract the connected components.
    First the input image is thresholded, 
    if the voxels values are denoted by v
    Background is defined by: v <  min ( || v >  max  if max is provided)
    Foreground is defined by: v >= min ( && v <= max  if max is provided)
    Then the connected components of the foreground are computed
    and selected with respect to extraction parameters CE.
    @param inr
    @param min self explanatory
    @param max self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void connectedComponents( Inrimage *inr, 
			  Voxel *min,  Voxel *max=NULL,
			  ConnectingElement *CE=NULL,
			  Voxel *bg=NULL, Voxel *fg=NULL )
  throw( Exception );

/** Extract the connected components and cast the result image.
    First the input image is thresholded, 
    if the voxels values are denoted by v
    Background is defined by: v <  min ( || v >  max  if max is provided)
    Foreground is defined by: v >= min ( && v <= max  if max is provided)
    Then the connected components of the foreground are computed
    and selected with respect to extraction parameters CE.
    @param inr
    @param newtype wanted type of result image
    @param min self explanatory
    @param max self explanatory
    @param CE parameters for connected components extraction
    @param bg if output = COMPONENT_BINARY, background color,
              If NULL, it will be 0
    @param fg if output = COMPONENT_BINARY, foreground color,
              If NULL, it will be the maximum encoding value
*/
void connectedComponents( Inrimage *inr, 
			  Inrimage::WORD_TYPE newtype,
			  Voxel *min,  Voxel *max=NULL,
			  ConnectingElement *CE=NULL,
			  Voxel *bg=NULL, Voxel *fg=NULL )
  throw( Exception );






/** Fill contours.
    The image is supposed to be binary (on the basis of background
    and foreground voxels). All the connected components
    are extracted (background and foreground). Then,
    background components are sorted into two categories,
    outside and inside (the first outside components are the ones
    that touch the image borders), the inside components are
    changed into foreground.
    @param inr
    @param CE parameters for connected components extraction
    @param background self explanatory (set to 0 if NULL)
    @param foreground self explanatory
    @param threeD (if true 3D computation)
*/
void fillContours( Inrimage *inr, 
		   ConnectingElement *CE = NULL,
		   Voxel *background = NULL, 
		   Voxel *foreground = NULL, 
		   bool threeD = true )
  throw( Exception );

/** Fill contours and cast the result image.
    The image is supposed to be binary (on the basis of background
    and foreground voxels). All the connected components
    are extracted (background and foreground). Then,
    background components are sorted into two categories,
    outside and inside (the first outside components are the ones
    that touch the image borders), the inside components are
    changed into foreground.
    @param inr
    @param newtype wanted type of result image
    @param CE parameters for connected components extraction
    @param background self explanatory (set to 0 if NULL)
    @param foreground self explanatory
    @param threeD (if true 3D computation)
*/
void fillContours( Inrimage *inr, 
		   Inrimage::WORD_TYPE newtype,
		   ConnectingElement *CE = NULL,
		   Voxel *background = NULL, 
		   Voxel *foreground = NULL, 
		   bool threeD = true )
  throw( Exception );




/** Reconstruct a binary image.
    The image is supposed to be binary (on the basis of background
    and foreground voxels). 
    Only the foreground components containing seeds (from
    the connecting element or/and from the seeds image)
    are kept.
    @param inr
    @param CE parameters for connected components extraction
    @param seeds seeds image, seeds are points of the image 
           whose value is strictly greater than 0
    @param background self explanatory (set to 0 if NULL)
    @param foreground self explanatory
*/
void reconstruct( Inrimage *inr,
		  ConnectingElement *CE, 
		  Inrimage *seeds=NULL,
		  Voxel *background=NULL, Voxel *foreground=NULL )
  throw( Exception );

/** Reconstruct a binary image and cast the result image.
    The image is supposed to be binary (on the basis of background
    and foreground voxels). 
    Only the foreground components containing seeds (from
    the connecting element or/and from the seeds image)
    are kept.
    @param inr
    @param newtype wanted type of result image
    @param CE parameters for connected components extraction
    @param seeds seeds image, seeds are points of the image 
           whose value is strictly greater than 0
    @param background self explanatory (set to 0 if NULL)
    @param foreground self explanatory
*/
void reconstruct( Inrimage *inr, Inrimage::WORD_TYPE newtype,
		  ConnectingElement *CE, 
		  Inrimage *seeds=NULL,
		  Voxel *background=NULL, Voxel *foreground=NULL )
  throw( Exception );


//@}





} // namespace yav

#endif // #ifndef CONNEXE_H
