#ifndef DISPLACEMENTFIELD3D_H
#define DISPLACEMENTFIELD3D_H

#pragma interface "inrimage/DisplacementField3D.h"
#include <basics/GeneralFormat.h>
#include <basics/Owner.h>
#include <framework/Client.h>
#include <inrimage/Inrimage.h>
#include <basics/Exception.h>

namespace yav {

  
/** \class DisplacementField3D
    \ingroup Inrimage
    \brief class for reel or voxel transformation */
  class DisplacementField3D : public GeneralFormat, public Owner<Inrimage>,
    public Pointed<DisplacementField3D>, public Client
    {
      
      /** @name Exception classes */
      /** \class DisplacementField3DException
          \ingroup Inrimage
      */
      class DisplacementField3DException : public Exception {
      public:
	DisplacementField3DException(const std::string& from)
	  : Exception(from, " General DisplacementField3D Exception ") {}
	virtual ~DisplacementField3DException() throw() {}
      };

      /** \class InvalidDisplacementField3DMode
          \ingroup Inrimage
      */
      class InvalidDisplacementField3DMode : public Exception {
      public:
	InvalidDisplacementField3DMode(const std::string& from)
	  : Exception(from, " DisplacementField3D mode is invalid ") {}
	virtual ~InvalidDisplacementField3DMode() throw() {}
      };

      /** \class InvalidDisplacementField3DImages
          \ingroup Inrimage
      */
      class InvalidDisplacementField3DImages : public Exception {
      public:
	InvalidDisplacementField3DImages(const std::string& from)
	  : Exception(from, " Some Images of the DisplacementField3D are null" ) {}
	virtual ~InvalidDisplacementField3DImages() throw(){}
      };

      /** \class InvalidDisplacementField3DImagesNb
          \ingroup Inrimage
      */
      class InvalidDisplacementField3DImagesNb : public Exception {
      public:
	InvalidDisplacementField3DImagesNb(const std::string& from)
	  : Exception(from, " Number of the images of the DisplacementField3D must be 3 ") {}
	virtual ~InvalidDisplacementField3DImagesNb() throw() {}
      };

      /** \class DisplacementField3DMismatchType
          \ingroup Inrimage
      */
      class DisplacementField3DMismatchType : public Exception {
      public:
	DisplacementField3DMismatchType(const std::string& from)
	  : Exception(from, " Types of the images of the DisplacementField3D must be the same") {}
	virtual ~DisplacementField3DMismatchType() throw(){}
      };

      /** \class DisplacementField3DMismatchSize
          \ingroup Inrimage
      */
      class DisplacementField3DMismatchSize : public Exception {
      public:
	DisplacementField3DMismatchSize(const std::string& from)
	  : Exception(from, " Size of the images of the DisplacementField3D must be the same") {}
	virtual ~DisplacementField3DMismatchSize() throw(){}
      };

      /** \class DisplacementField3DMismatchVoxelSize
          \ingroup Inrimage
      */
      class DisplacementField3DMismatchVoxelSize : public Exception {
      public:
	DisplacementField3DMismatchVoxelSize(const std::string& from)
	  : Exception(from, " The images of the DisplacementField3D have difeerent voxel sizes ") {}
	virtual ~DisplacementField3DMismatchVoxelSize() throw(){}
      };

      
    public:

      /** transformation can be real or voxel */
      enum FIELD_MODE {
	/// reel
	REAL,
	/// voxel
	VOXEL
      };


    public:
      /** @name Constructors */
      //@{
      /** create a transformation depending on its mode
	  The images of the deformation are needed : 2 in 2D and 3 in 3D 
	  In case of VOXEL mode, size of voxel is needed */
      
      DisplacementField3D ( FIELD_MODE m, Inrimage *imx, Inrimage *imy, Inrimage *imz, float vx=1, float vy=1, float vz=1 ) throw(DisplacementField3DMismatchType, DisplacementField3DMismatchSize, DisplacementField3DMismatchVoxelSize);

      /** build a new Displacement Field
	  @param filename 3D field displacement file name
	  @exception ImageNotFound */
      DisplacementField3D ( const char *filename )
	throw(GeneralFormat::FileNotFound,
	      Inrimage::ImageNotFound, InvalidDisplacementField3DImagesNb,
	      DisplacementField3DMismatchType,DisplacementField3DMismatchSize ) ;

      //@}

      /** @name destructor */
      //@{
      /// destroys any 3D inrimage
      ~DisplacementField3D();
      //@}
     

  
      /** an image owned by this 3D displacement field is being destroyed: destroy the displacement field
	  @param obj element before destruction */
      virtual void dying(const ObjectPointer<Inrimage> &obj) {
	delete this;
      }

      /// transform a voxel DisplacementField3D into a real DisplacementField3D
      void VoxelToReal(void) throw (DisplacementField3D::InvalidDisplacementField3DMode);
      /// transform a real DisplacementField3D into a voxel DisplacementField3D
      void RealToVoxel(float vx, float vy, float vz) throw (DisplacementField3D::InvalidDisplacementField3DMode,
							    Inrimage::InvalidVoxelSize);
  
      /// get the type of the image of deplacement
      Inrimage::WORD_TYPE getType(void) const { return type; }
  
      /// get transfo image voxel size in x dimension in case of VOXEL type
      float getVX() const throw(DisplacementField3D::InvalidDisplacementField3DMode);
      /// get transfo image voxel size in y dimension in case of VOXEL type
      float getVY() const throw(DisplacementField3D::InvalidDisplacementField3DMode);
      /// get transfo image voxel size in z dimension in case of VOXEL type
      float getVZ() const throw(DisplacementField3D::InvalidDisplacementField3DMode);
      /// set transfo image voxel size in x dimension in case of VOXEL type
      void setVX(float) throw(DisplacementField3D::InvalidDisplacementField3DMode);
      /// set transfo image voxel size in y dimension in case of VOXEL type
      void setVY(float) throw(DisplacementField3D::InvalidDisplacementField3DMode);
      /// set transfo image voxel size in z dimension in case of VOXEL type
      void setVZ(float) throw(DisplacementField3D::InvalidDisplacementField3DMode);
      
      /** general format read function for Inrimage names module */
      void readInrNames(GFInput &input) throw(InvalidDisplacementField3DImagesNb);
      /** general format skip function for Inrimage names module */
      void skipInrNames(GFInput &input);  
      /** general format write function for Inrimage names module */
      void writeInrNames(GFOutput &output);
       

    protected:
      /// 3D FIELD file name
      TclString filename;
      /// mode of the transformation
      FIELD_MODE mode;

      /// x, y and z deformation images pointers on deformation image
      ObjectPointer<Inrimage> *inrx;
      ObjectPointer<Inrimage> *inry;
      ObjectPointer<Inrimage> *inrz;
      
      /// type of the images of deformation
      Inrimage::WORD_TYPE type;

      /// size of the voxels in case of voxel type transformation
      float vx, vy, vz;
      

    private:
      /// constructors common code for I/O routines
      void initGeneralFormat(void);


      friend class TclDisplacementField3D;
      
    };
  
  

}

#endif
