#ifndef IMAGE3DBACKGROUNDMANAGER_H
#define IMAGE3DBACKGROUNDMANAGER_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/Image3DBackgroundManager.h"
#include <inrimage/Image3DCameraBackground.h>
#include <basics/DPOwner.h>

namespace yav {


class Image3DBackgroundManager;

class Image3DViewerInterface : public ListElement<Image3DViewerInterface> {
protected:
	/// the manager on which this viewer interface is pointing
    Image3DBackgroundManager *manager;
public:
	/// ask the Image3DViewer to be redrawn
	virtual void redrawViewer()=0;
    /** @name constructor */
	//@{
	/** creates a viewer interface given the background manager : add itself into the
	list of viewer interfaces */
	Image3DViewerInterface(Image3DBackgroundManager *_manager);
	//@}

	/** @name destructor */
	/// destructor:	 delete all backgrounds upon the destruction of the manager
	//@{
	virtual ~Image3DViewerInterface(){}
	//@}
	friend class Image3DBackgroundManager;
};

/** \class Image3DCameraBackground
\ingroup Inrimage
\brief a basic background that define data common between different 3D image backgrounds 
This data is managed by a Multiview */
class Image3DBackgroundManager : public DPListElement<Image3DBackgroundManager>,
	public DPOwner<Image3DBackgroundManager>,public Owner<Inrimage>

{
public:
	/// middle mouse motion action types
	enum MOUSE_MOTION_ACTION {
		MOVE_SLICES,
		CHANGE_WINDOWING
	};
	/// grey level transformations
	enum COLORMAP {
		AFFINE,
		THRESHOLD,
		PET_COLORS,
		JET_COLORS,
		SPECT_COLORS,
		HOTIRON_COLORS,
		RED,
		GREEN,
		BLUE,
		USER
	};
	/// grey level interpolation
	enum IMAGE_INTERPOLATION {
		NEAREST,
		BILINEAR
	};
public:
	/// HotIron colormap red component 
	static unsigned char HotIronColorMapRed[258];
	/// HotIron colormap green component 
	static unsigned char HotIronColorMapGreen[258];
	/// HotIron colormap  blue component 
	static unsigned char HotIronColorMapBlue[258];
	/// SPECT colormap red component 
	static unsigned char SPECTColorMapRed[258];
	/// SPECT colormap green component 
	static unsigned char SPECTColorMapGreen[258];
	/// SPECT colormap  blue component 
	static unsigned char SPECTColorMapBlue[258];
	/// PET colormap red component 
	static unsigned char PETColorMapRed[258];
	/// PET colormap green component 
	static unsigned char PETColorMapGreen[258];
	/// PET colormap  blue component 
	static unsigned char PETColorMapBlue[258];
	/// JET colormap red component 
	static unsigned char JetColorMapRed[258];
	/// JET colormap green component 
	static unsigned char JetColorMapGreen[258];
	/// JET colormap  blue component 
	static unsigned char JetColorMapBlue[258];    
	/// AFFINE colormap 
	static unsigned char AffineColorMap[258];
	/// THRESHOLD colormap 
	static unsigned char ThresholdColorMap[258];
	/// zero colormap 
	static unsigned char ZeroColorMap[258];

protected:
	/// the list of backgrounds that are synchronised by this manager
	ObjectsList<Image3DCameraBackground> backgroundList;
	/// the list of viewer interfaces that need to be redrawn whenever this manager (and his backgrounds) is updated
	ObjectsList<Image3DViewerInterface> viewerList;
	/// whether to display axis around cursor
	bool showAxis;
	/// what to do when the middle mouse button is pushed and the mouse is moved
	MOUSE_MOTION_ACTION mouseAction;
	/// multiview zoom factor
	double zoom;
	/// color name of axis
	float axisColor[3];
	/// record the mouse pixel when tracking mouse motion
	Vec2<double> previousMousePosition;
	/// minimum grey-level for windowing
	double minWindow;
	/// maximum grey-level for windowing
	double maxWindow;
	/// minimum grey value used to define the windowing
	double minGreyLevel;
	/// maximum grey value used to define the windowing
	double maxGreyLevel;
	/// minimum grey-level for windowing
	double minWindow2;
	/// maximum grey-level for windowing
	double maxWindow2;
	/// minimum grey value used to define the windowing
	double minGreyLevel2;
	/// maximum grey value used to define the windowing
	double maxGreyLevel2;
	/// grey level transformation type
	COLORMAP colorMap;
	/// grey level transformation type
	COLORMAP colorMap2;
	/// the red component colormap that depends on transfo
	const unsigned char *redColorMap;
	/// the green component colormap that depends on transfo
	const unsigned char *greenColorMap;
	/// the blue component colormap that depends on transfo
	const unsigned char *blueColorMap;
	/// the red component colormap that depends on transfo
	const unsigned char *redColorMap2;
	/// the green component colormap that depends on transfo
	const unsigned char *greenColorMap2;
	/// the blue component colormap that depends on transfo
	const unsigned char *blueColorMap2;
	/// the second image to be displayed
	ObjectPointer<Inrimage> secondImage;
	/// describes how to interpolate grey values
	IMAGE_INTERPOLATION interpolation;
	/// the opacity between the two images
	double opacity;

public:
	/** @name constructor */
	//@{
	Image3DBackgroundManager(Inrimage *im,Image3DCameraBackground *b=0);
	//@}

	/** @name desstructor */
	/// destructor:	 delete all backgrounds upon the destruction of the manager
	//@{
	virtual ~Image3DBackgroundManager();
	//@}

public:
	/// return handled inrimage
	Inrimage *getInrimage(void) const{
		return pointer();
	}
	/// return handled inrimage
	Inrimage *getSecondImage(void) const{
		return secondImage();
	}
	/// ask all connected viewers to be redrawn
	void redrawViewers();
	/// return the opacity used in compositing
	double getOpacity() const {
		return opacity;
	}
	/// set the opacity used in compositing
	void setOpacity(const double o); 
	/// set the second Image for image fusion
	void setSecondImage(Inrimage *img);
	/// add a background in the background list for synchronization
	void addBackground(Image3DCameraBackground *b);
	/// remove a background from the background list for synchronization
	void removeBackground(Image3DCameraBackground *b);
	/// returns an iterator on the first background
	ObjectsList<Image3DCameraBackground>::iterator firstBackground() {
		return backgroundList.begin();
	}
	/// returns an iterator on the first background
	ObjectsList<Image3DCameraBackground>::iterator lastBackground() {
		return backgroundList.end();
	}
	/// when a slider is moved, inform all synchronized manager that the pointer has moved
	void moveOtherManagers();
	/** requires all backgrounds to display an axis at a given world Position.
	b is the background that generated the event (may be 0) */
	void displayAxis(Image3DCameraBackground *b,
		Vec3<double> &worldPosition);
	/// requires to synchronize this background Manager with this one
	void synchronizeWithManager(Image3DBackgroundManager *manager);
	/// returns the min value used for the windowing
	double getMinWindow() const;
	/// returns the maximum value used for the windowing
	double getMaxWindow() const;
	/// returns the min value possible
	double getMinGreyLevel() const;
	/// returns the maximum value possible 
	double getMaxGreyLevel() const;
	/// returns the min value used for the windowing
	double getMinWindow2() const;
	/// returns the maximum value used for the windowing
	double getMaxWindow2() const;
	/// returns the min value possible
	double getMinGreyLevel2() const;
	/// returns the maximum value possible 
	double getMaxGreyLevel2() const;
	/// update the value of the grey level sliders
	virtual void updateSliders() const{}
	/// set to all backgrounds the  min value used for the windowing
	void setMinWindow(const double minW);
	/// set to all backgrounds the  maximum value used for the windowing
	void setMaxWindow(const double minW);
	/// set to all backgrounds the  min value used for the windowing
	void setMinWindow2(const double minW);
	/// set to all backgrounds the  maximum value used for the windowing
	void setMaxWindow2(const double minW);
	/// returns grey-level transformation mode
	COLORMAP getColormap(void) const { return colorMap; }
	/// returns grey-level transformation mode
	COLORMAP getColormap2(void) const { return colorMap2; }

	/** callback issued by a background informing that the grey-level of the image
	has changed */
	virtual void imageHasChangedTexture(Inrimage *inr);
	/** warn the background that the image has changed it size */
	virtual void imageHasChangedSize(Inrimage *inr);

	/// get the common colormap of all backgrounds
	COLORMAP getColorMap() const;
	/// get the common colormap of all backgrounds
	COLORMAP getColorMap2() const;
	/** set a given colormap for all backgrounds
	@param c the type of colormap
	@param red, green, blue the 3 user-defined components of the colormap if (t==USER) */
	void setColorMap(const COLORMAP c,
		const unsigned char *red=0,
		const  unsigned char *green=0,
		const  unsigned char *blue=0);
	/** set a given colormap for all backgrounds
	@param c the type of colormap
	@param red, green, blue the 3 user-defined components of the colormap if (t==USER) */
	void setColorMap2(const COLORMAP c,
		const unsigned char *red=0,
		const  unsigned char *green=0,
		const  unsigned char *blue=0);
	/** set for all associated background an interpolation method for image resampling*/
	void setInterpolationMethod(const IMAGE_INTERPOLATION interp);
	/** get the interpolation method for image resampling used by all backgrounds*/
	IMAGE_INTERPOLATION getInterpolationMethod() const;
	/// set the boolean controling the axis display when a background has mouse focus 
	void setShowAxis(const bool sa);
	/// get the boolean controling the axis display when a background has mouse focus 
	bool getShowAxis() const;
	/*** change the slice number of all background such that
	a given  point is seen by all backgrounds */ 
	virtual void moveToPoint(Vec3<double> &pos);
	/// get the array of 255 char corresponding to the red colormap
	virtual  const unsigned char *getRedColorMap() const {
		return redColorMap;
	}
	/// get the array of 255 char corresponding to the green colormap
	virtual  const unsigned char *getGreenColorMap() const {
		return greenColorMap;
	}
	/// get the array of 255 char corresponding to the blue colormap
	virtual  const unsigned char *getBlueColorMap() const {
		return blueColorMap;
	}
	/// get the array of 255 char corresponding to the red colormap
	virtual  const unsigned char *getRedColorMap2() const {
		return redColorMap2;
	}
	/// get the array of 255 char corresponding to the green colormap
	virtual  const unsigned char *getGreenColorMap2() const {
		return greenColorMap2;
	}
	/// get the array of 255 char corresponding to the blue colormap
	virtual  const unsigned char *getBlueColorMap2() const {
		return blueColorMap2;
	}
	/// min grey level scale owner callback
	virtual void dying(const ObjectPointer<Inrimage>& obj);
	/** update the label that describes the position and value of
	the voxel pointed by the current mouse position */
	virtual void updateLabel(const char *label);
	/** routine called when a mouse button is pressed and that it is not handle
	for scaling, selecting, rotating or translating the camera or an object */
	virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
		CameraTypes::KEY_MODIFIER modifier,
		Image3DCameraBackground *cam,int x,int y);
	/** routine called when a mouse button is released and that it is not handle
	for scaling, selecting, rotating or translating the camera or an object */
	virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
		CameraTypes::KEY_MODIFIER modifier,
		Image3DCameraBackground *cam,int x,int y);
	/** routine called when a mouse button is moved and that it is not handle
	for scaling, selecting, rotating or translating the camera or an object */
	virtual void mouseButtonMoved(Image3DCameraBackground *cam,int x,int y	);

	friend class Image3DViewerInterface;
}; 

} // namespace yav


#endif
