#ifndef IMAGE3DCAMERABACKGROUND_H
#define IMAGE3DCAMERABACKGROUND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/Image3DCameraBackground.h"
#include <inrimage/Inrimage.h>
#include <graphics2D/Camera2D.h>
#include <framework/SliceOrientation.h>

namespace yav {


	class Image3DBackgroundManager;
	class Image3DTextureInformation;

	/** \class Image3DCameraBackground
	\ingroup Inrimage
	\brief  background that draws a slice of a 3D image as an OpenGL Texture */
	class Image3DCameraBackground : public Camera2DBackground,
		public ListElement<Image3DCameraBackground>,
		public AutoLink<Image3DCameraBackground>
	{
	public:
		public:

		/** @name constructor */
		//@{
		/** a image background that is created by an inherited class
		@param cam the 2D camera associated with the background
		@param manager the background manager
		@param geom the slice geometry */
		Image3DCameraBackground(Camera2D *cam,Image3DBackgroundManager *manager,SLICE_ORIENTATION geom);

		/** @name destructors */
		//@{
		/// destructor: free buffers
		virtual ~Image3DCameraBackground();
		//@}
		


	protected:
		/// points on the manager that synchronizes the different views
		Image3DBackgroundManager *manager;
		/// the name of the background type
		TclString sliceName;
		/// the label associated with the current pointed voxel
		TclString voxelLabel;
		/// show the axis  around the current mouse position
		bool showAxis; 
		/// where to center the two axis
		Vec2<double> currentMousePosition;
				/// if the display list should be update
		bool updateDisplayList;
		/// image visible area
		unsigned int visible[6];
		/// slice geometry
		SLICE_ORIENTATION geometry;
		/// slice Number
		unsigned int sliceNumber;
		/// whether vectors should be drawn in case of a vectorial image
		bool showVectors;
		/** whether arrows or simple sticks (line segments) should be drawn for each vector */
		bool showArrows;
		/// the color of each vector
		float vectorColor[3];
		/// the amount of subsampling in the rows and columns direction when drawing vectors
		unsigned int vectorXYStep;
		/// the amount of subsampling in the Z direction when drawing vectors
		unsigned int vectorZStep;
		/// the vector display list ID
		GLuint vectorId;
		/// the scale used to draw each vector;
		double vectorScale;
		/// the width of arrow as a percentage of its length 
		double arrowWidth;
		/// the height of arrows as a percentage of its length 
		double arrowHeight;
		/// whether the current vector display list is valid or not
		bool recomputeVectorDisplayList;
		/// texture information on a main intensity image
		Image3DTextureInformation *mainTexture; 
		/// texture information on a second intensity image (may be null)
		Image3DTextureInformation *secondTexture; 
	protected:
		/// build if necessary the display list associated with vectors and draw it on screen
		void callVectorDisplayList();
		/// remove the display list associated with vectors (for vectorial images)
		void destroyVectorDisplayList();
		/** set the second image to overlay on the first image. If null 
		it implies that the second Image is removed */
		virtual void setSecondImage(Inrimage *inr);
		/** computes the slice number of the second texture based on the slice number
		of the main texture and some offset information */
		void setSecondSliceNumber();
	public:
		/// return the background manager
		Image3DBackgroundManager *getManager() const { return manager;}
		/// returns the texture object corresponding to the main texture
		Image3DTextureInformation *getMainTexture() const {
			return mainTexture;
		}
		/// returns the texture object corresponding to the second texture
		Image3DTextureInformation *getSecondTexture() const {
			return secondTexture;
		}
		/** routine called when a mouse button is pressed and that it is not handle
		for scaling, selecting, rotating or translating the camera or an object */
		virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
			CameraTypes::KEY_MODIFIER modifier,
			Camera2D *cam,int x,int y);
		/** routine called when a mouse button is released and that it is not handle
		for scaling, selecting, rotating or translating the camera or an object */
		virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
			CameraTypes::KEY_MODIFIER modifier,
			Camera2D *cam,int x,int y);
		/** routine called when a mouse button is moved and that it is not handle
		for scaling, selecting, rotating or translating the camera or an object */
		virtual void mouseButtonMoved(Camera2D *cam,int x,int y	);
	
	
		/// returns the main image to be displayed
		Inrimage *getInrimage() const;
		/** set visible area to given bounds (truncate to image size if needed)
		@param x1 visible area X first coordinate
		@param y1 visible area Y first coordinate
		@param z1 visible area Z first coordinate
		@param x2 visible area X second coordinate
		@param y2 visible area Y second coordinate
		@param z2 visible area Z second coordinate */
		// virtual void setVisibleArea(unsigned int x1, unsigned int y1, unsigned int z1,
		//			      unsigned int x2, unsigned int y2, unsigned int z2);

		/// return the slice geometry of the image
		SLICE_ORIENTATION getGeometry() const {return geometry;}

		/// return visible area
		virtual const unsigned int *getVisibleArea(void) const { return visible; }

		/// returns the boolean that indicates if vectors should be drawn with arrows or not 
		bool getShowArrows() const {
			return showArrows;
		}
		/// sets the boolean that indicates if vectors should be drawn with arrows or not 
		void setShowArrows(const bool b)  {
			showArrows=b;
			recomputeVectorDisplayList=true;
		}
		/// returns the arrow width
		double getArrowWidth() const {
			return arrowWidth;
		}
		/// sets the arrow width 
		void setArrowWidth(const double w) {
			arrowWidth=w;
			recomputeVectorDisplayList=true;
		}
		/// returns the arrow height
		double getArrowHeight() const {
			return arrowHeight;
		}
		/// sets the arrow height
		void setArrowHeight(const double w) {
			arrowHeight=w;
			recomputeVectorDisplayList=true;
		}
        /// returns the boolean that indicates if vectors should be drawn on vectorial images
		bool getShowVectors() const {
			return showVectors;
		}
		/// sets the boolean that indicates if vectors should be drawn on vectorial images
		void setShowVectors(const bool b)  {
			showVectors=b;
		}
		/// gets the scale at which each vector is drawn
		double getVectorScale() const {
			return vectorScale;
		}
		/// set the scale at which each vector is drawn
		void setVectorScale(const double vs) {
			vectorScale=vs;
			recomputeVectorDisplayList=true;
		}
		/// returns the rgb components used to draw each vector
		float getVectorColor(const unsigned i) const {
			if (i<3)
				return(vectorColor[i]);
			else
				return (0.0);
		}
		/// sets the rgb components used to draw each vector
		void setVectorColor(const float r, const float g, const float b) {
			vectorColor[0]=r;
			vectorColor[1]=g;
			vectorColor[2]=b;
			recomputeVectorDisplayList=true;
		}
		/// returns the amount of subsampling in the rows and columns direction when drawing vectors
		unsigned int getVectorXYStep() const {
			return vectorXYStep;
		}
		/// sets the amount of subsampling in the rows and columns direction when drawing vectors
		void setVectorXYStep(const unsigned int step) {
			vectorXYStep=step;
			recomputeVectorDisplayList=true;
		}
		/// returns the amount of subsampling in the Z direction when drawing vectors
		unsigned int getVectorZStep() const {
			return vectorZStep;
		}
		/// sets the amount of subsampling in the Z direction when drawing vectors
		void setVectorZStep(const unsigned int step) {
			vectorZStep=step;
			recomputeVectorDisplayList=true;
		}


		/// stop using this texture
		void disableTexture() const {
			glDisable(GL_TEXTURE_2D);
		}
		/** redraw camera image */
		virtual void redraw();

		/** warn the background that the image has changed it size */
		virtual void imageHasChangedSize(Inrimage *inr);
		/** warn the background that the image has changed it content (but not its size) */
		virtual void imageHasChangedTexture(Inrimage *inr);


		/** set grey-level transformation mode
		@param t new transformation mode
		@param red, green, blue		*/
		void setColormap();

		/** redraw when the opacity has changed*/
		void setOpacity();
		/*** returns  a label associated with current mouse position */
		virtual const char *getLabel(const Vec2<double> &pt);
		/*** transforms 2D plane coordinates into image coordinates. The image coordinates
		may have 4 components (x,y,z and time) for 4D images
		@param pt an input 2D point in plane coordinates 
		@param pixel output the pixel coordinates taking into account pixel size. The pixel
		coordinates may have negative coordinates */
		virtual void map2DWorldToImageCoordinates(const Vec2<double> &pt, Vec4<double> &pixel);
		/*** transforms image coordinates into 2D plane coordinates The image coordinates
		may have 4 components (x,y,z and time) for 4D images
		@param pixel output the pixel coordinates taking into account pixel size. The pixel
		coordinates may have negative coordinates 
		@param pt an input 2D point in plane coordinates 
		@return true if the voxel is in the slice plane */
		virtual bool mapImageTo2DWorldCoordinates(const Vec4<double> &pixel,Vec2<double> &pt );
		/*** change the slice number as to go through a known point*/ 
		virtual void moveToPoint(Vec3<double> &pos);
		/** defines the two corners of the camera viewport 
		(corresponding to the image size) */
		virtual void getViewportSize(Vec2<double> &vmin,Vec2<double> &vmax);
		/** returns the normal vector and a point lying on the image plane
		@param org the plane origin (output)
		@param normal the plane normal (output)*/
		virtual void getImagePlane(Vec3<double> &org, Vec3<double> &normal);
		/*** transforms 3D coordinates into plane coordinates 
		@param p an input 3D point lying in the image plane
		@param pt output the associated 2D coordinates in the image reference frame 
		@return returns true if the 3D lies effectively on the image plane */
		virtual bool map3DWorldTo2DWorldCoordinates(const Vec3<double> &p, Vec2<double> &pt);
		/*** transforms 2D plane coordinates into 3D world coordinates. 
		@param pt an input 2D point in plane coordinates 
		@param p output the 3D world coordinates taking into account the plane orientation and pixel size */
		virtual void map2DWorldTo3DWorldCoordinates(const Vec2<double> &pt, Vec3<double> &p);
		/*** transforms 2D plane coordinates into 3D world coordinates without rotation and translation
		@param pt an input 2D point in plane coordinates 
		@param p output the 3D world coordinates taking into account the plane orientation and pixel size */
		virtual void map2DWorldToRaw3DWorldCoordinates(const Vec2<double> &pt, Vec3<double> &p);
		/*** tests if a point in 3D world coordinates belongs to the plane of this background */
		virtual bool isPointInSlice(const Vec3<double> &p);
		/*** provide the slice name to characterize the slice type */
		virtual const char *getSliceName() const{
			return sliceName;
		}
		/*** returns the slice number if applicable */
		virtual int getSliceNumber() const{
			return sliceNumber;
		}
		/*** stop displaying axis around mouse */
		void hideAxis() {
			showAxis=false;
		}
		/** return the width of the main texture in world coordinates */ 
		double getRequiredWidth() const;
		/** return the width of the main texture in world coordinates */ 
		double getRequiredHeight() const;
		/** returns (vr+vc)/2 as an average pixel size */ 
		double getPixelSize() const;
		/*** change the slice Number and recomputes the texture */
		virtual void setSliceNumber(unsigned int s);
		/** get object string identifier : useful to recognize background without knowing its type
		@return object name */
		virtual TclString getName(void) const {
			return ("dumnmySlice");
		}

		friend class Image3DBackgroundManager;
	}; 

	/** the structure that stores all necessary information to display an additional image
	on top of the current background */
	class Image3DTextureInformation {
	public:
		/** @name constructor */
		//@{
		/** a image background that is created by an inherited class
		@param cam the 2D camera associated with the background
		@param manager the background manager
		@param geom the slice geometry */
		Image3DTextureInformation(Inrimage*inr,Image3DCameraBackground *b);
		/** @name destructors */
		//@{
		/// destructor: free buffers
		virtual ~Image3DTextureInformation();
		//@}
	protected:
		/// the background that uses this texture
		Image3DCameraBackground *background;
		/// the ratio between the interslice distance of the primal image with this image
		double sliceRatio;
		/// pixel size in the row direction
		double vr;
		/// pixel size in the column direction
		double vc;
		/// if the texture has been binded
		bool binded;
		/// the texture ID
		GLuint id;
		/// texture buffer (RGBA)
		unsigned int *buffer;
		/// number of rows: must be a power of two
		unsigned int nrows;
		/// number of columns: must be a power of two
		unsigned int ncols;
		/// the number of rows that are really meaningful
		unsigned int nbRealRows;
		/// the number of columns that are really meaningful
		unsigned int nbRealColumns;
		/// slice Number
		unsigned int sliceNumber;
		/// maximum slice number
		unsigned int maxSlice;
		/// offset in the x direction for the second Texture
		double offsetx;
		/// offset in the y direction for the secondTexture
		double offsety;
		/// whether the texture should be displayed in the background
		bool insideBackground;
		/// store the norm of the slice when displaying vectorial images
		double **normVector;
		/// delete the array that contains the L2 norm of each vector 
		void emptyNormVector();

	public:
		/// return the associated background
		Image3DCameraBackground *getBackground() const { return background;}
		/// update the OpenGL texture based on the new texture buffer
		void updateTexture() const;
		/// actually draws the image slice in the RGBA buffer
		void writeTexture(const Inrimage *inr,
			const unsigned char *redColorMap,
			const unsigned char *greenColorMap,
			const unsigned char *blueColorMap,
			const double minWindow,
			const double maxWindow,
			const double opacity);
		/// create the display list associated with the texture
		void createTexture();
		/// initialize variables
		void init(Inrimage *);
		/// whether the texture should be displayed in the background
		bool isInsideBackground() const {
			return insideBackground;
		}
		/// fill the texture of slice Z=cste
		void fillSliceZTexture(const Inrimage *inr,
			const unsigned char *redColorMap,
			const unsigned char *greenColorMap,
			const unsigned char *blueColorMap,
			const double minWindow,
			const double maxWindow,
			const double opacity);
		/// fill the texture of slice X=cste
		void fillSliceXTexture(const Inrimage *inr,
			const unsigned char *redColorMap,
			const unsigned char *greenColorMap,
			const unsigned char *blueColorMap,
			const double minWindow,
			const double maxWindow,
			const double opacity);
		/// fill the texture of slice Y=cste
		void fillSliceYTexture(const Inrimage *inr,
			const unsigned char *redColorMap,
			const unsigned char *greenColorMap,
			const unsigned char *blueColorMap,
			const double minWindow,
			const double maxWindow,
			const double opacity);
		/// return the nb of texture rows 
		unsigned int getNbRows() const { return nrows;}
		/// return the nb of texture columns
		unsigned int getNbColumns() const { return ncols;}
		/// return the nb of meaningful texture rows 
		unsigned int getNbRealRows() const { return nbRealRows;}
		/// return the nb of meaningful texture columns
		unsigned int getNbRealColumns() const { return nbRealColumns;}
		/// gets the space between two rows
		double getVr() const {return vr;}
		/// gets the space between two columns
		double getVc() const {return vc;}
		/// use this texture
		void enableTexture(const Inrimage *inr,
			const unsigned char *redColorMap,
			const unsigned char *greenColorMap,
			const unsigned char *blueColorMap,
			const double minWindow,
			const double maxWindow,
			const double opacity);
		/// use OpenGL to draw vectors on top of the slice background
		void drawVectors();
		/*** returns the slice number if applicable */
		virtual int getSliceNumber() const{
			return sliceNumber;
		}
		/// draw the textured quad 
		void drawTexture() const;
        /// draw the textured quad with 3D coordinates
		void drawTexture3D() const;
		friend class Image3DCameraBackground;
	};


} // namespace yav


#endif
