#ifndef IMAGEGEOMETRY_H
#define IMAGEGEOMETRY_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/ImageGeometry.h"
#include <inrimage/Inrimage.h>
#include <inrimage/Voxel.h>

namespace yav {


/** \class ImageGeometry
    \ingroup Inrimage
    \brief Base class for image geometry conversion */
class ImageGeometry {
 public:
  /// equals to zero
  static const double epsilon;

 protected:
  /// host image
  const Inrimage& inr;
  /// associated voxel
  Voxel *voxel;

 public:
  /** @name constructors */
  //@{
  /** builds an ImageGeometry parent object
      @param _inr studied inrimage */
  ImageGeometry(const Inrimage& _inr)
    : inr(_inr), voxel(Voxel::voxelByType(inr.getType())) {}
  /// copy constructor
  ImageGeometry(const ImageGeometry& imgg)
    : inr(imgg.inr), voxel(Voxel::voxelByType(inr.getType())) {}
  //@}

  /** @name destructor */
  //@{
  /** destroy object and allocated voxel */
  virtual ~ImageGeometry() { delete voxel; }
  //@}

  /// returns host inrimage
  const Inrimage& getInrimage(void) const { return inr; }
  /// returns voxel value
  Voxel& getVoxel(void) const { return *voxel; }

  /** assign new coordinate
      @param x coordinate value in cartesian space
      @param y coordinate value in cartesian space
      @param z coordinate value in cartesian space */
  virtual void set(double x, double y, double z) = 0;
  /** assign new coordinate from Image coordinates
      @param x coordinate value in image coordinate
      @param y coordinate value in image coordinate
      @param z coordinate value in image coordinate */
  virtual void setImage(double x, double y, double z) = 0;
  /// returns true if this coordinate falls inside image bounds, false otherwise
  virtual bool in(void) const = 0;
  /** converts coordinate in this geometry to cartesian coordinate
      @param pos 3 dimension double array to store cartesian coordinates */
  virtual  void toCartesian(double *pos) const = 0;
  /** converts coordinate in this geometry to image voxel position
      @param pos 3 dimension integer array to store voxel coordinate */
  virtual void toImage(int *pos) const = 0;
  /// returns tri-linearly interpolated greylevel value
  virtual double interpolate(void) const = 0;
  /** compute next image voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextVoxel(const Vec3<double>& direction) = 0;
    /** compute the first image border voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextBorderVoxel(const Vec3<double>& direction) = 0;
};





/** \class CartesianGeometry
    \ingroup Inrimage
    \brief cartesian geometry conversion */
class CartesianGeometry : public ImageGeometry {
 protected:
  /// x coordinate
  double x;
  /// y coordinate
  double y;
  /// z coordinate
  double z;
  /// image coordinate
  int ic[3];

 public:
  /** @name constructors */
  //@{
  /** builds parent, let this uninitialized (call set method)
      @param inr studied inrimage */
  CartesianGeometry(const Inrimage& inr) : ImageGeometry(inr) {
  }
  
  /// copy constructor
  CartesianGeometry(const CartesianGeometry& c)
    : ImageGeometry(c), x(c.x), y(c.y), z(c.z) {
  }
  //@}
  
  /** @name destructor */
  //@{
  virtual ~CartesianGeometry() {}
  //@}
  
  /** assign new coordinate
      @param x coordinate value in cartesian space
      @param y coordinate value in cartesian space
      @param z coordinate value in cartesian space */
  virtual void set(double x, double y, double z);
  /** assign new coordinate from Image coordinates
      @param x coordinate value in image coordinate
      @param y coordinate value in image coordinate
      @param z coordinate value in image coordinate */
  virtual void setImage(double x, double y, double z);
  /// returns true if this coordinate falls inside image bounds, false otherwise
  virtual bool in(void) const;
  /** converts coordinate in this geometry to cartesian coordinate
      @param pos 3 dimension double array to store cartesian coordinates */
  virtual  void toCartesian(double *pos) const {
    pos[0] = x; pos[1] = y; pos[2] = z;
  }
  /** converts coordinate in this geometry to image voxel position
      @param pos 3 dimension integer array to store voxel coordinate */
  virtual void toImage(int *pos) const;
  /// returns tri-linearly interpolated greylevel value
  virtual double interpolate(void) const;
  /** compute next image voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextVoxel(const Vec3<double>& direction);
      /** compute the first image border voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextBorderVoxel(const Vec3<double>& direction);
};




/** \class CylindricalGeometry
    \ingroup Inrimage
    \brief cylindrical geometry conversion */
class CylindricalGeometry : public ImageGeometry {
 protected:
  /// XDIM * VX / 2 constant
  double x0;
  /// radius
  double r;
  /// angle
  double theta;
  /// elevation
  double z;

 public:
  /** @name constructors */
  //@{
  /** builds parent, let this uninitialized (call set method)
      @param inr studied inrimage */
  CylindricalGeometry(const Inrimage& inr)
    : ImageGeometry(inr), x0(inr.getX() * inr.getVX() / 2.0) {
  }
  
  /// copy constructor
  CylindricalGeometry(const CylindricalGeometry& c)
    : ImageGeometry(c), x0(c.x0), r(c.r), theta(c.theta), z(c.z) {
  }
  //@}
  
  /** @name destructor */
  //@{
  virtual ~CylindricalGeometry() {}
  //@}
  
  /// returns radius
  double getRadius(void) const { return r; }
  /// returns angle
  double getAngle(void) const { return theta; }
  /// returns elevation
  double getZ(void) const { return z; }
  /// returns XDIM * VX / 2 constant
  double getX0(void) const { return x0; }
  /** assign new coordinate from Image coordinates
      @param x coordinate value in image coordinate
      @param y coordinate value in image coordinate
      @param z coordinate value in image coordinate */
  virtual void setImage(double x, double y, double z);

  /** assign new coordinate
      @param x coordinate value in cartesian space
      @param y coordinate value in cartesian space
      @param z coordinate value in cartesian space */
  virtual void set(double x, double y, double z);
  /// returns true if this coordinate falls inside image bounds, false otherwise
  virtual bool in(void) const;
  /** converts coordinate in this geometry to cartesian coordinate
      @param pos 3 dimension double array to store cartesian coordinates */
  virtual  void toCartesian(double *pos) const;
  /** converts coordinate in this geometry to image voxel position
      @param pos 3 dimension integer array to store voxel coordinate */
  virtual void toImage(int *pos) const;
  /// returns tri-linearly interpolated greylevel value
  virtual double interpolate(void) const;
  /** compute next image voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextVoxel(const Vec3<double>& direction);
      /** compute the first image border voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextBorderVoxel(const Vec3<double>& direction);


 private:
  // compute partial volume for tri-linear interpolation
  double volume(double ra, double rb, double dtheta, double dz) const;
};





/** \class SphericalGeometry
    \ingroup Inrimage
    \brief spherical geometry conversion */
class SphericalGeometry : public ImageGeometry {
 protected:
  double alpha0;
  /// XDIM * VX * sin(alpha0/2)
  double x0;
  double r;
  /// alpha : polar angle in the considered slice [0,alpha0]
  double alpha;
  /// theta : angle between the Y axis and the considered slice [0,Pi[
  double theta;

 public:
  /** @name constructors */
  //@{
  /** builds parent, let this uninitialized (call set method)
      @param inr studied inrimage */
  SphericalGeometry(const Inrimage& inr) : ImageGeometry(inr),x0(inr.getX()*inr.getVX()*sin(alpha0/2)) {
  }
  
  /// copy constructor
  SphericalGeometry(const SphericalGeometry& c)
    : ImageGeometry(c) {
  }
  //@}
  
  /** @name destructor */
  //@{
  virtual ~SphericalGeometry() {}
  //@}

  /// returns radius
  double getRadius(void) const {return r; }
  /// returns alpha
  double getAlpha(void) const {return alpha; }
  /// returns theta
  double getTheta(void) const {return theta; }
  /// returns x0
  //  double getX0(alpha0) const {return x0;}

  /** assign new coordinate from Image coordinates
      @param x coordinate value in image coordinate
      @param y coordinate value in image coordinate
      @param z coordinate value in image coordinate */
  virtual void setImage(double x, double y, double z);

  /** assign new coordinate
      @param x coordinate value in cartesian space
      @param y coordinate value in cartesian space
      @param z coordinate value in cartesian space */
  virtual void set(double x, double y, double z);
  /// returns true if this coordinate falls inside image bounds, false otherwise
  virtual bool in(void) const;
  /** converts coordinate in this geometry to cartesian coordinate
      @param pos 3 dimension double array to store cartesian coordinates */
  virtual  void toCartesian(double *pos) const;
  /** converts coordinate in this geometry to image voxel position
      @param pos 3 dimension integer array to store voxel coordinate */
  virtual void toImage(int *pos) const;
  /// returns tri-linearly interpolated greylevel value
  virtual double interpolate(void) const;
  /** compute next image voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextVoxel(const Vec3<double>& direction);
      /** compute the first image border voxel intersected by a line passing through
      this line and with given direction vector
      @param direction line direction vector
      @return distance to next voxel */
  virtual double nextBorderVoxel(const Vec3<double>& direction);
};


}


#endif
