/***************************************************************************/
/* InrChamferMap.h                                                         */
/***************************************************************************/
#ifndef INRCHAMFERMAP_H
#define INRCHAMFERMAP_H

#ifndef _YAV_
#define _YAV_
#endif

#include <inrimage/Inrimage.h>

namespace yav {

  ///
  enum { // option flags
    NO_PROPAGATION,
    COLOR_PROPAGATION,
    POSITION_PROPAGATION,
  };


/** \class InrChamferMap
    \ingroup Inrimage
*/
class InrChamferMap : public Inrimage
{
public :
  /** @name Constructors */
  //@{
  /** @param src      : binary image to compute chamfer map
      @param location : where to compute chamfer map : 
      @param location  = 0 -> computes the chamfer map of the background (data values <= 0)
      @param location  = 1 -> computes the chamfer map of the foreground (data values > 0)
      @param location  = 2 -> computes the chamfer map of both the background and the object
      (chamfer map of the background will be negative, chamfer of the foreground will be positive)
      @param maskSize : size of the mask. 
      For isotropic images, 3x3x3, 5x5x5, and 7x7x7 masks are pre-calculated
      For anisotropic images, only 3x3x3 mask is available
      @param maskDim : maskDim = 2 -> computes distance map slice by slice
                       maskDim = 3 -> computes 3D chamfer map
  */
  InrChamferMap(Inrimage * src,
		int location = 1,
		int maskSize = 3,
		int maskDim  = 3);
/** @param src      : binary image to compute chamfer map
    @param output   : image where the propagation is stored. It must be initialized (should be a copy of src)
    @param flag = NO_PROPAGATION    -> output remains unchanged
    @param flag = COLOR_PROPAGATION -> 
    for the foreground, propagates the value of the nearest background point for each point of the foreground
    for the background, propagates the value of the nearest foreground point for each point of the background
    @param flag = POSITION_PROPAGATION -> 
    for the foreground, propagates the position (x + dims[0]*y + dims[0]*dims[1]*z) of the nearest point of the background
    for the background, propagates the position (x + dims[0]*y + dims[0]*dims[1]*z) of the nearest point of the foreground
    for both foreground and background, propagates the positions, positive for the foreground, negative for the background
    @param location : where to compute chamfer map : 
    @param location  = 0 -> computes the chamfer map of the background (data values <= 0)
    @param location  = 1 -> computes the chamfer map of the foreground (data values > 0)
    @param location  = 2 -> computes the chamfer map of both the background and the object
    (chamfer map of the background will be negative, chamfer of the foreground will be positive)
    @param maskSize : size of the mask. 
    For isotropic images, 3x3x3, 5x5x5, and 7x7x7 masks are pre-calculated
    For anisotropic images, only 3x3x3 mask is available
    @param maskDim : maskDim = 2 -> computes distance map slice by slice
    maskDim = 3 -> computes 3D chamfer map
*/
 InrChamferMap(Inrimage * src,
		Inrimage * output,
		int flag = POSITION_PROPAGATION,
		int location = 1,
		int maskSize = 3,
		int maskDim  = 3);

  //@}

  /** @name Destructor */
  //@{
  ///
  ~InrChamferMap();
  //@}

  /** @name Public Methods */
  //@{
  ///
  void exactMap();
  ///
  float getExactValue(int pos);
  ///
  float getExactValue(double x, double y, double z);
  ///
  float getMaxRelativeError();
  //@}

protected :
  /** @name Protected Fields */
  //@{
  ///
  WORD_TYPE chamferType;

  ///
  void * myChamferMap;

  ///
  bool isExact;
  //@}


};

} // namespace yav
  

  

#endif // INRCHAMFERMAP_H
