#ifndef INRIMAGE_H
#define INRIMAGE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif

#pragma interface "inrimage/Inrimage.h"

#include <basics/Exception.h>
#include <basics/TclString.h>
#include <linearAlgebra/Vec2.h>
#include <linearAlgebra/Vec3.h>

#include <inrimage/BaseInrimage.h>
#include <basics/GenericHistogram.h>
#include <basics/Pointed.h>
#include <framework/SliceOrientation.h>
#include <inrimage/fparser.h>


namespace yav {

#define DEFAULT_SPHERE_FREQUENCY 3

class Voxel;
class VoxelIteratorObject;
class Voxel8u;
struct Interpolator;
struct Extrapolator; 
class ImageGeometry;
class InrimageView;
class Inrimage;


typedef VoxelIteratorObject& VoxelIterator;

/** \class Inrimage
    \ingroup Inrimage
    \brief Inrimages */
class Inrimage : public BaseInrimage, public Pointed<Inrimage>
{
 public:

  /** @name Exception classes */
  //@{
/** \class UnknownType
    \ingroup Inrimage
    \brief unknown inrimage type */
  class UnknownType : public Exception {
  public:
    UnknownType(const std::string& from);
  };
  
  /** \class InvalidArguments
    \ingroup Inrimage
    \brief invalid arguments transmitted to method */
  class InvalidArguments : public Exception {
  public:
    InvalidArguments(const std::string& from,
		     const std::string& error = "invalid arguments transmitted while calling this method");
  };

  /** \class InvalidCompressType
      \ingroup Inrimage
      \brief attempt to write image in an invalid compress type */
  class InvalidCompressType : public Exception {
     public:
	InvalidCompressType(const std::string& from);
  };
    
  /** \class MissingCompressInfo
      \ingroup Inrimage
      \brief missing compression information in header */
  class MissingCompressInfo : public Exception {
     public:
	MissingCompressInfo(const std::string& from);
  };
    
  //@}

  /** different handled inrimage types */
  enum WORD_TYPE {
    /// unsigned 8 bits
    WT_UNSIGNED_CHAR,
    /// signed 8 bits
    WT_SIGNED_CHAR,
    /// unsigned 16 bits
    WT_UNSIGNED_SHORT,
    /// signed 16 bits
    WT_SIGNED_SHORT,
    /// unsigned 32 bits
    WT_UNSIGNED_INT,
    /// signed 32 bits
    WT_SIGNED_INT, 
    /// unsigned 64 bits
    WT_UNSIGNED_LONG,
    /// signed 64 bits
    WT_SIGNED_LONG,
    /// float 32 bits
    WT_FLOAT,
    /// float 64 bits
    WT_DOUBLE,
    /// R, G, B, 8 bits each
    WT_RGB,
    /// R, G, B, A, 8 bits each
    WT_RGBA,
    /// Vector of 32 bit floats
    WT_FLOAT_VECTOR,
    /// Vector of 64 bit floats
    WT_DOUBLE_VECTOR
  };

 
  /// the different image kinds
  enum GEOMETRY {
    /// standard 3D image
    CARTESIAN,
    /// cylindrical geometry (Z = plane angle, Y = elevation, Z = in plane radius)
    CYLINDRICAL,
    /// spherical geometry (Z = plane angle, Y = in plane angle, X = radius)
    SPHERICAL
  };

  /// Compression type for writing
  enum CompressType {
     /// write using 8bits/voxel
     CT_8BITS,
     /// write using 16bits/voxel
     CT_16BITS
  };

 public:

  /** @name Constructors */
  //@{
  /** Reads inrimage from file whose path is given as argument.
      File can be an INRIMAGE(4) gzipped or not, an anlayse, a GIF or a RGB
      irix file. If image has a vectorial type, it can be read
      interlaced (default) or non interlaced by giving
      WM_NON_INTERLACED as third argument.
      @param name image file name
      @param headerOnly if true, only image header is read (data is set to 0)
      @param vm vectorial mode interleave
      @param filevm read file vectorial mode interleave
      @exception ImageNotFound image file not found
      @exception UnknownImageType unknown image file format
      @exception InvalidHeader invalid inrimage header encountered
      @exception InvalidData invalid inrimage data encountered
      @exception UnknownType unsupported inrimage type */
  Inrimage(const TclString& name,  bool headerOnly = false, 
	   VECTORIAL_MODE vm = VM_INTERLACED,
	   VECTORIAL_MODE filevm = VM_INTERLACED)
    throw(ImageNotFound, UnknownImageType,
	  InvalidHeader, InvalidData,
	  UnknownType, InvalidVoxelSize);

  /** Creates an inrimage with given size and word type.
      The inrimage data buffer is allocated
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize image dimension should be > 0
      @exception UnknownType unsupported inrimage type 
      @exception NotEnoughMemory unable to allocate memory to store image
      @exception InvalidVoxelSize voxels size should be > 0.0 */  
  Inrimage(unsigned int xdim,  unsigned int ydim,  unsigned int zdim,
	   unsigned int vdim = 1, unsigned int wdim = 1,
	   WORD_KIND wordKind = WK_FIXED, SIGN sign = SGN_SIGNED,
	   float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(InvalidSize, UnknownType,
	  NotEnoughMemory, InvalidVoxelSize);
  
  /** Creates an inrimage with given size and word type.
      If data is 0, the image is not allocated.
      If data is non 0, image data points to it.
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param data image data buffer
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize image dimension should be > 0
      @exception UnknownType unsupported inrimage type
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  Inrimage(unsigned int xdim, unsigned int ydim, unsigned int zdim,
	   void *data,  unsigned int vdim = 1, unsigned int wdim = 1,
	   WORD_KIND wordKind = WK_FIXED, SIGN sign = SGN_SIGNED,
	   float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(InvalidSize, UnknownType, InvalidVoxelSize);
  
  /** Creates an inrimage from its size and its type
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param wt inrimage word type
      @param vdim vectorial dimension (for WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR only, ignored otherwise)
      @param vm vectorial mode (for WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR only, ignored otherwise)
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize image dimension should be > 0
      @exception UnknownType unsupported inrimage type
      @exception NotEnoughMemory unable to allocate image buffer
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  Inrimage(unsigned int xdim, unsigned int ydim, unsigned int zdim,
	   WORD_TYPE wt, unsigned int vdim = 1,
	   VECTORIAL_MODE vm = VM_INTERLACED,
	   float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(InvalidSize, UnknownType,
	  NotEnoughMemory, InvalidVoxelSize);


  /** Creates a vectorial inrimage from scalar inrimages
      @param n number of inrimages
      @param inrs inrimages array
      @param wt type (only WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR)
      @param vm vectorial mode
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize image dimension should be > 0
      @exception InvalidArguments there should be at least one image and wt should be WT_FLOAT_VECTOR or WT_DOUBLE_VECTOR
      @exception NotEnoughMemory unable to allocate image buffer
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  Inrimage(unsigned int n,  Inrimage **inrs,
	   WORD_TYPE wt = WT_FLOAT_VECTOR, VECTORIAL_MODE vm = VM_INTERLACED,
	   float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(InvalidSize, InvalidArguments,
	  NotEnoughMemory, InvalidVoxelSize);

  /** Creates an replica of img.
      @param img image to copy
      @exception InvalidSize image dimension should be > 0
      @exception InvalidVectorialMode img has scalar vectorial mode for vectorial image or opposite
      @exception InvalidVoxelSize img has <= 0 voxel size
      @exception NotEnoughMemory unable to allocate image buffer */
  Inrimage(const Inrimage& img)
     throw(InvalidSize, InvalidVectorialMode,
	   InvalidVoxelSize, NotEnoughMemory);
  //@}

  /** @name Destructor */
  //@{
  /// deallocates image buffer
  virtual ~Inrimage();
  //@}

  /** Make *this a replica of img.
      @param img image to copy
      @exception InvalidSize image dimension should be > 0
      @exception InvalidVectorialMode img has scalar vectorial mode for vectorial image or opposite
      @exception InvalidVoxelSize img has <= 0 voxel size
      @exception NotEnoughMemory unable to allocate image buffer */
  Inrimage&
  operator=(const Inrimage& img)
     throw(Inrimage::InvalidSize, Inrimage::InvalidVectorialMode,
	   Inrimage::InvalidVoxelSize, Inrimage::NotEnoughMemory);


  /** copy header information (like position, orientation, voxel size,...) 
      from image inr
      @param inr the image from which the header is copied */
  void copyHeader(const Inrimage *inr);

  /** Changes image size. If image data buffer existed it is fried
      and reallocated.
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind
      @param sign
      @exception InvalidSize image dimension should be > 0
      @exception NotEnoughMemory unable to allocate image buffer */      
  virtual void resize(unsigned int xdim,  unsigned int ydim,
		      unsigned int zdim, unsigned int vdim = 1,
		      unsigned int wdim = 1,
		      WORD_KIND wordKind = WK_FIXED, SIGN sign = SGN_SIGNED)
		      throw(Inrimage::InvalidSize, Inrimage::NotEnoughMemory);

  /// returns number of columns
  unsigned int ncols(void) const { return xdim; }
  /// returns number of rows
  unsigned int nrows(void) const { return ydim; }
  // returns number of planes
  unsigned int nplanes(void) const { return zdim; }
   /// get inrimage file name
  const TclString& getName(void) const { return name; } 

  /** set inrimage file name
      @param _name new file name */
  void setName(const TclString& _name) {
    if(_name != EMPTY_STRING) name = _name;
  }
  /// returns inrimage type
  WORD_TYPE getType(void) const { return type; }
  /** set inrimage type
      @param type new image type
      @param vdim vectorial dimension (for WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR types only)
      @param vm
      @param resize whether image should be resized (and buffer reallocated to new image size) */
  void setType(WORD_TYPE type, unsigned int vdim = 1,
	       VECTORIAL_MODE vm = VM_INTERLACED, bool resize = true);


  /// returns pointer a 3D buffer for a scalar image or interleaved vectorial image
  void ***getArray(void) const { return array.threeD; }
  /// returns a 4D buffer for a vectorial non-interleaved image
  void ****getVarray(void) const { return array.fourD; }

  /// returns size of image words (each vector component size)
  unsigned int sizeofWords(void) const;
  /// returns size of each voxel (a vector is one voxel)
  unsigned int sizeofVoxels(void) const { return vdim * sizeofWords(); }

  /** Writes inrimage to file whose path is given as argument.
      If file name ends with '.gz' suffix, file is gzipped.
      @param name inrimage file name
      @exception UnableToCreateFile file creation failed for writting
      @exception UnableToWriteHeader write error encountered while writting header
      @exception UnableToWriteData write error encountered while writting data */
   void write(const char *name)
    throw(Inrimage::UnableToCreateFile,
	  Inrimage::UnableToWriteHeader, Inrimage::UnableToWriteData);

  /** Writes inrimage to file whose path is given as argument.
      If file name ends with '.gz' suffix, file is gzipped.
      @param name inrimage file name
      @param ct number of bits to convert image to before writing.
      @exception UnableToCreateFile file creation failed for writting
      @exception UnableToWriteHeader write error encountered while writting header
      @exception UnableToWriteData write error encountered while writting data */
  void write(const char *name, CompressType ct)
    throw(UnableToCreateFile, UnableToWriteHeader,
	  UnableToWriteData,InvalidCompressType);

   /** Allocates and fills with zero image data buffer. If data
       is not null, buffer is first fried.
       @exception NotEnoughMemory unable to allocate image buffer */
  virtual void allocate(void) throw(Inrimage::NotEnoughMemory);

  /// frees image data buffer
  virtual void free(void);

  /** Tests if a voxel position is within the bounds of the image 
      @param x the x coordinate of the voxel position
      @param y the y coordinate of the voxel position
      @param z the z coordinate of the voxel position
      @return true is the voxel is inside the image and false otherwise */
  bool inside(const int x, const int y, const int z);

  /** Set a new data buffer. If image data is not 0, it is
      first fried.
      Use setVectorialMode(VECTORIAL_MODE vm) if vectMode has changed 
      @param data new image data buffer
      @return old image data buffer
      @exception NotEnoughMemory unable to allocate image buffer */
  virtual void *setData(void *data)
    throw(NotEnoughMemory);

  /** Changes image vectorial mode. A vectorial image buffer can be
      stored either interlaced (default) or non interlaced.
      @param vm new vectorial mode
      @exception InvalidVectorialMode attempt to set a scalar mode to a vectorial image or opposite
      @exception NotEnoughMemory unable to allocate temporary buffer to swap vectors */
  virtual void setVectorialMode(VECTORIAL_MODE vm)
    throw(Inrimage::InvalidVectorialMode);

  /// returns image geometry
  GEOMETRY getGeometry(void) const { return geom; }

  /// set image geometry
  void setGeometry(GEOMETRY g) { geom = g; }

  /** set interpolator object
      @return old interpolator object */
  Interpolator const* getInterpolator() const {return interpolator;}
  /** set extrapolator object
      @return old extrapolator object */
  Extrapolator const* getExtrapolator() const {return extrapolator;}
  
  /** set interpolator object
      @param i new interpolator object
      @return old interpolator object */
  void setInterpolator(Interpolator* i);
  /** set extrapolator object
      @param e new extrapolator object
      @return old extrapolator object */
  void setExtrapolator(Extrapolator* e);
  
  /** creates an image coordinates converter for this image geometry
      @return image coordinates converter */
  ImageGeometry *newConverter(void) const;
  /** map a point in 3D world coordinates into voxel coordinates
  @param pos the input 3D world coordinates
  @param res the output voxel coordinates 
  @return a boolean that is true if the point is inside the image volume and false otherwise*/
  bool map3DWorldToVoxelCoordinates(const Vec3<double> pos,Vec3<double> &res) const; 
/** map a point in 3D voxel coordinates into world coordinates
  @param pos the input 3D voxel coordinates
  @param res the output world coordinates 
  @return a boolean that is true if the point is inside the image volume and false otherwise*/
  void map3DVoxelToWorldCoordinates(const Vec3<double> pos,Vec3<double> &res) const; 
  /** Return image voxel at (x,y,z), using interpolation or extrapolation as appropriate.
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate
      @returns Voxel representing given image voxel */
  Voxel* getRealVoxel(double x, double y, double z) const;
  /** Return image voxel at Vec3<double> = (x,y,z), using interpolation or extrapolation as appropriate.
      @param pos voxel coordinates
      @returns Voxel representing given image voxel */
  Voxel* getRealVoxel(Vec3<double> pos) const;


  /** Stores voxel (x, y, z) value in v, using interpolation or extrapolation as appropriate.
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate
      @param v voxel to store value */
  Voxel& getRealVoxel(double x, double y, double z, Voxel& v) const;

  /** Return image voxel at (x,y,z), using interpolation or extrapolation as appropriate, without taking global transformation into account.
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate
      @returns Voxel representing given image voxel */
  Voxel& getRealVoxelImageSpace(double x, double y, double z, Voxel& v) const;
  
  /** returns a Voxel representing given image voxel
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate */
  Voxel *getVoxel(unsigned int x, unsigned int y, unsigned int z) const;

  /** stores voxel (x, y, z) value in v
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate
      @param v voxel to store value */
  Voxel& getVoxel(unsigned int x, unsigned int y, unsigned int z, Voxel& v)
    const;
  
  /** returns a reference on a VoxelIteratorObject corresponding to the image
      type */
  VoxelIterator getVoxelIterator();
  /** set given image voxel with transmitted voxel value
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate
      @param v voxel value */
  void setVoxel(unsigned int x, unsigned int y, unsigned int z,
		const Voxel& v);

  /** Returns a Voxel representing given image voxel.
      Check that given coordinates do not exceed image size.
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate */
  Voxel *checkVoxel(unsigned int x, unsigned int y, unsigned int z) const;
  
  /** draw a 2D discrete line in given image plane
      @param x1 first pixel X coordinate
      @param y1 first pixel Y coordinate
      @param x2 last pixel X coordinate
      @param y2 last pixel Y coordinate
      @param z image plane
      @param v line voxels value */
  void line(int x1, int y1, int x2, int y2, unsigned int z, const Voxel& v);
/** draw a 3D discrete line between two points in 3D
      @param p1 Position of the first point in world coordinates (must be in the image)
      @param p2  Position of the second point in world coordinates 
      @param v line voxels value
	  @param forceProjection whether the 2 points should be projected on the image border if they are outside the image*/
  void line3D(Vec3<double> p1, Vec3<double> p2, const Voxel& v,bool forceProjection=false);

  /** Changes image word type. If image data buffer existed it is fried
      and reallocated.
      @param wordKind image word kind
      @param sign image word sign
      @exception NotEnoughMemory unable to allocate image buffer */
  virtual void setWords(WORD_KIND wordKind = WK_FIXED, SIGN sign = SGN_SIGNED)
    throw(Inrimage::NotEnoughMemory);


  /** indicates wether voxel lies in image bounds
      @param col voxel row (X)
      @param row voxel column (Y)
      @param plane voxel plane (Z) */
  bool in(unsigned int row, unsigned int col, unsigned int plane) const {
    return ((row < xdim) && (col < ydim) && (plane < zdim));
  }
  
  /** indicates wether a point lies in image bounds @param pos position vector
   */
  bool in(const Vec3<double>& pos) const;

  /** indicates wether inr has same size than this
      @param inr image to compare */
  bool sameSize(Inrimage const& inr) const {
    return (inr.xdim == xdim) && (inr.ydim == ydim) && (inr.zdim == zdim);
  }

  /** crops region between (x1, y1, z1) and (x2, y2, z2) in inrimage
      Crop expects x1 < x2, y1 < y2 and z1 < z2
      @param x1 first crop box corner X coordinate
      @param y1 first crop box corner Y coordinate
      @param z1 first crop box corner Z coordinate
      @param x2 second crop box corner X coordinate (x1 should be < x2)
      @param y2 secont crop box corner Y coordinate (y1 should be < y2)
      @param z2 second crop box corner Z coordinate (z1 should be < z2)
      @exception InvalidArgument x1 >= x2 or y1 >= y2 or z1 >= z2 or x2 >= xdim or y2 >= ydim or z2 >= zdim */
  void crop(unsigned int x1, unsigned int y1, unsigned int z1,
	    unsigned int x2, unsigned int y2, unsigned int z2)
  throw(Inrimage::InvalidArguments);

  /** put to 0 all voxels outside region between (x1, y1, z1) and (x2, y2, z2) in inrimage
      it expects x1 < x2, y1 < y2 and z1 < z2
      @param x1 first crop box corner X coordinate
      @param y1 first crop box corner Y coordinate
      @param z1 first crop box corner Z coordinate
      @param x2 second crop box corner X coordinate (x1 should be < x2)
      @param y2 secont crop box corner Y coordinate (y1 should be < y2)
      @param z2 second crop box corner Z coordinate (z1 should be < z2)
      @exception InvalidArgument x1 >= x2 or y1 >= y2 or z1 >= z2 or x2 >= xdim or y2 >= ydim or z2 >= zdim */
  void isolateRegion(unsigned int x1, unsigned int y1, unsigned int z1,
		     unsigned int x2, unsigned int y2, unsigned int z2)
    throw(Inrimage::InvalidArguments);
 /** put to 0 all voxels inside region between (x1, y1, z1) and (x2, y2, z2) in inrimage
      it expects x1 < x2, y1 < y2 and z1 < z2
      @param x1 first crop box corner X coordinate
      @param y1 first crop box corner Y coordinate
      @param z1 first crop box corner Z coordinate
      @param x2 second crop box corner X coordinate (x1 should be < x2)
      @param y2 secont crop box corner Y coordinate (y1 should be < y2)
      @param z2 second crop box corner Z coordinate (z1 should be < z2)
      @exception InvalidArgument x1 >= x2 or y1 >= y2 or z1 >= z2 or x2 >= xdim or y2 >= ydim or z2 >= zdim */
  void zeroRegion(unsigned int x1, unsigned int y1, unsigned int z1,
		  unsigned int x2, unsigned int y2, unsigned int z2)
    throw(Inrimage::InvalidArguments);
  /** compute inrimage grey level bounds
      @param min min GL voxel value (min should have been allocated with inrimage dynamic type)
      @param max max GL voxel value (max should have been allocated with inrimage dynamic type)
      @param vectorNorm if true, this method returns min and max vector norm for vectorial inrimages, otherwise it returns min and max vector component */
  void getBounds(Voxel *min, Voxel *max, bool vectorNorm = true) const;

  /** Set inrimage grey level bounds (rescales GL window).
      If z is >= 0, only one slice is set, otherwise the wall image buffer is
      set.
      @param min min GL value (min should have been allocated with inrimage dynamic type)
      @param max max GL value (max should have been allocated with inrimage dynamic type)
      @param z (if >= 0) slice number to modify */
  void setBounds(Voxel *min, Voxel *max, int z = -1);

  /// returns true if type is WT_FLOAT_VECTOR, WT_DOUBLE_VECTOR, WT_RGB or WT_RGBA, false otherwise
  static bool vectorial(WORD_TYPE);

  /// returns true if type is WT_FLOAT, WT_DOUBLE, WT_FLOAT_VECTOR or WT_DOUBLE_VECTOR, false otherwise
  static bool floating(WORD_TYPE);

  /** Split a vectorial inrimage in one inrimage per component.
      The inrimage array returned is a WT_FLOAT inrimages array
      if input image is WT_FLOAT_VECTOR, a WT_DOUBLE inrimages array
      if input is WT_DOUBLE_VECTOR, and WT_UNSIGNED_CHAR if input image
      is WT_RGB or WT_RGBA. Non vectorial inrimages lead to a copy.
      Input image is not modified
      @return inrimage array (one inrimage per vectorial dimension */
  Inrimage **splitVectors(void) const;

  /** Set vectorial components from an array of scalar inrimages
      @param inrs inrimages array (there should be one inrimage per vectorial dimension of this)
      @exception InvalidSize one of the inrimages size does not match this size
      @exception InvalidArguments one of the inrimages is vectorial */
  void setComponentsFrom(Inrimage **inrs) throw(Inrimage::InvalidSize,
						InvalidArguments);

  /** flip an inrimage in given direction
      @param geom flip axis */
  void flip(SLICE_ORIENTATION geom);

  /** switch two inrimage dimensions
      @param geom1 first dimension to switch
      @param geom2 second dimension to switch */
  void switchGeometry(SLICE_ORIENTATION geom1, SLICE_ORIENTATION geom2);

  /** mel given inrimage with this by substituing a box from this
      with a box from inr
      @param inr inrimage to mel with this
      @param xdest X coordinate of destination box
      @param ydest Y coordinate of destination box
      @param zdest Z coordinate of destination box
      @param xorg X coordinate of source box
      @param yorg Y coordinate of source box
      @param zorg Z coordinate of source box
      @param dx X size of box to mel
      @param dy Y size of box to mel
      @param dz Z size of box to mel
      @exception InvalidArguments inr type does not match this type or xorg > inr->xdim or yorg > inr->ydim or zorg > inr->zdim or xdest > xdim or ydest > ydim or zdest > zdim */
  void mel(const Inrimage *inr,
	   int xdest, int ydest, int zdest, 
	   int xorg, int yorg, int zorg,
	   unsigned int dx, unsigned int dy, unsigned int dz)
    throw(Inrimage::InvalidArguments);

  /// creates and return the  Histogram of this image
  GenericHistogram *createHistogram() const;
  GenericHistogram *createMaskedHistogram( const Inrimage* mask ) const;
  /// noise kinds
  enum Noise {
    /// white Gaussian noise
    WHITE_NOISE,
    /// Gaussian noise on image
    GAUSSIAN_NOISE,
    /// ultrasond-like
    US_NOISE
  };    

  /** creates a noisy cube inside this inrimage
      @param x1 first noisy box corner X coordinate
      @param y1 first noisy box corner Y coordinate
      @param z1 first noisy box corner Z coordinate
      @param x2 second noisy box corner X coordinate
      @param y2 second noisy box corner Y coordinate
      @param z2 second noisy box corner Z coordinate
      @param kind noise kind 
      @param mean noise value (if fromImage == false)
      @param dev mean dev value
      @exception InvalidArguments given box exceeds image size */
  void noise(unsigned int x1, unsigned int y1, unsigned int z1,
	     unsigned int x2, unsigned int y2, unsigned int z2,
	     Noise kind, double mean, double dev)
    throw(Inrimage::InvalidArguments /* if given box exceeds image size */);

  /** converts inrimage to any type
      @param type new inrimage type */
  void convertCast(WORD_TYPE type);

  /** converts inrimage to any type, resizing grey-levels
      @param type new inrimage type */
  void convertScale(WORD_TYPE type);


  /** Evaluates expression in function string. Supports the 
      following arithmetic operators in the following order: 
      ^(power),/(division) or *(multiplication) and +(addition) or -(subtraction)
      Operators with the same priority are processed from 
      left to right. All other operators have lowest priority 
      and therefor processed from left to right as well. 
      Special operators: 
      |<(minimum), |>(maximum) 
      Logical Operators(returns 1=true,0=false):  
      <=(Smaller or equal), >=(Larger or equal), ==(equal), 
      >(larger), <(smaller), &&(and), ||(or). 
      Standard functions (angles are in radians): 
      abs(), exp(), ceil()(rounds upwards), floor()(rounds downwards), 
      log(), sqrt(), sin(), cos(), tan(), asin(), acos(), atan(), 
      sinh(), cosh(), tanh(). 
      Global functions (returns a scalar and calculates 
      value based on the expression given as input to the function)): 
      mean, min(minimum value), max(maximum value), m
      ed(median), std(estimated standard deviation), var(estimated variance). 
      Variables: 
      x,y,z (pixel positions), inrX(X can be any number of a valid image). 
      Examples: 
        inr3 -eval inr1|<inr2  (each pixel of inr3 contains the
	   maximum of the corresponding pixels in inr1 and inr2), 
	inr3 -eval inr1 -eval (x-64)^2+(y-64)^2+(z-64)^2<200
	   (creates a sphere),
	inr3 -eval inr1-mean(inr1)(set mean to zero), 
	inr3 -eval (inr1-min(inr1))/(max(inr1)-min(inr1))*100
	   (normalizes image between 0 and 100), 
	inr3 -eval med(sqrt(inr1^2/inr2)/(x+y+z)^3)*inr2 (I don't know).  */
void eval(char *func,int nmbr, Var_list *op)
 throw(Inrimage::InvalidArguments);

  /** Creates a vector array corresponding to the position of unit vector on a 
      spherical dome of frequency "frequency". The size of the array is 
      20 * frequency *frequency
      @param frequency sphere tesselation level */
  static Vec3<double> *createDirectionArray(int frequency = DEFAULT_SPHERE_FREQUENCY);


  /** Undersamples image by an integer coefficient. Only scalar and interlaced vectorial images are handled.
      @param sx X sampling step (1 or more)
      @param sy Y sampling step (1 or more)
      @param sz Z sampling step (1 or more) */
  void underSample(unsigned int sx = 2, unsigned int sy = 2,
		   unsigned int sz = 1);

  /** Oversample an image. New image should be greater than original image or an exception is thrown.
      @param sx new image X dimension
      @param sy new image Y dimension
      @param sz new image Z dimension
      @exception InvalidSize (sx < xdim) || (sy < ydim) || (sz < zdim) */
  void overSample(unsigned int sx, unsigned int sy, unsigned int sz)
    throw(Inrimage::InvalidSize);

  /** Resamples an image to given image size. Undersampling is approximated by picking on trilinearly interpolated grey-level value in the image (i.e. no image averaging is computed).
      @param sx new image X dimension
      @param sy new image Y dimension
      @param sz new image Z dimension */
  void resample(unsigned int sx, unsigned int sy, unsigned int sz);

  /** set region in which operations are performed. not used by all operations.
      negative values mean lower or higher bound, as appropriate.
      @param x1 lower box x coordinate
      @param y1 lower box y coordinate
      @param z1 lower box z coordinate
      @param x2 higher box x coordinate
      @param y2 higher box y coordinate
      @param z2 higher box z coordinate */
  void setRegion(int x1,int y1,int z1,int x2,int y2,int z2);
  
  /** get region in which operations are performed.
      @param x1 lower box x coordinate
      @param y1 lower box y coordinate
      @param z1 lower box z coordinate
      @param x2 higher box x coordinate
      @param y2 higher box y coordinate
      @param z2 higher box z coordinate */
  void getRegion(int& x1,int& y1,int& z1,int& x2,int& y2,int& z2);
  
  /** get region [(x1,y1,z1),(x2,y2,z2)) in which operations are should be performed.
      Region returned is always valid, ie values >= 0 and < get[XYZ]().
      @param x1
      @param y1
      @param z1
      @param x2
      @param y2
      @param z2 the bounding box
      @return the old region in r */
  void getWorkRegion(unsigned int& x1,unsigned int& y1,unsigned int& z1,unsigned int& x2,unsigned int& y2,unsigned int& z2) const;

  /** copy region from another image.
      @param img the image to copy the region from. **/
  void copyRegion(Inrimage const& img);

  /** test if region is the whole image.
      @return true if it is so, false otherwise **/
  bool regionIsWholeImage() const;
  

 protected:
  /// image file name
  TclString name;
  /// inrimage type
  WORD_TYPE type;  
  
  /// inrimage geometry
  GEOMETRY geom;  

  union {
    void ***threeD;     /// 3D array on each image voxel -> array[z][y][x]
    void ****fourD;     /** 4D array on each vector component when vectors are
			    not interlaced -> varray[v][z][y][x] */
  } array;

  /// interpolator object used for image values at real coordinates inside the image.
  Interpolator const* interpolator;
  /// extrapolator object used for image values at real coordinates outside the image.
  Extrapolator const* extrapolator;

  /// region in which operations are performed, in world coordinates. not used by all operations.
  float region[6];
  
 private:
  /** set initial image type
      @param UnknownType unsuported inrimage type */
  void setType(void) throw(Inrimage::UnknownType);
  /** builds 3D index (array buffer, standard inrimage)
      @param NotEnoughMemory unable to allocate memory to store index */
  void index3D(void) throw(Inrimage::NotEnoughMemory);
  /** builds 4D index (varray buffer, non-interlaced vectorial inrimage)
      @param NotEnoughMemory unable to allocate memory to store index */
  void index4D(void) throw(Inrimage::NotEnoughMemory);
  /** builds 3D or 4D index depending on vectorial mode
      @param NotEnoughMemory unable to allocate memory to store index */
  void buildIndex() throw(Inrimage::NotEnoughMemory);
  /// deletes array buffer
  void deleteArray(void);
  /// set image geometry field looking at inrimage header tags
  void queryGeometry(void);



  
 

  /// buffer convertion routine
  void convertFromRGBToRGBABuffer(void);
  /// buffer convertion routine
  void convertFromRGBAToRGBBuffer(void);
  /// buffer convertion routine


  friend class TclInrimage;
  friend class Epinrimage;
  friend class TclEpinrimage;

};

} // namespace yav

#include <inrimage/Voxel.h>
#include <inrimage/VoxelIterator.h>

#endif
