#ifndef INRIMAGE4D_H
#define INRIMAGE4D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/Inrimage4D.h"
#include <basics/Exception.h>
#include <basics/GeneralFormat.h>
#include <basics/Pointer.h>
#include <inrimage/Inrimage.h>
#include <vector>

namespace yav {

class Inrimage4D;

/** defines the type of function called to test if an image is of a given
    format. The first parameter is an array of char of size 5 (ends with
    character 0) that describes the first file character (magic string). The
    second parameter is the filename. The output value is >=0 if the image is
    of that given format and <0 otherwise */
typedef int (*TEST_IMAGE4D_FORMAT)(char *,const char *);
/** defines the type of function called to read a 4D image
    from a file corresponding to a given format. The first parameter is the
    pointer on an empty Inrimage4D * after creation and the second parameter
    is the file name.  The output value is >=0  if  the whole image has been 
    read, and it is  <0 otherwise */
typedef int (*READ_IMAGE4D)(Inrimage4D *,const char *);
/** defines the type of function called to write a 4D image 
    to a file corresponding to a given format. The first parameter is the
    pointer on an Inrimage4D and the second parameter is the file name. 
    The output value is >=0  if  the whole image has been written correctly
    and it is  <0 otherwise */
typedef int (*WRITE_IMAGE4D)( Inrimage4D *,const char *);


typedef struct im4Dformat {
  /** a pointer on a function that tests if an image is of a given format */
  TEST_IMAGE4D_FORMAT testImageFormat;

  /** a pointer on a function that reads  a 4D image file */
  READ_IMAGE4D readImage;

  /** a pointer on a function that writes a 4D image of a given format */
  WRITE_IMAGE4D writeImage;
  /** the file extension of format (including a dot ".": if several 
     extensions may be used, they should be separed with a 
     comma ".inr,.inr.gz" */
  char fileExtension[IMAGE_FORMAT_NAME_LENGTH];
  /** the usual name given to a format : for instance "inrimage", "QCI" */
  char realName[IMAGE_FORMAT_NAME_LENGTH];
  /** pointing on next format */
  struct im4Dformat *next;
 } IMAGE4D_FORMAT, *PTRIMAGE4D_FORMAT;

/// add a format of a 4D image file in a static list of formats 
 void add4DImageFormat(IMAGE4D_FORMAT *f);

 /** returns the default format for reading a 4D Image 
     (based on General Format) */
 IMAGE4D_FORMAT *firstImage4DFormat();


/** \class Inrimage4D
    \ingroup Inrimage
    sequence of 3D inrimages */
class Inrimage4D : public Owner<Inrimage>, public Pointed<Inrimage4D>, 
		   public GeneralFormat {

 protected:
  /// 4D inrimage file name
  TclString name;
  /// list of 3D inrimages
  std::vector<Inrimage *> inrs;
  /// list of pointers on 3D inrimages
  std::vector<ObjectPointer<Inrimage> *> inrPointers;
  /// 4D index array to access [T][Z][Y][X] voxel
  void ****array;
  /// voxel size in the time dimension
  float vt;
  /// the format of the image (default General Format)
  IMAGE4D_FORMAT *format;
 public:
  /** @name exception */
  //@{
  /** \class ReadError
      \ingroup Inrimage
      \brief an inrimage file read error occured */
  class ReadError : public Exception {
  public:
    /// constructor
    ReadError(const std::string& from, const std::string& filename);
  };
  /** \class EmptyInrimageList
      \ingroup Inrimage
      \brief the list of  inrimages is empty */
  class EmptyInrimageList : public Exception {
  public:
    /// constructor
    EmptyInrimageList(const std::string& from) : Exception(from,"Empty inrimage list provided"){}
  };
  /** \class UnknownImageFormat
      \ingroup Inrimage
      \brief the image format is unknown */
  class UnknownImageFormat : public Exception {
  public:
    /// constructor
    UnknownImageFormat(const std::string& from) : Exception(from,"Unknown Image Format "){}
  };
  /** \class DifferentImageType
      \ingroup Inrimage
      \brief the images do not have the same image type */
  class DifferentImageType : public Exception {
  public:
    /// constructor
    DifferentImageType(const std::string& from) : Exception(from, "The images do not have the same image type"){}
  };
  /** \class DifferentImageSize
      \ingroup Inrimage
      \brief the images do not have the same image size */
  class DifferentImageSize : public Exception {
  public:
    /// constructor
    DifferentImageSize(const std::string& from) : Exception(from, "The images do not have the same image size"){}
  };
  //@}

  /** @name constructors */
  //@{
  /** builds a new sequence of 3D inrimages
      @param filename 4D inrimage file name
      @exception ImageNotFound image file not found
      @exception ReadError error while reading file */
  Inrimage4D(const char *filename)
    throw(GeneralFormat::FileNotFound,
	  Inrimage::ImageNotFound, ReadError,EmptyInrimageList, DifferentImageType, DifferentImageSize);
  /** builds a 4D inrimage from a sequence of 3D inrimages
      @param inrs list of constituting 3D inrimages */
   Inrimage4D(std::list<Inrimage *> inrs)
     throw(EmptyInrimageList, DifferentImageType, DifferentImageSize);
  //@}

  /** @name destructor */
  //@{
  /// destroys any 3D inrimage
  ~Inrimage4D();
  //@}

  /// returns 4D index array
  void ****getArray(void) const { return array; }
  /** compute inrimage grey level bounds
      @param min min GL voxel value (min should have been allocated with inrimage dynamic type)
      @param max max GL voxel value (max should have been allocated with inrimage dynamic type)
      @param vectorNorm if true, this method returns min and max vector norm for vectorial inrimages, otherwise it returns min and max vector component */
  void getBounds(Voxel *min, Voxel *max, bool vectorNorm = true) const;

 
  /// test if file contains 
  static int testGeneralFormat(char *magic,const char *fileName);
  /// read Inrimage 4D with General Format
  static int readGeneralFormat(Inrimage4D *,const char *fileName);
/// write Inrimage 4D with General Format
  static int writeGeneralFormat(Inrimage4D *,const char *fileName);

  /** add an image at the end of the list of images
      @param inr a valid 3d image
      @param buildArray whether the 4D array should be updated or not. Use false when incrementally building the list to avoid waste of time */
  void addInrimage(Inrimage *inr, bool buildArray=true) throw (DifferentImageType,DifferentImageSize);
  
  /** general format read function for Inrimage names module */
  void readInrNames(GFInput &input);
  /** general format skip function for Inrimage names module */
  void skipInrNames(GFInput &input);  
  /** general format write function for Inrimage names module */
  void writeInrNames(GFOutput &output);
  /** save the 4D image to a file : the format is chosen from the
      file extension (default General Format) */
  void write(const char *filename);
  /** returns a Voxel representing given image voxel
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate 
      @param t Time voxel coordinate */
  Voxel *getVoxel(unsigned int x, unsigned int y, unsigned int z,
		  unsigned int t) const;

  /** stores voxel (x, y, z) value in v
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate
      @param t Time voxel coordinate
      @param v voxel to store value */
  Voxel& getVoxel(unsigned int x, unsigned int y, unsigned int z, 
		  unsigned int t,Voxel& v)
    const;

  /** set given image voxel with transmitted voxel value
      @param x X voxel coordinate
      @param y Y voxel coordinate
      @param z Z voxel coordinate
      @param t Time voxel coordinate
      @param v voxel value */
  void setVoxel(unsigned int x, unsigned int y, unsigned int z,
		unsigned int t,	const Voxel& v);

  /** Returns 3D inrimage at instant i. If i exceed last instant,
      i's value is modified to last instant and last instant
      3D inrimage is returned
      @param i instant */
  Inrimage *getInrimage(unsigned int i);
  /// returns 3D inrimage at first instant
  Inrimage *getFirstInrimage() const { return inrs.front(); }
  /// returns inrimages vector
  std::vector<Inrimage *>& getInrimages(void) { return inrs; } 
  /// returns inrimages list
  std::list<Inrimage *> getInrimageList(void);
  /** an image owned by this 4D image is being destroyed : destroy 4D image
      @param obj element before destruction */
  virtual void dying(const ObjectPointer<Inrimage> & obj) {
    delete this;
  } 
 /// returns 4D inrimage file name
  const TclString& getName(void) const { return name; }

  /// set new 4D inrimage file name
  void setName(const char *_name) {
    if(_name && _name[0] != '\0') name = _name;
  }
  /// get VECTORIAL_MODE
  VECTORIAL_MODE getVectorialMode() const { 
    return getFirstInrimage()->getVectorialMode(); }
  /// returns inrimage type
  Inrimage::WORD_TYPE getType(void) const { return getFirstInrimage()->getType(); }
  /// get image time dimension
  unsigned int getTimeDimension() const { return inrs.size(); }
  /// get image x dimension
  unsigned int getX() const { return getFirstInrimage()->getX(); }
  /// get image y dimension
  unsigned int getY() const { return getFirstInrimage()->getY(); }
  /// get image z dimension
  unsigned int getZ() const { return getFirstInrimage()->getZ(); }
  /// get image vectorial dimension
  unsigned int getVdim(void) const { return getFirstInrimage()->getVdim(); }
  /// get image words size (in bytes)
  unsigned int getWdim(void) const { return getFirstInrimage()->getWdim(); }
  /// get image voxel size in x dimension
  float getVX() const { return getFirstInrimage()->getVX(); }
  /// get image voxel size in y dimension
  float getVY() const { return getFirstInrimage()->getVY(); }
  /// get image voxel size in z dimension
  float getVZ() const { return getFirstInrimage()->getVZ(); }
  /// get image voxel size in time dimension
  float getVT() const { return vt; }
  /// set the voxel size in the time dimension
  void setVT(const float &_vt) {
    vt=_vt;}
  /// set the voxel size including time dimension
  void setVoxelSize(float _vx, float _vy, float _vz, float _vt)  
    throw(BaseInrimage::InvalidVoxelSize);

  /** crops region between (x1, y1, z1) and (x2, y2, z2) in each 3D inrimage
      Crop expects x1 < x2, y1 < y2 and z1 < z2
      @param x1 first crop box corner X coordinate
      @param y1 first crop box corner Y coordinate
      @param z1 first crop box corner Z coordinate
      @param x2 second crop box corner X coordinate (x1 should be < x2)
      @param y2 secont crop box corner Y coordinate (y1 should be < y2)
      @param z2 second crop box corner Z coordinate (z1 should be < z2)
      @exception InvalidArgument x1 >= x2 or y1 >= y2 or z1 >= z2 or x2 >= xdim or y2 >= ydim or z2 >= zdim */
  void crop(unsigned int x1, unsigned int y1, unsigned int z1,
	    unsigned int x2, unsigned int y2, unsigned int z2)
  throw(Inrimage::InvalidArguments);
  
 protected:
  /// construtors used by Inrimage4DClient
  Inrimage4D(void);
  /// build 4D array to index each 4D vertex
  void buildArray(void);
  
 private:
  /// constructors common code
  void init(void);



  friend class TclInrimage4D;

};

} // namespace yav

#endif


