#ifndef INRIMAGEROI_H
#define INRIMAGEROI_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/InrimageROI.h"

#include <graphics2D/Camera2D.h>
#include <graphics2D/Data2D.h>
#include <graphics2D/SurfaceSliceable.h>
#include <inrimage/ROIVoxelList.h>

namespace yav {


/** \class InrimageROI
    \ingroup Inrimage
    \brief a region of interest in a volumetric image that can be drawn in a slice camera*/
class InrimageROI : public ROIVoxelList, public Data2D {

 public:
  /// the different opacity available for displaying a ROI
  enum OpacityLevel{
    FULL_OPACITY,
    OPACITY_75,
    OPACITY_50,
    OPACITY_25,
    OPACITY_12
  };
  /// the left mouse bouton action
  enum SelectionMode {BRUSH_MODE,ERASER_MODE,LINE_MODE,IDLE_MODE};
 protected:
  /// surface that has generated the region of interest
  SurfaceSliceable *surface;
  /// color of the region
  float roiColor[3];
  /// name used in the canvas to draw the region
  TclString traceName;
  ///  the opacity level used for drawing the ROI
  OpacityLevel opacityLevel;
  /// if the roi must be display with color or greyLevel 
  bool useGreyLevel;
  /// if the ROI should be hidden
  bool visibility;
  /// the different left mouse action
  SelectionMode selectionMode;
   /// the size of the brush when in brush mode
   unsigned int brushWidth;
   /// the size of the eraser when in eraser mode
   unsigned int eraserWidth;
   /// if the left button is beeing pushed 
   bool leftButtonOn;

 protected:
  /** initialize graphics variables */
  void init();
 public: 
  /** describes how to generate a region of interest from a surface */
  enum SurfaceRoiMethod{
    SURFACE_BOUNDARY, /* voxels on the surface */
	SURFACE_BOUNDARY_Z, /* voxels on the surface but only with axial contours*/
    SURFACE_REGION, /* voxels inside the region enclosed by the surface */
    SURFACE_REGION_OUTSIDE, /* voxels inside the region enclosed by the surface */
    SURFACE_NEAR_BOUNDARY /* voxels within a narrow band around the surface */
  };

  /** @name constructor */
  //@{
  /** build a new region of interest in a volumetric image
      @param scene the 2D scene where the object is located 
      @param svl a list of voxel located on the region boundary
      @param _vsize
  */
  InrimageROI(Scene2D *scene, const std::list<ROIVoxel *> &svl, 
	      Vec3<double> _vsize=Vec3<double>(0.,0.,0.));
  /** build a region of interest associated with a surface and an image
      @param scene the 2D scene where the object is located 
      @param image a 3D image defining the size of the slice image and its voxel size 
      @param s a surface sliceable that is sliced along the Z direction
      @param m defines if we are interested in the region enclosed by the surface or its boundary or a region near the boundary
      @param param size of the narrow band (in mm) near the surface boundary 
	  @param selectedZone if -1 then use all zones of the mesh. Otherwise draw only the selected zone*/
  InrimageROI(Scene2D *scene, 
	      Inrimage *image, 
	      SurfaceSliceable *s,
	      SurfaceRoiMethod m=InrimageROI::SURFACE_BOUNDARY, 
	      double param=1.0,int selectedZone= -1);
   /** build a region of interest around a given voxel in an image
      @param scene the 2D scene where the object is located 
      @param image a 3D image defining the size of the slice image and its voxel size 
      @param worldPosition a voxel in that image given in world coordinates
      @param radius the radius of the sphere defining the Region of interest */
InrimageROI(Scene2D *scene, 
			 Inrimage *inr, 
			 Vec3<double> worldPosition,
			 double radius); 
  /** build a region of interest from a binary image and a grey-value image
      @param scene the 2D scene where the object is located 
      @param image a 3D grey-value image where the ROI is defined : the grey-level value of each voxel is used from this original image as well as the image size
      @param binary a 3D binary image where each voxel of non-zero value is considered to belong to the region of interest. It must have the same size as the original image */
  InrimageROI(Scene2D *scene, 
	      Inrimage *image, 
	      const Inrimage *binary) throw(ImageMissing,
					    BinaryImageWrongSize);
  /** build an empty region of interest from a grey-value image
      @param scene the 2D scene where the object is located 
      @param image a 3D grey-value image where the ROI is defined : the grey-level value of each voxel is used from this original image as well as the image size */

  InrimageROI(Scene2D *scene, 
	      Inrimage *image) throw(ImageMissing);
  /** build a region of interest from a file 
      @param scene the 2D scene where the object is located 
      @param filename the name of the file    */
  InrimageROI(Scene2D *scene, char *filename);
   //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~InrimageROI(){}
  //@}
  
  /// returns the boolean indicating if the ROI is visible
  bool getVisibility() const {
    return visibility;
  }
  /// sets boolean indicating if the ROI is visible
  void setVisibility(const bool b) {
    visibility=b;
  }
  /// set the selection mode
  void setSelectionMode(const SelectionMode m) {
	  selectionMode=m;
  }
  /// return the current selection Mode
  SelectionMode getSelectionMode() const {
	  return selectionMode;
  }
  /// set the brush width
  void setBrushWidth(const unsigned int w) {
	  brushWidth=w;
  }
    /// set the eraser width
  void setEraserWidth(const unsigned int w) {
	  eraserWidth=w;
  }
   /// return the eraser width
  unsigned int getEraserWidth() const {
	  return eraserWidth;
  }
   /// return the eraser width
  unsigned int getBrushWidth() const {
	  return brushWidth;
  }
  /// return this object scene
  Scene2D *getScene(void) const { return scene; }
  /** add a list of voxels in the ROI  */
  virtual void addVoxel(std::list<ROIVoxel *> vlist);
 /** add a list of voxels in the ROI given their image coordinates
      @param plist the list of voxel coordinates
  */
  virtual void addVoxel(std::list< Vec3<unsigned int>  > &plist);
  /** remove a list of voxels in the ROI 
      @param plist the list of voxel coordinates
      @return a boolean : true if all voxels have been removed and false otherwise*/
  virtual bool removeVoxel(std::list< Vec3<unsigned int>  > &plist);

  /** Translates the 2D object
      @param t translation vector */
  virtual void translate(const Vec2<double>& t){}

  /** Rotates the 2D object: no rotation available for images
      @param r rotation matrix */
  virtual void rotate(const Rotation2D &r){}

  /** Scales the 2D object
      @param s scale factor */
  virtual void scale(const double s){}

  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec2<double>& center, unsigned int& n)  {
    center = Vec2<double>(0.0, 0.0);
    n = 0;
  }
  /// get the color of the region of interest
  void getColor( float &r, float &g, float &b) const {
    r=roiColor[0];g=roiColor[1];b=roiColor[2];
  }
  /// get the color of the region of interest
  void setColor( const float r,const  float g, const float b) {
    roiColor[0]=r;roiColor[1]=g;roiColor[2]=b;
  }
  /// if the roi must be display with color or greyLevel 
  bool getGreyLevelUsage() const {
    return useGreyLevel;
  }
  /// decides if the roi must be display with color or greyLevel 
  void setGreyLevelUsage( const bool v) {
    useGreyLevel=v;
  }
  /** default redraw function for sliceable objects
      @param cam camera to redraw */
  virtual void redraw(Camera2D *cam);
  
  /// returns the surface sliceable that has generated the ROI
  SurfaceSliceable *getSurface() const {
    return surface;
  }

  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec2<double>& center){return 0.0;}

  /** Return MAXDOUBLE, an image is considered as background
      and may not be selected */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const{
  return -1.0;}

  /// returns the opacity level for the ROI display
  OpacityLevel getOpacity() const {
    return opacityLevel; 
  }
  /// set the opacity level for the ROI display
  void setOpacity(const OpacityLevel o) {
    opacityLevel=o;
  }
  /** callback called when a button is pressed in a camera
      @param button indicates which button is pressed
      @param modifier indicates if a modifier (shift or control key) is pressed
      @param cam the 2D camera where occured the click button
      @param mx first coordinate of the pixel clicked 
      @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int mx,int my);
 /** callback called when a button is released in a camera
     @param button indicates which button is released
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				   CameraTypes::KEY_MODIFIER modifier,
				     Camera2D *cam,int mx,int my);
 /** callback called when a button is moved after having been pressed
     @param cam the 2D camera where occured the click button
     @param x first coordinate of the pixel clicked 
     @param y second coordinate of the pixel clicked */ 
  virtual void mouseButtonMoved(Camera2D *cam,int x,int y);


};

} // namespace yav

#endif


