#ifndef INRIMAGEVIEW_H
#define INRIMAGEVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/InrimageView.h"
#include <basics/Owner.h>
#include <basics/Pointer.h>
#include <graphics2D/SliceCamera.h>
#include <inrimage/InrimageSliceInterface.h>
#include <graphics2D/Sliceable.h>

namespace yav {


class InrimageView;
#ifdef TCL_INTERFACE
class TclInrimageView;
#endif

/** \class InrimageBackground
    \ingroup Inrimage
    \brief an interface class that adds specific functions for a background attached to an inrimage */
class InrimageBackground : public SliceCameraBackground {
 public:
  /** @name constructors */
  //@{
  /** Dummy constructor since no inrimage is specified
      @param cam camera attached to the background */
  InrimageBackground(SliceCamera *cam) :SliceCameraBackground(cam) {}
  //@}
  /// destruction
  virtual ~InrimageBackground(){}
  
  /// image type changed
  virtual void inrimageChanged(void)=0;
  /** redraw camera image */
  virtual void redraw()=0;
  /// reset pointer to camera
  virtual void refreshCamera(void)=0;

  friend class InrimageView;
  };


/** \class InrimageView
    \ingroup Inrimage
    \briefA viewer for an inrimage slice */
class InrimageView : public Sliceable,
		     public Owner<Inrimage>
{
 protected:
  /** @name constructors */
  //@{
  /** Constructor that does not call init() 
      @param inr1 inrimage to draw
      @param i dummy inrimage pointer
      @param scene  scene to build data in 
      @param _createGLScale boolean indicating if we should create  GLScales associated with each background */
  InrimageView(Inrimage *inr1, Inrimage *i,Scene2D *scene,
	       bool _createGLScale=true);
  /** Constructor that reads image from a file
      @param filename inrimage file name
      @param scene  scene to build data in */
  InrimageView(const TclString &filename, Scene2D *scene);
  //@}

  /// constructors common code
  void init(void);
 protected:
  /// the slice interface linked to the image
  InrimageSliceInterface *sliceInterface;
  /// the list of backgrounds associated with the inrimage view
  std::list<InrimageBackground *> backgroundList;
  /// whether we should create
  bool createGLScale;
  
 public:
    
  /** @name constructors */
  //@{
  /** Normal constructor by specifiying the inrimage to display
      @param inr inrimage to draw
      @param scene  scene to build data in 
      @param _createGLScale boolean indicating if we should create  GLScales associated with each background */
  InrimageView(Inrimage *inr, Scene2D *scene,bool _createGLScale=true);
  //@}

  /// destruction: do not delete visualized inrimage
  virtual ~InrimageView();
  
  /// die on inrimage deletion
  virtual void dying(const ObjectPointer < Inrimage > &);

  /** focus camera on image center
      @param cam camera to focus */
  void focusCamera(SliceCamera& cam);

  /// add a camera in inrimage view management if needed
  virtual void createdCamera(SliceCamera& cam);
  /// add a camera in inrimage view management
  virtual void doCreateCamera(SliceCamera& cam);

  /// remove a camera from inrimage view management
  void deletedCamera(SliceCamera& cam);

  /// return this object scene
  Scene2D *getScene(void) const { return scene; }


  /// returns eventually the TclImage associated with a given slice camera
  InrimageBackground *getBackgroundFromCamera(SliceCamera *cam);
  
  /** Translates the 2D object
      @param t translation vector */
  virtual void translate(const Vec2<double>& t) {}

  /** Rotates the 2D object: no rotation available for images
      @param r rotation matrix */
  virtual void rotate(const Rotation2D &r) {}

  /** Scales the 2D object
      @param s scale factor */
  virtual void scale(const double s) {}

  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec2<double>& center, unsigned int& n)  {
    center = Vec2<double>(0.0, 0.0);
    n = 0;
  }

  /// nothing done when an image is selected
  virtual void select(Camera2D *) {}

  /// redraw all background managed by this view
  virtual void redrawBackgrounds();

  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec2<double>& center)  { return 0.0; }

  /** Return MAXDOUBLE, an image is considered as background
      and may not be selected */
  virtual double computeClosestDistance(DataTypes::MouseClick& mc) const {
    return -1.0;
  }
  /// return iterator on the first background 
  std::list<InrimageBackground *>::iterator firstBackground() {
    return(backgroundList.begin());
  }
  /// return iterator on the last background 
  std::list<InrimageBackground *>::iterator lastBackground() {
    return(backgroundList.end());
  }
  /** returns the pointed inrimage */
  Inrimage *getInrimage() const;

  /** Modifies the pointed image and Updates all images */
  virtual void changeInrimage(Inrimage *inr);

  /// reset inrimageBackgrounds objects list pointers to cameras
  void refreshCameras(void);

  /// nothing done when an image is unselected
  virtual void unselect(Camera2D *) {}

  /** default redraw function for sliceable objects
      @param cam camera to redraw */
  virtual void redraw(Camera2D *cam);
  
  /// returns the slice interface associated with the image
  InrimageSliceInterface *getSliceInterface() const {
    return sliceInterface;
  }

  TclInrimageView *getInterface(void) const;

  friend class TclInrimageView;
  friend class TclInrimageBackground;
  friend class TclInrimage;
};


} // namespace yav


#endif

