#ifndef INTERPOLATOR_H
#define INTERPOLATOR_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/Interpolator.h"
#include <inrimage/Inrimage.h>

namespace yav {


class Inrimage;
class Voxel;

/// Base class for interpolation (estimates values between known values)
struct Interpolator {
  /// destructor
  virtual ~Interpolator() {}
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  virtual Voxel& operator()(Inrimage const& img, double x,
			    double y, double z, Voxel& v) const=0;
  /// returns a copy of the object.
  virtual Interpolator* clone() const=0;
  
  /** builds a tri-linear interpolator given image type
      @param geom image geometry
      @param t image word type
      @param vdim image vectorial dimension (1, 2 or 3, used only for cartesian inrimages) */
  static Interpolator *trilinearInterpolatorByType(Inrimage::GEOMETRY geom,
						   Inrimage::WORD_TYPE t,
						   int vdim);
};



/// Base class for extrapolation (infers values outside an interval of known values)
struct Extrapolator {
  /// destructor
  virtual ~Extrapolator() {}
  /** get extrapolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  virtual Voxel& operator()(Inrimage const& img, double x,
			    double y, double z, Voxel& v) const=0;
  /// returns a copy of the object.
  virtual Extrapolator* clone() const=0;
};


/// Nearest neighbor interpolation 
struct NearestInterpolator : public Interpolator {
  /// Overloaded function. See base class for information.
  Voxel& operator()(Inrimage const& img, double x, double y, double z,
		    Voxel& v) const;
  /// returns a copy of the object.
  Interpolator* clone() const;
};


  
/// Trilinear interpolation for images with scalar voxels. It will work for vectorial images but the result is probably not what you want.
struct TrilinInterpolator : public Interpolator {
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  Voxel& operator()(Inrimage const& img, double x, double y, double z,
		    Voxel& v) const;
  /// returns a copy of the object.
  Interpolator* clone() const;
  /** return a TypedTrilinInterpolator for an image of type t.
      @param t The type of values of the image to interpolate. */
  static TrilinInterpolator* interpolatorByType(Inrimage::WORD_TYPE t);
};



/// Trilinear interpolation when image type is known. Overestimates interpolated values by 0.5 on floating point types (adds 0.5 to interpolated value).
template <class V,typename T> struct TypedTrilinInterpolator : public TrilinInterpolator {
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  Voxel& operator()(Inrimage const& img, double x, double y, double z,
		    Voxel& v) const{
    (*this)(img,x,y,z,static_cast<T&>(static_cast<V&>(v).value()));
    return v;
  }
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v to store value
      @return v, the stored value */
  T& operator()(Inrimage const& img, double x, double y, double z, T& v) const;
  /// returns a copy of the object.
  Interpolator* clone() const;
};



/** \class TrilinVectInterpolator
    \ingroup Inrimage
    \brief Trilinear interpolation for images with vector voxels. It will not work for scalar images. */
class TrilinVectInterpolator : public Interpolator {
 public:
  /// Exception thrown when image voxel is not a vector.
  struct VoxelNotAVector : public Exception {
    /// Constructor. See base class to information.
    VoxelNotAVector(std::string const& f,std::string const& e="voxel is not a vector") : Exception(f,e) {}
  };

  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  Voxel& operator()(Inrimage const& img,double x,double y,double z,Voxel& v) const;

  /// returns a copy of the object.
  Interpolator* clone() const;

  /** return a TypedTrilinVectInterpolator for an image of type t.
      @param t The type of values of the image to interpolate.
      @param v
  */
  static TrilinVectInterpolator* interpolatorByType(Inrimage::WORD_TYPE t,
						    size_t v);
 protected:
  /** <b>r</b> += <b></b>v*<b>d</b>.
      <b>r</b> and <b>d</b> must be of be vectorial and of the same type.
      <b>r</b> and <b>d</b> may be equal.
      @param r result and initial value
      @param v voxel to add to r
      @param d value to scale v by
      @return r, the result voxel */
  static Voxel& addMulVoxDbl(Voxel& r,Voxel const& v,double d);
};


/// Trilinear vector interpolation when image type is known. Underestimes interpolated values for integer types (does not add 0.5 to interpolated value).
template<class V,typename T,size_t D>
struct TypedTrilinVectInterpolator : public TrilinVectInterpolator {
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  Voxel& operator()(Inrimage const& img,double x,
		    double y,double z,Voxel& v) const{
    (*this)(img,x,y,z,static_cast<T*>(static_cast<V&>(v).getValue()));
    return v;
  }
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x voxel position
      @param y y voxel position
      @param z z voxel position
      @param v to store value
      @return v, the stored value */
  T* operator()(Inrimage const& img,double x,double y,double z,T v[D]) const;
  /// returns a copy of the object.
  Interpolator* clone() const;
};


/** \class CylTrilinInterpolator
    \ingroup Inrimage
    \brief Trilinear interpolation for images with cylindrical geometry.
    It won't work for vectorial images. */
class CylTrilinInterpolator : public Interpolator {
 public:
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  Voxel& operator()(Inrimage const& img, double x, double y, double z,
		    Voxel& v) const;

  /// returns a copy of the object.
  Interpolator* clone() const;
};


/** \class SphTrilinInterpolator
    \ingroup Inrimage
    \brief Trilinear interpolation for images with spherical geometry.
    It won't work for vectorial images. */
class SphTrilinInterpolator : public Interpolator {
 public:
  /** get interpolated value at img(x,y,z).
      @param img image to interpolate
      @param x x cartesian coordinate
      @param y y cartesian coordinate
      @param z z cartesian coordinate
      @param v voxel to store value
      @return v, the voxel in which the value is stored */
  Voxel& operator()(Inrimage const& img, double x, double y, double z,
		    Voxel& v) const;

  /// returns a copy of the object.
  Interpolator* clone() const;
};



/// Extrapolate values to 0.
struct ToZeroExtrapolator : public Extrapolator {
  /// Overloaded function. See base class for information.
  Voxel& operator()(Inrimage const& img,double x,double y,
		    double z,Voxel& v) const;
  /// returns a copy of the object.
  Extrapolator* clone() const;
};


  
/// Extrapolate values nearest voxel.
struct NearestExtrapolator : public Extrapolator {
  /// Overloaded function. See base class for information.
  Voxel& operator()(Inrimage const& img,double x,double y,
		    double z,Voxel& v) const;
  /// returns a copy of the object.
  Extrapolator* clone() const;
};


} // namespace yav

#endif
