#ifndef LANDMARKSET_H
#define LANDMARKSET_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 )
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4081 )
#endif


#include <framework/Transformation.h>
#include <linearAlgebra/Vec3.h>
#include <basics/ListElement.h>
#include <framework/PointSet.h>

namespace yav {

class LandmarkSet;
class ImageLandmark;


  /** \class ImageLandmarkIterator
      \ingroup Inrimage
  */
  class ImageLandmarkIterator : public PointSetIterator {

 private:
   /// the landmarkset
   LandmarkSet *lset;
   /// the iterator on the imagelandmark
   ObjectsList<ImageLandmark>::iterator it;

 public:
   /// constructor
   ImageLandmarkIterator(LandmarkSet *);
   /// destructor
   ~ImageLandmarkIterator(void);
   
    // reset iterator to first point
    virtual void first();
    // increments iterator
    virtual void operator++();
    // increments iterator
    virtual void operator++(int);
    // returns true if end of list was reached
    virtual bool isAtEnd() const;
    
    // returns current point coordinates
    virtual const Vec3<double> get() const;
    // returns current point weight
    virtual double getWeight() const;
    
    // returns current point X coordinate
    virtual double getX() const;
    // returns current point Y coordinate
    virtual double getY() const;
    // returns current point Z coordinate
    virtual double getZ() const;
    
    // changes the coordinates of the current point
    virtual void set(const Vec3<double>& Coordinates);


 };


/** \class ImageLandmark
    \ingroup Inrimage
    class defining a landmark, ie a point located at a voxel in a 3D image 
*/
class ImageLandmark : public ListElement<ImageLandmark> {
 protected:  
  /// the landmark position in world coordinates
  Vec3<double> realPosition;
  /// the landmark position in image coordinates
  Vec3<unsigned int> voxelPosition;
  /// the landmark name
  TclString name;

 public:
  /** @name Constructors */
  //@{
  /*** creates a landmark from its position and name
       @param voxelPosition the voxel position in image coordinates ie the its voxel position (x,y,z)
       @param realPosition indicates the position in world coordinates of the landmark (taking into account the voxel size); by convention the landmark is position at the voxel center (offset .5 .5 .5)
       @param n the landmark name */
  ImageLandmark(Vec3<unsigned int> voxelposition, 
		Vec3<double> realPosition, 
		const TclString n=EMPTY_STRING);
  //@}

  /** @name destructor */
  virtual ~ImageLandmark(){}


  /* {
    voxelPosition = voxelposition;
    realPosition.setValue(voxelposition.getX() * voxelSize.getX(), voxelposition.getY() * voxelSize.getY(), voxelposition.getZ() * voxelSize.getZ() );
    name = (*n);
    }*/


 public:
  /// returns the landmark position in world coordinates
  Vec3<double> getRealPosition() const {
    return realPosition;
  }
  /// returns the landmark position in image coordinates
  Vec3<unsigned int> getVoxelPosition() const {
    return voxelPosition;
  }

  /// set the real position of the landmark
  void setRealPosition(const Vec3<double> &a)
   {
     realPosition = a;
   }
 
  /// set the voxel position of the landmark
  void setVoxelPosition(const Vec3<double> &a)
   {
     voxelPosition = a;
   }

  /// returns the name of the landmark
  TclString getName() const 
    {
      return name;
    }

  /// set the name of the landmark
  void setName(const TclString n) 
    {
      name=n;
    }    
  friend class LandmarkSet;


};

/** \class LandmarkSet
    \ingroup Inrimage
    class that defines a set of landmarks */
class LandmarkSet  {
 protected:
  /// the list of landmarks
  ObjectsList<ImageLandmark> landmarkList;
 public:
  /** @name constructor */
  //@{
  /** build an empty set of landmarks 
  */
  LandmarkSet();
  /** build an set of landmarks  from a file
      @param filename name of the landmark file
  */
  LandmarkSet(const char *filename);
 /** build from an STL list of  landmarks 
      @param list the list of landmarks : each landmark pointer should not be null  */
  LandmarkSet(std::list<ImageLandmark *> list);
  //@}

  /** @name destructor */
  virtual ~LandmarkSet();
  
 public:

  /// compute a transformation between two landmarkset
  /*!
    param: the landmarkset to register with
    param: the kind of transfo (rigid, similarity, affine)
  */
  Transformation *computeTransfo(LandmarkSet *, TclString);

  /// compute the list of error between two landmarkset
  /*!
    param: the landmarkset we want to evaluate
  */
  std::list<double> computeError(LandmarkSet *);

  /// returns the number of landmatks in the set
  unsigned int getNumberOfLandmarks(void) {
    return landmarkList.getCount();
  }
  /// add a landmark in the set
  void addLandmark(ImageLandmark *imagelandmark);
  /// detach landmark from list (without deleting it)
  void detachLandmark(ImageLandmark *imagelandmark);
  /// remove landmark from list (free landmark)
  void removeLandmark(ImageLandmark *imagelandmark);
  
  /// returns an iterator on the first image landmark
  ObjectsList<ImageLandmark>::iterator firstLandmark() {
    return landmarkList.begin();
  }
  /// returns an iterator on the last image landmark
  ObjectsList<ImageLandmark>::iterator lastLandmark() {
    return landmarkList.end();
  }
  /// write the set of landmarks in a simplex text file
  void write(char *name);
};


}


#endif
