#ifndef MULTIVIEW_H
#define MULTIVIEW_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/MultiView.h"
#include <tcl.h>
#include <basics/DPOwner.h>
#include <basics/Activable.h>
#include <modules/TclSon.h>
#include <graphics2D/TclSliceCamera.h>
#include <basics/Matrix4x4.h>

namespace yav {


/** \class MultiView
    \ingroup Inrimage
    \brief an object that synchronizes multiple SliceCamera */
class MultiView : public TclSon, public TclObject, 
		  public DPOwner<TclSliceCamera>,
                  public DPOwner<MultiView>,
                  public DPListElement<MultiView>,
		  public Activable {

 

 private:
  /// whether to display axis around cursor
  bool axis;
  /// multiview zoom factor
  double zoom;
  // multiview IVS flag
  bool ignoreVoxelSize;

  /// color name of axis
  TclString axisColor;
  /// color name of secondary axis
  TclString axisColor2;

  /// indicates whether this multiview should kill himself when its DPList becomes empty
  bool dieWhenEmpty;
      
  /// constructors common code
  void init(void);
  /** callback function for a DPList element being removed
      @param elt element being removed */
  virtual void dying(const TclSliceCamera *elt);

 public:
  /** @name constructors */
  //@{
  /** builds a multiview from a list of viewers
      @param m module this multiview is built in
      @param views viewers managed by this multiview
      @param window root window name */
  MultiView(TclModule *m,
	    const std::list<TclSliceCamera *>& views,
	    const TclString& window = EMPTY_STRING);

  /** builds an empty multiview
      @param m module this multiview is built in
      @param window root window name */
  MultiView(TclModule *m,
	    const TclString& window = EMPTY_STRING);
  //@}

  /** @name destructor */
  //@{
  /// destroys any attached object
  ~MultiView();
  //@}

  /// returns managed viewers list
  const std::list<TclSliceCamera *>& getViews(void) const {
    return DPOwner<TclSliceCamera>::elements;
  }

  /// map associating the transfo
  std::map<MultiView *, Matrix4x4<double> *> matrixMultiViewMap;


  /// overloads DPOwner<TclSliceCamera>::operator+=
  virtual void operator+=(TclSliceCamera *slicecam)
    throw(DPList<TclSliceCamera>::NullPointer,
	  DPList<TclSliceCamera>::AlreadyInList) { append(slicecam); }

  /// overloads DPOwner<TclSliceCamera>::operator-=
  virtual void operator-=(TclSliceCamera *slicecam)
    throw(DPList<TclSliceCamera>::NotFound) { remove(slicecam); }

  /** overloads DPOwner<TclSliceCamera>::append  */
  virtual void append(TclSliceCamera *elt)
    throw(DPList<TclSliceCamera>::NullPointer,
	  DPList<TclSliceCamera>::AlreadyInList);
  /** overloads DPOwner<TclSliceCamera>::remove */
  virtual void remove(TclSliceCamera *elt)
    throw(DPList<TclSliceCamera>::NotFound);

  /// get axis flag
  bool getAxis(void) const { return axis; }

  /// set axis flag
  void setAxis(bool _axis);

  // return object by ref or 0 if it is not found
  static MultiView *getObject(int argc, char **argv, int i = 1) {
    return
      dynamic_cast<MultiView *>(TclObjectKind::getObject(typeid(MultiView), argc, argv, i));
  }

  /// raise ambigours getRef function call
  virtual unsigned int getRef(void) const {
    return TclObject::getRef();
  }

  /// returns multiview zoom value
  double getScale(void) const { return zoom; }

  /// returns multiview IVS flag
  bool getIVS(void) const { return ignoreVoxelSize; }

  /// returns primary axis color
  const TclString& getAxisColor1(void) const { return axisColor; }

  /// returns secondary axis color
  const TclString& getAxisColor2(void) const { return axisColor2; }

  /// multiplicate multiview zoom factor by val (all attached views are zoomed)
  void setScale(double val);

  /// set multiview IVS flag (all attached views are concerned)
  void setIVS(bool val);

  /** set primary and secondary axis colors
      @param c1 primary axis color
      @param c2 secondary axis color */
  void setAxisColors(const TclString& c1, const TclString& c2) {
      if(c1 != EMPTY_STRING) axisColor = c1;
      if(c2 != EMPTY_STRING) axisColor2 = c2;
  }

  /// activates this multiview (activates all attached views)
  virtual void on(void);

  /// deactivates this multiview (deactivates all attached views)
  virtual void off(void);

  /// activation routine (void, only attached view activation is interesting)
  void activation(void);

  /// TCL new operator callback
  static int create(TclModule *m, int argc, char **argv);

  /// TCL -destroy command
  int destroyCmd(int argc, char **argv);
  /// TCL -add command
  int addCmd(int argc, char **argv);
  /// TCL -listmview command
  int listmviewCmd(int argc, char **argv);
  /// TCL -addmview command
  int addmviewCmd(int argc, char **argv);
  /// TCL -removeAllmview command
  int removeAllmviewCmd(int argc, char **argv);
  /// TCL -removemview command
  int removemviewCmd(int argc, char **argv);
  /// TCL -remove command
  int removeCmd(int argc, char **argv);
  /// TCL -views command
  int viewsCmd(int argc, char **argv);
  /// TCL -parentWidget command
  int parentWidgetCmd(int argc, char **argv);
  /// TCL -mousevoxel command
  int mousevoxelCmd(int argc, char **argv);
  /// TCL -changeview command
  int changeViewCmd(int argc, char **argv);
  /// TCL -axis command
  int axisCmd(int argc, char **argv);
  /// TCL -axiscolor command
  int axisColorCmd(int argc, char **argv);
  /// TCL -activate command
  int activateCmd(int argc, char **argv);
  /// TCL -ignoreVoxelSize command
  int ignoreVoxelSizeCmd(int argc, char **argv);
  /// TCL -scale command
  int scaleCmd(int argc, char **argv);
  /// TCL -redraw command
  int redrawCmd(int argc, char **argv);
  /// TCL -min command
  int minCmd(int argc, char **argv);
  /// TCL -max command
  int maxCmd(int argc, char **argv);
  /// TCL -inrimage command
  int inrimageCmd(int argc, char **argv);
  /// TCL -visible command
  int visibleCmd(int argc, char **argv);
  /// TCL -scene command
  int sceneCmd(int argc, char **argv);
  /// TCL -die command
  int dieCmd(int argc, char **argv);

};


} // namespace yav


#endif
