#ifndef MULTIVIEW4D_H
#define MULTIVIEW4D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/MultiView4D.h"
#include <tcl.h>
#include <basics/DPOwner.h>
#include <basics/Activable.h>
#include <modules/TclSon.h>
#include <graphics2D/TclTimeSliceCamera.h>


namespace yav {


/** \class MultiView4D
    \ingroup Inrimage
    \brief an object that synchronizes multiple TimeSliceCamera */
class MultiView4D : public TclSon, public TclObject, 
		  public DPOwner<TclTimeSliceCamera>,
		  public Activable {

 private:
  /// whether to display axis around cursor
  bool axis;
  /// multiview zoom factor
  double zoom;
  // multiview IVS flag
  bool ignoreVoxelSize;

  /// color name of axis
  TclString axisColor;
  /// color name of secondary axis
  TclString axisColor2;

  /// indicates whether this multiview should kill himself when its DPList becomes empty
  bool dieWhenEmpty;
      
  /// constructors common code
  void init(void);
  /** callback function for a DPList element being removed
      @param elt element being removed */
  virtual void dying(const TclTimeSliceCamera *elt);

 public:
  /** @name constructors */
  //@{
  /** builds a multiview from a list of viewers
      @param m module this multiview is built in
      @param views viewers managed by this multiview
      @param window root window name */
  MultiView4D(TclModule *m,
	    const std::list<TclTimeSliceCamera *>& views,
	    const TclString& window = EMPTY_STRING);

  /** builds an empty multiview
      @param m module this multiview is built in
      @param window root window name */
  MultiView4D(TclModule *m,
	    const TclString& window = EMPTY_STRING);
  //@}

  /** @name destructor */
  //@{
  /// destroys any attached object
  ~MultiView4D();
  //@}

  /// returns managed viewers list
  const std::list<TclTimeSliceCamera *>& getViews(void) const {
    return elements;
  }

  /// overloads DPOwner<TclTimeSliceCamera>::operator+=
  virtual void operator+=(TclTimeSliceCamera *slicecam)
    throw(DPList<TclTimeSliceCamera>::NullPointer,
	  DPList<TclTimeSliceCamera>::AlreadyInList) { append(slicecam); }

  /// overloads DPOwner<TclTimeSliceCamera>::operator-=
  virtual void operator-=(TclTimeSliceCamera *slicecam)
    throw(DPList<TclTimeSliceCamera>::NotFound) { remove(slicecam); }

  /** overloads DPOwner<TclTimeSliceCamera>::append  */
  virtual void append(TclTimeSliceCamera *elt)
    throw(DPList<TclTimeSliceCamera>::NullPointer,
	  DPList<TclTimeSliceCamera>::AlreadyInList);
  /** overloads DPOwner<TclTimeSliceCamera>::remove */
  virtual void remove(TclTimeSliceCamera *elt)
    throw(DPList<TclTimeSliceCamera>::NotFound);

  /// get axis flag
  bool getAxis(void) const { return axis; }

  /// set axis flag
  void setAxis(bool _axis);

  /// raise ambigours getRef function call
  virtual unsigned int getRef(void) const {
    return TclObject::getRef();
  }

  /// returns multiview zoom value
  double getScale(void) const { return zoom; }

  /// returns multiview IVS flag
  bool getIVS(void) const { return ignoreVoxelSize; }

  /// returns primary axis color
  const TclString& getAxisColor1(void) const { return axisColor; }

  /// returns secondary axis color
  const TclString& getAxisColor2(void) const { return axisColor2; }

  /// multiplicate multiview zoom factor by val (all attached views are zoomed)
  void setScale(double val);

  /// set multiview IVS flag (all attached views are concerned)
  void setIVS(bool val);

  /** set primary and secondary axis colors
      @param c1 primary axis color
      @param c2 secondary axis color */
  void setAxisColors(const TclString& c1, const TclString& c2) {
      if(c1 != EMPTY_STRING) axisColor = c1;
      if(c2 != EMPTY_STRING) axisColor2 = c2;
  }

  /// activates this multiview (activates all attached views)
  virtual void on(void);

  /// deactivates this multiview (deactivates all attached views)
  virtual void off(void);

  /// activation routine (void, only attached view activation is interesting)
  void activation(void);

  /// TCL new operator callback
  static int create(TclModule *m, int argc, char **argv);

  /// TCL -destroy command
  int destroyCmd(int argc, char **argv);
  /// TCL -add command
  int addCmd(int argc, char **argv);
  /// TCL -remove command
  int removeCmd(int argc, char **argv);
  /// TCL -views command
  int viewsCmd(int argc, char **argv);
  /// TCL -parentWidget command
  int parentWidgetCmd(int argc, char **argv);
  /// TCL -mousevoxel command
  int mousevoxelCmd(int argc, char **argv);
  /// TCL -changeview command
  int changeViewCmd(int argc, char **argv);
  /// TCL -axis command
  int axisCmd(int argc, char **argv);
  /// TCL -axiscolor command
  int axisColorCmd(int argc, char **argv);
  /// TCL -activate command
  int activateCmd(int argc, char **argv);
  /// TCL -ignoreVoxelSize command
  int ignoreVoxelSizeCmd(int argc, char **argv);
  /// TCL -scale command
  int scaleCmd(int argc, char **argv);
  /// TCL -redraw command
  int redrawCmd(int argc, char **argv);
  /// TCL -min command
  int minCmd(int argc, char **argv);
  /// TCL -max command
  int maxCmd(int argc, char **argv);
  /// TCL -inrimage command
  int inrimageCmd(int argc, char **argv);
  /// TCL -visible command
  int visibleCmd(int argc, char **argv);
  /// TCL -scene command
  int sceneCmd(int argc, char **argv);
  /// TCL -die command
  int dieCmd(int argc, char **argv);
};


} // namespace yav


#endif
