#ifndef ROIVOXELLIST_H
#define ROIVOXELLIST_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif


#include <inrimage/Inrimage.h>
#include <inrimage/connectivity.h>
#include <basics/GeneralFormat.h>

namespace yav {


/** basics information related to a voxel : position and intensity */
struct ROIVoxel {
  /// voxel value
  double voxelValue;
  /// voxel position
  Vec3<unsigned int> position;  
};

/** \class ROIVoxelList
    \ingroup Inrimage
    \brief a region of interest in a volumetric image */
class ROIVoxelList : public GeneralFormat, public Owner<Inrimage> {
 protected:
  /// voxel size
  Vec3<double> voxelSize;
  /// 2D image slice associated with the region
  Inrimage *slice;
  /// offset for the slice
  Vec3<unsigned int> offset;
  /// the number of rows, columns and planes
  Vec3<unsigned int> imageSize;
  /// voxel map to find a voxel given its coordinates
  std::map<unsigned int,ROIVoxel *> voxelMap;
  /// file name
  TclString filename;
  ///  name of the region of interest
  TclString realName;
 protected:
  /** initialize General Format modules */
  void initFormat();
  /** creates the voxel map from the voxel list*/
  void buildVoxelMap(const std::list<ROIVoxel *> &l);
 public: 
  /** @name Exception classes */
  //@{
  /** \class ImageMissing
      \ingroup Inrimage
      \brief image missing exception */
  class ImageMissing : public Exception {
  public:
    ImageMissing(const std::string& from) : Exception(from,"creating an ROIVoxelList with a null image pointer") {}
  };  
  /** \class BinaryImageWrongSize
      \ingroup Inrimage
      \brief binary image does not have the same size as the grey-level image */
  class BinaryImageWrongSize : public Exception {
  public:
    BinaryImageWrongSize(const std::string& from) : Exception(from,"attempt to create an ROIVoxelList from a binary image that does not have the same size as the grey-level image") {}
  };  
  //@}
  /** @name constructor */
  //@{
  /** build a new region of interest in a volumetric image
      @param svl a list of voxel located on the region boundary
      @param _vsize gives the voxel size for display 
      @param _imageSize the number of rows, colums and planes
      @param im the image that gives the grey-level values*/
  ROIVoxelList(const std::list<ROIVoxel *> &svl, 
	       Vec3<double> _vsize=Vec3<double>(0.,0.,0.),
	       Vec3<unsigned int> _imageSize=Vec3<unsigned int>(1,1,1),
	       Inrimage *im=0);
  /** build an empty region of interest in a volumetric image
      @param im the image that gives the grey-level values*/
  ROIVoxelList(Inrimage *im)  throw(ImageMissing);
  /** build a region of interest from a binary image and a grey-value image
      @param image a 3D grey-value image where the ROI is defined : the grey-level value of each voxel is used from this original image as well as the image size
      @param binary a 3D binary image where each voxel of non-zero value is considered to belong to the region of interest. It must have the same size as the original image */
  ROIVoxelList(Inrimage *image, 
	       const Inrimage *binary) throw(ImageMissing,
					     BinaryImageWrongSize);
  /** build a region of interest from a file 
      @param filename the name of the file    */
  ROIVoxelList(char *filename);
 private:
  /** actual routine for doing erosion/dilation */
  void dilateOrErode(CONNECTIVITY_TYPE connectivity, 
		     unsigned int niter, bool dilate);

 protected:
  /** constructor that is only used by inherited class : builds an empty
      list of voxels 
      @param _vsize voxel size 
      @param _imageSize the number of rows, columns and planes */
  ROIVoxelList(Vec3<double> _vsize=Vec3<double>(0.,0.,0.),
	       Vec3<unsigned int> _imageSize=Vec3<unsigned int>(1,1,1)); 
   //@}
 public:
  /** @name destructor */
  //@{
  /// empty
  virtual ~ROIVoxelList();
  //@}
  /// return this object file name
  TclString getFilename(void) const { return filename; }
   /// return this region name
  TclString getRealName(void) const { return realName; }
  /// set the region name
  void setRealName(const TclString n) {
    realName=n; }
  /// return the grey-level image associated with this ROI
  Inrimage *getImage(void) const { return getPointer(); }
  /// return the grey-level image associated with this ROI
  void setImage(Inrimage *inr) { 
    pointer=inr; }
  /** creates and returns a new binary image (of type unsigned char) 
      having the same size as  its associated grey-level image and where 
      voxels of the ROI are set to 255 */
  Inrimage *createBinaryImage() const;

  /// write the main description of a ROI
  void writeMain(GFOutput &output);
  /// read the main description of a ROI
  void readMain(GFInput &input);
  /// skip the main description of a ROI
  void skipMain(GFInput &input);
  /// write the ROI as a file
  void write(TclString fname);
  /// return first voxel in list
  std::map<unsigned int,ROIVoxel *>::const_iterator getFirstVoxel() const {
    return voxelMap.begin();
  } 
  /// return first voxel in list
  std::map<unsigned int,ROIVoxel *>::const_iterator getLastVoxel() const {
    return voxelMap.end();
  } 

  /** add a list of voxels in the ROI  */
  virtual void addVoxel(std::list<ROIVoxel *> vlist);
  /** add a list of voxels in the ROI given their image coordinates without checking that the voxel is already present
      @param plist the list of voxel coordinates
  */
  virtual void addVoxel(std::list< Vec3<unsigned int>  > &plist);
  /** remove a list of voxels in the ROI 
      @param plist the list of voxel coordinates
      @return a boolean : true if all voxels have been removed and false otherwise*/
  virtual bool removeVoxel(std::list< Vec3<unsigned int>  > &plist);
  /** returns the voxel located at (x,y,z) and 0 if no voxel is available at that location */
  ROIVoxel *getVoxel(const unsigned int x,const unsigned int y,const unsigned int z) const;
  /// called when image is destroyed
  virtual void dying(const ObjectPointer<Inrimage>& obj);
  /** dilate the set of voxels 
      @param connectivity type of connectivity for the dilation
      @param nbIterations the number of dilation operations */
  void dilate(CONNECTIVITY_TYPE connectivity,unsigned int nbIterations=1);
  /** erode the set of voxels 
      @param connectivity type of connectivity for the dilation
      @param nbIterations the number of dilation operations */
  void erode(CONNECTIVITY_TYPE connectivity,unsigned int nbIterations=1);
  /// creates and return the  Histogram of this ROI
  GenericHistogram *createHistogram() const;
  /** compute the mean and variance of the ROI in a robust manner
      @param mean the computed mean value
      @param variance the computed variance
      @param rejectRate the rate (between 0 and 1 ) of rejection of outliers. For rejectRate=0 then Euclidean distance is used. If rejectRate=1 then the mean is the median value */ 
  void computeMeanVariance(double &mean, double &variance, 
			   double rejectRate=0.0) const;
  /** compute the centroid of all voxels in world coordinates as well as the smallest radius of the sphere englobing all voxels */ 
  void getCenterAndRadius(Vec3<double> &center, double &radius) const;
};




} // namespace yav

#endif


