#ifndef STRUCTURINGELEMENT_H
#define STRUCTURINGELEMENT_H


#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/StructuringElement.h"

#include <inrimage/Inrimage.h>

#include <inrimage/connectivity.h>


namespace yav {







typedef struct {
  int x;
  int y;
  int z;
  int o;
} typeStructuringElementPoint;



/** \class StructuringElement
    \ingroup Inrimage
    \brief Class for structuring elements 

    This class is used either for 
    mathematical morphology procedures
    or for median filtering.

    A structuring element is either a pre-defined
    connectivity or a list of points (ie neighbors).
*/
class StructuringElement {

  /** @name Exception classes 
   */
  //@{
 public:
#ifdef LIBINRIMAGE
  /** \class ElementException
      \ingroup Inrimage
  */
  class ElementException : public std::exception {
  public:
    ElementException() {}
    virtual ~ElementException() {}
  };
  /** \class InvalidElement
      \ingroup Inrimage
  */
  class InvalidElement : public ElementException {
    public: 
    InvalidElement() : ElementException() {}
  };

#else
  /** \class InvalidElement
      \ingroup Inrimage
      \brief structuring element is not valid */
  class InvalidElement : public Exception {
  public:
    InvalidElement(const std::string& from);
    InvalidElement(const std::string& from, 
		   const std::string& error);
  };
#endif // #ifdef LIBINRIMAGE
  //@}



 protected:
  /** @name Parameters
   */
  //@{  
  /// a pre-defined connectivity
  CONNECTIVITY_TYPE connectivity;
  /// the minimal value of the SE center in image
  int centre_value;
  /// the number of points in the list, if any
  int nb;
  /// a pointer to the list of points
  typeStructuringElementPoint *list;
  //@}

 private:
  /// initialisation des champs
  void init( void );
  /// construction a partir d'une image
  int buildFromImage( const Inrimage *image );


 public:

  /** @name Public methods
   */
  //@{  

  /** get the connectivity of the structuring
      element
  */
  int  getConnectivity( void ) { return connectivity; }

  /** set a pointer to the list of neighbors
   */
  void setArrayOfPoints( typeStructuringElementPoint *l ) { list = l; }

  /** get the pointer to the list of neighbors
   */
  typeStructuringElementPoint *getArrayOfPoints( void ) { return list; }

  /** set the number of neighbors
   */
  void setNumberOfPoints( const int n ) { nb = n; }

  /** get the number of neighbors
   */
  int  getNumberOfPoints( void ) { return nb; }

  /** replicates the structuring element
   */
  StructuringElement& operator=(const StructuringElement& s);

  /** Transform any structuring element into a 
      list of neighbors
  */
  void convertToList( void );

  /** Invert the structuring element,
      ie compute the symetric
  */
  void inverse( void );

  //@}



 public:
  void print( void );


 public:
  /** @name Constructors */
  //@{
  /** Create structuring element from connectivity
      @param C connectivity (should be 4, 8, 6, 10 or 26)
      @exception InvalidElement invalid connectivity 
  */
  StructuringElement( const CONNECTIVITY_TYPE C=TWENTY_SIX_CONNECTIVITY )
    throw( InvalidElement );
  /** Create structuring element from a rectangular neighborhood
      The center is supposed to be x/2 y/2 z/2
      @param x window size along X dimension
      @param y window size along Y dimension
      @param z window size along Z dimension
      @exception InvalidElement invalid connectivity 
  */
  StructuringElement( const int x, const int y, const int z = 1 )
    throw( InvalidElement );
  /** Create structuring element from an image
      whose path is given as argument 
      @param name image file name
  */
  StructuringElement( const char *name );
  /** Create structuring element from an image
      @param image image
  */
  StructuringElement( const Inrimage *image )
    throw( InvalidElement );
  /** Create an replica of elem
      @param elem StructuringElement to copy
  */
  StructuringElement( const StructuringElement &elem );


  //@}


  /** @name Destructor */
  //@{
  /// deallocates image buffer
  virtual ~StructuringElement();
  //@}



};
  





/** @name Mathematical Morphology Functions 
 */

//@{


/** Erosion of an image. 
    The erosion is defined as the minimum value over the
    neighborhood.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void erosion(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Erosion of an image. 
    The erosion is defined as the minimum value over the
    neighborhood.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void erosion(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception, ErrMemFull, ErrUnknown, StructuringElement::InvalidElement );

/** Erosion of an image. 
    The erosion is defined as the minimum value over the
    neighborhood.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void erosion(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY, 
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Erosion of an image. 
    The erosion is defined as the minimum value over the
    neighborhood.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void erosion(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY, 
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );

/** Erosion of an image. 
    The erosion is defined as the minimum value over the
    neighborhood.
    @param image
    @param radius     Radius of the euclidian structuring element
		@param iterations Number of iterations to be performed with the
		                  radius
		@param dimension  Dimension of the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void erosion(Inrimage *image, 
						 int radius, 
						 int iterations = 1,
						 int dimension=3, 
						 bool binary = false)
  throw( Exception );



/** Dilation of an image. 
    The dilation is defined as the maximum value over the
    neighborhood.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void dilation(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Dilation of an image. 
    The dilation is defined as the maximum value over the
    neighborhood.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void dilation(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception, ErrMemFull, ErrUnknown, StructuringElement::InvalidElement );

/** Dilation of an image. 
    The dilation is defined as the maximum value over the
    neighborhood.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void dilation(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Dilation of an image. 
    The dilation is defined as the maximum value over the
    neighborhood.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void dilation(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );



/** Dilation of an image. 
    The dilation is defined as the maximum value over the
    neighborhood.
    @param image
    @param radius     Radius of the euclidian structuring element
		@param iterations Number of iterations to be performed with the
		                  radius
		@param dimension  Dimension of the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void dilation(Inrimage *image, int radius, int iterations = 1,
							int dimension=3, bool binary = false)
  throw( Exception );




/** Opening of an image. 
    The opening is defined as an erosion with the structuring element
    followed by a dilation with the symetric of the structuring
    element. Incorrectly, it is often defined as an erosion followed
    by a dilation, which is only true if the structuring element
    is symetrical.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void opening(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Opening of an image. 
    The opening is defined as an erosion with the structuring element
    followed by a dilation with the symetric of the structuring
    element. Incorrectly, it is often defined as an erosion followed
    by a dilation, which is only true if the structuring element
    is symetrical.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void opening(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception, StructuringElement::InvalidElement );

/** Opening of an image. 
    The opening is defined as an erosion with the structuring element
    followed by a dilation with the symetric of the structuring
    element. Incorrectly, it is often defined as an erosion followed
    by a dilation, which is only true if the structuring element
    is symetrical.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void opening(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Opening of an image. 
    The opening is defined as an erosion with the structuring element
    followed by a dilation with the symetric of the structuring
    element. Incorrectly, it is often defined as an erosion followed
    by a dilation, which is only true if the structuring element
    is symetrical.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void opening(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );




/** Closing of an image. 
    The closing is defined as a dilation with the structuring element
    followed by an erosion with the symetric of the structuring
    element. Incorrectly, it is often defined as a dilation followed
    by an erosion, which is only true if the structuring element
    is symetrical.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void closing(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Closing of an image. 
    The closing is defined as a dilation with the structuring element
    followed by an erosion with the symetric of the structuring
    element. Incorrectly, it is often defined as a dilation followed
    by an erosion, which is only true if the structuring element
    is symetrical.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void closing(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception, StructuringElement::InvalidElement );

/** Closing of an image. 
    The closing is defined as a dilation with the structuring element
    followed by an erosion with the symetric of the structuring
    element. Incorrectly, it is often defined as a dilation followed
    by an erosion, which is only true if the structuring element
    is symetrical.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void closing(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Closing of an image. 
    The closing is defined as a dilation with the structuring element
    followed by an erosion with the symetric of the structuring
    element. Incorrectly, it is often defined as a dilation followed
    by an erosion, which is only true if the structuring element
    is symetrical.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void closing(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );




/** Top Hat Transform of an image. 
    The Top Hat Transform is defined as the substraction of the
    input image by its opening.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void topHat(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Top Hat Transform of an image. 
    The Top Hat Transform is defined as the substraction of the
    input image by its opening.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void topHat(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );

/** Top Hat Transform of an image. 
    The Top Hat Transform is defined as the substraction of the
    input image by its opening.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void topHat(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Top Hat Transform of an image. 
    The Top Hat Transform is defined as the substraction of the
    input image by its opening.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void topHat(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );




/** Inverse Top Hat Transform of an image. 
    The inverse Top Hat Transform is defined as the substraction of the
    closing by the input image.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void invTopHat(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Inverse Top Hat Transform of an image. 
    The inverse Top Hat Transform is defined as the substraction of the
    closing by the input image.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void invTopHat(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception, StructuringElement::InvalidElement );

/** Inverse Top Hat Transform of an image. 
    The inverse Top Hat Transform is defined as the substraction of the
    closing by the input image.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void invTopHat(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Inverse Top Hat Transform of an image. 
    The inverse Top Hat Transform is defined as the substraction of the
    closing by the input image.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void invTopHat(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );




/** Morphological Gradient of an image. 
    It is defined by the substraction of the dilation with
    the erosion. Strictly speaking, one should divided it by 
    two, but it is not done here.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalGradient(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Morphological Gradient of an image. 
    It is defined by the substraction of the dilation with
    the erosion. Strictly speaking, one should divided it by 
    two, but it is not done here.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalGradient(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );

/** Morphological Gradient of an image. 
    It is defined by the substraction of the dilation with
    the erosion. Strictly speaking, one should divided it by 
    two, but it is not done here.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalGradient(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Morphological Gradient of an image. 
    It is defined by the substraction of the dilation with
    the erosion. Strictly speaking, one should divided it by 
    two, but it is not done here.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalGradient(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );




/** Morphological Lower Gradient of an image. 
    It is defined by the substraction of the original
    image with the erosion.
    two, but it is not done here.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalLowerGradient(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Morphological Lower Gradient of an image. 
    It is defined by the substraction of the original
    image with the erosion.
    two, but it is not done here.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalLowerGradient(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );

/** Morphological Lower Gradient of an image. 
    It is defined by the substraction of the original
    image with the erosion.
    two, but it is not done here.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalLowerGradient(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Morphological Lower Gradient of an image. 
    It is defined by the substraction of the original
    image with the erosion.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalLowerGradient(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );





/** Morphological Upper Gradient of an image. 
    It is defined by the substraction of the dilation with the 
    the original image.
    two, but it is not done here.
    @param image
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalUpperGradient(  Inrimage *image, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Morphological Upper Gradient of an image. 
    It is defined by the substraction of the dilation with the 
    the original image.
    two, but it is not done here.
    @param inputImage
    @param outputImage
    @param se
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalUpperGradient(  Inrimage *inputImage, Inrimage *outputImage, 
	       StructuringElement *se,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );

/** Morphological Upper Gradient of an image. 
    It is defined by the substraction of the dilation with the 
    the original image.
    @param image
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalUpperGradient(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );
 
/** Morphological Upper Gradient of an image. 
    It is defined by the substraction of the dilation with the 
    the original image.
    @param inputImage
    @param outputImage
    @param C
    @param iterations number of iterations to be performed
           with the structuring element
    @param binary if true specialized binary operations are used
           (may be interesting especially in 3D, and/or for a large 
            number of iterations)
*/
void morphologicalUpperGradient(  Inrimage *inputImage, Inrimage *outputImage, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY,
	       int iterations = 1,
	       bool binary = false )
  throw( Exception );



/** Watershed of an image. 
    @param gradientImage
    @param inputLabelsImage
    @param outputLabelsImage must be of same type as inputLabelsImage

    Image types have to be unsigned char, unsigned short, or unsigned int.
    
 */
void watershed( Inrimage *gradientImage,
		Inrimage *inputLabelsImage,
		Inrimage *outputLabelsImage )
  throw( Exception, ErrMemFull, ErrUnknown );


//@}




//
// MEDIAN FILTERING
//

/** @name Median Filtering Functions 
 */

//@{


/** apply a median filtering on image
    @param image image to be filtered
    @param se structuring element defining the neighborhood from
              which the median value is computed 
	      
    The central point must belong to the structuring element
    (hence there is at least one defined value), else an
    execption is thrown.
*/
void medianFiltering(  Inrimage *image, 
			StructuringElement *se )
   throw( Exception );
 
 /** apply a median filtering on inputImage image and stores the result in
     outputImage image. The outputImage must be allocated BEFORE the function
     call.
     @param inputImage
     @param outputImage
     @param se structuring element defining the neighborhood from
            which the median value is computed 
	      
    The central point must belong to the structuring element
    (hence there is at least one defined value), else an
    execption is thrown.
 */
void medianFiltering(  Inrimage *inputImage, Inrimage *outputImage,
			StructuringElement *se )
   throw( Exception, ErrMemFull, ErrUnknown, StructuringElement::InvalidElement );
 
 /** apply a median filtering on image
    @param image image to be filtered
    @param C the neighborhood from which the median value is computed 
	      
    The central point must belong to the structuring element
    (hence there is at least one defined value), else an
    execption is thrown.
*/
void medianFiltering(  Inrimage *image, 
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY )
  throw( Exception );
 
/** apply a median filtering on image
    @param inputImage
    @param outputImage
    @param C the neighborhood from which the median value is computed 
	      
    The central point must belong to the structuring element
    (hence there is at least one defined value), else an
    execption is thrown.
*/
void medianFiltering( Inrimage *inputImage, Inrimage *outputImage,
	       CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY )
  throw( Exception );


//@}





} // namespace yav

#endif // #ifndef STRUCTURINGELEMENT_H
