#ifndef TCLBOUNDINGBOX3D_H
#define TCLBOUNDINGBOX3D_H

#pragma interface "inrimage/TclBoundingBox3D.h"
#include <graphics2D/Data2D.h>
#include <graphics2D/Camera2D.h>
#include <graphics2D/TclScene2DObject.h>


namespace yav {

/// threshold distance in world coordinates used to select a corner of the box
#define BOUNDINGBOX3D_SELECTION_DISTANCE 10.0

class TclBoundingBox3D;

/** \class BoundingBox3D
    \ingroup Inrimage
    \brief a 3D box in a 2D scene that is used to select a Region of Interest */
class BoundingBox3D : public Data2D  {
 private:
  /// the two opposite corners in 3D world
  Vec3<double> corner[2];

  /// drawing color
  float color[3];
  /// whether has selected a corner
  bool selectedCorner;
  /// which corner should be modified when the selected corner is interactively moved
  int modificationIndex[2];

 public:
  /** @name constructor */
  //@{
  /** creates a box
      @param m module to create TCL object
      @param scene scene that object lies in
      @param x1 box first x coordinate
      @param y1 box first y coordinate
      @param z1 box first z coordinate
      @param x2 box second x coordinate
      @param y2 box second y coordinate
      @param z2 box second z coordinate
      @param color box color
      @param menu mouse button popup menu */
  BoundingBox3D(Scene2D *scene, 
		 const Vec3<double> c[2],
		 const float color[3]);
  //@}

  /** @name destructor */
  //@{
  /// destroys this box
  virtual ~BoundingBox3D();
  //@}


/** callback called when a button is pressed in a camera
     @param button indicates which button is pressed
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonPressed(CameraTypes::WHICH_BUTTON button,
				  CameraTypes::KEY_MODIFIER modifier,
				  Camera2D *cam,int mx,int my);
 /** callback called when a button is released in a camera
     @param button indicates which button is released
     @param modifier indicates if a modifier (shift or control key) is pressed
     @param cam the 2D camera where occured the click button
     @param mx first coordinate of the pixel clicked 
     @param my second coordinate of the pixel clicked */ 
  virtual void mouseButtonReleased(CameraTypes::WHICH_BUTTON button,
				   CameraTypes::KEY_MODIFIER modifier,
				     Camera2D *cam,int mx,int my);
 /** callback called when a button is moved after having been pressed
     @param cam the 2D camera where occured the click button
     @param x first coordinate of the pixel clicked 
     @param y second coordinate of the pixel clicked */ 
  virtual void mouseButtonMoved(Camera2D *cam,int x,int y);


  /** Translates the 2D object @param t translation vector */
  virtual void translate(const Vec2<double>& t) {}

  /** Rotates the 2D object: no rotation available for images
      @param r rotation matrix */
  virtual void rotate(const Rotation2D &r) {}

  /** Scales the 2D object
      @param s scale factor */
  virtual void scale(const double s) {}

  /// Returns the graphic object center and weight.
  virtual void updateCenter(Vec2<double>& center, unsigned int& n)  {
    center = Vec2<double>(0.0, 0.0);
    n = 0;
  }

  /// Returns the graphic object radius given its center
  virtual double updateRadius(const Vec2<double>& center)  { return 0.0; }

 
 /// nothing done when an image is selected
  virtual void select(Camera2D *) {}

  /// nothing done when an image is unselected
  virtual void unselect(Camera2D *) {}

  /** default redraw function for sliceable objects
      @param cam camera to redraw */
  virtual void redraw(Camera2D *cam);

  friend class TclBoundingBox3D;
};


/** \class TclBoundingBox3D
    \ingroup Inrimage
    \brief The interface around a BoundingBox3D  */
class TclBoundingBox3D : public TclScene2DObject  {
 private:

 public:
  /** @name constructor */
  //@{
  /** creates a box
      @param m module to create TCL object
      @param scene scene that object lies in
      @param c two opposite corners defining the 3D box (in 3D world coordinates)
      @param color box color (RGB)
      @param menu mouse button popup menu */
  TclBoundingBox3D(TclModule *m,
		 Scene2D *scene, 
		 const Vec3<double> c[2],
		 const float color[3],
		 const TclString& menu = EMPTY_STRING);
  //@}
protected:
  /// right mouse button popup menu
  TclString menu;
public:

  /** @name destructor */
  //@{
  /// destroys this box
  virtual ~TclBoundingBox3D();
  //@}
  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// returns the associated bounding box
  BoundingBox3D *getClient() const {
	  return ((BoundingBox3D *) client);
  }
  /// create the tcl sub options for that object
  static void createCommands(TclModule *m); 
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -boxSize option
  int boxSizeCmd(int argc, char **argv);
};


} // namespace yav


#endif
