#ifndef TCLINRIMAGE_H
#define TCLINRIMAGE_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclInrimage.h"
#include <basics/Pointed.h>
#include <modules/TclObject.h>
#include <graphics/TclSceneObject.h>
#include <graphics2D/TclScene2D.h>
#include <inrimage/StructuringElement.h>

namespace yav {


  //class BoundingBox;

/** \class TclInrimage
    \ingroup Inrimage
    \brief Tcl interface for any inrimage */
class TclInrimage : public TclObject, public Owner<Inrimage> 
  {
 protected:
  /** @name constructors */
  //@{
  /** Reads inrimage from file whose path is given as argument.
      File can be an INRIMAGE(4) gzipped or not, an anlayse, a GIF or a RGB
      irix file.
      @param info dynamic object type
      @param m module this image was created in
      @param name inrimage file name
      @param headerOnly only image header is read if headerOnly is set to true
      @param vm if image has a vectorial type, it can be read interlaced (default) or non interlaced by giving WM_NON_INTERLACED as third argument
      @param filevm if VM_NON_INTERLACED, data is considered as non-interlaced
      @exception ImageNotFound image file not found
      @exception UnknownImageType unsupported inrimage type
      @exception InvalidHeader invalid inrimage header
      @exception InvalidData invalid image data
      @exception UnknownType unknown file format */
  TclInrimage(const std::type_info& info,
	      TclModule *m, const TclString& name,
	      bool headerOnly = false, 
	      VECTORIAL_MODE vm = VM_INTERLACED,
	      VECTORIAL_MODE filevm = VM_INTERLACED)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);

  /** Creates an inrimage with given size and word type.
      The inrimage data buffer is allocated
      @param info dynamic object type
      @param m this image was created in
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
  */
  TclInrimage(const std::type_info& info, TclModule *m,
	      unsigned int xdim, unsigned int ydim, unsigned int zdim,
	      unsigned int vdim = 1, unsigned int wdim = 1,
	      WORD_KIND wordKind = WK_FIXED, SIGN sign = SGN_SIGNED,
	      float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize);
  
  /** creates an inrimage from its size and its type
      @param info dynamic object type
      @param m module this image was created in
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param wt image type
      @param vdim vectorial dimension (for WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR only, ignored otherwise)
      @param vm vectorial mode (for WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR only, ignored otherwise)
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize xdim, ydim and zdim should be > 0
      @exception UnknownType unsupported inrimage type
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  TclInrimage(const std::type_info& info, TclModule *m,
	      unsigned int xdim, unsigned int ydim, unsigned int zdim,
	      Inrimage::WORD_TYPE wt, unsigned int vdim = 1,
	      VECTORIAL_MODE vm = VM_NON_INTERLACED,
	      float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize);
  //@}

 public:
  /** @name constructors */
  //@{
  /** Creates an interface object for an existing inrimage.
      @param m module this image was created in
      @param info dynamic object type
      @param data inrimage object */
  TclInrimage(TclModule *m, const std::type_info& info, Inrimage *data);
  /** Reads inrimage from file whose path is given as argument.
      File can be an INRIMAGE(4) gzipped or not, an anlayse, a GIF or a RGB
      irix file.
      @param m module this image was created in
      @param name inrimage file name
      @param headerOnly only image header is read if headerOnly is set to true
      @param vm if image has a vectorial type, it can be read interlaced (default) or non interlaced by giving WM_NON_INTERLACED as third argument
      @param filevm if VM_NON_INTERLACED, data is considered as non-interlaced
      @exception ImageNotFound image file not found
      @exception UnknownImageType unsupported inrimage type
      @exception InvalidHeader invalid inrimage header
      @exception InvalidData invalid image data
      @exception UnknownType unknown file format */
  TclInrimage(TclModule *m, const TclString& name,
	      bool headerOnly = false, 
	      VECTORIAL_MODE vm = VM_INTERLACED,
	      VECTORIAL_MODE filevm = VM_INTERLACED)
    throw(Inrimage::ImageNotFound, Inrimage::UnknownImageType,
	  Inrimage::InvalidHeader, Inrimage::InvalidData,
	  Inrimage::UnknownType);

  /** Creates an inrimage with given size and word type.
      The inrimage data buffer is allocated
      @param m this image was created in
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
  */
  TclInrimage(TclModule *m,
	      unsigned int xdim, unsigned int ydim, unsigned int zdim,
	      unsigned int vdim = 1, unsigned int wdim = 1,
	      WORD_KIND wordKind = WK_FIXED,
	      SIGN sign = SGN_SIGNED,
	      float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize);
  
  /** Creates an inrimage with given size and word type.
      If data is 0, the image is not allocated.
      If data is non 0, image data points to it.
      @param m module this image was created in
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param data image data buffer
      @param vdim image vectorial dimension
      @param wdim image word size in bytes
      @param wordKind image word kind
      @param sign image word sign
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize xdim, ydim and zdim should be > 0
      @exception UnknownType unsupported inrimage type
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  TclInrimage(TclModule *m,
	      unsigned int xdim, unsigned int ydim, unsigned int zdim,
	      void *data, unsigned int vdim = 1, unsigned int wdim = 1,
	      WORD_KIND wordKind = WK_FIXED,
	      SIGN sign = SGN_SIGNED,
	      float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize);

  /** creates an inrimage from its size and its type
      @param m module this image was created in
      @param xdim image number of columns
      @param ydim image number of rows
      @param zdim image number of planes
      @param wt image type
      @param vdim vectorial dimension (for WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR only, ignored otherwise)
      @param vm vectorial mode (for WT_FLOAT_VECTOR and WT_DOUBLE_VECTOR only, ignored otherwise)
      @param vx distance between voxels in the X direction
      @param vy distance between voxels in the Y direction
      @param vz distance between voxels in the Z direction
      @exception InvalidSize xdim, ydim and zdim should be > 0
      @exception UnknownType unsupported inrimage type
      @exception InvalidVoxelSize voxels size should be > 0.0 */
  TclInrimage(TclModule *m,
	      unsigned int xdim, unsigned int ydim, unsigned int zdim,
	      Inrimage::WORD_TYPE wt, unsigned int vdim = 1,
	      VECTORIAL_MODE vm = VM_NON_INTERLACED,
	      float vx = 1.0, float vy = 1.0, float vz = 1.0)
    throw(Inrimage::InvalidSize, Inrimage::UnknownType, Inrimage::InvalidVoxelSize);

  /** creates an replica of img
      @param img image to copy
      @exception InvalidSize image dimension should be > 0
      @exception InvalidVectorialMode img has scalar vectorial mode for vectorial image or opposite
      @exception InvalidVoxelSize img has <= 0 voxel size
      @exception NotEnoughMemory unable to allocate image buffer */
  TclInrimage(const TclInrimage& img)
    throw(Inrimage::InvalidSize, Inrimage::InvalidVectorialMode,
	  Inrimage::InvalidVoxelSize, Inrimage::NotEnoughMemory);
  //@}

  /** @name destructor */
  //@{
  // deallocates image buffers and destroy object
  virtual ~TclInrimage();
  //@}

  /// returns TclObject client
  Inrimage *getClient(void) const { return pointer(); }
  /// returns TclObject client
  //Inrimage *getInrimage(void) const { return (Inrimage *) client; }
  /// return image X dimension
  unsigned int getX(void) const { return getClient()->xdim; }
  /// return image Y dimension
  unsigned int getY(void) const { return getClient()->ydim; }
  /// return image Z dimension
  unsigned int getZ(void) const { return getClient()->zdim; }
  /// return image X voxel size
  double getVX(void) const { return getClient()->vx; }
  /// return image Y voxel size
  double getVY(void) const { return getClient()->vy; }
  /// return image Z voxel size
  double getVZ(void) const { return getClient()->vz; }
  /// return image words type
  Inrimage::WORD_TYPE getType(void) const {
    return getClient()->type;
  }
  /// return image vectorial dimension
  unsigned int getVdim(void) const { return getClient()->vdim; }
  /// get image words size (in bytes)
  unsigned int getWdim(void) const { return getClient()->vdim; }
  /// return image data buffer
  void *getData(void) const { return getClient()->data; }
  /// returns pointer a 3D buffer for a scalar image or interleaved vectorial image
  void ***getArray(void) const { return getClient()->array.threeD; }
  /// returns a 4D buffer for a vectorial non-interleaved image
  void ****getVarray(void) const { return getClient()->array.fourD; }

  /// get image word size
  unsigned int getWordSize() const { return getClient()->wdim; }
  /// get word kind
  WORD_KIND getWordKind() const { return getClient()->wordKind; }
  /// get word sign
  SIGN getWordSign() const { return getClient()->sign; }
  /// get VECTORIAL_MODE
  VECTORIAL_MODE getVectorialMode() const { return getClient()->vectMode; }


  /// refresh inrimage in any viewer it is displayed
  void redraw(void) const;

  /** refresh inrimage in any viewer it is displayed
      resize the viewer to the image size */
  void resizeAndRedraw(void) const;

  /** refresh inrimage in any viewer it is displayed
      resize the viewer to the image voxel size */
  void resizeVoxelsAndRedraw(void) const;

  /** warns all backgrounds that the image content has changed (not its size) */
  void changeAndRedraw(void);

  /** saves image as
      @param newname new image file name */
  void save(const TclString& newname);

  /// saves image using name
  void save(void) const { getClient()->write(getClient()->name); }

  /// crop command enum
  enum CROPARG {
    /// initialize new box for cropping
    CROP_INIT,
    /// cancel: delete crop box
    CROP_DROP,
    /// done crop along crop box
    CROP_DONE
  };
    
  static int parseBoxOptions(TclModule *m, TclInrimage *inr,
				       int argc, char **argv,
				       unsigned int& x1, unsigned int& y1,
				       unsigned int& z1,
				       unsigned int& x2, unsigned int& y2,
				       unsigned int& z2);
    
  /// returns the inrimage type corresponding to given string
  static Inrimage::WORD_TYPE typeFromString(char *);

  /// returns a static string corresponding to given type
  static char *typeToString(Inrimage::WORD_TYPE);

  /// returns the connectivity type corresponding to given string
  static yav::CONNECTIVITY_TYPE connectivityFromString(char *str);

  /// loadImage command callback
  static int load(TclModule *m, int argc, char **argv);
  /** loads an inrimage, catch exceptions and returns an error
      message in module result string if an error occured
      @param name image file name
      @param m module to create interface
      @param interlace whether vectorial data should be interlaced
      @param fileinterlace whether source file is interlaced
      @return inrimage loaded, 0 if an error occured */
  static TclInrimage *load(const TclString& name, TclModule *m,
			   bool interlace = true, bool fileinterlace = true);

  /// createInrimage command callback
  static int create(TclModule *m, int argc, char **argv);

  /// view command callback
  static int view(TclModule *m, int argc, char **argv);

  /// image format command callback
  static int imageFormatCmd(TclModule *m, int argc, char **argv);

  /// get TclInrimage from image pointer
  static TclInrimage *getInterface(Inrimage *inr);

  // return object by ref or 0 if it is not found
  static TclInrimage *getObject(unsigned int n) {
    return dynamic_cast<TclInrimage *>(TclObjectKind::getObject(typeid(TclInrimage),
								n));
  }
  /// die on inrimage  deletion
  virtual void dying(const ObjectPointer < Inrimage > &g) {
    if (g()==pointer()) {
      pointer= 0;	
      delete this;
    }
  }

  // return object by name (argv[i]) or 0 with an error message
  // in module interpretor if it is not found
  static TclInrimage *getObject(int argc, char **argv, int i = 1) {
    return
      dynamic_cast<TclInrimage *>(TclObjectKind::getObject(typeid(TclInrimage),
							   argc, argv, i));
  }

  // return object id by command line arguments
  static const char *getTCLObjectName(void) {
    return TclObjectKind::getObjectName(typeid(TclInrimage));
  }

  // return object id by command line arguments
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// file scene read method
  static void read(const TclString &filename, Scene2D *scn);

  /// -destroy command option
  int destroyCmd(int argc, char **argv);
  /// -name command option
  int nameCmd(int argc, char **argv);
  /// -size command option
  int sizeCmd(int argc, char **argv);
  /// -save command option
  int saveCmd(int argc, char **argv);
  /// -anisotropicDiffusion command option
  int anisotropicDiffusionCmd(int argc, char **argv);  
  /// -copy command option
  int copyCmd(int argc, char **argv);
  /// -histogram command option
  int histogramCmd(int argc, char **argv);
  /// -maskedHistogram command option
  int maskedHistogramCmd(int argc, char **argv);
  /// -redraw command option
  int redrawCmd(int argc, char **argv);
  /// -probabilityMap command option
  int probabilityMapCmd(int argc, char **argv); 
  /// -vectorial command option
  int vectorialCmd(int argc, char **argv);
  /// -floating command option
  int floatingCmd(int argc, char **argv);
  /// -glbounds command option
  int glboundsCmd(int argc, char **argv);
  /// -crop command option
  int cropCmd(int argc, char **argv);
  /// -isolateRegion command option
  int isolateRegionCmd(int argc, char **argv); 
  /// -zeroRegion command option
  int zeroRegionCmd(int argc, char **argv);
  /// -voxel command option
  int voxelCmd(int argc, char **argv);
  /// -voxelValue command option
  int voxelValueCmd(int argc, char **argv);
  /// -voxelInterpolatedValue command option
  int voxelInterpolatedValueCmd(int argc, char **argv);
  /// -windowing command option
  int windowingCmd(int argc, char **argv);
  /// -flip command option
  int flipCmd(int argc, char **argv);
    /// -line3D command option
  int line3DCmd(int argc, char **argv);
  /// -blockMean command option
  int blockMeanCmd(int argc, char **argv);
  /// -blockStandardDeviationCmd command option
  int blockStandardDeviationCmd(int argc, char **argv);
  /// -blockMedian command option
  int blockMedianCmd(int argc, char **argv);
  /// -maskCorrelation command option
  int maskCorrelationCmd(int argc, char **argv);
  /// -switch command option
  int switchCmd(int argc, char **argv);
  /// -mel command option
  int melCmd(int argc, char **argv);
  /// -type command option
  int typeCmd(int argc, char **argv);
  /// -artih command option
  int arithCmd(int argc, char **argv);
  /// -logic command option
  int logicCmd(int argc, char **argv);
  /// -threshold command option
  int thresholdCmd(int argc, char **argv);
  /// -lookup command option
  int lookupCmd(int argc, char **argv);
  int evalCmd(int argc,char **argv);
  /// -cropHistogram command option
  int cropHistogrammCmd(int argc, char **argv);
  /// -noise command option
  int noiseCmd(int argc, char **argv);
  /// -connexe command option
  int connexeCmd(int argc, char **argv);
  /// -chamfer command option
  int chamferCmd(int argc, char **argv);
  /// -fill command option
  int fillCmd(int argc, char **argv);
  /// -morpho command option
  int morphoCmd(int argc, char **argv);
  /// -gradient command option
  int gradientCmd(int argc, char **argv);
  /// -convolute command option
  int convoluteCmd(int argc, char **argv);
  /// -recfilters command option
  int recfilterCmd(int argc, char **argv);
  /// -laplacian command option
  int laplacianCmd(int argc, char **argv);
  /// -split command option
  int splitCmd(int argc, char **argv);
  /// -component command option
  int componentsCmd(int argc, char **argv);
  /// -dft command option
  int dftCmd(int argc, char **argv);
  /// -undersample command option
  int undersampleCmd(int argc, char **argv);
  /// -oversample command option
  int oversampleCmd(int argc, char **argv);
  /// -createSlicePhoto command option
  int createSlicePhotoCmd(int argc, char **argv);
  /// -resample command option
  int resampleCmd(int argc, char **argv);
  /// -transform command option
  int transformCmd(int argc, char **argv);
  /// -interpolation command option
  int interpCmd(int argc, char **argv);
  /// -extrapolation command option
  int extrapCmd(int argc, char **argv);
  /// -region command option
  int regionCmd(int argc, char **argv);
  /// -curvature command option
  int curvatureCmd(int argc, char **argv);
  /// -worldToVoxel command option
  int worldToVoxelCmd(int argc, char **argv);
  /// -voxelToWorld command option
  int voxelToWorldCmd(int argc, char **argv);
  /// -reformat command option
  int reformatCmd(int argc, char **argv);
  /// -transformationMatrix command option
  int globalTransformationCmd(int argc, char **argv);
  int globalTransformationToVectorCmd(int argc, char **argv);
     
    /// creates the Tcl commands when the module is loaded
  static void createCommands(TclModule *m);

 private:
  /// inrimage type strings
  static char *types[];
};

} // namespace yav


#endif
