#ifndef TCLINRIMAGE4D_H
#define TCLINRIMAGE4D_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclInrimage4D.h"
#include <modules/TclModule.h>
#include <graphics/SceneItem.h>
#include <inrimage/TclInrimage.h>
#include <inrimage/Inrimage4D.h>

namespace yav {

/** \class TclInrimage4D
    \ingroup Inrimage
    \brief 3D + T camera interface */
class TclInrimage4D : public TclObject, public Owner<Inrimage4D> {
 public:
  /** @name constructors */
  //@{
  /** Reads 4D inrimage from file whose path is given as argument.
      @param m module this image is created in
      @param name 4D inrimage file name
      @param info dynamic object type
      @param data client (if alreaedy created)
      @exception ImageNotFound image file not found
      @exception ReadError error while reading file */
  TclInrimage4D(TclModule *m, const char *name,
		const std::type_info& info = typeid(TclInrimage4D),
		Inrimage4D *data = 0)
    throw(GeneralFormat::FileNotFound,
	  Inrimage::ImageNotFound, Inrimage4D::ReadError);
  /** builds a 4D inrimage interface from a sequence of 3D inrimages
      @param m module this image is created in
      @param inrs list of constituting 3D inrimages
  */
  TclInrimage4D(TclModule *m, std::list<Inrimage *> inrs);

  /** creates 4D inrimage interface for an existing client
      @param m module this image was created in
      @param info dynamic object type
      @param data inrimage 4D client */
  TclInrimage4D(TclModule *m,
		const std::type_info& info, Inrimage4D *data);
  //@}

  /** @name destructor */
  //@{
  /// empty
  ~TclInrimage4D();
  //@}

  /// returns interface's client inrimage
  Inrimage4D *getClient(void) const { return pointer(); }

  /// image format command callback
  static int imageFormatCmd(TclModule *m, int argc, char **argv);


  /// -destroy TCL command option
  int destroyCmd(int argc, char **argv);
  /// -type TCL command option
  int typeCmd(int argc, char **argv);
  /// -name TCL command option
  int nameCmd(int argc, char **argv);
  /// -save TCL command option
  int saveCmd(int argc, char **argv);
  /// -glbounds TCL command option
  int glboundsCmd(int argc, char **argv);
  /// -voxel TCL command option
  int voxelCmd(int argc, char **argv);
  /// -voxel TCL command option
  int gradientCmd(int argc, char **argv);
 /// -redraw command option
  int redrawCmd(int argc, char **argv);
  /// -slicecams command option
  int slicecamsCmd(int argc, char **argv);
  /// -anisotropicDiffusion command option
  int anisotropicDiffusionCmd(int argc, char **argv);  
  /// -crop command option
  int cropCmd(int argc, char **argv);
  
  /// redraw inrimage
  void redraw(void) const;
  /** refresh inrimage in any viewer it is displayed
      resize the viewer to the image size */
  void resizeAndRedraw(void) const;
  
  /** refresh inrimage in any viewer it is displayed
      resize the viewer to the image voxel size */
  void resizeVoxelsAndRedraw(void) const;
  
  /** reset GL bounds to inrimage min and max
      refresh inrimage in any viewer it is displayed */
  void rescaleGLAndRedraw(void) const;
  
  /// redraw new inrimage in any viewer it is displayed
  void changeAndRedraw(void);
  

  /// new inr4D command callback
  static int create(TclModule *m, int argc, char **argv);
  // return object id by command line arguments
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// loadImage command callback
  static int load(TclModule *m, int argc, char **argv);
  /** loads an inrimage, catch exceptions and returns an error
      message in module result string if an error occured
      @param name image file name
      @param scn  2D scene where to create 4D inrimage  */
  static void load(const TclString& name, Scene2D *scn);
  /// returns object by ref or 0 if it is not found
  static TclInrimage4D *getObject(unsigned int n) {
    return (TclInrimage4D *)
      TclObjectKind::getObject(typeid(TclInrimage4D), n);
  }
  /** return object by name (argv[i]) or 0 with an error message
      in module interpretor if it is not found */
  static TclInrimage4D *getObject(int argc, char **argv, int i = 1) {
    return (TclInrimage4D *)
      TclObjectKind::getObject(typeid(TclInrimage4D), argc, argv, i);
  }
  /// returns object id by command line arguments
  static const char *getTCLObjectName(void) {
    return TclObjectKind::getObjectName(typeid(TclInrimage4D));
  }

  /// die on inrimage 4D deletion
  virtual void dying(const ObjectPointer < Inrimage4D > &g) {
    if (g()==pointer()) {
      pointer= 0;	
      delete this;
    }
  }
  /// get TclInrimage4D from image pointer
  static TclInrimage4D *getInterface(Inrimage4D *inr);
 
 private:
  /// constructors common code
  void init(TclModule *m);  
  /** Apply given command on each 3D inrimage composing the 4D inrimage.
      If argc = 0, the command is supposed to be a get command and
      it is executed only on first inrimage.
      If argv > 0, the command is supposed to be a set command and
      it is executed on each 3D inrimage. Result of the get command
      is then returned.
      @param cmd TclInrimage command to call
      @param argc number of command line arguments
      @param argv command line arguments */
  int applyCmd(int (TclInrimage::*cmd)(int, char **),
	       int argc, char **argv);
};

} // namespace yav

#endif // TCLINRIMAGE4D_H
