#ifndef TCLINRIMAGE4DBACKGROUND_H
#define TCLINRIMAGE4DBACKGROUND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclInrimage4DBackground.h"
#include <inrimage/Inrimage4DView.h>
#include <inrimage/TclInrimageBackground.h>


namespace yav {


/** \class VoxelCoordinate4D
    \ingroup Inrimage
    \brief 3D image coordinates */
class VoxelCoordinate4D {
 public:
  /// initialise coordinates to (0, 0, 0)
  VoxelCoordinate4D() : x(0), y(0), z(0), t(0) {}
  /// x coordinate (column)
  unsigned int x;
  /// y coordinate (row)
  unsigned int y;
  /// z coordinate (plane)
  unsigned int z;
  /// t coordinate (time)
  unsigned int t;
};

/** \class CartesianCoordinate4D
    \ingroup Inrimage
    \brief 3D cartesian point */
class CartesianCoordinate4D {
 public:
  /// initialise coordinates to (0, 0, 0)
  CartesianCoordinate4D() : x(0.0), y(0.0), z(0.0), t(0.0) {}
  /// x coordinate
  double x;
  /// y coordinate
  double y;
  /// z coordinate
  double z;
  /// time coordinate
  double t;
};
 
/** \class TclInrimage4DBackground
    \ingroup Inrimage
    \brief The background of a slicecamera that is linked with an InrimageView
*/
class TclInrimage4DBackground : public Inrimage4DBackground, 
  public BasicInrimageBackground,
  public Scalable,
  public TclObject,
  public Owner<MinGLScale>,
  public Owner<MaxGLScale>
{
  protected:
    /** @name constructor */
    //@{
    /** a image background that is created by an inherited class
	@param m module to create TCL object
	@param type the Tcl type of the object
	@param inrv an inrimage view attached to an Inrimage
	@param cam slice camera to display image in 	
	@param minScale a TclScale controlling the minimum windowing value
	@param maxScale a TclScale controlling the maximum windowing value */
    TclInrimage4DBackground(TclModule *m,
			    const std::type_info& type,
			    Inrimage4DView *inrv, TimeSliceCamera *cam,
			    MinGLScale *minScale = 0,
			    MaxGLScale *maxScale = 0);
 public:
    /** builds a Tcl object corresponding to  photo data block attached to 
	an inrimage in a given slice camera
	@param m module to create TCL object
	@param inrv an inrimage view attached to an Inrimage
	@param cam slice camera to display image in 
	@param useMinGLScale if a  TclScale controlling the minimum grey-level windowing value should be created if minScale==0
	@param minScale a TclScale controlling the minimum windowing value
	@param useMaxGLScale if a  TclScale controlling the maximum grey-level windowing value should be created if maxScale==0
	@param maxScale a TclScale controlling the maximum windowing value 
	@param color color of the widget */
    TclInrimage4DBackground(TclModule *m,
			  Inrimage4DView *inrv, TimeSliceCamera *cam,
			  bool useMinGLScale = true,
			  MinGLScale *minScale = 0,
			  bool useMaxGLScale = true,
			  MaxGLScale *maxScale = 0, 
			  const char *color = "#E8E0D2");
    //@}

    /** @name destructors */
    //@{
    /// destructor: free buffers
    virtual ~TclInrimage4DBackground();
    //@}

 protected:
    
#ifdef TCL_INTERFACE
    /// TK photo image associated handle
    Tk_PhotoHandle handle;
    /// TK photo image associated data slice
    Tk_PhotoImageBlock *block;
#endif
    /// TK photo image ID
    TclString TKimageID; 
    /// min and max GL values for inrimage
    Voxel *glmin, *glmax;
    /// owner inrimage view object
    Inrimage4DView *inrv;
    /// geometry the TK photo block was allocated for
    TimeSliceCamera::TIME_GEOMETRY blockGeom;
    
    /// image visible area
    unsigned int visible[8];
    /// the 4 scale vector in the X,Y,Z,T direction depending on ignoreVoxelSize
    double scaleVector[4];
     /// the 4 voxel Size in the X,Y,Z,T direction 
    double voxelSize[4];
    /// time image size (in screen pixels)
    unsigned int st;
    /// voxel/pixel time ratio
    double rt;
 public: 
#ifdef TCL_INTERFACE
  /** set TK photo handle. Can be done only once. (setHandle is indirectly 
      called by constructor). */
    void setHandle(Tk_PhotoHandle _handle) { handle = _handle; }
    
    /// returns image TK handle
    Tk_PhotoHandle getHandle(void) const { return handle; }
#endif
 /** Converts a canvas coordinate (i, j) in the corresponding image voxel
      (x, y, z)
      @param i mouse X coordinate
      @param j mouse Y coordinate
      @return a new VoxelCoordinate structure containing x, y, and z */
  virtual VoxelCoordinate4D *coordinates(unsigned int i, unsigned int j) const;
  
  /** Converts a voxel (x, y, z) to its screen coordinate (i, j)
      @param vc voxel image coordinates
      @param i voxel screen X coordinate
      @param j voxel screen Y coordinate */
  virtual void screen(VoxelCoordinate4D *vc, int& i, int& j) const;
  
  /** Converts a canvas coordinate (i, j) to a cartesian point
      @param i mouse X coordinate
      @param j mouse Y coordinate
      @return a new CartesianCoordinate structure */
  virtual CartesianCoordinate4D *cartesianCoordinates(unsigned int i,
						      unsigned int j) const;
  
  /** Converts a cartesian point to its screen coordinate (i, j)
      @param vc cartesian point
      @param i voxel screen X coordinate
      @param j voxel screen Y coordinate */
  virtual void screen(CartesianCoordinate4D *vc, int& i, int& j) const;
  /// returns the 3 orthogonal of the X, Y and Z axis in the world coordinates (from image coordinates) 
  virtual Rotation3D getWorldFrame() const;
 /// returns the offset position of the lower corner of the image 
  virtual Vec3<double> getImageOffset() const;
  /// set visible area to whole image
  virtual void setVisibleArea(void);
    /** return the image minimum intensity value */
  virtual double getMin(const TclScale *) const  {
    return glmin->getValueAsDouble();}
  
  /** return the image minimum intensity value */
  virtual double getMax(const TclScale *) const  {
    return glmax->getValueAsDouble();}
   /// get the array of 255 char corresponding to the red colormap
  virtual  const unsigned char *getRedColorMap() const {
    return redColorMap;
  }
  /// get the array of 255 char corresponding to the green colormap
  virtual  const unsigned char *getGreenColorMap() const {
    return greenColorMap;
  }
  /// get the array of 255 char corresponding to the blue colormap
  virtual  const unsigned char *getBlueColorMap() const {
    return blueColorMap;
  }
  /** set visible area to given bounds (truncate to image size if needed)
      @param x1 visible area X first coordinate
      @param y1 visible area Y first coordinate
      @param z1 visible area Z first coordinate
      @param t1 visible area time first coordinate
      @param x2 visible area X second coordinate
      @param y2 visible area Y second coordinate
      @param z2 visible area Z second coordinate
      @param t2 visible area time second coordinate */
  virtual void setVisibleArea(unsigned int x1, unsigned int y1, 
			      unsigned int z1, unsigned int t1,
			      unsigned int x2, unsigned int y2, 
			      unsigned int z2, unsigned int t2);
  
  /** set visible area to given bounds (truncate to image size if needed)
      @param x1 visible area X first coordinate
      @param y1 visible area Y first coordinate
      @param z1 visible area Z first coordinate
      @param x2 visible area X second coordinate
      @param y2 visible area Y second coordinate
      @param z2 visible area Z second coordinate */
  virtual void setVisibleArea(unsigned int x1, unsigned int y1, 
			      unsigned int z1, 
			      unsigned int x2, unsigned int y2, 
			      unsigned int z2);
  
  /** get object string identifier
      @return object name */
  virtual TclString objectName(void) const {
    return name();}
 
  /** scale value changed
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleValue(const TclScale *scale, double val);


  /// resize viewer to visible region and redraw
  virtual void visibleAreaChanged(void);
  
  /// resize image voxels
  void imageVoxelSizeChanged(void);
  
  /// resize GL window and redraw
  virtual void imageGLChanged(void);
  
  /// image type changed
  void imageTypeChanged(void);
  
  /// image type changed
  virtual void inrimageChanged(void);

  /// return handled inrimage
  Inrimage4D *getInrimage(void) const;
  
  /// set min voxel value to val
  void setMinThreshold(double val) throw(Voxel::InvalidValue);
  
  /// set max voxel value to val
  void setMaxThreshold(double val) throw(Voxel::InvalidValue);
  
  /// set min and max voxel value to min and max
  void setGlThresholds(double min, double max) throw(Voxel::InvalidValue);
  
  /// set min and max voxel value to min and max
  void setGlThresholds(const TclString& min, const TclString& max)
    throw(Voxel::InvalidValue);
  
  /// min grey level scale owner callback
  virtual void dying(const ObjectPointer<MinGLScale>& obj) {}

  /// max grey level scale owner callback
  virtual void dying(const ObjectPointer<MaxGLScale>& obj) {}
  
  
  /// add a minimum grey-level scale to the background
  void addMinGLScale(MinGLScale *min);
  /// add a maximum grey-level scale to the background
  void addMaxGLScale(MaxGLScale *max);
  
  /// return the min scale associated with the camera background
  MinGLScale *getMinScale() const;
  /// return the max scale associated with the camera background
  MaxGLScale *getMaxScale() const;
  /// return visible area
  virtual const unsigned int *getVisibleArea(void) const { return visible; }
  /** return the image minimum intensity value */
  virtual double getMinIntensity() const  {
    return glmin->getValueAsDouble();}
  /** return the image maximum intensity value */
  virtual double getMaxIntensity() const  {
    return glmax->getValueAsDouble();} 
  /// return scale in the 4 axis between image and world coordinates
  const double  *getScaleVector() { 
    if (ignoreVoxelSize)
      {
	scaleVector[0]=1;
	scaleVector[1]=1;
	scaleVector[2]=1;
	scaleVector[3]=1;
      }
    else
      {
	scaleVector[0]=getInrimage()->getVX();
	scaleVector[1]=getInrimage()->getVY();
	scaleVector[2]=getInrimage()->getVZ();
	scaleVector[3]=getInrimage()->getVT();
      }
    return scaleVector; }
  /// return voxel size
  virtual const double  *getVoxelSize() { 
    return voxelSize; }
  /** redraw camera image */
  virtual void redraw();
  
  /// reset pointer to camera
  void refreshCamera(void);
  
  /// returns grey-level transformation mode
  GLTRANSFO getGLTransfo(void) const { return transfo; }
  /** set grey-level transformation mode
      @param t new transformation mode
      @param red, green, blue
  */
  void setGLTransfo(GLTRANSFO t,const unsigned char *red=0,
		    const  unsigned char *green=0,const  unsigned char *blue=0);
  /// return the scale in the x direction
  virtual float  getVX() const{
    return (getInrimage()->getVX());}
  /// return the scale in the y direction
  virtual float  getVY() const {
    return (getInrimage()->getVY());}
  /// return the scale in the z direction
  virtual float  getVZ() const {
    return (getInrimage()->getVZ());}
  /// return the number of slices in the X direction 
  virtual unsigned int getX() const {
    return (getInrimage()->getX());}
  /// return the number of slices in the Y direction 
  virtual unsigned int getY() const {
    return (getInrimage()->getY());}
  /// return the number of slices in the Z direction 
  virtual unsigned int getZ() const {
    return (getInrimage()->getZ());}
  /// return the label coresponding to pixel (x,y)
  virtual TclString getLabel(int &x,int &y);
 /// return the position in world coordinates of a pixel located at (x,y)
  virtual void getWorldCoordinates(const int x,const int y, 
				   double &px, double &py, 
				   double &pz ) const;  
  /// return the position in image coordinates of a pixel located at (x,y)
  virtual void getImageCoordinates(const int x,const int y, 
				   unsigned int &vx, unsigned int &vy, 
				   unsigned int &vz ) const ;  
  /// return the position in image coordinates of a pixel located at (x,y)
  virtual void getImageCoordinates(const int x,const int y, 
				   unsigned int &vx, unsigned int &vy, 
				   unsigned int &vz, unsigned int &vt ) const ;  
   /** Given a point in image (voxel) coordinates returns the position in 
      screen coordinates */
  virtual void getPixelPositionFromImageCoordinates(const unsigned int vx,
						    const unsigned int vy,
						    const unsigned int vz,
						    int &x, int&y) const;
   /** Given a point in image (voxel) coordinates returns the position in 
      screen coordinates */
  virtual void getPixelPositionFromImageCoordinates(const unsigned int vx,
						    const unsigned int vy,
						    const unsigned int vz,
						    const unsigned int vt,
						    int &x, int&y) const;
  /** Given a point in world coordinates returns the position in 
      screen coordinates */
  virtual void getPixelPositionFromWorldCoordinates(const double px,
						    const double py,
						    const double pz,
						    int &x, int&y) const;
  /// returns IgnoreVoxelSize flag
  bool getIVS(void) const { return ignoreVoxelSize; }
  /// Changes IgnoreVoxelSize flag. Cause recompute method to be called.
  void setIVS(const bool val);
  /// get the time slice camera associated with this background
  TimeSliceCamera *getCamera() const {
    return ( TimeSliceCamera *) slicecam;
  }



  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -scalemode option
  int scalemodeCmd(int argc, char **argv);
  /// -ignoreVoxelSize option
  int ignoreVoxelSizeCmd(int argc, char **argv);
  /// -greyLevel option
  int greyLevelCmd(int argc, char **argv);
  /// -interpolation greyLevel option
  int interpolationCmd(int argc, char **argv);

  protected:
  /** create all Tcl widgets associated with this background
      @param useMinScale if we should handle a min grey-level scale
      @param minScale the min scale that should be binded with the background
      @param useMaxScale if we should handle a max grey-level scale
      @param maxScale the max scale that should be binded with the background
      @param color the color of the background when creating scale widgets */
  void init(bool useMinScale,MinGLScale *minScale,
	    bool useMaxScale,MaxGLScale *maxScale, const char *color);
  
  /// compute inrimage grey levels bounds
  virtual void glBounds(void);
  /// compute image voxel/screen pixel ratio and allocates required memory
  virtual void voxelRatio(void);
  /// allocates TK photo image block
  virtual void realloc(void);
  

  friend class Inrimage4DView;
};


} // namespace yav


#endif
