#ifndef TCLINRIMAGEBACKGROUND_H
#define TCLINRIMAGEBACKGROUND_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclInrimageBackground.h"
#include <tk.h>
#include <inrimage/InrimageView.h>
#include <inrimage/GLScale.h>


namespace yav {

/** \class VoxelCoordinate
    \ingroup Inrimage
    \brief 3D image coordinates */
class VoxelCoordinate {
 public:
  /// initialise coordinates to (0, 0, 0)
  VoxelCoordinate() : x(0), y(0), z(0) {}
  /// x coordinate (column)
  unsigned int x;
  /// y coordinate (row)
  unsigned int y;
  /// z coordinate (plane)
  unsigned int z;
};

/** \class CartesianCoordinate
    \ingroup Inrimage
    \brief 3D cartesian point */
class CartesianCoordinate {
 public:
  /// initialise coordinates to (0, 0, 0)
  CartesianCoordinate() : x(0.0), y(0.0), z(0.0) {}
  /// x coordinate
  double x;
  /// y coordinate
  double y;
  /// z coordinate
  double z;
};


/** \class BasicInrimageBackground
    \ingroup Inrimage
    \brief a basic background that define common data between TclInrimageBackground and TclInrimage4DBackground */
class BasicInrimageBackground 
  {
  public:
    /// grey level transformations
    enum GLTRANSFO {
      AFFINE,
      THRESHOLD,
      PET_COLORS,
      JET_COLORS,
      SPECT_COLORS,
      HOTIRON_COLORS,
      BLUE,
      USER
    };
    /// grey level interpolation
    enum IMAGE_INTERPOLATION {
      NEAREST,
      BILINEAR
    };
 public:
  /** @name constructor */
  //@{
  BasicInrimageBackground();
  //@}

  /** @name desstructor */
  /// destructor: free buffers
  //@{
  virtual ~BasicInrimageBackground(){}
  //@}

 public:
  /// HotIron colormap red component 
  static unsigned char HotIronColorMapRed[257];
  /// HotIron colormap green component 
  static unsigned char HotIronColorMapGreen[257];
  /// HotIron colormap  blue component 
  static unsigned char HotIronColorMapBlue[257];
  /// SPECT colormap red component 
  static unsigned char SPECTColorMapRed[257];
  /// SPECT colormap green component 
  static unsigned char SPECTColorMapGreen[257];
  /// SPECT colormap  blue component 
  static unsigned char SPECTColorMapBlue[257];
  /// PET colormap red component 
  static unsigned char PETColorMapRed[257];
  /// PET colormap green component 
  static unsigned char PETColorMapGreen[257];
  /// PET colormap  blue component 
  static unsigned char PETColorMapBlue[257];
  /// JET colormap red component 
  static unsigned char JetColorMapRed[257];
  /// JET colormap green component 
  static unsigned char JetColorMapGreen[257];
  /// JET colormap  blue component 
  static unsigned char JetColorMapBlue[257];    
  /// AFFINE colormap 
  static unsigned char AffineColorMap[257];
  /// THRESHOLD colormap 
  static unsigned char ThresholdColorMap[257];
  /// zero colormap 
  static unsigned char ZeroColorMap[257];
 protected:
  /// grey level transformation type
  GLTRANSFO transfo;
  /// describes how to interpolate grey values
  IMAGE_INTERPOLATION interpolation;
  /// whether to display real voxel size
  bool ignoreVoxelSize;
  /// the red component colormap that depends on transfo
  const unsigned char *redColorMap;
  /// the green component colormap that depends on transfo
  const unsigned char *greenColorMap;
  /// the blue component colormap that depends on transfo
  const unsigned char *blueColorMap;

  /// X image size (in screen pixels)
  unsigned int sx;
  /// Y image size (in screen pixels)
  unsigned int sy;
  /// Z image size (in screen pixels)
  unsigned int sz;
  /// voxel/pixel X ratio
  double rx;
  /// voxel/pixel Y ratio
  double ry;
  /// voxel/pixel Z ratio
  double rz;
  
 public:
  /// returns voxel/pixel X ratio
  double getRX(void) const { return rx; }
  /// returns voxel/pixel Y ratio
  double getRY(void) const { return ry; }
  /// returns voxel/pixel Z ratio
  double getRZ(void) const { return rz; }

  
}; 

/** \class TclInrimageBackground
    \ingroup Inrimage
    \brief The background of a slicecamera that is linked with an InrimageView
*/
class TclInrimageBackground : public InrimageBackground, 
  public BasicInrimageBackground,
  public Scalable,
  public TclObject,
  public Owner<MinGLScale>,
  public Owner<MaxGLScale>
{
 public:
  protected:
    /** @name constructor */
    //@{
    /** a image background that is created by an inherited class
	@param m module to create TCL object
	@param type the Tcl type of the object
	@param inrv an inrimage view attached to an Inrimage
	@param cam slice camera to display image in 	
	@param minScale a TclScale controlling the minimum windowing value
	@param maxScale a TclScale controlling the maximum windowing value */
    TclInrimageBackground(TclModule *m,
			  const std::type_info& type,
			  InrimageView *inrv, SliceCamera *cam,
			  MinGLScale *minScale = 0,
			  MaxGLScale *maxScale = 0);
  public:
    /** builds a Tcl object corresponding to  photo data block attached to 
	an inrimage in a given slice camera
	@param m module to create TCL object
	@param inrv an inrimage view attached to an Inrimage
	@param cam slice camera to display image in 
	@param useMinGLScale if a  TclScale controlling the minimum grey-level windowing value should be created if minScale==0
	@param minScale a TclScale controlling the minimum windowing value
	@param useMaxGLScale if a  TclScale controlling the maximum grey-level windowing value should be created if maxScale==0
	@param maxScale a TclScale controlling the maximum windowing value 
	@param color color of the widget */
    TclInrimageBackground(TclModule *m,
			  InrimageView *inrv, SliceCamera *cam,
			  bool useMinGLScale = true,
			  MinGLScale *minScale = 0,
			  bool useMaxGLScale = true,
			  MaxGLScale *maxScale = 0, 
			  const char *color = "#E8E0D2");
    //@}

    /** @name destructors */
    //@{
    /// destructor: free buffers
    virtual ~TclInrimageBackground();
    //@}

 protected:
    
    
#ifdef TCL_INTERFACE
    /// TK photo image associated handle
    Tk_PhotoHandle handle;
    /// TK photo image associated data slice
    Tk_PhotoImageBlock *block;
#endif
    /// TK photo image ID
    TclString TKimageID; 
    /// min and max GL values for inrimage
    Voxel *glmin, *glmax;
    /// owner inrimage view object
    InrimageView *inrv;
    /// geometry the TK phot block was allocated for
    SliceCamera::SLICE_ORIENTATION blockGeom;
    
    /// image visible area
    unsigned int visible[6];
    /// the 3 scale vector in the X,Y,Z direction depending on ignoreVoxelSize
    double scaleVector[3];
    /// the 3 voxel Size in the X,Y,Z direction 
    double voxelSize[3];
    
 protected:  
    /** create all Tcl widgets associated with this background
	@param useMinScale if we should handle a min grey-level scale
	@param minScale the min scale that should be binded with the background
	@param useMaxScale if we should handle a max grey-level scale
	@param maxScale the max scale that should be binded with the background
      @param color the color of the background when creating scale widgets */
  void init(bool useMinScale,MinGLScale *minScale,
	     bool useMaxScale,MaxGLScale *maxScale, const char *color); 
  
  /// compute inrimage grey levels bounds
  virtual void glBounds(void);
  /// compute image voxel/screen pixel ratio and allocates required memory
  virtual void voxelRatio(void);
  /// allocates TK photo image block
  virtual void realloc(void);
 public:  
#ifdef TCL_INTERFACE
  /** set TK photo handle. Can be done only once. (setHandle is indirectly 
      called by constructor). */
  void setHandle(Tk_PhotoHandle _handle) { handle = _handle; }
  
  /// returns image TK handle
  Tk_PhotoHandle getHandle(void) const { return handle; }
#endif
  /// get the array of 255 char corresponding to the red colormap
  virtual  const unsigned char *getRedColorMap() const {
    return redColorMap;
  }
  /// get the array of 255 char corresponding to the green colormap
  virtual  const unsigned char *getGreenColorMap() const {
    return greenColorMap;
  }
  /// get the array of 255 char corresponding to the blue colormap
  virtual  const unsigned char *getBlueColorMap() const {
    return blueColorMap;
  }
  /** Converts a canvas coordinate (i, j) in the corresponding image voxel
      (x, y, z)
      @param i mouse X coordinate
      @param j mouse Y coordinate
      @return a new VoxelCoordinate structure containing x, y, and z */
  virtual VoxelCoordinate *coordinates(unsigned int i, unsigned int j) const;
  
  /** Converts a voxel (x, y, z) to its screen coordinate (i, j)
      @param vc voxel image coordinates
      @param i voxel screen X coordinate
      @param j voxel screen Y coordinate */
  virtual void screen(VoxelCoordinate *vc, int& i, int& j) const;
  
  /** Converts a canvas coordinate (i, j) to a cartesian point
      @param i mouse X coordinate
      @param j mouse Y coordinate
      @return a new CartesianCoordinate structure */
  virtual CartesianCoordinate *cartesianCoordinates(unsigned int i,
						    unsigned int j) const;
  
  /** Converts a cartesian point to its screen coordinate (i, j)
      @param vc cartesian point
      @param i voxel screen X coordinate
      @param j voxel screen Y coordinate */
  virtual void screen(CartesianCoordinate *vc, int& i, int& j) const;
  /// returns the 3 orthogonal of the X, Y and Z axis in the world coordinates (from image coordinates)
  virtual Rotation3D getWorldFrame() const;
  /// returns the offset position of the lower corner of the image 
  virtual Vec3<double> getImageOffset() const;
  
  /// set visible area to whole image
  virtual void setVisibleArea(void);
    /** return the image minimum intensity value */
  virtual double getMin(const TclScale *) const  {
    return glmin->getValueAsDouble();}
  
  /** return the image minimum intensity value */
  virtual double getMax(const TclScale *) const  {
    return glmax->getValueAsDouble();}
  /** get required scale step for this scalable
      @param scale calling scale object
      @return step value */
  virtual double getStep(const TclScale *scale) const;
  /** return the image minimum intensity value */
  virtual double getMinIntensity() const  {
    return glmin->getValueAsDouble();}
  /** return the image maximum intensity value */
  virtual double getMaxIntensity() const  {
    return glmax->getValueAsDouble();} 

  /// add a minimum grey-level scale to the background
  void addMinGLScale(MinGLScale *min);
  /// add a maximum grey-level scale to the background
  void addMaxGLScale(MaxGLScale *max);
  
  /// return the min scale associated with the camera background
  MinGLScale *getMinScale() const;
  /// return the max scale associated with the camera background
  MaxGLScale *getMaxScale() const;
  /** set visible area to given bounds (truncate to image size if needed)
      @param x1 visible area X first coordinate
      @param y1 visible area Y first coordinate
      @param z1 visible area Z first coordinate
      @param x2 visible area X second coordinate
      @param y2 visible area Y second coordinate
      @param z2 visible area Z second coordinate */
  virtual void setVisibleArea(unsigned int x1, unsigned int y1, unsigned int z1,
			      unsigned int x2, unsigned int y2, unsigned int z2);
  
  /// raise ambigous getRef() function call
  unsigned int getRef(void) const {
    return TclObject::getRef();
  }
  /// returns object id by command line arguments
  static const char *getTCLObjectName(void) {
    return TclObjectKind::getObjectName(typeid(TclInrimageBackground));
  }

  /** get object string identifier
      @return object name */
  virtual TclString objectName(void) const {
    return name();}
 
  /** scale value changed
      @param scale calling scale object
      @param val new scale value */
  virtual void scaleValue(const TclScale *scale, double val);


  /// resize viewer to visible region and redraw
  virtual void visibleAreaChanged(void);
  
  /// resize image voxels
  void imageVoxelSizeChanged(void);
  
  /// resize GL window and redraw
  virtual void imageGLChanged(void);
  
  /// image type changed
  void imageTypeChanged(void);
  
  /// image type changed
  virtual void inrimageChanged(void);

  /// return handled inrimage
  Inrimage *getInrimage(void) const;
  
  /// set min voxel value to val
  void setMinThreshold(double val) throw(Voxel::InvalidValue);
  
  /// set max voxel value to val
  void setMaxThreshold(double val) throw(Voxel::InvalidValue);
  
  /// set min and max voxel value to min and max
  void setGlThresholds(double min, double max) throw(Voxel::InvalidValue);
  
  /// set min and max voxel value to min and max
  void setGlThresholds(const TclString& min, const TclString& max)
    throw(Voxel::InvalidValue);
  
  /// min grey level scale owner callback
  virtual void dying(const ObjectPointer<MinGLScale>& obj) {}

  /// max grey level scale owner callback
  virtual void dying(const ObjectPointer<MaxGLScale>& obj) {}
  
  
  /// return visible area
  virtual const unsigned int *getVisibleArea(void) const { return visible; }

  /// return scale in the 3 axis between image and world coordinates
  virtual const double  *getScaleVector() { 
    if (ignoreVoxelSize)
      {
	scaleVector[0]=1;
	scaleVector[1]=1;
	scaleVector[2]=1;
      }
    else
      {
	scaleVector[0]=getInrimage()->getVX();
	scaleVector[1]=getInrimage()->getVY();
	scaleVector[2]=getInrimage()->getVZ();
      }
    return scaleVector; }
  /// return voxel size
  virtual const double  *getVoxelSize() { 
    return voxelSize; }
  
  /** redraw camera image */
  virtual void redraw();
  
  /// reset pointer to camera
  virtual void refreshCamera(void);
  
  /// returns grey-level transformation mode
  GLTRANSFO getGLTransfo(void) const { return transfo; }
  /** set grey-level transformation mode
      @param t new transformation mode
      @param red, green, blue
  */
  void setGLTransfo(GLTRANSFO t,const unsigned char *red=0,
		    const  unsigned char *green=0,const  unsigned char *blue=0);
  /// return the scale in the x direction
  virtual float  getVX() const{
    return (getInrimage()->getVX());}
  /// return the scale in the y direction
  virtual float  getVY() const {
    return (getInrimage()->getVY());}
  /// return the scale in the z direction
  virtual float  getVZ() const {
    return (getInrimage()->getVZ());}
  /// return the number of slices in the X direction 
  virtual unsigned int getX() const {
    return (getInrimage()->getX());}
  /// return the number of slices in the Y direction 
  virtual unsigned int getY() const {
    return (getInrimage()->getY());}
  /// return the number of slices in the Z direction 
  virtual unsigned int getZ() const {
    return (getInrimage()->getZ());}
  /// return the label coresponding to pixel (x,y)
  virtual TclString getLabel(int &x,int &y);
 /// return the position in world coordinates of a pixel located at (x,y)
  virtual void getWorldCoordinates(const int x,const int y, 
				   double &px, double &py, 
				   double &pz ) const;  
  /// return the position in image coordinates of a pixel located at (x,y)
  virtual void getImageCoordinates(const int x,const int y, 
				   unsigned int &vx, unsigned int &vy, 
				   unsigned int &vz ) const ;  
   /** Given a point in image (voxel) coordinates returns the position in 
      screen coordinates */
  virtual void getPixelPositionFromImageCoordinates(const unsigned int vx,
						    const unsigned int vy,
						    const unsigned int vz,
						    int &x, int&y) const;
  /** Given a point in world coordinates returns the position in 
      screen coordinates */
  virtual void getPixelPositionFromWorldCoordinates(const double px,
						    const double py,
						    const double pz,
						    int &x, int&y) const;
  /// returns IgnoreVoxelSize flag
  bool getIVS(void) const { return ignoreVoxelSize; }
  /// Changes IgnoreVoxelSize flag. Cause recompute method to be called.
  void setIVS(const bool val);




  /// new TCL command callback
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// new TCL command callback
  static int create(TclModule *m, int argc, char **argv);
  /// -destroy option
  int destroyCmd(int argc, char **argv);
  /// -scalemode option
  int scalemodeCmd(int argc, char **argv);
  /// -ignoreVoxelSize option
  int ignoreVoxelSizeCmd(int argc, char **argv);
   /// -greyLevel option
  int greyLevelCmd(int argc, char **argv);
  /// -interpolation greyLevel option
  int interpolationCmd(int argc, char **argv);
  
  friend class InrimageView;
};


} // namespace yav


#endif
