#ifndef TCLROIEDIT_H
#define TCLROIEDIT_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/TclROIEdit.h"

#include <inrimage/TclInrimageROI.h>

namespace yav {

class TclInrimage;
  
/** \class TclROIEdit
    \ingroup Inrimage
    \brief InrimageROI  interface with edition ability */
class TclROIEdit : public TclInrimageROI, public ListElement<TclROIEdit> {
 public:
  /** @name constructors */
  //@{
  /** builds object client from an inrimage then this interface
      @param scene scene to build client in 
      @param image displayed inrimage
      @param s
      @param m
      @param param
  */
  TclROIEdit(TclScene2D *scene, Inrimage *image, 
	     SurfaceSliceable *s,
	     InrimageROI::SurfaceRoiMethod m=InrimageROI::SURFACE_BOUNDARY,  
	     double param=1.0);
  /** builds a region of interest from a grey-level image and a binary image
      @param scene the 2D scene where the object is located 
      @param image a 3D grey-value image where the ROI is defined : the grey-level value of each voxel is used from this original image as well as the image size
      @param binary a 3D binary image where each voxel of non-zero value is considered to belong to the region of interest. It must have the same size as the original image */
  TclROIEdit(TclScene2D *scene,     
	     Inrimage *image, 
	     const Inrimage *binary) throw(InrimageROI::ImageMissing,
					   InrimageROI::BinaryImageWrongSize);
  /** builds an empty region of interest from a grey-level image 
      @param scene the 2D scene where the object is located 
      @param image a 3D grey-value image where the ROI is defined : the grey-level value of each voxel is used from this original image as well as the image size */

  TclROIEdit(TclScene2D *scene,     
	     Inrimage *image) throw(InrimageROI::ImageMissing);

  
  /** builds a region of interest from a file
      @param scene the 2D scene where the object is located 
      @param filename file name */
   TclROIEdit(TclScene2D *scene,     
	      char *filename);
   //@}
   
   /// constructors common code
   void init(const char *install, const char *uninstall);
 protected:
   /// initialize local variable
   void init(); 
 public:
  /** @name destructors */
  //@{
  /// destroys client and interface
   virtual ~TclROIEdit();
  //@}

   /// returns client
   InrimageROI *getClient() const {
     return((InrimageROI *)client);
   }

  /// raise ambigous list() function call
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// create option commands for this object
  static void createCommands(TclModule *m);

  /// TCL object creation function
  static int create(TclModule *m, int argc, char **argv);
  /** load function
      @param filename Inrimage ROI file name
      @param scn 2D scene to store the ROI */
  static void load(const TclString &filename, Scene2D *scn);

  /// -selectionMode TCL command option
   int selectionModeCmd(int argc, char **);
  /// -brusgWidth TCL command option
   int brushWidthCmd(int argc, char **);
  /// -eraserWidth TCL command option
   int eraserWidthCmd(int argc, char **);
};

/** \class TclROIList
    \ingroup Inrimage
    \brief an object that associates several TclROIEdit with an image */
class TclROIList : public TclObject, public Owner<Inrimage> {
protected:
  /// the list of region of interest associated with an image
  ObjectsList<TclROIEdit> roiList; 
  /// filename 
  TclString filename;
  // constructor common code
  void init();
  // scene 2D where the roi will be put
  TclScene2D *scene;
  /// token for timer
  Tcl_TimerToken token;
public:
  /** @name Exception classes */
  //@{
  /** \class ImageMissing
      \ingroup Inrimage
      \brief image missing exceptions */
  class ImageMissing : public Exception {
  public:
    ImageMissing(const std::string& from) : Exception(from,"creating an TclROIList with a null image pointer") {}
  };
  //@}
  /** @name constructor */
  //@{

  /** creates an empty list of ROI
      @param inr the grey-level image
      @param scene a 2D scene */
  TclROIList(TclInrimage *inr, TclScene2D *scene) throw(ImageMissing);
  /** creates a list of ROI from a file
      @param scene a 2D scene
      @param name */
  TclROIList(TclScene2D *scene,TclString name) throw(ImageMissing);
  //@}

  /** @name destructor */
  //@{
  /// empty
  virtual ~TclROIList();
  //@}
 public:
  /// callback for timer for saving data
  static void timerHandler(ClientData);
  /// return this object file name
  TclString getFilename(void) const { return filename; }
  
   /// raise ambigous list() function call
  static int list(TclModule *m, int argc, char **argv) {
    return TclObject::list(m, argc, argv);
  }
  /// create option commands for this object
  static void createCommands(TclModule *m);

  /// returns first ROI
  ObjectsList<TclROIEdit>::iterator firstROI() {
    return roiList.begin();
  }
  /// returns last ROI
  ObjectsList<TclROIEdit>::iterator lastROI() {
    return roiList.end();
  }

  /// called when image is destroyed
  virtual void dying(const ObjectPointer<Inrimage>& obj);
  /// TCL object creation function
  static int create(TclModule *m, int argc, char **argv);
  /// write the content of the list of rois
  void write(TclString fname);

  /// -newROI TCL command
  int newROICmd(int argc, char **argv);
  /// -scene TCL command
  int sceneCmd(int argc, char **argv);
  /// -rois TCL command
  int roisCmd(int argc, char **argv);
  /// -write TCL command
  int writeCmd(int argc, char **argv);
  /// -destroy TCL command
  int destroyCmd(int argc, char **argv);

};
  
} // namespace yav

#endif
