#ifndef VOXEL_H
#define VOXEL_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/Voxel.h"
#include <basics/Exception.h>
#include <basics/TclString.h>
#include <basics/Constants.h>
#include <inrimage/Inrimage.h>

namespace yav {


/** \class Voxel
    \ingroup Inrimage
    \brief a generic class for any image type voxel */
class Voxel {
 public:
  /** @name Exception */
  //@{
  /// scanning invalid voxel string representation
  class InvalidValue : public Exception {
  public:
    InvalidValue(const std::string& from);
  };
  //@}

 public:
  /// empy
  Voxel() {}
  /// empty
  virtual ~Voxel() {}

  /** assign this voxel from another one
      @param v voxel value to get */
  virtual Voxel& operator=(const Voxel& v) = 0;

  /// returns this voxel value converted to a double
  virtual double getValueAsDouble(void) const = 0;
  /** assign this voxel value from a double
      @param d new voxel value */
  virtual void setValueFromDouble(double d) = 0;
  /// returns this voxel value converted to an integer
  virtual LONGINT getValueAsLong(void) const = 0;
  /** -----
   */
  virtual Inrimage::WORD_TYPE getType(void) = 0;

  /** converts this voxel value to a string
      @param value output string */
  virtual void print(TclString& value) const = 0;
  /** reads this voxel value from a string
      @param value input string
      @exception InvalidValue invalid input string format */
  virtual void scan(const TclString& value) throw(InvalidValue) = 0;

  /// set voxel value to min possible value
  virtual void setMin(void) = 0;
  /// set voxel value to max possible value
  virtual void setMax(void) = 0;

  /// returns voxel max possible value
  virtual double maxValue(void) = 0;
  /// returns voxel min possible value
  virtual double minValue(void) = 0;
  /// check if voxel is null or not
  virtual bool isNull(void) =0;

  /** creates a voxel whose dynamic type is given
      Value is set to 0
      @param type dynamic voxel type
      @param vdim for vectorial types only: 
       creates a scalar voxel type if vdim == 1 
       and a vectorial voxel type otherwise 
  */
  static Voxel *voxelByType(Inrimage::WORD_TYPE type, unsigned int vdim = 1);
};


/** \class Voxel8u
    \ingroup Inrimage
    \brief unsigned byte voxel */
class Voxel8u : public Voxel {
 private:
  unsigned char val;

 public:
  Voxel8u(unsigned char v) : Voxel(), val(v) {}
  Voxel8u(const Voxel8u& v) : Voxel() { val = v.val; }
  ~Voxel8u() {}

  Voxel& operator=(const Voxel& v);
  Voxel8u& operator=(const Voxel8u& v) { val = v.val; return *this; }

  unsigned char& value(void) { return val; }
  unsigned char getValue(void) const { return val; }
  void setValue(unsigned char v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_UNSIGNED_CHAR; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_UNSIGNED_CHAR; }
  void setMax(void) { val = MAX_UNSIGNED_CHAR; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_UNSIGNED_CHAR; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_UNSIGNED_CHAR; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class Voxel8s
    \ingroup Inrimage
    \brief signed byte voxel */
class Voxel8s : public Voxel {
 private:
  char val;

 public:
  Voxel8s(char v) : Voxel(), val(v) {}
  Voxel8s(const Voxel8s& v) : Voxel() { val = v.val; }
  ~Voxel8s() {}

  Voxel& operator=(const Voxel& v);
  Voxel8s& operator=(const Voxel8s& v) { val = v.val; return *this; }

  char& value(void) { return val; }
  char getValue(void) const { return val; }
  void setValue(char v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_SIGNED_CHAR; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_SIGNED_CHAR; }
  void setMax(void) { val = MAX_SIGNED_CHAR; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_SIGNED_CHAR; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_SIGNED_CHAR; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class Voxel16s
    \ingroup Inrimage
    \brief signed short integer voxel */
class Voxel16s : public Voxel {
 private:
  short int val;

 public:
  Voxel16s(short int v) : Voxel(), val(v) {}
  Voxel16s(const Voxel16s& v) : Voxel(){ val = v.val; }
  ~Voxel16s() {}

  Voxel& operator=(const Voxel& v);
  Voxel16s& operator=(const Voxel16s& v) { val = v.val; return *this; }

  short int& value(void) { return val; }
  short int getValue(void) const { return val; }
  void setValue(short int v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void)  { return Inrimage::WT_SIGNED_SHORT; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_SIGNED_SHORT; }
  void setMax(void) { val = MAX_SIGNED_SHORT; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_SIGNED_SHORT; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_SIGNED_SHORT; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class Voxel16u
    \ingroup Inrimage
    \brief unsigned short integer voxel */
class Voxel16u : public Voxel {
 private:
  unsigned short int val;

 public:
  Voxel16u(unsigned short int v) : Voxel(), val(v) {}
  Voxel16u(const Voxel16u& v) : Voxel(){ val = v.val; }
  ~Voxel16u() {}

  Voxel& operator=(const Voxel& v);
  Voxel16u& operator=(const Voxel16u& v) { val = v.val; return *this; }

  unsigned short int& value(void) { return val; }
  unsigned short int getValue(void) const { return val; }
  void setValue(unsigned short int v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_UNSIGNED_SHORT; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_UNSIGNED_SHORT; }
  void setMax(void) { val = MAX_UNSIGNED_SHORT; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_UNSIGNED_SHORT; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_UNSIGNED_SHORT; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class Voxel32s
    \ingroup Inrimage
    \brief integer voxel */
class Voxel32s : public Voxel {
 private:
  int val;

 public:
  Voxel32s(int v) : Voxel(), val(v) {}
  Voxel32s(const Voxel32s& v) : Voxel(){ val = v.val; }
  ~Voxel32s() {}

  Voxel& operator=(const Voxel& v);
  Voxel32s& operator=(const Voxel32s& v) { val = v.val; return *this; }

  int& value(void) { return val; }
  int getValue(void) const { return val; }
  void setValue(int v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_SIGNED_INT; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_SIGNED_INT; }
  void setMax(void) { val = MAX_SIGNED_INT; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_SIGNED_INT; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_SIGNED_INT; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class Voxel32u
    \ingroup Inrimage
    \brief unsigned integer voxel */
class Voxel32u : public Voxel {
 private:
  unsigned int val;

 public:
  Voxel32u(unsigned int v) : Voxel(), val(v) {}
  Voxel32u(const Voxel32u& v) : Voxel() { val = v.val; }
  ~Voxel32u() {}

  Voxel& operator=(const Voxel& v);
  Voxel32u& operator=(const Voxel32u& v) { val = v.val; return *this; }

  unsigned int& value(void) { return val; }
  unsigned int getValue(void) const { return val; }
  void setValue(unsigned int v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_UNSIGNED_INT; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_UNSIGNED_INT; }
  void setMax(void) { val = MAX_UNSIGNED_INT; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_UNSIGNED_INT; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_UNSIGNED_INT; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class Voxel64s
    \ingroup Inrimage
    \brief long integer voxel */
class Voxel64s : public Voxel {
 private:
  LONGINT val;

 public:
  Voxel64s(LONGINT v) : Voxel(), val(v) {}
  Voxel64s(const Voxel64s& v) : Voxel(){ val = v.val; }
  ~Voxel64s() {}

  Voxel& operator=(const Voxel& v);
  Voxel64s& operator=(const Voxel64s& v) { val = v.val; return *this; }

  LONGINT& value(void) { return val; }
  LONGINT getValue(void) const { return val; }
  void setValue(LONGINT v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_SIGNED_LONG; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = (LONGINT) MIN_SIGNED_LONG; }
  void setMax(void) { val = (LONGINT) MAX_SIGNED_LONG; }

  /// returns voxel max possible value
  double maxValue(void) { return (double) MAX_SIGNED_LONG; }
  /// returns voxel min possible value
  double minValue(void) { return (double) MIN_SIGNED_LONG; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class Voxel64u
    \ingroup Inrimage
    \brief unsigned long integer voxel */
class Voxel64u : public Voxel {
 private:
  unsigned LONGINT val;

 public:
  Voxel64u(unsigned LONGINT v) : Voxel(), val(v) {}
  Voxel64u(const Voxel64u& v) : Voxel(){ val = v.val; }
  ~Voxel64u() {}

  Voxel& operator=(const Voxel& v);
  Voxel64u& operator=(const Voxel64u& v) { val = v.val; return *this; }

  unsigned LONGINT& value(void) { return val; }
  unsigned LONGINT getValue(void) const { return val; }
  void setValue(unsigned LONGINT v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_UNSIGNED_LONG; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val =(unsigned LONGINT) MIN_UNSIGNED_LONG; }
  void setMax(void) { val = (unsigned LONGINT) MAX_UNSIGNED_LONG; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_UNSIGNED_INT; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_UNSIGNED_INT; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0);}
};

/** \class VoxelFloat
    \ingroup Inrimage
    \brief float voxel */
class VoxelFloat : public Voxel {
 private:
  float val;

 public:
  VoxelFloat(float v) : Voxel(), val(v) {}
  VoxelFloat(const VoxelFloat& v) : Voxel() { val = v.val; }
  ~VoxelFloat() {}

  Voxel& operator=(const Voxel& v);
  VoxelFloat& operator=(const VoxelFloat& v) { val = v.val; return *this; }

  float& value(void) { return val; }
  float getValue(void) const { return val; }
  void setValue(float v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_FLOAT; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_FLOAT; }
  void setMax(void) { val = MAX_FLOAT; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_FLOAT; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_FLOAT; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0.0);}
};

/** \class VoxelDouble
    \ingroup Inrimage
    \brief double voxel */
class VoxelDouble : public Voxel {
 private:
  double val;

 public:
  VoxelDouble(double v) : Voxel(), val(v) {}
  VoxelDouble(const VoxelDouble& v) : Voxel() { val = v.val; }
  ~VoxelDouble() {}

  Voxel& operator=(const Voxel& v);
  VoxelDouble& operator=(const VoxelDouble& v) { val = v.val; return *this; }

  double& value(void) { return val; }
  double getValue(void) const { return val; }
  void setValue(double v) { val = v; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_DOUBLE; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val = MIN_DOUBLE; }
  void setMax(void) { val = MAX_DOUBLE; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_DOUBLE; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_DOUBLE; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val == 0.0);}
};

/** \class VoxelRGB
    \ingroup Inrimage
    \brief RGB (red green blue) voxel */
class VoxelRGB : public Voxel {
 private:
  unsigned char val[3];

 public:
  VoxelRGB(unsigned char r, unsigned char g, unsigned char b)
    : Voxel() {
    val[0] = r; val[1] = g; val[2] = b;
  }
  VoxelRGB(const VoxelRGB& v) : Voxel() {
    val[0] = v.val[0]; val[1] = v.val[1]; val[2] = v.val[2];
  }
  ~VoxelRGB() {}

  Voxel& operator=(const Voxel& v);
  VoxelRGB& operator=(const VoxelRGB& v) {
    val[0] = v.val[0]; val[1] = v.val[1]; val[2] = v.val[2];
    return *this;
  }

  const unsigned char *getValue(void) const { return val; }
  unsigned char *getValue(void) { return val; }
  void setValue(unsigned char r, unsigned char g, unsigned char b)
    { val[0] = r; val[1] = g; val[2] = b; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_RGB; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val[0] = MIN_UNSIGNED_CHAR; val[1] = MIN_UNSIGNED_CHAR;
                   val[2] = MIN_UNSIGNED_CHAR; }
  void setMax(void) { val[0] = MAX_UNSIGNED_CHAR; val[1] = MAX_UNSIGNED_CHAR; 
                   val[2] = MAX_UNSIGNED_CHAR; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_UNSIGNED_CHAR; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_UNSIGNED_CHAR; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val[0] == 0 || val[1] == 0 || val[2] == 0);}
};

/** \class VoxelRGBA
    \ingroup Inrimage
    \brief RGBA (Red-green-blue + alpha) voxel */
class VoxelRGBA : public Voxel {
 private:
  unsigned char val[4];

 public:
  VoxelRGBA(unsigned char r, unsigned char g, unsigned char b,
	    unsigned char a = MAX_UNSIGNED_CHAR)
    : Voxel() {
    val[0] = r; val[1] = g; val[2] = b; val[3] = a;
  }
  VoxelRGBA(const VoxelRGBA& v) : Voxel() {
    val[0] = v.val[0]; val[1] = v.val[1]; val[2] = v.val[2]; val[3] = v.val[3];
  }
  ~VoxelRGBA() {}

  Voxel& operator=(const Voxel& v);
  VoxelRGBA& operator=(const VoxelRGBA& v) {
    val[0] = v.val[0]; val[1] = v.val[1]; val[2] = v.val[2]; val[3] = v.val[3];
    return *this;
  }

  const unsigned char *getValue(void) const { return val; }
  unsigned char *getValue(void) { return val; }
  void setValue(unsigned char r, unsigned char g, unsigned char b,
		unsigned char a = 1.0)
    { val[0] = r; val[1] = g; val[2] = b; val[3] = a; }
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_RGBA; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void) { val[0] = MIN_UNSIGNED_CHAR; val[1] = MIN_UNSIGNED_CHAR;
                   val[2] = MIN_UNSIGNED_CHAR; val[3] = MIN_UNSIGNED_CHAR; }
  void setMax(void) { val[0] = MAX_UNSIGNED_CHAR; val[1] = MAX_UNSIGNED_CHAR;
                   val[2] = MAX_UNSIGNED_CHAR; val[3] = MAX_UNSIGNED_CHAR; }

  /// returns voxel max possible value
  double maxValue(void) { return MAX_UNSIGNED_CHAR; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_UNSIGNED_CHAR; }  
  /// check if voxel is null or not
  bool isNull(void) { return( val[0] == 0 || val[1] == 0 || val[2] == 0 || val[3] == 0);}
};

/** \class VoxelFV
    \ingroup Inrimage
    \brief float vector voxel */
class VoxelFV : public Voxel {
 private:
  float *val;
  unsigned int vdim;

 public:
  VoxelFV(float *_val, unsigned int _vdim);
  VoxelFV(float  _val, unsigned int _vdim);
  VoxelFV(const VoxelFV& v);
  ~VoxelFV();

  Voxel& operator=(const Voxel& v);
  VoxelFV& operator=(const VoxelFV&);

  const float *getValue(void) const { return val; }
  float *getValue(void) { return val; }
  float& operator[](unsigned int i) { return val[i]; }
  float const& operator[](unsigned int i)  const { return val[i]; }
  unsigned int getVdim(void) const { return vdim; }
  void setValue(float *_val, unsigned int _vdim);
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_FLOAT_VECTOR; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void);
  void setMax(void);

  /// returns voxel max possible value
  double maxValue(void) { return MAX_FLOAT; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_FLOAT; }  
  /// check if voxel is null or not
  bool isNull(void) { 
    bool a = 0;
    unsigned int i;
    for (i=0;i<vdim;++i)
      a = a || (val[i] == 0.0);
    return( a);}
};

/** \class VoxelDV
    \ingroup Inrimage
    \brief double vector voxel */
class VoxelDV : public Voxel {
 private:
  double *val;
  unsigned int vdim;

 public:
  VoxelDV(double *_val, unsigned int _vdim);
  VoxelDV(double  _val, unsigned int _vdim);
  VoxelDV(const VoxelDV& v);
  ~VoxelDV();

  Voxel& operator=(const Voxel& v);
  VoxelDV& operator=(const VoxelDV&);

  const double *getValue(void) const { return val; }
  double *getValue(void) { return val; }
  double& operator[](unsigned int i) { return val[i]; }
  double const& operator[](unsigned int i) const { return val[i]; }
  unsigned int getVdim(void) const { return vdim; }
  void setValue(double *_val, unsigned int _vdim);
  double getValueAsDouble(void) const;
  void setValueFromDouble(double d);
  LONGINT getValueAsLong(void) const;
  Inrimage::WORD_TYPE getType(void) { return Inrimage::WT_DOUBLE_VECTOR; }
  
  void print(TclString& value) const;
  void scan(const TclString& value) throw(InvalidValue);

  void setMin(void);
  void setMax(void);

  /// returns voxel max possible value
  double maxValue(void) { return MAX_DOUBLE; }
  /// returns voxel min possible value
  double minValue(void) { return MIN_DOUBLE; }  
  /// check if voxel is null or not
  bool isNull(void) { 
    bool a = 0;
    unsigned int i;
    for (i=0;i<vdim;++i)
      a = a || (val[i] == 0.0);
    return( a);}
};

} // namespace yav

#endif
