#ifndef _VOXELITERATOR_H 
#define _VOXELITERATOR_H 
 
#ifdef _MSC_VER 
#pragma warning ( disable : 4068 4786 4081 ) 
#endif 
 
#pragma interface  "inrimage/VoxelIterator.h"
#include <inrimage/Inrimage.h>

namespace yav { 
 


/** \class VoxelIteratorObject
    \ingroup Inrimage
    VoxelIteratorObject : a virtual class that provides the prototype functions
    for a generic iterator on a 3D image : increment, decrement, acces to voxel
    value */ 
class VoxelIteratorObject
{ 
 public: 
 /**  Exception classes */ 
  //@{ 

  /** \class OutOfRange
      \ingroup Inrimage
      \brief Try to get a voxel which is not in the image  */
  class OutOfRange : public Exception{
    public:
    /// Out of range voxel
    OutOfRange(const std::string& from): Exception(from,
   "Accessing voxel out of image range"){
     }
  };
  /** \class MissingInrimage
      \ingroup Inrimage
      \brief Missing inrimage */
  class MissingInrimage : public Exception{
    public:
    /// missing object VoxelIterator
    MissingInrimage(const std::string& from): Exception(from,
       "Voxel Iterator built with an empty image"){
     }
  };
  /** \class WrongImageType
      \ingroup Inrimage
      \brief Missing inrimage */
  class WrongImageType : public Exception{
    public:
    /// wrong image type
    WrongImageType(const std::string& from): Exception(from,
       "Voxel Iterator built with an image with an unsuited image type"){
     }
  };

  //@} 


  /**  Constructors */
  //@{
  /** copy constructor 
      @param it should be a valid voxel iterator
      @exception out of range exception thrown
  */
  VoxelIteratorObject(const VoxelIteratorObject &it) throw (OutOfRange);

   /** initialize a voxel iterator on an image
       @param inr a valid non-null 3D image pointer 
       @exception missing image  exception thrown when inr=0
  */
  VoxelIteratorObject(Inrimage *inr) throw (MissingInrimage);
  
  //@} 
 
    /**  Destructor */ 
  //@{ 
  /// desallocates an object 
  virtual ~VoxelIteratorObject(){}
  //@} 
 

 public:
  /** goes to the following element ( do not return any value for speed-up) */
  virtual void operator++(int) throw (OutOfRange)=0;

  /** goes to the following element ( do not return any value for speed-up) */
  virtual void operator++(void) throw (OutOfRange)=0;

  /** goes to the previous element ( do not return any value for speed-up) */
  virtual void operator--(int) throw (OutOfRange)=0;

  /** goes to the previous element ( do not return any value for speed-up) */
  virtual void operator--(void) throw (OutOfRange)=0;

  /** returns the array size of the pointed value : 1 if the image is scalar */
  unsigned int getArraySize() const;

  /** returns current value as double */
  virtual double operator*() const=0;

  /** returns current value as long integer */
  virtual LONGINT  operator()(void) const=0;
  /** modifies the current voxel value from a double value (use static cast)*/
  virtual void setValueFromDouble(const double v)=0;

  /** modifies the current voxel value from a long int (use static cast)*/
  virtual void setValueFromLong(const long int v)=0;

  /** get the current (x,y,z) position  of the iterator iterator 
      @param x column index starting from 0 
      @param y row index starting from 0 
      @param z plane index starting from 0 */
  void getPosition(unsigned int &x, unsigned int &y, unsigned int &z);


  /** get the current  position  as the offset from the first voxel */
  long int  getPosition() const {
    return index;
  }
  /** set the current position of the iterator given the offset from the first voxel */
  virtual void setPosition(const long int ind) throw(OutOfRange)=0;

  /** set the current position of the iterator the x,y and z voxel coordinate */
  virtual void setPosition(const unsigned int x,const unsigned int y,
			   const unsigned int z) throw(OutOfRange)=0;
  /// goes to first image voxel 
  virtual void first(void)=0;

  /// goes to last image voxel
  virtual void last(void)=0;

  /** returns true if the iterator is pointing on the last+1 voxel */
  bool isAtEnd(void) const { return (index==maxIndex); }

  /** returns true if the iterator is pointing on the first -1  voxel */
  bool isAtBeginning(void) const { return (index<0); }

  /** assign this voxel from another one
      @param v voxel value to get */
  virtual VoxelIteratorObject& operator=(const VoxelIteratorObject& v) = 0;
  /** returns true if the iterator t points on the same item
      as this iterator  */
  virtual bool operator==(const VoxelIteratorObject &t)const;
  /** returns true if the iterator t points on a different
      item as this iterator  */
  virtual bool operator!=(const VoxelIteratorObject &t)const;

  /// free the object 
  virtual void free() {
    delete this;
  }
  /** returns the pointed inrimage */
  virtual Inrimage *getInrimage() const =0 ;

 protected: 
  /** gives the index of the voxel in the image array (0 for first voxel ) */
  long int index;
  /// indicates the total number of voxels in the image 
  long int maxIndex;
}; 

typedef VoxelIteratorObject& VoxelIterator;


/** \class VoxelIteratorScalar
    \ingroup Inrimage
    \brief VoxelIteratorScalar class : iterator on images

    consider image as a
    scalar ; for a vectorial image, the index corresponds to the position of a
    single scalar voxel (for an RGB image, the maximum index is
    3*sixeX*sizeY*sizeZ  */
template <class T> class VoxelIteratorScalar : public VoxelIteratorObject, public Owner<Inrimage> 
{ 
 
 public: 
 
  /**  Constructors */
  //@{
  /** First public constructor : create from an image
      @param im : a valid pointer on a scalar image
      @exception  exception thrown on null pointer or if image is not scalar */ 
  VoxelIteratorScalar(Inrimage *im)  throw(VoxelIteratorObject::MissingInrimage,
VoxelIteratorObject::WrongImageType);
  /**  public constructor : create from another iterator
      @param it : a valid iterator on a scalar image
      @exception  exception thrown if pointer out of range or if image is not scalar */ 
  VoxelIteratorScalar(const VoxelIteratorObject &it) throw (VoxelIteratorObject::OutOfRange,VoxelIteratorObject::WrongImageType);
  //@} 

  /**  Destructor */ 
  //@{ 
  /// desallocates an object 
  virtual ~VoxelIteratorScalar(){}
  //@} 
 
  /** returns the pointed inrimage */
  virtual Inrimage *getInrimage() const { 
    return((Inrimage *)pointer()); }
  /** goes to the following element ( do not return any value for speed-up) */
  virtual void operator++(int) throw (VoxelIteratorObject::OutOfRange);

  /** goes to the following element ( do not return any value for speed-up) */
  virtual void operator++(void) throw (VoxelIteratorObject::OutOfRange);

  /** goes to the previous element ( do not return any value for speed-up) */
  virtual void operator--(int) throw (VoxelIteratorObject::OutOfRange);

  /** goes to the previous element ( do not return any value for speed-up) */
  virtual void operator--(void) throw (VoxelIteratorObject::OutOfRange);

  /** returns current value as double */
  virtual double operator*() const;
  /** returns current value as long integer */
  virtual LONGINT  operator()(void) const;

  /** modifies the current voxel value from a double value (use static cast)*/
  virtual void setValueFromDouble(const double v);

  /** modifies the current voxel value from a long int (use static cast)*/
  virtual void setValueFromLong(const long int v);

  /// goes to first image voxel 
  virtual void first(void);
  
  /// goes to last image voxel
  virtual void last(void);
  /** assign this voxel from another one
      @param v voxel value to get */
  virtual VoxelIteratorObject& operator=(const VoxelIteratorObject& v);

  /// die on inrimage deletion
  virtual void dying(const ObjectPointer < Inrimage > &) {
    delete this;
  }

  /** set the current position of the iterator given the offset from the first voxel */
  virtual void setPosition(const long int ind) throw(VoxelIteratorObject::OutOfRange);

  /** set the current position of the iterator the x,y and z voxel coordinate */
  virtual void setPosition(const unsigned int x,const unsigned int y,
			   const unsigned int z) throw(VoxelIteratorObject::OutOfRange);

 protected: 
   /// pointer on image buffer
   T *buffer;


}; 

#include <inrimage/VoxelIterator.txx>

} // namespace yav 
 
#endif 
