#ifndef ANISOTROPIC_DIFFUSION_H
#define ANISOTROPIC_DIFFUSION_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/anisotropicDiffusion.h"

#ifdef USE_MTL
#include <mtl/mtl.h>
#include <mtl/matrix.h>
#include <mtl/utils.h>
#endif

#include <inrimage/Inrimage.h>
#include <inrimage/Inrimage4D.h>

namespace yav {

   
  /// diffusitvity functions used in anisotropic diffusion
  enum DIFFUSION_FUNCTION {
    /// functions given by Perona and Malik paper
    PERONA_MALIK, PERONA_MALIK_2, PERONA_MALIK_3,
    /// function given by Aubert
    AUBERT,
    /// function given by Green
    GREEN,
    /// Function given by Weickert
    WEICKERT
  };
  
 ///type of anisotropic diffusion
  enum DIFFUSION_TYPE {
    /// scalar diffusion
    SCALAR,
    /// matrix diffusion
    MATRIX,
    /// matrix diffusion by Karl Krissian for vessels
    MATRIX_KRISSIAN,
    /// flow diffusion by Karl Krissian for vessels
    KRISSIAN
  };

  ///type of numerical scheme
  enum NUMERICAL_SCHEME {
    /// AOS scheme , an efficient semi-implicit scheme, only available with scalar diffusion
    AOS_scheme,
    ///explicit scheme, need a lower dt
    EXPLICIT
  };


  

#ifdef USE_MTL
  typedef mtl::matrix<float, mtl::banded<>, mtl::banded<>, mtl::row_major>::type MTLBanded;
  typedef mtl::dense1D<float> MTLVector;
  typedef mtl::compressed1D<float> MTLVecSparse;
  typedef mtl::dense1D<float>::iterator MTLVecIt;
  typedef mtl::dense1D<float>::const_iterator MTLConstVecIt;
#endif
  

  
#ifdef USE_MTL
  /** anisotropic diffusion in 2D and 3D images. It uses the Additive Operator
      Splitting semi-implicit numerical or explicit scheme and a quantile of
      the cumulated gradient histogram as a threshold to know if a point is an
      edge or not.
      @param inr
      @param output output Inrimage.
      @param dt time step in the numerical scheme.
      @param tmax diffusion time (dt/tmax) gives the number of iterations.
      @param quantx quantile on the cumulated gradient histogram in the x direction. Smaller quant gives sharper edges but the image is less smoothed.
      @param quanty quantile on the cumulated gradient histogram in the y direction. Smaller quant gives sharper edges but the image is less smoothed.
      @param quantz quantile on the cumulated gradient histogram in the z direction. Smaller quant gives sharper edges but the image is less smoothed.
      @param k allow to choose a constant value instead of quant. Smaller k gives sharper edges but the image is less smoothed.
      @param smooth boolean to chose if the gradient is smoothed before the diffusivity is computed or not. (It should increase the stability of the numerical scheme). Must be true except fot scalar diffusion
      @param sigma standard deviation of the eventual smoothing of the gradient.
      @param df diffusion function used.
      @param disk boolean to chose if temporary buffers must be writen on disk to save memory possible only for scalar diffusion
      @param  td diffusion type we want to use
      @param ns numerical scheme we want to apply*/
  void anisotropicDiffusion(const Inrimage *inr, Inrimage *& output, const float & dt = 0.1, const float & tmax = 3.0, const float & quantx = 80.0,const float & quanty = 80.0,const float & quantz = 80.0, float  k =-1.0, const bool smooth = true, const float & sigma = 1.0, const DIFFUSION_FUNCTION & df = PERONA_MALIK, const bool disk = false, const DIFFUSION_TYPE & td = SCALAR,const NUMERICAL_SCHEME & ns =EXPLICIT );


  // 4D images

  void anisotropicDiffusion(const Inrimage4D *inr, Inrimage4D *& output, const float & dt, const float & tmax, const float & q, const bool smooth, const float & sigma, const DIFFUSION_FUNCTION & p, const bool disk);


#endif



#ifdef USE_MTL
  // Functions for Anisotropic Diffusion
  /// AOS Numerical scheme for fast semi-implicit computation
  void AOS(MTLVector & r, MTLVector & g, const float & dt, const float & d, const unsigned int dim); 
  /// explicite scheme
  void explicite(MTLVector & r, MTLVector & g, const float & dt, const unsigned int dim);
  ///Diffusion Matrix in 2D
  void calculD(const MTLVector & du1s, const MTLVector & du2s, MTLVector & d11,  MTLVector & d12, MTLVector &d21, MTLVector &d22,const float & quant, const  DIFFUSION_FUNCTION & p);
  ///Diffusion Matrix in 3D
  void calculD3D(const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s, MTLVector & d11,  MTLVector & d12, MTLVector & d13,  MTLVector &d21, MTLVector &d22, MTLVector &d23, MTLVector &d31, MTLVector &d32,MTLVector &d33,const float & quant, const  DIFFUSION_FUNCTION & p);

  /// Explicit scheme for 2d matrix Krissian algorithm
  void explicitematKrissian(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s,const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION & p);
  /// Explicit scheme for 3d matrix Krissian algorithm
  void explicite3DmatKrissian(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s ,const MTLVector & du11s, const MTLVector & du12s, const MTLVector & du13s , const MTLVector & du22s, const MTLVector & du23s ,const MTLVector & du33s , const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION &p);
  /// Diffusion matrix for K. Krissian diffusion in 2d
  void calculDKarl(const MTLVector & du1s, const MTLVector & du2s, const MTLVector & u, MTLVector & d11,  MTLVector & d12, MTLVector &d21, MTLVector &d22,const float & quant, const  float & p, const unsigned int & xdim);
  /// Diffusion matrix for K. Krissian diffusion in 3d
  void calculDKarl3D(const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s, const MTLVector & du11s, const MTLVector & du12s, const MTLVector & du13s,const MTLVector & du22s, const MTLVector & du23s, const MTLVector & du33s,const MTLVector & u, MTLVector & d11,  MTLVector & d12,MTLVector & d13, MTLVector &d21,MTLVector &d22,MTLVector & d23, MTLVector &d31,MTLVector &d32,MTLVector & d33, const float & quant, const  DIFFUSION_FUNCTION & p, const unsigned int & xdim);
  ///Point to point diffusivity function
  float diffusion(const float & u, const float & seuil, const DIFFUSION_FUNCTION & p);
  /// explicit scheme for 2D matricial diffusion
 void explicitematrice(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s,const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION & p);
 /// explicite scheme for Krissian flow diffusion
 void expliciteKrissian(const MTLVector & u, MTLVector & r,  const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s, const MTLVector & du11s, const MTLVector & du12s, const MTLVector & du13s, const MTLVector & du22s,  const MTLVector & du23s, const MTLVector & du33s, const float & dt, const float k, const DIFFUSION_FUNCTION & p, unsigned int xdim, unsigned int ydim, unsigned int zdim);
 /// explicit scheme for 3D matricial diffusion
 void explicitematrice3D(const MTLVector & u, MTLVector & r , const MTLVector & du1s, const MTLVector & du2s, const MTLVector & du3s , const float & dt,const unsigned int xdim,const unsigned int ydim, const float quant, const DIFFUSION_FUNCTION & p);
  /// Gaussian smoothing to ensure the stability of the process
  void AOSGauss(MTLVector & r, const MTLVector & u, const float & dt, const float & d) ;
 /// Diffusivity functions computation
  void diffusivity(MTLVector & u, const float & seuil, const DIFFUSION_FUNCTION & p);
  /// Derivation by finite differences, absolute value of the result
  void derive(MTLVector & du, const MTLVector & u, const double& h = 1.0);
  /// Computation of a quantile of the cumulated gradient histogram
  float quantile(const MTLVector & u, const float & quant) ;
  /// Write MTL vector to disk
  void writeMTLVector(const TclString name,const MTLVector & m);
  /// Read MTL Vector from disk
  void readMTLVector(const TclString name, MTLVector & m);

#endif

}

#endif
