#ifndef BLOCKPROCESSING_H
#define BLOCKPROCESSING_H

#pragma interface "inrimage/blockProcessing.h"

#include <inrimage/Inrimage.h>




namespace yav {

 /** compute intensity standard deviation of a block surrounding each voxel
 @param inputImage
 @param blockSizeXY the size of the block in pixel size in the XY plane (must be an odd number) 
 @param blockSizeZ the size of the block in pixel size in the Z direction (must be an odd number) 
 @return an image of the same type and same size as the input image */
 Inrimage *computeBlockStandardDeviation(Inrimage *inputImage,unsigned int blockSizeXY,unsigned int blockSizeZ); 

 /** compute intensity mean of a block surrounding each voxel
 @param inputImage
 @param blockSizeXY the size of the block in pixel size in the XY plane (must be an odd number) 
 @param blockSizeZ the size of the block in pixel size in the Z direction (must be an odd number) 
 @return an image of the same type and same size as the input image */
 Inrimage *computeBlockMean(Inrimage *inputImage,unsigned int blockSizeXY,unsigned int blockSizeZ); 
 
 /** compute intensity median of a block surrounding each voxel
 @param inputImage
 @param blockSizeXY the size of the block in pixel size in the XY plane (must be an odd number) 
 @param blockSizeZ the size of the block in pixel size in the Z direction (must be an odd number) 
 @return an image of the same type and same size as the input image */
 Inrimage *computeBlockMedian(Inrimage *inputImage,unsigned int blockSizeXY,unsigned int blockSizeZ); 

  /** compute the sum of the square differences of 2 block on 2 images 
 @param image1 first image
 @param image2 second image
 @param blockSizeX the size of the block in pixel size in the X direction (must be an odd number) 
 @param blockSizeY the size of the block in pixel size in the Y direction (must be an odd number) 
 @param blockSizeZ the size of the block in pixel size in the Z direction (must be an odd number) 
 @param p1x minimum x value of the block in the 1st image
 @param p1y minimum y value of the block in the 1st image
 @param p1z minimum z value of the block in the 1st image
 @param p2x minimum x value of the block in the 2nd image
 @param p2y minimum y value of the block in the 2nd image
 @param p2z minimum z value of the block in the 2nd image
 @return the sum of the square difference divided by the block size */
 double computeBlockSSD(const Inrimage *image1,const Inrimage *image2,
	 const unsigned int blockSizeX,const unsigned int blockSizeY,const unsigned int blockSizeZ,
	 const unsigned int p1x,const unsigned int p1y,const unsigned int p1z,
	 const unsigned int p2x,const unsigned int p2y,const unsigned int p2z);

  /** compute the sum of the absolute differences of 2 block on 2 images 
 @param image1 first image
 @param image2 second image
 @param blockSizeX the size of the block in pixel size in the X direction (must be an odd number) 
 @param blockSizeY the size of the block in pixel size in the Y direction (must be an odd number) 
 @param blockSizeZ the size of the block in pixel size in the Z direction (must be an odd number) 
 @param p1x minimum x value of the block in the 1st image
 @param p1y minimum y value of the block in the 1st image
 @param p1z minimum z value of the block in the 1st image
 @param p2x minimum x value of the block in the 2nd image
 @param p2y minimum y value of the block in the 2nd image
 @param p2z minimum z value of the block in the 2nd image
 @return the sum of the absolute difference divided by the block size */
 double computeBlockSAD(const Inrimage *image1,const Inrimage *image2,
	 const unsigned int blockSizeX,const unsigned int blockSizeY,const unsigned int blockSizeZ,
	 const unsigned int p1x,const unsigned int p1y,const unsigned int p1z,
	 const unsigned int p2x,const unsigned int p2y,const unsigned int p2z);
   /** compute the correlation coefficient of 2 block on 2 images 
 @param image1 first image
 @param image2 second image
 @param blockSizeX the size of the block in pixel size in the X direction (must be an odd number) 
 @param blockSizeY the size of the block in pixel size in the Y direction (must be an odd number) 
 @param blockSizeZ the size of the block in pixel size in the Z direction (must be an odd number) 
 @param p1x minimum x value of the block in the 1st image
 @param p1y minimum y value of the block in the 1st image
 @param p1z minimum z value of the block in the 1st image
 @param p2x minimum x value of the block in the 2nd image
 @param p2y minimum y value of the block in the 2nd image
 @param p2z minimum z value of the block in the 2nd image
 @return the correlation coefficient */
 double computeBlockCorrelation(const Inrimage *image1,const Inrimage *image2,
	 const unsigned int blockSizeX,const unsigned int blockSizeY,const unsigned int blockSizeZ,
	 const unsigned int p1x,const unsigned int p1y,const unsigned int p1z,
	 const unsigned int p2x,const unsigned int p2y,const unsigned int p2z);
 
 /** compute the SSD between a given mask and a block centered on each voxel
 @param image input image
 @param maskImage the mask image that must have the same type and voxel size as the input image. It should also be smaller than the input image
 @return an image of type double and the same size as the input image where each voxel is the SSD with the mask */
 Inrimage *computeSSDWithMask(const Inrimage *image, const Inrimage *maskImage);
/** compute the SAD between a given mask and a block centered on each voxel
 @param image input image
 @param maskImage the mask image that must have the same type and voxel size as the input image. It should also be smaller than the input image
 @return an image of type double and the same size as the input image where each voxel is the SAD with the mask */
 Inrimage *computeSADWithMask(const Inrimage *image, const Inrimage *maskImage);

 /** compute the correlation ratio between a given mask and a block centered on each voxel
 @param image input image
 @param maskImage the mask image that must have the same type and voxel size as the input image. It should also be smaller than the input image
 @return an image of type double and the same size as the input image where each voxel is the correlation ratio with the mask */
 Inrimage *computeCorrelationRatioWithMask(const Inrimage *image, const Inrimage *maskImage);

 bool findBlockCorrespondance(const Inrimage *refImage, const Inrimage *targetImage,
	 const unsigned int blockSizeX,const unsigned int blockSizeY,const unsigned int blockSizeZ,
	 const unsigned int pix, const unsigned int piy, const unsigned int piz,
	 const unsigned int minptx, const unsigned int minpty, const unsigned int minptz,
	 const unsigned int searchSizeX,const unsigned int searchSizeY,const unsigned int searchSizeZ,
	 const unsigned int searchStepX,const unsigned int searchStepY,const unsigned int searchStepZ,
 	 unsigned int &bestptx, unsigned int &bestpty, unsigned int &bestptz,
	 double variance[6]);
}


#endif
