#ifndef RECURSIVE_FILTER_H
#define RECURSIVE_FILTER_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/recursiveFilters.h"

#include <inrimage/Inrimage.h>

namespace yav {

  /** @name Recursive Filtering Exceptions
   */
  
  //@{

  /** \class InvalidDericheArgument
      \ingroup Inrimage
      \brief invalid argument for deriche filter */
  class InvalidDericheArgument : public Exception {
  public:
    InvalidDericheArgument(const std::string& from, const std::string& arg);
  };
  
  /** \class Uninitialized
      \ingroup Inrimage
      \brief recursive filter descriptor was not initialized */
  class Uninitialized : public Exception {
  public:
    Uninitialized(const std::string& from, const std::string& arg);
  };

  //@}



  /** @name Recursive, Convolution, Gradient and FFT Filtering Structures
   */

  //@{

  /** type of recursive filtering performed : */
  enum FILTER_DERIVATIVE {  
    /** No filtering is performed */
    NO_FILTERING,
    /** the intensity  of the image is smoothed */
    SMOOTHING_FILTER,
    /** The first derivative of the intensity is smoothed */
    FIRST_DERIVATIVE,
    /** The second derivative of the intensity is smoothed */
    SECOND_DERIVATIVE,
    /** The third derivative of the intensity is smoothed */
    THIRD_DERIVATIVE
  };

  
  /** the different type of recursive filtering. Each type corresponds to a
      proposed kernel to convolute with */
  enum RECURSIVE_FILTER_TYPE { 
    /** The approximation of the Gaussian proposed by Deriche 
	that tends to make the filter more isotropic */
    DERICHE_GAUSSIAN,
    /** The approximation of the Gaussian proposed by Marta
	Fidrich that tends to make the filter more isotropic */
    MARTA_GAUSSIAN,
    /** The initial Deriche recursive filter */
    DERICHE_STEP_EDGE
  };


  /// convolution filters
  enum FILTER_TYPE {
    /// gaussian filtering
    GAUSSIAN,
    /// X smoothing
    SMOOTH_X,
    /// Y smoothing
    SMOOTH_Y,
    /// Z smoothing
    SMOOTH_Z,
    /// X first derivative
    DERIVATIVE_X,
    /// Y first derivative
    DERIVATIVE_Y,
    /// Z first derivative
    DERIVATIVE_Z,
    /// X second derivative
    SECOND_DERIVATIVE_X,
    /// Y second derivative
    SECOND_DERIVATIVE_Y,
    /// Z second derivative
    SECOND_DERIVATIVE_Z
  };



  /** recursive filter descriptor */
  struct RECURSIVE_FILTER {
    /** The type of filter */
    RECURSIVE_FILTER_TYPE filter_type;
    /** What should we compute */
    FILTER_DERIVATIVE filter_derivative;
  
    /** The spatial extent of the filter : this is linked
	to the spatial extent of filter by the relation
	alpha = SIGMA_TO_ALPHA/(2*sigma+1)*/
    double alpha;
    /** The number of pixels that is used to enlarge the image
	for filtering. This is set  by gradient3DRecursiveImage()
	and gradient2DRecursiveImage() to floor(sigma) */
    int margin;
    /** Coefficient used in the recursive implementatation :
	set automatically when  choosing the filter type */
    double sd1;
    /** Coefficient used in the recursive implementatation :
       set automatically when choosing the filter type */
    double sd2;
    /** Coefficient used in the recursive implementatation :
       set automatically when choosing the filter type */
    double sd3;
    /** Coefficient used in the recursive implementatation :
       set automatically when choosing the filter type */
    double sd4;
  
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sp0;
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sp1;
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sp2;
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sp3;

    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sn0;
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sn1;
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sn2;
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sn3;
    /** Coefficient used in the recursive implementatation :
	set automatically when choosing the filter type */
    double sn4;
  };
  //@}


  /** @name Gradient Structures */
  //@{
  /** gradient extraction output */
  enum GRADIENT_DIRECTION {
    /// no gradient direction returned
    GD_NONE,
    /// gradient vectors (in a vectorial image) returned
    GD_VECTORS,
    /// sampling of gradient vectors (in a scalar image) returned
    GD_SAMPLE
  };
  //@}

  /** @name FFT Structures */
  //@{
  /// DFT computation kind
  enum FFT_DIRECTION {
    /// direct DFT
    DFT_DIRECT = 1,
    /// inverse DFT
    DFT_INVERSE = -1
  };
  //@}

  
  /** @name Recursive Filtering Functions */


  
  //@{
  /** Convolutes this inrimage with given filter.
      @param inr
      @param filterType filter type
      @param filter_size filter size
      @param twoD whether to use a 2D gaussian
      @exception quedalle */
  void convolute(Inrimage *inr, FILTER_TYPE filterType, int filter_size, bool twoD = false);
  //@}






  /** @name General Recursive Filtering Function */
  //@{
  /** applies a recursive filter on this image
      @param inr
      @param rfilter recursive filter
      @param filterType filter type
      @param sigma recursive filter standard deviation
      @param twoD whether to use a 2D filter
      @param proportionalSigma scale sigma in Y and Z directions according to voxel size ratio with X direction
      @param hdelingMultFactor magic hdeling multiplicative factor (filter margin = hdelingMultFactor * sigma) */
  void recursiveFilter(Inrimage *inr, RECURSIVE_FILTER_TYPE rfilter,
		       FILTER_TYPE filterType,
		       double sigma, bool twoD = false,
		       bool proportionalSigma=true,
		       double hdelingMultFactor=1.0);
  //@}



    /** @name Laplacian Function */
  //@{
  /** computes image laplacian using recursive filters
      @param inr
      @param filter_type filter type
      @param sigma recursive filter standard deviation
      @param twoD whether to use a 2D filter
      @param proportionalSigma scale sigma in Y and Z directions according to voxel size ratio with X direction */
  void laplacian(Inrimage *inr, RECURSIVE_FILTER_TYPE filter_type,
		 double sigma, bool twoD = false,
		 bool proportionalSigma=true);
  //@}

  /** @name Gradient Functions */
  //@{
  /** computes this inrimage gradient using recursive filtering and
      return result in directions buffer
      @param inr
      @param directions gradient direction vectors (a non interlaced WT_FLOAT_VECTOR image with vdim == 3). The image is allocated if a 0 pointer is transmitted.
      @param sigma recursive filter standard deviation
      @param filter recursive filter type
      @param twoD hether to compute 2D gradient
      @param proportionalSigma scale sigma in Y and Z directions according to voxel size ratio with X direction */
  void gradient(const Inrimage *inr, Inrimage* & directions, double sigma, RECURSIVE_FILTER_TYPE filter,
		bool twoD = false, bool proportionalSigma=true) ;

  /** computes this inrimage gradient and norm using recursive filtering and
      return result in out buffer
      @param inr
      @param out returned gradient norm image (this image is allocated by the function)
      @param directions Gradient direction vectors (if dir != GD_NONE). The image is allocated if a 0 pointer is transmitted.
      @param sigma recursive filter standard deviation
      @param filter recursive filter type
      @param dir type of gradient vectors to return
      @param max whether to filter maximum gradient values along gradient direction
      @param twoD hether to compute 2D gradient
      @param lt low threshold on gradient norm
      @param proportionalSigma scale sigma in Y and Z directions according to voxel size ratio with X direction */
  void gradient(const Inrimage *inr, Inrimage *& out, Inrimage *& directions, 
		double sigma, RECURSIVE_FILTER_TYPE filter,
		GRADIENT_DIRECTION dir, bool max,
		bool twoD = false, double lt = 0.0, bool proportionalSigma=true);

  /** computes this inrimage gradient using sobel filters and
      return result in out buffer
      @param inr
      @param out returned gradient norm image (this image is allocated by the function)
      @param directions Gradient direction vectors (if dir != GD_NONE). The image is allocated if a 0 pointer is transmitted.
      @param sigma sobel filters size
      @param dir type of gradient vectors to return
      @param max whether to filter maximum gradient values along gradient direction
      @param twoD whether to compute 2D gradient
      @param lt low threshold on gradient norm
  */
  void sobel(const Inrimage *inr, Inrimage *& out, Inrimage *& directions,
	     unsigned int sigma, GRADIENT_DIRECTION dir,
	     bool max, bool twoD = false,  double lt = 0.0) ;

  //@}

  /** @name 2D Curvature Computation Function */
  //@{
  /** compute image curvature in XY plane
      (Ixx*Iy^2 - 2*Ix*Iy*Ixy + Iyy*Ix^2) / (Ix^2 + Iy^2)^(3/2) */
  void curvature(Inrimage *inr);
  //@}

  /** @name FFT Function */
  //@{
  /** Applied on a scalar image, computes the direct Discrete Fourrier
      Transform (DFT) and return a 2 components vectorial image whose
      first component is real coefficients and second components is
      imaginary coefficients of DFT.
      Applied on a 2 components vectorial image, this function computes
      the inverse DFT and returns a scalar coefficient assuming that the
      result is real.
      @param inr
      @param twoD 2D (slice by slice) DFT computation if true, 3D DFT otherwise
      @exception InvalidSize in DFT-1, image size is not a power of 2 */
  void fft(Inrimage *inr, bool twoD = false) throw(Inrimage::InvalidSize);
  //@} 
  
 


  /// allocates recursive filter descriptor
  RECURSIVE_FILTER *recursiveFilterAlloc(void);
  /// computes recursive filter coefficients
  void computeSn(RECURSIVE_FILTER *);

  /// filters one image line
  void filterRecursiveLine(RECURSIVE_FILTER *rf1, double *input, double *output, double *work1, unsigned int size) throw (Uninitialized);
  /// initializes recursive filter values
  RECURSIVE_FILTER *InitRecursiveFilter(double , RECURSIVE_FILTER_TYPE , FILTER_DERIVATIVE , int ) throw (InvalidDericheArgument);

  //@{
  /// gradient computing
  void gradient2DSobelImage(const Inrimage *inr, Inrimage *& dirvect, unsigned int filter_size);
  /// gradient computing
  void gradient3DSobelImage(const Inrimage *inr, Inrimage *& dirvect, unsigned int filter_size);
  //@}

  //@{
  /// gradient computing
  void keepGradientExtrema(const Inrimage *inr, Inrimage *norm, Inrimage *dirvect, double lt);
  /// gradient computing
  void createImageGradientDirection(const Inrimage *inr, Inrimage *& dir, Inrimage *dirvect,
				    unsigned int frequency) ;


  /// gradient computing
  void gradient2DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect,
				RECURSIVE_FILTER_TYPE filter, double sigma,
				bool proportionalSigma=true) ;
  /// gradient computing
  void gradient3DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect,
				RECURSIVE_FILTER_TYPE filter, double sigma,
				bool proportionalSigma=true) ;


  /// gradient computing. no orthogonal smoothing, ie partial derivatives.
  void fastGradient2DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect,
				    RECURSIVE_FILTER_TYPE filter, double sigma,
				    bool proportionalSigma=true);
  /// gradient computing. no orthogonal smoothing, ie partial derivatives.
  void fastGradient3DRecursiveImage(const Inrimage *inr, Inrimage ** dirvect,
				    RECURSIVE_FILTER_TYPE filter, double sigma,
				    bool proportionalSigma=true);

  //@}
  
  

}


#endif
