#ifndef RESAMPLE_H
#define RESAMPLE_H

#pragma interface "inrimage/resample.h"

#include <inrimage/Inrimage.h>




namespace yav {

  
  /** different type of resampling */
  enum RESAMPLE_TYPE {
    /** */
    NEAREST = 0,
    /** */
    TRILINEAR =1,
    /** */
    CSPLINE =2
  };

  /** different type of filtering performed with cpline resampling */
  enum RESAMPLE_FILTER_DERIVATIVE {  
    /**  */
    SAMPLE_NO_DERIVATIVE = 0,
    /** The first derivative  */
    SAMPLE_FIRST_DERIVATIVE = 1,
    /** The second derivative  */
    SAMPLE_SECOND_DERIVATIVE = 2
  };


  //! reformat an image
  /*!
    reformat an image considering 
   */
#ifdef GRAPHICS
  Inrimage * reformat(Inrimage *inr, unsigned int  orix, unsigned int  oriy, unsigned int  oriz, unsigned int  endx, unsigned int  endy, unsigned int  endz, 
	  SLICE_ORIENTATION geom, Matrix4x4<double> *&mat, bool verticalAxis=true);
#endif
      

  /** Resampling function with 4x4 matrix. We resample an image with a 4x4 
      Matrix of double using a nearest interpolation. It returns a new resampled image 
      @param inr image to be resampled
      @param mat 4x4 matrix 
      @param wt type of the new resampled inrimage 
      @param xdim x dimension of the new resampled inrimage 
      @param ydim y dimension of the new resampled inrimage 
      @param zdim z dimension of the new resampled inrimage 
      @param vx voxel size in the X direction
      @param vy voxel size in the Y direction
      @param vz voxel size in the Z direction
      @param twoD whether to resample in 2D or 3D
      @param gain gain added to the new resampled inrimage 
      @param bias bias added to the new resampled inrimage
      @return a new image which is the input image resampled with mat */      

  Inrimage *resample4x4Nearest ( Inrimage *inr, Matrix4x4<double> &mat, Inrimage::WORD_TYPE wt, unsigned int xdim, unsigned int ydim, unsigned int zdim, double vx=1.0, double vy=1.0, double vz=1.0, bool twoD = false, float gain = 1, float bias = 0 );


  /** Resampling function with 4x4 matrix. We resample an image with a 4x4 
      Matrix of double using a nearest interpolation. It returns a new resampled image 
      @param input image to be resampled
      @param output image resampled
      @param mat 4x4 matrix 
      @param twoD whether to resample in 2D or 3D
      @param gain gain added to the new resampled inrimage 
      @param bias bias added to the new resampled inrimage */

  void resample4x4Nearest ( Inrimage *input,Inrimage *output, Matrix4x4<double> &mat, bool twoD = false, float gain = 1, float bias = 0 );



  /** Resampling function with 4x4 matrix. We resample an image with a 4x4 
      Matrix of double using a trilinear interpolation. It returns a new resampled image 
      @param inr image to be resampled
      @param mat 4x4 matrix 
      @param wt type of the new resampled inrimage 
      @param xdim x dimension of the new resampled inrimage 
      @param ydim y dimension of the new resampled inrimage 
      @param zdim z dimension of the new resampled inrimage 
      @param vx voxel size in the X direction
      @param vy voxel size in the Y direction
      @param vz voxel size in the Z direction
      @param twoD whether to resample in 2D or 3D
      @param gain gain added to the new resampled inrimage 
      @param bias bias added to the new resampled inrimage
      @return a new image which is the input image resampled with mat */      

  Inrimage *resample4x4TriLinear ( Inrimage *inr, Matrix4x4<double> &mat, Inrimage::WORD_TYPE wt, unsigned int xdim, unsigned int ydim, unsigned int zdim, double vx=1.0, double vy=1.0, double vz=1.0, bool twoD = false, float gain = 1, float bias = 0 );
  

  /** Resampling function with 4x4 matrix. We resample an image with a 4x4 
      Matrix of double using a trilinear interpolation. It returns a new resampled image 
      @param inr image to be resampled
      @param output resampled image 
      @param mat 4x4 matrix 
      @param twoD whether to resample in 2D or 3D
      @param gain gain added to the new resampled inrimage 
      @param bias bias added to the new resampled inrimage */

  void resample4x4TriLinear ( Inrimage *inr, Inrimage *output,Matrix4x4<double> &mat, bool twoD = false, float gain = 1, float bias = 0 );
  


  /** Resampling function with 4x4 matrix. We resample an image with a 4x4 
      Matrix of double using a cubic spline interpolation. It returns a new resampled image 
      @param inr image to be resampled
      @param mat 4x4 matrix 
      @param wt type of the new resampled inrimage 
      @param xdim x dimension of the new resampled inrimage 
      @param ydim y dimension of the new resampled inrimage 
      @param zdim z dimension of the new resampled inrimage 
      @param derivative order of derivation computed in each direction
      @param vx voxel size in the X direction
      @param vy voxel size in the Y direction
      @param vz voxel size in the Z direction
      @return a new image which is the input image resampled with mat */      

  Inrimage *resample4x4CSpline ( Inrimage *inr, Matrix4x4<double> &mat, Inrimage::WORD_TYPE wt, unsigned int xdim, unsigned int ydim, unsigned int zdim, RESAMPLE_FILTER_DERIVATIVE derivative[3], double vx=1.0, double vy=1.0, double vz=1.0 );

  /** Resampling function with 4x4 matrix. We resample an image with a 4x4 
      Matrix of double using a cubic spline interpolation. It returns a new resampled image 
      @param inr image to be resampled
      @param output resampled image
      @param mat 4x4 matrix 
      @param derivative order of derivation computed in each direction
      @return a new image which is the input image resampled with mat */      

  void resample4x4CSpline ( Inrimage *inr, Inrimage *output, Matrix4x4<double> &mat, RESAMPLE_FILTER_DERIVATIVE derivative[3]);
  

  /* On applique a l'image d'entree d'abbord la matrice inverse puis le champ de deformation inverse
   Cela est donc equivalent a faire un reechantillonage avec la matrice puis appliquer le champ de deformation 
   a cette nouvelle image avec une matrice identite. */
  

  /** Resampling function with 4x4 matrix and deformations. We first apply the 4x4 matrix and then the field of deformations. It is equivalent to resample it with the 4x4 matrix and the resample the new image with the deformation field and an identity matrix. The interpolation is of nearest type. Inversing the matrix and the field may be necessary depending on what you do.
      It returns a new resampled image. 
      @param inr image to be resampled
      @param defx x image of 3D field deformation ( floating image )
      @param defy y image of 3D field deformation ( floating image )
      @param defz z image of 3D field deformation ( floating image )
      @param mat 4x4 matrix used first
      @param wt type of the new resampled inrimage 
      @param vx voxel size in the X direction
      @param vy voxel size in the Y direction
      @param vz voxel size in the Z direction
      @param gain
      @param bias
      @return a new image which the input image resampled with mat */      

  Inrimage *resampleDefNearest ( Inrimage *inr, Inrimage *defx, Inrimage *defy, Inrimage *defz, Matrix4x4<double> &mat, Inrimage::WORD_TYPE wt, double vx=1.0, double vy=1.0, double vz=1.0, float gain = 1, float bias = 0  );
  

  /** Resampling function with 4x4 matrix and deformations. We first apply the 4x4 matrix and then the field of deformations. It is equivalent to resample it with the 4x4 matrix and the resample the new image with the deformation field and an identity matrix. The interpolation is of trilinear type.Inversing the matrix and the field may be necessary depending on what you do.
      It returns a new resampled image. 
      @param inr image to be resampled
      @param defx x image of 3D field deformation ( floating image )
      @param defy y image of 3D field deformation ( floating image )
      @param defz z image of 3D field deformation ( floating image )
      @param mat 4x4 matrix used first
      @param wt type of the new resampled inrimage 
      @param vx voxel size in the X direction
      @param vy voxel size in the Y direction
      @param vz voxel size in the Z direction
      @param gain
      @param bias
      @return a new image which the input image resampled with mat */      

  Inrimage *resampleDefTriLinear ( Inrimage *inr, Inrimage *defx, Inrimage *defy, Inrimage *defz, Matrix4x4<double> &mat, Inrimage::WORD_TYPE wt, double vx=1.0, double vy=1.0, double vz=1.0, float gain = 1, float bias = 0 );
  
  /** Resampling function with 4x4 matrix and deformations. We first apply the 4x4 matrix and then the field of deformations. It is equivalent to resample it with the 4x4 matrix and the resample the new image with the deformation field and an identity matrix. The interpolation is of cubic spline type.Inversing the matrix and the field may be necessary depending on what you do.
      It returns a new resampled image. 
      @param inr image to be resampled
      @param defx x image of 3D field deformation ( floating image )
      @param defy y image of 3D field deformation ( floating image )
      @param defz z image of 3D field deformation ( floating image )
      @param mat 4x4 matrix used first
      @param wt type of the new resampled inrimage
      @param vx voxel size in the X direction
      @param vy voxel size in the Y direction
      @param vz voxel size in the Z direction
      @param derivative order of derivation computed in each direction
      @return a new image which the input image resampled with mat */      
  Inrimage *resampleDefCSpline ( Inrimage *inr, Inrimage *defx, Inrimage *defy, Inrimage *defz, Matrix4x4<double> &mat, Inrimage::WORD_TYPE wt, RESAMPLE_FILTER_DERIVATIVE derivative[3],double vx=1.0, double vy=1.0, double vz=1.0 );
  

  
}


#endif
