#ifndef THRESHOLD_H
#define THRESHOLD_H

#ifdef _MSC_VER
#pragma warning ( disable : 4068 4786 4081 )
#endif

#pragma interface "inrimage/threshold.h"


#include <inrimage/Inrimage.h>

namespace yav {



/** @name Voxel-based Thresholding Functions
 */

//@{




  /** thresholds image grey levels above min value
      If the voxels values are denoted by v
      Background is defined by: v <  min
      Foreground is defined by: v >= min
      Background voxels will be set to 0, while foreground
      voxels will be set to the maximum value (depending on the word type)
      @param inr
      @param min min threshold value */
  void threshold(Inrimage *inr, double min )
    throw( Exception );

  /** thresholds image grey levels between min and max values
      If the voxels values are denoted by v
      Background is defined by: v <  min || v >  max
      Foreground is defined by: v >= min && v <= max
      Background voxels will be set to 0, while foreground
      voxels will be set to the maximum value (depending on the word type)
      @param inr
      @param min min threshold value
      @param max max threshold value */
  void threshold(Inrimage *inr, double min, double max )
    throw( Exception );

  /** thresholds image grey levels between min and max values
      If the voxels values are denoted by v
      Background is defined by: v <  min ( || v >  max  if max is provided)
      Foreground is defined by: v >= min ( && v <= max  if max is provided)
      Background voxels will be set to 0, while foreground
      voxels will be set to the maximum value (depending on the word type)

      If the voxel max value is NULL, this function is equivalent
      to void threshold(Inrimage *inr, Voxel *min );
      @param inr
      @param min min threshold value
      @param max max threshold value */
  void threshold(Inrimage *inr, Voxel *min, Voxel *max=NULL )
    throw( Exception );

  /** thresholds image grey levels above min value
      If the voxels values are denoted by v
      Background is defined by: v <  min
      Foreground is defined by: v >= min
      Background voxels will be set to bg color, while foreground
      voxels will be set to fg color
      @param inr
      @param min min threshold value
      @param bg background color
      @param fg foreground color */
  void threshold(Inrimage *inr, Voxel *min, Voxel *bg, Voxel *fg)
    throw( Exception );

  /** thresholds image grey levels between min and max values
      If the voxels values are denoted by v
      Background is defined by: v <  min || v >  max
      Foreground is defined by: v >= min && v <= max
      Background voxels will be set to bg color, while foreground
      voxels will be set to fg color
      If the voxel max value is NULL, this function is equivalent
      to void threshold(Inrimage *inr, Voxel *min, Voxel *bg, Voxel *fg );
      @param inr
      @param min min threshold value
      @param max max threshold value
      @param bg background color
      @param fg foreground color */
  void threshold(Inrimage *inr, Voxel *min, Voxel *max, Voxel *bg, Voxel *fg)
    throw( ErrTypeMismatch );
  
  /** thresholds image grey levels between min and max values,
      and changes its type.
      If the voxels values are denoted by v
      Background is defined by: v <  min || v >  max
      Foreground is defined by: v >= min && v <= max
      Background voxels will be set to bg color, while foreground
      voxels will be set to fg color
      If the voxel max value is NULL, this function is equivalent
      to void threshold(Inrimage *inr, Voxel *min, Voxel *bg, Voxel *fg );
      @param inr
      @param type
      @param min min threshold value
      @param max max threshold value
      @param bg background color
      @param fg foreground color */
  void threshold(Inrimage *inr, Inrimage::WORD_TYPE type,
		 Voxel *min, Voxel *max, Voxel *bg, Voxel *fg)
    throw( ErrTypeMismatch );
  









/** Binarizes an image. 
    If a pixel value v is not a background pixel
    then it is set to foreground.
    In other words, voxels of value 'bg' are keeped to this value
    while all the others values are changed to 'fg'.
    @param inr
    @param bg background color
    @param fg foreground color */
  void binarize( Inrimage *inr, double bg, double fg )
    throw( Exception );

/** Binarizes an image. 
    If a pixel value v is not a background pixel
    then it is set to foreground.
    In other words, voxels of value 'bg' are keeped to this value
    while all the others values are changed to 'fg'.
    @param inr
    @param bg background color
    @param fg foreground color */
  void binarize( Inrimage *inr, Voxel *bg, Voxel *fg )
    throw( Exception );

/** Binarizes an image. 
    If a pixel value v is not a background pixel
    then it is set to foreground.
    In other words, voxels of value 'theBg' are changed to value 'resBg'
    while all the others values are changed to 'resFg'.
    @param inr
    @param theBg input background color
    @param resBg output background color
    @param resFg output foreground color */
  void binarize( Inrimage *inr, Voxel *theBg, Voxel *resBg, Voxel *resFg ) 
    throw( ErrTypeMismatch );

/** Binarizes an image and changes its type.
    If a pixel value v is not a background pixel
    then it is set to foreground.
    In other words, voxels of value 'theBg' are changed to value 'resBg'
    while all the others values are changed to 'resFg'.
    @param inr
    @param type inrimage type
    @param theBg input background color
    @param resBg output background color
    @param resFg output foreground color */
  void binarize( Inrimage *inr, Inrimage::WORD_TYPE type, 
		 Voxel *theBg, Voxel *resBg, Voxel *resFg ) 
    throw( ErrTypeMismatch );
  /** ---
      @param inr
      @param type
      @param theBg
      @param resBg
      @param resFg */
  void binarizeToOtherFormat( Inrimage *inr, Inrimage::WORD_TYPE type, 
			      Voxel *theBg, Voxel *resBg, Voxel *resFg ) 
    throw( Exception );
  









  /** Applies a double thresholding on this image. If a pixel value v < lt then
     it is set to bg. If lt <= v < ht then it is set to lf and if v>= ht
     then it is set to hf
     @param inr
     @param lt low threshold
     @param ht high threshold
     @param bg background value
     @param lf low foreground
     @param hf high foreground */   
  void ternarize(Inrimage *inr, double lt, double ht, double bg, double lf, double hf)
    throw( Exception );

  /** Applies a double thresholding on this image. If a pixel value v < lt then
     it is set to bg. If lt <= v < ht then it is set to lf and if v>= ht
     then it is set to hf
     @param inr
     @param lt low threshold
     @param ht high threshold
     @param bg background value
     @param lf low foreground
     @param hf high foreground */   
  void ternarize(Inrimage *inr, Voxel *lt, Voxel *ht, Voxel *bg, Voxel *lf, Voxel *hf)
    throw( ErrTypeMismatch );

  /** Applies a double thresholding on this image.
      We call v a pixel of the input image and w its corresponding pixel in
      the output image. If a pixel value v < lt then w is set to background.
      If lt <= v < ht then w is set to lf and if v>= ht then w is set to hf
      @param inr
      @param type new inrimage type
      @param lt low threshold
      @param ht high threshold
      @param bg background value
      @param lf low foreground
      @param hf high foreground */     
  void ternarize( Inrimage *inr, Inrimage::WORD_TYPE type, 
		 double lt, double ht, double bg, double lf, double hf )
    throw( Exception );

  /** Applies a double thresholding on this image.
      We call v a pixel of the input image and w its corresponding pixel in
      the output image. If a pixel value v < lt then w is set to background.
      If lt <= v < ht then w is set to lf and if v>= ht then w is set to hf
      @param inr
      @param type new inrimage type
      @param lt low threshold
      @param ht high threshold
      @param bg background value
      @param lf low foreground
      @param hf high foreground */     
  void ternarize( Inrimage *inr, Inrimage::WORD_TYPE type, 
		 Voxel *lt, Voxel *ht, Voxel *bg, Voxel *lf, Voxel *hf )
    throw( ErrTypeMismatch );
  /** ---
      @param inr
      @param type
      @param lt
      @param ht
      @param bg
      @param lf
      @param hf
  */
  void ternarizeToOtherFormat(Inrimage *inr, Inrimage::WORD_TYPE type, 
			      Voxel *lt, Voxel *ht, Voxel *bg, Voxel *lf, Voxel *hf)
    throw( Exception );


//@}

/* end of Voxel-based Thresholding Functions */










  /** apply a windowing of an image given min  and max values : the output image is an unsigned char image having the same size as the original image : original voxels having a grey value less than min are set to min, those having a grey value greater than max are set to 255 and a linear scaling is applied otherwise
     @param input input image 
     @param min min threshold value
     @param max max threshold value */
  Inrimage *windowing(Inrimage *input, Voxel *min, Voxel *max);





  /** thresholds image grey levels between min and max values
      voxels lower than min or higher than max value are set to bg color
      other voxels remain unchanged
      @param inr
      @param min min threshold value
      @param max max threshold value
      @param bg background color */
  void cropHistogramm(Inrimage *inr, Voxel *min, Voxel *max, Voxel *bg);
  








}




#endif
