/*=========================================================================
  Program:   Insight Segmentation & Registration Toolkit
  Module:    $RCSfile: mipsInrimageImageIO.cxx,v $
  Language:  C++
  Date:      $Date: 2006/09/27 12:17:44 $
  Version:   $Revision: 1.12 $

  Copyright (c) 2002 Insight Consortium. All rights reserved.
  See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

  This software is distributed WITHOUT ANY WARRANTY; without even
  the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the above copyright notices for more information.

  =========================================================================*/

#include "mipsInrimageImageIO.h"
#include "itkExceptionObject.h"
#include "itkByteSwapper.h"
#include <itkImageIOBase.h> 

#include <stdio.h>
#include <stdlib.h>
#include <fstream>


#if defined(_WIN32) && (defined(_MSC_VER) || defined(__BORLANDC__))
#include <stdlib.h>
#define _unlink unlink
#else
#include <unistd.h>
#endif



//
// GetExtension from itkAnalyzeImageIO.cxx
//
static std::string
GetExtension( const std::string& filename ) {

  const char* c_filename = filename.c_str();
  std::string ext = "";

  int len = strlen (c_filename);
  if( len<4 )
  {
    return ext;
  }

  ext = c_filename + len - 4;

  // search for .inr
  if( ext==".inr" )
  {
    return ext;
  }

  // else we try to find .inr.gz
  if( len < 7 )
  {
    return ext;
  }
  
  ext = c_filename + len -7;

  return ext;
  

  /*
  // This assumes that the final '.' in a file name is the delimiter
  // for the file's extension type
  const std::string::size_type it = filename.find_last_of( "." );
  
  // This determines the file's type by creating a new string
  // who's value is the extension of the input filename
  // eg. "myimage.gif" has an extension of "gif"
  std::string fileExt( filename, it+1, filename.length() );
  
  // For backwards compatibility.  If a filename has no extension
  if( it == filename.max_size() || fileExt[0] == '/' )    
  {
    std::string fileExt( "" );
    return( fileExt );
  }
  
  // Force the name to lower case
  //this->correctCase( fileExt );

  return( fileExt );
  */
}



//
// GetRootName from itkAnalyzeImageIO.cxx
//
static std::string
GetRootName( const std::string& filename ) 
{
  const std::string fileExt = GetExtension(filename);

  // Create a base filename
  // i.e Image.gif --> Image
  if( fileExt.length() > 0 ) 
    {
      const std::string::size_type it = filename.find_last_of( fileExt );
      std::string baseName( filename, 0, it-fileExt.length() );
      return( baseName );
    } 
  else 
    {
      // Case when the extension is nothing
      return( filename );
    }
}

// ImageFileReader::GenerateOutputInformation 
//   ( in Insight/Code/IO/itkImageFileReader.txx )
// - call ImageIOFactory::CreateImageIO() 
//     ( in Insight/Code/IO/itkImageIOFactory.cxx )
//   if m_ImageIO.IsNull()
//   - ImageIOFactory::CreateImageIO() ( in Insight/Code/IO/itkImageIOFactory.cxx )
//     will call either CanReadFile() or CanWriteFile()
// - set file name m_ImageIO->SetFileName();
// - call m_ImageIO->ReadImageInformation();
// - set dimensions, spacing, origine
// - output->SetMetaDataDictionary()
//
// ImageFileReader::GenerateData()
// - ...
// - set dimensions
// - if no buffer conversion required
//   if ( m_ImageIO->GetPixelType() == typeid(TOutputImage::PixelType) &&
//        (m_ImageIO->GetNumberOfComponents() 
//         == ConvertPixelTraits::GetNumberOfComponents()) )
//      m_ImageIO->Read(buffer);
//   else
//      m_ImageIO->Read(loadBuffer);
//      this->DoConvertBuffer(loadBuffer, region.GetNumberOfPixels());


namespace itk
{


  // Default constructor
  InrimageImageIO::InrimageImageIO()
  {
    // by default, only have 2 dimensions
    this->SetNumberOfDimensions( 2 );

    // Initialization
    this->SetOrigin( 0, 0.0 );
    this->SetOrigin( 1, 0.0 );

    this->SetSpacing( 0, 1.0 );
    this->SetSpacing( 1, 1.0 );
    
    // pixel type is UCHAR
    // m_PixelType is a protected member of ImageIOBase (IODataType itk::ImageIOBase::m_PixelType)
    this->SetPixelType( SCALAR );
    this->SetNumberOfComponents( 1 );
    this->SetComponentType( DOUBLE );
    

    // Byte order
    this->SetByteOrder( ( ByteSwapper<int>::SystemIsBigEndian() == true ) ? BigEndian : LittleEndian );
    
    // additionnal information
    // (inrimage only)
    m_file = NULL;
    //length of header of a multiple of 256
    m_NumberBlocksInHeader = 1;

  }



  InrimageImageIO::~InrimageImageIO()
  {
    //Purposefully left blank
  }



  void InrimageImageIO::PrintSelf(std::ostream& os, Indent indent) const
  {
    Superclass::PrintSelf(os, indent);
    os << "LOCAL PARAMETERS" << std::endl;
    os << indent << "hdr=" << m_NumberBlocksInHeader << std::endl;
    os << "GLOBAL PARAMETERS" << std::endl;
    os << indent << "ByteOrder=" << m_ByteOrder << std::endl;
    os << indent << "PixelType " << m_PixelType << std::endl;
  }




const std::type_info& InrimageImageIO::GetPixelType()
{
  switch( m_PixelType )
    {
    case CHAR:
      return typeid(char);
    case UCHAR:
      return typeid(unsigned char);
    case SHORT:
      return typeid(short);
    case USHORT:
      return typeid(unsigned short);
    case INT:
      return typeid(int);
    case UINT:
      return typeid(unsigned int);
    case LONG:
      return typeid(long);
    case ULONG:
      return typeid(unsigned long);
    case FLOAT:
      return typeid(float);
    case DOUBLE:
      return typeid(double);
    case UNKNOWNPIXELTYPE :
      itkExceptionMacro ("Unknown pixel type: " << m_PixelType);
    default:
      itkExceptionMacro ("Invalid type: " << m_PixelType );
      return this->ConvertToTypeInfo(m_PixelType);   
    }
  return typeid(ImageIOBase::UnknownType);
}

  const std::type_info& InrimageImageIO::ConvertToTypeInfo(IOPixelType t) const
  {

    switch(t)
    {
    case UCHAR:
      return typeid(unsigned char);
    case CHAR:
      return typeid(char);
    case USHORT:
      return typeid(unsigned short);
    case SHORT:
      return typeid(short);
    case UINT:
      return typeid(unsigned int);
    case INT:
      return typeid(int);
    case ULONG:
      return typeid(unsigned long);
    case LONG:
      return typeid(long);
    case FLOAT:
      return typeid(float);
    case DOUBLE:
      return typeid(double);
    /*case RGB:
      return typeid(RGBPixel<unsigned char>);
      case RGBA:
      return typeid(RGBAPixel<unsigned char>);
    */
    default:
      itkExceptionMacro ("Invalid type: " << m_PixelType );
    }
  return typeid(ImageIOBase::UNKNOWNPIXELTYPE);

  }


/* NO MORE USEFUL FOR ITK > 1.8
   unsigned int InrimageImageIO::GetComponentSize() const
   {
   switch(m_ComponentType)
   {
   case CHAR:
   return sizeof(char);
   case UCHAR:
   return sizeof(unsigned char);
   case SHORT:
   return sizeof(short);
   case USHORT:
   return sizeof(unsigned short);
   case INT:
   return sizeof(int);
   case UINT:
   return sizeof(unsigned int);
   case LONG:
   return sizeof(long);
   case ULONG:
   return sizeof(unsigned long);
   case FLOAT:
   return sizeof(float);
   case DOUBLE:
   return sizeof(double);
   case UNKNOWNCOMPONENTTYPE :
   default:
   itkExceptionMacro ("Invalid type: " << m_ComponentType );
   return 0;
   }
   return 1;
   }
*/




  //
  // called from CreateImageIO ( Insight/Code/IO/itkImageIOFactory.cxx )
  // called from ImageFileReader::GenerateOutputInformation ( Insight/Code/IO/itkImageFileReader.txx )
  bool InrimageImageIO::CanReadFile( const char* FileNameToRead )
  {
    m_FileName = FileNameToRead;

    // don't care about extensions
    // should be 'inr' but can be everything else
    
    // try to open the file
    m_file = ::gzopen( m_FileName.c_str(), "rb" );
    if( m_file == NULL ) 
    {
      std::string fileExt = GetExtension(m_FileName);
      // if it doesn't work, try by substracting extension .gz
      if( !fileExt.compare(".inr.gz") )
      {
	std::string ImageFileName = GetRootName(m_FileName);
	m_file = ::gzopen( ImageFileName.c_str(), "rb" );
	if( m_file == NULL ) 
	{
	  return false;
	}
	m_FileName = ImageFileName;
      }
      // if it still doesn't work, try by adding extension .gz
      else 
      {
	std::string ImageFileName = m_FileName;
	ImageFileName += ".gz";
	m_file = ::gzopen( ImageFileName.c_str(), "rb" );
	if( m_file == NULL ) 
	{
	  return false;
	}
	m_FileName = ImageFileName;
      }
    }
    
    // here m_file is open
    char buf_header[257];
    int nread;


    // find magic string at beginning
    nread = ::gzread( m_file, buf_header, 256 );
    if ( nread != 256 ) {
      ::gzclose( m_file);
      return false;
    }

    if ( strncmp( buf_header, "#INRIMAGE-4#{", strlen("#INRIMAGE-4#{") ) != 0 ) {
      ::gzclose( m_file);
      return false;
    }
    
    buf_header[256] = '\0';
    m_header = buf_header;
    m_NumberBlocksInHeader = 1;
    
    // find magic string at end of header
    while ( strncmp( &(buf_header[252]), "##}\n", strlen("##}\n") ) != 0 ) {
      nread = ::gzread( m_file, buf_header, 256 );
      if ( nread != 256 ) {
	::gzclose( m_file);
	return false;
      }
      
      buf_header[256] = '\0';
      m_header += buf_header;
      m_NumberBlocksInHeader ++;
    }
    
    // cout << header <<endl;

    // check endianess
    // (as in itkAnalyzeImageIO.cxx)

    ::gzclose( m_file);

    return true;
  }



  void InrimageImageIO::ReadImageInformation()
  {
    // m_header should have been read (in CanReadFile)
    if( m_header.length() == 0)
      this->CanReadFile (this->GetFileName() );
    
    
    // file endianess
    // default = Sytem
    if ( ByteSwapper<int>::SystemIsBigEndian() == true )
      this->SetByteOrderToBigEndian();
    else
      this->SetByteOrderToLittleEndian();

    // pixel type
    // to be read
    int pixsize = 0;
    std::string type;
    int ncomponents = this->GetNumberOfComponents();

    // parsing the header line by line 
    std::istringstream issh( m_header );
    std::string line;

    //std::cout << "GeTLINE: " << std::getline( issh, line ) << std::endl;
    
    
    while ( std::getline( issh, line ) ) {

      // skip comments and empty lines
      if ( line[0] == '#') continue;
      if ( line.length() == 0) continue;

      // cout << "1#" << line << "#1" << endl;

      // translate remaining lines
      // go to "="
      std::istringstream issl( line );
      std::string id;
      std::getline( issl, id, '=' );
      
      int i;
      double r;
      std::string s;
      

      
      // to have information about the used methods
      // see http://www.itk.org/Insight/Doxygen/html/classitk_1_1ImageIOBase.html
      // documentation of "Abstract superclass defines image IO interface" itk::ImageIOBase


      switch ( id[0] ) {

      default :
	break;

      case 'C' :
	if ( std::string( id, 0, strlen("CPU")).compare( "CPU" ) == 0 ) {
	  issl >> s;
	  if ( std::string( s, 0, strlen("decm")).compare( "decm" ) == 0 
	       || std::string( s, 0, strlen("alpha")).compare( "alpha" ) == 0 
	       || std::string( s, 0, strlen("linux")).compare( "linuxdecm" ) == 0 
	       || std::string( s, 0, strlen("pc")).compare( "pc" ) == 0 ) {
	    this->SetByteOrderToLittleEndian();
	  }
	  else if ( std::string( s, 0, strlen("sun")).compare( "sun" ) == 0 
		    || std::string( s, 0, strlen("sgi")).compare( "sgi" ) == 0 ) {
	    this->SetByteOrderToBigEndian();
	  }
	  else {
	    itkExceptionMacro ( "Endianess (CPU) Unknown" );
	    return;
	  }
	}
	break;

      case 'P' :
	if ( std::string( id, 0, strlen("PIXSIZE")).compare( "PIXSIZE" ) == 0 ) {
	  issl >> pixsize;
	  issl >> s;
          
	  // cerr << pixsize << s << endl;
	  if ( (pixsize == 0) || (pixsize % 8 != 0 ) ||
	       (std::string( s, 0, strlen("bits")).compare( "bits" ) != 0) ) {
            ExceptionObject exception(__FILE__, __LINE__);
            exception.SetDescription("Pixel Size (PIXSIZE) Unknown");
            throw exception;
	  }
	}
	break;

      case 'S' :
	// SCALE = 
	;
	break;

      case 'T' :
	if ( std::string( id, 0, strlen("TYPE")).compare( "TYPE" ) == 0 ) {
	  issl >> type;
	}
	else if ( std::string( id, 0, strlen("TX")).compare( "TX" ) == 0 ) {
	  issl >> r;
	  this->SetOrigin( 0, r );
	}
	else if ( std::string( id, 0, strlen("TY")).compare( "TY" ) == 0 ) {
	  issl >> r;
	  this->SetOrigin( 1, r );
	}
	else if ( std::string( id, 0, strlen("TZ")).compare( "TZ" ) == 0 ) {
	  issl >> r;
	  if (this->GetNumberOfDimensions() > 2)
	    this->SetOrigin( 2, r );
	}
	break;

      case 'X' :
	if ( std::string( id, 0, strlen("XDIM")).compare( "XDIM" ) == 0 ) {
	  issl >> i;
	  this->SetDimensions( 0, i );
	}
	break;

      case 'Y' :
	if ( std::string( id, 0, strlen("YDIM")).compare( "YDIM" ) == 0 ) {
	  issl >> i;
	  this->SetDimensions( 1, i );
	}
	break;

      case 'Z' :
	if ( std::string( id, 0, strlen("ZDIM")).compare( "ZDIM" ) == 0 ) {
	  issl >> i;
	  if ( i > 1 ) {
	    this->SetNumberOfDimensions( 3 );
	    this->SetDimensions( 2, i );
	  }
	}
	break;

      case 'V' :
	if ( std::string( id, 0, strlen("VDIM")).compare( "VDIM" ) == 0 ) {
	  issl >> i;
	  if ( i > 1 ) {
	    // ??? this->SetComponentType( VECTOR );
	    ncomponents = i;
	    this->SetNumberOfComponents( ncomponents );
	  }
	}
	else if ( std::string( id, 0, strlen("VX")).compare( "VX" ) == 0 ) {
	  issl >> r;
	  this->SetSpacing( 0, r );
	}
	else if ( std::string( id, 0, strlen("VY")).compare( "VY" ) == 0 ) {
	  issl >> r;
	  this->SetSpacing( 1, r );
	}
	else if ( std::string( id, 0, strlen("VZ")).compare( "VZ" ) == 0 ) {
	  issl >> r;
	  if ( this->GetNumberOfDimensions() > 2 )
	    this->SetSpacing( 2, r );
	}
	
	break;
      }

    }

    // header parsing is done
    // TYPE is "unsigned fixed", "signed fixed", or "float"
    
    
    if ( std::string( type, 0, strlen("unsigned")).compare( "unsigned" ) == 0 ) {
      if ( pixsize == 8 ) this->SetComponentType( UCHAR );
      else if ( pixsize ==  16 ) this->SetComponentType( USHORT );
      else if ( pixsize ==  32 ) this->SetComponentType( UINT );
      else if ( pixsize ==  64 ) this->SetComponentType( ULONG );
      else {
	ExceptionObject exception(__FILE__, __LINE__);
	exception.SetDescription("Pixel Size Not Recognized");
	throw exception;
      }
    }
    else if ( std::string( type, 0, strlen("signed")).compare( "signed" ) == 0 ) {
      if ( pixsize == 8 ) this->SetComponentType( CHAR );
      else if ( pixsize ==  16 ) this->SetComponentType( SHORT );
      else if ( pixsize ==  32 ) this->SetComponentType( INT );
      else if ( pixsize ==  64 ) this->SetComponentType( LONG );
      else {
	ExceptionObject exception(__FILE__, __LINE__);
	exception.SetDescription("Pixel Size Not Recognized");
	throw exception;
      }
    }
    else if ( std::string( type, 0, strlen("float")).compare( "float" ) == 0 ) {
      if ( pixsize ==  32 ) this->SetComponentType( FLOAT );
      else if ( pixsize ==  64 ) this->SetComponentType( DOUBLE );
      else {
	ExceptionObject exception(__FILE__, __LINE__);
	exception.SetDescription("Pixel Size Not Recognized");
	throw exception;
      }
    }

    this->SetPixelType(SCALAR);
    
    if ( 1 ) {
      std::cout << "type " << type << std::endl;
      std::cout << "pixsize " << pixsize << std::endl;
      std::cout << "ComponentType " << this->GetComponentType() << std::endl;
      std::cout << "PixelType ";
      if ( this->GetComponentTypeInfo() == typeid(unsigned char) )
          std::cout << "UCHAR";

      else if ( this->GetComponentTypeInfo() == typeid(unsigned short) )
          std::cout << "USHORT";

      else if ( this->GetComponentTypeInfo() == typeid(signed short) )
          std::cout << "SHORT";

      else 
	std::cout << "not handled in switch";
      std::cout << std::endl;
    }

    // I am not sure that this has to be done
    // GM
    if ( 0 ) {
      if ( this->GetNumberOfComponents() == 3  ) {
	switch ( this->GetComponentType() ) {
	default :
	  break;
	case UCHAR :
	  this->SetPixelType( RGB );
	  this->SetNumberOfComponents( 1 );
	  break;
	}
      }
      else if ( this->GetNumberOfComponents() == 4  ) {
	switch ( this->GetComponentType() ) {
	default :
	  break;
	case UCHAR :
	  this->SetPixelType( RGBA );
	  this->SetNumberOfComponents( 1 );
	  break;
	}
      }
    }
    
    return;
  }



  void InrimageImageIO::Read(void* buffer)
  {
    unsigned int nread;

    if ( 1 ) {
      std::cerr << "* ComponentType " << this->GetComponentType() << std::endl;
      std::cerr << "PixelType ";
      if ( this->GetComponentTypeInfo() == typeid(unsigned char) )
	std::cerr << "UCHAR";
      else if ( this->GetComponentTypeInfo() == typeid(unsigned short) )
	std::cerr << "USHORT";
      else if ( this->GetComponentTypeInfo() == typeid(signed short) )
	std::cerr << "=SSHORT";
      else 
	std::cerr << "not handled in switch";
      std::cerr << std::endl;
    }

    m_file = ::gzopen( m_FileName.c_str(), "rb" );
    if( m_file == NULL ) {
      ExceptionObject exception(__FILE__, __LINE__);
      exception.SetDescription("Unable to open file" );
      throw exception;
    }

    char buf_header[257];
    for( unsigned int i=0; i< m_NumberBlocksInHeader; i++ ) {
      nread = ::gzread( m_file, buf_header, 256 );
      if ( nread != 256 ) {
	::gzclose( m_file);
	ExceptionObject exception(__FILE__, __LINE__);
	exception.SetDescription("Unable to skip header" );
	throw exception;
      }
    }
    
    char * p = static_cast<char *>(buffer);
    nread = ::gzread( m_file, p, this->GetImageSizeInBytes()+10000 );
    if ( nread != this->GetImageSizeInBytes() ) {
      ::gzclose( m_file);
      ExceptionObject exception(__FILE__, __LINE__);
      exception.SetDescription("Unable to read buffer" );
      throw exception;
    }

    ::gzclose( m_file);

    SwapBytesIfNecessary( buffer, this->GetImageSizeInComponents() );

  }
 


  /**
   *
   */
  bool InrimageImageIO::CanWriteFile(const char * FileNameToWrite)
  {
    m_FileName=FileNameToWrite;

    if ( m_FileName != "" )
    {
      const std::string fileExt = GetExtension( m_FileName );
      if( strcmp (fileExt.c_str(),".inr.gz")==0 || strcmp (fileExt.c_str(),".inr")==0 )
        return true;
      
    }
    
    return false;
  }

  void
  InrimageImageIO
  ::WriteImageInformation(void)
  {
    return;
  }


  /**
   *
   */
  void
  InrimageImageIO
  ::Write( const void* buffer)
  {

    // try to open the file
    bool isGz = true;
    std::string fileExt = GetExtension(m_FileName);
    // if it doesn't work, try by substracting extension .gz
    if( fileExt==(".inr.gz") )
    {
      m_file = ::gzopen( m_FileName.c_str(), "wb" );
      if( m_file == NULL ) 
        throw ExceptionObject(__FILE__,__LINE__,"Error in opening file for writing");
    }
    else
    {
      if( fileExt==(".inr") )
      {
        isGz = false;
        m_file = fopen( m_FileName.c_str(), "wb" );
        if( m_file == NULL ) 
          throw ExceptionObject(__FILE__,__LINE__,"Error in opening file for writing");
      }
      else
        throw ExceptionObject(__FILE__,__LINE__,"Unrecognized extension.");
    }
    

    std::string type = "signed fixed";
    char scale[20];
    switch( this->GetComponentType() )
    {
        case UCHAR:
        case USHORT:
        case UINT:
        case ULONG:
          type = "unsigned fixed";
          sprintf(scale, "SCALE=2**0\n");
          break;

        case CHAR:
        case SHORT:
        case INT:
        case LONG:
          type = "signed fixed";
          sprintf(scale, "SCALE=2**0\n");
          break;

        case FLOAT:
        case DOUBLE:
          type = "float";
          scale[0] = '\0';
          break;

        default:
          throw ExceptionObject (__FILE__,__LINE__,"Error: pixel type not recognized.");
    }


    int pixsize = 8;
    switch( this->GetComponentType() )
    {
        case UCHAR:
        case CHAR:
          pixsize = 8;
          break;
          
        case USHORT:
        case SHORT:
          pixsize = 16;
          break;
          
        case UINT:
        case INT:
        case FLOAT:
          pixsize = 32;
          break;
          
        case ULONG:
        case LONG:
        case DOUBLE:
          pixsize = 64;
          break;

        default:
          pixsize=0;
          break;
    }

    std::string endianness;
    if( this->GetByteOrder() == LittleEndian )
      endianness = "decm";
    else
      endianness = "sun";
  
    char buf[257];
    /* write header information */
    sprintf(buf, "#INRIMAGE-4#{\nXDIM=%i\nYDIM=%i\nZDIM=%i\nVDIM=%d\nTYPE=%s\nPIXSIZE=%i bits\n%sCPU=%s\nVX=%f\nVY=%f\nVZ=%f\nTX=%f\nTY=%f\nTZ=%f\n#GEOMETRY=CARTESIAN\n",
	    this->GetDimensions (0), this->GetDimensions (1), this->GetDimensions (2), this->GetNumberOfComponents(),
	    type.c_str(), pixsize, scale, endianness.c_str(),
            this->GetSpacing (0), this->GetSpacing (1), this->GetSpacing (2),
            this->GetOrigin (0), this->GetOrigin (1), this->GetOrigin (2));


    int pos = strlen (buf);
    size_t npix = 0;
    // write the header:
    if( isGz )
      npix = ::gzwrite (m_file, (void *)buf, pos);
    else
      npix = fwrite ((void *)buf, 1, pos, (FILE*)m_file);
    
    if( npix!=strlen (buf))
    {
      if( isGz )
        ::gzclose( m_file);
      else
        fclose ((FILE*)m_file);
      throw ExceptionObject (__FILE__,__LINE__,"Error: bad number of bytes written.");
    }

    /* write end of header */
    pos = pos % 256;
    if(pos > 252) {
      for(int i = pos; i < 256; i++)
        if( isGz )
        {
          if( ::gzwrite (m_file, (void *)("\n"), 1) != 1 )
          {
            ::gzclose (m_file);
            throw ExceptionObject (__FILE__,__LINE__,"Error: bad number of bytes written.");
          }
        }
        else
        {
          if( fwrite((void *)("\n"), 1, 1, (FILE*)m_file) !=1 )
          {
            fclose ((FILE*)m_file);
            throw ExceptionObject (__FILE__,__LINE__,"Error: bad number of bytes written.");
          }
        }
      
      pos = 0;
    }
    buf[0] = '\0';
    for(int i = pos; i < 252; i++) strcat(buf, "\n");
    strcat(buf, "##}\n");

    if( isGz )
    {
      if( ::gzwrite (m_file, (void *)buf, strlen(buf)) !=  (int)(strlen(buf)) )
      {
        ::gzclose (m_file);
        throw ExceptionObject (__FILE__,__LINE__,"Error: bad number of bytes written.");
      }      
    }
    else
    {
      if( fwrite ((void *)buf, 1, strlen (buf), (FILE*)m_file) != (size_t)(strlen(buf)) )
      {
        fclose ((FILE*)m_file);
        throw ExceptionObject (__FILE__,__LINE__,"Error: bad number of bytes written.");
      }      
    }
    
    
    
    // write the buffer:
    if( isGz )
      npix = ::gzwrite (m_file, (void *)buffer, this->GetImageSizeInBytes());
    else
      npix = fwrite ((void *)buffer, 1, this->GetImageSizeInBytes(), (FILE*)m_file);
    if( (unsigned int)npix!=this->GetImageSizeInBytes())
    {
      if( isGz )
        ::gzclose( m_file);
      else
        fclose ((FILE*)m_file);
      throw ExceptionObject (__FILE__,__LINE__,"Error: bad number of bytes written.");
    }
    

    if( isGz )
      ::gzclose( m_file);
    else
      fclose ((FILE*)m_file);
    
  }



  /************************************************************
   *
   * private methods
   *
   ************************************************************/
  
  void
  InrimageImageIO::SwapBytesIfNecessary( void* buffer, 
					 unsigned long numberOfPixels )
  {
    if ( ImageIOBase::GetByteOrder() == LittleEndian ) 
      {
        switch(m_PixelType) 
          {
          case CHAR:
            ByteSwapper<char>::SwapRangeFromSystemToLittleEndian((char*)buffer,
                                                                 numberOfPixels );
            break;
          case UCHAR:
            ByteSwapper<unsigned char>::SwapRangeFromSystemToLittleEndian
              ((unsigned char*)buffer, numberOfPixels );
            break;
          case SHORT:
            ByteSwapper<short>::SwapRangeFromSystemToLittleEndian
              ((short*)buffer, numberOfPixels );
            break;
          case USHORT:
            ByteSwapper<unsigned short>::SwapRangeFromSystemToLittleEndian
              ((unsigned short*)buffer, numberOfPixels );
            break;
          case INT:
            ByteSwapper<int>::SwapRangeFromSystemToLittleEndian
              ((int*)buffer, numberOfPixels );
            break;
          case UINT:
            ByteSwapper<unsigned int>::SwapRangeFromSystemToLittleEndian
              ((unsigned int*)buffer, numberOfPixels );
            break;
          case LONG:
            ByteSwapper<long>::SwapRangeFromSystemToLittleEndian
              ((long*)buffer, numberOfPixels );
            break;
          case ULONG:
            ByteSwapper<unsigned long>::SwapRangeFromSystemToLittleEndian
              ((unsigned long*)buffer, numberOfPixels );
            break;
          case FLOAT:
            ByteSwapper<float>::SwapRangeFromSystemToLittleEndian((float*)buffer,
                                                                  numberOfPixels );
            break;
          case DOUBLE:
            ByteSwapper<double>::SwapRangeFromSystemToLittleEndian
              ((double*)buffer, numberOfPixels );
            break;
          default:
            ExceptionObject exception(__FILE__, __LINE__);
            exception.SetDescription("Pixel Type Unknown");
            throw exception;
          }
      }
    else 
      {
        switch(m_PixelType) 
          {
          case CHAR:
            ByteSwapper<char>::SwapRangeFromSystemToBigEndian((char *)buffer,
                                                              numberOfPixels );
            break;
          case UCHAR:
            ByteSwapper<unsigned char>::SwapRangeFromSystemToBigEndian
              ((unsigned char *)buffer, numberOfPixels );
            break;
          case SHORT:
            ByteSwapper<short>::SwapRangeFromSystemToBigEndian
              ((short *)buffer, numberOfPixels );
            break;
          case USHORT:
            ByteSwapper<unsigned short>::SwapRangeFromSystemToBigEndian
              ((unsigned short *)buffer, numberOfPixels );
            break;
          case INT:
            ByteSwapper<int>::SwapRangeFromSystemToBigEndian
              ((int *)buffer, numberOfPixels );
            break;
          case UINT:
            ByteSwapper<unsigned int>::SwapRangeFromSystemToBigEndian
              ((unsigned int *)buffer, numberOfPixels );
            break;
          case LONG:
            ByteSwapper<long>::SwapRangeFromSystemToBigEndian
              ((long *)buffer, numberOfPixels );
            break;
          case ULONG:
            ByteSwapper<unsigned long>::SwapRangeFromSystemToBigEndian
              ((unsigned long *)buffer, numberOfPixels );
            break;
          case FLOAT:
            ByteSwapper<float>::SwapRangeFromSystemToBigEndian
              ((float *)buffer, numberOfPixels );
            break;
          case DOUBLE:
            ByteSwapper<double>::SwapRangeFromSystemToBigEndian
              ((double *)buffer, numberOfPixels );
            break;
          default:
            ExceptionObject exception(__FILE__, __LINE__);
            exception.SetDescription("Pixel Type Unknown");
            throw exception;
          }
      }
  }



  
  
} // end namespace itk
