#include <inrimage/StructuringElement.h>

using namespace yav;


typedef enum {
  _EROSION_,
  _DILATION_,
  _OPENING_,
  _CLOSING_,
  _TOPHAT_,
  _INVTOPHAT_,
  _GRADIENT_,
  _LOWER_GRADIENT_,
  _UPPER_GRADIENT_,
  _WATERSHED_,
  _MEDIAN_
} MM_OPERATION;



static char *usage = "[image-in] [image-out]\n\
   [-se %s | -c %d | -w %d %d [%d]] [-i %d] [-b]\n\
   [-si %s]\n\
   [-ero | -dil | -ope | -clo | -th | -ith | ...]\n\
   [... | -grad | -lgrad | -ugrad | -ws | -med]\n\
   [-help]"; 

static char *detail = "\
 si 'image-in' est '-', on prendra stdin\n\
 si 'image-out' est absent, on prendra stdout\n\
 si les deux sont absents, on prendra stdin et stdout\n\
\n\
 -structuring-element | -se %s\n\
    # utilise l'image '%s' comme element structurant\n\
 -connectivity | -con | -c %d\n\
    # definit la connexite a utiliser comme element structurant (defaut = 26)\n\
 -window | -w %d %d [%d]\n\
    # definit un element structurant parallelepipedique\n\
 -iterations | -iter | -i %d\n\
    # nombre d'iterations\n\
 -binary-mode | -bin | -b\n\
    # des operations logiques sont utilisees\n\
 -seeds-image | -si %s\n\
    # seeds image = labels image for watershed\n\
 -erosion | -ero\n\
 -dilatation | -dilation | -dil\n\
 -opening | -ope | -ouverture | -ouv\n\
    # erosion puis dilatation\n\
 -closing | -clo | -fermeture | -fer\n\
    # dilatation puis erosion\n\
 -tophat | -th\n\
    # image - opening\n\
 -invtophat | -ith\n\
    # closing - image\n\
 -gradient | -grad\n\
    # morphological gradient = dilation - erosion\n\
 -lower-gradient | -low-grad | -lgrad\n\
    # lower morphological gradient = image - erosion\n\
 -upper-gradient | -up-grad | -ugrad\n\
    # upper morphological gradient = dilation - image\n\
 -watershed | -ws\n\
 -median | -med\n\
    # median filtering\n\
  $Revision: 1.3 $ $Date: 2003/12/12 09:30:51 $ $Author: hdeling $\n";

static char *program = NULL;

static void _errorMessage( char *str, int flag )
{
  CERR<<" Usage: "<<program<<" "<<usage<<"\n"<<ENDL;
  if ( flag == 1 )
    CERR<<detail<<ENDL;
  CERR<<" Error: "<<str<<ENDL;
  exit( -1 );
}



int main(int argc, char *argv[])
{
  char *inputImage  = NULL;
  char *outputImage = NULL;
  char *seedsImage  = NULL;
  char *eltImage    = NULL;

  Inrimage *image = NULL;
  Inrimage *seeds = NULL;

  StructuringElement *structelt = NULL;

  int connectivity = 0;
  CONNECTIVITY_TYPE C = TWENTY_SIX_CONNECTIVITY;

  int window[3] = {1,1,1};
  int iterations = 1;
  bool binary_mode = false;

  MM_OPERATION op = _DILATION_;

  program = argv[0];
  if ( argc == 1 ) _errorMessage( "\n", 0 );
    
  //
  // lecture des arguments
  //

  for ( int nimages = 0, i=1; i<argc; i++ ) {
    int status;
    // arguments
    if ( argv[i][0] == '-' ) {
      if ( argv[i][1] == '\0' ) {
	// standard input
	if ( nimages == 0 ) {
	  nimages ++;
	}
	else {
	  _errorMessage( "standard input specified twice\n", 0 );
	}
      }
      else if ( strcmp ( argv[i], "-help" ) == 0 ) {
	_errorMessage( "\n", 1 );
      }
      


      else if ( strcmp ( argv[i], "-structuring-element"  ) == 0 ||
		strcmp ( argv[i], "-se"  ) == 0 ) {
	i++;
	if ( i >= argc ) _errorMessage( "parsing -structuring-element ...\n", 0 );
	eltImage = argv[i];
      }
      else if ( strcmp ( argv[i], "-connectivity"  ) == 0 ||
		strcmp ( argv[i], "-con"  ) == 0 ||
		strcmp ( argv[i], "-c"  ) == 0 ) {
	i++;
	if ( i >= argc )   _errorMessage( "parsing -connectivity ...\n", 0 );
	status = sscanf( argv[i],"%d",&connectivity );
	if ( status != 1 ) _errorMessage( "parsing -connectivity ...\n", 0 );
      }
      else if ( strcmp ( argv[i], "-window"  ) == 0 ||
		strcmp ( argv[i], "-w"  ) == 0 ) {
	i++;
	if ( i >= argc )   _errorMessage( "parsing -window ...\n", 0 );
	status = sscanf( argv[i],"%d",&window[0] );
	if ( status != 1 ) _errorMessage( "parsing -window ...\n", 0 );
	i++;
	if ( i >= argc )   _errorMessage( "parsing -window ...\n", 0 );
	status = sscanf( argv[i],"%d",&window[1] );
	if ( status != 1 ) _errorMessage( "parsing -window ...\n", 0 );
	// peut-on encore lire un nombre
	if ( i+1 < argc && argv[i][0] >= '0' && argv[i][0] <= '9' ) {
	  i++;
	  if ( i >= argc )   _errorMessage( "parsing -window ...\n", 0 );
	  status = sscanf( argv[i],"%d",&window[2] );
	  if ( status != 1 ) _errorMessage( "parsing -window ...\n", 0 );
	}

      }
      else if ( strcmp ( argv[i], "-iterations"  ) == 0 ||
		strcmp ( argv[i], "-iter"  ) == 0 ||
		strcmp ( argv[i], "-i"  ) == 0 ) {
	i++;
	if ( i >= argc )   _errorMessage( "parsing -iterations ...\n", 0 );
	status = sscanf( argv[i],"%d",&iterations );
	if ( status != 1 ) _errorMessage( "parsing -iterations ...\n", 0 );
	if ( iterations <= 0 ) 
	  _errorMessage( "negative number of iterations ...\n", 0 );
      }


      else if ( strcmp ( argv[i], "-seeds-image" ) == 0 ||
		strcmp ( argv[i], "-si" ) == 0 ) {
	i++;
	if ( i >= argc )   _errorMessage( "parsing -seeds-image ...\n", 0 );
	seedsImage = argv[i];
      }


      else if ( strcmp ( argv[i], "-binary-mode"  ) == 0 ||
		strcmp ( argv[i], "-bin"  ) == 0 ||
		strcmp ( argv[i], "-b"  ) == 0 ) {
	binary_mode = true;
      }



      else if ( strcmp ( argv[i], "-erosion"  ) == 0 ||
		strcmp ( argv[i], "-ero"  ) == 0 ) {
	op = _EROSION_;
      }
      else if ( strcmp ( argv[i], "-dilatation"  ) == 0 ||
		strcmp ( argv[i], "-dilation"  ) == 0 ||
		strcmp ( argv[i], "-dil"  ) == 0 ) {
	op = _DILATION_;
      }
      else if ( strcmp ( argv[i], "-opening"  ) == 0 ||
		strcmp ( argv[i], "-ope"  ) == 0 ||
		strcmp ( argv[i], "-ouverture"  ) == 0 ||
		strcmp ( argv[i], "-ouv"  ) == 0 ) {
	op = _OPENING_;
      }
      else if ( strcmp ( argv[i], "-closing"  ) == 0 ||
		strcmp ( argv[i], "-clo"  ) == 0 ||
		strcmp ( argv[i], "-fermeture"  ) == 0 ||
		strcmp ( argv[i], "-fer"  ) == 0 ) {
	op = _CLOSING_;
      }
      else if ( strcmp ( argv[i], "-tophat"  ) == 0 ||
		strcmp ( argv[i], "-th"  ) == 0 ) {
	op = _TOPHAT_;
      }
      else if ( strcmp ( argv[i], "-invtophat"  ) == 0 ||
		strcmp ( argv[i], "-ith"  ) == 0 ) {
	op = _INVTOPHAT_;
      }
      else if ( strcmp ( argv[i], "-gradient"  ) == 0 ||
		strcmp ( argv[i], "-grad"  ) == 0 ) {
	op = _GRADIENT_;
      }
      else if ( strcmp ( argv[i], "-lower-gradient"  ) == 0 ||
		strcmp ( argv[i], "-low-grad"  ) == 0 ||
		strcmp ( argv[i], "-lgrad"  ) == 0 ) {
	op = _LOWER_GRADIENT_;
      }
      else if ( strcmp ( argv[i], "-upper-gradient"  ) == 0 ||
		strcmp ( argv[i], "-up-grad"  ) == 0 ||
		strcmp ( argv[i], "-ugrad"  ) == 0 ) {
	op = _UPPER_GRADIENT_;
      }
      else if ( strcmp ( argv[i], "-watershed"  ) == 0 ||
		strcmp ( argv[i], "-ws"  ) == 0 ) {
	op = _WATERSHED_;
      }
      else if ( strcmp ( argv[i], "-median"  ) == 0 ||
		strcmp ( argv[i], "-med"  ) == 0 ) {
	op = _MEDIAN_;
      }


      else {
	char message[256];
	sprintf( message, "unknown option '%s'\n", argv[i] );
	_errorMessage( message, 0 );
      }
    }
    // image names
    else {
      if ( nimages == 0 ) {
	inputImage = argv[i];
	nimages ++;
      }
      else if ( nimages == 1 ) {
	outputImage = argv[i];
	nimages ++;
      }
      else {
	_errorMessage( "too much file names when parsing\n", 0 );
      }
    }
  }
  //
  // fin de la lecture des arguments
  //
  


  try {

    //
    // lecture de l'image
    //
    image = new Inrimage( inputImage );


    //
    // element structurant
    //
    if ( eltImage != NULL ) {
      structelt = new StructuringElement( eltImage );
    } 
    else if ( window[0] != 1 || window[1] != 1 || window[2] !=1 ) {
      structelt = new StructuringElement( window[0], window[1], window[2] );
    }
    else if ( connectivity > 0 ) {
      switch ( connectivity ) {
      case  4 : C = FOUR_CONNECTIVITY;   break;
      case  8 : C = EIGHT_CONNECTIVITY;   break;
      case  6 : C = SIX_CONNECTIVITY;   break;
      case 10 : C = TEN_CONNECTIVITY;   break;
      case 18 : C = EIGHTEEN_CONNECTIVITY;   break;
      case 26 : C = TWENTY_SIX_CONNECTIVITY;   break;
      }
      structelt = new StructuringElement( C );
    }
    else {
      structelt = new StructuringElement;
    }

    // structelt->print();
    
    //
    // 
    //
    if ( iterations <= 0 ) _errorMessage( "", 0 );


    //
    //
    //
    switch( op ) {
    default :
      _errorMessage( "operation not implemented yet ...\n", 0 );
      
    case _EROSION_  :
      if ( binary_mode == false ) {
	if ( iterations <= 1 ) {
	  erosion( image,  structelt );
	}
	else {
	  erosion( image, structelt, iterations );
	}
      }
      else {
	erosion( image, structelt, iterations, binary_mode );
      }
      break;

    case _DILATION_  :

      dilation( image, structelt, iterations, binary_mode );
      break;

    case _OPENING_  :

      opening( image, structelt, iterations, binary_mode );
      break;

    case _CLOSING_  :

      closing( image, structelt, iterations, binary_mode );
      break;

    case _TOPHAT_  :

      topHat( image, structelt, iterations, binary_mode );
      break;

    case _INVTOPHAT_  :

      invTopHat( image, structelt, iterations, binary_mode );
      break;

    case _GRADIENT_  :

      morphologicalGradient( image, structelt, iterations, binary_mode );
      break;

    case _LOWER_GRADIENT_  :

      morphologicalLowerGradient( image, structelt, iterations, binary_mode );
      break;

    case _UPPER_GRADIENT_  :

      morphologicalUpperGradient( image, structelt, iterations, binary_mode );
      break;

    case _WATERSHED_ :

      if ( seedsImage == NULL ) {
	delete image;
	_errorMessage( "A seeds image has to be specified\n", 0 );
      }
      seeds = new Inrimage( seedsImage );
      watershed( image, seeds, seeds );
      *image = *seeds;
      delete seeds;
      break;

    case _MEDIAN_  :

      medianFiltering( image, structelt );
      break;

    }


    //
    // ecriture
    //
    image->write( outputImage );
  }

  catch( Exception e)  {
    CERR << e << ENDL;
    exit( -1 );
  }

    
  //
  //
  //
  delete image;

  return( 0 );
}
