#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <basics/AutoLink.h>
#include <modules/TclModule.h>
#include <inrimage/TclInrimage.h>

#ifdef STATIC_OBJECT
#include <graphics2D/Graphics2DCmd.h>
#include <inrimage/InrimageCmd.h>
#endif

#ifdef EMBEDED_TCL_SCRIPTS
#include "yav++.tcl.h"
#endif

using namespace yav;

static void readStdin(ClientData, int);
static TclString command, initrc = EMPTY_STRING;
static Tcl_Interp *globinterp;



class AutoExit : public WarnedObject {
public:
  AutoExit() : WarnedObject(typeid(TclCamera2D)) {}
  void emptyObjectsList(void) { TclModule::exit(); }
  void createdObject(TclObject *, const std::type_info&) {}
  void deletedObject(TclObject *, const std::type_info&) {}
};


static int letsgo(TclModule *core, int argc, char **argv);
static void usage(char *name);



static char releaseInfo[] = "\
\n\
                              zcompare\n\
\n\
         A viewer that overlays two volumetric images.\n\
         (in fact an extensible C++ front-end for TCL/TK)\n\
\n\
                          Version 1.0 (beta)\n\
\n\
authors:\n\
  Herv Delingette (Herve.Delingette@sophia.inria.fr)\n\
  Johan Montagnat (Johan.Montagnat@sophia.inria.fr)\n\
\n\
  INRIA Sophia-Antipolis, Epidaure project\n\
  2004 route des lucioles, BP93, 06902 Sophia-Antipolis, France\n\
  http://www.inria.fr/epidaure/Epidaure-eng.html\n\
\n\
Using TCL/TK 8.3\n\
  http://www.scriptics.com/software/8.3.html\n\
\n";


/* Main: let's go */
int main(int argc, char *argv[]) {
  bool verbose = false;
  /// do not load the file dynamically with dlopen()
  TclModule::loadDynamically=false;

  try {
    if(argc > 1 && !strcmp(argv[1], "-help")) usage(argv[0]);
    for(int i = 1; i < argc; i++) {
      if(!strcmp(argv[i], "-verbose")) verbose = true;
      else if(!strcmp(argv[i], "-version")) {
	COUT << releaseInfo << ENDL;
      }
    }
    if (argc<3)
      usage(argv[0]);
#ifdef STATIC_OBJECT
    INIT_MODULE_FUNCTION j=Tclgraphics2d_Init;
    TclModule::getModuleNameMap().insert(std::pair<char *,INIT_MODULE_FUNCTION>(((char *)TclGraphics2DInit::tclGraphics2DName),j));
    j=Tclinrimage_Init;
    TclModule::getModuleNameMap().insert(std::pair<char *,INIT_MODULE_FUNCTION>(((char *)TclInrimageInit::tclInrimageName),j));
#endif
    TclModule::init(letsgo, verbose, argc, argv, "tclGraphics2D", "tclInrimage", 0);
  }
  catch(...) {
    CERR << argv[0] << ": fatal: Unexpected exception caught" << ENDL;
    exit(-1);
  }

  return 0;
}


void usage(char *name) {
  fprintf(stderr, "usage: %s image_file_1 image_file_2 ...\n\
  image_file_1, image_file_2... : several volumetric images (at least 2) of the same size \n\
  -help                       : program usage\n\
  -version                    : program version\n",name);
  
exit(0);
}



// Starting point called by TK
int letsgo(TclModule *core, int argc, char **argv) {
  int i, j;

  
  globinterp = core->getInterp(); 
 
  core->tickle("wm withdraw .");


#ifdef EMBEDED_TCL_SCRIPTS
  /* do not load tcl scripts dynamically anymore */
  Tcl_Eval(core->getInterp(), yav___tcl);
#else
  try {
    core->loadSource("yav++.tcl");
  }
  catch(TclModule::ShellNotFound&) {
    CERR << "Could not find yav++.tcl" << ENDL
	 << "Recompile libModules with the right path " << ENDL;
    exit(-1);
  }
#endif
  
  std::list<TclInrimage *> inrlist;
  TclInrimage *inr;
  // else display all inrimages
  for(j=0,i = 1; i < argc; i++) {
    if ((!strcmp(argv[i], "-verbose")) || (!strcmp(argv[i], "-help")) || (!strcmp(argv[i], "-version")))
      {
      }
    else {
      try {
	if(argv[i][0] == '-') inr = new TclInrimage(core, (const char *) 0);
	else inr = new TclInrimage(core, argv[i]);
	if (inr)
	  inrlist.push_back(inr);
      }
      catch(Inrimage::UnknownImageType) {
	fprintf(stderr, "%s: warning: unknown image type \'%s\'\n",
		argv[0], argv[i]);
	fprintf(stderr,"Exiting...\n"); exit(0);
      }
      catch(Inrimage::ImageNotFound) {
	fprintf(stderr, "%s: warning: file \'%s\' not found\n", argv[0], argv[i]);
	fprintf(stderr,"Exiting...\n"); exit(0);
      }
      catch(Inrimage::InvalidHeader) {
	fprintf(stderr, "%s: warning: invalid header in \'%s\'\n",
		argv[0], argv[i]);
	fprintf(stderr,"Exiting...\n"); exit(0);
      }
      catch(Inrimage::InvalidData) {
	fprintf(stderr, "%s: warning: invalid data in \'%s\' (truncated file?)\n",
		argv[0], argv[i]);
	fprintf(stderr,"Exiting...\n"); exit(0);
      }
      catch(Inrimage::UnknownType) {
	fprintf(stderr, "%s: warning: unhandled inrimage type \'%s\'\n",
		argv[0], argv[i]);
	fprintf(stderr,"Exiting...\n"); exit(0);
      }
    } 
  }
  std::list<TclInrimage *>::iterator it;
  for(i=0,it=inrlist.begin();it!=inrlist.end();it++,i++)
    {
      core->tickle("view %s%u", 
		   TclInrimage::getTCLObjectName(), (*it)->getRef());
      if (i>0)
	{
		for (j=1;j<=i;++j) {
	  core->tickle("evaluate \"multiview%u -synchronize multiview%u \"",j,
		       i+1);
		}
	}
      else
	inr=(*it);
    }

  // Synchronisation des slicescale
  /*  core->tickle("foreach {m} [which inrview] { set i [$m -inrimage]; foreach {v} [ $i -slicecams ] { foreach {_m} [which inrview] { set _i [$_m -inrimage]; foreach {_v} [$_i -slicecams] { if ![ string compare [$v -geometry] [$_v -geometry ] ] { if [ string compare $v $_v ] { [$v -slicescale] -add $_v } } } } } }");
   */
  core->tickle("evaluate \"multiview1 -showAxis true\"");
  
  AutoExit autoexit;
  TclModule::run();


  exit(0);
  return TCL_OK;
}



/* read stdin in text interface */
static void readStdin(ClientData clientData, int mask) {
  static int partial = 0;
  Tcl_Channel outChannel, inChannel = (Tcl_Channel) clientData;
  Tcl_DString line;

  Tcl_DStringInit(&line);

  if(Tcl_Gets(inChannel, &line) < 0) {
    outChannel = Tcl_GetChannel(globinterp, "stdout", NULL);
    Tcl_Write(outChannel, "\n", 1);
    Tcl_Exit(0);
    return;
  }

  
  command += Tcl_DStringValue(&line);
  TclString cmd = command + "\n";
  Tcl_DStringFree(&line);
    
  if(!Tcl_CommandComplete((char *) cmd)) partial = 1;
  else {
    partial = 0;
    Tcl_CreateChannelHandler(inChannel, 0, readStdin, (ClientData) inChannel);
    Tcl_DStringInit(&line);
    switch(Tcl_RecordAndEval(globinterp, (char *) cmd, TCL_EVAL_GLOBAL)) {
    case TCL_OK:
      break;
    case TCL_ERROR:
      Tcl_DStringAppend(&line, "*Error* ", -1);
      break;
    case TCL_RETURN:
      Tcl_DStringAppend(&line, "*Return* ", -1);
      break;
    case TCL_BREAK:
      Tcl_DStringAppend(&line, "*Break* ", -1);
      break;
    case TCL_CONTINUE:
      Tcl_DStringAppend(&line, "*Continue* ", -1);
      break;
    }
    Tcl_DStringAppend(&line, globinterp->result, -1);
    if(*Tcl_DStringValue(&line)) puts(Tcl_DStringValue(&line));
    Tcl_DStringFree(&line);
    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, readStdin,
			     (ClientData) inChannel);
    command = "";
  }

  TclModule::Prompt(globinterp, partial);
  Tcl_ResetResult(globinterp);
}

