#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <basics/AutoLink.h>
#include <modules/TclModule.h>
#include <inrimage/TclInrimage4D.h>
#include <inrimage/MultiView4D.h>

#ifdef STATIC_OBJECT
#include <graphics2D/Graphics2DCmd.h>
#include <inrimage/InrimageCmd.h>
#endif

#ifdef EMBEDED_TCL_SCRIPTS
#include "yav++.tcl.h"
#endif

using namespace yav;

static void readStdin(ClientData, int);
static TclString command, initrc = EMPTY_STRING;
static Tcl_Interp *globinterp;



class AutoExit : public WarnedObject {
public:
  AutoExit() : WarnedObject(typeid(TclSliceCamera)) {}
  void emptyObjectsList(void) { TclModule::exit(); }
  void createdObject(TclObject *, const std::type_info&) {}
  void deletedObject(TclObject *, const std::type_info&) {}
};


static int letsgo(TclModule *core, int argc, char **argv);
static void usage(char *name);



static char releaseInfo[] = "\
\n\
                              zviewer4D\n\
\n\
         A script interpretor for volumetric image processing.\n\
         (in fact an extensible C++ front-end for TCL/TK)\n\
\n\
                          Version 1.1 (beta)\n\
\n\
authors:\n\
  Herv Delingette (Herve.Delingette@sophia.inria.fr)\n\
  Johan Montagnat (Johan.Montagnat@sophia.inria.fr)\n\
\n\
  INRIA Sophia-Antipolis, Epidaure project\n\
  2004 route des lucioles, BP93, 06902 Sophia-Antipolis, France\n\
  http://www.inria.fr/epidaure/Epidaure-eng.html\n\
\n\
Using TCL/TK 8.3\n\
  http://www.scriptics.com/software/8.3.html\n\
\n";


/* Main: let's go */
int main(int argc, char *argv[]) {
  bool verbose = false, rcopt = false;
  char *rfile = 0;
  unsigned int nbModules;
  TclString *modules;
  /// do not load the file dynamically with dlopen()
  TclModule::loadDynamically=false;

  try {
    if(argc > 1 && !strcmp(argv[1], "-help")) usage(argv[0]);
    for(int i = 1; i < argc; i++) {
      if(!strcmp(argv[i], "-rc")) {
	rcopt = true;
	if(++i >= argc) rfile = 0;
	else {
	  if(argv[i][0] == '-') rfile = 0;
	  else rfile = argv[i];
	}
      }
      else if(!strcmp(argv[i], "-verbose")) verbose = true;
      else if(!strcmp(argv[i], "-version")) {
	COUT << releaseInfo << ENDL;
      }
    }

#ifdef STATIC_OBJECT
    INIT_MODULE_FUNCTION j=Tclgraphics2d_Init;
    TclModule::getModuleNameMap().insert(std::pair<char *,INIT_MODULE_FUNCTION>(((char *)TclGraphics2DInit::tclGraphics2DName),j));
    j=Tclinrimage_Init;
    TclModule::getModuleNameMap().insert(std::pair<char *,INIT_MODULE_FUNCTION>(((char *)TclInrimageInit::tclInrimageName),j));
#endif
    if(rcopt && rfile) TclModule::readrc(rfile, initrc, nbModules, modules);
   TclModule::init(letsgo, verbose, argc, argv, "tclGraphics2D", "tclInrimage", 0);
  }
  catch(...) {
    CERR << argv[0] << ": fatal: Unexpected exception caught" << ENDL;
    exit(-1);
  }

  return 0;
}


void usage(char *name) {
#ifndef WIN32
  fprintf(stderr, "usage: %s [ image4D_file | [image3D_file_1 [image3D_file_2 [...]] | -vt %%f | script_file | -command %%s | -text | -graphic]\n\
  image4D_file                    : a 4D image to display\n\
  image3D_file_1, image3D_file_2, : set of 3D images of the same size and type that are put in a 4D image \n\
  -vt %%f                         : the time voxel size \n\
  script_file                     : evaluate TCL script in file and exit\n\
  -command %%s                    : evaluate given command and exit\n\
  -init %%s                       : evaluate given string before anything else\n\
  -rc [%%s]                       : load given ressource file (none if no ressource file\n\
                                   name is given).\n\
                                   By default, $HOME/.zviewerrc file is read.\n\
  -load %%s...                    : load given modules\n\
  -version                        : print out release informations\n\
  -verbose                        : print out lot of message (libraries search paths...)\n\
  -text|-graphic                  : select either text or graphic (tk) interface\n",
	  name);
#endif
  exit(0);

}


enum ARGS {
  NONE,
  CMD,
  TEXT,
  GRAPHIC,
  FILES
};

// Starting point called by TK
int letsgo(TclModule *core, int argc, char **argv) {
  Tcl_Channel inChannel;
  TclInrimage *inr;
  double vt= -1.0;
  int i;
  char *cmdstr = 0;
  ARGS cmd = NONE;
  int ninit = 0;

  for(i = 1; i < argc; i++) {

    if(strlen(argv[i]) > 0) {
      if(!strcmp(argv[i], "-") || argv[i][0] != '-') {
	if(cmd != NONE) usage(argv[0]);
	cmd = FILES;
	argv += i;
	argc -= i;
	break;
      }
      else {
	if(!strcmp(argv[i], "-command")) {
	  if(cmd != NONE) usage(argv[0]);
	  if(++i < argc && argv[i][0] != '-') {
	    cmdstr = argv[i];
	    cmd = CMD;
	  }
	  else usage(argv[0]);
	}
	else if(!strcmp(argv[i], "-text")) {
	  if(cmd != NONE) usage(argv[0]);
	  cmd = TEXT;
	  }
	else if(!strcmp(argv[i], "-vt")) {
	  if(++i < argc) {
	    vt=(double) atof(argv[i]);
	  }
	}
	else if(!strcmp(argv[i], "-graphic")) {
	  if(cmd != NONE) usage(argv[0]);
	  cmd = GRAPHIC;
	}
	else if(!strcmp(argv[i], "-init")) {
	  if(++i < argc && argv[i][0] != '-')
	    ninit = i;
	}	
	else if(!strcmp(argv[i], "-load")) {
	  if(++i < argc) {
	    int j;
	    for(j = i; j < argc && argv[j][0] != '-'; j++) {
	      try {
		(void) core->loadNewModule(core->getInterp(), argv[j]);
	      }
	      catch(TclModule::ModuleAlreadyExists) {
		COUT << ENDL << "yav++: warning: attempt to load module "
		     << argv[j] << " twice" << ENDL;
	      }
	      catch(Exception& e) {
		CERR << ENDL << e << ENDL;
		exit(-1);
	      }
	    }
	    i = j - 1;
	  }
	}
	else if(!strcmp(argv[i], "-version")) ;
	else if(!strcmp(argv[i], "-verbose")) ;
	else if(!strcmp(argv[i], "-rc"))  {
	  if(i + 1 < argc && argv[i+1][0] != '-') i++;
	}
	else {
	  fprintf(stderr, "%s: error: invalid option \'%s\'\n", argv[0], argv[i]);
	  exit(-1);
	}
      }
    }
  }
  
  globinterp = core->getInterp(); 
 
  /* handle initialization command */
  if(ninit > 0) {
    TclString str;
    str.sprintf("modules::evaluate \"%s\"",argv[ninit]);
    core->tickle(argv[ninit]);
  }
  core->tickle("wm withdraw .");

  if(initrc != EMPTY_STRING) {
    if(core->tickle((char *) initrc) == TCL_ERROR) {
      CERR << argv[0] << " warning: error while executing .zviewerrc init command."
	   << ENDL;
    }
  }

  switch(cmd) {
  case NONE:
  case GRAPHIC:

#ifdef EMBEDED_TCL_SCRIPTS
  /* do not load tcl scripts dynamically anymore */
  Tcl_Eval(core->getInterp(), yav___tcl);
#else
    try {
      core->loadSource("yav++.tcl");
    }
    catch(TclModule::ShellNotFound&) {
      CERR << "Could not find yav++.tcl" << ENDL
	   << "Recompile libModules with the right path " << ENDL;
      exit(-1);
    }
#endif

    /* launch graphic interface */
    core->tickle("yav::interface");
    TclModule::run();
    break;

  case TEXT:
    inChannel = Tcl_GetStdChannel(TCL_STDIN);
    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, readStdin,
			     (ClientData) inChannel);
    TclModule::Prompt(globinterp, 0);
    Tcl_ResetResult(globinterp);

    TclModule::run();
    break;

  case CMD:
    core->tickle(cmdstr);
    break;

  case FILES:
   
    std::list<Inrimage *> inrList;
    // else display all inrimages
    for(i = 0; i < argc; i++) {
      if(!strcmp(argv[i], "-vt")) {
	if(++i < argc) {
	  vt=(double) atof(argv[i]);
	}
      }
      else {
	try {
	  TclInrimage4D *inr4D= new TclInrimage4D(core,argv[i]);
	  if (inr4D)
	    {
	      if (vt>0)
		{
		  inr4D->getClient()->setVT(vt);
		}
	      core->tickle("inrimage::view4D %s%u",
			   TclInrimage4D::getTCLObjectName(), inr4D->getRef());
	      
	    }
	}
	catch(...) {
	  try {
	    if(argv[i][0] == '-') inr = new TclInrimage(core, (const char *) 0);
	    else inr = new TclInrimage(core, argv[i]);
	    inrList.push_back(inr->getClient());
	  }
	  catch(Inrimage::UnknownImageType) {
	    // if file is not an image, consider this is a script file
	    fprintf(stderr, "%s: warning: unknown image type \'%s\'\n",
		    argv[0], argv[i]);
	    if(i==0) {
	      core->tickle("script %s", argv[i]);
	      TclModule::exit(0);
	    }
	    continue;
	  }
	  catch(Inrimage::ImageNotFound) {
	    fprintf(stderr, "%s: warning: file \'%s\' not found\n", argv[0], argv[i]);
	    continue;
	  }
	  catch(Inrimage::InvalidHeader) {
	    fprintf(stderr, "%s: warning: invalid header in \'%s\'\n",
		    argv[0], argv[i]);
	    continue;
	  }
	  catch(Inrimage::InvalidData) {
	    fprintf(stderr, "%s: warning: invalid data in \'%s\' (truncated file?)\n",
		    argv[0], argv[i]);
	  continue;
	  }
	catch(Inrimage::UnknownType) {
	  fprintf(stderr, "%s: warning: unhandled inrimage type \'%s\'\n",
		  argv[0], argv[i]);
	  continue;
	} 
	}
      }
      
    }
    if (inrList.size()>0)
      {
	TclInrimage4D *inr4D= new TclInrimage4D(core,inrList);
	if (vt>0)
	  {
	    inr4D->getClient()->setVT(vt);
	  }
	core->tickle("inrimage::view4D %s%u", TclInrimage4D::getTCLObjectName(), 
		     inr4D->getRef());
      }
    
    AutoExit autoexit;
    TclModule::run();
    break;
  }

  exit(0);
  return TCL_OK;
}



/* read stdin in text interface */
static void readStdin(ClientData clientData, int mask) {
  static int partial = 0;
  Tcl_Channel outChannel, inChannel = (Tcl_Channel) clientData;
  Tcl_DString line;

  Tcl_DStringInit(&line);

  if(Tcl_Gets(inChannel, &line) < 0) {
    outChannel = Tcl_GetChannel(globinterp, "stdout", NULL);
    Tcl_Write(outChannel, "\n", 1);
    Tcl_Exit(0);
    return;
  }

  
  command += Tcl_DStringValue(&line);
  TclString cmd = command + "\n";
  Tcl_DStringFree(&line);
    
  if(!Tcl_CommandComplete((char *) cmd)) partial = 1;
  else {
    partial = 0;
    Tcl_CreateChannelHandler(inChannel, 0, readStdin, (ClientData) inChannel);
    Tcl_DStringInit(&line);
    switch(Tcl_RecordAndEval(globinterp, (char *) cmd, TCL_EVAL_GLOBAL)) {
    case TCL_OK:
      break;
    case TCL_ERROR:
      Tcl_DStringAppend(&line, "*Error* ", -1);
      break;
    case TCL_RETURN:
      Tcl_DStringAppend(&line, "*Return* ", -1);
      break;
    case TCL_BREAK:
      Tcl_DStringAppend(&line, "*Break* ", -1);
      break;
    case TCL_CONTINUE:
      Tcl_DStringAppend(&line, "*Continue* ", -1);
      break;
    }
    Tcl_DStringAppend(&line, globinterp->result, -1);
    if(*Tcl_DStringValue(&line)) puts(Tcl_DStringValue(&line));
    Tcl_DStringFree(&line);
    Tcl_CreateChannelHandler(inChannel, TCL_READABLE, readStdin,
			     (ClientData) inChannel);
    command = "";
  }

  TclModule::Prompt(globinterp, partial);
  Tcl_ResetResult(globinterp);
}

