#include <classification/FuzzyCMeans.h>
#include <basics/Constants.h>
#include <string>
#include <math.h>
#include <iostream>
#include <stdio.h>
#include <vector>
#include <algorithm>

namespace yav{

FuzzyCMeans::FuzzyCMeans(double** data, unsigned int nClust, unsigned int nData, unsigned int nDim, unsigned int* dataRange, unsigned int minDim, bool once, double expo, unsigned int maxIte, double minImpro, double* fuzzyPartition, double* initCluster)
{
	this->data = data;
	this->nClust = nClust;
	this->nData = nData;
	this->nDim = nDim;
	this->dataRange = dataRange;
	this->minDim = minDim;
	this->expo = expo;
	this->maxIte = maxIte;
	this->minImpro = minImpro;

	sizeMem = nClust*nData;
	sizeCen = nClust*nDim;
	myExpo = 1.0/(expo-1.0);
	isMaxClust = false;
	isMaxData = false;
	isMaxData2 = false;
	isIndData = false;

	mf = new double [sizeMem];
	sumMF = new double [nDim];
	isReliable = new bool [nData];
	weightRange = new double [nData];
	minRange = new unsigned int [nData];
	maxRange = new unsigned int [nData];
	maxClust = new double [nClust];
	maxData = new double [nData];
	maxData2 = new double [nData];
	indData = new unsigned int [nData];

	// FCM initialization
	reliability();

	// Initial fuzzy partition
	if (fuzzyPartition==NULL)
		memb = initMembership();
	else
		memb = fuzzyPartition;

	// Initial cluster centers
	if (initCluster==NULL)
		centers = new double [sizeCen];
	else
	{
		centers = initCluster;
		dist();
		membership();
	}

	// If not to be used more than once, use the srand()
	if (once)
		srand((unsigned int) time(NULL));
}

FuzzyCMeans::~FuzzyCMeans()
{
	if (memb) delete [] memb;
	if (centers) delete [] centers;
	if (indData) delete [] indData;
	if (maxData2) delete [] maxData2;
	if (maxData) delete [] maxData;
	if (maxClust) delete [] maxClust;
	if (maxRange) delete [] maxRange;
	if (minRange) delete [] minRange;
	if (weightRange) delete [] weightRange;
	if (isReliable) delete [] isReliable;
	if (sumMF) delete [] sumMF;
	if (mf) delete [] mf;
}

void FuzzyCMeans::reliability()
{
	// Check the data reliability
	if (dataRange==NULL)
	{	
		nDataUsed = nData;

		// By default, all the data are reliable
		for (unsigned int i=0; i<nData; i++)
		{
			isReliable[i] = true;
			weightRange[i] = 1.0;
			minRange[i] = 0;
			maxRange[i] = nDim-1;
		}
	}
	else
	{
		nDataUsed = 0;

		for (unsigned int i=0; i<nData; i++)
		{
			if (((dataRange[(2*i)+1]-dataRange[2*i])+1)>=minDim)
			{
				isReliable[i] = true;
				nDataUsed++;
			}
			else
				isReliable[i] = false;

			minRange[i] = dataRange[2*i];
			maxRange[i] = dataRange[(2*i)+1];
			weightRange[i] = (double)nDim/(double)((dataRange[(2*i)+1]-dataRange[2*i])+1);
		}
	}
}

double* FuzzyCMeans::initMembership()
{
	unsigned int i, j;
	double tmp;

	// Generate the probabilities
	memb = new double [sizeMem];
	
	for (j=0; j<nData; j++)
	{
		if(isReliable[j])
		{
			tmp = 0.0;

			for (i=0; i<nClust; i++)
			{
				memb[(i*nData)+j] = ((double)rand()/((double)RAND_MAX+1.0));
				tmp += memb[(i*nData)+j];
			}
	
			for (i=0; i<nClust; i++)
				memb[(i*nData)+j] /= tmp;
		}
		else
			for (i=0; i<nClust; i++)
				memb[(i*nData)+j] = 0.0;
	}

	return memb;
}

void FuzzyCMeans::membership()
{
	unsigned int i, j;
	double tmp;

	// Compute new U, suppose expo != 1
	// In the meantime, compute the sum of tmp rows in tempValue2 (=tempUnew)
	for (j=0; j<nData; j++)
	{
		if(isReliable[j])
		{
			tmp = 0.0;
	
			for (i=0; i<nClust; i++)
			{
				newMemb[(i*nData)+j] = 1/(pow(objArray[(i*nData)+j],myExpo));
				tmp += newMemb[(i*nData)+j];
			}
			// Ensure a sum of cluster membership equal to 1 for each data
			for (i=0; i<nClust; i++)
				newMemb[(i*nData)+j] /= tmp;
		}
		else
			for (i=0; i<nClust; i++)
				newMemb[(i*nData)+j] = 0.0;
	}
}

void FuzzyCMeans::dist()
{
	unsigned int i, j, k;
	double tmp;

	// Fill the objArrayput matrix
	// objArray[i][j] is the L2 norm distance between each sample j and each center of cluster i
	for (i=0; i<nClust; i++)
	{
		for (j=0; j<nData; j++)
		{
			objArray[(i*nData)+j] = 0.0;

			if(isReliable[j])
			{	
				for (k=minRange[j]; k<=maxRange[j]; k++)
				{
					tmp = data[j][k]-centers[(i*nDim)+k];
					objArray[(i*nData)+j] += (weightRange[j]*tmp*tmp);
				}
			}
		}
	}
}

void FuzzyCMeans::center()
{
	unsigned int i, j, k;

	// Set the center[i] to 0
	for (i=0; i<sizeCen; i++)
		centers[i] = 0.0;

	// Compute the clusters center
	for (i=0; i<nClust; i++)
	{
		for (j=0; j<nDim; j++)
			sumMF[j] = 0.0;

		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				mf[(i*nData)+j] = pow(memb[(i*nData)+j],expo);
		
				for (k=minRange[j]; k<=maxRange[j]; k++)
				{
					centers[(i*nDim)+k] += mf[(i*nData)+j]*data[j][k];
					// In the meantime, compute the sum of mf rows in sumMF
					sumMF[k] += mf[(i*nData)+j];
				}
			}
		}

		for (k=0; k<nDim; k++)
			centers[(i*nDim)+k] /= sumMF[k];
	}
}

double FuzzyCMeans::step()
{
	unsigned int i;
	double tmpObjFn;

	// Compute the clusters center
	center();
	// compute the distance between the center and each sample 
	dist();
	// Objective function
	tmpObjFn = 0.0;

	for (i=0; i<sizeMem; i++)
		tmpObjFn += objArray[i]*mf[i];

	// Compute the distances
	membership();
	// Compute the confidence variation and update U
	confVar = 0.0;

	for (i=0; i<sizeMem; i++)
	{
		confVar += (newMemb[i]-memb[i])*(newMemb[i]-memb[i]);
		memb[i] = newMemb[i];
	}

	return tmpObjFn;
}

void FuzzyCMeans::fcm()
{
	double tmpObjFn;

	objArray = new double [sizeMem];
	newMemb = new double [sizeMem];

	// Main loop
	for (nIte=0; nIte<maxIte; nIte++)
	{
		tmpObjFn = step();

		// Check termination condition
		if ((nIte>0) && (fabs(tmpObjFn-objFn)<minImpro))
			break;
		else
			objFn = tmpObjFn;
	}
	
	// Keep the value of the last objective function
	objFn = tmpObjFn;

	delete [] newMemb; delete [] objArray;
}

double FuzzyCMeans::getVo()
{
	unsigned int i;
	double ove, max1, max2;

	// Compute the VOS
	ove = 0.0;

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			max1 = getMaxData()[i];
			max2 = getMaxData2()[i];
			ove += 2.0*(max2/(max1+max2));
		}
	}	

	ove /= (double)nDataUsed;

	return ove;
}

double FuzzyCMeans::getVs()
{
	unsigned int i, j, k, n;
	double up, tmp, sep;

	sep = 0.0;

	for (i=0; i<nClust-1; i++)
	{
		for (j=i+1; j<nClust; j++)
		{
			tmp = 0.0;
			up = 0.0;
			n = 0;

			for (k=0; k<nData; k++)
			{
				if(isReliable[k])
				{
					if(i==getIndData()[k])
					{
						up += memb[(j*nData)+k]/(memb[(i*nData)+k]+memb[(j*nData)+k]);
						n++;
					}
					else if(j==getIndData()[k])
					{
						up += memb[(i*nData)+k]/(memb[(i*nData)+k]+memb[(j*nData)+k]);
						n++;
					}
				}
			}

			tmp = up/(double)n;

			if (tmp>sep)
				sep = tmp;
		}
	}

	return sep;
}


double FuzzyCMeans::getVpi()
{
	unsigned int i, j, k, sum;
	double vpi, tmp1, tmp2, tmp3;

	double* perfo = new double [nClust*nData];
	double* average = new double [nDim];

	// Compute the average of all data
	for (i=0; i<nDim; i++)
	{
		tmp1 = 0.0;
		sum = 0;

		for (j=0; j<nData; j++)
		{
			if((isReliable[j])&&(i>=minRange[j])&&(i<=maxRange[j]))
			{
				tmp1 += data[j][i];
				sum++;
			}
		}

		average[i] = tmp1/(double)sum;
	}

	// Get performance index
	for (i=0; i<nClust; i++)
	{
		for (j=0; j<nData; j++)
		{
			perfo[(i*nData)+j] = 0.0;

			if(isReliable[j])
			{
				for (k=minRange[j]; k<=maxRange[j]; k++)
				{
					tmp2 = data[j][k]-centers[(i*nDim)+k];
					tmp3 = centers[(i*nDim)+k]-average[k];
					perfo[(i*nData)+j] += (weightRange[j]*((tmp2*tmp2)-(tmp3*tmp3)));
				}
			}
		}
	}

	vpi = 0.0;

	for (i=0; i<nClust; i++)
		for (j=0; j<nData; j++)
			if(isReliable[j])
				vpi += mf[(i*nData)+j]*perfo[(i*nData)+j];

	// Delete part
	delete [] average;
	delete [] perfo;

	return vpi;
}

double FuzzyCMeans::getVfvq()
{
	unsigned int i, j, k;
	double tmp1, tmp2, qe, mind, vfvq;

	// Reconstructed data (x')
	double* data2 = new double [nData*nDim];

	for (i=0; i<(nData*nDim); i++)
		data2[i] = 0.0;

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			for (j=0; j<nDim; j++)
				sumMF[j] = 0.0;

			for (j=0; j<nClust; j++)
			{
				for (k=minRange[i]; k<=maxRange[i]; k++)
				{
					data2[(i*nDim)+k] += mf[(j*nData)+i]*centers[(j*nDim)+k];
					sumMF[k] += mf[(j*nData)+i];
				}
			}
	
			for (k=0; k<nDim; k++)
				data2[(i*nDim)+k] /= sumMF[k];
		}
	}

	// Quantization error (qe)
	qe = 0.0;

	for (i=0; i<nData; i++)
	{
		if(isReliable[i])
		{
			tmp2 = 0.0;
	
			for (j=minRange[i]; j<=maxRange[i]; j++)
			{
				tmp1 = data[i][j]-data2[(i*nDim)+j];
				tmp2 += weightRange[i]*tmp1*tmp1; 
			}
			
			qe += tmp2;
		}
	}

	// Minimum distance between clusters
	mind = MAX_DOUBLE;

	for (i=0; i<(nClust-1); i++)
	{
		for (j=i+1; j<nClust; j++)
		{
			tmp2 = 0.0;

			for (k=0; k<nDim; k++)
			{
				tmp1 = centers[(i*nDim)+k]-centers[(j*nDim)+k];
				tmp2 += tmp1*tmp1;
			}

			if (sqrt(tmp2)<mind)
				mind = sqrt(tmp2);
		}
	} 
	
	// Compute Vfvq
	vfvq = qe/(nDataUsed*mind);

	delete [] data2;

	return vfvq;
}

double FuzzyCMeans::getVpc()
{
	unsigned int i, j;
	double vpc;

	vpc = 0.0;

	for (i=0; i<nClust; i++)
		for (j=0; j<nData; j++)
			if(isReliable[j])
				vpc += (memb[(i*nData)+j]*memb[(i*nData)+j]);

	vpc /= nDataUsed;

	return vpc;
}

double FuzzyCMeans::getVpe()
{
	unsigned int i, j;
	double vpe;

	vpe = 0.0;

	for (i=0; i<nClust; i++)
		for (j=0; j<nData; j++)
			if(isReliable[j])
				vpe += memb[(i*nData)+j]*log(memb[(i*nData)+j]);

	vpe = -1.0*(vpe/nDataUsed);

	return vpe;
}

unsigned int FuzzyCMeans::getNbrClust()
{
	return nClust;
}

unsigned int FuzzyCMeans::getNbrData()
{
	return nData;
}

unsigned int FuzzyCMeans::getNbrDim()
{
	return nDim;
}

unsigned int FuzzyCMeans::getMaxIte()
{
	return maxIte;
}

unsigned int FuzzyCMeans::getNbrIteUsed()
{
	return nIte+1;
}

double FuzzyCMeans::getExpo()
{
	return expo;
}

double FuzzyCMeans::getMinImpro()
{
	return minImpro;
}

double FuzzyCMeans::getObjFn()
{
	return objFn;
}

double** FuzzyCMeans::getData()
{
	return data;
}

double* FuzzyCMeans::getMembership()
{
	return memb;
}

double* FuzzyCMeans::getCenters()
{
	return centers;
}

bool* FuzzyCMeans::getReliable()
{
	return isReliable;
}

double* FuzzyCMeans::getMaxClust()
{
	if(isMaxClust)
		return maxClust;
	else
	{
		unsigned int i, j;
		double tmp;
	
		// Compute maximum membership by cluster
		for (i=0; i<nClust; i++)
		{
			tmp = 0.0;
	
			for (j=0; j<nData; j++)
				if ((isReliable[j])&&(memb[(i*nData)+j]>tmp))
					tmp = memb[(i*nData)+j];
			
			maxClust[i] = tmp;
		}

		isMaxClust = true;	
		return maxClust;
	}
}

double* FuzzyCMeans::getMaxData()
{
	if(isMaxData)
		return maxData;
	else
	{
		unsigned int i, j;
		double tmp;
	
		// Compute maximum membership by data
		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				tmp = 0.0;
		
				for (i=0; i<nClust; i++)
					if (memb[(i*nData)+j]>tmp)
						tmp = memb[(i*nData)+j];
				
				maxData[j] = tmp;
			}
			else
				maxData[j] = 0.0;
		}

		isMaxData = true;	
		return maxData;
	}
}

double* FuzzyCMeans::getMaxData2()
{
	if(isMaxData2)
		return maxData2;
	else
	{
		unsigned int i, j;
		double tmp;
	
		// Compute maximum membership by data
		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				tmp = 0.0;
		
				for (i=0; i<nClust; i++)
					if ((memb[(i*nData)+j]>tmp)&&(memb[(i*nData)+j]!=maxData[j]))
						tmp = memb[(i*nData)+j];
				
				maxData2[j] = tmp;
			}
			else
				maxData2[j] = 0.0;
		}

		isMaxData2 = true;	
		return maxData2;
	}
}

unsigned int* FuzzyCMeans::getIndData()
{

	if(isIndData)
		return indData;
	else
	{
		unsigned int i, j, ind;
		double tmp;
	
		// Compute maximum membership by data
		for (j=0; j<nData; j++)
		{
			if(isReliable[j])
			{
				tmp = 0.0;
		
				for (i=0; i<nClust; i++)
				{
					if (memb[(i*nData)+j]>tmp)
					{
						tmp = memb[(i*nData)+j];
						ind = i;
					}
				}
		
				indData[j] = ind;
			}
			else
				indData[j] = 0;
		}

		isIndData = true;	
		return indData;
	}
}
}
