#ifndef FUZZYCMEANS_H
#define FUZZYCMEANS_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

namespace yav{

class FuzzyCMeans
{
 protected:
	/// Number of iterations used
	unsigned int nIte;
	/// Maximum number of iterations allowed
	unsigned int maxIte;
	/// Number of clusters
	unsigned int nClust;
	/// Number of data
	unsigned int nData;
	/// Data dimension
	unsigned int nDim;
	/// Data minimum dimension (to be defined as reliable)
	unsigned int minDim;
	/// If data range specified, number of data defined as reliable
	unsigned int nDataUsed;
	/// Size of membership in memory
	unsigned int sizeMem;
	/// Size of centers in memory
	unsigned int sizeCen;
	/// Fuzziness exponent
	double expo;
	/// Defined as 1/(expo-1)
	double myExpo;
	/// Minimum of improvement between 2 iterations
	double minImpro;
	/// Objective function value at iteration i
	double objFn;
	/// Confidence variation between iteration i and i-1
	double confVar;
	/// Specify whether maxClust has been assigned
	bool isMaxClust;
	/// Specify whether maxData has been assigned
	bool isMaxData;
	/// Specify whether maxData2 has been assigned
	bool isMaxData2;
	/// Specify whether IndData has been assigned
	bool isIndData;

	/// For each data, index of the cluster with the highest membership
	unsigned int* indData;
	/// For each data, range of the data to be taken into account [min1 max1 min2 max2...]
	unsigned int* dataRange;
	/// For each data, index of first dimension to be taken into account
	unsigned int* minRange;
	/// For each data, index of last dimension to be taken into account
	unsigned int* maxRange;
	/// MF matrix after exponential modification of memb
	double* mf;
	/// Membership
	double* memb;
	/// Temporary membership
	double* newMemb;
	/// Cluster centers
	double* centers;
	/// Objective array
	double* objArray;
	/// Maximum membership for each cluster
	double* maxClust;
	/// Maximum membership for each data
	double* maxData;
	/// Second maximum membership for each data
	double* maxData2;
	/// For each data, weighting factor for the center computation
	double* weightRange;
	/// For each coordinate, sum of the MF taken into account in the center computation
	double* sumMF;
	/// For each data, a boolean specify whether the data is reliable
	bool* isReliable;

	/// Data to classify
	double** data;

	/** Fuzzy C-Mean initialization */
	void reliability();

	/** Generates initial membership. By default, INITMEMBERSHIP() randomly generates a fuzzy partition matrix. The summation of each column is equal to 1, as required by FCM.*/
	double* initMembership();

	/** Computes the membership U.*/
	void membership();

	/** Performs one iteration of FCM.*/
	double step();

	/** Computes the distance.*/
	virtual void dist();

	/** Computes the cluster centers.*/
	virtual void center();
 public:
	/** @name Constructors */
	/** Initializes the Fuzzy C-Means algorithm.
	@param data data (data[number_of_data][number_of_coordinates])
	@param nClust number of clusters
	@param nData number of data
	@param nDim data dimension
	@param dataRange for each data, range to be taken into account [min1 max1 min2 max2...]
	@param minDim data minimum dimension to be considered as reliable (default=5)
	@param once boolean taken into account for the pseudo-random numbers generation used by U. Should be always true (default=true). This boolean is supposed to be false only when used by BESTNUMBERCLUSTERSFCM class.
	@param expo fuzziness exponent (default=2.0)
	@param maxIte maximum number of iterations allowed (default=100)
	@param minImpro minimum amount of improvement between 2 iterations (default=0.00001)
	@param fuzzyPartition initial membership (default=NULL)
	@param initCluster initial cluster centers (default=NULL)*/
	FuzzyCMeans(double** data, unsigned int nClust, unsigned int nData, unsigned int nDim, unsigned int* dataRange=NULL, unsigned int minDim=1, bool once=true, double expo=2.0, unsigned int maxIte=100, double minImpro=0.00001, double* fuzzyPartition=NULL, double* initCluster=NULL);

	/** @name Destructor */
	/** Deallocates the different arrays used in memory.*/
	virtual ~FuzzyCMeans();

	/** @name Other functions */
	/** Launches the Fuzzy C-Means. The process stops when the maximum number of iterations maxIte is reached, or when the minimum amount of improvement between 2 iterations is less than minImpro.*/
	void fcm();

	/** Returns the number of clusters.*/
	unsigned int getNbrClust();

	/** Returns the number of data.*/
	unsigned int getNbrData();

	/** Returns the data dimension.*/
	unsigned int getNbrDim();

	/** Returns the maximum number of iterations allowed.*/
	unsigned int getMaxIte();

	/** Returns the number of iterations used.*/
	unsigned int getNbrIteUsed();

	/** Returns the fuzziness exponent.*/
	double getExpo();

	/** Returns the minimum amount of improvement.*/
	double getMinImpro();

	/** Returns the last value of the objective function before stability.*/
	double getObjFn();

	/** Returns the data.*/
	double** getData();

	/** Returns the membership U which contains the grade of membership of each data in each cluster. The values 0 and 1 indicate no membership and full membership respectively. Grades between 0 and 1 indicate that the data has partial membership in a cluster.*/
	double* getMembership();

	/** Returns the cluster centers.*/
	double* getCenters();

	/** Returns the reliable data.*/
	bool* getReliable();

	/** Returns the maximum membership for each cluster.*/
	double* getMaxClust();

	/** Returns the maximum membership for each data.*/
	double* getMaxData();

	/** Returns the second maximum membership for each data.*/
	double* getMaxData2();

	/** Returns the index of the cluster with the highest membership for each data.*/
	unsigned int* getIndData();

	/** Computes the overlap.*/
	double getVo();

	/** Computes the separation.*/
	double getVs();

	/** Returns the performance index (Vpi).*/
	double getVpi();

	/** Returns the fuzzy vector quantization (Vfvq).*/
	double getVfvq();

	/** Returns the partition coefficient (Vpc).*/
	double getVpc();

	/** Returns the partition entropy (Vpe).*/
	double getVpe();
};

}

#endif
